/****************************************************************************
   Copyright & License:
   ====================
   
   Copyright 2009 - 2020 gAGE/UPC & ESA
   
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at
   
   http://www.apache.org/licenses/LICENSE-2.0
   
   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
 ***************************************************************************/

/****************************************************************************
 * Copyright: gAGE/UPC & ESA
 * Project: EDUNAV GNSS Lab Tool 
 * Supervisor: Jaume Sanz Subirana (group of Astronomy and GEomatics - gAGE/UPC)
 * Author: Pere Ramos-Bosch (group of Astronomy and GEomatics - gAGE/UPC)
 * Curator: Adria Rovira-Garcia ( gAGE/UPC )
 * Developers: Deimos Ibanez Segura ( gAGE/UPC )
 *             Jesus Romero Sanchez ( gAGE/UPC )
 *          glab.gage @ upc.edu
 * File: dataHandling.c
 * Code Management Tool File Version: 5.5  Revision: 1
 * Date: 2020/12/11
 ***************************************************************************/

/****************************************************************************
 * MODULE DESCRIPTION
 *
 * Name: dataHandling
 * Language: C
 *
 * Purpose:
 *  The purpose of this module is to centralize all the data structures of the 
 *  modules and allow an easy interface to them.
 *
 * Function:
 *  This class is in charge of declaring all the data structures used in all the
 *  modules. It defines all the required functions to access the information 
 *  inside the data structures, and a group of general functions.
 *
 * Dependencies:
 *  None
 *
 * Files modified:
 *  None
 *
 * Files read:
 *  None
 *
 * Resources usage:
 *  See Design Document
 ****************************************************************************/

/****************************
 *       RELEASE_HISTORY
 * -------------
 *          gLAB v1.3.1
 * Release: 2010/01/22
 * Change Log: First public version.
 * -------------
 *          gLAB v1.3.2
 * Release: 2010/02/23
 * Change Log: Corrected problem when processing with PCC (PC computed with C1).
 *             changeBase modified to be consistent with orientation matrix.
 *             Inertial/Non-inertial velocity switch included.
 *             minArcLength changed from 3 to 4.
 *             Included DCB data sources.
 *             Included SINEX, DCBs and GPS_Receiver_Type data files.
 *             Included strict radome option to enable/disable a relaxed matching
 *                  on ANTEX files.
 *             Corrected a bug when using PCC combination.
 *             When the receiver is of type 2 (no P1), the C1 will be provided instead
 *             Forward/Backward processing included.
 *             Corrected bug in t2doy function affecting the 31st of December and 
 *                  1st of January.
 * -------------
 *          gLAB v1.4.0
 * Release: 2010/06/21
 * Change Log: C1/P1 flexibility option is added to allow flexible handling of C1 and P1
 *             Included usable measurement index in TMeasOrder structure to select/deselect
 *                  individual measurements.
 *             Included prealignCP option to set/unset carrier-phase prealignment.
 *             Corrected a bug in t2doy function.
 * -------------
 *          gLAB v1.4.1
 * Release: 2010/07/09
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.2
 * Release: 2010/07/31
 * Change Log: No changes in this file.
 * -------------
 *          gLAB v1.4.3
 * Release: 2010/08/31
 * Change Log: Maximum number of filter iterations added.
 *             Increased robustness to non sequencialy transmitted ephemerides present
 *                  in RINEX Navigation Message files.
 * -------------
 *          gLAB v1.4.4
 * Release: 2010/09/22
 * Change Log: Included new DCB management.
 * -------------
 *          gLAB v1.4.5
 * Release: 2010/10/31
 * Change Log: Tropospheric correction option is extended to select independently the nominal and mapping used.
 *             Dilution Of Precision extension: (GDOP, PDOP, TDOP, VDOP, HDOP) new variables.
 *             Default precise orbit interpolation degree changed from 9 to 10.
 *             SIGMA_INF splitted into different SIGMA_INF_DR and SIGMA_INF_DT: receiver coordinates and clock.
 * -----------
 * Release: 2012/12/31
 * Change Log: gLAB version released together with the Educational Book:
 *             "GNSS Data Processing" (Vol. 1 and Vol. 2). ESA TM-23.
 *             Authors: J. Sanz Subirana, J.M. Juan Zornoza and M. Hernandez-Pajares
 *             ISBN: 978-92-9221-885-0 (two volumes)
 *             ISSN: 1013-7076
 *             December 2012.
 * -----------
 *          gLAB v2.2.0
 * Release: 2014/09/22
 * Change Log: Added functions for initialising IONEX and F-PPP structures (both for ionosphere and DCB data).
 *             Added functions for handling IONEX and F-PPP structures (both for ionosphere and DCB data).
 *             Added function for computing the ionosphere mapping value.
 *             Added function for computing the pierce point value.
 *             Added functions for initialising ESA's troposphere model structures.
 *             Added functions for handling ESA's troposphere model structures.
 *             Modified some initialiase functions for handling RINEX 3.02 files.
 * -----------
 *          gLAB v2.2.1
 * Release: 2014/12/23
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.2
 * Release: 2015/03/02
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.3
 * Release: 2015/05/28
 * Change Log: Added initialization values in initConstellationElement, initAntenna and initOptions functions for phase variation correction.
 * -----------
 *          gLAB v2.2.4
 * Release: 2015/07/01
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.5
 * Release: 2015/10/20
 * Change Log: No changes in this file
 * -----------
 *          gLAB v2.2.6
 * Release: 2015/11/25
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v2.2.7
 * Release: 2016/03/01
 * Change Log: Fixed "printError" function call, which failed to compile in some compilers
 * -----------
 *          gLAB v2.2.8
 * Release: 2016/04/12
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v3.0.0
 * Release: 2016/09/05 
 * Change Log: Added SBAS processing.
 *             Added user-defined error struct.
 *             Deleted satWas function.
 *             Added satIs function. It returns the index of the satellite in its dictionary.
 *             Added tsec function. It returns the MJDN + SoD in seconds.
 *             Modified structure definitions for handling Cycle Slip variables.
 *             Added QZSS and IRNSS constellations.
 * -----------
 *          gLAB v3.1.0
 * Release: 2016/09/26
 * Change Log: gpsws2ttime function now accepts negative seconds of week.
 *             Changed selectBRDCblock function to enable/disable the check of the transmission time of message.
 *             Fixed incorrect broadcast block selection when a block with a determined IODE was requested, and
 *                there was more than one block for the same satellite with the same IODE in the navigation file.
 *             Fixed rare problem when computing IONEX pierce point at high elevation.
 * -----------
 *          gLAB v3.1.1
 * Release: 2016/10/21
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v4.0.0
 * Release: 2017/03/03
 * Change Log: Added functions to initialise the RTCM structures.
 *             Added functions to initialise SBASplots structures.
 *             Added function date2doy.
 *             Added function doy2date.
 *             Added function SolutionModeNum2SolutionModeChar.
 *             Added ConcatenateSP3 function.
 *             Added comparison functions for qsort function.
 *             When searching a station in the SINEX file, it will only use the first four letters of the marker name
 *               in the RINEX observation file instead of the complete name.
 *             Fixed initialization of tropMapping variable (it was intialized to 'SimpleMominal' instead of 
 *               'SimpleMapping') in initOptions function (it didn't affect the code as both have the same integer value).
 *             Fixed incorrect memcpy call in t2tm function.
 * -----------
 *          gLAB v4.1.0
 * Release: 2017/04/07
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v4.2.0
 * Release: 2017/05/22
 * Change Log:   Added function 'mkdir_recursive' for creating directories recursively.
 *               Changed default SNR threshold from 35dbHz to 30dbHz.
 * -----------
 *          gLAB v5.0.0
 * Release: 2017/06/30
 * Change Log:   Fixed bug that made gLAB try to create an empty folder in Linux when the output path was absolute.
 * -----------
 *          gLAB v5.1.0
 * Release: 2017/11/24
 * Change Log:   Added functions 'ConvertCalendarDate', 'ConvertDoYDate', 'ConvertGPSWeekDate'
 *                 and 'ConvertMJDNDate' for date conversion.
 *               Added 'ConvertCartesianCoord', 'ConvertGeodeticCoord' and 'ConvertSphericalCoord'
 *                 for coordinate conversion.
 *               When 4 or more consecutive messages are wrong, UDREI data will timeout (for PA
 *                 mode only) instead of erasing all fast correction data.
 * -----------
 *          gLAB v5.1.1
 * Release: 2017/12/22
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v5.1.2
 * Release: 2018/01/12
 * Change Log:   Added initialization of 'numthreads' variable in 'initOptions' variable to the number of CPUs
 *                 in the machine.
 * -----------
 *          gLAB v5.1.3
 * Release: 2018/01/19
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v5.2.0
 * Release: 2018/03/09
 * Change Log: Added functions 'InitRangeList', 'freeRangeList', 'AddRangeList' and 'ExtendRangeList' for
 *               computing ranges of satellites grouped by a property (e.g. weight type).
 *             Added function 'SatRangeList2String' for transforming a range list to a string.
 *             Added function 'WeightType2String' to transform from weight type to string.
 *             Added function 'SNRCombModeNum2String' to transform from SNR combination type to string.
 * -----------
 *          gLAB v5.3.0
 * Release: 2018/06/08
 * Change Log:   Added data table with EGM96 grid with 1x1 degree resolution. This is the geoid model used in
 *                 Google Earth.
 *               Removed 'satIs' function. The function has been replaced by a index lookup  in a matrix array. 
 *                 All the function calls to 'satIs' in other files have also been replaced. This change
 *                 reduces computation time around 10% in all processing modes except for SBAS plots. 
 *               In function 'getMeasModelValue', the search for a satellite position is done using a dictionary
 *                 instead of a for loop (faster search).
 *               Changed the default marker name from an empty string to "----". If the marker name in the
 *                 RINEX observation file was empty or not present, the marker name was an empty string, thus
 *                 making all the output SBAS messages and the last line of the summary (in all modes) to have
 *                 the marker name (or receiver name) to be an empty string, which, in practice, made the mesages
 *                 have one column less, which changed the column number for all items after the marker name (and
 *                 in consequence, breaking all programs and scripts looking for a specific column).
 * -----------
 *          gLAB v5.4.0
 * Release: 2018/11/16
 * Change Log:   Added function 'qsort_compare_float' for comparing float values for the qsort
 *                 (quick sort) C function.
 *               Added function 'SBASGEOselection' for selection a GEO in SBAS mode.
 *               Added function 'SBASGEOAlarmCheck' for checking alarm conditions in SBAS mode.
 *                 Furthermore, gLAB is capable of handling any number of alarms is different
 *                 GEOs at the same time, while before, it could handle only two (which could
 *                 create errors if there more than two GEOs with alarm at the same time).
 *               Added function 'SaveDataforGEOselection' for saving modelled data for each GEO
 *                 with solution (when option select best GEO is enabled).
 *               Added function 'SelectBestGEO' for selecting best GEO from the GEOs with solution
 *                 (when option select best GEO is enabled).
 *               Several 'static' variables in different functions have been changes to "const"
 *                 (when possible) to ensure there are no race conditions with multithreading.
 *               Several 'static' variables in different functions (specially in function 
 *                 'meastype2measstr') now have below the OpenMP directive '#pragma omp threadprivate()'
 *                 to avoid race conditions in multithreading.
 *               Modified 'ConcatenateSP3' function. Before, when saving the orbit and clock data,
 *                 it removed all epochs with 0s or invalid data. Now, it saves all data equally
 *                 spaced in time (using the interval from the SP3 or clock file), adding 0s
 *                 when necessary. In this way, it is easy to find the current sample to be used,
 *                 as it is possible to compute the position of the sample by dividing the time
 *                 passed from the start of SP3 data to the current epoch by the time interval,
 *                 whereas before it had to cycle sample by sample until it found the desired one.
 *                 For big files (such as 5 second clock files), it is a big speed boost when
 *                 searching for the required samples.
 *               Memory usage for DOP percentile computation has been halved by changing the 
 *                 storage of the DOP values from type 'double' to 'float'. This only affects
 *                 to the storage of the values, as all the computation to get the DOP values
 *                 are computed using 'double' precision.
 * -----------
 *          gLAB v5.4.1
 * Release: 2019/02/15
 * Change Log:   Added an INFO message with the elevation of the SBAS GEO satellites available
 *                 (and if it was selected or not) when the selection of GEO with highest
 *                 elevation is done.
 * -----------
 *          gLAB v5.4.2
 * Release: 2019/03/08
 * Change Log:   When selection GEO with highest elevation, if a priori position is 
 *                 set to "calculate", in the first iteration it will do a temporary
 *                 GEO selection (and it will show it in the printed INFO message),
 *                 and it will select the GEO with highest elevation when the a
 *                 priori position has converged. It will be done regardless GEO
 *                 switching is enabled or disabled.
 *               If SBAS file contains data from GEOs of multiple service providers
 *                 (e.g. EGNOS, WAAS, etc), it will compute the elevation of all
 *                 GEOs, instead of only computing the elevation of the GEOs
 *                 available in the almanac message of the first GEO read.
 * -----------
 *          gLAB v5.4.3
 * Release: 2019/03/20
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v5.4.4
 * Release: 2019/05/10
 * Change Log: No changes in this file.
 * -----------
 *          gLAB v5.5.0
 * Release: 2020/11/13
 * Change Log:   Added NeQuick code from JRC.
 *               When selecting GEO with highest elevation, if any GEO is under alarm it will be excluded
 *                 from selection even if it is the one with highest elevation (with an explicit INFO
 *                 message informaing about the alarm state), except for the case that there is only GEO
 *                 in the SBAS data file (in this case it will stated that is selected because it is the
 *                 only GEO available).
 *               Fix occasional crash in precise orbit concatenation due to incorrect conditions check.
 *               Fix crash in precise orbit concatenation when using a precise clock file with at least
 *                5 concatenated days and in one of the days (not the first or the last) a satellite was
 *                available but in the other days it wasn't.
 *               Fixed memory leak from removed SP3 data when a precise orbit data was used with four or
 *                more concatenated days.
 *               Fixed all GEO elevations computed to value '0.0' for selection of GEO with highest elevatiom.
 *                 It ocurred when the SBAS data file started before the first epoch of the RINEX observation
 *                 file and on the first epoch of the observations file the SBAS data buffers were already full.
 *               Fixed incorrect computation of number of steps for IONEX maps when step size for 
 *                 latitude or longitude was equal to -1.
 * -----------
 *          gLAB v5.5.1
 * Release: 2020/12/11
 * Change Log: No changes in this file.
 * -----------
 *       END_RELEASE_HISTORY
 *****************************/

/* External classes */
#include "dataHandling.h"
#include "input.h"
#include "output.h"
#include <string.h>
#include <ctype.h>
#if defined _OPENMP
 #include <omp.h>
#endif

extern char printbufferMODEL[MAX_SATELLITES_VIEWED][MAX_INPUT_LINE];
extern char printbufferSBASCORR[MAX_SATELLITES_VIEWED][MAX_INPUT_LINE];
extern char printbufferSBASVAR[MAX_SATELLITES_VIEWED][MAX_INPUT_LINE];
extern char printbufferSBASIONO[MAX_SATELLITES_VIEWED][MAX_INPUT_LINE];
extern char printbufferSBASUNSEL[MAX_SATELLITES_VIEWED][MAX_INPUT_LINE];
extern int  linesstoredSATSEL[MAX_SATELLITES_VIEWED];
extern char printbufferSATSEL[MAX_SATELLITES_VIEWED][MAX_LINES_BUFFERED][MAX_INPUT_LINE];

extern int	ReadL1CAsItself;
const double EGM96grid[181][361];

/**************************************
 * Declarations of internal operations
 **************************************/

/*****************************************************************************
 * Name        : initOptions
 * Description : Initialises to the defaults a TOptions structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TOptions  *options              O  N/A  TOptions structure
 *****************************************************************************/
void initOptions (TOptions *options) {
	int	i,j,k;
	
	options->fdkml=NULL;
	options->fdkml0=NULL;
	options->KMLTimeStamps = 0;
	options->KMLEGM96height=1;
	options->KMLTimeStampDec = 30.;
	options->numKMLTimeRanges = 0;
	options->referenceStations = 1;
	options->KMLtimeranges=NULL;
	options->fdstdESALOI=NULL;
	options->SP3Letter='L';
	options->SP3PRN=9;
	options->YearSoDRefFile=1;
	options->GPSWeekRefFile=0;
	options->CalendarTimeRefFile=0;
	options->GeodeticCoordRefFile=0;
	options->WriteSPPsolInDGNSSmodeSP3=0;
	options->WriteSPPsolInDGNSSmodeRefFile=0;
	strcpy(options->model_dir,"model_files");
	strcpy(options->tropoGaldir,"TropoGal");
	strcpy(options->NeQuickdir,"NeQuick");

	// Verbose options
	options->printModel = -1;
	options->printPrefit = -1;
	options->printPostfit = -1;
	options->printFilterSolution = -1;
	options->printInfo = 1;
	options->printCycleslips = 1;
	options->printSatellites = -1;
	options->printInput = -1;
	options->printMeas = -1;
	options->printOutput = 1;
	options->printSatSel = -1;
	options->printSatDiff = 1;
	options->printSatStat = 1;
	options->printSatStatTot = 1;
	options->printSatPvt = 1;

	options->printSBASOUT    = 1;
	options->printSBASCORR   = 0;
	options->printSBASVAR    = 0;
	options->printSBASIONO   = 0;
	options->printSBASUNSEL  = 0;
	options->printSBASUNUSED = 0;
	options->printInBuffer   = 0;
	options->printUserError  = 1;

	options->printDGNSS = 1;
	options->printDGNSSUNUSED = 0;

	options->printSummary = 1;

	options->ProgressEndCharac='\r';

	options->satVel = svITRF;

	// Processing and modeling options
	options->workMode = wmDOPROCESSING;
	options->receiverPositionSource = rpUNKNOWN;
	options->aprioriReceiverPositionGeod[0]=0;
	options->aprioriReceiverPositionGeod[1]=0;
	options->aprioriReceiverPositionGeod[2]=-999999; //This is an invalid value to know that user has set no input for this value
	options->strictradome = 0;
	options->antennaData = adNONE;
	options->ARPData = arpUNKNOWN;
	for (j=0;j<3;j++) {
		options->receiverARP[j] = 0;
		for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) {
			options->receiverPCO[i][j] = 0;
		}
	}
	options->deprecatedMode = 0;
	options->prealignCP = 1;
	options->flexibleC1P1 = 1;
	options->csDataGap = -1;
	options->csLLI = 1;
	options->csNcon = -1;
	options->csNconMin = -2;
	options->csNconAutoDisabled = 0;
	options->csLI = 1;
	options->csLImaxjump = 1;
	options->csLImax = 0.08;
	options->csLIt = 60;
	options->csLIsamples = 7;
	options->csBW = 1;
	options->csBWInitStd = 2.;
	options->csBWmin = 0.8;
	options->csBWwindow = 300;
	options->csBWkfactor = 5;
	options->csBWsamples = 2;
	options->csL1C1 = 0;
	options->csUnconsistencyCheck = 1;
	options->csUnconsistencyCheckThreshold = 20.;
	options->csL1C1kfactor = 5.0;
	options->csL1C1window = 300;
	options->csL1C1init = 3.0;
	options->csL1C1samples = 2;
	options->timeTrans = 1;
	options->earthRotation = 1;
	options->satellitePhaseCenter = 1;
	options->satellitePhaseVar = 0;
	options->receiverPhaseVar = 0;
	options->relativisticCorrection = 1;
	options->windUpCorrection = 1;
	options->gravitationalCorrection = 1;
	options->solidTidesCorrection = 1;
	options->satelliteClockCorrection = 1;
	options->ionoModel = NoIono;
	options->ionoRMS = 100000;  //High value because it is not used. A normal value is 1
	options->troposphericCorrection = 1;
	options->useSigmaIono = 1;
	options->tropNominal = UNB3Nominal;
	options->tropMapping = SimpleMapping;
	options->p1c1dcbModel = p1c1FLEXIBLE;
// Found: options->p1p2dcbModel = p1c1NONE;
	options->p1p2dcbModel = p1p2NONE;
	options->SBAScorrections = 0;
	options->onlySBASiono = 0;
	options->decimate = -1.;
	options->orbitInterpolationDegree = 10;
	options->clockInterpolationDegree = 0;
	options->satelliteHealth = -1;
	options->brdcBlockTransTime = -1;
	options->MaxURABroadcast=9999999.;
	options->discardEclipsedSatellites = 0;
	options->maxKalmanIterations = 20;

	// Filtering options
	for (i=0;i<MAX_FILTER_MEASUREMENTS;i++) {
		//options->fixedMeasurementWeight[i] = 1;
		//options->elevationMeasurementWeight[i][0] = 0;
		//options->elevationMeasurementWeight[i][1] = 10 * d2r;
	}
	options->totalFilterMeasurements = 2;
	options->measurement[0] = PC;
	options->measurement[1] = LC;
	options->smoothMeas[0] = NA;
	options->smoothMeas[1] = NA;
	options->smoothEpochs = -1;
	options->maxdr2 = 10000;
	options->estimateTroposphere = 1;
	options->usePhase = 1;
	options->filterParams[PHI_PAR][DR_UNK] = 1;
	options->filterParams[PHI_PAR][DT_UNK] = 0;
	options->filterParams[PHI_PAR][TROP_UNK] = 1;
	options->filterParams[PHI_PAR][BIAS_UNK] = 1;
	options->filterParams[Q_PAR][DR_UNK] = 0;
	options->filterParams[Q_PAR][DT_UNK] = SIGMA_INF_DT*SIGMA_INF_DT;
	options->filterParams[Q_PAR][TROP_UNK] = 1e-4;
	options->filterParams[Q_PAR][BIAS_UNK] = 0;
	options->filterParams[P0_PAR][DR_UNK] = SIGMA_INF_DR*SIGMA_INF_DR;
	options->filterParams[P0_PAR][DT_UNK] = SIGMA_INF_DT*SIGMA_INF_DT;
	options->filterParams[P0_PAR][TROP_UNK] = 0.5 * 0.5;
	options->filterParams[P0_PAR][BIAS_UNK] = 20 * 20;
	options->elevationMask = 5. * d2r;
	options->filterIterations = 1;
	options->NextSP3 = 0;
	options->concatenSP3 = 0;
	options->LastSamplePrevDayOrb=0;
	options->LastSamplePrevDayClk=0;
	options->numPreciseFiles = 0;
	options->OrbitsMaxGaps=-1;
	options->ClocksMaxGaps=-1;
	options->OrbitsToTMaxGaps=-1;
	options->ClocksToTMaxGaps=-1;
	options->OrbitsMaxDistance=-1.;
	options->ClocksMaxDistance=-1.;
	options->OrbitsToTMaxDistance=-1.;
	options->ClocksToTMaxDistance=-1.;
	options->checkPhaseCodeJumps = 1;
	
	
	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			options->includeSatellite[i][j] = 1;
			options->SNRvalues[i][j] = 33.;
			for (k=0;k<MAX_FILTER_MEASUREMENTS;k++) {
				options->SNRweightComb[i][j][k]=SNRWeightCombMean;
				options->SNRweightCombVal[i][j][k][0]=0.5;
				options->SNRweightCombVal[i][j][k][1]=0.5;
				options->WeightConstantsValues[i][j][k][0]=1.;
				options->WeightConstantsValues[i][j][k][1]=0.;
				options->WeightConstantsValues[i][j][k][2]=0.;
				options->weightMode[i][j][k]=FixedWeight;
			}
		}
		for (j=0;j<MAX_FREQUENCIES_PER_GNSS;j++) {
			options->usableFreq[i][j] = 1;
		}
	}

	options->SNRfilter = -1;
	options->useraddednoise = 0;
	options->useraddednoisepreprocess = 0;

	options->writeRinexBfile = 0;
	options->writeEMSfile = 0;
	options->writePegasusfile = 0;
	options->onlyconvertSBAS = 0;
	options->onlySBASplots = 0;
	options->pegasuspurerinex = 0;
	options->pegasusaligned = 0;
	options->pegasusfs = ';';
	strcpy(options->sbasdir,SBASOUTPUTPATH);


	options->usetype0messages = 1;
	options->alarmmessageastype2 = 0;
	options->precisionapproach = PAMODE;
	options->UserForcedNPA = 0;
	options->excludeSmoothingConvergence = -1;
	options->switchmode = -1;
	options->GEOfallback = 0;
	options->MaintainGEO = 1;
	options->adquisitiontime=GEOACQUISIONTIME;
	options->minimumswitchtime=-1;
	options->GEOindex = -1;
	options->GEOPRN = -1;
	options->GEOPRNSelected = -1;
	options->switchGEO = -1;
	options->selectBestGEO = -1;
	options->mixedGEOdata = -1;
	options->initcoordNPA = 1;
	options->UDREIthreshold = 999999;
	options->SigmaOffsetNoMT10 = 8.;
	options->airborneReceiverType = 2;
	options->stepdetector = -1;
	options->SBASmaritime=0;
	options->prefitOutlierDetectorAbs = 0;
	options->prefitOutlierDetectorAbsThreshold= 30.;
	options->prefitOutlierDetectorMedian = 0;
	options->prefitOutlierDetectorMedianThreshold = 30.;
	options->FCtmout[0] = options->FCtmout[1] = -1;
	options->RRCtmout[0] = options->RRCtmout[1] = -1;
	options->usersigmamultipath = 0;
	options->sigma2multipath[0] = options->sigma2multipath[1] = -99999.;
	options->sigma2multipath[2] = 0.1;
	options->sigma2noise = -1;
	options->sigma2divergence = -1;
	options->kfactor[NPAMODE][KHORPOS] = KHNPA;
	options->kfactor[NPAMODE][KVERPOS] = KV;
	options->kfactor[PAMODE][KHORPOS] = KHPA;
	options->kfactor[PAMODE][KVERPOS] = KV;
	options->stanfordesa = 0;
	options->stanfordesaLOI = 1;
	options->WIRHorThreshold = 0.7;
	options->WIRVerThreshold = 0.7;
	options->waitForNextDay = 1;
	options->useDatasummary = 1;
	options->SBASHourlyMaps = 0;
	options->NoAvailabilityPlot = 0;
	options->NoContRiskPlot = 0;
	options->NoContRiskMarPlot = 0;
	options->ComputeRiskPlots=1;
	options->NoIonoPlot = 0;
	options->HDOPPlot = 0;
	options->PDOPPlot = 0;
	options->GDOPPlot = 0;
	options->CombinedDOPPlot=0;
	options->DOPPercentile=0;
	options->minLatplots = 25.0;
	options->maxLatplots = 70.0;
	options->minLonplots = -30.0;
	options->maxLonplots = 40.0;
	options->AvailabilityPlotStep = 1.0;
	options->IonoPlotStep = 0.3;
	options->IonoPlotTimeStep = 300;
	options->ContRiskWindowSize = 15;
	options->ContRiskWindowSizeMar = 15;
 	options->VerAlarmLimit = 50.0;
 	options->HorAlarmLimit = 40.0;
	options->percentile=95.;
	options->SBASPlotsRecHeight = 0;
	options->NumSBASPlotsExclusionArea=0;
	options->SBASPlotsExclusionAreaDelimeters=NULL;
	options->NumSBASPlotsInclusionArea=0;
	options->SBASPlotsInclusionAreaDelimeters=NULL;
	options->NoFastCorrection = 0;
	options->NoRRCCorrection = 0;
	options->NoSlowCorrection = 0;
	options->NoIonoCorrection= 0;
	options->NoDeltaUDRE = 0;
	options->NoFastSigma = 0;
	options->NoFastDeg = 0;
	options->NoRRCSigma = 0;
	options->NoSlowSigma = 0;
	options->NoIonoSigma = 0;
	options->NoTropoSigma = 0;
	options->NoEnRouteSigma = 0;
	options->NoMT10=0;
	// DGNSS
	options->DGNSS = 0;
	options->dgnssMaxEph = 4.0 * 3600.0; // seconds
	options->dgnssSigma2noiseGPS = 0.25;
	options->dgnssSigma2multipath = 0.0625;
	options->epsURA3o = 0.01;
	options->sigmaInflation = 1;
	options->maxAgeDGNSS = 31.0; // seconds
	options->maxDGNSSPRC = 500.; // meters
	options->excludeSmoothingConvergenceUser = 0;
	options->excludeSmoothingConvergenceRef = 1;
	options->rtcmYear = 0;
	options->rtcmYearShort = 0;
	options->rtcmMonth = 0;
	options->rtcmDay = 0;
	options->rtcmHour = -1;
	options->RTCMmode = RTCMUNKNOWN;
	options->rinexVersion = 3;
	options->GDOPthreshold = -1;
	options->HDOPthreshold = -1;
	options->PDOPthreshold = -1;
	options->HDOPorPDOP = 0;
	options->SBASplotsDOP=SBASplotsNoDOP;
	options->maxGDOP = 30.0; 
	options->maxHDOP = 4.0; 
	options->maxPDOP = 6.0; 
	options->excludeSmoothingConvergenceTime = 360; // sec
	options->DF = 0;

	for (i=0;i<MAX_SBAS_PRN;i++) {
		options->GEOPRNunsel[i] = 0;
	}

	options->solutionMode = SPPMode;

	options->OutputToTerminal=0;
	options->outFileStream=stdout;
	options->terminalStream=stdout;

	options->StartEpoch.MJDN=-1;
	options->StartEpoch.SoD=0.;
	options->EndEpoch.MJDN=-1;
	options->EndEpoch.SoD=0.;
	options->firstEpochSummary.MJDN=-1;
	options->firstEpochSummary.SoD=0.;
	options->recPosProvided=0;
	strcpy(options->StationNetworkName,"Other");
	

	options->RefDegInterp=0;

	#if defined _OPENMP
		options->numthreads=omp_get_num_procs();
	#endif

}

/*****************************************************************************
 * Name        : initSatellite
 * Description : Initialise a TSatellite struct. It is only called after a cycle-slip is found
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch struct
 * int  satIndex                   I  N/A  Pointing to the sat[satIndex]
 *                                         TSatellite structure that should be 
 *                                         initialised
 * int  satCSindex                 I  N/A  Pointing to the sat[satCSIndex] 
 *                                         TCS structure that should be initialised
 *****************************************************************************/
void initSatellite (TEpoch *epoch, int satIndex, int satCSindex) {
	int	i;

	epoch->cycleslip.arcLength[satCSindex] = 1;
	epoch->cycleslip.outlierLI[satCSindex] = 0;
	epoch->cycleslip.outlierBw[satCSindex] = 0;
	epoch->cycleslip.CS[satCSindex] = 1;
	epoch->cycleslip.CSPrealignFlag[satCSindex] = 1;
	epoch->cycleslip.narc[satCSindex]++;
	epoch->cycleslip.hasBeenCycleslip[satCSindex] = 1;
	epoch->cycleslip.windUpRadAccumReceiver[satCSindex] = 0.0;
	epoch->cycleslip.windUpRadAccumSatellite[satCSindex] = 0.0;
	epoch->sat[satIndex].ionoSigma = 0;
	epoch->sat[satIndex].hasSBAScor = 1;
	epoch->sat[satIndex].hasSNR = 1;
	
	for (i=0;i<MAX_FILTER_MEASUREMENTS;i++) {
		epoch->cycleslip.smoothedMeas[satCSindex][i] = 0.0;
	}

	for (i=0;i<epoch->measOrder[epoch->sat[satIndex].GNSS].nDiffMeasurements;i++) {
		epoch->cycleslip.preAlign[satCSindex][i] = 0.0;
	}

	for (i=0;i<MAX_SLIDING_WINDOW;i++) {
		epoch->cycleslip.windowMW[satCSindex][i] = 0.0;
		epoch->cycleslip.windowMWtime[satCSindex][i] = 0.0;
		epoch->cycleslip.windowL1C1[satCSindex][i] = 0.0;
		epoch->cycleslip.windowL1C1time[satCSindex][i] = 0.0;
	}
}

/*****************************************************************************
 * Name        : initEpoch
 * Description : Initialise a TEpoch struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  O  N/A  TEpoch struct
 * TOptions  *options              O  N/A  TOptions structure
 *****************************************************************************/
void initEpoch (TEpoch *epoch, TOptions  *options) {
	int i,j,k;

	epoch->firstepoch.SoD=-1.;
	epoch->secondepoch.SoD=-1.;
	epoch->PreviousEpoch.MJDN = -1;
	epoch->PreviousEpoch.SoD = -1;
	epoch->t.MJDN=-1;
	epoch->t.SoD=-1.;
	epoch->numEpochsfile=1.;
	epoch->numEpochsprocessed=-1.;
	epoch->PreviousPercentage=0;
	epoch->CurrentPercentage=0.;
	epoch->NewPercentage=0;

	epoch->numSatellites = 0;
	epoch->numAproxPosRead = 0;
	epoch->lastSBASindex = 0;
	epoch->prevNumSatellites = 0;
	epoch->numGNSS = 0;
	epoch->receiver.recType = rtUNKNOWN;
	epoch->receiver.interval = -1;
	epoch->receiver.intervalEstimate= -1;
	epoch->receiver.intervalDecimalsFactor = 1;
	epoch->receiver.DecimateVal=99999;
	strcpy(epoch->receiver.name,DEFAULTMARKERNAME);
	epoch->receiver.type[0] = '\0';
	epoch->receiver.aproxPosition[0] = EARTH_RADIUS;
	epoch->receiver.aproxPosition[1] = 0;
	epoch->receiver.aproxPosition[2] = 0;
	epoch->receiver.aproxPositionError = 1e10;
	epoch->receiver.equivalentC1P1 = 0;
	epoch->receiver.receiverSolConverged = 1;
	epoch->receiver.aproxPositionRover[0] = 0.;
	epoch->receiver.aproxPositionRover[1] = 0.;
	epoch->receiver.aproxPositionRover[2] = 0.;

	epoch->prealign = 0;
	epoch->ZTD = 0;
	epoch->IPPlat = 0;
	epoch->IPPlon = 0;

	epoch->dgnss.Nref = 0;
	epoch->dgnss.CLKref = 0.0;
	epoch->dgnss.epsURA = 0.01;
	
	for (i=0;i<3;i++) {
		epoch->solidTideDisplacement[i] = 0;
		epoch->receiver.ARP[i] = 0;
		for (k=0;k<MAX_GNSS;k++) {
			for (j=0;j<MAX_FREQUENCIES_PER_GNSS;j++) {
				epoch->receiver.antenna.PCO[k][j][i] = 0.;
			}
		}
	}

	for (i=0;i<MAX_SATELLITES_VIEWED;i++) {
		epoch->satsviewed[i] = 0;
		epoch->sat[i].azimuth = 0.0;
		epoch->sat[i].elevation = 0.0;
		epoch->cycleslip.hasBeenCycleslip[i] = 0;
		epoch->cycleslip.use4smooth[i] = 1;
		epoch->cycleslip.preCheck[i] = 0;
		epoch->cycleslip.outlierLI[i] = 0;
		epoch->cycleslip.outlierBw[i] = 0;
		epoch->cycleslip.consistency[i] = 0;
		epoch->cycleslip.CS[i] = 0;
		epoch->cycleslip.CSPrealignFlag[i] = 0;
		epoch->cycleslip.narc[i] = 0;
		epoch->cycleslip.arcLength[i] = 0;
		epoch->cycleslip.Nconsecutive[i] = 0;
		epoch->cycleslip.initialLi[i] = 0.0;
		epoch->cycleslip.BWmean[i] = 0.0;
		epoch->cycleslip.BWmean300[i] = 0.0;
		epoch->cycleslip.BWsigma[i] = options->csBWInitStd*options->csBWInitStd;
		epoch->cycleslip.L1C1mean[i] = 0.0;
		epoch->cycleslip.L1C1sigma[i] = 0.0;
		epoch->cycleslip.L1Prev[i] = 0.0;
		epoch->cycleslip.C1Prev[i] = 0.0;
		epoch->cycleslip.lastModelledEpoch[i].MJDN=-1;
		epoch->cycleslip.lastModelledEpoch[i].SoD=0.;
		epoch->sat[i].ionoSigma = 0.0;
		for (j=0;j<MAX_MEASUREMENTS_PER_SATELLITE;j++) {
			epoch->cycleslip.preAlign[i][j] = 0.0;
		}
		for (j=0;j<MAX_FILTER_MEASUREMENTS;j++) {
			epoch->cycleslip.smoothedMeas[i][j] = 0.0;
		}
		for (j=0;j<MAX_SLIDING_WINDOW;j++) {
			epoch->cycleslip.windowMW[i][j] = 0.0;
			epoch->cycleslip.windowMWtime[i][j] = 0.0;
			epoch->cycleslip.windowL1C1[i][j] = 0.0;
			epoch->cycleslip.windowL1C1time[i][j] = 0.0;
		}
		// DGNSS
		epoch->dgnss.msg[i] = 0;
		epoch->dgnss.health[i] = 0;
		epoch->dgnss.IODmsg1[i] = 0;
		epoch->dgnss.IODmsg2[i] = 0;
		epoch->dgnss.PRC[i] = 0.0;
		epoch->dgnss.PRCp[i] = 0.0;
		epoch->dgnss.sigmaInflation[i] = 1.0;
		epoch->dgnss.sigmaURA3[i] = 0.0;
		epoch->dgnss.sigmaURA3p[i] = 0.0;
		epoch->dgnss.C1[i] = 0.0;
		epoch->dgnss.C1s[i] = 0.0;
		epoch->dgnss.UDRE[i] = 0.0;
		epoch->dgnss.tPRC[i].SoD = 0.0;
		epoch->dgnss.tPRC[i].MJDN = 0.0;
		epoch->dgnss.tdeltaPRC[i].SoD = 0.0;
		epoch->dgnss.tdeltaPRC[i].MJDN = 0.0;
		epoch->DGNSSstruct=0;
	}

	for (j=0;j<MAX_GNSS;j++) {
		for (i=0;i<MAX_MEASUREMENTS;i++) {
			epoch->measOrder[j].meas2Ind[i] = -1;
			epoch->measOrder[j].conversionFactor[i] = 0;
			epoch->measOrder[j].usable[i] = 1;
		}
		for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
			epoch->satInfo[j][i].lastEclipse.SoD = 0;
			epoch->satInfo[j][i].lastEclipse.MJDN = 0;
			epoch->satIndex[j][i]=-1;
			epoch->satCSIndex[j][i]=-1;
		}
		epoch->measOrder[j].hasSNRmeas=0;
		for(i=0;i<MAX_FREQUENCIES_PER_GNSS;i++) {
			epoch->measOrder[j].SNRmeaspos[i]=-1;
		}
	}

	epoch->ResetNumSwitch=0;
	epoch->currentGEOPRN=0;

	epoch->KMLData=NULL;
	epoch->KMLTime=NULL;
	epoch->numKMLData=0;
	epoch->receiver.numRecStation=0;
	epoch->receiver.RecStationCoord=NULL;
	epoch->receiver.ChangedStaEpoch=NULL;
	epoch->receiver.look4interval = 1;
	epoch->receiver.RefPositionAvail=1;

	//Summary variables
	epoch->TotalEpochs=0;
	epoch->TotalEpochsForward=0;
	epoch->TotalEpochsDecimated=0;
	epoch->TotalEpochsDecimatedForward=0;
	epoch->TotalEpochsFile=0;
	epoch->TotalEpochsSol=0;
	epoch->TotalEpochsDGNSS=0;
	epoch->TotalEpochsSPP=0;
	epoch->TotalEpochsRef=0;
	epoch->NumNoRefSumSkipped=0;
	epoch->NumNoRefSolSkipped=0;
	epoch->NumDOPSkipped  = 0;
	epoch->NumHDOPSkipped = 0;
	epoch->NumPDOPSkipped = 0;
	epoch->NumGDOPSkipped = 0;
	epoch->NumHDOPorPDOPSkipped = 0;
	epoch->NumSingularMatrixSkipped = 0;
	epoch->NumNoSatSkipped = 0;
	epoch->NumDataGapsSum = 0;
	epoch->NumEpochsMissingSum = 0;
	epoch->NumEpochsMissingTotal = 0;
	epoch->MaxDataGapSizeSum = 0;
	epoch->NumDataGapsTotal = 0;
	epoch->MaxDataGapSizeTotal = 0;
	epoch->HDOPPercentileSamples = -1.;
	epoch->PDOPPercentileSamples = -1.;
	epoch->GDOPPercentileSamples = -1.;
	epoch->TDOPPercentileSamples = -1.;
	epoch->VDOPPercentileSamples = -1.;
	epoch->HPEPercentileSamples = -1.;
	epoch->VPEPercentileSamples = -1.;
	epoch->MaxHError = -1.;
	epoch->MaxVError = -1.;
	epoch->MaxHDOP = -1.;
	epoch->MaxPDOP = -1.;
	epoch->MaxGDOP = -1.;
	epoch->MaxVDOP = -1.;
	epoch->MaxTDOP = -1.;
	epoch->HError = NULL;
	epoch->VError = NULL;
	epoch->HDOP = NULL;
	epoch->PDOP = NULL;
	epoch->GDOP = NULL;
	epoch->VDOP = NULL;
	epoch->TDOP = NULL;
	epoch->StartSummaryEpoch.MJDN=-1;
	epoch->StartSummaryEpoch.SoD=0.;
	epoch->LastSummaryEpoch.MJDN=-1;
	epoch->LastSummaryEpoch.SoD=0.;
	epoch->StartSummaryPercentileEpoch.MJDN=-1;
	epoch->StartSummaryPercentileEpoch.SoD=0.;
	epoch->LastSummaryPercentileEpoch.MJDN=-1;
	epoch->LastSummaryPercentileEpoch.SoD=0.;

	//Variables for backing up data when selecting the GEO with smallest protection levels
	epoch->Buffer_NumDataStored=0;
	epoch->Buffer_NumDataStoredUnderAlarmLimitsPA=0;
	epoch->Buffer_NumDataStoredPA=0;
	epoch->Buffer_NumDataStoredNPA=0;
	epoch->Buffer_PosDataStoredUnderAlarmLimitsPA=NULL;
	epoch->Buffer_PosDataStoredPA=NULL;
	epoch->Buffer_PosDataStoredNPA=NULL;
	epoch->Buffer_numsatdiscardedSBAS=NULL;
	epoch->Buffer_usableSatellites=NULL;
	epoch->Buffer_HPL=NULL;
	epoch->Buffer_VPL=NULL;
	epoch->Buffer_HDOP=NULL;
	epoch->Buffer_VDOP=NULL;
	epoch->Buffer_PDOP=NULL;
	epoch->Buffer_GDOP=NULL;
	epoch->Buffer_TDOP=NULL;
	epoch->Buffer_printMODEL=NULL;
	epoch->Buffer_printSBASIONO=NULL;
	epoch->Buffer_printSBASCORR=NULL;
	epoch->Buffer_printSBASVAR=NULL;
	epoch->Buffer_printSBASUNSEL=NULL;
	epoch->Buffer_printSATSEL=NULL;
	epoch->Buffer_NumSatSel=NULL;
	epoch->Buffer_SBASMode=NULL;
	epoch->Buffer_GEOindex=NULL;
	epoch->Buffer_sat=NULL;
}

/*****************************************************************************
 * Name        : initGNSSproducts
 * Description : Initialise a TGNSSproducts struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts*  products        O  N/A  TGNSSproducts struct
 *****************************************************************************/
void initGNSSproducts (TGNSSproducts *products) {
	products->BRDC = NULL;
	products->SP3 = NULL;
}

/*****************************************************************************
 * Name        : initBRDCproducts
 * Description : Initialise a TBRDCproducts struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCproducts  *BRDC            O  N/A  TBRDCproducts struct
 *****************************************************************************/
void initBRDCproducts (TBRDCproducts *BRDC) {
	int i,j;

	BRDC->numsats = 0;
	BRDC->numsatsPast = 0;
	BRDC->block = NULL;
	BRDC->blockPast = NULL;
	BRDC->numblocks = NULL;
	BRDC->numblocksPast = NULL;

	for (i=0;i<4;i++) {
		BRDC->ionA[i] = 0;	//Rinex v2
		BRDC->ionB[i] = 0;	//Rinex v2
		BRDC->ai[i] = 0;	//Rinex v3
		BRDC->bdsA[i] = 0;	//Rinex v3
		BRDC->bdsB[i] = 0;	//Rinex v3
	}

	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			BRDC->index[i][j] = -1;
			BRDC->indexPast[i][j] = -1;
		}
	}

	BRDC->LeapSecondsAvail=0;
	BRDC->AT_LS  = 0;
	BRDC->AT_LSF = 0;
	BRDC->WN_LSF = 0;
	BRDC->DN     = 0;

	//Rinex v3
	for(i=0;i<6;i++) {
		BRDC->timeSysCorr[i].acoff[0] = 0;
		BRDC->timeSysCorr[i].acoff[1] = 0;
		BRDC->timeSysCorr[i].timeref = 0;
		BRDC->timeSysCorr[i].weekref = 0;
	} 
}

/*****************************************************************************
 * Name        : initBRDCblock
 * Description : Initialise a TBRDCblock struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCblock  *BRDCblock          O  N/A  TBRDCblock struct
 *****************************************************************************/
void initBRDCblock (TBRDCblock *BRDCblock) {
	BRDCblock->PRN = 0;
	BRDCblock->clockbias = 0;
	BRDCblock->clockdrift = 0;
	BRDCblock->clockdriftrate = 0;
	BRDCblock->IODE = 0;
	BRDCblock->IODNav = 0;
	BRDCblock->crs = 0;
	BRDCblock->deltan = 0;
	BRDCblock->M0 = 0;
	BRDCblock->cuc = 0;
	BRDCblock->e = 0;
	BRDCblock->cus = 0;
	BRDCblock->sqrta = 0;
	BRDCblock->toe = 0;
	BRDCblock->cic = 0;
	BRDCblock->OMEGA = 0;
	BRDCblock->cis = 0;
	BRDCblock->i0 = 0;
	BRDCblock->crc = 0;
	BRDCblock->omega = 0;
	BRDCblock->OMEGADOT = 0;
	BRDCblock->IDOT = 0;
	BRDCblock->codesOnL2 = 0;
	BRDCblock->GPSweek = 0;
	BRDCblock->L2Pdataflag = 0;
	BRDCblock->dataSources = 0;
	BRDCblock->GALweek = 0;
	BRDCblock->BDSweek = 0;
	BRDCblock->SVaccuracy = 0;
	BRDCblock->SVhealth = 0;
	BRDCblock->TGD = 0;
	BRDCblock->TGD2 = 0;
	BRDCblock->IODC = 0;
	BRDCblock->SISASignal = 0;
	BRDCblock->BGDE5a = 0;
	BRDCblock->BGDE5b = 0;
	BRDCblock->transTime = 0;
	BRDCblock->fitInterval = 0;
	BRDCblock->transTimeGAL = 0;
	BRDCblock->satposX = 0;
	BRDCblock->satvelX = 0;
	BRDCblock->sataccX = 0;
	BRDCblock->satposY = 0;
	BRDCblock->satvelY = 0;
	BRDCblock->sataccY = 0;
	BRDCblock->freqnumber = 0;
	BRDCblock->satposZ = 0;
	BRDCblock->satvelZ = 0;
	BRDCblock->ageofoperation = 0;
	BRDCblock->IODNGEO = 0;
	BRDCblock->URAGEO = 0;
}

/*****************************************************************************
 * Name        : initSP3products
 * Description : Initialise a TSP3products struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSP3products  *SP3              O  N/A  TSP3products struct
 *****************************************************************************/
void initSP3products (TSP3products *SP3) {
	int i,j;

	SP3->orbits.numSatellites = 0;
	SP3->orbits.numSatellitesRead = 0;
	SP3->orbits.block = NULL;
	SP3->orbits.accuracy = NULL;
	SP3->orbits.numRecords = 0;
	SP3->orbits.interval = 0.;
	SP3->orbits.velocityAvailable = 0;
	SP3->orbits.startTime.MJDN = 0;
	SP3->orbits.startTime.SoD = 0.;
	SP3->orbits.startTimeConcat.MJDN = 0;
	SP3->orbits.startTimeConcat.SoD = 0.;

	SP3->clocks.numSatellites = 0;
	SP3->clocks.numSatellitesRead = 0;
	SP3->clocks.block = NULL;
	SP3->clocks.numRecords = 0;
	SP3->clocks.vsizeblock = 0;
	SP3->clocks.interval = 0.;
	SP3->clocks.startTime.MJDN = 0;
	SP3->clocks.startTime.SoD = 0.;
	SP3->clocks.startTimeConcat.MJDN = 0;
	SP3->clocks.startTimeConcat.SoD = 0.;

	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			SP3->clocks.index[i][j] = -1;
			SP3->orbits.index[i][j] = -1;
			SP3->orbits.numblocsConcat[i][j]=0;
			SP3->clocks.numblocsConcat[i][j]=0;
			SP3->orbits.numblocsValidConcat[i][j]=0;
			SP3->clocks.numblocsValidConcat[i][j]=0;
		}
	}

	for(i=0;i<MAX_GNSS*MAX_SATELLITES_PER_GNSS;i++) {
		SP3->orbits.Concatblock[i] = NULL;
		SP3->clocks.Concatblock[i] = NULL;
	}
}

/*****************************************************************************
 * Name        : initSP3orbitblock
 * Description : Initialise a TSP3orbitblock struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSP3orbitblock  *SP3            O  N/A  TSP3orbitblock struct
 *****************************************************************************/
void initSP3orbitblock (TSP3orbitblock *SP3orbitblock) {
	SP3orbitblock->v[0] = 0.;
	SP3orbitblock->v[1] = 0.;
	SP3orbitblock->v[2] = 0.;
	SP3orbitblock->x[0] = 0.;
	SP3orbitblock->x[1] = 0.;
	SP3orbitblock->x[2] = 0.;
	SP3orbitblock->xsigma[0] = 0.;
	SP3orbitblock->xsigma[1] = 0.;
	SP3orbitblock->xsigma[2] = 0.;
	SP3orbitblock->orbitflags[0] = '\0';
	SP3orbitblock->validSample=0;
}

/*****************************************************************************
 * Name        : initSP3clockblock
 * Description : Initialise a TSP3clockblock struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSP3clockblock  *SP3            O  N/A  TSP3clockblock struct
 *****************************************************************************/
void initSP3clockblock (TSP3clockblock *SP3clockblock) {
	SP3clockblock->clock = INVALID_CLOCK;
	SP3clockblock->clockrate = 0;
	SP3clockblock->clockdriftrate = 0;
	SP3clockblock->clockflags[0] = '\0';
	SP3clockblock->clocksigma = 0.;
	SP3clockblock->validSample=0;
}

/*****************************************************************************
 * Name        : initIONEXDCB
 * Description : Initialise a TIonexDCB structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonexDCB  *IonexDCB            O  N/A  TIonexDCB struct
 *****************************************************************************/
void initIONEXDCB (TIonexDCB *IonexDCB){

	int i,j;

	IonexDCB->DCBavailable = 0;

	
	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			IonexDCB->DCB[i][j] = -999999.;
			IonexDCB->DCBRMS[i][j] = 1e6;
		}
	}
}

/*****************************************************************************
 * Name        : initIONEX
 * Description : Initialise a TIONEX structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIONEX  *IONEX                  O  N/A  TIONEX struct
 *****************************************************************************/
void initIONEX (TIONEX *IONEX) {

	IONEX->map = NULL;
	IONEX->numMaps = 0;
	IONEX->RMSavailable = 0;	//RMS values are not mandatory
	IONEX->HGTavailable = 0;	//HGT values are not mandatory
	IONEX->conversionFactor = 0.1; // Default conversion factor
	IONEX->exponent = -1;
	IONEX->satNumber = 0 ;
	IONEX->stationNumber = 0 ;
	IONEX->mapDimension = 2 ; // typically 2D dimension maps
	IONEX->numStationData = 0;
	IONEX->baseRadius = EARTH_RADIUS;
	IONEX->ionexStation = NULL;
	strcpy(IONEX->mappingFunction,"COSZ");

	initIONEXDCB(&IONEX->ionexDCB);
}

/*****************************************************************************
 * Name        : initIonoMap
 * Description : Initialise a TIonoMap structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              O  N/A  TIonoMap struct
 *****************************************************************************/
void initIonoMap (TIonoMap *ionoMap) {
	ionoMap->dhgt = -1;
	ionoMap->dlat = -1;
	ionoMap->dlon = -1;

	ionoMap->RMS = NULL;
	ionoMap->TEC = NULL;
	ionoMap->HGT = NULL;
}

/*****************************************************************************
 * Name        : initIonexStation
 * Description : Initialise a TIonexStation structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonexStation  *ionexStation    O  N/A  TIonexStation struct
 *****************************************************************************/
void initIonexStation (TIonexStation *ionexStation) {
	ionexStation->stationInitials[0] = '\0';
	ionexStation->stationcode[0] = '\0';
	ionexStation->stationBias = 0;
	ionexStation->stationRMS = 0;
}

/*****************************************************************************
 * Name        : initFPPPDCB
 * Description : Initialise a TFPPPDCB structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPPDCB  *FPPPDCB              O  N/A  TFPPPDCB struct
 *****************************************************************************/
void initFPPPDCB (TFPPPDCB *FPPPDCB) {

	int i,j;	

	FPPPDCB->DCBavailable = 0;

	for (i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_SATELLITES_PER_GNSS;j++) {
			FPPPDCB->frac1[i][j] = 0;
			FPPPDCB->fracw[i][j] = 0;
			FPPPDCB->IFB[i][j] = -999999.;
		}
	}
}

/*****************************************************************************
 * Name        : initFPPPIONEX
 * Description : Initialise a TFPPPIONEX structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPP  *FPPP                    O  N/A  TFPPPIONEX struct
 *****************************************************************************/
void initFPPPIONEX (TFPPPIONEX *FPPP) {

	FPPP->numMaps = 0;
	FPPP->baseRadius = EARTH_RADIUS/1000; // In Km
	strcpy(FPPP->mappingFunction,"COSZ");

	initFPPPDCB(&FPPP->fpppDCB);

}

/*****************************************************************************
 * Name        : initFPPPMap
 * Description : Initialise a TFPPPMap structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPPMap  *FPPPMap              O  N/A  TFPPPMap struct
 *****************************************************************************/
void initFPPPMap (TFPPPMap *FPPPMap) {
	FPPPMap->dhgt = 0;
	FPPPMap->dlat = -1;
	FPPPMap->dlt = -1;
	
	FPPPMap->lat1 = -90;
	FPPPMap->lat2 = 90;
	
	FPPPMap->lt1 = 0;
	FPPPMap->lt2 = 360;

	FPPPMap->TEC = NULL;
	FPPPMap->SIGMA = NULL;
}


/*****************************************************************************
 * Name        : initFPPPMapData
 * Description : Fill with 9999 the TEC data vector of a TFPPPMap structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPPMap  *FPPPMap              O  N/A  TFPPPMap struct
 *****************************************************************************/
void initFPPPMapData (TFPPPMap *FPPPMap){
	int nh,nlat,nlt;
	int mapType=0;
	double hgt,lt,lat;
	for(nh=0; nh<NUMBERFPPPMAPS;nh++)
	{
		hgt = FPPPMap->hgt1 + nh*FPPPMap->dhgt;
		
		for(nlat=0; nlat<FPPPMap->nlat;nlat++)
		{
			lat = FPPPMap->lat1 + nlat*FPPPMap->dlat;
			
			for(nlt=0; nlt<FPPPMap->nlt;nlt++)
			{
				lt = FPPPMap->lt1 + nlt*FPPPMap->dlt;
				setFPPPValue(FPPPMap, mapType, hgt, lat, lt, 9999);
				setFPPPValue(FPPPMap, 1-mapType, hgt, lat, lt, 9999);
			}
		}
	}
}


/*****************************************************************************
 * Name        : initiPiercePoint
 * Description : Initialise a TPiercePoint structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonexStation  *ionexStation    O  N/A  TPiercePoint struct
 *****************************************************************************/
void initPiercePoint (TPiercePoint *PiercePoint){
	PiercePoint->x = 0;
	PiercePoint->y = 0;
	PiercePoint->z = 0;
	PiercePoint->fi = 0;
	PiercePoint->longitude = 0;
	PiercePoint->elevation = 0;
	PiercePoint->localtime = 0;
	PiercePoint->mappingFunction = 0;
	
}

/*****************************************************************************
 * Name        : initTropoGalData
 * Description : Initialise a TTROPOGal structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTROPOGal *TropoGal             O  N/A  TTROPOGal struct
 *****************************************************************************/
void initTropoGalData (TTROPOGal *TropoGal) {

	int i;

	TropoGal->lat_pix_positions = 121; 	//Default Gal value
	TropoGal->lon_pix_positions = 241;	//Default Gal value 
	TropoGal->delta_lat_pix = -1.5;		//Default Gal value
	TropoGal->delta_lon_pix = 1.5;		//Default Gal value 
	TropoGal->lat_pix_init = 90;		//Map from 90 to -90 degrees
	TropoGal->lat_pix_end = -90;		//Map from 90 to -90 degrees
	TropoGal->lon_pix_init = 0;		//Map from 0 to 360 degrees
	TropoGal->lon_pix_end = 360;		//Map from 0 to 360 degrees

	for(i=0;i<3;i++) {
		TropoGal->am_a_size[i] = 0;	
		TropoGal->es_a_size[i] = 0;	
		TropoGal->ld_a_size[i] = 0;	
		TropoGal->ps_a_size[i] = 0;	
		TropoGal->tm_a_size[i] = 0;	
	}

	for(i=0;i<4;i++) {
		TropoGal->es_b_size[i] = 0;	
		TropoGal->ld_b_size[i] = 0;	
		TropoGal->tm_b_size[i] = 0;	
	}

	TropoGal->hgt_pix_values = NULL;	
	TropoGal->lnd_values  = NULL;	
	TropoGal->am_a_values = NULL;
	TropoGal->es_a_values = NULL;
	TropoGal->ld_a_values = NULL;
	TropoGal->ps_a_values = NULL;
	TropoGal->tm_a_values = NULL;
	TropoGal->es_b_values = NULL;
	TropoGal->ld_b_values = NULL;
	TropoGal->tm_b_values = NULL;

}

/*****************************************************************************
 * Name        : initTropoGalParameters
 * Description : Initialise a TTROPOGalParameters structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTROPOGal *TropoGal             I  N/A  TTROPOGal struct
 * TTROPOGalParameters *GalParam   O  N/A  TTROPOGalParameters struct
 *****************************************************************************/
void initTropoGalParameters (TTROPOGal *TropoGal,TTROPOGalParameters *GalParam) {
	
	int i;

	for (i=0;i<4;i++) {
		GalParam->gm[i]          = 0; 
		GalParam->ph[i]          = 0;
		GalParam->eh[i]          = 0;
		GalParam->tmh[i]         = 0;
		GalParam->g0[i]          = 0;
		GalParam->ps[i]          = 0; 
		GalParam->tm[i]          = 0; 
		GalParam->am[i]          = 0; 
		GalParam->es[i]          = 0; 
		GalParam->ld[i]          = 0; 
		GalParam->zwd_map[i]     = 0;
		GalParam->zhd_map[i]     = 0;
		GalParam->ztd_map[i]     = 0;
		GalParam->validpoint[i]  = 1;
	}

	GalParam->DoY = 0;
	GalParam->hour_minute = 0;

	GalParam->a  = NULL;
	GalParam->b  = NULL;
	
	GalParam->a = malloc(sizeof(double)*TropoGal->ps_a_size[2]);
	//GalParam->b = malloc(sizeof(double)*TropoGal->tm_b_size[3]);
	GalParam->b = malloc(sizeof(double)*4);
}

/*****************************************************************************
 * Name        : initFilterSolution
 * Description : Initialise a TFilterSolution structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFilterSolution  *solution      O  N/A  TFilterSolution struct
 *****************************************************************************/
void initFilterSolution (TFilterSolution *solution) {
	int	i;

	solution->prevNumSatellites = 0;
	solution->prevNumSatellitesGPS = 0;

	for (i=0;i<MAX_UNK;i++) {
		solution->parameters[i] = 0;
	}

	for (i=0;i<MAX_VECTOR_UNK;i++) {
		solution->correlations[i] = 0;
	}
}

/*****************************************************************************
 * Name        : initConstellation
 * Description : Initialise a TConstellation structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TConstellation  *constellation  O  N/A  TConstellation struct
 *****************************************************************************/
void initConstellation (TConstellation *constellation) {
	constellation->numSatellites = 0;
	constellation->sat = NULL;
}

/*****************************************************************************
 * Name        : initConstellationElement
 * Description : Initialise a TConstellationElement structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TConstellationElement  *sat     O  N/A  TConstellationElement struct
 *****************************************************************************/
void initConstellationElement (TConstellationElement *sat) {
	int		i;
	
	sat->GNSS = GPS;
	sat->block = UNKNOWN_BLOCK;
	for (i=0;i<MAX_FREQUENCIES_PER_GNSS;i++)
		sat->PCO[i][0] = sat->PCO[i][1] = sat->PCO[i][2] = 0;
	sat->PRN = 0;
	sat->SVN = 0;
	sat->tDecommissioned.MJDN = 62502;  // 1st January 2030
	sat->tDecommissioned.SoD = 0;
	sat->tLaunch.MJDN = 0;
	sat->tLaunch.SoD = 0;

	//This data must be given in the ANTEX file
	sat->zen1 = 0;
	sat->zen2 = 0;
	sat->dzen = 0;
	sat->nzen = 1;		//Initalize to 1 to avoid division to zero, even though this value will be replaced by the data in the file

	//This values can appear or not in the file
	sat->azi1 = 0;
	sat->azi2 = 0;
	sat->dazi = 0;		//Set to 0. If greater than 0, then azimuth data is available
	sat->numazi = 0;		//Set to 0 because dazi is 0.

	sat->noazi = NULL;
	sat->azi = NULL;

}

/*****************************************************************************
 * Name        : initAntennaList
 * Description : Initialise a TAntennaList structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TAntennaList  *antennaList      O  N/A  TAntennaList struct
 *****************************************************************************/
void initAntennaList (TAntennaList *antennaList) {
	antennaList->numAnt = 0;
	antennaList->ant = NULL;
}

/*****************************************************************************
 * Name        : initAntenna
 * Description : Initialise a TAntenna structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TAntenna  *antenna              O  N/A  TAntenna struct
 *****************************************************************************/
void initAntenna (TAntenna *antenna) {
	int		i,j;
	
	for(i=0;i<MAX_GNSS;i++) {
		for (j=0;j<MAX_FREQUENCIES_PER_GNSS;j++) {
			antenna->PCO[i][j][0] = antenna->PCO[i][j][1] = antenna->PCO[i][j][2] = 0.;
		}
	}

	antenna->type[0] = '\0';
	antenna->type_ant[0] = '\0';
	antenna->type_radome[0] = '\0';

	//This data must be given in the ANTEX file
	antenna->azi1 = 0;
	antenna->azi2 = 0;
	antenna->dazi = 0;	
	antenna->numazi = 0;	

	//This values can appear or not in the file
	antenna->zen1 = 0;
	antenna->zen2 = 0;
	antenna->dzen = 0;	//Set to 0. If greater than 0, then azimuth data is available
	antenna->nzen = 0;

	antenna->noazi = NULL;
	antenna->azi = NULL;
}

/*****************************************************************************
 * Name        : initStat
 * Description : Initialise a TStat structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStat  *stat                    O  N/A  TStat struct
 *****************************************************************************/
void initStat (TStat *stat) {
	stat->n = 0;
	stat->mean = 0;
	stat->mean2 = 0;
}

/*****************************************************************************
 * Name        : initTGD
 * Description : Initialise a TGDData structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTGDdata  *tgdData              O  N/A  TTGDdata struct
 * enum P1C1DCBModel  p1c1dcbModel I  N/A  P1C1 DCB source (p1c1UNKNOWN if still unavailable)
 * enum P1P2DCBModel  p1p2dcbModel I  N/A  P1P2 DCB source (p1p2UNKNOWN if still unavailable)
 *****************************************************************************/
void initTGD (TTGDdata *tgdData, enum P1C1DCBModel p1c1dcbModel, enum P1P2DCBModel p1p2dcbModel) {
	tgdData->BRDC = NULL;
	tgdData->DCB = NULL;
	tgdData->p1c1dcbModel = p1c1dcbModel;
	tgdData->p1p2dcbModel = p1p2dcbModel;

//	if (tgdData->p1c1dcbModel==p1c1STRICT ||tgdData->p1p2dcbModel==p1p2DCB) {
		tgdData->DCB = malloc(sizeof(TDCBdata));
		initDCB(tgdData->DCB);
//	}

	initIONEXDCB(&tgdData->ionexDCB);
	initFPPPDCB(&tgdData->fpppDCB);
}

/*****************************************************************************
 * Name        : initDCB
 * Description : Initialise a TDCBdata structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TDCBdata  *dcbData              O  N/A  TDCBdata struct
 *****************************************************************************/
void initDCB (TDCBdata *dcbData) {
	int i,j,k;
	
	dcbData-> n = 0;
	
	for (i=0;i<MAX_DCB_TYPES;i++) {
		dcbData->block[i].startTime.MJDN = 0;
		dcbData->block[i].startTime.SoD = 0;
		dcbData->block[i].endTime.MJDN = 0;
		dcbData->block[i].endTime.SoD = 0;
		dcbData->block[i].measFrom = NA;
		dcbData->block[i].measTo = NA;
		for (j=0;j<MAX_GNSS;j++) {
			for (k=0;k<MAX_SATELLITES_PER_GNSS;k++) {
				dcbData->block[i].dcb[j][k] = 0;
				dcbData->block[i].available[j][k] = 0;
			}
		}
	}
}

/*****************************************************************************
 * Name        : initReceiverList
 * Description : Initialise a TReceiverList structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TReceiverList  *recList         O  N/A  TReceiverList struct
 *****************************************************************************/
void initReceiverList (TReceiverList *recList) {
	recList->n = 0;
	recList->rec = NULL;
}

/*****************************************************************************
 * Name        : initStationList
 * Description : Initialise a TStationList structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStationList  *stationList      O  N/A  TStationList struct
 *****************************************************************************/
void initStationList (TStationList *stationList) {
	stationList->n = 0;
	stationList->station = NULL;
}

/*****************************************************************************
 * Name        : initSBASdatabox
 * Description : Initialise a TSBASdatabox structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSBASdatabox  *SBASdatabox      O  N/A  TSBASdatabox struct
 *****************************************************************************/
void initSBASdatabox (TSBASdatabox *SBASdatabox) {

	SBASdatabox->SBASdata = NULL;
	SBASdatabox->SBASdata = malloc(sizeof(TSBASdata));
	
	SBASdatabox->HWIR = -1.;
	SBASdatabox->VWIR = -1.;
	SBASdatabox->HPLPercentileSamples = -1.;
	SBASdatabox->VPLPercentileSamples = -1.;
	SBASdatabox->MaxHPL = -1.;
	SBASdatabox->MaxVPL = -1.;
	SBASdatabox->decimation = 100.;	//This value is for the first epoch, to ensure there is no discontinuity

	SBASdatabox->TotalPAEpochs = 0;
	SBASdatabox->TotalPAEpochsAvail = 0;
	SBASdatabox->TotalNPAEpochs = 0;
	SBASdatabox->NumEpochsRisk = 0;
	SBASdatabox->NumEpochsRiskMar = 0;
	SBASdatabox->NumMI = 0;
	SBASdatabox->NumHMI = 0;
	SBASdatabox->NumVMI = 0;
	SBASdatabox->HPL = NULL;
	SBASdatabox->VPL = NULL;
	SBASdatabox->LastEpochAvailPrevCurrDisc.SoD = -1.;
	SBASdatabox->FirstEpochAvailAfterPrevDisc.SoD = -1.;

	//decimation, HerrorEpoch. VerrorEpoch, HWIREpoch, VWIREpoch, MaxHPLEpoch and MaxVPLEpoch does not need to be initialized
}


/*****************************************************************************
 * Name        : initSBASblock
 * Description : Initialise a TSBASblock structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSBASblock  *sbasblock          O  N/A  TSBASblock struct
 *****************************************************************************/
void initSBASblock (TSBASblock *sbasblock) {

	sbasblock->frequency = 1;			//Only appears in RINEXB file
	sbasblock->receiverindex = 0 ;			//Only appears in RINEXB file
	sbasblock->datalength = 32 ;			//Only appears in RINEXB file (in EMS message is 32)
	strcpy(sbasblock->sourceidentifier,"SBA");	//Only appears in RINEXB file

	sbasblock->dontuse = 0;			//Will be 1 when a type 0 message is received, and will last one minute at least
	sbasblock->problems = 0;			//Will be 1 when a type 0 message is received and it is filled with 0
	sbasblock->messageslost=0;

	sbasblock->IODF=NULL;				//Will have 1 position for type 2,3,4,5,24 message and 4 position for type 6. For other messages will be empty.
	sbasblock->IODP=-1;				//Type 6 does not have IODP. -1 means value not set.
	sbasblock->numsatellites=-1;			//-1 means value not set.


	sbasblock->PRNactive=NULL;			//Will only contain data when a type 1 message arrives. 
	sbasblock->pos2PRN=NULL;			//Will only contain data when a type 1 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->pos2GNSS=NULL;			//Will only contain data when a type 1 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->PRN2pos=NULL;			//Will only contain data when a type 1 message arrives. 

	sbasblock->PRC=NULL;				//Will only contain data when a type 2,3,4,5,24 message arrives. It will have 13 positions for type 2,3,4,5 and 6 for type 24.
	sbasblock->RRC=NULL;				//Will only contain data when a type 2,3,4,5,24 message arrives. It will have 13 positions for type 2,3,4,5 and 6 for type 24.
	sbasblock->UDREI=NULL;				//Will only contain data when a type 2,3,4,5,24 message arrives. It will have 13 positions for type 2,3,4,5 and 6 for type 24.
	sbasblock->UDRE=NULL;				//Will only contain data when a type 2,3,4,5,24 message arrives. It will have 13 positions for type 2,3,4,5 and 6 for type 24.
	sbasblock->UDREsigma=NULL;			//Will only contain data when a type 2,3,4,5,24 message arrives. It will have 13 positions for type 2,3,4,5 and 6 for type 24.

	sbasblock->UDREIacu=NULL;			//Will only contain data when a type 6 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->UDREacu=NULL;			//Will only contain data when a type 6 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->UDREacusigma=NULL;			//Will only contain data when a type 6 message arrives. It will have 51 positions when memory is allocated.

	sbasblock->tlat=-1;				//Only appears in message type 7. -1 means value not set.
	sbasblock->aiind=NULL;				//Will only contain data when a type 7 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->aifactor=NULL;			//Will only contain data when a type 7 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->timeoutintervalnonprecise=NULL;	//Will only contain data when a type 7 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->timeoutintervalprecise=NULL;	//Will only contain data when a type 7 message arrives. It will have 51 positions when memory is allocated.
	sbasblock->fastcorrupdateinterval=NULL;	//Will only contain data when a type 7 message arrives. It will have 51 positions when memory is allocated.
	

	sbasblock->degradationfactors=NULL;		//Will only contain data when a type 10 message arrives. It will have 16 positions when memory is allocated.

	sbasblock->longtermsaterrcorrections=NULL;	//Will only contain data when a type 25 or 24 message arrives. The size will change depending on the message.
	sbasblock->numlongtermsaterrcorrections=-1;	//Only appears in type 25 or 24 messages. -1 means value not set.

	sbasblock->BlockID=-1;				//Only appears in type 24 messages. -1 means value not set.

	sbasblock->geonavigationmessage=NULL;		//Will only contain data when a type 9 message arrives. It will have 17 positions when memory is allocated.

	sbasblock->geoalmanacsmessage=NULL;		//Will only contain data when a type 17 message arrives. It will have a matrix of 3x12, due there are 3 satellites in the message 
							//and 12 parameters per satellite
	sbasblock->numgeoalmanacs=-1;			//Only appears in type 17 messages. -1 means value not set.

	sbasblock->servicemessage=NULL;		//Will only contain data when a type 27 message arrives. It will have 9 positions when memory is allocated.
	sbasblock->regioncoordinates=NULL;		//Will only contain data when a type 27 message arrives. It will have a matrix of Nx9 positions when memory is allocated. 
							//N will vary between 1-5 dependeing on the the number of regions given in the message
	sbasblock->numregioncoordinates=-1;		//Only appears in type 27 messages. -1 means value not set

	sbasblock->networktimemessage=NULL;		//Will only contain data when a type 12 message arrives. It will have 12 positions when memory is allocated.

	sbasblock->clockephemeriscovariance=NULL;	//Will only contain data when a type 28 message arrives. It will have a matrix of Nx13 positions when memory is allocated, being N=1 or 2 
	sbasblock->numclockephemeriscovariance=-1;	//Only appears in message type 28. -1 means value not set

	sbasblock->igpmaskmessage=NULL;		//Will only contain data when a type 18 message arrives. It will have 204 positions when memory is allocated (201 for the IGP mask and 3 for parameters)

	sbasblock->ionodelayparameters=NULL;		//Will only contain data when a type 26 message arrives. It will have a matrix of 15x7 positions when memory is allocated.
	
}

/*****************************************************************************
 * Name        : initSBASdata
 * Description : Initialise a TSBASdata structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSBASdata  *SBASdata            O  N/A  TSBASdata struct
 * int pos                         I  N/A  Slot of TSBASdata to be initialized
 *****************************************************************************/
void initSBASdata (TSBASdata *SBASdata, int pos) {
	int i,j,k,l;


	//-1 -> Timeout not set  9999999-> Timeout infinite
	//Message type number                0   1  2  3  4  5  6   7  8   9  10 11    12 13 14 15 16      17   18 19 20 21 22 23 24  25  26    27  28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63
	int timeoutnonprecision[64] = {9999999,600,-1,-1,-1,-1,18,360,-1,360,360,-1,86400,-1,-1,-1,-1,9999999,1200,-1,-1,-1,-1,-1,-1,360,600,86400,360,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
	int timeoutprecision[64]    = {9999999,600,-1,-1,-1,-1,12,240,-1,240,240,-1,86400,-1,-1,-1,-1,9999999,1200,-1,-1,-1,-1,-1,-1,240,600,86400,240,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
	int msgupdateinterval[64]   = {     -1,120,-1,-1,-1,-1, 6,120,-1,120,120,-1,  300,-1,-1,-1,-1,    300, 300,-1,-1,-1,-1,-1,-1,120,300,  300,120,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};


	if(pos==0) {
		memcpy(SBASdata[pos].timeoutmessages[0],&timeoutnonprecision,sizeof(timeoutnonprecision));
		memcpy(SBASdata[pos].timeoutmessages[1],&timeoutprecision,sizeof(timeoutprecision));
		memcpy(SBASdata[pos].msgupdateinterval,&msgupdateinterval,sizeof(msgupdateinterval));
	} else {
		//This is done because time out may be different from the default one (received from parameters)
		//This way, the user timeouts are propagated to all GEOs
		memcpy(SBASdata[pos].timeoutmessages[0],SBASdata[0].timeoutmessages[0],sizeof(timeoutnonprecision));
		memcpy(SBASdata[pos].timeoutmessages[1],SBASdata[0].timeoutmessages[1],sizeof(timeoutprecision));
		memcpy(SBASdata[pos].msgupdateinterval,SBASdata[0].msgupdateinterval,sizeof(msgupdateinterval));
	}



	//Set all satellites unselected by default
	for(i=0;i<5;i++) {
		SBASdata[pos].numsatellites[i]=0;
		for(j=0;j<MAX_GNSS;j++) {
			for(k=0;k<MAX_SBAS_PRN;k++) {
				SBASdata[pos].PRNactive[i][j][k]=0;
				SBASdata[pos].PRN2pos[i][j][k]=-1;
				SBASdata[pos].useforbidden[j][k]=0;
			}
		}
	}
	//Set all values to -1 or 9999 (value not set)
	for(i=0;i<64;i++) {
		SBASdata[pos].oldlastmsgreceived[i].MJDN = -1;
		SBASdata[pos].oldlastmsgreceived[i].SoD = -1.;
		SBASdata[pos].lastmsgreceived[i].MJDN = -1;
		SBASdata[pos].lastmsgreceived[i].SoD = -1.;
		SBASdata[pos].oldmaskdiscarded[i]=0;
	}

	for(i=0;i<51;i++) {

		for(j=0;j<5;j++) {
			SBASdata[pos].lastlongtermdata[j][i].MJDN = -1;
			SBASdata[pos].lastlongtermdata[j][i].SoD = -1.;

			SBASdata[pos].prevlastlongtermdata[j][i].MJDN = -1;
			SBASdata[pos].prevlastlongtermdata[j][i].SoD = -1.;

			SBASdata[pos].lastcovmatrix[j][i].MJDN = -1;
			SBASdata[pos].lastcovmatrix[j][i].SoD = -1.;

			SBASdata[pos].fastcorrtimeout[0][j][i] = -1;
			SBASdata[pos].fastcorrtimeout[1][j][i] = -1;
			SBASdata[pos].fastcorrupdateinterval[j][i] = -1;
			SBASdata[pos].aifactor[j][i] = -1;
			SBASdata[pos].tlat[j] =  0;
			SBASdata[pos].lastfastdegfactor[j][i].MJDN = -1;
			SBASdata[pos].lastfastdegfactor[j][i].SoD = -1.;

		}


		SBASdata[pos].firstmessage.MJDN=-1;
		SBASdata[pos].firstmessage.SoD=-1.;
		//SBASdata[pos].oldlastfastcorrections[i].MJDN = -1;
		//SBASdata[pos].oldlastfastcorrections[i].SoD = -1;
		SBASdata[pos].oldlastintegrityinfo[i].MJDN = -1;
		SBASdata[pos].oldlastintegrityinfo[i].SoD = -1;
		//SBASdata[pos].oldlastfastdegfactor[i].MJDN = -1;
		//SBASdata[pos].oldlastfastdegfactor[i].SoD = -1;

		SBASdata[pos].lastintegrityinfo[i].MJDN = -1;
		SBASdata[pos].lastintegrityinfo[i].SoD = -1.;
		//SBASdata[pos].lastfastdegfactor[i].MJDN = -1;
		//SBASdata[pos].lastfastdegfactor[i].SoD = -1;

		SBASdata[pos].lastfastmessage[i] = -1;
		//SBASdata[pos].fastcorrtimeoutnonprecision[i] = -1;
		//SBASdata[pos].fastcorrtimeoutprecision[i] = -1;
		//SBASdata[pos].fastcorrupdateinterval[i] = -1;
		//SBASdata[pos].oldfastcorrtimeoutnonprecision[i] = -1;
		//SBASdata[pos].oldfastcorrtimeoutprecision[i] = -1;
		//SBASdata[pos].oldfastcorrupdateinterval[i] = -1;
		//SBASdata[pos].PRC[i] = 9999;
		//SBASdata[pos].RRC[i] = 9999;
		//SBASdata[pos].UDRE[i] = -1;
		//SBASdata[pos].UDREsigma[i] = -1;
		//SBASdata[pos].oldPRC[i] = -1;
		//SBASdata[pos].oldRRC[i] = -1;
		//SBASdata[pos].oldUDRE[i] = -1;
		//SBASdata[pos].oldUDREsigma[i] = -1;
		//SBASdata[pos].oldaifactor[i] = -1;
		SBASdata[pos].UDREI6[i]=-1;
		SBASdata[pos].UDRE6[i]=-1;
		SBASdata[pos].UDREsigma6[i]=-1;
		SBASdata[pos].oldUDREI6[i]=-1;
		SBASdata[pos].oldUDRE6[i]=-1;
		SBASdata[pos].oldUDREsigma6[i]=-1;

		//SBASdata[pos].IODFfastcorr[i]=-1;
		SBASdata[pos].IODFintegrity[i]=-1;
		//SBASdata[pos].oldIODFfastcorr[i]=-1;
		SBASdata[pos].oldIODFintegrity[i]=-1;

		for(j=0;j<5;j++) {
			SBASdata[pos].pos2PRN[j][i] = -1;
			SBASdata[pos].pos2GNSS[j][i] = -1;
			//SBASdata[pos].oldpos2PRN[j][i] = -1;
			//SBASdata[pos].oldpos2GNSS[j][i] = -1;
		}


		for(j=0;j<5;j++) {
			SBASdata[pos].poslastFC[j][i] = -1;
			SBASdata[pos].numFC[j][i] = 0;
			for(k=0;k<MAXSBASFASTCORR;k++) {
				SBASdata[pos].PRC[j][k][i] = 9999;
				SBASdata[pos].RRC[j][k][i] = 9999;
				SBASdata[pos].UDREI[j][k][i] = -1;
				SBASdata[pos].UDRE[j][k][i] = -1;
				SBASdata[pos].UDREsigma[j][k][i] = -1;
				SBASdata[pos].IODF[j][k][i] = -1;
				SBASdata[pos].lastfastcorrections[j][k][i].MJDN = -1;
				SBASdata[pos].lastfastcorrections[j][k][i].SoD = -1.;
			}
		}

		for(j=0;j<13;j++) {
			for(k=0;k<4;k++) {
				SBASdata[pos].longtermsaterrcorrections[k][i][j] = 9999;
				SBASdata[pos].prevlongtermsaterrcorrections[k][i][j] = 9999;
			}
		}

		for(j=0;j<13;j++) {
			for(k=0;k<5;k++) {
				SBASdata[pos].clockephemeriscovariance[k][i][j] = 9999;
			}
		}
	}
	
	if(pos==0) SBASdata[pos].PRN		=  0;
	else SBASdata[pos].PRN				= -1;

	SBASdata[pos].alarmtime.MJDN		= -1;
	SBASdata[pos].alarmtime.SoD			= -1.;
	SBASdata[pos].alarmTimeRemaining	=  0;	
	SBASdata[pos].numAlarmGEO			=  0;	
	SBASdata[pos].cleared				=  0;
	SBASdata[pos].numreceivers			=  0;
	SBASdata[pos].numSBASsatellites		=  0;
	SBASdata[pos].dontuse				=  0;
	SBASdata[pos].problems				=  0;
	SBASdata[pos].IODPmask				=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPmask			=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldnumsatellites		=  0;
	SBASdata[pos].IODPfastcorr			=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPfastcorr		=  4;   //Set to 4 to avoid invalid vector indexes
	SBASdata[pos].IODPlongterm			=  4;   //Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPlongterm		=  4;   //Set to 4 to avoid invalid vector indexes

	for(i=0;i<MAX_SBAS_PRN;i++) {
		SBASdata[pos].GEOPRN2pos[i]=-1;
		SBASdata[pos].failedmessages[i]=0;
		SBASdata[pos].missed4msg[i]=0;
		SBASdata[pos].alarmGEOPRN[i]=-1;
		SBASdata[pos].alarmGEOindex[i]=-1;
	}

	for(i=0;i<4;i++) {
		for(l=0;l<9;l++) {
			SBASdata[pos].numberofregions[l][i]=0;

			for(j=0;j<40;j++) {
				for(k=0;k<11;k++) {
					SBASdata[pos].regioncoordinates[l][i][j][k]=9999;
				}
			}
		}
	}

	for(i=0;i<8;i++) {
		for(j=0;j<9;j++) {
			SBASdata[pos].servicemessagesreceived[j][i]=-1;
			SBASdata[pos].maxprioritycode[j]=0;
			SBASdata[pos].totalservicemessagesreceived[j]=0;
		}
	}

	SBASdata[pos].IODPdegcorr	=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPdegcorr	=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].IODS		=  8;	//Set to 8 to avoid invalid vector indexes
	SBASdata[pos].oldIODS		=  8;	//Set to 8 to avoid invalid vector indexes

	for(i=0;i<16;i++) {
		SBASdata[pos].degradationfactors[i]=-1;
		SBASdata[pos].olddegradationfactors[i]=-1;
	}

	for(i=0;i<17;i++) {
		SBASdata[pos].geonavigationmessage[i]=9999;
		SBASdata[pos].oldgeonavigationmessage[i]=9999;
	}

	for(i=0;i<12;i++) {
		SBASdata[pos].networktimemessage[i]=-9999;
		for(j=0;j<3;j++) {
			SBASdata[pos].geoalmanacsmessage[j][i]=-1;
			SBASdata[pos].oldgeoalmanacsmessage[j][i]=-1;
		}
	}

	SBASdata[pos].IODPcovariance		= -1;
	SBASdata[pos].oldIODPcovariance		= -1;
	SBASdata[pos].numIGPbandbroadcast	= -1;
	SBASdata[pos].numIGPbandreceived	=  0;
	SBASdata[pos].oldnumIGPbandreceived	=  0;
	SBASdata[pos].IODImask				=  4; //Initialize to 4 because it is not a valid IODI and is a valid position in vector
	SBASdata[pos].oldIODImask			=  4; //Initialize to 4 because it is not a valid IODI and is a valid position in vector
	SBASdata[pos].IODIcorr				= -1;
	SBASdata[pos].oldIODIcorr			= -1;
	SBASdata[pos].numgeoalmanacs		=  0;
	SBASdata[pos].oldnumgeoalmanacs		=  0;


	for(i=0;i<11;i++) {
		SBASdata[pos].IGPbandreceived[i]= -1;
		SBASdata[pos].oldIGPbandreceived[i]= -1;

		for(j=0;j<202;j++) {
			for(k=0;k<5;k++) {
				SBASdata[pos].IGPinMask[k][i][j]=0;
			}
			SBASdata[pos].Ionodelay[i][j]=-1;
			SBASdata[pos].IonoGIVE[i][j]=-1;
			SBASdata[pos].Ionosigma[i][j]=-1;
			SBASdata[pos].lastiono[i][j].MJDN=-1;
			SBASdata[pos].lastiono[i][j].SoD=-1.;
			SBASdata[pos].IGP2Mask[i][j]=4;	//Initialize to 4 because it is not a valid IODI and is a valid position in vector

		}

		for(j=0;j<14;j++) {
			for(k=0;k<15;k++) {
				SBASdata[pos].BlockIDPlace2Grid[i][j][k]=-1;
				SBASdata[pos].oldBlockIDPlace2Grid[i][j][k]=-1;
			}
		}
	}

	SBASdata[pos].sigmamultipathtype=-1;
	SBASdata[pos].numsigmamultipath=0;
	SBASdata[pos].sigmamultipathdata=NULL;

	SBASdata[pos].numSat60=0;
}

/*****************************************************************************
 * Name        : ClearSBASdata
 * Description : Clear a TSBASdata struct
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSBASdata  *SBASdata            O  N/A  TSBASdata struct
 * int pos                         I  N/A  Slot of TSBASdata to be cleared
 *****************************************************************************/
void ClearSBASdata (TSBASdata *SBASdata, int pos) {

	int i,j,k,l;

	/*NOTE: The differences of this function from initSBASdata are:
		- we don't need to copy timeout values again
	- Values for message type 0 (DONTUSE) does not have to be erased
	- Common data (like number of satellites) which are stored in position 0 of TSBASdata vector does not have to be erased
	- Almanac data does not have to be erased
	- Number of satellites with fast and slow sigma of 60 metres does not have to be erased
	*/

	//Set all satellites unselected by default
	for(i=0;i<5;i++) {
		SBASdata[pos].numsatellites[i]=0;
		for(j=0;j<MAX_GNSS;j++) {
			for(k=0;k<MAX_SBAS_PRN;k++) {
				SBASdata[pos].PRNactive[i][j][k]=0;
				SBASdata[pos].PRN2pos[i][j][k]=-1;
				SBASdata[pos].failedmessages[k]=0;
				SBASdata[pos].missed4msg[k]=0;
				if(pos==0) continue;
				SBASdata[pos].GEOPRN2pos[k]=-1;
			}
		}
	}
	//Set all values to -1 or 9999 (value not set)
	for(i=0;i<64;i++) {
		SBASdata[pos].oldlastmsgreceived[i].MJDN = -1;
		SBASdata[pos].oldlastmsgreceived[i].SoD = -1.;
		SBASdata[pos].oldmaskdiscarded[i]=0;

		if(i==DONTUSE || i==GEOSATELLITEALMANACS) continue;

		SBASdata[pos].lastmsgreceived[i].MJDN = -1;
		SBASdata[pos].lastmsgreceived[i].SoD = -1.;
	}

	for(i=0;i<51;i++) {

		for(j=0;j<5;j++) {
			SBASdata[pos].lastlongtermdata[j][i].MJDN = -1;
			SBASdata[pos].lastlongtermdata[j][i].SoD = -1.;

			SBASdata[pos].prevlastlongtermdata[j][i].MJDN = -1;
			SBASdata[pos].prevlastlongtermdata[j][i].SoD = -1.;

			SBASdata[pos].lastcovmatrix[j][i].MJDN = -1;
			SBASdata[pos].lastcovmatrix[j][i].SoD = -1.;

			SBASdata[pos].fastcorrtimeout[0][j][i] = -1;
			SBASdata[pos].fastcorrtimeout[1][j][i] = -1;
			SBASdata[pos].fastcorrupdateinterval[j][i] = -1;
			SBASdata[pos].aifactor[j][i] = -1;
			SBASdata[pos].tlat[j] =  0;
			SBASdata[pos].lastfastdegfactor[j][i].MJDN = -1;
			SBASdata[pos].lastfastdegfactor[j][i].SoD = -1.;
		}


		SBASdata[pos].firstmessage.MJDN=-1;
		SBASdata[pos].firstmessage.SoD=-1.;
		//SBASdata[pos].oldlastfastcorrections[i].MJDN = -1;
		//SBASdata[pos].oldlastfastcorrections[i].SoD = -1.;
		SBASdata[pos].oldlastintegrityinfo[i].MJDN = -1;
		SBASdata[pos].oldlastintegrityinfo[i].SoD = -1.;
		//SBASdata[pos].oldlastfastdegfactor[i].MJDN = -1;
		//SBASdata[pos].oldlastfastdegfactor[i].SoD = -1.;

		SBASdata[pos].lastintegrityinfo[i].MJDN = -1;
		SBASdata[pos].lastintegrityinfo[i].SoD = -1.;
		//SBASdata[pos].lastfastdegfactor[i].MJDN = -1;
		//SBASdata[pos].lastfastdegfactor[i].SoD = -1.;

		SBASdata[pos].lastfastmessage[i]= -1;
		//SBASdata[pos].fastcorrtimeoutnonprecision[i] = -1;
		//SBASdata[pos].fastcorrtimeoutprecision[i] = -1;
		//SBASdata[pos].fastcorrupdateinterval[i] = -1;
		//SBASdata[pos].oldfastcorrtimeoutnonprecision[i] = -1;
		//SBASdata[pos].oldfastcorrtimeoutprecision[i] = -1;
		//SBASdata[pos].oldfastcorrupdateinterval[i] = -1;
		//SBASdata[pos].PRC[i] = 9999;
		//SBASdata[pos].RRC[i] = 9999;
		//SBASdata[pos].UDRE[i] = -1;
		//SBASdata[pos].UDREsigma[i] = -1;
		//SBASdata[pos].oldPRC[i] = -1;
		//SBASdata[pos].oldRRC[i] = -1;
		//SBASdata[pos].oldUDRE[i] = -1;
		//SBASdata[pos].oldUDREsigma[i] = -1;
		//SBASdata[pos].oldaifactor[i] = -1;
		SBASdata[pos].UDREI6[i]=-1;
		SBASdata[pos].UDRE6[i]=-1;
		SBASdata[pos].UDREsigma6[i]=-1;
		SBASdata[pos].oldUDREI6[i]=-1;
		SBASdata[pos].oldUDRE6[i]=-1;
		SBASdata[pos].oldUDREsigma6[i]=-1;

		//SBASdata[pos].IODFfastcorr[i]=-1;
		SBASdata[pos].IODFintegrity[i]=-1;
		//SBASdata[pos].oldIODFfastcorr[i]=-1;
		SBASdata[pos].oldIODFintegrity[i]=-1;

		for(j=0;j<5;j++) {
			SBASdata[pos].pos2PRN[j][i] = -1;
			SBASdata[pos].pos2GNSS[j][i] = -1;
			//SBASdata[pos].oldpos2PRN[j][i] = -1;
			//SBASdata[pos].oldpos2GNSS[j][i] = -1;
		}


		for(j=0;j<5;j++) {
			SBASdata[pos].poslastFC[j][i] = -1;
			SBASdata[pos].numFC[j][i] = 0;
			for(k=0;k<MAXSBASFASTCORR;k++) {
				SBASdata[pos].PRC[j][k][i] = 9999;
				SBASdata[pos].RRC[j][k][i] = 9999;
				SBASdata[pos].UDREI[j][k][i] = -1;
				SBASdata[pos].UDRE[j][k][i] = -1;
				SBASdata[pos].UDREsigma[j][k][i] = -1;
				SBASdata[pos].IODF[j][k][i] = -1;
				SBASdata[pos].lastfastcorrections[j][k][i].MJDN = -1;
				SBASdata[pos].lastfastcorrections[j][k][i].SoD = -1.;
			}
		}

		for(j=0;j<13;j++) {
			for(k=0;k<4;k++) {
				SBASdata[pos].longtermsaterrcorrections[k][i][j] = 9999;
				SBASdata[pos].prevlongtermsaterrcorrections[k][i][j] = 9999;
			}
		}

		for(j=0;j<13;j++) {
			for(k=0;k<5;k++) {
				SBASdata[pos].clockephemeriscovariance[k][i][j] = 9999;
			}
		}
	}
	
	SBASdata[pos].IODPmask				=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPmask			=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldnumsatellites		=  0;
	SBASdata[pos].IODPfastcorr			=  4;   //Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPfastcorr		=  4;   //Set to 4 to avoid invalid vector indexes
	SBASdata[pos].IODPlongterm			=  4;   //Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPlongterm		=  4;   //Set to 4 to avoid invalid vector indexes

	for(i=0;i<4;i++) {
		for(l=0;l<9;l++) {
			SBASdata[pos].numberofregions[l][i]=0;

			for(j=0;j<40;j++) {
				for(k=0;k<11;k++) {
					SBASdata[pos].regioncoordinates[l][i][j][k]=9999;
				}
			}
		}
	}

	for(i=0;i<8;i++) {
		for(j=0;j<9;j++) {
			SBASdata[pos].servicemessagesreceived[j][i]=-1;
			SBASdata[pos].maxprioritycode[j]=0;
			SBASdata[pos].totalservicemessagesreceived[j]=0;
		}
	}

	SBASdata[pos].IODPdegcorr	=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].oldIODPdegcorr	=  4;	//Set to 4 to avoid invalid vector indexes
	SBASdata[pos].IODS		=  8;	//Set to 8 to avoid invalid vector indexes
	SBASdata[pos].oldIODS		=  8;	//Set to 8 to avoid invalid vector indexes

	for(i=0;i<16;i++) {
		SBASdata[pos].degradationfactors[i]=-1;
		SBASdata[pos].olddegradationfactors[i]=-1;
	}

	for(i=0;i<17;i++) {
		SBASdata[pos].geonavigationmessage[i]=9999;
		SBASdata[pos].oldgeonavigationmessage[i]=9999;
	}

	for(i=0;i<12;i++) {
		SBASdata[pos].networktimemessage[i]=-9999;
	}

	SBASdata[pos].cleared				=  1;

	SBASdata[pos].IODPcovariance		= -1;
	SBASdata[pos].oldIODPcovariance		= -1;
	SBASdata[pos].numIGPbandbroadcast	= -1;
	SBASdata[pos].numIGPbandreceived	=  0;
	SBASdata[pos].oldnumIGPbandreceived	=  0;
	SBASdata[pos].IODImask				=  4; //Initialize to 4 because it is not a valid IODI and is a valid position in vector
	SBASdata[pos].oldIODImask			=  4; //Initialize to 4 because it is not a valid IODI and is a valid position in vector
	SBASdata[pos].IODIcorr				= -1;
	SBASdata[pos].oldIODIcorr			= -1;


	for(i=0;i<11;i++) {
		SBASdata[pos].IGPbandreceived[i]= -1;
		SBASdata[pos].oldIGPbandreceived[i]= -1;

		for(j=0;j<202;j++) {
			for(k=0;k<5;k++) {
				SBASdata[pos].IGPinMask[k][i][j]=0;
			}
			SBASdata[pos].Ionodelay[i][j]=-1;
			SBASdata[pos].IonoGIVE[i][j]=-1;
			SBASdata[pos].Ionosigma[i][j]=-1;
			SBASdata[pos].lastiono[i][j].MJDN=-1;
			SBASdata[pos].lastiono[i][j].SoD=-1.;
			SBASdata[pos].IGP2Mask[i][j]=4;	//Initialize to 4 because it is not a valid IODI and is a valid position in vector

		}

		for(j=0;j<14;j++) {
			for(k=0;k<15;k++) {
				SBASdata[pos].BlockIDPlace2Grid[i][j][k]=-1;
				SBASdata[pos].oldBlockIDPlace2Grid[i][j][k]=-1;
			}
		}
	}



}

/*****************************************************************************
 * Name        : initSBAScorrections
 * Description : Initialise a TSBAScorr structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSBAScorr  *SBAScorr            O  N/A  TSBAScorr struct
 *****************************************************************************/
void initSBAScorrections (TSBAScorr *SBAScorr) {
	SBAScorr->SatCoord[0]=SBAScorr->SatCoordbrdc[0]=SBAScorr->SatCoordCor[0]=0;
	SBAScorr->SatCoord[1]=SBAScorr->SatCoordbrdc[1]=SBAScorr->SatCoordCor[1]=0;
	SBAScorr->SatCoord[2]=SBAScorr->SatCoordbrdc[2]=SBAScorr->SatCoordCor[2]=0;

	SBAScorr->Clock=SBAScorr->Clockbrdc=SBAScorr->ClockCor=0;

	SBAScorr->uraindex=-1;

	SBAScorr->IOD=-1;
	SBAScorr->IODnoSBAS=-1;
	SBAScorr->ionocorrection=0;
	
	SBAScorr->PRC=0;
	SBAScorr->RRC=0;
	SBAScorr->RRCtime=0;
	SBAScorr->UDREI=-1;
	SBAScorr->UDREsigma=0;
	SBAScorr->fastcorrdegfactor=-1;
	SBAScorr->fastcorrtmout=-1;
	SBAScorr->fastcorrsmallertmout=-1;
	SBAScorr->tudre=-1;
	SBAScorr->tu=-1;
	SBAScorr->tPRC=-1;

	SBAScorr->longtermdegradation=0;
	SBAScorr->geonavdegradation=0;
	SBAScorr->enroutedegradation=0;
	SBAScorr->fastcorrdeg=0;
	SBAScorr->RRCdeg=0;
	SBAScorr->deltaudre=1;
	SBAScorr->deltaudresource=0;

	SBAScorr->mt10avail=-1;

	SBAScorr->sigma2flt=0;
	SBAScorr->sigma2iono=0;
	SBAScorr->sigma2tropo=0;
	SBAScorr->sigma2noise=0;
	SBAScorr->sigma2divergence=0;
	SBAScorr->sigma2multipath=0;
	SBAScorr->sigma2air=0;
	SBAScorr->SBASsatsigma2=0;

	SBAScorr->fastcorIODP=-1;
	SBAScorr->fastcorIODF=-1;
	SBAScorr->fastcorPRNpos=0;
	SBAScorr->longtermIODP=-1;
	SBAScorr->longtermPRNpos=0;
	SBAScorr->serviceIODS=-1;
	SBAScorr->covarianceIODP=-1;
	SBAScorr->covariancePRNpos=-2;
	SBAScorr->ionomodelflag=0;

	SBAScorr->SBASplotIonoAvail=0;
}

/*****************************************************************************
 * Name        : initStfdESA
 * Description : Initialise a TStdESA structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStdESA  *StdESA                O  N/A  TStdESA struct
 *****************************************************************************/
void initStfdESA (TStdESA  *StdESA) {

	unsigned int i,j;

	// Compute number of pixels and allocate space
	StdESA->numpixelhor = StdESA->xmax/StdESA->xstep;
	StdESA->numpixelver = StdESA->ymax/StdESA->ystep;
	StdESA->counthor = NULL;
	StdESA->countver = NULL;
	StdESA->counthor = malloc(sizeof(int *)*StdESA->numpixelver);
	StdESA->countver = malloc(sizeof(int *)*StdESA->numpixelver);
	StdESA->numsamplesMI=0;
	StdESA->numsamplesHMI=0;
	StdESA->numsamplesVMI=0;
	StdESA->numsamplesProcessed=0;
	StdESA->numsamplesComputed=0;
	StdESA->numsamplesSingular=0;
	StdESA->HWIR=0.;
	StdESA->VWIR=0.;
	for (i=0;i<StdESA->numpixelver;i++) {
		StdESA->counthor[i] = malloc(sizeof(int)*StdESA->numpixelhor);
		StdESA->countver[i] = malloc(sizeof(int)*StdESA->numpixelhor);
		for (j=0;j<StdESA->numpixelhor;j++) {
			StdESA->counthor[i][j] = 0;
			StdESA->countver[i][j] = 0;
		}
	}

}

/*****************************************************************************
 * Name        : initSBASPlotsMode
 * Description : Initialise a TSBASPlots and Tepoch structure for SBAS plots mode
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch *epoch                   O  N/A  TEpoch struct
 * TSBASPlots *SBASplots           O  N/A  TSBASPlots struct
 * TOptions  *options              I  N/A  TOptions struct
 * int *retvalue                   O  N/A  Return value. 1->OK -1->Error allocating memory
 *****************************************************************************/
void initSBASPlotsMode (TEpoch *epoch, TSBASPlots *SBASplots, TOptions  *options, int *retvalue)  {
	int 	i,j;
	
	//Set the number of satellites in view to 32 (GPS PRN 1-32)
	epoch->numSatellites=32;
	//Set last index for SBAS satellite
	if(options->NoAvailabilityPlot==0) {
		//We are computing Availability plot, so we will loop through all satellites
		epoch->lastSBASindex=31;
	} else {
		//We are only computing ionosphere availability plot, so we will use only satellite in position 0 of epoch->sat
		epoch->lastSBASindex=0;
	}

	//Set GPS measurements available only to C1C
	epoch->measOrder[GPS].nDiffMeasurements=1;
	epoch->measOrder[GPS].GNSS=GPS;
	epoch->measOrder[GPS].meas2Ind[C1C]=0;
	epoch->measOrder[GPS].conversionFactor[C1C]=1.0;
	epoch->measOrder[GPS].ind2Meas[0]=C1C;
	//Set the PRN in each position (we will set them in order, but it is not relevant)
	for (i=0;i<epoch->numSatellites;i++) {
		epoch->sat[i].GNSS=GPS;
		epoch->sat[i].PRN=i+1;
		epoch->sat[i].meas[0].value=-1;
	}

	//To minimize memory usage, we will make a table with the minimum size possible.
	//Depending on the resolution (1º, 0.1º or 0.01º we will need more or less memory)
	if(options->AvailabilityPlotStep<0.1) {
		//Resolution is below than 0.1º
		SBASplots->AvailabilityMemFactor=100;
		//Resolution is below than 1º and equal or higher than 0.1º
	} else if (options->AvailabilityPlotStep<1.) {
		SBASplots->AvailabilityMemFactor=10;
	} else {
		//Resolution is equal or higher than 1º
		SBASplots->AvailabilityMemFactor=1;
	}

	if(options->IonoPlotStep<0.1) {
		//Resolution is below than 0.1º
		SBASplots->IonoMemFactor=100;
		//Resolution is below than 1º and equal or higher than 0.1º
	} else if (options->IonoPlotStep<1.) {
		SBASplots->IonoMemFactor=10;
	} else {
		//Resolution is equal or higher than 1º
		SBASplots->IonoMemFactor=1;
	}		//Compute offset for vector position
	
	if(options->minLatplots<=0.) {
		SBASplots->offsetLatAvail=(int)(fabs(options->minLatplots)*(double)SBASplots->AvailabilityMemFactor);
		SBASplots->offsetLatIono=(int)(fabs(options->minLatplots)*(double)SBASplots->IonoMemFactor);
	} else {
		SBASplots->offsetLatAvail=(int)(-options->minLatplots*(double)SBASplots->AvailabilityMemFactor);
		SBASplots->offsetLatIono=(int)(-options->minLatplots*(double)SBASplots->IonoMemFactor);
	}
	if(options->minLonplots<=0.) {
		SBASplots->offsetLonAvail=(int)(fabs(options->minLonplots)*(double)SBASplots->AvailabilityMemFactor);
		SBASplots->offsetLonIono=(int)(fabs(options->minLonplots)*(double)SBASplots->IonoMemFactor);
	} else {
		SBASplots->offsetLonAvail=(int)(-options->minLonplots*(double)SBASplots->AvailabilityMemFactor);
		SBASplots->offsetLonIono=(int)(-options->minLonplots*(double)SBASplots->IonoMemFactor);
	}

	//Set Availability and Risk pointer to NULL
	SBASplots->SBASNumEpochsAvail=NULL;
	SBASplots->SBASNumEpochsRisk=NULL;
	SBASplots->SBASNumEpochsRiskMar=NULL;
	SBASplots->SBASNumEpochsDOP=NULL;
	SBASplots->LastEpochAvailPrevCurrDisc=NULL;
	SBASplots->FirstEpochAvailAfterPrevDisc=NULL;
	SBASplots->SBASNumEpochsAvailHourly=NULL;
	SBASplots->SBASNumEpochsRiskHourly=NULL;
	SBASplots->SBASNumEpochsRiskHourlyMar=NULL;
	SBASplots->SBASNumEpochsDOPHourly=NULL;
	SBASplots->LastEpochAvailPrevCurrDiscHourly=NULL;
	SBASplots->FirstEpochAvailAfterPrevDiscHourly=NULL;
	SBASplots->HDOPMean=NULL;
	SBASplots->PDOPMean=NULL;
	SBASplots->GDOPMean=NULL;
	SBASplots->HDOPMeanHourly=NULL;
	SBASplots->PDOPMeanHourly=NULL;
	SBASplots->GDOPMeanHourly=NULL;
	SBASplots->HDOPValues=NULL;
	SBASplots->PDOPValues=NULL;
	SBASplots->GDOPValues=NULL;
	SBASplots->HDOPValuesHourly=NULL;
	SBASplots->PDOPValuesHourly=NULL;
	SBASplots->GDOPValuesHourly=NULL;

	//Allocate memory for availability plot counter
	if(options->NoAvailabilityPlot==0) {
		SBASplots->AvailabilityLatSize=1+(int)((options->maxLatplots-options->minLatplots)*(double)(SBASplots->AvailabilityMemFactor));
		SBASplots->AvailabilityLonSize=1+(int)((options->maxLonplots-options->minLonplots)*(double)(SBASplots->AvailabilityMemFactor));
		SBASplots->SBASNumEpochsAvail=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
		if(SBASplots->SBASNumEpochsAvail==NULL) {
			*retvalue=-1;
			return;
		}
		SBASplots->SBASNumEpochsDOP=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
		if(SBASplots->SBASNumEpochsDOP==NULL) {
			*retvalue=-1;
			return;
		}
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			SBASplots->SBASNumEpochsAvail[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
			if (SBASplots->SBASNumEpochsAvail[i]==NULL) {
				*retvalue=-1;
				return;
			}
			SBASplots->SBASNumEpochsDOP[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
			if (SBASplots->SBASNumEpochsDOP[i]==NULL) {
				*retvalue=-1;
				return;
			}
		}

		//Allocate memory for continuity risk data
		if(options->ComputeRiskPlots==1) {
			SBASplots->SBASNumEpochsRisk=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->SBASNumEpochsRisk==NULL) {
				*retvalue=-1;
				return;
			}
			SBASplots->SBASNumEpochsRiskMar=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->SBASNumEpochsRiskMar==NULL) {
				*retvalue=-1;
				return;
			}
			SBASplots->LastEpochAvailPrevCurrDisc=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->LastEpochAvailPrevCurrDisc==NULL) {
				*retvalue=-1;
				return;
			}
			SBASplots->FirstEpochAvailAfterPrevDisc=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->FirstEpochAvailAfterPrevDisc==NULL) {
				*retvalue=-1;
				return;
			}
			for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
				SBASplots->SBASNumEpochsRisk[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
				if (SBASplots->SBASNumEpochsRisk[i]==NULL) {
					*retvalue=-1;
					return;
				}
				SBASplots->SBASNumEpochsRiskMar[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
				if (SBASplots->SBASNumEpochsRiskMar[i]==NULL) {
					*retvalue=-1;
					return;
				}
				SBASplots->LastEpochAvailPrevCurrDisc[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(double)); 
				if (SBASplots->LastEpochAvailPrevCurrDisc[i]==NULL) {
					*retvalue=-1;
					return;
				}
				SBASplots->FirstEpochAvailAfterPrevDisc[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(double)); 
				if (SBASplots->FirstEpochAvailAfterPrevDisc[i]==NULL) {
					*retvalue=-1;
					return;
				}
			}
		}

		//Repeat memory allocation for hourly maps
		if (options->SBASHourlyMaps==1)	{
			SBASplots->SBASNumEpochsAvailHourly=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->SBASNumEpochsAvailHourly==NULL) {
				*retvalue=-1;
				return;
			}
			SBASplots->SBASNumEpochsDOPHourly=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->SBASNumEpochsDOPHourly==NULL) {
				*retvalue=-1;
				return;
			}
			for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
				SBASplots->SBASNumEpochsAvailHourly[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
				if (SBASplots->SBASNumEpochsAvailHourly[i]==NULL) {
					*retvalue=-1;
					return;
				}
				SBASplots->SBASNumEpochsDOPHourly[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
				if (SBASplots->SBASNumEpochsDOPHourly[i]==NULL) {
					*retvalue=-1;
					return;
				}
			}

			//Allocate memory for continuity risk data
			if(options->ComputeRiskPlots==1) {
				SBASplots->SBASNumEpochsRiskHourly=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
				if(SBASplots->SBASNumEpochsRiskHourly==NULL) {
					*retvalue=-1;
					return;
				}
				SBASplots->SBASNumEpochsRiskHourlyMar=malloc(sizeof(int*)*SBASplots->AvailabilityLatSize);
				if(SBASplots->SBASNumEpochsRiskHourlyMar==NULL) {
					*retvalue=-1;
					return;
				}
				SBASplots->LastEpochAvailPrevCurrDiscHourly=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
				if(SBASplots->LastEpochAvailPrevCurrDiscHourly==NULL) {
					*retvalue=-1;
					return;
				}
				SBASplots->FirstEpochAvailAfterPrevDiscHourly=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
				if(SBASplots->FirstEpochAvailAfterPrevDiscHourly==NULL) {
					*retvalue=-1;
					return;
				}
				for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
					SBASplots->SBASNumEpochsRiskHourly[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
					if (SBASplots->SBASNumEpochsRiskHourly[i]==NULL) {
						*retvalue=-1;
						return;
					}
					SBASplots->SBASNumEpochsRiskHourlyMar[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(int)); //Allocate memory and set it to 0
					if (SBASplots->SBASNumEpochsRiskHourlyMar[i]==NULL) {
						*retvalue=-1;
						return;
					}
					SBASplots->LastEpochAvailPrevCurrDiscHourly[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(double)); 
					if (SBASplots->LastEpochAvailPrevCurrDiscHourly[i]==NULL) {
						*retvalue=-1;
						return;
					}
					SBASplots->FirstEpochAvailAfterPrevDiscHourly[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(double)); 
					if (SBASplots->FirstEpochAvailAfterPrevDiscHourly[i]==NULL) {
						*retvalue=-1;
						return;
					}
				}
			}
		}
	}

	//Allocate memory for Iono availability counter
	SBASplots->IonoNumEpochsAvail=NULL;
	SBASplots->IonoNumEpochsAvailHourly=NULL;
	if(options->NoIonoPlot==0) {
		//SBASplots->IonoLatSize=1+(int)((options->maxLatplots-options->minLatplots)*(double)(SBASplots->IonoMemFactor)/options->IonoPlotStep);
		//SBASplots->IonoLonSize=1+(int)((options->maxLonplots-options->minLonplots)*(double)(SBASplots->IonoMemFactor)/options->IonoPlotStep);
		SBASplots->IonoLatSize=1+(int)((options->maxLatplots-options->minLatplots)*(double)(SBASplots->IonoMemFactor));
		SBASplots->IonoLonSize=1+(int)((options->maxLonplots-options->minLonplots)*(double)(SBASplots->IonoMemFactor));
		SBASplots->IonoNumEpochsAvail=malloc(sizeof(int*)*SBASplots->IonoLatSize);
		if(SBASplots->IonoNumEpochsAvail==NULL) {
			*retvalue=-1;
			return;
		}
		for(i=0;i<SBASplots->IonoLatSize;i++) {
			SBASplots->IonoNumEpochsAvail[i]=calloc(SBASplots->IonoLonSize,sizeof(int)); //Allocate memory and set it to 0
			if (SBASplots->IonoNumEpochsAvail[i]==NULL) {
				*retvalue=-1;
				return;
			}
		}

		//Repeat memory allocation for hourly maps
		if (options->SBASHourlyMaps==1)	{
			SBASplots->IonoNumEpochsAvailHourly=malloc(sizeof(int*)*SBASplots->IonoLatSize);
			if(SBASplots->IonoNumEpochsAvailHourly==NULL) {
				*retvalue=-1;
				return;
			}
			for(i=0;i<SBASplots->IonoLatSize;i++) {
				SBASplots->IonoNumEpochsAvailHourly[i]=calloc(SBASplots->IonoLonSize,sizeof(int)); //Allocate memory and set it to 0
				if (SBASplots->IonoNumEpochsAvailHourly[i]==NULL) {
					*retvalue=-1;
					return;
				}
			}
		}
	}

	if(options->HDOPPlot==1||options->CombinedDOPPlot==1) {
		SBASplots->HDOPMean=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
		if(SBASplots->HDOPMean==NULL) {
			*retvalue=-1;
			return;
		}
		if(options->DOPPercentile) {
			SBASplots->HDOPValues=malloc(sizeof(float*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->HDOPValues==NULL) {
				*retvalue=-1;
				return;
			}
		}
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			SBASplots->HDOPMean[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(double)); //Allocate memory and set it to 0
			if (SBASplots->HDOPMean[i]==NULL) {
				*retvalue=-1;
				return;
			}
			if(options->DOPPercentile) {
				SBASplots->HDOPValues[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(float*)); 
				if (SBASplots->HDOPValues[i]==NULL) {
					*retvalue=-1;
					return;
				}
				for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
					SBASplots->HDOPValues[i][j]=calloc(86400,sizeof(float)); //Allocate memory and set it to 0
					if (SBASplots->HDOPValues[i][j]==NULL) {
						*retvalue=-1;
						return;
					}
				}
			}
		}

		//Repeat memory allocation for hourly maps
		if (options->SBASHourlyMaps==1)	{
			SBASplots->HDOPMeanHourly=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->HDOPMeanHourly==NULL) {
				*retvalue=-1;
				return;
			}
			if(options->DOPPercentile) {
				SBASplots->HDOPValuesHourly=malloc(sizeof(float*)*SBASplots->AvailabilityLatSize);
				if(SBASplots->HDOPValuesHourly==NULL) {
					*retvalue=-1;
					return;
				}
			}
			for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
				SBASplots->HDOPMeanHourly[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(double)); //Allocate memory and set it to 0
				if (SBASplots->HDOPMeanHourly[i]==NULL) {
					*retvalue=-1;
					return;
				}
				if(options->DOPPercentile) {
					SBASplots->HDOPValuesHourly[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(float*)); 
					if (SBASplots->HDOPValuesHourly[i]==NULL) {
						*retvalue=-1;
						return;
					}
					for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
						SBASplots->HDOPValuesHourly[i][j]=calloc(3600,sizeof(float)); //Allocate memory and set it to 0
						if (SBASplots->HDOPValuesHourly[i][j]==NULL) {
							*retvalue=-1;
							return;
						}
					}
				}
			}
		}
	}

	if(options->PDOPPlot==1||options->CombinedDOPPlot==1) {
		SBASplots->PDOPMean=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
		if(SBASplots->PDOPMean==NULL) {
			*retvalue=-1;
			return;
		}
		if(options->DOPPercentile) {
			SBASplots->PDOPValues=malloc(sizeof(float*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->PDOPValues==NULL) {
				*retvalue=-1;
				return;
			}
		}
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			SBASplots->PDOPMean[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(double)); //Allocate memory and set it to 0
			if (SBASplots->PDOPMean[i]==NULL) {
				*retvalue=-1;
				return;
			}	

			if(options->DOPPercentile) {
				SBASplots->PDOPValues[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(float*)); 
				if (SBASplots->PDOPValues[i]==NULL) {
					*retvalue=-1;
					return;
				}
				for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
					SBASplots->PDOPValues[i][j]=calloc(86400,sizeof(float)); //Allocate memory and set it to 0
					if (SBASplots->PDOPValues[i][j]==NULL) {
						*retvalue=-1;
						return;
					}
				}
			}
		}

		//Repeat memory allocation for hourly maps
		if (options->SBASHourlyMaps==1)	{
			SBASplots->PDOPMeanHourly=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->PDOPMeanHourly==NULL) {
				*retvalue=-1;
				return;
			}
			if(options->DOPPercentile) {
				SBASplots->PDOPValuesHourly=malloc(sizeof(float*)*SBASplots->AvailabilityLatSize);
				if(SBASplots->PDOPValuesHourly==NULL) {
					*retvalue=-1;
					return;
				}
			}
			for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
				SBASplots->PDOPMeanHourly[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(double)); //Allocate memory and set it to 0
				if (SBASplots->PDOPMeanHourly[i]==NULL) {
					*retvalue=-1;
					return;
				}
				if(options->DOPPercentile) {
					SBASplots->PDOPValuesHourly[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(float*)); 
					if (SBASplots->PDOPValuesHourly[i]==NULL) {
						*retvalue=-1;
						return;
					}
					for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
						SBASplots->PDOPValuesHourly[i][j]=calloc(3600,sizeof(float)); //Allocate memory and set it to 0
						if (SBASplots->PDOPValuesHourly[i][j]==NULL) {
							*retvalue=-1;
							return;
						}
					}
				}
			}
		}
	}

	if(options->GDOPPlot==1||options->CombinedDOPPlot==1) {
		SBASplots->GDOPMean=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
		if(SBASplots->GDOPMean==NULL) {
			*retvalue=-1;
			return;
		}
		if(options->DOPPercentile) {
			SBASplots->GDOPValues=malloc(sizeof(float*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->GDOPValues==NULL) {
				*retvalue=-1;
				return;
			}
		}
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			SBASplots->GDOPMean[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(double)); //Allocate memory and set it to 0
			if (SBASplots->GDOPMean[i]==NULL) {
				*retvalue=-1;
				return;
			}
			if(options->DOPPercentile) {
				SBASplots->GDOPValues[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(float*)); 
				if (SBASplots->GDOPValues[i]==NULL) {
					*retvalue=-1;
					return;
				}
				for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
					SBASplots->GDOPValues[i][j]=calloc(86400,sizeof(float)); //Allocate memory and set it to 0
					if (SBASplots->GDOPValues[i][j]==NULL) {
						*retvalue=-1;
						return;
					}
				}
			}
		}

		//Repeat memory allocation for hourly maps
		if (options->SBASHourlyMaps==1)	{
			SBASplots->GDOPMeanHourly=malloc(sizeof(double*)*SBASplots->AvailabilityLatSize);
			if(SBASplots->GDOPMeanHourly==NULL) {
				*retvalue=-1;
				return;
			}
			if(options->DOPPercentile) {
				SBASplots->GDOPValuesHourly=malloc(sizeof(float*)*SBASplots->AvailabilityLatSize);
				if(SBASplots->GDOPValuesHourly==NULL) {
					*retvalue=-1;
					return;
				}
			}
			for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
				SBASplots->GDOPMeanHourly[i]=calloc(SBASplots->AvailabilityLonSize,sizeof(double)); //Allocate memory and set it to 0
				if (SBASplots->GDOPMeanHourly[i]==NULL) {
					*retvalue=-1;
					return;
				}
				if(options->DOPPercentile) {
					SBASplots->GDOPValuesHourly[i]=malloc(SBASplots->AvailabilityLonSize*sizeof(float*)); 
					if (SBASplots->GDOPValuesHourly[i]==NULL) {
						*retvalue=-1;
						return;
					}
					for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
						SBASplots->GDOPValuesHourly[i][j]=calloc(3600,sizeof(float)); //Allocate memory and set it to 0
						if (SBASplots->GDOPValuesHourly[i][j]==NULL) {
							*retvalue=-1;
							return;
						}
					}
				}
			}
		}
	}
	
	*retvalue=1;
}

/*****************************************************************************
 * Name        : freeSBASPlotsData
 * Description : Free memory in TSBASPlots structure for SBAS plots mode
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TSBASPlots *SBASplots           O  N/A  TSBASPlots struct
 * TOptions  *options              I  N/A  TOptions struct
 *****************************************************************************/
void freeSBASPlotsData(TSBASPlots *SBASplots,TOptions  *options) {

	int i,j;

	if(options->NoAvailabilityPlot==0) {
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			free(SBASplots->SBASNumEpochsAvail[i]);
			if (options->SBASHourlyMaps==1) {
				free(SBASplots->SBASNumEpochsAvailHourly[i]);
			}
		}
		free(SBASplots->SBASNumEpochsAvail);
		if (options->SBASHourlyMaps==1) {
			free(SBASplots->SBASNumEpochsAvailHourly);
		}
		if(options->ComputeRiskPlots==1) {
			for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
				free(SBASplots->SBASNumEpochsRisk[i]);
				free(SBASplots->LastEpochAvailPrevCurrDisc[i]);
				free(SBASplots->FirstEpochAvailAfterPrevDisc[i]);
				if (options->SBASHourlyMaps==1) {
					free(SBASplots->SBASNumEpochsRiskHourly[i]);
					free(SBASplots->LastEpochAvailPrevCurrDiscHourly[i]);
					free(SBASplots->FirstEpochAvailAfterPrevDiscHourly[i]);
				}
			}
			free(SBASplots->SBASNumEpochsRisk);
			free(SBASplots->LastEpochAvailPrevCurrDisc);
			free(SBASplots->FirstEpochAvailAfterPrevDisc);
			if (options->SBASHourlyMaps==1) {
				free(SBASplots->SBASNumEpochsRiskHourly);
				free(SBASplots->LastEpochAvailPrevCurrDiscHourly);
				free(SBASplots->FirstEpochAvailAfterPrevDiscHourly);
			}
		}
	}
	if(options->NoIonoPlot==0) {
		for(i=0;i<SBASplots->IonoLatSize;i++) {
			free(SBASplots->IonoNumEpochsAvail[i]);
			if (options->SBASHourlyMaps==1) {
				free(SBASplots->IonoNumEpochsAvailHourly[i]);
			}
		}
		free(SBASplots->IonoNumEpochsAvail);
		if (options->SBASHourlyMaps==1) {
			free(SBASplots->IonoNumEpochsAvailHourly);
		}
	}
	if(options->HDOPPlot==1||options->CombinedDOPPlot==1) {
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			if(options->DOPPercentile) {
				for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
					free(SBASplots->HDOPValues[i][j]);
					if (options->SBASHourlyMaps==1) {
						free(SBASplots->HDOPValuesHourly[i][j]);
					}
				}
			}
			free(SBASplots->HDOPMean[i]);
			if(options->DOPPercentile) {
				free(SBASplots->HDOPValues[i]);
			}
			if (options->SBASHourlyMaps==1) {
				free(SBASplots->HDOPMeanHourly[i]);
				if(options->DOPPercentile) {
					free(SBASplots->HDOPValuesHourly[i]);
				}
			}
		}
		free(SBASplots->HDOPMean);
		if(options->DOPPercentile) {
			free(SBASplots->HDOPValues);
		}
		if (options->SBASHourlyMaps==1) {
			free(SBASplots->HDOPMeanHourly);
			if(options->DOPPercentile) {
				free(SBASplots->HDOPValuesHourly);
			}
		}
	}
	if(options->PDOPPlot==1||options->CombinedDOPPlot==1) {
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			if(options->DOPPercentile) {
				for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
					free(SBASplots->PDOPValues[i][j]);
					if (options->SBASHourlyMaps==1) {
						free(SBASplots->PDOPValuesHourly[i][j]);
					}
				}
			}
			free(SBASplots->PDOPMean[i]);
			if(options->DOPPercentile) {
				free(SBASplots->PDOPValues[i]);
			}
			if (options->SBASHourlyMaps==1) {
				free(SBASplots->PDOPMeanHourly[i]);
				if(options->DOPPercentile) {
					free(SBASplots->PDOPValuesHourly[i]);
				}
			}
		}
		free(SBASplots->PDOPMean);
		if(options->DOPPercentile) {
			free(SBASplots->PDOPValues);
		}
		if (options->SBASHourlyMaps==1) {
			free(SBASplots->PDOPMeanHourly);
			if(options->DOPPercentile) {
				free(SBASplots->PDOPValuesHourly);
			}
		}
	}
	if(options->GDOPPlot==1||options->CombinedDOPPlot==1) {
		for(i=0;i<SBASplots->AvailabilityLatSize;i++) {
			if(options->DOPPercentile) {
				for(j=0;j<SBASplots->AvailabilityLonSize;j++) {
					free(SBASplots->GDOPValues[i][j]);
					if (options->SBASHourlyMaps==1) {
						free(SBASplots->GDOPValuesHourly[i][j]);
					}
				}
			}
			free(SBASplots->GDOPMean[i]);
			if(options->DOPPercentile) {
				free(SBASplots->GDOPValues[i]);
			}
			if (options->SBASHourlyMaps==1) {
				free(SBASplots->GDOPMeanHourly[i]);
				if(options->DOPPercentile) {
					free(SBASplots->GDOPValuesHourly[i]);
				}
			}
		}
		free(SBASplots->GDOPMean);
		if(options->DOPPercentile) {
			free(SBASplots->GDOPValues);
		}
		if (options->SBASHourlyMaps==1) {
			free(SBASplots->GDOPMeanHourly);
			if(options->DOPPercentile) {
				free(SBASplots->GDOPValuesHourly);
			}
		}
	}
}


/*****************************************************************************
 * Name        : initUserError
 * Description : Initialise a TUserError structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TUserError *UserError           O  N/A  TUserError struct
 *****************************************************************************/
void initUserError (TUserError *UserError) {
	int i;

	for (i=0;i<NUMERRORFUNCTIONS;i++) {
		UserError->activeErrorsStart[i] = NULL;
		UserError->activeErrorsEnd[i] = NULL;
		UserError->ErrorParam[i] = NULL;
		UserError->measType[i] = NULL;
		UserError->System[i] = NULL;
		UserError->PRN[i] = NULL;
		UserError->numactiveErrors[i] = 0;
	}
	UserError->totalactiveErrors = 0;
}

/*****************************************************************************
 * Name        : initRTCM2
 * Description : Initialise a TRTCM2 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TRTCM2 *rtcm2                   O  N/A  TRTCM2 struct
 *****************************************************************************/
void initRTCM2 (TRTCM2 *rtcm2) {
	// Initialise control variables
	rtcm2->doWeHaveHeaderCorr = 0;
	rtcm2->doWeHaveHeaderAnt = 0;
	rtcm2->antMSG3 = 0;
	rtcm2->antMSG24 = 0;
	rtcm2->firstPass = 0;
	rtcm2->firstPass2 = 0;

	// Initialise the RTCM v2.x header
	rtcm2->header.message_type = 0;
	rtcm2->header.station_id = 0;
	rtcm2->header.modi_zcount = 0.0;
	rtcm2->header.sequence_no = 0;
	rtcm2->header.dataword_no = 0;
	strcpy(rtcm2->header.station_health,"ini");

	// Initialise data read
	rtcm2->dataRead.indicator_end = 0;
	rtcm2->dataRead.offset = 0;
	strcpy(rtcm2->dataRead.ch_str,"ini");

	// Initialise variables
	rtcm2->i = 0;
	rtcm2->checkControl = 1;
	rtcm2->last2bits[0] = 0;
	rtcm2->last2bits[1] = 1;
	rtcm2->my_state = Initial;
	rtcm2->prevZcount = 0.0;
}

/*****************************************************************************
 * Name        : initRTCM3
 * Description : Initialise a TRTCM3 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TRTCM3 *rtcm3                   O  N/A  TRTCM3 struct
 *****************************************************************************/
void initRTCM3 (TRTCM3 *rtcm3) {
	int i;

	// Initialise control variables
	rtcm3->doWeHaveHeader = 0;
	rtcm3->numObsHeaderRead = 0;
	rtcm3->doWeHaveAntennaPos = 0;
	rtcm3->doWeHaveAntennaName = 0;
	rtcm3->counter = 0;

	// Initialise the RTCM v3.x header
	rtcm3->header.message_type = 0;
	rtcm3->header.station_id = 0;
	rtcm3->header.epoch_time = 0;
	rtcm3->header.synchronous = '2';
	rtcm3->header.gps_no = 0;
	rtcm3->header.smooth_indicator = '2';
	rtcm3->header.smooth_interval = 0;

	// Initialise the RTCM v3.x message 1004
	for ( i = 0; i < 64; i++ ) {
		rtcm3->msg1004[i].satellite_id = 0;
		rtcm3->msg1004[i].l1_code_indicator = 0;
		rtcm3->msg1004[i].l1_pseudoR = 0;
		rtcm3->msg1004[i].l1_part_phaseR = 0;
		rtcm3->msg1004[i].l1_lock_indicator = 0;
		rtcm3->msg1004[i].l1_N_ambiguity = 0;
		rtcm3->msg1004[i].l1_cnr = 0;
		rtcm3->msg1004[i].l2_code_indicator = 0;
		rtcm3->msg1004[i].pseudoR_differ = 0;
		rtcm3->msg1004[i].l2Phase_l1Pseudo = 0;
		rtcm3->msg1004[i].l2_lock_indicator = 0;
		rtcm3->msg1004[i].l2_cnr = 0;
	}

	// Initialise the RTCM v3.x message 1006
	rtcm3->msg1006.message_type = 0;
	rtcm3->msg1006.station_id = 0;
	rtcm3->msg1006.gps_indicator = '2';
	rtcm3->msg1006.glonass_indicator = '2';
	rtcm3->msg1006.referStation_indicator = '2';
	rtcm3->msg1006.ecef_x = 0.0;
	rtcm3->msg1006.oscillator_indicator = '2';
	rtcm3->msg1006.ecef_y = 0.0;
	rtcm3->msg1006.ecef_z = 0.0;
	rtcm3->msg1006.antenna_height = 0.0;

	// Initialise the RTCM v3.x message 1008
	rtcm3->msg1008.message_type = 0;
	rtcm3->msg1008.station_id = 0;
	rtcm3->msg1008.descriptor_no = 0;
	strcpy(rtcm3->msg1008.descriptor,"ini");
	rtcm3->msg1008.setup_id = 0;
	rtcm3->msg1008.serialNum_no = 0;
	strcpy(rtcm3->msg1008.serialNum,"ini");

	for (i=0;i<MAX_SATELLITES_PER_GNSS;i++) {
		rtcm3->L1prevLLI[i] = -9999;
		rtcm3->L2prevLLI[i] = -9999;
	}

	strcpy(rtcm3->content_str,"ini");
	rtcm3->hour = malloc(sizeof(int));
	rtcm3->previousHour = malloc(sizeof(int));
}

/*****************************************************************************
 * Name        : initMSG1
 * Description : Initialise a TMSG1 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TMSG1 *sc                       O  N/A  TMSG1 struct
 * int n                           I  N/A  Number of satellites
 *****************************************************************************/
void initMSG1 (TMSG1 *sc, int n) {
	int i;

	for ( i=0;i<n;i++ ) {
		sc[i].scale_factor = 0;
//		sc[i].udre_lowerlimit = 0.0;
//		sc[i].udre_upperlimit = 0.0;
		sc[i].udre = 0;
		sc[i].satellite_id = 0;
		sc[i].prc = 0;
		sc[i].rrc = 0;
		sc[i].iod = 0;
	}
}

/*****************************************************************************
 * Name        : initMSG2
 * Description : Initialise a TMSG2 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TMSG2 *sc                       O  N/A  TMSG2 struct
 * int n                           I  N/A  Number of satellites
 *****************************************************************************/
void initMSG2 (TMSG2 *sc, int n) {
	int i;

	for (i=0;i<n;i++) {
		sc[i].scale_factor = 0;
//		sc[i].udre_lowerlimit = 0.0;
//		sc[i].udre_upperlimit = 0.0;
		sc[i].udre = 0;
		sc[i].satellite_id = 0;
		sc[i].delta_prc = 0;
		sc[i].delta_rrc = 0;
		sc[i].iod = 0;
	}
}

/*****************************************************************************
 * Name        : initMSG3
 * Description : Initialise a TMSG3 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TMSG3 *sc                       O  N/A  TMSG3 struct
 *****************************************************************************/
void initMSG3 (TMSG3 *sc) {
	sc->x_coord = 0.0;
	sc->y_coord = 0.0;
	sc->z_coord = 0.0;
}

/*****************************************************************************
 * Name        : initMSG24
 * Description : Initialise a TMSG24 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TMSG24 *sc                      O  N/A  TMSG24 struct
 *****************************************************************************/
void initMSG24 (TMSG24 *sc) {
	sc->x_coord = 0.0;
	sc->y_coord = 0.0;
	sc->z_coord = 0.0;
	sc->system_indicator = '2';
	sc->AH_indicator = '2';
	sc->antenna_height = 0.0;
}

/*****************************************************************************
 * Name        : changeBase
 * Description : Changes a 3D vector to the specified base
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *base                   I  N/A  Base (vectors of the base in the rows)
 * double  *vector                 IO N/A  Vector to be changed (output 
 *                                        overwrites this vector)
 *****************************************************************************/
void changeBase (double base[3][3], double vector[3]) {
	double	res[3];
	int		i,j;

	for (i=0;i<3;i++) {
		res[i] = 0;
		for (j=0;j<3;j++) {
			res[i] += base[i][j] * vector[j];
		}
	}

	memcpy (vector,res,sizeof(double)*3);
}

/*****************************************************************************
 * Name        : min
 * Description : Return the minimum between two numbers
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a                       I  N/A  First number
 * double  b                       I  N/A  Second number 
 * Returned value (double)         O  N/A  Minimum between a and b
 *****************************************************************************/
double min (double a, double b) {
	if ( a < b ) return a;
	else return b;
}

/*****************************************************************************
 * Name        : max
 * Description : Return the maximum between two numbers
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a                       I  N/A  First number
 * double  b                       I  N/A  Second number 
 * Returned value (double)         O  N/A  Maximum between a and b
 *****************************************************************************/
double max (double a, double b) {
	if (a<b) return b;
	else return a;
}

/*****************************************************************************
 * Name        : modulo
 * Description : Does the modulo mod of a number
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a                       I  N/A  Number
 * double  mod                     I  N/A  Modulo base
 * Returned value (double)         O  N/A  Modulo
 *****************************************************************************/
double modulo (double a, double mod) {
	return a - ((int)(a/mod))*mod;
}

/*****************************************************************************
 * Name        : mod
 * Description : Does the modulo for an integer 
 *              (handling correctly negative numbers)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int     a                       I  N/A  Number
 * int     N                       I  N/A  Modulo base
 * Returned value (int)            O  N/A  Modulo
 *****************************************************************************/
int mod (int a, int N) {
	int r = a % N;
	return r < 0 ? r + N : r;
}

/*****************************************************************************
 * Name        : unitary
 * Description : Makes a vector unitary
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *vect                   IO N/A  3 dimension vector (output
 *                                           overwrites this vector)
 *****************************************************************************/
void unitary (double *vect) {
	double	module;

	module = sqrt(vect[0]*vect[0] + vect[1]*vect[1] + vect[2]*vect[2]);

	vect[0] /= module;
	vect[1] /= module;
	vect[2] /= module;
}

/*****************************************************************************
 * Name        : vectProd
 * Description : Makes the vectorial product between two vectors
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *a                      I  N/A  First vector
 * double  *b                      I  N/A  Second vector
 * double  *res                    O  N/A  Result vector [res = a x b]
 *****************************************************************************/
void vectProd (double *a, double *b, double *res) {
	res[0] = a[1]*b[2] - a[2]*b[1];
	res[1] = a[2]*b[0] - a[0]*b[2];
	res[2] = a[0]*b[1] - a[1]*b[0];
}

/*****************************************************************************
 * Name        : scalarProd
 * Description : Makes the scalar product between two vectors
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *a                      I  N/A  First vector
 * double  *b                      I  N/A  Second vector
 * Returned value (double)         O  N/A  Result vector [res = a · b]
 *****************************************************************************/
double scalarProd (double *a, double *b) {
	return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

/*****************************************************************************
 * Name        : trim
 * Description : Trim a string (remove preceding and succeding spaces)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *line                     IO N/A  String to trim
 * Returned value (char*)          O  N/A  Pointer to *line
 *****************************************************************************/
char *trim (char *line) {
	unsigned int 	i;
	int 			pos=0;
	int 			lastnotspace=0;
	int 			state=0;
		// 0 -> Trimming first spaces
		// 1 -> Waiting to end string

	for (i=0;i<strlen(line);i++){
		if (state==0) {
			if (line[i]!=' ') {
				state = 1;
				line[pos] = line[i];
				pos++;
				lastnotspace = pos;
			}
		} else if (state==1) {
			line[pos] = line[i];
			pos++;
			if (line[i]!=' ' && line[i]!='\n') {
				lastnotspace = pos;
			}
		}
	}
	line[lastnotspace]='\0';
	return &line[0];
}

/*****************************************************************************
 * Name        : getstr
 * Description : Get a substring from a string and trim it
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *out                      O  N/A  Substring
 * char  *line                     I  N/A  Input string
 * int  ini                        I  N/A  Initial character position
 * int  length                     I  N/A  Length of the substring
 * Returned value (char*)          O  N/A  Pointer to *out
 *****************************************************************************/
char *getstr (char *out,char *line,int ini, int length) {
	int end;

	end = strlen(line);
	if (ini>=end) {
		out[0]='\0';
		return &out[0];
	}
	strncpy(out,&line[ini],length);
	out[length]='\0';
	trim(out);
	return &out[0];
}

/*****************************************************************************
 * Name        : getnumericalelements
 * Description : Gets the different numbers in a string separated with spaces
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *line                     I  N/A  Input string
 * int  *numelements               O  N/A  Number of elements in the string
 * double  *out                    O  N/A  Output vector
 *****************************************************************************/
void getnumericalelements (char *line, int *numelements, double *datavector) {
	int i,j,end, counter;
	int numberstarted; //0-> not reading number (spaces) 1->reading number byte per byte
	char aux[100];

	//Initialize variables
	j = 0;
	numberstarted = 0;
	counter = 0;
	end = strlen(line);

	//Intialize output vector
		
	for (i=0;i<end;i++){
		if (line[i]!=' ' && line[i]!='\n' && line[i]!=13) {
			if (numberstarted == 0) {
				numberstarted = 1;
			}	
			aux[j] = line[i];
			j++;
		} else if (numberstarted == 1) {
			aux[j] = '\0';
			counter++;
			datavector[counter-1] = atof(aux);
			j=0;
			numberstarted = 0;
		}
	}

	*numelements = counter;
}

/*****************************************************************************
 * Name        : checkConstellationChar
 * Description : Check the given character corresponds to the character of
 *                 a GNSS system
 * Parameters  :
 * Name                            |Da|Unit|Description
 * char c                           I  N/A  Character to be checked for a constellation
 * Returned value (int)             O  N/A  0-> Not a valid constellation letter
 *                                          1-> Valid constellation letter
 *****************************************************************************/
int checkConstellationChar(char c) {
	if (c!='G' && c!='E' && c!='R' && c!='S' && c!='C' && c!='J' && c!='I') return 0;
	else return 1;
}

/*****************************************************************************
 * Name        : whatIs
 * Description : Check whether an observation is carrier phase, pseudorange, 
 *               SNR or doppler
 * Parameters  :
 * Name                            |Da|Unit|Description
 * MeasurementType  meas            I  N/A  Measurement type (C1P, P1P...)
 * Returned value (MeasurementKind) O  N/A  Measurement kind (Pseudorange, CarrierPhase...)
 *****************************************************************************/
enum MeasurementKind whatIs (enum MeasurementType meas) {

	if (meas == NA) return UNKNOWN_MEAS;
	else if (meas>=ENDMEAS) {
		// Combination
		if (meas == LW || meas == LI || meas == LC || meas == G1C || meas == G1P || meas == G2C || meas == G2P || meas == DF || meas == BW) return CarrierPhase;
		else if (meas == PW || meas == PI || meas == PC || meas == PCC) return Pseudorange;
		else return UNKNOWN_MEAS;
	} else if (((meas-1)%4)==0) return Pseudorange;
	else if (((meas-2)%4)==0) return CarrierPhase;
	else if (((meas-3)%4)==0) return Doppler;
	else if (((meas)%4)==0) return SNR;
	else return UNKNOWN_MEAS;
}

/*****************************************************************************
 * Name        : getLambda
 * Description : Get the lambda corresponding to the GNSS and measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSS system
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (double)         O  m    Lambda on the measurement
 *****************************************************************************/
double getLambda (enum GNSSystem GNSS, enum MeasurementType meas) {
	if (GNSS == GPS) {
		if (meas <= S1N) return GPSl1;
		else if (meas <= S2N) return GPSl2;
		else if (meas <= S5X) return GPSl5;
	} else if (GNSS == Galileo) {
		if (meas <= S1N) return GALl1;
		else if (meas > S2N && meas <= S5X) return GALl5a;
		else if (meas <= S7X) return GALl5b;
		else if (meas <= S8X) return GALl5;
		else if (meas <= S6Z) return GALl6;
	} else if (GNSS == GLONASS) {
		return 1;
	} else if (GNSS == GEO) {
		if (meas <= S1N) return SBASl1;
		else if (meas > S2N && meas <= S5X) return SBASl5;
	}

	return 1;
}

/*****************************************************************************
 * Name        : getFrequency
 * Description : Get the frequency corresponding to the GNSS and measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSS system
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (double)         O  Hz   Frequency of the measurement
 *****************************************************************************/
double getFrequency (enum GNSSystem GNSS, enum MeasurementType meas) {
	return c0/getLambda (GNSS,meas);
}

/*****************************************************************************
 * Name        : getFrequencyInt
 * Description : Get the integer identifier value for the frequency 
 *               corresponding to the measurement
 * Parameters  :
 * Name                           |Da|Unit|Description
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (int)            O  N/A  Frequency identifier of the measurement:
 *							GPS:		'G01' - 1 - L1
 *										'G02' - 2 - L2 
 *										'G05' - 5 - L5
 *							GLONASS:	'R01' - 1 - G1
 *										'R02' - 2 - G2
 *							Galileo:	'E01' - 1 - E1
 *										'E05' - 5 - E5a
 *										'E07' - 7 - E5b
 *										'E08' - 8 - E5 (E5a-E5b)
 *										'E06' - 6 - E56   
 *****************************************************************************/
int getFrequencyInt (enum MeasurementType meas) {
	if (meas <= S1N) return 1;
	else if (meas <= S2N) return 2;
	else if (meas <= S5X) return 5;
	else if (meas <= S7X) return 7;
	else if (meas <= S8X) return 8;
	else if (meas <= S6Z) return 6;
	else return 0;
}

/*****************************************************************************
 * Name        : measstr2meastype
 * Description : Transform from string to internal Measurement type
 * Parameters  :
 * Name                            |Da|Unit|Description
 * char  *str                       I  N/A  Measurement string
 * Returned value (MeasurementType) O  N/A  Measurement type
 *****************************************************************************/
enum MeasurementType measstr2meastype (char *str) {
	// Possible options:
	//    1st digit: C,L,D,S,P,G,B   [7]
	//    2nd digit: 1,2,5,6,7,8,W,I,C,F    [10]
	//    3rd digit: <empty>,A,B,C,D,I,L,M,N,P,Q,S,W,X,Y,Z   [16]

	//This function is entered in single processing mode (at least the first time), and only the
	//first time the static variables are initialized, therefore there will be no race conditions

	static int	initialized = 0;
	static enum MeasurementType	s2t[7][10][16];
	int 		a1,a2,a3;

	if (!initialized) {
		for (a1=0;a1<7;a1++) {
			for (a2=0;a2<10;a2++) {
				for (a3=0;a3<16;a3++) {
					s2t[a1][a2][a3] = NA;
				}
			}
		}

		// Combinations
		s2t[1][6][0] = LW;
		s2t[4][6][0] = PW;
		s2t[6][6][0] = BW;
		s2t[1][7][0] = LI;
		s2t[4][7][0] = PI;
		s2t[1][8][0] = LC;
		s2t[4][8][0] = PC;
		s2t[4][8][3] = PCC;
		s2t[5][0][3] = G1C;
		s2t[5][0][9] = G1P;
		s2t[5][1][3] = G2C;
		s2t[5][1][9] = G2P;
		s2t[2][9][0] = DF;
		
		////// Automatically generated code block /////
		// RINEX 3.00
		s2t[0][0][1] = C1A; s2t[1][0][1] = L1A; s2t[2][0][1] = D1A; s2t[3][0][1] = S1A; s2t[0][0][2] = C1B; s2t[1][0][2] = L1B; s2t[2][0][2] = D1B; s2t[3][0][2] = S1B; s2t[0][0][3] = C1C; s2t[1][0][3] = L1C; s2t[2][0][3] = D1C; s2t[3][0][3] = S1C; s2t[0][0][9] = C1P; s2t[1][0][9] = L1P; s2t[2][0][9] = D1P; s2t[3][0][9] = S1P; s2t[0][0][12] = C1W; s2t[1][0][12] = L1W; s2t[2][0][12] = D1W; s2t[3][0][12] = S1W; s2t[0][0][13] = C1X; s2t[1][0][13] = L1X; s2t[2][0][13] = D1X; s2t[3][0][13] = S1X; s2t[0][0][14] = C1Y; s2t[1][0][14] = L1Y; s2t[2][0][14] = D1Y; s2t[3][0][14] = S1Y; s2t[0][0][15] = C1Z; s2t[1][0][15] = L1Z; s2t[2][0][15] = D1Z; s2t[3][0][15] = S1Z; s2t[0][0][7] = C1M; s2t[1][0][7] = L1M; s2t[2][0][7] = D1M; s2t[3][0][7] = S1M; s2t[1][0][8] = L1N; s2t[2][0][8] = D1N; s2t[3][0][8] = S1N;
		s2t[0][1][3] = C2C; s2t[1][1][3] = L2C; s2t[2][1][3] = D2C; s2t[3][1][3] = S2C; s2t[0][1][4] = C2D; s2t[1][1][4] = L2D; s2t[2][1][4] = D2D; s2t[3][1][4] = S2D; s2t[0][1][11] = C2S; s2t[1][1][11] = L2S; s2t[2][1][11] = D2S; s2t[3][1][11] = S2S; s2t[0][1][6] = C2L; s2t[1][1][6] = L2L; s2t[2][1][6] = D2L; s2t[3][1][6] = S2L; s2t[0][1][13] = C2X; s2t[1][1][13] = L2X; s2t[2][1][13] = D2X; s2t[3][1][13] = S2X; s2t[0][1][9] = C2P; s2t[1][1][9] = L2P; s2t[2][1][9] = D2P; s2t[3][1][9] = S2P; s2t[0][1][12] = C2W; s2t[1][1][12] = L2W; s2t[2][1][12] = D2W; s2t[3][1][12] = S2W; s2t[0][1][14] = C2Y; s2t[1][1][14] = L2Y; s2t[2][1][14] = D2Y; s2t[3][1][14] = S2Y; s2t[0][1][7] = C2M; s2t[1][1][7] = L2M; s2t[2][1][7] = D2M; s2t[3][1][7] = S2M; s2t[1][1][8] = L2N; s2t[2][1][8] = D2N; s2t[3][1][8] = S2N;
		s2t[0][2][5] = C5I; s2t[1][2][5] = L5I; s2t[2][2][5] = D5I; s2t[3][2][5] = S5I; s2t[0][2][10] = C5Q; s2t[1][2][10] = L5Q; s2t[2][2][10] = D5Q; s2t[3][2][10] = S5Q; s2t[0][2][13] = C5X; s2t[1][2][13] = L5X; s2t[2][2][13] = D5X; s2t[3][2][13] = S5X;
		s2t[0][4][5] = C7I; s2t[1][4][5] = L7I; s2t[2][4][5] = D7I; s2t[3][4][5] = S7I; s2t[0][4][10] = C7Q; s2t[1][4][10] = L7Q; s2t[2][4][10] = D7Q; s2t[3][4][10] = S7Q; s2t[0][4][13] = C7X; s2t[1][4][13] = L7X; s2t[2][4][13] = D7X; s2t[3][4][13] = S7X;
		s2t[0][5][5] = C8I; s2t[1][5][5] = L8I; s2t[2][5][5] = D8I; s2t[3][5][5] = S8I; s2t[0][5][10] = C8Q; s2t[1][5][10] = L8Q; s2t[2][5][10] = D8Q; s2t[3][5][10] = S8Q; s2t[0][5][13] = C8X; s2t[1][5][13] = L8X; s2t[2][5][13] = D8X; s2t[3][5][13] = S8X;
		s2t[0][3][1] = C6A; s2t[1][3][1] = L6A; s2t[2][3][1] = D6A; s2t[3][3][1] = S6A; s2t[0][3][2] = C6B; s2t[1][3][2] = L6B; s2t[2][3][2] = D6B; s2t[3][3][2] = S6B; s2t[0][3][3] = C6C; s2t[1][3][3] = L6C; s2t[2][3][3] = D6C; s2t[3][3][3] = S6C; s2t[0][3][13] = C6X; s2t[1][3][13] = L6X; s2t[2][3][13] = D6X; s2t[3][3][13] = S6X; s2t[0][3][15] = C6Z; s2t[1][3][15] = L6Z; s2t[2][3][15] = D6Z; s2t[3][3][15] = S6Z;
		////// End of automatically generated code block /////

		// RINEX 2.11
		s2t[0][0][0] = C1C; // C1
		s2t[4][0][0] = C1P; // P1
		s2t[1][0][0] = L1P; // L1
		s2t[2][0][0] = D1P; // D1
		s2t[3][0][0] = S1P; // S1
		
		s2t[0][1][0] = C2C;	// C2
		s2t[4][1][0] = C2P; // P2
		s2t[1][1][0] = L2P;	// L2
		s2t[2][1][0] = D2P;	// D2
		s2t[3][1][0] = S2P; // S2
		
		s2t[0][2][0] = C5X; // C5
		s2t[1][2][0] = L5X; // L5
		s2t[2][2][0] = D5X; // D5
		s2t[3][2][0] = S5X; // S5
		
		s2t[0][3][0] = C6X; // C6
		s2t[1][3][0] = L6X; // L6
		s2t[2][3][0] = D6X; // D6
		s2t[3][3][0] = S6X; // S6
		
		s2t[0][4][0] = C7X; // C7
		s2t[1][4][0] = L7X; // L7
		s2t[2][4][0] = D7X; // D7
		s2t[3][4][0] = S7X; // S7
		
		s2t[0][5][0] = C8X; // C8
		s2t[1][5][0] = L8X; // L8
		s2t[2][5][0] = D8X; // D8
		s2t[3][5][0] = S8X; // S8

		initialized = 1;
	}

	if (str[0]=='C') a1=0;
	else if (str[0]=='L') a1=1;
	else if (str[0]=='D') a1=2;
	else if (str[0]=='S') a1=3;
	else if (str[0]=='P') a1=4;
	else if (str[0]=='G') a1=5;
	else if (str[0]=='B') a1=6;
	else return NA;

	if (str[1]=='1') a2=0;
	else if (str[1]=='2') a2=1;
	else if (str[1]=='5') a2=2;
	else if (str[1]=='6') a2=3;
	else if (str[1]=='7') a2=4;
	else if (str[1]=='8') a2=5;
	else if (str[1]=='W') a2=6;
	else if (str[1]=='I') a2=7;
	else if (str[1]=='C') a2=8;
	else if (str[1]=='F') a2=9;
	else return NA;

	if (str[2]=='\0') a3=0;
	else if (str[2]=='A') a3=1;
	else if (str[2]=='B') a3=2;
	else if (str[2]=='C') a3=3;
	else if (str[2]=='D') a3=4;
	else if (str[2]=='I') a3=5;
	else if (str[2]=='L') a3=6;
	else if (str[2]=='M') a3=7;
	else if (str[2]=='N') a3=8;
	else if (str[2]=='P') a3=9;
	else if (str[2]=='Q') a3=10;
	else if (str[2]=='S') a3=11;
	else if (str[2]=='W') a3=12;
	else if (str[2]=='X') a3=13;
	else if (str[2]=='Y') a3=14;
	else if (str[2]=='Z') a3=15;
	else return NA;
	
	if (strncmp(str,"L1C",3)==0 && ReadL1CAsItself==0) {return s2t[1][0][0];} //Trick to make gLAB believe L1C is L1P 
	return s2t[a1][a2][a3];
}

/*****************************************************************************
 * Name        : meastype2measstr
 * Description : Transform from internal Measurement type to string
 * Parameters  :
 * Name                           |Da|Unit|Description
 * MeasurementType  meas           I  N/A  Measurement type
 * Returned value (char*)          O  N/A  Measurement string
 *****************************************************************************/
char *meastype2measstr (enum MeasurementType meas) {
	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[10];
	//Static variables below are initalized at compile and not modified during execution, so no race conditions occur
	#pragma omp threadprivate(str)
	static int	initialized = 0;
	#pragma omp threadprivate(initialized)
	static char	l[34];
	#pragma omp threadprivate(l)
	char		type;
	int			aux;
	
	strcpy(str,"NAN");

	if (!initialized) {
		initialized = 1;
		l[0]='A'; l[1]='B'; l[2]='C'; l[3]='P'; l[4]='W'; l[5]='X'; l[6]='Y'; l[7]='Z'; l[8]='M'; l[9]='N';
		l[10]='C'; l[11]='D'; l[12]='S'; l[13]='L'; l[14]='X'; l[15]='P'; l[16]='W'; l[17]='Y'; l[18]='M'; l[19]='N';
		l[20]='I'; l[21]='Q'; l[22]='X'; l[23]='I'; l[24]='Q'; l[25]='X'; l[26]='I'; l[27]='Q'; l[28]='X'; l[29]='A';
		l[30]='B'; l[31]='C'; l[32]='X'; l[33]='Z';
	}

	if (meas<ENDMEAS) {
		switch (whatIs(meas)) {
			case Pseudorange:
				type='C';
				break;
			case CarrierPhase:
				type='L';
				break;
			case Doppler:
				type='D';
				break;
			case SNR:
				type='S';
				break;
			default:
				return str;
				break;
		}
		aux = getFrequencyInt(meas);
		sprintf(str,"%1c%1d%1c",type,aux,l[(int)((meas-1)/4)]);
	} else {
		switch (meas) {
			case 138:
				strcpy(str,"LW");
				break;
			case 139:
				strcpy(str,"BW");
				break;
			case 140:
				strcpy(str,"PW");
				break;
			case 141:
				strcpy(str,"LI");
				break;
			case 142:
				strcpy(str,"PI");
				break;
			case 143:
				strcpy(str,"LC");
				break;
			case 144:
				strcpy(str,"PC");
				break;
			case 145:
				strcpy(str,"PCC");
				break;
			case 146:
				strcpy(str,"G1C");
				break;
			case 147:
				strcpy(str,"G1P");
				break;
			case 148:
				strcpy(str,"G2C");
				break;
			case 149:
				strcpy(str,"G2P");
				break;
			case 150:
				strcpy(str,"DF");
				break;
			default:
				return str;
				break;
		}
	}

	return str;
}

/*****************************************************************************
 * Name        : gnsschar2gnsstype
 * Description : Transform from string to internal GNSS System type
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *system                   I  N/A  GNSS System string
 * Returned value (GNSSystem)      O  N/A  GNSSystem type
 *****************************************************************************/
enum GNSSystem gnsschar2gnsstype (char system) {
	if (system=='G' || system==' ') return GPS;
	else if (system=='E') return Galileo;
	else if (system=='R') return GLONASS;
	else if (system=='S') return GEO;
	else if (system=='C') return BDS;
	else if (system=='J') return QZSS;
	else if (system=='I') return IRNSS;

	return GPS; // Default value
}

/*****************************************************************************
 * Name        : gnsstype2gnssstr
 * Description : Transform from internal GNSS System type to string
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSSystem type
 * Returned value (char*)          O  N/A  GNSS System string
 *****************************************************************************/
char *gnsstype2gnssstr (enum GNSSystem GNSS) {
	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[10];
	#pragma omp threadprivate(str)

	strcpy(str,"NAN");

	switch (GNSS) {
		case GPS:
			strcpy(str,"GPS");
			break;
		case Galileo:
			strcpy(str,"GAL");
			break;
		case GLONASS:
			strcpy(str,"GLO");
			break;
		case GEO:
			strcpy(str,"GEO");
			break;
		case BDS:
			strcpy(str,"BDS");
			break;
		case QZSS:
			strcpy(str,"QZS");
			break;
		case IRNSS:
			strcpy(str,"IRN");
			break;
		default:
			strcpy(str,"NAN");
	}

	return str;
}

/*****************************************************************************
 * Name        : gnsstype2gnssname
 * Description : Transform from internal GNSS System type to GNSS long name
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSSystem type
 * Returned value (char*)          O  N/A  GNSS System string
 *****************************************************************************/
char *gnsstype2gnssname (enum GNSSystem GNSS) {
	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[10];
	#pragma omp threadprivate(str)

    switch (GNSS) {
        case GPS:
            strcpy(str,"GPS");
            break;
        case Galileo:
            strcpy(str,"Galileo");
            break;
        case GLONASS:
            strcpy(str,"GLONASS");
            break;
        case GEO:
            strcpy(str,"SBAS");
            break;
        case BDS:
            strcpy(str,"BeiDou");
            break;
        case QZSS:
            strcpy(str,"QZSS");
            break;
        case IRNSS:
            strcpy(str,"IRNSS");
            break;
        default:
			strcpy(str,"NAN");
    }

    return str;
}

/*****************************************************************************
 * Name        : gnsstype2char
 * Description : Transform from internal GNSS System type to char
 * Parameters  :
 * Name                           |Da|Unit|Description
 * GNSSystem  GNSS                 I  N/A  GNSSystem type
 * Returned value (char)           O  N/A  GNSS System char
 *****************************************************************************/
char gnsstype2char (enum GNSSystem GNSS) {
	
	switch (GNSS) {
		case GPS:
			return 'G';
			break;
		case Galileo:
			return 'E';
			break;
		case GLONASS:
			return 'R';
			break;
		case GEO:
			return 'S';
			break;
		case BDS:
			return 'C';
			break;
		case QZSS:
			return 'J';
			break;
		case IRNSS:
			return 'I';
			break;
		default:
			return 'N';
	}

	return 'N';
}

/*****************************************************************************
 * Name        : satblockstr2satblock
 * Description : Transform from a satellite block string to the internal 
 *               satellite block
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *str                      I  N/A  Satellite block string
 * Returned value (SatelliteBlock) O  N/A  Internal satellite block
 *****************************************************************************/
enum SatelliteBlock satblockstr2satblock (char *str) {
	if (strcmp(str,"BLOCK I")==0) {
		return GPS_BLOCK_I;
	} else if (strcmp(str,"BLOCK II")==0) {
		return GPS_BLOCK_II;
	} else if (strcmp(str,"BLOCK IIA")==0) {
		return GPS_BLOCK_IIA;
	} else if (strcmp(str,"BLOCK IIR")==0) {
		return GPS_BLOCK_IIR;
	} else if (strcmp(str,"BLOCK IIR-A")==0) {
		return GPS_BLOCK_IIR;
	} else if (strcmp(str,"BLOCK IIR-B")==0) {
		return GPS_BLOCK_IIR;
	} else if (strcmp(str,"BLOCK IIR-M")==0) {
		return GPS_BLOCK_IIRM;
	} else if (strcmp(str,"GLONASS")==0) {
		return GLONASS_BLOCK;
	} else if (strcmp(str,"GLONASS-M")==0) {
		return GLONASS_BLOCK_M;
	}
	
	return UNKNOWN_BLOCK;
}

/*****************************************************************************
 * Name        : satblockstr2satblock
 * Description : Transform from an internal satellite block to a satellite  
 *               block string
 * Parameters  :
 * Name                           |Da|Unit|Description
 * SatelliteBlock  satBlock        I  N/A  Internal satellite block
 * Returned value (char *)         O  N/A  Satellite block string
 *****************************************************************************/
char *satblock2satblockstr (enum SatelliteBlock satBlock) {
	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[20];
	#pragma omp threadprivate(str)

	strcpy(str,"NAN");

	switch (satBlock) {
		case GPS_BLOCK_I:
			strcpy(str,"BLOCK I");
			break;
		case GPS_BLOCK_II:
			strcpy(str,"BLOCK II");
			break;
		case GPS_BLOCK_IIA:
			strcpy(str,"BLOCK IIA");
			break;
		case GPS_BLOCK_IIR:
			strcpy(str,"BLOCK IIR");
			break;
		case GPS_BLOCK_IIRM:
			strcpy(str,"BLOCK IIR-M");
			break;
		case GLONASS_BLOCK:
			strcpy(str,"GLONASS");
			break;
		case GLONASS_BLOCK_M:
			strcpy(str,"GLONASS-M");
			break;
		default:
			strcpy(str,"NAN");
	}

	return str;
}

/*****************************************************************************
 * Name        : timecorrstr2timecorrtype
 * Description : Transform from string to internal TimeSystemCorrection
 * Parameters  :
 * Name                                   |Da|Unit|Description
 * char  *str                              I  N/A  TimeSystemCorrection string
 * Returned value (TimeSystemCorrection)   O  N/A  TimeSystemCorrection type
 *****************************************************************************/
enum TimeSystemCorrection timecorrstr2timecorrtype (char *str) {
	if (strcmp(str,"GAUT")==0) {
				return GAUT;
		} else if (strcmp(str,"GPUT")==0) {
				return GPUT;
		} else if (strcmp(str,"SBUT")==0) {
				return SBUT;
		} else if (strcmp(str,"GLUT")==0) {
				return GLUT;
		} else if (strcmp(str,"GPGA")==0) {
				return GPGA;
		} else if (strcmp(str,"GLGP")==0) {
				return GLGP;
		} else if (strcmp(str,"QZGP")==0) {
				return QZGP;
		} else if (strcmp(str,"QZUT")==0) {
				return QZUT;
		} else if (strcmp(str,"BDUT")==0) {
				return BDUT;
		} else if (strcmp(str,"IRUT")==0) {
				return IRUT;
		} else if (strcmp(str,"IRGP")==0) {
				return IRGP;
		}

		return UNKNOWN_TIME_CORR;
}

/*****************************************************************************
 * Name        : timecorrtype2timecorrstr
 * Description : Transform from internal TimeSystemCorrection type to string
 * Parameters  :
 * Name                                 |Da|Unit|Description
 * enum TimeSystemCorrection timecorr    I  N/A  TimeSystemCorrection type
 * Returned value (char*)                O  N/A  TimeSystemCorrection string
 *****************************************************************************/
char *timecorrtype2timecorrstr (enum TimeSystemCorrection timecorr) {
	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[5];
	#pragma omp threadprivate(str)

	strcpy(str,"NAN");

	switch (timecorr) {
			case GAUT:
					strcpy(str,"GAUT");
					break;
			case GPUT:
					strcpy(str,"GPUT");
					break;
			case SBUT:
					strcpy(str,"SBUT");
					break;
			case GLUT:
					strcpy(str,"GLUT");
					break;
			case GPGA:
					strcpy(str,"GPGA");
					break;
			case GLGP:
					strcpy(str,"GLGP");
					break;
			case QZGP:
					strcpy(str,"QZGP");
					break;
			case QZUT:
					strcpy(str,"QZUT");
					break;
			case BDUT:
					strcpy(str,"BDUT");
					break;
			case IRUT:
					strcpy(str,"IRUT");
					break;
			case IRGP:
					strcpy(str,"IRGP");
					break;
			default:
					strcpy(str,"NAN");
	}

	return str;
}

/*****************************************************************************
 * Name        : SolutionModeNum2SolutionModeChar
 * Description : Transform from internal Solution mode type to string
 * Parameters  :
 * Name                            |Da|Unit|Description
 * enum SolutionMode numsol         I  N/A  Solution mode type
 * Returned value (char*)           O  N/A  Solution mode string
 *****************************************************************************/
char *SolutionModeNum2SolutionModeChar(enum SolutionMode numsol) {

	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[10];
	#pragma omp threadprivate(str)

	switch (numsol) {
		case SPPMode:
			strcpy(str,"SPP");
			break;
		case PPPMode:
			strcpy(str,"PPP");
			break;
		case SBASMode:
			strcpy(str,"SBAS");
			break;
		case DGNSSMode:
			strcpy(str,"DGNSS");
			break;
		default:
			strcpy(str,"Unknown");
			break;
	}

	return str;

}

/*****************************************************************************
 * Name        : WeightModeNum2WeightModestr
 * Description : Transform from internal Weight mode type to string
 * Parameters  :
 * Name                            |Da|Unit|Description
 * enum WeightMode weightMode       I  N/A  Weight mode type
 * Returned value (char*)           O  N/A  Weight mode string
 *****************************************************************************/
char *WeightModeNum2WeightModestr(enum WeightMode weightMode) {

	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[50];
	#pragma omp threadprivate(str)

	switch(weightMode) {
		case FixedWeight: case SBASFixedWeight: case DGNSSFixedWeight:
			strcpy(str,"Fixed");
			break;
		case ElevationWeight: case SBASElevWeight: case DGNSSElevWeight:
			strcpy(str,"Exponential elevation");
			break;
		case Elevation2Weight: case SBASElev2Weight: case DGNSSElev2Weight:
			strcpy(str,"Inverse of sin(elevation)");
			break;
		case SNRWeight: case SBASSNRWeight: case DGNSSSNRWeight:
			strcpy(str,"SNR");
			break;
		case SNRElevWeight: case SBASSNRElevWeight: case DGNSSSNRElevWeight:
			strcpy(str,"SNR and inverse of sin(elevation)");
			break;
		case SBASOnlyWeight:
			strcpy(str,"SBAS");
			break;
		case DGNSSOnlyWeight:
			strcpy(str,"DGNSS");
			break;
		default:
			strcpy(str,"Unknown");
			break;
	}

	return str;
}

/*****************************************************************************
 * Name        : getMeasModelValue
 * Description : Get the measurement and model value of a particular satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * GNSSystem  system               I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType  measType       I  N/A  Measurement type
 * double  *measurement            O  m    Measurement
 * double  *model                  O  m    Model
 * Returned value (int)            O  N/A  Status of the function
 *                                         1 => Measurement found
 *                                         0 => Not found
 *                                        -1 => Found, but not usable
 *****************************************************************************/
int getMeasModelValue (TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType, double *measurement, double *model) {
	int		ind;
	int		i,j;
	double	rawValue;
	double	conversionFactor;
	double	align;
	double	meas1,meas2;
	double	model1,model2;
	int		res1,res2;
	int		usable = 1;

	// Check if measurement is usable
	if ( !epoch->measOrder[system].usable[measType] ) usable = 0;

	if ( measType < ENDMEAS ) { // It is a measurement and NOT a combination
		// If it is a receiver type, that does not has P1, use C1 instead
		if ( epoch->receiver.recType == rtNOP1 && measType == C1P ) { 
			measType = C1C;
		}
		
		// Obtain the stored index of the measurement (-1 if not found)
		ind = epoch->measOrder[system].meas2Ind[measType];
		
		// If the measurement is not found, and it can make equivalences between C1 and P1
		if ( ind == -1 && epoch->receiver.equivalentC1P1 ) { 
			if ( measType == C1C ) {
				measType = C1P;
				ind = epoch->measOrder[system].meas2Ind[measType];
			} else if ( measType == C1P ) {
				measType = C1C;
				ind = epoch->measOrder[system].meas2Ind[measType];
			} 
		}
		if ( ind == -1 ) return 0; // Measurement not found
	} else if ( measType == PW ) {
		res1 = getMeasModelValue(epoch,system,PRN,C1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C2P,&meas2,&model2);
		if ( res1 && res2 ) {
			if ( measurement != NULL ) *measurement = (GPSmf1*meas1 + GPSmf2*meas2) / (GPSmf1 + GPSmf2);
			if ( model != NULL ) *model =  (GPSmf1*model1 + GPSmf2*model2) / (GPSmf1 + GPSmf2);
			if ( usable ) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if ( measType == LW ) {
		res1 = getMeasModelValue(epoch,system,PRN,L1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if ( res1 && res2 ) {
			if ( measurement != NULL ) *measurement = (GPSmf1*meas1 - GPSmf2*meas2) / (GPSmf1 - GPSmf2);
			if ( model != NULL ) *model =  (GPSmf1*model1 - GPSmf2*model2) / (GPSmf1 - GPSmf2);
			if ( usable ) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==BW) {
		res1 = getMeasModelValue(epoch,system,PRN,LW,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,PW,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = meas1 - meas2;
			if (model!=NULL) *model =  model1 - model2;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==PI) {
		res1 = getMeasModelValue(epoch,system,PRN,C2P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C1P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = meas1 - meas2;
			if (model!=NULL) *model =  model1 - model2;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==LI) {
		res1 = getMeasModelValue(epoch,system,PRN,L1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = meas1 - meas2;
			if (model!=NULL) *model =  model1 - model2;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==DF) {
		res1 = getMeasModelValue(epoch,system,PRN,L1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = meas1 + 2.0 * ALPHAGPS12 * ( meas1 - meas2 );
			if (model!=NULL) *model =  model1 + 2.0 * ALPHAGPS12 * ( model1 - model2 );
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==PC) {
		res1 = getMeasModelValue(epoch,system,PRN,C1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*GPSmf1*meas1 - GPSmf2*GPSmf2*meas2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*GPSmf1*model1 - GPSmf2*GPSmf2*model2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==PCC) {
		res1 = getMeasModelValue(epoch,system,PRN,C1C,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,C2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*GPSmf1*meas1 - GPSmf2*GPSmf2*meas2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*GPSmf1*model1 - GPSmf2*GPSmf2*model2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==LC) {
		res1 = getMeasModelValue(epoch,system,PRN,L1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (GPSmf1*GPSmf1*meas1 - GPSmf2*GPSmf2*meas2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (model!=NULL) *model =  (GPSmf1*GPSmf1*model1 - GPSmf2*GPSmf2*model2) / (GPSmf1*GPSmf1 - GPSmf2*GPSmf2);
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G1C) {
		res1 = getMeasModelValue(epoch,system,PRN,C1C,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L1P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G1P) {
		res1 = getMeasModelValue(epoch,system,PRN,C1P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L1P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G2C) {
		res1 = getMeasModelValue(epoch,system,PRN,C2C,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else if (measType==G2P) {
		res1 = getMeasModelValue(epoch,system,PRN,C2P,&meas1,&model1);
		res2 = getMeasModelValue(epoch,system,PRN,L2P,&meas2,&model2);
		if (res1 && res2) {
			if (measurement!=NULL) *measurement = (meas1 + meas2) / 2.0;
			if (model!=NULL) *model =  (model1 + model2) / 2.0;
			if (usable) return 1;
			else return -1;
		} else {
			return 0;
		}
	} else {
		return 0; // Unknown measurement
	}

	// Seek satellite
	i=epoch->satIndex[system][PRN];
	if (i!=-1) {
		// Satellite found
		rawValue = epoch->sat[i].meas[ind].value;
		j=epoch->satCSIndex[system][PRN];
		if ( measurement != NULL ) {
			align = epoch->cycleslip.preAlign[j][ind];
			conversionFactor = epoch->measOrder[system].conversionFactor[measType];
			*measurement =  (rawValue+align) * conversionFactor;
		}
		if ( model != NULL ) *model = epoch->sat[i].meas[ind].model;
		if ( measurement != NULL && rawValue == -1 ) {
			*measurement = -1;
			return 0;
		}
		if ( usable ) return 1;
		else return -1;
	
	}
	return 0;
}

/*****************************************************************************
 * Name        : getModelValue
 * Description : Get the model value of a particular satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * GNSSystem  system               I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType  measType       I  N/A  Measurement type
 * Returned value (double)         O  m    Model value (-1 invalid model)
 *****************************************************************************/
double getModelValue (TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType) {
	double	model;
	int		res;

	res = getMeasModelValue(epoch,system,PRN,measType,NULL,&model);

	if ( res == 1 ) {
		return model;
	} else {
		return -1;
	}
}

/*****************************************************************************
 * Name        : getMeasurementValue
 * Description : Get the model value of a particular satellite
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * GNSSystem  system               I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * MeasurementType  measType       I  N/A  Measurement type
 * Returned value (double)         O  m    Measurement value  (-1 invalid measurement)
 *****************************************************************************/
double getMeasurementValue (TEpoch *epoch, enum GNSSystem system, int PRN, enum MeasurementType measType) {
	double	measurement;
	int		res;

	res = getMeasModelValue(epoch,system,PRN,measType,&measurement,NULL);

	if (res==1) {
		return measurement;
	} else {
		return -1;
	}
}

/*****************************************************************************
 * Name        : SaveDataforGEOselection
 * Description : In SBAS mode, when option to select GEO with smallest
 *                 protection level is enabled, with every iteration that allows
 *                 to compute a solution, the data will be buffered so the
 *                 next iteration can be computed without losing the
 *                 values for current iteration
 *                 NOTE: The only value missing to be saved is the windup correction.
 *                 As this value is not computed in SBAS mode, it is not necessary.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TFilterSolution  *solution      IO N/A  Structure with the values of the protection levels and DOP 
 * int OverAlarmLimits             I  N/A  This variable is the output of function "calculateDOP"
 *                                           1 -> Solution under alarm limits
 *                                           2 -> Solution over  alarm limits
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void SaveDataforGEOselection (TEpoch *epoch, TFilterSolution *solution, int OverAlarmLimits, TOptions *options) {

	int 	i,j,k;
	int		pos,numsat;

	pos=epoch->Buffer_NumDataStored;
	numsat=epoch->numSatellites;

	for(k=0;k<epoch->Buffer_NumDataStored;k++) {
		//Check if current GEO is already stored. This can occur if current GEO had solution
		//and it could not change to any other GEO
		if (epoch->Buffer_GEOPRN[k]==epoch->currentGEOPRN) {
			for(i=0;i<numsat;i++) {
				printbufferMODEL[i][0]='\0';
				printbufferSBASIONO[i][0]='\0';
				printbufferSBASCORR[i][0]='\0';
				printbufferSBASVAR[i][0]='\0';
				printbufferSBASUNSEL[i][0]='\0';
				for(j=0;j<linesstoredSATSEL[i];j++) {
					printbufferSATSEL[i][j][0]='\0';
				}
				linesstoredSATSEL[i]=0;
			}
			return;
		}
	}

	epoch->Buffer_NumDataStored++; //Add 1 to the number of iterations stored

	//Allocate data for new iteration
	epoch->Buffer_numsatdiscardedSBAS=realloc(epoch->Buffer_numsatdiscardedSBAS,sizeof(int)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_usableSatellites=realloc(epoch->Buffer_usableSatellites,sizeof(int)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_HPL=realloc(epoch->Buffer_HPL,sizeof(double)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_VPL=realloc(epoch->Buffer_VPL,sizeof(double)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_HDOP=realloc(epoch->Buffer_HDOP,sizeof(double)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_VDOP=realloc(epoch->Buffer_VDOP,sizeof(double)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_PDOP=realloc(epoch->Buffer_PDOP,sizeof(double)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_GDOP=realloc(epoch->Buffer_GDOP,sizeof(double)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_TDOP=realloc(epoch->Buffer_TDOP,sizeof(double)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_SBASMode=realloc(epoch->Buffer_SBASMode,sizeof(int)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_GEOindex=realloc(epoch->Buffer_GEOindex,sizeof(int)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_GEOPRN=realloc(epoch->Buffer_GEOPRN,sizeof(int)*(epoch->Buffer_NumDataStored));

	epoch->Buffer_sat=realloc(epoch->Buffer_sat,sizeof(TSatellite*)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_sat[pos]=NULL;
	epoch->Buffer_sat[pos]=malloc(sizeof(TSatellite)*numsat);

	epoch->Buffer_NumSatSel=realloc(epoch->Buffer_NumSatSel,sizeof(int *)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_NumSatSel[pos]=NULL;
	epoch->Buffer_NumSatSel[pos]=malloc(sizeof(int)*numsat);

	epoch->Buffer_printMODEL=realloc(epoch->Buffer_printMODEL,sizeof(char *)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_printSBASIONO=realloc(epoch->Buffer_printSBASIONO,sizeof(char *)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_printSBASCORR=realloc(epoch->Buffer_printSBASCORR,sizeof(char *)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_printSBASVAR=realloc(epoch->Buffer_printSBASVAR,sizeof(char *)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_printSBASUNSEL=realloc(epoch->Buffer_printSBASUNSEL,sizeof(char *)*(epoch->Buffer_NumDataStored));
	epoch->Buffer_printSATSEL=realloc(epoch->Buffer_printSATSEL,sizeof(char *)*(epoch->Buffer_NumDataStored));

	epoch->Buffer_printMODEL[pos]=NULL;
	epoch->Buffer_printSBASIONO[pos]=NULL;
	epoch->Buffer_printSBASCORR[pos]=NULL;
	epoch->Buffer_printSBASVAR[pos]=NULL;
	epoch->Buffer_printSBASUNSEL[pos]=NULL;
	epoch->Buffer_printSATSEL[pos]=NULL;

	epoch->Buffer_printMODEL[pos]=malloc(sizeof(char *)*numsat);
	epoch->Buffer_printSBASIONO[pos]=malloc(sizeof(char *)*numsat);
	epoch->Buffer_printSBASCORR[pos]=malloc(sizeof(char *)*numsat);
	epoch->Buffer_printSBASVAR[pos]=malloc(sizeof(char *)*numsat);
	epoch->Buffer_printSBASUNSEL[pos]=malloc(sizeof(char *)*numsat);
	epoch->Buffer_printSATSEL[pos]=malloc(sizeof(char *)*numsat);

	for(i=0;i<numsat;i++) {
		epoch->Buffer_printMODEL[pos][i]=NULL;
		epoch->Buffer_printSBASIONO[pos][i]=NULL;
		epoch->Buffer_printSBASCORR[pos][i]=NULL;
		epoch->Buffer_printSBASVAR[pos][i]=NULL;
		epoch->Buffer_printSBASUNSEL[pos][i]=NULL;
		epoch->Buffer_printSATSEL[pos][i]=NULL;

		epoch->Buffer_printMODEL[pos][i]=malloc(sizeof(char)*MAX_INPUT_LINE);
		epoch->Buffer_printSBASIONO[pos][i]=malloc(sizeof(char)*MAX_INPUT_LINE);
		epoch->Buffer_printSBASCORR[pos][i]=malloc(sizeof(char)*MAX_INPUT_LINE);
		epoch->Buffer_printSBASVAR[pos][i]=malloc(sizeof(char)*MAX_INPUT_LINE);
		epoch->Buffer_printSBASUNSEL[pos][i]=malloc(sizeof(char)*MAX_INPUT_LINE);
		epoch->Buffer_printSATSEL[pos][i]=malloc(sizeof(char *)*linesstoredSATSEL[i]);

		for(j=0;j<linesstoredSATSEL[i];j++) {
			epoch->Buffer_printSATSEL[pos][i][j]=NULL;
			epoch->Buffer_printSATSEL[pos][i][j]=malloc(sizeof(char)*MAX_INPUT_LINE);
		}
	}

	//Save data
	epoch->Buffer_numsatdiscardedSBAS[pos]=epoch->numsatdiscardedSBAS;
	epoch->Buffer_usableSatellites[pos]=epoch->usableSatellites;
	epoch->Buffer_HPL[pos]=solution->HPL;
	epoch->Buffer_VPL[pos]=solution->VPL;
	epoch->Buffer_HDOP[pos]=solution->HDOP;
	epoch->Buffer_VDOP[pos]=solution->VDOP;
	epoch->Buffer_PDOP[pos]=solution->PDOP;
	epoch->Buffer_GDOP[pos]=solution->GDOP;
	epoch->Buffer_TDOP[pos]=solution->TDOP;
	epoch->Buffer_SBASMode[pos]=options->precisionapproach;
	epoch->Buffer_GEOindex[pos]=options->GEOindex;
	epoch->Buffer_GEOPRN[pos]=epoch->currentGEOPRN;


	for(i=0;i<numsat;i++) {

		epoch->Buffer_NumSatSel[pos][i]=linesstoredSATSEL[i];

		memcpy(&epoch->Buffer_sat[pos][i],&epoch->sat[i],sizeof(TSatellite));

		strcpy(epoch->Buffer_printMODEL[pos][i],printbufferMODEL[i]);
		strcpy(epoch->Buffer_printSBASIONO[pos][i],printbufferSBASIONO[i]);
		strcpy(epoch->Buffer_printSBASCORR[pos][i],printbufferSBASCORR[i]);
		strcpy(epoch->Buffer_printSBASVAR[pos][i],printbufferSBASVAR[i]);
		strcpy(epoch->Buffer_printSBASUNSEL[pos][i],printbufferSBASUNSEL[i]);

		//Empty printbuffers
		printbufferMODEL[i][0]='\0';
		printbufferSBASIONO[i][0]='\0';
		printbufferSBASCORR[i][0]='\0';
		printbufferSBASVAR[i][0]='\0';
		printbufferSBASUNSEL[i][0]='\0';

		for(j=0;j<linesstoredSATSEL[i];j++) {
			strcpy(epoch->Buffer_printSATSEL[pos][i][j],printbufferSATSEL[i][j]);
			//Empty printbufferSATSEL
			printbufferSATSEL[i][j][0]='\0';
		}
		linesstoredSATSEL[i]=0;
	}

	//Save indexes if values are below the alarm limits and are in PA mode
	if (OverAlarmLimits==1 && options->precisionapproach==PAMODE) {
		epoch->Buffer_PosDataStoredUnderAlarmLimitsPA=realloc(epoch->Buffer_PosDataStoredUnderAlarmLimitsPA,sizeof(int)*(epoch->Buffer_NumDataStoredUnderAlarmLimitsPA+1));
		epoch->Buffer_PosDataStoredUnderAlarmLimitsPA[epoch->Buffer_NumDataStoredUnderAlarmLimitsPA]=pos;
		epoch->Buffer_NumDataStoredUnderAlarmLimitsPA++;
	}

	if (options->precisionapproach==PAMODE) {
		//Save indexes if solution is in PAMODE
		epoch->Buffer_PosDataStoredPA=realloc(epoch->Buffer_PosDataStoredPA,sizeof(int)*(epoch->Buffer_NumDataStoredPA+1));
		epoch->Buffer_PosDataStoredPA[epoch->Buffer_NumDataStoredPA]=pos;
		epoch->Buffer_NumDataStoredPA++;	
	} else {
		//Save indexes if solution is in NPAMODE
		epoch->Buffer_PosDataStoredNPA=realloc(epoch->Buffer_PosDataStoredNPA,sizeof(int)*(epoch->Buffer_NumDataStoredNPA+1));
		epoch->Buffer_PosDataStoredNPA[epoch->Buffer_NumDataStoredNPA]=pos;
		epoch->Buffer_NumDataStoredNPA++;	
	}
}

/*****************************************************************************
 * Name        : SelectBestGEO
 * Description : In SBAS mode, when option to select GEO with smallest
 *                 protection level is enabled and all possible iterations
 *                 have been computed and buffered, this function will select
 *                 the one with smallest protection levels.
 *                 The criteria for selection is:
 *                 1 - Look for the iterations whose protection levels are under 
 *                      the alarm limits and in PA mode, and select the iteration
 *                      with smallest protection levels
 *
 *                 2 - If no iteration has protection levels under the alarm 
 *                      limits, select the iteration in PA mode with smallest 
 *                      protection levels
 *
 *                 3 - If no iteration in PA mode available, select the iteration
 *                      in NPA mode with smallest protection level
 *
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  IO N/A  TEpoch structure
 * TFilterSolution  *solution      IO N/A  Structure with the values of the protection levels and DOP 
 * TOptions  *options              IO N/A  TOptions structure
 *****************************************************************************/
void SelectBestGEO (TEpoch *epoch, TFilterSolution *solution, TOptions *options) {

	int 	i,j,k;
	int		smallestSumPos;
	int		numsat;
	int		prevGEOPRN;
	double	sumPL;
	double	smallestSumPL=9999999999999.;

	numsat=epoch->numSatellites;
	prevGEOPRN=epoch->currentGEOPRN;

	//Look first for iterations with smallest protection levels
	if (epoch->Buffer_NumDataStoredUnderAlarmLimitsPA>0) {
		for(i=0;i<epoch->Buffer_NumDataStoredUnderAlarmLimitsPA;i++) {
			j=epoch->Buffer_PosDataStoredUnderAlarmLimitsPA[i];
			sumPL=epoch->Buffer_HPL[j]+epoch->Buffer_VPL[j];
			if(sumPL<smallestSumPL) {
				smallestSumPL=sumPL;
				smallestSumPos=j;
			}
		}
	} else if (epoch->Buffer_NumDataStoredPA>0) {
		//No iteration with protection levels under alarm limits and in PA mode
		for(i=0;i<epoch->Buffer_NumDataStoredPA;i++) {
			j=epoch->Buffer_PosDataStoredPA[i];
			sumPL=epoch->Buffer_HPL[j]+epoch->Buffer_VPL[j];
			if(sumPL<smallestSumPL) {
				smallestSumPL=sumPL;
				smallestSumPos=j;
			}
		}
	} else {
		//No iteration in PA mode
		for(i=0;i<epoch->Buffer_NumDataStoredNPA;i++) {
			j=epoch->Buffer_PosDataStoredNPA[i];
			sumPL=epoch->Buffer_HPL[j]+epoch->Buffer_VPL[j];
			if(sumPL<smallestSumPL) {
				smallestSumPL=sumPL;
				smallestSumPos=j;
			}
		}
	}

	//Iteration selected. Recover data
	epoch->numsatdiscardedSBAS=epoch->Buffer_numsatdiscardedSBAS[smallestSumPos];
	epoch->usableSatellites=epoch->Buffer_usableSatellites[smallestSumPos];
	solution->HPL=epoch->Buffer_HPL[smallestSumPos];
	solution->VPL=epoch->Buffer_VPL[smallestSumPos];
	solution->HDOP=epoch->Buffer_HDOP[smallestSumPos];
	solution->VDOP=epoch->Buffer_VDOP[smallestSumPos];
	solution->PDOP=epoch->Buffer_PDOP[smallestSumPos];
	solution->GDOP=epoch->Buffer_GDOP[smallestSumPos];
	solution->TDOP=epoch->Buffer_TDOP[smallestSumPos];
	options->precisionapproach=epoch->Buffer_SBASMode[smallestSumPos];
	options->GEOindex=epoch->Buffer_GEOindex[smallestSumPos];
	epoch->currentGEOPRN=epoch->Buffer_GEOPRN[smallestSumPos];

	for(i=0;i<numsat;i++) {
		memcpy(&epoch->sat[i],&epoch->Buffer_sat[smallestSumPos][i],sizeof(TSatellite));
	}

	//Print buffered messages. The selected iteration should be the last one.
	//The other iterations will be printed with an '*' if option SBASUNUSED is enabled
	printBuffersBestGEO(epoch,smallestSumPos,prevGEOPRN,options);

	
	//Free memory
	for(i=0;i<epoch->Buffer_NumDataStored;i++) {
		for(j=0;j<epoch->numSatellites;j++) {
			for(k=0;k<epoch->Buffer_NumSatSel[i][j];k++) {
				free(epoch->Buffer_printSATSEL[i][j][k]);
			}
			free(epoch->Buffer_printMODEL[i][j]);
			free(epoch->Buffer_printSBASIONO[i][j]);
			free(epoch->Buffer_printSBASCORR[i][j]);
			free(epoch->Buffer_printSBASVAR[i][j]);
			free(epoch->Buffer_printSBASUNSEL[i][j]);
			free(epoch->Buffer_printSATSEL[i][j]);
		}
		free(epoch->Buffer_sat[i]);
		free(epoch->Buffer_NumSatSel[i]);
		free(epoch->Buffer_printMODEL[i]);
		free(epoch->Buffer_printSBASIONO[i]);
		free(epoch->Buffer_printSBASCORR[i]);
		free(epoch->Buffer_printSBASVAR[i]);
		free(epoch->Buffer_printSBASUNSEL[i]);
		free(epoch->Buffer_printSATSEL[i]);
	}

	free(epoch->Buffer_numsatdiscardedSBAS);
	free(epoch->Buffer_usableSatellites);
	free(epoch->Buffer_HPL);
	free(epoch->Buffer_VPL);
	free(epoch->Buffer_HDOP);
	free(epoch->Buffer_VDOP);
	free(epoch->Buffer_PDOP);
	free(epoch->Buffer_GDOP);
	free(epoch->Buffer_TDOP);
	free(epoch->Buffer_SBASMode);
	free(epoch->Buffer_GEOindex);
	free(epoch->Buffer_GEOPRN);
	free(epoch->Buffer_sat);
	free(epoch->Buffer_NumSatSel);
	free(epoch->Buffer_printMODEL);
	free(epoch->Buffer_printSBASIONO);
	free(epoch->Buffer_printSBASCORR);
	free(epoch->Buffer_printSBASVAR);
	free(epoch->Buffer_printSBASUNSEL);
	free(epoch->Buffer_printSATSEL);
	//Set all pointers to NULL
	epoch->Buffer_numsatdiscardedSBAS=NULL;
	epoch->Buffer_usableSatellites=NULL;
	epoch->Buffer_HPL=NULL;
	epoch->Buffer_VPL=NULL;
	epoch->Buffer_HDOP=NULL;
	epoch->Buffer_VDOP=NULL;
	epoch->Buffer_PDOP=NULL;
	epoch->Buffer_GDOP=NULL;
	epoch->Buffer_TDOP=NULL;
	epoch->Buffer_SBASMode=NULL;
	epoch->Buffer_GEOindex=NULL;
	epoch->Buffer_GEOPRN=NULL;
	epoch->Buffer_sat=NULL;
	epoch->Buffer_NumSatSel=NULL;
	epoch->Buffer_printMODEL=NULL;
	epoch->Buffer_printSBASIONO=NULL;
	epoch->Buffer_printSBASCORR=NULL;
	epoch->Buffer_printSBASVAR=NULL;
	epoch->Buffer_printSBASUNSEL=NULL;
	epoch->Buffer_printSATSEL=NULL;
	//Set counters to 0
	epoch->Buffer_NumDataStored=0;
	epoch->Buffer_NumDataStoredUnderAlarmLimitsPA=0;
	epoch->Buffer_NumDataStoredPA=0;
	epoch->Buffer_NumDataStoredNPA=0;
}

/*****************************************************************************
 * Name        : getProductsFirstEpochBRDC
 * Description : Return the first epoch TTime in a TGNSSproducts broadcast structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of first observation in file
 *****************************************************************************/
TTime getProductsFirstEpochBRDC (TGNSSproducts *products){
	TTime	t;
	int		i;
	
	t.MJDN = 0;
	t.SoD = 0.;
	
	if (products->type!=BRDC) return t;
	
	// Initialise with the first transmission time of first satellite
	memcpy(&t,&products->BRDC->block[0][0].TtransTime,sizeof(TTime));
	// Get lower transmission time
	for (i=0;i<products->BRDC->numsats;i++) {
		if (tdiff(&products->BRDC->block[i][0].TtransTime,&t)<0) {
			memcpy(&t,&products->BRDC->block[i][0].TtransTime,sizeof(TTime));
		}
	}
	return t;
}

/*****************************************************************************
 * Name        : getProductsFirstEpochSP3
 * Description : Return the first epoch TTime in a TGNSSproducts SP3 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of first observation in file
 *****************************************************************************/
TTime getProductsFirstEpochSP3 (TGNSSproducts *products) {
	TTime	t;
	
	t.MJDN = 0;
	t.SoD = 0;
	
	if (products->type!=SP3) return t;
	
	return products->SP3->orbits.startTime; // Header start time
}

/*****************************************************************************
 * Name        : getProductsFirstEpoch
 * Description : Return the first epoch TTime in a TGNSSproducts structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of first observation in file
 *****************************************************************************/
TTime getProductsFirstEpoch (TGNSSproducts *products) {
	if ( products->type == SP3 ) {
		return getProductsFirstEpochSP3(products);
	} else { // (products->type == BRDC)
		return getProductsFirstEpochBRDC(products);
	}
}

/*****************************************************************************
 * Name        : getProductsLastEpochBRDC
 * Description : Return the last epoch TTime in a TGNSSproducts broadcast structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of last observation in file
 *****************************************************************************/
TTime getProductsLastEpochBRDC (TGNSSproducts *products) {
	TTime	t;
	int		i;
	
	t.MJDN = 0;
	t.SoD = 0;
	
	if (products->type!=BRDC) return t;
	
	// Initialise with the last toe time of first satellite
	memcpy(&t,&products->BRDC->block[0][products->BRDC->numblocks[0]-1].Ttoe,sizeof(TTime));
	// Get lower transmission time
	for (i=0;i<products->BRDC->numsats;i++) {
		if (tdiff(&products->BRDC->block[i][products->BRDC->numblocks[i]-1].Ttoe,&t)>0) {
			memcpy(&t,&products->BRDC->block[i][products->BRDC->numblocks[i]-1].Ttoe,sizeof(TTime));
		}
	}
	// Broadcast can be valid for longer periods (up to 2 hours later from Toe)
	t = tdadd(&t,7200);
	return t;
}

/*****************************************************************************
 * Name        : getProductsLastEpochSP3
 * Description : Return the last epoch TTime in a TGNSSproducts SP3 structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of last observation in file
 *****************************************************************************/
TTime getProductsLastEpochSP3 (TGNSSproducts *products) {
	TTime	t;
	
	t.MJDN = 0;
	t.SoD = 0;
	
	if (products->type!=SP3) return t;

	//return products->SP3->orbits.block[0][products->SP3->orbits.numRecords - 1].t;
	return products->SP3->orbits.endTime;
}

/*****************************************************************************
 * Name        : getProductsLastEpoch
 * Description : Return the last epoch TTime in a TGNSSproducts structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 * Returned value (TTime)          O  N/A  Time of last observation in file
 *****************************************************************************/
TTime getProductsLastEpoch (TGNSSproducts *products) {
	if ( products->type == SP3 ) {
		return getProductsLastEpochSP3(products);
	} else { // (products->type == BRDC)
		return getProductsLastEpochBRDC(products);
	}
}

/*****************************************************************************
 * Name        : selectBRDCBlock
 * Description : Select the BRDC block with the following conditions
 *                  - Transmission time equal or lower than reference time
 *                  - Closest toe to reference time
 *                  - If SBAS IOD is not -1, select the block with that IODE
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TBRDCproducts  *products        I  N/A  TBRDCproducts structure used as base to return PVT
 * TTime  *t                       I  N/A  Reference time used
 * GNSSystem  GNSS                 I  N/A  GNSS system of the satellite
 * int  PRN                        I  N/A  PRN identifier of the satellite
 * TOptions  *options              I  N/A  TOptions structure
 * int SBASIOD                     I  N/A  SBAS IOD for long term corrections                                       
 * Returned value (TBRDCblock*)    O  N/A  Pointer to the corresponding block
 *****************************************************************************/
TBRDCblock *selectBRDCblock (TBRDCproducts *products, TTime *t, enum GNSSystem GNSS, int PRN, int SBASIOD, TOptions *options) {
	int			ind,indPast;
	int			i,j;
	int			limit;
	int			check;		
	int			selected = -1,timeSelected;
	double		diff;
	double		diffTransmitted;
	double		diffToe;
	double		diffToemax; 
	double		diffLastTransmitted = 9e9;
	TTime       tGLO;
	TBRDCblock	*block;


	ind = products->index[GNSS][PRN];
	indPast = products->indexPast[GNSS][PRN];

	if (ind==-1 && indPast==-1) return NULL;

	for(j=0;j<2;j++) {
		if (indPast==-1 && j==0) continue;
		if (ind==-1 && j==1) continue;
		if(j==0) limit=products->numblocksPast[indPast];
		else limit=products->numblocks[ind];

		for (i=0;i<limit;i++) {
			check=0;
			if(j==0)  block = &products->blockPast[indPast][i];
			else block = &products->block[ind][i];
			
			if (SBASIOD!=-1 && block->IODE!=SBASIOD) {continue;}
			diffTransmitted = tdiff(t,&block->TtransTime);
			if(GNSS==GLONASS) {
				tGLO.MJDN=block->Ttoe.MJDN;
				tGLO.SoD=block->Ttoe.SoD;
				//Transform from UTC to GPS time
				tGLO.SoD-=products->AT_LS;
				if(tGLO.SoD<0) {
					tGLO.SoD+=86400.;
					tGLO.MJDN--;
				}
				diffToe = tdiff(t,&tGLO);
				//Tranmission time of message of GLONASS broadcast block is directly saved in GPS time
			} else {
				diffToe = tdiff(t,&block->Ttoe);
			}
			diffToemax = 0.5 * 3600 * block->fitInterval; // Update diffToemax

			if (options->brdcBlockTransTime==0) {
				//Do not check that transmission time of message is equal or before current time
				if (fabs(diffTransmitted)<fabs(diffLastTransmitted) && diffToe<=diffToemax && diffToe>=-diffToemax) {
					check=1;
				}
			} else {
				//Check that transmission time of message is equal or before current time
				if (diffTransmitted>=0 && diffTransmitted<diffLastTransmitted && diffToe<=diffToemax && diffToe>=-diffToemax) {
					check=1;
				}
			}
			if (check) {
	//			if (block->SVhealth==0 || options->satelliteHealth==0) {
	//				diffLastTransmitted = diffTransmitted;
	//				selected = i;
	//			}

				// Update following ESA/EPO suggestion: 
				if (options->satelliteHealth==0) {
					diffLastTransmitted = diffTransmitted;
					selected = i;
					timeSelected=j;
				} else {
					// Last Transmitted Healthy Ephemerides
					if (GNSS==Galileo && block->SVhealth==0 && block->SISASignal!=-1) { 
						diffLastTransmitted = diffTransmitted;
						selected = i;
						timeSelected=j;
					} else if (GNSS!=Galileo && block->SVhealth==0) {
						diffLastTransmitted = diffTransmitted;
						selected = i;
						timeSelected=j;
					} else {// Last transmitted unhealthy block 
						diffLastTransmitted = diffTransmitted;
						selected = -1;
					}
				}
			}
		}
	}

	if (selected==-1) return NULL;

	// Check if selected is inside the fit interval
	if(timeSelected==0) {
		block = &products->blockPast[indPast][selected];
	} else {
		block = &products->block[ind][selected];
	}
	if(GNSS==GLONASS) {
		tGLO.MJDN=block->Ttoe.MJDN;
		tGLO.SoD=block->Ttoe.SoD;
		//Transform from UTC to GPS time
		tGLO.SoD-=products->AT_LS;
		if(tGLO.SoD<0) {
			tGLO.SoD+=86400.;
			tGLO.MJDN--;
		}
		diff = tdiff(t,&tGLO);
		//Tranmission time of message of GLONASS broadcast block is directly saved in GPS time
	} else {
		diff = tdiff(t,&block->Ttoe);
	}
	diffToemax = 0.5*3600*block->fitInterval; // Update diffToemax
	if (diff<=diffToemax && diff>=-diffToemax) return block;
	else return NULL;
}

/*****************************************************************************
 * Name        : MJDN
 * Description : Get the modified julian day from YY/MM/DD
 *               MJDN starts at 17/11/1858 at 0h UTC (JD=2400000.5)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * struct tm  tm                   I  N/A  Date
 * Returned value (int)            O  N/A  Integer days in Modified Julian Date
 *****************************************************************************/
int MJDN (struct tm *tm) {
	int y,m;

	y = tm->tm_year + 1900;
	m = tm->tm_mon + 1;
	if (m<=2) {
		y--;
		m+=12;
	}

	return (int)(365.25*y) + (int)(30.6001*(m+1)) + tm->tm_mday - 679019;
}

/*****************************************************************************
 * Name        : MonthName
 * Description : Get the month name from the month number
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int month                       I  N/A  Month number (1-12)
 * char *monthname                 O  N/A  Name of the month
 *****************************************************************************/
void MonthName (int month, char *monthname) {
	const char monthnames[12][10]={"January","February","March","April","May","June","July","August","September","October","November","December"};
	
	strcpy(monthname,monthnames[month-1]);
}



/*****************************************************************************
 * Name        : t2tm
 * Description : Get the calendar time from a TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure
 * tm  *tm                         O  N/A  tm structure
 * double  *seconds                O  N/A  Seconds of minute in double format
 *****************************************************************************/
void t2tm (TTime *t, struct tm *tm, double *seconds) {
	int		year;
	double	doy;
	time_t	tt;
	
	t2doy(t,&year,&doy);
	tm->tm_year = year - 1900;
	tm->tm_mday = (int)(doy);
	tm->tm_mon = 0;
	tm->tm_hour = (int)(t->SoD/3600.);
	tm->tm_min = (int)(modulo(t->SoD,3600.)/60.);
	tm->tm_sec = (int)(modulo(t->SoD,60.));
	tm->tm_isdst = -1;
	*seconds =  modulo(t->SoD,60.);
	tt = mktime(tm);
	//localtime_r(&tt,tm);
	memcpy(tm,localtime(&tt),sizeof(time_t));
}

/*****************************************************************************
 * Name        : t2tmnolocal
 * Description : Get the calendar time from a TTime structure without using
 *                localtime function
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure
 * tm  *tm                         O  N/A  tm structure
 * double  *seconds                O  N/A  Seconds of minute in double format
 *****************************************************************************/
void t2tmnolocal (TTime *t, struct tm *tm, double *seconds) {
	int             year;
	int             day_of_month;
	int             month_days;
	int             month_number;
	double          doy;

	t2doy(t,&year,&doy);

	tday_of_month(t,&day_of_month, &month_days, &month_number);

	//IMPORTANT NOTE!! Read man page ctime ('man ctime') to understand how the system tm structure is defined
	tm->tm_year = year - 1900;	// The number of years since 1900.
	tm->tm_mday = day_of_month;	// The day of the month, in the range 1 to 31.
	tm->tm_yday = (int)doy -1; 	// The number of days since January 1, in the range 0 to 365
	tm->tm_mon = month_number-1;	// The number of months since January, in the range 0 to 11.
	tm->tm_hour = (int)(t->SoD/3600.); // The number of hours past midnight, in the range 0 to 23.
	tm->tm_min = (int)(modulo(t->SoD,3600.)/60.); // The number of minutes after the hour, in the range 0 to 59.
	tm->tm_sec = (int)(modulo(t->SoD,60.)); // The number of seconds after the minute, normally in the range 0 to 59, but can be up to 60 to allow for leap seconds.
	tm->tm_isdst = -1;	// A  flag that indicates whether daylight saving time is in effect at the time described.  The value is positive if
				//    daylight saving time is in effect, zero if it is not, and negative if the information is not available.
	*seconds =  modulo(t->SoD,60);
}

/*****************************************************************************
 * Name        : t2doy
 * Description : Get the Year/doy from a TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure
 * int  *year                      O  N/A  Year
 * double  *doy                    O  N/A  Day of year
 *****************************************************************************/
void t2doy (TTime *t, int *year, double *doy) {
	const int 	MJD_1980 = 44239;
	int     	day;    
	int     	ref = 1980; 
	int     	day_aux;
	int     	iy;
	
	day = t->MJDN - MJD_1980;
	
	day_aux = modulo(day,1461.);
	if (day_aux>365) {
		day_aux -= 366; 
		*doy = modulo (day_aux,365.) + 1;
		iy = (int)((double)(day_aux)/365.) + 1;
	} else {
		*doy = (double)(day_aux + 1);
		iy = 0; 
	}
	*year = ref + 4*(int)((double)(day)/1461.) + iy;
	*doy += t->SoD/86400.;
}

/*****************************************************************************
 * Name        : date2doy
 * Description : Get the Year/doy from date year, month and day
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  year                       I  y    Year (must be of 4 digits)
 * int  month                      I  m    Month
 * int  day                        I  d    Day of month
 * Return value (int)              O  d    Day of Year
 *****************************************************************************/
int date2doy (int year, int month, int day) {
	int i, doy;
	int daysmonth[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	if ( year < 1000 ) return -1;

	if ( (( year%4 == 0 ) && ( year%100 != 0 )) || year%400 == 0 )
		daysmonth[2] += 1;

	doy = 0;
	for ( i=0;i<month;i++ )
		doy += daysmonth[i];
	doy += day;

	return doy;
}

/*****************************************************************************
 * Name        : doy2date
 * Description : Get the date year, month and day from Year/doy
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  year                       I  y    Year (4 digits)
 * int  doy                        I  d    Day of Year
 * int  *month                     O  m    Month
 * int  *day                       O  d    Day of month
 *****************************************************************************/
void doy2date (int year, int doy, int *month, int *day) {
	int i = -1;
	int res;
	int daysmonth[13] = {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	if ( doy > 366 ) return;

	if ( (( year%4 == 0 ) && ( year%100 != 0 )) || year%400 == 0 )
		daysmonth[2] += 1;

	res = doy;
	while ( res > 0 ) {
		i += 1;
		res -= daysmonth[i];
	}

	*month = i;
	*day = daysmonth[i] + res;
}

/*****************************************************************************
 * Name        : t2doystr
 * Description : Get a string with the Year/DoY/SoD from a TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure
 * Returned value (char*)          O  N/A  String with Year / Doy / SoD (length=17)
 *****************************************************************************/
char *t2doystr (TTime *t) {
	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[100];
	#pragma omp threadprivate(str)
	int			year;
	double		doy;
	
	t2doy(t,&year,&doy);
	sprintf(str,"%04d %03d %8.2f",year,(int)(doy),(doy-(int)(doy))*86400.0);

	return str;
}

/*****************************************************************************
 * Name        : gpsws2ttime
 * Description : Transforms between GPSweek and Seconds of Week to TTime structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  GPSweek                    I  N/A  GPS Week
 * double  SoW                     I  s    Seconds of week
 * Returned value (TTime)          O  N/A  TTime structure
 *****************************************************************************/
TTime gpsws2ttime (int GPSweek, double SoW) {
	TTime		t;
	int			DoW;

	// GPSweek = 0 and SoW = 0 corresponds to MJDN = 44244
	// Check if Seconds of week is negative (meaning that is from previous week)
	if(SoW<0) {
		SoW+=604800.;
		GPSweek--;
	} else if (SoW>=604800.) {
		SoW-=604800.;
		GPSweek++;
	}
	DoW = (int)(SoW/86400.0);
	t.MJDN = (int)(44244 + GPSweek * 7.0 + DoW);
	t.SoD = SoW - DoW*86400.0;

	return t;
}

/*****************************************************************************
 * Name        : yeardoy2MJDN
 * Description : Get the modified julian day from Year/DoY/SoD
 *               MJDN starts at 17/11/1858 at 0h UTC (JD=2400000.5)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  year                       I  N/A  Year
 * int  doy                        I  N/A  Day of Year
 * double  sod                     I  s    Seconds of day
 * Returned value (double)         O  N/A  MJDN
 ****************************************************************************/
double yeardoy2MJDN (int year, int doy, int sod) {
	double tmp;

	tmp = (int)(365.25 * (year-1)) + 428 + doy;
	tmp = tmp + 1720981.5 + (double)(sod)/86400. - 2400000.5;

	return tmp;
}


/*****************************************************************************
 * Name        : ttime2gpswsnoroll
 * Description : Transforms between TTime structure to GPSweek (without rollover)
 *                and Seconds of Week 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime *t                        I  N/A  TTime structure
 * int  GPSweek                    O  N/A  GPS Week
 * double  SoW                     O  s    Seconds of week
 * Returned value                  -  N/A  -
 *****************************************************************************/
void ttime2gpswsnoroll (TTime *t, int *GPSweek, double *SoW) {
	int gpsday,weekday;

	//44244 comes from adding 2400000.5 to convert to Julian Date and then substracting 2444244.5 to get the first day of GPS time
	gpsday = t->MJDN - 44244;
	weekday = gpsday%7;

	*GPSweek = (int)(gpsday - (gpsday%7)) / 7;
	*SoW = (double)(weekday) * 86400. + t->SoD;
}

/*****************************************************************************
 * Name        : cal2t
 * Description : Transforms between Calendar time (year, month, day of month, 
 *               hour, minute and second) to TTime structure
 *               MJDN starts at 17/11/1858 at 0h UTC (JD=2400000.5)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  year                       I  y    Year (can be of 4 or 2 digits)
 * int  month                      I  m    Month
 * int  day                        I  d    Day of month
 * int  hour                       I  h    Hours of day
 * int  minute                     I  m    Minutes in hour
 * double  second                  I  s    Seconds of minute
 * Returned value (TTime)          O  N/A  TTime structure
 *****************************************************************************/
TTime cal2t (int year, int month, int day, int hour, int minute, double second) {
	struct tm	tm;
	TTime		t;
	
	if ( year < 100 ) { // 2 digits
		if ( year <= 70 ) { // 20XX
			tm.tm_year = year + 100;
		} else { // 19XX
			tm.tm_year = year;
		}
	} else { // 4 digits
		tm.tm_year = year - 1900;
	}
	
	tm.tm_mon = month - 1;
	tm.tm_mday = day;
	tm.tm_hour = hour;
	tm.tm_min = minute;
	tm.tm_sec = (int)second;
	t.MJDN = MJDN (&tm);
	t.SoD = tm.tm_hour*3600 + tm.tm_min*60 + second;
	
	return t;
}

/*****************************************************************************
 * Name        : tday_of_month
 * Description : Compute day of month from Ttime structure
 *               
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       I  N/A  TTime structure 
 * int    *day_of_month            O  N/A  Day of the month (1..31)
 * int    *month_days              O  N/A  Number of days of the month
 * int    *month_number            O  N/A  Month number (1..12)
 * Returned value                  O  -    -
 *****************************************************************************/
void tday_of_month (TTime *t, int *day_of_month, int *month_days, int *month_number) {
	int			year;
	int			leapyear = 0;
	int			daysum = 0, prevdaysum = 0;
	double		doy;

	// The following array has the number of days for each month, for a normal year and for a leap year
	const int number_days_month[2][12]={{31,28,31,30,31,30,31,31,30,31,30,31},{31,29,31,30,31,30,31,31,30,31,30,31}};

	t2doy(t,&year,&doy);

	// Check if we are in a leap year
	if ( year%400 == 0 )		leapyear = 1;
	else if ( year%100 == 0 )	leapyear = 0;
	else if ( year%4 == 0 )		leapyear = 1;
	else						leapyear = 0;
 

	// Check for the month we are currently in 
	*month_number = 0;
	*month_days = 0;
	while ( *month_days == 0 ) {
		daysum += number_days_month[leapyear][*month_number];
		if ( (int)doy <= daysum ) {
			*month_days = number_days_month[leapyear][*month_number];
			*day_of_month = (int)doy - prevdaysum;
		}
		*month_number = *month_number + 1;
		prevdaysum = daysum;	
	}
}

/*****************************************************************************
 * Name        : getcurrentsystemtime
 * Description : Get the system time and date
 *               
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *t                       O  N/A  TTime structure 
 * tm  *tm                         O  N/A  tm structure
 * Returned value                  O  -    -
 *****************************************************************************/
void getcurrentsystemtime (struct tm *tm) {
	time_t tt;

	tt = time(NULL);
	memcpy(tm,localtime(&tt),sizeof(struct tm));
}


/*****************************************************************************
 * Name        : tdiff
 * Description : Difference between two internal time structures in seconds, 
 *               equivalent to t1-t2
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *a                       I  N/A  Time 1
 * TTime  *b                       I  N/A  Time 2
 * Returned value (double)         O  s    Seconds between t1 and t2
 *****************************************************************************/
double tdiff (TTime *a, TTime *b) {
	double aux;

	aux = ( a->MJDN - b->MJDN ) * 86400.0 + ( a->SoD - b->SoD );
	// Sanity check to avoid rounding problems
	if ( fabs(aux) < 1e-6 ) aux = 0.0;

	return aux;
}

/*****************************************************************************
 * Name        : tsec
 * Description : Return MJDN + SoD in seconds
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *a                       I  N/A  Time 1
 * Returned value (double)         O  s    Seconds of t1
 *****************************************************************************/
double tsec (TTime *a) {
	return (a->MJDN) * 86400.0 + a->SoD;
}

/*****************************************************************************
 * Name        : tdadd
 * Description : Add a double (in seconds) from a TTime structure 
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime  *a                       I  N/A  Reference time
 * double  b                       I  s    Time to add
 * Returned value (TTime)          O  N/A  a+b
 *****************************************************************************/
TTime tdadd (TTime *a, double b) {
	TTime		t;
	int			MJDNdiff;

	t.MJDN = a->MJDN;
	t.SoD = a->SoD + b;

	MJDNdiff = (int)(t.SoD/86400);

	t.MJDN += MJDNdiff;
	t.SoD -= MJDNdiff*86400;

	if (t.SoD<0) {
		t.MJDN--;
		t.SoD += 86400;
	}

	return t;
}

/*****************************************************************************
 * Name        : Gal_loc_month
 * Description : Locate months around a certain day (Galileo's Matlab function)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double doy                      I  N/A  Day of year [1..365.25]
 * int *im                         O  N/A  Indexes of the month [0 .. 11]
 * double *id                      O  N/A  Days of half month [-14.5 .. 380.75]
 * Returned value (int)            O  N/A  0 -> DoY value found in dom vector
 *                                         1 -> DoY value not found in dom vector
 *****************************************************************************/
int Gal_loc_month (double doy, int *im, double *id) {

	int i;
	int dom_positions = 14;
	double dom[14] = {-14.5,15.5,45.125,74.75,105.25,135.75,166.25,196.75,227.75,258.25,288.75,319.25,349.75,380.75};
	
	// Initialize to 0. If im[1]==9999 then the value of doy was found in dom vector
	im[0] = 0;
	im[1] = 9999;
	id[0] = id[1] = 0;
 
	for(i=0;i<dom_positions;i++) {
		if(dom[i] == doy) {
			im[0] = i;
			id[0] = dom[i];
			// Need to substract 1 because the matrix index starts at 0 (in Matlab starts at 1)
			im[0]--; 
			return 0;
		} else if (dom[i] > doy) {
			im[0] = i-1;
			im[1] = i;
			id[0] = dom[im[0]];
			id[1] = dom[im[1]];
			break;
		}
	}

	// DoY value not found in dom vector
	if ( im[0] == 0 ) {
		im[0] = 12;
	} else if ( im[0] == 12 ) {
		im[1] = 1;
	}

	// Need to substract 1 because the matrix index starts at 0 (in Matlab starts at 1)
	im[0]--;
	im[1]--;

	return 1;
}

/*****************************************************************************
 * Name        : XYZ2Spherical
 * Description : Transforms from XYZ to Spherical
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *positionXYZ            I  m    Station position in XYZ
 * double  *positionSph            O  rd/m Station position in Spherical
 *                                         0=>Latitude  1=>Longitude   2=>Radius
 *****************************************************************************/
void XYZ2Spherical (double  *positionXYZ, double *positionSph) {
	
	if (positionXYZ[2]==0.) {
		positionSph[0]=0.;
	} else {
		positionSph[0]=asin(positionXYZ[2]/(sqrt(positionXYZ[0]*positionXYZ[0]+positionXYZ[1]*positionXYZ[1]+positionXYZ[2]*positionXYZ[2])));
	}
    positionSph[1]=atan2(positionXYZ[1],positionXYZ[0]);
    positionSph[2]=sqrt(positionXYZ[0]*positionXYZ[0]+positionXYZ[1]*positionXYZ[1]+positionXYZ[2]*positionXYZ[2]);
}

/*****************************************************************************
 * Name        : Spherical2XYZ
 * Description : Transforms from Spherical to XYZ
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *positionSph            I  rd/m Station position in Spherical
 *                                         0=>Latitude  1=>Longitude   2=>Radius
 * double  *positionXYZ            O  m    Station position in XYZ
 *****************************************************************************/
void Spherical2XYZ (double  *positionSph, double *positionXYZ) {

	positionXYZ[0]=positionSph[2]*cos(positionSph[0])*cos(positionSph[1]);
    positionXYZ[1]=positionSph[2]*cos(positionSph[0])*sin(positionSph[1]);
    positionXYZ[2]=positionSph[2]*sin(positionSph[0]);
}

/*****************************************************************************
 * Name        : XYZ2NEU
 * Description : Transforms from XYZ to NEU
 *                  Z axis => Zenith (Up)  [m]
 *                  X axis => North                            - Latitude  [rad]
 *                  Y axis => Completing the system (East)     - Longitude [rad]
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *positionXYZ            I  m    Station position in XYZ
 * double  *positionNEU            O  rd/m Station position in NEU
 *                                         0=>N(Lat)  1=>E(Lon)   2=>U
 *****************************************************************************/
void XYZ2NEU (double *positionXYZ, double *positionNEU) {
	const double	tol = 1e-11;
	const double	a = EARTH_RADIUS;
	//const double	a = SBAS_EARTH_RADIUS;
	const double	f = 1/298.257223563;
	const double	b = a*(1-f);
	const double	e2 = (a*a-b*b)/(a*a);
	double			p;
	double			longitude;
	double			latitude,latitudePrev;
	double			xn;
	double			heightOverElipsoide;
	// WGS84 parameters (in meters)

	longitude = atan2(positionXYZ[1],positionXYZ[0]);
	p = sqrt(positionXYZ[0]*positionXYZ[0]+positionXYZ[1]*positionXYZ[1]);
	if (p==0) p=0.00001;
	latitude = atan((positionXYZ[2]/p)/(1-e2));

	latitudePrev = latitude + tol*2; // To overbound the following while
	while (fabs(latitude-latitudePrev) > tol) {
		latitudePrev = latitude;
		xn = a*a/sqrt((a*cos(latitude))*(a*cos(latitude)) + (b*sin(latitude))*(b*sin(latitude)));
		heightOverElipsoide = p/cos(latitude)-xn;
		if ((xn+heightOverElipsoide)!=0)
			latitude = atan((positionXYZ[2]/p)/(1-e2*(xn/(xn+heightOverElipsoide))));
		else
			latitude = 0;
	}
	
	positionNEU[0] = latitude;
	positionNEU[1] = longitude;
	positionNEU[2] = heightOverElipsoide;
}

/*****************************************************************************
 * Name        : NEU2XYZ
 * Description : Transforms from NEU to XYZ
 *                  Z axis => Zenith (Up)  [m]
 *                  X axis => North                            - Latitude  [rad]
 *                  Y axis => Completing the system (East)     - Longitude [rad]
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  *positionNEU            I  rd/m Station position in NEU
 *                                         0=>N(Lat)  1=>E(Lon)   2=>U
 * double  *positionXYZ            O  m    Station position in XYZ
 *****************************************************************************/
void NEU2XYZ (double *positionNEU, double *positionXYZ) {
    const double    a = EARTH_RADIUS;
    //const double  a = SBAS_EARTH_RADIUS;
    const double    f = 1/298.257223563;
    const double    b = a*(1-f);
    const double    e2 = (a*a-b*b)/(a*a);
    double          N;
    // WGS84 parameters (in meters)

	N=a/sqrt(1-e2*sin(positionNEU[0])*sin(positionNEU[0]));

    positionXYZ[0] = (N+positionNEU[2])*cos(positionNEU[0])*cos(positionNEU[1]);
    positionXYZ[1] = (N+positionNEU[2])*cos(positionNEU[0])*sin(positionNEU[1]);
    positionXYZ[2] = ((1-e2)*N+positionNEU[2])*sin(positionNEU[0]);
}

/*****************************************************************************
 * Name        : ConvertCalendarDate
 * Description : Convert calendar date to other date formats and exit 
 *                (date conversion only). Only dates after year 1980 are supported.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int year                        I  N/A  Current year
 * int month                       I  N/A  Current month number
 * int day                         I  N/A  Current day number
 * int hour                        I  N/A  Current hour
 * int minute                      I  N/A  Current minute
 * double seconds                  I  N/A  Current second
 *****************************************************************************/
void ConvertCalendarDate(int year, int month, int day, int hour, int minute, double seconds) {

	int 		GPSweek,DoW,DoY;
	double		SoD,SecondsComp;
	double 		SoW;
	TTime		t;
	struct tm	tm;

	//Fill TTime struct
	t=cal2t (year,month,day,hour,minute,seconds);
	//Fill tm struct
	t2tmnolocal(&t, &tm, &SecondsComp);
	//Get DoY
	DoY=date2doy(year,month,day);
	SoD=(double)hour*3600.+(double)minute*60+seconds;
	//Get GPSWeek
	ttime2gpswsnoroll (&t, &GPSweek, &SoW);
	DoW = (int)(SoW/86400.0);

	printDates(year,month,day,hour,minute,seconds,DoY,SoD,GPSweek,DoW,SoW,t.MJDN);

}

/*****************************************************************************
 * Name        : ConvertDoYDate
 * Description : Convert Year/DoY/SoD date to other date formats and exit 
 *                (date conversion only). Only dates after year 1980 are supported.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int year                        I  N/A  Current year
 * int doy                         I  N/A  Current day of year
 * double sod                      I  N/A  Current second of day
 *****************************************************************************/
void ConvertDoYDate(int year, int doy, double sod) {
	int 		GPSweek,DoW;
	double		SecondsComp;
	double 		SoW;
	TTime		t;
	struct tm	tm;

	//Fill TTime struct
  	t.MJDN=yeardoy2MJDN (year,doy, (int)sod);
	t.SoD=sod;
	//Fill tm struct
    t2tmnolocal(&t, &tm, &SecondsComp);
	//Get GPSWeek
	ttime2gpswsnoroll (&t, &GPSweek, &SoW);
	DoW = (int)(SoW/86400.0);

	printDates(year,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,SecondsComp,doy,sod,GPSweek,DoW,SoW,t.MJDN);
}

/*****************************************************************************
 * Name        : ConvertGPSWeekDate
 * Description : Convert GPSWeek/SoW date to other date formats and exit 
 *                (date conversion only). Only dates after year 1980 are supported.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int GPSWeek                     I  N/A  Current GPS week
 * double SoW                      I  N/A  Current second of week
 *****************************************************************************/
void ConvertGPSWeekDate(int GPSWeek, double SoW) {
	int 		DoW;
	int			year;
	double		DoY;
	double		SecondsComp;
	TTime		t;
	struct tm	tm;

	//Fill TTime struct
	t=gpsws2ttime(GPSWeek,SoW);
	//Fill tm struct
    t2tmnolocal(&t, &tm, &SecondsComp);
	//Get DoW
	DoW = (int)(SoW/86400.0);
	//Get Year/DoY
	t2doy(&t,&year,&DoY);

	printDates(year,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,SecondsComp,(int)DoY,t.SoD,GPSWeek,DoW,SoW,t.MJDN);

}

/*****************************************************************************
 * Name        : ConvertMJDNDate
 * Description : Convert MJDN/SoD date to other dateformats and exit 
 *                (date conversion only). Only dates after year 1980 are supported.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int GPSWeek                     I  N/A  Current MJDN
 * double SoD                      I  N/A  Current second of day
 *****************************************************************************/
void ConvertMJDNDate(int MJDN, double SoD) {

	int 		GPSweek,DoW;
	int			year;
	double		SecondsComp;
	double 		SoW,DoY;
	TTime		t;
	struct tm	tm;

	//Fill TTime struct
	t.MJDN=MJDN;
	t.SoD=SoD;
	//Fill tm struct
	t2tmnolocal(&t, &tm, &SecondsComp);
	//Get DoY
	t2doy(&t,&year,&DoY);
	//Get GPSWeek
	ttime2gpswsnoroll (&t, &GPSweek, &SoW);
	DoW = (int)(SoW/86400.0);

	printDates(year,tm.tm_mon+1,tm.tm_mday,tm.tm_hour,tm.tm_min,SecondsComp,(int)DoY,t.SoD,GPSweek,DoW,SoW,t.MJDN);

}

/*****************************************************************************
 * Name        : ConvertCartesianCoord
 * Description : Convert cartesian coordiantes to other coordinate sytems 
 *                (coordinates conversion only)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double X                        I  m    X
 * double Y                        I  m    Y
 * double Z                        I  m    Z
 *****************************************************************************/
void ConvertCartesianCoord(double x, double y, double z) {

	double 	positionXYZ[3];
	double	positionNEU[3];
	double	positionSph[3];

	positionXYZ[0]=x;
	positionXYZ[1]=y;
	positionXYZ[2]=z;

	XYZ2NEU(positionXYZ,positionNEU);
	XYZ2Spherical(positionXYZ,positionSph);

	printCoordinates(positionXYZ,positionNEU,positionSph);

}

/*****************************************************************************
 * Name        : ConvertGeodeticCoord
 * Description : Convert geodetic coordiantes to other coordinate sytems 
 *                (coordinates conversion only)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double Longitude                I  º     Longitude [-180..180]
 * double Latitude                 I  º     Latitude [-90..90]
 * double Height                   I  m     Height
 *****************************************************************************/
void ConvertGeodeticCoord(double lon, double lat, double height) {

    double  positionXYZ[3];
    double  positionNEU[3];
    double  positionSph[3];

    positionNEU[0]=lat*d2r;
    positionNEU[1]=lon*d2r;
    positionNEU[2]=height;

    NEU2XYZ(positionNEU,positionXYZ);
    XYZ2Spherical(positionXYZ,positionSph);

    printCoordinates(positionXYZ,positionNEU,positionSph);

}

/*****************************************************************************
 * Name        : ConvertSphericalCoord
 * Description : Convert Spherical coordiantes to other coordinate sytems (coordinates conversion only)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double Longitude                I  º     Longitude [-180..180]
 * double Latitude                 I  º     Latitude [-90..90]
 * double Radius                   I  m     Radius
 *****************************************************************************/
void ConvertSphericalCoord(double lon, double lat, double Radius) {

    double  positionXYZ[3];
    double  positionNEU[3];
    double  positionSph[3];

    positionSph[0]=lat*d2r;
    positionSph[1]=lon*d2r;
    positionSph[2]=Radius;

	Spherical2XYZ(positionSph,positionXYZ);
	XYZ2NEU(positionXYZ,positionNEU);

    printCoordinates(positionXYZ,positionNEU,positionSph);

}

/*****************************************************************************
 * Name        : numSteps
 * Description : Return the number of increments of da between values a2 and a1
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double  a1                      I  N/A  First value
 * double  a2                      I  N/A  Second value
 * double  da                      I  N/A  Increment between values
 * Returned value (int)            O  N/A  Number of increments
 *****************************************************************************/
int numSteps (double a1, double a2, double da) {
	if (da==0) return 1; // Uninitialised or invalid increment

	return (int)((a2-a1)/da+1); // +1 is to take into account that both a1 and a2 are included in the distance
}

/*****************************************************************************
 * Name        : allocateIonoMap
 * Description : Allocate memory for ionospheric map. The following fields of
 *               the structure must be already prepared for this function:
 *                    nhgt, nlat, nlon
 *               As they will be used to allocate memory in the members:
 *                    TEC & RMS
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              IO N/A  Ionospheric map structure
 *****************************************************************************/
void allocateIonoMap (TIonoMap *ionoMap) {
	int		totalSize;

	totalSize = ionoMap->nhgt * ionoMap->nlat * ionoMap->nlon;

	ionoMap->TEC = (double*)(realloc(ionoMap->TEC,sizeof(double)*totalSize));
	ionoMap->RMS = (double*)(realloc(ionoMap->RMS,sizeof(double)*totalSize));
}

/*****************************************************************************
 * Name        : E
 * Description : Return the value 10^exponent withouth using pow function
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  exponent                   I  N/A  Exponent
 * Returned value (double)         O  N/A  10^exponent
 *****************************************************************************/
double E (int exponent) {
	double	value = 1;
	int		i;

	if (exponent<0) {
		for (i=0;i>exponent;i--) value /= 10;
	} else if (exponent>0) {
		for (i=0;i<exponent;i++) value *= 10;
	}

	return value;
}

/*****************************************************************************
 * Name        : conv3Dto1D
 * Description : Transforms a 3 dimensional index to a single dimension index
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  i                          I  N/A  First dimension  
 * int  j                          I  N/A  Second dimension
 * int  k                          I  N/A  Third dimension
 * int  ni                         I  N/A  Size of first dimension
 * int  nj                         I  N/A  Size of second dimension
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int	conv3Dto1D (int i, int j, int k, int ni, int nj) {
	if (ni==0 || nj==0) return 0;

	return k*ni*nj + j*ni + i;
}

/*****************************************************************************
 * Name        : ionoConv3Dto1D
 * Description : Transforms a 3 dimensional index to a single dimension index
 *               obtaining the data from a TIonoMap structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * int  ihgt                       I  N/A  Height index  
 * int  ilat                       I  N/A  Latitude index
 * int  ilon                       I  N/A  Longitude index
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int ionoConv3Dto1D (TIonoMap *ionoMap, int ihgt, int ilat, int ilon) {
	return conv3Dto1D(ihgt,ilat,ilon,ionoMap->nhgt,ionoMap->nlat);
}

/*****************************************************************************
 * Name        : ionoHLLto1D
 * Description : Gives the 1D index of the double values of hgt, lat and lon
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lon                     I  deg  Longitude value
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int ionoHLLto1D (TIonoMap *ionoMap, double hgt, double lat, double lon) {
	int		ilat, ilon, ihgt;

	if (ionoMap->nhgt==1) {
		ihgt=0;
	} else {
		ihgt = (int)((hgt-ionoMap->hgt1)/ionoMap->dhgt);
	}

	if (ionoMap->nlat==1) {
		ilat=0;
	} else {
		ilat = (int)((lat-ionoMap->lat1)/ionoMap->dlat);
	}

	if (ionoMap->nlon==1) {
		ilon=0;
	} else {
		ilon = (int)((lon-ionoMap->lon1)/ionoMap->dlon);
	}

	return ionoConv3Dto1D(ionoMap, ihgt, ilat, ilon);
}

/*****************************************************************************
 * Name        : setIonoValue
 * Description : Sets the TEC or RMS value of a ionospheric map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              IO N/A  Ionospheric map structure
 * int  type                       I  N/A  TEC, RMS or HGT (0, 1 or 2)
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lon                     I  deg  Longitude value
 * double  value                   I  TECU Value to fill
 * Returned value (int)            O  N/A  Status of function
 *                                         0 => Failed to set the value
 *                                         1 => Value properly set
 *****************************************************************************/
int setIonoValue (TIonoMap *ionoMap, int type, double hgt, double lat, double lon, double value) {
	int ind;

	ind = ionoHLLto1D(ionoMap,hgt,lat,lon);
	if (ind<0) return 0;

	if (type==0) { // TEC
		ionoMap->TEC[ind] = value;
	} else if (type==1) { // RMS
		ionoMap->RMS[ind] = value;
	} else if (type==2) { // HGT
				ionoMap->HGT[ind] = value;
	} else {
		return 0;
	}

	return 1;
}

/*****************************************************************************
 * Name        : getIonoValue
 * Description : Returns the TEC or RMS value of a ionospheric map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * int  type                       I  N/A  TEC, RMS or HGT (0, 1 or 2)
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lon                     I  deg  Longitude value
 * Returned value (double)         O  TECU Value of the cell (in TECUs).
 *                                    9999 is used for invalid values,
 *                                    as defined in IONEX format)
 *****************************************************************************/
double getIonoValue (TIonoMap *ionoMap, int type, double hgt, double lat, double lon) {
	int ind;

	ind = ionoHLLto1D(ionoMap,hgt,lat,lon);

	if (ind<0) return 9999;

	if (type==0) { // TEC
		return ionoMap->TEC[ind];
	} else if (type==1) { // RMS
		return ionoMap->RMS[ind];
	} else if (type==2) { // HGT
				return ionoMap->HGT[ind];
	}

	return 9999;
}

/*****************************************************************************
 * Name        : getIonoMapTotalSize
 * Description : Returns the total size of a map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TIonoMap  *ionoMap              I  N/A  Ionospheric map structure
 * Returned value (int)            O  N/A  Map total size
 *****************************************************************************/
int getIonoMapTotalSize (TIonoMap *ionoMap) {
		return ionoMap->nhgt * ionoMap->nlat * ionoMap->nlon;
}

/*****************************************************************************
 * Name        : FPPPConv3Dto1D
 * Description : Transforms a 3 dimensional index to a single dimension index
 *               obtaining the data from a TFPPPMap structure
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPPMap  *FPPPMap              I  N/A  FPPP map structure
 * int  ihgt                       I  N/A  Height index  
 * int  ilat                       I  N/A  Latitude index
 * int  ilt                        I  N/A  Local time index
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int FPPPConv3Dto1D (TFPPPMap *FPPPMap, int ihgt, int ilat, int ilt) {
	return conv3Dto1D(ihgt,ilat,ilt,NUMBERFPPPMAPS,FPPPMap->nlat);
}

/*****************************************************************************
 * Name        : FPPPHLLto1D
 * Description : Gives the 1D index of the double values of hgt, lat and lon
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPPMap  *FPPPMap              I  N/A  FPPP map structure
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lt                      I  deg  Local time value
 * Returned value (int)            O  N/A  1D index
 *****************************************************************************/
int FPPPHLLto1D (TFPPPMap *FPPPMap, double hgt, double lat, double lt) {
	int		ilat, ilt, ihgt;

	if (hgt == FPPPMap->hgt1) { 
		ihgt = 0;
	} else if (hgt == FPPPMap->hgt2) {
		ihgt = 1;
	} else {
		return -1;
	}

	if (FPPPMap->nlat==1) {
		ilat=0;
	} else {
		ilat = (int)((lat-FPPPMap->lat1)/FPPPMap->dlat);
	}

	if (FPPPMap->nlt==1) {
		ilt=0;
	} else {
		ilt = (int)((lt-FPPPMap->lt1)/FPPPMap->dlt);
	}

	return FPPPConv3Dto1D(FPPPMap, ihgt, ilat, ilt);
}

/*****************************************************************************
 * Name        : setFPPPValue
 * Description : Sets the TEC or SIGMA value of a FPPP map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPPMap  *FPPPMap              IO N/A  FPPP map structure
 * int  type                       I  N/A  TEC, SIGMA (0 or 1)
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lt                      I  deg  Local time value
 * double  value                   I  TECU Value to fill
 * Returned value (int)            O  N/A  Status of function
 *                                         0 => Failed to set the value
 *                                         1 => Value properly set
 *****************************************************************************/
int setFPPPValue (TFPPPMap *FPPPMap, int type, double hgt, double lat, double lt, double value) {
	int ind;

	ind = FPPPHLLto1D(FPPPMap,hgt,lat,lt);
	if (ind<0) return 0;

	if (type==0) { // TEC
		FPPPMap->TEC[ind] = value;
	} else if (type==1) { // SIGMA
		FPPPMap->SIGMA[ind] = value;
	} else {
		return 0;
	}

	return 1;
}

/*****************************************************************************
 * Name        : getFPPPValue
 * Description : Returns the TEC or SIGMA value of a FPPP map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPPMap  *FPPPMap              I  N/A  FPPP map structure
 * int  type                       I  N/A  TEC, SIGMA (0 or 1)
 * double  hgt                     I  km   Height value  
 * double  lat                     I  deg  Latitude value
 * double  lt                      I  deg  Local time value
 * Returned value (double)         O  TECU Value of the cell (in TECUs).
 *                                    9999 is used for invalid values,
 *                                    as defined in IONEX format)
 *****************************************************************************/
double getFPPPValue (TFPPPMap *FPPPMap, int type, double hgt, double lat, double lt) {
	int ind;

	ind = FPPPHLLto1D(FPPPMap,hgt,lat,lt);

	if (ind<0) return 9999;

	if (type==0) { // TEC
		return FPPPMap->TEC[ind];
	} else if (type==1) { // SIGMA
		return FPPPMap->SIGMA[ind];
	}

	return 9999;
}

/*****************************************************************************
 * Name        : getFPPPMapTotalSize
 * Description : Returns the total size of a map
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TFPPP  *FPPPMap                 I  N/A  TFPPP structure
 * Returned value (int)            O  N/A  Map total size
 *****************************************************************************/
int getFPPPMapTotalSize (TFPPPMap *FPPPMap) {
	return NUMBERFPPPMAPS * FPPPMap->nlat * FPPPMap->nlt;
}

/*****************************************************************************
 * Name        : getIonoPiercePoint
 * Description : Computes the ionospheric pierce point for a 
 *                   satellite-receiver ray
 * Parameters  :
 * Name                             |Da|Unit|Description
 * TEpoch  *epoch                    I  N/A  TEpoch structure
 * TPiercePoint *PiercePoint         IO N/A  TPiercePoint structure
 * int  satIndex                     I  N/A  Index of the satellite to 
 *                                             obtain the correction
 * double earthRadius                I  Km   Earth radius (in Kilometers)
 * double altitude                   I  Km   IonoMap altitude (in Kilometers)
 * int mode                          I  N/A  0 => Use spherical coordinates for reference station (IONEX, FPPP)
 *                                           1 => Use geodetic coordinates for reference station (SBAS)
 * Returned value                    O  N/A  None
 *****************************************************************************/
void getIonoPiercePoint (TEpoch *epoch, TPiercePoint *PiercePoint, int satIndex, double earthRadius, double altitude, int mode){

	double rhox,rhoy,rhoz;
	double rho;
	double recx,recy,recz;
	double Rse;
	double raylambda;
	double xymodule;

	//For the calculations in the Pierce Point and in the rest of ionexModel and FPPPmodel, a geocentric elevation is used. That differs from gLAB's elevation, which is geodetic elevation

	earthRadius*= 1000; //Convert to meters 
	altitude*= 1000;    //Convert to meters
	
	if(mode==0) {
		recx=epoch->receiver.aproxPosition[0];
		recy=epoch->receiver.aproxPosition[1];
		recz=epoch->receiver.aproxPosition[2];
		earthRadius = sqrt(recx*recx+recy*recy+recz*recz); // For IONEX and F-PPP, the Earth Radius contains the Earth Radius plus the height from the user
	} else {
		//mode==1
		recx=earthRadius*cos(epoch->receiver.aproxPositionNEU[0])*cos(epoch->receiver.aproxPositionNEU[1]);
		recy=earthRadius*cos(epoch->receiver.aproxPositionNEU[0])*sin(epoch->receiver.aproxPositionNEU[1]);
		recz=earthRadius*sin(epoch->receiver.aproxPositionNEU[0]);
	}


	rhox = epoch->sat[satIndex].position[0] - recx;
	rhoy = epoch->sat[satIndex].position[1] - recy;
	rhoz = epoch->sat[satIndex].position[2] - recz;

	rho = sqrt(pow(rhox,2) + pow(rhoy,2) + pow (rhoz,2));
	Rse = (recx*rhox + recy*rhoy + recz*rhoz)/rho;
	raylambda = -Rse + sqrt(pow(earthRadius + altitude,2) - pow(earthRadius,2) + pow(Rse,2));

	PiercePoint->x = recx + raylambda*rhox/rho;
	PiercePoint->y = recy + raylambda*rhoy/rho;
	PiercePoint->z = recz + raylambda*rhoz/rho;
	
	xymodule = sqrt( pow(PiercePoint->x,2) + pow(PiercePoint->y,2) );
	
	PiercePoint->fi = atan2(PiercePoint->z,xymodule)*r2d;	//Latitude between -90..90º
	PiercePoint->longitude = atan2(PiercePoint->y,PiercePoint->x)*r2d; //Longitude between -180º..180º
	PiercePoint->elevation = asin(Rse/earthRadius)*r2d;

	epoch->IPPlat = PiercePoint->fi;
	epoch->IPPlon = PiercePoint->longitude;
}

/*****************************************************************************
 * Name        : getMappingFunction
 * Description : Computes the mapping function for a given grid at any height  
 * Parameters  :
 * Comment     : QFAC function is from "Mapping and Predicting the Earth's
 *                  Ionosphere Using the Global Positioning System", Stefan
 *                  Schaer Inaugural dissertation
 * Name                                 |Da|Unit|Description
 * TPiercePoint *PiercePoint             IO N/A  TPiercePoint structure
 * double earthRadius                    I  N/A  Earth radius from IONEX file
 * double altitude                       I  N/A  IonoMap altitude
 * TOptions  *options                    I  N/A  TOptions structure
 * Returned value                        O  N/A  None
 *****************************************************************************/
void getMappingFunction (TPiercePoint *PiercePoint, double earthRadius, double altitude, char functionType[5],TOptions *options){

	double	a0 = 1.0206;
	double	a1 = 0.4663;
	double	a2 = 3.5055;
	double	a3 = -1.8415;
	double  z = 90 + PiercePoint->elevation;

	earthRadius*= 1000; //Convert to meters
	altitude*= 1000;    //Convert to meters

	if (strncmp(functionType,"COSZ",4)==0 || strncmp(functionType,"NONE",4)==0){
		PiercePoint->mappingFunction = (altitude + earthRadius) / ( sqrt( pow(altitude + earthRadius,2) - pow(earthRadius*cos(PiercePoint->elevation*d2r),2) ) );
	}
	else if (strncmp(functionType,"QFAC",4)==0){
		PiercePoint->mappingFunction = a0 + a1*pow((z/90),2) + a2*pow((z/90),4) + a3*pow((z/90),6);
	}
	else{
		sprintf(messagestr,"Mapping function type \"%s\" in IONEX file not supported\n",functionType); 
		printError(messagestr,options);
	}

}


/*****************************************************************************
 * Name        : getConstellationElement
 * Description : Returns the TConstellationElement of a satellite for a given
 *               time
 * Parameters  :
 * Name                           |Da|Unit|Description
 * enum GNSSystem  GNSS            I  N/A  GNSS enumerator of the satellite
 * int  PRN                        I  N/A  PRN of the satellite
 * TTime  *t                       I  N/A  Reference time
 * TConstellation  *constellation  I  N/A  TConstellation structure
 * Returned value (TConstellationElement*)
 *                                 O  N/A  Pointer to the TConstellationElement
 *                                         selected
 *                                         NULL => Satellite not found
 *****************************************************************************/
TConstellationElement* getConstellationElement (enum GNSSystem GNSS, int PRN, TTime *t,TConstellation* constellation) {
	int i;

	for (i=0;i<constellation->numSatellites;i++) {
		if (constellation->sat[i].GNSS == GNSS && constellation->sat[i].PRN == PRN) {
			if (tdiff(t,&constellation->sat[i].tLaunch)>0 && tdiff(t,&constellation->sat[i].tDecommissioned)<0) {
				// If t is beyond tLaunch but lower than tDecommissioned
				return &constellation->sat[i];
			}
		}
	}
	return NULL;
}

/*****************************************************************************
 * Name        : getStation
 * Description : Get the station structure from its name (as a string)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *name                     I  N/A  Name of the station
 * TStationList  *stationList      I  N/A  List of stations to look for
 * Returned value (*TStatoin)      O  N/A  Pointer to the antenna inside TStationList
 *                                         If it is NULL, the station has not been
 *                                         found
 *****************************************************************************/
TStation *getStation (char *name, TStationList *stationList) {
	int		i;
	char	aux[5];

	//Copy only the first four characters of the station name, which will be the name in the SINEX
	strncpy(aux,name,4);
	aux[4]='\0';

	for (i=0;i<stationList->n;i++) {
		if (strcasecmp(aux,stationList->station[i].name)==0) return &stationList->station[i];
	}
	
	return NULL;
}

/*****************************************************************************
 * Name        : addStat
 * Description : Add a value to a TStat structure, computing the mean and square
 *               mean of all the values that have been inserted
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TStat  *stat                    IO N/A  TStat structure to operate
 * double  value                   I  N/A  Value to include to stat
 *****************************************************************************/
void addStat (TStat *stat, double value) {
	stat->mean = (stat->n*stat->mean + value)/(stat->n+1);
	stat->mean2 = (stat->n*stat->mean2 + value*value)/(stat->n+1);
	stat->n++;
}



/*****************************************************************************
 * Name        : ConcatenateSP3
 * Description : Concanate precise orbits and clocks from several (if available)
 *               and also take out all samples with invalid values
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int RefData                     I  N/A  Flag to indicate if data is for reference position
 *                                           (no clocks, only one satellite). Values are 0 or 1
 * TGNSSproducts  *products        IO N/A  TGNSSproducts structure
 * TGNSSproducts  *pastSP3Prod     I  N/A  TGNSSproducts structure for previous SP3 products
 * TGNSSproducts  *nextSP3Prod     I  N/A  TGNSSproducts structure for next SP3 products
 * TOptions  *options              I  N/A  TOptions structure to configure behaviour
 *****************************************************************************/
void ConcatenateSP3 (int RefData, TGNSSproducts *products, TGNSSproducts *pastSP3Prod, TGNSSproducts *nextSP3Prod,  TOptions  *options) {

	int 			i,GNSS,PRN;
	int				indJ,indK;
	int				ind,indPast,indNext,indTmp;
	int				FirstSampleToBeSkipped;
	int				MissingSamples;
	int				initPRN=1;
	double			diffTimePast,diffTimeNext;
	double			diffSample;
	TSP3orbits		*orbitsNext;
	TSP3orbits		*orbitsPast;
	TSP3orbits		*orbitsTmp;
	TSP3clocks		*clocksTmp;
	TSP3clocks		*clocksNext;
	TSP3clocks		*clocksPast;

	if (RefData==1) {
		initPRN=0;
	}
	
	//Free memory from previous concatenations
	//Concatenation from previous products will be in pastSP3Prod structure,
	//as data from products has been previously moved to pastSP3Prod
	for(i=0;i<MAX_GNSS*MAX_SATELLITES_PER_GNSS;i++) {
		if(pastSP3Prod->SP3->orbits.Concatblock[i]!=NULL) {
			free(pastSP3Prod->SP3->orbits.Concatblock[i]);
			pastSP3Prod->SP3->orbits.Concatblock[i]=NULL;
		}
		if (RefData==0) {
			if(pastSP3Prod->SP3->clocks.Concatblock[i]!=NULL) {
				free(pastSP3Prod->SP3->clocks.Concatblock[i]);
				pastSP3Prod->SP3->clocks.Concatblock[i]=NULL;
			}
		}
	}

	orbitsTmp = &products->SP3->orbits;
	orbitsPast = &pastSP3Prod->SP3->orbits;
	orbitsNext = &nextSP3Prod->SP3->orbits;
	//Get minimum interval orbit sample time
	orbitsTmp->intervalConcat=min(orbitsPast->interval,orbitsTmp->interval);
	orbitsTmp->intervalConcat=min(orbitsTmp->intervalConcat,orbitsNext->interval);
	//Start epoch will be the one from the previous day
	memcpy(&orbitsTmp->startTimeConcat,&orbitsPast->startTime,sizeof(TTime));
	//End epoch will be the one from the next day
	memcpy(&orbitsTmp->endTimeConcat,&orbitsNext->endTime,sizeof(TTime));

	if(RefData==0) {
		//Concatenate clocks (from SP3 or CLK file)
		clocksTmp = &products->SP3->clocks;
		clocksPast = &pastSP3Prod->SP3->clocks;
		clocksNext = &nextSP3Prod->SP3->clocks;
		//Get minimum interval clock sample time
		clocksTmp->intervalConcat=min(clocksPast->interval,clocksTmp->interval);
		clocksTmp->intervalConcat=min(clocksTmp->intervalConcat,clocksNext->interval);
		//Start epoch will be the one from the previous day
		memcpy(&clocksTmp->startTimeConcat,&clocksPast->startTime,sizeof(TTime));
		//Reset start time of clocks
		memcpy(&clocksTmp->endTimeConcat,&clocksNext->endTime,sizeof(TTime));
	}


	for(GNSS=0;GNSS<MAX_GNSS;GNSS++) {
		for(PRN=initPRN;PRN<MAX_SATELLITES_PER_GNSS;PRN++) {
			//Concatenate SP3 orbits
			
			// orbitsTmp corresponds to the current products
			ind = orbitsTmp->index[GNSS][PRN];

			// Take into account previous and following SP3 products 
			indPast = orbitsPast->index[GNSS][PRN];
			indNext = orbitsNext->index[GNSS][PRN];


			if (RefData==1) {
				ind=indPast=indNext=0;
			}
			
			if (ind==-1 && indPast==-1 && indNext==-1) continue;


			// Check how many data can be allocated in arrays
			if(indPast!=-1 && ind!=-1) {
				diffTimePast=tdiff(&orbitsTmp->block[ind][0].t,&orbitsPast->block[indPast][0].t);
			} else if (indPast!=-1 && ind==-1) {
				diffTimePast=86400.; 
			} else {
				diffTimePast=0.;
			}

			if(indNext!=-1 && ind!=-1) {
				diffTimeNext=tdiff(&orbitsNext->block[indNext][0].t,&orbitsTmp->block[ind][0].t);
			} else if (indNext!=-1 && ind==-1) {
				diffTimeNext=86400.; 
			} else {
				diffTimeNext=0.;
			}

			if(ind==-1) {
				//There is no data for today, but for previous or next day data is available
				//Add a new index number
				orbitsTmp->index[GNSS][PRN]=orbitsTmp->numSatellites;
				indTmp=-1;
				ind=orbitsTmp->numSatellites;
				orbitsTmp->numSatellites++;
				orbitsTmp->block = realloc(orbitsTmp->block,sizeof(TSP3orbitblock*)*orbitsTmp->numSatellites);
				orbitsTmp->block[ind] = malloc(sizeof(TSP3orbitblock)*orbitsTmp->numRecords);
				for (i=0;i<orbitsTmp->numRecords;i++) {
					initSP3orbitblock(&orbitsTmp->block[ind][i]);
					memcpy(&orbitsTmp->block[ind][i].t,&orbitsTmp->block[0][i].t,sizeof(TTime)); //Copy timestamps from first satellite in list (any satellite could be used, as timestamp is the same for all satellites)
				}
			} else {
				indTmp=ind;
			}

			//Check if there are past SP3 products than current time
			if ( diffTimePast >= 85500.0 && diffTimePast <= 87300.0 ) { //Leave a possible data gap of +-15 minutes (900 seconds)
				indJ=indK=0;
				MissingSamples=0;
				for (i=0;i<orbitsPast->numRecords;i++) {		
					//Samples must be saved with an equal time between each samples. Therefore, we need to save epochs with 0s or add samples with 0's
					//to account for data gaps
					if (i==0) {
						//The first sample must be compared with the start of the data
						diffSample=tdiff(&orbitsPast->block[indPast][i].t,&orbitsTmp->startTimeConcat);
					} else {
						diffSample=tdiff(&orbitsPast->block[indPast][i].t,&orbitsTmp->Concatblock[ind][indJ].t);
					}
					while(diffSample>orbitsTmp->intervalConcat) {
						//Data gap. Add epochs until all data gaps are filled with 0s
						indJ=i+MissingSamples;
						orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indJ+1));
						initSP3orbitblock(&orbitsTmp->Concatblock[ind][indJ]);
						if (indJ==0) {
							//If it is the first sample saved, the first epoch is the start epoch of data
							memcpy(&orbitsTmp->Concatblock[ind][indJ].t,&orbitsTmp->startTimeConcat,sizeof(TTime));
						} else {
							orbitsTmp->Concatblock[ind][indJ].t=tdadd(&orbitsTmp->Concatblock[ind][indJ-1].t,orbitsTmp->intervalConcat);
						}
						diffSample=tdiff(&orbitsPast->block[indPast][i].t,&orbitsTmp->Concatblock[ind][indJ].t);
						MissingSamples++;
					}
					indJ=i+MissingSamples;
					orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indJ+1));
					memcpy(&orbitsTmp->Concatblock[ind][indJ],&orbitsPast->block[indPast][i],sizeof(TSP3orbitblock));
					if (orbitsPast->block[indPast][i].x[0]!=0. || orbitsPast->block[indPast][i].x[1]!=0. || orbitsPast->block[indPast][i].x[2]!=0.) {
						orbitsTmp->numblocsValidConcat[GNSS][PRN]++;
						orbitsTmp->Concatblock[ind][indJ].validSample=1;
					}
				}
				indJ++;
				//Check if the last entry of previous file and first entry of current file are the same
				if(indTmp==-1) {
					//No samples for current day
					FirstSampleToBeSkipped=0;
				} else if( tdiff(&orbitsTmp->block[ind][0].t,&orbitsPast->block[indPast][orbitsPast->numRecords-1].t)==0.)  {
					//Last entry of previous day and first entry of current day are the same
					//Check if last entry of previous day has invalid coordinates
					if(orbitsPast->block[indPast][orbitsPast->numRecords-1].x[0]==0. && orbitsPast->block[indPast][orbitsPast->numRecords-1].x[1]==0. && orbitsPast->block[indPast][orbitsPast->numRecords-1].x[2]==0.) {
						FirstSampleToBeSkipped=0;
						indJ--; //Substract 1 to overwrite the previous record
						orbitsTmp->numblocsValidConcat[GNSS][PRN]--;
					} else {
						//Last entry of previous day is valid, check if we need to keep the sample from the previous day or from current day
						if(options->LastSamplePrevDayOrb==1) {
							//Use Sample from previous day
							FirstSampleToBeSkipped=1;
						} else {
							//Use Sample from next day
							FirstSampleToBeSkipped=0;
							indJ--; //Substract 1 to overwrite the previous record
							orbitsTmp->numblocsValidConcat[GNSS][PRN]--;
						}
					}
				} else {
					FirstSampleToBeSkipped=0;
				}
				if (indTmp!=-1) {
					MissingSamples=0;
					indK = indJ;
					for (i=FirstSampleToBeSkipped;i<orbitsTmp->numRecords;i++) {
						if(i==FirstSampleToBeSkipped) {
							diffSample=tdiff(&orbitsTmp->block[ind][i].t,&orbitsTmp->Concatblock[ind][indK-1].t);
						} else {
							diffSample=tdiff(&orbitsTmp->block[ind][i].t,&orbitsTmp->Concatblock[ind][indK].t);
						}
						while(diffSample>orbitsTmp->intervalConcat) {
							//Data gap. Add epochs until all data gaps are filled with 0s
							indK = indJ + i + MissingSamples - FirstSampleToBeSkipped;
							orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indK+1));
							initSP3orbitblock(&orbitsTmp->Concatblock[ind][indK]);
							orbitsTmp->Concatblock[ind][indK].t=tdadd(&orbitsTmp->Concatblock[ind][indK-1].t,orbitsTmp->intervalConcat);
							diffSample=tdiff(&orbitsTmp->block[ind][i].t,&orbitsTmp->Concatblock[ind][indK].t);
							MissingSamples++;
						}
						indK = indJ + i + MissingSamples - FirstSampleToBeSkipped;

						orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indK+1));
						memcpy(&orbitsTmp->Concatblock[ind][indK],&orbitsTmp->block[ind][i],sizeof(TSP3orbitblock));
						if (orbitsTmp->block[ind][i].x[0]!=0. || orbitsTmp->block[ind][i].x[1]!=0. || orbitsTmp->block[ind][i].x[2]!=0.) {
							orbitsTmp->numblocsValidConcat[GNSS][PRN]++;
							orbitsTmp->Concatblock[ind][indK].validSample=1;
						}
					}
				} else {
					//No data for current time
					indK = indJ;
				}
			} else { // No previous SP3 products
				indK=0;
				MissingSamples=0;
				if(indTmp!=-1) {
					for (i=0;i<orbitsTmp->numRecords;i++) {
						if (i==0) {
							//The first sample must be compared with the start of the data
							diffSample=tdiff(&orbitsTmp->block[ind][i].t,&orbitsTmp->startTimeConcat);
						} else {
							diffSample=tdiff(&orbitsTmp->block[ind][i].t,&orbitsTmp->Concatblock[ind][indK].t);
						}
						while(diffSample>orbitsTmp->intervalConcat) {
							//Data gap. Add epochs until all data gaps are filled with 0s
							indK=i+MissingSamples;
							orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indK+1));
							initSP3orbitblock(&orbitsTmp->Concatblock[ind][indK]);
							if (indK==0) {
								//If it is the first sample saved, the first epoch is the start epoch of data
								memcpy(&orbitsTmp->Concatblock[ind][indK].t,&orbitsTmp->startTimeConcat,sizeof(TTime));
							} else {
								orbitsTmp->Concatblock[ind][indK].t=tdadd(&orbitsTmp->Concatblock[ind][indK-1].t,orbitsTmp->intervalConcat);
							}
							diffSample=tdiff(&orbitsTmp->block[ind][i].t,&orbitsTmp->Concatblock[ind][indK].t);
							MissingSamples++;
						}
						indK=i+MissingSamples;
						orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indK+1));
						memcpy(&orbitsTmp->Concatblock[ind][indK],&orbitsTmp->block[ind][i],sizeof(TSP3orbitblock));
						if (orbitsTmp->block[ind][i].x[0]!=0. || orbitsTmp->block[ind][i].x[1]!=0. || orbitsTmp->block[ind][i].x[2]!=0.) {
							orbitsTmp->numblocsValidConcat[GNSS][PRN]++;
							orbitsTmp->Concatblock[ind][indK].validSample=1;
						}
					}
				}
			}
			// Fill arrays with future SP3 products
			if ( diffTimeNext >= 85500.0 && diffTimeNext <= 87300.0 ) { //Leave a possible data gap of +-15 minutes (900 seconds)
				indK++;
				//Check if the last entry of previous file and first entry of current file are the same
				if(indTmp==-1) {
					FirstSampleToBeSkipped=0;
				} else if(tdiff(&orbitsNext->block[indNext][0].t,&orbitsTmp->block[ind][orbitsTmp->numRecords-1].t)==0.) {
					//Last entry of current day and first entry of next day are the same
					//Check if last entry of current day has invalid coordinates
					if(orbitsTmp->block[ind][orbitsTmp->numRecords-1].x[0]==0. && orbitsTmp->block[ind][orbitsTmp->numRecords-1].x[1]==0. && orbitsTmp->block[ind][orbitsTmp->numRecords-1].x[2]==0.) {
						//Last entry of current day is invalid, so the last one saved is from a previous epoch. indK has to be increased.
						FirstSampleToBeSkipped=0;
						indK--; //Substract 1 to overwrite the previous record
						orbitsTmp->numblocsValidConcat[GNSS][PRN]--;
					} else {
						//Last entry of current day is valid, check if we need to keep the sample from the current day or from next day
						if(options->LastSamplePrevDayOrb==1) {
							//Use Sample from current day
							FirstSampleToBeSkipped=1;
						} else {
							//Use Sample from next day
							FirstSampleToBeSkipped=0;
							indK--; //Substract 1 to overwrite the previous record
							orbitsTmp->numblocsValidConcat[GNSS][PRN]--;
						}
					}
				} else {
					FirstSampleToBeSkipped=0;
				}
				MissingSamples=0;
				if (indPast!=-1 && indTmp==-1) {
					indK--;
				} else if (indPast==-1 && indTmp==-1) {
					indK=0;
				}
				indJ=indK;
				for (i=FirstSampleToBeSkipped;i<orbitsNext->numRecords;i++) {
					if (indPast==-1 && indTmp==-1) {
						//No samples from previous day and current day
						diffSample=tdiff(&orbitsNext->block[indNext][i].t,&orbitsTmp->startTimeConcat);
					} else if(i==FirstSampleToBeSkipped) {
						diffSample=tdiff(&orbitsNext->block[indNext][i].t,&orbitsTmp->Concatblock[ind][indJ-1].t);
					} else {
						diffSample=tdiff(&orbitsNext->block[indNext][i].t,&orbitsTmp->Concatblock[ind][indJ].t);
					}
					while(diffSample>orbitsTmp->intervalConcat) {
						//Data gap. Add epochs until all data gaps are filled with 0s
						indJ = indK + i + MissingSamples -FirstSampleToBeSkipped;
						orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indJ+1));
						initSP3orbitblock(&orbitsTmp->Concatblock[ind][indJ]);
						if (indJ==0) {
							//If it is the first sample saved, the first epoch is the start epoch of data
							memcpy(&orbitsTmp->Concatblock[ind][indJ].t,&orbitsTmp->startTimeConcat,sizeof(TTime));
						} else {
							orbitsTmp->Concatblock[ind][indJ].t=tdadd(&orbitsTmp->Concatblock[ind][indJ-1].t,orbitsTmp->intervalConcat);
						}
						diffSample=tdiff(&orbitsNext->block[indNext][i].t,&orbitsTmp->Concatblock[ind][indJ].t);
						MissingSamples++;
					}

					indJ = indK + i + MissingSamples - FirstSampleToBeSkipped;
					orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indJ+1));
					memcpy(&orbitsTmp->Concatblock[ind][indJ],&orbitsNext->block[indNext][i],sizeof(TSP3orbitblock));
					if (orbitsNext->block[indNext][i].x[0]!=0. || orbitsNext->block[indNext][i].x[1]!=0. || orbitsNext->block[indNext][i].x[2]!=0.) {
						orbitsTmp->numblocsValidConcat[GNSS][PRN]++;
						orbitsTmp->Concatblock[ind][indJ].validSample=1;
					}
				}
			} else {
				//No future SP3 products
				indJ=indK;
				if (indPast!=-1 && indTmp==-1) {
					indJ--;
				}
				//Check if we need to fill with 0 for the next day
				diffSample=tdiff(&orbitsTmp->endTimeConcat,&orbitsTmp->Concatblock[ind][indJ].t);
				while(diffSample>0.) {
					indJ++;
					orbitsTmp->Concatblock[ind]=realloc(orbitsTmp->Concatblock[ind],sizeof(TSP3orbitblock)*(indJ+1));
					initSP3orbitblock(&orbitsTmp->Concatblock[ind][indJ]);
					orbitsTmp->Concatblock[ind][indJ].t=tdadd(&orbitsTmp->Concatblock[ind][indJ-1].t,orbitsTmp->intervalConcat);
					diffSample=tdiff(&orbitsTmp->endTimeConcat,&orbitsTmp->Concatblock[ind][indJ].t);
				}
			}

			orbitsTmp->numblocsConcat[GNSS][PRN]=indJ+1;	
			if(RefData==1) break;


			ind = clocksTmp->index[GNSS][PRN];
			indPast = clocksPast->index[GNSS][PRN];
			indNext = clocksNext->index[GNSS][PRN];


			if (ind==-1 && indPast==-1 && indNext==-1) continue;


			// Check how many data can be allocated in arrays
			if(indPast!=-1 && ind!=-1) {
				diffTimePast=tdiff(&clocksTmp->block[ind][0].t,&clocksPast->block[indPast][0].t);
			} else if (indPast!=-1 && ind==-1) {
				diffTimePast=86400.; 
			} else {
				diffTimePast=0.;
			}

			if(indNext!=-1 && ind!=-1) {
				diffTimeNext=tdiff(&clocksNext->block[indNext][0].t,&clocksTmp->block[ind][0].t);
			} else if (indNext!=-1 && ind==-1) {
				diffTimeNext=86400.; 
			} else {
				diffTimeNext=0.;
			}

			if(ind==-1) {
				//There is no data for today, but for previous or next day data is available
				//Add a new index number and allocate memory for this slot
				clocksTmp->index[GNSS][PRN]=clocksTmp->numSatellites;
				indTmp=-1;
				ind=clocksTmp->numSatellites;
				clocksTmp->numSatellites++;
				clocksTmp->block = realloc(clocksTmp->block,sizeof(TSP3clockblock*)*clocksTmp->numSatellites);
				clocksTmp->block[ind] = malloc(sizeof(TSP3clockblock)*clocksTmp->vsizeblock);
				for (i=0;i<clocksTmp->vsizeblock;i++) {
					initSP3clockblock(&clocksTmp->block[ind][i]);
					memcpy(&clocksTmp->block[ind][i].t,&clocksTmp->block[0][i].t,sizeof(TTime)); //Copy timestamps from first satellite in list (any satellite could be used, as timestamp is the same for all satellites)
				}
			} else {
				indTmp=ind;
			}


			//We will create a new block with past, present and future clocks, and also without INVALID_CLOCK entries
			if ( diffTimePast >= 85500.0 && diffTimePast <= 87300.0 ) {	//Leave a possible data gap of +-15 minutes (900 seconds)
				MissingSamples=0;
				indJ=indK=0;
				for ( i = 0; i < clocksPast->numRecords; i++ ) {
					//Samples must be saved with an equal time between each samples. Therefore, we need to save epochs with 0s or add samples with 0's
					//to account for data gaps
					if (i==0) {
						//The first sample must be compared with the start of the data
						diffSample=tdiff(&clocksPast->block[indPast][i].t,&clocksTmp->startTimeConcat);
					} else {
						diffSample=tdiff(&clocksPast->block[indPast][i].t,&clocksTmp->Concatblock[ind][indJ].t);
					}
					while(diffSample>clocksTmp->intervalConcat) {
						//Data gap. Add epochs until all data gaps are filled with 0s
						indJ=i+MissingSamples;
						clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indJ+1));
						initSP3clockblock(&clocksTmp->Concatblock[ind][indJ]);
						if (indJ==0) {
							//If it is the first sample saved, the first epoch is the start epoch of data
							memcpy(&clocksTmp->Concatblock[ind][indJ].t,&clocksTmp->startTimeConcat,sizeof(TTime));
						} else {
							clocksTmp->Concatblock[ind][indJ].t=tdadd(&clocksTmp->Concatblock[ind][indJ-1].t,clocksTmp->intervalConcat);
						}
						diffSample=tdiff(&clocksPast->block[indPast][i].t,&clocksTmp->Concatblock[ind][indJ].t);
						MissingSamples++;
					}
					indJ=i+MissingSamples;
					clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indJ+1));
					memcpy(&clocksTmp->Concatblock[ind][indJ],&clocksPast->block[indPast][i],sizeof(TSP3clockblock));
					if (clocksPast->block[indPast][i].clock!=INVALID_CLOCK) {
						clocksTmp->numblocsValidConcat[GNSS][PRN]++;
						clocksTmp->Concatblock[ind][indJ].validSample=1;
					}
				}
				indJ++;
				//Check if the last entry of previous file and first entry of current file are the same
				if (indTmp==-1) {
					//No samples for current day
					FirstSampleToBeSkipped=0;
				} else if(tdiff(&clocksTmp->block[ind][0].t,&clocksPast->block[indPast][clocksPast->numRecords-1].t)==0.) {
					//Last entry of previous day and first entry of current day are the same
					//Check if last entry of previous day does not have INVALID_CLOCK
					if(clocksPast->block[indPast][clocksPast->numRecords-1].clock==INVALID_CLOCK) {
						FirstSampleToBeSkipped=0;
						indJ--; //Substract 1 to overwrite the previous record
						clocksTmp->numblocsValidConcat[GNSS][PRN]--;
					} else {
						//Last entry of previous day is valid, check if we need to keep the sample from the previous day or from current day
						if(options->LastSamplePrevDayClk==1) {
							//Use Sample from previous day
							FirstSampleToBeSkipped=1;
						} else {
							//Use Sample from next day
							FirstSampleToBeSkipped=0;
							indJ--; //Substract 1 to overwrite the previous record
							clocksTmp->numblocsValidConcat[GNSS][PRN]--;
						}
					}
				} else {
					FirstSampleToBeSkipped=0;
				}
				if (indTmp!=-1) {
					MissingSamples=0;
					indK = indJ;
					for ( i = FirstSampleToBeSkipped; i < clocksTmp->numRecords; i++ ) {
						if(i==FirstSampleToBeSkipped) {
							diffSample=tdiff(&clocksTmp->block[ind][i].t,&clocksTmp->Concatblock[ind][indK-1].t);
						} else {
							diffSample=tdiff(&clocksTmp->block[ind][i].t,&clocksTmp->Concatblock[ind][indK].t);
						}
						while(diffSample>clocksTmp->intervalConcat) {
							//Data gap. Add epochs until all data gaps are filled with 0s
							indK = indJ + i + MissingSamples - FirstSampleToBeSkipped;
							clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indK+1));
							initSP3clockblock(&clocksTmp->Concatblock[ind][indK]);
							clocksTmp->Concatblock[ind][indK].t=tdadd(&clocksTmp->Concatblock[ind][indK-1].t,clocksTmp->intervalConcat);
							diffSample=tdiff(&clocksTmp->block[ind][i].t,&clocksTmp->Concatblock[ind][indK].t);
							MissingSamples++;
						}
						indK = indJ + i + MissingSamples - FirstSampleToBeSkipped;

						clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indK+1));
						memcpy(&clocksTmp->Concatblock[ind][indK],&clocksTmp->block[ind][i],sizeof(TSP3clockblock));
						if (clocksTmp->block[ind][i].clock!=INVALID_CLOCK) {
							clocksTmp->numblocsValidConcat[GNSS][PRN]++;
							clocksTmp->Concatblock[ind][indK].validSample=1;
						}
					}
				} else {
					//No data for current time
					indK = indJ;
				}
			} else { // No previous SP3 clocks
				if(indTmp!=-1) {
					indK=0;
					MissingSamples=0;
					for ( i = 0; i < clocksTmp->numRecords; i++ ) {
						if (i==0) {
							//The first sample must be compared with the start of the data
							diffSample=tdiff(&clocksTmp->block[ind][i].t,&clocksTmp->startTimeConcat);
						} else {
							diffSample=tdiff(&clocksTmp->block[ind][i].t,&clocksTmp->Concatblock[ind][indK].t);
						}
						while(diffSample>clocksTmp->intervalConcat) {
							//Data gap. Add epochs until all data gaps are filled with 0s
							indK=i+MissingSamples;
							clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indK+1));
							initSP3clockblock(&clocksTmp->Concatblock[ind][indK]);
							if (indK==0) {
								//If it is the first sample saved, the first epoch is the start epoch of data
								memcpy(&clocksTmp->Concatblock[ind][indK].t,&clocksTmp->startTimeConcat,sizeof(TTime));
							} else {
								clocksTmp->Concatblock[ind][indK].t=tdadd(&clocksTmp->Concatblock[ind][indK-1].t,clocksTmp->intervalConcat);
							}
							diffSample=tdiff(&clocksTmp->block[ind][i].t,&clocksTmp->Concatblock[ind][indK].t);
							MissingSamples++;
						}
						indK=i+MissingSamples;
						clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indK+1));
						memcpy(&clocksTmp->Concatblock[ind][indK],&clocksTmp->block[ind][i],sizeof(TSP3clockblock));
						if (clocksTmp->block[ind][i].clock!=INVALID_CLOCK) {
							clocksTmp->numblocsValidConcat[GNSS][PRN]++;
							clocksTmp->Concatblock[ind][indK].validSample=1;
						}
					}
				}
			}
			// Fill arrays with future SP3 clocks
			if ( diffTimeNext >= 85500.0 && diffTimeNext <= 87300.0 ) { //Leave a possible data gap of +-15 minutes (900 seconds)
				//Check if the last entry of previous file and first entry of current file are the same
				indK++;
				if(indTmp==-1) {
					FirstSampleToBeSkipped=0;
				} else if(tdiff(&clocksNext->block[indNext][0].t,&clocksTmp->block[ind][clocksTmp->numRecords-1].t)==0.) {
					//Last entry of next day and first entry of current day are the same
					//Check if last entry of current day has INVALID_CLOCK
					if(clocksTmp->block[ind][clocksTmp->numRecords-1].clock==INVALID_CLOCK) {
						FirstSampleToBeSkipped=0;
						indK--; //Substract 1 to overwrite the previous record
						clocksTmp->numblocsValidConcat[GNSS][PRN]--;
					} else {
						//Last entry of previous day is valid, check if we need to keep the sample from the previous day or from current day
						if(options->LastSamplePrevDayClk==1) {
							//Use Sample from previous day
							FirstSampleToBeSkipped=1;
						} else {
							//Use Sample from next day
							FirstSampleToBeSkipped=0;
							indK--; //Substract 1 to overwrite the previous record
							clocksTmp->numblocsValidConcat[GNSS][PRN]--;
						}
					}
				} else {
					FirstSampleToBeSkipped=0;
				}
				MissingSamples=0;
				if (indPast!=-1 && indTmp==-1) {
					indK--;
				} else if (indPast==-1 && indTmp==-1) {
					indK=0;
				}
				indJ=indK;
				for ( i = FirstSampleToBeSkipped; i < clocksNext->numRecords; i++ ) {
					if (indPast==-1 && indTmp==-1) {
						//No samples from previous day and current day
						diffSample=tdiff(&clocksNext->block[indNext][i].t,&clocksTmp->startTimeConcat);
					} else if(i==FirstSampleToBeSkipped) {
						diffSample=tdiff(&clocksNext->block[indNext][i].t,&clocksTmp->Concatblock[ind][indJ-1].t);
					} else {
						diffSample=tdiff(&clocksNext->block[indNext][i].t,&clocksTmp->Concatblock[ind][indJ].t);
					}
					while(diffSample>clocksTmp->intervalConcat) {
						//Data gap. Add epochs until all data gaps are filled with 0s
						indJ = indK + i + MissingSamples - FirstSampleToBeSkipped;
						clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indJ+1));
						initSP3clockblock(&clocksTmp->Concatblock[ind][indJ]);
						if (indJ==0) {
							//If it is the first sample saved, the first epoch is the start epoch of data
							memcpy(&clocksTmp->Concatblock[ind][indJ].t,&clocksTmp->startTimeConcat,sizeof(TTime));
						} else {
							clocksTmp->Concatblock[ind][indJ].t=tdadd(&clocksTmp->Concatblock[ind][indJ-1].t,clocksTmp->intervalConcat);
						}
						diffSample=tdiff(&clocksNext->block[indNext][i].t,&clocksTmp->Concatblock[ind][indJ].t);
						MissingSamples++;
					}
					indJ = indK + i + MissingSamples -FirstSampleToBeSkipped;
					clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indJ+1));
					memcpy(&clocksTmp->Concatblock[ind][indJ],&clocksNext->block[indNext][i],sizeof(TSP3clockblock));
					if (clocksNext->block[indNext][i].clock!=INVALID_CLOCK) {
						clocksTmp->numblocsValidConcat[GNSS][PRN]++;
						clocksTmp->Concatblock[ind][indJ].validSample=1;
					}
				}
			} else {
				//No future SP3 products
				indJ=indK;
				if (indPast!=-1 && indTmp==-1) {
					indJ--;
				}
				//Check if we need to fill with 0 for the next day
				diffSample=tdiff(&clocksTmp->endTimeConcat,&clocksTmp->Concatblock[ind][indJ].t);
				while(diffSample>0.) {
					indJ++;
					clocksTmp->Concatblock[ind]=realloc(clocksTmp->Concatblock[ind],sizeof(TSP3clockblock)*(indJ+1));
					initSP3clockblock(&clocksTmp->Concatblock[ind][indJ]);
					clocksTmp->Concatblock[ind][indJ].t=tdadd(&clocksTmp->Concatblock[ind][indJ-1].t,clocksTmp->intervalConcat);
					diffSample=tdiff(&clocksTmp->endTimeConcat,&clocksTmp->Concatblock[ind][indJ].t);
				}
			}
			clocksTmp->numblocsConcat[GNSS][PRN]=indJ+1;	
		}//End for PRN loop
		if(RefData==1) break;
	}//End for GNSS loop

	products->type=SP3;
	pastSP3Prod->type=SP3;
	nextSP3Prod->type=SP3;
}

/*****************************************************************************
 * Name        : freeSP3data
 * Description : Free SP3 orbit and clock data. This is necesary when concatening
 *                 SP3/clock files
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TGNSSproducts  *products        I  N/A  TGNSSproducts structure
 *****************************************************************************/
void freeSP3data(TGNSSproducts *products) {

	int i;
	
	//Free orbit data
	for(i=0;i<products->SP3->orbits.numSatellitesRead;i++) {
		free(products->SP3->orbits.block[i]);
	}
	free(products->SP3->orbits.block);
	free(products->SP3->orbits.accuracy);

	products->SP3->orbits.block=NULL;
	products->SP3->orbits.accuracy=NULL;

	//Free clock data
	for(i=0;i<products->SP3->clocks.numSatellitesRead;i++) {
		free(products->SP3->clocks.block[i]);
	}
	free(products->SP3->clocks.block);

	products->SP3->clocks.block=NULL;

	free(products->SP3);

	products->SP3=NULL;
}



/*****************************************************************************
 * Name        : strbintostrhex
 * Description : Convert a binary string to a hexadecimal string
 *               (character by character)
 *               Note: If input string length is not a multiple of 4, it will
 *                     add trailing zeros at the end as necessary
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *binary                   I  N/A  String with the binary values (one bit per byte)
 * char  *hex                      O  N/A  String with the text values
 * int   capital                   I  N/A  To indicate if we want upper case or lower case characters
 *                                          0 => Lower case
 *                                          1 => Upper case
 *****************************************************************************/
void strbintostrhex (char *binary, char *hex, int capital) {

	int 	i,j;
	int 	len;
	int 	modulo;
	int 	ascii;
	char 	aux[2000],aux2[2000];
	char	*endptr;                //This is just for strtol function

	strcpy(aux,binary);	//Copy string to auxiliar variable, so we don't modify the input

	len=strlen(binary);	//Find input length
	modulo=len%4;		//Check if input bit length is a multiple of 4

	if(modulo!=0) {		//If it nos a multiple of 4 add trailing zeros
		for(i=0;i<4-len%4;i++) {
			sprintf(aux2,"%s0",aux);	
			strcpy(aux,aux2);
		}
	}

	j=0;			//j will indicate the position in the output string

	//Loop to process all bits in groups of 4
	for(i=0;i<(int)(len+modulo)/4;i++) {
		getstr(aux2,aux,i*4,4);		//Get 4 bits

		ascii=strtol(aux2,&endptr,2);	//Get the integer value

		if(ascii<10) ascii+=48;		//Add 48 to get the ascii code for 0-9
		else if(capital==1) ascii+=55;	//Add 55 to get the ascii code for A-F
		else ascii+=87;			//Add 87 to get the ascii code for a-f

		hex[j]=ascii;			//Put the ascii value in the hexadecimal string
		j++;
	}

	hex[j]='\0';				//Set the end of the string
}

/*****************************************************************************
 * Name        : SBASnavmode2SBASnavstr
 * Description : Transform from internal SBAS navigation mode to string
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int mode                        I  N/A  SBAS navigation mode
 * Returned value (char*)          O  N/A  SBAS navigation string
 *****************************************************************************/
char *SBASnavmode2SBASnavstr (int mode) {
	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[10];
	#pragma omp threadprivate(str)

	strcpy(str,"GPS");

	switch (mode) {
	case 0:
		strcpy(str,"NPA");
		break;
	case 1:
		strcpy(str,"PA");
		break;
	default:
		strcpy(str,"GPS");
	}

	return str;
}

/*****************************************************************************
 * Name        : strhextostrbin
 * Description : Convert a hexadecimal string to a binary string
 *               (character by character)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *hex                      I  N/A  String with the text values
 * char  *binary                   O  N/A  String with the binary values (one bit per byte)
 *****************************************************************************/
void strhextostrbin (char *hex,char *binary){

	int i,j,k;
	int ascii;		    	//Used to store the ASCII number of a character
	int length=strlen(hex);		//Find the length of the user's input

	char binary_reverse[5];
	
	for(i=0;i<length;i++){		//Repeat until user's input have all been read
					//i<length because the last character is "\0"
	
		ascii = hex[i];  	//Store a character in its ASCII number


		if(ascii>=48&&ascii<=57) {
			//ASCII number for 0-9
			ascii-=48;			//Substract 48 so the integer value stays between 0-9
		} else if(ascii>=65&&ascii<=70) {
			//ASCII code for A-F
			ascii-=55;			//Substract 55 so the integer value stays between 10-15
		} else if (ascii>=97&&ascii<=102) {
			//ASCII code for a-f
			ascii-=87;			//Substract 87 so the integer value stays between 10-15
		} else {
			//Not a valid hexademical character, put 0
			ascii=0;
		}
		
		j=0;    					//Counter used for arrays
		
		while(ascii!= 1) {				//Continue until ascii == 1
			if(ascii==0) {				//If ascii is 0 put a zero and escape while loop
				binary_reverse[j] = '0';	
				break;
			}	
			if(ascii%2== 0) {  			//If ascii is divisible by 2 then put a zero
				binary_reverse[j] = '0';	
			}
			else if(ascii%2==1) {  			//If it isn't divisible by 2 then put a 1
				binary_reverse[j] = '1';   	
			}
			ascii /= 2;				//Find the quotient of ascii / 2
			j++;    				//Add 1 to j for next loop
		}
		
		if(ascii==1) {    				//When ascii is 1, we have to add 1 to the beginning
			binary_reverse[j] = '1';
			j++;
		}
		
		if(j<4) { 					//Add zeros to the end of string if not 4 characters (1/2 byte)
			for(;j<4;j++) {  			//Add until binary_reverse[4] (4th element)
				binary_reverse[j] = '0';
			}
		}
	
		for(k=0;k<4;k++) { 				//Our array is reversed. put the numbers in the right order (last comes first)
			binary[4*i+k] = binary_reverse[3-k];
		}	
	}

	binary[4*i]='\0';					//Set the end of the string
}


/*****************************************************************************
 * Name        : strbintobinstr
 * Description : Convert the binary string (one bit per byte) to 
                   one string where each bit is one value of the message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *binary                   I  N/A  String with the binary values (one bit per byte)
 * int   binarylen                 I  N/A  Length of binary string
 * char  *bitstr                   O  N/A  String with the binary values (one bit per bit)
 * int   bitstrlen                 I  N/A  Length of bitstr string
 *****************************************************************************/
void strbintobinstr (char *binary, int binarylen, unsigned char *bitstr, int bitstrlen) {
	int i;
	int bit;
	
	//Initialize bitstr to 0
	for(i=0;i<bitstrlen;i++) {
		bitstr[i]=0x00;
	}

	for(i=0;i<binarylen;i++){
		if(binary[i]=='0') bit=0;
		else bit=1;
		bitstr[(int)(i/8)]+=bit<<(7-i%8);
	}
}

/*****************************************************************************
 * Name        : twocomplementstrbintointeger
 * Description : Convert a two complement binary value in string 
 *                 (one bit per byte) to an integer number
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char  *binary                   I  N/A  String with the binary values (one bit per byte)
 * int binarylen                   O  N/A  Length of binary string
 * Returned value (integer)        O  N/A  The integer value of two complement binary value
 *****************************************************************************/
int twocomplementstrbintointeger (char *binary, int binarylen) {
	int i;
	char *endptr;		//for strtol function
	char aux[250];		//string to put the complemented number (without the sign bit)

	if(binary[0]=='0') {	//Positive number
		return strtol(binary,&endptr,2);
	} else {		//Negative number
		for(i=1;i<binarylen;i++) {		//Loop for complementing all the bits except the sign bit
			if(binary[i]=='0') {
				aux[i-1]='1';
			} else {
				aux[i-1]='0';
			}
		}
		aux[i-1]='\0';	//Set the end of the string
		return -1*(1+strtol(aux,&endptr,2));
	}
}


/*****************************************************************************
 * Name        : checksumSBAS
 * Description : Compute CRC-24Q parity in SBAS message
 * Parameters  :
 * Name                           |Da|Unit|Description
 * unsigned char *msg              I  N/A  String with the binary values (one message bit per string bit)
 * Returned value (int)            O  N/A  Returns the value of the computed CRC
 *****************************************************************************/
int checksumSBAS (unsigned char *msg) {

	int key=0x1864CFB;						//25 bit polynomial for this CRC, with bits 0,1,3,4,5,6,7,10,11,14,17,18,23,24 equal to 1
	int chk=0;							//Contains the value of the checksum
	int pos=0;							//Indicates the bit number we are reading in the whole message (starting in bit 0)						
	int i;
	int maxpos=250;							//The total message is 250 bits long (including the CRC bits)
	int end=0;
	int position,shift;
	unsigned char aux[32];

	for (i=0;i<32;i++) aux[i]=0;					//Initialize aux to 0
	for (i=0;i<28;i++) aux[i]=*( (unsigned char *)(msg  + i));	//Copy first 224 bits of message to aux
	aux[28] = *( (unsigned char *)(msg  + 28)) & 0xC0;		//Copy last 2 bits of message to aux (which are the first two bits of aux[28]) (0xC0=11000000)
									//The data message is 224 bits, where CRC has to be computed. The rest of the bits of aux are 0
	while (!end) {
		while ((chk & 0x1000000)==0 && !end) {			//(chk & 0x1000000)==0 means bit 25 of chk equal to 0.  !end means to not have arrived to the end of the message (not arrived to bit 250)
									//This while is true while we don't arrive to the last bit or when bit 25 of chk (checksum) is not 0, which means
									//that our current value of chk has more than 24 bits, which is the maximum number for this CRC
				chk=chk<<1;					//Move chk one bit to the right (equivalent to multiply chk per 2)
			shift=7-(pos%8);				//shift indicates which bit in the current byte of the message we want to read (from right to left)
			position=1<<shift;				//position=2^shift. Position is the mask to apply in order to read the current bit (all the others bits in the byte will be 0)
			chk+=(aux[(int)(pos/8)] & position)>>shift;	//Select the current byte of the message, apply the mask and move the bit to the the least significant position,
									//(so the integer value is 0 or 1), and then add this value to chk
			pos++;						//pos indicates the current bit we are reading. Add 1 in order to read the next bit.
			if (pos==maxpos) end=1;				//If we have arrived to the last bit set end=1
			}
		if ((chk & 0x1000000)!=0) chk=chk^key;			//(chk & 0x1000000)!=0 means bit 25 of chk not equal 0 (bit 25=1). chk=chk^key means chk = chk XOR key
									//Doing this XOR, bit 25 will be 0 again, so the checksum will be 24 bits long. If we have not arrived to the end
									//of the message, we will continue reading bits and computing the checksum
	}

	return chk;
}

/*****************************************************************************
 * Name        : openSBASwritefiles
 * Description : Open all the files needed for writing the SBAS data
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  O  N/A  Vector of file descriptors for writing data
 * char  **filelist                O  N/A  Name of the SBAS files that will be written
 * char  *fileread                 O  N/A  Name of the SBAS file that is going to be read
 * int sourcefile                  I  N/A  Indicates from what type file the SBAS data has been read.
 *                                          0 => RINEXB
 *                                          1 => EMS
 * TOptions  *options              I  N/A  TOptions structure
 * Returned value (int)            O  N/A  Status of the function
 *                                         1  => Files opened properly or no files need to be opened
 *                                         0  => Problem reading EMS file when searching for year
 *                                         -1 => Problems opening EMS file when searching for year
 *****************************************************************************/
int openSBASwritefiles (FILE **fdlist, char **filelist, char *fileread, int sourcefile, TOptions  *options) {

	int				i;
	int 			len=0;
	int 			ret;
	int				doy;
	int				year,yearprev;
	int				GPSweek;
	double			SoW;
	double  		doyprev;
	TTime   		t,tprev;
	FILE			*fd;
	char   		 	line[MAX_INPUT_LINE];
	char 			aux[MAX_MESSAGE_STR];
	char			name[MAX_INPUT_LINE];
	char			fullname[MAX_INPUT_LINE];
	char			path[MAX_INPUT_LINE];
	char   			extensionlist[12][5]={".smt",".s00",".s01",".sfc",".s06",".s07",".s10",".s18",".ssc",".s26",".s27",".s28"};
	char 		  	*filepointer;
	struct 	stat 	path_stat;

	//Initialize filelist vector and file descriptor vector
	for(i=0;i<18;i++) {
		fdlist[i]=NULL;						//Set file descriptors to NULL
		filelist[i]=malloc(sizeof(char)*MAX_INPUT_LINE); 	//Allocate space for the filelist
		filelist[i][0]='\0';					//Set the filelist string to empty
	}

		//Check if any file has to be written.
		if (options->writeRinexBfile==0 && options->writeEMSfile==0 && options->writePegasusfile==0) return 1;
	
	//Check if output directory has a slash or backslash at the end, and if it has one, take it out.
	if(options->sbasdir[strlen(options->sbasdir)-1]=='/'||options->sbasdir[strlen(options->sbasdir)-1]=='\\') {
		options->sbasdir[strlen(options->sbasdir)-1]='\0';
	}

	//First we have to check if SBAS directory exists, where data will be saved
	//Create directory recursively if needed
	ret=mkdir_recursive(options->sbasdir);
	if(ret!=0) {
	   //Error when creating the directory, and it is not due to that the directory already exists
	   sprintf(aux,"When creating '%s' folder for writing SBAS data",options->sbasdir);
	   printError(aux,options);
	}
	//mkdir_recursive assumes the last part is a file, so it does not create a folder. Create it now
	#if !defined (__WIN32__)
			ret=mkdir(options->sbasdir,S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH); //Permission of the directory set to 755
	#else
			ret=mkdir(options->sbasdir);
	#endif
	if(ret!=0 && errno!=EEXIST) {
	   //Error when creating the directory, and it is not due to that the directory already exists
	   sprintf(aux,"When creating '%s' folder for writing SBAS data",options->sbasdir);
	   printError(aux,options);
	} else if (ret!=0 && errno==EEXIST) {
		//Check that what exists is a directory and not a file
		stat(options->sbasdir, &path_stat);
		if (!S_ISDIR(path_stat.st_mode)) {
		   return -1;
		}
	}
	//Save the path in path variable (Depending if it Windows or Linux it will have slash or backslash)
	strcpy(path,options->sbasdir);
	#if !defined (__WIN32__)
		   strcpy(&path[strlen(path)],"/");
	#else
		   strcpy(&path[strlen(path)],"\\");
	#endif
		

	//Extract the name of the file without the extension and the path
	#if defined (__WIN32__)
		filepointer=strrchr(fileread,'\\');
	#elif defined (__CYGWIN__)
		filepointer=strrchr(fileread,'/');
		//In cygwin, we may have linux or windows pathes
		if(filepointer==NULL) {
			filepointer=strrchr(fileread,'\\');
		}
	#else
		filepointer=strrchr(fileread,'/');
	#endif

	if (filepointer==NULL) {
		//There is no directory path in the file
		getstr(name,fileread,0,8);
		strcpy(fullname,fileread);
	} else {
		//There are directories in the path
		getstr(name,filepointer,1,8);
		strcpy(fullname,(char *)(filepointer+sizeof(char)));
	}
	//Check if a RINEX B file has to be written
	if(options->writeRinexBfile==1) {
		if(sourcefile==0) {
			//The data file is a RINEXB, the new file will have the same filename but with a ".v" at the end (v stands for verified)
			strcpy(filelist[fdRINEXB],path);		//Put path in filelist
			len=strlen(path);
			strcpy(&filelist[fdRINEXB][len],fullname);	//Put original name in filename
			len=strlen(filelist[fdRINEXB]);
			strcpy(&filelist[fdRINEXB][len],".v");		//Add the ".v" at the end of the filename
		} else {
			//The data file is a EMS file. We have to change the extension from ".ems" to ".yyb.v", being yy the year
			//To get the year, we have to read the first line of the EMS file
			if ((fd = fopen(fileread,"r"))) {
					if(getL(line,&len,fd)!=-1) {
					getstr(aux,line,4,2);
					year=atoi(aux);
					fclose(fd);
				} else {
					//Error reading EMS file
					return 0;
				}
			} else {
				//Error opening EMS file
				return -1;	
			}
			strcpy(filelist[fdRINEXB],path);		//Put path in filename
			len=strlen(path);
			strcpy(&filelist[fdRINEXB][len],name);		//Put original name (without extension) in filename
			len=strlen(filelist[fdRINEXB]);
			sprintf(aux,".%02db.v",year);
			strcpy(&filelist[fdRINEXB][len],aux);		//Add the ".v" at the end of the filename
		}
		fdlist[fdRINEXB]=fopen(filelist[fdRINEXB],"w");
		if (fdlist[fdRINEXB]==NULL) {
			//No need to close files here because this was the first one to be opened
			sprintf(aux,"When creating [%s] file for writing RINEX B file",filelist[fdRINEXB]);
			printError(aux,options);
		}
		
	}

	//Check if a EMS file has to be written
	if(options->writeEMSfile==1) {
		if(sourcefile==0) {
				//The data file is a RINEXB, the new file will have the same filename but with a ".v" at the end (v stands for verified)
				strcpy(filelist[fdEMS],path);			//Put path in filelist
				len=strlen(path);
				strcpy(&filelist[fdEMS][len],name);		//Put original name (without extension) in filename
				len=strlen(filelist[fdEMS]);
				strcpy(&filelist[fdEMS][len],".ems.v");		//Add the ".v" at the end of the filename
		} else {
				//The data file is a EMS file. The new file will have the extension ".ems.v" (v stands for verified)
				strcpy(filelist[fdEMS],path);			//Put path in filename
				len=strlen(path);
				strcpy(&filelist[fdEMS][len],fullname);		//Put original name in filename
				len=strlen(filelist[fdEMS]);
				strcpy(&filelist[fdEMS][len],".v");		//Add the ".v" at the end of the filename
		}
		fdlist[fdEMS]=fopen(filelist[fdEMS],"w");
		if (fdlist[fdEMS]==NULL) {
				closefiles(fdlist,1);	//Close all other files before exiting (The only one open will be the RINEX B file)
				sprintf(aux,"When creating [%s] file for writing EMS file",filelist[fdEMS]);
				printError(aux,options);
			}	
		}

	//Check if a Pegasus file has to be written
	if(options->writePegasusfile==1) {
		//Get DoY from filename
		getstr(aux,fullname,4,3);
		doy=atoi(aux);

		if(sourcefile==0) {
			//The data file is a RINEXB, we can extract the year directly from the filename
			getstr(aux,fullname,9,2);
			year=atoi(aux)+2000;
		} else {
			//The data file is a EMS file, we need to read the first line of the file to extract the time in order to get the year and the GPS week
			if ((fd = fopen(fileread,"r"))) {
				if(getL(line,&len,fd)!=-1) {
					getstr(aux,line,4,2);
					year=atoi(aux)+2000;
					fclose(fd);
				} else {
					//Error reading EMS file
					return 0;
				}
			} else {
				//Error opening EMS file
				return -1;	
			}
		}
		//We need to find the GPS Week from Year/DoY.
		t.SoD=0;				//The second of day will be 0 because we only need the week number for the filename
		t.MJDN=(int)yeardoy2MJDN(year,doy,t.SoD);	//First get the modified julian date
		ttime2gpswsnoroll(&t, &GPSweek, &SoW);	//Get the GPS Week number

		//Open SBAS Ephemeris file
		strcpy(filelist[fdEPHEMERIS],path);		//Put path in filename
		len=strlen(path);
		strcpy(&filelist[fdEPHEMERIS][len],name);	//Put original name (without extension) in filename
		len=strlen(filelist[fdEPHEMERIS]);
		sprintf(aux,"_BRDC%03d0.%02dH",doy,year%100);
		strcpy(&filelist[fdEPHEMERIS][len],aux);	//Add the _BRDCXXX0.YYH at the end of the filename
		fdlist[fdEPHEMERIS]=fopen(filelist[fdEPHEMERIS],"w");
		if (fdlist[fdEPHEMERIS]==NULL) {
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When creating [%s] file for writing BRDC Pegasus file",filelist[fdEPHEMERIS]);
			printError(aux,options);
		}

		//We need to find the year/doy from the previous day (taking into account leap years)
		tprev.MJDN=--t.MJDN;
		tprev.SoD=0;
		t2doy(&tprev,&yearprev,&doyprev);

		//Open SBAS Ephemeris file for previous day
		strcpy(filelist[fdEPHEMERISPREVDAY],path);		//Put path in filename
		len=strlen(path);
		strcpy(&filelist[fdEPHEMERISPREVDAY][len],name);	//Put original name (without extension) in filename
		len=strlen(filelist[fdEPHEMERISPREVDAY]);
		sprintf(aux,"_BRDC%03d0.%02dH",(int)doyprev,yearprev%100);
		strcpy(&filelist[fdEPHEMERISPREVDAY][len],aux);	//Add the _BRDCXXX0.YYH at the end of the filename
		fdlist[fdEPHEMERISPREVDAY]=fopen(filelist[fdEPHEMERISPREVDAY],"w");
		if (fdlist[fdEPHEMERISPREVDAY]==NULL) {
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When creating [%s] file for writing BRDC Pegasus file",filelist[fdEPHEMERISPREVDAY]);
			printError(aux,options);
		}

		//Open SBAS Almanac file
		strcpy(filelist[fdALMANAC],path);		//Put path in filename
		len=strlen(path);
		strcpy(&filelist[fdALMANAC][len],name);		//Put original name (without extension) in filename
		len=strlen(filelist[fdALMANAC]);
		sprintf(aux,"_GEO%04d.alm",GPSweek);
		strcpy(&filelist[fdALMANAC][len],aux);		//Add the _GEOYYYY.alm at the end of the filename (YYYY is the GPS Week)
		fdlist[fdALMANAC]=fopen(filelist[fdALMANAC],"w");
		if (fdlist[fdALMANAC]==NULL) {
			closefiles(fdlist,18);	//Close all other files before exiting
			sprintf(aux,"When creating [%s] file for writing Almanac Pegasus file",filelist[fdALMANAC]);
			printError(aux,options);
		}


		for(i=2;i<14;i++) {
			//Open the rest of the files. The extension for these files are fixed
			strcpy(filelist[i],path);		//Put path in filename
			len=strlen(path);
			strcpy(&filelist[i][len],name);		//Put original name (without extension) in filename
			len=strlen(filelist[i]);
			strcpy(&filelist[i][len],extensionlist[i-2]);		//Add the extension at the end of the filename
			fdlist[i]=fopen(filelist[i],"w");
			if (fdlist[i]==NULL) {
				closefiles(fdlist,18);	//Close all other files before exiting
				sprintf(aux,"When creating [%s] file for writing Pegasus files",filelist[i]);
				printError(aux,options);
			}
		}
	

	}

	//Open log file
	strcpy(filelist[fdLOGFILE],path);			//Put path in filename
	len=strlen(path);
	strcpy(&filelist[fdLOGFILE][len],fullname);		//Put original name in filename
	len=strlen(filelist[fdLOGFILE]);
	strcpy(&filelist[fdLOGFILE][len],".log");		//Add the ".log" at the end of the filename

	fdlist[fdLOGFILE]=fopen(filelist[fdLOGFILE],"w");
	if (fdlist[fdLOGFILE]==NULL) {
		closefiles(fdlist,18);	//Close all other files before exiting
		sprintf(aux,"When creating [%s] file for writing log file",filelist[fdLOGFILE]);
		printError(aux,options);
	}

	return 1;
}

/*****************************************************************************
 * Name        : closefiles
 * Description : Close all the files given in the file descriptor list
 * Parameters  :
 * Name                           |Da|Unit|Description
 * FILE  **fdlist                  IO N/A  Vector of file descriptors to be closed
 * int numfiles                    I  N/A  Number of files to be closed
 *****************************************************************************/
void closefiles (FILE  **fdlist, int numfiles) {

	int i;

	for(i=0;i<numfiles;i++) {
		if(fdlist[i]!=NULL) fclose(fdlist[i]);
	}
	
}

/*****************************************************************************
 * Name        : freeSBASblock
 * Description : Free memory sbasblock struct to avoid memory leak
 *               (each call to malloc needs its correspondent free call)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int  messagetype                I  N/A  Message type identifier number
 * TSBASblock  *sbasblock          O  N/A  TSBASblock struct
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void freeSBASblock (int messagetype,TSBASblock *sbasblock,TOptions  *options) {

	int i,num;

	switch(messagetype) {
		case(PRNMASKASSIGNMENTS):
			free(sbasblock->pos2PRN);
			free(sbasblock->pos2GNSS);
			for(i=0;i<MAX_GNSS;i++) {
				free(sbasblock->PRNactive[i]);
				free(sbasblock->PRN2pos[i]);
			}
			free(sbasblock->PRNactive);
			free(sbasblock->PRN2pos);
			break;
		case(DONTUSE):
			if(options->alarmmessageastype2==0) break;
			/* Falls through. */ //To avoid warning -Wimplicit-fallthrough=
		case(FASTCORRECTIONS2):case(FASTCORRECTIONS3):case(FASTCORRECTIONS4):case(FASTCORRECTIONS5):
			free(sbasblock->IODF);
			free(sbasblock->PRC);
			free(sbasblock->UDREI);
			free(sbasblock->UDRE);
			free(sbasblock->UDREsigma);
			break;
		case(INTEGRITYINFO):
			free(sbasblock->IODF);
			free(sbasblock->UDREIacu);
			free(sbasblock->UDREacu);
			free(sbasblock->UDREacusigma);
			break;
		case(FASTCORRECTIONSDEGRADATIONFACTOR):
			free(sbasblock->aiind);
			free(sbasblock->aifactor);
			free(sbasblock->timeoutintervalnonprecise);
			free(sbasblock->timeoutintervalprecise);
			free(sbasblock->fastcorrupdateinterval);
			break;
		case(DEGRADATIONPARAMETERS):
			free(sbasblock->degradationfactors);
			break;
		case(LONGTERMSATELLITECORRECTIONS):
			for(i=0;i<sbasblock->numlongtermsaterrcorrections;i++) {
				free(sbasblock->longtermsaterrcorrections[i]);
			}
			free(sbasblock->longtermsaterrcorrections);
			break;
		case(MIXEDFASTLONGTERMCORRECTIONS):
			free(sbasblock->IODF);
			free(sbasblock->PRC);
			free(sbasblock->UDREI);
			free(sbasblock->UDRE);
			free(sbasblock->UDREsigma);
			
			for(i=0;i<sbasblock->numlongtermsaterrcorrections;i++) {
				free(sbasblock->longtermsaterrcorrections[i]);
			}
			free(sbasblock->longtermsaterrcorrections);
			break;
		case(GEONAVIGATIONMESSAGE):
			free(sbasblock->geonavigationmessage);
			break;
		case(GEOSATELLITEALMANACS):
			if(sbasblock->numgeoalmanacs==0) num=1;
			else num=sbasblock->numgeoalmanacs;

			for(i=0;i<num;i++) {
				free(sbasblock->geoalmanacsmessage[i]);
			}
			free(sbasblock->geoalmanacsmessage);
			break;
		case(SBASSERVICEMESSAGE):
			if(sbasblock->numregioncoordinates==0) num=1;
			else num=sbasblock->numregioncoordinates;

			free(sbasblock->servicemessage);
			for(i=0;i<num;i++) {
				free(sbasblock->regioncoordinates[i]);
			}
			free(sbasblock->regioncoordinates);
			break;
		case(SBASNETWORKTIMEPARAMETERS):
			free(sbasblock->networktimemessage);
			break;
		case(CLOCKEPHEMERISCOVARIANCEMATRIX):
			if(sbasblock->numclockephemeriscovariance==0) num=1;
			else num=sbasblock->numclockephemeriscovariance;

			for(i=0;i<num;i++) {
				free(sbasblock->clockephemeriscovariance[i]);
			}			
			free(sbasblock->clockephemeriscovariance);
			break;
		case(IONOSPHERICGRIDPOINTSMASKS):
			free(sbasblock->igpmaskmessage);
			break;
		case(IONOSPHERICDELAYCORRECTIONS):
			for(i=0;i<15;i++) {
				free(sbasblock->ionodelayparameters[i]);
			}
			free(sbasblock->ionodelayparameters);
			break;
		default:
			break;
	}
}


/*****************************************************************************
 * Name        : SBASMissingMessageClearData
 * Description : When 4 or more consecutive messages are missed, all fast 
                 corrections data for all PRN have to be erased and UDREI set
                 14 (Don't Use)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int GEOindex                    I  N/A  Index of the GEO which had 4 or more missed messages
 * TSBASdata  *SBASdata            IO N/A  Struct to save SBAS data
 *****************************************************************************/
void SBASMissingMessageClearData (int GEOindex, TSBASdata  *SBASdata) {

	int i,j,k;

	//We have lost 4 or more consecutive messages
	//We have to assume that sigma UDRE for all satellites are set to "Not monitored". Thus, we have to erase all data from all satellites
	//Loop through all Masks
	for(i=0;i<5;i++) {
		//Loop through all SBAS fast corrections
		for(j=0;j<MAXSBASFASTCORR;j++) {
			//Loop for all satellites in mask
			for(k=0;k<51;k++) {
				SBASdata[GEOindex].poslastFC[i][k] = -1;
				SBASdata[GEOindex].numFC[i][k] = 0;
				SBASdata[GEOindex].PRC[i][j][k] = 9999;
				SBASdata[GEOindex].RRC[i][j][k] = 9999;
				SBASdata[GEOindex].UDRE[i][j][k] = -1;
				SBASdata[GEOindex].UDREsigma[i][j][k] = -1;
				SBASdata[GEOindex].IODF[i][j][k] = -1;
				SBASdata[GEOindex].lastfastcorrections[i][j][k].MJDN = -1;
				SBASdata[GEOindex].lastfastcorrections[i][j][k].SoD = -1;
				if (j==0 && i==0) {
					SBASdata[GEOindex].lastfastmessage[k]=-1;
					SBASdata[GEOindex].IODFintegrity[k]=-1;
					SBASdata[GEOindex].oldIODFintegrity[k]=-1;
					SBASdata[GEOindex].UDREI6[k]=-1;
					SBASdata[GEOindex].UDRE6[k]=-1;
					SBASdata[GEOindex].UDREsigma6[k]=-1;
					SBASdata[GEOindex].oldUDREI6[k]=-1;
					SBASdata[GEOindex].oldUDRE6[k]=-1;
					SBASdata[GEOindex].oldUDREsigma6[k]=-1;
					SBASdata[GEOindex].lastintegrityinfo[k].MJDN = -1;
					SBASdata[GEOindex].lastintegrityinfo[k].SoD = -1.;
					SBASdata[GEOindex].oldlastintegrityinfo[k].MJDN = -1;
					SBASdata[GEOindex].oldlastintegrityinfo[k].SoD = -1.;
				}

				if(i==SBASdata[GEOindex].IODPfastcorr) {
					SBASdata[GEOindex].useforbidden[SBASdata[GEOindex].pos2GNSS[i][k]][SBASdata[GEOindex].pos2PRN[i][k]]=14;
				}
			}
		}
	}
}		

/*****************************************************************************
 * Name        : SBASErrorMessageTimeoutFastCorr
 * Description : When 4 or more consecutive messages are wrong, we need to timeout
 *                the fast corrections (for PA only) for this GEO and set it as not
 *                monitored (not available for navigation).
 *                This behaviour is done by following the following rules in MOPS-D
 *
 *                2.1.1.4.9 Message Timeout Periods:
 *                  For approach LNAV/VNAV, LP, LPV when no valid SBAS message has been received
 *                  for 4 seconds (indicating a probable communications link problem or SBAS signal
 *                  blockage), all UDREI data from that SBAS satellite shall timeout
 *
 *                2.1.1.5.2 SBAS UNHEALTHY Designation
 *                  The equipment shall designate any GPS or SBAS satellite as SBAS UNHEALTHY upon
 *                  the occurrence of any of the following conditions:
 *                   a) The equipment has successfully decoded a UDREI of 15, indicating that the
 *                      SBAS has assessed the satellite’s signal as unusable;
 *                   b) The step detection function has declared a step error;
 *                   c) For SBAS satellites, URA index is 15 or higher; or
 *                   d) For SBAS satellites, failure of parity on 4 successive messages.
 *                   
 *                 The SBAS UNHEALTHY status for that satellite shall be changed only after the
 *                 condition has cleared (including time-out of UDREI data) and none of the above
 *                 conditions exist. When an SBAS satellite is designated as UNHEALTHY due to any one
 *                 of the above conditions, the integrity and correction data can continue to be applied.
 *                
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int GEOindex                    I  N/A  Index of the GEO which had 4 or more missed messages
 * TTime *currentepoch             I  N/A  Current epoch
 * TSBASdata  *SBASdata            IO N/A  Struct to save SBAS data
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void SBASErrorMessageTimeoutFastCorr (int GEOindex, TTime *currentepoch, TSBASdata  *SBASdata, TOptions  *options) {
	int		i,j,k,l;
	double	timediffMsg,timediffTimeOut;
	double	UDREItmout;

	//UDREI data hast to timeout only on PA mode, so the reception time of the fast corrections will set just in the epoch where
	//they are already timed out
	UDREItmout=(double)(SBASdata[GEOindex].timeoutmessages[PAMODE][INTEGRITYINFO]);

	//Forced timeout of UDREI data is only for PA, so if user forced NPA mode do nothing
	if (options->UserForcedNPA==0) {
		//Loop through all Masks
		for(i=0;i<5;i++) {
			//Loop through all SBAS fast corrections
			for(j=0;j<MAXSBASFASTCORR;j++) {
				//Loop for all satellites in mask
				for(k=0;k<51;k++) {
					if (SBASdata[GEOindex].lastfastcorrections[i][j][k].MJDN!=-1) {
						timediffMsg=tdiff(currentepoch,&SBASdata[GEOindex].lastfastcorrections[i][j][k]);
						timediffTimeOut=timediffMsg-UDREItmout;
						if (timediffTimeOut<=1E-4) { //The 1E-4 is to avoid problems with decimals
							//UDRE data not timed out. Change the message reception time so during the modelling
							//process all UDRE data is timed out
							//The .1 second is to make message reception be one tenth of a second behind of the timeout (so for 0.1 data rate files they also timeout)
							SBASdata[GEOindex].lastfastcorrections[i][j][k]=tdadd(&SBASdata[GEOindex].lastfastcorrections[i][j][k],timediffTimeOut-.1);
						}//If data is timed out, leave the reception time as it is
					}
				}
			}
		}
	}
	//Set GEO to do not use (not available for ranging)
	if (GEOindex==0) {
		//If it is the mixed GEO, we need to set all GEOs to do not use (not available for ranging)
		for(l=1;l<=SBASdata[0].numSBASsatellites;l++) { //Loop through all GEOs
			SBASdata[GEOindex].useforbidden[GEO][SBASdata[l].PRN]=15;
		}
	} else {
		SBASdata[GEOindex].useforbidden[GEO][SBASdata[GEOindex].PRN]=15;
	}
}

/*****************************************************************************
 * Name        : SBASGEOselection
 * Description : Select a GEO according to user options. The selection has to
 *                 be done in a separate function (this one) in single threaded
 *                 execution, to avoid race conditions.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TSBASdata  *SBASdata            IO N/A  Struct to save SBAS data
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void SBASGEOselection (TEpoch  *epoch, TSBASdata  *SBASdata, TOptions  *options) {

	int 		i,j;
	int 		numGEOcoord=0,satGEOindex,numGEOelev,maxelevationpos=-1;
	int			GEOSatRead[MAXGEOPRN];
	static int	numGEOelevprev=1;
	double 		difftime, position[3];
	double 		elevation,azimuth,maxelevation=-999.;
	char		ElevText[MAX_NUM_SBAS][MAX_INPUT_LINE];
	
	if(options->GEOPRNSelected!=2*MAXGEOPRN) {
		if(SBASdata[0].numSBASsatellites==0) {
			//No GEO available. Leave options->GEOindex negative to keep checking this condition
			options->GEOindex=-1;
			return;
		} else if (options->GEOPRNSelected==0) {
			//Use data from all GEO
			options->GEOindex=0;
			options->GEOPRN=SBASdata[options->GEOindex].PRN;
			epoch->currentGEOPRN=SBASdata[options->GEOindex].PRN;
			options->GEOPRNSelected=2*MAXGEOPRN;
		} else if (options->GEOPRNSelected==1) {
			//Use the GEO PRN from the first SBAS data read, which will be in position one of the SBASdata vector
			options->GEOindex=1;
			options->GEOPRN=SBASdata[options->GEOindex].PRN;
			epoch->currentGEOPRN=SBASdata[options->GEOindex].PRN;
			options->GEOPRNSelected=2*MAXGEOPRN;
		} else if (options->GEOPRNSelected==2||options->GEOPRNSelected==-1) {
			//If it is the first epoch, fill station orientation (needed for computing elevation)
			if (tdiff(&epoch->t,&epoch->firstepoch)<1E-4) {
				// Station NEU coordinates
				fillGroundStationNEU(epoch);
				// Station orientation
				fillGroundStationOrientation(epoch);
			}

			//We have to use the satellite with highest elevation at the moment we have an almanac message available
			//Reset flag to indicate if a GEO elevation has already been computed with an almanac message
			for(i=0;i<MAXGEOPRN;i++) {
				GEOSatRead[i]=0;
			}
			//Read all GEO almanacs available
			numGEOelev=0;
			for(i=1;i<=SBASdata[0].numSBASsatellites;i++) {
				//Check GEO is not under alarm
 				if (SBASdata[0].alarmGEOPRN[SBASdata[i].PRN]==SBASdata[i].PRN && SBASdata[i].alarmTimeRemaining>0 && GEOSatRead[SBASdata[i].PRN]==0) {
     				sprintf(ElevText[numGEOelev],"SBAS GEO %3d is under alarm",SBASdata[i].PRN);
     				GEOSatRead[SBASdata[i].PRN]=1;
     				numGEOelev++;
     				continue;
 				}
				//Check if we have an available almanac message
				if(SBASdata[i].lastmsgreceived[GEOSATELLITEALMANACS].MJDN!=-1) {
					//Almanac message available
					for(j=0;j<SBASdata[i].numgeoalmanacs;j++) {
						//Compute GEO elevation for all GEOs in almanac message
						//Check GEO is not under alarm
						if (SBASdata[0].alarmGEOPRN[(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER]]==(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER] &&
							SBASdata[SBASdata[0].GEOPRN2pos[(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER]]].alarmTimeRemaining>0) {
							sprintf(ElevText[numGEOelev],"SBAS GEO %3d is under alarm",(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER]);
							GEOSatRead[(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER]]=1;
							numGEOelev++;
							continue;
						}
						satGEOindex=SBASdata[0].GEOPRN2pos[(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER]];
						//Check that we have data for the GEO given in the almanac. Otherwise we would take
						//the risk that the GEO with highest elevation in the almanac is not provided in the
						//SBAS file, therefore selecting a GEO that would not have any data
						if (satGEOindex!=-1) {
							if (GEOSatRead[(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER]]==0) {
								//GEO elevation has not been computed before
								difftime=epoch->t.SoD-SBASdata[i].geoalmanacsmessage[j][T0ALMANAC];
								//Check for day rollover in time difference. As a threshold, we use the second at half of the day (86400/2=43200)
								if(difftime>43200) difftime-=86400; 	//This occurs when we are at the end of the day and the time of applicability is in the next day
								if(difftime<-43200) difftime+=86400;	//This occurs when we are at the beginning of the day and the time of applicability is in the previous day (because we have missed the previous message)
								position[0]=SBASdata[i].geoalmanacsmessage[j][XGALMANAC]+SBASdata[i].geoalmanacsmessage[j][XGALMANACRATEOFCHANGE]*difftime;
								position[1]=SBASdata[i].geoalmanacsmessage[j][YGALMANAC]+SBASdata[i].geoalmanacsmessage[j][YGALMANACRATEOFCHANGE]*difftime;
								position[2]=SBASdata[i].geoalmanacsmessage[j][ZGALMANAC]+SBASdata[i].geoalmanacsmessage[j][ZGALMANACRATEOFCHANGE]*difftime;
							
								getAzimuthElevation(epoch->receiver.orientation, epoch->receiver.aproxPosition, position, &azimuth, &elevation);
								sprintf(ElevText[numGEOelev],"SBAS GEO %3d available with elevation %5.2f degrees",(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER],elevation*r2d);
								GEOSatRead[(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER]]=1;
								if(elevation>maxelevation) {
									//The elevation is higher than the satellites computed in the previous ones
									maxelevation=elevation;
									maxelevationpos=numGEOelev;
									options->GEOindex=satGEOindex;
									epoch->currentGEOPRN=SBASdata[options->GEOindex].PRN;
								}
								numGEOelev++;
							}
						}
					}
				}
			}
			//NOTE: If in the almanac message there is missing data from the GEO from which we have received data, this GEO 
			if(options->GEOindex<0) {
				//We still have no GEO selected, it could be due to we do not have an almanac message or due to the almanac message did not contain enough data
				//Try to use the GEO navigation messages if we have them all
				//Reset flag to indicate if a GEO elevation has already been computed with an almanac message
				for(i=0;i<MAXGEOPRN;i++) {
					GEOSatRead[i]=0;
				}
				for(i=0;i<numGEOelev;i++) {
					ElevText[i][0]='\0';
				}
				//Check that we have GEO navigation message from all GEO from which we have received data
				for(i=1;i<=SBASdata[0].numSBASsatellites;i++) {
					//Check GEO is not under alarm
					if (SBASdata[0].alarmGEOPRN[SBASdata[i].PRN]==SBASdata[i].PRN && SBASdata[i].alarmTimeRemaining>0) {
						numGEOcoord++;
						continue;
					}
					if(SBASdata[i].lastmsgreceived[GEONAVIGATIONMESSAGE].MJDN==-1) {
						difftime=tdiff(&epoch->t,&SBASdata[i].lastmsgreceived[GEONAVIGATIONMESSAGE]);
						if(difftime>(double)(SBASdata[i].timeoutmessages[options->precisionapproach][GEONAVIGATIONMESSAGE]) ) {
							SBASdata[i].lastmsgreceived[GEONAVIGATIONMESSAGE].MJDN=-1;
							break;
						} else {
							numGEOcoord++;
						}
					} else {
						break;
					}
				}
				if(numGEOcoord==SBASdata[0].numSBASsatellites) {
					//We have all the navigation messages
					numGEOelev=0;
					for(i=1;i<=SBASdata[0].numSBASsatellites;i++) {
						//Check GEO is not under alarm
						if (SBASdata[0].alarmGEOPRN[SBASdata[i].PRN]==SBASdata[i].PRN && SBASdata[i].alarmTimeRemaining>0) {
							sprintf(ElevText[numGEOelev],"SBAS GEO %3d is under alarm",SBASdata[i].PRN);
							numGEOelev++;
							continue;
						}
						//Compute GEO elevation
						difftime=epoch->t.SoD-SBASdata[i].geonavigationmessage[T0NAV];
						//Check for day rollover in time difference. As a threshold, we use the second at half of the day (86400/2=43200)
						if(difftime>43200) difftime-=86400; 	//This occurs when we are at the end of the day and the time of applicability is in the next day
						if(difftime<-43200) difftime+=86400;	//This occurs when we are at the beginning of the day and the time of applicability is in the previous day (because we have missed the previous message)
						position[0]=SBASdata[i].geonavigationmessage[XG]+SBASdata[i].geonavigationmessage[XGRATEOFCHANGE]*difftime+0.5*SBASdata[i].geonavigationmessage[XGACCELERATION]*difftime*difftime;
						position[1]=SBASdata[i].geonavigationmessage[YG]+SBASdata[i].geonavigationmessage[YGRATEOFCHANGE]*difftime+0.5*SBASdata[i].geonavigationmessage[YGACCELERATION]*difftime*difftime;
						position[2]=SBASdata[i].geonavigationmessage[ZG]+SBASdata[i].geonavigationmessage[ZGRATEOFCHANGE]*difftime+0.5*SBASdata[i].geonavigationmessage[ZGACCELERATION]*difftime*difftime;

						getAzimuthElevation(epoch->receiver.orientation, epoch->receiver.aproxPosition, position, &azimuth, &elevation);
						sprintf(ElevText[numGEOelev],"GEO %3d available with elevation %5.2f degrees",(int)SBASdata[i].geoalmanacsmessage[j][PRNNUMBER],elevation*r2d);
						if(elevation>maxelevation) {
							//The elevation is higher than the satellites computed in the previous ones
							maxelevation=elevation;
							maxelevationpos=numGEOelev;
							options->GEOindex=i;
							epoch->currentGEOPRN=SBASdata[options->GEOindex].PRN;
						}
						numGEOelev++;
					}
				} else {
					//Missing almanac data or GEO navigation messages. Leave options->GEOindex negative in order to keep checking this condition
					options->GEOindex=-3;
					return;
				}
			}
			//Check if solution has converged. If not, do not mark the GEO as selected, so when
			//solution has converged, the GEO with higher elevation is selected
			if (epoch->receiver.receiverSolConverged==1) {
				options->GEOPRNSelected=2*MAXGEOPRN;
			}
			//Print messages with GEO elevation and the selected one
			//Only print it when more GEOs are in the list or solution has converged
			//Check first if any satellite was selected (there maybe GEOs but they may be under alarm)
			if (maxelevationpos==-1) {
				//No satellite selected due to all are under alarm
				//If there is only one GEO, select it. Otherwise keep waiting until one of them exits alarm state
				if (SBASdata[0].numSBASsatellites==1) {
					options->GEOindex=1;
					epoch->currentGEOPRN=SBASdata[options->GEOindex].PRN;
					sprintf(messagestr,"%s (%s)",ElevText[0],options->GEOPRNSelected==2*MAXGEOPRN?"Selected as it is the only GEO available":"Temporary selection until solution converges as it is the only GEO available");
					printInfo(messagestr,options);
				} else {
					options->GEOindex=-3;
					return;
				}
			} else if (options->GEOPRNSelected==2*MAXGEOPRN || numGEOelevprev!=numGEOelev) {
				for(i=0;i<numGEOelev;i++) {
					if(i==maxelevationpos) continue;
					sprintf(messagestr,"%s (Not selected)",ElevText[i]);
					printInfo(messagestr,options);
				}
				sprintf(messagestr,"%s (%s)",ElevText[maxelevationpos],options->GEOPRNSelected==2*MAXGEOPRN?"Selected":"Temporary selection until solution converges");
				printInfo(messagestr,options);
				options->GEOPRN=SBASdata[options->GEOindex].PRN;
			}
			numGEOelevprev=numGEOelev;
		} else { //MINGEOPRN<=options->GEOPRN<=MAXGEOPRN
			//The user has given the GEO PRN
			options->GEOindex=SBASdata[0].GEOPRN2pos[options->GEOPRN];
			epoch->currentGEOPRN=SBASdata[options->GEOindex].PRN;
			if(options->GEOindex<0) {
				//We have no data from this satellite. Maybe in future epochs we receive data
				//until we do not receive data, we will leave options->GEOindex negative in order to keep checking this condition
				options->GEOindex=-2;
				return;
			} else {
				options->GEOPRNSelected=2*MAXGEOPRN;
			}
		}
	} 
}

/*****************************************************************************
 * Name        : SBASGEOAlarmCheck
 * Description : Check if any GEO has received an alarm and update its time
 *                remaining for finishing the alarm state. It has to be
 *                executed before an epoch is processed.
 *                It has to be executed by a single thread, to avoid race conditions
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TEpoch  *epoch                  I  N/A  TEpoch structure
 * TSBASdata  *SBASdata            IO N/A  Struct to save SBAS data
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void SBASGEOAlarmCheck (TEpoch  *epoch, TSBASdata  *SBASdata, TOptions  *options) {

	int 	i,j;
	int 	GEOPositionWithLatestAlarm;
	double 	difftime;

	if(options->usetype0messages==1 && SBASdata[0].numAlarmGEO>0) {
		//We have received a type 0 messages and we are treating it as it is, or
		//we are treating type 0 messages as type 2, but we have received a type 0 message will all zeros (0)
		//In this case, we need to erase all data from the current GEO (if we use mixed data, we will also erase the mixed data)

		SBASdata[0].alarmTimeRemaining=-1; //Reset time remaining so it is saved for mixed GEO

		for(i=0;i<SBASdata[0].numAlarmGEO;i++) {
			//Check for the position in the SBASdata for the GEO from which we received the type 0 message
			j=SBASdata[0].alarmGEOindex[i];

			//Check if we have cleared the data from the GEO
			if(SBASdata[0].cleared==0) {
				SBASdata[0].cleared=1;
				//Erase all data from mixed GEO except for almanac data only if mixed GEO is enabled
				if (options->mixedGEOdata==1) {
					ClearSBASdata(SBASdata,0);
				}
			}
			if(SBASdata[j].cleared==0) {
				SBASdata[j].cleared=1;
				//Erase all data from this GEO except for almanac data
				ClearSBASdata(SBASdata,j);
			}

			//When we receive a type 0 message, we have to wait for one minute until we can use it again
			difftime=tdiff(&epoch->t,&SBASdata[j].lastmsgreceived[DONTUSE]); 
			if((1-(difftime-(int)difftime))<1E-4) difftime=(int)difftime+1; //This is to account for problems when converting to integer, when a number is 1.99999998
			if((int)difftime>=60) {
				//The minute cooldown has passed. Clear alarm condition
				SBASdata[j].dontuse=0;
				SBASdata[j].problems=0;
				SBASdata[0].numAlarmGEO--;
				SBASdata[0].alarmGEOPRN[SBASdata[j].PRN]=-1;
				SBASdata[0].alarmGEOindex[i]=-1;
			}

			//Save time remaining to finish alarm (for printing in SBASUNSEL message)
			SBASdata[j].alarmTimeRemaining=60-(int)difftime;
			if (SBASdata[j].alarmTimeRemaining>SBASdata[0].alarmTimeRemaining) {
				SBASdata[0].alarmTimeRemaining=SBASdata[j].alarmTimeRemaining;
				GEOPositionWithLatestAlarm=j;
			}
		}

		//Save time of message type 0 for mixed GEO (for printing in SBASUNSEL message)
		SBASdata[0].alarmtime=SBASdata[GEOPositionWithLatestAlarm].lastmsgreceived[DONTUSE];

		if (SBASdata[0].numAlarmGEO==0) {
			//When no alarms are remaining, we can disable the alarm for the mixed satellite
			//The mixed satellite, (even though it is disabled) is used to account for the alarms
			SBASdata[0].dontuse=0;
			SBASdata[0].cleared=0;
			SBASdata[0].problems=0;
		} else {
			//GEOs with alarm remaining. The satellite unselected may be the one in the middle of the
			//index list, therefore we need to check it and move the next satellites to the first 
			//position so the satellites with alarm are at the beginning of the list
			for(i=0;i<SBASdata[0].numAlarmGEO;i++) {
				if (SBASdata[0].alarmGEOindex[i]==-1) {
					for(j=i+1;j<MAX_SBAS_PRN;j++) {
						if(SBASdata[0].alarmGEOindex[j]!=-1) {
							SBASdata[0].alarmGEOindex[i]=SBASdata[0].alarmGEOindex[j];
							SBASdata[0].alarmGEOindex[j]=-1;
							break;
						}
					}
				}
			}
		}
	}
}

/*****************************************************************************
 * Name        : SBAShandlespecialevents
 * Description : If we have decimation greater than one second, we still need 
 *                to handle special events every second such as alarm messages,
 *                or UDREI=15. This option is called only on decimated epochs,
 *                otherwise these events are already handled at computeSBAS function
 *                This function is called by a single thread.
 * Parameters  :
 * Name                           |Da|Unit|Description
 * TTime currentepoch              I  N/A  Current epoch
 * TSBASdata  *SBASdata            IO N/A  Struct to save SBAS data
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void SBAShandlespecialevents (TTime currentepoch, TSBASdata  *SBASdata, TOptions  *options) {

	int						i,j,k;
	int						UDREI,IODP,IODPfastcorr,posFC;
	double 					difftime;
	unsigned int			nearestInt;
	static unsigned int		initialized=0;
	static unsigned int		Decimateval;
	TTime   				tudre;
	
	if (initialized==0) {
		Decimateval = (int)(options->decimate*10000.);
		initialized=1;
	}
	
	//Check if this epoch is decimated
	if (options->decimate==0.) return;
	else {
		nearestInt = (int)(currentepoch.SoD*10000.);
		if ( (nearestInt%Decimateval) == 0 ) return;
	}

	//Check if we have received a type 0 message
	if(options->usetype0messages==1 && SBASdata[0].numAlarmGEO>0) {
		//We have received a type 0 messages and we are treating it as it is, or
		//we are treating type 0 messages as type 2, but we have received a type 0 message will all zeros (0)
		//In this case, we need to erase all data from the current GEO (if we use mixed data, we will also erase the mixed data)

		for(i=0;i<SBASdata[0].numAlarmGEO;i++) {
			//Check for the position in the SBASdata for the GEO from which we received the type 0 message
			j=SBASdata[0].alarmGEOindex[i];

			//Check if we have cleared the data from the GEO
			if(SBASdata[0].cleared==0) {
				SBASdata[0].cleared=1;
				//Erase all data from mixed GEO except for almanac data only if mixed GEO is enabled
				if (options->mixedGEOdata==1) {
					ClearSBASdata(SBASdata,0);
				}
			}
			if(SBASdata[j].cleared==0) {
				SBASdata[j].cleared=1;
				//Erase all data from this GEO except for almanac data
				ClearSBASdata(SBASdata,j);
			}

			//When we receive a type 0 message, we have to wait for one minute until we can use it again
			difftime=tdiff(&currentepoch,&SBASdata[j].lastmsgreceived[DONTUSE]);
			if((1-(difftime-(int)difftime))<1E-4) difftime=(double)((int)difftime+1); //This is to account for problems when converting to integer, when a number is 1.99999998
			if((int)difftime>=60) {
				//The minute cooldown has passed. Clear alarm condition
				SBASdata[j].dontuse=0;
				SBASdata[j].problems=0;
				SBASdata[0].numAlarmGEO--;
				SBASdata[0].alarmGEOPRN[SBASdata[j].PRN]=-1;
				SBASdata[0].alarmGEOindex[i]=-1;
			}
		}

		if (SBASdata[0].numAlarmGEO==0) {
			//When no alarms are remaining, we can disable the alarm for the mixed satellite
			//The mixed satellite, (even though it is disabled) is used to account for the alarms
			SBASdata[0].dontuse=0;
			SBASdata[0].cleared=0;
			SBASdata[0].problems=0;
		} else {
			//GEOs with alarm remaining. The satellite unselected may be the one in the middle of the
			//index list, therefore we need to check it and move the next satellites to the first 
			//position so the satellites with alarm are at the beginning of the list
			for(i=0;i<SBASdata[0].numAlarmGEO;i++) {
				if (SBASdata[0].alarmGEOindex[i]==-1) {
					for(j=i+1;j<MAX_SBAS_PRN;j++) {
						if(SBASdata[0].alarmGEOindex[j]!=-1) {
							SBASdata[0].alarmGEOindex[i]=SBASdata[0].alarmGEOindex[j];
							SBASdata[0].alarmGEOindex[j]=-1;
							break;
						}
					}
				}
			}
		}
	}

	//Next conditions will be done if we have selected a GEO PRN
	if(options->GEOindex>=0) {

		//Select IODP
		if(SBASdata[options->GEOindex].IODPfastcorr==-1) return; //No IODP available. Nothing else can be checked
		IODPfastcorr=SBASdata[options->GEOindex].IODPfastcorr; //IODPfastcorr is the IODP for current mask, which will be the same for all the for loop
		//Check for UDREI=14 or UDREI=15
		for(i=0;i<SBASdata[options->GEOindex].numsatellites[IODPfastcorr];i++) {
			IODP=SBASdata[options->GEOindex].IODPfastcorr; //We have to save again the IODP because in any satellite we may use the IODP for the current or the previous mask
			posFC=SBASdata[options->GEOindex].poslastFC[IODP][i];
			if (posFC==-1) {
				//No data in current mask, try previous
				IODP=SBASdata[options->GEOindex].oldIODPfastcorr;
				posFC=SBASdata[options->GEOindex].poslastFC[IODP][i];
				if (posFC==-1) {
					//No data for this satelllite
					continue;
				}
			}

			UDREI=0;
			if(SBASdata[options->GEOindex].lastfastmessage[i]==INTEGRITYINFO) {
				//Last fast correction data received is from message type 6
				if(SBASdata[options->GEOindex].IODFintegrity[i]==3) {
					//Alarm. Use IODF from message type 6 regardles of other IODF
					if(SBASdata[options->GEOindex].UDREI6[i]>=14) UDREI=SBASdata[options->GEOindex].UDREI6[i];
				} else if (SBASdata[options->GEOindex].IODFintegrity[i]!=SBASdata[options->GEOindex].IODF[IODP][posFC][i]) {
					//IODF from message type 6 does not match the IODF from messages type 2-5,24 and IODF!=3
					//We cannot use the UDREsigma from message type 6, so we keep using the last one from messages type 2-5,24
					if(SBASdata[options->GEOindex].UDREI[IODP][posFC][i]>=14) UDREI=SBASdata[options->GEOindex].UDREI[IODP][posFC][i];
				} else {
					if(SBASdata[options->GEOindex].UDREI6[i]>=14) UDREI=SBASdata[options->GEOindex].UDREI6[i];
				}
				tudre=SBASdata[options->GEOindex].lastintegrityinfo[i];
			} else {
				//Last fast correction data received is from message type is from type 2-5,24
				if(SBASdata[options->GEOindex].UDREI[IODP][posFC][i]>=14) UDREI=SBASdata[options->GEOindex].UDREI[IODP][posFC][i];
				tudre=SBASdata[options->GEOindex].lastfastcorrections[IODP][posFC][i];
			}
			//Check UDREI has not timed out
			difftime=tdiff(&currentepoch,&tudre);
			if(difftime<=(double)(SBASdata[options->GEOindex].timeoutmessages[options->precisionapproach][INTEGRITYINFO])) {
				//UDRE not timed out. 
				if(UDREI>=14) {
					//We have received an UDREI=14 or UDREI=15. We need to erase all the data
					for(k=0;k<5;k++) {
						SBASdata[options->GEOindex].poslastFC[k][i] = -1;
						SBASdata[options->GEOindex].numFC[k][i] = 0;
						if(k==SBASdata[options->GEOindex].IODPfastcorr) {
							SBASdata[options->GEOindex].useforbidden[SBASdata[options->GEOindex].pos2GNSS[k][i]][SBASdata[options->GEOindex].pos2PRN[k][i]]=UDREI;
						}
						for(j=0;j<MAXSBASFASTCORR;j++) {
							SBASdata[options->GEOindex].PRC[k][j][i] = 9999;
							SBASdata[options->GEOindex].RRC[k][j][i] = 9999;
							SBASdata[options->GEOindex].UDREI[k][j][i] = -1;
							SBASdata[options->GEOindex].UDRE[k][j][i] = -1;
							SBASdata[options->GEOindex].UDREsigma[k][j][i] = -1;
							SBASdata[options->GEOindex].IODF[k][j][i] = -1;
							SBASdata[options->GEOindex].lastfastcorrections[k][j][i].MJDN = -1;
							SBASdata[options->GEOindex].lastfastcorrections[k][j][i].SoD = -1;
						}
					}
				} else {
					k=SBASdata[options->GEOindex].IODPfastcorr;
					if (SBASdata[options->GEOindex].useforbidden[SBASdata[options->GEOindex].pos2GNSS[k][i]][SBASdata[options->GEOindex].pos2PRN[k][i]]>=14) {
						//We have received UDREI<=13, we need to clear the useforbidden if it was active
						SBASdata[options->GEOindex].useforbidden[SBASdata[options->GEOindex].pos2GNSS[k][i]][SBASdata[options->GEOindex].pos2PRN[k][i]]=0;
					}
				}
			}
		}
	}
	return;

}

/*****************************************************************************
 * Name        : IGP2latlon
 * Description : Given an IGP in a band number in the SBAS iono grid, compute
 *                its latitude and longitude (for SBAS iono)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * int bandnumber                  I  N/A  Band number (0-10)
 * int IGP                         I  N/A  IGP number (0-200)
 * int *IGPlat                     O  º    Latitude of the IGP (-90º..90º)
 * int *IGPlon                     O  º    Longitude of the IGP (-180º..180º)
 * Returned value (int)            O  N/A  1 => IGP in map
 *                                         0 => IGP not in map
 *****************************************************************************/
int IGP2latlon (int bandnumber, int IGP, int *IGPlat, int *IGPlon) {

	int 	dlon=0;

	IGP++;

	if (bandnumber<=8) {

		if (bandnumber==0) {
			dlon=-160;
			if (IGP==28) {
				*IGPlon=-180;
				*IGPlat=85;
			} else if((IGP>=29) && (IGP<=128)) {
				IGP=IGP-1;
			}
		}                                                        


		if (bandnumber==1) {
			dlon=-120;
			if (IGP==1) {
				*IGPlon=-140;
				*IGPlat=-85;
			} else if((IGP>=2) && (IGP<=128)) {
				IGP=IGP-1;
			}
		}                                                   


		if (bandnumber==2) {
			dlon=-80;
			if (IGP==78) {
				*IGPlon=-90;
				*IGPlat=85;
			} else if((IGP>=79) && (IGP<=128)) {
				IGP=IGP-1;
			}
		}                                                    


		if (bandnumber==3) {  
			dlon=-40;
			if (IGP==51) {
				*IGPlon=-10;
				*IGPlat=-85;
			} else if((IGP>=52) && (IGP<=128)) {     
				IGP=IGP-1;
			}
		}

		if (bandnumber==4) {
			dlon=0;
		}                                

		if (bandnumber==5) {
			dlon=40;
			if (IGP==101) {
				*IGPlon=-10;
				*IGPlat=-85;
			} else if((IGP>=102) && (IGP<=128)) {
				IGP=IGP-1;
			}
		}  

		if (bandnumber==6) {
			dlon=80;
			if (IGP==178) {
				*IGPlon=-10;
				*IGPlat=85;
			} else if((IGP>=128) && (IGP<=177)) {
				IGP=IGP+1;
			}
		}   

		if (bandnumber==7) {
			dlon=120;
			if (IGP==151) {
				*IGPlon=130;
				*IGPlat=-85;
			} else if((IGP>=128) && (IGP<=150)) {
				IGP=IGP+1;
			}
		}                                                         

		if (bandnumber==8) {
			dlon=160;
			if(IGP>=128) {
				IGP=IGP+1;
			}
		}                                                     

		if ((IGP<1) || (IGP>201)) {
			*IGPlon=9999;
			*IGPlat=9999;
			return 0;
		}                                                                                                                                           

	//       ...........................
		if(IGP==1) {
			*IGPlon=-20;
			*IGPlat=-75;
		} else if(IGP==2) { 
			*IGPlon=-20;
			*IGPlat=-65;
		} else if(IGP<=25) {
			*IGPlon=-20;
			*IGPlat=-55+5*(IGP-3);      
		} else if(IGP==26) {
			*IGPlon=-20;
			*IGPlat=65;
		} else if(IGP==27) {
			*IGPlon=-20;
			*IGPlat=75;                      
	//         ...........................
		} else if(IGP<=50) {
			*IGPlon=-15;
			*IGPlat=-55+5*(IGP-28);
	//         ...........................
		} else if(IGP==51) {
			*IGPlon=-10;
			*IGPlat=-75;                  
		} else if(IGP==52) {
			*IGPlon=-10;
			*IGPlat=-65;                  
		} else if(IGP<=75) {
			*IGPlon=-10;
			*IGPlat=-55+5*(IGP-53);
		} else if(IGP==76) {
			*IGPlon=-10;
			*IGPlat=65;
		} else if(IGP==77) {
			*IGPlon=-10;
			*IGPlat=75;
	//         ...........................                                                  
		} else if(IGP<=100) {
			*IGPlon=-5;
			*IGPlat=-55+5*(IGP-78);
	//         ...........................     
		} else if(IGP==101) {
			*IGPlon=0;
			*IGPlat=-75;
		} else if(IGP==102) {
			*IGPlon=0;
			*IGPlat=-65;
		} else if(IGP<=125) {
			*IGPlon=0;
			*IGPlat=-55+5*(IGP-103);
		} else if(IGP==126) {
			*IGPlon=0;
			*IGPlat=65;
		} else if(IGP==127) {
			*IGPlon=0;
			*IGPlat=75;
		} else if(IGP==128) {
			*IGPlon=0;
			*IGPlat=85;                      
	//         ...........................                      
		} else if(IGP<=151) {
			*IGPlon=5;
			*IGPlat=-55+5*(IGP-129);
	//         ...........................
		} else if(IGP==152) {
			*IGPlon=10;
			*IGPlat=-75;
		} else if(IGP==153) {
			*IGPlon=10;
			*IGPlat=-65;
		} else if(IGP<=176) {
			*IGPlon=10;
			*IGPlat=-55+5*(IGP-154);
		} else if(IGP==177) {
			*IGPlon=10;
			*IGPlat=65;
		} else if(IGP==178) {
			*IGPlon=10;
			*IGPlat=75;
	//         ...........................
		} else if(IGP<=201) {
			*IGPlon=15;
			*IGPlat=-55+5*(IGP-179);
	//       ...........................                        
		}
		*IGPlon=*IGPlon+dlon;

	//       ------------- Bands 9 and 10 -----------------
	} else if (bandnumber>=9) {
		if ((IGP>=1) && (IGP<37)) {
			*IGPlat=60;
			*IGPlon=(IGP-1)*5-180;
		} else if ((IGP>=37) && (IGP<73)) {
			*IGPlat=60;
			*IGPlon=(IGP-37)*5;
		} else if ((IGP>=73) && (IGP<91)) {
			*IGPlat=65;
			*IGPlon=(IGP-73)*10-180;
		} else if ((IGP>=91) && (IGP<109)) {
			*IGPlat=65;
			*IGPlon=(IGP-91)*10;
		} else if ((IGP>=109) && (IGP<127)) {
			*IGPlat=70;
			*IGPlon=(IGP-109)*10-180;
		} else if ((IGP>=127) && (IGP<145)) {
			*IGPlat=70;
			*IGPlon=(IGP-127)*10;
		} else if ((IGP>=145) && (IGP<163)) {
			*IGPlat=75;
			*IGPlon=(IGP-145)*10-180;
		} else if ((IGP>=163) && (IGP<181)) {
			 *IGPlat=75;
			 *IGPlon=(IGP-163)*10;
		} else if ((IGP>=181) && (IGP<187)) {
			*IGPlat=85;
			*IGPlon=(IGP-181)*30-180;
		} else if ((IGP>=187) && (IGP<=192)) {
			*IGPlat=85;
			*IGPlon=(IGP-187)*30;
		} else {
			*IGPlat=9999;
			*IGPlon=9999;
			return 0;
		}


		if (bandnumber==10) {
			*IGPlat=-*IGPlat;
			if (IGP>=181) *IGPlon=*IGPlon+10;
		}

	}
	return 1;
}

/*****************************************************************************
 * Name        : IPP2Sqr
 * Description : Given the latitude and longitude of an IPP, compute the left
 *               down corner (number 3) latitude and longitude IGP involved in the IPP 
 *               interpolation and the increment for the other corners (for SBAS iono)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double IPPlat                   I  º    Latitude of the IPP (-90º..90º)
 * double IPPlon                   I  º    Longitude of the IPP (-180º..180º or 0º..360º)
 * double *IGPlat                  O  º    Latitude of the left-down corner IGP (-90º..90º)
 * double *IGPlon                  O  º    Longitude of the left-down corner IGP (0º..360º)
 * double *increment               O  º    Distance to the IGP in the other corners (in degrees)
 * Returned value (int)            O  N/A  1 => Valid IGP
 *                                         0 => IGP not in map
 *****************************************************************************/

int IPP2Sqr (double IPPlat, double IPPlon, double *IGPlat, double *IGPlon, double *increment) {
 
	/*

	2-----1         3: (*IGPlat,*IGPlon)
	|     |         2: (*IGPlat,*IGPlon+*increment)
	| IPP |         4: (*IGPlat+*increment,*IGPlon)
	3-----4         1: (*IGPlat+*increment,*IGPlon+*increment)

	NOTE: when |IPPlat|>85 deg, *IGPlon is the longitude of the closest IGP to the west of the IPP*/

	double	difflat, tmplon=0.;
 

	//Compute the IGP for |IGPlat|>80. 
	//We look for the IGP closest to the west of the IPP
	
	if ((IPPlat>=85.)||(IPPlat<-85.)) {     

		*increment=0;

		*IGPlon=modulo(IPPlon+360,360);

		if (IPPlat>=85.) {  
			*IGPlat=85;
			if (*IGPlon<360) tmplon=270;
			if (*IGPlon<270) tmplon=180;
			if (*IGPlon<180) tmplon=90;
			if (*IGPlon<90)  tmplon=0;
			if (*IGPlon<0) return 0;

		} else if (IPPlat<-85.) {      

			*IGPlat=-85.;
			if (*IGPlon<310) tmplon=220;
			if (*IGPlon<220) tmplon=130;
			if (*IGPlon<130) tmplon=40;
			if (*IGPlon<40)  tmplon=310;
			if (*IGPlon<=360) tmplon=310;
			if (*IGPlon<0) return 0;
		}  

		*IGPlon=tmplon;
	   
	} else {

		//Compute the IGP in the other cases.
		*increment=10;
		difflat=5;
		if ((IPPlat>=-60.) && (IPPlat<60.)) {
			*increment=5;
			difflat=0;
		}
	 
		*IGPlon=IPPlon+360;
		*IGPlon=modulo(*IGPlon,360);
		*IGPlon=(int)(*IGPlon/(*increment))*(*increment);
	 
		*IGPlat=IPPlat+360-difflat;
		*IGPlat=modulo(*IGPlat,360);
		*IGPlat=(int)(*IGPlat/(*increment))*(*increment)+difflat;
		if (*IGPlat>180.) *IGPlat=*IGPlat-360;

	}

	return 1;
}

/*****************************************************************************
 * Name        : IGPLatLon2BandNumberIGP
 * Description : Given the latitude and longitude of an IGP, compute the 
 *                band number and IGP for bands 0-8 and bands 9-10 (for SBAS iono)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double IGPlat                   I  º    Latitude of the IGP (-90º..90º)
 * double IGPlon                   I  º    Longitude of the IGP (0º..360º)
 * int *Bandnumber                 O  N/A  Band number (0-10)
 * int *IGP                        O  N/A  IGP number for bands 0-8 (1-201)
 * int *IGP90                      O  N/A  IGP number for bands 9-10 (1-192)
 * Returned value (int)            O  N/A  1 => IGP found
 *                                         0 => IGP not in map
 *****************************************************************************/
int IGPLatLon2BandNumberIGP (double IGPlat, double IGPlon, int *Bandnumber, int *IGP, int *IGP90) {
 
	//-------------------------------------------------------
	//This subroutine computes the Bandnumber and IGP from the longitude (IGPlon) and latitude (IGPlat) of IGP.
	//The subroutine also checks if (IGPlat,IGPlon) corresponds to a valid IGP.


	//INPUT:
	//*IGP coordinates (IGPlat,IGPlon)


	//OUTPUT:
	//*Bandnumber= Band Number
	//*IGP= IGP in band
	//*IGP90= IGP in band 9 or 10

	//NOTE 1: The same IGP can belong to Bandnumber=0 to 8 and at the same time to Bandnumber=9.
	//In this case, the routine provides both, [*Bandnumber][*IGP] and *IGP90.

	//If IGP90=0, thence, the IGP does not belong to Band 9 nor 10.

	//........................................................

	double 	templon,templon2,templat,templat2,xx,yy;
	int 	nlon,nlat,lonstep,latstep,m;
 
 
	*Bandnumber=0;
	*IGP=0;
	*IGP90=0;

	//IGP check -------------------------------------------

	//Compute the corresponding IGP

	if (fabs(IGPlat)>75) {
		//Latitude over 75º or below -75º
		templon=modulo(IGPlon+360,360);
		nlon=(int)round(templon);
		if (IGPlat==85.0) {  
			//IGPlat equal to 85º
			if (nlon==180) {
				*Bandnumber=0;
				*IGP=28;
				*IGP90=181;
			} else if (nlon==270) {  
				*Bandnumber=2;
				*IGP=78;
				*IGP90=184;
			} else if (nlon==0) {      
				*Bandnumber=4;
				*IGP=128;
				*IGP90=187;
			} else if (nlon==90) {      
				*Bandnumber=6;
				*IGP=178;
				*IGP90=190;
			} else if (nlon==210) {
				*IGP90=182;
			} else if (nlon==240) {
				*IGP90=183;
			} else if (nlon==300) {
				*IGP90=185;
			} else if (nlon==330) {
				*IGP90=186;
			} else if (nlon==30) {
				*IGP90=188;
			} else if (nlon==60) {
				*IGP90=189;
			} else if (nlon==120) {
				*IGP90=191;
			} else if (nlon==150) {
				*IGP90=192;
			}
			return 1;
		} else if (IGPlat==-85.0) {
			//IGPlat equal to -85º
			if (nlon==220) {
				*Bandnumber=1;
				*IGP=1;
				*IGP90=182;
			} else if (nlon==310) {
				*Bandnumber=3;
				*IGP=51;
				*IGP90=185;
			} else if (nlon==40) {
				*Bandnumber=5;
				*IGP=101;
				*IGP90=188;
			} else if (nlon==130) {
				*Bandnumber=7;
				*IGP=151;
				*IGP90=191;
			} else if (nlon==190) {
				*IGP90=181;
			} else if (nlon==250) {
				*IGP90=183;
			} else if (nlon==280) {
				*IGP90=184;
			} else if (nlon==340) {
				*IGP90=186;
			} else if (nlon==10) {
				*IGP90=187;
			} else if (nlon==70) {
				*IGP90=189;
			} else if (nlon==100) {
				*IGP90=190;
			} else if (nlon==160) {
				*IGP90=192;
			}
			return 1;
		} else { 
			//IGPlat over 85º or below -85º
			return 0;
		}
	}


	//Checking IGPlat: [-75:75]
	//Note: we include IGPlat=+/-60 because, at this latitude, there are IGPs every 5 Deg (we need just to decide if it is IGP or not). 
	//If we need to find the left-down corner, we should exclude IGPlat=+/-60 (i.e., to put "</>" instead of "<=/>=" in the condition "if").

	if ((IGPlat>=-60.0)&&(IGPlat<=60.0)) {
		lonstep=5;
		latstep=5;
	} else {
		lonstep=10;
		latstep=5;
	}
 

	templon=modulo(IGPlon+360,360);
	templon=(int)(templon/lonstep)*lonstep;
	nlon=(int)round(templon);
 
	templat=modulo(IGPlat+360,360);
	templat=(int)(templat/latstep)*latstep;
	nlat=(int)round(templat);
	if (nlat>180) nlat=nlat-360;

	//Map to the interval [0:360]    
	templon2=modulo(IGPlon+360,360);
	templat2=modulo(IGPlat+360,360);
 
	//Comparing to check if it is a IGP
	xx=fabs(templon2-templon);
	yy=fabs(templat2-templat);
 
	if ((xx>0)||(yy>0)) {
		return 0;
	}

	//----------------------------------------------------

	if (nlon<20) {
		*Bandnumber=4;
		m=(nlon+20)/5;
	} else if (nlon<60) {
		*Bandnumber=5;
		m=(nlon-20)/5;
	} else if (nlon<100) {
		*Bandnumber=6;
		m=(nlon-60)/5;
	} else if (nlon<140) {
		*Bandnumber=7;
		m=(nlon-100)/5;
	} else if (nlon<180) {
		*Bandnumber=8;
		m=(nlon-140)/5;
	} else if (nlon<220) {
		*Bandnumber=0;
		m=(nlon-180)/5;
	} else if (nlon<260) {
		*Bandnumber=1;
		m=(nlon-220)/5;
	} else if (nlon<300) {
		*Bandnumber=2;
		m=(nlon-260)/5;
	} else if (nlon<340) {
		*Bandnumber=3;
		m=(nlon-300)/5;
	//} else if (nlon<=360) {
	} else {
		*Bandnumber=4;
		m=(nlon-340)/5;
	}
 
	if (m==0) {
		if (nlat==-75) {
			*IGP=1;
		} else if (nlat==-65) {
			*IGP=2;
		} else if (nlat==65) {
			*IGP=26;
		} else if (nlat==75) {
			*IGP=27;
		} else if ((nlat>=-55)&&(nlat<=55)) {
			*IGP=(nlat+55)/5+3;
		}
	} else if (m==1) {
		*IGP=(nlat+55)/5+28;
	} else if (m==2) {
		if (nlat==-75) {
			*IGP=51;
		} else if (nlat==-65) {
			*IGP=52;
		} else if (nlat==65) {
			*IGP=76;
		} else if (nlat==75) {
			*IGP=77;
		} else if ((nlat>=-55)&&(nlat<=55)) {
			*IGP=(nlat+55)/5+53;
		}
	} else if (m==3) {
		*IGP=(nlat+55)/5+78;
	} else if (m==4) {
		if (nlat==-75) {
			*IGP=101;
		} else if (nlat==-65) {
			*IGP=102;
		} else if (nlat==65) {
			*IGP=126;
		} else if (nlat==75) {
			*IGP=127;
		} else if (nlat==85) {
			*IGP=128;
		} else if ((nlat>=-55)&&(nlat<=55)) {
			*IGP=(nlat+55)/5+103;
		}
	} else if (m==5) {
		 *IGP=(nlat+55)/5+129;
	} else if (m==6) {
		if (nlat==-75) {
			*IGP=152;
		} else if (nlat==-65) {
			*IGP=153;
		} else if (nlat==65) {
			*IGP=177;
		} else if (nlat==75) {
			*IGP=178;
		} else if ((nlat>=-55)&&(nlat<=55)) {
			*IGP=(nlat+55)/5+154;
		}
	} else if (m==7) { 
		*IGP=(nlat+55)/5+179;
	}
	if (*Bandnumber==0) {
		if ((*IGP>=28)&&(*IGP<=128)) *IGP=*IGP+1;
		if ((m==0)&&(nlat==85)) *IGP=28;
	} else if (*Bandnumber==1) {
		if ((*IGP>=1)&&(*IGP<=128)) *IGP=*IGP+1;
		if ((m==0)&&(nlat==-85)) *IGP=1;
	} else if (*Bandnumber==2) {
		if ((*IGP>=78)&&(*IGP<=128)) *IGP=*IGP+1;
		if ((m==2)&&(nlat==85)) *IGP=78;
	} else if (*Bandnumber==3) {
		if ((*IGP>=51)&&(*IGP<=128)) *IGP=*IGP+1;
		if ((m==2)&&(nlat==-85)) *IGP=51;
	} else if (*Bandnumber==5) {
		if ((*IGP>=101)&&(*IGP<=128)) *IGP=*IGP+1;
		if ((m==4)&&(nlat==-85)) *IGP=101;
	} else if (*Bandnumber==6) {
		if ((*IGP>=128)&&(*IGP<=178)) *IGP=*IGP-1;
		if ((m==6)&&(nlat==85)) *IGP=178;
	} else if (*Bandnumber==7) {
		if ((*IGP>=128)&&(*IGP<=151)) *IGP=*IGP-1;
		if ((m==6)&&(nlat==-85)) *IGP=151;
	} else if (*Bandnumber==8) {
		if (*IGP>=128) *IGP=*IGP-1;
	}

	//New IONO bands 9 and 10: -----------------
	if ((nlat==60)||(nlat==-60)) {
		if (nlon>=180) {
			*IGP90=(nlon-180)/5+1;
		} else {
			*IGP90=(nlon+180)/5+1;
		}
	} else if ((nlat==65)||(nlat==-65)) {
		if (nlon>=180) {
			*IGP90=(nlon-180)/10+73;
		} else {
			*IGP90=(nlon+180)/10+73;
		}
	} else if ((nlat==70)||(nlat==-70)) {
		if (nlon>=180) {
			*IGP90=(nlon-180)/10+109;
		} else {
			*IGP90=(nlon+180)/10+109;
		}
	} else if ((nlat==75)||(nlat==-75)) {
		if (nlon>=180) {
			*IGP90=(nlon-180)/10+145;
		} else {
			*IGP90=(nlon+180)/10+145;
		}
	}
	//------------------------------------------

	if (*IGP+*IGP90<=0) {
		return 0;	
	}

	return 1;
}

/*****************************************************************************
 * Name        : IGPsearch
 * Description : Given the latitude and longitude of an IPP, the coordinates 
 *               of the left down corner IGP and the latitude and longitude
 *               increment (so as to define a region of IGP ), gather the number 
 *               in mask of the four IGP and check if a interpolation can be done
 *               (over a square or a triangle) (for SBAS iono)
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double IPPlat                   I  º    Latitude of the IPP (-90º..90º)
 * double IPPlon                   I  º    Longitude of the IPP (-180º..180º or 0º..360º)
 * double IGPlat                   I  º    Left-down corner IGP latitude (-90º..90º)
 * double IGPlon                   I  º    Left-down corner IGP longitude (0º..360º)
 * double lat_inc                  I  º    Latitude increment (in degrees)
 * double lon_inc                  I  º    Longitude increment (in degrees) 
 * int IGPinMask[5][11][202]       I  N/A  Matrix indicating whether an IGP is monitored or not 
 * int IGP2Mask[11][202]           I  N/A  Matrix indicating to what mask correspond the last received for this IPP
 * int *InterpolationMode          O  N/A  Indicates the available InterpolationMode
 *                                           0 Full square in mask with IPP inside
 *                                           [1:4] Triangle in mask with IPP inside (the number indicates the excluded vertex)
 *                                           -1  Triangle in mask with IPP OUTSIDE
 *                                           -10 Less than three IGPs of the square in mask
 *                                           -20 IPP OUTSIDE of the square
 * int *IGPnumber                  O  N/A  Vector indicating the number in the mask of the IGPs around the IPP 
 * int *IGPBandNumber              O  N/A  Vector indicating the band number of the IGPs around the IPP 
 *****************************************************************************/
void IGPsearch (double IPPlat, double IPPlon, double IGPlat, double IGPlon, double lat_inc, double lon_inc, int IGPinMask[5][11][202], int IGP2Mask[11][202], int *InterpolationMode, int *IGPnumber, int *IGPBandNumber) {

	double 	IPPlon_aux,IGPlat_aux1,IGPlat_aux2,IGPlon_aux1,IGPlon_aux2,x,y,W;
	int 	inside,numIGPvalid;
	int 	Bandnumber1,Bandnumber2,Bandnumber3,Bandnumber4,IGP1,IGP2,IGP3,IGP4,IGP90;
	int 	IGP1valid,IGP2valid,IGP3valid,IGP4valid;
 
	//WARNING:
	//This routine can return IGP=0 when the IGP is not in the MT18 mask for bands 9 and 10 and, at the same time, it does not belong to any MOPS grid IGP for bands 0 to 8.

	IPPlon_aux=modulo(IPPlon+360,360);


	//LOOK SQUARE:
	IGPlon_aux1=IGPlon;
	IGPlat_aux1=IGPlat;
	IGPlon_aux2=IGPlon+lon_inc;
	IGPlat_aux2=IGPlat+lat_inc;
 
	*InterpolationMode=0;
	inside=0;
	if ((IPPlon_aux>=IGPlon_aux1)&&(IPPlon_aux<=IGPlon_aux2)) inside=1;
	if ((IPPlat>=IGPlat_aux1)&&(IPPlat<=IGPlat_aux2)) inside=inside+1;
	if (inside!=2) {
		*InterpolationMode=-20;
		return;
	}
 
	IGPLatLon2BandNumberIGP(IGPlat_aux2,IGPlon_aux2,&Bandnumber1,&IGP1,&IGP90);
	if (IGP90>0) {
		if (IGPinMask[IGP2Mask[9][IGP90]][9][IGP90]==1) {
			Bandnumber1=9;
			IGP1=IGP90;
		} else if (IGPinMask[IGP2Mask[10][IGP90]][10][IGP90]==1) {
			Bandnumber1=10;
			IGP1=IGP90;
		}
	}

	IGPLatLon2BandNumberIGP(IGPlat_aux2,IGPlon_aux1,&Bandnumber2,&IGP2,&IGP90);
	if (IGP90>0) {
		if (IGPinMask[IGP2Mask[9][IGP90]][9][IGP90]==1) {
			Bandnumber2=9;
			IGP2=IGP90;
		} else if (IGPinMask[IGP2Mask[10][IGP90]][10][IGP90]==1) {
			Bandnumber2=10;
			IGP2=IGP90;
		}
	}

	IGPLatLon2BandNumberIGP(IGPlat_aux1,IGPlon_aux1,&Bandnumber3,&IGP3,&IGP90);
	if (IGP90>0) {
		if (IGPinMask[IGP2Mask[9][IGP90]][9][IGP90]==1) {
			Bandnumber3=9;
			IGP3=IGP90;
		} else if (IGPinMask[IGP2Mask[10][IGP90]][10][IGP90]==1) {
			Bandnumber3=10;
			IGP3=IGP90;
		}
	}

	IGPLatLon2BandNumberIGP(IGPlat_aux1,IGPlon_aux2,&Bandnumber4,&IGP4,&IGP90);
	if (IGP90>0) {
		if (IGPinMask[IGP2Mask[9][IGP90]][9][IGP90]==1) {
			Bandnumber4=9;
			IGP4=IGP90;
		} else if (IGPinMask[IGP2Mask[10][IGP90]][10][IGP90]==1) {
			Bandnumber4=10;
			IGP4=IGP90;
		}
	}

	IGPnumber[1]=IGP1;
	IGPnumber[2]=IGP2;
	IGPnumber[3]=IGP3;
	IGPnumber[4]=IGP4;
	IGPBandNumber[1]=Bandnumber1;
	IGPBandNumber[2]=Bandnumber2;
	IGPBandNumber[3]=Bandnumber3;
	IGPBandNumber[4]=Bandnumber4;
 

	//NOTICE that, after introducing the Bands 9 and 10 we will look for IGPs that are only in the Band 9 or 10 (and they are not in the others bands). 
	//Thence, if these IGPs are not in the mask (i.e, IGPinMask[IGP2Mask[9][IGP90]][9][IGP90]!=1 or IGPinMask[IGP2Mask[10][IGP90]][10][IGP90]!=1), the IGP will set as "0", because it does not exist for the other bands!!!. 
	//This is the reason why we must to check if the IGP1,IGP2,IGP3 and IGP4 are zero before to use them as matrix index.

	IGP1valid=0;
	IGP2valid=0;
	IGP3valid=0;
	IGP4valid=0;
	if (IGP1>0) IGP1valid=IGPinMask[IGP2Mask[Bandnumber1][IGP1]][Bandnumber1][IGP1];
	if (IGP2>0) IGP2valid=IGPinMask[IGP2Mask[Bandnumber2][IGP2]][Bandnumber2][IGP2];
	if (IGP3>0) IGP3valid=IGPinMask[IGP2Mask[Bandnumber3][IGP3]][Bandnumber3][IGP3];
	if (IGP4>0) IGP4valid=IGPinMask[IGP2Mask[Bandnumber4][IGP4]][Bandnumber4][IGP4];
		 
	numIGPvalid=IGP1valid+IGP2valid+IGP3valid+IGP4valid;

	if (numIGPvalid==4)  {
		*InterpolationMode=0;
	} else if (numIGPvalid==3)  {
		//::::::::::::  LOOK TRIANGLE  :::::::::::::
		x=(IPPlon_aux-IGPlon)/lon_inc;
		y=(IPPlat-IGPlat)/lat_inc;

		*InterpolationMode=1;
		if (IGP2valid==0) {
			*InterpolationMode=2;
			x=1-x;
		} else if (IGP3valid==0) {
			*InterpolationMode=3;
			x=1-x;
			y=1-y;
		} else if (IGP4valid==0) {
			*InterpolationMode=4;
			y=1-y;
		}
		//:::::::::::::::::::::::::::::::::::::::::: 
		W=1-x-y;
		//If W>=0, the IPP is in the triangle, and it can be used for interpolation
		if(W<0) {
			*InterpolationMode=-1;
		}
	} else if (numIGPvalid<3) {
		*InterpolationMode=-10;
	}
 
	return;

}

/*****************************************************************************
 * Name        : SBASionoInterpolation
 * Description : For a given region around an IPP (defined by the IGP given),
 *               interpolate the vertical delay, the sigma error and the GIVE
 *               value according to the interpolation mode given (square or triangle)
 *               (for SBAS iono)
 *               
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double IPPlat                   I  º    Latitude of the IPP (-90º..90º)
 * double IPPlon                   I  º    Longitude of the IPP (-180º..180º or 0º..360º)
 * double IGPlat1                  I  º    Left-down corner IGP latitude (-90º..90º)
 * double IGPlon1                  I  º    Left-down corner IGP longitude (0º..360º)
 * double IGPlat2                  I  º    Right-up corner IGP latitude (-90º..90º)
 * double IGPlon2                  I  º    Right-up corner IGP longitude (0º..360º)
 * int *InterpolationMode          I  N/A  0     Interpolation in square
 *                                         [1:4] Interpolation in triangle (the number indicates the excluded vertex)
 *                                         -1    No interpolation in square
 *                                         [-1:-4] No Interpolation in triangle (the number indicates the excluded vertex)
 * double IGPDelays[5]             I  m    Vertical delays in the four grid points [1-4] from MT 26 (position 0 is not used) (in meters)
 * double IGPSigma2[5]             I  m^2  Sigma^2 in the four grid points [1-4] after applying formula A-58 in MOPS-C (position 0 is not used) (in meters^2)
 * double *VerticalDelay           O  m    Interpolated ionospheric vertical delay (in meters)
 * double *Sigma2                  O  m^2  Interpolated sigma error (in meters^2)
 * double **IPP_Weight             O  N/A  Weights for each of the corners involved in the IPP interpolation
 *****************************************************************************/
void SBASionoInterpolation (double IPPlat,double IPPlon,double IGPlat1,double IGPlon1,double IGPlat2,double IGPlon2,int *InterpolationMode,double IGPDelays[5],double IGPSigma2[5],double *VerticalDelay,double *Sigma2,double *IPP_Weight) {
 
	/*---------------------------------------------------------------

	INPUT:
	-----
	(IPPlat,IPPlon): IPP latitude and longitude (deg)

	* Square or triangle interpolation:
	(IGPlat1,IGPlat2,IGPlon1,IGPlon2):surrounding square grid-points latitude and longitude coordinates (deg)
	
	(IGPlat2,IGPlon1)   (IGPlat2,IGPlon2)
          v2..............v1
          .               .
          .               .
          .               .
          .               .
          .               .
          .               .
          .               .
          v3..............v4
	(IGPlat1,IGPlon1)   (IGPlat1,IGPlon2)
          WEST               EAST

	* Circle or semicircle interpolation (when |IPPlat|>85deg):
	(IPPlon3,xx,xx,xx):surrounding square grid-points longitude and where [3] is the closest IGP to the west of the IPP

        	   2          IPPlon3= Longitude of IGP[3]
        	/     \
          3         1
        W   \ *   /     
        	   4
        	 East
	
	
	InterpolationMode:
	InterpolationMode=-1 NO interpolation (skip)
	InterpolationMode=0  square interpolation
	InterpolationMode<>0 triangle interpolation, where the value of "InterpolationMode"
	indicates the "bad" corner (or quadrant)  |2 1|
	                                          |3 4|

	IGPDelays: Vertical Delay values at the surrounding square grid-points
  	   IGPDelays[2]|IGPDelays[1]
	  	   ----x----
	   IGPDelays[3]|IGPDelays[4]

	IGPSigma2: Sigma values at the surrounding square grid-points
	   IGPSigma2[2]|IGPSigma2[1]
	           ----x----
	   IGPSigma2[3]|IGPSigma2[4]

	Note:
	IGPSigma2 = (IGPGIVE+eps_iono)**2   (RSSiono=0)   (formula A-58 in MOPS-C)
	            IGPGIVE**2+eps_iono**2  (RSSiono=1)



	IT ALSO INCLUDES the interpolation for |IPPlat|>80deg (only with the four IGPs monitored [MOPS-B])


	A) INTERPOLATION OVER SQUARES or TRIANGLES:

	
	* Four points:
	-----------      

	(IGPlat2,IGPlon1)   (IGPlat2,IGPlon2)
	  W2=(1-x)*y          W1=x*y
          v2..............v1
          .               .
          .               .
          .               .
          .               .
          .     *         .
          .     |         .
          .     |         .
          v3____..........v4
	  W3=(1-x)*(1-y)      W4=x*(1-y)
	 (IGPlat1,IGPlon1)   (IGPlat1,IGPlon2)
	   WEST               EAST

	
	
	*VerticalDelay= v1*W1 + v2*W2 + v3*W3 + v4*W4
	
	
	where:
	
	(x,y): are taken from corner [3]
	
	
	
	* Three points: Four expressions.
	------------     
	
	a)
	
         v2        v1
        W2=y      W1=0
          |\
          |  \
          | *  \
          |______\
         v3        v4
	    W3=1-x-y     W4=x
	
	
	where:
	(x,y): are taken from corner [3]
	*VerticalDelay= v1*W1 + v2*W2 + v3*W3 + v4*W4
	

	NOTE: W2+W1 --> W2
	W3+W1 --> W3
	W4+W1 --> W4
	

	
	b)
        v2       v1
        W2=0     W1=y
          /|
          /  |
          /    |
          /      |
          --------
          v3     v4
        W3=1-x   W4=x-y
	
	
	(x,y): are taken from corner [3]
	*VerticalDelay= v1*W1 + v2*W2 + v3*W3 + v4*W4
	
	
	
	c)
          v2       v1
	   W2=1-x     W1=x+y-1
          ---------
          \       |
          \     |
          \   |
          \ |
          |
          v3       v4
         W3=0     W4=1-y
	
	
	(x,y): are taken from corner [3]
	*VerticalDelay= v1*W1 + v2*W2 + v3*W3 + v4*W4
	
	
	
	d)
	
          v2         v1
        W2=y-x      W1=x
          ________
          |        /
          |      /
          |    /
          |  /
          |/
          v3        v4
        W3=1-y     W4=0
	
	
	(x,y): are taken from corner [3]
	*VerticalDelay= v1*W1 + v2*W2 + v3*W3 + v4*W4
	
	



	B) INTERPOLATION OVER CIRCLES or SEMICIRCLES [|fi|>85deg]      
	
	* Four points:
	-----------

              2
           /     \      [3] closest IGP to the west of the IPP
         3         1
   	   W   \ *   /       y=(fabs(lat)-85)/10
	          4
              E          x=(lon - lon3)/90*(1-2*y)+y
	
	W1=x*y
	W2=(1-x)*y
	W3=(1-x)*(1-y)
	W4=x*(1-y)

	*VerticalDelay= v1*W1 + v2*W2 + v3*W3 + v4*W4



	* Three points: Two expressions:
	----------- 


	1) Not available IGP1:

              2
           /           [3] closest IGP to the west of the IPP
         3    |  
        W  \ *         y=(fabs(lat)-85)/10
              4
              E        x=(lon - lon3)/90*(1-2*y)+y

	W2=(1-x)*y
	W3=(1-x)*(1-y)
	W4=x

	NOTE: W3+W4 --> W4

	*VerticalDelay= v2*W2 + v3*W3 + v4*W4


	2) Not available IGP2:

	
	[3] closest IGP to the west of the IPP
          3 ------ 1
        W   \ *  /       y=(fabs(lat)-85)/10
              4
  	          E          x=(lon - lon3)/90*(1-2*y)+y

	W1=x*y
	W3=(1-x)
	W4=x*(1-y)

	NOTE: W2+W3 --> W3

	*VerticalDelay= v1*W1 + v3*W3 + v4*W4

	NOTE:
	If IGP3 or IGP4 are not available, the interpolation is not
	possible, because [3] is the closest IGP to the west of the IPP



	
	OUTPUT:
	------
	*VerticalDelay: Vertical Delay at the IPP (in meters)
	*Sigma2: User Ionospheric Vertical Error at the IPP (in meters^2)

	InterpolationMode= -1 ===> NOT MONITORED
	(in this case, also *Sigma2=-1)

	COMMENTS;
	--------
	The interpolation weights [W1,W2,W3,W4] are defined like the
	components of vector "IGPDelays", and corresponds to the 1st,2n,3st
	and 4th quadrants:
	i.e.:   W2|W1
	        --x--
	        W3|W4



	..................................................................*/


	double 	W1,W2,W3,W4,x,y,IPPlon_aux;
 
	IPP_Weight[1]=W1=0;
	IPP_Weight[2]=W2=0;
	IPP_Weight[3]=W3=0;
	IPP_Weight[4]=W4=0;

	IPPlon_aux=modulo(IPPlon+360,360);

	if ((IPPlat>=85.)||(IPPlat<-85.)) {
		//Interpolation for |IPPlat|>80deg
		//Notice that for |IPPlat|>80, IGPlon1 has been assigned to the IGPlon value provided by the function "IPP2Sqr" for this case.
		//That is, it corresponds to the closest IGP to IPP from west

		y=(fabs(IPPlat)-85.)/10;
		x=(IPPlon_aux-IGPlon1)/90*(1-2*y)+y;
		if ((x>1)||(y>1)) {
			*Sigma2=-1;
			*InterpolationMode=-1;
			return;
		}


		if (*InterpolationMode==0) {
			//Interpolating with four points
			W1=x*y;
			W2=(1-x)*y;
			W3=(1-x)*(1-y);
			W4=x*(1-y);
			*VerticalDelay=IGPDelays[1]*W1+IGPDelays[2]*W2+IGPDelays[3]*W3+IGPDelays[4]*W4;
			*Sigma2=IGPSigma2[1]*W1+IGPSigma2[2]*W2+IGPSigma2[3]*W3+IGPSigma2[4]*W4;
			IPP_Weight[1]=W1;
			IPP_Weight[2]=W2;
			IPP_Weight[3]=W3;
			IPP_Weight[4]=W4;
		 } else {              
			//Interpolating with Three points
			//MOPS does not define the weights for the three points interpolation when |fi|>85.(we assume these equations!!!)
			//NOTE: This interpolation is only valid in MOPS A. In this version, it will never execute this code

			if (*InterpolationMode==1) {
				W2=(1-x)*y;
				W3=(1-x)*(1-y);
				W4=x;
				*VerticalDelay=IGPDelays[2]*W2+IGPDelays[3]*W3+IGPDelays[4]*W4;
				*Sigma2=IGPSigma2[2]*W2+IGPSigma2[3]*W3+IGPSigma2[4]*W4;
				IPP_Weight[1]=0;
				IPP_Weight[2]=W2;
				IPP_Weight[3]=W3;
				IPP_Weight[4]=W4;
			} else if (*InterpolationMode==2) {
				W1=x*y;
				W3=(1-x);
				W4=x*(1-y);
				*VerticalDelay=IGPDelays[1]*W1+IGPDelays[3]*W3+IGPDelays[4]*W4;
				*Sigma2=IGPSigma2[1]*W1+IGPSigma2[3]*W3+IGPSigma2[4]*W4;
				IPP_Weight[1]=W1;
				IPP_Weight[2]=0;
				IPP_Weight[3]=W3;
				IPP_Weight[4]=W4;
			} else if (*InterpolationMode>2) {     
				*InterpolationMode=-1;
			} else if (*InterpolationMode<0) {
				*InterpolationMode=-1;
			}
		}
		return;
	}


	x=(IPPlon_aux-IGPlon1)/(IGPlon2-IGPlon1);
	y=(IPPlat -IGPlat1)/(IGPlat2-IGPlat1);

	if ((x>1)||(y>1)) {
		*Sigma2=-1;
		*InterpolationMode=-1;
		return;
	}

	if (*InterpolationMode<0) {
		*Sigma2=-1;
		*InterpolationMode=-1;
		return;
	}
	if (*InterpolationMode==0) {
		//Rectangle
		W1=x*y;
		W2=(1.-x)*y;
		W3=(1.-x)*(1.-y);
		W4=x*(1.-y);
		*VerticalDelay=IGPDelays[1]*W1+IGPDelays[2]*W2+IGPDelays[3]*W3+IGPDelays[4]*W4;
		*Sigma2=IGPSigma2[1]*W1+IGPSigma2[2]*W2+IGPSigma2[3]*W3+IGPSigma2[4]*W4;
		IPP_Weight[1]=W1;
		IPP_Weight[2]=W2;
		IPP_Weight[3]=W3;
		IPP_Weight[4]=W4;
	} else {
		//Triangle
		if (*InterpolationMode==2) {
			x=1.-x;
		} else if (*InterpolationMode==3) {
			x=1.-x;
			y=1.-y;
		} else if (*InterpolationMode==4) {
			y=1.-y;
		}
		W1=y;
		W2=1.-x-y;
		W3=x;
		//*Sigma2=-1 ==> OUTSIDE TRIANGLE: No correction
		if (W2<0) {
			 *Sigma2=-1;
			 *InterpolationMode=*InterpolationMode*(-1);
		} else {
			if (*InterpolationMode==1) {
				*VerticalDelay=IGPDelays[2]*W1+IGPDelays[3]*W2+IGPDelays[4]*W3;
				*Sigma2=IGPSigma2[2]*W1+IGPSigma2[3]*W2+IGPSigma2[4]*W3;
				IPP_Weight[1]=0;
				IPP_Weight[2]=W1;
				IPP_Weight[3]=W2;
				IPP_Weight[4]=W3;
			}
			if (*InterpolationMode==2) {
				*VerticalDelay=IGPDelays[1]*W1+IGPDelays[3]*W3+IGPDelays[4]*W2;
				*Sigma2=IGPSigma2[1]*W1+IGPSigma2[3]*W3+IGPSigma2[4]*W2;
				IPP_Weight[1]=W1;
				IPP_Weight[2]=0;
				IPP_Weight[3]=W3;
				IPP_Weight[4]=W2;
			}
			if (*InterpolationMode==3) {
				*VerticalDelay=IGPDelays[1]*W2+IGPDelays[2]*W3+IGPDelays[4]*W1;
				*Sigma2=IGPSigma2[1]*W2+IGPSigma2[2]*W3+IGPSigma2[4]*W1;
				IPP_Weight[1]=W2;
				IPP_Weight[2]=W3;
				IPP_Weight[3]=0;
				IPP_Weight[4]=W1;
			}
			if (*InterpolationMode==4) {
				*VerticalDelay=IGPDelays[1]*W3+IGPDelays[2]*W2+IGPDelays[3]*W1;
				*Sigma2=IGPSigma2[1]*W3+IGPSigma2[2]*W2+IGPSigma2[3]*W1;
				IPP_Weight[1]=W3;
				IPP_Weight[2]=W2;
				IPP_Weight[3]=W1;
				IPP_Weight[4]=0;
			}
		}
	} 
															   

	return;
}

/*****************************************************************************
 * Name        : InsideOutside
 * Description : For a given region (squared or triangular), check if the coordinate
 *               given (in latitude and longitude) is inside the region for
 *                SBAS message type 27
 *               
 * Parameters  :
 * Name                           |Da|Unit|Description
 * double latpos                   I  º    Coordinate latitude (-90º..90º)
 * double lonpos                   I  º    Coordinate longitude (0º..360º, -180º..180º)
 * double lat1                     I  º    Latitude  point 1 (-90º..90º)
 * double lon1                     I  º    Longitude point 1 (0º..360º, -180º..180º)
 * double lat2                     I  º    Latitude  point 2 (-90º..90º)
 * double lon2                     I  º    Longitude point 2 (0º..360º, -180º..180º)
 * double lat3                     I  º    Latitude  point 3 (-90º..90º) 
 * double lon3                     I  º    Longitude point 3 (0º..360º, -180º..180º)
 * double lat4                     I  º    Latitude  point 4 (-90º..90º)
 * double lon4                     I  º    Longitude point 4 (0º..360º, -180º..180º)
 *                                         Point4 is ignored if the region is triangular
 * int  shape                      I  N/A  0 -> triangular region
 *                                         1 -> squared region
 * Returned value (int)            O  N/A  1  => Coordinate inside region
 *                                         0  => Coordinate outside region
 *                                         -1 => Not a valid region
 *****************************************************************************/
int InsideOutside (double latpos, double lonpos, double lat1, double lon1, double lat2, double lon2, double lat3, double lon3, double lat4, double lon4, int shape) {

	/*In MOPS, message type 27 sends coordinates 1 and 2, but it doesn't define in which of these orders is given:
	 *
	 *       1             3                  2             4                   3             1                  4             2
	 *  (lat1,lon1)   (lat1,lon2)        (lat2,lon2)   (lat2,lon1)         (lat1,lon2)   (lat1,lon1)        (lat2,lon1)   (lat2,lon2)
	 *       x             x                  x             x                   x             x                  x             x
	 *                              or                               or                                 or
	 *       x             x                  x             x                   x             x                  x             x
	 *  (lat2,lon1)   (lat2,lon2)        (lat1,lon2)   (lat1,lon1)         (lat2,lon2)   (lat2,lon1)        (lat1,lon1)   (lat1,lon2)
	 *       4             2                  3             1                   2             4                  1             3
	 *
	 *  Coordinate 3 takes the coordinate 1 latitude and the coordinate 2 longitude.
	 *  For a square region, coordinate 4 takes the coordinate 2 latitude and the coordinate 1 longitude.
	 *  We will reorder the points so as it stays with the first shape shown above.
	 *
	 *  Additionally, the regions will not be longer than 179º, having this shape:
	 *
	 *    |lon2-lon1|<=179º           |lon2-lon1|>179º
	 *    
	 *   1 ·---------· 3              ---· 1   3 ·---
	 *     |/////////|                ///|       |///
	 *     |/////////|                ///|       |///
	 *   4 ·---------· 2              ---· 4   2 ·---
	 */




	int	vertexoutside=4;	//This variable states which vertex will not be used in the triangle interpolation
	double	tmplat,tmplon;
	double	x,y,Weight;

	//Set all longitudes between 0..360º
	lonpos=modulo(lonpos+360,360);
	lon1=modulo(lon1+360,360);
	lon2=modulo(lon2+360,360);
	lon3=modulo(lon3+360,360);
	lon4=modulo(lon4+360,360);

	//Check that coordinate 1 is in the upper part. If it is not the case, reorder the points
	if(lat1<lat2) {
		//Coordinate 1 is in the lower part. We need to reorder it
		tmplat=lat1;
		tmplon=lon1;
		lat1=lat2;
		lon1=lon2;
		lat2=tmplat;
		lon2=tmplon;
		lat3=lat1;
		lon3=lon2;
		lat4=lat2;
		lon4=lon1;

		if(lon1>lon2) vertexoutside=3;
		else vertexoutside=1;
	}

	//Check that coordinate 1 is in the upper-left corner. If it is not the case, reorder the points
	if(lon1>lon2) {
		//coordinate 1 is in the upper-right corner. We need to reorder it
		tmplon=lon1;
		lon1=lon2;
		lon2=tmplon;
		lon3=lon2;
		lon4=lon1;

		if(vertexoutside==4) vertexoutside=2;
	}

	if(lat1-lat2<1 || lon2-lon1<1) {
		//Distance between points in longitude or latitude are less than 1º. It is not a valid region
		return -1;
	}

	if(shape==1) {
		//Squared region
		if(latpos>=lat2 && latpos<=lat1) {
			if(lon2-lon1>179) {
				//The region goes through the extremes
				if(lonpos>=lon2 || lonpos<=lon1) return 1;
				else return 0;
			} else {
				//The region does not go through the extremes
				if(lonpos<=lon2 && lonpos>=lon1) return 1;
				else return 0;
			}
		} else {
			//Position outside region
			return 0;
		}
	} else if (shape==0) {
		//Triangular region
		if(lon2-lon1>179) {
			//The region goes through the extremes, so in this case it is easier to work with longitudes between -180..180º
			if(lonpos>180) lonpos-=360;
			if(lon1>180) lon1-=360;
			if(lon2>180) lon2-=360;
			if(lon3>180) lon3-=360;
			if(lon4>180) lon4-=360;
		}

		x=(lonpos-lon4)/(lon3-lon4);
		y=(latpos-lat4)/(lat3-lat4);

		if(vertexoutside==1) {
            /*This case:
                 1     3     
       (lat1,lon1)     (lat1,lon2)
                 x     x
                      /|             
                     / |             
                    /  |             
                   /   |             
                  /    |             
                 x-----x
       (lat2,lon1)     (lat2,lon2)
                 4     2     
            */
			 x=1-x;
		} else if (vertexoutside==2) {
            /*This case:
                 1     3     
       (lat1,lon1)     (lat1,lon2)
                 x-----x
                 |    /              
                 |   /               
                 |  /                
                 | /                 
                 |/                  
                 x     x
       (lat2,lon1)     (lat2,lon2)
                 4     2     
           */
			y=1-y;
		} else if (vertexoutside==3) {
           /*This case:
                 1     3     
       (lat1,lon1)     (lat1,lon2)
                 x     x
                 |\
                 | \
                 |  \
                 |   \
                 |    \
                 x-----x
       (lat2,lon1)     (lat2,lon2)
                 4     2     
         */
        //Nothing to change in this case
		} else if (vertexoutside==4) {
         /*This case:
                 1     3     
       (lat1,lon1)     (lat1,lon2)
                 x-----x
                  \    |             
                   \   |             
                    \  |             
                     \ |             
                      \|             
                 x     x
       (lat2,lon1)     (lat2,lon2)
                 4     2     
        */
			x=1-x;
			y=1-y;
		}
		Weight=1-x-y;
		if(Weight>=0) return 1;
		else return 0;
	} else {
		//Not a valid shape
		return -1;
	}
}

/*****************************************************************************
 * Name        : AWGN_generator
 * Description : Generate an AWGN sample using Box-Muller Transformation method
 *               
 * Parameters  : 
 * Name                           |Da|Unit|Description
 * double stddev                   I  m    Standard deviation of the AWGN to be generated
 * Returned value (double)         O  m    AWGN sample value (in meters) 
 *****************************************************************************/
double AWGN_generator (double stddev) {

	double		mean = 0;
	double		sample, val1, val2 = 0;

	while ( val2 == 0 ) {
		#if defined (__WIN32__)
			val2 = (double)(rand()) / ((double)RAND_MAX);
		#else
			val2 = (double)random() / ((double)RAND_MAX);
		#endif
	}

	#if defined (__WIN32__)
		val1 = cos((2.0*Pi)*(double)(rand()) / ((double)RAND_MAX));
	#else
		val1 = cos((2.0*Pi)*(double)random() / ((double)RAND_MAX));
	#endif
	sample = sqrt(-2.0*log(val2)) * val1;

	//The sample is now with mean 0 and variance=1
	//We can change the mean and variance as desired. In this case the mean will be 0 always
	sample = mean + sample * stddev;

	return sample;
}

/*****************************************************************************
 * Name        : qsort_compare_float
 * Description : Function to compare two doubles for qsort function 
 *               (from C library <stdlib.h>). The requisite for this function
 *               stated in the qsort man page is that when a<b, a negative
 *               value should be returned, when a=b zero should be a returned,
 *               and when a>b, a positive value should be returned
 *
 * Parameters  :
 * Name                           |Da|Unit|Description
 * const void *a                   I  N/A  First double value as 'const void' pointer 
 * const void *b                   I  N/A  Second double value as 'const void' pointer 
 * Returned value (int)            O  N/A  -1 -> a < b
 *                                          0 -> a=b
 *                                          1 -> a > b
 *****************************************************************************/
int qsort_compare_float(const void *a,const void *b) {
    float *x = (float *) a;
    float *y = (float *) b;
    if (*x < *y) return -1;
    else if (*x > *y) return 1;
    return 0;
    //NOTE: In some codes, instead of setting the return value with if-else,
    //they use the following line:
    //return (*x-*y);
    //This instructions is valid, except in the cases in the limits of decimal representation.
    //For example, if a = MAX_FLOAT_VALUE and b = -10, then a-b will overflow the result and 
    // return an undefined value
}

/*****************************************************************************
 * Name        : qsort_compare_double
 * Description : Function to compare two doubles for qsort function 
 *               (from C library <stdlib.h>). The requisite for this function
 *               stated in the qsort man page is that when a<b, a negative
 *               value should be returned, when a=b zero should be a returned,
 *               and when a>b, a positive value should be returned
 *
 * Parameters  :
 * Name                           |Da|Unit|Description
 * const void *a                   I  N/A  First double value as 'const void' pointer 
 * const void *b                   I  N/A  Second double value as 'const void' pointer 
 * Returned value (int)            O  N/A  -1 -> a < b
 *                                          0 -> a=b
 *                                          1 -> a > b
 *****************************************************************************/
int qsort_compare_double(const void *a,const void *b) {
    double *x = (double *) a;
    double *y = (double *) b;
    if (*x < *y) return -1;
    else if (*x > *y) return 1;
    return 0;
    //NOTE: In some codes, instead of setting the return value with if-else,
    //they use the following line:
    //return (*x-*y);
    //This instructions is valid, except in the cases in the limits of decimal representation.
    //For example, if a = MAX_DOUBLE_VALUE and b = -10, then a-b will overflow the result and 
    // return an undefined value
}

/*****************************************************************************
 * Name        : qsort_compare_int
 * Description : Function to compare two integers for qsort function 
 *               (from C library <stdlib.h>). The requisite for this function
 *               stated in the qsort man page is that when a<b, a negative
 *               value should be returned, when a=b zero should be a returned,
 *               and when a>b, a positive value should be returned
 *
 * Parameters  :
 * Name                           |Da|Unit|Description
 * const void *a                   I  N/A  First integer value as 'const void' pointer 
 * const void *b                   I  N/A  Second integer value as 'const void' pointer 
 * Returned value (int)            O  N/A  -1 -> a < b
 *                                          0 -> a=b
 *                                          1 -> a > b
 *****************************************************************************/
int qsort_compare_int(const void *a,const void *b) {
    int *x = (int *) a;
    int *y = (int *) b;
    if (*x < *y) return -1;
    else if (*x > *y) return 1;
    return 0;
    //NOTE: In some codes, instead setting the return value with if-else,
    //they use the following line:
    //return (*x-*y);
    //This instructions is valid, except in the cases in the limits of integer representation.
    //For example, if a = MAX_INTEGER_VALUE and b = -10, then a-b will overflow the result and 
    // return an undefined value
}

/*****************************************************************************
 * Name        : mkdir_recursive
 * Description : Function to create directories recursively. The function 
 *                assumes that the string after the last slash (or backslash
 *                in Windows) is the filename, so it does not use the filename
 *                to create a directory.
 *
 * Parameters  :
 * Name                           |Da|Unit|Description
 * char *path                      I  N/A  Directory
 * Returned value (int)            O  N/A  -1 -> Error when creating directory
 *                                          0 -> No error
 *****************************************************************************/
int mkdir_recursive(char *filepath) {

	int 			i;
	int 			len;
	int				ret;
	int				badDir=0;
	char			slash;
	struct 	stat 	path_stat;

	#if (defined __WIN32__) || (defined __CYGWIN__)
		int			colon=0;
	#endif

	len=strlen(filepath);
	for (i=0; i<len-1; i++) {
		#if defined (__WIN32__)
			if (filepath[i] == '\\') {
				slash='\\';
				if(i==0) {
					badDir=1;
					break;
				}
				//For windows, if path is absolute, we need to skip first part of path (the C: part)
				if (colon==1) {
					colon=0;
					continue;
				}
		#elif defined (__CYGWIN__)
			if (filepath[i] == '/'|| filepath[i] == '\\') {
				if (filepath[i] == '/') {
					if (i==0) continue; //If is a full path in Linux path style, skip the first part
					slash='/';
				} else if (filepath[i] == '\\') {
					slash='\\';
					if(i==0) {
						badDir=1;
						break;
					}
					//For windows, if path is absolute, we need to skip first part of path (the C: part)
					if (colon==1) {
						colon=0;
						continue;
					}
				}
		#else
			if (filepath[i] == '/') {
				if (i==0) continue; //If is a full path in Linux, skip the first part
				slash='/';
		#endif
				/* Temporarily truncate */
				filepath[i] = '\0';

				#if !defined (__WIN32__)
						ret=mkdir(filepath,S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH); //Permission of the directory set to 755
				#else
						ret=mkdir(filepath);
				#endif
				if(ret!=0 && errno!=EEXIST) {
				   return -1;
				} else if (ret!=0 && errno==EEXIST) {
					//Check that what exists is a directory and not a file
					stat(filepath, &path_stat);
					if (!S_ISDIR(path_stat.st_mode)) {
					   return -1;
					}
				}

				filepath[i] = slash;
		#if defined (__WIN32__) || (defined __CYGWIN__)
			} else if (filepath[i] == ':' && i==1) {
				colon=1;
		#endif
			}
	}
	if(badDir==1) {
		//In Windows, directory starts with backslash, which is illegal
		for(i=1;i<len-1; i++) {
			if (filepath[i] == '\\') {
				filepath[i] = '\0';
				return -1;
			}
		}
	}
	return 0;
}

/*****************************************************************************
 * Name        : InitRangeList
 * Description : Initialize a TRangeList structure
 * Parameters  :
 * Name                            |Da|Unit|Description
 * TRangeList *List                 O  N/A  Structure with the list of satellite ranges
 *****************************************************************************/
void InitRangeList (TRangeList *List) {

    List->size=1;
    List->listSatGNSS=NULL;
    List->listSatPRN=NULL;
    List->numJumpslist=NULL;
    List->listSatGNSS=malloc(sizeof(enum GNSSystem*));
    List->listSatGNSS[0]=malloc(sizeof(enum GNSSystem*));
    List->listSatGNSS[0][0]=malloc(sizeof(enum GNSSystem)*2);
    List->listSatPRN=malloc(sizeof(int*));
    List->listSatPRN[0]=malloc(sizeof(int*));
    List->listSatPRN[0][0]=malloc(sizeof(int)*2);
    List->numJumpslist=malloc(sizeof(int));
    List->listSatPRN[0][0][0]=1;
    List->listSatPRN[0][0][1]=32; //This is for the last range, which will not have end value
    List->listSatGNSS[0][0][0]=GPS;
    List->listSatGNSS[0][0][1]=GPS; //This is for the last range, which will not have end value
    List->numJumpslist[0]=1;
}

/*****************************************************************************
 * Name        : freeRangeList
 * Description : Free memory of a TRangeList structure
 * Parameters  :
 * Name                            |Da|Unit|Description
 * TRangeList *List                 O  N/A  Structure with the list of satellite ranges
 *****************************************************************************/
void freeRangeList (TRangeList *List) {

    int i,j;

    for(i=0;i<List->size;i++) {
        for(j=0;j<List->numJumpslist[i];j++) {
            free(List->listSatGNSS[i][j]);
            free(List->listSatPRN[i][j]);
        }
        free(List->listSatGNSS[i]);
        free(List->listSatPRN[i]);
    }
    free(List->listSatGNSS);
    free(List->listSatPRN);
    free(List->numJumpslist);
    List->listSatGNSS=NULL;
    List->listSatPRN=NULL;
    List->numJumpslist=NULL;
    List->size=1;
}

/*****************************************************************************
 * Name        : AddRangeList
 * Description : Add a new block of ranges in the TRangeList structure
 * Parameters  :
 * Name                            |Da|Unit|Description
 * enum GNSSystem GNSS              I  N/A  Constellation of the first satellite in the first range
 * int PRN                          I  N/A  PRN  of the first satellite in the first range
 * TRangeList *List                 O  N/A  Structure with the list of satellite ranges
 *****************************************************************************/
void AddRangeList (enum GNSSystem GNSS, int PRN, TRangeList *List) {

	int i;
	
	i=List->size;
	List->size++;
	List->listSatGNSS=realloc(List->listSatGNSS,sizeof(enum GNSSystem*)*List->size);
	List->listSatGNSS[i]=malloc(sizeof(enum GNSSystem*));
	List->listSatGNSS[i][0]=malloc(sizeof(enum GNSSystem)*2);
	List->listSatPRN=realloc(List->listSatPRN,sizeof(int*)*List->size);
	List->listSatPRN[i]=malloc(sizeof(int*));
	List->listSatPRN[i][0]=malloc(sizeof(int)*2);
	List->numJumpslist=realloc(List->numJumpslist,sizeof(int)*List->size);

	List->listSatGNSS[i][0][0]=GNSS;
	List->listSatPRN[i][0][0]=PRN;
	List->listSatGNSS[i][0][1]=GPS; //This is the last constellation, due to the last range will not have an end
	List->listSatPRN[i][0][1]=32; //Value to see that no end range value was set
	List->numJumpslist[i]=1;
}

/*****************************************************************************
 * Name        : ExtendRangeList
 * Description : Add a new range in an existing block of ranges of a TRangeList structure
 * Parameters  :
 * Name                            |Da|Unit|Description
 * enum GNSSystem GNSS              I  N/A  Constellation of the first satellite in the range
 * int PRN                          I  N/A  PRN of the first satellite in the range
 * int pos                          I  N/A  Position of the block where to add the new range
 * TRangeList *List                 O  N/A  Structure with the list of satellite ranges
 *****************************************************************************/
void ExtendRangeList (enum GNSSystem GNSS, int PRN, int pos, TRangeList *List) {
	
	List->listSatGNSS[pos]=realloc(List->listSatGNSS[pos],sizeof(enum GNSSystem*)*(List->numJumpslist[pos]+1));
	List->listSatGNSS[pos][List->numJumpslist[pos]]=malloc(sizeof(enum GNSSystem)*2);
	List->listSatPRN[pos]=realloc(List->listSatPRN[pos],sizeof(int*)*(List->numJumpslist[pos]+1));
	List->listSatPRN[pos][List->numJumpslist[pos]]=malloc(sizeof(int)*2);
	List->listSatGNSS[pos][List->numJumpslist[pos]][0]=GNSS;
	List->listSatPRN[pos][List->numJumpslist[pos]][0]=PRN;
	List->listSatGNSS[pos][List->numJumpslist[pos]][1]=GPS; //This is the last constellation, due to the last range will not have an end value set
	List->listSatPRN[pos][List->numJumpslist[pos]][1]=32; //Value to see that no end range value was set
	List->numJumpslist[pos]++;
}

/*****************************************************************************
 * Name        : SatRangeList2String
 * Description : Get the list of ranges and transform them to a string. The
 *                 ranges will be grouped by satellite constellations
 * Parameters  :
 * Name                            |Da|Unit|Description
 * int i                            I  N/A  Position of the current range list
 * TRangeList *List                 I  N/A  Structure with the list of satellite ranges
 * char *auxstr                     O  N/A  Output string with the range list
 *****************************************************************************/
void SatRangeList2String (int i, TRangeList *List,  char *auxstr) {

	int 			j;
	unsigned int	k; //to avoid warning: comparison between signed and unsigned integer expression
	const int		listMaxSatGNSS[MAX_GNSS]={MAX_SAT_GPS,MAX_SAT_GAL,MAX_SAT_GLO,MAX_SAT_GEO,MAX_SAT_BDS,MAX_SAT_QZS,MAX_SAT_IRN};
	char 			auxstr2[200];

    auxstr[0]='\0';
	for(j=0;j<List->numJumpslist[i];j++) {
		if(List->listSatGNSS[i][j][0]==List->listSatGNSS[i][j][1] && List->listSatPRN[i][j][0]==List->listSatPRN[i][j][1]) {
			//Range of only one satellite
			sprintf(auxstr2,"%s %c%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][0]),List->listSatPRN[i][j][0]);
			sprintf(auxstr,"%s",auxstr2);
		} else if (List->listSatGNSS[i][j][0]!=List->listSatGNSS[i][j][1]) {
			//Constellation changed
			if (List->listSatGNSS[i][j][1]-List->listSatGNSS[i][j][0]==1) {
				//One constellation change only
				if (List->listSatPRN[i][j][1]==999) {
					sprintf(auxstr2,"%s %c%02d-%02d %c01-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][0]),List->listSatPRN[i][j][0],listMaxSatGNSS[List->listSatGNSS[i][j][0]],gnsstype2char(List->listSatGNSS[i][j][1]),listMaxSatGNSS[List->listSatGNSS[i][j][1]]);
				} else {
					sprintf(auxstr2,"%s %c%02d-%02d %c01-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][0]),List->listSatPRN[i][j][0],listMaxSatGNSS[List->listSatGNSS[i][j][0]],gnsstype2char(List->listSatGNSS[i][j][1]),List->listSatPRN[i][j][1]);
				}
				sprintf(auxstr,"%s",auxstr2);
			} else {
				//Multiple constellation change
				//First constellation change is printed outside of the for loop
				sprintf(auxstr2,"%s %c%02d-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][0]),List->listSatPRN[i][j][0],listMaxSatGNSS[List->listSatGNSS[i][j][0]]);
				sprintf(auxstr,"%s",auxstr2);
				for(k=1;k<(List->listSatGNSS[i][j][1]-List->listSatGNSS[i][j][0]);k++) {
					//Intermidiate constellation changes
					sprintf(auxstr2,"%s %c01-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][0]+k),listMaxSatGNSS[List->listSatGNSS[i][j][0]+k]);
					sprintf(auxstr,"%s",auxstr2);
				}
				//Last constellation change
				if (List->listSatPRN[i][j][1]==999) {
					sprintf(auxstr2,"%s %c01-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][1]),listMaxSatGNSS[List->listSatGNSS[i][j][1]]);
				} else {
					sprintf(auxstr2,"%s %c01-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][1]),List->listSatPRN[i][j][1]);
				}
			}
		} else if (List->listSatPRN[i][j][1]==999) {
			sprintf(auxstr2,"%s %c%02d-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][0]),List->listSatPRN[i][j][0],listMaxSatGNSS[List->listSatGNSS[i][j][1]]);
			sprintf(auxstr,"%s",auxstr2);
		} else {
			sprintf(auxstr2,"%s %c%02d-%02d",auxstr,gnsstype2char(List->listSatGNSS[i][j][0]),List->listSatPRN[i][j][0],List->listSatPRN[i][j][1]);
			sprintf(auxstr,"%s",auxstr2);
		}
	}
}

/*****************************************************************************
 * Name        : WeightType2String
 * Description : Create as string with the weight type and their values for
 *                 printing an INFO message with the user selected weights
 *               
 * Parameters  :
 * Name                           |Da|Unit|Description
 * enum GNSSystem GNSS             I  N/A  Constellation
 * int PRN                         I  N/A  PRN number
 * int NumMeas                     I  N/A  Number of measurement in the filter
 * int *SNRWeightused              O  N/A  Flag to indicate if a SNR weight is used
 * char *str                       O  N/A  String to be printed
 * TOptions  *options              I  N/A  TOptions structure
 *****************************************************************************/
void WeightType2String (enum GNSSystem GNSS, int PRN, int NumMeas, int *SNRWeightused, char *str, TOptions *options) {
	
	char ConstantVal[20];
	if (options->WeightConstantsValues[GNSS][PRN][NumMeas][0]==-2) {
		strcpy(ConstantVal,"URA");
	} else {
		sprintf(ConstantVal,"%.6g",options->WeightConstantsValues[GNSS][PRN][NumMeas][0]);
	}

	switch(options->weightMode[GNSS][PRN][NumMeas]) {
		case SBASOnlyWeight:
			sprintf(str,"StdDev^2: SBAS variance");
			break;
		case DGNSSOnlyWeight:
			sprintf(str,"StdDev^2: DGNSS variance");
			break;
		case FixedWeight:
			sprintf(str,"StdDev: %s",ConstantVal);
			break;
		case SBASFixedWeight: 
			sprintf(str,"StdDev^2: SBAS variance + (%s)^2",ConstantVal);
			break;
		case DGNSSFixedWeight:
			sprintf(str,"StdDev^2: DGNSS variance + (%s)^2",ConstantVal);
			break;
		case ElevationWeight:
			sprintf(str,"StdDev: %s + %.6g*e^(elev/%.6g)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1],options->WeightConstantsValues[GNSS][PRN][NumMeas][2]*r2d);
			break;
		case SBASElevWeight:
			sprintf(str,"StdDev^2: SBAS variance + (%s + %.6g*e^(elev/%.6g)^2",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1],options->WeightConstantsValues[GNSS][PRN][NumMeas][2]*r2d);
			break;
		case DGNSSElevWeight:
			sprintf(str,"StdDev^2: DGNSS variance + (%s + %.6g*e^(elev/%.6g))^2",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1],options->WeightConstantsValues[GNSS][PRN][NumMeas][2]*r2d);
			break;
		case Elevation2Weight:
			sprintf(str,"StdDev^2: %s + %.6g/(sin(elev)^2)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			break;
		case SBASElev2Weight:
			sprintf(str,"StdDev^2: SBAS variance + %s + %.6g/(sin(elev)^2)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			break;
		case DGNSSElev2Weight:
			sprintf(str,"StdDev^2: DGNSS variance + %s + %.6g/(sin(elev)^2)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			break;
		case SNRWeight:
			sprintf(str,"StdDev^2: %s + %.6g*10^(-SNR/10)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			*SNRWeightused=1;
			break;
		case SBASSNRWeight:
			sprintf(str,"StdDev^2: SBAS variance + %s + %.6g*10^(-SNR/10)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			*SNRWeightused=1;
			break;
		case DGNSSSNRWeight:
			sprintf(str,"StdDev^2: DGNSS variance + %s + %.6g*10^(-SNR/10)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			*SNRWeightused=1;
			break;
		case SNRElevWeight:
			sprintf(str,"StdDev^2: %s + %.6g*10^(-SNR/10)/(sin(elev)^2)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			*SNRWeightused=1;
			break;
		case SBASSNRElevWeight:
			sprintf(str,"StdDev^2: SBAS variance + %s + %.6g*10^(-SNR/10)/(sin(elev)^2)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			*SNRWeightused=1;
			break;
		case DGNSSSNRElevWeight:
			sprintf(str,"StdDev^2: DGNSS variance + %s + %.6g*10^(-SNR/10)/(sin(elev)^2)",ConstantVal,options->WeightConstantsValues[GNSS][PRN][NumMeas][1]);
			*SNRWeightused=1;
			break;
		default:
			//This the unknown weight, which is not possible. Leave the default just to avoid compiler warning
			str[0]='\0';
			break;
	}
}

/*****************************************************************************
 * Name        : SNRCombModeNum2String
 * Description : Transform from internal SNR combination weight mode type to
 *                 a string for printing an INFO message
 * Parameters  :
 * Name                            |Da|Unit|Description
 * enum SNRWeightComb SNRweightComb I  N/A  SNR Combination weight mode
 * double K1                        I  N/A  Combining factor for first frequency
 * double K2                        I  N/A  Combining factor for second frequency
 * Returned value (char*)           O  N/A  SNR Combination weight mode string
 *****************************************************************************/
char *SNRCombModeNum2String(enum SNRWeightComb SNRweightComb, double K1, double K2) {

	//Static variables are shared between threads (as they are saved in the data segment).
	//To avoid race conditions, we need to set the directive '#pragma omp threadprivate()'
	//directive to make OpenMP create a local (static) copy for each thread
	static char str[50];
	#pragma omp threadprivate(str)

    switch(SNRweightComb) {
        case SNRWeightCombi:
            strcpy(str,"Use SNR from measurement of first frequency");
            break;
        case SNRWeightCombj:
            strcpy(str,"Use SNR from measurement of second frequency");
            break;
        case SNRWeightCombMax:
            strcpy(str,"Use maximum SNR of both measurements");
            break;
        case SNRWeightCombMin:
            strcpy(str,"Use minimum SNR of both measurements");
            break;
        case SNRWeightCombMean:
            strcpy(str,"Use the mean SNR of both measurements");
            break;
        case SNRWeightCombUser:
            sprintf(str,"SNR = SNRi*%.3g + SNRj*%.3g",K1,K2);
            break;
        default:
            strcpy(str,"Unknown");
            break;
    }

    return str;
}


/*****************************************************************************
 * Name        : getEGM96Height
 * Description : Get the geoid height from EGM96 model for a given position by
 *                linear interpolation of the EGM6 grid
 * Parameters  :
 * Name                            |Da|Unit|Description
 * double *position                 I  º/m  Position in ENU (N: [-90..90], E: [-180..180], H: meters)
 * Returned value (double)          O    m  EGM96 Geoid height
 *****************************************************************************/
double getEGM96Height (double *position) {

	double 	latitude,longitude;
	double	difflat,difflon;
	double	height;
	int		latpos[2],lonpos[2];

	//Convert latitude to 0..180
	latitude=position[0]+90.;
	longitude=position[1];

	//Convert longitude to 0..360 only if longitude is negative
	if (longitude<0.) {
		longitude+=360.;
	}

	latpos[0]=(int)latitude;
	latpos[1]=(latpos[0]+1)%181;

	lonpos[0]=(int)longitude;
	lonpos[1]=(lonpos[0]+1)%361;

	difflat=latitude-(double)(latpos[0]);
	difflon=longitude-(double)(lonpos[0]);


	height=(1.-difflat)*(1.-difflon)*EGM96grid[latpos[0]][lonpos[0]] + difflat*(1.-difflon)*EGM96grid[latpos[1]][lonpos[0]] +
		(1.-difflat)*difflon*EGM96grid[latpos[0]][lonpos[1]]  + difflat*difflon*EGM96grid[latpos[1]][lonpos[1]];

	return height;
}


/*****************************************************************************
 * Name        : EGM96grid
 * Description : Table with EGM96 geoid model grid values (in metres) with a resolution
 *               of 1x1 degree (this is the geoid model used by Google Earth). 
 *               The values are gathered from the following web page: 
 *               http://earth-info.nga.mil/GandG/wgs84/gravitymod/egm96/egm96.html
 *               The website provides the data grid in a 0.25x0.25 degree resolution
 *               in a plain text named "ww15mgh.grd". In order to reduce the size of
 *               the data, the data set has been reduced to a 1x1 degree resolution.
 *               The values are saved in the variable below by ordering them by latitude
 *               and longitude. For example, the first block of data, is the grid values
 *               at latitude -90 for longitudes 0 to 360. The following is at latiude -89
 *               for longitudes 0 to 360, and so on. Note that there is a value on
 *               the right side of the normal 10x10 table, this is the value at longitude=0.
 *               In order to create the list of data below from the "ww15mgh.grd" file,
 *               execute the following command line in Linux or Cygwin (note that the
 *               backslashes at the end are to make the command line interpreter
 *               to escape the newlines):
 *
  awk 'BEGIN{lat=90;lon=-0.25} NF>0 && NR>1 {for(i=1;i<=NF;i++){lon+=.25;if(lon>360)\
  {lon=0.0;lat-=.25} val[lat,lon]=$i;}} END {printf "const double EGM96grid[181][361]={{\n";\
  for(i=-90;i<=90;i++) {if(i>-90){printf"\n"} num=0; for(j=0;j<=360;j++) {num++; printf "%8.3f",val[i,j];\
  if(j==360){printf "}";if(i!=90) {printf ",\n{"}} else{printf ",";if(j==179){num--} else if(num%10==0)\
  {printf "\n"}}}} printf "};"}' ww15mgh.grd
 *              
 *****************************************************************************/
const double EGM96grid[181][361]={{
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534,
 -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534, -29.534},
{
 -27.880, -27.872, -27.865, -27.858, -27.853, -27.848, -27.844, -27.841, -27.838, -27.837,
 -27.836, -27.836, -27.837, -27.838, -27.841, -27.844, -27.848, -27.853, -27.858, -27.865,
 -27.871, -27.879, -27.887, -27.896, -27.906, -27.916, -27.926, -27.938, -27.949, -27.961,
 -27.974, -27.987, -28.001, -28.015, -28.029, -28.043, -28.058, -28.073, -28.088, -28.104,
 -28.119, -28.135, -28.150, -28.166, -28.181, -28.197, -28.212, -28.227, -28.241, -28.256,
 -28.270, -28.284, -28.297, -28.310, -28.322, -28.334, -28.346, -28.357, -28.368, -28.378,
 -28.388, -28.397, -28.406, -28.415, -28.424, -28.432, -28.441, -28.449, -28.457, -28.465,
 -28.473, -28.482, -28.491, -28.500, -28.509, -28.520, -28.530, -28.541, -28.553, -28.566,
 -28.579, -28.593, -28.609, -28.625, -28.642, -28.660, -28.679, -28.699, -28.720, -28.742,
 -28.765, -28.789, -28.814, -28.841, -28.868, -28.896, -28.925, -28.955, -28.985, -29.017,
 -29.049, -29.082, -29.116, -29.150, -29.184, -29.220, -29.255, -29.291, -29.327, -29.364,
 -29.401, -29.437, -29.474, -29.511, -29.548, -29.584, -29.621, -29.657, -29.693, -29.728,
 -29.763, -29.798, -29.832, -29.865, -29.898, -29.930, -29.961, -29.992, -30.021, -30.050,
 -30.078, -30.105, -30.131, -30.156, -30.179, -30.202, -30.224, -30.244, -30.263, -30.281,
 -30.298, -30.313, -30.327, -30.340, -30.351, -30.361, -30.370, -30.378, -30.384, -30.388,
 -30.392, -30.394, -30.395, -30.394, -30.392, -30.389, -30.385, -30.380, -30.373, -30.366,
 -30.357, -30.348, -30.338, -30.326, -30.314, -30.302, -30.289, -30.275, -30.261, -30.246,
 -30.232, -30.217, -30.202, -30.187, -30.172, -30.157, -30.143, -30.129, -30.115, -30.102, -30.089,
 -30.078, -30.066, -30.056, -30.046, -30.037, -30.029, -30.021, -30.015, -30.009, -30.005,
 -30.001, -29.999, -29.997, -29.997, -29.997, -29.999, -30.001, -30.005, -30.009, -30.015,
 -30.021, -30.028, -30.037, -30.046, -30.056, -30.067, -30.079, -30.092, -30.105, -30.119,
 -30.134, -30.150, -30.166, -30.183, -30.200, -30.217, -30.235, -30.253, -30.272, -30.290,
 -30.309, -30.327, -30.346, -30.364, -30.382, -30.399, -30.416, -30.432, -30.448, -30.463,
 -30.476, -30.489, -30.501, -30.512, -30.522, -30.531, -30.538, -30.544, -30.549, -30.552,
 -30.554, -30.554, -30.553, -30.551, -30.547, -30.541, -30.534, -30.526, -30.516, -30.504,
 -30.491, -30.476, -30.460, -30.442, -30.422, -30.401, -30.378, -30.354, -30.327, -30.299,
 -30.270, -30.238, -30.205, -30.170, -30.133, -30.094, -30.053, -30.011, -29.967, -29.922,
 -29.874, -29.825, -29.775, -29.723, -29.670, -29.616, -29.561, -29.505, -29.448, -29.391,
 -29.334, -29.276, -29.219, -29.161, -29.104, -29.048, -28.993, -28.939, -28.886, -28.835,
 -28.785, -28.737, -28.692, -28.648, -28.606, -28.567, -28.530, -28.496, -28.464, -28.434,
 -28.407, -28.383, -28.360, -28.340, -28.323, -28.307, -28.293, -28.282, -28.272, -28.263,
 -28.256, -28.250, -28.246, -28.242, -28.239, -28.237, -28.235, -28.234, -28.232, -28.231,
 -28.230, -28.229, -28.227, -28.225, -28.222, -28.219, -28.216, -28.212, -28.207, -28.202,
 -28.196, -28.189, -28.182, -28.174, -28.165, -28.156, -28.146, -28.136, -28.125, -28.114,
 -28.102, -28.090, -28.078, -28.065, -28.053, -28.040, -28.027, -28.014, -28.001, -27.988,
 -27.976, -27.963, -27.951, -27.940, -27.928, -27.917, -27.907, -27.897, -27.888, -27.880},
{
 -24.897, -24.818, -24.741, -24.665, -24.591, -24.521, -24.454, -24.390, -24.330, -24.272,
 -24.217, -24.164, -24.112, -24.060, -24.007, -23.953, -23.897, -23.839, -23.777, -23.712,
 -23.644, -23.572, -23.497, -23.420, -23.341, -23.261, -23.182, -23.103, -23.028, -22.956,
 -22.890, -22.830, -22.779, -22.736, -22.703, -22.681, -22.670, -22.671, -22.684, -22.708,
 -22.744, -22.792, -22.849, -22.916, -22.992, -23.075, -23.164, -23.259, -23.357, -23.458,
 -23.561, -23.663, -23.765, -23.865, -23.963, -24.057, -24.147, -24.233, -24.314, -24.390,
 -24.461, -24.527, -24.589, -24.646, -24.699, -24.750, -24.798, -24.844, -24.890, -24.936,
 -24.983, -25.032, -25.084, -25.139, -25.199, -25.264, -25.334, -25.409, -25.490, -25.578,
 -25.671, -25.769, -25.873, -25.981, -26.093, -26.209, -26.327, -26.448, -26.570, -26.693,
 -26.817, -26.940, -27.063, -27.185, -27.306, -27.425, -27.543, -27.659, -27.773, -27.885,
 -27.996, -28.105, -28.213, -28.318, -28.422, -28.525, -28.625, -28.724, -28.822, -28.917,
 -29.011, -29.104, -29.194, -29.283, -29.369, -29.454, -29.537, -29.619, -29.698, -29.775,
 -29.851, -29.925, -29.997, -30.067, -30.136, -30.203, -30.267, -30.331, -30.392, -30.452,
 -30.509, -30.565, -30.620, -30.672, -30.723, -30.772, -30.819, -30.865, -30.909, -30.951,
 -30.992, -31.032, -31.070, -31.107, -31.142, -31.176, -31.208, -31.239, -31.268, -31.295,
 -31.321, -31.344, -31.365, -31.382, -31.397, -31.408, -31.415, -31.418, -31.415, -31.407,
 -31.393, -31.372, -31.345, -31.310, -31.268, -31.219, -31.161, -31.096, -31.024, -30.944,
 -30.858, -30.766, -30.668, -30.566, -30.459, -30.350, -30.238, -30.125, -30.011, -29.897, -29.785,
 -29.674, -29.565, -29.459, -29.357, -29.258, -29.163, -29.072, -28.986, -28.905, -28.829,
 -28.760, -28.696, -28.640, -28.590, -28.548, -28.514, -28.489, -28.473, -28.466, -28.468,
 -28.480, -28.502, -28.533, -28.574, -28.622, -28.679, -28.743, -28.814, -28.889, -28.969,
 -29.051, -29.135, -29.220, -29.303, -29.385, -29.464, -29.538, -29.607, -29.669, -29.725,
 -29.773, -29.813, -29.844, -29.866, -29.879, -29.883, -29.878, -29.865, -29.843, -29.813,
 -29.775, -29.731, -29.681, -29.626, -29.567, -29.505, -29.440, -29.373, -29.307, -29.240,
 -29.175, -29.111, -29.050, -28.992, -28.937, -28.887, -28.841, -28.799, -28.761, -28.728,
 -28.700, -28.675, -28.654, -28.637, -28.622, -28.609, -28.597, -28.586, -28.574, -28.561,
 -28.545, -28.526, -28.503, -28.475, -28.440, -28.399, -28.351, -28.295, -28.231, -28.159,
 -28.080, -27.994, -27.902, -27.803, -27.700, -27.594, -27.485, -27.374, -27.264, -27.155,
 -27.049, -26.946, -26.848, -26.755, -26.669, -26.591, -26.520, -26.457, -26.403, -26.358,
 -26.322, -26.295, -26.277, -26.268, -26.267, -26.275, -26.289, -26.311, -26.339, -26.371,
 -26.408, -26.446, -26.486, -26.525, -26.562, -26.595, -26.622, -26.642, -26.653, -26.653,
 -26.642, -26.617, -26.580, -26.529, -26.464, -26.386, -26.296, -26.195, -26.085, -25.968,
 -25.845, -25.719, -25.594, -25.471, -25.352, -25.242, -25.141, -25.051, -24.975, -24.914,
 -24.868, -24.837, -24.822, -24.823, -24.838, -24.865, -24.904, -24.952, -25.006, -25.066,
 -25.127, -25.189, -25.248, -25.302, -25.350, -25.389, -25.419, -25.438, -25.445, -25.441,
 -25.426, -25.399, -25.361, -25.313, -25.257, -25.194, -25.125, -25.051, -24.975, -24.897},
{
 -21.973, -21.917, -21.844, -21.753, -21.645, -21.520, -21.380, -21.229, -21.070, -20.906,
 -20.744, -20.587, -20.440, -20.307, -20.191, -20.093, -20.014, -19.951, -19.902, -19.863,
 -19.827, -19.791, -19.746, -19.690, -19.616, -19.522, -19.407, -19.272, -19.121, -18.956,
 -18.785, -18.615, -18.453, -18.308, -18.185, -18.092, -18.031, -18.006, -18.017, -18.062,
 -18.137, -18.237, -18.356, -18.487, -18.622, -18.756, -18.882, -18.997, -19.098, -19.184,
 -19.256, -19.317, -19.369, -19.418, -19.468, -19.523, -19.589, -19.669, -19.765, -19.877,
 -20.006, -20.149, -20.304, -20.467, -20.634, -20.802, -20.966, -21.123, -21.271, -21.409,
 -21.536, -21.653, -21.762, -21.866, -21.966, -22.068, -22.173, -22.284, -22.404, -22.534,
 -22.674, -22.826, -22.988, -23.159, -23.337, -23.522, -23.710, -23.900, -24.092, -24.284,
 -24.475, -24.667, -24.859, -25.052, -25.247, -25.447, -25.651, -25.861, -26.078, -26.303,
 -26.534, -26.772, -27.015, -27.262, -27.512, -27.761, -28.009, -28.254, -28.492, -28.723,
 -28.946, -29.158, -29.360, -29.551, -29.731, -29.901, -30.060, -30.210, -30.352, -30.485,
 -30.611, -30.731, -30.845, -30.955, -31.060, -31.162, -31.260, -31.356, -31.450, -31.542,
 -31.633, -31.722, -31.811, -31.900, -31.987, -32.075, -32.162, -32.249, -32.335, -32.420,
 -32.503, -32.583, -32.659, -32.730, -32.795, -32.852, -32.902, -32.941, -32.971, -32.990,
 -32.999, -32.998, -32.990, -32.974, -32.954, -32.932, -32.910, -32.891, -32.875, -32.865,
 -32.861, -32.862, -32.868, -32.876, -32.882, -32.884, -32.877, -32.859, -32.824, -32.773,
 -32.702, -32.613, -32.508, -32.389, -32.262, -32.131, -32.002, -31.883, -31.777, -31.689, -31.622,
 -31.578, -31.554, -31.549, -31.558, -31.576, -31.596, -31.613, -31.622, -31.618, -31.598,
 -31.563, -31.514, -31.456, -31.393, -31.332, -31.281, -31.246, -31.233, -31.247, -31.290,
 -31.360, -31.455, -31.568, -31.691, -31.815, -31.929, -32.024, -32.089, -32.119, -32.107,
 -32.053, -31.958, -31.826, -31.665, -31.484, -31.295, -31.109, -30.936, -30.787, -30.668,
 -30.583, -30.534, -30.518, -30.529, -30.558, -30.597, -30.633, -30.656, -30.656, -30.626,
 -30.561, -30.461, -30.328, -30.168, -29.989, -29.803, -29.621, -29.454, -29.312, -29.204,
 -29.133, -29.100, -29.103, -29.134, -29.186, -29.246, -29.303, -29.345, -29.363, -29.348,
 -29.297, -29.211, -29.091, -28.944, -28.781, -28.613, -28.450, -28.306, -28.189, -28.108,
 -28.067, -28.067, -28.105, -28.176, -28.272, -28.382, -28.494, -28.597, -28.679, -28.732,
 -28.747, -28.719, -28.646, -28.528, -28.369, -28.172, -27.944, -27.691, -27.423, -27.144,
 -26.863, -26.583, -26.311, -26.049, -25.799, -25.562, -25.338, -25.129, -24.934, -24.754,
 -24.589, -24.443, -24.316, -24.212, -24.133, -24.084, -24.067, -24.085, -24.140, -24.232,
 -24.360, -24.521, -24.712, -24.925, -25.153, -25.386, -25.614, -25.827, -26.012, -26.159,
 -26.259, -26.304, -26.289, -26.211, -26.071, -25.871, -25.619, -25.322, -24.991, -24.638,
 -24.275, -23.915, -23.569, -23.246, -22.955, -22.701, -22.486, -22.310, -22.173, -22.069,
 -21.995, -21.944, -21.912, -21.891, -21.879, -21.870, -21.863, -21.855, -21.847, -21.839,
 -21.831, -21.825, -21.822, -21.823, -21.829, -21.840, -21.856, -21.876, -21.900, -21.927,
 -21.955, -21.982, -22.007, -22.028, -22.044, -22.052, -22.051, -22.039, -22.013, -21.973},
{
 -18.154, -18.173, -18.189, -18.195, -18.181, -18.139, -18.059, -17.938, -17.776, -17.576,
 -17.349, -17.108, -16.868, -16.644, -16.452, -16.300, -16.194, -16.130, -16.102, -16.097,
 -16.099, -16.091, -16.059, -15.991, -15.881, -15.728, -15.535, -15.312, -15.070, -14.821,
 -14.578, -14.353, -14.153, -13.983, -13.847, -13.746, -13.679, -13.643, -13.636, -13.656,
 -13.699, -13.763, -13.844, -13.938, -14.040, -14.144, -14.247, -14.341, -14.425, -14.493,
 -14.546, -14.583, -14.607, -14.622, -14.634, -14.648, -14.670, -14.705, -14.758, -14.832,
 -14.927, -15.044, -15.180, -15.332, -15.498, -15.672, -15.851, -16.032, -16.214, -16.395,
 -16.576, -16.757, -16.943, -17.134, -17.333, -17.542, -17.764, -17.997, -18.243, -18.500,
 -18.767, -19.043, -19.326, -19.614, -19.907, -20.202, -20.500, -20.797, -21.093, -21.387,
 -21.676, -21.961, -22.241, -22.515, -22.787, -23.056, -23.328, -23.604, -23.888, -24.182,
 -24.489, -24.808, -25.141, -25.484, -25.835, -26.192, -26.552, -26.913, -27.273, -27.631,
 -27.988, -28.342, -28.695, -29.046, -29.395, -29.740, -30.079, -30.410, -30.730, -31.035,
 -31.323, -31.591, -31.839, -32.067, -32.275, -32.464, -32.638, -32.799, -32.949, -33.092,
 -33.230, -33.366, -33.502, -33.639, -33.778, -33.918, -34.058, -34.195, -34.327, -34.449,
 -34.559, -34.652, -34.729, -34.789, -34.836, -34.874, -34.910, -34.953, -35.010, -35.087,
 -35.187, -35.311, -35.455, -35.610, -35.767, -35.912, -36.034, -36.120, -36.164, -36.158,
 -36.103, -35.999, -35.851, -35.664, -35.447, -35.206, -34.947, -34.678, -34.405, -34.135,
 -33.878, -33.640, -33.433, -33.264, -33.140, -33.067, -33.044, -33.070, -33.138, -33.239, -33.361,
 -33.495, -33.631, -33.762, -33.886, -34.002, -34.114, -34.224, -34.338, -34.458, -34.587,
 -34.724, -34.867, -35.014, -35.162, -35.311, -35.459, -35.609, -35.765, -35.928, -36.102,
 -36.285, -36.474, -36.662, -36.837, -36.986, -37.096, -37.150, -37.136, -37.045, -36.871,
 -36.613, -36.274, -35.862, -35.390, -34.873, -34.331, -33.785, -33.260, -32.784, -32.381,
 -32.078, -31.894, -31.842, -31.925, -32.134, -32.447, -32.828, -33.234, -33.612, -33.912,
 -34.089, -34.111, -33.962, -33.649, -33.198, -32.657, -32.083, -31.541, -31.092, -30.786,
 -30.653, -30.698, -30.902, -31.225, -31.607, -31.980, -32.279, -32.447, -32.444, -32.255,
 -31.889, -31.375, -30.763, -30.112, -29.480, -28.923, -28.480, -28.174, -28.010, -27.973,
 -28.038, -28.168, -28.327, -28.481, -28.604, -28.682, -28.712, -28.700, -28.662, -28.614,
 -28.574, -28.557, -28.569, -28.611, -28.675, -28.747, -28.808, -28.836, -28.812, -28.715,
 -28.533, -28.258, -27.887, -27.426, -26.888, -26.291, -25.660, -25.023, -24.410, -23.851,
 -23.372, -22.996, -22.737, -22.599, -22.578, -22.659, -22.822, -23.039, -23.286, -23.537,
 -23.776, -23.994, -24.191, -24.374, -24.559, -24.759, -24.987, -25.249, -25.540, -25.843,
 -26.133, -26.375, -26.532, -26.567, -26.452, -26.174, -25.730, -25.139, -24.431, -23.650,
 -22.844, -22.063, -21.349, -20.737, -20.245, -19.879, -19.629, -19.476, -19.394, -19.352,
 -19.324, -19.286, -19.223, -19.129, -19.003, -18.853, -18.690, -18.526, -18.373, -18.241,
 -18.138, -18.065, -18.022, -18.007, -18.014, -18.036, -18.067, -18.100, -18.131, -18.156,
 -18.172, -18.178, -18.175, -18.166, -18.154, -18.141, -18.133, -18.132, -18.140, -18.154},
{
 -15.493, -15.402, -15.386, -15.452, -15.590, -15.771, -15.950, -16.080, -16.117, -16.031,
 -15.811, -15.473, -15.055, -14.608, -14.189, -13.850, -13.622, -13.515, -13.510, -13.569,
 -13.641, -13.672, -13.619, -13.458, -13.186, -12.823, -12.400, -11.958, -11.533, -11.152,
 -10.827, -10.557, -10.330, -10.130,  -9.944,  -9.765,  -9.596,  -9.446,  -9.331,  -9.267,
  -9.263,  -9.322,  -9.435,  -9.587,  -9.755,  -9.917, -10.052, -10.146, -10.196, -10.204,
 -10.180, -10.137, -10.091, -10.053, -10.034, -10.037, -10.062, -10.104, -10.158, -10.215,
 -10.268, -10.313, -10.348, -10.374, -10.396, -10.421, -10.455, -10.507, -10.584, -10.692,
 -10.834, -11.014, -11.234, -11.492, -11.788, -12.119, -12.480, -12.863, -13.260, -13.661,
 -14.059, -14.448, -14.825, -15.194, -15.563, -15.940, -16.335, -16.754, -17.199, -17.665,
 -18.142, -18.616, -19.072, -19.500, -19.894, -20.256, -20.595, -20.921, -21.250, -21.591,
 -21.952, -22.332, -22.726, -23.128, -23.529, -23.924, -24.314, -24.704, -25.102, -25.521,
 -25.969, -26.454, -26.976, -27.533, -28.115, -28.711, -29.309, -29.896, -30.463, -31.004,
 -31.515, -31.997, -32.451, -32.879, -33.283, -33.662, -34.017, -34.342, -34.632, -34.881,
 -35.084, -35.238, -35.346, -35.416, -35.464, -35.507, -35.565, -35.654, -35.783, -35.949,
 -36.139, -36.329, -36.492, -36.601, -36.639, -36.604, -36.511, -36.390, -36.282, -36.227,
 -36.257, -36.381, -36.586, -36.832, -37.060, -37.208, -37.223, -37.079, -36.784, -36.386,
 -35.963, -35.611, -35.419, -35.449, -35.719, -36.190, -36.773, -37.347, -37.779, -37.958,
 -37.820, -37.369, -36.673, -35.861, -35.090, -34.520, -34.276, -34.420, -34.946, -35.775, -36.775,
 -37.793, -38.682, -39.336, -39.704, -39.797, -39.683, -39.464, -39.252, -39.141, -39.192,
 -39.420, -39.795, -40.257, -40.728, -41.138, -41.436, -41.603, -41.651, -41.621, -41.567,
 -41.544, -41.594, -41.733, -41.952, -42.221, -42.489, -42.705, -42.824, -42.817, -42.679,
 -42.424, -42.084, -41.699, -41.308, -40.943, -40.617, -40.331, -40.071, -39.814, -39.539,
 -39.232, -38.891, -38.531, -38.178, -37.867, -37.631, -37.496, -37.470, -37.538, -37.666,
 -37.799, -37.875, -37.836, -37.639, -37.264, -36.728, -36.074, -35.374, -34.711, -34.165,
 -33.799, -33.642, -33.686, -33.885, -34.161, -34.423, -34.578, -34.555, -34.313, -33.850,
 -33.203, -32.437, -31.629, -30.860, -30.191, -29.660, -29.270, -28.999, -28.803, -28.630,
 -28.432, -28.177, -27.853, -27.474, -27.076, -26.706, -26.419, -26.262, -26.269, -26.450,
 -26.795, -27.268, -27.818, -28.382, -28.899, -29.315, -29.592, -29.713, -29.681, -29.516,
 -29.249, -28.911, -28.527, -28.110, -27.659, -27.163, -26.605, -25.977, -25.286, -24.557,
 -23.842, -23.206, -22.719, -22.443, -22.416, -22.639, -23.074, -23.649, -24.267, -24.828,
 -25.245, -25.465, -25.477, -25.315, -25.049, -24.766, -24.554, -24.475, -24.553, -24.765,
 -25.045, -25.296, -25.409, -25.287, -24.869, -24.139, -23.135, -21.944, -20.686, -19.492,
 -18.485, -17.754, -17.342, -17.241, -17.398, -17.724, -18.116, -18.474, -18.722, -18.816,
 -18.751, -18.553, -18.274, -17.972, -17.700, -17.496, -17.370, -17.312, -17.292, -17.272,
 -17.215, -17.095, -16.906, -16.657, -16.377, -16.102, -15.869, -15.706, -15.631, -15.641,
 -15.719, -15.836, -15.957, -16.047, -16.080, -16.045, -15.945, -15.799, -15.637, -15.493},
{
 -14.636, -14.776, -14.844, -14.835, -14.763, -14.651, -14.517, -14.370, -14.200, -13.991,
 -13.726, -13.402, -13.039, -12.674, -12.359, -12.137, -12.031, -12.028, -12.081, -12.118,
 -12.061, -11.843, -11.433, -10.839, -10.107,  -9.311,  -8.532,  -7.839,  -7.275,  -6.854,
  -6.558,  -6.357,  -6.216,  -6.109,  -6.029,  -5.985,  -5.997,  -6.085,  -6.259,  -6.510,
  -6.812,  -7.122,  -7.396,  -7.596,  -7.702,  -7.716,  -7.657,  -7.553,  -7.434,  -7.318,
  -7.208,  -7.096,  -6.970,  -6.823,  -6.660,  -6.503,  -6.382,  -6.328,  -6.357,  -6.460,
  -6.604,  -6.738,  -6.806,  -6.767,  -6.610,  -6.361,  -6.076,  -5.831,  -5.697,  -5.726,
  -5.935,  -6.307,  -6.795,  -7.338,  -7.880,  -8.384,  -8.835,  -9.239,  -9.619,  -9.998,
 -10.393, -10.808, -11.237, -11.669, -12.093, -12.508, -12.922, -13.351, -13.808, -14.303,
 -14.834, -15.387, -15.939, -16.469, -16.965, -17.429, -17.874, -18.326, -18.805, -19.323,
 -19.877, -20.445, -20.998, -21.510, -21.965, -22.372, -22.757, -23.161, -23.625, -24.176,
 -24.822, -25.544, -26.307, -27.071, -27.804, -28.491, -29.137, -29.766, -30.404, -31.073,
 -31.778, -32.508, -33.232, -33.917, -34.528, -35.045, -35.461, -35.786, -36.037, -36.237,
 -36.407, -36.563, -36.716, -36.872, -37.031, -37.194, -37.357, -37.514, -37.657, -37.777,
 -37.870, -37.935, -37.977, -38.008, -38.039, -38.075, -38.112, -38.131, -38.105, -38.001,
 -37.793, -37.474, -37.055, -36.571, -36.065, -35.579, -35.143, -34.765, -34.436, -34.140,
 -33.875, -33.663, -33.558, -33.638, -33.982, -34.645, -35.623, -36.850, -38.192, -39.483,
 -40.560, -41.311, -41.704, -41.796, -41.719, -41.639, -41.710, -42.024, -42.591, -43.335, -44.126,
 -44.815, -45.285, -45.478, -45.410, -45.157, -44.829, -44.533, -44.342, -44.283, -44.336,
 -44.447, -44.557, -44.617, -44.605, -44.525, -44.406, -44.284, -44.193, -44.152, -44.169,
 -44.235, -44.333, -44.441, -44.535, -44.595, -44.604, -44.553, -44.437, -44.261, -44.034,
 -43.770, -43.482, -43.179, -42.867, -42.545, -42.216, -41.884, -41.560, -41.264, -41.012,
 -40.819, -40.685, -40.595, -40.520, -40.425, -40.283, -40.081, -39.828, -39.552, -39.290,
 -39.078, -38.933, -38.849, -38.794, -38.723, -38.589, -38.358, -38.026, -37.614, -37.168,
 -36.741, -36.376, -36.096, -35.891, -35.723, -35.536, -35.272, -34.890, -34.375, -33.747,
 -33.053, -32.357, -31.720, -31.190, -30.783, -30.488, -30.268, -30.073, -29.860, -29.602,
 -29.301, -28.978, -28.672, -28.423, -28.256, -28.179, -28.176, -28.218, -28.274, -28.324,
 -28.364, -28.405, -28.467, -28.568, -28.708, -28.870, -29.016, -29.104, -29.092, -28.960,
 -28.705, -28.349, -27.924, -27.461, -26.983, -26.498, -26.007, -25.509, -25.015, -24.555,
 -24.177, -23.937, -23.879, -24.024, -24.348, -24.788, -25.243, -25.603, -25.768, -25.680,
 -25.335, -24.786, -24.130, -23.486, -22.961, -22.630, -22.515, -22.587, -22.770, -22.968,
 -23.088, -23.058, -22.845, -22.453, -21.918, -21.295, -20.640, -20.002, -19.413, -18.894,
 -18.454, -18.104, -17.852, -17.711, -17.688, -17.781, -17.971, -18.220, -18.477, -18.689,
 -18.807, -18.806, -18.682, -18.459, -18.173, -17.862, -17.555, -17.261, -16.975, -16.678,
 -16.355, -16.002, -15.631, -15.270, -14.954, -14.713, -14.558, -14.480, -14.451, -14.432,
 -14.392, -14.315, -14.208, -14.101, -14.031, -14.028, -14.108, -14.259, -14.450, -14.636},
{
 -12.657, -12.758, -12.759, -12.659, -12.487, -12.289, -12.098, -11.927, -11.756, -11.555,
 -11.301, -11.000, -10.694, -10.450, -10.329, -10.356, -10.502, -10.682, -10.776, -10.668,
 -10.280,  -9.598,  -8.672,  -7.599,  -6.489,  -5.442,  -4.521,  -3.755,  -3.142,  -2.664,
  -2.305,  -2.052,  -1.902,  -1.849,  -1.889,  -2.008,  -2.195,  -2.434,  -2.720,  -3.048,
  -3.418,  -3.819,  -4.233,  -4.624,  -4.952,  -5.179,  -5.285,  -5.273,  -5.172,  -5.021,
  -4.856,  -4.696,  -4.535,  -4.353,  -4.133,  -3.877,  -3.611,  -3.384,  -3.243,  -3.214,
  -3.281,  -3.392,  -3.474,  -3.458,  -3.311,  -3.045,  -2.713,  -2.394,  -2.163,  -2.070,
  -2.138,  -2.357,  -2.701,  -3.137,  -3.633,  -4.159,  -4.684,  -5.182,  -5.633,  -6.032,
  -6.391,  -6.738,  -7.110,  -7.537,  -8.034,  -8.591,  -9.179,  -9.764, -10.321, -10.844,
 -11.351, -11.875, -12.451, -13.102, -13.832, -14.621, -15.437, -16.246, -17.021, -17.750,
 -18.439, -19.099, -19.740, -20.367, -20.981, -21.577, -22.158, -22.734, -23.325, -23.950,
 -24.622, -25.342, -26.098, -26.872, -27.646, -28.411, -29.161, -29.892, -30.600, -31.273,
 -31.902, -32.490, -33.058, -33.645, -34.296, -35.037, -35.857, -36.703, -37.486, -38.116,
 -38.532, -38.732, -38.775, -38.765, -38.812, -38.993, -39.329, -39.775, -40.246, -40.647,
 -40.904, -40.986, -40.911, -40.731, -40.517, -40.332, -40.220, -40.190, -40.213, -40.226,
 -40.149, -39.902, -39.441, -38.774, -37.978, -37.182, -36.538, -36.177, -36.167, -36.494,
 -37.066, -37.750, -38.416, -38.980, -39.424, -39.790, -40.155, -40.592, -41.141, -41.788,
 -42.481, -43.141, -43.696, -44.106, -44.367, -44.514, -44.605, -44.694, -44.819, -44.989, -45.188,
 -45.389, -45.565, -45.708, -45.824, -45.935, -46.058, -46.197, -46.338, -46.455, -46.519,
 -46.517, -46.458, -46.369, -46.286, -46.238, -46.234, -46.263, -46.300, -46.319, -46.310,
 -46.277, -46.241, -46.216, -46.205, -46.184, -46.112, -45.946, -45.662, -45.270, -44.819,
 -44.381, -44.024, -43.794, -43.687, -43.661, -43.645, -43.574, -43.410, -43.156, -42.852,
 -42.555, -42.313, -42.147, -42.045, -41.971, -41.881, -41.748, -41.567, -41.358, -41.148,
 -40.962, -40.798, -40.636, -40.436, -40.161, -39.794, -39.344, -38.850, -38.361, -37.919,
 -37.539, -37.203, -36.866, -36.473, -35.981, -35.376, -34.679, -33.929, -33.175, -32.452,
 -31.776, -31.152, -30.581, -30.075, -29.660, -29.372, -29.235, -29.247, -29.374, -29.554,
 -29.717, -29.807, -29.797, -29.696, -29.536, -29.354, -29.184, -29.043, -28.941, -28.884,
 -28.884, -28.951, -29.081, -29.249, -29.402, -29.476, -29.414, -29.197, -28.854, -28.460,
 -28.111, -27.883, -27.803, -27.836, -27.901, -27.896, -27.747, -27.429, -26.984, -26.496,
 -26.069, -25.783, -25.673, -25.718, -25.848, -25.974, -26.004, -25.875, -25.562, -25.076,
 -24.458, -23.764, -23.048, -22.358, -21.728, -21.187, -20.761, -20.475, -20.350, -20.391,
 -20.574, -20.838, -21.094, -21.242, -21.204, -20.946, -20.497, -19.934, -19.356, -18.847,
 -18.448, -18.148, -17.906, -17.679, -17.454, -17.263, -17.166, -17.221, -17.442, -17.780,
 -18.122, -18.329, -18.281, -17.921, -17.282, -16.471, -15.635, -14.908, -14.373, -14.034,
 -13.832, -13.675, -13.480, -13.202, -12.854, -12.492, -12.187, -11.996, -11.935, -11.976,
 -12.066, -12.143, -12.173, -12.154, -12.119, -12.115, -12.177, -12.311, -12.489, -12.657},
{
 -10.648, -10.529, -10.339, -10.107,  -9.869,  -9.660,  -9.510,  -9.433,  -9.418,  -9.432,
  -9.422,  -9.337,  -9.143,  -8.839,  -8.454,  -8.031,  -7.612,  -7.216,  -6.838,  -6.446,
  -5.999,  -5.456,  -4.791,  -3.996,  -3.093,  -2.135,  -1.195,  -0.360,   0.294,   0.725,
   0.938,   0.979,   0.910,   0.784,   0.622,   0.415,   0.148,  -0.176,  -0.517,  -0.811,
  -1.005,  -1.084,  -1.088,  -1.084,  -1.132,  -1.243,  -1.376,  -1.465,  -1.454,  -1.333,
  -1.136,  -0.915,  -0.703,  -0.494,  -0.252,   0.053,   0.416,   0.786,   1.080,   1.228,
   1.204,   1.042,   0.823,   0.637,   0.552,   0.590,   0.734,   0.942,   1.177,   1.415,
   1.647,   1.860,   2.026,   2.098,   2.026,   1.778,   1.363,   0.827,   0.234,  -0.360,
  -0.934,  -1.493,  -2.057,  -2.636,  -3.221,  -3.794,  -4.351,  -4.920,  -5.552,  -6.296,
  -7.168,  -8.141,  -9.156, -10.152, -11.100, -12.009, -12.908, -13.819, -14.735, -15.625,
 -16.455, -17.207, -17.897, -18.558, -19.224, -19.910, -20.618, -21.345, -22.097, -22.890,
 -23.737, -24.631, -25.543, -26.435, -27.283, -28.094, -28.901, -29.732, -30.588, -31.428,
 -32.194, -32.854, -33.436, -34.037, -34.784, -35.766, -36.969, -38.255, -39.405, -40.207,
 -40.558, -40.524, -40.326, -40.248, -40.516, -41.193, -42.154, -43.145, -43.907, -44.289,
 -44.309, -44.124, -43.932, -43.863, -43.928, -44.039, -44.085, -44.019, -43.888, -43.785,
 -43.769, -43.795, -43.718, -43.372, -42.687, -41.772, -40.896, -40.384, -40.465, -41.163,
 -42.293, -43.548, -44.642, -45.421, -45.896, -46.182, -46.410, -46.652, -46.903, -47.117,
 -47.260, -47.336, -47.386, -47.447, -47.531, -47.616, -47.668, -47.674, -47.652, -47.643, -47.687,
 -47.800, -47.973, -48.180, -48.399, -48.617, -48.824, -49.009, -49.151, -49.231, -49.242,
 -49.199, -49.132, -49.072, -49.030, -48.995, -48.942, -48.853, -48.732, -48.607, -48.511,
 -48.459, -48.436, -48.397, -48.289, -48.078, -47.765, -47.389, -47.006, -46.668, -46.401,
 -46.200, -46.036, -45.872, -45.684, -45.470, -45.245, -45.035, -44.858, -44.716, -44.583,
 -44.422, -44.192, -43.873, -43.478, -43.051, -42.653, -42.336, -42.124, -42.000, -41.912,
 -41.800, -41.615, -41.335, -40.959, -40.502, -39.977, -39.396, -38.774, -38.141, -37.536,
 -37.001, -36.554, -36.183, -35.846, -35.490, -35.086, -34.635, -34.169, -33.730, -33.336,
 -32.975, -32.607, -32.190, -31.707, -31.184, -30.679, -30.262, -29.981, -29.841, -29.800,
 -29.787, -29.730, -29.585, -29.347, -29.048, -28.738, -28.460, -28.239, -28.086, -28.007,
 -28.015, -28.129, -28.352, -28.653, -28.958, -29.170, -29.216, -29.084, -28.837, -28.587,
 -28.437, -28.425, -28.503, -28.567, -28.516, -28.310, -27.983, -27.623, -27.309, -27.069,
 -26.868, -26.634, -26.313, -25.907, -25.482, -25.137, -24.954, -24.956, -25.096, -25.285,
 -25.430, -25.472, -25.408, -25.267, -25.079, -24.852, -24.562, -24.190, -23.759, -23.348,
 -23.078, -23.040, -23.235, -23.531, -23.710, -23.558, -22.980, -22.060, -21.031, -20.172,
 -19.664, -19.507, -19.530, -19.484, -19.186, -18.610, -17.899, -17.284, -16.951, -16.946,
 -17.146, -17.316, -17.221, -16.730, -15.870, -14.805, -13.757, -12.917, -12.370, -12.082,
 -11.937, -11.802, -11.592, -11.299, -10.978, -10.708, -10.538, -10.469, -10.455, -10.436,
 -10.377, -10.284, -10.196, -10.163, -10.211, -10.332, -10.485, -10.615, -10.677, -10.648},
{
  -9.301,  -9.116,  -8.906,  -8.614,  -8.231,  -7.801,  -7.400,  -7.085,  -6.869,  -6.722,
  -6.597,  -6.449,  -6.249,  -5.974,  -5.599,  -5.111,  -4.520,  -3.868,  -3.218,  -2.626,
  -2.112,  -1.653,  -1.201,  -0.705,  -0.140,   0.489,   1.147,   1.775,   2.298,   2.638,
   2.740,   2.602,   2.289,   1.916,   1.602,   1.425,   1.398,   1.480,   1.619,   1.780,
   1.944,   2.092,   2.195,   2.226,   2.192,   2.149,   2.170,   2.296,   2.493,   2.678,
   2.787,   2.846,   2.976,   3.312,   3.892,   4.606,   5.238,   5.595,   5.623,   5.417,
   5.141,   4.906,   4.711,   4.483,   4.176,   3.835,   3.573,   3.487,   3.589,   3.802,
   4.039,   4.271,   4.534,   4.860,   5.202,   5.415,   5.334,   4.882,   4.128,   3.246,
   2.409,   1.697,   1.090,   0.526,  -0.023,  -0.535,  -0.987,  -1.419,  -1.949,  -2.714,
  -3.777,  -5.072,  -6.432,  -7.681,  -8.739,  -9.648, -10.527, -11.477, -12.512, -13.566,
 -14.550, -15.425, -16.225, -17.022, -17.869, -18.760, -19.659, -20.544, -21.447, -22.435,
 -23.552, -24.768, -25.981, -27.081, -28.017, -28.830, -29.604, -30.401, -31.212, -31.976,
 -32.652, -33.274, -33.948, -34.779, -35.793, -36.903, -37.955, -38.809, -39.415, -39.819,
 -40.127, -40.442, -40.835, -41.349, -42.017, -42.854, -43.834, -44.858, -45.766, -46.400,
 -46.695, -46.724, -46.676, -46.744, -47.002, -47.363, -47.644, -47.717, -47.631, -47.608,
 -47.890, -48.532, -49.293, -49.722, -49.419, -48.318, -46.791, -45.493, -45.019, -45.584,
 -46.917, -48.430, -49.552, -50.013, -49.921, -49.609, -49.397, -49.413, -49.578, -49.729,
 -49.762, -49.692, -49.617, -49.619, -49.704, -49.804, -49.851, -49.837, -49.823, -49.891, -50.075,
 -50.331, -50.566, -50.700, -50.724, -50.703, -50.721, -50.822, -50.977, -51.111, -51.150,
 -51.072, -50.916, -50.748, -50.619, -50.543, -50.500, -50.461, -50.412, -50.360, -50.323,
 -50.319, -50.353, -50.408, -50.453, -50.450, -50.375, -50.225, -50.018, -49.785, -49.545,
 -49.299, -49.030, -48.726, -48.392, -48.046, -47.708, -47.382, -47.057, -46.726, -46.401,
 -46.111, -45.871, -45.652, -45.375, -44.947, -44.321, -43.538, -42.717, -41.999, -41.477,
 -41.158, -40.980, -40.860, -40.741, -40.604, -40.443, -40.246, -39.985, -39.638, -39.208,
 -38.730, -38.245, -37.789, -37.377, -37.010, -36.689, -36.402, -36.113, -35.758, -35.265,
 -34.606, -33.832, -33.063, -32.424, -31.963, -31.607, -31.204, -30.620, -29.837, -28.976,
 -28.225, -27.720, -27.452, -27.286, -27.053, -26.677, -26.225, -25.859, -25.724, -25.854,
 -26.158, -26.490, -26.739, -26.873, -26.923, -26.935, -26.937, -26.938, -26.954, -27.004,
 -27.084, -27.141, -27.082, -26.830, -26.389, -25.865, -25.407, -25.107, -24.936, -24.758,
 -24.436, -23.935, -23.362, -22.896, -22.680, -22.735, -22.968, -23.254, -23.525, -23.785,
 -24.072, -24.389, -24.685, -24.891, -24.980, -24.978, -24.936, -24.867, -24.724, -24.438,
 -23.987, -23.435, -22.893, -22.443, -22.079, -21.715, -21.267, -20.732, -20.206, -19.825,
 -19.669, -19.699, -19.779, -19.748, -19.497, -19.007, -18.334, -17.577, -16.842, -16.230,
 -15.815, -15.617, -15.578, -15.570, -15.450, -15.136, -14.653, -14.120, -13.679, -13.402,
 -13.255, -13.131, -12.931, -12.633, -12.305, -12.043, -11.898, -11.839, -11.777, -11.631,
 -11.386, -11.097, -10.834, -10.632, -10.467, -10.284, -10.048,  -9.776,  -9.516,  -9.301},
{
  -7.084,  -6.976,  -6.689,  -6.205,  -5.629,  -5.090,  -4.653,  -4.283,  -3.907,  -3.481,
  -3.031,  -2.621,  -2.288,  -2.015,  -1.736,  -1.385,  -0.938,  -0.429,   0.070,   0.501,
   0.850,   1.156,   1.482,   1.870,   2.314,   2.769,   3.181,   3.518,   3.762,   3.891,
   3.869,   3.667,   3.304,   2.870,   2.499,   2.302,   2.318,   2.512,   2.828,   3.234,
   3.724,   4.266,   4.768,   5.101,   5.200,   5.127,   5.052,   5.123,   5.341,   5.562,
   5.635,   5.582,   5.634,   6.080,   7.005,   8.163,   9.080,   9.363,   8.962,   8.199,
   7.517,   7.164,   7.034,   6.802,   6.226,   5.363,   4.537,   4.084,   4.101,   4.402,
   4.700,   4.848,   4.928,   5.116,   5.474,   5.846,   5.966,   5.675,   5.039,   4.297,
   3.665,   3.200,   2.803,   2.340,   1.760,   1.109,   0.460,  -0.163,  -0.792,  -1.481,
  -2.281,  -3.231,  -4.361,  -5.664,  -7.071,  -8.450,  -9.681, -10.733, -11.679, -12.639,
 -13.682, -14.784, -15.878, -16.924, -17.943, -18.970, -20.005, -21.011, -21.970, -22.934,
 -23.998, -25.214, -26.519, -27.768, -28.834, -29.703, -30.470, -31.246, -32.076, -32.924,
 -33.756, -34.595, -35.518, -36.561, -37.646, -38.605, -39.294, -39.705, -39.986, -40.337,
 -40.885, -41.610, -42.391, -43.122, -43.794, -44.499, -45.344, -46.362, -47.463, -48.477,
 -49.244, -49.691, -49.868, -49.899, -49.902, -49.932, -49.969, -49.975, -49.953, -49.974,
 -50.124, -50.417, -50.747, -50.930, -50.859, -50.631, -50.555, -50.965, -51.967, -53.293,
 -54.414, -54.856, -54.486, -53.574, -52.575, -51.834, -51.417, -51.194, -51.024, -50.887,
 -50.844, -50.931, -51.093, -51.245, -51.355, -51.474, -51.667, -51.942, -52.235, -52.467, -52.608,
 -52.674, -52.694, -52.681, -52.647, -52.619, -52.629, -52.679, -52.721, -52.698, -52.602,
 -52.492, -52.450, -52.505, -52.612, -52.697, -52.714, -52.677, -52.614, -52.531, -52.415,
 -52.277, -52.182, -52.205, -52.359, -52.547, -52.615, -52.462, -52.119, -51.729, -51.422,
 -51.218, -51.008, -50.663, -50.155, -49.588, -49.118, -48.820, -48.628, -48.392, -48.016,
 -47.538, -47.098, -46.803, -46.607, -46.318, -45.728, -44.772, -43.593, -42.455, -41.595,
 -41.091, -40.863, -40.765, -40.681, -40.567, -40.417, -40.212, -39.908, -39.482, -38.973,
 -38.489, -38.153, -38.035, -38.110, -38.278, -38.412, -38.407, -38.204, -37.790, -37.180,
 -36.403, -35.491, -34.490, -33.470, -32.518, -31.696, -30.985, -30.270, -29.415, -28.371,
 -27.251, -26.262, -25.552, -25.084, -24.663, -24.110, -23.424, -22.793, -22.436, -22.420,
 -22.601, -22.734, -22.651, -22.377, -22.098, -22.045, -22.364, -23.038, -23.880, -24.603,
 -24.948, -24.823, -24.357, -23.824, -23.467, -23.338, -23.293, -23.118, -22.709, -22.141,
 -21.600, -21.228, -21.036, -20.918, -20.763, -20.542, -20.329, -20.240, -20.355, -20.663,
 -21.072, -21.451, -21.699, -21.792, -21.779, -21.721, -21.620, -21.385, -20.888, -20.073,
 -19.045, -18.047, -17.339, -17.057, -17.151, -17.443, -17.747, -17.947, -18.002, -17.899,
 -17.645, -17.302, -17.026, -17.024, -17.408, -18.064, -18.640, -18.736, -18.165, -17.094,
 -15.957, -15.178, -14.911, -14.989, -15.082, -14.944, -14.548, -14.038, -13.579, -13.233,
 -12.951, -12.651, -12.299, -11.930, -11.605, -11.356, -11.156, -10.935, -10.637, -10.258,
  -9.851,  -9.484,  -9.176,  -8.878,  -8.520,  -8.083,  -7.640,  -7.307,  -7.143,  -7.084},
{
  -4.850,  -4.953,  -4.647,  -3.991,  -3.271,  -2.710,  -2.299,  -1.891,  -1.399,  -0.881,
  -0.437,  -0.081,   0.269,   0.676,   1.126,   1.561,   1.967,   2.372,   2.788,   3.170,
   3.473,   3.726,   4.023,   4.430,   4.903,   5.311,   5.554,   5.641,   5.662,   5.683,
   5.694,   5.633,   5.464,   5.218,   4.975,   4.813,   4.775,   4.863,   5.049,   5.304,
   5.616,   5.989,   6.415,   6.838,   7.161,   7.308,   7.281,   7.163,   7.062,   7.050,
   7.163,   7.433,   7.890,   8.521,   9.223,   9.820,  10.140,  10.102,   9.739,   9.160,
   8.497,   7.858,   7.301,   6.819,   6.358,   5.876,   5.396,   5.006,   4.784,   4.726,
   4.739,   4.729,   4.691,   4.700,   4.820,   5.005,   5.106,   4.977,   4.596,   4.072,
   3.554,   3.117,   2.720,   2.267,   1.689,   0.995,   0.259,  -0.436,  -1.028,  -1.508,
  -1.947,  -2.509,  -3.409,  -4.774,  -6.514,  -8.316,  -9.820, -10.868, -11.605, -12.357,
 -13.369, -14.643, -15.984, -17.192, -18.213, -19.138, -20.094, -21.150, -22.282, -23.418,
 -24.499, -25.518, -26.530, -27.605, -28.760, -29.928, -31.011, -31.974, -32.881, -33.836,
 -34.890, -36.001, -37.095, -38.123, -39.066, -39.885, -40.519, -40.945, -41.245, -41.574,
 -42.067, -42.745, -43.540, -44.376, -45.247, -46.199, -47.254, -48.352, -49.361, -50.160,
 -50.735, -51.200, -51.712, -52.327, -52.931, -53.320, -53.375, -53.179, -52.945, -52.850,
 -52.912, -53.032, -53.104, -53.064, -52.867, -52.505, -52.118, -52.060, -52.719, -54.157,
 -55.885, -57.079, -57.144, -56.148, -54.725, -53.557, -52.918, -52.652, -52.511, -52.428,
 -52.465, -52.602, -52.672, -52.550, -52.337, -52.309, -52.643, -53.239, -53.819, -54.201, -54.426,
 -54.640, -54.873, -55.005, -54.936, -54.755, -54.671, -54.792, -54.990, -55.045, -54.892,
 -54.708, -54.724, -54.955, -55.151, -55.017, -54.493, -53.809, -53.286, -53.087, -53.119,
 -53.167, -53.087, -52.909, -52.785, -52.837, -53.040, -53.211, -53.127, -52.678, -51.963,
 -51.222, -50.680, -50.407, -50.323, -50.318, -50.359, -50.477, -50.648, -50.728, -50.517,
 -49.921, -49.045, -48.132, -47.389, -46.851, -46.382, -45.789, -44.962, -43.944, -42.906,
 -42.061, -41.545, -41.336, -41.257, -41.089, -40.719, -40.198, -39.660, -39.180, -38.705,
 -38.128, -37.426, -36.717, -36.178, -35.902, -35.816, -35.737, -35.503, -35.090, -34.633,
 -34.300, -34.138, -33.991, -33.618, -32.912, -32.017, -31.194, -30.553, -29.946, -29.158,
 -28.184, -27.245, -26.493, -25.765, -24.727, -23.310, -21.880, -20.868, -20.218, -19.353,
 -17.844, -16.121, -15.379, -16.588, -19.518, -22.759, -24.775, -25.028, -24.177, -23.302,
 -22.982, -23.061, -23.108, -22.938, -22.673, -22.430, -22.105, -21.521, -20.710, -19.972,
 -19.600, -19.589, -19.628, -19.399, -18.871, -18.309, -18.010, -18.050, -18.257, -18.416,
 -18.451, -18.433, -18.444, -18.481, -18.484, -18.424, -18.324, -18.181, -17.915, -17.417,
 -16.662, -15.786, -15.041, -14.659, -14.730, -15.161, -15.715, -16.127, -16.239, -16.061,
 -15.731, -15.377, -15.045, -14.737, -14.515, -14.499, -14.723, -14.999, -14.975, -14.394,
 -13.318, -12.106, -11.151, -10.645, -10.533, -10.647, -10.828, -10.952, -10.909, -10.646,
 -10.229,  -9.831,  -9.598,  -9.529,  -9.488,  -9.344,  -9.077,  -8.753,  -8.410,  -8.023,
  -7.577,  -7.128,  -6.753,  -6.432,  -6.036,  -5.487,  -4.909,  -4.563,  -4.597,  -4.850},
{
  -2.015,  -2.543,  -2.412,  -1.675,  -0.833,  -0.272,   0.079,   0.525,   1.173,   1.823,
   2.245,   2.463,   2.693,   3.083,   3.576,   4.046,   4.487,   4.998,   5.617,   6.213,
   6.624,   6.861,   7.133,   7.622,   8.251,   8.736,   8.879,   8.784,   8.732,   8.864,
   9.044,   9.040,   8.796,   8.471,   8.253,   8.194,   8.228,   8.291,   8.357,   8.422,
   8.499,   8.640,   8.879,   9.141,   9.249,   9.098,   8.795,   8.582,   8.585,   8.709,
   8.794,   8.847,   9.051,   9.524,  10.125,  10.517,  10.450,   9.965,   9.350,   8.904,
   8.709,   8.595,   8.316,   7.777,   7.123,   6.591,   6.270,   6.006,   5.574,   4.947,
   4.356,   4.060,   4.081,   4.182,   4.142,   3.989,   3.925,   4.024,   4.066,   3.748,
   3.035,   2.226,   1.628,   1.208,   0.644,  -0.264,  -1.311,  -2.057,  -2.272,  -2.155,
  -2.149,  -2.584,  -3.508,  -4.773,  -6.199,  -7.647,  -9.039, -10.354, -11.617, -12.846,
 -14.048, -15.284, -16.704, -18.401, -20.192, -21.601, -22.232, -22.198, -22.130, -22.650,
 -23.823, -25.191, -26.337, -27.325, -28.515, -30.034, -31.582, -32.809, -33.744, -34.718,
 -35.914, -37.151, -38.171, -38.994, -39.842, -40.764, -41.519, -41.903, -42.063, -42.362,
 -42.966, -43.707, -44.385, -45.080, -46.044, -47.319, -48.603, -49.544, -50.102, -50.561,
 -51.200, -52.048, -52.947, -53.785, -54.576, -55.335, -55.980, -56.393, -56.567, -56.596,
 -56.544, -56.370, -56.042, -55.638, -55.250, -54.785, -54.010, -52.914, -51.969, -51.843,
 -52.740, -54.074, -54.922, -54.815, -54.074, -53.389, -53.178, -53.371, -53.689, -53.962,
 -54.163, -54.274, -54.241, -54.068, -53.867, -53.805, -53.994, -54.436, -55.028, -55.614, -56.082,
 -56.428, -56.717, -56.970, -57.101, -57.009, -56.731, -56.453, -56.366, -56.500, -56.715,
 -56.820, -56.712, -56.436, -56.149, -56.011, -56.053, -56.132, -56.028, -55.636, -55.045,
 -54.444, -53.945, -53.552, -53.276, -53.213, -53.441, -53.828, -54.029, -53.753, -53.045,
 -52.276, -51.800, -51.629, -51.468, -51.081, -50.602, -50.417, -50.731, -51.253, -51.359,
 -50.617, -49.176, -47.644, -46.578, -46.072, -45.788, -45.336, -44.568, -43.568, -42.460,
 -41.333, -40.324, -39.649, -39.434, -39.537, -39.614, -39.398, -38.892, -38.253, -37.572,
 -36.842, -36.119, -35.598, -35.415, -35.430, -35.306, -34.864, -34.276, -33.883, -33.840,
 -34.017, -34.179, -34.173, -33.915, -33.349, -32.539, -31.753, -31.266, -31.035, -30.679,
 -29.896, -28.828, -27.841, -26.980, -25.847, -24.177, -22.350, -21.049, -20.330, -19.350,
 -17.331, -14.777, -13.423, -14.703, -18.243, -22.002, -23.943, -23.573, -21.973, -20.591,
 -20.134, -20.435, -20.987, -21.388, -21.408, -20.960, -20.171, -19.383, -18.919, -18.797,
 -18.720, -18.392, -17.808, -17.230, -16.912, -16.888, -16.983, -16.978, -16.763, -16.409,
 -16.111, -16.028, -16.113, -16.134, -15.923, -15.599, -15.461, -15.624, -15.820, -15.658,
 -15.058, -14.363, -13.997, -14.079, -14.411, -14.750, -14.983, -15.058, -14.918, -14.598,
 -14.269, -14.040, -13.694, -12.824, -11.325,  -9.691,  -8.658,  -8.512,  -8.841,  -9.003,
  -8.767,  -8.425,  -8.329,  -8.457,  -8.476,  -8.134,  -7.510,  -6.879,  -6.425,  -6.110,
  -5.792,  -5.411,  -5.035,  -4.749,  -4.538,  -4.316,  -4.064,  -3.873,  -3.818,  -3.812,
  -3.676,  -3.363,  -3.025,  -2.803,  -2.595,  -2.173,  -1.564,  -1.162,  -1.353,  -2.015},
{
   1.642,   1.310,   1.357,   1.769,   2.365,   2.978,   3.514,   3.940,   4.303,   4.715,
   5.216,   5.670,   5.911,   6.003,   6.245,   6.853,   7.709,   8.513,   9.120,   9.608,
  10.064,  10.455,  10.776,  11.158,  11.696,  12.255,  12.601,  12.710,  12.756,  12.793,
  12.622,  12.119,  11.531,  11.252,  11.347,  11.489,  11.407,  11.179,  11.018,  10.947,
  10.879,  10.861,  11.015,  11.223,  11.142,  10.659,  10.134,  10.014,  10.272,  10.471,
  10.360,  10.203,  10.391,  10.901,  11.318,  11.314,  10.922,  10.356,   9.786,   9.380,
   9.363,   9.763,  10.174,  10.010,   9.121,   7.996,   7.228,   6.867,   6.454,   5.667,
   4.757,   4.241,   4.279,   4.484,   4.383,   3.948,   3.579,   3.572,   3.738,   3.642,
   3.168,   2.648,   2.384,   2.221,   1.794,   1.104,   0.566,   0.415,   0.320,  -0.221,
  -1.250,  -2.354,  -3.272,  -4.189,  -5.367,  -6.712,  -7.929,  -8.992, -10.237, -11.960,
 -14.056, -16.109, -17.730, -18.760, -19.254, -19.385, -19.386, -19.511, -19.996, -20.997,
 -22.489, -24.220, -25.862, -27.287, -28.655, -30.159, -31.739, -33.154, -34.307, -35.344,
 -36.396, -37.399, -38.292, -39.207, -40.242, -41.147, -41.542, -41.512, -41.677, -42.485,
 -43.659, -44.616, -45.291, -46.184, -47.563, -48.999, -49.881, -50.183, -50.459, -51.163,
 -52.240, -53.349, -54.267, -54.980, -55.575, -56.158, -56.779, -57.339, -57.646, -57.632,
 -57.431, -57.163, -56.766, -56.178, -55.573, -55.166, -54.806, -54.095, -53.038, -52.315,
 -52.611, -53.765, -54.909, -55.423, -55.441, -55.379, -55.328, -55.220, -55.277, -55.820,
 -56.628, -56.953, -56.397, -55.489, -55.107, -55.501, -56.173, -56.643, -56.957, -57.379, -57.905,
 -58.302, -58.464, -58.483, -58.436, -58.294, -58.073, -57.896, -57.859, -57.925, -58.013,
 -58.104, -58.238, -58.442, -58.689, -58.892, -58.902, -58.597, -58.033, -57.439, -57.021,
 -56.778, -56.576, -56.343, -56.119, -55.935, -55.740, -55.496, -55.234, -54.975, -54.637,
 -54.121, -53.454, -52.768, -52.113, -51.358, -50.331, -49.050, -47.758, -46.730, -46.041,
 -45.518, -44.927, -44.199, -43.490, -42.995, -42.715, -42.443, -41.973, -41.268, -40.396,
 -39.361, -38.135, -36.882, -36.029, -35.948, -36.550, -37.262, -37.523, -37.259, -36.844,
 -36.609, -36.523, -36.366, -36.078, -35.789, -35.571, -35.332, -35.022, -34.773, -34.715,
 -34.701, -34.412, -33.728, -32.883, -32.172, -31.633, -31.105, -30.504, -29.890, -29.264,
 -28.490, -27.497, -26.444, -25.547, -24.797, -23.988, -23.040, -22.125, -21.393, -20.697,
 -19.750, -18.572, -17.660, -17.603, -18.498, -19.737, -20.366, -19.773, -18.190, -16.630,
 -16.209, -17.315, -19.280, -20.871, -21.229, -20.446, -19.338, -18.658, -18.487, -18.335,
 -17.790, -17.016, -16.569, -16.710, -17.038, -16.889, -16.080, -15.101, -14.533, -14.427,
 -14.376, -14.110, -13.801, -13.731, -13.872, -13.958, -13.896, -13.872, -14.027, -14.194,
 -14.084, -13.658, -13.152, -12.782, -12.540, -12.341, -12.203, -12.180, -12.210, -12.184,
 -12.122, -12.133, -12.171, -12.017, -11.587, -11.071, -10.601,  -9.921,  -8.672,  -6.976,
  -5.449,  -4.527,  -4.014,  -3.479,  -2.864,  -2.414,  -2.147,  -1.799,  -1.346,  -1.166,
  -1.436,  -1.644,  -1.103,   0.145,   1.304,   1.754,   1.685,   1.658,   1.828,   1.899,
   1.715,   1.523,   1.611,   1.938,   2.265,   2.467,   2.546,   2.456,   2.121,   1.642},
{
   5.903,   5.640,   5.676,   5.916,   6.334,   6.889,   7.382,   7.644,   7.820,   8.213,
   8.829,   9.330,   9.509,   9.612,  10.028,  10.786,  11.554,  12.087,  12.469,  12.862,
  13.230,  13.457,  13.642,  14.023,  14.601,  15.064,  15.201,  15.184,  15.259,  15.253,
  14.723,  13.627,  12.582,  12.264,  12.672,  13.177,  13.232,  12.853,  12.432,  12.298,
  12.487,  12.794,  12.939,  12.796,  12.510,  12.330,  12.303,  12.276,  12.201,  12.261,
  12.602,  13.080,  13.426,  13.512,  13.357,  12.976,  12.405,  11.788,  11.284,  10.936,
  10.726,  10.702,  10.902,  11.174,  11.197,  10.727,   9.787,   8.671,   7.747,   7.174,
   6.777,   6.298,   5.785,   5.520,   5.523,   5.423,   5.006,   4.593,   4.552,   4.686,
   4.563,   4.267,   4.256,   4.495,   4.390,   3.813,   3.475,   3.802,   3.949,   2.697,
   0.108,  -2.446,  -3.925,  -4.599,  -5.264,  -6.159,  -7.062,  -7.946,  -9.017, -10.294,
 -11.624, -13.014, -14.564, -16.078, -17.137, -17.661, -18.104, -18.932, -20.123, -21.394,
 -22.748, -24.442, -26.418, -28.169, -29.377, -30.385, -31.708, -33.272, -34.551, -35.349,
 -35.991, -36.721, -37.418, -38.050, -38.885, -39.955, -40.771, -41.016, -41.180, -41.973,
 -43.352, -44.689, -45.793, -47.091, -48.681, -49.903, -50.209, -50.020, -50.244, -51.193,
 -52.425, -53.487, -54.345, -55.096, -55.717, -56.236, -56.816, -57.463, -57.899, -57.927,
 -57.740, -57.616, -57.479, -57.066, -56.490, -56.252, -56.516, -56.685, -56.070, -54.920,
 -54.367, -55.172, -56.810, -58.057, -58.299, -57.897, -57.444, -57.162, -57.119, -57.463,
 -58.092, -58.456, -58.145, -57.508, -57.252, -57.573, -58.072, -58.464, -58.866, -59.357, -59.737,
 -59.902, -60.033, -60.205, -60.165, -59.765, -59.315, -59.193, -59.337, -59.440, -59.405,
 -59.287, -59.030, -58.685, -58.708, -59.536, -60.873, -61.891, -62.222, -62.307, -62.618,
 -63.010, -63.122, -63.008, -62.892, -62.548, -61.468, -59.673, -57.885, -56.712, -56.013,
 -55.308, -54.484, -53.733, -52.999, -51.942, -50.481, -48.967, -47.683, -46.514, -45.290,
 -44.184, -43.459, -42.984, -42.343, -41.414, -40.609, -40.401, -40.673, -40.699, -39.859,
 -38.341, -36.980, -36.329, -36.128, -35.888, -35.704, -36.027, -36.743, -37.061, -36.538,
 -35.742, -35.519, -35.838, -35.801, -34.753, -33.071, -31.775, -31.506, -31.987, -32.377,
 -32.078, -31.190, -30.236, -29.549, -29.041, -28.498, -27.912, -27.402, -26.953, -26.425,
 -25.780, -25.151, -24.632, -24.126, -23.493, -22.790, -22.238, -21.911, -21.564, -20.909,
 -19.999, -19.220, -18.858, -18.835, -18.895, -18.878, -18.703, -18.323, -17.899, -17.808,
 -18.184, -18.575, -18.416, -17.818, -17.453, -17.571, -17.526, -16.568, -14.845, -13.273,
 -12.554, -12.687, -13.346, -14.256, -15.045, -15.160, -14.354, -13.134, -12.359, -12.341,
 -12.619, -12.621, -12.309, -12.030, -11.988, -12.060, -12.037, -11.846, -11.561, -11.280,
 -11.018, -10.686, -10.213,  -9.683,  -9.247,  -8.932,  -8.650,  -8.393,  -8.239,  -8.151,
  -7.987,  -7.748,  -7.617,  -7.661,  -7.703,  -7.605,  -7.458,  -7.359,  -7.206,  -6.888,
  -6.473,  -6.003,  -5.319,  -4.343,  -3.373,  -2.757,  -2.364,  -1.759,  -0.948,  -0.501,
  -0.709,  -0.980,  -0.451,   0.968,   2.552,   3.710,   4.525,   5.288,   5.935,   6.244,
   6.306,   6.426,   6.662,   6.824,   6.854,   6.901,   6.975,   6.845,   6.408,   5.903},
{
   9.258,   8.533,   9.011,  10.055,  10.625,  10.628,  10.761,  11.373,  12.096,  12.545,
  12.838,  13.263,  13.805,  14.250,  14.562,  14.904,  15.321,  15.651,  15.796,  15.888,
  16.069,  16.282,  16.407,  16.472,  16.579,  16.735,  16.882,  16.983,  16.989,  16.825,
  16.467,  15.963,  15.426,  15.101,  15.260,  15.800,  16.110,  15.685,  14.838,  14.460,
  14.989,  15.873,  16.235,  15.917,  15.584,  15.720,  15.899,  15.501,  14.856,  14.941,
  15.907,  16.724,  16.594,  15.913,  15.412,  15.085,  14.589,  14.058,  13.779,  13.516,
  12.879,  12.084,  11.734,  12.005,  12.566,  13.037,  13.078,  12.328,  10.875,   9.537,
   9.040,   9.082,   8.832,   8.187,   7.790,   7.870,   7.905,   7.553,   7.147,   6.998,
   6.862,   6.493,   6.154,   6.120,   6.101,   5.714,   5.199,   5.041,   4.999,   4.186,
   2.262,  -0.044,  -1.836,  -3.008,  -4.010,  -5.091,  -6.141,  -7.067,  -7.987,  -9.127,
 -10.604, -12.222, -13.595, -14.625, -15.653, -16.911, -18.185, -19.337, -20.714, -22.571,
 -24.520, -26.072, -27.330, -28.633, -29.918, -30.967, -31.943, -33.120, -34.359, -35.298,
 -35.873, -36.295, -36.726, -37.280, -38.134, -39.324, -40.561, -41.509, -42.140, -42.704,
 -43.497, -44.723, -46.310, -47.755, -48.443, -48.375, -48.307, -48.884, -49.906, -50.765,
 -51.359, -52.080, -53.058, -53.961, -54.608, -55.253, -56.079, -56.842, -57.268, -57.477,
 -57.731, -58.012, -58.163, -58.194, -58.190, -58.147, -58.164, -58.421, -58.598, -57.818,
 -55.835, -53.935, -53.752, -55.297, -56.984, -57.622, -57.538, -57.568, -57.878, -58.199,
 -58.527, -58.967, -59.293, -59.200, -58.803, -58.545, -58.721, -59.338, -60.231, -61.076, -61.530,
 -61.638, -61.882, -62.525, -63.160, -63.288, -63.068, -63.019, -63.188, -63.206, -63.038,
 -63.046, -63.281, -63.369, -63.232, -63.317, -63.862, -64.425, -64.479, -64.127, -63.864,
 -63.878, -63.908, -63.726, -63.419, -63.150, -62.904, -62.609, -62.354, -62.232, -62.040,
 -61.381, -60.082, -58.284, -56.183, -54.031, -52.288, -51.246, -50.491, -49.376, -48.025,
 -47.113, -46.622, -45.777, -44.449, -43.561, -43.666, -44.097, -43.959, -43.119, -41.841,
 -40.357, -39.123, -38.617, -38.572, -38.263, -37.682, -37.436, -37.463, -37.015, -36.105,
 -35.716, -36.202, -36.592, -35.959, -34.580, -33.236, -32.161, -31.186, -30.342, -29.751,
 -29.261, -28.628, -27.883, -27.234, -26.764, -26.427, -26.183, -25.966, -25.643, -25.162,
 -24.594, -23.938, -23.102, -22.187, -21.498, -21.094, -20.676, -20.088, -19.554, -19.206,
 -18.803, -18.167, -17.519, -17.138, -17.032, -17.019, -16.848, -16.293, -15.511, -15.102,
 -15.328, -15.512, -14.792, -13.247, -11.667, -10.436,  -9.298,  -8.091,  -7.039,  -6.250,
  -5.534,  -4.930,  -5.041,  -6.427,  -8.713, -10.606, -11.074, -10.398,  -9.756,  -9.797,
 -10.074,  -9.987,  -9.675,  -9.565,  -9.555,  -9.273,  -8.827,  -8.631,  -8.656,  -8.517,
  -8.156,  -7.833,  -7.564,  -7.125,  -6.561,  -6.121,  -5.786,  -5.335,  -4.777,  -4.303,
  -3.930,  -3.549,  -3.166,  -2.837,  -2.517,  -2.183,  -1.913,  -1.746,  -1.680,  -1.771,
  -1.991,  -2.077,  -1.837,  -1.495,  -1.376,  -1.414,  -1.306,  -1.022,  -0.804,  -0.728,
  -0.521,   0.184,   1.505,   3.110,   4.426,   5.269,   6.143,   7.563,   9.198,  10.204,
  10.395,  10.458,  10.861,  11.194,  11.013,  10.695,  10.804,  11.005,  10.464,   9.258},
{
  10.905,  10.927,  12.056,  13.417,  14.113,  14.298,  14.603,  15.108,  15.555,  15.960,
  16.435,  16.866,  17.195,  17.561,  17.953,  18.190,  18.315,  18.541,  18.842,  19.000,
  18.901,  18.566,  18.170,  18.030,  18.223,  18.428,  18.527,  18.810,  19.167,  18.966,
  18.208,  17.762,  18.006,  18.355,  18.391,  18.297,  18.117,  17.796,  17.802,  18.539,
  19.457,  19.765,  19.591,  19.555,  19.692,  19.565,  19.028,  18.370,  17.993,  18.120,
  18.552,  18.754,  18.440,  17.875,  17.427,  17.144,  16.934,  16.827,  16.879,  17.014,
  17.082,  17.048,  17.049,  17.187,  17.387,  17.572,  17.679,  17.287,  15.811,  13.555,
  11.863,  11.604,  12.237,  12.853,  13.180,  13.122,  12.438,  11.335,  10.425,   9.853,
   9.313,   8.802,   8.507,   8.254,   7.821,   7.365,   6.954,   6.302,   5.360,   4.513,
   3.840,   2.869,   1.320,  -0.453,  -1.949,  -3.090,  -4.199,  -5.532,  -6.957,  -8.211,
  -9.383, -10.764, -12.300, -13.733, -15.152, -16.752, -18.324, -19.699, -21.321, -23.556,
 -25.864, -27.422, -28.175, -28.687, -29.364, -30.228, -31.133, -31.982, -32.871, -33.907,
 -34.898, -35.620, -36.307, -37.314, -38.478, -39.460, -40.357, -41.253, -41.749, -41.771,
 -42.230, -43.877, -45.926, -46.807, -46.178, -45.389, -45.852, -47.484, -49.047, -49.703,
 -49.759, -49.965, -50.600, -51.474, -52.416, -53.435, -54.528, -55.563, -56.379, -56.966,
 -57.454, -57.871, -58.088, -58.179, -58.498, -59.067, -59.280, -58.708, -57.741, -56.789,
 -55.334, -52.712, -49.664, -48.159, -49.490, -52.935, -56.413, -58.299, -58.582, -58.377,
 -58.503, -58.863, -59.064, -59.075, -59.102, -59.275, -59.575, -59.883, -60.156, -60.592, -61.355,
 -62.204, -62.829, -63.402, -64.200, -64.893, -64.970, -64.707, -64.873, -65.518, -65.854,
 -65.498, -65.051, -65.128, -65.443, -65.353, -64.865, -64.530, -64.565, -64.655, -64.545,
 -64.316, -64.060, -63.745, -63.424, -63.216, -63.036, -62.671, -62.168, -61.765, -61.388,
 -60.672, -59.635, -58.839, -58.553, -58.282, -57.542, -56.569, -55.710, -54.695, -53.179,
 -51.477, -50.134, -49.316, -49.048, -49.213, -48.950, -47.269, -44.751, -43.200, -43.058,
 -42.769, -41.360, -39.862, -39.320, -39.151, -38.464, -37.529, -36.961, -36.710, -36.623,
 -36.956, -37.701, -38.173, -37.751, -36.492, -34.894, -33.414, -32.231, -31.305, -30.622,
 -30.211, -29.841, -29.189, -28.424, -27.995, -27.770, -27.201, -26.269, -25.347, -24.388,
 -23.168, -21.913, -20.884, -19.971, -19.190, -18.774, -18.535, -18.126, -17.774, -17.706,
 -17.370, -16.390, -15.523, -15.564, -16.164, -16.573, -16.496, -15.777, -14.435, -13.195,
 -12.777, -12.740, -12.147, -10.937,  -9.612,  -8.286,  -6.989,  -6.090,  -5.590,  -4.790,
  -3.349,  -1.912,  -1.162,  -1.029,  -1.342,  -2.430,  -4.372,  -6.265,  -7.149,  -7.250,
  -7.426,  -7.736,  -7.589,  -6.988,  -6.503,  -6.221,  -5.786,  -5.248,  -4.943,  -4.805,
  -4.579,  -4.338,  -4.204,  -4.051,  -3.853,  -3.784,  -3.811,  -3.764,  -3.735,  -3.845,
  -3.873,  -3.643,  -3.335,  -2.993,  -2.347,  -1.428,  -0.682,  -0.338,  -0.296,  -0.537,
  -1.016,  -1.413,  -1.543,  -1.568,  -1.449,  -0.889,  -0.056,   0.518,   0.875,   1.514,
   2.332,   2.704,   2.660,   2.967,   3.970,   5.178,   6.078,   6.695,   7.253,   7.829,
   8.504,   9.521,  10.996,  12.547,  13.524,  13.676,  13.300,  12.659,  11.750,  10.905},
{
  13.100,  13.842,  14.836,  15.862,  16.758,  17.657,  18.434,  18.532,  17.964,  17.613,
  17.910,  18.298,  18.505,  18.917,  19.368,  19.313,  19.190,  19.667,  20.134,  19.948,
  19.793,  19.899,  19.376,  18.860,  20.262,  22.714,  23.159,  21.506,  20.510,  20.803,
  20.753,  20.302,  20.663,  21.524,  21.790,  21.572,  21.455,  21.503,  21.866,  22.721,
  23.500,  23.512,  22.912,  22.187,  21.509,  21.039,  21.004,  21.279,  21.532,  21.594,
  21.364,  20.833,  20.243,  19.721,  19.181,  18.852,  18.977,  19.045,  18.698,  18.875,
  20.240,  21.488,  21.297,  20.490,  20.279,  20.322,  20.175,  20.103,  19.378,  16.814,
  13.560,  12.126,  12.787,  13.938,  14.918,  15.703,  15.562,  14.408,  13.471,  13.190,
  12.669,  11.722,  11.092,  10.718,   9.965,   9.032,   8.325,   7.415,   6.048,   4.995,
   4.734,   4.479,   3.414,   1.832,   0.473,  -0.489,  -1.478,  -2.992,  -5.008,  -6.878,
  -8.206,  -9.481, -11.231, -13.174, -14.913, -16.619, -18.399, -20.014, -21.683, -23.969,
 -26.558, -28.286, -28.664, -28.493, -28.785, -29.652, -30.521, -31.114, -31.748, -32.581,
 -33.280, -33.839, -34.756, -35.962, -36.873, -37.765, -39.228, -40.294, -39.617, -38.371,
 -39.032, -41.469, -43.249, -43.408, -43.149, -43.490, -44.567, -46.374, -48.405, -49.548,
 -49.444, -49.012, -49.160, -49.905, -50.890, -51.967, -53.127, -54.319, -55.452, -56.416,
 -57.054, -57.323, -57.464, -57.740, -58.103, -58.404, -58.590, -58.441, -57.610, -56.186,
 -54.700, -53.514, -52.490, -51.383, -50.733, -51.899, -55.077, -57.971, -58.375, -57.246,
 -56.970, -57.775, -58.266, -58.195, -58.424, -59.195, -60.021, -60.555, -60.778, -60.828, -60.938,
 -61.320, -62.030, -62.894, -63.637, -64.221, -64.843, -65.444, -65.725, -65.729, -65.749,
 -65.753, -65.579, -65.410, -65.406, -65.333, -65.068, -64.853, -64.759, -64.559, -64.245,
 -64.020, -63.849, -63.565, -63.220, -62.934, -62.643, -62.276, -61.890, -61.484, -60.992,
 -60.425, -59.787, -59.003, -58.185, -57.607, -57.201, -56.576, -55.763, -55.211, -54.890,
 -54.251, -53.244, -52.389, -51.777, -51.058, -50.276, -49.626, -48.702, -47.172, -45.650,
 -44.674, -43.588, -41.828, -40.296, -39.934, -40.026, -39.442, -38.394, -37.617, -37.005,
 -36.249, -35.637, -35.405, -35.229, -34.812, -34.272, -33.826, -33.473, -33.010, -32.376,
 -31.817, -31.361, -30.563, -29.337, -28.181, -27.159, -25.948, -24.939, -24.595, -24.208,
 -23.200, -22.446, -22.468, -22.241, -21.385, -20.964, -20.962, -20.034, -18.420, -17.824,
 -18.232, -18.082, -17.267, -16.911, -16.916, -16.416, -15.661, -15.252, -14.760, -13.730,
 -12.681, -12.036, -11.381, -10.391,  -9.340,  -8.379,  -7.133,  -5.325,  -3.401,  -2.037,
  -1.100,   0.096,   1.471,   2.317,   2.619,   2.596,   1.518,  -1.004,  -3.493,  -4.436,
  -4.378,  -4.475,  -4.570,  -4.130,  -3.460,  -2.951,  -2.471,  -1.991,  -1.722,  -1.575,
  -1.359,  -1.255,  -1.453,  -1.746,  -1.975,  -2.347,  -2.943,  -3.487,  -3.833,  -4.147,
  -4.466,  -4.612,  -4.541,  -4.327,  -3.948,  -3.441,  -2.989,  -2.686,  -2.464,  -2.279,
  -2.090,  -1.885,  -1.815,  -1.961,  -2.031,  -1.678,  -0.923,   0.013,   0.926,   1.694,
   2.565,   3.989,   5.692,   6.715,   6.838,   6.900,   7.348,   7.659,   7.581,   7.797,
   9.016,  11.107,  13.260,  14.573,  14.594,  13.673,  12.704,  12.348,  12.582,  13.100},
{
  12.465,  12.831,  13.289,  14.224,  15.703,  17.063,  17.697,  17.439,  16.519,  15.652,
  15.459,  15.582,  15.444,  15.476,  16.025,  16.249,  16.006,  16.516,  17.530,  17.478,
  17.079,  18.064,  19.419,  19.954,  21.149,  23.043,  22.930,  20.989,  20.423,  21.445,
  21.808,  21.564,  21.889,  22.462,  23.161,  24.514,  25.728,  25.818,  25.525,  25.628,
  25.618,  25.265,  24.804,  23.668,  21.763,  20.844,  21.935,  23.354,  23.606,  23.650,
  24.289,  24.282,  22.892,  21.464,  21.082,  21.065,  20.675,  20.267,  20.175,  20.090,
  19.888,  19.829,  19.888,  19.930,  20.222,  20.844,  21.135,  20.365,  18.463,  16.034,
  14.086,  13.190,  12.875,  12.684,  13.058,  14.073,  14.791,  14.825,  14.800,  14.903,
  14.703,  14.154,  13.472,  12.663,  11.781,  10.936,  10.020,   8.900,   7.625,   6.367,
   5.443,   4.976,   4.485,   3.453,   2.071,   0.716,  -0.783,  -2.647,  -4.554,  -6.152,
  -7.598,  -9.170, -10.837, -12.446, -13.997, -15.609, -17.312, -18.920, -20.372, -22.079,
 -24.226, -26.022, -26.757, -27.041, -27.778, -28.800, -29.656, -30.544, -31.545, -32.137,
 -32.178, -32.353, -33.123, -34.044, -34.706, -35.565, -36.909, -37.750, -37.278, -36.714,
 -37.639, -39.277, -39.974, -40.183, -41.154, -42.433, -43.402, -45.007, -47.748, -50.260,
 -51.158, -50.603, -49.909, -50.193, -51.097, -51.576, -51.844, -52.719, -54.004, -55.264,
 -56.359, -56.770, -56.676, -57.156, -57.807, -57.478, -57.002, -57.093, -56.191, -54.056,
 -53.223, -54.678, -56.444, -56.714, -55.586, -54.655, -55.237, -56.051, -55.012, -53.341,
 -53.822, -56.061, -57.632, -58.005, -58.286, -59.090, -60.203, -61.093, -61.497, -61.680, -61.764,
 -61.638, -61.695, -62.362, -63.281, -64.124, -65.098, -65.999, -66.291, -66.119, -65.872,
 -65.498, -65.140, -65.138, -65.165, -64.767, -64.309, -64.167, -64.000, -63.612, -63.299,
 -63.095, -62.809, -62.528, -62.305, -61.962, -61.538, -61.185, -60.778, -60.244, -59.797,
 -59.480, -59.103, -58.629, -58.069, -57.317, -56.450, -55.714, -55.112, -54.461, -53.732,
 -52.998, -52.277, -51.629, -51.112, -50.577, -49.776, -48.735, -47.684, -46.704, -45.815,
 -45.164, -44.700, -44.163, -43.597, -43.100, -42.300, -41.116, -40.177, -39.489, -38.229,
 -36.514, -35.286, -34.441, -33.398, -32.647, -32.595, -32.509, -31.999, -31.529, -31.137,
 -30.615, -30.249, -29.985, -29.234, -28.042, -26.951, -25.965, -24.844, -23.753, -22.978,
 -22.525, -22.066, -21.369, -21.010, -21.654, -22.407, -21.664, -19.639, -18.008, -17.407,
 -17.149, -16.694, -16.065, -15.455, -14.982, -14.547, -14.023, -13.439, -12.831, -12.318,
 -12.004, -11.488, -10.319,  -8.785,  -7.260,  -5.647,  -3.932,  -2.102,  -0.047,   1.621,
   2.293,   2.783,   4.021,   5.567,   6.757,   7.070,   5.604,   2.747,   0.590,  -0.111,
  -0.526,  -1.000,  -1.001,  -0.857,  -0.955,  -0.922,  -0.790,  -0.949,  -1.094,  -0.905,
  -0.807,  -1.064,  -1.358,  -1.504,  -1.575,  -1.585,  -1.617,  -1.772,  -1.895,  -1.838,
  -1.717,  -1.621,  -1.496,  -1.360,  -1.246,  -1.110,  -0.998,  -1.012,  -1.097,  -1.113,
  -1.018,  -0.903,  -0.901,  -0.970,  -0.916,  -0.751,  -0.615,  -0.397,  -0.011,   0.396,
   1.018,   2.259,   3.910,   5.521,   6.983,   7.892,   7.943,   8.172,   9.403,  10.339,
  10.067,  10.313,  12.214,  13.951,  14.044,  13.489,  13.200,  12.822,  12.407,  12.465},
{
  13.448,  12.856,  12.278,  12.716,  13.417,  13.906,  14.096,  13.633,  12.815,  12.250,
  11.825,  11.374,  11.145,  11.353,  12.169,  13.385,  14.331,  14.980,  15.651,  15.943,
  16.160,  17.738,  20.288,  21.793,  22.179,  22.456,  22.022,  20.584,  19.645,  19.979,
  20.839,  21.542,  21.635,  21.064,  20.783,  21.450,  22.263,  22.345,  21.993,  21.774,
  21.625,  21.556,  22.013,  22.714,  22.793,  22.654,  23.366,  24.451,  24.934,  25.213,
  25.706,  25.745,  24.990,  24.032,  23.500,  23.475,  23.406,  22.859,  22.439,  22.676,
  22.634,  21.654,  21.168,  22.512,  24.531,  25.311,  24.598,  23.005,  20.594,  17.634,
  15.436,  15.005,  15.614,  15.955,  15.918,  15.983,  15.942,  15.548,  15.302,  15.550,
  15.783,  15.454,  14.663,  13.751,  12.910,  12.205,  11.446,  10.339,   8.991,   7.740,
   6.633,   5.619,   4.635,   3.304,   1.407,  -0.598,  -2.268,  -3.641,  -4.786,  -5.860,
  -7.256,  -8.847, -10.100, -11.225, -12.657, -14.027, -15.193, -16.627, -18.069, -19.020,
 -20.047, -21.678, -23.508, -25.066, -26.094, -26.612, -27.216, -28.229, -29.317, -30.267,
 -30.891, -31.090, -31.502, -32.459, -33.263, -33.697, -34.240, -34.810, -35.259, -36.065,
 -37.349, -38.351, -38.613, -39.022, -40.453, -42.107, -43.175, -44.388, -45.944, -47.041,
 -47.872, -48.688, -49.035, -49.673, -51.155, -51.889, -51.516, -51.941, -53.401, -54.709,
 -55.803, -56.504, -56.626, -57.164, -57.766, -57.162, -56.657, -56.969, -55.838, -53.924,
 -54.730, -57.400, -58.865, -59.410, -59.595, -58.693, -57.811, -57.866, -57.558, -56.842,
 -57.135, -58.197, -58.993, -59.325, -59.308, -59.436, -60.184, -60.977, -61.186, -61.030, -60.824,
 -60.988, -62.038, -63.491, -64.413, -64.838, -65.021, -64.752, -64.472, -64.949, -65.538,
 -65.093, -64.146, -63.899, -64.068, -63.800, -63.343, -63.174, -62.969, -62.521, -62.189,
 -62.036, -61.763, -61.399, -61.181, -61.043, -60.739, -60.198, -59.565, -59.016, -58.592,
 -58.214, -57.841, -57.436, -56.917, -56.312, -55.681, -54.933, -54.098, -53.368, -52.660,
 -51.846, -51.151, -50.588, -49.806, -48.855, -48.068, -47.364, -46.507, -45.543, -44.616,
 -43.823, -43.196, -42.667, -42.155, -41.589, -40.970, -40.458, -40.102, -39.678, -39.107,
 -38.608, -38.201, -37.658, -37.001, -36.353, -35.581, -34.845, -34.492, -34.013, -32.950,
 -32.152, -31.771, -30.367, -28.127, -27.042, -26.959, -26.069, -24.605, -23.901, -23.688,
 -23.122, -22.351, -21.806, -21.511, -21.117, -20.080, -18.476, -17.233, -16.613, -15.796,
 -14.702, -14.185, -14.242, -14.249, -14.122, -13.762, -13.014, -12.202, -11.435, -10.706,
 -10.303,  -9.694,  -8.320,  -7.121,  -6.240,  -4.484,  -2.523,  -1.354,   0.394,   2.757,
   3.826,   4.647,   7.186,   9.763,  10.698,  10.820,   9.984,   7.641,   5.289,   3.891,
   2.773,   1.932,   1.475,   0.665,  -0.434,  -1.137,  -1.503,  -1.640,  -1.446,  -1.260,
  -1.197,  -0.951,  -0.674,  -0.601,  -0.467,  -0.189,   0.016,   0.166,   0.323,   0.396,
   0.453,   0.616,   0.763,   0.762,   0.750,   0.846,   0.864,   0.678,   0.509,   0.499,
   0.565,   0.747,   1.008,   1.088,   1.033,   1.107,   1.208,   1.206,   1.230,   1.217,
   1.118,   1.327,   2.098,   3.278,   4.809,   6.582,   8.267,   9.786,  10.939,  11.149,
  10.742,  10.790,  11.062,  10.757,  10.477,  10.738,  10.919,  11.334,  12.616,  13.448},
{
  14.464,  14.414,  14.264,  14.619,  14.793,  14.875,  14.871,  14.591,  14.465,  14.387,
  14.062,  14.352,  15.693,  16.826,  16.837,  16.750,  17.584,  18.122,  17.031,  15.701,
  15.908,  17.223,  18.798,  20.328,  21.156,  21.201,  21.169,  20.999,  20.472,  20.365,
  20.784,  20.688,  20.390,  21.026,  21.722,  21.411,  20.929,  20.403,  19.439,  19.509,
  20.876,  21.755,  22.876,  25.470,  27.139,  26.662,  26.410,  26.877,  26.607,  26.208,
  26.665,  27.406,  28.035,  28.396,  27.868,  26.692,  26.017,  25.924,  25.559,  25.267,
  25.739,  25.829,  25.164,  25.997,  28.245,  28.418,  26.014,  23.767,  21.996,  19.505,
  17.281,  16.600,  16.860,  17.172,  17.131,  16.555,  16.086,  16.157,  15.966,  15.476,
  15.575,  15.693,  14.979,  14.142,  13.633,  12.972,  12.170,  11.433,  10.539,   9.363,
   8.244,   7.532,   6.818,   5.562,   4.112,   2.662,   0.880,  -0.913,  -2.428,  -4.111,
  -5.931,  -7.434,  -8.773, -10.269, -11.752, -13.031, -14.377, -15.807, -16.805, -17.521,
 -18.279, -18.641, -19.380, -21.698, -23.851, -24.262, -24.779, -26.114, -26.860, -27.628,
 -28.964, -29.551, -29.867, -30.926, -31.743, -32.267, -33.073, -33.576, -34.458, -36.139,
 -36.867, -36.881, -37.613, -38.148, -38.887, -41.532, -44.096, -44.102, -43.062, -43.056,
 -43.893, -44.822, -45.870, -47.395, -49.271, -50.802, -51.345, -51.295, -52.058, -53.765,
 -54.940, -55.436, -56.166, -56.660, -56.254, -55.623, -55.562, -55.759, -55.464, -55.273,
 -56.242, -57.343, -57.389, -57.280, -57.465, -57.451, -57.780, -58.657, -59.261, -59.321,
 -58.875, -58.222, -58.193, -58.626, -58.886, -59.386, -60.124, -60.464, -60.780, -61.405, -61.886,
 -62.411, -63.207, -63.684, -63.784, -63.942, -63.929, -63.578, -63.259, -63.127, -63.204,
 -63.471, -63.427, -62.770, -62.015, -61.656, -61.659, -61.799, -61.818, -61.649, -61.418,
 -61.054, -60.581, -60.397, -60.498, -60.358, -59.842, -59.198, -58.467, -57.747, -57.332,
 -57.058, -56.476, -55.773, -55.382, -55.043, -54.372, -53.567, -52.904, -52.298, -51.588,
 -50.831, -50.078, -49.169, -48.212, -47.482, -46.679, -45.660, -44.981, -44.523, -43.625,
 -42.546, -41.733, -41.001, -40.294, -39.775, -39.576, -39.574, -39.030, -37.776, -36.812,
 -36.485, -36.193, -35.740, -35.226, -34.745, -34.453, -34.148, -33.642, -33.163, -32.843,
 -32.666, -32.476, -31.952, -31.243, -30.667, -30.041, -29.253, -28.238, -26.922, -26.098,
 -26.254, -26.642, -27.045, -26.961, -24.734, -21.519, -20.637, -21.583, -21.642, -20.670,
 -19.251, -17.105, -14.743, -12.981, -11.857, -11.024, -10.312,  -9.994,  -9.989,  -9.534,
  -8.542,  -7.356,  -5.857,  -4.497,  -3.831,  -3.088,  -1.582,   0.296,   2.152,   3.688,
   5.080,   7.317,  10.079,  11.917,  12.976,  13.233,  11.130,   7.672,   5.797,   5.219,
   4.359,   3.803,   3.523,   2.282,   0.556,  -0.525,  -0.961,  -0.827,  -0.229,   0.278,
   0.590,   0.919,   1.174,   1.415,   1.717,   1.912,   2.130,   2.448,   2.542,   2.515,
   2.681,   2.756,   2.590,   2.472,   2.309,   1.962,   1.711,   1.618,   1.509,   1.447,
   1.501,   1.619,   1.794,   1.946,   1.990,   1.997,   2.022,   2.054,   2.152,   2.420,
   2.867,   3.313,   3.574,   3.752,   4.100,   4.724,   5.565,   6.569,   7.717,   8.781,
   9.504,  10.004,  10.351,  10.454,  10.546,  10.508,  10.144,  10.719,  12.857,  14.464},
{
  10.528,  10.363,  10.369,  10.803,  11.489,  12.449,  13.127,  13.377,  13.556,  13.581,
  14.179,  15.483,  15.637,  14.923,  15.423,  16.351,  16.305,  15.894,  15.404,  14.913,
  15.082,  15.681,  16.026,  16.184,  16.557,  17.175,  17.725,  18.122,  18.480,  19.236,
  20.986,  22.637,  22.862,  23.108,  24.116,  23.912,  22.397,  21.301,  20.912,  21.367,
  22.729,  23.756,  24.295,  25.430,  26.841,  27.814,  28.243,  27.529,  26.456,  27.265,
  29.431,  31.048,  32.532,  33.533,  32.256,  30.034,  29.509,  29.927,  29.375,  28.274,
  27.667,  27.243,  27.067,  27.915,  28.975,  28.762,  27.239,  25.155,  23.377,  22.010,
  20.298,  18.604,  17.546,  16.551,  15.714,  15.575,  15.836,  16.357,  16.910,  17.005,
  16.694,  16.034,  15.169,  14.431,  13.767,  13.352,  12.977,  12.028,  11.248,  10.889,
  10.021,   9.288,   9.056,   8.402,   7.689,   6.687,   4.526,   2.299,   0.210,  -2.496,
  -4.728,  -6.247,  -7.957,  -9.590, -11.201, -13.019, -14.577, -15.862, -16.858, -17.624,
 -18.392, -18.630, -18.776, -20.145, -22.089, -23.327, -24.164, -25.074, -25.742, -26.278,
 -27.199, -28.120, -28.747, -29.592, -30.422, -31.154, -32.214, -33.194, -34.542, -36.383,
 -36.946, -36.828, -37.317, -37.030, -37.202, -40.109, -43.035, -43.227, -42.069, -41.884,
 -43.030, -43.744, -44.062, -45.696, -47.397, -48.425, -50.206, -51.382, -51.236, -51.843,
 -52.839, -53.456, -54.280, -54.170, -53.356, -53.494, -53.495, -53.147, -54.101, -55.200,
 -55.009, -54.911, -55.865, -57.150, -58.005, -58.172, -57.935, -57.797, -57.862, -57.820,
 -57.621, -57.624, -57.860, -58.202, -58.852, -59.567, -59.733, -59.639, -60.107, -60.902, -61.002,
 -60.607, -61.041, -62.190, -62.678, -62.563, -62.671, -62.530, -61.839, -61.342, -60.985,
 -60.300, -59.741, -59.617, -59.576, -59.360, -59.038, -58.915, -59.012, -58.990, -58.785,
 -58.531, -58.428, -58.533, -58.356, -57.774, -57.359, -57.123, -56.719, -56.292, -55.945,
 -55.456, -54.779, -54.127, -53.572, -53.003, -52.456, -51.906, -51.195, -50.521, -50.045,
 -49.455, -48.691, -47.927, -47.048, -46.138, -45.451, -44.797, -43.977, -43.137, -42.374,
 -41.583, -40.698, -40.031, -39.760, -39.333, -38.537, -37.723, -36.824, -36.121, -35.997,
 -35.708, -35.048, -34.701, -34.341, -33.754, -33.549, -33.452, -32.957, -32.408, -31.985,
 -31.496, -31.018, -30.635, -30.174, -29.486, -28.689, -28.081, -27.613, -26.813, -25.942,
 -25.626, -25.296, -24.813, -24.844, -24.245, -22.405, -20.901, -19.444, -17.354, -16.340,
 -16.327, -15.640, -14.811, -14.487, -14.012, -13.159, -12.040, -10.748,  -9.452,  -8.249,
  -7.367,  -6.691,  -5.525,  -3.825,  -2.539,  -1.633,   0.266,   2.747,   3.977,   4.571,
   5.922,   7.494,   9.216,  11.297,  12.173,  11.016,   9.221,   8.032,   7.286,   6.519,
   6.146,   6.534,   6.761,   5.983,   4.590,   3.391,   2.922,   3.132,   3.584,   4.057,
   4.418,   4.547,   4.501,   4.354,   4.218,   4.171,   4.151,   4.235,   4.470,   4.598,
   4.545,   4.566,   4.650,   4.515,   4.158,   3.758,   3.324,   2.908,   2.647,   2.511,
   2.463,   2.548,   2.691,   2.803,   2.905,   3.000,   3.108,   3.300,   3.579,   3.898,
   4.234,   4.543,   4.840,   5.136,   5.307,   5.445,   5.810,   6.289,   6.738,   7.228,
   7.641,   8.132,   8.975,   9.537,   9.311,   8.912,   8.827,   9.190,  10.021,  10.528},
{
  10.660,  11.018,  11.217,  11.632,  12.038,  12.338,  12.795,  13.473,  13.988,  14.070,
  14.615,  15.627,  15.369,  14.371,  14.549,  15.101,  14.808,  14.248,  14.150,  14.702,
  15.187,  15.046,  15.033,  15.322,  15.614,  16.108,  16.344,  16.417,  16.964,  17.333,
  17.658,  19.213,  21.672,  23.292,  23.286,  23.221,  24.382,  25.269,  25.116,  25.187,
  25.362,  24.834,  24.474,  25.016,  25.537,  25.834,  27.065,  28.687,  29.791,  31.976,
  35.043,  36.330,  36.854,  38.369,  38.374,  36.042,  34.427,  34.218,  33.521,  32.206,
  31.135,  30.159,  29.707,  29.914,  29.130,  27.787,  27.410,  26.605,  25.172,  24.343,
  22.967,  20.945,  19.615,  18.751,  17.999,  17.143,  16.553,  17.274,  18.076,  17.390,
  16.210,  15.340,  14.600,  14.044,  13.483,  12.751,  11.622,  10.438,  10.449,  10.540,
   9.412,   9.371,  10.036,   8.673,   7.123,   6.126,   3.989,   2.357,   1.033,  -1.543,
  -3.376,  -4.917,  -7.564,  -9.397, -10.615, -12.072, -13.064, -14.267, -15.532, -16.621,
 -18.010, -18.370, -17.884, -18.059, -19.212, -21.711, -23.886, -24.689, -25.829, -26.484,
 -26.615, -27.986, -28.761, -28.654, -29.395, -30.085, -31.312, -33.580, -34.949, -35.717,
 -36.619, -37.093, -37.245, -36.988, -37.392, -39.467, -41.195, -41.453, -41.223, -41.005,
 -41.432, -42.631, -43.944, -45.212, -46.396, -47.654, -48.871, -49.619, -50.563, -51.667,
 -51.859, -51.989, -52.729, -52.627, -51.760, -51.325, -50.855, -50.545, -51.482, -52.803,
 -53.372, -53.813, -54.629, -55.027, -54.666, -54.550, -54.862, -54.860, -54.940, -55.334,
 -55.577, -56.060, -56.760, -57.077, -57.365, -57.849, -58.182, -58.531, -58.952, -58.811, -58.358,
 -58.880, -60.115, -60.880, -61.475, -61.901, -61.307, -60.054, -58.970, -58.277, -57.902,
 -57.698, -57.625, -57.513, -57.319, -57.113, -56.649, -56.111, -55.793, -55.670, -55.927,
 -56.223, -56.099, -55.876, -55.654, -55.363, -55.110, -54.832, -54.597, -54.280, -53.687,
 -53.072, -52.519, -52.014, -51.595, -51.045, -50.354, -49.663, -49.024, -48.528, -48.071,
 -47.599, -47.126, -46.462, -45.702, -45.039, -44.354, -43.587, -42.751, -41.853, -40.976,
 -40.190, -39.559, -38.968, -38.252, -37.596, -37.193, -36.822, -36.314, -35.917, -35.609,
 -35.102, -34.561, -34.044, -33.455, -32.944, -32.265, -31.596, -31.459, -31.162, -30.474,
 -30.111, -29.813, -29.385, -29.016, -28.491, -28.031, -27.597, -26.939, -26.421, -25.742,
 -24.863, -24.412, -23.809, -22.852, -22.182, -21.551, -20.807, -19.907, -18.675, -17.542,
 -16.638, -15.887, -15.243, -14.428, -13.719, -13.049, -12.072, -11.279, -10.619,  -9.827,
  -9.152,  -8.233,  -6.920,  -5.249,  -2.912,  -0.641,   0.899,   2.138,   3.386,   4.937,
   6.409,   7.005,   7.379,   7.964,   8.485,   9.205,   9.324,   8.583,   8.478,   8.632,
   8.428,   9.053,   9.857,   9.333,   7.992,   6.813,   6.239,   6.236,   6.511,   6.916,
   7.168,   7.230,   7.208,   7.010,   6.859,   6.815,   6.749,   6.915,   7.181,   7.223,
   7.195,   7.163,   7.128,   7.138,   6.992,   6.662,   6.325,   5.960,   5.589,   5.375,
   5.294,   5.196,   5.145,   5.217,   5.249,   5.211,   5.228,   5.269,   5.361,   5.579,
   5.825,   6.022,   6.185,   6.291,   6.333,   6.336,   6.294,   6.294,   6.457,   6.741,
   7.196,   7.738,   8.008,   8.267,   8.946,   9.663,   9.978,   9.986,  10.154,  10.660},
{
  10.931,  11.346,  11.990,  12.865,  13.143,  13.218,  13.861,  14.576,  14.877,  15.049,
  15.578,  15.922,  15.624,  15.671,  16.062,  15.846,  15.479,  15.595,  15.814,  15.883,
  15.909,  15.764,  15.451,  15.335,  15.593,  16.043,  16.580,  17.239,  17.791,  18.271,
  18.415,  18.117,  19.553,  22.119,  22.228,  21.717,  22.630,  22.819,  22.876,  23.923,
  24.860,  25.516,  26.007,  26.025,  25.721,  26.537,  28.901,  30.834,  31.995,  33.005,
  33.821,  35.690,  38.580,  40.727,  41.107,  38.981,  35.391,  33.043,  32.170,  31.292,
  30.397,  29.585,  28.838,  28.369,  27.201,  25.926,  25.550,  24.384,  23.327,  24.053,
  24.598,  24.511,  24.084,  22.345,  20.211,  18.766,  18.085,  18.171,  18.209,  17.258,
  15.949,  15.769,  15.847,  14.912,  14.283,  13.679,  11.753,  10.259,  10.332,  10.050,
   8.843,   8.429,   7.997,   5.818,   4.065,   3.295,   1.152,  -0.734,  -1.169,  -2.104,
  -2.699,  -3.180,  -5.647,  -8.125,  -9.480, -10.980, -12.085, -13.303, -15.147, -16.550,
 -17.716, -18.332, -17.908, -17.493, -18.452, -21.647, -24.713, -25.656, -26.722, -27.824,
 -27.657, -28.019, -28.908, -29.181, -29.668, -30.644, -31.876, -33.944, -36.387, -37.323,
 -36.708, -36.413, -36.781, -37.526, -38.791, -39.597, -39.826, -40.285, -40.380, -40.295,
 -40.805, -41.666, -42.918, -43.913, -44.743, -46.514, -47.976, -48.642, -49.336, -49.320,
 -48.891, -48.729, -48.704, -48.987, -48.913, -48.472, -48.758, -49.749, -50.460, -50.372,
 -50.622, -51.386, -51.030, -50.273, -50.425, -51.029, -52.025, -52.849, -53.013, -53.383,
 -53.923, -54.359, -55.036, -55.541, -55.713, -56.077, -56.622, -57.080, -57.482, -57.671, -57.674,
 -58.164, -58.732, -58.374, -57.661, -56.999, -56.349, -56.213, -56.108, -55.775, -55.634,
 -55.381, -55.194, -55.164, -54.806, -54.418, -54.276, -54.146, -53.885, -53.713, -53.806,
 -53.673, -53.269, -52.943, -52.571, -52.313, -52.119, -51.739, -51.491, -51.295, -50.952,
 -50.554, -49.939, -49.272, -48.683, -47.958, -47.342, -46.904, -46.400, -46.054, -45.906,
 -45.647, -45.248, -44.717, -44.102, -43.513, -42.890, -42.283, -41.693, -40.958, -40.069,
 -39.127, -38.317, -37.698, -37.126, -36.613, -36.265, -36.018, -35.672, -35.280, -34.977,
 -34.557, -34.056, -33.661, -33.164, -32.487, -31.867, -31.327, -30.621, -29.801, -29.117,
 -28.665, -28.505, -28.233, -27.622, -27.153, -26.939, -26.766, -26.346, -25.648, -25.090,
 -24.623, -24.034, -23.112, -22.000, -21.357, -20.807, -20.001, -19.351, -18.447, -17.212,
 -16.339, -15.539, -14.298, -13.302, -12.802, -11.916, -10.841, -10.173,  -9.503,  -8.608,
  -7.489,  -6.280,  -5.055,  -3.516,  -2.025,  -0.587,   1.565,   3.679,   4.792,   6.190,
   8.352,   9.283,   9.059,   9.411,   9.875,  10.557,  11.783,  11.933,  11.428,  11.376,
  11.226,  11.522,  12.372,  12.602,  11.581,   9.807,   8.821,   8.701,   8.641,   8.822,
   8.956,   8.905,   8.955,   9.087,   9.469,   9.826,   9.797,   9.606,   9.426,   9.251,
   9.058,   8.888,   8.721,   8.590,   8.636,   8.664,   8.434,   8.130,   7.973,   7.829,
   7.551,   7.261,   7.081,   6.982,   6.886,   6.711,   6.461,   6.352,   6.425,   6.433,
   6.499,   6.707,   6.808,   6.968,   7.157,   7.137,   7.182,   7.190,   7.042,   7.181,
   7.620,   8.102,   8.449,   8.700,   9.174,   9.826,  10.408,  10.640,  10.646,  10.931},
{
  12.894,  13.963,  14.534,  14.739,  14.829,  15.054,  15.200,  15.444,  15.863,  16.031,
  16.567,  17.481,  18.247,  18.777,  18.410,  17.792,  17.562,  17.049,  16.946,  17.413,
  17.533,  17.542,  17.588,  17.419,  17.354,  17.720,  18.034,  18.210,  18.666,  19.375,
  19.949,  19.256,  19.298,  21.867,  22.993,  22.679,  23.363,  23.380,  23.309,  23.829,
  24.232,  24.843,  25.408,  26.348,  27.149,  27.357,  28.227,  29.440,  30.181,  30.178,
  30.889,  33.133,  34.267,  35.252,  36.953,  35.986,  33.664,  32.331,  31.037,  29.722,
  28.360,  27.127,  26.255,  25.489,  25.122,  25.098,  24.751,  23.879,  23.598,  23.653,
  22.641,  21.669,  21.261,  20.478,  19.544,  18.768,  17.962,  16.601,  15.331,  15.066,
  15.154,  15.559,  15.744,  15.052,  14.204,  12.956,  11.886,  11.541,  10.358,   8.581,
   7.539,   6.561,   4.893,   3.073,   1.329,   0.141,  -0.089,  -1.257,  -2.843,  -2.271,
  -1.440,  -2.006,  -3.244,  -5.573,  -8.066, -10.020, -12.111, -13.583, -14.361, -14.910,
 -15.249, -16.236, -17.592, -18.949, -20.825, -22.080, -22.511, -23.616, -24.605, -24.923,
 -26.059, -27.363, -28.009, -29.001, -29.991, -30.611, -31.640, -33.404, -35.108, -35.904,
 -36.408, -36.905, -37.409, -38.104, -38.565, -39.170, -40.077, -41.088, -41.967, -42.038,
 -41.796, -42.036, -43.019, -44.243, -44.962, -45.307, -45.358, -45.584, -45.865, -45.557,
 -45.244, -45.677, -47.382, -48.622, -47.772, -47.110, -47.182, -46.817, -46.856, -47.308,
 -47.588, -47.612, -47.503, -47.924, -48.538, -48.875, -49.396, -50.052, -50.830, -51.653,
 -52.080, -52.460, -53.071, -53.711, -54.424, -54.759, -54.618, -54.900, -55.345, -55.306, -54.955,
 -54.190, -53.500, -53.432, -53.565, -53.675, -53.688, -53.607, -53.521, -53.298, -52.996,
 -52.867, -52.756, -52.481, -52.426, -52.427, -52.141, -51.927, -51.821, -51.740, -51.559,
 -51.103, -50.726, -50.486, -50.192, -49.905, -49.721, -49.595, -49.197, -48.699, -48.362,
 -47.832, -47.318, -46.866, -46.122, -45.499, -45.049, -44.590, -44.107, -43.453, -43.234,
 -43.397, -43.120, -42.720, -42.442, -42.042, -41.686, -41.388, -40.911, -40.184, -39.323,
 -38.449, -37.610, -36.872, -36.310, -35.822, -35.285, -34.807, -34.392, -34.009, -33.613,
 -33.128, -32.802, -32.453, -31.953, -31.602, -31.072, -30.441, -29.832, -29.049, -28.399,
 -27.698, -27.113, -26.762, -26.254, -25.900, -25.488, -25.090, -24.794, -24.228, -23.836,
 -23.336, -22.726, -22.021, -20.914, -20.214, -19.666, -19.372, -18.833, -16.879, -15.758,
 -15.702, -14.798, -13.839, -13.136, -12.333, -11.519, -10.579,  -9.530,  -8.441,  -7.479,
  -6.538,  -5.433,  -4.173,  -3.014,  -1.937,  -0.438,   1.005,   2.439,   4.838,   7.394,
   9.203,  10.436,  11.255,  12.152,  12.880,  13.783,  15.313,  15.642,  15.094,  14.811,
  14.202,  14.407,  15.444,  16.203,  15.913,  14.136,  12.979,  12.512,  11.696,  11.336,
  11.037,  11.185,  11.809,  12.062,  12.251,  12.148,  12.000,  11.924,  11.585,  11.502,
  11.637,  11.545,  11.168,  10.988,  10.991,  10.616,  10.190,   9.904,   9.607,   9.320,
   9.015,   8.829,   8.701,   8.551,   8.431,   8.366,   8.317,   8.060,   7.753,   7.607,
   7.331,   7.109,   7.236,   7.373,   7.517,   7.672,   7.709,   7.825,   7.913,   8.023,
   8.286,   8.660,   9.137,   9.497,   9.873,  10.348,  10.780,  11.222,  11.833,  12.894},
{
  13.775,  15.276,  16.613,  17.272,  17.094,  16.980,  17.027,  16.873,  16.730,  16.940,
  17.494,  17.975,  18.334,  18.848,  19.497,  19.747,  19.140,  18.554,  18.897,  19.059,
  18.759,  18.890,  19.025,  19.270,  19.420,  19.423,  19.782,  19.866,  20.276,  20.764,
  21.132,  21.245,  20.047,  20.540,  22.804,  23.599,  24.145,  25.079,  25.577,  25.971,
  26.407,  27.033,  27.589,  27.636,  27.702,  28.160,  28.759,  29.547,  30.371,  30.513,
  30.106,  30.102,  30.317,  30.790,  31.126,  30.475,  30.224,  29.971,  29.088,  28.603,
  27.665,  26.722,  26.579,  26.868,  27.345,  26.894,  25.858,  25.468,  25.441,  25.001,
  23.894,  22.601,  21.552,  20.765,  20.408,  20.261,  19.658,  18.853,  18.352,  17.561,
  16.222,  15.099,  14.286,  13.857,  13.782,  13.108,  12.074,  11.037,   9.740,   8.414,
   7.069,   6.002,   5.134,   4.332,   3.927,   3.432,   2.524,   1.481,   0.421,  -0.567,
  -1.012,  -1.737,  -4.120,  -6.787,  -8.285,  -9.251, -10.559, -11.930, -13.140, -15.107,
 -17.216, -18.461, -19.057, -19.314, -19.717, -20.094, -20.685, -22.364, -24.653, -26.778,
 -28.877, -30.655, -31.556, -31.932, -32.357, -32.864, -33.102, -33.687, -34.608, -35.056,
 -35.483, -35.515, -35.871, -37.257, -38.302, -39.385, -40.653, -41.894, -43.070, -43.698,
 -44.282, -44.405, -44.354, -44.687, -44.907, -45.162, -45.309, -45.807, -46.246, -45.860,
 -45.580, -45.432, -44.984, -44.251, -43.792, -43.714, -43.596, -43.892, -44.238, -44.329,
 -44.371, -44.408, -44.965, -45.726, -46.289, -46.864, -47.483, -48.314, -48.975, -49.295,
 -50.055, -50.706, -51.071, -51.850, -51.921, -51.691, -52.211, -52.336, -51.869, -51.562, -51.524,
 -51.468, -51.445, -51.504, -51.454, -51.383, -51.285, -51.215, -51.051, -50.867, -50.710,
 -50.423, -50.305, -50.166, -49.934, -49.729, -49.484, -49.352, -49.234, -48.987, -48.689,
 -48.573, -48.398, -47.995, -47.612, -47.198, -46.883, -46.659, -46.473, -46.319, -45.918,
 -45.422, -44.893, -44.231, -43.706, -43.316, -42.785, -42.220, -41.946, -41.957, -41.923,
 -41.561, -41.092, -40.757, -40.540, -40.546, -40.415, -39.864, -39.297, -38.823, -38.166,
 -37.353, -36.605, -35.847, -35.111, -34.615, -34.067, -33.459, -33.126, -32.700, -32.229,
 -31.941, -31.465, -30.997, -30.616, -30.273, -30.010, -29.508, -28.904, -28.292, -27.648,
 -27.020, -26.429, -25.976, -25.517, -25.091, -24.538, -23.995, -23.601, -23.162, -22.834,
 -22.285, -21.938, -21.305, -20.218, -19.621, -17.863, -16.793, -17.130, -15.567, -14.610,
 -14.684, -13.661, -12.810, -11.880, -10.970, -10.324,  -9.394,  -8.398,  -7.209,  -6.024,
  -4.965,  -3.804,  -2.440,  -0.980,   0.300,   1.304,   2.373,   3.673,   4.817,   6.204,
   8.375,  10.970,  12.824,  13.833,  14.783,  15.330,  16.393,  17.580,  17.672,  18.101,
  18.429,  18.100,  18.084,  18.830,  18.859,  16.398,  14.635,  14.533,  13.643,  12.973,
  13.241,  14.140,  14.731,  14.344,  13.972,  13.995,  14.297,  14.330,  14.193,  14.226,
  14.175,  14.144,  13.961,  13.595,  12.933,  12.480,  12.507,  12.065,  11.634,  11.534,
  11.281,  11.003,  10.723,  10.499,  10.321,  10.189,   9.983,   9.820,   9.709,   9.363,
   9.083,   8.812,   8.651,   8.654,   8.647,   8.713,   8.595,   8.562,   8.677,   8.843,
   9.093,   9.287,   9.729,  10.219,  10.515,  10.839,  11.269,  11.730,  12.456,  13.775},
{
  13.406,  14.385,  15.894,  16.934,  17.228,  17.334,  17.405,  17.630,  17.919,  18.184,
  18.453,  18.802,  19.261,  19.712,  20.165,  20.495,  20.857,  20.751,  19.954,  19.790,
  19.981,  20.284,  20.643,  20.779,  21.126,  21.232,  21.399,  21.644,  21.991,  22.492,
  22.732,  22.813,  22.604,  22.804,  23.695,  25.008,  26.054,  26.572,  27.265,  27.829,
  28.325,  28.957,  29.257,  29.045,  28.928,  28.863,  28.946,  29.445,  29.206,  28.779,
  28.749,  28.943,  29.145,  28.792,  28.822,  28.984,  29.057,  29.489,  29.691,  29.592,
  29.073,  28.517,  28.334,  28.060,  27.448,  27.034,  26.748,  26.443,  26.392,  25.890,
  25.203,  24.386,  23.367,  22.674,  22.228,  21.979,  21.543,  20.987,  20.073,  18.829,
  17.607,  16.275,  15.388,  14.690,  13.780,  12.889,  11.921,  10.900,   9.729,   8.246,
   6.643,   5.370,   4.411,   3.426,   2.726,   2.396,   1.603,   0.411,   0.019,  -0.337,
  -1.750,  -4.371,  -7.723,  -9.325, -10.140, -11.418, -11.844, -13.174, -15.562, -17.655,
 -19.295, -19.984, -20.640, -21.763, -22.258, -22.243, -22.965, -24.647, -26.651, -28.332,
 -29.756, -31.189, -32.265, -33.079, -33.925, -34.844, -35.738, -36.595, -37.419, -38.154,
 -38.987, -39.646, -40.187, -40.710, -41.024, -41.130, -40.903, -40.697, -40.698, -40.943,
 -41.419, -41.963, -42.873, -43.705, -43.834, -43.365, -42.676, -42.436, -42.382, -42.172,
 -42.018, -41.785, -41.520, -41.112, -40.613, -40.504, -40.966, -41.256, -41.019, -41.057,
 -41.459, -42.130, -42.667, -43.199, -44.073, -44.643, -45.338, -46.009, -46.552, -47.156,
 -47.522, -47.852, -47.878, -48.311, -49.121, -49.520, -49.839, -50.203, -50.534, -50.608, -50.615,
 -50.530, -50.339, -50.151, -49.832, -49.700, -49.583, -49.378, -49.114, -48.759, -48.429,
 -48.029, -47.732, -47.581, -47.499, -47.223, -46.981, -46.884, -46.553, -46.242, -45.868,
 -45.439, -45.148, -44.996, -44.833, -44.534, -44.067, -43.471, -43.169, -43.028, -42.718,
 -42.323, -41.988, -41.789, -41.534, -41.041, -40.517, -40.349, -40.359, -40.043, -39.578,
 -39.243, -39.182, -39.279, -39.068, -38.646, -38.180, -37.812, -37.414, -36.837, -36.251,
 -35.640, -35.075, -34.508, -33.964, -33.403, -32.871, -32.357, -31.787, -31.490, -31.155,
 -30.628, -30.113, -29.644, -29.404, -29.091, -28.650, -28.192, -27.645, -27.164, -26.667,
 -26.168, -25.711, -25.269, -24.735, -24.221, -23.802, -23.190, -22.871, -22.580, -21.967,
 -21.369, -20.636, -20.021, -19.301, -18.294, -17.339, -16.818, -16.197, -14.887, -13.995,
 -13.393, -12.718, -12.173, -11.516, -10.866,  -9.904,  -8.873,  -7.756,  -6.469,  -5.437,
  -4.158,  -2.623,  -1.197,   0.005,   1.187,   2.389,   3.345,   4.459,   5.673,   6.875,
   8.277,   9.663,  11.386,  13.871,  16.133,  17.350,  17.994,  18.427,  19.405,  20.685,
  21.061,  21.328,  21.431,  21.515,  21.822,  20.133,  17.858,  16.655,  15.404,  15.430,
  16.300,  16.395,  16.520,  16.733,  16.793,  16.796,  16.533,  16.355,  16.313,  15.987,
  15.702,  15.622,  15.437,  15.121,  14.804,  14.657,  14.513,  14.504,  14.454,  14.010,
  13.728,  13.350,  12.574,  12.121,  12.259,  12.295,  11.632,  11.238,  11.262,  10.786,
  10.516,  10.459,  10.200,  10.084,  10.051,   9.962,   9.770,   9.763,   9.747,   9.678,
   9.782,  10.144,  10.588,  10.781,  11.001,  11.120,  11.415,  12.053,  12.724,  13.406},
{
  13.503,  13.940,  14.464,  15.039,  16.001,  16.820,  17.421,  17.824,  18.421,  19.253,
  19.944,  20.620,  21.250,  21.692,  21.927,  22.175,  22.181,  22.353,  22.355,  21.941,
  21.813,  22.179,  22.436,  22.449,  22.571,  22.521,  22.840,  23.278,  23.522,  23.920,
  24.372,  24.642,  24.483,  24.767,  25.782,  26.671,  27.431,  28.134,  28.645,  29.277,
  29.917,  30.354,  30.660,  30.669,  30.497,  30.495,  30.717,  30.724,  30.111,  29.413,
  29.196,  29.207,  29.159,  29.093,  29.196,  29.660,  30.060,  30.342,  30.510,  30.434,
  30.184,  29.763,  29.442,  29.221,  28.618,  27.933,  27.459,  26.874,  26.470,  26.063,
  25.628,  25.082,  24.683,  24.381,  24.040,  23.597,  22.743,  21.970,  20.955,  19.718,
  18.471,  17.818,  17.670,  16.950,  15.632,  14.044,  12.717,  11.503,   9.901,   8.273,
   6.971,   5.717,   4.379,   3.156,   2.093,   0.354,  -1.061,  -1.558,  -2.935,  -4.468,
  -5.836,  -7.229,  -8.454,  -9.455, -10.264, -11.572, -12.585, -13.336, -14.525, -16.047,
 -17.850, -19.538, -20.764, -21.854, -22.737, -23.374, -24.219, -25.397, -26.670, -27.846,
 -28.953, -30.060, -31.238, -32.014, -32.472, -33.105, -33.567, -34.195, -35.128, -36.117,
 -37.254, -38.085, -38.576, -39.262, -40.137, -40.251, -40.240, -40.763, -41.075, -41.085,
 -40.978, -40.930, -40.499, -40.020, -39.812, -39.501, -39.439, -39.433, -39.188, -39.020,
 -38.840, -38.649, -38.397, -37.632, -37.207, -37.825, -37.889, -37.409, -37.770, -38.355,
 -38.904, -39.621, -40.342, -41.028, -41.701, -42.441, -42.941, -43.251, -43.839, -44.423,
 -44.838, -45.610, -47.177, -48.631, -48.928, -49.503, -50.186, -49.939, -49.799, -49.884, -49.787,
 -49.572, -49.232, -48.972, -48.840, -48.655, -48.406, -47.971, -47.372, -46.899, -46.582,
 -46.129, -45.784, -45.368, -44.868, -44.559, -44.145, -43.816, -43.656, -43.322, -42.987,
 -42.807, -42.266, -41.825, -41.529, -41.161, -40.853, -40.651, -40.316, -39.857, -39.766,
 -39.636, -39.427, -39.140, -38.859, -38.567, -38.180, -37.929, -37.801, -37.641, -37.568,
 -37.473, -37.192, -36.871, -36.570, -36.429, -36.004, -35.579, -35.235, -34.779, -34.525,
 -33.996, -33.457, -33.095, -32.615, -32.124, -31.598, -31.004, -30.296, -29.815, -29.682,
 -29.348, -28.746, -28.212, -27.825, -27.531, -27.205, -26.804, -26.315, -25.992, -25.707,
 -25.290, -24.856, -24.223, -23.745, -23.288, -22.844, -22.473, -21.926, -21.494, -20.972,
 -20.179, -19.503, -18.899, -18.019, -17.368, -16.698, -15.768, -15.217, -14.421, -13.363,
 -12.764, -12.204, -11.669, -11.163, -10.442,  -9.569,  -8.307,  -7.082,  -6.087,  -5.010,
  -3.950,  -2.589,  -1.016,   0.496,   1.891,   2.874,   3.942,   5.023,   6.227,   7.623,
   8.831,  10.310,  11.878,  13.421,  14.442,  14.870,  15.460,  17.008,  19.377,  20.743,
  21.634,  22.850,  23.632,  23.954,  23.423,  22.319,  21.563,  20.855,  20.003,  19.545,
  19.093,  18.806,  18.815,  18.622,  18.772,  18.458,  18.242,  18.364,  17.855,  18.142,
  18.410,  17.722,  17.259,  17.305,  17.328,  17.055,  16.807,  16.888,  16.772,  16.679,
  16.682,  16.209,  15.858,  15.350,  14.768,  14.297,  13.914,  13.609,  13.122,  12.993,
  12.734,  11.946,  11.545,  11.600,  11.241,  10.966,  11.079,  10.820,  10.614,  11.071,
  11.519,  11.621,  11.575,  11.549,  11.605,  11.655,  11.963,  12.288,  12.886,  13.503},
{
  13.701,  14.094,  14.419,  14.977,  15.661,  16.579,  17.593,  18.531,  19.514,  19.977,
  20.579,  21.822,  22.895,  23.315,  23.536,  23.826,  23.701,  23.742,  23.888,  24.240,
  24.062,  23.652,  24.066,  24.428,  24.233,  24.222,  24.507,  24.795,  25.386,  25.774,
  26.137,  26.458,  26.472,  26.622,  27.260,  28.067,  28.868,  29.565,  30.213,  31.001,
  31.559,  31.962,  32.060,  32.034,  31.913,  31.668,  31.461,  31.430,  31.343,  30.957,
  30.669,  30.555,  30.370,  30.393,  30.511,  30.923,  31.096,  31.125,  30.996,  30.834,
  30.727,  30.338,  30.160,  29.721,  29.187,  28.472,  28.068,  27.951,  27.743,  27.460,
  27.109,  26.401,  25.786,  25.176,  24.670,  24.293,  23.377,  22.498,  21.813,  21.539,
  21.448,  21.185,  20.148,  18.693,  16.917,  14.710,  13.020,  11.717,  10.378,   8.882,
   7.440,   5.877,   3.944,   2.045,   0.628,  -0.308,  -1.327,  -2.145,  -3.175,  -4.741,
  -6.210,  -7.134,  -7.925,  -8.523,  -9.080,  -9.973, -11.464, -13.211, -13.964, -14.834,
 -16.162, -17.549, -19.075, -20.665, -21.983, -23.020, -24.095, -25.173, -26.152, -27.256,
 -28.353, -29.216, -30.001, -30.917, -31.796, -32.299, -32.719, -33.574, -34.376, -34.718,
 -35.246, -35.791, -36.190, -36.494, -36.906, -37.273, -37.540, -37.849, -38.408, -39.004,
 -38.809, -37.914, -37.319, -37.050, -36.830, -36.820, -36.663, -36.443, -36.054, -35.996,
 -35.983, -35.359, -34.537, -34.150, -34.779, -35.076, -34.775, -34.983, -35.434, -35.865,
 -36.601, -37.277, -37.988, -38.736, -39.373, -40.158, -41.097, -41.924, -43.040, -45.112,
 -46.847, -47.780, -48.072, -48.047, -48.426, -48.723, -48.886, -49.353, -49.325, -48.933, -48.909,
 -48.843, -48.644, -48.280, -47.931, -47.600, -47.114, -46.571, -46.086, -45.540, -45.121,
 -44.709, -44.377, -43.936, -43.321, -42.838, -42.416, -42.178, -41.628, -41.344, -40.967,
 -40.330, -39.903, -39.570, -39.082, -38.606, -38.206, -37.902, -37.701, -37.402, -37.091,
 -36.827, -36.665, -36.484, -36.409, -36.164, -35.981, -35.811, -35.747, -35.565, -35.335,
 -35.103, -34.625, -34.336, -34.143, -33.883, -33.425, -32.924, -32.578, -32.341, -31.941,
 -31.550, -31.193, -30.846, -30.568, -30.163, -29.821, -29.520, -29.175, -28.774, -27.976,
 -27.262, -26.990, -26.676, -26.260, -25.923, -25.725, -25.560, -25.410, -25.168, -24.872,
 -24.429, -23.833, -23.307, -22.794, -22.249, -21.794, -21.273, -20.886, -20.532, -19.988,
 -19.486, -18.763, -17.817, -17.048, -16.420, -15.666, -15.043, -14.504, -13.809, -12.901,
 -12.266, -11.925, -11.215, -10.409,  -9.774,  -8.837,  -7.831,  -6.890,  -5.735,  -4.561,
  -3.313,  -2.202,  -0.680,   0.859,   2.269,   3.450,   4.389,   5.711,   7.113,   8.496,
   9.900,  11.516,  12.857,  14.171,  15.335,  15.141,  14.392,  14.181,  15.067,  17.722,
  20.660,  21.906,  22.234,  22.331,  22.600,  22.981,  21.930,  20.788,  20.184,  20.167,
  20.293,  19.547,  20.854,  22.414,  22.803,  22.500,  21.795,  20.763,  18.997,  18.648,
  19.246,  19.987,  20.194,  20.430,  19.828,  19.353,  19.627,  19.365,  19.682,  19.385,
  19.037,  18.791,  18.518,  18.289,  17.934,  17.673,  16.977,  16.469,  16.363,  15.820,
  14.840,  14.316,  13.971,  13.854,  13.521,  13.052,  12.982,  12.894,  13.201,  13.269,
  12.844,  13.005,  13.139,  12.816,  12.781,  12.731,  12.873,  13.087,  13.213,  13.701},
{
  14.329,  14.546,  14.737,  15.045,  15.448,  16.177,  17.118,  17.993,  18.884,  19.755,
  20.770,  21.876,  22.852,  23.646,  23.848,  24.022,  24.182,  24.410,  24.697,  24.911,
  25.170,  25.288,  24.973,  25.177,  26.056,  26.095,  26.201,  27.067,  27.601,  27.776,
  27.775,  28.060,  28.386,  28.761,  29.226,  29.875,  30.625,  31.072,  31.626,  32.284,
  32.552,  32.968,  33.152,  33.133,  33.012,  32.970,  32.401,  32.247,  32.194,  31.859,
  31.628,  31.477,  31.334,  31.368,  31.491,  31.472,  31.429,  31.464,  31.641,  31.897,
  31.864,  31.390,  31.015,  30.522,  30.259,  30.071,  29.988,  29.668,  29.222,  28.546,
  28.306,  27.791,  26.882,  26.512,  25.963,  25.627,  25.453,  24.762,  24.068,  23.817,
  23.402,  22.913,  22.383,  21.089,  19.031,  16.793,  14.461,  12.448,  10.707,   9.378,
   7.972,   6.314,   4.709,   3.084,   2.003,   1.363,   0.302,  -0.739,  -2.026,  -3.744,
  -5.179,  -6.017,  -6.881,  -7.660,  -8.004,  -8.859, -10.060, -11.440, -12.622, -13.555,
 -14.709, -16.003, -17.300, -18.702, -19.942, -21.173, -22.240, -23.223, -24.243, -25.292,
 -26.241, -27.319, -28.125, -28.883, -29.628, -30.188, -30.720, -31.306, -31.985, -32.623,
 -33.287, -33.807, -34.218, -34.608, -34.985, -35.001, -35.107, -35.123, -35.039, -34.983,
 -34.974, -34.957, -34.627, -34.451, -34.235, -33.808, -33.808, -33.247, -33.000, -33.130,
 -32.693, -31.834, -31.317, -31.512, -32.216, -32.830, -32.778, -32.927, -33.448, -34.097,
 -35.019, -35.877, -36.849, -38.053, -39.072, -39.703, -41.055, -43.409, -44.758, -45.016,
 -45.516, -46.388, -46.887, -47.460, -47.917, -47.832, -47.880, -47.977, -48.243, -48.032, -47.660,
 -47.367, -47.156, -47.018, -46.511, -46.128, -45.755, -45.399, -45.181, -44.776, -44.014,
 -43.412, -42.973, -42.785, -42.407, -42.001, -41.442, -40.745, -40.104, -39.621, -39.033,
 -38.518, -38.035, -37.379, -36.854, -36.410, -35.922, -35.470, -35.234, -34.996, -34.633,
 -34.349, -34.187, -34.085, -33.909, -33.715, -33.606, -33.343, -33.166, -32.829, -32.594,
 -32.345, -32.243, -32.100, -31.763, -31.522, -30.979, -30.345, -30.030, -29.628, -29.002,
 -28.564, -28.074, -27.997, -28.263, -28.686, -28.341, -27.370, -26.556, -25.931, -25.783,
 -25.448, -25.156, -24.960, -24.841, -24.632, -24.437, -24.315, -24.204, -24.026, -23.637,
 -22.984, -22.303, -21.738, -21.448, -21.331, -20.803, -20.129, -19.559, -18.827, -18.374,
 -18.013, -17.349, -16.608, -15.910, -15.304, -14.632, -14.106, -13.582, -12.847, -12.174,
 -11.516, -10.812, -10.012,  -9.199,  -8.412,  -7.656,  -6.774,  -5.996,  -5.075,  -4.132,
  -3.067,  -1.868,  -0.483,   0.662,   1.864,   3.349,   5.028,   6.615,   8.103,   9.364,
  10.608,  12.080,  13.504,  15.426,  16.941,  17.274,  17.931,  18.376,  17.894,  16.616,
  15.639,  15.404,  16.447,  20.504,  23.171,  23.618,  23.508,  22.296,  21.343,  20.790,
  20.826,  20.854,  23.434,  25.395,  25.916,  25.721,  24.433,  21.074,  18.180,  19.510,
  19.971,  21.164,  21.556,  20.650,  21.797,  22.179,  22.280,  22.479,  22.257,  22.198,
  21.658,  20.614,  18.765,  16.653,  16.640,  16.811,  16.491,  17.213,  17.788,  18.026,
  17.930,  17.459,  16.797,  16.179,  15.896,  15.920,  15.861,  15.459,  15.189,  14.926,
  14.696,  14.648,  14.464,  13.989,  13.863,  14.062,  14.138,  14.067,  14.150,  14.329},
{
  15.132,  15.512,  15.552,  15.547,  15.894,  16.516,  16.896,  17.492,  18.241,  19.078,
  19.934,  20.926,  21.985,  22.824,  23.249,  23.640,  24.340,  24.901,  25.255,  25.338,
  25.447,  25.608,  26.099,  26.289,  26.273,  27.169,  27.746,  28.268,  28.627,  28.984,
  28.907,  29.009,  29.427,  30.237,  30.959,  31.634,  32.027,  32.438,  32.661,  33.323,
  33.728,  34.191,  34.478,  34.522,  34.443,  34.436,  34.398,  33.949,  33.481,  33.084,
  32.927,  32.761,  32.739,  32.864,  32.768,  32.583,  32.339,  32.392,  32.534,  32.681,
  32.551,  32.176,  31.533,  31.005,  30.769,  30.787,  30.894,  31.114,  30.761,  30.025,
  29.355,  28.686,  27.693,  27.264,  27.439,  27.431,  27.545,  27.268,  26.443,  25.559,
  24.978,  24.978,  23.811,  22.486,  21.122,  17.385,  14.179,  12.410,  11.095,   9.921,
   8.876,   7.954,   6.610,   5.102,   4.093,   2.952,   1.547,   0.333,  -0.956,  -2.225,
  -3.480,  -4.385,  -5.479,  -6.735,  -7.371,  -7.926,  -8.876, -10.063, -11.158, -12.338,
 -13.579, -14.696, -15.952, -17.094, -18.363, -19.630, -20.702, -21.793, -22.709, -23.695,
 -24.625, -25.546, -26.216, -26.882, -27.603, -28.328, -28.809, -29.504, -30.338, -30.975,
 -31.401, -31.932, -32.350, -32.671, -32.914, -32.875, -32.947, -33.008, -32.922, -32.848,
 -32.984, -32.666, -32.206, -31.758, -31.370, -31.017, -30.913, -30.497, -30.493, -30.233,
 -29.654, -29.110, -29.148, -29.671, -30.815, -30.599, -30.728, -31.111, -31.242, -33.048,
 -34.958, -35.873, -36.922, -37.988, -38.790, -40.129, -41.208, -41.718, -42.698, -43.621,
 -44.474, -45.138, -45.819, -46.282, -46.462, -46.830, -46.997, -46.933, -46.691, -46.458, -46.305,
 -46.116, -45.771, -45.385, -44.893, -44.577, -44.411, -44.276, -43.639, -43.148, -42.861,
 -42.550, -42.086, -41.450, -40.845, -40.405, -39.789, -39.186, -38.632, -37.890, -37.334,
 -36.717, -35.998, -35.332, -34.814, -34.239, -33.736, -33.199, -32.771, -32.581, -32.403,
 -32.191, -31.958, -31.642, -31.417, -31.240, -31.157, -30.962, -30.575, -30.368, -30.231,
 -30.151, -29.939, -29.535, -29.257, -29.020, -28.614, -28.095, -27.563, -27.109, -26.713,
 -26.674, -26.726, -26.390, -25.739, -24.888, -24.606, -24.380, -24.183, -24.017, -23.647,
 -23.295, -23.052, -22.945, -23.016, -23.018, -23.196, -23.007, -22.768, -22.655, -22.560,
 -22.274, -21.497, -20.663, -20.136, -19.893, -19.535, -18.908, -18.373, -17.612, -16.932,
 -16.344, -15.896, -15.367, -14.414, -13.925, -13.560, -12.849, -12.240, -11.530, -11.120,
 -10.240,  -9.455,  -8.759,  -7.968,  -7.206,  -6.575,  -5.752,  -5.001,  -4.256,  -3.512,
  -2.470,  -1.251,   0.031,   1.308,   2.254,   4.024,   5.644,   6.633,   8.065,   9.433,
  11.653,  13.362,  14.765,  15.823,  16.620,  17.402,  17.766,  18.400,  19.026,  20.199,
  21.014,  20.660,  20.426,  20.183,  20.408,  20.622,  21.269,  21.305,  20.022,  19.614,
  19.266,  18.024,  18.817,  17.324,  16.085,  15.910,  17.711,  20.167,  21.260,  21.863,
  21.183,  20.366,  20.622,  21.848,  22.674,  22.587,  23.194,  23.736,  23.707,  23.544,
  23.094,  21.765,  17.788,  12.902,  10.302,  13.338,  18.484,  19.620,  20.225,  20.839,
  21.094,  21.035,  20.619,  19.687,  19.164,  18.417,  18.100,  17.609,  17.173,  16.702,
  16.458,  16.228,  15.714,  15.532,  15.467,  15.206,  15.454,  15.566,  15.272,  15.132},
{
  17.828,  17.859,  17.503,  17.438,  17.342,  17.416,  17.665,  17.772,  18.232,  18.582,
  19.012,  19.731,  20.671,  21.535,  22.330,  22.987,  23.620,  24.066,  24.696,  25.250,
  25.722,  26.138,  26.436,  27.034,  27.719,  27.659,  28.079,  29.008,  29.495,  29.814,
  30.247,  30.428,  30.954,  31.700,  32.493,  32.898,  33.201,  33.595,  34.171,  34.760,
  35.362,  35.807,  35.892,  35.937,  35.954,  36.075,  36.043,  35.798,  35.286,  34.740,
  34.177,  34.083,  34.176,  34.127,  34.150,  34.013,  33.620,  33.614,  33.548,  33.604,
  33.334,  33.106,  32.606,  32.145,  31.394,  31.622,  31.873,  32.050,  31.904,  31.076,
  29.980,  29.534,  29.010,  29.285,  30.037,  30.188,  29.919,  29.398,  28.506,  27.395,
  26.064,  24.948,  24.267,  22.552,  19.408,  16.276,  14.378,  13.239,  12.455,  11.582,
  10.309,   9.244,   8.259,   6.990,   5.634,   4.629,   3.382,   1.812,   0.370,  -0.864,
  -1.910,  -2.994,  -4.286,  -5.574,  -6.813,  -7.597,  -8.383,  -9.388, -10.659, -11.911,
 -13.001, -14.139, -15.323, -16.486, -17.560, -18.745, -19.747, -20.780, -21.737, -22.634,
 -23.594, -24.347, -25.194, -25.905, -26.534, -27.306, -27.911, -28.581, -29.132, -29.544,
 -29.957, -30.271, -30.465, -30.466, -30.612, -30.688, -30.641, -30.772, -30.744, -30.484,
 -30.368, -30.203, -29.625, -29.236, -28.920, -28.716, -28.366, -28.026, -27.966, -27.383,
 -27.227, -27.160, -27.293, -28.442, -28.904, -28.857, -29.161, -30.001, -33.940, -33.619,
 -33.336, -34.481, -35.678, -36.857, -37.882, -38.905, -39.762, -40.614, -41.286, -42.339,
 -43.063, -43.854, -44.457, -44.617, -44.737, -45.010, -45.261, -45.211, -45.019, -44.723, -44.608,
 -44.457, -44.474, -44.231, -43.514, -43.128, -42.949, -42.662, -42.394, -41.904, -41.461,
 -41.085, -40.584, -39.778, -39.322, -38.815, -38.258, -37.487, -36.740, -35.961, -35.186,
 -34.569, -34.047, -33.379, -32.863, -32.341, -31.799, -31.241, -30.687, -30.256, -29.979,
 -29.789, -29.646, -29.410, -29.300, -29.007, -28.777, -28.552, -28.384, -28.301, -28.100,
 -27.766, -27.357, -27.137, -26.773, -26.503, -26.594, -26.582, -26.019, -25.265, -24.600,
 -24.021, -23.607, -23.225, -22.894, -22.566, -22.632, -22.476, -22.193, -21.797, -21.401,
 -21.060, -21.162, -21.099, -21.029, -21.568, -22.044, -22.294, -22.129, -21.402, -20.448,
 -19.578, -19.382, -19.552, -18.981, -18.127, -17.233, -16.624, -16.226, -15.880, -15.633,
 -15.106, -14.434, -13.961, -13.180, -12.322, -12.346, -11.761, -11.143, -10.756,  -9.937,
  -9.268,  -8.824,  -8.100,  -6.921,  -6.004,  -5.550,  -4.827,  -4.019,  -3.065,  -2.162,
  -1.537,  -0.636,   0.608,   1.764,   2.981,   4.352,   5.838,   7.355,   9.260,  11.237,
  12.480,  13.451,  14.927,  16.024,  16.791,  17.151,  17.798,  18.596,  19.604,  19.239,
  19.090,  19.967,  20.668,  20.654,  20.447,  20.384,  20.477,  20.201,  20.294,  19.895,
  20.206,  20.431,  20.713,  20.967,  20.824,  19.986,  19.643,  19.319,  19.572,  20.077,
  20.511,  20.392,  20.806,  21.428,  21.953,  22.000,  22.759,  23.226,  23.766,  24.102,
  23.790,  23.171,  22.404,  19.996,  14.587,   9.420,  14.749,  18.391,  19.713,  20.577,
  20.934,  20.876,  21.264,  20.861,  21.320,  20.643,  20.396,  19.951,  19.922,  19.664,
  19.287,  18.780,  18.258,  18.040,  18.000,  17.998,  18.007,  17.915,  17.879,  17.828},
{
  20.762,  20.624,  20.371,  19.743,  19.200,  19.038,  18.825,  18.992,  19.498,  19.134,
  19.207,  19.492,  19.988,  20.554,  21.632,  22.811,  23.253,  24.065,  25.277,  25.704,
  26.287,  26.749,  27.634,  27.200,  28.079,  28.309,  28.904,  28.988,  30.188,  30.773,
  31.697,  32.286,  32.603,  33.080,  33.718,  34.231,  34.575,  35.346,  36.202,  36.763,
  37.133,  37.461,  37.398,  37.361,  37.487,  37.586,  37.601,  37.487,  37.159,  36.588,
  36.278,  35.700,  35.270,  35.174,  35.086,  35.046,  34.951,  34.549,  34.562,  34.403,
  34.060,  33.955,  33.731,  33.514,  33.034,  31.695,  31.460,  31.544,  30.639,  30.885,
  30.377,  29.902,  29.777,  29.941,  29.986,  30.054,  30.303,  29.373,  28.211,  27.614,
  26.029,  24.892,  23.420,  20.692,  17.770,  16.031,  15.013,  14.895,  13.514,  12.509,
  11.581,  10.469,   9.337,   8.368,   7.036,   5.704,   4.391,   2.876,   1.028,  -0.174,
  -1.264,  -2.308,  -3.547,  -4.834,  -5.882,  -6.943,  -8.068,  -9.185, -10.166, -11.487,
 -12.690, -13.813, -14.969, -16.171, -17.274, -18.348, -19.196, -20.266, -21.129, -22.017,
 -22.781, -23.716, -24.364, -25.005, -25.694, -26.316, -27.043, -27.443, -27.764, -28.221,
 -28.471, -28.599, -28.678, -28.726, -28.746, -28.761, -28.764, -28.648, -28.576, -28.281,
 -28.433, -27.897, -27.219, -26.908, -26.633, -26.385, -25.951, -25.964, -25.598, -25.508,
 -25.791, -25.536, -26.181, -26.879, -26.912, -27.569, -27.590, -32.141, -31.571, -28.302,
 -30.922, -31.817, -33.907, -34.959, -35.976, -36.693, -37.955, -38.641, -39.445, -40.557,
 -41.230, -41.657, -42.088, -42.301, -42.196, -42.375, -42.436, -42.610, -42.670, -42.709, -42.718,
 -42.736, -42.520, -42.298, -42.018, -41.746, -41.471, -41.198, -40.867, -40.459, -39.773,
 -39.040, -38.419, -37.869, -37.409, -36.816, -36.057, -35.212, -34.500, -33.880, -33.315,
 -32.616, -32.103, -31.476, -30.902, -30.257, -29.728, -29.235, -28.789, -28.224, -27.897,
 -27.679, -27.409, -27.315, -27.244, -27.082, -26.755, -26.584, -26.393, -26.199, -26.139,
 -26.112, -25.835, -25.485, -25.039, -24.534, -23.679, -23.070, -22.869, -22.674, -22.321,
 -21.912, -21.435, -20.893, -20.644, -20.706, -20.596, -20.305, -19.916, -19.581, -19.587,
 -19.566, -19.465, -19.515, -19.995, -19.737, -19.042, -19.132, -19.340, -19.256, -18.716,
 -18.463, -17.994, -17.368, -17.010, -16.893, -16.115, -15.508, -14.846, -14.536, -14.134,
 -13.686, -13.289, -12.796, -12.077, -10.982, -10.754, -10.756,  -9.951,  -9.085,  -8.489,
  -8.195,  -7.554,  -6.797,  -6.209,  -5.642,  -4.732,  -3.553,  -2.513,  -1.990,  -1.246,
  -0.170,   0.534,   1.465,   2.665,   3.712,   4.949,   6.756,   8.417,   9.862,  11.070,
  12.509,  13.369,  14.603,  15.093,  15.958,  16.203,  17.050,  18.423,  19.051,  19.550,
  19.860,  20.264,  20.353,  20.012,  19.787,  19.802,  19.696,  19.292,  19.071,  19.200,
  19.210,  19.629,  20.074,  20.212,  19.862,  19.763,  19.555,  19.329,  19.164,  19.511,
  19.907,  19.879,  20.235,  21.005,  21.392,  21.628,  22.356,  22.828,  23.638,  23.655,
  24.028,  22.820,  22.382,  19.528,  15.214,   9.604,  14.545,  18.277,  19.530,  19.806,
  19.790,  19.482,  19.338,  19.435,  19.371,  18.639,  19.002,  19.083,  19.325,  19.318,
  20.265,  21.086,  21.129,  21.301,  21.236,  20.741,  21.113,  21.003,  20.764,  20.762},
{
  22.627,  22.106,  21.913,  22.194,  21.642,  20.935,  20.440,  19.870,  19.446,  19.592,
  20.006,  20.188,  20.470,  21.343,  22.066,  23.395,  24.642,  25.418,  26.437,  27.054,
  27.769,  28.134,  28.887,  29.060,  29.140,  29.469,  29.556,  29.928,  30.813,  31.589,
  32.448,  33.468,  34.001,  34.192,  34.913,  35.658,  36.465,  37.262,  38.196,  38.689,
  38.907,  39.066,  38.993,  39.153,  39.140,  39.266,  39.257,  39.032,  38.765,  38.304,
  37.762,  37.371,  36.919,  36.212,  36.263,  36.318,  36.182,  35.661,  35.581,  35.464,
  35.101,  34.638,  34.518,  34.612,  34.565,  35.147,  35.205,  34.708,  34.497,  34.665,
  33.934,  34.281,  33.774,  32.355,  31.428,  30.815,  30.445,  29.156,  28.538,  27.620,
  26.045,  23.116,  20.425,  18.639,  18.060,  17.056,  16.179,  15.799,  14.524,  13.477,
  12.464,  11.290,  10.079,   8.983,   7.750,   6.255,   4.946,   3.625,   2.228,   0.684,
  -0.741,  -1.627,  -2.843,  -4.206,  -5.482,  -6.593,  -7.622,  -8.875,  -9.901, -11.165,
 -12.432, -13.647, -14.915, -16.057, -17.051, -18.175, -19.042, -20.008, -20.851, -21.646,
 -22.382, -23.096, -23.764, -24.351, -25.060, -25.604, -26.151, -26.488, -26.778, -26.937,
 -27.177, -27.250, -27.275, -27.181, -27.091, -27.111, -27.079, -27.121, -26.820, -26.664,
 -26.656, -25.845, -25.125, -24.689, -24.616, -24.101, -24.076, -23.827, -23.751, -24.483,
 -24.298, -24.310, -25.052, -25.342, -25.423, -25.995, -26.179, -30.436, -29.604, -27.000,
 -29.576, -29.475, -31.133, -32.781, -33.981, -34.898, -35.595, -36.450, -37.195, -37.807,
 -38.270, -39.054, -39.404, -39.684, -39.791, -39.855, -39.746, -39.780, -39.944, -40.207, -40.197,
 -40.114, -39.970, -39.627, -39.447, -39.442, -39.271, -39.021, -38.504, -37.896, -37.264,
 -36.683, -36.388, -36.017, -35.312, -34.662, -33.900, -33.257, -32.598, -31.969, -31.425,
 -30.706, -30.005, -29.606, -29.000, -28.333, -27.710, -27.288, -27.000, -26.612, -26.280,
 -25.896, -25.545, -25.211, -25.046, -25.066, -24.857, -24.897, -24.843, -24.368, -23.391,
 -22.671, -22.180, -21.945, -21.845, -21.761, -21.551, -21.318, -20.991, -20.689, -20.351,
 -20.185, -19.969, -19.638, -19.158, -18.630, -18.426, -18.349, -18.603, -18.802, -18.506,
 -18.241, -17.823, -17.069, -16.855, -17.255, -17.408, -16.876, -16.694, -16.402, -16.130,
 -16.044, -15.814, -15.593, -15.574, -15.262, -14.606, -13.816, -13.376, -12.984, -12.594,
 -12.266, -12.041, -11.625, -11.299, -11.155, -10.750, -10.006,  -9.275,  -9.195,  -8.616,
  -7.799,  -7.059,  -6.585,  -5.995,  -5.045,  -3.670,  -2.732,  -2.061,  -1.049,  -0.345,
   0.613,   1.595,   2.652,   3.578,   4.559,   5.868,   6.453,   7.463,   8.198,   8.081,
   9.020,  12.182,  13.808,  13.245,  14.292,  15.257,  16.186,  16.737,  17.810,  18.850,
  19.687,  19.632,  19.843,  19.899,  18.797,  18.224,  19.168,  18.956,  18.077,  18.040,
  18.121,  18.685,  18.905,  18.446,  18.565,  18.436,  18.030,  18.021,  18.514,  19.041,
  19.617,  19.620,  20.210,  20.509,  20.887,  21.385,  21.961,  22.608,  23.095,  23.105,
  22.981,  22.331,  20.494,  16.142,   8.242,   8.660,  16.166,  18.846,  19.641,  19.821,
  20.026,  19.884,  20.348,  20.922,  21.067,  21.151,  21.326,  21.599,  21.888,  22.584,
  23.093,  24.108,  24.192,  24.363,  24.577,  24.186,  23.484,  23.067,  22.963,  22.627},
{
  24.433,  24.013,  23.853,  24.179,  24.267,  23.615,  22.505,  21.651,  21.528,  21.792,
  22.100,  22.091,  22.434,  23.019,  24.115,  24.992,  26.214,  27.152,  27.946,  29.127,
  29.436,  30.119,  30.637,  31.176,  31.160,  30.751,  31.618,  31.466,  31.976,  32.798,
  33.640,  34.548,  35.470,  36.120,  36.743,  37.457,  38.248,  38.978,  39.765,  40.631,
  41.196,  41.412,  41.519,  41.301,  41.270,  41.370,  41.414,  41.409,  40.991,  40.360,
  39.694,  39.230,  38.706,  38.517,  38.174,  37.753,  37.454,  37.236,  36.855,  36.301,
  35.666,  35.498,  35.142,  34.889,  35.009,  35.900,  35.759,  34.776,  34.252,  33.758,
  33.472,  33.428,  33.431,  32.876,  32.264,  31.196,  30.547,  29.894,  28.632,  26.622,
  24.397,  22.174,  19.993,  19.405,  18.566,  17.623,  17.104,  15.951,  14.987,  13.930,
  12.958,  11.682,  10.443,   9.209,   7.806,   6.542,   5.301,   4.221,   2.780,   1.278,
   0.001,  -1.120,  -2.244,  -3.803,  -5.015,  -6.179,  -7.407,  -8.757,  -9.735, -10.795,
 -12.153, -13.487, -14.736, -15.923, -17.022, -18.079, -18.892, -19.680, -20.580, -21.235,
 -21.991, -22.731, -23.368, -23.964, -24.537, -25.100, -25.504, -25.728, -25.893, -25.881,
 -25.974, -25.847, -25.723, -25.619, -25.631, -25.620, -25.619, -25.655, -25.289, -25.070,
 -24.875, -23.860, -23.175, -22.848, -22.773, -22.204, -22.261, -22.021, -22.907, -23.318,
 -23.024, -23.131, -23.811, -23.925, -24.065, -24.467, -24.533, -25.220, -26.184, -26.693,
 -27.558, -28.402, -29.485, -30.848, -31.979, -31.598, -30.845, -31.937, -33.050, -34.061,
 -35.171, -36.036, -36.700, -36.811, -37.119, -36.836, -36.684, -36.621, -37.004, -37.037, -37.167,
 -37.378, -37.520, -37.263, -37.139, -36.997, -36.774, -36.163, -35.642, -35.136, -34.807,
 -34.449, -34.031, -33.201, -32.635, -32.251, -32.049, -31.310, -30.446, -29.983, -29.590,
 -29.005, -28.179, -27.381, -26.874, -26.528, -26.270, -25.766, -25.230, -24.769, -24.305,
 -23.983, -23.881, -23.558, -23.194, -22.868, -23.064, -22.719, -21.975, -21.440, -20.930,
 -20.393, -20.015, -19.866, -19.855, -19.865, -19.767, -19.657, -19.639, -19.392, -19.059,
 -18.640, -18.459, -18.513, -18.093, -17.298, -16.915, -16.452, -15.288, -14.514, -14.323,
 -14.142, -14.121, -14.148, -14.219, -14.269, -14.408, -14.512, -14.464, -14.379, -14.479,
 -14.515, -14.448, -14.356, -14.374, -14.154, -13.657, -12.988, -12.343, -12.097, -11.949,
 -11.473, -11.271, -10.557, -10.579, -11.067, -10.940,  -9.928,  -9.146,  -8.793,  -7.989,
  -7.290,  -6.705,  -6.082,  -4.976,  -3.808,  -2.976,  -2.448,  -1.683,  -0.591,   0.736,
   1.765,   2.935,   3.733,   4.620,   5.112,   4.961,   4.506,   4.303,   6.096,  11.438,
  14.435,  14.777,  15.627,  14.310,  13.186,  13.789,  14.665,  15.861,  16.770,  17.776,
  18.332,  18.874,  18.436,  17.482,  18.335,  18.118,  17.126,  17.226,  17.314,  17.916,
  17.901,  17.670,  17.502,  17.052,  17.132,  17.292,  17.252,  17.373,  17.783,  18.321,
  18.527,  18.887,  19.219,  19.407,  20.179,  20.555,  21.105,  21.341,  21.464,  21.300,
  20.646,  21.148,  16.545,   4.718,   5.398,  14.807,  18.577,  19.759,  20.061,  20.526,
  20.186,  20.199,  20.377,  21.080,  21.330,  21.741,  21.830,  22.018,  22.549,  23.014,
  23.518,  24.314,  24.902,  25.535,  25.184,  25.490,  24.835,  24.515,  24.603,  24.433},
{
  27.098,  25.824,  25.577,  25.803,  26.250,  26.014,  25.907,  25.845,  25.300,  24.516,
  24.237,  24.209,  24.446,  25.351,  25.906,  26.929,  27.794,  28.563,  29.783,  30.680,
  31.591,  32.039,  32.720,  33.130,  33.292,  33.734,  32.843,  32.873,  33.563,  34.030,
  34.610,  35.412,  36.379,  37.358,  38.185,  39.003,  39.957,  41.127,  42.603,  43.515,
  44.486,  44.614,  44.612,  44.481,  44.141,  43.920,  44.018,  43.819,  43.334,  41.886,
  40.810,  40.781,  40.883,  40.609,  39.741,  39.232,  38.904,  38.283,  37.945,  37.259,
  36.717,  36.202,  35.702,  35.057,  34.817,  34.803,  34.504,  34.459,  34.616,  34.590,
  34.721,  34.716,  34.097,  34.087,  34.068,  33.286,  32.694,  31.850,  29.646,  27.592,
  24.737,  22.616,  20.969,  20.362,  18.674,  18.855,  17.618,  16.575,  15.309,  14.196,
  12.909,  11.624,  10.304,   8.969,   7.863,   6.669,   5.447,   4.223,   2.458,   1.590,
   0.291,  -0.590,  -2.214,  -3.449,  -4.963,  -6.062,  -7.134,  -8.509,  -9.804, -10.928,
 -12.082, -13.398, -14.715, -15.827, -17.024, -17.947, -18.941, -19.606, -20.399, -21.070,
 -21.631, -22.355, -22.915, -23.560, -24.129, -24.682, -24.993, -24.896, -24.854, -24.896,
 -24.837, -24.703, -24.649, -24.487, -24.516, -24.476, -24.426, -24.289, -23.798, -23.480,
 -23.155, -22.242, -21.568, -21.268, -21.034, -20.677, -20.507, -21.185, -22.068, -21.769,
 -21.523, -21.965, -22.421, -22.469, -22.391, -23.101, -23.115, -23.176, -21.567, -23.238,
 -26.434, -26.157, -26.995, -27.200, -27.624, -26.885, -25.558, -25.419, -26.317, -28.052,
 -29.912, -32.064, -33.027, -33.508, -33.645, -33.437, -33.264, -33.492, -33.579, -34.005, -34.086,
 -34.563, -34.669, -34.661, -34.552, -34.353, -33.918, -33.524, -33.152, -32.770, -32.329,
 -31.844, -31.389, -31.032, -30.871, -30.497, -29.966, -29.515, -28.901, -28.068, -27.315,
 -26.928, -26.411, -25.910, -25.380, -24.915, -24.288, -24.008, -23.800, -23.455, -23.175,
 -23.017, -22.967, -22.953, -22.520, -22.057, -21.436, -20.781, -20.391, -19.980, -19.510,
 -19.058, -18.584, -18.468, -18.298, -18.375, -18.371, -18.276, -18.334, -18.232, -17.815,
 -17.311, -16.450, -15.855, -15.230, -14.911, -14.351, -13.585, -13.140, -12.732, -12.628,
 -12.564, -12.716, -12.937, -13.059, -13.034, -13.065, -13.342, -13.448, -13.393, -13.480,
 -13.447, -13.425, -13.343, -13.314, -13.051, -12.625, -12.180, -11.815, -11.456, -11.119,
  -9.786,  -9.967,  -9.708,  -9.824, -10.479, -10.390,  -9.660,  -8.231,  -7.210,  -6.195,
  -5.497,  -5.090,  -4.878,  -4.296,  -3.313,  -2.666,  -1.516,  -0.938,   0.697,   1.526,
   3.079,   3.893,   4.819,   4.502,   3.359,   3.454,   6.869,   9.641,  10.882,  12.736,
  13.628,  13.788,  14.157,  13.538,  12.251,  11.280,  11.515,  13.041,  15.209,  16.815,
  16.897,  17.958,  18.344,  16.638,  14.916,  13.595,  14.163,  14.594,  14.671,  14.834,
  15.591,  15.851,  15.646,  15.592,  15.725,  15.631,  15.514,  15.882,  16.536,  16.306,
  16.529,  18.919,  21.351,  22.755,  21.478,  16.571,  10.845,  11.667,  10.284,   8.469,
   6.682,   4.085,   3.092,   8.224,  14.917,  18.347,  20.445,  21.203,  20.263,  20.486,
  20.818,  21.078,  21.492,  21.975,  22.332,  22.453,  22.841,  23.191,  23.915,  24.449,
  25.086,  25.263,  25.685,  26.142,  26.506,  26.650,  27.057,  27.469,  27.003,  27.098},
{
  27.238,  27.236,  26.714,  27.246,  27.707,  27.547,  27.382,  26.556,  26.074,  25.942,
  26.653,  27.029,  27.722,  27.599,  27.974,  28.757,  29.464,  30.759,  31.495,  32.334,
  33.330,  34.373,  35.168,  35.879,  36.201,  35.811,  36.510,  34.366,  35.466,  36.004,
  35.569,  36.125,  36.937,  38.088,  39.207,  40.169,  41.298,  42.585,  43.950,  45.382,
  46.516,  47.414,  47.087,  46.543,  45.625,  45.258,  45.061,  44.282,  44.809,  43.322,
  42.965,  44.195,  44.714,  42.582,  41.400,  40.537,  40.320,  39.357,  38.272,  37.901,
  37.710,  37.090,  36.194,  35.660,  34.713,  34.485,  34.692,  35.050,  35.102,  34.790,
  34.691,  34.780,  35.283,  35.008,  34.994,  35.315,  34.395,  32.575,  30.598,  27.270,
  24.432,  23.306,  23.514,  22.085,  20.247,  19.619,  18.182,  16.908,  15.781,  14.476,
  13.224,  11.749,  10.534,   9.216,   7.996,   6.773,   5.602,   4.404,   2.835,   1.273,
   0.065,  -0.801,  -2.271,  -3.536,  -4.834,  -6.089,  -7.252,  -8.581,  -9.842, -11.138,
 -12.160, -13.554, -14.820, -16.007, -17.201, -18.156, -19.017, -19.652, -20.280, -21.061,
 -21.727, -22.109, -22.678, -23.241, -23.841, -24.270, -24.611, -24.388, -24.171, -24.114,
 -23.791, -23.654, -23.651, -23.436, -23.398, -23.313, -23.140, -22.864, -22.361, -22.155,
 -21.547, -20.836, -20.321, -19.897, -19.783, -19.556, -19.318, -20.738, -21.246, -20.739,
 -20.637, -21.495, -21.053, -20.976, -21.455, -21.539, -22.365, -21.904, -21.848, -19.480,
 -22.697, -22.138, -22.753, -23.191, -23.737, -22.069, -22.329, -22.110, -22.317, -22.970,
 -23.738, -24.437, -25.931, -27.648, -29.145, -29.746, -30.029, -30.609, -30.602, -30.816, -31.151,
 -31.530, -31.569, -31.423, -31.397, -31.372, -31.124, -30.781, -30.630, -30.308, -29.699,
 -29.421, -29.104, -28.856, -28.381, -28.039, -27.679, -27.170, -26.635, -26.072, -25.582,
 -25.156, -24.799, -24.295, -23.772, -23.404, -22.998, -23.085, -23.317, -23.013, -22.509,
 -21.786, -21.341, -20.880, -20.523, -20.142, -19.824, -19.500, -19.195, -18.755, -18.119,
 -18.082, -17.924, -17.706, -17.391, -17.085, -18.121, -18.817, -18.522, -18.168, -17.697,
 -17.124, -16.059, -15.175, -14.656, -14.191, -13.755, -13.303, -12.750, -12.280, -12.045,
 -11.785, -11.655, -11.802, -11.993, -12.061, -12.246, -12.296, -12.191, -12.376, -12.395,
 -12.425, -12.559, -12.570, -12.410, -11.961, -11.924, -11.843, -11.658, -11.057, -10.303,
  -9.775,  -9.048,  -8.842,  -8.917,  -9.190,  -9.049,  -8.320,  -7.261,  -6.411,  -5.734,
  -5.265,  -4.842,  -4.343,  -3.730,  -3.158,  -1.968,  -0.769,   0.389,   1.686,   2.567,
   3.678,   4.741,   4.807,   4.200,   3.676,   8.809,  11.568,  10.327,   9.264,   9.695,
  11.675,  12.190,  12.319,  11.918,  10.988,   9.979,   9.384,  10.268,  10.938,  10.228,
  10.109,  10.312,   9.014,   7.479,   6.894,   8.774,  11.788,  13.436,  11.769,  11.451,
  10.900,  12.050,  13.723,  14.008,  13.657,  13.487,  14.061,  15.932,  15.852,  15.305,
  17.216,  18.483,  18.282,  16.624,  11.834,  11.996,  13.206,  13.232,  13.444,  14.057,
  13.599,  13.107,  13.533,  14.756,  17.025,  19.165,  20.940,  20.578,  20.972,  21.032,
  21.081,  21.514,  21.848,  22.359,  22.652,  23.144,  23.421,  23.914,  24.567,  25.286,
  25.487,  25.628,  26.229,  26.901,  27.051,  27.370,  27.690,  27.705,  27.587,  27.238},
{
  27.168,  27.060,  27.074,  26.646,  26.574,  26.127,  26.365,  26.396,  27.017,  26.723,
  27.013,  28.154,  28.618,  29.875,  31.002,  31.543,  32.374,  33.035,  33.914,  34.538,
  35.298,  35.903,  36.476,  37.078,  37.750,  37.852,  37.953,  37.307,  35.558,  37.317,
  37.032,  37.019,  37.740,  38.722,  39.741,  40.902,  42.273,  43.518,  45.148,  46.330,
  47.267,  48.297,  47.881,  48.490,  50.892,  50.084,  46.548,  45.796,  47.002,  45.143,
  44.213,  44.857,  43.020,  43.049,  41.964,  41.454,  41.065,  40.715,  39.433,  37.831,
  37.306,  37.289,  36.311,  35.416,  34.940,  34.963,  34.744,  34.734,  34.779,  34.472,
  34.373,  34.880,  36.818,  39.559,  39.101,  37.087,  35.364,  34.059,  32.463,  29.521,
  28.346,  26.061,  23.963,  22.791,  21.838,  20.072,  18.849,  17.279,  15.919,  14.806,
  13.459,  12.115,  10.581,   9.272,   8.052,   6.694,   5.332,   3.993,   2.778,   1.438,
  -0.043,  -1.392,  -2.592,  -4.140,  -5.170,  -6.443,  -7.791,  -9.041, -10.264, -11.503,
 -12.675, -13.758, -15.147, -16.265, -17.452, -18.443, -19.197, -19.931, -20.570, -21.322,
 -21.429, -22.162, -22.586, -23.123, -23.655, -24.014, -24.256, -23.826, -23.549, -23.362,
 -23.005, -22.988, -22.785, -22.565, -22.405, -22.149, -21.910, -21.683, -21.034, -21.047,
 -20.170, -20.052, -19.365, -18.963, -18.868, -18.661, -19.190, -20.288, -19.732, -19.674,
 -19.599, -20.159, -19.879, -20.232, -19.819, -19.969, -20.197, -20.177, -20.032, -19.548,
 -17.440, -18.966, -18.953, -20.179, -20.792, -20.391, -18.562, -19.249, -18.978, -18.315,
 -18.442, -19.495, -20.248, -21.144, -22.938, -25.713, -26.507, -27.083, -27.520, -27.742, -27.997,
 -28.321, -28.526, -28.477, -28.566, -28.703, -28.313, -28.177, -27.938, -27.553, -26.950,
 -26.669, -26.353, -26.153, -25.957, -25.550, -25.186, -24.903, -24.520, -24.074, -23.596,
 -23.095, -22.862, -22.494, -22.085, -22.549, -22.930, -22.078, -21.225, -21.001, -20.886,
 -20.487, -20.138, -19.946, -19.661, -19.242, -18.905, -18.777, -18.307, -17.924, -18.225,
 -18.330, -18.413, -19.217, -20.243, -20.393, -19.625, -18.683, -18.126, -17.408, -16.867,
 -16.380, -15.734, -14.984, -14.381, -13.732, -13.133, -12.501, -12.048, -11.650, -11.421,
 -11.187, -11.004, -11.204, -11.157, -11.240, -11.346, -11.469, -11.687, -11.683, -11.766,
 -11.920, -12.102, -12.081, -11.856, -11.768, -11.597, -11.157, -10.838, -10.163,  -9.819,
  -9.469,  -9.068,  -8.372,  -8.308,  -8.193,  -7.950,  -7.788,  -7.214,  -6.490,  -5.931,
  -5.345,  -4.545,  -3.921,  -3.213,  -2.257,  -1.156,   0.036,   1.241,   2.315,   3.391,
   4.469,   4.998,   4.642,   4.696,   9.078,  11.276,  10.147,  10.464,   8.916,   9.334,
  10.801,  12.033,  12.228,  12.333,  12.620,  12.786,  12.415,  12.092,  11.289,  10.560,
  10.507,  10.478,   9.750,   8.184,   6.397,   5.026,   5.062,   5.751,   6.811,   5.535,
   3.606,   5.564,   7.760,   6.490,   4.309,   4.101,   5.735,   6.040,   3.919,   2.789,
   2.903,   5.384,   9.784,  11.034,  12.219,  13.492,  15.315,  15.231,  15.943,  16.015,
  15.558,  15.533,  15.841,  16.028,  17.431,  19.223,  19.398,  19.775,  20.117,  20.866,
  21.392,  21.808,  22.488,  22.793,  23.290,  23.881,  24.394,  24.848,  25.463,  26.068,
  26.372,  26.532,  26.944,  27.393,  27.802,  28.226,  27.793,  27.557,  27.312,  27.168},
{
  26.391,  26.242,  26.194,  26.346,  26.004,  25.595,  25.635,  25.794,  25.964,  26.439,
  27.080,  28.613,  29.866,  30.967,  31.430,  32.671,  33.128,  33.759,  34.648,  35.061,
  35.502,  35.839,  36.415,  36.469,  36.552,  37.140,  36.986,  38.243,  36.682,  36.174,
  39.020,  38.750,  38.452,  39.014,  39.735,  41.184,  42.524,  43.977,  45.093,  46.156,
  47.044,  49.219,  48.713,  47.499,  46.969,  47.001,  47.243,  46.872,  45.583,  44.993,
  44.436,  44.329,  43.284,  43.441,  43.024,  42.389,  41.574,  40.726,  39.821,  38.742,
  37.848,  37.071,  36.760,  36.435,  36.067,  35.722,  35.478,  35.433,  34.815,  34.481,
  34.520,  36.712,  38.256,  37.833,  38.156,  36.809,  36.174,  33.447,  30.418,  28.806,
  28.041,  26.888,  25.051,  23.443,  21.984,  20.562,  18.995,  17.577,  16.070,  14.948,
  13.818,  12.202,  10.875,   9.379,   7.982,   6.471,   4.979,   3.620,   2.100,   0.812,
  -0.571,  -1.937,  -3.269,  -4.542,  -5.908,  -7.032,  -8.458,  -9.779, -11.018, -12.128,
 -13.320, -14.438, -15.554, -16.843, -17.829, -18.817, -19.387, -20.011, -20.665, -21.184,
 -21.553, -22.131, -22.616, -23.031, -23.615, -23.815, -24.029, -23.427, -23.040, -22.691,
 -22.449, -22.132, -21.896, -21.645, -21.382, -20.980, -20.778, -20.463, -20.014, -19.893,
 -19.124, -19.407, -18.841, -18.395, -18.243, -17.933, -18.810, -19.384, -19.060, -18.669,
 -19.055, -19.168, -18.837, -18.537, -19.119, -19.148, -18.778, -18.158, -18.185, -17.806,
 -15.322, -15.695, -17.192, -16.828, -17.826, -17.854, -15.449, -15.914, -15.558, -15.201,
 -16.005, -16.194, -16.514, -17.465, -18.802, -20.802, -23.096, -23.725, -24.215, -24.229, -24.654,
 -24.912, -25.335, -25.532, -25.658, -25.836, -25.720, -25.308, -24.689, -24.216, -23.846,
 -23.625, -23.836, -23.840, -23.492, -23.185, -23.088, -22.888, -22.451, -21.947, -21.478,
 -21.026, -21.005, -21.983, -22.123, -21.285, -20.350, -20.279, -20.171, -19.791, -19.407,
 -19.033, -18.752, -18.675, -18.608, -18.225, -18.221, -18.422, -18.485, -18.142, -18.090,
 -18.853, -19.516, -19.204, -18.678, -18.563, -18.520, -18.256, -17.863, -17.368, -16.563,
 -15.988, -15.366, -14.565, -13.936, -13.454, -12.904, -12.370, -11.875, -11.294, -10.904,
 -10.675, -10.536, -10.635, -10.681, -10.786, -10.994, -11.139, -11.345, -11.535, -11.701,
 -11.613, -11.544, -11.413, -11.149, -11.332, -10.988, -10.606, -10.036,  -9.839,  -9.518,
  -9.094,  -8.539,  -8.150,  -8.006,  -7.723,  -7.464,  -7.211,  -6.652,  -6.010,  -5.312,
  -4.811,  -3.981,  -2.987,  -2.067,  -1.139,  -0.091,   0.867,   1.784,   2.805,   4.108,
   5.098,   5.364,   4.505,   7.113,  12.788,  13.084,  12.689,  10.722,   9.546,   9.302,
   9.861,  10.293,  10.336,  11.143,  12.392,  13.174,  13.696,  13.865,  13.526,  11.863,
  11.660,  11.746,  12.079,  11.427,  10.129,   9.085,   7.674,   6.644,   6.198,   5.743,
   5.572,   5.138,   4.432,   4.091,   4.715,   5.010,   5.291,   5.183,   4.965,   5.806,
   6.682,   7.988,  10.296,  10.728,  10.475,  12.648,  15.023,  14.553,  15.166,  15.542,
  14.870,  14.533,  15.326,  17.691,  17.402,  18.332,  18.774,  19.397,  19.982,  20.419,
  21.020,  21.650,  22.411,  22.910,  23.554,  24.294,  24.759,  25.153,  25.672,  26.323,
  26.989,  27.312,  27.390,  27.968,  27.889,  27.597,  27.381,  26.961,  26.803,  26.391},
{
  26.405,  25.680,  25.353,  25.153,  25.197,  24.695,  25.146,  25.476,  25.697,  25.567,
  26.766,  26.876,  27.253,  29.473,  30.884,  31.200,  31.974,  32.430,  32.637,  33.691,
  34.060,  34.203,  34.654,  35.329,  35.593,  35.937,  37.058,  37.291,  37.508,  35.986,
  38.414,  39.577,  39.735,  39.442,  39.921,  40.892,  42.095,  43.349,  44.393,  45.533,
  46.068,  46.372,  46.593,  46.455,  46.556,  46.493,  46.147,  45.506,  45.112,  44.605,
  44.354,  44.337,  44.033,  43.913,  43.123,  42.829,  41.434,  40.863,  40.117,  39.240,
  38.423,  37.993,  37.437,  36.892,  36.025,  36.296,  36.892,  36.552,  36.258,  34.726,
  37.729,  38.602,  37.949,  37.650,  36.648,  36.017,  33.405,  31.118,  30.182,  28.772,
  27.789,  26.638,  25.058,  23.563,  21.961,  20.433,  18.937,  17.525,  16.047,  14.590,
  13.274,  11.925,  10.520,   9.192,   7.702,   6.149,   4.552,   3.043,   1.568,   0.096,
  -1.337,  -2.677,  -3.929,  -5.364,  -6.809,  -8.019,  -9.414, -10.719, -11.973, -12.946,
 -14.108, -15.300, -16.349, -17.420, -18.356, -19.415, -20.057, -20.547, -20.940, -21.497,
 -21.976, -22.592, -23.045, -23.402, -23.593, -23.689, -24.011, -23.486, -22.911, -22.511,
 -22.141, -21.891, -21.455, -21.122, -20.636, -20.325, -19.991, -19.501, -19.041, -18.740,
 -18.831, -18.731, -18.232, -17.761, -17.778, -17.367, -18.222, -18.356, -18.169, -17.609,
 -18.075, -18.277, -17.904, -17.736, -17.623, -17.653, -17.631, -17.521, -16.673, -15.911,
 -15.293, -13.716, -11.941, -12.132, -13.829, -14.101, -11.420, -11.578, -11.842, -12.135,
 -12.766, -13.057, -13.612, -13.794, -14.635, -15.783, -18.168, -19.647, -20.489, -20.879, -20.999,
 -21.240, -21.125, -21.404, -22.090, -22.595, -22.692, -22.189, -21.603, -21.203, -20.977,
 -21.212, -21.275, -21.147, -21.120, -21.094, -20.930, -20.535, -20.296, -19.874, -19.732,
 -20.449, -20.810, -20.127, -19.645, -19.604, -19.285, -18.947, -18.600, -18.019, -17.671,
 -17.503, -17.442, -17.118, -17.391, -17.861, -17.693, -16.778, -17.408, -18.432, -18.542,
 -17.830, -18.190, -18.382, -18.326, -18.194, -18.210, -18.039, -17.565, -17.001, -16.307,
 -15.463, -14.843, -14.049, -13.397, -12.691, -12.107, -11.580, -11.126, -10.736, -10.441,
 -10.265, -10.239, -10.238, -10.339, -10.450, -10.418, -10.403, -10.716, -10.930, -10.858,
 -10.803, -10.876, -11.010, -11.107, -10.873, -10.629, -10.299, -10.109,  -9.780,  -8.967,
  -8.202,  -7.683,  -7.557,  -7.415,  -7.357,  -6.962,  -6.479,  -6.063,  -5.182,  -4.361,
  -3.603,  -2.914,  -1.993,  -1.160,  -0.324,   0.566,   1.565,   2.410,   3.410,   4.528,
   5.351,   5.042,   4.609,   9.192,  13.698,  14.942,  14.837,  12.613,  10.834,  10.376,
   9.808,   9.953,   9.969,  10.815,  12.306,  13.331,  13.271,  11.770,  12.204,  12.848,
  12.555,  12.657,  11.999,  10.762,   9.795,   9.604,   8.541,   7.083,   6.000,   5.537,
   5.797,   5.938,   5.397,   5.898,   6.285,   7.583,   8.766,   7.850,   6.514,   6.041,
   6.271,   6.523,   7.437,   7.589,   8.091,   9.924,   9.944,  10.862,  11.720,  12.170,
  12.874,  13.327,  15.281,  16.386,  17.108,  17.280,  18.041,  18.662,  19.590,  20.203,
  20.990,  21.849,  22.630,  23.453,  24.257,  24.828,  25.337,  25.880,  26.282,  26.466,
  26.809,  27.286,  27.584,  27.873,  27.379,  27.037,  26.845,  26.592,  26.526,  26.405},
{
  25.022,  24.794,  24.914,  24.467,  24.562,  24.802,  24.473,  25.119,  23.581,  24.658,
  25.100,  26.493,  26.336,  26.998,  28.346,  29.694,  30.727,  31.463,  32.004,  32.538,
  33.226,  33.751,  34.147,  34.591,  35.068,  35.116,  35.666,  36.728,  37.135,  38.018,
  37.602,  40.460,  40.121,  40.163,  40.615,  40.737,  41.815,  42.765,  43.229,  44.323,
  44.974,  44.988,  44.978,  44.856,  44.527,  44.850,  45.218,  45.257,  45.185,  44.876,
  44.541,  44.466,  44.343,  43.972,  43.430,  42.827,  42.052,  40.969,  40.293,  39.256,
  38.223,  37.471,  37.019,  36.254,  39.017,  40.417,  38.728,  36.915,  38.766,  40.635,
  39.925,  37.367,  36.646,  36.342,  34.971,  32.970,  31.947,  30.801,  29.546,  28.746,
  27.216,  25.744,  24.308,  22.857,  21.615,  20.167,  18.680,  17.216,  15.782,  14.145,
  12.701,  11.403,  10.033,   8.589,   7.157,   5.650,   4.170,   2.455,   0.999,  -0.558,
  -2.186,  -3.619,  -5.044,  -6.343,  -7.728,  -9.262, -10.398, -11.744, -13.052, -14.149,
 -15.248, -16.386, -17.417, -18.324, -19.226, -20.084, -20.776, -21.285, -21.735, -22.006,
 -22.668, -22.993, -23.311, -23.620, -23.728, -23.635, -23.697, -23.466, -23.097, -22.821,
 -22.409, -22.114, -21.618, -20.991, -20.486, -19.921, -19.433, -18.900, -18.310, -17.925,
 -18.521, -18.077, -17.584, -17.140, -17.149, -17.072, -17.887, -17.995, -17.479, -16.865,
 -17.065, -16.904, -16.827, -16.768, -17.073, -16.905, -16.710, -16.514, -15.790, -14.906,
 -14.001, -13.060, -12.770, -10.967,  -9.566, -12.317,  -9.740,  -8.693,  -9.616, -10.209,
 -10.074,  -9.861,  -9.872, -10.148, -10.703, -11.765, -12.624, -13.219, -13.987, -15.207, -16.589,
 -16.843, -16.900, -15.304, -17.320, -19.481, -19.764, -19.180, -18.938, -18.718, -18.710,
 -18.942, -18.805, -18.581, -18.552, -18.356, -18.355, -18.449, -18.047, -18.124, -18.453,
 -18.113, -18.188, -18.320, -18.145, -17.989, -17.324, -16.743, -16.463, -16.330, -16.145,
 -16.015, -16.409, -17.055, -16.751, -16.497, -17.060, -17.775, -17.803, -17.828, -17.431,
 -17.042, -17.463, -17.584, -17.881, -17.833, -17.699, -17.473, -16.989, -16.553, -15.826,
 -15.029, -14.341, -13.569, -12.915, -12.448, -11.934, -11.551, -10.971, -10.561, -10.242,
  -9.951,  -9.797,  -9.598,  -9.726,  -9.982,  -9.996, -10.150, -10.408, -10.563, -10.620,
 -10.688, -10.741, -10.615, -10.441, -10.362, -10.231,  -9.811,  -9.365,  -9.144,  -8.583,
  -8.049,  -7.527,  -7.038,  -6.802,  -6.676,  -6.262,  -5.811,  -5.227,  -4.277,  -3.378,
  -2.654,  -2.049,  -1.072,  -0.401,   0.394,   1.144,   2.139,   3.195,   4.065,   4.949,
   5.550,   5.193,   5.609,  10.007,  13.496,  15.815,  16.107,  13.684,  12.879,  11.452,
  10.392,  10.430,  10.509,  10.537,  11.345,  12.073,  12.962,  12.893,  12.710,  12.383,
  11.900,  11.193,  10.503,   9.197,   7.817,   6.172,   4.380,   3.436,   3.157,   2.474,
   2.227,   2.379,   2.244,   2.274,   3.189,   4.527,   6.024,   6.377,   5.719,   3.419,
   2.807,   3.260,   3.954,   5.227,   5.664,   6.762,   7.518,   8.708,  10.012,  11.097,
  11.171,  12.433,  13.878,  14.971,  15.827,  16.851,  17.869,  18.943,  19.888,  20.422,
  21.092,  21.811,  22.595,  23.498,  24.636,  25.468,  26.015,  26.281,  26.743,  26.901,
  27.225,  27.270,  27.181,  26.891,  26.587,  26.408,  26.021,  25.752,  25.291,  25.022},
{
  23.761,  23.590,  23.342,  23.726,  24.092,  24.617,  24.494,  24.255,  25.107,  23.199,
  25.440,  26.264,  26.855,  27.670,  28.432,  28.919,  29.606,  30.281,  31.073,  32.258,
  32.657,  33.061,  33.315,  33.523,  34.374,  34.166,  34.917,  35.451,  36.809,  38.196,
  38.913,  39.546,  42.287,  41.602,  42.232,  42.383,  42.613,  43.018,  42.213,  43.339,
  44.024,  44.373,  44.576,  44.221,  43.767,  43.936,  44.775,  45.077,  44.956,  44.999,
  44.975,  44.856,  44.580,  44.203,  43.660,  43.016,  42.429,  41.379,  40.542,  39.710,
  38.624,  37.818,  38.157,  37.578,  36.902,  37.221,  37.607,  39.246,  41.023,  40.841,
  40.075,  37.643,  35.523,  34.085,  33.315,  32.152,  31.206,  29.932,  28.854,  27.886,
  26.443,  25.083,  23.703,  22.258,  20.809,  19.554,  18.084,  16.737,  15.206,  13.726,
  12.225,  10.773,   9.372,   7.779,   6.262,   4.859,   3.310,   1.850,   0.193,  -1.562,
  -2.884,  -4.513,  -5.991,  -7.506,  -8.845, -10.203, -11.728, -12.874, -14.127, -15.400,
 -16.621, -17.658, -18.658, -19.450, -20.312, -21.170, -21.565, -22.309, -22.782, -23.158,
 -23.382, -23.887, -23.984, -24.154, -24.196, -24.040, -23.936, -23.723, -23.369, -22.857,
 -22.395, -21.822, -21.268, -20.753, -20.287, -19.651, -19.015, -18.437, -17.783, -17.390,
 -18.191, -17.598, -17.162, -16.644, -16.506, -16.623, -17.221, -17.099, -16.248, -15.768,
 -14.012, -15.099, -16.147, -16.152, -15.955, -15.865, -15.788, -15.482, -14.842, -13.896,
 -12.450, -11.663, -10.715, -10.286, -10.999,  -7.869,  -8.838,  -6.222,  -7.446,  -7.990,
  -7.962,  -7.719,  -7.255,  -7.022,  -7.152,  -7.086,  -8.343,  -9.009,  -9.182,  -8.844,  -9.612,
 -11.736, -14.017, -15.018, -15.919, -16.443, -16.427, -16.484, -16.573, -16.423, -16.326,
 -16.452, -16.403, -16.099, -16.218, -16.107, -15.838, -16.124, -16.145, -16.092, -16.375,
 -16.451, -16.526, -16.317, -15.964, -15.493, -15.143, -14.886, -14.576, -14.630, -15.386,
 -16.001, -15.958, -15.416, -16.182, -16.460, -16.508, -16.846, -17.319, -17.051, -17.005,
 -17.186, -17.213, -17.274, -17.337, -17.311, -17.112, -16.815, -16.470, -15.845, -15.226,
 -14.617, -13.943, -13.258, -12.532, -11.928, -11.464, -11.036, -10.779, -10.549, -10.395,
 -10.302, -10.087,  -9.797,  -9.739,  -9.515,  -9.409,  -9.485,  -9.624,  -9.795,  -9.936,
 -10.049, -10.100, -10.260, -10.212, -10.106,  -9.725,  -9.453,  -9.238,  -8.814,  -8.125,
  -7.357,  -6.880,  -6.588,  -6.615,  -6.263,  -5.759,  -5.082,  -4.178,  -3.123,  -2.509,
  -1.736,  -1.253,  -0.697,   0.241,   1.125,   1.771,   2.646,   3.472,   4.447,   5.478,
   6.066,   5.606,   5.801,  10.229,  14.709,  16.335,  16.922,  16.049,  14.583,  12.911,
  11.900,  11.547,  11.393,  10.904,  10.911,  11.270,  11.571,  11.437,  11.302,  10.814,
  10.266,   9.565,   8.653,   7.200,   5.529,   3.091,   0.653,  -0.311,  -0.714,  -0.640,
  -0.702,  -0.884,  -0.927,  -0.595,   0.018,   0.678,   1.373,   1.787,   1.502,   1.319,
   1.338,   1.822,   2.662,   3.099,   4.138,   4.464,   5.593,   6.310,   6.795,   7.527,
   9.400,  11.691,  12.575,  13.524,  14.725,  15.996,  17.491,  18.783,  19.556,  20.504,
  21.346,  22.535,  23.430,  24.084,  24.871,  25.819,  26.551,  26.937,  27.389,  27.588,
  27.415,  26.559,  26.419,  25.727,  25.449,  25.093,  24.778,  24.448,  24.096,  23.761},
{
  23.319,  22.886,  22.596,  22.471,  22.633,  23.232,  23.834,  24.205,  25.961,  24.912,
  26.439,  26.011,  26.229,  27.119,  28.075,  28.625,  29.510,  29.918,  30.646,  31.927,
  32.301,  32.456,  32.419,  32.871,  33.208,  33.218,  33.871,  34.242,  35.280,  37.196,
  39.401,  40.035,  41.385,  43.548,  43.643,  44.048,  43.471,  43.353,  42.709,  43.135,
  43.733,  44.086,  43.702,  43.257,  43.905,  44.044,  44.949,  45.361,  44.727,  44.554,
  44.910,  44.876,  44.523,  44.083,  43.855,  43.259,  42.589,  41.658,  40.670,  39.660,
  38.457,  37.567,  36.910,  36.299,  35.509,  35.875,  38.627,  39.492,  39.952,  39.764,
  38.559,  37.248,  34.764,  32.812,  31.818,  30.860,  29.650,  28.567,  27.543,  26.542,
  25.371,  24.121,  22.919,  21.500,  20.000,  18.625,  17.368,  15.897,  14.507,  13.025,
  11.368,   9.754,   8.208,   6.654,   5.030,   3.582,   2.281,   0.758,  -0.881,  -2.601,
  -4.241,  -5.608,  -7.326,  -8.808, -10.240, -11.765, -13.091, -14.484, -15.598, -16.944,
 -18.024, -19.100, -20.007, -20.916, -21.735, -22.377, -22.991, -23.533, -24.017, -24.318,
 -24.462, -24.722, -24.894, -24.846, -24.818, -24.557, -24.412, -24.094, -23.816, -23.280,
 -22.671, -22.113, -21.515, -20.855, -20.243, -19.525, -18.757, -18.160, -17.529, -17.103,
 -17.913, -17.424, -17.018, -16.725, -16.072, -16.295, -15.733, -14.772, -13.842, -13.499,
 -14.281, -15.038, -14.954, -14.766, -14.968, -14.903, -14.516, -13.859, -13.539, -12.802,
 -11.412, -10.345,  -8.857,  -8.279, -10.592,  -9.587,  -6.133,  -3.223,  -3.853,  -4.642,
  -5.090,  -5.786,  -5.366,  -5.310,  -5.198,  -5.153,  -5.615,  -5.193,  -5.070,  -4.270,  -4.896,
  -7.830, -11.253, -11.921, -12.525, -12.901, -13.083, -13.491, -13.552, -13.715, -13.875,
 -14.028, -14.033, -14.109, -14.180, -14.155, -14.004, -13.790, -13.918, -14.360, -14.369,
 -14.466, -14.282, -13.925, -13.666, -13.390, -13.502, -13.460, -13.721, -14.119, -13.882,
 -13.568, -14.052, -14.920, -15.290, -15.572, -15.768, -16.142, -16.426, -16.692, -16.823,
 -17.074, -17.284, -17.190, -17.010, -16.758, -16.409, -15.955, -15.371, -15.008, -14.811,
 -14.247, -13.617, -12.928, -12.466, -12.080, -11.795, -11.490, -11.036, -10.753, -10.422,
 -10.108,  -9.897,  -9.763,  -9.777,  -9.657,  -9.432,  -9.391,  -9.555,  -9.604,  -9.753,
  -9.948,  -9.866,  -9.780,  -9.635,  -9.476,  -9.359,  -9.165,  -8.806,  -8.236,  -7.718,
  -7.322,  -6.851,  -6.770,  -6.432,  -5.693,  -5.028,  -4.294,  -3.359,  -2.288,  -1.521,
  -1.140,  -0.744,  -0.059,   0.831,   1.559,   2.294,   3.243,   3.962,   4.776,   5.897,
   6.742,   7.269,   6.903,  10.886,  15.004,  15.561,  17.402,  17.753,  15.690,  15.223,
  13.865,  11.604,  11.460,  11.499,  11.437,  11.377,  11.293,  11.080,  11.000,   9.990,
   8.441,   6.619,   4.975,   3.749,   2.182,   0.857,  -0.027,  -0.955,  -1.642,  -1.959,
  -2.308,  -2.762,  -2.886,  -2.809,  -2.113,  -1.333,  -1.126,  -1.132,  -0.714,  -0.322,
  -0.212,   0.164,   0.756,   1.515,   2.371,   3.318,   4.445,   5.834,   7.171,   8.452,
   9.978,  11.173,  12.568,  13.999,  15.370,  16.223,  17.519,  18.749,  19.946,  21.041,
  22.046,  22.832,  24.067,  24.865,  25.578,  26.217,  27.122,  27.724,  28.076,  28.095,
  27.213,  26.642,  26.157,  25.262,  24.635,  24.333,  23.990,  23.671,  23.523,  23.319},
{
  22.775,  22.182,  21.805,  21.860,  22.110,  22.784,  23.578,  25.014,  26.093,  24.808,
  25.798,  25.938,  26.005,  26.712,  27.025,  28.179,  29.019,  30.242,  30.844,  31.372,
  31.924,  32.271,  32.575,  32.534,  32.565,  33.014,  33.244,  33.561,  34.144,  35.409,
  37.492,  40.313,  40.972,  43.267,  44.702,  45.406,  44.921,  44.456,  46.260,  43.349,
  44.768,  43.865,  43.021,  43.058,  43.947,  45.165,  46.380,  46.152,  44.632,  44.600,
  46.184,  47.874,  47.603,  45.400,  43.528,  43.073,  42.347,  41.436,  40.340,  39.227,
  37.664,  36.528,  35.507,  35.228,  34.061,  34.420,  36.351,  37.503,  37.447,  36.487,
  35.680,  34.536,  33.235,  31.643,  30.650,  29.752,  28.651,  27.599,  26.305,  25.296,
  24.195,  22.838,  21.685,  20.459,  19.034,  17.735,  16.385,  14.955,  13.468,  11.878,
  10.333,   8.633,   7.046,   5.538,   3.695,   2.223,   0.852,  -0.785,  -2.452,  -4.027,
  -5.691,  -7.348,  -8.936, -10.487, -11.987, -13.297, -14.667, -15.969, -17.395, -18.614,
 -19.646, -20.618, -21.492, -22.358, -23.380, -23.859, -24.597, -24.704, -25.094, -25.379,
 -25.450, -25.596, -25.727, -25.605, -25.387, -25.269, -24.901, -24.688, -24.462, -23.763,
 -23.262, -22.559, -21.916, -21.309, -20.449, -19.659, -18.939, -18.123, -17.436, -17.148,
 -17.663, -17.050, -16.625, -16.518, -15.894, -16.132, -14.937, -13.352, -12.275, -12.924,
 -14.314, -14.318, -14.126, -13.501, -13.030, -12.999, -12.718, -12.131, -11.542, -10.996,
 -10.695,  -8.849,  -7.163,  -6.688,  -7.010,  -7.730,  -3.660,  -1.501,   1.088,   0.876,
  -0.904,  -2.040,  -2.443,  -2.475,  -2.422,  -2.608,  -2.750,  -3.138,  -3.322,  -3.528,  -4.735,
  -6.526,  -8.230,  -8.375,  -9.064,  -9.456,  -9.748,  -9.827, -10.329, -10.806, -11.463,
 -11.703, -11.918, -12.078, -12.264, -11.956, -12.017, -12.196, -12.233, -12.337, -12.196,
 -11.998, -11.878, -11.922, -11.543, -11.825, -11.749, -12.001, -12.266, -12.966, -13.348,
 -13.746, -13.967, -14.086, -14.291, -14.825, -15.110, -15.426, -16.011, -16.447, -16.644,
 -16.636, -16.730, -16.458, -16.190, -16.180, -16.043, -15.541, -15.119, -14.708, -14.499,
 -14.001, -13.621, -13.127, -12.673, -12.035, -11.574, -11.342, -11.198, -11.062, -10.787,
 -10.469, -10.176,  -9.936,  -9.832,  -9.631,  -9.463,  -9.395,  -9.474,  -9.533,  -9.583,
  -9.549,  -9.662,  -9.441,  -9.317,  -9.155,  -8.945,  -8.557,  -8.111,  -7.492,  -7.243,
  -6.826,  -6.060,  -5.989,  -5.646,  -5.008,  -4.524,  -3.861,  -2.993,  -2.065,  -1.213,
  -0.446,   0.062,   0.576,   1.189,   2.082,   3.040,   4.154,   4.820,   5.724,   6.809,
   7.737,   8.558,   8.946,   9.810,  13.460,  17.947,  20.140,  19.988,  17.482,  15.780,
  14.474,  13.036,  12.114,  11.607,  11.548,  11.333,  11.556,  11.408,  11.373,   9.467,
   6.789,   5.553,   3.781,   1.598,   0.119,  -0.773,  -1.486,  -2.237,  -2.932,  -3.603,
  -3.900,  -3.962,  -3.938,  -3.783,  -3.143,  -2.557,  -2.784,  -3.317,  -3.382,  -2.764,
  -2.471,  -2.262,  -1.823,  -1.077,   0.075,   1.557,   2.935,   4.405,   5.809,   7.392,
   9.086,  10.529,  11.986,  13.650,  14.807,  15.981,  17.421,  18.916,  20.275,  21.499,
  22.511,  23.479,  24.590,  25.714,  26.420,  27.401,  27.620,  27.264,  27.146,  27.252,
  27.013,  26.623,  25.335,  24.845,  24.287,  24.141,  23.632,  23.177,  22.860,  22.775},
{
  22.479,  21.504,  21.324,  21.231,  21.742,  22.287,  23.041,  24.014,  24.159,  25.002,
  25.368,  26.021,  26.679,  27.343,  27.479,  28.500,  29.112,  30.449,  31.388,  32.005,
  32.384,  32.748,  33.032,  33.073,  32.848,  32.906,  33.018,  33.457,  33.321,  34.328,
  35.496,  37.691,  39.963,  41.743,  44.549,  45.450,  45.874,  45.672,  45.670,  45.147,
  45.564,  44.993,  46.470,  45.870,  43.911,  46.320,  47.213,  46.733,  44.653,  47.419,
  51.917,  50.171,  47.302,  44.427,  43.209,  42.652,  42.205,  41.158,  39.834,  38.455,
  36.792,  35.297,  34.452,  33.691,  32.519,  32.599,  33.113,  33.973,  33.991,  33.090,
  32.677,  32.011,  31.159,  30.141,  29.335,  28.695,  27.799,  26.563,  25.410,  23.975,
  22.677,  21.599,  20.287,  19.219,  17.933,  16.738,  15.154,  13.633,  12.009,  10.535,
   8.965,   7.410,   5.768,   4.062,   2.436,   0.705,  -0.866,  -2.618,  -4.362,  -5.938,
  -7.475,  -9.182, -10.954, -12.469, -13.916, -15.145, -16.406, -17.670, -19.030, -20.502,
 -21.405, -22.397, -23.293, -24.112, -24.939, -25.507, -25.884, -26.010, -26.171, -26.279,
 -26.389, -26.385, -26.478, -26.340, -26.100, -25.880, -25.640, -25.265, -24.963, -24.379,
 -23.804, -23.153, -22.480, -21.700, -20.875, -19.935, -19.114, -18.327, -17.587, -17.301,
 -17.589, -17.021, -16.381, -16.034, -15.508, -15.289, -13.966, -13.744, -13.746, -13.567,
 -13.509, -13.194, -12.627, -11.878, -11.506, -11.024, -10.742, -10.483, -10.114,  -9.314,
  -8.279,  -7.556,  -6.503,  -4.771,  -4.564,  -5.307,  -2.462,   3.601,   3.523,   5.219,
   5.770,   2.591,   0.793,   0.573,   0.523,   0.872,   0.624,   0.081,  -0.351,  -1.105,  -1.921,
  -2.800,  -3.683,  -3.767,  -4.019,  -4.527,  -5.326,  -6.301,  -7.432,  -8.170,  -8.669,
  -9.364,  -9.621, -10.253, -10.306, -10.482, -10.648, -10.457, -10.367, -10.143,  -9.906,
  -9.843,  -9.753,  -9.746,  -9.188,  -9.928,  -9.774, -11.047, -11.482, -11.756, -12.259,
 -12.686, -12.845, -13.578, -13.649, -13.670, -14.570, -15.186, -15.414, -15.720, -15.692,
 -15.723, -16.168, -16.220, -16.145, -15.648, -15.347, -14.942, -14.940, -14.922, -14.372,
 -13.785, -13.453, -12.987, -12.619, -12.284, -11.745, -11.421, -10.956, -10.815, -10.692,
 -10.502, -10.411, -10.222, -10.060,  -9.816,  -9.725,  -9.527,  -9.564,  -9.637,  -9.561,
  -9.523,  -9.474,  -9.284,  -9.008,  -8.847,  -8.635,  -8.289,  -7.870,  -7.319,  -6.454,
  -5.720,  -5.662,  -5.396,  -4.803,  -3.964,  -3.600,  -3.247,  -2.634,  -2.147,  -1.318,
  -0.488,   0.148,   0.826,   1.660,   2.373,   3.606,   4.508,   5.415,   6.623,   7.597,
   8.531,   9.757,  10.784,  11.256,  14.942,  17.583,  22.910,  23.445,  18.753,  16.085,
  15.133,  14.371,  13.019,  11.730,  11.653,  11.923,  11.560,  11.037,  10.266,   8.724,
   5.423,   4.378,   2.246,   0.404,  -0.241,  -0.700,  -1.647,  -2.610,  -3.326,  -3.663,
  -4.122,  -4.062,  -3.716,  -3.561,  -3.307,  -3.547,  -4.238,  -4.572,  -4.773,  -4.185,
  -3.726,  -3.252,  -2.740,  -2.112,  -1.307,   0.198,   1.838,   3.274,   5.050,   7.034,
   8.890,  10.365,  11.868,  13.270,  14.567,  15.888,  17.332,  18.866,  20.283,  21.532,
  22.745,  23.880,  24.890,  25.705,  26.628,  27.008,  26.870,  26.475,  26.622,  26.094,
  25.338,  24.732,  24.572,  24.074,  23.468,  23.238,  23.308,  22.689,  22.752,  22.479},
{
  21.243,  21.160,  21.433,  21.588,  22.345,  21.921,  22.849,  23.629,  24.672,  25.723,
  26.060,  26.320,  26.778,  27.378,  28.200,  28.456,  29.412,  30.033,  31.142,  32.032,
  32.462,  32.759,  33.238,  33.449,  33.327,  33.301,  33.258,  33.211,  33.257,  33.372,
  34.290,  35.281,  37.990,  39.174,  42.333,  43.888,  46.218,  46.454,  46.635,  46.411,
  46.354,  45.954,  45.588,  45.261,  45.364,  45.642,  46.176,  45.757,  44.824,  44.489,
  44.892,  44.569,  43.762,  43.201,  42.752,  41.916,  41.210,  40.152,  38.726,  37.141,
  35.605,  34.147,  33.032,  32.229,  31.766,  31.719,  31.433,  31.094,  31.032,  31.044,
  30.883,  30.334,  29.593,  28.582,  27.766,  26.795,  26.117,  25.237,  23.970,  22.882,
  21.420,  20.177,  19.041,  17.807,  16.422,  15.029,  13.707,  11.994,  10.585,   9.019,
   7.587,   6.042,   4.365,   2.679,   0.873,  -0.734,  -2.591,  -4.548,  -6.159,  -7.973,
  -9.424, -10.881, -12.583, -14.124, -15.701, -16.957, -18.534, -19.878, -21.034, -22.456,
 -23.486, -24.292, -25.210, -25.968, -26.539, -27.126, -27.312, -27.409, -27.345, -27.482,
 -27.350, -27.425, -27.407, -27.356, -26.987, -26.654, -26.364, -25.997, -25.736, -25.127,
 -24.438, -23.618, -22.976, -22.243, -21.313, -20.413, -19.329, -18.517, -17.648, -17.132,
 -17.202, -16.809, -16.177, -15.787, -15.291, -14.607, -13.726, -13.147, -13.252, -12.393,
 -11.796, -11.771, -11.113, -10.654,  -9.699,  -9.272,  -8.941,  -8.318,  -8.189,  -7.628,
  -6.651,  -5.851,  -4.355,  -3.466,  -2.516,  -2.834,  -5.782,  -0.690,   4.173,   5.204,
   7.694,   6.839,   5.508,   4.975,   5.271,   5.701,   5.419,   4.911,   4.574,   3.828,   3.240,
   3.010,   2.668,   2.574,   1.443,  -0.111,  -1.450,  -3.012,  -4.076,  -5.132,  -6.067,
  -6.824,  -7.619,  -8.038,  -8.317,  -8.553,  -8.439,  -8.303,  -8.209,  -8.006,  -8.089,
  -7.650,  -6.929,  -7.877,  -8.394,  -8.468,  -9.180,  -9.576, -10.279, -10.863, -11.539,
 -11.923, -12.078, -12.724, -13.013, -13.430, -13.994, -14.459, -14.836, -14.889, -15.025,
 -15.096, -15.455, -15.816, -15.737, -15.191, -15.167, -14.951, -14.560, -14.398, -14.283,
 -13.803, -13.369, -13.058, -12.594, -12.204, -11.818, -11.634, -11.394, -11.235, -10.932,
 -10.853, -10.674, -10.661, -10.423, -10.201, -10.023,  -9.803,  -9.763,  -9.751,  -9.692,
  -9.671,  -9.399,  -9.081,  -8.873,  -8.530,  -8.368,  -7.959,  -7.425,  -6.958,  -6.642,
  -6.159,  -5.503,  -4.663,  -4.036,  -3.681,  -3.451,  -2.973,  -2.354,  -1.765,  -1.019,
  -0.274,   0.429,   1.231,   2.196,   3.054,   3.886,   4.855,   5.811,   6.443,   7.210,
   8.862,  10.408,  11.133,  10.791,  13.389,  16.116,  19.349,  18.874,  17.759,  16.927,
  15.847,  14.694,  13.546,  12.308,  11.925,  11.909,  11.252,  11.030,  10.798,   8.724,
   4.965,   2.918,   1.645,   0.303,   0.109,  -0.390,  -1.471,  -2.407,  -2.775,  -3.122,
  -3.331,  -3.709,  -3.776,  -3.634,  -3.480,  -3.644,  -4.375,  -4.857,  -5.025,  -4.836,
  -4.423,  -3.967,  -3.346,  -2.787,  -1.616,  -0.149,   1.399,   2.922,   4.820,   6.801,
   8.617,  10.250,  11.806,  13.358,  14.984,  16.548,  17.744,  19.226,  20.834,  22.031,
  23.112,  23.930,  24.893,  25.890,  26.328,  26.211,  26.068,  25.974,  25.856,  25.658,
  25.344,  25.260,  24.938,  24.831,  24.268,  23.646,  22.505,  21.777,  21.480,  21.243},
{
  20.774,  21.131,  21.171,  21.070,  21.296,  22.085,  22.969,  24.111,  24.250,  25.305,
  26.398,  27.135,  27.348,  27.524,  28.047,  28.109,  29.369,  29.686,  30.473,  31.441,
  31.888,  32.140,  32.651,  32.928,  33.375,  33.634,  33.626,  33.229,  33.319,  33.204,
  33.655,  34.236,  35.138,  37.844,  39.551,  41.659,  43.924,  45.429,  45.719,  45.694,
  46.194,  45.620,  44.485,  44.434,  44.479,  44.514,  44.189,  44.154,  44.107,  43.915,
  43.542,  43.164,  42.848,  42.829,  41.875,  40.664,  39.372,  38.504,  37.160,  35.724,
  34.170,  32.982,  32.414,  31.701,  31.117,  30.575,  30.188,  29.887,  29.842,  29.672,
  28.999,  28.548,  27.749,  26.893,  26.019,  25.194,  24.218,  23.217,  22.351,  21.402,
  20.322,  18.988,  17.558,  16.325,  15.103,  13.448,  12.121,  10.618,   9.045,   7.618,
   6.068,   4.412,   2.681,   0.900,  -0.875,  -2.774,  -4.529,  -6.435,  -8.173,  -9.774,
 -11.698, -13.140, -14.440, -16.142, -17.708, -19.198, -20.629, -22.020, -23.318, -24.456,
 -25.467, -26.239, -27.202, -27.698, -28.167, -28.679, -28.772, -28.875, -28.679, -28.786,
 -28.569, -28.613, -28.399, -28.177, -27.753, -27.480, -27.068, -26.662, -26.458, -25.909,
 -25.124, -24.163, -23.296, -22.589, -21.724, -20.563, -19.542, -18.690, -17.759, -17.010,
 -16.785, -16.247, -15.818, -15.119, -14.965, -13.007,  -9.746,  -8.084,  -9.376, -10.685,
  -9.014,  -8.861,  -9.269,  -8.881,  -8.198,  -7.395,  -7.245,  -6.403,  -5.674,  -5.834,
  -5.067,  -4.131,  -2.927,  -1.477,  -0.871,   0.164,   1.381,  -0.350,   1.986,   5.106,
   7.793,   9.135,   9.499,  10.669,  10.396,  11.055,  11.603,  11.051,  10.564,  10.224,  10.247,
   9.792,   9.260,   9.780,   9.231,   6.369,   3.251,   1.282,  -0.231,  -2.080,  -3.511,
  -4.233,  -4.631,  -5.249,  -5.548,  -5.608,  -5.565,  -5.456,  -5.589,  -4.111,  -5.520,
  -6.674,  -6.989,  -6.721,  -7.107,  -7.214,  -8.197,  -8.503,  -9.270,  -9.750, -10.349,
 -10.943, -11.434, -11.714, -12.217, -12.418, -13.188, -13.654, -14.042, -14.341, -14.503,
 -14.803, -14.995, -14.677, -14.333, -14.589, -14.493, -14.262, -14.094, -14.193, -14.005,
 -13.681, -13.486, -13.133, -12.754, -12.327, -11.934, -11.614, -11.442, -11.280, -11.280,
 -11.190, -11.099, -11.022, -10.814, -10.678, -10.351, -10.144,  -9.947,  -9.889,  -9.768,
  -9.712,  -9.405,  -9.119,  -9.022,  -8.820,  -8.322,  -7.995,  -7.467,  -6.835,  -6.348,
  -5.784,  -5.438,  -5.006,  -4.325,  -3.999,  -3.579,  -3.026,  -2.447,  -1.847,  -1.217,
  -0.254,   0.828,   1.750,   2.627,   3.324,   4.202,   5.286,   6.328,   7.182,   8.302,
   9.201,  10.024,  10.711,  10.318,  12.349,  13.930,  16.666,  18.079,  18.037,  18.771,
  17.541,  16.010,  14.507,  13.593,  12.121,  11.396,  11.387,  11.070,  11.095,   9.861,
   6.350,   5.584,   2.415,   0.909,   0.584,  -0.086,  -1.008,  -2.051,  -2.261,  -2.337,
  -2.559,  -2.761,  -3.144,  -3.351,  -3.805,  -4.667,  -5.360,  -5.972,  -6.369,  -6.251,
  -5.700,  -5.272,  -4.413,  -3.165,  -1.819,  -0.475,   0.984,   2.890,   4.960,   6.881,
   8.433,   9.989,  11.655,  13.166,  14.723,  16.305,  17.803,  19.102,  20.572,  21.922,
  22.980,  23.804,  24.831,  25.688,  25.456,  25.579,  25.747,  25.504,  24.904,  24.697,
  24.525,  24.117,  23.818,  23.330,  23.231,  22.309,  21.833,  22.689,  22.172,  20.774},
{
  18.793,  20.380,  21.243,  21.128,  21.180,  21.845,  22.872,  23.849,  24.518,  24.953,
  26.397,  27.514,  27.784,  27.986,  27.649,  28.417,  29.004,  29.388,  29.846,  30.297,
  30.988,  31.589,  32.583,  33.265,  33.432,  33.717,  34.059,  33.478,  33.704,  33.549,
  32.996,  33.291,  33.900,  35.388,  37.417,  39.526,  40.580,  42.885,  43.277,  43.685,
  44.833,  45.415,  43.785,  43.487,  43.615,  43.655,  43.344,  43.399,  43.582,  43.661,
  43.280,  42.604,  42.256,  41.937,  40.553,  39.029,  37.878,  36.882,  36.262,  34.894,
  33.350,  32.277,  31.511,  30.699,  30.275,  29.680,  29.515,  29.118,  28.729,  28.211,
  27.450,  26.683,  25.601,  24.777,  24.283,  23.571,  22.531,  21.652,  20.732,  19.760,
  18.725,  17.719,  16.476,  14.821,  12.963,  11.994,  10.593,   9.099,   7.587,   5.988,
   4.389,   2.726,   0.854,  -0.996,  -2.804,  -4.668,  -6.522,  -8.432, -10.545, -12.133,
 -13.930, -15.922, -16.951, -18.463, -20.037, -21.440, -22.709, -24.079, -25.216, -26.369,
 -27.443, -28.217, -28.870, -29.362, -29.798, -30.063, -30.353, -30.204, -30.048, -30.000,
 -29.994, -29.841, -29.550, -29.414, -28.791, -28.327, -27.849, -27.328, -27.073, -26.474,
 -25.604, -24.660, -23.811, -23.065, -21.995, -20.847, -19.683, -18.636, -17.889, -16.619,
 -16.199, -15.763, -15.001, -14.911, -14.091,  -8.970,  -5.227,  -4.055,  -5.051,  -8.223,
  -8.741,  -8.111,  -7.501,  -6.904,  -6.422,  -5.483,  -4.754,  -4.435,  -3.962,  -3.296,
  -2.884,  -1.543,  -0.396,   0.639,   2.018,   2.571,   4.347,   5.140,   7.276,   8.345,
   9.474,  11.149,  12.836,  11.676,  10.406,  12.340,  13.510,  14.460,  14.833,  14.324,  13.704,
  13.205,  12.435,  11.796,  10.307,   8.742,   6.650,   4.316,   2.619,   1.941,   0.441,
  -0.709,  -2.159,  -2.479,  -2.583,  -2.705,  -3.226,  -3.323,  -2.865,  -3.856,  -4.498,
  -4.328,  -4.807,  -5.343,  -5.901,  -6.036,  -6.742,  -7.422,  -8.230,  -8.906,  -9.161,
  -9.960, -10.614, -11.336, -11.790, -12.175, -12.524, -12.737, -13.321, -13.704, -13.984,
 -14.310, -14.180, -14.313, -14.230, -14.217, -14.383, -14.283, -14.190, -13.916, -13.839,
 -13.598, -13.313, -12.953, -12.661, -12.355, -12.011, -11.946, -11.821, -11.732, -11.619,
 -11.559, -11.477, -11.301, -11.115, -10.836, -10.628, -10.350, -10.120,  -9.971,  -9.939,
  -9.802,  -9.499,  -9.295,  -8.960,  -8.701,  -8.424,  -7.954,  -7.663,  -7.506,  -7.111,
  -6.298,  -5.492,  -5.384,  -4.902,  -4.199,  -3.514,  -2.950,  -2.323,  -1.720,  -1.040,
  -0.212,   0.799,   1.783,   2.441,   3.277,   4.301,   5.186,   6.224,   7.207,   8.155,
   9.078,  10.219,  11.204,  10.073,  11.704,  14.558,  16.802,  19.864,  20.016,  20.285,
  20.329,  18.932,  16.121,  14.253,  13.037,  12.718,  11.924,  11.498,  11.734,  11.498,
   9.003,   5.595,   3.502,   1.372,   0.501,   0.009,  -0.884,  -1.758,  -2.167,  -2.318,
  -2.622,  -2.792,  -3.111,  -3.574,  -4.364,  -5.425,  -6.311,  -6.936,  -7.197,  -6.906,
  -6.309,  -5.506,  -4.503,  -3.334,  -2.047,  -0.783,   0.888,   3.208,   5.118,   6.844,
   8.370,   9.767,  11.190,  12.552,  14.156,  15.969,  17.581,  19.203,  20.410,  21.620,
  22.737,  23.812,  24.823,  25.358,  25.179,  25.272,  25.161,  24.776,  24.416,  23.948,
  23.509,  23.382,  22.840,  22.504,  22.220,  21.198,  21.980,  20.852,  18.988,  18.793},
{
  23.573,  21.356,  21.521,  19.511,  20.181,  20.674,  21.633,  22.925,  24.000,  24.926,
  25.761,  26.345,  27.409,  28.238,  28.346,  28.588,  28.734,  29.255,  29.394,  29.658,
  30.094,  30.590,  31.397,  32.494,  34.127,  35.063,  35.263,  34.801,  34.478,  33.603,
  32.863,  33.114,  33.436,  33.799,  35.463,  36.999,  38.213,  39.905,  41.099,  41.734,
  42.531,  43.619,  44.036,  43.285,  43.571,  43.534,  42.746,  42.843,  42.776,  42.751,
  42.081,  41.604,  40.472,  40.227,  38.939,  37.366,  36.383,  35.365,  34.677,  33.442,
  32.111,  31.004,  29.947,  29.611,  29.225,  28.472,  28.296,  28.072,  27.702,  27.095,
  26.301,  25.165,  24.169,  23.046,  22.067,  21.657,  20.927,  20.110,  19.184,  18.130,
  16.990,  15.980,  14.704,  13.343,  11.800,   9.918,   8.931,   7.486,   5.828,   4.178,
   2.454,   0.570,  -1.252,  -3.081,  -5.025,  -6.806,  -8.839, -10.824, -12.700, -14.494,
 -16.107, -18.059, -19.426, -20.788, -22.155, -23.558, -25.027, -26.051, -27.403, -28.423,
 -29.220, -30.074, -30.647, -31.014, -31.230, -31.513, -31.650, -31.406, -31.318, -31.212,
 -31.151, -30.962, -30.872, -30.455, -29.886, -29.331, -28.851, -28.294, -27.818, -27.133,
 -26.272, -25.198, -24.130, -23.296, -22.308, -20.974, -19.795, -18.632, -17.644, -16.694,
 -16.157, -15.400, -15.262, -14.478, -11.623,  -5.565,  -1.927,  -1.341,  -1.741,  -5.100,
  -6.905,  -6.719,  -5.929,  -4.811,  -3.933,  -3.265,  -2.496,  -1.993,  -1.267,  -0.569,
  -0.101,   0.857,   2.047,   3.421,   4.372,   5.651,   7.329,   8.729,  10.722,  11.751,
  12.668,  14.200,  15.150,  14.960,  10.729,   8.993,   9.704,  11.324,  14.074,  15.900,  15.302,
  14.670,  13.731,  12.713,  11.304,  10.213,   8.543,   7.167,   5.305,   4.240,   2.600,
   1.620,   0.577,   0.391,   0.061,  -0.937,   0.115,  -0.136,  -1.700,  -2.175,  -2.465,
  -2.630,  -3.599,  -3.578,  -4.149,  -4.723,  -5.337,  -5.825,  -6.503,  -7.129,  -7.991,
  -8.831,  -9.963, -10.646, -11.433, -11.714, -11.639, -11.874, -12.349, -12.935, -13.357,
 -13.326, -13.564, -13.609, -13.805, -13.821, -13.802, -13.844, -13.733, -13.591, -13.554,
 -13.206, -12.894, -12.666, -12.382, -12.180, -12.031, -11.978, -12.215, -12.355, -12.337,
 -12.177, -12.012, -11.814, -11.537, -11.272, -10.894, -10.576, -10.337, -10.127, -10.056,
  -9.945,  -9.809,  -9.412,  -9.175,  -8.934,  -8.589,  -8.490,  -8.170,  -7.628,  -7.388,
  -7.402,  -6.222,  -5.081,  -4.918,  -4.246,  -3.560,  -2.824,  -2.032,  -1.512,  -0.931,
  -0.357,   0.505,   1.385,   2.190,   3.065,   3.924,   4.908,   5.897,   7.056,   8.132,
   9.228,  10.352,  11.369,  10.717,  11.752,  15.742,  17.073,  20.066,  21.224,  21.697,
  23.061,  22.721,  19.322,  15.734,  14.114,  13.293,  12.642,  12.303,  12.425,  12.209,
  11.606,   9.730,   5.790,   2.333,   0.903,   0.049,  -0.609,  -1.342,  -2.026,  -2.734,
  -2.987,  -3.224,  -3.646,  -4.462,  -5.423,  -6.457,  -7.216,  -7.706,  -7.615,  -7.162,
  -6.400,  -5.289,  -4.334,  -3.234,  -1.932,  -0.647,   0.661,   2.648,   4.717,   6.358,
   7.999,   9.738,  11.407,  12.821,  14.409,  16.040,  17.673,  19.070,  20.387,  21.773,
  22.710,  23.586,  24.594,  25.136,  24.892,  24.939,  24.778,  24.454,  24.258,  23.826,
  23.720,  23.353,  23.119,  22.336,  21.973,  21.411,  20.203,  19.774,  19.650,  23.573},
{
  17.031,  18.016,  18.293,  19.079,  19.475,  19.881,  20.593,  21.710,  23.171,  24.637,
  25.697,  26.679,  26.967,  27.088,  28.887,  28.686,  28.779,  29.023,  29.425,  29.365,
  29.555,  29.908,  30.222,  31.204,  32.608,  34.837,  35.528,  35.385,  34.636,  33.264,
  33.250,  33.167,  33.210,  32.710,  33.316,  34.632,  35.997,  36.911,  38.340,  39.479,
  40.311,  41.663,  42.680,  42.670,  43.362,  43.715,  42.639,  42.593,  42.123,  41.608,
  40.996,  40.885,  39.963,  39.140,  37.190,  36.418,  35.127,  33.586,  32.493,  31.418,
  30.185,  29.362,  28.514,  28.269,  27.792,  27.265,  27.057,  26.829,  26.444,  25.819,
  24.875,  23.513,  22.365,  21.343,  20.293,  19.260,  19.053,  18.482,  17.702,  16.426,
  15.111,  13.718,  12.564,  11.134,   9.542,   7.955,   6.507,   5.433,   3.722,   1.857,
   0.057,  -1.691,  -3.632,  -5.457,  -7.445,  -9.367, -11.338, -13.216, -14.928, -16.847,
 -18.557, -20.181, -21.855, -23.070, -24.452, -25.585, -26.968, -28.242, -29.238, -30.390,
 -31.040, -31.754, -32.308, -32.695, -32.712, -32.969, -32.793, -32.741, -32.782, -32.597,
 -32.482, -32.290, -31.954, -31.522, -30.990, -30.297, -29.647, -29.025, -28.472, -27.841,
 -26.871, -25.994, -24.738, -23.458, -22.301, -21.033, -19.684, -18.452, -17.165, -16.191,
 -15.775, -15.274, -14.034, -11.652,  -5.904,  -2.332,  -1.377,  -1.036,  -0.259,  -2.531,
  -4.609,  -4.229,  -3.211,  -1.846,  -1.480,  -0.666,   0.079,   0.827,   1.049,   1.610,
   2.543,   3.649,   4.655,   5.616,   7.108,   8.193,   9.398,  12.540,  14.383,  15.439,
  15.831,  16.241,  17.782,  16.287,  12.837,  13.304,  16.848,  15.938,  15.402,  17.285,  18.896,
  18.282,  16.598,  14.966,  13.030,  11.601,  10.171,   7.309,   6.300,   5.450,   4.134,
   4.322,   3.190,   2.453,   2.349,   3.910,   1.915,   0.374,   0.326,  -0.595,  -0.947,
  -1.143,  -1.474,  -1.460,  -2.317,  -3.134,  -3.595,  -3.921,  -5.101,  -5.879,  -7.417,
  -8.309,  -9.551, -10.359, -10.889, -11.017, -11.040, -11.184, -11.738, -12.162, -12.465,
 -12.795, -13.102, -13.371, -13.439, -13.367, -13.239, -13.452, -13.454, -13.186, -13.119,
 -12.985, -12.729, -12.623, -12.333, -12.162, -12.068, -12.076, -12.488, -12.718, -12.786,
 -12.656, -12.316, -12.016, -11.724, -11.498, -11.132, -10.849, -10.570, -10.275, -10.287,
 -10.127,  -9.855,  -9.605,  -9.185,  -8.758,  -8.476,  -8.000,  -7.825,  -7.668,  -7.439,
  -6.632,  -4.958,  -4.334,  -3.832,  -3.269,  -2.691,  -2.183,  -1.779,  -1.285,  -0.940,
  -0.445,  -0.072,   0.426,   1.047,   1.791,   2.740,   3.847,   5.094,   6.361,   7.699,
   9.037,  10.319,  11.142,  11.120,  11.317,  15.713,  18.506,  21.453,  22.141,  22.547,
  23.980,  23.298,  19.799,  16.255,  14.143,  13.580,  12.695,  12.695,  13.462,  12.620,
  12.388,  11.145,   9.444,   5.162,   2.104,   0.646,  -0.264,  -0.987,  -1.722,  -2.671,
  -3.313,  -3.872,  -4.572,  -5.411,  -6.179,  -7.085,  -7.620,  -7.840,  -7.646,  -7.103,
  -6.181,  -5.212,  -4.339,  -3.358,  -2.333,  -1.119,   0.267,   2.131,   4.170,   6.184,
   7.964,   9.524,  11.135,  12.668,  14.195,  15.990,  17.501,  18.928,  20.335,  21.438,
  22.171,  23.085,  24.246,  24.502,  24.389,  24.415,  24.258,  24.026,  23.686,  23.441,
  22.992,  22.887,  22.374,  22.295,  20.431,  20.129,  19.686,  18.740,  17.750,  17.031},
{
  17.474,  17.639,  18.301,  18.796,  19.345,  19.615,  20.033,  20.827,  22.081,  23.628,
  24.771,  25.629,  26.421,  26.635,  26.091,  27.889,  27.769,  28.775,  29.365,  29.264,
  29.418,  29.355,  29.838,  30.241,  31.928,  34.104,  35.035,  34.514,  34.116,  33.317,
  32.979,  32.943,  34.210,  33.728,  33.173,  33.085,  33.627,  34.441,  35.237,  36.592,
  38.227,  39.577,  40.867,  42.023,  42.612,  42.759,  43.013,  42.294,  41.590,  40.969,
  40.493,  40.080,  38.797,  37.746,  36.109,  34.697,  33.408,  31.993,  30.624,  29.638,
  28.655,  27.846,  27.316,  27.118,  26.726,  26.086,  25.743,  25.641,  25.343,  24.268,
  23.221,  21.989,  20.575,  19.529,  18.450,  17.815,  17.010,  16.978,  16.356,  14.716,
  13.083,  11.668,  10.010,   8.780,   7.290,   5.847,   4.123,   2.412,   0.968,  -0.579,
  -2.478,  -4.329,  -6.203,  -8.068,  -9.996, -11.975, -13.763, -15.663, -17.539, -19.253,
 -21.037, -22.609, -24.243, -25.439, -26.680, -28.008, -29.177, -30.185, -31.083, -31.967,
 -32.551, -33.168, -33.671, -33.926, -34.173, -34.106, -34.197, -34.210, -34.099, -33.805,
 -33.574, -33.247, -32.900, -32.436, -31.830, -31.071, -30.386, -29.691, -28.933, -28.315,
 -27.493, -26.331, -25.496, -24.200, -22.618, -21.098, -19.499, -18.045, -16.740, -15.381,
 -15.067, -13.533, -11.686,  -7.415,  -1.966,  -1.725,  -1.002,  -0.032,   0.795,   0.169,
  -2.335,  -2.061,  -0.672,   0.584,   1.424,   2.019,   2.774,   3.262,   3.825,   4.112,
   4.853,   5.872,   7.285,   8.382,  10.101,  11.751,  12.241,  15.323,  17.607,  18.551,
  19.136,  19.367,  19.721,  18.905,  15.432,  12.876,  15.995,  18.284,  19.026,  18.794,  20.541,
  20.184,  18.614,  16.883,  15.151,  12.796,  10.559,   9.340,   8.195,   6.892,   6.166,
   5.962,   5.485,   4.779,   5.946,   2.689,   3.090,   2.640,   1.878,   1.188,   0.935,
   0.784,   0.147,  -0.271,  -0.757,  -1.692,  -2.286,  -3.098,  -4.242,  -5.603,  -6.922,
  -8.234,  -8.870,  -9.581,  -9.967, -10.351, -10.321, -10.917, -11.362, -11.568, -12.035,
 -12.301, -12.339, -12.482, -12.637, -12.810, -12.863, -12.875, -12.908, -12.792, -12.745,
 -12.645, -12.437, -12.171, -12.214, -12.090, -12.137, -12.467, -12.683, -12.944, -12.902,
 -12.685, -12.593, -12.224, -12.022, -11.732, -11.347, -10.988, -10.727, -10.484, -10.317,
 -10.267, -10.040,  -9.758,  -9.368,  -8.948,  -8.357,  -7.980,  -7.723,  -7.437,  -7.116,
  -6.468,  -5.748,  -4.638,  -3.813,  -3.144,  -2.599,  -2.032,  -1.607,  -1.218,  -1.121,
  -0.727,  -0.365,   0.080,   0.398,   1.034,   2.229,   3.276,   4.465,   5.549,   6.855,
   8.081,   9.377,  10.719,  10.968,   9.575,  13.400,  17.317,  21.500,  23.292,  22.746,
  22.671,  21.084,  18.592,  16.347,  14.687,  14.091,  13.248,  13.045,  13.519,  13.555,
  13.253,  12.948,  12.206,   9.582,   5.104,   1.783,   0.385,  -0.399,  -1.641,  -2.633,
  -3.549,  -4.418,  -5.101,  -5.787,  -6.434,  -6.970,  -7.428,  -7.506,  -7.389,  -6.836,
  -6.027,  -5.186,  -4.393,  -3.173,  -2.210,  -1.106,   0.094,   2.123,   4.172,   6.314,
   8.072,   9.582,  11.211,  12.915,  14.421,  15.708,  17.107,  18.685,  19.602,  20.882,
  21.838,  22.552,  23.287,  23.883,  23.865,  23.667,  23.223,  23.233,  23.214,  23.167,
  22.270,  22.154,  21.878,  21.853,  19.967,  18.942,  18.432,  17.917,  17.522,  17.474},
{
  16.804,  17.348,  18.016,  18.945,  19.445,  19.811,  20.359,  21.040,  22.059,  22.890,
  23.851,  24.868,  25.929,  26.356,  28.275,  26.652,  27.558,  27.226,  27.526,  28.848,
  29.232,  29.302,  29.838,  29.688,  30.254,  32.397,  33.502,  33.136,  32.397,  32.678,
  31.983,  31.795,  32.260,  32.260,  32.156,  32.178,  31.987,  32.245,  32.803,  34.153,
  36.083,  37.498,  38.586,  39.765,  40.413,  40.832,  41.192,  42.614,  41.941,  41.241,
  40.262,  39.882,  37.546,  36.578,  35.056,  33.107,  31.428,  29.976,  28.909,  27.773,
  27.042,  26.263,  25.890,  25.640,  25.680,  25.429,  24.954,  24.544,  23.759,  22.774,
  21.511,  20.140,  18.954,  17.683,  16.626,  16.033,  15.721,  14.987,  15.332,  12.895,
  11.149,   9.458,   7.946,   6.402,   4.908,   3.435,   1.907,   0.031,  -1.754,  -3.485,
  -5.137,  -6.885,  -8.641, -10.661, -12.449, -14.436, -16.342, -18.083, -19.991, -21.759,
 -23.362, -24.922, -26.563, -27.811, -29.018, -30.266, -31.313, -32.143, -32.863, -33.418,
 -33.756, -34.385, -34.914, -35.212, -35.362, -35.328, -35.299, -35.370, -35.307, -34.863,
 -34.502, -34.122, -33.775, -32.997, -32.523, -31.819, -31.087, -30.295, -29.424, -28.644,
 -27.753, -26.923, -25.557, -24.024, -23.088, -21.682, -19.858, -18.086, -16.040, -14.709,
 -13.052, -10.218,  -6.272,  -2.647,  -1.034,   0.795,   1.953,   1.627,   2.693,   0.793,
  -0.889,   0.103,   1.823,   2.886,   4.182,   4.634,   5.304,   6.108,   6.553,   6.774,
   7.316,   8.242,   9.434,  11.007,  13.024,  15.550,  15.535,  18.243,  21.042,  22.406,
  22.112,  21.942,  22.361,  23.680,  23.299,  25.113,  24.561,  19.726,  21.300,  20.200,  21.633,
  22.077,  20.858,  19.399,  17.231,  15.554,  12.990,  11.647,  10.986,   9.932,   8.700,
   7.767,   9.603,   8.322,   5.897,   5.801,   4.965,   4.197,   3.376,   2.940,   2.619,
   2.382,   1.346,   0.660,   0.046,  -0.952,  -1.906,  -3.042,  -4.278,  -5.700,  -6.691,
  -7.607,  -8.137,  -8.615,  -8.950,  -9.468, -10.196, -10.550, -10.881, -11.218, -11.464,
 -11.701, -11.872, -12.113, -12.249, -12.278, -12.435, -12.669, -12.580, -12.460, -12.321,
 -12.350, -12.379, -12.280, -12.223, -12.206, -12.203, -12.356, -12.627, -12.947, -12.915,
 -12.894, -12.713, -12.412, -12.119, -11.829, -11.490, -11.132, -10.702, -10.370, -10.280,
 -10.112, -10.033,  -9.705,  -9.502,  -9.065,  -8.492,  -7.897,  -7.669,  -7.422,  -6.938,
  -6.376,  -5.658,  -4.661,  -3.996,  -3.448,  -2.832,  -2.385,  -1.874,  -1.691,  -1.308,
  -0.977,  -0.514,  -0.134,   0.510,   0.807,   1.979,   3.156,   4.379,   5.651,   6.701,
   7.928,   9.263,  10.306,  10.602,   9.342,  12.994,  16.970,  20.480,  24.361,  22.664,
  20.329,  18.480,  17.235,  15.859,  14.413,  13.883,  13.815,  14.071,  13.422,  12.307,
  12.655,  13.080,  12.606,  10.733,   7.669,   3.483,   1.329,  -0.179,  -1.560,  -2.698,
  -3.538,  -4.363,  -4.878,  -5.376,  -6.055,  -6.696,  -7.097,  -7.262,  -7.184,  -6.638,
  -6.324,  -5.862,  -4.858,  -3.794,  -2.715,  -1.257,   0.558,   2.473,   4.174,   6.414,
   8.174,   9.661,  11.209,  12.662,  14.198,  15.673,  16.996,  18.152,  19.094,  19.990,
  20.814,  21.734,  22.537,  23.257,  23.093,  22.865,  22.691,  22.609,  22.236,  22.198,
  21.772,  21.828,  20.407,  19.819,  19.016,  18.585,  17.540,  17.035,  16.845,  16.804},
{
  16.597,  16.947,  17.562,  18.551,  19.421,  20.174,  20.982,  21.635,  22.507,  22.708,
  23.321,  24.113,  25.140,  25.526,  26.014,  27.357,  26.582,  27.046,  27.338,  26.861,
  27.802,  29.370,  29.088,  28.643,  29.254,  30.717,  31.445,  31.353,  31.284,  31.019,
  31.399,  31.038,  31.245,  31.410,  32.228,  31.103,  30.092,  30.254,  31.169,  32.279,
  33.756,  34.821,  35.782,  37.201,  38.416,  38.833,  39.142,  40.913,  40.721,  40.916,
  40.554,  39.463,  37.344,  35.864,  33.721,  32.330,  30.619,  28.608,  27.070,  26.381,
  25.527,  25.082,  24.864,  24.685,  24.600,  24.267,  23.828,  23.177,  22.296,  21.208,
  20.014,  18.612,  17.336,  16.190,  15.019,  14.326,  14.022,  13.445,  13.903,  12.230,
   9.000,   7.369,   5.591,   4.327,   2.614,   0.927,  -0.462,  -2.337,  -4.206,  -6.035,
  -7.701,  -9.557, -11.234, -13.284, -14.900, -16.999, -18.906, -20.710, -22.422, -24.165,
 -25.662, -27.182, -28.571, -30.036, -31.273, -32.378, -33.225, -33.847, -34.121, -34.367,
 -34.888, -35.452, -36.152, -36.256, -36.474, -36.653, -36.300, -36.881, -36.338, -36.246,
 -35.499, -35.268, -34.585, -34.087, -33.494, -32.720, -31.844, -30.927, -29.835, -28.920,
 -27.792, -26.748, -25.559, -24.230, -22.455, -20.831, -19.275, -17.220, -15.441, -11.729,
  -7.929,  -3.801,  -0.710,   1.500,   2.961,   3.864,   6.677,   6.520,   6.489,   7.076,
   6.400,   2.491,   4.141,   5.426,   6.456,   7.433,   8.266,   8.715,   9.158,   9.144,
  10.206,  11.018,  12.266,  13.432,  16.701,  18.448,  18.822,  21.183,  23.866,  25.078,
  24.891,  24.137,  24.911,  26.090,  28.302,  30.039,  29.203,  25.870,  22.123,  20.767,  20.806,
  22.742,  23.001,  21.741,  20.439,  18.099,  15.811,  14.216,  13.284,  12.820,  11.071,
  11.886,  12.152,   8.911,   8.078,   7.188,   6.351,   5.185,   4.356,   4.321,   3.895,
   3.121,   1.858,   0.692,  -0.257,  -1.286,  -2.338,  -3.094,  -4.102,  -5.185,  -6.268,
  -7.132,  -7.404,  -7.679,  -8.409,  -9.391,  -9.726, -10.225, -10.726, -10.766, -10.876,
 -11.069, -11.212, -11.318, -11.690, -11.898, -11.996, -11.928, -11.939, -11.744, -11.962,
 -12.164, -12.318, -12.207, -12.274, -12.042, -12.049, -12.113, -12.488, -12.711, -12.729,
 -12.765, -12.546, -12.420, -12.054, -11.774, -11.362, -10.862, -10.469, -10.035,  -9.995,
  -9.855,  -9.852,  -9.678,  -9.453,  -9.077,  -8.695,  -8.262,  -7.775,  -7.300,  -6.895,
  -5.966,  -5.269,  -4.492,  -3.811,  -3.198,  -2.750,  -2.397,  -2.313,  -2.068,  -1.804,
  -1.493,  -1.118,  -0.598,   0.107,   0.991,   2.002,   3.042,   4.187,   5.400,   6.469,
   7.540,   9.036,  10.284,  10.940,   8.839,  14.327,  19.892,  20.896,  24.476,  23.923,
  21.223,  18.674,  17.644,  16.057,  15.546,  15.218,  15.776,  15.137,  14.370,  14.617,
  14.948,  14.912,  13.223,  11.023,   8.589,   4.260,   2.062,   0.834,  -1.095,  -2.480,
  -3.384,  -3.891,  -4.563,  -5.141,  -5.731,  -6.402,  -6.828,  -7.085,  -6.901,  -6.668,
  -6.548,  -6.049,  -5.096,  -4.114,  -3.010,  -0.903,   0.619,   2.369,   4.161,   6.415,
   7.993,   9.468,  10.938,  12.532,  14.025,  15.234,  16.425,  17.824,  18.834,  19.729,
  20.454,  21.438,  21.977,  21.896,  21.902,  21.864,  21.642,  21.358,  21.176,  21.011,
  20.854,  20.241,  20.047,  19.437,  18.700,  17.316,  16.350,  15.888,  16.386,  16.597},
{
  16.734,  17.054,  17.604,  17.949,  19.152,  20.307,  21.274,  22.412,  23.115,  23.378,
  23.731,  24.425,  24.919,  25.477,  26.193,  26.233,  26.182,  26.921,  27.263,  27.143,
  28.893,  28.631,  29.374,  28.865,  28.273,  29.730,  30.270,  29.792,  29.443,  29.619,
  29.282,  29.889,  30.229,  30.604,  31.018,  29.807,  28.573,  28.399,  29.095,  30.051,
  30.947,  31.342,  32.244,  34.085,  36.366,  36.922,  36.725,  36.995,  37.605,  38.699,
  38.520,  37.705,  36.245,  35.616,  33.029,  31.368,  29.582,  27.648,  25.731,  25.080,
  24.574,  24.224,  24.321,  24.212,  24.153,  23.430,  22.531,  21.717,  20.675,  19.429,
  18.168,  16.937,  15.758,  14.732,  13.723,  12.896,  12.232,  11.786,  11.393,  10.293,
   8.000,   4.976,   3.625,   1.849,   0.335,  -1.362,  -3.004,  -4.607,  -6.632,  -8.367,
 -10.240, -12.104, -13.744, -15.660, -17.422, -19.420, -21.358, -23.199, -24.999, -26.406,
 -27.853, -29.318, -30.692, -32.032, -33.374, -34.300, -34.836, -35.520, -36.627, -37.390,
 -37.656, -38.600, -38.679, -39.004, -38.876, -38.504, -38.283, -38.944, -38.541, -38.397,
 -37.678, -36.958, -36.328, -35.448, -34.595, -33.766, -32.562, -31.809, -30.363, -29.425,
 -28.287, -27.172, -26.038, -24.655, -22.744, -20.885, -18.233, -13.475,  -8.585,  -6.333,
  -3.588,  -0.877,   2.071,   3.788,   5.494,   6.684,   8.790,  12.393,  13.516,  14.101,
  11.825,   5.584,   6.553,   8.001,   9.118,  10.294,  12.216,  11.343,  11.876,  11.959,
  12.625,  14.008,  15.687,  17.165,  19.252,  21.176,  22.331,  24.059,  25.059,  25.707,
  26.641,  27.442,  27.541,  29.011,  31.581,  33.466,  32.910,  30.085,  22.866,  21.427,  19.643,
  21.401,  23.105,  23.125,  22.378,  20.216,  18.427,  16.886,  15.466,  14.884,  14.432,
  12.791,  11.671,  11.225,   9.967,   8.161,   7.469,   6.096,   5.502,   4.897,   4.021,
   3.066,   1.868,   0.767,  -0.209,  -1.255,  -2.045,  -2.969,  -3.862,  -4.883,  -5.550,
  -6.045,  -6.507,  -7.208,  -7.857,  -8.445,  -9.112,  -9.640,  -9.984, -10.061, -10.189,
 -10.521, -10.757, -10.773, -11.241, -11.684, -11.919, -11.947, -11.728, -11.766, -11.805,
 -11.939, -11.768, -11.503, -11.676, -11.803, -11.756, -12.008, -12.411, -12.155, -12.511,
 -12.383, -12.300, -12.187, -11.928, -11.524, -11.205, -10.702, -10.373, -10.003,  -9.866,
  -9.839,  -9.864,  -9.537,  -9.109,  -8.775,  -8.436,  -7.954,  -7.178,  -6.712,  -6.181,
  -5.512,  -4.946,  -4.387,  -3.931,  -3.380,  -2.882,  -2.856,  -2.705,  -2.487,  -2.335,
  -2.069,  -1.651,  -0.716,   0.039,   0.974,   2.056,   3.265,   4.525,   5.713,   6.772,
   8.029,   9.205,  10.672,  11.164,   9.475,  12.411,  20.647,  20.842,  25.297,  25.345,
  22.925,  20.113,  17.882,  17.432,  17.173,  16.849,  16.561,  16.258,  16.598,  16.878,
  16.351,  15.302,  14.382,  12.219,   9.823,   8.055,   4.181,   2.316,  -0.365,  -2.203,
  -3.283,  -4.131,  -4.772,  -5.317,  -5.728,  -6.215,  -6.732,  -7.171,  -7.081,  -6.872,
  -6.578,  -6.119,  -5.311,  -4.195,  -2.449,  -0.850,   0.515,   2.450,   4.107,   5.950,
   7.349,   8.929,  10.675,  12.177,  13.473,  14.868,  15.990,  16.749,  17.863,  18.836,
  19.741,  20.631,  20.890,  20.657,  20.738,  20.546,  20.583,  20.312,  19.527,  19.466,
  19.314,  19.254,  18.621,  18.447,  18.007,  17.713,  16.902,  16.407,  16.183,  16.734},
{
  16.967,  17.591,  18.234,  18.620,  19.198,  20.120,  20.968,  21.917,  22.884,  23.609,
  24.177,  24.887,  25.311,  25.543,  25.584,  25.897,  26.637,  27.222,  28.404,  29.276,
  30.850,  31.136,  30.313,  27.502,  27.679,  27.782,  27.968,  28.004,  28.143,  28.161,
  27.799,  28.067,  29.520,  30.034,  30.456,  29.439,  27.146,  26.584,  26.998,  27.421,
  28.037,  28.335,  29.686,  31.871,  33.774,  34.526,  34.687,  34.064,  34.239,  34.498,
  34.527,  34.206,  33.963,  33.380,  32.872,  30.658,  28.854,  26.848,  25.180,  23.765,
  23.569,  23.561,  23.603,  23.667,  23.232,  22.513,  21.528,  20.344,  18.963,  17.784,
  16.588,  15.364,  14.304,  13.215,  12.017,  11.233,  10.609,   9.928,   9.156,   7.988,
   6.040,   3.870,   1.250,  -0.035,  -1.815,  -3.443,  -5.446,  -7.067,  -8.902, -10.686,
 -12.370, -14.277, -16.009, -17.972, -19.728, -21.852, -23.759, -25.491, -27.196, -28.470,
 -29.752, -31.152, -32.577, -33.497, -35.507, -37.064, -38.144, -38.806, -39.371, -39.850,
 -39.970, -39.864, -40.603, -40.615, -40.127, -39.493, -38.458, -37.894, -38.079, -38.487,
 -38.507, -38.303, -37.082, -36.426, -35.587, -34.879, -33.578, -32.704, -31.232, -30.176,
 -29.210, -28.052, -25.984, -24.252, -21.408, -18.493, -10.740,  -6.752,  -4.841,  -3.385,
  -1.050,   0.942,   3.031,   4.607,   6.394,   7.909,   9.506,  11.819,  15.704,  18.586,
  16.643,   9.663,   9.137,  10.500,  11.739,  13.044,  13.809,  14.024,  14.599,  14.742,
  15.745,  17.856,  18.696,  20.380,  22.157,  24.356,  25.130,  25.270,  26.641,  27.879,
  28.921,  29.424,  30.618,  32.731,  35.514,  36.786,  36.442,  34.459,  32.875,  28.277,  23.158,
  17.004,  21.553,  24.052,  24.102,  22.874,  20.773,  19.476,  17.876,  16.259,  16.238,
  14.180,  13.666,  12.189,  10.778,   9.398,   8.507,   6.942,   6.011,   5.061,   3.902,
   3.144,   1.718,   1.033,  -0.115,  -1.071,  -1.550,  -2.582,  -3.284,  -3.735,  -4.295,
  -5.250,  -6.277,  -6.740,  -7.229,  -7.864,  -8.631,  -9.134,  -9.432,  -9.506,  -9.744,
  -9.974, -10.257, -10.772, -11.141, -11.556, -11.423, -11.272, -11.247, -11.299, -11.435,
 -11.541, -11.483, -11.277, -11.334, -10.994, -11.608, -11.673, -11.751, -12.120, -12.225,
 -11.975, -12.006, -12.125, -11.617, -11.155, -10.796, -10.813, -10.512, -10.135,  -9.884,
  -9.735,  -9.595,  -9.164,  -8.942,  -8.317,  -7.962,  -7.565,  -7.211,  -6.639,  -6.183,
  -5.709,  -5.184,  -4.836,  -4.453,  -3.982,  -3.315,  -3.044,  -2.943,  -2.847,  -2.619,
  -2.237,  -1.689,  -1.163,  -0.435,   0.670,   2.119,   3.389,   4.856,   6.000,   7.252,
   8.435,   9.797,  11.023,  11.942,  11.155,   9.047,  18.537,  20.712,  24.786,  26.959,
  26.192,  22.846,  19.764,  18.696,  18.039,  17.715,  17.676,  17.535,  16.991,  16.837,
  17.226,  17.038,  15.286,  12.741,  11.306,   9.174,   7.140,   4.134,   1.495,  -1.459,
  -3.391,  -4.522,  -5.261,  -5.660,  -5.983,  -6.482,  -6.830,  -7.222,  -7.324,  -6.883,
  -6.749,  -6.336,  -5.462,  -4.140,  -2.840,  -1.289,   0.652,   1.975,   3.782,   5.557,
   7.353,   9.117,  10.394,  11.806,  13.283,  14.315,  15.540,  16.624,  17.600,  18.275,
  18.966,  19.614,  19.710,  19.518,  19.550,  19.280,  18.925,  18.499,  18.012,  18.061,
  17.940,  17.726,  17.216,  17.165,  17.496,  17.530,  17.168,  16.676,  17.003,  16.967},
{
  17.767,  18.255,  19.057,  19.614,  19.921,  20.208,  20.775,  21.464,  22.214,  23.157,
  24.075,  24.950,  25.451,  25.774,  25.569,  25.500,  26.321,  28.041,  29.606,  31.041,
  31.538,  30.816,  29.993,  29.129,  28.566,  26.017,  27.418,  27.510,  27.013,  26.334,
  26.327,  27.055,  28.297,  29.659,  29.746,  28.915,  26.522,  24.883,  25.099,  25.144,
  25.600,  26.066,  27.101,  30.291,  32.517,  33.176,  32.871,  31.870,  31.114,  31.117,
  30.898,  30.691,  30.432,  30.843,  30.329,  30.357,  28.630,  26.222,  24.739,  23.152,
  22.269,  22.093,  22.084,  21.994,  21.721,  20.862,  19.805,  18.570,  17.353,  16.190,
  15.041,  14.032,  12.746,  11.547,  10.513,   9.572,   8.644,   7.898,   6.909,   5.505,
   3.632,   1.766,  -0.546,  -2.405,  -4.008,  -5.784,  -7.066,  -9.246, -11.047, -13.017,
 -14.755, -16.591, -18.332, -20.301, -21.928, -23.934, -26.028, -27.726, -29.269, -30.405,
 -31.545, -32.754, -34.094, -37.120, -38.035, -39.742, -40.493, -40.499, -40.097, -39.987,
 -38.893, -38.224, -38.649, -40.078, -38.113, -34.807, -33.102, -32.179, -31.899, -33.668,
 -34.648, -34.549, -33.886, -33.035, -32.996, -33.176, -33.059, -32.553, -31.537, -30.612,
 -28.386, -25.044, -22.202, -19.022, -14.951, -10.529,  -6.662,  -3.837,  -2.067,  -0.432,
   1.131,   2.138,   4.077,   6.031,   8.151,   9.682,  11.757,  13.558,  16.853,  19.524,
  20.421,  17.015,  11.270,  13.288,  14.455,  15.201,  16.270,  16.834,  17.283,  17.192,
  19.417,  20.969,  22.439,  24.124,  24.951,  26.120,  26.006,  27.108,  28.217,  30.777,
  31.823,  31.679,  34.303,  37.676,  38.837,  38.148,  37.552,  37.761,  37.988,  36.732,  28.442,
  18.289,  19.917,  24.725,  26.083,  25.442,  23.436,  21.496,  19.630,  18.611,  17.194,
  15.641,  14.462,  12.918,  11.571,  10.354,   9.004,   7.700,   6.322,   5.429,   4.791,
   3.433,   2.200,   1.457,   0.370,  -0.310,  -1.195,  -2.046,  -2.812,  -3.259,  -3.805,
  -4.587,  -5.589,  -6.060,  -6.695,  -7.619,  -8.157,  -8.565,  -8.994,  -9.441,  -9.646,
  -9.958, -10.253, -10.633, -11.030, -11.121, -11.361, -10.930, -11.064, -11.014, -11.059,
 -11.222, -11.185, -11.184, -10.809, -11.090, -11.280, -11.656, -12.227, -11.711, -11.866,
 -11.519, -11.935, -11.661, -11.616, -11.168, -11.031, -10.823, -10.442, -10.147,  -9.809,
  -9.588,  -9.437,  -9.262,  -8.787,  -8.153,  -7.428,  -6.986,  -6.454,  -6.169,  -6.059,
  -5.786,  -5.501,  -4.861,  -4.423,  -3.814,  -3.345,  -2.964,  -2.864,  -2.547,  -2.296,
  -2.160,  -1.724,  -0.931,  -0.128,   0.929,   2.204,   3.549,   4.712,   5.968,   7.344,
   8.352,   9.703,  11.303,  12.556,  12.814,   9.523,  11.735,  20.708,  24.837,  30.058,
  25.802,  22.080,  20.908,  20.215,  19.665,  18.851,  18.153,  17.756,  17.320,  17.324,
  16.706,  15.920,  15.192,  13.508,  12.530,  10.329,   8.827,   7.689,   3.169,  -0.245,
  -3.058,  -5.039,  -5.835,  -6.065,  -6.282,  -6.959,  -7.309,  -7.619,  -7.336,  -7.169,
  -7.026,  -6.335,  -5.167,  -4.046,  -2.707,  -1.198,   0.485,   1.726,   4.927,   6.571,
   6.292,   7.948,   9.259,  11.077,  12.255,  13.675,  14.676,  15.844,  16.456,  17.087,
  17.813,  18.151,  18.264,  18.617,  18.684,  18.398,  17.886,  17.537,  17.235,  16.865,
  16.856,  16.610,  16.656,  16.669,  17.829,  17.386,  16.994,  17.153,  17.802,  17.767},
{
  18.370,  19.837,  19.765,  20.048,  20.557,  20.724,  21.250,  21.842,  22.567,  23.476,
  24.035,  24.639,  25.458,  25.870,  25.914,  25.875,  26.712,  28.629,  31.187,  31.621,
  32.708,  32.512,  31.749,  30.525,  29.572,  27.859,  27.202,  29.114,  27.752,  26.137,
  25.710,  25.914,  26.938,  29.109,  29.401,  28.909,  26.461,  23.321,  22.913,  23.009,
  23.215,  24.231,  25.055,  28.000,  31.378,  32.986,  31.428,  29.516,  28.051,  27.612,
  27.533,  27.184,  26.754,  27.020,  27.781,  28.538,  27.080,  24.923,  23.749,  22.005,
  20.913,  20.347,  20.014,  20.030,  19.515,  18.663,  17.688,  16.586,  15.501,  14.508,
  13.503,  12.243,  11.063,  10.061,   8.727,   7.869,   6.917,   5.871,   4.727,   3.123,
   1.476,  -0.583,  -2.420,  -4.451,  -6.212,  -7.668,  -9.063, -10.518, -12.963, -14.963,
 -16.892, -18.881, -20.591, -22.427, -24.146, -26.084, -28.070, -29.611, -31.085, -32.139,
 -32.950, -34.873, -36.623, -37.597, -39.895, -40.891, -40.943, -40.490, -39.729, -39.512,
 -36.850, -36.341, -37.126, -38.134, -35.634, -32.519, -30.618, -28.125, -27.927, -29.882,
 -30.380, -29.791, -29.958, -29.069, -27.202, -26.322, -27.159, -27.695, -27.670, -26.915,
 -23.947, -21.750, -18.195, -14.064, -11.394,  -8.012,  -4.164,  -1.656,   0.613,   2.872,
   3.714,   4.024,   5.676,   7.848,   9.736,  11.768,  14.161,  16.575,  19.131,  22.141,
  24.353,  22.309,  17.846,  15.355,  17.087,  17.792,  18.723,  19.161,  19.891,  20.128,
  23.300,  24.326,  25.604,  27.171,  27.255,  27.202,  28.362,  30.075,  31.745,  32.329,
  33.491,  35.723,  38.809,  38.750,  38.622,  39.241,  39.740,  40.113,  40.710,  40.563,  36.892,
  24.092,  17.699,  24.700,  27.338,  27.256,  25.633,  23.328,  20.789,  19.637,  17.223,
  16.220,  15.522,  13.615,  12.112,  10.586,   9.270,   7.973,   6.940,   6.261,   5.282,
   4.003,   2.724,   1.772,   1.065,   0.435,  -0.474,  -1.550,  -2.291,  -2.799,  -3.156,
  -3.882,  -4.439,  -5.133,  -6.079,  -7.284,  -7.997,  -8.862,  -9.326,  -9.744,  -9.959,
 -10.431, -10.556, -10.800, -11.272, -11.592, -11.869, -11.828, -11.741, -11.304, -11.040,
 -11.123, -11.068, -10.922, -11.086, -11.005, -11.553, -11.742, -11.932, -12.105, -11.470,
 -11.577, -11.292, -11.358, -11.303, -11.120, -10.896, -10.370, -10.156, -10.082,  -9.511,
  -9.131,  -8.913,  -8.588,  -8.280,  -7.874,  -7.457,  -7.062,  -6.736,  -6.326,  -6.030,
  -5.804,  -5.532,  -5.245,  -4.765,  -4.316,  -3.687,  -3.243,  -2.738,  -2.264,  -2.161,
  -1.708,  -1.271,  -0.590,   0.075,   1.079,   2.254,   3.290,   4.877,   6.599,   7.443,
  10.050,   9.599,  10.565,  12.425,  14.020,  13.314,   8.441,  19.159,  25.003,  31.957,
  24.428,  21.838,  22.359,  22.826,  21.803,  21.204,  19.402,  18.144,  17.264,  17.106,
  16.908,  16.511,  16.309,  15.564,  14.625,  12.785,   9.601,   8.433,   5.927,   2.140,
  -1.129,  -3.419,  -4.858,  -5.607,  -5.989,  -6.724,  -7.198,  -7.601,  -7.378,  -7.429,
  -7.092,  -5.957,  -5.232,  -3.885,  -2.673,  -0.836,   0.712,   2.822,   2.911,   4.745,
   6.233,   7.138,   8.555,   9.495,  10.678,  11.532,  12.772,  13.678,  14.662,  15.684,
  16.416,  17.174,  17.522,  17.687,  18.129,  17.873,  17.331,  16.761,  16.420,  16.104,
  16.064,  16.126,  16.162,  16.457,  16.779,  17.840,  17.204,  18.053,  18.774,  18.370},
{
  20.028,  19.761,  22.960,  20.934,  21.618,  21.813,  22.204,  22.901,  23.023,  23.664,
  24.290,  24.863,  25.615,  26.575,  26.714,  27.174,  27.770,  31.070,  31.180,  32.835,
  33.553,  34.167,  32.252,  31.107,  28.943,  27.626,  28.673,  29.470,  29.566,  28.603,
  25.333,  24.551,  24.928,  25.612,  27.337,  28.835,  25.918,  21.658,  20.449,  20.591,
  21.016,  22.104,  23.266,  26.113,  30.967,  30.260,  28.235,  26.222,  25.263,  24.673,
  24.079,  23.597,  22.973,  23.499,  23.718,  24.919,  25.384,  23.574,  22.123,  21.415,
  19.335,  18.554,  17.972,  17.881,  17.275,  16.527,  15.577,  14.734,  13.747,  12.790,
  11.646,  10.576,   9.343,   8.191,   7.052,   6.243,   5.167,   4.019,   2.519,   0.710,
  -1.146,  -2.900,  -4.881,  -6.577,  -8.350, -10.255, -11.144, -12.412, -14.676, -17.049,
 -19.087, -21.167, -22.854, -24.535, -26.538, -27.989, -29.762, -31.925, -33.396, -34.070,
 -35.036, -36.540, -38.793, -39.365, -40.092, -40.847, -40.395, -40.043, -40.628, -39.898,
 -37.750, -37.951, -38.555, -38.227, -35.786, -33.094, -30.416, -27.113, -26.918, -27.675,
 -28.227, -28.336, -28.248, -26.591, -25.565, -23.308, -22.108, -21.315, -21.133, -20.666,
 -19.137, -16.552, -14.558, -14.474, -11.013,  -6.754,  -2.361,   0.861,   2.627,   7.085,
   7.182,   8.162,   8.343,   9.067,  11.503,  14.017,  16.737,  18.869,  21.994,  24.827,
  26.895,  26.023,  25.087,  20.192,  19.427,  20.522,  24.360,  21.978,  23.538,  23.780,
  26.629,  27.644,  28.518,  29.706,  29.359,  30.380,  29.897,  33.921,  33.859,  34.590,
  36.392,  37.392,  38.477,  40.053,  40.065,  40.165,  40.959,  41.590,  42.036,  43.305,  42.507,
  34.707,  18.264,  23.690,  27.961,  28.482,  27.211,  24.276,  22.706,  20.222,  18.921,
  17.598,  15.907,  14.353,  12.412,  11.160,   9.995,   8.859,   7.666,   6.763,   5.611,
   4.522,   3.035,   1.990,   1.145,   0.527,  -0.262,  -0.636,  -1.577,  -2.183,  -2.985,
  -3.537,  -3.909,  -4.523,  -5.805,  -7.074,  -8.081,  -8.892,  -9.348,  -9.691,  -9.939,
 -10.267, -10.622, -10.846, -11.214, -11.571, -12.017, -12.078, -12.039, -11.845, -11.623,
 -11.551, -11.477, -11.177, -11.038, -11.157, -11.271, -11.307, -11.118, -11.129, -11.154,
 -11.109, -10.999, -10.866, -10.734, -10.610, -10.325, -10.033,  -9.661,  -9.593,  -9.336,
  -9.068,  -8.780,  -8.345,  -7.989,  -7.546,  -7.320,  -6.982,  -6.605,  -6.305,  -6.104,
  -5.848,  -5.597,  -5.072,  -4.812,  -4.372,  -3.484,  -2.897,  -2.322,  -1.878,  -1.571,
  -1.165,  -0.729,  -0.222,   0.403,   1.495,   2.507,   3.612,   4.633,   5.529,   6.383,
   7.614,   8.770,  10.754,  11.665,  13.543,  15.189,  10.331,  14.902,  26.202,  31.814,
  22.758,  20.643,  22.800,  25.473,  25.167,  23.351,  20.306,  18.842,  17.511,  16.968,
  17.536,  17.484,  16.976,  15.850,  14.879,  13.191,  11.002,   7.828,   7.296,   4.474,
  -0.291,  -2.653,  -4.091,  -5.039,  -6.105,  -6.993,  -7.614,  -7.633,  -7.773,  -7.553,
  -7.100,  -6.444,  -5.144,  -3.159,  -1.945,  -0.571,   0.974,   1.792,   3.435,   4.277,
   5.839,   6.600,   7.409,   8.400,   9.273,  10.756,  11.972,  12.948,  13.656,  14.493,
  15.274,  15.901,  16.354,  16.577,  17.284,  17.383,  16.718,  16.180,  15.802,  15.660,
  15.567,  15.713,  15.908,  16.418,  17.246,  17.350,  18.688,  18.819,  19.157,  20.028},
{
  20.044,  20.307,  22.922,  21.138,  22.933,  22.621,  22.763,  23.404,  24.088,  24.301,
  24.967,  25.372,  26.531,  27.104,  27.673,  28.637,  30.501,  30.447,  32.208,  33.015,
  34.923,  37.119,  35.859,  34.524,  34.105,  33.892,  33.448,  33.060,  31.017,  30.733,
  27.185,  24.263,  23.110,  23.124,  24.162,  25.332,  24.203,  19.956,  18.039,  18.144,
  18.549,  19.953,  21.124,  23.752,  26.327,  25.874,  24.966,  23.434,  22.620,  21.446,
  20.799,  20.237,  19.607,  19.911,  20.412,  21.053,  21.871,  20.673,  20.956,  20.761,
  18.544,  16.419,  16.114,  15.414,  15.056,  14.397,  13.779,  12.914,  12.028,  11.021,
   9.853,   8.645,   7.675,   6.452,   5.307,   4.358,   3.248,   1.970,   0.199,  -1.787,
  -3.683,  -5.633,  -7.167,  -9.038, -10.537, -12.559, -14.254, -14.635, -16.050, -18.959,
 -21.134, -23.240, -24.988, -26.214, -28.135, -29.728, -29.477, -30.650, -32.235, -33.756,
 -34.627, -35.821, -37.194, -40.195, -40.436, -41.046, -40.389, -40.629, -40.673, -40.106,
 -39.305, -39.092, -39.113, -38.281, -36.104, -34.522, -31.414, -27.085, -26.512, -25.909,
 -25.458, -25.887, -26.268, -23.809, -24.174, -21.822, -20.430, -19.332, -20.135, -19.435,
 -17.724, -14.821, -12.249, -12.264,  -8.952,  -3.681,   0.223,   2.182,   4.343,   8.556,
   9.689,  11.919,  12.659,  12.231,  13.151,  16.500,  20.025,  21.657,  23.708,  25.988,
  28.226,  29.709,  29.164,  27.127,  22.107,  23.418,  23.688,  24.825,  25.827,  28.395,
  29.197,  30.318,  30.912,  31.506,  31.049,  31.642,  32.498,  36.343,  36.705,  36.971,
  38.761,  39.842,  40.948,  42.272,  41.693,  41.765,  42.343,  42.703,  43.121,  44.194,  44.204,
  41.390,  23.321,  20.297,  28.352,  30.204,  28.565,  25.445,  24.469,  21.036,  19.963,
  18.691,  16.718,  15.034,  13.064,  11.658,  10.819,   9.588,   8.268,   6.968,   5.825,
   4.429,   2.955,   2.140,   1.084,   0.645,   0.207,  -0.497,  -1.545,  -2.252,  -2.873,
  -3.319,  -3.537,  -4.989,  -6.124,  -7.134,  -8.220,  -8.800,  -9.282,  -9.522,  -9.754,
 -10.015, -10.297, -10.648, -11.015, -11.516, -11.993, -12.302, -12.326, -12.333, -12.152,
 -11.936, -11.874, -11.467, -11.209, -10.966, -10.766, -10.731, -10.841, -10.856, -10.832,
 -10.776, -10.657, -10.621, -10.453, -10.186, -10.007,  -9.669,  -9.238,  -8.963,  -8.894,
  -8.724,  -8.454,  -8.054,  -7.820,  -7.466,  -7.112,  -6.824,  -6.472,  -6.199,  -5.831,
  -5.593,  -5.413,  -4.942,  -4.497,  -3.987,  -3.114,  -2.334,  -1.506,  -1.061,  -0.925,
  -0.555,  -0.200,   0.197,   0.698,   1.679,   2.636,   3.646,   4.863,   6.087,   7.095,
   8.597,  10.035,  11.878,  13.241,  14.572,  15.667,  12.447,  14.622,  29.447,  32.281,
  25.622,  21.265,  23.826,  25.296,  26.677,  24.522,  21.609,  19.536,  17.860,  16.834,
  17.037,  17.074,  16.372,  15.465,  14.450,  12.770,  10.766,   8.190,   5.830,   4.527,
  -0.997,  -2.668,  -3.797,  -4.923,  -6.230,  -7.559,  -7.588,  -7.381,  -8.016,  -7.481,
  -7.131,  -6.376,  -4.558,  -1.846,  -0.761,   0.459,   1.279,   2.282,   3.121,   4.181,
   5.552,   5.713,   6.681,   7.770,   8.775,   9.865,  11.057,  11.733,  12.546,  13.298,
  13.994,  14.604,  15.126,  15.733,  16.215,  16.427,  16.226,  15.962,  15.660,  15.551,
  15.416,  15.336,  15.400,  15.901,  16.870,  18.276,  18.521,  19.237,  19.414,  20.044},
{
  20.940,  21.184,  21.914,  24.234,  22.642,  22.759,  22.981,  23.830,  24.469,  25.152,
  25.765,  25.643,  26.602,  27.692,  28.559,  29.020,  31.119,  30.889,  32.143,  32.916,
  35.139,  34.901,  35.396,  34.612,  34.527,  33.867,  34.987,  35.622,  35.513,  32.035,
  30.251,  25.052,  22.302,  22.265,  22.378,  22.349,  21.635,  17.330,  15.876,  15.842,
  16.077,  16.364,  17.652,  20.198,  22.452,  22.480,  21.988,  20.901,  18.774,  17.533,
  17.158,  16.898,  16.596,  16.754,  17.604,  18.108,  18.062,  17.748,  18.831,  18.049,
  17.752,  15.158,  14.115,  13.261,  13.052,  12.566,  11.869,  11.088,  10.196,   9.108,
   8.036,   6.926,   5.632,   4.476,   3.378,   2.376,   1.162,  -0.446,  -2.196,  -4.443,
  -6.402,  -8.328, -10.097, -11.698, -13.177, -15.042, -16.759, -17.469, -17.747, -20.788,
 -22.974, -25.174, -26.878, -28.234, -28.100, -29.559, -32.457, -33.727, -34.379, -35.563,
 -36.906, -37.067, -38.562, -39.276, -39.345, -39.568, -40.296, -40.326, -40.187, -40.095,
 -39.783, -39.259, -38.355, -37.481, -35.569, -31.425, -30.299, -26.181, -25.103, -24.736,
 -24.343, -23.638, -23.100, -24.026, -21.342, -20.487, -19.950, -18.813, -20.238, -18.100,
 -16.238, -13.019,  -8.810,  -7.304,  -5.687,  -2.234,   1.243,   3.782,   6.625,  10.978,
  10.214,  13.575,  15.677,  16.132,  16.612,  18.355,  21.563,  24.488,  25.351,  27.243,
  29.576,  32.306,  32.950,  30.590,  25.284,  26.153,  28.792,  27.699,  28.693,  30.628,
  31.868,  32.741,  33.181,  33.864,  33.534,  33.533,  34.622,  36.692,  39.484,  40.135,
  41.292,  42.039,  43.031,  44.577,  43.293,  43.416,  43.377,  43.784,  44.316,  44.901,  46.172,
  45.056,  32.668,  19.095,  27.754,  30.975,  29.413,  28.040,  26.409,  22.826,  21.450,
  19.390,  17.480,  15.682,  13.880,  12.136,  11.424,  10.012,   8.380,   6.908,   5.727,
   4.142,   3.040,   1.795,   1.238,   0.554,  -0.064,  -0.713,  -1.299,  -1.635,  -1.894,
  -2.717,  -3.419,  -4.804,  -5.961,  -6.978,  -7.721,  -8.187,  -8.542,  -8.624,  -9.206,
  -9.493, -10.075, -10.766, -11.325, -11.893, -12.232, -12.570, -12.834, -12.804, -12.664,
 -12.409, -12.031, -11.725, -11.398, -11.240, -11.152, -10.836, -10.800, -10.690, -10.651,
 -10.596, -10.472, -10.218, -10.003,  -9.864,  -9.476,  -9.187,  -8.877,  -8.633,  -8.438,
  -8.220,  -7.961,  -7.591,  -7.215,  -6.877,  -6.585,  -6.244,  -5.927,  -5.633,  -5.249,
  -4.939,  -4.547,  -4.267,  -3.732,  -3.208,  -2.323,  -1.564,  -0.894,  -0.296,  -0.266,
   0.225,   0.599,   0.894,   1.445,   2.102,   3.091,   4.073,   4.938,   6.379,   7.296,
   8.475,   9.933,  11.315,  12.703,  14.432,  15.872,  13.300,  16.796,  31.432,  33.246,
  26.964,  23.719,  25.283,  25.039,  26.579,  25.669,  21.710,  19.591,  17.937,  17.082,
  16.814,  16.102,  14.960,  14.532,  13.633,  12.726,  11.135,   8.244,   5.750,   4.393,
   0.928,  -3.127,  -5.182,  -6.136,  -7.895,  -8.027,  -7.943,  -8.092,  -8.956,  -8.041,
  -7.055,  -5.405,  -3.737,  -0.704,   1.224,   2.714,   2.862,   2.376,   2.804,   3.992,
   5.386,   4.973,   5.527,   6.348,   7.437,   8.402,   9.323,  10.246,  11.001,  11.733,
  12.458,  13.255,  13.845,  14.439,  15.028,  15.801,  15.904,  15.357,  15.326,  15.233,
  15.087,  15.157,  15.470,  15.990,  16.630,  17.562,  18.324,  18.961,  20.208,  20.940},
{
  21.409,  22.997,  22.944,  25.249,  22.478,  23.471,  23.564,  24.222,  24.913,  25.625,
  25.904,  25.942,  26.532,  27.990,  29.431,  31.714,  30.758,  31.533,  33.847,  34.822,
  33.695,  33.039,  32.954,  32.924,  32.527,  32.289,  34.296,  33.209,  36.054,  34.574,
  30.214,  28.194,  23.908,  21.191,  21.063,  20.392,  19.567,  15.554,  13.917,  13.427,
  13.593,  13.900,  14.720,  15.247,  18.067,  19.151,  18.928,  17.670,  16.392,  14.934,
  14.436,  13.585,  13.623,  14.525,  14.865,  14.723,  13.975,  13.730,  13.975,  15.358,
  14.922,  13.669,  12.907,  11.565,  10.706,  10.397,  10.019,   9.234,   8.305,   7.224,
   6.075,   4.869,   3.785,   2.556,   1.229,   0.150,  -1.264,  -3.066,  -5.096,  -7.187,
  -9.295, -10.725, -12.697, -14.758, -16.027, -17.791, -19.158, -20.246, -20.848, -22.650,
 -25.197, -27.389, -29.799, -30.524, -31.466, -32.114, -33.490, -35.054, -35.565, -36.500,
 -37.669, -38.185, -38.140, -38.943, -38.831, -39.222, -39.556, -39.447, -39.922, -39.921,
 -39.551, -38.619, -37.187, -36.208, -32.685, -29.165, -27.495, -24.396, -23.813, -22.482,
 -22.596, -22.392, -21.565, -20.620, -20.387, -17.679, -18.070, -17.339, -18.365, -16.637,
 -14.407, -10.285,  -6.130,  -3.517,  -1.150,   0.429,   3.202,   5.848,   7.975,  11.752,
  12.594,  15.313,  17.624,  19.181,  20.860,  22.339,  24.109,  26.035,  27.663,  28.218,
  31.691,  33.923,  35.543,  34.641,  29.681,  29.254,  29.404,  29.943,  32.341,  32.798,
  33.445,  34.959,  35.997,  36.334,  35.717,  35.984,  37.463,  39.271,  42.462,  42.679,
  43.845,  44.773,  45.429,  45.952,  44.947,  44.955,  45.120,  45.252,  45.470,  45.637,  47.528,
  46.256,  39.853,  26.232,  24.655,  30.738,  30.521,  30.297,  26.626,  24.615,  22.536,
  20.527,  18.629,  16.451,  14.902,  13.523,  11.776,  10.160,   8.551,   7.000,   5.591,
   4.236,   3.083,   2.042,   1.283,   0.544,  -0.256,  -0.846,  -1.548,  -1.706,  -1.441,
  -2.171,  -3.657,  -4.731,  -5.650,  -6.275,  -7.007,  -7.416,  -7.672,  -8.216,  -8.780,
  -9.306, -10.040, -10.875, -11.408, -12.063, -12.614, -13.134, -13.287, -13.378, -13.163,
 -12.843, -12.309, -11.997, -11.713, -11.448, -11.210, -10.816, -10.739, -10.616, -10.455,
 -10.153,  -9.976,  -9.746,  -9.535,  -9.250,  -9.004,  -8.752,  -8.358,  -8.062,  -7.859,
  -7.563,  -7.290,  -7.063,  -6.716,  -6.340,  -6.025,  -5.612,  -5.231,  -4.925,  -4.449,
  -4.129,  -3.767,  -3.352,  -2.914,  -2.478,  -1.630,  -0.907,   0.140,   0.494,   0.855,
   0.972,   1.154,   1.613,   2.137,   2.611,   3.526,   4.657,   5.396,   6.536,   7.996,
   9.029,  10.171,  11.527,  13.440,  14.665,  15.745,  13.706,  15.168,  32.594,  36.671,
  31.787,  27.033,  24.801,  23.801,  23.945,  25.597,  22.552,  21.223,  19.125,  17.789,
  17.135,  15.326,  13.996,  12.757,  11.223,   9.533,   8.577,   6.125,   4.245,   2.771,
   3.222,   0.819,  -3.905,  -6.361,  -7.315,  -7.765,  -8.247,  -9.116,  -9.586,  -8.687,
  -7.201,  -6.046,  -4.056,  -1.330,  -0.144,   0.482,   0.502,   1.473,   2.635,   3.488,
   4.053,   4.141,   4.646,   4.914,   5.761,   6.873,   7.838,   8.656,   9.441,  10.248,
  11.045,  11.852,  12.632,  13.376,  14.181,  15.021,  15.257,  15.088,  14.982,  14.860,
  15.009,  15.256,  15.542,  15.950,  16.651,  17.498,  18.400,  19.111,  20.392,  21.409},
{
  21.247,  22.284,  24.099,  24.186,  23.033,  23.658,  24.142,  24.359,  24.668,  25.319,
  25.760,  26.109,  26.835,  28.499,  29.943,  31.624,  31.246,  32.311,  32.489,  32.115,
  31.271,  30.714,  32.178,  32.113,  30.796,  31.145,  31.267,  31.242,  34.344,  35.394,
  28.887,  26.734,  24.618,  19.295,  18.321,  18.110,  17.595,  13.161,  11.721,  11.255,
  10.999,  11.622,  11.911,  11.346,  13.217,  14.569,  14.888,  14.852,  14.549,  13.375,
  11.130,  10.184,  10.343,  11.228,  11.355,  10.881,  10.171,   9.636,   9.863,  10.737,
  11.454,  10.489,  10.843,  10.856,   9.717,   8.795,   7.918,   7.283,   6.418,   5.285,
   4.102,   2.970,   1.655,   0.374,  -0.796,  -2.217,  -4.195,  -5.981,  -8.201, -10.026,
 -11.865, -14.099, -15.449, -17.417, -19.007, -20.273, -21.920, -22.962, -24.535, -26.986,
 -28.527, -30.900, -32.543, -33.039, -33.458, -34.179, -35.518, -36.846, -35.893, -36.639,
 -38.261, -38.627, -38.433, -38.201, -38.355, -38.931, -39.001, -39.087, -39.404, -39.646,
 -39.257, -37.819, -36.453, -33.612, -28.097, -26.281, -23.436, -22.593, -21.428, -20.721,
 -20.454, -19.992, -18.046, -17.058, -16.777, -16.012, -16.261, -16.439, -16.116, -14.720,
 -12.107,  -9.198,  -4.116,  -0.993,   1.725,   3.821,   5.387,   7.571,   9.716,  11.534,
  13.599,  16.289,  18.854,  20.477,  22.821,  24.518,  26.228,  27.792,  30.086,  30.633,
  32.827,  34.792,  37.332,  37.059,  34.338,  31.154,  32.841,  32.126,  33.831,  35.432,
  35.484,  37.011,  38.445,  38.779,  38.295,  38.093,  39.715,  41.470,  46.101,  44.558,
  45.797,  46.627,  46.824,  47.232,  46.741,  46.474,  46.706,  46.579,  46.451,  46.410,  47.515,
  47.331,  45.525,  34.582,  23.098,  30.494,  31.478,  31.220,  28.495,  26.307,  23.337,
  21.362,  19.169,  17.089,  15.090,  13.599,  11.729,  10.191,   8.644,   7.183,   5.805,
   4.461,   3.118,   2.187,   1.115,   0.257,  -0.436,  -1.193,  -1.641,  -1.576,  -1.708,
  -2.670,  -4.032,  -4.642,  -5.100,  -5.373,  -6.059,  -6.595,  -7.078,  -7.541,  -8.198,
  -9.861, -10.714, -11.598, -12.353, -12.950, -13.294, -13.478, -13.508, -13.468, -13.152,
 -12.916, -12.540, -12.107, -11.831, -11.351, -10.988, -10.652, -10.410, -10.089,  -9.952,
  -9.797,  -9.532,  -9.247,  -8.978,  -8.714,  -8.375,  -7.997,  -7.759,  -7.358,  -6.943,
  -6.695,  -6.439,  -6.141,  -5.892,  -5.689,  -5.345,  -5.054,  -4.631,  -4.214,  -3.885,
  -3.470,  -3.180,  -2.682,  -2.180,  -1.704,  -0.942,  -0.292,   0.419,   1.118,   1.365,
   1.667,   1.819,   2.319,   2.750,   3.162,   3.886,   4.966,   5.847,   6.537,   7.843,
   9.255,  11.151,  12.024,  13.603,  15.317,  16.733,  15.799,  14.101,  30.768,  37.574,
  35.160,  31.142,  28.095,  25.633,  25.821,  24.422,  23.592,  22.161,  20.146,  18.655,
  17.325,  15.691,  13.772,  11.251,   9.681,   9.073,   8.330,   6.973,   6.211,   4.303,
   1.999,   1.266,  -3.009,  -6.316,  -8.015,  -8.684,  -9.085,  -9.217,  -9.460,  -8.639,
  -7.660,  -7.942,  -6.179,  -4.440,  -3.198,  -1.438,   0.141,   0.748,   2.463,   2.176,
   2.173,   2.254,   2.646,   3.402,   4.473,   5.428,   6.477,   7.412,   8.249,   8.999,
   9.918,  10.784,  11.692,  12.420,  13.186,  13.818,  14.543,  14.485,  14.613,  14.794,
  14.877,  15.067,  15.399,  15.971,  16.566,  17.530,  18.391,  19.130,  20.320,  21.247},
{
  20.852,  22.086,  23.245,  24.752,  24.020,  24.010,  24.538,  24.644,  24.779,  25.191,
  25.543,  25.610,  26.651,  28.495,  30.080,  31.542,  32.488,  31.879,  32.068,  31.913,
  30.187,  30.876,  31.122,  32.884,  31.022,  29.856,  30.281,  30.441,  32.366,  31.757,
  28.155,  25.688,  23.923,  18.744,  17.558,  16.659,  15.215,  12.324,   9.849,   8.466,
   8.564,   9.643,   9.927,   9.112,   8.474,  11.983,  12.325,  11.982,  12.074,  10.620,
   7.991,   7.185,   6.937,   6.841,   7.507,   7.737,   7.609,   7.055,   6.870,   6.819,
   7.222,   8.095,   8.866,   8.164,   8.150,   7.018,   6.725,   5.839,   4.155,   3.327,
   1.972,   0.890,  -0.307,  -1.564,  -3.124,  -5.122,  -7.113,  -9.161, -10.961, -13.213,
 -14.847, -16.538, -18.616, -19.992, -21.714, -23.170, -24.485, -25.853, -27.622, -29.907,
 -30.985, -33.754, -34.630, -35.281, -36.454, -37.191, -37.698, -38.376, -37.001, -37.682,
 -39.388, -38.993, -38.432, -37.651, -38.055, -38.547, -38.905, -38.852, -39.452, -39.121,
 -38.102, -36.806, -33.667, -28.901, -24.116, -21.458, -20.353, -19.857, -19.521, -19.021,
 -18.069, -16.213, -15.422, -14.125, -13.566, -13.666, -13.241, -12.382, -12.401, -11.983,
 -10.259,  -8.592,  -5.796,   0.042,   3.533,   5.143,   7.510,   9.025,  10.744,  13.214,
  15.034,  17.512,  19.897,  22.345,  24.746,  27.358,  29.005,  30.148,  32.165,  33.234,
  35.217,  38.060,  40.406,  39.827,  37.915,  35.362,  35.947,  34.854,  36.455,  36.490,
  37.620,  39.336,  40.636,  41.449,  41.172,  40.242,  41.708,  43.547,  45.710,  46.842,
  47.499,  48.240,  48.277,  48.637,  48.839,  48.328,  47.805,  47.491,  47.112,  47.290,  47.799,
  48.852,  47.804,  40.900,  24.847,  30.033,  35.302,  31.753,  29.799,  27.352,  24.443,
  21.812,  19.191,  17.196,  15.165,  13.397,  12.185,  10.386,   8.736,   7.436,   6.268,
   4.829,   3.607,   2.182,   1.069,   0.236,  -0.390,  -1.037,  -1.395,  -1.011,  -1.803,
  -2.995,  -3.729,  -4.169,  -4.701,  -5.296,  -4.383,  -5.093,  -6.392,  -8.051,  -9.125,
 -10.524, -11.290, -11.982, -12.580, -13.096, -13.403, -13.587, -13.550, -13.330, -12.973,
 -12.620, -12.246, -11.858, -11.485, -11.135, -10.831, -10.454, -10.004,  -9.823,  -9.471,
  -9.281,  -9.109,  -8.712,  -8.262,  -7.932,  -7.452,  -6.827,  -6.763,  -6.442,  -6.022,
  -5.919,  -5.563,  -5.312,  -4.938,  -4.583,  -4.425,  -4.005,  -3.973,  -3.400,  -3.160,
  -2.874,  -2.460,  -1.931,  -1.403,  -0.873,  -0.439,   0.096,   0.685,   1.292,   1.615,
   2.198,   2.740,   2.857,   3.422,   4.069,   4.498,   4.988,   6.092,   6.656,   8.031,
   9.421,  10.863,  12.363,  13.621,  15.264,  16.743,  17.073,  12.947,  25.786,  35.090,
  38.266,  34.975,  28.518,  29.066,  26.559,  24.078,  23.874,  23.221,  21.333,  19.847,
  18.243,  16.482,  13.965,  11.005,   8.609,   7.934,   7.841,   6.946,   6.396,   5.898,
   1.615,   1.238,  -1.594,  -6.291,  -7.855, -10.449, -10.331, -10.772, -11.036, -10.767,
  -9.863,  -8.956,  -6.953,  -5.533,  -4.081,  -2.910,  -1.610,  -0.988,  -0.182,  -0.161,
   0.626,   1.147,   1.711,   2.379,   3.305,   4.356,   5.284,   6.233,   7.128,   7.892,
   8.888,   9.738,  10.627,  11.579,  12.431,  13.331,  14.135,  14.243,  14.235,  14.367,
  14.624,  15.027,  15.425,  16.049,  16.708,  17.295,  17.986,  19.094,  19.995,  20.852},
{
  20.564,  21.615,  22.425,  24.201,  26.116,  25.421,  24.686,  24.768,  25.331,  25.186,
  25.288,  25.321,  26.420,  28.214,  31.186,  31.601,  33.354,  33.245,  30.946,  31.906,
  29.771,  29.373,  28.433,  28.662,  28.443,  28.733,  28.819,  28.941,  29.259,  29.125,
  28.342,  24.140,  22.382,  18.097,  15.416,  14.273,  12.853,  11.395,   7.341,   6.099,
   6.127,   7.011,   7.436,   6.817,   5.754,   9.279,  10.177,   9.299,   8.040,   6.332,
   4.911,   4.202,   3.712,   3.628,   4.183,   4.902,   5.193,   4.907,   4.074,   3.683,
   3.862,   4.249,   5.264,   5.879,   5.713,   5.676,   4.699,   2.612,   2.626,   1.202,
   0.004,  -1.160,  -2.365,  -3.956,  -5.824,  -7.874,  -9.998, -12.129, -13.968, -15.734,
 -18.007, -19.400, -21.222, -22.866, -24.445, -25.974, -27.083, -27.385, -28.975, -32.160,
 -33.726, -35.786, -36.881, -37.625, -38.700, -39.483, -39.888, -39.959, -39.999, -39.846,
 -39.784, -39.074, -38.974, -38.259, -37.988, -37.869, -38.008, -37.792, -37.887, -38.136,
 -37.224, -34.428, -30.094, -25.989, -23.006, -20.305, -17.952, -17.628, -17.979, -17.254,
 -15.262, -13.558, -12.570, -12.422, -10.837, -10.238, -10.260,  -8.247,  -7.152,  -5.220,
  -1.797,   0.011,   0.312,   0.512,   4.313,   6.749,   8.805,  11.654,  13.569,  15.988,
  17.691,  20.089,  21.928,  24.273,  26.489,  29.446,  31.683,  33.940,  35.211,  36.212,
  38.001,  41.217,  43.171,  42.639,  38.816,  37.144,  36.807,  37.152,  38.366,  40.566,
  40.506,  42.083,  43.424,  44.303,  44.003,  42.886,  45.006,  46.382,  47.989,  49.221,
  49.706,  50.248,  50.156,  50.252,  50.099,  49.599,  48.931,  48.100,  47.529,  47.361,  48.063,
  49.893,  49.082,  44.898,  29.076,  30.521,  33.612,  32.529,  30.593,  27.955,  24.625,
  21.516,  19.103,  17.470,  15.176,  13.375,  12.000,  10.224,   8.802,   7.465,   6.763,
   5.417,   4.102,   2.366,   1.560,   0.348,  -0.125,  -0.828,  -1.101,  -1.274,  -2.240,
  -3.262,  -3.610,  -4.107,  -4.014,  -3.951,  -5.217,  -6.296,  -7.403,  -8.965,  -9.969,
 -10.819, -11.718, -12.319, -12.941, -13.285, -13.372, -13.293, -13.273, -13.057, -12.729,
 -12.294, -11.882, -11.575, -11.286, -10.877, -10.360,  -9.852,  -9.377,  -8.938,  -8.577,
  -8.247,  -7.968,  -7.630,  -7.254,  -6.878,  -6.390,  -6.093,  -5.767,  -5.419,  -5.243,
  -4.978,  -4.771,  -4.528,  -3.984,  -3.609,  -3.421,  -3.243,  -2.992,  -2.618,  -2.324,
  -2.128,  -1.740,  -1.417,  -0.842,  -0.240,   0.041,   0.752,   0.790,   1.121,   1.718,
   2.085,   2.741,   3.390,   3.758,   4.342,   5.311,   6.133,   6.802,   7.709,   8.646,
   9.990,  11.712,  13.250,  14.212,  16.200,  17.726,  18.542,  13.359,  22.669,  34.594,
  41.244,  41.047,  36.927,  35.922,  26.365,  24.739,  24.158,  23.766,  22.292,  20.510,
  18.919,  16.949,  14.612,  11.311,   7.592,   5.925,   5.310,   6.505,   5.892,   4.045,
   0.776,   0.399,  -2.679,  -7.050,  -8.138, -10.558, -10.229, -10.143, -11.012, -11.527,
  -9.465,  -9.158,  -7.912,  -6.635,  -5.287,  -4.004,  -3.320,  -2.849,  -2.357,  -1.850,
  -1.361,  -0.766,   0.297,   1.333,   2.297,   3.185,   4.089,   4.964,   5.880,   6.858,
   7.779,   8.713,   9.681,  10.837,  11.817,  13.033,  13.594,  13.676,  14.065,  14.313,
  14.587,  14.938,  15.414,  15.872,  16.551,  17.432,  18.107,  18.884,  19.687,  20.564},
{
  20.098,  21.142,  22.109,  23.339,  24.905,  26.892,  28.629,  24.342,  24.728,  24.776,
  24.599,  24.629,  25.843,  27.868,  30.606,  32.328,  33.627,  33.624,  29.678,  28.358,
  29.044,  28.406,  26.670,  26.678,  25.617,  26.239,  26.242,  25.873,  25.634,  26.711,
  26.103,  20.004,  19.006,  15.585,  13.053,  12.009,  10.037,   9.985,   6.199,   5.413,
   4.139,   4.521,   5.000,   4.435,   5.257,   7.991,   8.844,   6.850,   4.757,   3.352,
   2.482,   2.021,   1.961,   1.852,   2.337,   2.616,   2.689,   2.623,   2.300,   2.048,
   1.885,   1.732,   1.781,   2.581,   2.206,   1.571,   1.150,   1.254,   0.280,  -1.186,
  -2.068,  -3.549,  -5.120,  -6.934,  -8.715, -10.833, -13.126, -15.024, -17.206, -19.018,
 -20.456, -22.491, -23.946, -25.401, -27.063, -28.869, -29.970, -29.256, -31.253, -34.191,
 -35.869, -37.697, -38.682, -39.574, -40.431, -41.212, -42.014, -42.022, -41.574, -41.244,
 -39.026, -39.847, -39.618, -38.858, -37.952, -37.434, -36.997, -37.149, -37.687, -36.609,
 -34.763, -31.469, -26.687, -23.834, -22.459, -19.473, -17.492, -16.613, -16.634, -13.844,
 -11.635, -10.974, -10.595, -10.106,  -8.183,  -7.244,  -6.193,  -1.866,   0.227,   3.121,
   3.609,   1.203,   1.837,   4.390,   7.982,   9.634,  12.246,  14.470,  16.969,  19.472,
  20.906,  22.792,  23.985,  26.335,  28.282,  31.177,  34.558,  38.071,  39.151,  40.265,
  42.150,  44.679,  45.007,  44.581,  41.829,  39.726,  39.574,  40.198,  40.903,  41.426,
  43.287,  44.844,  46.106,  46.676,  46.268,  45.836,  46.915,  49.156,  50.313,  51.418,
  51.436,  51.385,  51.955,  51.727,  51.279,  50.636,  50.079,  49.172,  48.524,  48.600,  49.055,
  50.516,  49.972,  47.367,  33.646,  31.799,  33.889,  33.530,  31.393,  27.727,  24.667,
  21.544,  19.580,  17.278,  15.192,  13.557,  11.849,  10.104,   8.993,   8.307,   7.306,
   6.045,   4.495,   3.099,   1.917,   1.242,   0.497,  -0.309,  -0.922,  -1.619,  -2.230,
  -3.174,  -4.098,  -3.316,  -4.610,  -5.687,  -6.766,  -7.687,  -8.873,  -9.910, -10.726,
 -11.411, -11.918, -12.292, -12.898, -13.087, -12.900, -12.889, -12.855, -12.722, -12.505,
 -11.891, -11.296, -11.005, -10.771, -10.252,  -9.892,  -9.306,  -8.941,  -8.337,  -7.783,
  -7.443,  -7.057,  -6.425,  -5.980,  -5.848,  -5.661,  -5.510,  -5.175,  -4.975,  -4.840,
  -4.709,  -4.559,  -4.106,  -3.725,  -3.353,  -2.948,  -2.677,  -2.046,  -1.855,  -1.979,
  -1.704,  -1.267,  -0.814,  -0.484,  -0.061,   0.134,   0.699,   1.042,   1.226,   1.490,
   1.981,   2.495,   3.506,   3.856,   4.479,   5.270,   6.409,   6.435,   7.759,   9.468,
   9.356,  11.115,  13.026,  14.531,  15.778,  17.436,  18.283,  15.638,  20.212,  34.258,
  40.663,  42.812,  41.330,  36.121,  29.347,  26.572,  25.487,  24.648,  22.900,  21.241,
  19.113,  16.812,  14.297,  10.572,   7.036,   4.877,   4.174,   4.848,   4.913,   3.854,
   2.013,  -0.106,  -2.465,  -5.754,  -9.589, -10.922, -10.520, -10.243, -11.067, -10.884,
 -10.984,  -9.815,  -8.647,  -7.286,  -6.019,  -5.346,  -4.689,  -4.191,  -4.093,  -3.689,
  -3.114,  -2.269,  -1.447,  -0.495,   0.706,   1.701,   2.631,   3.643,   4.610,   5.574,
   6.702,   7.808,   8.942,   9.986,  11.085,  12.291,  12.957,  13.179,  13.522,  13.727,
  13.970,  14.430,  15.027,  15.677,  16.338,  17.095,  17.922,  18.540,  19.195,  20.098},
{
  19.847,  20.632,  21.822,  22.729,  23.921,  25.756,  28.327,  25.604,  23.617,  23.915,
  23.915,  23.804,  25.429,  27.657,  30.384,  31.618,  33.277,  32.760,  29.299,  27.643,
  27.603,  27.020,  25.547,  24.401,  24.309,  23.783,  22.380,  21.264,  21.118,  19.442,
  20.967,  15.112,  15.017,  11.892,  10.874,  10.714,   8.347,   6.681,   5.670,   4.325,
   3.425,   3.119,   3.003,   1.986,   4.739,   8.033,   9.039,   6.852,   2.898,   1.268,
   0.768,   0.608,   0.730,   0.586,   0.594,   1.111,   1.101,   0.881,   0.383,   0.428,
   0.493,   0.482,   0.282,   0.025,  -0.229,  -0.369,  -1.022,  -1.856,  -2.981,  -3.664,
  -4.669,  -6.458,  -8.062,  -9.925, -11.937, -13.714, -15.990, -18.174, -20.050, -22.264,
 -23.636, -25.248, -26.940, -28.305, -29.806, -31.823, -32.651, -32.194, -33.138, -36.310,
 -38.290, -39.487, -40.540, -41.438, -42.100, -42.621, -42.990, -43.232, -43.068, -42.565,
 -41.687, -40.875, -39.940, -39.131, -37.902, -37.830, -37.076, -37.115, -35.356, -33.988,
 -33.560, -30.080, -24.930, -21.998, -21.131, -18.287, -15.962, -16.284, -15.268, -12.179,
  -8.796,  -7.332,  -6.328,  -6.998,  -6.319,  -3.864,  -2.897,  -2.565,   0.211,   2.449,
   3.100,   3.495,   4.739,   6.667,   9.524,  13.389,  16.394,  19.446,  21.232,  23.911,
  25.805,  26.588,  27.870,  28.757,  31.639,  34.457,  37.836,  41.330,  44.317,  45.601,
  46.178,  46.978,  46.761,  46.816,  42.892,  43.125,  41.463,  43.441,  44.192,  44.493,
  47.970,  47.006,  48.275,  48.548,  47.891,  47.574,  49.208,  50.686,  52.721,  53.586,
  53.278,  52.371,  53.101,  52.941,  52.423,  51.960,  51.166,  50.314,  49.394,  49.214,  49.760,
  51.343,  51.229,  49.680,  39.276,  28.172,  33.784,  33.650,  31.183,  27.596,  24.702,
  21.770,  19.419,  16.956,  15.218,  13.805,  12.010,  10.416,   9.914,   9.048,   7.906,
   6.588,   5.103,   3.588,   2.672,   1.863,   0.957,   0.044,  -1.096,  -2.219,  -2.586,
  -4.091,  -3.874,  -5.255,  -5.027,  -6.785,  -7.805,  -8.630,  -9.472, -10.169, -10.849,
 -11.422, -11.736, -12.124, -12.372, -12.734, -12.600, -12.769, -12.753, -12.297, -11.485,
 -11.155, -10.996, -10.718, -10.454,  -9.981,  -9.465,  -8.994,  -8.329,  -7.737,  -7.259,
  -6.714,  -6.232,  -5.692,  -5.428,  -5.213,  -5.219,  -4.990,  -4.610,  -4.651,  -4.597,
  -4.454,  -4.244,  -3.871,  -3.489,  -3.140,  -2.921,  -2.626,  -2.329,  -2.075,  -1.589,
  -1.215,  -0.938,  -0.716,  -0.382,  -0.098,   0.318,   0.572,   1.256,   1.436,   1.821,
   2.047,   2.758,   2.825,   3.449,   4.330,   4.600,   5.127,   7.747,   7.863,   8.409,
   9.508,  11.024,  12.093,  14.115,  16.565,  18.184,  18.863,  16.471,  18.319,  35.752,
  38.421,  41.184,  43.064,  38.966,  29.307,  26.651,  25.650,  25.294,  23.255,  21.055,
  19.083,  16.935,  13.305,   9.649,   6.088,   4.174,   2.808,   2.453,   3.017,   4.111,
   1.577,  -1.420,  -1.567,  -3.362,  -5.665, -10.188, -11.619, -11.278, -11.777, -11.301,
 -10.525,  -9.235,  -8.593,  -7.718,  -7.178,  -6.774,  -6.069,  -5.744,  -5.574,  -4.932,
  -4.598,  -3.913,  -3.119,  -2.194,  -1.029,   0.200,   1.187,   2.235,   3.267,   4.276,
   5.383,   6.572,   7.814,   9.059,  10.257,  11.541,  12.334,  12.572,  12.980,  13.302,
  13.535,  13.736,  14.219,  14.815,  15.639,  16.458,  17.342,  18.205,  19.094,  19.847},
{
  18.959,  20.144,  21.210,  22.350,  23.911,  26.050,  27.073,  25.050,  22.854,  22.905,
  22.918,  23.389,  25.204,  27.927,  29.205,  30.666,  32.511,  32.075,  29.226,  27.118,
  26.235,  25.663,  24.040,  22.929,  22.194,  20.625,  18.644,  16.690,  18.589,  18.437,
  15.193,  11.071,  12.455,   9.064,   8.282,   8.808,   5.921,   4.479,   3.695,   2.964,
   2.296,   1.585,   1.305,  -0.081,   4.932,   7.268,   7.655,   5.851,   2.331,  -0.748,
  -1.814,  -1.815,  -1.361,  -1.137,  -0.752,  -0.702,  -0.073,  -0.730,  -1.633,  -1.689,
  -1.626,  -1.484,  -1.384,  -1.872,  -2.259,  -2.692,  -3.406,  -4.278,  -5.259,  -6.092,
  -7.565,  -9.446, -11.358, -13.000, -15.041, -17.177, -19.136, -21.572, -23.341, -25.114,
 -27.100, -28.550, -29.833, -31.246, -32.989, -34.489, -35.260, -35.087, -36.351, -38.332,
 -40.597, -41.520, -42.584, -43.090, -43.217, -43.849, -44.277, -44.443, -44.599, -43.990,
 -42.964, -41.804, -40.712, -39.554, -38.660, -36.965, -37.099, -34.962, -32.379, -32.370,
 -33.132, -30.402, -24.986, -20.904, -18.891, -16.170, -12.847, -11.402, -10.279,  -7.521,
  -5.008,  -4.323,  -3.569,  -3.829,  -3.052,  -0.616,   0.290,   1.973,   3.366,   3.692,
   3.348,   4.508,   7.081,   9.880,  12.902,  17.001,  21.843,  24.286,  26.206,  28.380,
  31.368,  32.388,  31.903,  32.829,  34.710,  38.297,  41.954,  45.500,  47.382,  47.680,
  49.255,  49.060,  48.667,  47.857,  44.564,  44.463,  45.821,  46.622,  46.047,  46.941,
  47.780,  49.222,  50.605,  50.047,  49.735,  50.072,  51.495,  53.586,  55.335,  55.170,
  54.420,  53.156,  53.183,  53.607,  53.783,  53.376,  52.441,  51.246,  50.232,  49.628,  50.457,
  51.871,  51.573,  51.065,  44.552,  24.212,  31.900,  32.711,  30.775,  27.728,  24.721,
  22.188,  19.649,  17.218,  15.470,  14.182,  12.353,  11.134,  10.914,   9.712,   8.248,
   6.677,   5.609,   4.421,   3.387,   2.063,   1.481,   0.173,  -1.002,  -2.291,  -3.704,
  -4.041,  -3.800,  -5.311,  -6.407,  -7.570,  -8.390,  -8.859,  -9.396,  -9.973, -10.537,
 -11.113, -11.520, -11.999, -12.084, -12.792, -12.436, -12.323, -12.217, -11.818, -11.439,
 -10.959, -10.615, -10.346, -10.163,  -9.786,  -9.209,  -8.629,  -8.020,  -7.354,  -6.780,
  -6.231,  -5.682,  -5.204,  -4.812,  -4.604,  -4.492,  -4.366,  -4.184,  -4.216,  -4.102,
  -4.066,  -3.911,  -3.855,  -3.616,  -3.226,  -3.070,  -2.867,  -2.639,  -2.375,  -1.896,
  -1.518,  -1.373,  -0.968,  -0.696,  -0.253,   0.132,   0.468,   0.738,   1.039,   1.469,
   1.923,   2.291,   2.821,   3.274,   4.356,   4.476,   5.556,   6.233,   7.945,   8.094,
   9.457,  10.536,  12.434,  14.179,  15.758,  18.248,  18.943,  17.135,  16.955,  32.146,
  34.178,  39.763,  43.174,  43.192,  30.769,  26.556,  23.895,  23.942,  22.887,  20.271,
  18.424,  16.138,  12.679,   8.775,   5.461,   2.453,   0.820,  -0.996,  -0.006,  -0.131,
  -1.021,  -3.140,  -2.466,  -3.375,  -4.489,  -6.162,  -7.944,  -8.752, -10.339, -11.585,
 -10.548,  -9.655,  -9.200,  -8.565,  -7.912,  -7.355,  -6.780,  -6.399,  -6.235,  -5.970,
  -5.681,  -5.182,  -4.651,  -3.904,  -2.947,  -1.777,  -0.709,   0.410,   1.526,   2.677,
   4.046,   5.409,   6.688,   7.994,   9.268,  10.964,  11.756,  12.085,  12.578,  12.826,
  12.963,  13.242,  13.643,  13.993,  14.694,  15.611,  16.117,  16.712,  17.727,  18.959},
{
  18.146,  19.147,  20.463,  21.693,  22.839,  24.699,  25.850,  24.853,  23.789,  21.959,
  22.423,  22.994,  24.397,  27.802,  29.349,  30.016,  31.783,  32.401,  29.793,  26.503,
  24.964,  23.086,  21.820,  20.929,  19.435,  18.333,  17.472,  15.893,  15.084,  14.780,
  14.073,  11.525,   9.544,   5.901,   5.563,   5.544,   2.882,   1.428,   0.423,   0.689,
  -0.286,  -0.144,  -0.504,  -1.423,   3.333,   3.887,   6.443,   2.854,   1.018,  -2.888,
  -4.217,  -4.644,  -4.018,  -2.923,  -2.569,  -2.622,  -2.665,  -2.825,  -3.804,  -3.746,
  -3.977,  -3.763,  -3.805,  -3.851,  -4.206,  -5.103,  -6.013,  -6.806,  -7.693,  -8.833,
 -10.581, -12.647, -14.431, -16.837, -18.379, -20.654, -22.696, -24.942, -26.980, -28.400,
 -30.194, -31.903, -33.234, -34.617, -35.853, -36.997, -37.577, -37.736, -37.969, -40.835,
 -42.535, -43.418, -44.233, -44.781, -44.779, -44.864, -44.964, -45.838, -45.956, -45.324,
 -44.005, -42.723, -41.252, -39.315, -37.434, -36.991, -36.280, -33.784, -31.820, -31.487,
 -32.224, -30.748, -26.577, -21.314, -16.382, -13.090, -10.374,  -8.737,  -7.247,  -5.225,
  -5.120,  -2.025,  -0.489,   1.227,   1.609,   2.899,   3.561,   6.187,   8.179,   9.129,
  11.401,  12.905,  15.490,  19.153,  21.742,  23.980,  26.481,  28.010,  30.388,  33.129,
  35.923,  37.448,  37.240,  37.308,  39.102,  42.290,  46.097,  48.824,  49.758,  50.330,
  51.466,  50.727,  50.934,  50.240,  50.385,  48.001,  48.459,  48.959,  48.139,  49.883,
  50.503,  51.515,  52.028,  52.201,  51.825,  52.102,  53.469,  56.917,  58.061,  56.550,
  51.606,  48.799,  49.988,  49.514,  51.521,  53.343,  53.588,  52.011,  50.847,  50.213,  50.481,
  52.193,  51.809,  52.145,  51.026,  29.268,  27.751,  31.212,  30.335,  27.858,  25.010,
  22.643,  19.899,  17.618,  15.699,  14.142,  12.618,  11.797,  11.302,   9.943,   8.969,
   7.535,   6.027,   4.626,   3.480,   2.503,   1.679,   0.837,  -1.156,  -0.807,  -3.059,
  -3.876,  -5.403,  -6.644,  -7.222,  -7.628,  -8.141,  -8.753,  -9.034,  -9.689, -10.382,
 -10.930, -11.320, -11.104, -11.896, -10.647, -12.148, -11.780, -11.428, -11.402, -11.167,
 -10.927, -10.789, -10.358, -10.149,  -9.726,  -9.135,  -8.365,  -7.820,  -7.140,  -6.431,
  -5.825,  -5.205,  -4.754,  -4.383,  -4.007,  -3.796,  -3.873,  -3.765,  -3.732,  -3.661,
  -3.553,  -3.610,  -3.575,  -3.500,  -3.284,  -3.200,  -3.107,  -2.949,  -2.823,  -2.429,
  -2.147,  -2.079,  -1.712,  -1.346,  -0.877,  -0.623,  -0.102,   0.212,   0.335,   0.740,
   1.242,   1.768,   2.335,   2.837,   3.598,   4.448,   5.628,   7.024,   7.101,   8.015,
   8.719,  10.140,  11.684,  14.100,  15.874,  17.622,  19.825,  18.414,  15.033,  33.283,
  36.310,  40.278,  43.769,  43.434,  37.253,  24.723,  23.543,  22.578,  21.079,  19.473,
  17.817,  15.907,  13.017,   9.090,   4.753,  -0.060,  -1.792,  -2.656,  -4.412,  -5.005,
  -4.808,  -5.195,  -3.986,  -2.179,  -3.181,  -4.054,  -5.636,  -5.855,  -6.358, -10.598,
 -10.804,  -9.718,  -9.479,  -8.506,  -7.885,  -7.532,  -7.332,  -6.520,  -6.383,  -6.340,
  -6.211,  -5.978,  -5.764,  -5.200,  -4.260,  -3.334,  -2.201,  -1.120,   0.068,   1.465,
   2.868,   4.179,   5.402,   6.730,   8.105,   9.528,  11.039,  12.067,  12.462,  12.558,
  12.622,  12.716,  13.074,  13.602,  14.203,  15.015,  16.005,  16.768,  17.493,  18.146},
{
  17.336,  18.136,  19.201,  20.519,  21.555,  22.533,  24.353,  24.518,  23.733,  22.510,
  22.387,  23.014,  24.348,  26.778,  27.831,  29.281,  29.821,  30.653,  28.795,  26.912,
  24.947,  21.256,  19.297,  19.257,  17.732,  16.316,  15.550,  14.360,  12.553,  10.968,
  10.393,  10.160,   7.559,   3.615,   3.230,   1.769,  -0.504,  -1.510,  -2.502,  -2.316,
  -2.123,  -2.885,  -3.026,  -1.672,   0.085,   0.084,   3.654,   3.336,  -0.252,  -5.310,
  -6.696,  -6.665,  -6.530,  -5.782,  -5.069,  -3.848,  -4.023,  -5.985,  -5.856,  -5.343,
  -5.459,  -6.006,  -6.084,  -6.363,  -6.762,  -7.340,  -8.099,  -9.102, -10.322, -11.673,
 -13.717, -15.912, -17.892, -19.852, -22.098, -24.156, -26.447, -28.326, -30.466, -31.840,
 -33.493, -35.136, -36.736, -38.051, -38.654, -39.330, -40.199, -39.941, -39.849, -43.515,
 -44.553, -45.113, -46.042, -46.381, -46.285, -45.881, -45.795, -46.593, -47.192, -46.527,
 -45.070, -43.562, -41.748, -39.536, -34.946, -34.002, -35.739, -32.976, -31.114, -30.994,
 -30.688, -28.850, -26.551, -22.105, -15.200, -11.878,  -8.570,  -7.477,  -7.081,  -5.963,
  -2.321,   0.026,   3.265,   4.890,   4.504,   5.427,   6.698,  10.091,  12.527,  14.226,
  14.795,  15.585,  18.529,  21.349,  25.114,  27.253,  29.720,  31.402,  33.575,  37.267,
  40.445,  40.986,  40.757,  41.656,  43.323,  45.796,  49.123,  51.405,  52.111,  53.604,
  53.438,  52.929,  53.755,  53.183,  51.909,  50.751,  51.076,  51.893,  51.271,  53.501,
  52.215,  53.604,  53.958,  53.943,  52.540,  52.615,  59.328,  60.329,  58.150,  54.393,
  51.364,  60.203,  60.852,  60.072,  58.288,  56.648,  54.195,  53.228,  51.217,  50.128,  50.471,
  52.094,  52.155,  52.918,  52.418,  43.869,  25.581,  29.865,  30.265,  28.108,  25.567,
  22.887,  20.548,  17.890,  15.442,  13.642,  12.550,  12.162,  11.260,  10.848,   9.952,
   8.564,   7.560,   4.861,   4.040,   3.566,   2.056,   0.990,  -0.401,  -1.186,  -2.615,
  -3.717,  -5.149,  -6.160,  -6.381,  -7.068,  -7.747,  -8.585,  -9.002,  -9.352, -10.142,
  -9.512, -10.462, -10.904, -10.714, -10.935, -11.157, -11.456, -11.367, -11.307, -11.210,
 -11.107, -11.035, -10.758, -10.302,  -9.746,  -9.149,  -8.356,  -7.476,  -6.757,  -6.025,
  -5.357,  -4.794,  -4.255,  -3.872,  -3.571,  -3.278,  -3.239,  -3.228,  -3.120,  -3.088,
  -3.039,  -3.200,  -3.266,  -3.181,  -3.087,  -3.129,  -3.174,  -3.104,  -3.089,  -2.924,
  -2.695,  -2.555,  -2.294,  -2.004,  -1.566,  -1.140,  -0.772,  -0.571,  -0.236,   0.186,
   0.692,   1.226,   1.693,   2.248,   3.110,   4.030,   5.742,   6.463,   7.039,   8.413,
   8.768,   9.799,  11.325,  13.420,  15.699,  17.240,  19.188,  19.501,  14.598,  33.308,
  36.830,  43.256,  43.572,  42.001,  38.888,  27.491,  23.224,  21.411,  20.519,  20.184,
  17.745,  15.544,  14.108,   9.253,   4.336,  -0.562,  -2.844,  -4.397,  -5.691,  -6.197,
  -6.506,  -6.299,  -5.118,  -2.832,  -2.733,  -3.886,  -4.632,  -5.222,  -6.042,  -7.652,
 -10.395, -10.023,  -9.643,  -9.523,  -8.704,  -7.985,  -7.165,  -6.425,  -6.064,  -6.255,
  -6.356,  -6.418,  -6.269,  -5.955,  -5.244,  -4.578,  -3.488,  -2.190,  -0.855,   0.515,
   1.837,   3.115,   4.474,   6.070,   7.569,   9.108,  10.458,  11.414,  12.174,  12.290,
  12.527,  12.631,  12.848,  13.301,  13.775,  14.446,  15.288,  15.902,  16.872,  17.336},
{
  16.808,  17.373,  18.137,  19.023,  20.361,  21.972,  22.283,  22.725,  23.227,  22.956,
  22.403,  23.088,  24.512,  25.866,  25.377,  26.991,  28.852,  28.287,  28.643,  27.515,
  24.054,  20.509,  17.666,  17.636,  16.229,  14.651,  14.109,  13.348,  12.236,  10.291,
   7.793,   4.655,   4.251,   3.179,   0.277,  -1.764,  -4.500,  -5.450,  -4.836,  -6.218,
  -6.924,  -6.321,  -6.299,  -4.853,  -3.651,  -3.033,   0.123,   2.588,  -0.880,  -6.095,
  -9.080,  -9.197,  -8.774,  -8.952,  -8.144,  -1.082,  -2.046,  -9.648,  -9.499,  -8.218,
  -7.445,  -7.599,  -8.413,  -8.647,  -9.062,  -9.798, -10.716, -11.652, -13.052, -14.796,
 -16.964, -19.309, -21.577, -23.260, -25.779, -27.634, -29.931, -31.806, -33.549, -35.463,
 -36.952, -38.508, -40.361, -41.484, -42.147, -42.350, -42.810, -42.582, -43.071, -45.955,
 -46.188, -46.229, -47.030, -47.666, -47.488, -47.053, -46.599, -47.162, -47.821, -47.231,
 -45.597, -44.312, -42.160, -39.479, -36.646, -35.348, -34.874, -32.964, -31.257, -29.602,
 -28.152, -26.905, -23.944, -21.917, -17.952, -11.547,  -8.551,  -7.742,  -6.175,  -3.294,
   1.090,   3.713,   6.408,   6.992,   7.617,   9.212,  11.480,  13.878,  15.930,  17.485,
  20.232,  22.265,  23.856,  25.896,  27.975,  30.792,  31.859,  33.828,  36.273,  40.227,
  43.896,  43.649,  43.407,  44.356,  46.689,  49.852,  52.548,  53.939,  55.042,  55.296,
  55.637,  55.249,  56.034,  55.582,  52.765,  52.141,  53.862,  53.633,  54.063,  57.243,
  53.925,  57.093,  56.855,  54.270,  54.235,  62.224,  60.190,  60.235,  57.145,  49.570,
  61.519,  63.779,  63.756,  62.715,  61.742,  60.111,  58.987,  53.961,  53.361,  50.619,  50.179,
  51.775,  51.999,  52.593,  52.880,  52.254,  35.159,  26.933,  29.614,  28.229,  25.861,
  22.938,  20.290,  17.501,  15.548,  14.107,  13.518,  12.284,  11.710,  10.909,  11.038,
   8.745,   6.819,   6.303,   4.910,   3.836,   2.302,   1.052,   0.425,  -0.416,  -2.156,
  -3.003,  -4.114,  -5.325,  -6.197,  -6.987,  -7.802,  -8.004,  -8.887,  -9.377,  -9.973,
 -10.042, -10.214, -10.447, -11.191, -11.439, -11.473, -11.771, -11.753, -11.723, -11.651,
 -11.515, -11.207, -10.677, -10.107,  -9.540,  -8.716,  -7.844,  -7.013,  -6.212,  -5.664,
  -5.069,  -4.479,  -3.954,  -3.586,  -3.246,  -3.004,  -2.805,  -2.682,  -2.619,  -2.667,
  -2.770,  -2.815,  -3.052,  -3.075,  -3.118,  -3.239,  -3.359,  -3.433,  -3.277,  -3.381,
  -3.281,  -3.174,  -2.950,  -2.777,  -2.312,  -1.880,  -1.614,  -1.435,  -0.996,  -0.684,
   0.000,   0.393,   0.945,   1.385,   2.160,   3.019,   4.452,   6.289,   7.684,   8.241,
   9.073,  10.489,  12.055,  13.875,  15.778,  17.950,  19.572,  20.040,  15.553,  31.616,
  37.887,  42.778,  42.392,  42.917,  39.661,  29.388,  22.405,  22.815,  20.963,  20.900,
  18.210,  15.883,  13.861,   9.141,   4.241,  -0.721,  -3.977,  -5.797,  -6.880,  -7.124,
  -7.122,  -7.063,  -5.575,  -5.439,  -4.289,  -3.804,  -5.215,  -5.796,  -5.981,  -7.925,
 -11.423,  -8.858,  -9.817,  -8.103,  -5.676,  -5.323,  -6.386,  -4.877,  -6.007,  -6.217,
  -5.586,  -6.616,  -6.331,  -6.219,  -5.868,  -4.990,  -4.124,  -3.037,  -1.699,  -0.281,
   1.222,   2.774,   4.177,   5.590,   7.088,   8.394,   9.857,  11.681,  12.382,  12.477,
  12.642,  12.819,  13.085,  13.223,  13.594,  14.171,  14.773,  15.620,  16.112,  16.808},
{
  16.339,  16.785,  16.982,  17.573,  18.372,  19.429,  20.479,  20.790,  21.197,  22.698,
  23.065,  23.512,  25.081,  24.724,  25.459,  25.870,  26.204,  26.372,  25.099,  23.455,
  21.570,  18.527,  16.393,  15.281,  13.829,  12.508,  11.134,  10.621,  10.384,   9.964,
   7.253,   4.730,   0.996,   1.165,  -1.184,  -4.036,  -4.025,  -9.344, -11.044,  -9.384,
 -10.480,  -9.610,  -9.408,  -7.836,  -6.700,  -6.410,  -2.206,   1.160,  -1.203,  -6.641,
 -11.206, -11.251, -11.308, -11.171, -10.702, -10.763, -11.689, -12.423,  -6.875, -12.313,
 -10.458, -10.952, -11.214, -10.837, -11.834, -12.470, -13.281, -14.697, -16.483, -18.361,
 -20.596, -23.174, -25.322, -27.280, -29.422, -31.600, -33.478, -35.432, -36.989, -38.774,
 -40.365, -41.870, -43.982, -44.970, -45.844, -46.734, -46.238, -44.813, -44.851, -47.915,
 -48.157, -48.077, -47.654, -48.362, -48.712, -48.332, -47.849, -47.605, -47.748, -47.212,
 -46.072, -43.960, -42.187, -39.925, -37.187, -35.113, -33.837, -32.297, -30.054, -27.664,
 -25.772, -23.887, -22.191, -18.734, -17.091, -12.406,  -7.216,  -6.219,  -2.822,   0.652,
   4.407,   8.487,   9.455,   9.698,  11.171,  13.550,  15.531,  17.720,  20.568,  22.993,
  24.383,  24.230,  24.593,  27.511,  31.048,  33.081,  34.319,  37.018,  39.085,  43.208,
  45.995,  45.264,  45.481,  48.264,  52.184,  55.249,  56.329,  56.609,  56.504,  57.346,
  57.400,  57.341,  57.519,  56.362,  54.242,  54.883,  55.290,  54.965,  55.727,  57.629,
  56.661,  57.682,  55.889,  55.718,  60.094,  58.949,  61.083,  58.107,  51.310,  54.715,
  63.960,  64.892,  64.829,  64.091,  62.703,  61.112,  59.104,  56.798,  52.268,  50.929,  49.744,
  50.906,  51.396,  51.484,  52.089,  51.967,  43.059,  24.113,  28.978,  28.912,  26.208,
  23.238,  21.093,  17.841,  15.811,  14.545,  13.131,  12.307,  11.269,  10.228,   9.328,
   8.802,   9.629,   6.824,   5.211,   4.313,   2.792,   1.867,   0.790,  -0.106,  -0.868,
  -1.952,  -3.350,  -4.236,  -5.851,  -5.771,  -7.460,  -7.910,  -8.333,  -9.060,  -8.976,
  -9.855, -10.002, -10.479, -11.054, -11.430, -11.413, -11.638, -11.529, -11.651, -11.405,
 -10.967, -10.675, -10.318,  -9.861,  -9.258,  -8.451,  -7.785,  -6.964,  -6.235,  -5.528,
  -4.856,  -4.202,  -3.692,  -3.223,  -2.961,  -2.703,  -2.404,  -2.238,  -2.245,  -2.311,
  -2.525,  -2.742,  -2.907,  -3.047,  -3.197,  -3.424,  -3.597,  -3.901,  -3.920,  -4.009,
  -3.981,  -3.966,  -3.860,  -3.690,  -3.228,  -2.878,  -2.580,  -2.473,  -2.036,  -1.467,
  -1.017,  -0.401,   0.081,   0.638,   1.306,   2.341,   3.457,   5.230,   6.876,   8.162,
   9.374,  10.897,  12.501,  14.635,  16.493,  18.283,  20.196,  18.781,  15.762,  33.577,
  41.781,  43.204,  43.169,  43.597,  41.768,  29.757,  23.164,  23.578,  22.116,  19.708,
  17.703,  16.492,  13.895,   8.976,   3.829,  -1.083,  -5.400,  -6.555,  -8.022,  -8.115,
  -8.053,  -9.089,  -7.870,  -8.122,  -6.951,  -6.280,  -9.677,  -6.315,  -5.628,  -6.679,
 -10.129,  -7.230,  -8.906,  -7.185,  -8.310,  -7.302,  -5.884,  -5.366,  -5.710,  -6.023,
  -5.751,  -6.681,  -6.790,  -6.690,  -6.362,  -5.648,  -4.755,  -3.488,  -2.108,  -0.627,
   0.972,   2.578,   4.206,   5.627,   6.886,   8.246,   9.820,  11.568,  12.038,  12.176,
  12.374,  12.620,  12.950,  13.141,  13.478,  13.981,  14.688,  15.415,  16.132,  16.339},
{
  16.371,  16.271,  16.273,  16.519,  16.821,  17.607,  18.566,  19.248,  19.255,  19.039,
  19.606,  22.296,  24.463,  24.367,  25.660,  24.663,  22.785,  22.204,  21.969,  20.878,
  19.539,  16.614,  13.942,  12.899,  12.140,  11.189,   9.582,   7.747,   5.764,   5.686,
   5.958,   5.589,   1.758,  -1.619,  -4.876,  -6.365,  -6.721,  -7.249, -11.054, -12.082,
 -12.939, -12.141, -11.340, -10.239,  -9.556,  -8.221,  -2.871,  -1.018,  -1.914,  -6.863,
 -12.877, -13.315, -13.570, -13.405, -13.340, -13.736, -14.986, -12.322,  -9.636, -12.753,
 -13.063, -14.080, -14.763, -15.527, -15.071, -15.599, -16.593, -18.243, -20.005, -22.083,
 -24.515, -27.167, -29.302, -31.321, -33.611, -35.561, -37.169, -39.013, -40.781, -42.040,
 -43.723, -45.170, -47.381, -48.575, -48.810, -50.051, -51.142, -48.436, -46.933, -50.584,
 -50.548, -50.698, -49.895, -48.777, -50.001, -49.157, -48.804, -47.652, -47.303, -46.404,
 -45.381, -43.534, -41.836, -39.641, -37.349, -35.029, -33.098, -31.151, -29.096, -26.757,
 -23.693, -22.174, -20.465, -16.581, -14.534, -13.835,  -7.309,  -3.547,  -0.304,   3.647,
   7.754,  10.297,  11.422,  13.432,  15.869,  17.818,  19.703,  22.142,  23.501,  26.724,
  28.525,  28.364,  28.908,  31.394,  34.295,  35.676,  37.378,  39.660,  41.951,  44.920,
  47.288,  47.750,  48.740,  52.441,  55.547,  58.179,  58.522,  58.355,  57.766,  58.048,
  58.049,  58.214,  59.017,  57.365,  56.859,  57.380,  57.501,  57.306,  56.055,  58.414,
  57.835,  56.989,  57.415,  59.003,  59.103,  60.794,  59.957,  57.806,  53.331,  63.933,
  66.054,  65.852,  65.376,  64.237,  62.893,  61.137,  59.267,  57.085,  55.492,  51.144,  50.512,
  50.300,  50.604,  50.602,  50.614,  49.716,  46.825,  25.508,  29.178,  29.072,  28.909,
  24.748,  20.970,  17.860,  15.822,  13.548,  12.427,  11.637,  11.046,  10.461,  10.440,
   9.297,   7.585,   6.163,   5.086,   4.297,   3.372,   2.195,   2.350,   1.695,   0.559,
  -0.443,  -2.136,  -3.484,  -4.609,  -5.973,  -6.293,  -6.910,  -8.008,  -8.378,  -9.100,
  -9.248, -10.066, -10.145, -10.320, -10.638, -10.960, -11.383, -11.332, -11.180, -11.067,
 -10.599, -10.433, -10.026,  -9.350,  -8.751,  -8.197,  -7.486,  -6.783,  -6.090,  -5.390,
  -4.828,  -4.254,  -3.628,  -3.082,  -2.729,  -2.427,  -2.132,  -2.009,  -2.034,  -2.184,
  -2.405,  -2.583,  -2.754,  -2.972,  -3.302,  -3.589,  -3.964,  -4.135,  -4.430,  -4.517,
  -4.772,  -4.781,  -4.597,  -4.555,  -4.092,  -3.796,  -3.493,  -3.304,  -2.839,  -2.325,
  -1.980,  -1.443,  -1.026,  -0.079,   0.510,   1.549,   2.880,   4.431,   6.146,   7.802,
   9.884,  11.762,  13.378,  15.338,  17.218,  18.825,  20.589,  13.866,  20.387,  35.431,
  43.511,  43.168,  44.059,  43.292,  39.149,  30.252,  21.987,  23.056,  21.506,  18.897,
  17.332,  15.735,  12.488,   8.064,   3.621,  -1.502,  -4.946,  -6.569,  -8.415,  -8.328,
 -10.772, -11.087,  -9.134,  -9.211, -10.452,  -9.680, -11.310, -10.842,  -8.638,  -7.658,
  -6.615,  -7.677, -10.926,  -9.321,  -8.307,  -7.249,  -6.043,  -5.488,  -5.402,  -5.422,
  -5.694,  -6.387,  -6.928,  -7.171,  -7.021,  -6.242,  -4.913,  -3.403,  -1.870,  -0.167,
   1.361,   3.183,   4.839,   6.030,   7.128,   8.159,   9.747,  11.159,  11.456,  12.143,
  12.420,  12.379,  12.445,  12.682,  12.835,  13.354,  14.167,  15.247,  15.979,  16.371},
{
  15.504,  15.632,  15.655,  15.752,  15.775,  16.271,  17.106,  17.555,  17.480,  17.490,
  18.388,  20.461,  22.243,  24.619,  25.577,  23.765,  21.058,  19.766,  19.177,  18.494,
  17.272,  14.188,  12.351,  11.103,   9.965,   9.140,   7.587,   4.059,   3.254,   2.002,
   3.090,   4.034,   1.971,  -3.881,  -7.920,  -9.594, -10.140,  -9.252, -10.677, -13.348,
 -14.381, -14.986, -14.012, -13.656, -11.028,  -9.514,  -7.154,  -4.768,  -5.049,  -7.032,
 -13.381, -15.304, -15.820, -15.659, -15.781, -15.577, -16.086, -16.888, -16.768, -16.312,
 -16.494, -16.490, -17.296, -18.318, -19.282, -19.575, -20.533, -21.950, -23.704, -26.224,
 -28.970, -31.127, -33.414, -35.746, -38.203, -39.610, -40.746, -42.361, -43.697, -45.475,
 -46.938, -48.543, -50.202, -51.730, -52.921, -52.375, -53.684, -51.958, -48.115, -52.990,
 -52.858, -52.204, -52.137, -50.529, -49.997, -50.793, -49.330, -48.005, -46.571, -45.314,
 -43.794, -42.569, -41.337, -38.824, -36.320, -34.088, -31.923, -29.724, -27.308, -25.100,
 -22.000, -19.384, -17.100, -14.122, -11.650, -10.015,  -7.376,  -5.022,   0.293,   5.690,
   8.468,  11.538,  14.178,  17.480,  19.787,  21.834,  24.076,  27.261,  28.264,  28.818,
  31.876,  32.094,  32.961,  35.039,  36.909,  39.161,  41.025,  43.502,  45.454,  47.610,
  49.649,  49.929,  51.451,  53.834,  56.731,  59.535,  59.662,  59.050,  57.734,  58.738,
  60.027,  60.055,  59.729,  59.320,  59.387,  60.218,  59.940,  58.911,  57.219,  59.120,
  58.489,  57.862,  58.567,  61.259,  59.724,  59.835,  60.211,  57.478,  57.668,  65.307,
  66.746,  66.135,  65.241,  64.084,  62.722,  60.924,  59.348,  57.206,  56.348,  52.530,  50.025,
  50.569,  49.610,  49.104,  48.947,  48.360,  46.773,  33.445,  25.923,  29.433,  27.200,
  24.913,  21.162,  17.657,  15.684,  14.237,  13.297,  13.415,  11.498,  10.379,   8.893,
   7.844,   7.428,   6.795,   5.708,   4.459,   4.559,   3.446,   2.744,   2.971,   3.471,
   3.440,  -0.141,  -2.627,  -3.765,  -4.487,  -5.577,  -6.539,  -6.420,  -7.264,  -8.431,
  -8.711,  -9.053,  -9.655, -10.298, -10.855, -11.191, -11.223, -11.127, -10.878, -10.676,
 -10.586, -10.369,  -9.967,  -9.508,  -8.932,  -8.133,  -7.436,  -6.723,  -6.166,  -5.523,
  -4.876,  -4.259,  -3.641,  -3.108,  -2.672,  -2.354,  -2.066,  -2.049,  -2.051,  -2.201,
  -2.359,  -2.662,  -2.876,  -3.213,  -3.475,  -3.712,  -4.175,  -4.564,  -5.035,  -5.178,
  -5.298,  -5.429,  -5.404,  -5.264,  -4.970,  -4.811,  -4.479,  -4.248,  -3.749,  -3.318,
  -2.866,  -2.267,  -1.795,  -1.091,  -0.290,   0.833,   2.202,   3.706,   5.443,   7.817,
  10.305,  12.053,  14.353,  15.880,  16.978,  17.399,  13.443,  18.076,  27.481,  38.298,
  44.400,  44.311,  45.042,  41.421,  37.528,  30.405,  21.613,  21.602,  21.856,  20.660,
  15.891,  13.272,   9.925,   5.950,   2.030,  -1.105,  -3.386,  -6.106,  -7.943,  -9.057,
 -13.258, -13.172, -11.478, -11.608, -12.102, -10.965, -10.541, -10.134, -10.133,  -9.021,
  -7.560,  -5.901, -10.773,  -6.678,  -8.885,  -7.851,  -6.980,  -6.378,  -5.933,  -5.680,
  -6.158,  -6.755,  -7.267,  -7.372,  -6.968,  -6.097,  -4.673,  -2.612,  -1.428,  -0.182,
   1.611,   3.415,   4.839,   6.132,   7.192,   8.233,   9.529,  10.247,  10.884,  11.598,
  12.261,  12.728,  12.692,  12.698,  12.769,  13.052,  13.610,  14.673,  15.094,  15.504},
{
  14.713,  14.847,  14.998,  15.071,  15.094,  15.167,  15.736,  15.881,  16.093,  16.650,
  17.649,  21.282,  21.840,  23.382,  23.555,  22.282,  20.689,  19.091,  17.835,  16.505,
  15.293,  12.736,  11.133,   9.988,   8.741,   7.106,   7.024,   6.725,  -0.041,  -0.522,
   1.048,   0.528,  -3.298,  -8.901, -10.708, -10.772, -11.451, -11.743, -12.871, -12.770,
 -14.881, -18.452, -16.639, -15.893, -14.223, -12.145, -13.120, -11.276,  -8.427,  -9.143,
 -12.770, -17.019, -17.478, -17.516, -17.828, -17.367, -17.500, -18.765, -19.287, -16.871,
 -17.437, -20.040, -20.750, -22.023, -23.061, -23.803, -24.431, -25.875, -27.691, -30.315,
 -33.042, -35.284, -37.874, -40.114, -42.181, -44.051, -44.987, -46.062, -47.314, -49.028,
 -50.319, -51.754, -53.864, -54.921, -55.990, -56.620, -55.878, -54.424, -50.991, -54.885,
 -54.458, -54.002, -52.747, -52.062, -49.333, -51.338, -48.882, -47.686, -45.606, -44.186,
 -42.131, -41.829, -39.804, -37.459, -34.845, -32.734, -30.561, -27.835, -25.046, -22.494,
 -20.096, -17.263, -14.337, -10.313,  -8.686,  -5.705,  -6.035,  -3.526,  -0.496,   7.180,
  12.246,  15.765,  19.355,  21.931,  24.294,  25.781,  28.757,  31.443,  32.873,  30.889,
  33.408,  34.525,  36.472,  37.980,  39.131,  42.588,  44.623,  45.402,  47.531,  50.190,
  51.509,  52.312,  53.922,  56.827,  58.534,  61.114,  61.834,  59.333,  60.471,  61.140,
  62.759,  62.576,  62.952,  61.311,  61.252,  61.200,  61.477,  60.275,  60.149,  59.546,
  59.417,  58.446,  60.000,  58.712,  59.826,  60.199,  60.131,  57.879,  61.996,  66.438,
  66.409,  65.875,  64.737,  63.380,  61.951,  60.563,  58.923,  57.687,  55.312,  54.244,  51.434,
  50.438,  48.535,  47.759,  47.067,  46.343,  45.278,  36.399,  23.597,  28.365,  27.420,
  23.869,  20.172,  17.385,  15.697,  14.117,  12.081,  11.675,  11.322,   9.954,   9.351,
   8.056,   7.227,   6.681,   6.217,   5.748,   5.393,   4.397,   5.364,   5.470,   3.024,
   1.394,  -0.563,  -2.875,  -2.426,  -3.148,  -3.275,  -4.236,  -5.913,  -7.009,  -7.318,
  -7.720,  -8.875,  -9.924, -10.770, -11.086, -11.019, -10.924, -10.991, -10.687, -10.351,
 -10.008,  -9.747,  -9.312,  -8.895,  -8.375,  -7.782,  -7.334,  -6.819,  -6.216,  -5.628,
  -5.063,  -4.499,  -3.891,  -3.388,  -2.886,  -2.570,  -2.315,  -2.264,  -2.282,  -2.409,
  -2.580,  -2.694,  -3.103,  -3.441,  -3.452,  -4.170,  -4.473,  -4.881,  -5.490,  -5.682,
  -5.912,  -5.997,  -5.901,  -5.835,  -5.808,  -5.650,  -5.496,  -5.255,  -4.752,  -4.382,
  -3.687,  -3.078,  -2.548,  -1.903,  -1.086,  -0.124,   1.353,   3.019,   5.010,   7.324,
  10.080,  13.024,  15.236,  15.779,  14.125,  11.838,  18.983,  30.436,  39.357,  44.503,
  44.558,  44.382,  42.720,  36.123,  24.885,  21.083,  22.600,  22.273,  21.642,  19.098,
  14.393,  10.411,   7.125,   3.976,   0.824,  -1.288,  -3.311,  -4.485,  -6.513,  -9.112,
 -11.781, -12.496, -12.859, -14.624, -14.588, -12.365, -12.227, -11.396, -10.227,  -9.264,
  -8.280, -12.013, -10.103,  -9.905,  -8.986,  -8.361,  -7.628,  -7.022,  -6.677,  -6.479,
  -6.800,  -7.160,  -7.584,  -7.592,  -7.121,  -6.002,  -4.193,  -2.340,  -1.156,   0.115,
   1.952,   3.707,   5.018,   6.251,   7.417,   8.706,   9.240,  10.134,  10.590,  10.996,
  11.616,  12.280,  12.683,  12.762,  13.048,  13.017,  13.336,  14.044,  14.507,  14.713},
{
  14.061,  14.130,  14.191,  14.152,  14.295,  14.390,  14.633,  14.741,  14.877,  15.258,
  16.277,  19.670,  22.758,  23.857,  23.975,  21.637,  20.169,  18.425,  17.055,  15.583,
  14.025,  11.980,  10.099,   8.338,   7.307,   6.684,   4.659,   3.043,   1.663,  -2.655,
  -5.203,  -9.399,  -8.223, -10.352, -11.831, -10.537, -10.280, -13.682, -15.499, -15.734,
 -15.452, -21.550, -19.694, -21.432, -19.413, -17.883, -17.663, -16.658, -12.921, -10.283,
 -14.040, -17.586, -19.352, -19.722, -19.914, -19.386, -19.942, -20.703, -21.890, -20.462,
 -19.341, -21.256, -24.594, -25.701, -26.561, -27.052, -28.243, -29.485, -31.793, -33.873,
 -36.947, -39.509, -42.311, -44.742, -46.389, -47.815, -49.526, -50.089, -51.290, -52.292,
 -53.674, -54.814, -56.627, -58.090, -58.782, -59.162, -58.227, -56.702, -54.180, -57.494,
 -56.264, -55.987, -54.070, -52.880, -51.527, -49.641, -48.771, -46.252, -44.754, -43.074,
 -40.499, -39.761, -37.299, -35.423, -33.229, -30.743, -28.185, -25.246, -21.993, -19.437,
 -17.512, -14.808, -11.377,  -8.927,  -7.834,  -5.653,  -3.300,  -0.966,   2.136,   8.077,
  13.906,  19.845,  22.443,  24.825,  27.032,  29.174,  33.042,  34.158,  34.618,  33.818,
  35.033,  36.803,  39.050,  40.693,  42.810,  45.405,  47.392,  48.975,  49.779,  50.856,
  52.457,  53.700,  55.938,  58.842,  61.832,  62.777,  62.054,  61.043,  62.880,  63.270,
  64.178,  63.902,  63.127,  60.686,  61.415,  61.294,  61.505,  62.753,  62.188,  61.939,
  61.177,  60.561,  60.433,  60.077,  58.613,  61.077,  59.757,  62.155,  64.424,  65.827,
  65.997,  65.148,  63.752,  62.481,  60.873,  59.468,  57.699,  55.978,  53.969,  52.632,  51.589,
  49.163,  47.513,  46.284,  45.111,  44.143,  42.253,  34.319,  22.275,  28.428,  26.158,
  22.749,  19.884,  17.031,  15.720,  13.654,  12.392,  11.584,  10.670,  10.190,   9.184,
   8.370,   7.285,   6.688,   6.647,   6.290,   5.756,   4.833,   5.086,   3.544,   1.975,
   0.579,  -1.269,   0.800,  -0.107,  -1.774,  -4.208,  -4.772,  -5.139,  -5.346,  -6.302,
  -7.910,  -8.861,  -9.795, -10.488, -10.627, -10.655, -10.551, -10.320,  -9.898,  -9.639,
  -9.599,  -9.421,  -9.110,  -8.826,  -8.412,  -7.980,  -7.517,  -6.945,  -6.516,  -5.998,
  -5.509,  -4.880,  -4.232,  -3.750,  -3.315,  -2.982,  -2.808,  -2.701,  -2.773,  -2.848,
  -3.060,  -3.204,  -3.327,  -3.573,  -4.580,  -4.978,  -5.145,  -5.406,  -5.601,  -5.888,
  -6.046,  -6.142,  -6.283,  -6.463,  -6.397,  -6.435,  -6.350,  -6.180,  -5.642,  -5.206,
  -4.499,  -3.999,  -3.405,  -2.825,  -2.078,  -0.974,   0.384,   2.387,   4.526,   6.701,
   9.750,  12.950,  14.757,  12.678,  15.241,  30.432,  38.486,  42.364,  45.051,  45.456,
  45.605,  41.000,  31.677,  24.888,  22.161,  22.306,  23.897,  22.418,  21.071,  17.031,
  13.262,   8.639,   4.912,   1.893,  -0.181,  -2.209,  -4.284,  -4.937,  -6.078,  -8.244,
 -10.629, -12.625, -13.222, -13.887, -14.055, -12.864, -11.957, -10.803,  -9.453,  -9.472,
  -8.477,  -9.334, -13.978, -11.315,  -9.136,  -8.428,  -8.129,  -7.578,  -7.199,  -6.855,
  -6.935,  -7.187,  -7.384,  -7.405,  -6.668,  -5.513,  -4.224,  -2.855,  -1.565,  -0.064,
   1.513,   3.315,   4.907,   6.187,   7.314,   8.385,   9.159,   9.578,  10.450,  10.890,
  11.309,  11.842,  12.716,  13.895,  13.050,  13.144,  13.328,  13.668,  13.875,  14.061},
{
  13.421,  13.408,  13.418,  13.391,  13.570,  13.610,  13.861,  14.402,  14.638,  14.987,
  16.191,  17.937,  20.970,  22.393,  24.410,  22.453,  20.305,  18.037,  16.414,  14.740,
  13.039,  10.873,   8.759,   6.925,   5.328,   4.455,   2.882,   1.515,  -0.146,  -3.096,
  -7.980,  -9.057,  -9.677, -10.872, -10.077, -12.240, -13.711, -16.147, -17.881, -18.889,
 -20.155, -23.311, -21.241, -23.297, -24.457, -26.089, -23.013, -19.795, -18.044, -13.726,
 -14.761, -19.425, -21.683, -22.314, -22.578, -21.960, -22.199, -22.978, -24.336, -27.158,
 -23.788, -23.752, -27.559, -29.020, -30.032, -30.944, -31.671, -33.110, -35.661, -38.139,
 -40.756, -43.409, -46.454, -49.252, -50.716, -52.013, -53.263, -54.290, -55.287, -56.278,
 -57.279, -58.202, -58.918, -60.157, -61.165, -61.151, -58.473, -56.494, -56.433, -59.403,
 -57.770, -57.486, -55.467, -53.935, -52.650, -48.583, -48.092, -43.988, -43.288, -41.479,
 -39.592, -36.924, -34.075, -32.191, -31.018, -28.367, -24.649, -21.674, -18.374, -15.270,
 -14.067, -12.069,  -8.020,  -5.196,  -2.641,  -1.318,   1.308,   3.531,   6.447,   9.770,
  15.436,  21.324,  25.630,  28.139,  30.233,  31.843,  34.763,  36.615,  37.472,  36.834,
  39.084,  40.509,  42.548,  44.757,  46.959,  48.441,  50.731,  52.675,  52.218,  52.347,
  53.736,  55.892,  58.162,  61.251,  63.994,  65.219,  62.830,  63.347,  64.295,  64.072,
  63.429,  61.965,  61.397,  61.770,  62.294,  62.658,  63.263,  63.161,  63.914,  63.210,
  63.614,  61.765,  61.119,  62.237,  63.012,  59.022,  54.301,  62.143,  63.563,  64.969,
  64.938,  64.113,  62.593,  60.899,  59.472,  58.015,  56.190,  54.560,  52.547,  51.054,  49.128,
  47.849,  46.354,  45.058,  43.716,  41.801,  37.978,  26.791,  26.920,  28.697,  26.001,
  22.523,  19.891,  17.225,  15.437,  13.749,  12.892,  11.823,  10.509,   9.769,   9.000,
   8.234,   7.307,   7.080,   6.234,   6.008,   5.707,   4.517,   3.442,   2.418,   1.247,
   1.048,   2.210,   0.175,  -1.772,  -2.767,  -3.703,  -4.858,  -5.213,  -5.733,  -6.722,
  -7.958,  -9.124,  -9.592,  -9.938, -10.117, -10.113,  -9.968,  -9.881,  -9.568,  -9.634,
  -9.488,  -9.344,  -9.169,  -8.875,  -8.567,  -8.262,  -7.893,  -7.490,  -6.997,  -6.567,
  -6.074,  -5.556,  -4.826,  -4.344,  -3.885,  -3.542,  -3.358,  -3.288,  -3.296,  -3.339,
  -3.389,  -3.578,  -3.946,  -4.504,  -5.069,  -5.631,  -6.044,  -6.494,  -6.792,  -6.593,
  -6.732,  -6.804,  -6.781,  -6.770,  -6.971,  -6.979,  -6.853,  -6.732,  -6.217,  -5.628,
  -5.189,  -4.734,  -4.303,  -3.803,  -2.891,  -1.728,  -0.074,   1.656,   4.105,   6.950,
   9.851,  11.654,  11.037,  15.450,  30.370,  38.459,  43.637,  45.825,  46.533,  46.584,
  44.522,  32.931,  27.297,  26.508,  26.042,  25.316,  22.409,  19.983,  19.117,  16.401,
  12.632,   7.977,   3.173,  -0.150,  -1.958,  -4.359,  -5.000,  -5.937,  -7.231,  -9.230,
 -11.526, -13.543, -15.030, -13.644, -12.755, -13.234, -12.789, -10.640,  -9.954, -10.895,
  -8.423, -13.252, -13.717, -11.942, -10.630,  -9.414,  -8.272,  -7.665,  -7.059,  -7.056,
  -7.148,  -7.254,  -7.275,  -7.005,  -6.225,  -5.317,  -4.227,  -2.932,  -1.557,  -0.076,
   1.448,   3.045,   4.624,   5.987,   7.265,   8.701,   9.539,   9.726,  10.401,  10.975,
  11.505,  12.003,  12.393,  12.758,  13.045,  13.434,  13.297,  13.510,  13.380,  13.421},
{
  12.683,  12.631,  12.485,  12.539,  12.784,  12.930,  13.264,  14.005,  14.597,  14.960,
  15.558,  17.237,  18.972,  21.322,  21.556,  22.419,  20.720,  17.959,  15.941,  13.918,
  11.955,   9.127,   6.025,   5.428,   3.908,   2.717,   1.103,  -0.081,  -2.396,  -4.585,
  -7.079, -11.342, -11.470, -11.560, -10.743, -15.556, -17.063, -18.903, -20.051, -20.470,
 -22.058, -24.009, -23.057, -25.008, -26.387, -26.579, -25.482, -24.261, -18.404, -15.688,
 -16.316, -21.902, -23.746, -24.465, -25.087, -25.113, -25.298, -25.657, -26.898, -29.030,
 -29.118, -26.918, -30.573, -32.281, -33.770, -34.713, -35.715, -37.382, -39.920, -42.391,
 -44.822, -47.458, -50.472, -52.754, -54.522, -56.258, -57.867, -58.988, -59.539, -60.039,
 -60.901, -61.641, -62.035, -62.181, -63.915, -64.050, -62.724, -60.981, -59.502, -59.956,
 -59.499, -58.903, -56.458, -54.810, -52.663, -49.519, -46.078, -42.446, -41.467, -40.498,
 -37.892, -34.574, -29.971, -28.171, -26.942, -24.647, -20.610, -17.385, -14.426,  -9.412,
  -9.586,  -7.682,  -4.422,  -0.509,   1.010,   3.137,   5.574,   7.714,  10.386,  14.334,
  19.364,  21.649,  26.951,  29.352,  32.355,  34.010,  35.743,  37.662,  38.925,  41.878,
  42.481,  44.716,  47.734,  50.216,  51.413,  52.059,  53.677,  55.115,  54.914,  55.230,
  56.872,  58.703,  61.583,  63.732,  66.365,  64.652,  63.916,  64.844,  63.626,  62.308,
  62.212,  62.609,  62.969,  63.384,  63.552,  64.575,  66.109,  64.713,  64.234,  64.470,
  64.723,  64.493,  63.629,  62.676,  61.564,  61.718,  52.161,  56.759,  64.605,  64.022,
  63.664,  62.580,  61.082,  59.489,  57.750,  56.024,  54.387,  52.513,  51.144,  49.516,  47.459,
  46.052,  44.478,  43.098,  38.267,  35.763,  32.306,  31.479,  33.265,  29.815,  29.363,
  23.556,  19.473,  17.016,  15.642,  14.213,  12.806,  12.582,  11.718,  10.399,   9.137,
   8.161,   7.427,   6.923,   6.361,   5.708,   4.919,   4.049,   3.336,   1.977,   2.650,
   1.047,  -0.882,  -1.863,  -2.506,  -3.030,  -3.947,  -5.138,  -5.601,  -5.796,  -6.822,
  -7.143,  -7.656,  -8.518,  -9.025,  -9.338,  -9.690,  -9.652,  -9.607,  -9.499,  -9.502,
  -9.425,  -9.296,  -9.307,  -9.085,  -8.784,  -8.635,  -8.285,  -7.910,  -7.558,  -7.166,
  -6.680,  -6.044,  -5.512,  -4.986,  -4.549,  -4.215,  -4.062,  -3.923,  -3.969,  -4.035,
  -4.183,  -4.326,  -4.669,  -5.080,  -5.632,  -6.404,  -7.067,  -7.509,  -7.672,  -7.743,
  -7.602,  -7.685,  -7.763,  -7.809,  -7.711,  -7.648,  -7.477,  -7.062,  -6.557,  -6.008,
  -5.652,  -5.120,  -4.658,  -4.186,  -3.488,  -2.335,  -0.687,   1.271,   3.500,   6.319,
   8.505,   8.769,  11.663,  24.460,  37.657,  37.101,  40.306,  44.807,  49.123,  44.159,
  33.725,  31.139,  28.831,  29.353,  28.379,  24.806,  21.131,  18.807,  17.640,  14.906,
  12.283,  10.168,   4.095,  -1.536,  -4.199,  -6.269,  -7.599,  -8.796,  -9.534,  -9.562,
 -10.635, -14.600, -16.752, -14.055, -12.681, -13.124, -13.133, -13.562, -11.951, -10.269,
  -9.221, -11.657, -12.888, -12.471, -11.238, -10.020,  -8.534,  -7.866,  -7.349,  -7.412,
  -7.243,  -7.173,  -6.990,  -6.602,  -6.021,  -5.131,  -4.069,  -2.740,  -1.191,   0.370,
   2.013,   3.530,   4.908,   6.263,   7.767,   8.830,   9.333,   9.592,  10.189,  10.596,
  11.324,  12.003,  12.596,  13.111,  13.288,  13.030,  12.899,  13.031,  12.826,  12.683},
{
  11.942,  11.798,  11.826,  11.840,  12.027,  12.399,  12.844,  13.472,  14.202,  15.060,
  15.867,  16.669,  18.321,  21.982,  20.704,  22.303,  21.632,  18.473,  15.278,  13.072,
  11.000,   7.864,   5.136,   3.513,   2.150,   1.177,   0.167,  -2.330,  -4.645,  -4.673,
  -6.085, -10.171, -14.336, -12.395, -15.370, -18.885, -17.190, -21.205, -21.816, -21.368,
 -21.590, -25.446, -27.099, -27.757, -27.145, -22.107, -27.321, -26.891, -20.627, -18.313,
 -20.562, -25.500, -26.202, -27.204, -27.876, -27.378, -27.535, -28.969, -30.166, -31.675,
 -34.081, -31.284, -34.035, -35.799, -37.361, -38.454, -39.403, -41.621, -43.991, -46.242,
 -48.869, -51.339, -54.356, -56.756, -58.712, -60.499, -62.190, -63.670, -64.072, -64.044,
 -64.143, -64.909, -65.321, -64.587, -66.038, -66.752, -65.451, -65.562, -62.766, -60.386,
 -61.013, -59.247, -57.125, -55.688, -52.752, -47.551, -43.018, -42.107, -39.858, -38.026,
 -36.574, -31.366, -26.474, -23.794, -22.028, -19.307, -16.572, -12.792,  -9.525,  -5.591,
  -4.376,  -2.584,   0.615,   2.950,   4.927,   7.415,  10.537,  12.252,  14.510,  18.198,
  23.024,  25.657,  29.028,  32.432,  35.340,  36.246,  37.247,  39.117,  41.965,  44.016,
  45.728,  49.140,  51.513,  53.836,  54.801,  55.515,  56.875,  57.586,  58.280,  58.581,
  59.712,  60.989,  63.780,  66.734,  65.139,  64.335,  65.670,  64.939,  64.868,  64.648,
  64.435,  64.334,  64.227,  64.090,  64.741,  66.960,  68.383,  67.655,  65.787,  65.636,
  63.638,  65.238,  63.806,  62.654,  61.938,  59.468,  47.984,  58.977,  62.156,  62.324,
  61.636,  60.441,  59.104,  57.257,  55.786,  53.599,  52.157,  50.294,  48.098,  45.909,  41.152,
  43.763,  41.405,  36.420,  36.288,  33.708,  31.888,  29.163,  27.730,  25.177,  24.284,
  21.463,  19.341,  17.043,  15.458,  13.691,  13.316,  13.776,  13.173,  11.696,   9.529,
   8.815,   8.007,   7.197,   6.678,   5.510,   4.398,   3.887,   3.083,   1.688,   0.841,
   0.106,  -0.486,  -1.405,  -2.526,  -3.274,  -4.029,  -4.348,  -4.874,  -5.228,  -5.893,
  -6.040,  -6.781,  -7.583,  -8.345,  -8.789,  -9.153,  -9.103,  -9.056,  -9.242,  -9.212,
  -9.232,  -9.343,  -9.311,  -9.365,  -9.106,  -8.837,  -8.552,  -8.326,  -8.069,  -7.887,
  -7.598,  -7.272,  -6.737,  -6.035,  -5.581,  -5.232,  -5.036,  -4.884,  -4.749,  -4.737,
  -4.882,  -5.117,  -5.416,  -6.031,  -6.539,  -7.396,  -8.006,  -8.522,  -8.685,  -8.892,
  -8.844,  -8.807,  -8.780,  -8.907,  -8.615,  -8.492,  -8.178,  -7.778,  -7.353,  -6.954,
  -6.316,  -5.611,  -4.869,  -4.197,  -3.523,  -2.389,  -1.027,   1.259,   3.552,   5.702,
   6.255,   6.645,  16.848,  31.969,  35.765,  34.334,  36.390,  41.484,  36.632,  31.042,
  30.175,  30.547,  31.129,  29.194,  26.835,  23.899,  20.606,  18.690,  17.040,  13.394,
  11.519,  11.066,   4.692,  -2.332,  -4.985,  -8.014,  -8.469,  -8.344, -11.124, -12.152,
 -11.741, -15.855, -17.952, -14.031, -12.817, -14.264, -15.671, -12.706, -11.997, -10.850,
 -11.125, -11.104, -12.317, -12.528, -11.602, -10.185,  -8.771,  -8.154,  -8.220,  -8.365,
  -8.395,  -7.923,  -7.415,  -7.071,  -6.371,  -5.237,  -4.048,  -2.730,  -1.118,   0.658,
   2.352,   4.032,   5.457,   6.836,   8.667,   9.311,   9.458,   9.758,  10.022,  10.397,
  10.796,  11.639,  12.309,  14.263,  13.373,  13.289,  13.328,  12.467,  12.117,  11.942},
{
  11.145,  11.082,  11.096,  11.210,  11.389,  11.697,  12.149,  12.914,  13.809,  14.745,
  15.806,  16.623,  17.239,  19.394,  20.884,  21.653,  21.372,  19.251,  14.629,  11.975,
   9.822,   7.181,   5.085,   2.751,   1.026,  -0.303,  -1.707,  -2.670,  -5.347,  -7.226,
  -8.684, -10.246, -13.239, -13.288, -16.181, -18.815, -18.878, -22.632, -24.062, -23.633,
 -23.436, -26.124, -26.724, -25.931, -23.729, -26.892, -28.059, -28.573, -24.911, -24.050,
 -26.311, -28.525, -28.703, -29.567, -30.008, -29.949, -30.827, -31.848, -33.455, -34.769,
 -36.541, -34.424, -36.409, -39.375, -41.058, -42.413, -43.757, -45.736, -48.372, -50.715,
 -52.695, -54.955, -58.173, -60.956, -62.773, -64.475, -66.239, -67.669, -68.245, -68.371,
 -68.145, -68.484, -68.765, -68.156, -67.672, -69.109, -68.310, -67.200, -64.865, -61.565,
 -62.739, -58.849, -58.389, -55.771, -52.296, -47.934, -43.856, -39.167, -37.874, -36.004,
 -33.436, -28.257, -24.068, -21.091, -17.211, -13.797, -12.734,  -9.300,  -6.512,  -2.915,
   0.377,   3.293,   5.655,   7.134,   7.826,  10.343,  13.216,  14.930,  17.190,  20.578,
  23.389,  24.985,  29.507,  34.984,  36.514,  38.432,  40.938,  43.283,  44.724,  45.951,
  49.110,  52.294,  53.545,  56.428,  56.760,  58.875,  60.478,  60.660,  61.781,  61.968,
  62.437,  63.469,  66.535,  69.328,  68.005,  66.948,  67.879,  67.210,  66.536,  65.337,
  65.248,  65.556,  65.824,  66.448,  66.671,  64.557,  67.145,  68.195,  66.912,  63.435,
  66.799,  64.262,  61.736,  60.932,  59.761,  55.787,  46.261,  58.243,  60.770,  60.467,
  59.537,  57.903,  56.558,  52.529,  51.610,  49.753,  49.120,  46.318,  45.937,  42.257,  42.816,
  40.850,  39.176,  34.992,  33.987,  31.177,  30.541,  27.392,  26.498,  24.622,  23.012,
  20.494,  18.350,  16.895,  15.410,  14.781,  14.439,  14.137,  14.438,  12.642,  10.315,
   9.295,   8.011,   7.311,   6.652,   5.737,   4.779,   4.191,   3.197,   2.607,   2.419,
   1.083,   0.478,  -0.070,  -1.638,  -2.355,  -3.545,  -4.085,  -4.746,  -4.633,  -4.877,
  -4.845,  -5.445,  -6.281,  -7.010,  -7.658,  -8.218,  -8.504,  -8.655,  -8.733,  -8.937,
  -9.234,  -9.270,  -9.407,  -9.465,  -9.330,  -9.236,  -9.306,  -9.050,  -8.853,  -8.780,
  -8.555,  -8.352,  -7.882,  -7.240,  -6.747,  -6.361,  -6.103,  -5.880,  -5.701,  -5.705,
  -5.890,  -6.087,  -6.397,  -6.858,  -7.615,  -8.507,  -9.162,  -9.370,  -9.572,  -9.734,
  -9.905,  -9.968,  -9.896,  -9.713,  -9.684,  -9.364,  -9.186,  -8.864,  -8.471,  -7.956,
  -7.037,  -6.284,  -5.480,  -4.470,  -3.476,  -2.275,  -0.816,   1.282,   3.410,   3.823,
   1.934,  10.235,  24.029,  31.575,  32.546,  30.987,  31.445,  30.952,  27.868,  27.473,
  28.817,  29.323,  29.829,  26.998,  22.973,  20.833,  19.409,  18.370,  15.249,  10.845,
   9.143,   8.129,   2.022,  -4.608,  -7.527,  -9.556,  -9.436, -10.683, -13.996, -16.348,
 -14.045, -15.390, -16.929, -15.588, -15.773, -16.022, -15.690, -13.136, -11.979, -11.571,
 -11.783, -10.659, -12.320, -12.517, -12.089, -10.590,  -9.395,  -8.662,  -9.210,  -9.062,
  -8.669,  -8.066,  -7.731,  -7.206,  -6.161,  -4.943,  -3.575,  -2.186,  -0.371,   1.521,
   3.156,   4.679,   6.111,   7.388,   8.680,   9.632,   9.865,  10.264,  10.342,  10.510,
  11.055,  11.727,  12.185,  11.866,  12.962,  12.728,  12.289,  11.796,  11.241,  11.145},
{
  11.092,  10.733,  10.563,  10.474,  10.493,  10.703,  11.290,  11.945,  12.801,  13.517,
  14.483,  15.230,  16.663,  18.226,  18.497,  19.860,  19.485,  15.996,  12.643,  10.365,
   8.378,   6.150,   4.070,   1.311,  -1.122,  -0.690,  -2.549,  -4.212,  -5.577,  -9.234,
 -10.049, -11.045, -12.221, -14.001, -14.547, -16.822, -18.165, -22.424, -24.569, -24.281,
 -24.017, -27.032, -26.803, -28.662, -29.225, -30.300, -31.223, -30.438, -31.190, -30.456,
 -30.344, -30.681, -30.415, -31.264, -31.964, -32.231, -33.121, -34.082, -35.653, -37.336,
 -38.192, -36.180, -37.864, -42.434, -44.884, -46.935, -47.531, -49.732, -52.124, -54.883,
 -56.778, -58.338, -61.689, -65.094, -67.165, -68.567, -70.150, -71.483, -72.037, -72.151,
 -71.992, -71.584, -72.030, -71.808, -70.094, -69.726, -69.609, -68.009, -66.674, -62.858,
 -64.579, -59.625, -57.311, -54.899, -51.591, -47.787, -43.686, -39.273, -36.695, -33.352,
 -28.624, -25.930, -22.655, -18.729, -13.780,  -8.364,  -7.986,  -7.320,  -5.275,  -2.442,
  -0.161,   2.693,   4.979,   4.606,   4.318,   7.216,  10.806,  13.390,  15.738,  17.591,
  22.827,  27.071,  28.989,  29.966,  31.208,  36.798,  41.559,  45.749,  47.967,  48.946,
  51.997,  54.565,  57.043,  58.940,  58.662,  60.157,  62.279,  62.579,  63.826,  65.522,
  66.327,  66.727,  69.363,  72.354,  71.806,  70.070,  70.213,  69.299,  68.378,  69.000,
  71.548,  74.733,  73.960,  73.717,  71.572,  70.132,  65.433,  69.023,  67.994,  64.705,
  64.203,  58.105,  62.149,  55.653,  55.246,  52.449,  54.335,  57.887,  58.639,  57.482,
  56.060,  52.234,  50.521,  49.168,  46.470,  45.155,  42.633,  40.755,  40.308,  38.889,  39.243,
  37.798,  33.678,  32.639,  30.811,  29.358,  27.966,  27.403,  26.326,  24.034,  21.939,
  20.408,  18.546,  16.836,  16.646,  14.603,  14.766,  14.984,  14.682,  12.250,  10.760,
   9.232,   8.028,   7.880,   6.908,   5.881,   5.720,   4.860,   4.711,   4.256,   3.473,
   2.623,   1.668,   0.726,  -0.495,  -1.537,  -2.425,  -3.059,  -3.421,  -3.595,  -3.835,
  -3.707,  -4.006,  -4.672,  -5.577,  -6.599,  -7.432,  -7.883,  -8.267,  -8.632,  -8.996,
  -9.278,  -9.447,  -9.593,  -9.739,  -9.753,  -9.808,  -9.647,  -9.622,  -9.722,  -9.887,
  -9.854,  -9.485,  -9.150,  -8.680,  -8.128,  -7.559,  -7.242,  -7.047,  -6.893,  -6.768,
  -6.894,  -7.130,  -7.426,  -7.861,  -8.525,  -9.453, -10.098, -10.484, -10.568, -10.702,
 -10.898, -10.925, -10.916, -10.641, -10.570, -10.154, -10.088,  -9.816,  -9.346,  -8.714,
  -7.777,  -6.899,  -6.028,  -4.937,  -3.734,  -2.283,  -0.572,   1.328,   2.276,  -0.052,
   2.243,  14.975,  26.677,  31.677,  30.958,  28.697,  27.833,  26.232,  27.028,  28.394,
  27.692,  27.152,  26.514,  24.436,  20.896,  19.397,  17.875,  14.753,  11.055,   6.779,
   2.284,  -1.908,  -4.416,  -8.972, -10.094, -10.425, -10.789, -12.027, -15.596, -18.501,
 -17.111, -15.738, -16.499, -16.208, -17.097, -17.016, -16.479, -15.813, -12.667, -11.758,
 -13.412, -11.568, -10.316, -10.401, -11.610, -11.186, -10.640, -10.159,  -9.815,  -9.189,
  -8.747,  -8.063,  -7.772,  -7.076,  -5.991,  -4.738,  -3.271,  -1.446,   0.540,   2.490,
   4.323,   5.739,   7.100,   8.410,   9.759,  10.661,  11.874,  11.444,  11.324,  11.163,
  11.450,  12.095,  12.525,  12.626,  12.712,  12.691,  12.107,  11.577,  11.934,  11.092},
{
  11.477,  10.329,  10.033,   9.890,   9.780,   9.923,  10.216,  10.675,  11.271,  11.930,
  12.982,  14.320,  15.327,  17.424,  17.230,  17.204,  16.402,  13.119,  10.560,   8.374,
   6.333,   4.160,   1.916,  -1.592,  -4.390,  -5.234,  -4.977,  -7.489,  -6.384,  -9.328,
 -12.651, -11.682, -11.387, -11.897, -16.749, -15.318, -18.540, -21.520, -23.453, -23.793,
 -26.714, -31.052, -29.794, -31.801, -32.410, -32.638, -33.384, -33.880, -32.786, -33.197,
 -32.479, -30.190, -32.430, -32.654, -33.364, -34.182, -34.890, -36.145, -38.286, -39.475,
 -38.094, -38.977, -41.541, -45.841, -48.761, -50.510, -52.376, -53.601, -56.221, -58.556,
 -60.861, -62.613, -64.980, -68.439, -71.149, -72.711, -74.022, -75.289, -75.933, -76.194,
 -75.940, -75.151, -75.116, -74.595, -73.546, -71.180, -70.130, -69.290, -68.168, -64.918,
 -64.511, -60.127, -58.132, -54.050, -51.876, -46.860, -42.898, -38.598, -35.436, -31.967,
 -26.696, -23.151, -20.808, -15.987, -11.260,  -6.048,  -4.677,  -7.770,  -6.581,  -3.742,
  -2.172,   0.856,   4.806,   9.709,  10.214,  11.007,  13.714,  16.273,  19.300,  27.170,
  36.360,  33.589,  27.231,  26.926,  33.261,  33.094,  36.139,  42.028,  46.665,  49.787,
  52.229,  55.261,  57.786,  60.347,  61.227,  62.761,  64.934,  64.210,  65.607,  67.514,
  69.201,  70.022,  72.079,  74.296,  73.661,  70.927,  69.206,  69.072,  75.329,  78.171,
  78.443,  76.643,  74.346,  73.064,  72.596,  71.217,  69.797,  67.369,  65.925,  62.998,
  63.230,  62.729,  58.420,  59.676,  59.820,  58.238,  58.336,  57.415,  55.647,  53.949,
  50.628,  49.562,  46.958,  44.349,  42.602,  41.495,  39.323,  37.734,  36.219,  35.070,  35.210,
  33.142,  31.217,  29.722,  28.515,  27.422,  27.291,  26.385,  24.513,  22.434,  21.455,
  19.877,  18.306,  17.248,  15.843,  14.939,  15.501,  15.160,  13.987,  12.574,  10.868,
   9.670,   8.267,   7.983,   7.305,   7.011,   6.575,   6.203,   6.057,   5.464,   4.912,
   3.551,   2.371,   1.579,   0.312,  -0.552,  -1.379,  -1.693,  -2.078,  -2.899,  -2.540,
   0.706,  -2.067,  -3.908,  -4.819,  -6.040,  -6.756,  -7.491,  -8.162,  -8.837,  -9.244,
  -9.534,  -9.996, -10.082, -10.290, -10.406, -10.702, -10.610, -10.760, -10.804, -11.028,
 -11.012, -11.012, -10.713, -10.261,  -9.583,  -8.949,  -8.520,  -8.304,  -8.052,  -7.832,
  -7.981,  -8.167,  -8.447,  -8.879,  -9.369, -10.196, -11.168, -11.440, -11.762, -11.961,
 -11.968, -11.850, -11.760, -11.506, -11.211, -11.117, -10.932, -10.703, -10.240,  -9.363,
  -8.469,  -7.483,  -6.473,  -5.308,  -3.957,  -2.426,  -0.624,   0.754,   0.387,  -0.787,
   9.197,  20.177,  24.226,  28.332,  26.513,  23.260,  26.349,  26.310,  27.267,  27.401,
  26.278,  24.797,  23.441,  21.391,  18.491,  15.913,  14.459,  11.129,   7.234,   3.321,
  -0.813,  -4.002,  -3.940,  -7.075, -10.236, -11.526, -12.420, -12.752, -14.751, -18.860,
 -19.403, -18.862, -18.510, -17.214, -16.003, -14.816, -16.776, -16.529, -14.236, -11.872,
 -12.348, -10.583,  -9.060,  -9.376, -11.554, -11.432, -11.217, -10.796, -10.184,  -9.548,
  -8.932,  -8.213,  -7.453,  -6.841,  -5.939,  -4.404,  -2.633,  -0.783,   1.566,   3.870,
   5.306,   6.860,   8.020,   9.223,  10.473,  11.995,  13.254,  12.834,  12.758,  12.658,
  12.984,  13.062,  13.084,  13.359,  13.381,  13.365,  12.745,  12.627,  12.519,  11.477},
{
  11.602,  10.127,   9.905,   9.404,   9.152,   9.386,   9.382,   9.484,   9.841,  10.287,
  10.964,  11.443,  12.786,  15.666,  15.554,  14.950,  12.684,   9.789,   7.984,   6.413,
   3.927,   1.580,  -0.484,  -3.139,  -5.625,  -7.556,  -9.657,  -8.873,  -8.392, -12.226,
 -13.808, -13.169, -11.259, -12.184, -13.597, -13.424, -19.301, -21.351, -23.807, -25.244,
 -31.129, -33.349, -33.389, -34.383, -34.702, -35.044, -37.016, -33.617, -38.067, -34.567,
 -34.386, -33.971, -33.776, -35.927, -35.422, -36.060, -36.762, -38.258, -40.137, -41.084,
 -40.308, -44.286, -46.356, -49.164, -51.964, -54.322, -55.945, -57.902, -60.128, -62.578,
 -64.217, -65.920, -68.528, -72.126, -74.986, -76.301, -77.828, -79.072, -79.821, -79.890,
 -79.066, -78.090, -77.044, -76.902, -75.620, -73.913, -72.290, -70.215, -68.185, -64.730,
 -64.627, -60.204, -58.004, -54.541, -49.877, -45.718, -42.461, -37.396, -32.338, -29.510,
 -25.350, -20.644, -16.803, -12.867, -10.025,  -9.393,  -8.951,  -5.856,  -2.772,   1.020,
   7.220,  11.403,  15.538,  20.761,  24.158,  27.485,  30.686,  34.751,  33.271,  35.991,
  39.719,  37.484,  34.638,  32.270,  37.336,  41.986,  41.786,  37.418,  34.138,  40.950,
  46.467,  51.742,  56.275,  59.849,  61.965,  64.513,  66.660,  66.472,  67.491,  68.885,
  70.472,  71.361,  73.286,  74.611,  74.385,  74.402,  70.884,  77.019,  82.188,  77.721,
  78.413,  77.650,  77.115,  73.608,  73.430,  71.217,  70.327,  68.518,  66.950,  63.066,
  62.308,  61.538,  54.496,  59.604,  58.921,  56.327,  55.675,  53.817,  52.320,  49.307,
  47.984,  45.129,  42.803,  41.545,  40.775,  39.432,  37.872,  35.923,  34.504,  33.325,  31.891,
  29.368,  28.531,  27.150,  25.662,  24.521,  24.337,  23.511,  22.927,  21.643,  20.264,
  18.856,  17.422,  16.759,  15.944,  15.204,  14.988,  14.781,  13.077,  12.305,  11.198,
   9.483,  12.384,   8.688,   7.953,   7.976,   7.665,   7.315,   7.297,   6.992,   5.306,
   4.605,   3.481,   2.212,   1.086,   0.412,   0.144,  -0.266,  -0.837,  -0.471,   2.357,
  -1.519,  -2.657,  -3.168,  -4.397,  -5.559,  -6.618,  -7.829,  -8.497,  -9.129,  -9.688,
 -10.216, -10.404, -11.057, -11.345, -11.456, -11.459, -11.521, -11.894, -12.277, -12.572,
 -12.516, -12.329, -11.940, -11.646, -11.156, -10.565, -10.059,  -9.670,  -9.388,  -9.180,
  -9.160,  -9.231,  -9.360,  -9.700, -10.184, -10.702, -11.642, -12.158, -12.284, -12.591,
 -12.571, -12.652, -12.571, -12.391, -12.194, -12.043, -11.906, -11.383, -10.793,  -9.979,
  -9.038,  -8.019,  -6.813,  -5.461,  -3.917,  -2.189,  -0.309,   0.348,  -2.256,   5.812,
  13.758,  18.928,  19.277,  17.668,  19.218,  20.675,  25.139,  24.918,  26.175,  26.406,
  24.359,  22.609,  20.333,  17.301,  13.779,  12.030,  10.986,   7.724,   4.106,   0.947,
  -2.526,  -5.353,  -7.695, -10.445, -13.189, -14.884, -15.590, -16.214, -17.445, -20.533,
 -22.076, -21.278, -20.688, -19.887, -18.909, -17.716, -15.643, -14.491, -13.751, -12.139,
 -10.895,  -9.383,  -7.585,  -7.078,  -7.574, -11.143, -11.050, -10.927, -10.369,  -9.618,
  -8.627,  -7.946,  -7.163,  -6.126,  -5.121,  -3.656,  -2.078,  -0.214,   2.313,   4.330,
   6.251,   7.854,   9.436,  10.492,  11.764,  13.258,  14.218,  13.954,  13.758,  13.795,
  13.968,  14.254,  14.177,  13.954,  14.086,  14.059,  14.081,  13.817,  12.857,  11.602},
{
  13.059,  12.546,  10.429,   9.648,   9.006,   8.763,   8.487,   8.362,   8.417,   8.526,
   9.002,  10.281,  11.138,  13.215,  13.150,  12.048,  10.016,   7.110,   5.528,   4.035,
   1.269,  -1.450,  -3.429,  -5.479,  -6.965,  -7.818, -12.094, -13.589, -11.365, -11.672,
 -11.391, -14.031, -13.854, -15.458, -16.573, -14.931, -17.204, -21.973, -24.773, -26.750,
 -29.374, -33.689, -35.345, -36.273, -36.543, -37.679, -39.639, -35.514, -39.892, -37.521,
 -36.194, -36.926, -36.261, -37.097, -37.899, -38.398, -38.854, -40.744, -42.407, -42.577,
 -44.458, -47.409, -49.918, -52.693, -55.208, -57.743, -59.558, -61.203, -62.785, -66.486,
 -68.849, -69.460, -71.883, -75.880, -78.776, -80.231, -81.512, -82.451, -83.344, -84.088,
 -83.273, -81.506, -80.876, -79.614, -79.010, -76.652, -74.587, -72.293, -69.609, -65.591,
 -64.717, -60.815, -58.066, -53.551, -49.846, -44.802, -42.019, -35.311, -30.412, -26.286,
 -22.660, -18.726, -14.674, -13.219, -12.152, -11.687,  -6.802,   1.904,  12.361,  19.057,
  23.520,  25.340,  26.548,  30.445,  33.203,  33.992,  33.469,  35.471,  36.796,  37.276,
  37.895,  37.933,  41.267,  43.313,  45.255,  45.824,  47.617,  45.588,  42.174,  41.725,
  45.506,  46.677,  51.040,  58.995,  63.176,  66.447,  67.204,  68.354,  68.685,  69.818,
  70.995,  72.174,  74.238,  74.775,  74.530,  73.796,  75.442,  84.229,  74.844,  67.704,
  71.729,  72.816,  68.732,  70.623,  69.428,  68.921,  67.709,  69.513,  65.470,  63.346,
  61.454,  56.951,  58.690,  58.015,  56.006,  53.530,  51.459,  50.330,  48.623,  46.766,
  44.179,  41.756,  40.526,  39.404,  38.647,  37.248,  35.722,  34.258,  33.087,  31.301,  28.605,
  27.062,  26.149,  24.807,  23.589,  22.190,  21.117,  21.440,  20.418,  20.273,  18.884,
  17.923,  17.195,  16.860,  16.576,  16.291,  14.935,  14.564,  14.165,  12.390,  12.197,
  10.443,   9.589,   9.515,   9.202,   9.235,   8.863,   8.815,   8.276,   7.562,   6.065,
   5.079,   3.994,   3.105,   2.170,   1.722,   1.729,   1.197,   0.341,   2.843,   0.833,
  -1.489,  -1.672,  -2.544,  -3.941,  -5.280,  -6.437,  -7.549,  -8.286,  -9.116,  -9.888,
 -10.441, -10.990, -11.678, -11.929, -12.344, -12.720, -12.868, -13.233, -13.518, -13.436,
 -13.473, -13.381, -13.379, -13.014, -12.758, -12.140, -11.576, -11.117, -10.844, -10.525,
 -10.405, -10.256, -10.405, -10.627, -10.935, -11.428, -12.092, -12.911, -13.151, -13.122,
 -13.129, -12.986, -13.003, -13.041, -12.995, -12.772, -12.403, -11.894, -11.273, -10.300,
  -9.356,  -8.252,  -6.877,  -5.373,  -3.837,  -1.950,  -0.261,  -0.200,  -1.858,   8.543,
  14.608,  19.770,  13.362,  11.924,  14.606,  20.568,  22.784,  23.560,  25.617,  24.246,
  22.424,  20.125,  17.947,  14.548,  10.933,  10.018,   8.117,   4.420,   1.316,  -1.339,
  -4.134,  -7.487, -10.443, -13.513, -16.340, -18.042, -19.064, -19.439, -20.083, -21.769,
 -23.277, -22.955, -23.055, -23.170, -22.621, -21.737, -19.194, -16.170, -12.539, -11.072,
  -9.917,  -7.931,  -6.313,  -5.317,  -5.270,  -9.543, -10.685, -10.633, -10.515,  -9.222,
  -8.059,  -7.652,  -6.940,  -5.914,  -4.740,  -3.727,  -1.954,   0.139,   2.057,   4.300,
   6.422,   8.312,   9.709,  10.850,  12.114,  14.334,  15.076,  15.100,  15.221,  15.017,
  14.882,  14.687,  14.816,  14.626,  15.010,  14.996,  15.169,  15.207,  14.591,  13.059},
{
  15.223,  12.820,  10.989,   9.593,   9.111,   8.690,   7.872,   7.522,   7.415,   7.441,
   8.046,  10.681,  12.795,  11.099,   9.760,   8.885,   7.395,   4.675,   3.110,   0.550,
  -2.368,  -4.765,  -6.780,  -9.444, -10.433, -11.187, -13.183, -15.855, -15.325, -10.940,
 -11.023, -12.946, -14.459, -15.702, -16.335, -18.736, -20.382, -21.859, -24.531, -26.759,
 -30.361, -34.031, -36.494, -37.302, -38.236, -39.119, -40.348, -43.055, -43.131, -38.502,
 -38.646, -38.188, -38.323, -39.090, -39.210, -40.490, -41.645, -43.079, -43.757, -45.459,
 -48.641, -51.039, -53.394, -55.965, -58.472, -60.974, -62.831, -64.643, -66.058, -69.138,
 -72.612, -72.743, -74.282, -78.078, -82.000, -83.983, -85.192, -85.685, -86.095, -86.586,
 -85.663, -84.164, -82.282, -81.274, -80.663, -78.333, -76.010, -73.424, -70.331, -65.317,
 -63.865, -61.285, -57.625, -52.139, -48.453, -44.450, -39.741, -33.754, -28.999, -23.329,
 -18.388, -16.848, -14.935, -11.839,  -9.350,  -6.018,   7.619,  20.129,  21.680,  22.256,
  24.376,  24.987,  27.910,  31.081,  33.550,  36.847,  39.470,  40.810,  42.863,  44.209,
  44.402,  48.961,  48.114,  47.518,  48.612,  50.191,  50.333,  50.058,  49.514,  49.521,
  45.318,  41.798,  46.077,  53.909,  62.905,  67.124,  68.115,  69.843,  69.639,  70.126,
  72.213,  73.311,  74.952,  76.373,  77.424,  76.712,  78.264,  73.312,  59.791,  58.705,
  64.837,  66.839,  69.737,  68.387,  60.354,  63.079,  69.853,  67.249,  62.123,  58.720,
  56.387,  58.723,  58.330,  55.364,  53.166,  51.794,  50.040,  48.220,  45.988,  44.101,
  42.023,  40.189,  38.659,  37.413,  36.674,  35.192,  32.457,  31.477,  29.697,  29.342,  26.724,
  25.180,  23.868,  22.371,  21.204,  20.685,  20.129,  18.840,  19.524,  19.120,  17.746,
  16.583,  16.814,  16.671,  16.845,  16.234,  15.017,  14.406,  13.788,  13.133,  12.493,
  11.596,  10.873,  10.574,  10.660,  10.041,   9.827,   9.390,   9.241,   7.697,   7.078,
   5.890,   5.192,   4.280,   3.648,   3.154,   2.920,   2.639,   1.926,   0.488,  -0.305,
  -0.308,  -0.954,  -2.217,  -3.550,  -4.904,  -6.138,  -7.277,  -8.404,  -9.147,  -9.728,
 -10.481, -11.058, -11.639, -12.113, -12.977, -13.313, -13.635, -14.029, -14.350, -14.538,
 -14.683, -14.625, -14.353, -14.139, -13.948, -13.659, -13.423, -12.607, -12.164, -11.817,
 -11.650, -11.545, -11.477, -11.676, -11.965, -12.298, -12.702, -13.086, -13.190, -13.357,
 -13.388, -13.307, -13.430, -13.562, -13.397, -13.135, -12.784, -12.309, -11.569, -10.699,
  -9.516,  -8.202,  -6.694,  -5.080,  -3.328,  -1.676,  -0.338,  -1.880,   2.812,   8.719,
  17.188,  19.485,  11.386,  12.133,  16.228,  20.244,  21.593,  23.261,  23.608,  22.700,
  22.146,  18.840,  16.129,  13.031,   9.434,   7.034,   4.341,   1.221,  -1.643,  -3.585,
  -6.177,  -9.437, -12.303, -15.656, -18.297, -19.747, -20.542, -20.913, -21.290, -22.124,
 -22.811, -23.608, -24.423, -24.563, -23.825, -23.111, -21.327, -18.431, -14.054, -11.337,
 -10.472,  -8.630,  -6.081,  -3.997,  -4.421,  -9.233,  -9.523,  -9.481,  -9.213,  -8.404,
  -7.481,  -7.042,  -5.715,  -4.727,  -3.576,  -2.327,  -0.538,   1.238,   3.297,   5.285,
   7.577,   9.479,  10.494,  12.112,  12.854,  14.096,  15.097,  15.771,  16.049,  16.405,
  16.388,  15.940,  15.426,  15.141,  15.376,  15.891,  16.130,  16.133,  16.103,  15.223},
{
  16.411,  15.290,  11.786,  10.888,   9.411,   8.648,   7.869,   7.189,   7.209,   7.301,
   7.983,   8.987,  10.655,   7.473,   5.594,   5.173,   3.669,   1.484,  -0.850,  -3.867,
  -6.429,  -7.626,  -9.741, -12.677, -14.581, -14.893, -15.095, -15.602, -15.123, -12.566,
 -13.173, -13.498, -15.470, -18.367, -18.226, -20.346, -19.732, -18.554, -23.038, -28.048,
 -31.845, -34.491, -37.493, -38.636, -39.337, -39.993, -40.966, -41.887, -41.520, -41.763,
 -40.838, -40.360, -40.366, -39.039, -41.172, -42.137, -43.421, -44.356, -45.144, -48.685,
 -51.340, -53.956, -56.312, -58.881, -61.426, -63.940, -65.977, -67.615, -69.343, -71.826,
 -75.750, -77.224, -75.888, -80.454, -85.512, -87.499, -88.267, -88.575, -89.255, -89.869,
 -89.854, -88.632, -86.640, -85.750, -84.199, -81.078, -76.844, -74.666, -71.452, -65.851,
 -63.258, -60.956, -56.490, -50.942, -47.612, -43.559, -37.328, -31.345, -27.070, -21.479,
 -16.738, -16.668, -12.496, -10.484,  -3.998,   7.611,  14.027,  18.627,  21.157,  23.896,
  26.294,  28.397,  31.102,  33.957,  37.091,  39.933,  42.250,  44.126,  46.371,  47.217,
  49.795,  49.011,  51.427,  52.077,  54.592,  55.078,  54.167,  54.362,  53.494,  52.540,
  50.106,  34.941,  43.563,  52.503,  57.547,  66.134,  68.022,  70.102,  70.066,  70.331,
  72.856,  74.572,  78.517,  81.409,  82.389,  80.519,  72.099,  71.121,  69.821,  74.951,
  81.327,  74.534,  60.484,  57.041,  56.169,  71.228,  64.831,  62.507,  60.192,  60.049,
  59.944,  58.325,  55.823,  54.283,  51.739,  49.155,  47.316,  45.332,  44.057,  42.292,
  40.402,  38.648,  37.200,  35.830,  34.075,  32.690,  32.334,  30.510,  28.292,  26.801,  26.036,
  23.810,  22.456,  21.178,  20.174,  19.179,  19.396,  17.900,  18.749,  17.557,  17.325,
  16.775,  16.870,  17.500,  16.701,  16.103,  15.914,  14.876,  14.756,  14.275,  13.020,
  12.215,  11.638,  11.770,  11.874,  11.107,  10.939,  10.494,  10.314,   9.627,   8.660,
   7.443,   6.852,   5.808,   5.380,   4.834,   4.203,   3.458,   2.644,   1.429,   0.625,
   0.158,  -1.207,  -2.442,  -3.725,  -5.059,  -6.316,  -7.272,  -7.996,  -8.933,  -9.540,
 -10.285, -11.082, -11.804, -12.682, -13.534, -14.076, -14.413, -14.601, -14.887, -15.233,
 -15.406, -15.662, -15.510, -15.432, -15.241, -14.905, -14.698, -14.144, -13.368, -13.067,
 -12.873, -12.771, -12.522, -12.613, -12.884, -13.060, -13.335, -13.727, -13.813, -13.822,
 -13.780, -13.824, -13.861, -13.814, -13.573, -13.345, -12.858, -12.316, -11.603, -10.555,
  -9.441,  -8.056,  -6.405,  -4.599,  -2.817,  -1.132,  -0.029,  -2.859,   8.417,   9.899,
  16.723,  18.818,  14.262,  12.603,  14.569,  17.763,  19.701,  22.206,  23.119,  22.652,
  21.279,  17.754,  14.371,  11.659,   7.671,   3.120,   0.799,  -1.586,  -4.513,  -6.176,
  -8.757, -11.423, -13.664, -16.756, -19.283, -20.649, -21.296, -21.514, -21.931, -22.878,
 -23.608, -24.334, -25.519, -25.385, -23.819, -22.620, -21.506, -18.954, -14.741, -11.843,
 -10.619,  -7.548,  -6.511,  -5.013,  -5.390,  -8.466,  -7.986,  -8.100,  -7.653,  -6.944,
  -6.490,  -5.653,  -4.672,  -3.468,  -2.084,  -0.658,   0.709,   2.484,   4.038,   5.647,
   8.192,  10.125,  11.721,  12.870,  13.981,  15.158,  16.454,  17.464,  17.934,  17.395,
  17.018,  16.263,  15.868,  15.922,  15.591,  15.900,  16.750,  17.126,  17.129,  16.411},
{
  16.838,  15.546,  13.499,  11.683,  10.193,   9.241,   8.443,   7.625,   7.341,   7.121,
   5.531,   5.623,   4.414,   4.232,   3.099,   2.021,   0.299,  -2.369,  -5.602,  -8.310,
  -9.731, -11.797, -13.555, -15.005, -15.899, -16.502, -16.427, -16.321, -15.092, -13.284,
 -14.782, -14.663, -16.077, -19.163, -20.154, -18.771, -18.800, -17.816, -22.755, -26.906,
 -29.931, -34.426, -37.531, -39.603, -40.713, -41.755, -42.039, -42.893, -43.280, -43.094,
 -42.761, -42.082, -41.775, -41.361, -43.127, -41.229, -41.306, -44.408, -48.705, -51.286,
 -53.947, -56.273, -58.929, -61.446, -64.283, -66.487, -68.638, -70.731, -72.066, -74.809,
 -78.454, -81.347, -81.293, -83.587, -88.384, -90.014, -91.128, -91.327, -91.830, -93.125,
 -92.927, -91.684, -88.421, -86.704, -86.333, -83.290, -78.733, -75.978, -72.200, -67.237,
 -64.339, -60.455, -55.814, -50.417, -45.640, -41.506, -34.863, -29.658, -24.154, -19.172,
 -18.756, -14.938,  -8.891,  -3.999,   7.620,  13.225,  15.950,  19.670,  22.874,  25.614,
  27.840,  30.460,  32.988,  36.105,  39.504,  42.651,  45.070,  46.913,  47.114,  50.277,
  53.974,  53.601,  57.115,  57.074,  55.795,  56.293,  57.506,  57.946,  57.663,  55.787,
  54.817,  43.155,  48.859,  58.730,  57.415,  67.526,  67.518,  69.888,  69.257,  71.609,
  75.784,  81.112,  83.627,  80.506,  76.693,  70.403,  68.271,  73.513,  78.240,  80.377,
  81.823,  79.819,  74.007,  57.793,  59.515,  63.989,  65.911,  65.782,  62.878,  61.145,
  59.360,  56.819,  54.396,  51.615,  49.576,  47.433,  45.297,  43.470,  42.116,  40.909,
  39.458,  37.120,  35.050,  33.873,  32.702,  31.141,  29.161,  28.448,  27.188,  25.594,  24.356,
  22.789,  21.480,  20.499,  19.577,  18.823,  18.257,  17.942,  18.350,  17.163,  16.757,
  16.807,  16.902,  17.006,  16.819,  16.966,  16.460,  16.287,  14.869,  14.340,  14.005,
  13.409,  12.867,  12.685,  12.521,  12.167,  11.850,  11.838,  10.764,  10.770,  10.118,
   9.304,   8.235,   7.783,   6.688,   5.750,   4.823,   3.770,   3.065,   1.979,   0.864,
  -0.136,  -1.239,  -2.530,  -3.881,  -5.237,  -6.231,  -7.126,  -8.084,  -9.093,  -9.926,
 -10.732, -11.623, -12.320, -13.192, -13.805, -14.416, -14.830, -15.164, -15.728, -16.281,
 -16.741, -16.766, -16.623, -16.672, -16.473, -16.210, -15.821, -15.534, -14.871, -14.297,
 -14.028, -13.846, -13.693, -13.530, -13.636, -13.773, -13.890, -14.083, -14.046, -14.278,
 -14.086, -13.975, -14.021, -13.823, -13.628, -13.245, -12.735, -12.075, -11.268, -10.206,
  -9.088,  -7.635,  -6.009,  -4.098,  -2.270,  -0.436,   0.808,  -1.952,  11.235,  16.603,
  18.323,  11.432,  10.401,  11.414,  14.610,  18.557,  20.504,  21.102,  21.533,  19.600,
  18.738,  15.149,  12.535,   9.912,   6.069,   0.157,  -3.252,  -4.368,  -5.922,  -7.632,
 -10.613, -13.434, -15.777, -18.469, -20.161, -20.933, -21.588, -21.859, -22.295, -23.737,
 -24.569, -24.843, -26.300, -26.404, -25.313, -23.000, -21.562, -19.274, -15.417, -12.337,
  -9.647,  -8.010,  -7.169,  -6.318,  -6.529,  -8.434,  -8.002,  -7.247,  -6.326,  -5.169,
  -4.784,  -4.184,  -3.219,  -1.805,  -0.972,   0.108,   1.687,   3.120,   4.860,   6.666,
   8.790,  10.677,  12.404,  13.915,  15.013,  16.146,  17.465,  18.239,  18.416,  18.339,
  18.187,  17.780,  17.529,  17.064,  16.836,  16.583,  16.570,  17.037,  17.396,  16.838},
{
  17.664,  16.568,  15.282,  13.659,  11.846,  10.701,   9.403,   8.341,   7.621,   4.554,
   5.286,   5.295,   2.715,   3.476,   2.861,   1.060,  -2.046,  -5.508,  -8.504, -10.484,
 -12.792, -14.364, -15.184, -16.878, -17.661, -18.094, -17.525, -16.492, -14.994, -12.341,
 -11.881, -14.623, -16.724, -19.256, -20.437, -19.232, -19.140, -17.978, -20.729, -24.208,
 -31.048, -32.783, -37.061, -39.769, -41.262, -42.903, -43.625, -44.186, -44.276, -44.221,
 -44.190, -43.787, -44.105, -44.041, -43.612, -41.576, -43.800, -48.907, -50.864, -53.542,
 -55.843, -58.690, -61.316, -64.132, -66.604, -68.608, -70.540, -73.041, -74.815, -77.452,
 -81.139, -84.177, -86.235, -86.799, -90.424, -92.178, -93.285, -93.943, -94.456, -95.162,
 -93.944, -94.161, -90.747, -87.372, -87.065, -84.832, -80.315, -76.313, -72.597, -66.335,
 -62.878, -60.124, -55.149, -49.342, -43.999, -39.064, -33.100, -27.664, -21.754, -18.623,
 -18.736, -10.372,  -2.837,   7.362,  11.303,  14.182,  17.187,  21.365,  23.886,  26.960,
  29.368,  31.965,  35.317,  38.694,  41.982,  43.891,  48.119,  49.921,  48.083,  51.002,
  56.689,  57.393,  60.801,  59.324,  55.996,  58.451,  58.875,  58.806,  59.135,  59.148,
  57.761,  56.643,  54.477,  65.159,  66.820,  70.023,  75.594,  80.779,  82.330,  81.022,
  77.301,  78.736,  78.267,  76.549,  74.999,  75.182,  77.267,  79.020,  80.567,  80.531,
  79.809,  77.861,  74.172,  65.784,  63.466,  63.228,  64.504,  63.718,  61.193,  58.848,
  56.229,  54.136,  51.743,  49.668,  47.803,  45.390,  43.624,  42.600,  40.839,  39.682,
  37.951,  35.379,  34.067,  32.251,  30.394,  28.908,  28.798,  26.606,  26.369,  25.387,  23.801,
  21.709,  20.587,  19.698,  19.038,  18.714,  18.157,  18.228,  17.949,  17.773,  16.168,
  16.640,  16.701,  16.237,  16.847,  17.106,  16.573,  15.604,  15.133,  14.254,  13.950,
  14.281,  13.715,  13.999,  13.587,  13.703,  12.854,  12.421,  12.712,  12.131,  11.648,
  10.856,   9.939,   8.928,   7.545,   6.742,   5.587,   4.330,   3.144,   2.207,   0.920,
  -0.279,  -1.536,  -2.885,  -4.075,  -5.238,  -6.037,  -7.384,  -8.614,  -9.368, -10.434,
 -11.473, -12.158, -12.851, -13.802, -14.684, -15.314, -15.852, -16.428, -17.159, -17.514,
 -17.882, -17.978, -18.082, -17.960, -17.760, -17.508, -17.112, -16.669, -16.357, -15.893,
 -15.549, -15.294, -14.985, -14.729, -14.563, -14.512, -14.455, -14.443, -14.441, -14.309,
 -14.132, -14.074, -13.834, -13.692, -13.271, -12.793, -12.279, -11.484, -10.587,  -9.495,
  -8.385,  -6.983,  -5.231,  -3.439,  -1.485,   0.539,   1.748,   0.079,   8.132,  15.855,
  17.178,  12.875,  10.194,  11.955,  15.957,  19.968,  21.319,  20.848,  20.274,  18.205,
  16.266,  12.639,   9.807,   7.669,   3.609,  -1.706,  -5.498,  -5.696,  -6.824,  -7.818,
  -9.710, -11.938, -16.489, -19.917, -20.145, -20.016, -21.195, -22.091, -22.536, -24.189,
 -25.425, -24.685, -26.418, -27.480, -26.465, -23.425, -21.058, -19.115, -14.934, -13.172,
  -9.543,  -8.477,  -9.459,  -7.717,  -8.171,  -7.082,  -5.955,  -5.839,  -5.535,  -3.943,
  -3.461,  -3.149,  -1.930,  -0.873,   0.189,   1.350,   2.823,   4.564,   5.753,   7.475,
   9.719,  11.290,  12.822,  14.456,  15.841,  17.024,  18.290,  19.065,  18.986,  18.737,
  18.560,  18.277,  18.201,  17.996,  17.802,  17.722,  17.661,  18.047,  18.050,  17.664},
{
  18.217,  18.234,  17.237,  15.487,  14.082,  12.881,  11.327,   9.911,   7.748,   7.567,
   7.046,   4.294,   3.764,   4.735,   4.359,   2.143,  -2.433,  -7.503, -10.739, -13.130,
 -15.168, -14.966, -18.309, -19.555, -18.931, -19.400, -18.054, -16.417, -13.950,  -9.406,
  -8.983, -12.763, -16.437, -18.277, -18.512, -17.751, -18.337, -16.080, -19.503, -24.984,
 -29.865, -31.462, -37.760, -39.880, -40.740, -43.223, -44.925, -45.181, -46.008, -45.848,
 -45.374, -44.867, -45.486, -46.327, -46.116, -47.266, -49.437, -51.196, -53.103, -55.358,
 -57.878, -61.270, -63.870, -66.411, -68.682, -71.032, -72.950, -74.888, -77.136, -80.037,
 -83.794, -86.605, -88.816, -90.089, -91.955, -94.888, -96.114, -97.003, -97.266, -97.585,
 -97.704, -95.853, -92.777, -88.249, -88.920, -86.689, -81.844, -77.253, -72.893, -67.477,
 -62.991, -59.037, -54.041, -47.802, -42.601, -36.975, -31.302, -26.676, -21.444, -22.610,
 -15.032,  -9.765,   4.160,   8.368,  12.084,  14.803,  18.286,  21.773,  25.062,  28.112,
  31.392,  33.964,  37.544,  40.600,  43.132,  45.119,  50.100,  52.559,  53.708,  56.046,
  60.713,  61.225,  61.360,  57.860,  57.871,  59.767,  60.048,  58.785,  57.934,  59.713,
  59.909,  58.588,  63.454,  68.542,  68.675,  68.664,  69.122,  76.710,  78.075,  76.423,
  77.238,  77.573,  77.106,  77.502,  78.182,  78.842,  79.405,  79.868,  79.443,  78.162,
  77.093,  75.505,  71.220,  67.300,  61.892,  63.412,  62.746,  61.044,  58.578,  56.159,
  53.778,  51.565,  49.651,  47.709,  44.768,  43.547,  42.011,  40.416,  39.489,  38.492,
  36.707,  34.070,  32.765,  31.793,  29.551,  27.266,  26.094,  25.863,  25.149,  24.311,  23.029,
  21.359,  20.163,  19.895,  18.959,  18.782,  18.001,  17.997,  18.014,  17.359,  16.059,
  16.367,  16.297,  15.845,  15.783,  15.492,  15.242,  15.178,  14.939,  14.566,  14.373,
  14.470,  14.750,  14.580,  14.443,  13.973,  13.774,  13.782,  13.677,  12.738,  12.012,
  11.169,  10.180,   9.102,   8.041,   7.097,   5.778,   4.561,   3.574,   2.519,   1.181,
  -0.065,  -1.343,  -2.776,  -4.005,  -5.351,  -6.800,  -7.536,  -8.559,  -9.643, -10.838,
 -11.667, -12.735, -13.948, -15.082, -15.948, -16.760, -17.214, -17.814, -18.307, -18.840,
 -19.362, -19.475, -19.571, -19.458, -19.303, -18.879, -18.418, -17.943, -17.545, -17.111,
 -16.830, -16.414, -16.136, -15.843, -15.559, -15.329, -15.090, -14.868, -14.551, -14.286,
 -14.075, -13.854, -13.507, -13.200, -12.796, -12.233, -11.520, -10.685,  -9.538,  -8.373,
  -7.312,  -5.987,  -4.311,  -2.503,  -0.465,   1.537,   3.771,   3.155,   4.963,  10.183,
  22.230,  16.238,   9.929,  14.172,  18.079,  20.343,  19.790,  19.189,  17.787,  16.642,
  14.310,  10.374,   8.350,   6.005,   2.003,  -1.729,  -4.423,  -7.226,  -9.328, -11.888,
 -14.652, -14.737, -14.881, -18.278, -17.871, -18.540, -20.119, -22.276, -22.173, -23.806,
 -24.954, -26.061, -26.432, -26.624, -26.553, -23.525, -22.151, -20.369, -15.815, -13.081,
 -10.781, -11.023, -10.790, -10.049,  -8.818,  -8.069,  -6.592,  -5.631,  -4.757,  -3.733,
  -2.782,  -2.009,  -0.806,   0.456,   1.408,   2.332,   3.546,   4.758,   6.266,   8.000,
   9.616,  11.319,  13.057,  14.709,  16.153,  17.603,  18.768,  19.707,  19.240,  19.159,
  18.969,  18.813,  18.568,  18.214,  17.811,  17.607,  17.673,  17.951,  18.272,  18.217},
{
  18.196,  18.361,  17.811,  16.908,  16.287,  15.101,  13.967,  12.455,  11.371,  10.487,
   6.526,   5.356,   7.300,   6.980,   4.401,   1.539,  -3.124,  -7.303, -11.015, -14.556,
 -17.660, -19.525, -20.584, -20.833, -20.582, -20.283, -18.837, -17.178, -13.702,  -6.610,
  -8.357, -12.163, -16.292, -17.536, -18.401, -16.910, -18.877, -15.433, -20.029, -24.951,
 -28.786, -30.646, -37.644, -40.519, -41.072, -43.700, -45.387, -46.922, -47.171, -47.423,
 -46.812, -46.313, -47.305, -47.989, -47.961, -49.195, -51.123, -53.113, -55.086, -57.113,
 -60.110, -63.215, -65.759, -68.505, -70.516, -72.385, -74.757, -76.814, -79.058, -82.503,
 -85.890, -89.003, -91.174, -91.856, -92.862, -97.331, -97.538, -99.249, -99.953,-100.841,
-100.371, -98.091, -94.397, -92.159, -90.359, -87.100, -82.682, -78.327, -72.327, -67.706,
 -63.940, -58.985, -53.731, -47.166, -41.872, -35.570, -30.139, -26.179, -24.639, -17.362,
 -15.937,  -3.401,   5.147,   7.677,  11.646,  15.268,  18.841,  22.295,  25.654,  29.281,
  32.590,  35.619,  39.378,  42.422,  45.144,  47.592,  51.129,  53.354,  51.171,  56.867,
  63.793,  64.492,  61.228,  61.451,  61.181,  62.223,  63.931,  61.523,  59.343,  63.761,
  68.127,  68.866,  70.554,  70.688,  69.716,  72.121,  72.104,  73.647,  74.879,  73.707,
  71.214,  71.138,  73.477,  76.485,  77.568,  78.306,  77.637,  77.597,  76.228,  73.807,
  72.988,  70.092,  66.444,  63.264,  61.641,  61.503,  61.083,  59.583,  56.886,  54.337,
  51.263,  49.217,  47.911,  45.912,  43.996,  42.132,  40.630,  39.597,  38.205,  36.821,
  34.663,  33.037,  32.314,  29.532,  27.277,  27.506,  27.087,  24.518,  24.097,  22.785,  21.880,
  20.611,  19.700,  19.261,  18.717,  19.481,  17.454,  17.491,  16.513,  15.771,  16.149,
  16.419,  15.494,  15.298,  15.229,  15.030,  15.090,  14.959,  14.418,  14.688,  15.265,
  15.386,  15.380,  14.840,  15.044,  14.900,  14.778,  14.222,  13.767,  12.807,  12.172,
  11.519,  10.521,   9.654,   8.693,   7.663,   6.482,   5.399,   4.342,   2.755,   1.383,
  -0.081,  -1.569,  -3.225,  -4.417,  -5.737,  -6.785,  -7.629,  -9.032, -10.253, -11.490,
 -12.599, -13.792, -14.866, -16.002, -17.054, -17.834, -18.543, -19.354, -20.026, -20.408,
 -20.786, -21.195, -21.208, -20.993, -20.716, -20.248, -19.735, -19.267, -18.840, -18.397,
 -17.893, -17.488, -17.065, -16.662, -16.326, -16.047, -15.639, -15.216, -14.720, -14.300,
 -13.916, -13.448, -13.019, -12.567, -12.084, -11.460, -10.582,  -9.677,  -9.244,  -7.978,
  -6.571,  -5.169,  -3.510,  -1.781,   0.297,   2.643,   4.997,   6.263,  10.866,  18.492,
  23.432,  23.414,  13.751,  15.982,  18.872,  19.527,  18.825,  17.022,  15.527,  14.994,
  13.608,  10.136,   6.889,   4.862,   0.894,  -2.179,  -4.920,  -7.397,  -9.462, -12.022,
 -15.147, -16.386, -17.194, -18.681, -22.341, -21.642, -21.686, -22.717, -23.051, -23.941,
 -24.536, -25.205, -24.877, -25.708, -26.351, -22.296, -20.269, -19.035, -16.894, -15.757,
 -15.850, -12.906, -12.360, -11.360,  -9.532,  -8.198,  -7.083,  -5.986,  -4.480,  -3.527,
  -2.458,  -0.894,   0.783,   2.359,   4.183,   5.317,   5.985,   7.557,   8.368,   9.837,
  11.878,  13.057,  14.204,  15.349,  17.002,  18.239,  19.561,  19.983,  19.876,  19.729,
  19.487,  19.166,  18.815,  18.227,  17.727,  17.593,  17.512,  17.838,  18.391,  18.196},
{
  17.846,  17.446,  17.591,  17.157,  16.712,  16.232,  15.677,  14.390,  14.324,  12.113,
   9.303,   8.141,   8.246,   6.231,   3.820,   0.082,  -4.257,  -7.689, -13.061, -14.998,
 -18.760, -20.620, -21.231, -21.359, -21.303, -19.783, -18.437, -16.259, -11.885,  -7.340,
  -8.251, -12.372, -14.464, -16.564, -17.429, -15.285, -14.724, -15.681, -19.718, -24.078,
 -28.080, -31.366, -33.533, -39.688, -41.620, -42.830, -44.924, -47.533, -48.813, -48.425,
 -48.509, -48.051, -48.442, -49.096, -49.507, -50.654, -52.706, -54.413, -56.289, -58.733,
 -61.738, -64.281, -67.162, -69.944, -72.331, -74.307, -76.592, -78.248, -81.066, -84.633,
 -87.840, -90.757, -93.646, -93.275, -94.580, -99.292, -99.216,-100.363,-101.061,-101.820,
-101.251,-100.094, -97.215, -94.597, -91.795, -88.198, -83.855, -79.340, -73.942, -67.460,
 -62.219, -58.800, -53.537, -46.726, -40.566, -34.657, -29.434, -27.047, -22.571, -16.648,
 -11.869,  -0.385,   3.317,   7.506,  10.955,  15.029,  18.929,  22.032,  26.036,  30.053,
  33.751,  36.931,  41.254,  45.283,  48.242,  50.195,  53.740,  53.783,  53.808,  55.312,
  61.864,  61.140,  64.379,  60.564,  50.058,  43.752,  45.464,  54.522,  67.793,  70.802,
  72.366,  73.940,  76.510,  79.491,  77.891,  75.191,  74.609,  72.421,  69.956,  69.321,
  72.106,  72.791,  73.389,  72.873,  73.336,  73.276,  72.968,  72.313,  71.140,  68.897,
  64.406,  61.828,  60.487,  59.829,  58.955,  59.566,  59.227,  57.938,  55.530,  53.078,
  50.288,  47.916,  45.590,  44.109,  42.369,  41.082,  39.454,  38.870,  37.101,  35.205,
  33.372,  32.027,  30.597,  27.757,  28.364,  27.506,  26.363,  23.842,  23.634,  22.056,  20.980,
  20.329,  20.124,  18.597,  18.842,  16.950,  17.191,  17.012,  16.124,  15.832,  15.743,
  14.855,  14.319,  15.740,  14.891,  14.487,  15.010,  14.825,  14.791,  15.267,  15.314,
  15.473,  15.820,  15.763,  15.757,  15.549,  15.148,  14.342,  13.642,  12.875,  12.275,
  11.572,  11.127,  10.430,   9.239,   7.946,   6.566,   5.117,   3.639,   2.103,   1.014,
  -0.267,  -1.702,  -3.270,  -4.831,  -6.150,  -7.393,  -8.605,  -9.797, -10.864, -12.208,
 -13.494, -14.521, -15.866, -17.083, -18.055, -18.988, -20.016, -20.679, -21.353, -21.923,
 -22.246, -22.468, -22.508, -22.460, -22.122, -21.618, -20.967, -20.510, -19.966, -19.485,
 -18.921, -18.385, -17.904, -17.468, -16.988, -16.560, -16.075, -15.503, -14.877, -14.303,
 -13.668, -13.084, -12.539, -11.986, -11.499, -10.762,  -9.851,  -9.599,  -5.803,  -3.595,
  -3.608,  -3.112,  -2.271,  -1.098,   0.879,   3.536,   6.742,   8.896,  10.032,  15.868,
  23.746,  23.147,  14.986,  17.122,  18.671,  19.657,  19.055,  16.792,  15.475,  14.154,
  12.057,   8.408,   5.051,   2.122,  -0.655,  -3.385,  -5.905,  -7.530,  -9.076, -11.809,
 -15.125, -17.868, -20.209, -21.158, -23.581, -24.084, -23.285, -23.579, -23.019, -24.958,
 -25.035, -25.323, -25.388, -23.806, -22.262, -22.190, -22.839, -21.751, -19.791, -17.866,
 -16.009, -14.071, -12.135,  -9.598,  -7.607,  -5.808,  -4.245,  -2.454,  -0.562,   1.257,
   3.586,   4.640,   5.741,   5.933,   6.732,   7.485,   8.494,   8.309,   8.751,  10.496,
  13.079,  15.272,  16.564,  18.233,  17.821,  18.509,  19.479,  19.721,  20.004,  19.993,
  19.891,  19.564,  19.238,  18.769,  18.252,  17.821,  17.663,  17.630,  17.622,  17.846},
{
  17.162,  16.997,  16.757,  16.942,  16.617,  16.254,  16.804,  16.331,  13.662,  12.343,
   9.000,   9.154,   7.836,   6.074,   3.243,  -1.156,  -3.744,  -8.208, -11.410, -15.079,
 -17.336, -19.364, -19.952, -20.823, -20.906, -19.028, -17.391, -15.482, -11.529,  -6.690,
  -9.817, -11.860, -12.084, -14.069, -16.410, -16.299, -14.050, -12.102, -19.258, -22.920,
 -26.981, -30.343, -32.985, -35.341, -41.311, -43.372, -45.018, -47.756, -49.225, -49.549,
 -48.937, -49.168, -49.270, -50.034, -50.660, -51.707, -52.893, -55.396, -57.721, -60.083,
 -62.749, -65.524, -67.895, -70.659, -73.069, -75.251, -77.714, -79.801, -82.858, -86.464,
 -89.475, -92.225, -94.834, -94.642, -96.448,-100.316,-100.655,-102.115,-102.606,-102.894,
-102.688,-101.659, -98.650, -94.316, -92.866, -88.790, -84.087, -79.038, -73.734, -68.665,
 -63.236, -58.098, -53.021, -46.945, -39.827, -34.089, -29.623, -29.074, -19.016, -12.513,
  -7.164,  -1.964,   1.980,   6.350,  10.414,  14.418,  18.051,  22.043,  26.013,  30.414,
  33.904,  37.184,  40.794,  45.363,  50.179,  54.232,  54.842,  54.745,  52.813,  53.638,
  58.348,  61.248,  63.128,  56.136,  54.811,  50.794,  46.844,  68.556,  72.155,  73.505,
  76.604,  76.337,  72.741,  73.762,  73.017,  71.986,  67.969,  68.395,  70.105,  71.190,
  72.228,  72.307,  72.288,  72.123,  70.758,  69.202,  67.927,  66.041,  65.546,  62.062,
  63.325,  61.307,  59.258,  58.042,  58.737,  59.055,  58.609,  56.400,  54.109,  51.697,
  49.257,  47.442,  44.768,  43.543,  42.194,  40.622,  39.076,  37.882,  36.155,  33.833,
  31.946,  31.137,  28.620,  27.944,  27.546,  25.349,  24.657,  23.382,  22.755,  21.911,  21.153,
  20.051,  19.691,  18.746,  18.712,  17.599,  17.179,  16.337,  15.212,  15.189,  15.218,
  14.624,  14.752,  14.615,  14.374,  15.143,  15.902,  15.523,  15.915,  15.603,  16.312,
  16.641,  16.411,  16.385,  16.625,  15.697,  14.931,  14.364,  14.019,  13.457,  12.727,
  11.715,  11.118,  10.061,   8.958,   7.634,   6.334,   4.881,   3.439,   2.082,   0.676,
  -0.897,  -2.423,  -4.116,  -5.766,  -6.863,  -8.364,  -9.435, -10.579, -11.755, -13.154,
 -14.495, -15.677, -16.969, -18.265, -19.467, -20.459, -21.303, -22.125, -22.783, -23.146,
 -23.593, -23.949, -23.996, -23.817, -23.417, -22.804, -22.163, -21.686, -21.099, -20.424,
 -19.809, -19.172, -18.719, -18.191, -17.555, -16.983, -16.465, -15.729, -14.970, -14.238,
 -13.482, -12.770, -12.028, -11.297, -10.732,  -9.966,  -8.793,  -8.208,  -5.959,  -4.286,
  -3.592,  -2.634,  -1.592,  -0.431,   0.898,   3.025,   6.166,   9.560,   8.863,  14.391,
  22.030,  24.530,  17.319,  18.048,  19.591,  20.685,  18.982,  18.562,  16.978,  13.712,
  10.630,   6.960,   3.378,   0.443,  -2.263,  -4.676,  -6.858,  -8.360,  -9.969, -12.627,
 -15.817, -18.872, -21.303, -23.068, -24.915, -25.554, -24.762, -24.158, -23.655, -25.693,
 -26.783, -26.444, -26.060, -24.872, -23.154, -24.792, -24.768, -23.138, -20.176, -18.568,
 -16.488, -14.357, -12.448, -10.476,  -8.434,  -6.544,  -3.870,  -1.791,   0.420,   2.156,
   4.282,   6.024,   7.598,   8.608,  10.096,  10.195,  10.181,  10.654,  11.591,  12.605,
  12.859,  14.834,  16.515,  18.261,  18.763,  19.316,  19.476,  19.001,  19.126,  19.457,
  19.540,  19.331,  19.426,  19.466,  18.851,  18.106,  17.740,  17.469,  17.305,  17.162},
{
  16.992,  16.742,  16.210,  16.466,  16.232,  16.002,  15.432,  15.955,  14.059,  13.445,
  11.431,  10.954,   9.161,   6.566,   3.743,  -0.314,  -2.665,  -6.206, -10.020, -12.667,
 -15.490, -17.872, -19.080, -18.733, -18.880, -17.853, -16.123, -14.365, -12.869, -11.415,
 -12.472, -13.023, -13.583, -13.704, -15.725, -12.138, -15.821, -14.132, -19.298, -23.673,
 -26.236, -29.016, -31.161, -33.241, -36.186, -41.997, -44.108, -47.352, -49.776, -50.778,
 -49.266, -49.990, -50.387, -50.927, -51.800, -52.752, -53.800, -55.727, -58.203, -60.824,
 -63.573, -65.831, -68.247, -70.696, -73.337, -76.030, -78.475, -80.621, -84.430, -87.922,
 -91.154, -94.021, -95.748, -95.694, -97.359,-100.195,-101.565,-102.842,-103.495,-103.375,
-103.379,-102.479,-100.444, -96.088, -94.863, -91.094, -85.904, -79.744, -73.607, -67.821,
 -61.227, -57.573, -52.372, -46.517, -40.257, -35.312, -32.145, -25.725, -18.904, -11.930,
  -6.713,  -3.164,   1.433,   4.482,   8.552,  12.950,  16.916,  21.201,  25.156,  28.755,
  32.824,  36.720,  40.681,  44.526,  49.606,  54.504,  58.143,  57.704,  57.685,  55.995,
  56.354,  64.002,  64.533,  63.179,  66.814,  67.688,  58.459,  65.387,  71.211,  72.402,
  73.326,  74.328,  74.425,  74.390,  73.115,  70.967,  70.649,  70.858,  70.796,  70.927,
  70.936,  70.894,  70.667,  70.214,  69.128,  68.039,  66.999,  66.345,  65.905,  62.435,
  62.227,  59.178,  57.633,  56.552,  57.493,  57.469,  56.769,  54.978,  53.271,  50.421,
  48.401,  46.375,  44.695,  43.089,  41.613,  40.267,  38.658,  37.030,  35.176,  32.923,
  31.118,  30.897,  28.148,  29.151,  25.843,  25.890,  24.025,  22.696,  21.305,  20.382,  19.198,
  19.886,  19.297,  19.359,  17.902,  17.187,  16.540,  15.591,  14.987,  14.651,  14.552,
  14.270,  14.425,  14.283,  14.657,  14.968,  15.114,  15.605,  15.846,  15.934,  16.111,
  16.859,  17.205,  17.445,  16.294,  15.622,  15.158,  14.675,  14.068,  13.413,  12.345,
  11.399,  10.553,   9.706,   8.227,   6.932,   5.786,   4.506,   2.984,   1.267,  -0.453,
  -1.980,  -3.540,  -5.095,  -6.622,  -7.850,  -9.232, -10.589, -11.800, -13.054, -14.529,
 -15.736, -16.847, -18.092, -19.324, -20.578, -21.781, -22.708, -23.554, -24.259, -24.831,
 -25.154, -25.249, -25.318, -25.183, -24.721, -24.144, -23.507, -22.925, -22.152, -21.439,
 -20.652, -20.223, -19.390, -18.601, -17.942, -17.352, -16.620, -15.842, -15.031, -14.277,
 -13.417, -12.539, -11.693, -10.851, -10.141,  -8.863,  -7.343,  -6.204,  -4.563,  -3.685,
  -2.755,  -1.796,  -0.733,   0.610,   1.944,   3.455,   5.724,   8.454,   9.541,  12.117,
  17.646,  27.656,  24.490,  19.095,  19.666,  22.215,  20.860,  20.357,  17.421,  13.683,
  10.283,   6.100,   1.808,   1.153,  -1.309,  -5.896,  -7.891,  -9.232, -11.327, -14.435,
 -16.977, -20.275, -23.003, -24.939, -26.435, -27.553, -27.578, -26.516, -25.561, -26.828,
 -28.145, -27.257, -26.085, -26.482, -28.256, -26.740, -24.501, -22.110, -19.889, -17.689,
 -15.287, -12.824, -10.507,  -8.149,  -5.794,  -3.792,  -1.163,   1.228,   3.559,   5.047,
   6.159,   7.128,   8.308,   8.974,   9.890,  10.279,  10.879,  11.935,  12.780,  12.554,
  12.229,  12.260,  14.631,  16.228,  16.660,  17.207,  17.236,  17.587,  18.726,  19.474,
  19.937,  19.677,  19.142,  18.663,  18.444,  18.063,  17.714,  17.585,  17.384,  16.992},
{
  16.895,  16.948,  16.158,  15.867,  15.606,  15.280,  15.102,  15.972,  14.148,  13.439,
  12.674,  11.653,   9.427,   7.388,   4.407,   0.287,  -0.843,  -3.087,  -6.514, -10.069,
 -14.135, -15.890, -17.552, -15.890, -15.666, -14.730, -13.774, -13.083, -12.702, -11.423,
 -10.932, -12.620, -14.285, -14.430, -15.754, -15.343, -16.531, -16.114, -19.739, -22.748,
 -24.565, -27.679, -29.548, -30.366, -32.683, -36.292, -40.355, -45.566, -49.929, -51.348,
 -49.658, -50.323, -50.861, -52.064, -52.594, -53.873, -55.300, -56.472, -58.428, -61.198,
 -63.488, -65.849, -68.303, -70.859, -73.749, -76.114, -78.510, -81.418, -85.488, -88.925,
 -92.300, -95.032, -96.847, -96.608, -98.589,-100.405,-102.393,-103.651,-103.671,-103.521,
-103.731,-103.170,-101.749, -99.244, -95.477, -89.939, -84.754, -80.258, -73.433, -66.477,
 -60.768, -56.460, -51.324, -47.350, -43.444, -38.876, -32.076, -27.033, -21.020, -10.454,
  -6.779,  -3.932,  -0.163,   3.779,   7.551,  11.715,  15.413,  19.845,  23.070,  26.490,
  29.999,  34.325,  38.076,  41.257,  45.141,  53.321,  56.205,  57.832,  57.086,  53.885,
  54.462,  58.716,  58.998,  60.424,  64.072,  70.406,  62.148,  56.185,  68.269,  69.304,
  72.358,  74.110,  74.322,  74.306,  72.654,  71.584,  71.088,  70.471,  70.092,  69.740,
  69.250,  69.541,  69.235,  68.123,  66.941,  66.229,  65.612,  65.300,  64.155,  61.957,
  59.971,  57.008,  55.761,  56.291,  56.181,  56.271,  54.858,  54.078,  51.907,  50.306,
  47.325,  45.855,  44.257,  42.095,  40.413,  39.129,  37.765,  35.819,  33.890,  31.698,
  30.687,  29.885,  27.180,  28.349,  25.686,  24.541,  23.315,  21.228,  20.816,  19.844,  19.200,
  19.033,  18.784,  18.039,  17.579,  17.044,  15.918,  15.364,  14.793,  13.859,  13.856,
  14.320,  13.939,  13.999,  14.207,  14.630,  15.291,  15.764,  15.321,  15.959,  16.739,
  17.355,  18.362,  17.966,  16.306,  15.392,  14.105,  13.807,  13.788,  12.880,  11.553,
  10.539,   9.560,   8.614,   7.390,   6.383,   5.020,   3.636,   1.949,   0.281,  -1.038,
  -2.515,  -4.048,  -5.688,  -7.518,  -9.169, -10.502, -11.745, -12.979, -14.350, -15.669,
 -16.911, -18.241, -19.535, -20.791, -22.112, -23.202, -24.259, -25.060, -25.616, -26.079,
 -26.543, -26.658, -26.636, -26.431, -25.906, -25.250, -24.611, -23.929, -23.150, -22.526,
 -21.779, -20.973, -20.051, -19.079, -18.231, -17.591, -16.783, -15.850, -14.891, -13.753,
 -12.809, -11.538, -10.440,  -9.503,  -8.532,  -7.355,  -6.203,  -4.958,  -3.925,  -2.850,
  -1.956,  -0.990,   0.092,   1.354,   2.887,   4.638,   6.504,   9.042,  11.276,  10.550,
  12.585,  19.544,  26.364,  24.361,  19.699,  20.394,  20.087,  17.544,  13.822,  10.702,
   7.759,   4.223,  -0.146,  -1.367,  -4.061,  -6.485,  -8.608, -10.063, -11.925, -15.525,
 -18.970, -21.702, -24.060, -25.949, -27.652, -29.007, -29.336, -28.659, -28.391, -28.634,
 -29.402, -28.446, -29.960, -31.513, -29.277, -26.675, -24.361, -21.956, -19.803, -17.400,
 -15.162, -12.692,  -9.909,  -7.738,  -5.174,  -2.403,   0.184,   2.652,   5.127,   6.512,
   7.473,   8.428,   9.163,   9.665,  10.445,  10.837,  11.235,  11.541,  11.669,  12.141,
  12.792,  13.140,  13.664,  14.394,  14.846,  15.959,  17.973,  20.126,  19.963,  19.558,
  18.622,  18.360,  18.284,  18.188,  19.283,  19.028,  17.827,  17.001,  16.676,  16.895},
{
  16.616,  16.553,  15.951,  15.076,  14.330,  15.548,  17.403,  17.036,  15.824,  15.328,
  14.569,  13.161,  11.343,   8.256,   5.894,   3.495,   1.096,  -0.471,  -3.274,  -6.918,
 -10.131, -12.894, -13.627, -12.937, -12.359, -11.730, -11.500, -11.617, -11.562, -10.876,
 -11.434, -12.937, -14.064, -13.163, -13.252, -15.407, -16.509, -17.511, -19.611, -21.455,
 -23.769, -26.704, -29.229, -30.012, -31.031, -34.601, -37.364, -40.641, -47.004, -49.509,
 -50.201, -51.695, -50.847, -52.513, -52.543, -53.814, -55.791, -57.140, -59.212, -61.068,
 -63.183, -65.580, -68.202, -70.592, -73.012, -75.838, -78.772, -82.346, -86.080, -89.755,
 -92.512, -95.469, -97.549, -95.940, -99.221,-101.016,-103.011,-104.203,-103.471,-103.920,
-104.088,-103.570,-101.675, -98.786, -95.019, -90.839, -85.108, -80.260, -73.787, -66.976,
 -60.329, -54.978, -51.071, -48.831, -44.285, -36.796, -31.191, -28.612, -19.321, -12.209,
  -8.616,  -5.126,  -1.066,   2.027,   5.488,   9.358,  13.346,  17.857,  21.383,  24.221,
  27.503,  31.103,  34.751,  39.681,  44.907,  49.957,  53.673,  56.617,  54.819,  54.624,
  56.886,  59.202,  61.830,  64.867,  67.831,  70.052,  67.323,  55.833,  61.556,  67.432,
  72.731,  73.359,  74.421,  73.398,  71.925,  70.606,  69.979,  69.381,  68.763,  68.533,
  67.937,  68.177,  68.104,  66.630,  65.625,  64.783,  64.170,  63.749,  62.815,  60.985,
  57.337,  55.745,  54.288,  54.342,  55.420,  54.868,  54.431,  52.971,  50.270,  49.486,
  47.343,  45.727,  42.932,  41.017,  39.540,  37.703,  35.652,  33.957,  31.900,  30.387,
  29.888,  28.619,  26.357,  27.852,  24.526,  23.989,  22.164,  20.812,  20.074,  19.487,  18.749,
  18.304,  17.888,  16.959,  17.065,  16.434,  15.928,  14.863,  13.853,  14.009,  13.629,
  13.566,  13.678,  13.909,  13.892,  14.866,  14.817,  15.526,  16.396,  16.671,  16.788,
  17.924,  17.158,  16.764,  15.965,  14.795,  14.150,  13.718,  12.342,  11.430,  10.425,
   9.572,   8.715,   7.747,   6.546,   5.267,   3.992,   2.527,   0.959,  -0.205,  -1.694,
  -3.475,  -5.011,  -6.831,  -8.622, -10.058, -11.499, -13.006, -14.418, -15.846, -17.217,
 -18.416, -19.671, -20.847, -22.177, -23.396, -24.533, -25.481, -26.316, -27.077, -27.583,
 -27.858, -28.217, -27.959, -27.733, -27.076, -26.500, -25.903, -25.205, -24.475, -23.551,
 -22.718, -21.824, -20.796, -19.541, -18.670, -17.706, -16.742, -15.701, -14.624, -13.506,
 -12.434, -11.384, -10.454,  -9.036,  -7.738,  -6.617,  -5.466,  -4.320,  -3.189,  -2.203,
  -0.989,   0.291,   1.072,   2.320,   3.873,   5.497,   7.444,  10.141,  11.834,  12.655,
  10.941,  15.393,  23.863,  26.881,  18.380,  17.507,  16.084,  11.822,   8.811,   5.827,
   3.053,   1.233,  -3.252,  -6.318,  -6.957,  -6.293, -10.162, -12.689, -14.430, -17.488,
 -20.265, -22.937, -25.607, -28.060, -29.410, -30.444, -30.396, -30.447, -30.855, -30.567,
 -29.285, -28.979, -30.764, -30.656, -28.685, -26.888, -24.860, -22.856, -20.553, -17.940,
 -15.244, -12.391,  -9.544,  -6.815,  -4.166,  -1.543,   1.112,   3.596,   5.364,   6.433,
   7.578,   8.424,   9.122,   9.929,  10.795,  11.398,  11.978,  12.955,  13.433,  14.657,
  14.543,  15.442,  16.253,  16.753,  17.530,  18.058,  19.331,  20.293,  20.662,  20.438,
  21.106,  20.500,  20.063,  19.537,  19.079,  18.142,  18.087,  17.601,  17.044,  16.616},
{
  17.673,  17.426,  15.816,  14.806,  16.558,  19.235,  20.866,  19.374,  18.235,  20.480,
  16.015,  15.020,  13.952,  12.298,   9.800,   7.870,   4.763,   2.909,  -2.799,  -6.313,
  -8.206,  -9.424,  -9.624, -10.578, -10.656, -10.546, -10.624, -11.045, -11.173, -10.877,
 -11.722, -12.507, -13.161, -11.621, -11.177, -13.815, -16.085, -16.660, -16.406, -18.133,
 -22.428, -26.079, -29.885, -30.496, -30.568, -34.095, -36.852, -38.786, -42.804, -46.861,
 -49.067, -52.660, -54.319, -52.219, -53.388, -53.588, -55.091, -57.047, -58.618, -60.573,
 -62.669, -65.101, -67.168, -69.779, -73.030, -76.534, -80.003, -83.452, -86.821, -90.374,
 -93.199, -95.854, -97.639, -95.742, -99.614,-101.100,-102.925,-104.153,-104.199,-106.377,
-105.697,-104.791,-102.643, -99.335, -95.041, -89.991, -85.797, -80.008, -73.093, -66.832,
 -60.266, -54.761, -51.390, -50.575, -44.418, -42.686, -32.634, -22.450, -20.083, -15.450,
 -10.514,  -6.538,  -2.178,   0.725,   2.903,   6.203,  10.218,  14.981,  18.955,  21.912,
  25.245,  28.843,  32.756,  36.528,  42.493,  47.186,  52.227,  55.528,  55.448,  56.882,
  58.526,  61.194,  63.428,  65.911,  68.268,  69.117,  70.163,  61.785,  65.198,  66.449,
  71.136,  72.191,  72.110,  72.473,  70.658,  69.714,  68.803,  68.111,  67.475,  67.289,
  66.631,  66.662,  66.877,  65.757,  64.598,  63.980,  63.002,  62.464,  62.089,  59.461,
  57.120,  54.986,  53.320,  52.939,  53.592,  54.051,  52.231,  52.570,  49.404,  48.206,
  46.944,  44.656,  41.442,  40.576,  37.983,  36.019,  34.247,  32.710,  31.159,  29.989,
  28.351,  27.728,  25.127,  26.198,  23.852,  23.294,  21.352,  19.239,  18.916,  18.112,  17.086,
  16.952,  16.224,  16.171,  17.035,  16.140,  14.760,  13.970,  13.632,  13.870,  13.083,
  13.174,  13.630,  13.498,  13.720,  14.027,  14.947,  16.270,  16.865,  15.867,  16.883,
  17.487,  17.298,  16.124,  15.349,  14.557,  13.867,  12.694,  11.577,  10.717,   9.897,
   8.806,   7.270,   5.942,   5.049,   3.942,   2.965,   1.803,   0.161,  -1.270,  -2.920,
  -4.488,  -6.204,  -7.961,  -9.779, -11.513, -13.213, -14.564, -16.034, -17.397, -18.716,
 -19.915, -21.067, -22.147, -23.445, -24.617, -25.824, -26.973, -27.826, -28.462, -28.951,
 -29.267, -29.370, -29.289, -29.028, -28.518, -27.965, -27.200, -26.558, -25.641, -24.654,
 -23.649, -22.543, -21.425, -20.272, -19.155, -17.968, -16.784, -15.515, -14.213, -13.016,
 -11.849, -10.790,  -9.683,  -8.776,  -7.788,  -6.606,  -5.422,  -4.117,  -2.926,  -1.574,
  -0.415,   1.365,   2.908,   3.666,   4.664,   6.768,   8.504,  10.469,  13.071,  13.635,
  13.089,  10.132,  17.340,  26.670,  17.578,  18.950,  10.216,   6.286,   4.492,   1.869,
  -1.062,  -3.363,  -6.014,  -6.781,  -7.578,  -8.586, -11.773, -14.322, -15.493, -18.296,
 -22.017, -24.387, -26.886, -30.992, -32.593, -32.627, -32.167, -32.521, -33.639, -32.293,
 -29.491, -30.149, -31.201, -31.795, -29.672, -28.177, -25.613, -22.990, -20.610, -18.087,
 -15.295, -12.563,  -9.676,  -6.819,  -3.948,  -1.404,   1.550,   3.737,   5.157,   6.424,
   7.820,   8.742,   9.599,  10.461,  11.611,  12.115,  12.993,  14.124,  14.951,  15.681,
  15.713,  16.555,  17.113,  17.970,  19.002,  19.347,  20.318,  21.045,  21.547,  22.389,
  24.295,  25.177,  23.947,  23.079,  22.399,  20.507,  19.901,  20.252,  18.966,  17.673},
{
  20.453,  19.542,  17.326,  16.182,  18.130,  20.892,  19.099,  18.789,  20.255,  21.327,
  20.145,  18.332,  15.763,  14.098,  12.047,   9.020,   5.939,   4.301,  -1.130,  -5.191,
  -7.529,  -7.728,  -7.288,  -8.736,  -8.876,  -9.884, -11.120, -11.623, -10.885, -10.812,
 -11.460, -12.287, -13.121, -12.521, -12.835, -13.371, -14.984, -14.392, -14.688, -17.290,
 -20.885, -24.052, -29.165, -29.856, -31.328, -33.942, -36.250, -38.789, -40.935, -46.408,
 -48.652, -52.663, -54.008, -54.406, -53.397, -53.649, -54.836, -56.287, -57.261, -59.552,
 -61.658, -63.939, -66.439, -69.969, -73.888, -77.502, -80.808, -84.405, -87.594, -90.636,
 -93.885, -96.013, -97.416, -96.493, -99.858,-100.931,-102.407,-103.421,-104.683,-106.594,
-105.992,-104.557,-102.563, -99.234, -94.691, -90.096, -85.521, -79.932, -73.424, -66.715,
 -60.384, -54.471, -52.405, -50.247, -49.277, -41.775, -29.403, -25.693, -20.949, -17.239,
 -12.602,  -7.989,  -3.173,  -1.237,   0.499,   3.131,   7.622,  11.861,  16.684,  19.636,
  22.852,  26.792,  31.108,  34.387,  39.233,  44.130,  49.404,  54.423,  57.107,  59.152,
  61.821,  61.879,  63.829,  65.904,  67.058,  63.696,  65.074,  53.469,  55.527,  67.603,
  69.890,  69.478,  69.736,  70.540,  68.990,  68.299,  67.846,  67.059,  66.518,  66.290,
  65.510,  65.206,  65.527,  64.866,  63.652,  63.016,  61.916,  61.276,  60.936,  59.093,
  54.925,  53.599,  53.061,  52.453,  53.648,  51.676,  51.181,  50.659,  49.064,  46.597,
  44.746,  42.734,  40.009,  40.224,  36.058,  34.329,  32.768,  30.295,  29.718,  28.561,
  27.697,  26.634,  24.848,  24.114,  23.176,  22.196,  21.277,  17.946,  18.063,  16.729,  16.106,
  16.129,  15.839,  15.934,  15.802,  14.531,  13.185,  13.030,  13.061,  12.785,  12.233,
  12.452,  13.114,  13.117,  14.072,  14.465,  15.261,  15.927,  15.882,  16.538,  16.339,
  16.626,  16.171,  15.408,  14.983,  14.208,  13.001,  11.908,  10.774,   9.593,   8.636,
   7.623,   6.306,   5.211,   3.893,   2.861,   1.168,  -0.165,  -1.472,  -2.949,  -4.319,
  -5.963,  -7.904,  -9.777, -11.653, -13.210, -14.722, -16.053, -17.341, -18.636, -19.939,
 -21.290, -22.559, -23.731, -25.053, -26.194, -27.348, -28.231, -29.148, -29.677, -30.236,
 -30.486, -30.781, -30.828, -30.545, -30.084, -29.473, -28.604, -27.746, -26.715, -25.532,
 -24.343, -23.244, -22.067, -20.759, -19.428, -17.941, -16.680, -15.261, -13.862, -12.549,
 -11.378, -10.242,  -9.195,  -8.073,  -6.958,  -5.947,  -4.964,  -3.991,  -2.697,  -1.376,
   0.231,   1.717,   3.873,   5.594,   6.066,   7.610,   9.357,  10.401,  12.147,  13.356,
  14.384,  11.971,  16.469,  27.074,  23.061,  21.723,  13.210,   2.666,  -0.250,  -1.564,
  -4.799,  -6.499,  -4.548,  -8.402,  -8.739,  -9.954, -15.171, -13.388, -14.123, -18.916,
 -23.649, -25.442, -27.337, -32.582, -34.083, -33.322, -33.711, -34.264, -33.698, -36.260,
 -35.000, -34.620, -33.904, -32.880, -31.179, -28.554, -26.463, -23.733, -21.127, -18.399,
 -15.694, -12.835,  -9.846,  -6.820,  -3.873,  -1.184,   2.199,   3.774,   5.508,   6.623,
   7.942,   8.773,   9.615,  10.809,  12.105,  13.276,  14.436,  15.391,  17.049,  16.841,
  17.085,  17.383,  17.956,  18.892,  19.893,  20.253,  21.272,  22.563,  23.264,  26.902,
  30.284,  29.610,  28.746,  27.015,  25.615,  24.238,  25.172,  24.441,  22.479,  20.453},
{
  25.912,  24.700,  20.984,  18.527,  19.340,  19.610,  18.549,  20.377,  21.128,  21.736,
  22.538,  20.108,  17.547,  14.671,  12.415,  10.118,   7.923,   5.029,   1.203,  -2.758,
  -5.639,  -6.113,  -5.858,  -7.062,  -7.504,  -9.317, -11.345, -10.600,  -9.767, -10.341,
 -11.451, -11.882, -12.676, -13.194, -13.375, -12.716, -13.229, -12.686, -13.129, -14.193,
 -18.227, -22.869, -26.443, -29.050, -30.155, -32.393, -34.423, -36.616, -39.501, -42.027,
 -48.502, -52.086, -53.493, -54.728, -54.210, -53.606, -53.640, -55.374, -56.624, -58.489,
 -60.221, -62.454, -66.255, -70.300, -74.292, -77.914, -81.514, -85.232, -88.455, -91.200,
 -93.826, -96.194, -97.153, -96.603,-100.164,-100.881,-101.900,-102.330,-104.391,-105.007,
-100.427, -97.878,-100.330, -98.451, -95.080, -90.340, -85.422, -79.345, -73.350, -67.148,
 -59.782, -54.751, -55.418, -54.257, -52.431, -40.405, -33.675, -29.777, -23.338, -18.852,
 -14.787, -10.370,  -7.428,  -5.238,  -3.097,   0.446,   5.730,   9.515,  14.236,  17.300,
  21.199,  24.124,  28.497,  31.255,  34.888,  41.553,  46.609,  53.092,  53.021,  56.383,
  59.608,  66.521,  65.250,  65.849,  67.253,  70.126,  64.085,  45.248,  60.250,  67.609,
  68.196,  67.396,  67.146,  68.720,  66.325,  65.557,  66.661,  65.743,  65.513,  64.777,
  64.299,  64.210,  64.453,  63.767,  62.990,  62.586,  61.244,  60.137,  59.464,  57.703,
  56.729,  53.349,  53.088,  51.590,  52.024,  50.077,  49.422,  50.214,  48.374,  44.950,
  42.769,  41.428,  38.333,  36.455,  34.913,  32.406,  30.906,  29.355,  29.171,  28.208,
  27.680,  24.616,  27.383,  23.644,  22.623,  20.753,  21.385,  18.289,  18.145,  16.481,  15.587,
  15.757,  16.172,  15.750,  14.660,  12.973,  12.447,  11.417,  12.013,  11.642,  11.619,
  12.074,  12.844,  13.183,  14.138,  14.240,  15.251,  15.140,  15.860,  16.757,  15.812,
  16.201,  15.287,  14.209,  13.334,  12.535,  11.745,  10.734,   9.717,   8.188,   7.153,
   6.459,   5.160,   3.661,   2.587,   1.479,   0.043,  -1.326,  -2.738,  -4.574,  -6.375,
  -8.162,  -9.755, -11.418, -13.124, -14.777, -16.356, -17.602, -18.852, -20.123, -21.683,
 -22.937, -24.293, -25.437, -26.601, -27.879, -28.847, -29.926, -30.844, -31.331, -31.803,
 -31.927, -32.143, -32.140, -31.762, -31.372, -30.719, -29.916, -28.978, -27.767, -26.222,
 -24.997, -23.801, -22.613, -21.198, -19.673, -18.004, -16.590, -15.041, -13.562, -12.197,
 -10.957,  -9.759,  -8.642,  -7.573,  -6.398,  -5.257,  -4.200,  -3.169,  -2.057,  -0.694,
   0.882,   2.536,   4.252,   6.117,   7.593,   8.830,  10.021,  12.807,  12.897,  13.718,
  14.105,  12.096,  19.252,  26.218,  20.064,  13.390,  20.135,   5.787,  -3.132,  -4.634,
  -7.139,  -9.549,  -7.880,  -6.216,  -7.294, -13.159, -17.218, -16.488, -18.842, -22.718,
 -26.133, -27.373, -28.235, -31.621, -32.110, -31.783, -32.627, -33.098, -35.616, -38.493,
 -38.238, -37.568, -36.179, -34.485, -32.439, -29.603, -27.052, -24.148, -21.553, -18.855,
 -16.320, -13.094,  -9.894,  -6.789,  -3.881,  -0.627,   1.996,   3.651,   5.163,   6.530,
   7.872,   9.299,   9.989,  10.880,  12.412,  13.310,  15.237,  15.960,  17.254,  18.446,
  17.774,  18.156,  18.815,  19.799,  20.810,  21.666,  22.756,  24.505,  28.050,  30.043,
  30.753,  30.580,  29.246,  29.004,  27.363,  27.220,  27.542,  27.162,  26.936,  25.912},
{
  26.234,  26.558,  25.957,  24.702,  24.026,  23.241,  22.326,  21.712,  21.908,  21.333,
  20.179,  20.026,  19.005,  16.806,  15.100,  11.949,   8.266,   5.880,   2.623,  -0.369,
  -1.890,  -2.680,  -3.475,  -4.985,  -6.115,  -8.577,  -8.675,  -8.248,  -8.451,  -9.501,
 -10.559, -11.292, -11.660, -12.908, -12.488, -11.802, -10.916, -10.726, -11.399, -11.028,
 -14.108, -19.336, -23.311, -25.359, -26.215, -28.403, -31.084, -33.118, -35.695, -39.071,
 -43.523, -51.120, -53.357, -54.190, -56.069, -54.441, -53.164, -54.019, -55.736, -56.977,
 -59.064, -62.387, -66.445, -70.376, -74.362, -78.331, -81.949, -85.344, -88.703, -91.251,
 -93.588, -95.538, -96.805, -96.854, -99.836,-100.411,-100.994,-100.675,-103.329,-102.733,
 -96.745, -92.728, -95.258, -96.049, -93.747, -89.558, -84.724, -79.102, -73.673, -67.255,
 -60.389, -56.165, -57.735, -59.321, -49.846, -41.808, -34.539, -30.510, -25.321, -21.067,
 -16.730, -13.935, -11.618,  -8.569,  -5.412,  -1.517,   3.146,   6.924,  10.941,  14.986,
  17.674,  21.764,  26.055,  31.034,  33.319,  36.182,  43.540,  48.284,  51.525,  53.399,
  55.120,  58.542,  67.063,  66.460,  69.252,  71.314,  61.921,  51.153,  61.977,  66.622,
  66.701,  65.794,  65.909,  66.501,  67.650,  62.784,  62.665,  61.494,  63.922,  64.115,
  63.392,  62.534,  62.543,  62.616,  61.977,  60.859,  60.420,  59.348,  58.423,  58.455,
  52.586,  52.298,  55.287,  51.954,  49.894,  50.381,  48.415,  45.774,  49.059,  43.456,
  41.215,  38.735,  37.218,  35.001,  33.076,  31.521,  30.243,  28.416,  26.500,  27.184,
  25.737,  25.780,  25.864,  23.453,  22.155,  20.874,  19.018,  17.746,  17.293,  16.342,  15.584,
  15.028,  14.750,  15.335,  13.751,  11.420,  11.294,  10.632,  11.155,  10.985,  11.123,
  11.794,  12.165,  13.278,  13.498,  14.011,  14.404,  13.802,  14.896,  15.252,  15.250,
  14.511,  13.499,  12.929,  11.842,  10.932,  10.144,   8.972,   7.847,   7.027,   6.133,
   5.202,   3.870,   2.535,   1.393,  -0.003,  -1.424,  -2.917,  -4.603,  -6.222,  -7.966,
  -9.779, -11.198, -12.825, -14.555, -16.444, -18.035, -19.620, -21.145, -22.580, -23.839,
 -25.209, -26.529, -27.669, -28.908, -30.088, -31.150, -31.934, -32.483, -33.003, -33.267,
 -33.531, -33.408, -33.523, -33.137, -32.780, -32.120, -31.217, -30.048, -28.605, -27.082,
 -25.679, -24.541, -23.134, -21.616, -19.856, -18.170, -16.449, -14.800, -13.250, -11.831,
 -10.483,  -9.315,  -8.149,  -7.102,  -5.919,  -4.627,  -3.309,  -2.240,  -1.271,   0.058,
   1.477,   3.131,   4.907,   6.417,   8.474,  10.723,  10.894,  11.570,  11.754,  11.960,
  11.813,  13.955,  17.022,  21.663,  14.158,   4.358,  11.383,   3.184,  -4.780,  -7.014,
  -8.371,  -9.890, -11.551, -10.463, -11.854, -15.579, -20.913, -23.619, -25.891, -27.301,
 -28.582, -29.634, -30.513, -31.721, -32.764, -33.222, -34.324, -39.220, -39.996, -40.314,
 -40.251, -39.070, -37.537, -35.051, -32.478, -29.952, -27.576, -24.720, -21.847, -18.837,
 -15.951, -12.758,  -9.525,  -6.945,  -4.060,  -1.242,   0.927,   2.818,   4.448,   5.755,
   7.059,   8.658,  10.111,  11.537,  12.792,  13.928,  15.451,  17.424,  17.034,  18.485,
  18.981,  19.444,  20.199,  21.024,  21.848,  23.057,  26.736,  31.245,  30.881,  31.342,
  32.063,  31.274,  30.458,  29.951,  28.189,  27.701,  27.753,  28.002,  26.761,  26.234},
{
  25.914,  25.574,  26.021,  25.834,  25.344,  25.027,  24.098,  23.681,  22.085,  20.872,
  20.136,  19.385,  18.225,  16.363,  14.310,  11.841,   8.611,   5.693,   3.976,   1.935,
  -0.587,  -0.794,  -2.324,  -3.136,  -4.669,  -5.203,  -6.108,  -6.620,  -7.307,  -8.511,
  -9.863, -10.109, -10.516, -11.676, -11.531,  -9.737,  -7.740,  -8.381,  -8.533, -10.619,
 -12.381, -15.611, -18.715, -20.502, -21.917, -24.624, -28.057, -29.030, -30.560, -34.192,
 -38.655, -47.127, -51.651, -52.654, -53.792, -54.937, -52.022, -52.874, -54.007, -55.491,
 -58.346, -62.227, -66.282, -70.401, -74.445, -78.395, -81.463, -84.643, -87.775, -90.622,
 -93.254, -95.302, -96.303, -96.914, -99.150,-100.127,-100.049, -97.444, -99.427,-100.373,
 -96.020, -94.181, -95.360, -94.720, -92.422, -88.371, -84.462, -78.386, -73.229, -66.680,
 -61.782, -58.308, -58.099, -59.754, -49.250, -42.784, -35.906, -31.927, -27.193, -22.776,
 -18.779, -16.916, -14.271, -10.771,  -7.083,  -3.446,  -0.020,   4.218,   7.779,  12.506,
  16.293,  20.310,  24.852,  28.884,  33.401,  35.967,  38.254,  45.877,  50.348,  51.643,
  53.832,  56.199,  59.613,  67.233,  70.642,  71.351,  64.718,  55.077,  61.455,  64.790,
  64.336,  63.517,  64.034,  63.978,  64.061,  62.583,  63.063,  65.328,  58.489,  63.251,
  62.728,  61.875,  61.330,  61.590,  59.281,  60.107,  59.754,  58.682,  58.680,  57.721,
  52.293,  49.915,  52.482,  48.978,  47.851,  48.956,  45.549,  44.293,  42.336,  41.055,
  38.556,  37.454,  35.266,  32.791,  31.279,  29.834,  28.578,  27.769,  28.429,  26.582,
  25.320,  25.878,  22.678,  22.466,  21.652,  20.183,  18.898,  17.685,  16.445,  15.522,  14.686,
  13.982,  13.463,  13.131,  12.141,  10.407,  10.004,  10.259,  10.756,  10.841,  11.454,
  12.288,  12.251,  12.805,  13.089,  13.103,  14.538,  14.328,  13.795,  13.498,  14.126,
  12.869,  12.138,  11.345,  10.098,   9.440,   8.470,   7.367,   6.429,   5.646,   4.735,
   3.616,   2.181,   0.910,  -0.148,  -1.339,  -2.696,  -4.248,  -5.924,  -7.700,  -9.442,
 -11.058, -12.649, -14.274, -15.782, -17.835, -19.460, -21.585, -23.256, -24.897, -26.046,
 -27.291, -28.518, -29.816, -31.050, -32.277, -33.284, -34.182, -34.716, -35.155, -35.370,
 -35.425, -35.392, -35.171, -34.720, -34.094, -33.274, -32.155, -30.906, -29.454, -28.094,
 -26.555, -25.030, -23.392, -21.809, -20.043, -18.182, -16.310, -14.672, -13.043, -11.568,
 -10.320,  -9.011,  -7.602,  -6.475,  -5.367,  -4.128,  -2.934,  -1.536,  -0.447,   0.886,
   2.129,   3.666,   5.283,   7.134,   9.192,  10.786,  11.912,  14.169,  15.276,  14.540,
  15.043,  14.609,  12.334,   8.879,   4.214,   1.792,   4.407,   0.044,  -5.145,  -8.780,
  -9.881, -10.862, -12.615, -14.433, -17.918, -21.433, -24.764, -26.905, -28.803, -29.505,
 -31.021, -32.742, -35.046, -36.836, -38.300, -37.410, -38.835, -43.167, -42.968, -42.047,
 -41.526, -39.960, -37.385, -34.629, -32.148, -29.795, -27.986, -24.853, -21.276, -18.378,
 -14.832, -12.050,  -9.796,  -7.390,  -5.430,  -3.435,  -1.747,  -0.182,   2.087,   4.324,
   5.860,   7.199,   8.734,  10.201,  11.667,  13.021,  14.984,  16.681,  18.666,  18.677,
  19.216,  19.956,  21.273,  22.698,  23.988,  30.599,  32.665,  30.264,  32.219,  33.879,
  34.905,  33.646,  30.900,  30.078,  28.789,  27.328,  27.642,  27.156,  26.042,  25.914},
{
  24.272,  23.570,  25.190,  24.851,  24.359,  23.769,  24.508,  25.011,  23.126,  22.753,
  20.630,  19.103,  18.578,  16.488,  13.704,  12.036,   8.914,   5.505,   3.572,   0.709,
  -0.867,  -1.452,  -1.289,  -0.711,  -1.813,  -3.052,  -3.976,  -5.493,  -5.958,  -7.478,
  -8.283,  -7.715,  -9.198, -10.065,  -7.793,  -4.041,  -6.645,  -7.091,  -7.035,  -8.136,
 -12.253, -12.328, -14.179, -16.541, -18.244, -22.029, -24.644, -26.842, -27.947, -30.473,
 -34.465, -41.718, -48.516, -51.051, -51.842, -53.153, -50.825, -51.236, -52.578, -54.697,
 -58.230, -61.926, -66.464, -70.220, -73.666, -77.371, -80.411, -83.778, -86.666, -89.697,
 -92.292, -94.383, -95.497, -96.137, -98.313, -99.265, -96.143, -96.422, -96.784, -96.695,
 -96.065, -95.402, -97.548, -95.126, -92.154, -87.797, -84.385, -78.312, -73.447, -66.764,
 -62.257, -61.180, -59.274, -60.868, -51.485, -44.019, -39.079, -33.975, -29.294, -25.073,
 -21.905, -19.891, -16.869, -13.149,  -9.402,  -6.096,  -2.904,   1.524,   6.156,  10.354,
  13.060,  17.714,  22.610,  26.341,  30.803,  34.719,  38.453,  42.261,  48.832,  48.990,
  54.269,  55.266,  52.945,  60.356,  63.685,  67.663,  68.014,  50.729,  60.184,  62.245,
  61.499,  60.753,  60.858,  61.640,  61.167,  61.716,  61.653,  61.712,  58.601,  60.622,
  60.429,  60.847,  60.456,  57.357,  54.723,  55.520,  56.006,  54.071,  54.482,  50.707,
  55.623,  46.676,  51.185,  45.814,  42.841,  43.238,  42.550,  41.693,  39.605,  37.808,
  35.357,  34.858,  33.924,  32.379,  30.330,  29.078,  30.132,  29.386,  28.295,  24.590,
  27.579,  26.386,  21.928,  21.615,  20.411,  19.301,  18.242,  17.807,  16.384,  14.890,  13.509,
  12.755,  12.675,  11.898,  10.638,  10.151,   9.852,   9.948,  10.518,  11.135,  11.134,
  11.642,  12.274,  13.164,  13.434,  13.314,  13.443,  13.743,  12.894,  12.845,  11.550,
  11.525,  11.248,   9.630,   8.585,   7.579,   6.733,   5.896,   5.046,   4.072,   2.751,
   1.708,   0.818,  -0.240,  -1.427,  -2.708,  -3.980,  -5.522,  -7.130,  -8.790, -10.595,
 -12.328, -13.955, -15.497, -17.322, -19.297, -21.343, -23.143, -24.928, -26.386, -27.673,
 -29.070, -30.596, -31.869, -33.193, -34.330, -35.225, -36.123, -36.829, -37.238, -37.444,
 -37.399, -37.186, -36.828, -36.212, -35.446, -34.413, -33.249, -31.918, -30.115, -28.574,
 -26.935, -25.100, -23.466, -21.686, -19.926, -18.075, -16.416, -14.847, -13.110, -11.514,
 -10.366,  -8.995,  -7.519,  -6.082,  -4.819,  -3.486,  -2.387,  -1.046,   0.013,   1.317,
   2.672,   4.400,   5.889,   6.275,   6.811,  10.093,  14.872,  13.263,  10.722,  12.352,
  14.482,   9.980,   1.487,   2.233,  -2.232,  -1.771,  -2.013, -11.214,  -4.599,  -8.479,
 -11.709, -12.699, -14.897, -18.330, -23.030, -29.188, -33.321, -35.211, -34.196, -33.221,
 -35.691, -39.915, -40.224, -42.286, -43.846, -42.178, -43.913, -43.856, -43.138, -42.403,
 -41.158, -39.300, -37.010, -34.909, -32.537, -29.550, -26.185, -22.907, -19.591, -16.404,
 -14.455, -12.329, -10.221,  -8.066,  -6.085,  -4.362,  -2.607,  -0.522,   1.175,   2.886,
   4.772,   6.768,   8.395,  10.211,  11.839,  13.221,  14.518,  16.124,  18.600,  19.779,
  18.796,  20.941,  23.052,  25.141,  28.611,  32.346,  32.068,  29.829,  32.683,  34.067,
  34.937,  33.553,  30.506,  29.079,  28.367,  27.005,  27.091,  25.697,  24.794,  24.272},
{
  23.262,  21.963,  23.883,  24.082,  23.284,  23.077,  24.169,  24.881,  23.888,  23.894,
  21.569,  19.019,  18.132,  16.953,  14.584,  12.599,  10.228,   7.937,   5.661,   3.734,
   1.623,   0.094,  -0.453,  -0.009,  -0.226,  -1.302,  -2.305,  -3.270,  -4.551,  -5.989,
  -5.324,  -5.758,  -6.735,  -7.791,  -4.444,  -3.994,  -5.010,  -6.420,  -7.431,  -8.274,
  -9.728, -12.455, -13.437, -14.519, -18.174, -20.486, -22.862, -24.822, -25.992, -28.290,
 -32.070, -36.516, -44.219, -47.700, -49.528, -50.829, -50.226, -50.220, -50.753, -54.244,
 -58.224, -62.220, -65.982, -69.546, -72.476, -75.927, -79.219, -82.622, -85.465, -88.392,
 -91.153, -93.025, -93.524, -94.693, -96.744, -97.286, -95.019, -94.895, -95.356, -94.909,
 -96.033, -98.390, -97.193, -94.583, -91.399, -87.744, -84.424, -78.984, -73.759, -67.568,
 -63.260, -62.916, -60.660, -63.776, -53.093, -46.101, -42.208, -35.792, -31.694, -28.006,
 -25.356, -22.697, -19.424, -15.734, -12.188,  -9.052,  -5.068,  -1.483,   2.760,   7.962,
  10.843,  14.662,  19.156,  24.161,  28.674,  32.467,  36.634,  39.686,  44.822,  49.922,
  52.010,  55.887,  54.297,  61.728,  64.799,  64.685,  64.066,  49.711,  58.561,  59.894,
  59.414,  58.579,  58.246,  58.137,  58.783,  59.575,  59.510,  59.703,  60.112,  55.067,
  61.322,  56.076,  53.968,  51.897,  53.264,  55.253,  53.500,  51.226,  51.174,  46.013,
  44.975,  43.754,  43.767,  42.294,  41.319,  41.325,  40.061,  39.242,  37.929,  36.027,
  34.926,  35.562,  33.095,  32.254,  29.821,  27.518,  28.204,  25.860,  24.619,  26.444,
  24.966,  23.141,  21.913,  21.015,  20.202,  18.987,  17.613,  16.343,  14.845,  14.324,  12.684,
  12.035,  11.962,  11.327,  10.360,  10.259,  10.373,   9.939,  10.509,  11.793,  11.679,
  11.979,  12.247,  12.675,  13.280,  14.819,  13.388,  12.890,  12.491,  11.141,  10.242,
   9.761,   8.587,   8.024,   7.066,   6.153,   5.491,   4.507,   3.425,   2.323,   1.080,
   0.168,  -0.706,  -1.608,  -2.621,  -3.865,  -5.149,  -6.672,  -8.066,  -9.554, -11.405,
 -13.394, -15.096, -16.890, -18.901, -20.859, -22.841, -24.285, -26.149, -27.688, -29.262,
 -30.748, -32.206, -33.615, -34.915, -36.114, -37.190, -37.916, -38.497, -39.102, -39.423,
 -39.402, -39.082, -38.594, -37.806, -36.774, -35.668, -34.305, -32.741, -31.049, -29.057,
 -26.935, -25.504, -23.916, -22.078, -20.050, -18.026, -16.393, -14.675, -13.007, -11.611,
 -10.231,  -9.018,  -7.498,  -5.877,  -4.633,  -3.227,  -2.098,  -0.712,   0.429,   1.283,
   2.826,   4.240,   3.805,   6.285,  10.498,  13.705,   9.382,   6.287,   2.591,   0.841,
   1.139,  -1.612,  -4.462,  -2.569,  -4.783,  -4.390,  -1.266, -12.556, -14.191, -10.821,
 -13.484, -15.501, -16.594, -22.521, -28.268, -31.586, -40.218, -45.096, -42.514, -41.849,
 -41.013, -41.915, -44.354, -45.819, -44.928, -44.943, -44.976, -44.800, -44.051, -42.806,
 -41.154, -39.489, -37.320, -34.948, -32.354, -28.963, -25.612, -22.298, -18.702, -16.334,
 -14.683, -12.718, -11.022,  -9.051,  -7.147,  -5.200,  -3.242,  -1.173,   0.903,   2.785,
   4.600,   6.499,   8.442,  10.173,  11.388,  12.934,  14.489,  15.738,  15.747,  17.071,
  18.852,  20.878,  25.424,  28.759,  30.376,  31.335,  32.780,  32.141,  33.081,  32.657,
  32.379,  31.810,  29.736,  28.368,  27.905,  27.018,  26.407,  25.278,  24.105,  23.262},
{
  23.412,  22.519,  21.991,  22.824,  21.650,  21.308,  22.182,  22.413,  22.158,  22.135,
  21.015,  19.259,  18.077,  16.352,  14.839,  12.759,  11.350,   9.695,   7.552,   5.846,
   4.240,   3.354,   2.838,   2.220,   1.500,   0.161,  -0.924,  -2.499,  -3.146,  -2.586,
  -2.019,  -3.350,  -4.579,  -5.126,  -3.891,  -4.914,  -4.549,  -3.451,  -6.454,  -6.547,
  -8.925, -10.783, -11.547, -13.356, -17.102, -19.775, -21.244, -23.982, -25.411, -27.710,
 -27.935, -33.254, -40.175, -44.040, -46.813, -47.594, -47.961, -49.178, -51.612, -54.230,
 -58.008, -61.545, -64.518, -67.935, -71.236, -74.607, -77.756, -80.844, -84.018, -86.679,
 -89.554, -90.716, -91.519, -93.106, -95.110, -93.965, -92.344, -93.514, -94.476, -92.610,
 -93.495, -99.016, -96.413, -93.372, -90.140, -87.543, -83.869, -78.802, -73.772, -68.168,
 -64.208, -63.484, -61.458, -62.827, -53.007, -48.188, -44.463, -37.764, -33.746, -30.225,
 -27.666, -25.178, -21.993, -17.992, -14.798, -11.924,  -7.548,  -3.074,   1.067,   5.647,
   8.636,  11.932,  16.218,  20.371,  25.159,  30.259,  34.878,  39.297,  41.551,  47.600,
  52.056,  54.411,  55.089,  59.888,  61.835,  65.336,  55.470,  51.713,  57.189,  57.813,
  57.051,  56.355,  56.054,  55.787,  56.083,  58.032,  57.946,  57.794,  58.869,  50.489,
  50.860,  43.494,  37.894,  40.894,  46.358,  50.752,  50.458,  49.687,  47.426,  45.657,
  44.268,  43.432,  41.915,  41.176,  40.185,  39.119,  37.306,  38.501,  35.571,  34.379,
  33.622,  33.149,  31.965,  28.999,  29.109,  27.198,  26.926,  27.983,  25.515,  22.543,
  25.223,  22.088,  21.329,  20.134,  18.692,  17.118,  15.917,  15.283,  14.348,  12.943,  12.248,
  11.075,  10.904,  11.086,  10.749,  11.260,  10.810,  10.448,  10.598,  10.475,  12.389,
  11.787,  11.941,  12.554,  13.343,  13.063,  12.909,  11.766,  10.783,   9.932,   9.234,
   8.298,   7.538,   6.448,   5.644,   4.913,   3.951,   2.954,   1.821,   0.384,  -0.523,
  -1.221,  -1.995,  -2.933,  -4.105,  -5.340,  -6.446,  -7.626,  -8.966, -10.710, -12.425,
 -14.524, -16.420, -18.371, -20.013, -22.032, -23.779, -25.695, -27.604, -29.188, -30.613,
 -32.318, -33.974, -35.417, -36.501, -37.687, -38.806, -39.561, -40.208, -40.725, -40.836,
 -40.961, -40.612, -39.958, -39.020, -37.963, -36.627, -35.367, -33.679, -31.774, -30.188,
 -28.393, -26.356, -24.194, -22.106, -20.093, -18.016, -16.024, -14.133, -12.531, -11.241,
  -9.980,  -8.780,  -7.401,  -6.002,  -4.620,  -3.142,  -1.859,  -0.419,   0.753,   1.603,
   2.687,   1.210,  -0.708,   6.036,   8.398,   8.875,   5.894,   3.364,   1.122,  -0.803,
  -2.089,  -3.647,  -5.897,  -8.607,  -8.867,   0.010,  -5.200, -12.780, -17.356, -16.701,
 -17.275, -21.616, -23.257, -26.063, -29.860, -29.996, -32.558, -38.578, -41.798, -46.452,
 -44.963, -44.798, -47.397, -47.163, -45.839, -45.380, -45.302, -44.983, -43.903, -42.547,
 -41.139, -39.383, -37.399, -34.770, -30.790, -27.480, -24.752, -22.422, -20.224, -18.448,
 -16.719, -14.868, -12.556, -10.509,  -8.370,  -6.313,  -4.212,  -2.063,   0.078,   2.274,
   4.462,   6.556,   8.551,  10.407,  12.265,  13.929,  15.106,  15.673,  16.258,  17.475,
  19.080,  19.944,  26.159,  28.193,  28.830,  31.051,  32.285,  35.299,  33.569,  32.030,
  31.576,  30.824,  28.872,  28.745,  29.157,  27.628,  26.048,  25.413,  24.333,  23.412},
{
  24.155,  23.588,  22.407,  22.397,  20.171,  19.886,  20.792,  21.035,  21.603,  21.395,
  20.507,  18.843,  17.383,  15.956,  14.408,  13.004,  12.026,  10.634,   8.302,   7.459,
   5.436,   5.211,   4.950,   4.540,   3.913,   2.957,   1.255,  -0.043,  -0.038,  -0.172,
  -0.617,  -1.318,  -2.907,  -3.892,  -4.673,  -5.454,  -4.028,  -2.352,  -1.881,  -3.246,
  -7.977,  -9.002,  -9.757, -11.710, -15.072, -16.827, -18.965, -21.311, -23.435, -25.809,
 -28.455, -30.413, -34.328, -36.869, -40.523, -42.681, -44.660, -46.402, -50.635, -53.604,
 -56.849, -60.390, -63.030, -66.622, -69.774, -72.989, -76.135, -79.290, -82.157, -84.945,
 -87.879, -89.004, -89.614, -91.923, -92.884, -91.086, -89.960, -88.890, -91.113, -91.033,
 -93.008, -96.467, -94.624, -91.575, -88.876, -86.623, -83.259, -78.702, -73.611, -68.605,
 -64.251, -63.154, -62.413, -61.683, -53.989, -48.914, -46.104, -39.257, -35.970, -32.348,
 -29.490, -27.343, -23.817, -19.907, -17.232, -14.073,  -9.935,  -5.101,   1.192,   4.184,
   5.960,   8.640,  13.435,  17.514,  22.342,  26.834,  30.672,  35.967,  38.218,  43.140,
  48.928,  51.156,  52.480,  55.139,  58.447,  60.343,  47.427,  52.089,  55.140,  55.374,
  54.732,  54.318,  53.869,  53.810,  53.846,  55.340,  55.709,  55.974,  56.081,  54.030,
  52.845,  48.140,  46.371,  48.303,  38.267,  35.438,  41.714,  46.808,  47.108,  46.347,
  44.558,  42.601,  41.395,  39.775,  38.818,  37.766,  36.089,  34.486,  33.493,  32.927,
  31.551,  30.117,  30.952,  27.742,  27.292,  29.651,  25.948,  25.515,  22.997,  23.307,
  21.942,  20.193,  19.180,  18.226,  17.025,  15.571,  14.427,  14.037,  13.253,  12.380,  10.808,
  10.446,  10.955,  11.313,  10.891,  10.683,  10.461,  11.337,  10.651,  10.356,  10.765,
  11.546,  11.831,  12.348,  12.370,  11.845,  10.910,  10.254,   9.732,   9.100,   7.946,
   7.531,   6.146,   5.249,   4.634,   3.664,   2.604,   1.524,   0.394,  -0.821,  -1.534,
  -2.405,  -3.210,  -4.313,  -5.282,  -6.391,  -7.512,  -8.632, -10.087, -11.941, -13.452,
 -15.364, -17.505, -19.298, -21.350, -23.064, -24.904, -26.697, -28.658, -30.694, -32.274,
 -33.914, -35.479, -37.020, -38.220, -39.349, -40.334, -41.130, -41.638, -42.004, -42.201,
 -42.157, -41.818, -41.069, -40.303, -39.181, -37.746, -36.429, -34.585, -32.868, -30.840,
 -28.957, -26.965, -24.832, -22.455, -20.282, -18.114, -15.920, -13.886, -12.124, -10.407,
  -9.195,  -8.458,  -7.363,  -6.128,  -4.646,  -3.342,  -1.812,  -0.733,  -0.069,  -2.309,
  -3.456,  -2.449,  -0.253,   5.671,   8.558,   7.289,   5.521,   3.070,   2.554,   0.311,
  -1.633,  -4.189,  -5.925, -10.265, -16.480, -20.105, -18.161, -15.614, -21.502, -22.703,
 -24.595, -25.702, -29.506, -33.884, -35.898, -32.125, -37.926, -38.030, -42.985, -49.391,
 -49.045, -45.443, -48.004, -46.971, -45.952, -44.930, -44.709, -44.197, -43.356, -42.499,
 -41.012, -38.420, -35.243, -32.030, -29.042, -25.942, -23.678, -22.247, -20.464, -18.871,
 -16.774, -14.779, -12.841, -10.870,  -8.597,  -6.394,  -4.035,  -1.671,   0.763,   3.140,
   5.549,   7.771,   9.885,  11.849,  13.575,  14.568,  15.326,  16.485,  17.976,  18.811,
  20.003,  21.272,  27.278,  28.217,  29.243,  30.187,  30.297,  33.011,  33.382,  33.016,
  32.078,  30.570,  29.113,  28.829,  28.605,  27.856,  26.792,  26.059,  25.136,  24.155},
{
  25.095,  23.959,  22.965,  22.366,  20.802,  19.779,  19.493,  19.929,  20.888,  20.969,
  19.677,  18.694,  17.326,  15.424,  13.911,  13.225,  12.213,  10.716,   9.655,   8.111,
   6.766,   6.400,   7.183,   7.214,   7.607,   5.523,   3.654,   2.469,   1.947,   1.989,
   0.954,  -0.539,  -1.533,  -2.426,  -3.533,  -4.005,  -3.396,  -3.119,  -1.214,  -1.819,
  -4.688,  -8.007,  -9.041, -10.961, -11.331, -14.769, -17.970, -19.199, -21.380, -23.461,
 -25.906, -28.087, -31.252, -34.173, -36.768, -39.790, -42.392, -43.892, -47.171, -52.000,
 -56.743, -58.735, -61.532, -65.067, -68.254, -71.152, -74.228, -77.341, -80.134, -82.865,
 -85.671, -86.694, -87.515, -89.494, -88.869, -91.263, -87.666, -86.029, -86.507, -89.350,
 -91.754, -93.413, -92.369, -89.610, -86.993, -86.170, -81.987, -78.061, -73.372, -68.505,
 -64.643, -62.259, -62.231, -61.107, -57.366, -49.978, -46.776, -40.676, -38.292, -33.958,
 -31.763, -29.178, -25.302, -22.218, -19.700, -16.084, -12.022,  -7.265,  -2.300,   1.242,
   3.443,   6.204,  10.641,  15.101,  19.191,  23.808,  28.028,  32.109,  35.814,  38.280,
  37.842,  46.697,  49.791,  51.367,  53.136,  46.375,  44.967,  51.194,  52.625,  52.563,
  52.036,  51.891,  51.616,  52.016,  52.283,  53.643,  53.988,  54.642,  54.956,  53.788,
  52.936,  52.422,  53.394,  54.110,  52.992,  45.781,  31.760,  39.501,  46.510,  45.521,
  44.053,  42.164,  40.626,  39.453,  37.737,  36.131,  33.783,  34.831,  32.375,  31.304,
  29.980,  29.503,  28.548,  27.906,  25.940,  25.760,  24.336,  21.597,  21.254,  20.289,
  19.520,  18.402,  17.248,  15.998,  15.445,  14.456,  13.930,  13.298,  11.916,  11.963,   9.945,
  10.402,  10.973,  10.988,  10.075,   9.706,   9.573,  10.189,   9.593,   9.598,   9.649,
  10.420,  11.591,  11.619,  11.692,  11.122,  10.512,   9.543,   8.488,   7.712,   6.870,
   6.038,   4.954,   3.825,   2.915,   2.294,   1.061,  -0.080,  -1.306,  -2.245,  -2.755,
  -3.558,  -4.267,  -5.158,  -6.223,  -7.184,  -8.532,  -9.944, -11.361, -12.804, -14.453,
 -16.379, -18.170, -20.122, -22.032, -23.972, -25.982, -28.148, -30.018, -32.061, -33.673,
 -35.158, -36.527, -38.054, -39.336, -40.582, -41.661, -42.580, -43.042, -43.219, -43.242,
 -43.135, -43.086, -42.285, -41.309, -40.043, -38.561, -36.935, -35.314, -33.299, -31.485,
 -29.701, -27.528, -25.251, -22.794, -20.487, -18.172, -15.906, -13.873, -12.184, -10.514,
  -8.900,  -7.629,  -7.101,  -5.929,  -4.899,  -4.056,  -4.037,  -5.930,  -4.832,  -2.850,
  -1.186,  -0.414,   4.336,   6.801,   7.059,   5.509,   4.598,   1.887,   0.459,  -1.568,
  -3.915,  -6.434,  -8.231, -11.492, -16.899, -21.884, -24.407, -25.857, -26.388, -30.007,
 -34.747, -38.362, -39.973, -40.277, -36.670, -35.319, -38.209, -41.799, -38.744, -48.076,
 -48.433, -46.111, -47.447, -46.342, -45.561, -44.724, -44.240, -43.731, -42.788, -41.398,
 -39.660, -36.962, -33.702, -30.653, -27.148, -25.165, -23.741, -22.164, -20.216, -18.334,
 -16.577, -14.778, -12.883, -10.790,  -8.687,  -6.046,  -3.415,  -0.776,   1.761,   4.265,
   6.595,   8.846,  11.206,  13.012,  14.448,  15.743,  16.481,  17.499,  18.931,  20.121,
  21.409,  22.742,  28.622,  29.796,  30.966,  30.740,  31.018,  30.707,  31.330,  32.345,
  32.508,  30.759,  29.065,  28.481,  27.885,  28.115,  26.814,  26.062,  26.068,  25.095},
{
  23.751,  22.591,  23.025,  22.361,  21.103,  20.147,  19.081,  19.285,  20.143,  20.431,
  20.178,  18.923,  16.934,  15.232,  13.536,  13.683,  12.541,  11.205,  10.096,   9.383,
   8.198,   8.326,   9.591,   9.143,   8.626,   7.084,   5.393,   3.808,   3.956,   3.322,
   2.179,   1.076,  -0.176,  -1.050,  -1.604,  -1.779,  -2.742,  -4.783,  -5.438,  -1.557,
  -3.661,  -6.290,  -8.390,  -9.862,  -7.649,  -9.855, -13.512, -16.791, -19.048, -21.517,
 -24.073, -26.305, -29.289, -31.011, -33.698, -36.175, -38.751, -41.444, -45.364, -50.578,
 -54.346, -57.545, -59.694, -62.965, -66.332, -69.349, -71.915, -74.804, -77.860, -80.715,
 -83.279, -84.267, -86.018, -86.512, -86.368, -87.689, -86.030, -85.011, -86.637, -88.954,
 -90.304, -91.541, -90.604, -87.959, -86.384, -86.147, -80.241, -76.977, -72.971, -68.510,
 -64.578, -61.807, -61.843, -60.054, -58.728, -50.516, -46.208, -42.390, -38.144, -35.033,
 -33.153, -30.011, -26.585, -24.179, -21.367, -18.078, -14.087,  -8.831,  -4.468,  -2.050,
  -0.473,   3.311,   8.397,  12.485,  16.587,  21.075,  25.511,  29.938,  33.586,  33.769,
  36.434,  43.503,  47.889,  50.390,  52.478,  43.900,  46.293,  49.570,  50.113,  49.956,
  49.843,  49.940,  49.611,  50.048,  50.621,  51.830,  52.155,  52.821,  53.069,  53.111,
  53.022,  52.610,  52.784,  54.450,  54.722,  54.949,  42.874,  32.106,  42.883,  44.324,
  43.485,  41.723,  39.626,  38.428,  36.860,  34.170,  36.494,  33.810,  31.406,  29.887,
  28.835,  29.816,  26.921,  25.901,  27.293,  23.665,  24.053,  23.954,  21.793,  18.921,
  17.540,  16.553,  15.532,  14.234,  14.105,  13.622,  13.704,  12.422,  11.159,   9.019,   9.262,
   9.678,   9.761,   9.523,   9.224,   8.894,   8.786,   8.875,   9.391,   9.004,  10.241,
  10.513,  11.036,  10.965,  11.208,  10.022,   9.540,   8.505,   7.538,   6.545,   5.920,
   5.022,   3.971,   2.835,   1.931,   0.960,  -0.493,  -1.762,  -2.705,  -3.404,  -3.801,
  -4.424,  -5.276,  -6.447,  -7.341,  -8.454,  -9.695, -11.009, -12.273, -13.772, -15.498,
 -17.174, -18.711, -20.712, -22.958, -25.046, -27.166, -29.279, -31.512, -33.319, -34.974,
 -36.333, -37.614, -39.089, -40.472, -41.620, -42.653, -43.517, -43.882, -44.045, -44.172,
 -44.065, -43.695, -43.068, -42.124, -40.543, -38.919, -37.417, -35.884, -33.708, -32.024,
 -29.830, -27.771, -25.459, -23.176, -20.820, -18.384, -16.109, -14.463, -12.511, -10.658,
  -8.926,  -7.555,  -6.724,  -6.172,  -6.304,  -9.336,  -7.440,  -4.189,  -2.915,   0.467,
   2.576,   3.642,   5.349,   5.550,   5.575,   3.445,   2.327,   0.571,  -0.594,  -2.804,
  -5.947,  -7.564, -12.088, -14.670, -17.714, -20.811, -22.513, -24.877, -29.353, -32.964,
 -34.773, -36.087, -36.596, -37.206, -37.385, -37.459, -37.749, -41.252, -39.031, -46.629,
 -53.415, -48.688, -48.250, -47.714, -47.401, -45.952, -44.368, -43.209, -41.867, -40.544,
 -38.226, -35.304, -32.043, -28.993, -25.734, -24.621, -23.085, -21.433, -19.952, -17.680,
 -15.787, -13.805, -11.493,  -9.170,  -6.992,  -4.584,  -2.034,   0.371,   3.053,   5.678,
   7.966,  10.456,  12.940,  15.114,  16.543,  17.619,  18.238,  19.028,  20.242,  21.483,
  23.553,  25.710,  29.012,  30.199,  30.491,  31.241,  32.188,  30.377,  30.695,  31.291,
  31.220,  30.115,  29.236,  28.554,  28.207,  29.075,  26.595,  25.264,  24.616,  23.751},
{
  22.245,  22.114,  22.790,  22.268,  20.681,  20.817,  20.190,  19.383,  19.409,  20.108,
  19.774,  18.788,  16.708,  15.278,  14.532,  13.685,  12.896,  11.686,  10.491,  10.028,
  10.061,   9.268,  10.720,  10.621,   9.609,   8.815,   7.304,   6.605,   5.522,   4.712,
   3.651,   2.900,   1.800,   0.291,   0.055,   0.014,  -0.934,  -2.816,  -2.834,   1.058,
  -3.579,  -5.935,  -7.279,  -8.962,  -5.347,  -8.346, -12.188, -14.618, -15.833, -17.942,
 -21.102, -23.966, -26.846, -29.138, -31.622, -34.035, -36.528, -40.119, -44.670, -50.386,
 -53.580, -56.723, -58.007, -60.875, -64.095, -67.039, -69.878, -72.649, -75.999, -79.147,
 -80.708, -82.164, -83.718, -82.569, -83.972, -83.478, -83.502, -84.855, -87.109, -88.429,
 -86.792, -87.762, -88.767, -86.696, -84.402, -83.145, -79.294, -76.026, -72.477, -68.520,
 -64.136, -61.180, -60.263, -62.015, -55.954, -50.176, -46.636, -43.097, -38.873, -35.976,
 -34.152, -31.292, -29.043, -26.994, -24.488, -21.198, -16.802, -12.113,  -7.348,  -5.427,
  -2.657,   1.521,   5.876,   9.406,  13.667,  18.009,  22.651,  26.964,  31.448,  31.930,
  40.614,  42.516,  41.154,  41.604,  44.059,  42.282,  45.467,  47.087,  47.347,  47.630,
  47.512,  48.014,  48.452,  48.737,  49.402,  50.329,  50.667,  51.316,  51.907,  52.134,
  52.366,  52.181,  52.196,  53.704,  54.827,  54.991,  49.670,  36.966,  39.610,  43.449,
  43.862,  41.648,  40.090,  38.182,  36.300,  34.603,  30.606,  30.169,  29.456,  29.449,
  28.586,  27.750,  25.136,  25.695,  23.656,  22.636,  22.468,  18.989,  17.615,  19.183,
  16.437,  15.332,  14.350,  13.099,  12.641,  12.018,  12.278,  11.334,  10.176,   8.892,   8.807,
   8.730,   8.461,   8.665,   8.196,   7.924,   8.316,   8.430,   8.387,   9.270,   8.084,
   9.681,  10.393,  11.008,   9.788,   9.560,   8.752,   8.041,   7.126,   6.223,   5.333,
   4.500,   3.440,   2.699,   1.871,   0.796,  -0.794,  -2.193,  -3.222,  -3.824,  -4.295,
  -5.075,  -6.422,  -7.645,  -8.770, -10.230, -11.678, -13.124, -14.274, -15.184, -16.504,
 -18.247, -20.114, -22.239, -24.195, -26.103, -28.265, -30.569, -32.534, -34.502, -36.113,
 -37.581, -39.111, -40.059, -41.320, -42.423, -43.516, -44.311, -44.720, -44.955, -44.844,
 -44.912, -44.369, -43.615, -42.387, -41.067, -39.494, -37.837, -35.977, -33.871, -32.351,
 -30.569, -28.615, -26.013, -23.479, -21.005, -18.663, -16.751, -14.763, -12.885, -11.061,
  -9.589,  -9.416,  -9.082,  -9.198, -10.170,  -5.231,  -5.357,   0.948,   3.820,   2.127,
   1.753,   2.288,   4.204,   4.284,   3.375,   1.040,  -0.624,  -1.169,  -3.246,  -5.725,
  -7.663,  -9.534, -13.830, -15.269, -17.959, -21.000, -21.248, -23.969, -28.213, -30.155,
 -31.982, -34.686, -36.897, -37.740, -39.787, -39.122, -39.674, -42.056, -40.270, -55.380,
 -55.591, -49.108, -49.513, -49.379, -48.842, -47.015, -44.495, -42.845, -40.846, -38.256,
 -35.567, -32.953, -30.662, -28.057, -25.770, -24.503, -23.140, -21.365, -19.168, -16.927,
 -14.748, -12.660, -10.239,  -7.842,  -5.431,  -2.954,  -0.259,   2.564,   5.371,   7.846,
  10.289,  12.830,  15.454,  17.537,  20.440,  24.834,  25.235,  21.321,  21.612,  23.172,
  24.752,  27.826,  30.472,  29.379,  30.921,  31.997,  33.193,  29.307,  29.658,  30.109,
  29.775,  29.084,  28.343,  28.009,  28.333,  29.100,  28.228,  25.640,  24.020,  22.245},
{
  24.140,  23.332,  24.188,  23.071,  21.466,  21.613,  20.819,  20.505,  20.876,  21.291,
  20.885,  18.831,  17.222,  16.707,  15.413,  14.118,  13.041,  12.295,  10.872,  10.453,
  11.305,  10.654,  11.496,  12.146,  11.097,   9.907,   8.755,   8.437,   7.436,   6.216,
   5.088,   4.607,   3.059,   2.337,   1.846,   0.960,   0.268,   0.481,   1.353,  -0.581,
  -3.289,  -4.078,  -5.370,  -5.987,  -4.542,  -8.618, -11.807, -14.846, -17.389, -19.302,
 -21.086, -22.699, -25.776, -28.683, -30.876, -33.372, -36.154, -39.512, -43.759, -47.022,
 -52.462, -55.936, -56.370, -58.671, -61.704, -64.629, -67.349, -70.499, -74.046, -75.460,
 -77.830, -79.894, -80.491, -79.735, -80.160, -80.389, -81.181, -83.222, -84.564, -84.568,
 -83.431, -81.124, -81.912, -84.028, -83.393, -81.195, -78.091, -75.051, -71.792, -68.078,
 -64.080, -60.969, -59.619, -61.373, -56.348, -50.713, -46.961, -43.839, -39.928, -36.752,
 -35.141, -32.094, -30.512, -29.583, -26.825, -23.453, -19.343, -14.269, -11.745,  -9.048,
  -6.312,  -1.119,   3.717,   7.226,  13.080,  16.279,  19.750,  24.317,  28.715,  30.801,
  41.033,  42.621,  37.840,  33.232,  41.375,  44.490,  44.695,  44.895,  45.276,  45.472,
  46.067,  46.015,  46.409,  46.421,  47.162,  48.649,  49.134,  49.750,  50.658,  51.111,
  51.165,  51.634,  51.536,  52.964,  53.655,  53.682,  52.657,  39.536,  36.164,  43.177,
  42.363,  40.585,  41.066,  36.715,  34.886,  36.304,  30.915,  29.643,  29.188,  27.642,
  28.153,  24.575,  23.705,  25.701,  22.724,  21.505,  20.312,  19.596,  18.140,  16.590,
  15.694,  14.822,  13.220,  12.348,  11.844,  11.062,  11.268,  10.245,   9.291,   8.276,   7.321,
   7.428,   7.604,   7.557,   7.739,   7.633,   7.574,   7.657,   8.344,   8.676,   9.648,
   9.901,  10.315,   9.456,   9.576,   9.252,   8.295,   7.627,   6.974,   5.860,   5.146,
   4.495,   4.165,   3.726,   2.988,   1.436,  -0.018,  -1.643,  -3.162,  -4.345,  -4.833,
  -5.786,  -7.233,  -8.401,  -9.687, -11.347, -12.706, -14.334, -15.789, -17.367, -19.074,
 -20.831, -22.335, -23.800, -25.592, -27.485, -29.395, -31.572, -33.578, -35.320, -37.013,
 -38.615, -39.929, -41.150, -42.195, -42.953, -43.928, -44.753, -45.141, -45.662, -45.674,
 -45.302, -44.635, -44.065, -43.177, -41.574, -39.759, -38.284, -36.471, -34.476, -32.815,
 -30.638, -29.218, -26.494, -23.556, -21.217, -19.345, -17.241, -15.399, -14.560, -14.293,
 -11.996,  -7.707,  -2.562,  -3.420,  -7.602,  -6.303,  -2.494,  -1.508,  -2.647,  -1.857,
  -1.492,  -3.105,  -1.735,  -0.477,  -1.011,  -1.643,  -2.394,  -3.611,  -5.112,  -7.157,
 -10.162, -12.732, -14.551, -16.872, -19.330, -22.735, -24.789, -27.042, -29.246, -31.468,
 -33.913, -36.669, -39.281, -40.739, -42.138, -41.603, -41.374, -41.163, -42.406, -61.322,
 -54.690, -49.441, -46.050, -46.043, -47.286, -46.129, -44.210, -42.021, -39.910, -37.762,
 -34.479, -32.026, -28.905, -26.637, -25.234, -23.624, -21.856, -20.237, -18.158, -15.901,
 -13.598, -11.252,  -8.370,  -5.400,  -3.228,  -0.501,   2.550,   5.191,   7.711,  10.109,
  12.627,  14.989,  17.414,  19.244,  21.213,  22.615,  31.100,  25.145,  23.584,  24.978,
  25.823,  27.106,  29.723,  30.052,  31.536,  33.221,  31.784,  27.665,  27.857,  28.525,
  28.598,  27.929,  27.911,  27.766,  28.090,  28.526,  28.712,  28.147,  26.847,  24.140},
{
  27.104,  26.426,  25.330,  23.606,  22.402,  22.690,  22.097,  21.650,  22.493,  23.873,
  22.985,  21.135,  19.977,  18.870,  17.109,  14.891,  13.607,  12.873,  10.981,  10.034,
  11.301,  10.986,  11.883,  12.948,  12.399,  11.193,  10.156,   9.749,   8.500,   7.703,
   6.592,   5.453,   4.283,   3.523,   2.844,   2.606,   2.089,   0.727,   0.144,  -1.139,
  -1.881,  -2.966,  -3.543,  -3.819,  -4.340,  -8.488, -12.281, -16.229, -19.810, -21.895,
 -22.782, -23.773, -25.137, -28.492, -32.009, -33.860, -35.650, -39.140, -42.497, -44.266,
 -49.706, -54.978, -54.604, -56.469, -59.246, -61.718, -64.314, -68.789, -70.885, -72.083,
 -74.583, -77.417, -75.545, -77.433, -78.057, -76.942, -78.019, -79.440, -78.600, -79.833,
 -80.012, -77.210, -75.383, -77.989, -80.114, -80.167, -76.902, -73.326, -70.711, -67.155,
 -63.282, -60.626, -59.531, -60.017, -57.811, -51.304, -47.862, -44.153, -41.133, -38.152,
 -36.560, -33.128, -31.592, -29.650, -27.129, -24.596, -19.816, -17.360, -14.519, -10.533,
  -7.635,  -4.829,   0.587,   4.371,   7.915,  12.329,  16.425,  21.133,  25.224,  28.128,
  34.306,  43.318,  40.548,  32.912,  39.762,  42.920,  42.304,  42.435,  43.001,  43.296,
  43.825,  43.939,  44.127,  44.399,  45.280,  46.718,  47.615,  48.650,  49.566,  49.966,
  50.845,  50.670,  50.641,  52.326,  52.603,  52.075,  48.828,  38.184,  31.384,  40.059,
  44.069,  40.425,  41.043,  36.857,  37.228,  32.430,  31.134,  29.924,  28.197,  26.991,
  25.818,  24.291,  22.765,  23.676,  21.474,  20.342,  19.053,  18.202,  16.592,  15.511,
  14.384,  13.225,  11.725,  10.853,  10.539,   9.650,   9.293,   8.897,   8.063,   7.038,   6.885,
   6.750,   5.948,   7.028,   7.026,   7.644,   7.922,   8.268,   9.095,   9.331,   9.772,
   9.702,  10.141,   9.434,   9.011,   8.348,   7.944,   6.987,   6.368,   6.277,   5.562,
   5.499,   5.254,   4.796,   3.560,   1.802,   0.037,  -1.406,  -3.132,  -4.316,  -5.406,
  -6.597,  -7.831,  -9.088, -10.437, -12.252, -13.667, -15.455, -17.361, -18.938, -20.537,
 -21.981, -23.395, -25.395, -27.393, -29.274, -31.099, -33.075, -34.534, -36.312, -37.978,
 -39.366, -40.364, -41.513, -42.779, -43.661, -44.525, -45.261, -45.857, -45.758, -46.056,
 -45.929, -45.151, -43.914, -43.508, -42.207, -40.476, -38.590, -36.638, -34.818, -33.024,
 -31.449, -29.587, -26.949, -24.047, -21.824, -19.572, -20.007, -19.008, -14.297,  -8.485,
  -4.809,  -3.399,  -1.887,  -6.489,  -8.646,  -7.178,  -5.046,  -4.827,  -6.087,  -4.754,
  -3.591,  -5.384,  -8.567,  -8.584,  -7.605,  -5.668,  -6.084,  -6.824,  -8.262, -10.349,
 -12.285, -13.847, -16.545, -18.741, -21.061, -25.496, -29.100, -30.074, -32.862, -37.650,
 -41.444, -42.986, -44.461, -45.229, -45.195, -43.350, -41.771, -40.995, -55.832, -62.025,
 -52.918, -49.118, -49.355, -48.015, -46.678, -44.039, -42.101, -40.721, -38.802, -36.087,
 -33.613, -31.008, -27.907, -25.342, -24.217, -22.492, -21.033, -19.051, -16.967, -14.950,
 -12.730, -10.073,  -7.024,  -4.091,  -1.301,   1.614,   4.713,   7.614,  10.333,  12.699,
  14.667,  16.368,  18.313,  21.204,  29.185,  25.150,  27.099,  25.238,  25.323,  26.238,
  26.619,  27.458,  29.778,  31.613,  32.165,  33.849,  30.446,  27.350,  26.637,  27.671,
  27.603,  27.588,  27.795,  27.831,  27.467,  28.130,  28.668,  28.075,  27.605,  27.104},
{
  28.980,  28.284,  26.929,  25.512,  23.756,  23.438,  22.998,  22.839,  24.230,  27.037,
  25.402,  22.949,  21.997,  20.076,  18.005,  16.294,  15.226,  14.868,  13.151,  12.230,
  11.837,  10.981,  10.974,  11.688,  12.617,  12.194,  11.361,  10.616,   9.032,   8.344,
   7.460,   6.456,   5.635,   5.339,   4.312,   3.980,   3.506,   2.987,   0.495,  -1.238,
  -1.096,  -2.272,  -2.002,  -0.441,  -3.549,  -9.151, -14.086, -17.825, -21.959, -25.308,
 -26.825, -27.731, -28.915, -32.067, -34.441, -34.230, -34.795, -35.560, -39.878, -43.653,
 -47.753, -53.802, -52.691, -53.931, -55.997, -58.480, -62.168, -65.639, -66.248, -68.966,
 -71.990, -70.406, -71.170, -74.969, -74.023, -74.598, -74.901, -74.634, -74.827, -75.037,
 -75.333, -74.971, -71.740, -69.963, -72.082, -76.078, -73.910, -71.266, -68.947, -66.600,
 -63.089, -59.965, -59.214, -58.736, -54.845, -49.931, -47.967, -44.406, -41.563, -39.291,
 -37.238, -34.321, -32.481, -31.907, -29.982, -24.756, -21.199, -19.484, -16.043, -12.626,
  -8.483,  -6.365,  -3.548,   0.835,   4.901,   9.283,  13.368,  17.980,  22.140,  24.882,
  30.586,  39.063,  37.027,  31.476,  35.516,  38.355,  39.628,  39.925,  41.007,  41.478,
  41.420,  41.812,  42.281,  42.555,  43.485,  44.692,  46.058,  47.290,  48.268,  49.314,
  50.042,  50.132,  49.941,  51.560,  51.219,  49.936,  44.180,  36.946,  30.938,  39.007,
  40.858,  42.708,  37.612,  36.275,  33.466,  32.110,  30.479,  28.384,  26.316,  25.122,
  24.110,  22.413,  21.452,  20.953,  20.105,  18.618,  17.584,  17.535,  15.923,  14.187,
  12.422,  14.570,  12.334,  10.674,   9.778,   8.968,   8.013,   7.275,   7.178,   6.490,   6.085,
   6.046,   6.030,   6.506,   8.257,   9.300,   8.930,   8.311,   8.895,   8.588,   8.994,
   9.811,   8.766,   9.083,   8.297,   7.707,   7.741,   7.246,   6.597,   6.800,   6.844,
   6.946,   6.843,   5.600,   3.458,   1.324,  -0.524,  -1.578,  -3.089,  -4.610,  -5.902,
  -7.225,  -8.425,  -9.562, -11.073, -13.019, -14.757, -16.401, -18.368, -19.997, -21.834,
 -23.319, -24.745, -26.438, -28.323, -29.960, -31.976, -33.926, -35.638, -37.395, -38.700,
 -39.834, -41.115, -42.191, -43.121, -44.114, -45.033, -45.819, -46.442, -46.805, -46.757,
 -46.344, -45.679, -44.761, -43.865, -42.420, -40.844, -39.055, -36.980, -35.127, -33.349,
 -31.666, -29.890, -27.634, -24.737, -22.910, -23.410, -18.330, -14.294,  -8.504,  -7.010,
  -5.691,  -4.101,  -5.795, -12.111, -11.410, -12.716, -11.244, -10.093,  -8.023,  -6.843,
  -6.263,  -8.185, -10.969, -10.937, -11.698, -12.240, -12.676, -11.238, -12.137, -14.964,
 -15.805, -15.111, -15.042, -20.030, -22.868, -24.847, -27.965, -29.335, -33.755, -38.171,
 -41.289, -45.442, -42.476, -41.688, -45.061, -45.685, -42.948, -49.142, -61.905, -57.849,
 -50.897, -48.410, -47.013, -45.491, -43.993, -41.785, -40.665, -39.372, -37.599, -35.396,
 -32.651, -29.837, -26.815, -24.029, -22.639, -20.738, -19.288, -17.666, -15.613, -13.661,
 -11.566,  -9.299,  -6.354,  -3.391,  -0.146,   3.123,   6.383,   9.477,  12.186,  14.432,
  16.186,  17.992,  19.697,  21.805,  23.628,  24.693,  25.430,  25.968,  26.681,  26.140,
  26.296,  26.472,  28.367,  32.799,  32.356,  34.679,  30.850,  29.518,  28.628,  28.298,
  28.037,  27.610,  27.307,  26.863,  26.833,  27.013,  27.316,  28.631,  28.812,  28.980},
{
  30.837,  30.341,  30.179,  27.653,  25.827,  24.999,  24.003,  23.268,  24.776,  28.501,
  26.481,  24.652,  22.852,  20.576,  18.568,  17.911,  17.702,  17.831,  16.926,  15.752,
  12.916,  11.465,  11.221,  11.655,  12.698,  12.449,  12.362,  12.525,  10.535,   9.142,
   7.925,   7.636,   7.188,   6.409,   5.411,   4.885,   5.429,   4.790,   1.809,   0.964,
  -0.255,  -0.529,   0.775,  -0.172,  -4.652, -10.146, -14.789, -18.700, -23.147, -26.843,
 -29.328, -31.364, -32.286, -34.320, -36.191, -36.632, -36.035, -34.719, -36.685, -42.018,
 -43.200, -49.995, -50.044, -51.555, -54.073, -56.451, -58.375, -60.000, -63.010, -67.359,
 -65.318, -66.701, -68.752, -69.341, -70.083, -70.413, -71.145, -71.781, -71.395, -71.625,
 -72.426, -69.934, -68.186, -65.161, -66.734, -67.497, -70.040, -68.212, -66.755, -64.817,
 -62.929, -60.109, -59.613, -56.597, -51.028, -49.314, -48.121, -43.961, -41.484, -39.341,
 -37.499, -34.562, -33.413, -30.570, -29.038, -26.914, -23.615, -21.452, -17.835, -13.898,
 -10.467,  -6.918,  -3.850,   0.127,   3.546,   6.744,  10.496,  14.571,  19.393,  22.061,
  23.796,  30.166,  32.601,  30.213,  33.263,  35.324,  36.751,  38.000,  38.838,  39.314,
  39.758,  39.879,  40.493,  41.049,  41.825,  42.622,  44.870,  46.353,  47.473,  48.606,
  49.307,  49.347,  49.287,  50.886,  50.117,  47.996,  42.227,  34.772,  32.978,  38.277,
  38.105,  37.783,  36.628,  35.401,  33.054,  31.105,  28.554,  26.191,  25.330,  22.677,
  22.119,  21.019,  20.185,  19.013,  17.275,  18.989,  15.160,  16.589,  14.011,  12.076,
  11.254,  10.366,   9.369,   8.576,   7.473,   7.256,   6.682,   6.376,   5.592,   5.504,   5.387,
   5.198,   5.316,   5.317,   6.186,   6.776,   8.277,   8.792,   8.262,   9.840,   9.376,
   9.934,   8.979,   8.656,   8.232,   7.984,   7.864,   7.590,   7.277,   7.514,   7.480,
   7.276,   7.269,   6.118,  13.203,   7.873,  -1.555,  -2.485,  -3.405,  -5.187,  -6.779,
  -7.856,  -9.093, -10.403, -12.098, -13.933, -15.533, -17.134, -19.166, -20.858, -22.468,
 -23.845, -25.336, -27.035, -28.999, -30.908, -32.855, -34.678, -36.159, -37.771, -39.169,
 -40.429, -41.657, -42.931, -43.727, -44.443, -45.654, -46.324, -46.823, -47.351, -47.189,
 -46.779, -46.213, -45.400, -44.190, -42.795, -41.228, -39.250, -37.262, -34.968, -33.674,
 -31.692, -30.316, -27.994, -27.938, -24.708, -19.195, -15.972, -12.968, -10.749,  -6.591,
  -5.209,  -4.246,  -6.469, -12.995, -13.345, -13.514, -11.877, -11.294, -10.023,  -8.827,
  -7.394,  -6.787, -12.327, -12.358, -13.952, -13.241, -14.300, -16.173, -16.362, -19.553,
 -22.293, -24.345, -26.453, -26.542, -27.496, -27.188, -28.046, -27.275, -28.882, -35.709,
 -44.231, -55.379, -62.241, -60.947, -55.453, -55.498, -60.784, -62.033, -57.375, -51.697,
 -48.248, -48.085, -47.377, -45.293, -42.795, -41.476, -40.495, -38.779, -36.998, -34.331,
 -31.682, -28.639, -25.131, -22.536, -20.673, -19.030, -17.432, -15.631, -13.800, -11.992,
 -10.011,  -7.795,  -5.172,  -2.279,   0.958,   4.413,   7.950,  11.030,  13.622,  15.426,
  16.951,  18.283,  20.001,  22.053,  23.349,  24.396,  25.692,  26.800,  26.683,  26.176,
  26.228,  26.035,  28.729,  33.071,  33.895,  35.095,  32.112,  31.305,  30.734,  29.617,
  28.494,  27.559,  26.839,  26.276,  25.929,  26.444,  27.022,  28.157,  29.965,  30.837},
{
  31.445,  31.433,  31.546,  30.007,  28.106,  26.666,  25.489,  24.178,  24.607,  26.649,
  26.504,  25.056,  22.833,  21.220,  20.012,  20.147,  21.541,  22.845,  24.088,  20.298,
  14.434,  12.137,  11.896,  12.433,  13.283,  13.656,  13.576,  13.519,  11.873,  10.783,
   9.543,   8.622,   8.645,   8.002,   6.952,   6.321,   7.074,   5.667,   2.647,   1.960,
   1.925,   1.880,   3.462,  -1.151,  -5.087,  -9.909, -15.181, -19.773, -24.113, -28.025,
 -30.885, -32.930, -34.250, -35.340, -36.372, -36.503, -35.424, -33.307, -35.484, -38.660,
 -42.720, -47.845, -49.320, -49.920, -51.848, -54.730, -55.696, -57.496, -61.190, -62.673,
 -60.900, -63.904, -65.377, -66.217, -65.973, -66.722, -66.988, -67.450, -67.938, -68.392,
 -67.924, -66.904, -65.169, -65.129, -63.041, -63.243, -63.900, -64.177, -62.245, -60.897,
 -59.527, -57.037, -54.920, -52.811, -50.097, -51.185, -48.132, -43.621, -41.573, -39.250,
 -37.448, -35.306, -33.711, -31.367, -29.683, -28.058, -25.120, -23.408, -20.635, -16.623,
 -12.674,  -9.310,  -5.892,  -1.877,   1.801,   5.360,   9.273,  13.242,  16.598,  20.100,
  21.269,  22.427,  30.723,  27.974,  31.119,  32.589,  34.461,  35.469,  36.471,  37.391,
  37.807,  38.466,  39.010,  39.813,  40.659,  41.601,  43.607,  45.457,  46.523,  48.078,
  48.859,  48.793,  48.759,  50.196,  49.238,  46.608,  40.519,  31.172,  35.643,  41.115,
  40.881,  37.463,  35.701,  32.993,  31.367,  28.303,  27.953,  26.807,  23.815,  20.932,
  21.635,  20.914,  19.446,  18.376,  15.180,  14.335,  13.473,  12.805,  11.368,   9.885,
   8.987,   6.965,   7.936,   5.704,   5.548,   5.166,   6.507,   5.390,   4.861,   4.455,   4.001,
   4.543,   4.224,   3.657,   4.340,   4.718,   5.877,   7.782,   8.195,   9.279,   8.965,
   8.852,   9.479,   8.974,   8.691,   8.613,   8.757,   8.591,   8.253,   7.861,   7.547,
   6.597,   5.368,   6.130,  18.149,  12.436,  -1.837,  -2.781,  -3.842,  -5.443,  -6.966,
  -8.442,  -9.965, -11.325, -12.849, -14.811, -16.546, -18.046, -19.698, -21.536, -22.822,
 -24.202, -25.892, -27.948, -29.494, -31.374, -33.425, -34.997, -36.397, -37.809, -39.386,
 -40.883, -42.040, -43.281, -44.405, -45.298, -46.167, -46.954, -47.434, -47.569, -47.576,
 -47.134, -46.358, -45.569, -44.431, -43.011, -41.478, -39.651, -37.646, -35.724, -33.938,
 -32.142, -30.325, -29.049, -29.377, -20.856, -15.905, -14.311, -11.836,  -9.508,  -7.020,
  -5.961,  -5.580,  -9.682, -15.108, -19.200, -17.323, -15.978, -13.493, -11.654, -10.723,
 -10.087,  -8.738, -12.657, -15.646, -14.806, -17.054, -17.478, -18.082, -19.107, -19.246,
 -20.908, -22.599, -24.633, -26.136, -27.440, -35.951, -36.585, -41.133, -44.850, -47.155,
 -51.000, -58.834, -62.227, -64.611, -62.910, -59.238, -57.270, -52.976, -49.849, -47.213,
 -46.111, -45.447, -46.130, -45.018, -42.581, -40.988, -40.001, -37.900, -35.505, -32.982,
 -30.161, -26.902, -23.982, -20.668, -18.396, -17.043, -15.206, -13.767, -12.023, -10.238,
  -8.318,  -6.237,  -3.668,  -1.007,   2.015,   5.805,   9.345,  12.630,  15.290,  16.998,
  17.920,  18.637,  19.913,  21.416,  22.638,  23.430,  24.552,  25.555,  25.350,  25.261,
  25.671,  27.676,  31.206,  32.595,  33.633,  33.649,  33.617,  32.959,  31.860,  30.315,
  28.811,  27.271,  26.799,  26.719,  26.909,  27.031,  26.556,  28.540,  30.065,  31.445},
{
  30.261,  31.283,  31.718,  31.606,  30.013,  29.152,  27.853,  26.177,  26.040,  27.712,
  27.053,  25.013,  23.772,  22.225,  21.100,  22.067,  24.675,  28.835,  29.591,  24.499,
  16.606,  14.245,  13.392,  12.730,  13.216,  14.913,  15.147,  13.799,  12.526,  11.327,
  10.155,   9.392,   9.505,   9.187,   8.507,   8.654,   7.837,   6.432,   4.088,   3.388,
   4.492,   5.798,   3.880,  -0.433,  -3.833,  -8.183, -13.296, -19.076, -24.198, -28.013,
 -31.342, -33.140, -34.652, -35.230, -35.917, -36.040, -35.264, -34.146, -32.945, -35.051,
 -40.368, -43.019, -46.449, -47.575, -49.835, -54.943, -55.463, -56.639, -58.235, -57.019,
 -56.819, -59.089, -60.783, -61.645, -62.870, -63.495, -63.319, -63.312, -63.657, -64.448,
 -64.719, -64.695, -64.220, -63.505, -61.861, -59.935, -60.191, -61.556, -59.326, -57.567,
 -57.172, -56.322, -54.493, -52.270, -51.302, -51.803, -47.658, -43.326, -41.746, -39.446,
 -37.267, -35.773, -34.090, -32.454, -31.107, -29.378, -27.477, -24.180, -21.799, -18.849,
 -14.684, -10.715,  -7.977,  -4.205,  -0.896,   3.444,   8.107,  12.443,  15.311,  17.927,
  18.538,  20.733,  27.500,  25.901,  29.019,  30.783,  32.560,  34.070,  35.344,  35.904,
  36.193,  37.205,  37.952,  38.858,  39.721,  40.707,  41.968,  44.538,  45.902,  47.468,
  48.324,  48.294,  48.431,  48.993,  47.551,  43.811,  34.819,  29.460,  37.633,  36.670,
  36.493,  33.707,  32.784,  32.574,  28.862,  29.737,  25.702,  25.238,  23.286,  19.436,
  19.278,  18.007,  15.989,  16.409,  13.812,  11.993,  11.589,  10.417,   9.253,   8.467,
   7.159,   6.247,   5.091,   4.344,   4.593,   4.469,   5.191,   5.560,   3.804,   3.866,   2.642,
   2.487,   2.170,   1.831,   3.233,   3.829,   5.021,   5.857,   6.903,   7.669,   8.268,
   8.965,   9.188,   9.694,   9.314,   9.156,   9.080,   8.804,   8.088,   7.185,   6.429,
   6.071,  12.054,  15.271,   9.833,  -0.362,  -1.941,  -2.525,  -4.317,  -6.089,  -7.880,
  -9.268, -10.734, -12.426, -13.904, -15.660, -17.431, -18.721, -20.203, -21.995, -23.244,
 -24.737, -26.592, -28.421, -30.316, -32.148, -33.573, -35.059, -36.472, -37.646, -39.528,
 -41.018, -42.150, -43.390, -44.540, -45.575, -46.317, -47.087, -47.541, -47.702, -47.487,
 -47.336, -46.624, -45.948, -44.587, -43.110, -41.757, -39.998, -38.073, -36.139, -34.202,
 -32.234, -30.571, -30.165, -27.966, -21.793, -18.418, -15.055, -12.608, -11.052, -10.317,
 -10.042, -12.678, -15.379, -19.087, -22.105, -21.419, -20.846, -14.238, -13.279, -13.624,
 -12.400, -11.715, -11.306, -17.810, -19.041, -19.712, -20.276, -21.693, -22.843, -23.508,
 -21.382, -23.245, -25.276, -28.532, -34.213, -41.624, -43.581, -45.523, -44.019, -48.024,
 -51.175, -51.247, -51.252, -52.113, -51.842, -51.376, -50.777, -50.517, -49.740, -48.882,
 -47.174, -45.092, -44.366, -43.774, -42.499, -40.770, -38.726, -36.333, -34.013, -31.264,
 -28.592, -25.580, -22.411, -18.864, -16.451, -14.920, -13.427, -11.951, -10.243,  -8.247,
  -6.403,  -4.239,  -1.822,   1.078,   3.887,   7.334,  10.593,  13.652,  15.998,  17.444,
  17.734,  17.820,  18.741,  20.051,  21.077,  22.376,  23.320,  23.905,  23.981,  23.872,
  24.735,  28.140,  31.615,  34.733,  35.801,  35.292,  35.455,  35.160,  32.981,  31.549,
  30.011,  28.666,  27.801,  27.281,  27.385,  27.390,  26.989,  28.908,  29.214,  30.261},
{
  30.432,  31.117,  31.502,  32.335,  31.334,  31.820,  31.410,  29.701,  29.836,  29.742,
  27.404,  27.014,  25.195,  23.673,  23.179,  24.184,  26.439,  28.132,  28.108,  25.715,
  19.863,  16.810,  14.688,  13.343,  13.357,  14.794,  15.399,  13.787,  13.033,  11.988,
  10.947,   9.938,   9.243,   9.205,  10.308,  10.523,   8.445,   6.226,   4.951,   4.498,
   5.790,   5.899,   4.151,   1.105,  -2.317,  -5.739,  -9.883, -15.872, -22.583, -27.327,
 -30.836, -32.774, -33.895, -34.316, -34.430, -34.530, -35.050, -33.136, -31.639, -31.600,
 -36.154, -41.055, -41.935, -45.088, -44.330, -48.811, -52.291, -53.072, -53.737, -54.026,
 -53.999, -55.971, -57.688, -59.575, -59.764, -59.951, -60.767, -60.087, -59.243, -59.538,
 -60.325, -61.259, -62.138, -61.658, -60.037, -58.122, -57.708, -59.061, -58.274, -57.080,
 -56.594, -55.908, -54.064, -51.111, -51.250, -51.373, -48.031, -44.201, -41.951, -40.005,
 -37.901, -36.761, -34.274, -32.666, -30.396, -30.018, -28.726, -25.965, -23.015, -20.234,
 -17.398, -13.591,  -9.540,  -5.743,  -2.269,   1.199,   5.472,  10.243,  13.961,  17.282,
  18.852,  21.698,  23.402,  24.078,  26.768,  28.288,  30.392,  32.325,  33.921,  35.110,
  35.165,  36.014,  36.972,  38.288,  39.471,  40.052,  40.848,  43.831,  45.325,  46.939,
  47.593,  47.473,  48.928,  47.185,  42.930,  32.953,  25.733,  32.055,  36.921,  37.272,
  35.874,  34.043,  32.218,  29.583,  27.028,  25.505,  23.242,  21.873,  19.622,  19.187,
  16.527,  16.032,  15.458,  12.500,  11.391,  10.332,   7.981,   7.354,   6.534,   6.009,
   4.699,   3.732,   2.986,   3.176,   1.814,   3.795,   2.015,   2.098,   2.636,   1.559,   1.378,
   1.412,   0.910,   1.352,   1.972,   2.697,   4.161,   5.439,   6.194,   7.284,   7.889,
   8.422,   8.918,   8.804,   8.581,   8.676,   8.335,   7.876,   7.397,   7.611,  14.262,
  10.866,   6.970,   1.632,  -0.706,  -0.073,  -0.577,  -2.581,  -5.020,  -7.195,  -8.789,
 -10.259, -11.113, -12.973, -14.712, -16.452, -17.866, -19.600, -20.960, -22.596, -23.948,
 -25.481, -27.250, -28.843, -30.860, -32.395, -33.774, -35.201, -36.489, -37.927, -39.466,
 -41.064, -42.064, -43.346, -44.539, -45.486, -46.223, -46.940, -47.402, -47.538, -47.406,
 -47.147, -46.633, -45.728, -44.719, -43.302, -41.883, -40.099, -38.166, -36.568, -34.610,
 -32.555, -31.033, -29.030, -25.320, -22.394, -18.080, -15.201, -13.383, -12.173, -12.421,
 -14.440, -15.871, -19.342, -23.682, -25.630, -24.493, -24.203, -17.773, -16.267, -15.260,
 -14.823, -13.890, -13.494, -17.772, -20.411, -18.958, -16.867, -17.940, -20.041, -21.350,
 -22.847, -23.553, -28.776, -31.906, -36.796, -37.785, -41.748, -42.849, -48.933, -50.740,
 -50.708, -50.577, -51.049, -51.537, -51.870, -52.095, -51.615, -51.085, -50.228, -49.747,
 -48.756, -46.986, -45.322, -43.859, -41.975, -40.204, -38.237, -35.683, -33.506, -30.634,
 -27.278, -24.024, -20.631, -17.224, -14.359, -12.920, -11.398,  -9.744,  -8.180,  -6.250,
  -4.261,  -2.252,   0.030,   2.719,   5.581,   8.625,  11.419,  13.783,  15.486,  16.715,
  17.120,  17.375,  18.034,  19.068,  20.477,  21.842,  23.137,  23.775,  23.943,  24.037,
  24.860,  27.131,  32.665,  35.952,  35.765,  35.589,  36.827,  36.896,  35.458,  33.103,
  31.854,  30.468,  28.915,  27.294,  26.130,  25.485,  25.237,  27.284,  28.555,  30.432},
{
  29.865,  31.596,  32.046,  33.284,  33.168,  34.425,  36.313,  33.566,  33.258,  31.482,
  29.286,  28.892,  26.539,  25.094,  25.388,  26.087,  26.578,  27.267,  26.639,  24.964,
  21.806,  18.545,  16.084,  14.242,  13.686,  14.487,  14.914,  13.969,  13.221,  12.708,
  11.228,  10.070,   9.700,   9.495,  10.923,  10.530,   7.943,   6.517,   5.555,   5.930,
   6.769,   6.278,   4.580,   2.052,   0.225,  -3.249,  -7.509, -12.522, -19.599, -25.813,
 -29.788, -31.814, -33.241, -33.491, -33.236, -32.524, -32.687, -29.953, -27.549, -29.770,
 -37.645, -39.531, -43.417, -46.487, -45.061, -47.295, -47.217, -47.641, -49.893, -50.987,
 -51.755, -53.305, -54.922, -56.367, -56.508, -56.489, -57.424, -57.544, -59.244, -59.526,
 -57.911, -58.549, -59.873, -59.676, -57.256, -55.479, -56.099, -56.384, -57.152, -56.040,
 -55.710, -54.766, -53.780, -50.594, -50.849, -51.015, -48.460, -44.133, -42.592, -40.911,
 -38.670, -37.002, -34.785, -32.788, -31.217, -30.102, -28.941, -27.746, -25.512, -22.356,
 -18.751, -14.882, -11.631,  -7.715,  -4.052,  -0.191,   3.744,   7.775,  13.076,  16.274,
  18.608,  26.463,  19.619,  16.777,  18.564,  21.516,  26.308,  29.781,  32.117,  34.086,
  35.124,  35.597,  36.728,  37.653,  38.757,  39.523,  39.437,  42.002,  44.336,  46.324,
  46.894,  47.646,  47.436,  40.385,  32.943,  26.550,  31.228,  35.019,  35.983,  35.485,
  33.524,  30.912,  29.256,  31.460,  25.091,  24.408,  22.647,  20.575,  19.070,  15.306,
  14.386,  13.382,  11.841,  10.593,   8.980,   7.906,   6.522,   5.077,   4.070,   3.193,
   2.165,   1.732,   1.288,   1.043,   1.118,  -0.010,   0.532,   0.265,  -0.461,  -0.967,  -0.938,
  -1.053,  -0.436,   0.614,   1.492,   1.946,   3.460,   4.316,   5.236,   6.489,   6.980,
   7.040,   6.557,   6.073,   6.764,   8.382,  10.019,  10.649,  12.384,   7.473,   4.126,
   4.229,   2.983,   2.624,   2.682,   1.913,  -0.188,  -3.185,  -5.963,  -7.960,  -9.367,
 -10.724, -12.132, -13.850, -15.565, -17.069, -18.471, -20.019, -21.075, -22.431, -24.012,
 -25.798, -27.419, -28.927, -30.595, -32.504, -33.895, -35.136, -36.485, -38.305, -39.893,
 -41.036, -42.003, -43.464, -44.722, -45.538, -46.201, -46.898, -47.171, -47.511, -47.312,
 -46.942, -46.199, -45.620, -44.697, -43.139, -41.746, -40.164, -38.650, -36.529, -33.071,
 -32.913, -31.116, -28.493, -25.251, -20.852, -17.688, -15.591, -14.352, -13.203, -13.217,
 -15.582, -17.751, -21.552, -25.783, -28.832, -27.441, -26.470, -26.939, -25.388, -19.682,
 -15.823, -16.673, -17.786, -19.072, -22.528, -22.877, -21.691, -21.570, -23.001, -23.667,
 -25.137, -24.223, -26.836, -30.369, -36.005, -39.063, -44.131, -46.704, -48.922, -50.376,
 -51.026, -51.406, -51.920, -51.783, -52.343, -52.585, -52.626, -52.189, -51.183, -50.355,
 -49.020, -47.494, -45.958, -44.025, -41.681, -39.314, -36.855, -34.246, -31.353, -28.820,
 -25.763, -22.280, -18.862, -15.650, -12.354, -10.803,  -9.356,  -7.756,  -5.781,  -4.167,
  -2.268,  -0.524,   1.643,   3.894,   6.525,   8.870,  11.280,  13.455,  15.080,  16.259,
  16.867,  17.269,  18.295,  19.796,  21.085,  22.365,  23.389,  24.219,  24.912,  25.877,
  26.875,  28.549,  33.299,  35.634,  37.562,  37.154,  38.961,  38.115,  36.254,  34.657,
  32.602,  30.988,  29.442,  27.886,  25.918,  25.230,  25.428,  26.694,  27.388,  29.865},
{
  29.163,  30.563,  31.603,  32.620,  33.658,  35.793,  36.965,  36.582,  34.752,  34.439,
  32.656,  30.358,  28.185,  26.433,  25.774,  25.918,  26.850,  27.281,  26.077,  25.083,
  22.411,  19.548,  17.340,  15.334,  14.486,  14.763,  14.414,  13.690,  13.379,  13.082,
  12.255,  11.684,  11.500,  10.353,  11.289,  11.416,   8.790,   7.500,   7.658,   8.391,
   7.916,   7.785,   6.216,   3.911,   1.592,  -1.827,  -5.438, -10.780, -17.502, -24.150,
 -28.219, -30.803, -32.567, -33.131, -33.170, -32.679, -29.779, -26.614, -29.918, -35.644,
 -37.495, -39.131, -41.852, -44.948, -46.617, -48.091, -45.091, -45.599, -46.474, -47.965,
 -48.264, -49.365, -50.827, -51.727, -52.574, -54.114, -54.535, -54.774, -56.334, -57.613,
 -58.451, -59.045, -58.692, -59.265, -58.115, -56.866, -56.403, -55.490, -56.283, -55.577,
 -55.421, -54.755, -54.879, -52.276, -50.532, -50.659, -48.554, -45.874, -42.816, -40.611,
 -38.851, -37.042, -35.525, -33.170, -31.687, -31.709, -31.280, -29.273, -26.711, -23.742,
 -20.179, -16.648, -13.258,  -9.984,  -6.069,  -1.912,   2.145,   6.108,   9.712,  14.150,
  17.438,  24.261,  15.414,  17.619,  21.432,  23.365,  27.221,  25.055,  28.926,  32.413,
  35.144,  36.256,  37.200,  37.244,  36.824,  37.911,  39.137,  41.722,  43.715,  45.447,
  46.357,  47.685,  44.197,  34.846,  28.743,  36.126,  36.730,  35.440,  33.908,  31.894,
  32.656,  29.701,  27.976,  24.881,  24.308,  21.467,  19.893,  19.196,  16.134,  15.933,
  12.652,  12.250,  10.729,   8.880,   6.903,   4.888,   3.868,   2.834,   2.493,   1.167,
   0.612,  -0.185,  -0.577,  -1.047,  -1.292,  -1.388,  -1.827,  -1.893,  -2.659,  -2.487,  -2.130,
  -1.554,  -0.509,  -0.108,   0.444,   1.504,   1.778,   2.856,   3.522,   4.110,   4.451,
   5.560,   7.289,  12.246,  14.056,   9.571,   6.925,   4.725,   4.865,   4.462,   3.643,
   3.695,   3.761,   3.157,   2.397,   0.876,  -1.675,  -4.510,  -7.092,  -8.762, -10.080,
 -11.397, -12.914, -14.188, -15.614, -16.947, -18.593, -19.908, -21.068, -22.500, -24.007,
 -25.804, -27.573, -29.244, -30.863, -32.571, -34.462, -35.894, -36.854, -38.584, -39.994,
 -41.341, -42.477, -43.479, -44.462, -45.458, -46.390, -46.946, -47.205, -47.140, -47.124,
 -46.837, -45.956, -45.061, -44.336, -42.729, -41.608, -40.513, -36.864, -35.354, -32.610,
 -33.243, -30.387, -27.386, -22.359, -19.306, -18.293, -16.768, -15.427, -14.209, -13.160,
 -17.905, -20.374, -23.399, -28.108, -31.337, -30.608, -28.209, -28.171, -28.536, -27.855,
 -24.407, -20.883, -22.682, -24.328, -26.143, -25.362, -24.716, -25.644, -26.459, -26.369,
 -26.916, -27.047, -31.758, -34.474, -39.141, -43.020, -45.134, -47.559, -49.726, -51.199,
 -52.034, -52.340, -51.809, -51.497, -51.908, -52.190, -52.651, -52.322, -51.476, -50.164,
 -48.833, -47.389, -45.797, -43.866, -41.460, -38.631, -36.330, -33.720, -30.697, -27.700,
 -24.512, -20.846, -17.198, -13.490, -11.376,  -8.570,  -6.667,  -4.935,  -3.360,  -1.506,
   0.147,   1.672,   3.341,   5.320,   7.291,   9.438,  11.472,  13.608,  15.774,  17.135,
  17.992,  18.703,  19.505,  20.471,  21.727,  22.980,  23.952,  25.095,  26.525,  28.108,
  30.075,  31.167,  33.373,  34.747,  37.243,  39.815,  39.193,  38.410,  38.007,  36.367,
  33.976,  32.595,  30.554,  30.195,  28.534,  27.249,  26.297,  26.311,  27.796,  29.163},
{
  27.348,  28.159,  28.279,  30.289,  33.426,  35.128,  35.964,  36.022,  35.717,  34.874,
  32.577,  30.499,  28.569,  26.872,  25.632,  25.256,  27.011,  28.047,  26.481,  24.702,
  22.090,  20.291,  18.853,  17.256,  15.552,  14.870,  13.629,  12.950,  13.166,  13.127,
  12.857,  12.576,  12.528,  11.712,  13.007,  11.606,   9.821,   9.396,  10.236,  10.170,
   9.510,   9.272,   7.760,   5.274,   1.970,  -1.520,  -4.993,  -9.579, -15.755, -22.180,
 -26.843, -29.863, -31.067, -32.310, -32.932, -33.122, -28.215, -26.930, -28.699, -30.183,
 -31.810, -32.266, -33.813, -35.657, -38.392, -41.995, -43.167, -43.307, -44.504, -46.268,
 -45.884, -46.248, -47.997, -48.591, -48.107, -51.821, -53.563, -54.341, -56.079, -58.643,
 -59.528, -59.982, -60.398, -61.706, -62.252, -61.513, -60.552, -58.801, -56.245, -55.034,
 -53.810, -53.327, -52.344, -51.022, -49.795, -50.540, -48.838, -45.496, -41.328, -40.017,
 -38.927, -36.807, -34.720, -32.686, -30.965, -31.800, -31.595, -29.523, -27.152, -24.505,
 -21.409, -17.422, -13.976, -10.748,  -7.472,  -3.614,   0.524,   5.016,   8.932,  12.350,
  14.959,  17.954,  19.787,  21.980,  25.200,  28.146,  26.365,  25.108,  24.479,  28.762,
  33.203,  36.269,  34.286,  35.683,  36.792,  36.865,  39.113,  40.976,  43.065,  44.327,
  45.608,  46.811,  41.764,  32.709,  37.631,  38.298,  35.254,  34.104,  32.339,  30.936,
  28.007,  27.108,  25.891,  24.480,  21.127,  19.111,  16.735,  15.106,  13.535,  12.511,
  10.128,   8.931,   8.628,   7.050,   4.854,   2.838,   2.096,   1.081,   0.293,  -0.613,
  -1.452,  -2.189,  -2.837,  -3.296,  -3.446,  -3.962,  -3.959,  -4.186,  -3.856,  -3.343,  -2.742,
  -2.078,  -1.107,  -0.930,  -0.571,  -0.014,   0.807,   0.489,   4.631,   5.554,   7.464,
   6.138,  14.278,   7.444,   3.079,   3.933,   3.983,   3.791,   3.786,   3.516,   3.271,
   2.769,   2.258,   1.286,  -0.141,  -1.914,  -4.304,  -6.529,  -8.264,  -9.705, -11.023,
 -12.382, -13.667, -14.967, -16.068, -17.375, -18.593, -19.786, -21.161, -22.902, -24.668,
 -26.093, -28.068, -29.757, -31.403, -33.071, -34.386, -35.956, -37.520, -38.912, -40.569,
 -41.586, -42.662, -43.791, -44.819, -45.544, -46.322, -46.743, -47.112, -46.826, -46.555,
 -46.160, -45.436, -44.586, -43.134, -42.576, -41.265, -39.311, -36.638, -34.142, -35.104,
 -32.632, -29.789, -26.463, -21.864, -20.632, -19.664, -18.539, -16.343, -14.750, -16.250,
 -19.374, -22.659, -22.297, -27.310, -32.330, -32.107, -30.364, -28.232, -27.377, -27.006,
 -28.003, -27.685, -25.695, -24.481, -27.109, -23.348, -23.219, -23.572, -24.480, -28.308,
 -28.416, -31.537, -32.988, -38.978, -43.279, -44.446, -46.734, -48.790, -50.842, -52.148,
 -52.419, -51.653, -50.944, -50.637, -51.108, -51.316, -51.871, -51.793, -50.726, -49.447,
 -48.238, -47.079, -45.147, -43.113, -40.880, -37.986, -35.496, -32.627, -29.254, -26.168,
 -22.480, -18.932, -15.014, -11.367,  -8.325,  -6.513,  -4.127,  -2.132,  -0.470,   1.126,
   3.027,   4.687,   6.194,   7.802,   9.325,  11.115,  12.833,  14.587,  16.805,  18.479,
  19.339,  20.544,  21.126,  21.372,  21.990,  23.348,  24.736,  26.172,  28.063,  30.526,
  32.979,  32.962,  31.238,  33.608,  33.970,  37.128,  39.455,  39.508,  38.726,  37.939,
  35.906,  33.805,  31.560,  30.530,  30.376,  29.373,  28.160,  27.385,  27.299,  27.348},
{
  25.617,  25.538,  26.031,  27.742,  30.368,  32.418,  33.889,  33.149,  33.239,  31.979,
  31.143,  29.130,  28.266,  26.785,  26.257,  26.239,  28.153,  29.539,  28.154,  26.094,
  23.884,  21.972,  21.164,  19.169,  16.913,  15.404,  13.805,  12.938,  12.697,  12.463,
  13.077,  12.683,  12.594,  13.301,  13.941,  11.576,  11.368,  12.641,  12.170,  11.558,
  11.251,  10.150,   7.898,   4.877,   1.506,  -1.297,  -4.729,  -9.926, -15.298, -20.990,
 -26.188, -28.480, -30.199, -31.243, -31.976, -33.318, -31.034, -26.165, -22.012, -23.677,
 -25.641, -27.786, -29.902, -30.625, -31.420, -34.822, -40.145, -42.045, -43.799, -45.315,
 -45.309, -44.602, -45.578, -47.551, -47.504, -48.953, -53.479, -54.903, -56.740, -59.354,
 -61.135, -62.389, -63.860, -65.692, -66.131, -65.285, -64.377, -61.812, -57.767, -53.348,
 -50.152, -48.734, -48.416, -52.523, -54.711, -52.108, -49.250, -45.554, -42.043, -39.217,
 -37.869, -36.823, -35.594, -32.704, -30.882, -31.042, -30.438, -28.376, -26.480, -23.910,
 -20.716, -18.524, -16.081, -12.855,  -8.657,  -5.637,  -1.489,   3.069,   7.289,  10.356,
  13.099,  16.376,  19.136,  22.074,  25.744,  27.917,  29.129,  28.907,  28.159,  25.055,
  28.254,  33.368,  36.186,  36.475,  36.313,  37.262,  39.293,  40.333,  42.183,  43.285,
  44.488,  42.552,  38.836,  35.251,  39.188,  38.387,  36.369,  33.829,  32.055,  28.186,
  27.125,  26.002,  24.300,  21.805,  19.027,  16.988,  14.395,  12.612,  11.440,  10.190,
   8.323,   6.743,   6.273,   4.824,   2.745,   1.787,   0.667,  -0.444,  -1.510,  -2.721,
  -3.464,  -4.025,  -4.692,  -4.957,  -5.157,  -5.135,  -5.059,  -4.762,  -4.279,  -4.062,  -3.811,
  -3.160,  -2.639,  -2.336,  -2.315,  -0.666,   6.091,   2.433,   4.801,   5.704,   3.034,
   5.272,   4.606,   3.026,   3.716,   4.162,   3.646,   2.883,   2.673,   2.292,   1.704,
   0.862,  -0.197,  -1.774,  -3.321,  -4.927,  -6.541,  -8.093,  -9.368, -10.726, -12.040,
 -13.123, -14.363, -15.462, -16.542, -17.901, -19.166, -20.376, -21.852, -23.125, -24.796,
 -26.676, -28.251, -29.910, -31.506, -33.079, -34.611, -36.336, -37.796, -39.365, -40.842,
 -41.987, -43.022, -44.127, -45.090, -45.576, -46.014, -46.461, -46.221, -46.329, -46.189,
 -45.674, -45.014, -44.184, -43.222, -41.754, -41.269, -37.186, -36.266, -35.306, -34.535,
 -31.417, -28.713, -24.374, -21.477, -21.049, -20.582, -20.293, -17.936, -18.739, -19.984,
 -23.432, -24.558, -22.696, -27.050, -32.572, -31.118, -29.432, -28.711, -28.561, -27.437,
 -26.206, -25.700, -25.305, -25.631, -27.861, -25.442, -24.893, -24.490, -26.254, -28.446,
 -30.412, -32.881, -39.057, -43.671, -43.199, -45.257, -47.701, -49.634, -51.425, -52.543,
 -52.244, -51.321, -50.222, -49.940, -50.025, -50.060, -50.845, -50.874, -49.790, -48.606,
 -47.359, -45.815, -43.894, -41.843, -39.419, -36.508, -33.596, -30.706, -27.658, -24.327,
 -20.921, -17.083, -13.466,  -9.574,  -6.148,  -3.643,  -1.591,   0.724,   2.325,   3.841,
   5.792,   7.447,   9.295,  10.776,  11.971,  13.331,  15.357,  16.625,  18.522,  20.440,
  21.550,  22.625,  23.578,  24.263,  24.074,  24.585,  25.683,  27.288,  29.296,  32.061,
  34.198,  33.727,  30.811,  31.244,  33.609,  35.979,  38.815,  39.081,  40.152,  39.918,
  37.278,  34.847,  33.029,  31.677,  31.556,  31.322,  30.349,  28.410,  26.763,  25.617},
{
  25.845,  25.691,  24.934,  25.887,  27.193,  29.146,  30.188,  29.679,  29.754,  29.649,
  29.343,  28.704,  28.163,  26.897,  27.149,  28.042,  29.622,  30.860,  29.715,  26.641,
  25.218,  23.260,  21.819,  20.116,  18.028,  15.743,  14.514,  14.474,  14.148,  13.626,
  13.325,  12.814,  12.975,  14.219,  12.667,  12.159,  14.383,  15.406,  13.797,  12.219,
  11.317,   9.645,   7.239,   4.027,   0.529,  -2.264,  -5.785, -10.420, -15.359, -19.967,
 -24.735, -27.083, -28.749, -27.982, -25.591, -26.400, -30.069, -27.843, -19.779, -21.905,
 -23.630, -24.611, -27.200, -29.462, -31.147, -31.304, -35.150, -40.412, -44.913, -43.454,
 -45.257, -45.221, -44.019, -46.081, -47.804, -46.888, -50.292, -53.750, -56.664, -60.367,
 -63.189, -65.501, -67.101, -67.279, -66.691, -65.446, -58.612, -51.090, -47.085, -41.015,
 -41.045, -41.830, -41.576, -47.324, -55.841, -55.732, -51.005, -45.540, -43.708, -39.586,
 -36.248, -35.095, -34.668, -34.453, -32.166, -31.076, -30.926, -29.427, -26.945, -24.286,
 -21.607, -19.476, -17.710, -14.906, -10.829,  -7.149,  -3.774,   0.662,   4.777,   9.379,
  11.711,  14.887,  17.443,  20.629,  24.076,  27.475,  30.014,  31.154,  31.978,  26.041,
  24.248,  28.972,  32.971,  34.852,  35.340,  37.185,  38.598,  40.076,  41.728,  42.522,
  43.658,  41.868,  45.581,  33.341,  33.008,  35.144,  33.784,  32.244,  29.478,  27.698,
  25.232,  23.168,  22.249,  18.146,  16.216,  15.327,  13.285,  11.141,   9.553,   8.235,
   6.872,   4.913,   4.035,   2.531,   1.471,   0.111,  -0.696,  -1.886,  -2.877,  -4.042,
  -4.707,  -5.613,  -6.156,  -6.351,  -6.180,  -5.677,  -5.442,  -4.697,  -3.951,  -3.864,  -4.365,
  -4.521,  -3.588,  -1.111,  -1.196,  -3.475,  -3.368,  -2.762,  -1.468,  -0.605,   1.034,
   1.695,   2.239,   2.862,   2.688,   2.581,   1.650,   1.110,   0.463,   0.278,  -0.608,
  -1.835,  -3.227,  -4.598,  -6.006,  -7.324,  -8.500,  -9.521, -10.657, -11.667, -12.712,
 -13.936, -15.110, -16.142, -17.054, -18.261, -19.611, -21.116, -22.622, -24.363, -25.934,
 -27.548, -28.981, -30.540, -31.938, -33.289, -34.781, -36.390, -37.784, -39.223, -40.637,
 -41.842, -43.118, -43.922, -44.635, -45.334, -45.489, -45.880, -45.907, -45.639, -45.214,
 -44.765, -44.206, -43.359, -42.894, -42.586, -37.749, -37.388, -35.725, -35.627, -33.150,
 -30.578, -27.151, -23.668, -23.130, -22.164, -21.350, -20.262, -20.255, -21.072, -21.983,
 -25.285, -26.483, -25.323, -27.982, -31.050, -30.235, -28.880, -29.177, -28.540, -27.054,
 -26.010, -26.092, -27.023, -29.268, -27.163, -25.039, -25.922, -25.994, -27.155, -29.534,
 -30.186, -33.446, -37.559, -44.466, -43.823, -45.183, -47.837, -49.984, -51.600, -52.099,
 -51.619, -50.518, -49.136, -48.700, -48.765, -48.863, -49.465, -49.377, -48.701, -47.684,
 -46.242, -44.660, -42.784, -40.821, -38.216, -35.289, -32.572, -29.525, -26.335, -22.898,
 -19.331, -15.407, -11.646,  -8.187,  -3.844,  -0.821,   0.875,   3.367,   5.314,   6.802,
   8.409,  10.166,  12.007,  13.716,  15.309,  16.584,  18.269,  20.028,  21.354,  22.784,
  23.367,  24.594,  25.686,  26.009,  26.011,  25.444,  25.873,  27.652,  29.630,  31.485,
  33.704,  34.192,  33.272,  31.773,  31.707,  36.074,  38.258,  39.099,  39.396,  39.872,
  38.499,  38.009,  35.005,  33.336,  33.261,  32.977,  31.933,  29.836,  27.335,  25.845},
{
  28.325,  27.841,  26.527,  27.412,  28.160,  27.687,  28.047,  27.527,  28.955,  28.740,
  28.412,  28.291,  28.573,  28.994,  29.421,  29.953,  30.246,  30.212,  28.533,  26.261,
  24.534,  22.952,  21.544,  21.178,  19.708,  17.209,  15.352,  15.000,  14.514,  15.046,
  14.561,  13.856,  14.083,  13.473,  10.781,  13.322,  17.860,  16.536,  14.024,  11.957,
  10.327,   8.390,   5.705,   2.860,  -0.178,  -3.094,  -6.578, -10.734, -14.782, -18.624,
 -23.139, -25.661, -24.016, -20.640, -18.940, -19.361, -20.790, -19.159, -17.860, -19.725,
 -20.464, -20.992, -24.686, -29.928, -31.677, -32.248, -31.644, -37.913, -43.812, -42.497,
 -44.794, -47.560, -46.960, -46.767, -46.379, -46.520, -48.456, -52.653, -56.283, -59.848,
 -62.844, -62.951, -58.198, -54.532, -52.512, -45.210, -33.135, -28.944, -29.405, -28.662,
 -29.868, -31.230, -30.404, -34.348, -41.369, -48.239, -48.956, -42.191, -39.965, -36.802,
 -33.506, -32.064, -32.953, -34.666, -37.263, -37.037, -34.598, -31.318, -28.445, -25.245,
 -22.510, -20.340, -19.120, -16.404, -12.169,  -8.730,  -5.210,  -1.217,   3.218,   7.445,
  10.040,  13.465,  16.360,  19.604,  22.912,  25.831,  29.283,  30.613,  31.217,  29.278,
  24.898,  25.904,  32.661,  34.839,  35.058,  37.139,  38.145,  39.628,  40.724,  42.356,
  43.431,  40.203,  48.476,  32.239,  30.191,  32.397,  32.985,  30.172,  27.598,  24.483,
  22.181,  20.988,  18.360,  17.478,  15.066,  12.883,  11.049,   9.342,   8.075,   6.214,
   4.894,   3.906,   3.153,   1.442,   0.086,  -0.946,  -2.063,  -2.920,  -4.042,  -4.591,
  -5.833,  -6.633,  -6.987,  -6.810,  -6.973,  -6.353,  -5.580,  -4.779,  -4.816,  -4.971,  -5.170,
  -4.288,  -0.311,  -1.781,   0.612,  -3.566,  -4.704,  -2.896,  -1.821,  -0.747,  -0.464,
   0.366,   0.845,   0.583,   0.220,  -0.223,  -0.926,  -1.526,  -1.964,  -2.505,  -3.404,
  -4.803,  -6.090,  -7.283,  -8.386,  -9.206, -10.136, -10.833, -11.552, -12.557, -13.619,
 -14.414, -15.355, -16.481, -17.725, -19.082, -20.540, -22.158, -23.731, -25.223, -26.834,
 -28.314, -29.644, -30.810, -32.282, -33.599, -34.861, -36.265, -37.587, -39.179, -40.498,
 -41.904, -42.833, -43.629, -44.039, -44.364, -44.827, -45.024, -44.997, -44.823, -44.334,
 -44.030, -43.291, -42.506, -43.366, -37.387, -37.482, -35.111, -35.914, -33.864, -31.789,
 -29.142, -24.948, -23.528, -23.857, -23.314, -21.887, -21.209, -21.017, -22.965, -24.404,
 -26.300, -27.929, -28.144, -27.242, -28.021, -27.152, -25.850, -25.637, -25.329, -25.779,
 -27.360, -30.195, -31.493, -29.546, -26.354, -24.584, -25.798, -27.664, -29.539, -31.274,
 -33.574, -36.365, -38.493, -45.423, -45.416, -45.907, -48.249, -50.339, -51.755, -51.822,
 -51.325, -49.941, -48.462, -47.625, -47.342, -47.165, -47.085, -46.861, -46.402, -45.631,
 -44.315, -42.654, -40.736, -38.730, -36.282, -33.678, -31.137, -28.021, -24.638, -21.004,
 -17.529, -13.845, -10.086,  -6.173,  -2.087,   1.698,   3.990,   5.901,   8.007,   9.742,
  11.434,  13.253,  14.928,  16.766,  18.620,  20.204,  21.489,  23.213,  24.779,  25.803,
  25.558,  25.797,  26.638,  26.818,  26.812,  26.781,  27.156,  28.337,  29.625,  31.414,
  33.567,  38.058,  45.418,  44.216,  40.889,  41.228,  39.493,  39.458,  39.137,  38.996,
  38.945,  37.753,  36.454,  35.968,  36.021,  35.790,  34.081,  31.431,  29.782,  28.325},
{
  30.965,  28.600,  27.728,  28.285,  27.771,  26.968,  27.069,  27.080,  27.320,  27.964,
  28.505,  28.480,  29.077,  30.354,  30.674,  31.016,  30.410,  29.249,  28.517,  26.146,
  24.702,  23.202,  22.143,  22.784,  21.249,  19.677,  17.707,  15.835,  15.292,  16.219,
  15.773,  15.124,  16.041,  13.579,  13.477,  14.796,  19.593,  16.975,  14.142,  11.590,
   9.510,   7.513,   5.089,   2.486,  -0.181,  -3.301,  -6.915, -10.742, -14.510, -18.237,
 -20.572, -21.075, -14.869, -12.366, -13.467, -11.135,  -7.978,  -7.234, -11.159, -16.904,
 -17.279, -17.226, -22.419, -25.073, -26.919, -29.554, -29.549, -34.236, -42.389, -38.997,
 -44.844, -50.233, -49.338, -48.075, -47.080, -46.812, -48.775, -51.328, -54.544, -57.470,
 -55.211, -47.955, -39.585, -33.954, -30.808, -29.278, -29.585, -30.098, -31.128, -33.225,
 -33.586, -32.946, -32.080, -31.179, -30.052, -39.941, -43.934, -40.303, -36.733, -34.035,
 -31.976, -30.762, -32.393, -37.849, -42.559, -41.823, -38.931, -34.321, -30.438, -27.431,
 -23.731, -21.369, -20.434, -17.689, -13.229,  -9.258,  -6.137,  -2.310,   1.871,   5.641,
   8.408,  11.505,  14.985,  18.731,  21.599,  24.097,  27.057,  29.244,  30.280,  30.209,
  26.293,  24.359,  28.726,  33.080,  35.787,  36.130,  37.703,  39.118,  39.997,  42.101,
  42.589,  36.436,  36.724,  23.492,  30.224,  31.585,  30.191,  27.900,  24.873,  22.705,
  21.304,  19.571,  17.353,  15.607,  13.008,  11.180,   9.716,   8.517,   7.217,   4.582,
   3.896,   3.450,   2.258,   0.561,  -1.058,  -2.591,  -3.436,  -4.578,  -5.352,  -5.973,
  -6.392,  -6.727,  -6.736,  -6.663,  -6.546,  -5.983,  -5.762,  -5.558,  -5.672,  -6.394,  -5.251,
  -2.969,  -5.086,  -5.795,  -5.372,  -6.017,  -4.117,  -4.252,  -3.202,  -2.764,  -2.532,
  -2.379,  -2.383,  -2.528,  -2.802,  -3.281,  -3.896,  -4.470,  -4.362,  -5.553,  -6.560,
  -7.513,  -8.644,  -9.577, -10.472, -11.233, -11.737, -12.226, -12.637, -13.418, -14.675,
 -15.661, -16.588, -17.467, -18.544, -20.015, -21.633, -23.091, -24.734, -26.023, -27.295,
 -28.739, -30.045, -31.161, -32.513, -33.772, -35.196, -36.295, -37.806, -39.216, -40.518,
 -41.789, -42.590, -42.984, -43.876, -44.028, -44.140, -44.228, -44.206, -44.011, -43.151,
 -42.662, -41.933, -41.788, -40.697, -37.795, -35.285, -36.666, -34.543, -32.755, -31.051,
 -28.589, -24.921, -23.793, -24.518, -23.777, -22.595, -22.299, -21.789, -23.639, -23.903,
 -25.430, -27.338, -28.885, -28.275, -27.429, -26.588, -25.863, -25.353, -25.099, -24.535,
 -25.215, -28.908, -28.955, -27.368, -27.559, -27.915, -28.530, -29.201, -29.652, -32.042,
 -34.896, -37.026, -41.224, -46.324, -45.885, -46.761, -48.193, -50.711, -51.982, -51.881,
 -51.175, -49.485, -47.996, -46.960, -46.078, -45.280, -44.666, -44.355, -43.955, -42.905,
 -41.732, -40.208, -38.351, -36.561, -34.362, -32.003, -29.093, -26.299, -22.920, -19.342,
 -15.720, -12.010,  -8.078,  -4.177,  -0.493,   3.637,   6.859,   9.017,  11.125,  12.997,
  14.855,  16.392,  18.269,  20.100,  21.771,  23.354,  24.799,  26.258,  27.382,  28.215,
  28.123,  27.494,  27.954,  28.609,  28.522,  28.649,  29.227,  30.444,  31.596,  32.408,
  33.706,  39.078,  37.064,  36.692,  36.019,  42.236,  42.125,  40.986,  41.404,  42.223,
  42.086,  39.907,  40.048,  40.253,  40.426,  40.126,  37.812,  36.354,  33.521,  30.965},
{
  35.389,  32.533,  30.134,  29.022,  27.461,  26.463,  26.381,  26.407,  26.524,  26.717,
  27.749,  29.004,  30.290,  31.289,  31.063,  29.809,  29.526,  28.869,  28.358,  26.941,
  25.833,  24.049,  22.826,  22.319,  22.278,  21.221,  20.579,  18.165,  16.834,  16.653,
  16.214,  15.500,  16.305,  16.883,  16.521,  17.320,  19.290,  16.960,  14.395,  11.824,
   9.797,   7.119,   4.860,   2.491,  -0.320,  -4.183,  -8.492, -12.531, -15.071, -17.356,
 -17.681, -13.200,  -6.978,  -8.173,  -7.788,  -6.458,  -5.862,  -7.028, -13.407, -15.810,
 -15.055, -18.535, -21.522, -22.579, -23.818, -26.990, -27.914, -28.073, -33.607, -33.051,
 -38.805, -49.334, -49.078, -47.274, -45.822, -46.482, -48.413, -50.196, -48.350, -39.381,
 -34.529, -29.347, -25.351, -25.539, -27.358, -27.751, -29.748, -29.613, -30.672, -32.911,
 -34.079, -34.914, -34.820, -34.209, -35.086, -38.469, -38.006, -36.075, -35.142, -33.885,
 -31.653, -31.127, -32.222, -41.015, -43.693, -42.342, -40.583, -37.827, -32.462, -27.322,
 -23.688, -22.146, -22.107, -19.020, -14.930, -10.758,  -7.067,  -3.048,   1.158,   4.206,
   7.023,  10.356,  14.063,  17.497,  20.599,  23.207,  25.363,  28.390,  29.680,  30.562,
  30.395,  25.716,  26.786,  31.990,  34.399,  36.017,  37.491,  38.564,  39.470,  42.094,
  42.338,  34.539,  25.043,  22.705,  29.865,  29.962,  28.272,  25.778,  23.454,  21.181,
  19.332,  17.296,  15.546,  13.995,  12.329,  10.753,   8.923,   7.535,   5.975,   4.257,
   3.289,   1.641,  -0.194,  -1.407,  -2.523,  -3.688,  -5.182,  -6.222,  -6.449,  -6.948,
  -6.843,  -6.951,  -7.074,  -7.168,  -6.389,  -6.857,  -7.040,  -6.917,  -6.973,  -4.933,  -7.504,
  -8.148,  -8.315,  -7.771,  -7.250,  -7.500,  -6.412,  -6.323,  -5.673,  -5.348,  -5.337,
  -5.507,  -5.767,  -5.534,  -5.694,  -5.973,  -6.464,  -6.811,  -6.938,  -7.932,  -8.706,
  -9.586, -10.566, -11.371, -12.115, -12.456, -12.706, -13.306, -13.954, -14.846, -15.841,
 -16.674, -17.432, -18.459, -19.765, -21.460, -22.878, -24.134, -25.449, -26.833, -28.113,
 -29.432, -30.560, -31.793, -32.960, -34.312, -35.520, -36.717, -37.897, -39.275, -40.465,
 -41.341, -42.081, -42.427, -42.895, -43.416, -43.462, -43.240, -43.296, -42.906, -42.404,
 -41.901, -41.034, -40.227, -38.398, -34.618, -36.520, -35.438, -32.916, -31.498, -29.825,
 -27.914, -25.061, -25.147, -24.540, -24.008, -22.479, -22.693, -23.968, -23.947, -23.141,
 -23.544, -26.719, -27.726, -28.729, -28.543, -28.501, -28.025, -27.818, -27.507, -27.256,
 -27.842, -27.934, -27.782, -28.102, -28.572, -28.637, -29.114, -29.424, -30.124, -32.390,
 -35.801, -38.804, -41.382, -46.466, -45.821, -46.822, -49.453, -51.485, -52.131, -51.679,
 -50.664, -48.738, -46.792, -45.704, -44.846, -43.489, -42.338, -41.707, -41.213, -40.795,
 -39.582, -38.043, -36.460, -34.719, -32.558, -30.126, -27.592, -24.581, -21.042, -17.633,
 -13.944, -10.187,  -6.433,  -2.524,   1.281,   5.132,   8.848,  12.025,  14.226,  16.299,
  18.289,  20.196,  21.882,  23.403,  24.831,  26.240,  27.620,  28.932,  29.877,  30.194,
  32.966,  32.657,  30.350,  30.592,  30.096,  29.816,  30.501,  31.952,  33.359,  34.150,
  34.798,  35.500,  36.659,  38.927,  39.500,  40.616,  44.465,  42.043,  39.657,  43.405,
  47.290,  49.234,  47.999,  47.269,  46.259,  43.658,  41.367,  38.972,  37.537,  35.389},
{
  39.102,  36.899,  34.456,  32.069,  29.163,  27.412,  26.753,  26.617,  27.292,  28.181,
  29.019,  31.021,  31.888,  31.472,  30.891,  29.556,  28.521,  28.431,  28.132,  26.870,
  26.931,  26.192,  24.197,  22.740,  21.741,  21.498,  20.898,  19.342,  16.598,  14.804,
  15.258,  15.533,  16.326,  17.302,  17.056,  17.542,  20.190,  17.657,  15.689,  14.214,
  12.351,   8.806,   5.287,   2.179,  -1.309,  -5.786, -10.325, -14.250, -16.147, -15.800,
 -11.346,  -5.293,  -3.041,  -3.668,  -4.698,  -5.040,  -6.561,  -8.552, -14.314, -13.880,
 -14.771, -19.094, -21.080, -22.276, -23.435, -24.546, -25.586, -24.571, -26.396, -30.810,
 -37.340, -48.384, -46.857, -44.757, -43.813, -46.259, -48.005, -43.521, -35.948, -24.372,
 -22.274, -24.246, -25.863, -28.396, -29.115, -30.055, -30.749, -31.252, -32.575, -34.387,
 -35.397, -35.605, -36.338, -36.569, -36.197, -37.136, -38.083, -37.353, -35.716, -33.985,
 -32.061, -31.241, -30.976, -32.766, -40.560, -42.709, -41.894, -39.249, -35.125, -29.411,
 -24.386, -22.513, -22.382, -19.524, -15.776, -12.119,  -7.792,  -4.381,  -0.654,   2.799,
   5.760,   9.442,  13.253,  16.112,  19.035,  21.600,  24.271,  27.216,  29.285,  30.499,
  31.243,  29.264,  26.630,  29.804,  33.208,  35.644,  37.208,  37.920,  39.227,  42.374,
  42.313,  35.484,  19.519,  24.353,  28.874,  28.313,  26.356,  24.448,  21.845,  19.504,
  17.630,  16.162,  14.256,  12.976,  11.083,  10.036,   8.763,   7.469,   5.594,   4.091,
   2.775,   0.527,  -1.359,  -2.854,  -3.886,  -5.058,  -6.681,  -7.481,  -7.658,  -7.693,
  -7.740,  -8.171,  -8.566,  -8.464,  -8.224,  -8.337,  -6.452,  -8.144,  -8.411,  -8.501,  -8.125,
  -8.576,  -8.414,  -8.213,  -8.131,  -8.304,  -8.228,  -8.015,  -7.825,  -7.858,  -8.078,
  -8.036,  -8.121,  -8.190,  -8.060,  -8.320,  -8.256,  -8.498,  -8.556,  -9.469, -10.315,
 -11.165, -11.992, -12.749, -13.173, -13.359, -13.450, -13.918, -14.648, -15.446, -16.480,
 -18.089, -19.614, -21.174, -22.894, -23.930, -24.816, -25.625, -26.603, -27.721, -28.753,
 -29.957, -31.010, -32.264, -33.336, -34.528, -35.872, -36.994, -38.423, -39.314, -40.136,
 -40.885, -41.307, -41.678, -42.179, -42.268, -42.352, -42.211, -42.244, -41.795, -41.472,
 -40.885, -39.550, -38.494, -34.144, -34.869, -35.721, -34.382, -32.218, -30.344, -28.380,
 -27.405, -25.780, -25.529, -24.484, -23.908, -23.034, -24.397, -24.501, -24.482, -24.099,
 -24.625, -27.181, -27.298, -27.709, -28.275, -28.822, -27.520, -28.003, -28.031, -28.492,
 -29.256, -29.481, -28.545, -28.284, -28.849, -28.609, -28.857, -29.648, -31.017, -33.632,
 -35.548, -39.579, -42.693, -43.594, -44.410, -48.331, -50.963, -51.653, -51.554, -51.071,
 -49.681, -47.719, -45.960, -44.824, -43.287, -41.733, -40.468, -39.558, -38.930, -38.646,
 -37.660, -36.183, -34.880, -33.046, -30.700, -28.267, -25.804, -22.825, -19.450, -15.731,
 -12.196,  -8.432,  -4.579,  -0.843,   2.957,   6.747,  10.561,  14.476,  17.551,  19.552,
  21.613,  23.534,  25.062,  26.747,  28.041,  29.210,  30.320,  31.537,  32.474,  32.419,
  33.784,  32.237,  32.216,  31.672,  30.860,  30.377,  31.642,  33.351,  34.636,  35.678,
  37.072,  37.246,  37.905,  38.674,  41.097,  44.012,  44.226,  44.135,  42.935,  46.000,
  51.050,  53.102,  53.869,  52.336,  49.718,  47.860,  45.506,  43.343,  41.763,  39.102},
{
  41.083,  38.970,  38.013,  35.868,  32.442,  29.614,  28.247,  27.890,  28.717,  30.031,
  32.060,  32.540,  32.318,  33.021,  31.772,  30.270,  28.693,  27.188,  26.711,  26.957,
  29.200,  30.234,  27.446,  24.742,  21.601,  18.231,  13.858,  11.348,  10.659,  12.453,
  16.110,  18.877,  17.496,  16.770,  16.446,  20.302,  20.915,  19.778,  18.837,  17.641,
  14.922,  11.351,   6.831,   2.373,  -2.168,  -6.555,  -9.907, -12.139, -14.092, -12.388,
  -3.974,  -1.135,  -1.414,  -1.699,  -2.778,  -5.111,  -6.482, -10.044, -12.075, -12.219,
 -13.916, -17.674, -20.511, -22.447, -23.572, -24.324, -24.547, -25.167, -26.242, -31.090,
 -39.525, -44.747, -43.488, -43.307, -44.405, -46.328, -45.595, -35.361, -28.191, -24.592,
 -24.422, -26.369, -28.419, -28.678, -30.117, -30.807, -31.594, -33.294, -34.970, -37.120,
 -37.009, -36.794, -37.196, -37.888, -38.237, -38.640, -39.136, -38.622, -36.590, -34.462,
 -33.683, -32.469, -32.114, -31.435, -34.876, -40.917, -40.755, -37.951, -33.932, -28.623,
 -26.251, -23.843, -23.509, -20.720, -17.101, -13.170,  -9.240,  -5.602,  -1.791,   1.690,
   5.071,   8.968,  11.940,  15.173,  17.973,  20.825,  23.724,  26.314,  28.998,  30.116,
  31.804,  31.243,  25.814,  31.368,  32.508,  33.159,  35.537,  37.244,  38.618,  41.856,
  41.582,  33.148,  16.345,  24.602,  28.507,  28.187,  26.036,  24.022,  21.732,  19.775,
  16.977,  15.796,  13.163,  12.113,  10.854,   9.101,   7.561,   6.785,   6.042,   3.908,
   1.337,  -0.585,  -2.533,  -4.313,  -5.631,  -6.646,  -7.857,  -8.323,  -8.646,  -8.427,
  -8.419,  -9.471,  -8.429,  -5.157,  -9.329,  -8.964,  -9.214,  -9.093,  -9.083,  -8.935,  -9.128,
  -9.521,  -9.031,  -9.055,  -9.641,  -9.635,  -9.160,  -9.610,  -9.610,  -9.777,  -9.898,
 -10.020, -10.111, -10.013,  -9.946,  -9.728,  -9.470,  -8.932,  -9.935, -10.672, -11.323,
 -12.390, -13.260, -13.953, -14.127, -14.095, -14.102, -14.591, -15.640, -16.761, -18.105,
 -19.801, -21.105, -22.522, -23.841, -24.929, -26.026, -27.157, -28.007, -28.868, -29.750,
 -30.738, -31.666, -33.070, -34.210, -35.536, -36.555, -37.570, -38.507, -39.606, -40.045,
 -40.459, -40.728, -41.279, -41.628, -41.600, -41.708, -41.856, -41.551, -40.965, -40.781,
 -39.057, -38.296, -36.502, -32.501, -35.331, -34.478, -32.646, -31.158, -29.662, -28.153,
 -26.793, -25.647, -24.928, -24.176, -23.456, -26.447, -24.084, -25.723, -26.156, -26.960,
 -28.051, -28.489, -27.723, -26.727, -27.529, -27.055, -26.897, -26.739, -27.337, -27.164,
 -27.842, -28.696, -28.548, -27.953, -27.206, -27.744, -28.818, -30.738, -32.787, -34.468,
 -36.607, -38.748, -40.914, -43.204, -45.971, -49.992, -51.401, -51.360, -50.766, -49.966,
 -48.404, -46.804, -45.149, -43.412, -36.970, -39.526, -38.564, -37.366, -36.537, -36.520,
 -35.792, -34.993, -33.563, -31.231, -28.955, -26.452, -23.791, -20.876, -17.722, -14.074,
 -10.294,  -6.732,  -3.179,   0.591,   4.435,   8.366,  12.048,  15.778,  19.869,  22.836,
  24.601,  26.713,  28.350,  29.897,  31.437,  32.594,  33.531,  34.520,  35.583,  35.863,
  35.551,  37.948,  33.730,  34.069,  33.357,  32.666,  33.036,  34.150,  35.682,  37.935,
  39.478,  39.865,  40.612,  40.152,  40.479,  41.659,  42.922,  43.123,  43.550,  45.571,
  48.298,  49.358,  50.408,  55.495,  53.836,  51.136,  48.180,  47.114,  44.136,  41.083},
{
  45.020,  41.651,  40.139,  39.096,  35.690,  32.045,  29.939,  28.934,  29.407,  30.806,
  32.830,  33.083,  33.102,  32.466,  31.357,  29.779,  28.794,  27.655,  27.158,  27.023,
  27.904,  27.817,  28.706,  22.974,  14.446,   9.542,   7.480,   6.977,   6.757,   8.888,
  11.527,  14.373,  15.483,  15.698,  15.961,  20.325,  23.566,  23.767,  20.772,  19.007,
  16.578,  12.856,   8.310,   3.323,  -1.306,  -4.387,  -5.385,  -3.315,  -3.308,  -0.104,
   4.108,   3.136,   1.378,  -0.601,  -3.641,  -5.431,  -7.875, -10.346, -11.640, -12.404,
 -14.135, -19.341, -20.548, -20.944, -20.618, -22.565, -23.979, -22.814, -25.638, -28.802,
 -36.855, -45.377, -47.382, -47.395, -45.340, -40.425, -33.652, -26.721, -22.914, -24.100,
 -25.472, -26.804, -28.987, -30.403, -31.287, -32.015, -33.042, -34.369, -35.634, -37.045,
 -37.603, -37.614, -37.903, -38.754, -39.494, -40.004, -40.309, -39.733, -37.431, -35.768,
 -34.535, -33.852, -34.049, -32.769, -34.149, -38.067, -38.524, -35.979, -33.981, -31.388,
 -28.410, -25.906, -23.373, -20.737, -17.938, -14.870, -11.015,  -6.483,  -2.429,   1.124,
   4.587,   8.004,  11.201,  14.828,  17.676,  20.984,  23.965,  26.118,  28.235,  30.708,
  32.165,  32.673,  30.691,  34.322,  34.908,  33.537,  34.273,  35.670,  36.997,  40.636,
  43.158,  33.620,  14.570,  24.106,  27.625,  27.310,  25.410,  22.985,  20.817,  17.959,
  16.208,  14.445,  12.752,  10.497,   8.889,   7.141,   5.713,   5.133,   4.858,   2.947,
   0.332,  -1.672,  -4.125,  -5.996,  -7.312,  -8.173,  -8.795,  -8.972,  -8.990,  -9.365,
 -10.128, -10.476,  -6.580,  -9.489,  -9.368,  -8.608,  -9.157,  -9.276,  -9.158,  -9.056,  -9.421,
  -9.678, -10.187, -10.405, -10.413,  -9.984, -10.681, -10.991, -11.046, -11.160, -11.280,
 -11.366, -11.273, -11.182, -11.025, -10.927, -10.891, -11.209, -11.871, -12.526, -12.959,
 -13.502, -14.239, -14.911, -15.119, -15.138, -15.342, -16.213, -17.156, -18.174, -19.730,
 -21.129, -22.519, -23.653, -24.738, -25.769, -26.994, -27.855, -28.531, -29.469, -30.458,
 -31.595, -32.659, -33.661, -34.931, -36.057, -37.215, -38.169, -38.840, -39.434, -40.194,
 -40.767, -41.097, -41.274, -41.427, -41.668, -41.443, -40.955, -40.710, -40.366, -38.420,
 -37.702, -37.119, -33.181, -34.244, -34.136, -32.773, -31.894, -30.714, -29.135, -27.515,
 -26.000, -23.853, -23.945, -22.316, -22.441, -23.523, -22.494, -24.935, -27.310, -30.097,
 -30.402, -29.778, -27.674, -26.530, -27.744, -27.648, -27.349, -27.128, -27.414, -27.904,
 -29.043, -30.020, -30.106, -30.125, -30.063, -28.620, -29.284, -31.255, -32.923, -34.311,
 -36.184, -39.374, -39.885, -44.186, -47.554, -49.582, -50.710, -50.552, -49.768, -48.581,
 -47.319, -45.872, -44.076, -42.202, -39.560, -38.323, -37.549, -36.328, -35.175, -34.764,
 -33.823, -31.904, -30.907, -29.284, -27.211, -24.709, -21.602, -18.355, -15.509, -12.089,
  -8.541,  -4.720,  -1.095,   2.565,   6.479,  10.168,  13.671,  17.139,  20.921,  25.129,
  28.048,  29.889,  31.653,  33.020,  34.447,  35.963,  37.091,  37.978,  38.520,  39.354,
  39.539,  38.406,  37.853,  38.103,  37.217,  35.349,  34.742,  35.162,  36.270,  39.592,
  43.128,  45.191,  48.065,  45.124,  42.653,  41.054,  42.261,  42.467,  40.777,  39.407,
  45.962,  48.280,  50.596,  52.161,  54.610,  52.410,  51.696,  49.614,  48.303,  45.020},
{
  48.293,  46.206,  44.998,  42.115,  38.945,  34.360,  30.694,  29.298,  29.607,  31.529,
  33.460,  33.722,  33.438,  32.984,  32.230,  31.795,  30.528,  29.100,  27.733,  26.096,
  25.473,  22.736,  18.511,  13.159,   9.019,   9.130,   5.654,   1.286,   1.395,   5.161,
   7.537,   9.980,  12.808,  14.292,  15.355,  17.920,  27.636,  25.538,  22.277,  19.504,
  17.642,  14.439,  10.380,   5.698,   1.331,   0.251,   4.182,   4.943,   5.392,   6.095,
   5.902,   3.915,  -0.558,  -2.591,  -5.479,  -7.474,  -9.706, -11.080, -12.358, -13.500,
 -17.196, -20.375, -21.111, -19.763, -19.935, -20.648, -22.493, -22.108, -23.493, -27.769,
 -34.988, -41.624, -47.769, -44.306, -37.554, -32.337, -24.886, -24.580, -22.814, -23.665,
 -25.153, -26.402, -28.447, -30.333, -31.880, -33.522, -34.381, -35.753, -36.951, -37.959,
 -38.661, -39.473, -40.421, -40.838, -40.796, -40.849, -41.096, -40.007, -38.524, -37.724,
 -36.364, -35.663, -35.706, -35.664, -36.421, -37.978, -38.168, -36.315, -34.954, -33.186,
 -28.295, -25.074, -22.897, -20.682, -18.601, -15.363, -11.316,  -6.873,  -2.527,   1.133,
   4.690,   7.928,  11.002,  14.462,  17.916,  20.994,  23.575,  25.895,  28.060,  29.661,
  31.069,  32.313,  31.903,  34.017,  36.959,  36.515,  39.600,  36.136,  37.136,  39.672,
  39.302,  24.859,  10.793,  22.765,  26.909,  25.496,  24.005,  21.476,  19.496,  17.018,
  15.809,  13.714,  11.120,   8.780,   7.339,   5.401,   4.199,   3.661,   2.870,   1.699,
  -0.140,  -2.072,  -4.252,  -6.478,  -7.896,  -8.498,  -9.104,  -9.560,  -9.481,  -9.785,
 -11.753, -11.711,  -9.027, -11.313,  -9.032,  -8.218,  -8.806,  -8.928,  -8.510,  -8.158,  -9.612,
 -10.343,  -9.387, -10.498,  -9.993, -11.248, -11.920, -12.137, -12.254, -12.478, -12.671,
 -12.507, -12.235, -11.845, -11.999, -12.115, -12.482, -12.867, -13.244, -13.693, -14.235,
 -14.755, -15.020, -15.533, -15.899, -16.114, -16.706, -17.758, -18.936, -19.927, -20.997,
 -22.148, -23.107, -23.970, -25.102, -26.216, -27.405, -28.643, -29.488, -30.614, -31.555,
 -32.477, -33.596, -34.802, -35.636, -36.888, -38.152, -39.055, -39.797, -40.460, -40.629,
 -40.975, -41.160, -41.249, -41.432, -41.475, -41.263, -40.632, -39.712, -39.053, -37.231,
 -37.176, -34.158, -31.850, -32.429, -32.594, -31.351, -30.193, -28.836, -26.387, -24.322,
 -23.169, -22.340, -22.867, -21.062, -22.942, -23.945, -24.025, -26.177, -28.169, -29.451,
 -29.624, -27.266, -26.717, -28.649, -29.585, -28.527, -27.786, -27.046, -27.668, -27.286,
 -28.320, -28.266, -28.961, -30.067, -30.489, -30.955, -30.516, -31.056, -32.450, -33.553,
 -35.575, -38.881, -39.885, -41.492, -46.960, -47.775, -48.357, -48.581, -48.050, -47.192,
 -46.091, -44.701, -43.486, -41.678, -39.510, -38.046, -36.677, -35.238, -34.274, -33.319,
 -32.046, -30.090, -29.118, -27.727, -25.209, -22.361, -19.551, -16.199, -13.274,  -9.715,
  -6.086,  -2.708,   0.755,   4.336,   7.850,  11.341,  14.842,  18.430,  22.014,  25.762,
  29.324,  32.453,  35.103,  36.400,  37.790,  39.046,  40.243,  40.677,  41.410,  42.110,
  41.940,  42.015,  41.513,  39.503,  38.287,  38.269,  37.623,  37.528,  38.384,  40.163,
  41.987,  43.092,  44.249,  45.854,  45.228,  43.224,  41.925,  41.245,  40.098,  37.413,
  37.994,  43.148,  45.137,  45.669,  47.548,  51.533,  51.448,  51.025,  50.145,  48.293},
{
  49.580,  48.666,  46.912,  44.076,  42.108,  37.402,  36.085,  37.322,  38.465,  37.213,
  36.839,  36.441,  36.376,  35.842,  34.107,  33.889,  33.128,  29.407,  28.211,  26.413,
  23.041,  17.233,  11.647,   9.244,  16.803,  19.955,  16.864,   9.732,   5.655,   5.671,
   6.814,   6.975,  17.308,  27.906,  24.295,  19.447,  24.190,  25.788,  24.782,  22.167,
  19.715,  16.922,  13.497,   9.633,   6.326,   5.885,   8.703,  11.241,  11.243,   9.295,
   6.475,   1.294,  -1.998,  -4.580,  -7.340,  -9.115, -10.778, -11.933, -14.442, -16.341,
 -16.827, -21.349, -22.110, -22.689, -22.433, -22.129, -22.262, -22.593, -25.435, -28.757,
 -32.439, -35.427, -37.133, -34.622, -31.863, -28.435, -26.622, -22.993, -21.543, -22.266,
 -24.771, -26.351, -29.056, -31.544, -33.409, -34.941, -36.180, -37.262, -38.484, -39.462,
 -40.410, -41.277, -41.811, -42.509, -42.762, -42.296, -42.259, -41.921, -41.005, -39.731,
 -39.783, -39.138, -37.924, -38.362, -38.886, -39.465, -37.533, -36.246, -35.380, -33.502,
 -31.517, -27.118, -22.385, -20.737, -18.991, -15.558, -11.191,  -6.782,  -2.354,   1.554,
   4.387,   7.297,  10.593,  13.685,  17.113,  19.827,  22.238,  25.322,  28.147,  29.167,
  29.942,  31.020,  32.720,  35.077,  36.063,  36.355,  36.730,  37.492,  40.882,  39.367,
  33.735,  22.738,  12.689,  20.461,  23.573,  23.421,  21.937,  20.964,  18.957,  16.658,
  14.782,  12.357,   9.939,   8.229,   6.403,   5.265,   4.093,   2.623,   1.015,  -0.529,
  -1.611,  -2.716,  -4.747,  -6.479,  -8.195,  -8.950,  -9.506,  -9.650,  -9.626, -10.310,
 -11.863,  -8.826,  -4.060, -12.719, -10.101,  -9.917, -10.209,  -9.776,  -9.482,  -9.872, -10.142,
 -10.287, -10.168,  -9.755, -10.410, -12.160, -12.462, -12.770, -12.947, -13.062, -13.334,
 -13.247, -13.244, -13.177, -13.084, -13.120, -13.176, -13.612, -13.923, -14.339, -14.704,
 -15.197, -15.623, -16.083, -16.329, -16.909, -17.640, -18.672, -19.788, -20.714, -21.737,
 -22.517, -23.206, -24.463, -25.504, -26.827, -27.882, -29.137, -30.397, -31.504, -32.526,
 -33.276, -34.280, -35.154, -36.410, -37.707, -38.738, -39.107, -39.869, -40.272, -40.316,
 -40.704, -40.952, -41.102, -41.078, -40.740, -40.140, -39.455, -39.227, -36.262, -33.954,
 -32.509, -31.893, -31.647, -31.427, -30.368, -29.419, -26.831, -24.669, -24.286, -24.452,
 -23.080, -20.846, -22.109, -19.937, -21.938, -23.942, -24.900, -26.757, -27.821, -28.250,
 -27.015, -27.501, -26.993, -30.050, -32.353, -31.634, -29.878, -28.485, -28.099, -28.349,
 -28.619, -28.418, -28.526, -29.890, -31.025, -29.739, -31.460, -33.018, -31.829, -32.229,
 -34.867, -37.226, -38.902, -39.744, -42.778, -45.232, -46.532, -46.201, -45.846, -45.200,
 -44.585, -43.838, -42.181, -40.329, -38.623, -37.488, -36.140, -34.277, -32.535, -31.304,
 -30.433, -29.435, -27.237, -25.545, -23.057, -20.281, -17.518, -14.718, -11.287,  -7.512,
  -4.212,  -0.658,   2.147,   5.926,   9.343,  12.749,  16.231,  19.850,  23.419,  26.727,
  30.407,  33.752,  36.892,  39.180,  40.821,  42.184,  43.071,  43.970,  44.507,  44.351,
  45.341,  45.944,  44.363,  43.168,  42.503,  42.000,  41.249,  40.816,  40.238,  41.028,
  42.091,  43.823,  47.459,  48.104,  46.412,  45.747,  47.748,  47.396,  44.058,  43.020,
  41.038,  40.652,  40.030,  40.116,  43.626,  48.364,  49.912,  49.938,  50.431,  49.580},
{
  47.460,  48.107,  47.887,  46.268,  44.068,  43.444,  43.335,  43.291,  43.029,  42.088,
  40.039,  39.322,  39.411,  37.965,  36.779,  37.763,  31.055,  28.620,  28.503,  26.515,
  20.530,  14.199,  11.431,  18.650,  27.551,  30.870,  29.289,  24.622,  16.483,  12.443,
  15.178,  14.589,  20.881,  26.707,  24.730,  23.257,  25.229,  26.980,  25.755,  23.978,
  22.302,  19.864,  16.489,  13.996,  13.335,  13.440,  14.988,  14.971,  11.993,  10.020,
   5.497,   4.962,   0.894,  -1.862,  -3.983,  -7.606, -10.088, -12.170, -13.229, -15.265,
 -18.251, -22.887, -26.659, -28.253, -28.733, -29.555, -30.333, -31.195, -31.830, -35.111,
 -28.764, -28.397, -30.462, -29.691, -31.038, -29.518, -25.646, -21.173, -24.098, -25.302,
 -29.875, -32.950, -33.279, -32.083, -34.294, -36.170, -37.038, -37.987, -39.180, -40.588,
 -41.855, -42.868, -43.899, -44.546, -46.000, -47.623, -46.890, -47.599, -47.401, -43.953,
 -45.708, -44.352, -42.555, -43.336, -42.258, -40.034, -36.851, -36.057, -33.942, -31.353,
 -28.832, -26.324, -21.623, -17.143, -16.510, -15.473, -11.595,  -6.754,  -1.937,   1.553,
   4.307,   7.676,  10.452,  12.944,  15.883,  18.838,  21.553,  24.296,  27.087,  29.600,
  27.716,  28.457,  30.003,  32.916,  33.064,  34.320,  35.649,  39.826,  42.320,  41.430,
  38.186,  30.788,  18.972,  16.163,  21.916,  22.400,  22.024,  20.797,  18.808,  16.385,
  13.999,  11.705,  10.088,   8.067,   5.801,   4.649,   3.303,   2.109,   1.698,   0.221,
  -2.074,  -3.837,  -4.997,  -6.247,  -7.340,  -8.417,  -9.391,  -9.910,  -9.760, -10.215,
 -11.933, -11.138, -11.053, -12.298, -10.991, -10.891, -10.584, -10.599, -10.263, -10.693, -11.171,
 -11.297, -11.366, -11.772, -11.631, -12.229, -13.027, -13.209, -13.218, -13.560, -13.566,
 -13.619, -13.850, -13.821, -13.286, -13.099, -13.332, -13.846, -14.287, -14.615, -15.019,
 -15.369, -15.660, -16.044, -16.697, -17.242, -18.290, -19.486, -20.248, -21.079, -21.724,
 -22.553, -23.782, -24.721, -25.566, -27.055, -28.814, -30.076, -31.042, -32.213, -33.245,
 -34.151, -35.186, -36.067, -36.868, -37.497, -38.694, -38.979, -39.194, -39.865, -40.105,
 -40.146, -40.375, -40.458, -40.449, -39.959, -38.686, -37.859, -36.380, -33.256, -34.083,
 -31.684, -28.918, -30.432, -28.793, -28.741, -27.302, -24.706, -23.479, -23.496, -21.204,
 -20.373, -21.433, -20.048, -18.862, -18.673, -20.602, -24.016, -27.098, -28.356, -29.397,
 -28.748, -28.024, -28.966, -29.635, -29.862, -29.003, -29.279, -29.018, -29.044, -29.111,
 -29.190, -29.388, -29.797, -30.442, -30.276, -31.896, -31.908, -32.282, -34.152, -32.823,
 -32.332, -35.995, -37.463, -40.296, -39.912, -42.979, -44.446, -44.346, -43.984, -43.114,
 -42.972, -42.056, -40.878, -39.151, -37.712, -36.823, -35.448, -33.477, -31.382, -30.299,
 -28.682, -27.275, -25.859, -23.449, -20.984, -18.176, -15.655, -11.314,  -8.784,  -5.421,
  -2.944,   0.245,   3.928,   7.351,  10.692,  14.022,  17.691,  21.511,  24.754,  27.918,
  31.211,  34.466,  37.694,  40.912,  43.396,  45.627,  46.545,  47.043,  47.837,  48.132,
  48.277,  48.502,  48.453,  47.737,  48.161,  46.854,  45.233,  43.523,  42.705,  41.409,
  41.516,  42.751,  45.105,  47.277,  48.779,  45.761,  46.101,  46.962,  43.760,  43.175,
  44.757,  45.405,  41.645,  41.669,  42.412,  45.523,  47.540,  47.663,  46.838,  47.460},
{
  46.235,  45.242,  44.456,  44.511,  44.542,  43.043,  43.467,  42.751,  42.697,  43.112,
  42.474,  42.077,  41.920,  40.279,  38.398,  41.384,  30.834,  28.950,  28.259,  27.235,
  23.124,  21.025,  25.490,  31.255,  35.626,  37.645,  37.017,  34.466,  30.479,  27.555,
  29.386,  28.515,  33.139,  32.900,  32.362,  28.885,  29.579,  29.763,  27.104,  25.272,
  22.996,  21.553,  18.907,  18.438,  19.033,  19.625,  17.319,  15.984,  11.809,   7.863,
   4.637,  -2.919,  -7.336,  -9.472, -10.146, -10.183, -10.482, -11.459, -13.530, -16.742,
 -23.099, -28.105, -30.517, -32.541, -34.865, -36.736, -38.386, -41.053, -43.533, -44.904,
 -38.518, -32.662, -29.737, -28.750, -29.454, -30.646, -27.607, -28.377, -33.845, -41.784,
 -48.186, -50.722, -50.618, -50.645, -46.258, -42.145, -39.287, -41.753, -41.822, -44.393,
 -46.144, -46.380, -50.425, -55.690, -57.961, -57.078, -54.504, -53.604, -50.054, -47.384,
 -45.848, -43.554, -43.201, -43.047, -43.552, -42.068, -39.744, -36.792, -33.777, -31.371,
 -29.851, -24.801, -20.671, -16.406, -15.599, -15.600, -12.261,  -8.034,  -3.478,   0.390,
   4.283,   7.249,   9.991,  13.164,  15.757,  18.500,  20.538,  23.290,  25.893,  28.724,
  26.918,  27.524,  29.132,  31.457,  31.643,  32.656,  34.538,  36.770,  39.586,  42.368,
  42.581,  39.027,  28.451,  16.178,  18.006,  22.151,  21.822,  20.322,  18.371,  15.654,
  13.672,  11.448,   9.817,   7.842,   5.991,   4.140,   2.369,   1.026,   0.996,  -0.357,
  -2.303,  -3.629,  -4.707,  -5.678,  -7.332,  -8.423,  -8.739,  -9.455,  -9.897, -11.210,
 -13.192, -12.299, -12.932, -12.582, -11.648, -11.013, -10.638, -10.727, -11.158, -11.856, -12.042,
 -11.909, -12.397, -12.699, -12.464, -12.887, -12.960, -13.088, -12.942, -12.939, -12.824,
 -13.021, -13.003, -13.117, -13.190, -12.961, -13.222, -13.661, -14.204, -14.687, -14.975,
 -15.201, -15.778, -16.184, -16.996, -17.634, -18.428, -19.643, -20.129, -20.963, -22.126,
 -22.610, -24.018, -25.334, -26.193, -27.305, -29.064, -30.447, -31.438, -32.539, -33.987,
 -34.701, -35.772, -36.683, -37.181, -37.457, -38.033, -38.814, -38.870, -39.191, -39.639,
 -39.781, -39.816, -39.554, -39.289, -39.006, -37.982, -36.625, -33.354, -32.691, -32.350,
 -26.771, -27.129, -28.038, -27.154, -26.574, -25.644, -23.192, -22.638, -23.044, -21.541,
 -21.308, -20.511, -18.577, -17.033, -17.291, -19.775, -22.837, -25.634, -27.567, -28.210,
 -29.011, -28.944, -29.679, -30.117, -30.645, -30.280, -29.843, -30.422, -30.873, -29.507,
 -29.006, -30.028, -31.190, -31.731, -32.102, -31.330, -31.948, -32.104, -32.285, -33.327,
 -32.687, -33.776, -36.455, -39.009, -38.591, -41.888, -42.930, -42.502, -42.753, -42.471,
 -42.198, -41.646, -40.214, -38.541, -36.781, -35.117, -33.586, -32.017, -30.620, -28.497,
 -27.504, -25.353, -23.613, -21.748, -19.480, -16.437, -13.083, -10.211,  -6.815,  -3.706,
  -0.882,   2.016,   5.288,   8.545,  11.747,  15.538,  19.182,  22.761,  26.302,  29.197,
  32.518,  35.521,  38.422,  41.722,  44.543,  47.267,  49.432,  50.505,  50.550,  50.636,
  51.495,  52.014,  52.492,  52.101,  52.320,  49.196,  48.569,  48.556,  48.016,  46.624,
  45.844,  45.647,  46.954,  47.169,  50.747,  52.187,  48.142,  45.799,  50.640,  48.848,
  51.849,  51.801,  50.500,  47.313,  47.245,  48.400,  50.842,  49.950,  46.726,  46.235},
{
  49.156,  47.015,  46.072,  45.338,  44.446,  43.578,  42.678,  42.135,  41.625,  42.876,
  43.499,  43.169,  43.318,  43.700,  42.250,  42.496,  37.560,  31.298,  29.470,  25.130,
  23.303,  25.067,  29.204,  35.799,  39.142,  38.886,  38.538,  38.207,  36.429,  35.551,
  36.680,  37.041,  36.927,  35.977,  35.371,  35.700,  35.006,  34.238,  30.609,  27.850,
  25.829,  23.366,  23.479,  24.181,  23.647,  20.481,  18.466,  16.204,  13.048,   4.931,
  -7.703, -11.902, -12.491, -13.231, -15.984, -17.355, -15.688, -14.951, -19.156, -25.725,
 -28.906, -30.796, -32.656, -34.579, -36.438, -37.720, -37.727, -39.003, -43.407, -44.447,
 -40.601, -33.433, -30.305, -28.089, -27.635, -30.406, -33.390, -45.687, -51.503, -53.443,
 -53.771, -54.080, -54.972, -56.236, -56.227, -55.598, -54.727, -48.575, -44.649, -43.740,
 -48.307, -53.877, -58.643, -59.999, -58.415, -52.694, -48.364, -47.404, -46.550, -45.401,
 -44.247, -44.789, -46.203, -48.993, -46.919, -44.619, -42.385, -39.124, -35.713, -32.186,
 -29.788, -25.047, -19.565, -16.696, -15.615, -15.363, -12.513,  -8.768,  -4.587,  -0.687,
   3.011,   6.695,   9.802,  12.697,  15.309,  17.900,  20.456,  22.890,  25.576,  25.586,
  25.810,  27.198,  27.961,  28.904,  31.162,  30.558,  32.362,  35.931,  36.248,  38.126,
  41.619,  41.355,  35.366,  24.233,  16.680,  21.160,  22.238,  20.746,  18.620,  16.293,
  13.400,  11.682,  10.165,   8.406,   6.299,   4.480,   2.623,   0.842,  -0.436,  -1.756,
  -3.215,  -3.589,  -4.006,  -5.200,  -7.305,  -8.745,  -9.421,  -9.749, -10.644, -12.366,
  -9.567,  -7.565, -14.378, -12.913, -11.690, -11.402, -11.142, -11.499, -12.151, -12.621, -12.516,
 -12.426, -12.620, -12.988, -12.582, -12.696, -12.744, -12.945, -12.589, -12.294, -12.256,
 -12.222, -12.203, -12.355, -12.264, -12.175, -12.245, -12.714, -13.358, -13.994, -14.440,
 -14.914, -15.474, -15.751, -16.735, -17.379, -17.898, -19.165, -19.667, -20.980, -22.160,
 -22.458, -23.590, -24.623, -26.285, -26.988, -28.920, -30.124, -31.375, -32.600, -33.427,
 -34.595, -35.231, -35.677, -36.367, -36.825, -37.422, -38.025, -38.358, -38.633, -38.624,
 -38.902, -38.812, -38.616, -37.943, -37.303, -37.135, -33.025, -31.936, -30.994, -25.939,
 -24.360, -25.461, -25.462, -24.683, -24.007, -22.832, -21.271, -18.959, -20.545, -19.780,
 -19.390, -16.768, -15.116, -16.198, -17.370, -21.326, -23.212, -24.613, -25.876, -26.605,
 -27.084, -28.668, -29.796, -31.547, -32.308, -32.537, -32.371, -31.970, -30.795, -30.262,
 -30.201, -31.235, -32.447, -34.039, -34.667, -33.923, -33.544, -33.471, -32.258, -32.406,
 -33.429, -33.502, -34.855, -37.211, -37.491, -38.857, -42.060, -42.026, -41.276, -40.926,
 -40.643, -40.611, -39.737, -37.766, -35.835, -34.389, -32.200, -30.488, -28.407, -27.554,
 -26.697, -24.452, -22.436, -20.152, -17.852, -14.768, -11.122,  -7.959,  -4.985,  -2.034,
   0.906,   3.754,   6.702,   9.396,  12.607,  16.039,  19.725,  23.311,  26.951,  30.501,
  33.541,  36.535,  39.135,  41.791,  44.611,  47.801,  50.549,  52.835,  54.083,  55.008,
  56.041,  55.495,  55.354,  55.458,  54.788,  53.150,  51.548,  50.543,  50.028,  48.626,
  47.906,  48.571,  47.128,  48.163,  48.519,  50.144,  48.123,  45.075,  44.342,  49.007,
  52.736,  54.779,  55.377,  54.971,  52.206,  50.838,  51.202,  51.740,  50.401,  49.156},
{
  49.913,  48.806,  48.802,  47.426,  45.654,  44.710,  44.277,  43.562,  45.074,  46.097,
  43.900,  43.186,  43.082,  42.787,  43.028,  43.180,  44.374,  37.090,  29.527,  29.055,
  26.731,  28.379,  33.303,  38.473,  39.545,  39.648,  39.785,  40.044,  39.284,  40.028,
  39.885,  38.985,  37.378,  36.522,  36.234,  35.653,  35.648,  33.849,  30.616,  30.397,
  29.310,  28.121,  26.890,  26.122,  23.986,  20.675,  18.292,  13.538,  10.623,  -0.492,
 -11.868, -17.313, -16.910, -17.996, -20.606, -19.542, -19.237, -21.494, -23.967, -26.711,
 -29.351, -31.153, -33.055, -33.985, -35.524, -37.030, -37.701, -35.789, -34.652, -35.885,
 -36.376, -35.348, -33.004, -30.255, -29.660, -35.826, -46.759, -50.931, -51.701, -50.660,
 -51.457, -53.548, -57.462, -59.025, -60.476, -61.516, -62.565, -62.960, -62.850, -58.306,
 -55.482, -55.267, -56.105, -55.623, -52.480, -50.947, -48.687, -47.213, -46.408, -48.352,
 -51.897, -51.792, -51.532, -50.900, -48.222, -45.095, -41.606, -39.137, -35.749, -32.032,
 -28.467, -25.220, -19.237, -16.443, -14.017, -13.474, -12.244,  -9.364,  -5.026,  -1.374,
   1.948,   5.995,   9.966,  13.156,  14.957,  17.468,  20.209,  23.026,  24.291,  22.982,
  24.776,  24.783,  25.842,  27.670,  29.770,  30.632,  31.354,  32.791,  34.509,  36.699,
  39.119,  41.209,  38.049,  24.724,  15.470,  19.808,  22.030,  21.475,  19.890,  17.685,
  15.650,  13.378,  11.420,   9.140,   7.086,   5.321,   3.684,   1.993,   0.182,  -1.279,
  -2.848,  -3.759,  -4.622,  -5.722,  -7.373,  -8.512,  -9.205,  -9.836, -10.897, -12.920,
 -14.709, -13.111, -13.974, -12.544, -11.668, -11.935, -11.997, -12.321, -12.630, -13.030, -12.747,
 -12.229, -12.636, -12.915, -12.877, -12.790, -12.532, -12.380, -11.986, -11.667, -11.590,
 -11.311, -11.409, -11.676, -11.671, -11.606, -11.906, -12.524, -13.027, -13.399, -13.849,
 -13.748, -14.762, -15.181, -15.891, -16.591, -17.384, -18.331, -19.306, -20.691, -21.503,
 -21.906, -23.139, -24.104, -25.506, -26.547, -28.098, -29.722, -31.191, -32.002, -32.898,
 -33.777, -34.443, -34.802, -35.379, -35.858, -36.076, -36.708, -37.015, -37.248, -37.237,
 -36.944, -36.780, -36.682, -36.142, -35.663, -33.591, -29.662, -29.786, -27.231, -23.671,
 -24.748, -23.667, -22.301, -22.018, -20.946, -20.550, -20.497, -18.290, -17.789, -20.195,
 -17.917, -16.628, -14.447, -13.719, -15.962, -19.635, -22.202, -23.959, -25.242, -26.227,
 -26.643, -27.794, -29.588, -31.601, -32.851, -33.407, -33.411, -33.254, -32.665, -31.925,
 -32.271, -32.310, -33.458, -34.578, -34.915, -34.537, -34.148, -34.827, -34.314, -32.254,
 -32.926, -34.129, -33.146, -35.499, -36.329, -37.024, -37.071, -39.413, -39.839, -38.779,
 -38.252, -37.976, -37.951, -36.638, -35.697, -33.157, -32.105, -30.515, -27.424, -26.889,
 -25.494, -23.653, -21.652, -19.353, -16.257, -13.310, -10.176,  -6.733,  -3.320,  -0.374,
   2.495,   5.100,   7.742,  11.124,  13.863,  16.843,  20.456,  23.789,  27.013,  30.367,
  33.324,  36.381,  39.186,  42.519,  45.448,  48.604,  51.568,  53.832,  55.925,  57.268,
  57.658,  58.478,  58.244,  57.150,  56.443,  54.659,  53.552,  51.920,  50.667,  51.082,
  49.476,  49.286,  48.918,  48.385,  49.050,  50.421,  50.916,  48.614,  49.395,  53.380,
  54.109,  54.919,  55.081,  54.292,  54.330,  54.138,  53.306,  52.668,  51.817,  49.913},
{
  51.091,  49.616,  47.837,  48.638,  48.376,  45.486,  44.986,  44.224,  45.807,  48.223,
  45.726,  44.352,  44.248,  44.180,  45.255,  46.366,  44.921,  37.884,  39.073,  34.161,
  34.187,  36.194,  39.531,  40.470,  40.758,  40.268,  39.887,  39.971,  38.778,  38.296,
  38.505,  38.510,  37.691,  36.716,  34.646,  33.882,  32.134,  30.346,  29.007,  29.789,
  29.469,  28.816,  27.274,  25.964,  22.407,  20.256,  19.194,  10.657,   3.137,  -7.695,
 -15.891, -20.601, -18.034, -15.373, -15.610, -15.284, -17.660, -21.009, -24.103, -26.270,
 -28.329, -30.356, -32.541, -33.778, -35.224, -36.392, -36.927, -37.644, -39.394, -41.347,
 -41.082, -41.576, -40.162, -38.559, -34.735, -35.460, -40.597, -44.895, -46.389, -49.639,
 -55.718, -59.764, -61.506, -62.403, -64.040, -65.545, -66.515, -66.517, -66.297, -66.456,
 -66.646, -65.968, -64.378, -62.995, -61.834, -58.938, -55.661, -54.977, -56.990, -56.961,
 -55.322, -54.294, -52.731, -49.922, -47.233, -44.625, -43.169, -39.767, -36.301, -32.062,
 -28.800, -25.060, -19.705, -16.085, -13.175, -10.692, -10.045,  -8.466,  -4.390,  -0.752,
   2.711,   4.529,   8.405,  11.325,  14.146,  17.210,  20.383,  23.868,  24.167,  23.407,
  22.967,  23.884,  24.943,  26.177,  28.989,  30.194,  30.990,  31.452,  32.518,  34.420,
  37.738,  40.368,  37.377,  23.564,  14.508,  16.913,  20.693,  20.805,  19.799,  18.389,
  16.643,  14.530,  12.275,  10.125,   8.137,   6.318,   4.669,   2.876,   1.323,  -0.403,
  -1.835,  -3.189,  -4.480,  -5.637,  -6.774,  -8.340,  -8.879,  -9.763, -11.119, -12.787,
 -10.680, -15.222, -13.483, -11.853, -11.237, -11.909, -12.440, -12.543, -12.738, -12.366, -12.520,
 -12.795, -12.274, -12.583, -12.484, -12.276, -11.958, -11.668, -11.523, -11.078, -10.896,
 -10.612, -10.830, -10.894, -11.017, -11.350, -11.453, -11.978, -12.278, -12.610, -13.099,
 -13.097, -13.346, -14.244, -14.756, -15.573, -16.433, -17.439, -18.331, -19.683, -20.774,
 -21.795, -22.345, -23.775, -24.901, -26.290, -27.510, -28.476, -29.816, -30.446, -31.466,
 -32.286, -33.022, -33.940, -34.458, -34.975, -35.369, -36.107, -36.207, -36.467, -36.635,
 -36.583, -36.480, -35.881, -35.004, -33.730, -30.970, -28.372, -27.657, -23.423, -23.328,
 -23.160, -22.678, -21.096, -19.722, -19.352, -19.121, -18.692, -17.721, -16.281, -18.435,
 -16.662, -15.633, -13.011, -12.278, -17.207, -19.194, -21.011, -22.958, -24.249, -25.052,
 -26.179, -27.075, -27.863, -30.247, -32.077, -33.085, -32.621, -33.100, -33.235, -33.271,
 -32.778, -32.724, -34.039, -34.658, -34.467, -33.753, -34.417, -34.598, -34.545, -33.945,
 -32.690, -34.968, -34.505, -34.671, -34.040, -33.760, -34.871, -34.677, -34.895, -34.597,
 -34.265, -34.364, -34.336, -33.813, -33.864, -32.802, -31.325, -29.890, -28.505, -26.955,
 -24.943, -22.890, -20.761, -18.212, -15.080, -11.757,  -8.891,  -5.796,  -2.073,   1.179,
   4.479,   6.963,  10.150,  13.491,  16.119,  19.352,  22.523,  25.564,  28.817,  32.340,
  35.134,  38.105,  40.819,  43.432,  45.934,  49.030,  52.340,  54.938,  56.869,  58.756,
  59.329,  59.197,  59.515,  58.541,  57.636,  56.617,  55.695,  54.169,  53.536,  51.872,
  50.885,  51.690,  50.562,  49.433,  48.392,  47.031,  45.930,  45.837,  45.815,  48.540,
  54.009,  55.882,  55.093,  54.816,  54.471,  52.391,  52.386,  52.723,  52.456,  51.091},
{
  50.314,  49.832,  48.677,  47.538,  46.319,  45.808,  45.025,  44.773,  46.017,  48.052,
  47.020,  46.020,  46.068,  47.154,  47.572,  48.151,  47.591,  45.798,  39.786,  35.293,
  37.167,  42.536,  43.995,  42.733,  42.225,  40.967,  40.649,  39.347,  38.116,  37.431,
  36.454,  35.772,  36.097,  35.371,  33.528,  30.682,  28.318,  26.351,  24.856,  24.428,
  24.190,  24.136,  24.623,  25.697,  23.199,  18.466,  11.180,   5.424,   4.660,  -6.319,
 -15.991, -16.263, -11.441, -11.910, -12.953, -13.630, -17.307, -21.345, -24.671, -26.684,
 -28.207, -30.199, -31.865, -33.389, -34.145, -35.673, -37.067, -39.786, -42.159, -44.226,
 -41.211, -45.270, -49.216, -43.840, -37.570, -34.197, -34.429, -33.377, -35.203, -44.374,
 -54.184, -59.281, -61.212, -63.093, -64.547, -65.623, -65.011, -63.690, -61.901, -61.622,
 -62.497, -63.645, -62.515, -61.444, -61.021, -59.665, -57.777, -56.310, -56.336, -55.645,
 -55.625, -54.810, -52.883, -50.310, -48.031, -45.663, -41.977, -40.348, -37.970, -33.170,
 -27.841, -23.983, -20.619, -16.777, -13.779, -10.665,  -8.243,  -6.575,  -3.120,   0.653,
   2.802,   4.734,   6.959,  10.018,  14.160,  17.975,  20.708,  24.367,  26.288,  26.623,
  23.528,  22.907,  23.745,  24.878,  25.795,  26.966,  28.375,  29.624,  30.282,  32.305,
  35.303,  35.918,  29.321,  17.019,  11.211,  10.494,  16.296,  18.685,  18.743,  18.468,
  17.274,  15.728,  13.696,  11.617,   9.727,   7.467,   5.614,   3.884,   2.376,   0.502,
  -0.931,  -2.615,  -3.702,  -5.093,  -6.154,  -7.031,  -7.954,  -8.942, -10.505, -12.122,
  -7.250,  -7.932, -13.489, -11.323, -11.188, -12.062, -12.306, -12.449, -12.117, -11.597, -11.692,
 -11.905, -12.097, -11.875, -11.689, -11.288, -11.204, -10.795, -10.544, -10.227, -10.157,
  -9.893, -10.027,  -9.816,  -9.510,  -9.948, -10.466, -11.076, -11.100, -11.095, -11.593,
 -11.974, -12.286, -13.131, -13.770, -14.549, -15.233, -16.173, -16.800, -17.589, -18.587,
 -19.816, -20.999, -22.073, -23.224, -24.361, -25.415, -26.359, -27.297, -28.473, -29.266,
 -29.967, -30.626, -31.340, -31.946, -32.570, -32.907, -32.956, -32.928, -32.556, -32.102,
 -31.923, -31.664, -31.652, -31.811, -32.531, -29.710, -25.342, -24.879, -22.688, -21.539,
 -21.285, -20.394, -19.176, -18.023, -17.096, -17.492, -17.349, -16.912, -13.973, -12.887,
 -14.122, -15.563, -11.549, -10.703, -15.133, -18.207, -19.650, -20.907, -22.431, -23.289,
 -24.484, -26.006, -27.177, -28.132, -30.262, -31.840, -32.421, -33.349, -33.469, -33.341,
 -33.030, -33.088, -34.353, -34.626, -34.044, -35.896, -35.400, -34.055, -34.406, -34.525,
 -33.319, -33.684, -34.015, -33.240, -33.950, -32.312, -31.406, -32.153, -31.583, -30.688,
 -31.055, -29.408, -28.126, -30.041, -30.636, -29.664, -28.603, -28.269, -27.223, -25.678,
 -23.848, -21.494, -19.133, -16.126, -12.938,  -9.847,  -7.311,  -4.227,  -0.858,   3.419,
   6.530,   9.329,  11.862,  15.579,  18.567,  21.309,  24.360,  27.323,  30.362,  33.583,
  36.708,  39.716,  42.171,  44.616,  47.068,  49.739,  52.676,  55.452,  57.634,  59.482,
  60.741,  60.604,  60.487,  59.806,  59.109,  58.047,  57.279,  56.437,  56.102,  54.055,
  52.127,  51.802,  51.382,  51.086,  49.629,  48.571,  48.003,  48.103,  47.912,  49.108,
  54.540,  55.951,  55.873,  56.167,  54.537,  53.721,  53.679,  52.940,  51.249,  50.314},
{
  49.985,  50.005,  51.152,  49.946,  47.110,  46.589,  45.603,  45.549,  45.813,  50.617,
  47.223,  47.718,  48.386,  49.299,  48.498,  46.227,  47.390,  43.624,  39.050,  36.853,
  40.781,  44.511,  44.481,  44.752,  43.346,  41.084,  39.782,  39.535,  38.473,  33.713,
  30.448,  28.393,  27.341,  27.678,  29.121,  26.731,  24.576,  21.725,  20.114,  19.437,
  18.764,  18.191,  20.179,  20.348,  18.835,  14.051,  11.506,   9.426,  -1.376,  -7.977,
 -11.229, -11.810, -10.923, -11.991, -13.893, -16.555, -19.980, -23.163, -25.279, -26.956,
 -27.746, -29.330, -31.466, -33.087, -34.343, -35.459, -38.570, -40.653, -42.861, -42.807,
 -40.121, -35.959, -37.686, -37.469, -35.984, -35.323, -37.045, -38.447, -37.837, -38.134,
 -38.815, -43.919, -52.277, -55.724, -55.741, -57.231, -60.244, -62.559, -62.445, -63.612,
 -63.023, -61.742, -62.519, -63.117, -62.763, -60.955, -57.710, -56.238, -55.846, -55.895,
 -55.417, -54.121, -52.691, -50.144, -48.087, -45.806, -43.350, -39.789, -36.179, -32.727,
 -29.346, -26.288, -22.666, -19.127, -15.585, -11.383,  -7.496,  -4.932,  -2.370,   0.036,
   2.366,   4.280,   6.005,   8.425,  12.627,  17.574,  21.037,  22.542,  24.573,  26.435,
  24.571,  22.961,  23.109,  23.774,  24.724,  25.819,  27.197,  28.120,  29.052,  30.465,
  34.537,  32.730,  22.499,  20.894,  14.635,  10.981,   8.986,   9.999,  13.798,  16.468,
  16.992,  16.401,  15.053,  12.977,  11.112,   9.149,   6.704,   4.802,   3.366,   1.610,
  -0.109,  -1.415,  -2.788,  -4.291,  -5.474,  -6.083,  -6.974,  -7.773,  -9.257, -11.726,
  -9.678, -11.089, -12.529, -10.707, -10.708, -11.281, -11.941, -11.673, -10.924, -10.885, -11.195,
 -11.496, -11.440, -10.853, -10.648, -10.454, -10.039,  -9.383,  -8.952,  -8.793,  -8.684,
  -8.568,  -8.577,  -8.199,  -8.228,  -8.613,  -8.845,  -9.501,  -9.981, -10.175, -10.356,
 -10.721, -11.258, -11.909, -12.564, -13.155, -13.837, -14.910, -15.792, -16.615, -17.656,
 -18.903, -20.183, -21.393, -22.194, -23.141, -24.149, -25.111, -25.926, -26.730, -27.660,
 -28.590, -29.537, -30.108, -30.685, -31.181, -31.575, -31.676, -31.632, -31.243, -31.023,
 -30.626, -30.512, -30.172, -30.044, -30.161, -26.661, -24.125, -22.807, -21.369, -19.960,
 -19.376, -18.263, -17.163, -16.044, -15.040, -14.571, -14.739, -14.530, -13.594, -14.279,
 -13.550, -14.352, -13.445, -11.795, -13.697, -17.907, -18.833, -19.005, -20.851, -22.296,
 -23.694, -24.947, -25.399, -28.201, -29.094, -28.311, -32.139, -31.769, -32.904, -33.394,
 -33.530, -34.058, -34.260, -33.927, -34.126, -35.168, -35.723, -35.544, -35.412, -34.976,
 -33.373, -32.410, -33.282, -32.894, -30.580, -32.012, -29.756, -29.952, -29.321, -28.453,
 -28.266, -26.892, -26.276, -25.396, -26.644, -26.292, -26.253, -26.344, -25.892, -24.451,
 -22.915, -20.011, -16.691, -12.998, -10.316,  -8.144,  -5.723,  -1.032,   3.589,   5.205,
   9.463,  11.785,  15.248,  18.091,  20.741,  23.528,  26.287,  29.160,  32.166,  35.028,
  38.450,  41.618,  43.978,  46.041,  48.402,  50.653,  53.313,  55.873,  58.067,  60.233,
  61.541,  61.472,  61.460,  60.878,  60.387,  59.750,  59.365,  58.362,  57.464,  56.727,
  55.149,  54.150,  53.639,  52.735,  51.132,  50.530,  50.154,  50.978,  52.036,  52.578,
  55.626,  56.313,  57.102,  56.810,  54.489,  53.621,  54.027,  51.708,  50.245,  49.985},
{
  52.108,  51.846,  50.728,  49.958,  49.680,  48.893,  47.975,  45.828,  45.739,  47.476,
  48.546,  48.848,  48.411,  47.124,  43.147,  43.084,  43.123,  43.273,  42.868,  44.137,
  45.690,  44.537,  43.828,  43.810,  42.330,  41.153,  39.567,  37.944,  35.717,  30.903,
  29.434,  28.491,  25.597,  22.512,  20.609,  20.152,  19.461,  20.051,  17.979,  14.943,
  17.115,  21.772,  23.316,  21.591,  16.774,  11.916,   4.940,  -1.051,  -6.393,  -7.770,
  -8.968, -10.022, -11.745, -13.780, -16.748, -18.592, -20.314, -22.711, -25.224, -26.728,
 -27.690, -29.399, -31.513, -32.424, -34.437, -36.990, -39.174, -40.963, -42.406, -42.090,
 -39.620, -41.338, -41.383, -42.109, -43.775, -42.764, -41.620, -41.200, -42.067, -44.404,
 -46.672, -46.340, -47.930, -49.158, -50.684, -47.700, -48.976, -52.649, -61.991, -64.730,
 -63.018, -64.248, -64.079, -61.527, -57.334, -57.243, -59.599, -56.760, -56.153, -55.314,
 -53.369, -51.793, -49.998, -48.813, -45.887, -43.775, -42.571, -40.413, -37.948, -34.398,
 -31.277, -28.047, -24.955, -21.659, -17.660, -13.028,  -8.360,  -4.342,  -2.189,  -0.702,
   1.283,   3.452,   5.458,   7.647,  10.619,  15.223,  18.812,  21.006,  22.224,  23.995,
  25.405,  25.279,  26.275,  26.041,  26.798,  26.052,  25.688,  26.932,  28.098,  28.799,
  31.627,  33.140,  30.337,  30.896,  28.293,  28.077,  21.354,  10.537,   2.790,   5.614,
  12.132,  14.661,  14.407,  13.504,  12.341,  10.850,   8.674,   6.917,   4.715,   3.090,
   1.453,   0.047,  -1.774,  -3.243,  -4.561,  -5.326,  -5.622,  -6.534,  -8.017, -10.813,
  -8.361, -10.389, -11.110,  -9.167,  -9.809, -11.304, -10.537,  -9.679, -10.091,  -9.629,  -9.689,
  -9.525,  -9.264,  -9.212,  -9.002,  -9.011,  -8.375,  -7.637,  -6.968,  -7.184,  -7.109,
  -6.945,  -6.906,  -6.886,  -6.802,  -7.189,  -7.512,  -8.008,  -8.090,  -8.318,  -8.706,
  -9.476, -10.083, -10.563, -11.176, -11.877, -12.478, -13.418, -14.392, -15.556, -16.684,
 -17.836, -18.785, -19.797, -20.685, -21.532, -22.309, -23.139, -24.144, -25.249, -26.177,
 -27.066, -28.130, -28.975, -29.646, -30.162, -30.360, -30.395, -30.505, -30.315, -29.987,
 -29.670, -29.290, -29.073, -29.285, -29.236, -24.110, -23.214, -21.243, -20.071, -19.272,
 -18.329, -17.543, -16.205, -16.022, -14.916, -13.664, -12.971, -11.866, -10.674, -10.103,
 -10.136, -13.595, -12.496, -13.632, -15.003, -15.136, -17.603, -18.734, -21.115, -23.714,
 -25.776, -26.558, -26.161, -27.205, -29.010, -29.039, -30.253, -30.935, -32.370, -34.306,
 -34.796, -35.226, -34.713, -33.963, -33.787, -33.969, -35.169, -35.757, -36.349, -36.699,
 -36.311, -35.393, -35.316, -33.550, -32.079, -31.517, -29.240, -28.223, -27.380, -26.730,
 -26.533, -25.625, -24.416, -24.900, -23.290, -22.192, -21.537, -20.962, -20.950, -20.026,
 -19.752, -18.050, -13.933,  -9.824,  -8.298,  -6.617,  -2.479,   3.678,   9.570,  13.012,
  16.154,  17.415,  18.555,  20.451,  22.972,  25.774,  28.560,  31.422,  34.003,  36.619,
  39.837,  42.698,  45.005,  47.290,  49.688,  51.736,  53.975,  56.482,  58.598,  60.869,
  62.257,  62.382,  62.265,  62.320,  62.174,  61.605,  61.697,  60.799,  59.654,  57.687,
  58.340,  57.715,  56.511,  55.561,  53.139,  50.804,  51.824,  55.259,  52.968,  53.558,
  55.319,  56.552,  56.227,  56.099,  55.206,  53.615,  51.426,  50.233,  51.390,  52.108},
{
  48.706,  49.203,  50.571,  51.977,  50.915,  50.381,  51.084,  51.578,  48.571,  45.549,
  46.414,  44.595,  42.135,  40.782,  42.194,  43.049,  44.225,  46.350,  46.822,  46.377,
  45.268,  43.787,  43.254,  40.494,  39.958,  38.556,  36.894,  36.039,  34.973,  33.248,
  32.364,  29.301,  24.510,  21.533,  20.391,  16.699,  17.437,  15.254,  12.709,  15.834,
  20.900,  20.146,  18.426,  13.829,   6.502,   2.029,  -2.073,  -5.499,  -7.392,  -8.958,
 -10.584, -11.722, -12.605, -14.779, -16.424, -17.881, -20.210, -22.967, -25.445, -27.341,
 -28.908, -30.163, -32.212, -33.639, -35.516, -37.186, -38.430, -38.778, -37.526, -38.966,
 -41.208, -42.449, -44.147, -44.219, -44.839, -45.057, -45.960, -46.852, -47.365, -48.622,
 -49.422, -51.217, -51.426, -51.518, -52.265, -55.799, -61.163, -66.301, -63.653, -60.411,
 -60.805, -59.501, -57.950, -58.992, -59.746, -61.526, -57.832, -55.964, -54.463, -53.319,
 -52.005, -50.846, -49.440, -47.001, -45.812, -44.058, -41.886, -39.754, -37.316, -35.093,
 -32.127, -28.427, -25.465, -22.330, -18.422, -13.863, -10.218,  -6.106,  -3.561,  -1.565,
   0.312,   2.325,   4.352,   6.656,   9.415,  12.740,  16.011,  18.755,  20.379,  22.291,
  25.271,  25.381,  24.454,  24.603,  25.004,  26.952,  26.666,  26.269,  26.945,  27.744,
  29.552,  30.624,  31.342,  31.694,  30.541,  30.534,  30.973,  29.627,  18.909,   2.230,
  -2.953,   2.040,   8.598,  11.454,  12.233,  11.414,   9.848,   8.227,   6.261,   4.397,
   3.316,   1.516,  -0.331,  -1.452,  -2.951,  -4.145,  -4.763,  -5.315,  -7.532, -10.514,
  -7.658, -10.766, -10.076,  -8.576,  -9.472, -10.240,  -8.786,  -8.735,  -8.579,  -8.270,  -7.974,
  -7.787,  -7.725,  -7.494,  -7.067,  -6.467,  -5.967,  -5.967,  -5.681,  -5.663,  -5.258,
  -4.673,  -4.545,  -4.742,  -4.645,  -4.850,  -5.238,  -5.702,  -5.871,  -6.055,  -6.620,
  -7.390,  -7.990,  -8.599,  -9.378, -10.404, -11.415, -12.216, -12.920, -13.987, -15.122,
 -16.249, -17.306, -18.205, -19.034, -19.804, -20.801, -21.612, -22.520, -23.794, -25.024,
 -26.037, -26.846, -27.525, -28.333, -28.948, -29.209, -29.291, -29.118, -28.927, -28.542,
 -28.177, -27.715, -27.734, -28.710, -27.883, -23.402, -22.472, -20.567, -19.494, -17.952,
 -17.550, -16.998, -17.081, -15.495, -13.119, -11.491, -11.183, -10.896,  -9.262,  -7.904,
  -9.767, -12.816, -10.207, -14.641, -14.572, -12.935, -16.704, -19.772, -23.184, -24.917,
 -25.330, -25.754, -26.159, -26.595, -27.722, -28.918, -29.376, -30.527, -32.905, -35.521,
 -36.827, -36.129, -35.792, -34.454, -33.708, -35.515, -35.443, -36.395, -36.235, -36.758,
 -37.098, -36.772, -35.783, -34.098, -30.870, -29.181, -28.984, -27.949, -27.494, -26.505,
 -25.678, -24.262, -23.539, -23.342, -23.373, -22.941, -21.334, -19.435, -18.036, -14.615,
 -12.825, -12.301, -10.826,  -8.525,  -6.622,  -3.094,   2.603,   8.258,  11.945,  15.565,
  16.939,  17.827,  20.494,  23.436,  25.370,  27.553,  30.173,  32.986,  35.070,  37.692,
  40.445,  43.511,  45.905,  48.403,  50.838,  52.806,  54.584,  57.150,  59.206,  60.922,
  62.784,  63.833,  63.715,  63.855,  63.635,  63.118,  62.567,  62.591,  60.887,  59.893,
  58.045,  58.353,  58.497,  55.745,  54.662,  55.243,  53.070,  49.869,  47.417,  47.332,
  51.317,  50.813,  48.652,  47.762,  49.560,  46.230,  46.928,  48.663,  48.555,  48.706},
{
  47.140,  48.970,  50.685,  52.593,  52.964,  50.720,  52.180,  52.963,  44.668,  41.503,
  39.049,  40.733,  42.678,  43.195,  44.444,  45.537,  46.475,  46.312,  46.065,  45.082,
  44.007,  42.885,  43.254,  41.763,  38.406,  36.509,  34.317,  32.271,  32.970,  32.149,
  30.976,  29.374,  27.059,  25.829,  25.691,  21.651,  17.655,  16.098,  14.151,  12.611,
  12.924,  12.761,  11.277,   8.209,   4.202,   0.569,  -2.133,  -4.279,  -6.251,  -8.902,
 -10.968, -12.476, -14.077, -16.175, -18.239, -20.177, -22.106, -23.817, -25.982, -27.790,
 -28.946, -30.078, -31.597, -33.351, -35.144, -36.529, -37.686, -38.404, -39.489, -40.640,
 -42.372, -43.149, -43.542, -44.265, -45.354, -47.276, -48.466, -49.605, -49.227, -45.465,
 -40.262, -43.903, -52.379, -55.944, -58.216, -62.579, -66.003, -66.743, -64.781, -61.271,
 -59.302, -56.896, -57.402, -56.688, -56.141, -52.932, -52.558, -50.850, -49.210, -48.987,
 -49.338, -48.903, -47.337, -46.974, -45.765, -43.640, -41.639, -39.591, -37.015, -34.296,
 -31.935, -28.734, -25.595, -22.088, -18.123, -14.792, -12.127,  -8.439,  -4.999,  -1.770,
   0.371,   1.939,   4.087,   5.624,   8.129,  11.002,  13.553,  16.226,  17.957,  20.292,
  22.695,  23.763,  23.018,  23.532,  25.212,  28.680,  29.927,  26.219,  25.613,  26.637,
  28.156,  28.599,  28.418,  28.042,  27.386,  25.606,  25.708,  26.460,  27.966,  24.029,
  13.590,   1.264,  -3.962,   2.664,   9.904,  11.631,  11.402,   9.843,   7.985,   6.394,
   4.658,   3.106,   1.570,  -0.288,  -1.891,  -3.432,  -3.964,  -4.877,  -7.250, -10.188,
  -3.284,  -9.655,  -8.699,  -7.643,  -8.566,  -7.491,  -7.760,  -7.094,  -6.801,  -6.931,  -6.432,
  -6.278,  -5.691,  -5.638,  -5.195,  -4.718,  -4.308,  -3.904,  -3.920,  -3.601,  -3.296,
  -2.906,  -2.735,  -2.641,  -2.510,  -2.798,  -3.107,  -3.171,  -3.367,  -3.878,  -4.233,
  -5.001,  -5.710,  -6.518,  -7.454,  -8.292,  -9.091, -10.026, -11.027, -12.150, -13.273,
 -14.323, -15.134, -15.969, -16.914, -17.801, -18.920, -20.080, -21.247, -22.398, -23.638,
 -24.572, -25.545, -26.336, -27.006, -27.436, -27.892, -27.734, -27.759, -26.749, -26.248,
 -26.550, -26.442, -26.519, -27.498, -26.905, -22.561, -21.835, -20.387, -20.094, -19.165,
 -17.179, -16.304, -15.281, -15.024, -14.016, -13.005, -11.080,  -9.422,  -8.557,  -7.482,
 -11.822, -11.195, -13.699, -14.833, -14.621, -15.451, -17.159, -19.522, -21.604, -22.541,
 -23.958, -24.647, -25.391, -26.696, -27.304, -27.710, -28.414, -31.051, -31.370, -33.788,
 -35.926, -36.664, -37.238, -36.119, -34.980, -36.402, -37.071, -37.379, -37.244, -36.589,
 -35.908, -35.312, -34.489, -34.044, -32.075, -29.968, -29.588, -28.049, -26.505, -26.190,
 -25.536, -24.677, -23.708, -22.650, -21.990, -21.141, -19.552, -18.198, -17.432, -15.398,
 -12.812, -10.243,  -6.615,  -3.908,  -1.695,   3.321,   6.164,   9.743,  14.045,  17.046,
  19.469,  20.909,  23.481,  26.258,  27.854,  29.688,  31.925,  34.126,  37.195,  39.194,
  41.680,  44.206,  46.337,  48.990,  50.640,  53.197,  55.129,  57.479,  59.525,  61.268,
  63.099,  64.766,  64.687,  64.758,  64.969,  64.413,  63.814,  62.830,  61.769,  60.760,
  59.381,  58.540,  58.957,  58.817,  57.225,  55.462,  56.170,  53.879,  52.998,  51.037,
  47.793,  46.464,  45.555,  44.927,  44.856,  44.392,  46.617,  47.560,  47.214,  47.140},
{
  48.169,  48.786,  50.028,  50.290,  50.246,  49.311,  49.330,  51.283,  51.853,  47.574,
  48.953,  49.278,  47.630,  44.762,  46.672,  46.977,  46.046,  45.433,  44.912,  44.650,
  43.669,  42.745,  42.269,  42.273,  41.328,  39.493,  37.340,  32.466,  31.186,  30.877,
  29.128,  26.920,  24.625,  23.296,  21.664,  19.705,  17.423,  14.532,  13.224,  12.159,
  10.809,   9.115,   7.320,   5.686,   3.435,   1.063,  -1.481,  -3.783,  -5.663,  -7.998,
 -10.426, -12.435, -14.755, -16.119, -17.794, -19.909, -22.027, -24.293, -25.865, -27.184,
 -28.767, -30.064, -31.004, -33.041, -34.793, -36.224, -37.407, -38.345, -38.825, -41.152,
 -41.463, -41.864, -42.586, -43.603, -44.037, -46.005, -47.964, -50.145, -50.839, -50.705,
 -51.273, -50.321, -52.189, -50.515, -51.266, -55.777, -59.804, -60.752, -60.295, -57.507,
 -54.264, -52.931, -51.465, -49.932, -48.866, -51.803, -49.740, -49.317, -49.435, -49.145,
 -47.841, -45.806, -44.577, -44.723, -44.101, -42.469, -40.537, -38.559, -36.388, -33.858,
 -31.536, -28.342, -25.440, -22.150, -18.838, -16.181, -13.483, -10.635,  -7.318,  -3.076,
   0.116,   2.135,   3.757,   5.369,   7.215,   9.999,  12.162,  14.209,  16.762,  18.797,
  20.460,  21.139,  21.786,  22.248,  23.684,  26.580,  29.054,  28.526,  26.092,  24.621,
  25.545,  26.500,  26.411,  25.718,  24.122,  23.080,  23.014,  22.913,  23.174,  23.699,
  25.003,  23.752,  17.184,   6.863,   3.587,  10.242,  12.119,  12.158,  10.482,   8.696,
   6.945,   4.887,   3.087,   1.149,  -0.617,  -1.520,  -2.527,  -4.096,  -7.389,  -4.472,
  -4.702,  -8.580,  -6.906,  -6.066,  -6.341,  -6.279,  -6.167,  -5.755,  -5.170,  -5.335,  -4.947,
  -4.427,  -3.814,  -3.278,  -2.934,  -2.722,  -2.417,  -1.781,  -1.714,  -1.309,  -0.917,
  -0.752,  -0.394,  -0.434,  -0.525,  -0.541,  -0.469,  -0.315,  -0.529,  -1.179,  -1.960,
  -2.779,  -3.530,  -4.122,  -4.932,  -5.813,  -6.693,  -7.800,  -8.739,  -9.788, -10.969,
 -11.904, -12.881, -13.665, -14.682, -15.910, -17.041, -18.314, -19.565, -20.781, -21.879,
 -22.969, -23.937, -24.676, -25.446, -25.660, -25.789, -25.841, -25.522, -24.980, -24.595,
 -24.951, -25.114, -25.221, -26.064, -25.796, -22.724, -20.457, -19.855, -19.683, -20.384,
 -20.659, -18.197, -16.499, -15.843, -14.887, -13.375, -12.018, -11.668, -11.244, -11.122,
 -12.677, -13.909, -14.745, -15.858, -15.852, -15.706, -16.486, -19.065, -20.769, -22.183,
 -23.617, -25.206, -26.288, -26.759, -26.861, -27.286, -27.812, -27.562, -30.159, -31.013,
 -32.245, -33.865, -35.433, -35.502, -35.717, -37.496, -37.931, -37.907, -37.555, -36.926,
 -36.632, -36.356, -35.345, -33.849, -33.102, -32.514, -31.870, -28.841, -27.580, -26.106,
 -24.801, -24.547, -23.634, -21.945, -20.162, -19.373, -18.210, -16.869, -14.592, -13.244,
 -11.602,  -9.475,  -5.876,  -1.681,   1.467,   4.400,   7.456,  10.261,  13.752,  17.574,
  20.133,  23.713,  26.235,  29.085,  29.490,  30.962,  33.487,  35.322,  37.736,  40.294,
  42.328,  44.815,  46.909,  49.280,  51.753,  53.469,  55.461,  57.377,  59.505,  61.434,
  63.087,  65.018,  65.709,  65.730,  65.928,  65.414,  64.066,  63.051,  61.468,  60.855,
  60.815,  60.480,  58.985,  58.577,  56.714,  55.001,  54.190,  52.202,  49.996,  49.060,
  48.116,  47.452,  47.336,  46.670,  46.214,  48.421,  49.390,  48.342,  47.805,  48.169},
{
  47.587,  47.510,  47.513,  47.902,  48.196,  48.285,  49.635,  49.310,  48.389,  47.703,
  48.668,  49.611,  49.545,  48.629,  48.473,  47.768,  46.594,  45.356,  45.104,  43.717,
  42.814,  42.160,  41.583,  41.317,  39.103,  39.401,  36.162,  33.107,  32.543,  30.868,
  28.662,  27.505,  25.371,  23.718,  22.318,  21.078,  19.790,  17.790,  15.420,  13.552,
  11.390,   9.024,   6.992,   5.130,   2.467,  -1.057,  -3.511,  -5.539,  -8.324, -11.148,
 -13.687, -15.269, -16.382, -17.617, -19.173, -20.398, -21.392, -22.137, -22.907, -25.346,
 -26.605, -28.595, -30.329, -32.063, -33.883, -34.726, -35.306, -36.333, -37.444, -38.778,
 -39.414, -41.348, -42.375, -42.797, -43.729, -44.721, -45.832, -47.223, -47.485, -48.448,
 -49.385, -49.042, -49.111, -47.988, -47.500, -46.718, -47.996, -53.042, -53.722, -51.441,
 -47.324, -43.082, -45.396, -47.476, -49.375, -49.307, -48.373, -47.326, -45.696, -44.976,
 -44.478, -43.397, -43.756, -43.530, -43.199, -42.238, -40.230, -37.758, -35.472, -33.068,
 -31.007, -28.338, -25.517, -22.708, -19.934, -17.040, -14.796, -11.877,  -8.291,  -4.346,
  -0.638,   1.643,   3.634,   4.855,   6.650,   9.253,  11.586,  13.398,  15.861,  18.120,
  18.573,  19.562,  20.297,  20.662,  22.506,  24.519,  26.049,  28.228,  27.309,  25.229,
  23.766,  23.677,  24.913,  24.531,  21.640,  21.134,  20.454,  21.082,  21.244,  21.219,
  21.549,  22.533,  24.712,  23.648,  16.063,   4.343,   9.694,  11.979,  11.827,  10.480,
   8.835,   6.896,   4.641,   2.584,   1.141,   0.065,  -0.864,  -3.166,  -6.641,  -2.995,
  -4.377,  -6.042,  -4.957,  -4.049,  -4.505,  -4.624,  -4.115,  -3.849,  -3.702,  -3.295,  -2.748,
  -2.580,  -1.977,  -1.408,  -0.867,  -0.587,   0.073,   0.468,   0.516,   1.003,   1.480,
   1.341,   2.038,   1.764,   1.851,   1.805,   1.729,   1.855,   1.703,   1.128,   0.355,
  -0.294,  -1.021,  -1.593,  -2.433,  -3.401,  -4.445,  -5.330,  -6.273,  -7.206,  -8.273,
  -9.361, -10.354, -11.437, -12.546, -13.839, -15.073, -16.228, -17.551, -18.659, -19.755,
 -20.694, -21.630, -22.443, -23.282, -23.885, -24.027, -23.888, -23.690, -23.276, -23.469,
 -23.473, -23.718, -24.071, -24.525, -24.478, -23.070, -20.539, -20.166, -18.239, -19.521,
 -19.650, -18.592, -17.465, -15.973, -15.090, -14.396, -12.718, -12.827, -11.543, -11.879,
 -12.451, -14.184, -14.644, -15.835, -16.314, -15.929, -16.895, -18.729, -20.899, -22.090,
 -23.716, -25.302, -26.380, -26.984, -27.658, -28.889, -29.641, -28.979, -31.297, -31.699,
 -32.536, -35.366, -36.057, -36.753, -37.239, -37.232, -36.903, -36.853, -36.696, -36.641,
 -37.246, -36.514, -35.732, -34.863, -33.938, -32.685, -31.065, -29.195, -27.942, -27.748,
 -25.762, -24.224, -21.932, -20.934, -20.432, -19.402, -18.059, -17.173, -15.041, -13.535,
 -11.001,  -7.803,  -3.928,  -0.165,   3.299,   6.475,   8.498,  11.194,  14.709,  17.416,
  20.523,  25.060,  27.799,  31.982,  35.015,  35.528,  35.147,  36.091,  38.285,  41.208,
  43.604,  45.575,  47.662,  50.016,  52.229,  53.868,  55.682,  57.475,  59.797,  61.625,
  62.896,  64.984,  66.012,  65.986,  65.606,  65.160,  64.049,  62.228,  62.106,  61.910,
  61.141,  60.298,  59.111,  58.102,  57.026,  56.074,  55.219,  52.579,  50.165,  49.863,
  50.142,  48.695,  48.037,  48.238,  50.297,  49.724,  49.315,  48.517,  48.125,  47.587},
{
  47.370,  46.237,  45.615,  46.000,  46.384,  47.377,  48.018,  49.141,  49.306,  48.101,
  46.634,  45.956,  45.186,  44.741,  45.159,  46.447,  46.296,  44.001,  43.596,  43.887,
  42.729,  41.302,  39.929,  38.639,  38.413,  35.937,  33.961,  33.795,  32.628,  31.076,
  29.717,  28.115,  25.093,  23.794,  22.490,  21.264,  19.895,  17.974,  16.107,  14.155,
  11.844,   9.246,   6.835,   4.350,   0.784,  -3.035,  -5.029,  -8.507, -11.823, -12.915,
 -14.187, -15.811, -16.913, -18.108, -19.294, -20.389, -21.503, -21.441, -20.388, -21.737,
 -24.239, -26.904, -28.782, -31.160, -32.600, -32.839, -32.781, -33.878, -34.799, -36.149,
 -38.290, -39.856, -40.696, -40.858, -42.206, -42.898, -43.227, -42.905, -42.750, -43.253,
 -44.784, -45.989, -46.924, -50.292, -51.014, -51.400, -51.473, -48.907, -46.253, -42.643,
 -42.488, -44.317, -48.283, -49.441, -48.618, -47.558, -46.693, -45.511, -44.357, -43.831,
 -43.509, -43.703, -43.926, -43.188, -42.590, -41.351, -39.806, -36.766, -34.393, -32.338,
 -30.491, -28.056, -25.740, -23.324, -20.748, -18.256, -15.627, -12.666,  -9.360,  -5.680,
  -1.872,   1.515,   3.054,   4.955,   6.565,   9.092,  11.444,  13.621,  16.049,  17.406,
  18.116,  18.905,  19.097,  19.772,  20.679,  21.909,  23.388,  25.300,  26.075,  25.690,
  23.504,  21.629,  23.795,  22.783,  21.620,  21.833,  19.800,  19.957,  19.911,  19.886,
  20.172,  20.839,  21.171,  23.905,  24.990,  15.253,   2.410,   6.704,  10.261,  10.403,
   9.389,   8.226,   6.461,   4.508,   3.239,   1.918,   0.710,  -1.907,  -3.910,  -0.303,
  -5.569,  -3.030,  -2.168,  -2.934,  -2.200,  -2.062,  -2.033,  -1.736,  -1.827,  -1.567,  -1.045,
  -0.492,  -0.003,   0.539,   0.780,   1.296,   1.950,   2.555,   3.273,   3.596,   3.849,
   3.903,   4.099,   4.314,   4.357,   4.186,   4.096,   3.897,   3.672,   3.291,   2.900,
   2.141,   1.407,   0.527,  -0.217,  -0.968,  -1.725,  -2.657,  -3.503,  -4.311,  -5.265,
  -6.516,  -7.696,  -8.669,  -9.780, -10.988, -12.332, -13.717, -14.836, -16.039, -17.208,
 -18.367, -19.223, -19.966, -20.775, -21.343, -21.638, -21.613, -21.713, -21.752, -21.818,
 -21.932, -22.026, -22.958, -22.549, -22.974, -20.301, -20.262, -21.612, -17.851, -17.923,
 -18.185, -17.497, -16.633, -16.352, -15.403, -15.096, -12.981, -13.982, -14.267, -14.565,
 -14.663, -15.929, -16.236, -16.733, -17.242, -17.670, -17.766, -18.327, -20.637, -21.999,
 -23.735, -24.970, -27.178, -27.290, -28.285, -29.200, -30.797, -31.572, -30.647, -33.131,
 -34.551, -34.803, -36.212, -37.171, -37.453, -37.385, -37.016, -37.684, -37.805, -37.916,
 -38.194, -37.684, -36.169, -34.535, -33.445, -32.739, -31.792, -29.937, -27.643, -27.507,
 -26.919, -24.187, -22.297, -20.796, -19.950, -19.348, -17.878, -16.749, -15.639, -14.214,
  -8.642,  -4.234,  -1.501,   1.911,   5.254,   7.926,   9.465,  12.947,  16.672,  20.574,
  25.154,  27.226,  28.971,  31.575,  34.888,  35.863,  36.419,  37.454,  39.735,  42.073,
  44.086,  46.325,  48.865,  50.724,  52.354,  54.097,  56.047,  57.962,  59.940,  61.732,
  63.386,  65.054,  65.604,  65.093,  64.617,  63.650,  62.780,  62.624,  62.563,  62.136,
  61.507,  60.661,  59.799,  58.432,  57.263,  56.382,  54.695,  54.366,  53.143,  52.129,
  52.391,  53.423,  53.577,  52.400,  51.463,  50.653,  49.590,  48.888,  48.277,  47.370},
{
  47.048,  45.552,  44.643,  44.714,  45.423,  46.672,  47.150,  48.396,  48.233,  48.414,
  48.345,  46.995,  46.210,  46.606,  46.836,  46.411,  46.198,  43.452,  43.159,  43.032,
  42.772,  40.773,  38.812,  36.707,  33.114,  32.465,  33.543,  33.731,  32.703,  30.726,
  28.436,  25.988,  23.263,  22.832,  20.769,  18.545,  17.004,  15.402,  12.603,  10.270,
   9.566,   8.577,   7.370,   4.702,   1.410,  -2.642,  -5.434,  -6.938,  -8.645, -11.655,
 -13.591, -15.283, -16.019, -16.107, -16.429, -18.706, -21.280, -21.508, -19.875, -18.612,
 -21.778, -24.293, -26.508, -29.059, -30.523, -30.976, -31.142, -32.077, -32.516, -33.993,
 -35.462, -36.505, -37.682, -38.577, -39.567, -39.997, -40.086, -39.911, -40.356, -42.327,
 -44.326, -44.907, -45.260, -46.057, -46.825, -44.366, -44.111, -41.840, -40.696, -42.176,
 -42.891, -44.376, -45.512, -47.181, -46.456, -45.589, -45.287, -45.336, -44.613, -43.507,
 -43.362, -43.198, -42.482, -41.501, -41.037, -40.403, -39.438, -36.869, -33.769, -31.944,
 -29.776, -27.586, -25.259, -23.205, -21.033, -18.634, -16.064, -13.399, -10.460,  -6.922,
  -3.361,   0.408,   3.268,   4.560,   6.752,   8.927,  11.179,  12.867,  14.285,  15.666,
  16.102,  17.627,  19.346,  19.625,  20.066,  20.660,  21.272,  22.789,  23.904,  24.157,
  23.262,  21.446,  22.433,  22.142,  21.172,  21.010,  20.014,  19.841,  20.159,  20.278,
  21.001,  20.756,  21.001,  21.689,  24.642,  24.598,  12.253,   1.457,   3.723,   7.995,
   9.176,   8.798,   7.830,   6.818,   5.244,   4.005,   2.333,  -0.522,   0.596,   1.164,
  -1.751,   0.277,   0.102,  -0.402,  -0.236,   0.026,  -0.152,   0.337,   0.228,   0.176,   0.356,
   0.885,   1.249,   1.822,   1.840,   2.694,   3.560,   4.062,   4.999,   5.592,   5.682,
   5.965,   6.167,   6.573,   6.663,   6.646,   6.511,   6.234,   5.730,   5.446,   5.104,
   4.606,   4.025,   3.380,   2.638,   2.007,   1.234,   0.459,  -0.576,  -1.604,  -2.586,
  -3.800,  -4.943,  -6.066,  -7.231,  -8.105,  -9.368, -10.730, -12.018, -13.154, -14.310,
 -15.580, -16.708, -17.627, -18.332, -19.045, -19.411, -19.615, -20.035, -20.044, -20.551,
 -20.839, -21.535, -21.418, -19.565, -17.300, -17.378, -20.566, -18.875, -16.704, -16.168,
 -16.313, -16.279, -15.794, -15.383, -15.312, -14.701, -14.471, -14.551, -15.125, -16.155,
 -16.312, -16.665, -16.977, -17.176, -17.811, -18.618, -18.751, -20.026, -21.493, -21.522,
 -23.290, -25.304, -27.070, -28.057, -29.399, -30.555, -32.144, -33.375, -34.415, -35.250,
 -36.255, -37.110, -37.558, -38.081, -38.614, -38.693, -37.850, -38.385, -38.592, -39.139,
 -39.390, -38.937, -38.051, -36.958, -35.790, -34.675, -34.368, -33.111, -29.618, -26.881,
 -26.774, -26.173, -23.657, -20.041, -18.923, -19.973, -19.545, -17.795, -16.150, -13.761,
  -9.638,  -3.343,   1.619,   4.614,   6.972,   9.488,  12.220,  15.496,  21.733,  23.794,
  22.200,  23.321,  26.266,  29.104,  32.864,  35.086,  37.177,  39.425,  41.483,  43.430,
  45.235,  47.304,  49.555,  51.613,  53.092,  54.709,  56.190,  58.070,  60.228,  61.735,
  63.782,  64.714,  65.092,  64.486,  64.182,  64.026,  63.378,  63.140,  62.688,  62.344,
  61.441,  60.294,  58.977,  58.023,  56.926,  55.786,  56.474,  57.675,  58.253,  57.635,
  55.954,  54.983,  53.880,  52.801,  51.730,  50.816,  50.445,  49.067,  48.255,  47.048},
{
  45.038,  45.048,  44.892,  44.569,  45.712,  46.709,  48.087,  48.733,  48.497,  48.236,
  48.029,  47.189,  47.303,  46.695,  45.781,  44.796,  44.127,  44.025,  42.673,  41.969,
  40.009,  37.209,  35.166,  32.941,  31.034,  31.923,  32.377,  31.988,  31.206,  29.283,
  27.298,  24.798,  23.131,  21.303,  19.478,  15.996,  13.989,  12.378,  10.993,  10.943,
  10.526,   9.525,   7.756,   5.482,   3.434,   0.420,  -4.604,  -7.245,  -8.315, -10.037,
 -11.801, -13.562, -14.870, -15.055, -15.518, -16.929, -19.619, -20.910, -18.282, -17.158,
 -19.360, -22.293, -24.280, -26.105, -28.266, -29.175, -29.287, -30.437, -31.133, -32.380,
 -33.374, -33.878, -34.528, -35.251, -36.471, -37.843, -39.059, -40.304, -42.036, -43.609,
 -44.506, -45.320, -44.968, -43.563, -42.656, -42.506, -40.678, -41.105, -40.418, -40.808,
 -42.053, -42.360, -44.688, -45.649, -46.093, -45.950, -44.768, -43.147, -42.644, -42.289,
 -41.668, -41.191, -40.899, -40.380, -39.978, -39.806, -39.034, -37.186, -34.263, -32.193,
 -29.537, -27.261, -25.376, -23.151, -21.006, -18.803, -16.455, -13.628, -10.953,  -7.840,
  -4.504,  -0.629,   2.681,   4.627,   6.554,   8.714,  10.669,  11.805,  11.962,  13.106,
  14.964,  16.818,  18.555,  19.985,  20.791,  20.553,  20.394,  20.588,  21.807,  22.715,
  22.490,  20.872,  20.734,  21.642,  20.994,  19.219,  18.812,  18.778,  19.313,  20.009,
  20.277,  21.130,  21.200,  21.627,  22.214,  23.801,  23.850,  15.605,   5.351,   2.317,
   6.275,   8.420,   8.490,   8.125,   6.824,   5.469,   3.900,   3.185,   4.164,   1.223,
   2.614,   2.218,   1.926,   0.657,   0.869,   0.863,   0.536,  -0.964,  -2.989,  -4.982,  -5.705,
  -5.800,  -4.903,  -3.818,  -2.420,  -0.537,   1.488,   2.041,   3.335,   4.840,   5.494,
   6.550,   7.376,   8.040,   8.493,   8.610,   8.728,   8.430,   8.099,   7.644,   7.369,
   7.123,   6.753,   6.079,   5.144,   4.543,   3.566,   2.613,   1.671,   0.727,  -0.035,
  -0.905,  -2.018,  -2.900,  -4.178,  -5.263,  -6.512,  -7.674,  -8.939, -10.138, -11.536,
 -12.546, -13.872, -14.935, -15.914, -16.576, -17.248, -17.475, -17.865, -18.203, -18.042,
 -20.232, -19.588, -16.358, -14.891, -16.830, -17.479, -16.838, -16.618, -15.990, -16.043,
 -15.932, -16.183, -15.886, -15.436, -14.147, -15.554, -16.874, -17.534, -17.352, -17.374,
 -17.322, -17.412, -18.851, -19.427, -19.333, -19.600, -20.026, -21.447, -22.991, -23.789,
 -24.956, -26.678, -28.456, -29.196, -29.994, -31.177, -32.515, -33.968, -34.886, -35.009,
 -36.326, -37.890, -38.928, -39.358, -39.832, -39.938, -39.474, -38.876, -38.800, -39.502,
 -40.189, -40.086, -39.478, -37.730, -35.735, -34.540, -33.335, -31.388, -28.849, -26.294,
 -25.004, -23.567, -23.501, -22.518, -22.686, -20.580, -19.043, -17.905, -15.661, -12.755,
  -9.796,  -5.082,   0.644,   4.160,   8.089,  10.944,  12.907,  16.458,  22.083,  24.159,
  22.967,  23.476,  26.092,  29.242,  32.376,  35.272,  37.886,  40.031,  42.206,  44.505,
  47.054,  49.277,  51.136,  52.775,  54.096,  55.553,  57.255,  59.141,  60.574,  62.429,
  63.052,  64.125,  63.924,  63.659,  63.445,  62.968,  62.881,  62.686,  62.770,  61.655,
  61.107,  59.654,  58.217,  57.512,  56.608,  56.070,  57.154,  57.471,  59.089,  58.389,
  57.563,  56.090,  54.917,  53.820,  52.943,  52.053,  50.182,  48.049,  46.171,  45.038},
{
  45.301,  44.938,  44.441,  44.694,  45.111,  45.442,  45.700,  46.870,  48.385,  47.941,
  47.138,  46.464,  45.847,  44.492,  43.576,  42.758,  42.374,  40.745,  39.458,  38.109,
  36.729,  35.823,  34.133,  31.339,  29.524,  29.809,  29.968,  29.884,  27.660,  25.338,
  24.603,  24.750,  22.831,  20.435,  17.879,  16.039,  14.288,  12.876,  11.656,  11.073,
   9.907,   8.804,   7.130,   6.041,   4.705,   2.822,  -0.533,  -4.074,  -6.367,  -8.435,
 -10.247, -11.463, -12.846, -14.239, -15.323, -16.707, -18.095, -18.376, -16.428, -15.505,
 -18.272, -20.463, -22.740, -24.452, -26.237, -26.983, -27.052, -28.305, -28.706, -29.865,
 -31.199, -32.063, -33.383, -34.543, -35.772, -37.636, -39.136, -40.157, -40.733, -41.551,
 -43.018, -43.185, -42.498, -41.964, -41.078, -41.060, -41.344, -40.050, -40.233, -39.613,
 -40.554, -42.360, -43.057, -43.464, -43.402, -43.456, -43.311, -42.890, -41.523, -40.099,
 -39.569, -39.456, -38.447, -38.307, -37.811, -38.313, -38.607, -37.061, -34.455, -31.908,
 -29.477, -27.485, -25.353, -23.478, -21.306, -18.721, -16.307, -13.585, -11.018,  -8.267,
  -5.281,  -1.552,   2.052,   4.145,   5.990,   7.777,   9.166,  10.050,  10.721,  11.978,
  13.592,  15.810,  17.128,  18.499,  20.219,  20.227,  19.692,  19.417,  20.341,  21.420,
  21.417,  19.814,  19.210,  20.149,  18.447,  17.471,  17.495,  17.572,  18.127,  18.450,
  19.053,  19.485,  20.374,  21.637,  22.687,  22.672,  24.505,  25.220,  20.060,   5.983,
   0.739,   4.680,   7.407,   8.014,   7.965,   7.621,   6.234,   6.347,   5.582,   4.935,
   3.698,   2.537,   0.843,  -1.006,  -3.899,  -5.422,  -8.959, -10.263,  -8.360,  -5.295,  -7.689,
  -5.778,  -7.282,  -8.140,  -9.649, -10.810, -10.883,  -9.613,  -7.320,  -4.679,  -1.535,
   2.120,   5.491,   7.551,   8.771,   9.708,  10.151,  10.431,  10.464,  10.350,  10.006,
   9.551,   8.750,   7.904,   7.186,   6.698,   6.047,   5.093,   4.346,   3.352,   2.606,
   1.813,   0.904,  -0.183,  -1.180,  -2.512,  -3.298,  -4.737,  -6.336,  -7.549,  -8.839,
 -10.323, -11.449, -12.725, -13.686, -14.368, -14.774, -15.175, -15.709, -16.562, -17.337,
 -16.453, -15.445, -15.445, -15.066, -14.666, -13.969, -13.577, -14.798, -15.192, -15.836,
 -16.390, -15.961, -15.653, -14.152, -13.466, -16.846, -17.504, -18.714, -18.799, -19.052,
 -19.620, -19.969, -20.971, -21.045, -20.893, -21.153, -22.232, -23.045, -24.544, -26.567,
 -28.447, -29.648, -30.547, -31.114, -31.575, -32.609, -33.834, -34.881, -35.909, -36.989,
 -37.746, -39.315, -39.989, -39.920, -40.440, -41.382, -41.601, -41.188, -41.360, -42.133,
 -41.918, -40.673, -39.348, -37.314, -35.682, -34.293, -33.254, -32.064, -29.365, -25.969,
 -24.157, -22.325, -21.204, -19.865, -18.655, -17.678, -16.537, -15.208, -13.137, -10.279,
  -7.582,  -4.925,  -1.769,   3.050,   8.817,  11.332,  13.646,  16.362,  20.108,  23.632,
  24.226,  24.693,  27.164,  30.649,  33.501,  36.326,  38.631,  40.794,  42.989,  45.410,
  47.770,  50.036,  51.791,  53.298,  54.908,  56.573,  58.064,  59.386,  61.040,  62.212,
  63.343,  63.023,  62.959,  62.861,  62.916,  62.965,  63.056,  63.097,  62.174,  61.522,
  59.969,  58.738,  57.676,  57.073,  58.999,  60.145,  58.363,  58.554,  59.554,  59.033,
  57.783,  56.137,  54.802,  54.465,  53.962,  52.937,  50.644,  48.106,  46.144,  45.301},
{
  46.300,  45.371,  44.538,  44.059,  43.604,  43.418,  43.694,  44.141,  44.503,  44.853,
  44.395,  44.308,  43.060,  41.808,  40.854,  39.731,  38.764,  37.069,  34.870,  33.099,
  32.511,  32.028,  30.466,  29.090,  28.027,  27.672,  27.334,  26.140,  23.557,  21.069,
  21.752,  23.322,  21.188,  19.671,  18.218,  16.079,  14.175,  13.000,  11.891,  10.693,
   9.587,   8.175,   7.148,   6.509,   5.506,   3.504,   1.625,  -0.047,  -1.080,  -2.524,
  -4.185,  -5.724,  -7.175,  -8.235,  -9.317, -11.680, -14.282, -14.039, -12.503, -13.454,
 -16.251, -17.969, -20.747, -22.988, -24.566, -24.927, -24.892, -26.004, -27.101, -28.287,
 -29.108, -30.187, -31.389, -33.224, -34.380, -35.686, -37.316, -38.662, -39.688, -40.504,
 -40.720, -41.037, -41.857, -42.836, -43.667, -43.476, -42.585, -41.911, -40.903, -38.987,
 -38.854, -39.808, -40.947, -41.972, -42.699, -40.960, -40.312, -39.879, -38.895, -37.628,
 -36.815, -36.520, -37.888, -38.010, -39.147, -39.964, -39.319, -36.454, -34.272, -31.887,
 -29.605, -27.446, -24.747, -23.237, -21.005, -18.941, -16.329, -13.970, -11.603,  -8.716,
  -6.660,  -3.762,  -0.202,   2.388,   4.479,   5.975,   6.605,   7.989,   9.342,  10.833,
  12.346,  13.997,  15.884,  17.584,  18.817,  18.974,  18.183,  18.396,  19.489,  19.742,
  20.022,  19.283,  18.678,  18.152,  17.606,  16.397,  16.462,  17.054,  17.826,  18.834,
  19.310,  19.600,  20.151,  20.960,  22.057,  23.233,  23.835,  24.414,  23.858,  12.117,
   1.903,  -1.056,   2.916,   6.642,   8.441,   8.843,   8.846,   8.034,   5.861,   3.305,
   0.796,  -2.592,  -5.787,  -2.213,   1.225,   5.067,   5.864,   3.967,   5.723,   5.600,   3.969,
   2.409,   3.862,   5.068,   6.880,   7.631,   7.249,   6.050,   6.053,   3.450,   1.591,
  -2.052,  -2.751,  -0.422,   2.441,   5.775,   8.211,   9.685,  10.756,  11.341,  11.373,
  10.988,  10.519,  10.010,   9.689,   9.389,   8.819,   8.160,   7.059,   6.590,   5.694,
   4.946,   3.743,   2.331,   1.392,   0.476,  -0.858,  -2.326,  -3.992,  -5.319,  -6.701,
  -7.951,  -9.008, -10.230, -11.238, -11.873, -12.481, -13.423, -14.436, -13.264, -14.113,
 -14.187, -14.202, -14.134, -12.842, -11.708, -12.942, -13.603, -14.172, -14.303, -14.758,
 -14.992, -14.728, -13.114, -12.240, -15.499, -18.130, -19.459, -19.682, -20.001, -19.551,
 -20.016, -20.646, -21.550, -21.880, -22.298, -22.939, -23.963, -24.340, -26.815, -30.026,
 -30.901, -32.100, -32.731, -32.691, -32.776, -33.948, -35.237, -36.467, -37.433, -38.989,
 -40.649, -41.772, -42.642, -42.687, -42.420, -42.867, -43.230, -43.210, -43.292, -43.546,
 -43.200, -41.683, -40.346, -38.641, -36.308, -34.345, -32.621, -31.039, -29.721, -28.200,
 -25.939, -22.760, -20.139, -18.496, -17.349, -16.125, -14.713, -12.814, -10.046,  -7.756,
  -4.917,  -1.147,   1.828,   4.064,   8.859,  12.135,  14.559,  17.394,  20.634,  22.247,
  23.558,  26.311,  29.082,  31.850,  34.146,  36.484,  39.185,  41.686,  44.583,  47.059,
  48.685,  50.473,  52.309,  53.935,  55.547,  56.952,  58.304,  59.504,  61.078,  61.442,
  61.960,  61.893,  61.638,  61.999,  61.916,  62.264,  61.772,  61.579,  60.856,  60.935,
  58.997,  57.938,  57.492,  58.028,  60.508,  60.964,  60.239,  59.760,  59.412,  58.362,
  57.754,  56.475,  55.733,  54.518,  54.035,  53.850,  51.868,  49.163,  47.757,  46.300},
{
  46.367,  45.013,  44.131,  42.913,  42.290,  42.007,  41.611,  41.230,  41.177,  40.853,
  41.003,  40.701,  40.459,  39.402,  37.483,  35.775,  34.287,  33.332,  30.947,  29.817,
  30.627,  30.485,  29.055,  28.342,  27.903,  26.803,  25.826,  23.990,  22.102,  20.400,
  19.331,  18.844,  17.991,  17.348,  16.766,  15.905,  13.928,  12.385,  11.127,   9.959,
   9.005,   8.298,   7.582,   6.594,   5.210,   4.575,   3.778,   2.883,   1.481,   0.172,
  -1.011,  -2.094,  -2.767,  -4.022,  -5.915,  -8.583, -11.223, -11.642, -11.325, -11.799,
 -14.538, -16.586, -18.637, -20.403, -22.538, -23.666, -23.915, -24.672, -26.043, -27.317,
 -27.833, -28.770, -30.328, -32.085, -33.442, -35.140, -36.756, -37.776, -38.806, -39.259,
 -39.742, -40.239, -40.469, -40.519, -40.882, -41.629, -41.199, -40.131, -39.499, -39.231,
 -39.661, -40.864, -40.586, -40.442, -39.884, -39.702, -39.127, -38.027, -37.126, -36.342,
 -36.236, -36.964, -37.889, -37.865, -37.146, -35.881, -35.333, -35.153, -34.227, -31.236,
 -29.142, -26.737, -24.772, -22.815, -20.468, -18.704, -17.003, -14.886, -12.471, -10.098,
  -8.027,  -6.398,  -4.190,  -1.788,   0.194,   2.305,   4.505,   5.862,   7.556,   9.399,
  11.005,  12.712,  14.542,  16.534,  17.785,  18.480,  17.283,  17.538,  17.958,  18.478,
  18.815,  18.698,  18.052,  17.787,  17.307,  15.738,  15.938,  16.656,  17.759,  18.235,
  18.430,  18.940,  19.769,  20.570,  21.282,  22.708,  23.454,  24.091,  24.217,  20.759,
  16.923,   5.061,  -2.507,   1.175,   8.441,   9.603,   8.523,   5.468,   0.944,  -2.636,
  -0.118,   2.953,   4.610,   6.825,   4.712,   2.265,   1.843,   1.344,   1.391,   2.143,   4.946,
   1.649,   1.955,   2.385,   3.174,   3.260,   4.512,   6.463,   8.421,  10.150,  12.526,
  13.938,  13.531,   9.942,   3.704,   0.037,   0.443,   2.178,   4.362,   6.671,   8.912,
  10.478,  11.042,  11.369,  11.638,  11.480,  10.861,  10.085,   9.171,   8.267,   7.270,
   6.342,   5.745,   4.679,   3.477,   2.404,   1.197,  -0.226,  -1.680,  -3.152,  -4.532,
  -5.832,  -6.684,  -7.784,  -8.764,  -9.240, -11.234, -11.693,  -9.179, -10.747, -11.580,
 -12.180, -13.038, -12.212, -11.324, -11.557, -12.149, -13.256, -13.478, -13.797, -14.277,
 -14.530, -14.195, -14.657, -16.439, -17.815, -19.590, -20.773, -21.785, -21.919, -22.040,
 -22.022, -22.334, -23.324, -23.649, -24.486, -25.128, -26.425, -27.388, -28.034, -29.758,
 -31.546, -33.113, -34.039, -34.349, -35.284, -36.362, -37.637, -38.532, -39.667, -40.700,
 -42.038, -43.172, -44.288, -45.085, -45.164, -44.658, -44.752, -44.873, -44.712, -44.626,
 -43.884, -42.554, -40.910, -39.095, -36.758, -34.288, -32.285, -30.382, -28.846, -26.928,
 -24.921, -23.464, -22.185, -19.235, -17.608, -16.314, -15.086, -13.859, -10.465,  -6.961,
  -3.084,  -0.265,   3.632,   6.808,  10.225,  12.578,  15.401,  18.783,  20.752,  22.806,
  25.362,  28.283,  31.343,  33.923,  36.334,  38.845,  41.320,  43.523,  45.434,  47.417,
  49.661,  51.399,  53.627,  56.086,  57.641,  59.057,  59.484,  59.819,  59.990,  60.108,
  60.303,  60.750,  61.259,  61.497,  61.413,  60.816,  60.296,  60.137,  60.287,  60.326,
  60.085,  59.632,  59.165,  58.641,  60.171,  61.495,  60.425,  60.584,  59.949,  58.548,
  57.535,  56.868,  55.878,  55.760,  54.585,  54.571,  52.806,  50.801,  48.543,  46.367},
{
  45.821,  44.036,  42.692,  41.445,  41.365,  40.648,  39.777,  39.540,  39.194,  39.449,
  39.506,  39.290,  38.592,  37.155,  35.772,  34.742,  33.721,  31.646,  30.065,  29.154,
  29.276,  28.659,  28.280,  27.518,  26.798,  26.051,  25.177,  24.012,  22.557,  20.691,
  19.581,  18.303,  17.088,  16.634,  16.112,  15.788,  14.684,  12.854,  11.363,  10.236,
   9.265,   9.343,   8.039,   7.033,   6.794,   6.093,   4.961,   4.117,   2.989,   1.561,
   0.407,   0.022,  -1.154,  -2.938,  -5.363,  -7.543, -10.227, -10.766,  -9.983,  -9.605,
 -11.909, -14.147, -16.775, -18.135, -20.014, -22.007, -23.379, -24.347, -25.932, -26.649,
 -27.310, -28.492, -29.727, -31.372, -32.882, -34.188, -35.672, -36.608, -37.688, -38.169,
 -38.552, -38.729, -38.750, -38.900, -39.503, -39.424, -39.369, -39.924, -38.345, -36.030,
 -36.801, -38.991, -39.502, -39.019, -38.560, -37.850, -36.842, -35.719, -36.123, -37.437,
 -39.183, -39.718, -38.905, -37.834, -35.609, -33.971, -33.975, -32.557, -32.071, -32.877,
 -31.154, -27.403, -23.693, -22.350, -20.681, -18.992, -17.134, -14.955, -12.778, -10.904,
  -8.983,  -7.605,  -6.069,  -4.010,  -1.669,   0.996,   3.546,   5.183,   6.668,   8.004,
   9.797,  11.658,  13.104,  14.058,  15.531,  16.616,  16.292,  16.480,  16.957,  17.436,
  17.853,  16.926,  16.014,  17.345,  16.508,  15.315,  15.827,  16.739,  17.725,  18.146,
  18.782,  19.176,  19.685,  20.067,  20.494,  21.862,  22.146,  24.297,  25.086,  23.575,
  21.529,  14.453,   4.631,  -0.653,   3.757,   5.686,   2.494,   0.168,   2.095,   4.388,
   4.329,   3.422,   2.335,   1.955,   2.421,   3.245,   2.695,   1.946,   1.605,   3.985,   4.928,
  -0.115,   1.941,   2.764,   3.746,   4.270,   4.922,   5.859,   6.925,   8.316,   9.878,
  10.879,  12.593,  15.037,  17.563,  17.655,  16.218,  12.575,   8.328,   7.172,   7.743,
   7.889,   8.082,   8.359,   9.173,  10.347,  10.908,  10.916,  10.758,  10.467,  10.638,
  10.222,   8.538,   6.667,   5.246,   4.211,   3.032,   1.466,   0.168,  -1.196,  -2.145,
  -3.209,  -3.972,  -4.940,  -5.877,  -7.892,  -9.092,  -6.429,  -7.879,  -8.848,  -9.114,
 -10.313,  -9.654,  -9.539, -10.897, -11.803, -12.079, -12.684, -13.513, -13.850, -13.685,
 -14.617, -15.387, -16.322, -17.535, -18.646, -20.116, -21.657, -22.915, -23.481, -24.009,
 -24.617, -25.077, -25.902, -25.772, -26.265, -27.277, -29.191, -29.966, -30.458, -30.927,
 -32.017, -34.470, -35.882, -36.750, -37.706, -38.697, -39.484, -40.387, -41.558, -42.640,
 -43.831, -44.843, -45.612, -45.973, -45.702, -45.100, -44.623, -44.520, -44.396, -44.008,
 -43.370, -42.054, -40.909, -39.166, -37.005, -34.676, -32.326, -30.714, -28.638, -26.518,
 -24.391, -22.571, -21.085, -20.017, -18.441, -17.599, -16.179, -14.652, -12.864,  -9.152,
  -5.067,  -1.217,   2.021,   6.086,  10.630,  13.927,  15.644,  17.655,  21.098,  23.644,
  26.754,  29.673,  32.320,  35.358,  37.761,  39.977,  42.349,  44.330,  46.533,  48.843,
  51.146,  53.401,  55.395,  57.894,  59.369,  59.630,  59.928,  60.137,  60.021,  60.319,
  60.739,  60.683,  60.937,  61.030,  61.117,  60.801,  60.067,  60.248,  60.423,  61.042,
  60.719,  60.778,  59.713,  58.983,  58.015,  58.021,  58.500,  58.751,  59.358,  58.750,
  58.241,  57.519,  56.889,  56.262,  55.260,  53.858,  52.610,  51.154,  48.096,  45.821},
{
  46.184,  44.361,  43.024,  42.459,  41.790,  40.924,  40.626,  40.296,  40.626,  40.381,
  40.019,  38.628,  37.468,  35.793,  34.340,  33.733,  32.698,  31.384,  29.471,  27.570,
  26.520,  25.978,  25.430,  24.964,  24.514,  24.171,  23.413,  22.101,  21.479,  20.446,
  19.388,  18.547,  17.466,  16.804,  16.249,  15.684,  15.231,  15.040,  13.983,  12.013,
  11.094,  10.176,   8.568,   7.234,   6.460,   6.429,   5.718,   5.138,   4.080,   2.505,
   1.604,   0.655,  -0.263,  -1.967,  -3.888,  -6.179,  -8.432,  -9.680,  -9.461,  -8.542,
  -9.512, -12.248, -14.148, -16.641, -18.037, -19.885, -21.625, -22.553, -23.650, -24.799,
 -26.277, -27.369, -28.826, -30.415, -31.903, -33.423, -34.104, -34.761, -35.316, -36.062,
 -36.263, -35.814, -36.092, -36.802, -36.790, -36.967, -37.657, -36.759, -35.433, -36.045,
 -37.582, -38.118, -38.104, -37.573, -37.156, -37.867, -38.200, -37.709, -38.064, -39.628,
 -40.230, -39.700, -38.653, -36.836, -35.260, -34.486, -32.938, -32.474, -31.568, -32.303,
 -29.784, -26.304, -24.748, -22.629, -20.437, -18.471, -16.588, -14.707, -12.649, -10.920,
  -9.359,  -8.208,  -6.074,  -4.229,  -1.775,   0.330,   2.140,   3.816,   5.382,   7.207,
   9.198,  11.801,  13.459,  14.126,  14.257,  14.537,  14.666,  15.210,  15.661,  15.605,
  15.715,  15.225,  15.764,  16.463,  15.917,  15.957,  16.945,  17.706,  18.109,  18.838,
  19.310,  19.600,  19.823,  19.623,  19.502,  20.932,  21.212,  24.025,  24.668,  22.948,
  22.356,  20.307,  14.103,   0.673,  -0.835,   0.615,   8.276,  13.546,   8.902,   6.560,
   4.859,   2.475,   2.298,   1.671,  -0.469,  -2.174,   0.814,   4.041,   3.535,   1.729,  -0.694,
   1.031,   2.232,   3.348,   3.998,   4.870,   5.264,   6.310,   6.495,   7.410,   9.224,
  10.333,  12.166,  13.905,  14.269,  15.527,  16.533,  16.692,  17.346,  17.659,  17.103,
  15.502,  14.369,  13.154,  11.724,   9.941,   9.138,  10.252,  11.519,  11.675,  11.046,
   9.484,   8.115,   7.364,   6.637,   5.556,   4.528,   3.382,   2.194,   1.405,   0.394,
  -0.756,  -1.924,  -3.251,  -5.457,  -6.645,  -3.841,  -3.643,  -4.773,  -6.000,  -7.051,
  -7.716,  -8.423,  -8.731,  -9.300, -10.222, -10.625, -11.742, -12.492, -13.682, -15.389,
 -16.466, -16.781, -17.275, -17.898, -19.003, -20.482, -21.762, -23.365, -24.408, -25.146,
 -26.025, -26.968, -27.670, -28.199, -28.848, -29.489, -30.547, -31.153, -31.970, -32.798,
 -34.070, -35.168, -37.053, -38.757, -39.796, -40.661, -41.639, -42.395, -43.388, -44.601,
 -45.577, -46.053, -46.543, -47.037, -46.400, -45.329, -44.757, -44.523, -44.379, -44.446,
 -44.282, -42.928, -40.947, -38.613, -36.608, -34.506, -32.525, -30.932, -28.628, -26.231,
 -24.467, -22.792, -20.500, -17.478, -16.154, -14.798, -13.377, -12.245, -10.759,  -8.263,
  -4.465,  -0.997,   3.325,   7.922,  12.506,  15.054,  15.977,  18.422,  21.899,  25.065,
  27.942,  30.635,  33.849,  36.569,  38.770,  40.858,  43.068,  45.167,  47.361,  49.575,
  51.821,  53.988,  56.048,  58.232,  59.741,  60.226,  60.193,  60.420,  60.357,  60.221,
  60.473,  60.931,  60.671,  61.041,  61.301,  60.759,  60.030,  60.196,  59.763,  60.830,
  61.696,  61.499,  60.204,  58.459,  57.566,  57.153,  56.336,  56.252,  56.139,  57.907,
  58.392,  57.588,  57.227,  55.766,  55.094,  54.424,  53.029,  50.926,  48.258,  46.184},
{
  46.676,  45.252,  44.565,  43.830,  43.021,  42.127,  41.818,  41.312,  40.872,  40.477,
  39.343,  37.660,  36.597,  36.318,  34.841,  33.053,  31.814,  29.869,  27.717,  25.903,
  24.840,  24.584,  24.335,  23.441,  22.697,  21.834,  21.484,  21.045,  19.944,  19.291,
  18.506,  17.406,  17.359,  17.422,  16.830,  15.711,  15.383,  15.499,  14.267,  12.731,
  11.565,  10.350,   9.100,   8.308,   7.874,   6.795,   6.420,   6.191,   5.094,   3.908,
   2.607,   0.966,  -0.575,  -2.366,  -4.614,  -6.014,  -6.626,  -6.728,  -6.984,  -7.663,
  -7.724, -10.240, -13.267, -15.605, -17.565, -18.647, -20.439, -21.656, -22.806, -24.007,
 -25.238, -26.546, -28.256, -29.900, -30.729, -31.346, -32.492, -33.516, -34.049, -34.594,
 -34.834, -35.109, -35.492, -35.790, -36.134, -36.244, -35.772, -35.993, -37.458, -38.070,
 -38.184, -38.164, -37.862, -39.019, -39.191, -39.455, -39.785, -40.547, -41.068, -40.661,
 -40.260, -38.960, -38.136, -36.582, -35.646, -34.278, -33.510, -32.559, -32.185, -28.789,
 -28.000, -26.965, -25.769, -23.623, -21.774, -19.603, -17.816, -14.554, -11.390,  -9.690,
  -8.871,  -7.984,  -6.706,  -4.470,  -1.881,  -0.224,   1.701,   3.799,   5.965,   8.200,
  10.469,  11.295,  12.351,  13.029,  13.484,  14.082,  14.574,  15.050,  14.953,  14.627,
  14.179,  14.408,  14.803,  15.146,  16.248,  17.383,  17.620,  17.961,  18.466,  18.934,
  19.081,  18.824,  18.395,  17.846,  18.303,  19.962,  20.817,  22.091,  23.983,  22.631,
  20.965,  19.388,  14.594,  10.701,   6.602,   8.388,   9.380,   8.330,   8.103,   6.662,
   6.019,   4.252,   2.437,   2.768,   1.426,   0.517,  -0.039,  -0.496,  -0.406,   0.557,   1.682,
   1.959,   2.868,   2.843,   2.967,   3.153,   3.368,   3.951,   4.728,   7.620,   9.670,
  11.579,  12.796,  12.973,  13.692,  14.288,  14.706,  15.000,  15.273,  15.663,  16.934,
  16.344,  14.376,  13.069,  12.440,  13.289,  13.996,  10.837,   6.852,   6.230,   7.422,
   8.335,   8.626,   8.567,   8.218,   7.509,   6.381,   5.587,   4.601,   2.911,   1.666,
   0.532,  -0.624,  -2.282,  -3.934,  -0.681,  -0.316,  -1.286,  -2.278,  -2.171,  -3.247,
  -5.544,  -6.743,  -6.902,  -7.693,  -8.368,  -9.611, -10.875, -12.667, -13.901, -14.847,
 -15.941, -17.434, -18.365, -19.311, -20.045, -20.974, -22.504, -23.670, -25.255, -27.033,
 -28.251, -28.957, -29.095, -29.131, -29.732, -30.657, -31.326, -32.145, -33.354, -34.448,
 -35.159, -37.031, -38.359, -39.719, -41.022, -41.993, -43.007, -44.180, -45.100, -46.384,
 -47.537, -47.382, -47.504, -47.628, -47.718, -47.442, -46.769, -46.467, -45.182, -44.033,
 -43.767, -42.590, -40.881, -38.899, -36.959, -35.086, -33.351, -31.339, -29.394, -27.474,
 -26.361, -24.551, -20.808, -16.955, -14.742, -12.521, -10.872,  -9.960,  -9.082,  -6.280,
  -0.241,   5.314,   6.336,   7.859,  11.241,  14.218,  16.970,  19.536,  22.735,  25.718,
  28.379,  31.088,  33.897,  36.589,  39.091,  41.317,  43.357,  45.417,  47.494,  49.456,
  51.504,  53.801,  55.906,  58.254,  60.099,  60.982,  61.036,  61.188,  60.872,  60.668,
  60.719,  60.951,  61.043,  60.871,  61.162,  61.907,  62.766,  62.989,  61.099,  61.092,
  61.241,  61.347,  61.978,  61.588,  60.127,  57.624,  56.572,  55.242,  55.780,  56.814,
  57.779,  58.138,  57.118,  55.809,  55.036,  54.047,  52.204,  49.989,  48.075,  46.676},
{
  47.262,  46.356,  45.533,  44.198,  43.171,  42.530,  41.680,  40.885,  40.445,  39.426,
  38.490,  37.940,  37.256,  36.169,  33.912,  32.420,  30.485,  28.159,  26.519,  24.540,
  22.961,  22.128,  21.951,  21.279,  20.439,  20.706,  21.033,  19.963,  18.801,  18.589,
  17.644,  16.319,  16.038,  16.648,  16.091,  14.434,  13.293,  13.424,  13.315,  12.865,
  12.105,  10.570,   9.204,   8.771,   8.354,   8.618,   7.852,   6.977,   6.417,   5.132,
   3.753,   2.106,   0.476,  -1.420,  -3.353,  -4.664,  -5.014,  -5.060,  -5.842,  -6.407,
  -7.041,  -9.754, -12.385, -14.352, -16.486, -17.811, -19.234, -20.969, -22.212, -23.497,
 -24.844, -26.358, -27.857, -28.871, -29.430, -30.046, -30.761, -32.084, -33.379, -33.879,
 -33.986, -34.392, -34.998, -35.429, -35.895, -36.311, -36.106, -36.886, -37.855, -38.160,
 -38.174, -38.507, -38.611, -38.758, -38.525, -40.243, -40.523, -40.391, -40.278, -40.180,
 -39.020, -37.686, -36.570, -35.920, -35.327, -34.821, -34.416, -33.611, -33.233, -29.788,
 -26.610, -24.432, -23.534, -22.794, -21.345, -19.053, -17.112, -14.580, -12.986, -11.964,
 -10.156,  -9.757,  -7.771,  -4.980,  -3.791,  -1.565,   1.192,   3.866,   5.554,   7.472,
   8.861,   9.589,  11.086,  11.492,  11.575,  12.421,  13.529,  14.891,  15.060,  14.550,
  14.292,  14.215,  13.802,  14.853,  14.998,  15.792,  17.037,  17.228,  17.515,  17.702,
  17.595,  17.249,  16.526,  16.490,  17.045,  18.131,  19.119,  20.424,  20.896,  20.763,
  20.258,  19.382,  17.662,  14.537,  10.842,  10.482,  10.218,   9.223,   7.725,   6.519,
   6.221,   5.379,   3.199,   2.335,   1.996,   1.560,   0.857,   0.922,   1.044,   1.524,   1.847,
   2.134,   2.160,   2.544,   2.832,   2.315,   5.277,   9.853,  11.218,  11.356,  11.759,
  11.974,  12.168,  12.983,  13.435,  14.052,  14.568,  14.895,  15.056,  14.851,  14.329,
  14.563,  15.534,  15.050,  12.031,  11.683,  13.325,  14.446,  15.325,  14.007,  10.916,
   8.310,   7.274,   7.862,   8.120,   7.920,   7.169,   6.366,   5.158,   3.777,   2.619,
   1.251,   0.341,  -0.879,   0.842,   2.076,   0.845,   0.884,   1.149,   0.969,  -1.551,
  -2.640,  -3.919,  -4.635,  -5.818,  -7.653,  -8.368, -10.262, -12.124, -13.472, -14.363,
 -15.405, -17.422, -19.255, -20.137, -21.713, -22.074, -23.237, -25.098, -26.358, -27.649,
 -29.041, -29.458, -29.808, -29.886, -31.051, -32.277, -33.015, -33.662, -34.680, -36.138,
 -37.836, -39.391, -40.655, -41.925, -43.092, -44.927, -46.110, -46.695, -47.127, -47.597,
 -48.074, -48.518, -48.525, -48.198, -47.984, -47.819, -47.546, -46.862, -44.838, -43.698,
 -43.764, -42.219, -40.562, -38.550, -36.713, -34.850, -32.920, -31.026, -29.193, -28.072,
 -26.582, -23.426, -19.618, -16.432, -13.340, -10.518,  -9.945,  -8.975,  -6.818,  -1.238,
   4.007,   5.630,   7.864,  10.619,  13.097,  15.729,  18.433,  20.824,  23.510,  26.526,
  29.298,  31.974,  34.025,  36.682,  39.341,  41.512,  43.781,  45.545,  47.220,  48.909,
  50.971,  53.029,  55.312,  57.264,  59.512,  60.970,  61.767,  61.937,  61.903,  61.754,
  61.349,  61.354,  61.161,  61.132,  61.053,  61.060,  61.051,  61.564,  61.752,  61.331,
  60.303,  59.795,  60.402,  59.584,  60.360,  59.600,  56.516,  55.440,  55.871,  55.808,
  57.689,  58.110,  57.029,  55.603,  54.925,  53.723,  52.302,  50.065,  48.496,  47.262},
{
  47.762,  47.067,  45.788,  44.724,  43.635,  42.778,  42.158,  41.829,  40.740,  39.184,
  38.129,  36.775,  35.538,  34.405,  32.508,  31.152,  29.280,  27.248,  25.097,  22.483,
  20.671,  20.053,  19.912,  19.576,  19.280,  19.368,  19.674,  19.246,  17.963,  16.656,
  15.546,  15.324,  15.533,  15.349,  15.472,  15.331,  14.440,  14.186,  13.733,  12.780,
  11.935,  11.486,  10.942,  10.080,   9.332,   9.047,   8.803,   7.872,   7.115,   5.898,
   5.131,   3.775,   2.344,   0.799,  -0.626,  -1.685,  -2.511,  -3.363,  -4.196,  -4.808,
  -4.864,  -8.383, -11.377, -13.345, -14.860, -16.484, -17.962, -19.876, -21.363, -22.432,
 -23.780, -24.957, -25.765, -27.091, -28.427, -29.240, -29.887, -30.423, -31.605, -32.696,
 -33.395, -33.788, -34.415, -35.137, -35.737, -35.890, -35.460, -35.274, -36.073, -36.532,
 -37.227, -38.129, -37.879, -36.719, -36.408, -37.775, -38.647, -39.579, -39.577, -39.540,
 -38.734, -37.335, -36.243, -34.815, -33.530, -33.651, -33.639, -33.421, -32.818, -31.615,
 -29.016, -26.560, -24.095, -21.607, -20.345, -19.112, -17.010, -15.549, -15.905, -14.099,
 -12.612, -10.943,  -8.843,  -6.673,  -4.830,  -1.660,   0.828,   2.355,   3.660,   5.732,
   7.126,   8.401,   9.827,  10.009,  10.373,  10.822,  11.317,  12.413,  13.798,  14.877,
  14.255,  13.284,  13.125,  13.287,  13.491,  14.006,  14.855,  15.354,  15.661,  15.996,
  16.517,  16.724,  16.309,  16.223,  16.835,  16.423,  17.085,  18.566,  18.698,  18.028,
  18.272,  18.755,  16.830,  15.222,  12.567,  10.274,   9.621,   8.766,   7.294,   6.135,
   6.580,   5.987,   3.004,   1.500,   1.704,   1.203,   1.008,   0.627,   1.005,   0.921,   1.550,
   0.858,   1.159,   1.696,   4.008,   4.236,   7.250,   9.223,   9.458,  10.182,  10.607,
  10.890,  11.525,  12.043,  12.896,  13.610,  14.098,  14.500,  15.326,  15.427,  15.425,
  15.362,  15.120,  15.169,  15.585,  14.507,  12.883,  12.511,  11.716,  12.612,  13.893,
  13.205,  10.149,   6.307,   5.995,   7.095,   7.053,   6.506,   5.268,   3.902,   2.822,
   3.437,   2.471,   2.973,   4.055,   2.858,   1.803,   2.626,   2.433,   1.087,  -0.004,
  -0.998,  -2.220,  -3.910,  -5.256,  -6.283,  -8.867, -11.763, -13.257, -14.649, -15.115,
 -16.919, -18.799, -20.362, -21.205, -22.898, -23.835, -24.472, -26.290, -28.239, -29.170,
 -30.096, -30.518, -31.628, -32.336, -32.721, -33.749, -34.869, -35.755, -36.573, -38.130,
 -39.821, -41.312, -42.590, -44.007, -45.173, -46.190, -47.116, -47.732, -48.057, -48.196,
 -48.677, -48.878, -48.416, -48.150, -47.601, -47.117, -46.815, -46.180, -44.693, -43.768,
 -42.435, -40.645, -38.396, -36.573, -34.855, -33.317, -31.722, -29.708, -28.362, -28.163,
 -25.796, -22.284, -19.305, -15.759, -11.956,  -9.121,  -7.780,  -6.982,  -2.774,   3.266,
   4.717,   6.309,   8.589,  11.212,  13.919,  16.775,  19.177,  22.085,  24.604,  26.942,
  29.659,  32.483,  35.043,  37.666,  40.221,  41.647,  43.428,  45.326,  47.107,  48.768,
  50.542,  52.439,  54.377,  56.270,  58.174,  60.270,  61.842,  62.732,  63.030,  62.778,
  62.698,  62.547,  62.083,  61.698,  61.335,  61.012,  60.970,  60.994,  60.906,  61.393,
  61.142,  60.699,  59.591,  59.527,  59.984,  59.558,  57.092,  55.790,  55.638,  55.691,
  58.153,  57.328,  57.129,  55.866,  55.201,  52.786,  50.256,  49.024,  48.596,  47.762},
{
  46.927,  46.723,  46.040,  44.971,  44.465,  44.391,  44.184,  45.263,  43.606,  41.661,
  40.336,  37.784,  35.119,  32.938,  30.912,  29.320,  27.720,  25.776,  23.652,  21.339,
  19.467,  17.998,  18.877,  20.005,  19.810,  18.921,  18.067,  17.422,  16.702,  15.875,
  15.179,  14.825,  15.194,  16.471,  17.211,  16.207,  15.253,  14.496,  13.774,  13.586,
  13.042,  12.895,  12.697,  11.206,  10.071,   9.106,   8.516,   7.750,   7.276,   6.738,
   5.689,   4.616,   3.552,   2.227,   1.190,   0.123,  -1.198,  -2.428,  -2.954,  -3.040,
  -4.763,  -8.503, -10.372, -12.226, -14.043, -15.438, -16.568, -18.011, -19.533, -20.797,
 -22.254, -23.659, -25.117, -26.625, -27.600, -28.424, -29.185, -29.733, -30.890, -31.343,
 -32.186, -32.720, -33.045, -33.909, -34.184, -34.314, -34.527, -34.501, -35.011, -35.768,
 -36.827, -36.471, -35.182, -34.103, -33.509, -33.687, -34.422, -34.912, -35.505, -36.080,
 -36.181, -36.033, -35.511, -33.967, -32.351, -31.778, -31.788, -31.211, -30.589, -29.651,
 -28.546, -27.871, -27.370, -25.430, -22.665, -20.308, -19.101, -18.788, -18.664, -17.819,
 -15.910, -13.936, -11.199,  -8.112,  -6.330,  -4.767,  -3.184,  -1.271,   0.260,   1.361,
   2.176,   3.781,   5.640,   7.119,   8.049,   8.506,   9.676,  10.883,  11.811,  12.973,
  13.422,  13.223,  13.007,  12.544,  12.839,  13.264,  13.892,  14.278,  15.015,  16.155,
  17.164,  17.571,  17.339,  16.875,  16.330,  16.399,  16.167,  15.889,  15.969,  16.520,
  16.797,  17.208,  16.276,  14.494,  14.117,  11.994,   9.568,   7.460,   6.147,   5.555,
   6.761,   5.114,   1.245,   0.159,   0.145,   0.034,  -0.084,   0.029,   0.110,   0.510,   0.625,
   0.796,   4.494,   6.029,   6.361,   6.669,   7.675,   8.416,   9.036,   9.477,  10.141,
  10.752,  11.455,  11.807,  12.495,  12.832,  13.422,  13.942,  14.884,  15.139,  15.360,
  15.355,  15.000,  15.285,  15.486,  15.333,  13.863,  12.085,  10.903,  10.205,  11.483,
  12.870,  12.468,  10.957,   8.420,   6.802,   6.382,   6.242,   7.294,   8.135,   7.431,
   5.090,   7.147,   9.252,   5.598,   5.409,   5.203,   3.267,   2.009,   1.604,   0.354,
  -1.008,  -2.558,  -3.956,  -5.410,  -8.448, -10.750, -12.353, -13.373, -14.670, -16.836,
 -17.994, -19.039, -19.627, -20.083, -21.228, -24.222, -26.877, -28.679, -29.928, -31.059,
 -32.177, -32.925, -33.303, -33.960, -34.799, -35.279, -36.286, -37.993, -39.504, -40.677,
 -41.855, -43.328, -44.810, -46.379, -47.777, -48.897, -49.186, -48.952, -48.637, -48.295,
 -48.006, -47.870, -47.814, -47.611, -46.566, -45.798, -45.612, -45.119, -43.809, -44.006,
 -41.684, -38.629, -36.101, -34.416, -32.752, -31.106, -29.375, -27.862, -27.010, -26.465,
 -24.197, -21.030, -17.913, -14.825, -11.016,  -7.179,  -4.723,  -1.421,   2.748,   4.905,
   5.275,   7.277,   9.926,  12.553,  14.631,  17.526,  20.583,  23.143,  25.271,  27.951,
  30.561,  33.577,  36.324,  37.750,  39.729,  43.277,  45.357,  46.365,  47.392,  48.694,
  50.078,  51.534,  53.246,  55.244,  57.125,  58.919,  60.832,  62.465,  63.331,  63.859,
  63.684,  63.435,  63.181,  62.648,  62.123,  61.525,  60.947,  60.777,  60.664,  60.282,
  60.681,  60.666,  59.858,  58.898,  58.811,  58.248,  56.389,  55.703,  55.446,  56.206,
  55.312,  55.394,  56.501,  55.893,  54.540,  53.645,  52.465,  50.518,  48.469,  46.927},
{
  48.227,  47.361,  46.826,  45.367,  44.783,  44.966,  44.452,  45.307,  44.699,  41.977,
  40.457,  38.595,  35.762,  32.610,  30.724,  29.331,  27.167,  25.200,  22.891,  20.260,
  18.615,  18.779,  19.383,  19.513,  19.017,  17.680,  16.416,  15.598,  15.289,  15.408,
  15.988,  15.931,  15.795,  15.834,  16.123,  15.537,  13.752,  11.790,  11.391,  11.745,
  11.428,  11.292,  11.232,  11.163,  10.977,  10.002,   8.474,   7.338,   6.953,   6.847,
   6.032,   4.859,   3.816,   3.146,   1.587,   0.349,  -1.093,  -2.479,  -3.055,  -2.634,
  -3.866,  -8.109, -10.246, -10.953, -12.961, -14.554, -15.924, -17.298, -18.225, -19.766,
 -21.313, -22.815, -23.718, -24.769, -25.861, -26.357, -27.510, -28.435, -28.780, -30.026,
 -30.595, -30.966, -31.535, -31.510, -31.569, -31.724, -32.116, -32.344, -33.008, -34.103,
 -33.911, -32.835, -32.147, -31.216, -31.408, -33.081, -34.526, -34.626, -34.429, -34.176,
 -34.245, -34.067, -33.495, -32.914, -32.061, -31.123, -31.094, -30.532, -29.414, -28.314,
 -27.674, -27.536, -27.699, -27.821, -27.412, -25.678, -22.986, -20.236, -19.024, -18.230,
 -16.614, -14.617, -11.683,  -9.531,  -7.718,  -5.769,  -4.113,  -3.158,  -2.470,  -1.077,
   0.589,   2.438,   3.737,   4.494,   5.147,   5.295,   6.307,   8.593,  10.072,  11.272,
  12.185,  12.581,  12.383,  12.561,  13.697,  13.741,  13.994,  14.142,  14.449,  15.302,
  16.206,  16.689,  16.828,  16.739,  16.598,  16.030,  14.926,  14.373,  14.069,  13.743,
  14.436,  15.493,  14.387,  13.070,  13.201,  12.543,  11.674,  10.584,   8.665,   7.661,
   8.921,   4.271,  -0.910,  -0.924,  -1.270,  -1.884,  -1.841,  -1.018,  -0.331,   0.052,   0.445,
   2.698,   4.522,   4.553,   5.089,   6.177,   7.217,   7.289,   8.164,   8.747,   9.151,
   9.732,  10.493,  11.396,  11.787,  11.887,  12.096,  12.883,  13.431,  14.513,  15.195,
  14.803,  14.665,  14.991,  15.041,  14.796,  14.385,  12.576,   9.219,  10.719,  13.464,
  13.349,  13.363,  12.437,  11.930,  11.622,  11.053,  10.267,  10.502,  10.449,  10.503,
  11.501,  11.007,   9.677,   8.271,   6.667,   5.327,   4.188,   2.844,   1.941,   0.454,
  -0.984,  -2.552,  -3.825,  -4.676,  -6.776, -10.322, -11.941, -12.413, -13.882, -15.902,
 -17.112, -18.336, -20.388, -22.357, -24.236, -26.186, -28.055, -29.440, -30.362, -31.046,
 -31.838, -32.631, -33.966, -35.114, -35.865, -37.152, -38.846, -40.292, -41.093, -42.152,
 -42.899, -44.201, -45.593, -46.765, -47.859, -49.050, -49.501, -49.154, -48.150, -47.040,
 -46.226, -46.540, -46.981, -46.755, -45.479, -44.088, -43.240, -42.633, -42.291, -42.354,
 -40.768, -37.853, -34.574, -31.920, -29.931, -28.113, -26.546, -25.334, -24.294, -23.101,
 -21.465, -19.486, -16.880, -13.087,  -8.422,  -4.864,  -1.720,   1.241,   3.512,   4.817,
   6.908,   9.256,  11.257,  13.844,  16.514,  19.083,  21.683,  24.357,  26.784,  28.522,
  30.422,  32.792,  35.964,  39.190,  41.214,  43.654,  45.497,  47.544,  48.009,  48.765,
  49.599,  50.807,  52.332,  54.099,  55.812,  57.795,  59.594,  61.204,  62.788,  63.799,
  64.373,  64.306,  64.085,  63.681,  63.264,  62.736,  62.116,  61.004,  60.629,  60.658,
  60.375,  59.862,  59.714,  59.741,  59.927,  59.581,  58.468,  57.691,  56.596,  55.457,
  55.575,  55.616,  55.417,  55.372,  54.208,  53.724,  52.911,  50.981,  49.185,  48.227},
{
  49.454,  47.685,  47.043,  46.038,  45.372,  45.949,  45.957,  45.594,  45.209,  43.124,
  40.397,  38.163,  35.574,  33.576,  32.050,  30.658,  28.865,  25.821,  23.220,  21.474,
  20.337,  19.478,  18.744,  18.869,  19.044,  18.266,  16.530,  15.477,  15.465,  15.954,
  16.453,  16.462,  15.672,  15.072,  15.153,  14.036,  12.637,  11.657,  12.143,  12.268,
  11.482,  11.377,  11.146,  10.676,   9.788,   9.369,   9.142,   8.112,   7.458,   6.956,
   6.178,   5.153,   3.470,   2.599,   1.217,  -0.458,  -2.151,  -3.091,  -3.795,  -2.942,
  -3.351,  -6.376,  -8.773, -10.205, -11.016, -13.161, -15.035, -16.775, -18.349, -19.357,
 -20.278, -21.285, -22.641, -23.595, -23.776, -24.646, -26.411, -27.339, -27.079, -27.823,
 -28.865, -29.361, -29.383, -29.474, -29.458, -29.196, -29.373, -29.252, -30.168, -30.880,
 -30.248, -29.725, -29.991, -30.416, -31.212, -31.723, -32.703, -34.114, -34.202, -33.623,
 -32.926, -32.663, -32.365, -31.381, -30.524, -29.963, -29.896, -29.793, -29.041, -28.496,
 -27.265, -26.231, -26.020, -25.798, -26.264, -25.675, -24.061, -21.408, -19.373, -18.905,
 -17.266, -14.656, -12.701, -11.397, -10.008,  -8.239,  -6.690,  -5.788,  -4.792,  -3.263,
  -1.486,   0.060,   1.537,   2.849,   3.464,   3.766,   4.973,   7.689,   9.410,  10.521,
  11.204,  11.702,  12.539,  12.656,  13.424,  13.581,  13.605,  14.024,  14.532,  14.921,
  15.408,  15.637,  16.014,  16.317,  16.552,  16.190,  14.823,  13.774,  12.993,  13.193,
  13.652,  13.891,  13.644,  12.168,  11.329,  11.887,  11.637,  11.549,  11.278,  10.255,
  10.007,   9.902,   8.233,   5.243,   3.403,   1.122,  -0.575,  -1.525,  -0.093,   3.537,   4.645,
   4.167,   3.681,   3.821,   4.906,   6.104,   6.442,   6.905,   7.193,   7.391,   7.938,
   8.451,   9.223,   9.913,  10.516,  10.994,  11.329,  11.781,  12.261,  12.799,  13.678,
  13.750,  13.656,  13.863,  13.979,  14.575,  14.659,  14.803,  10.679,   5.871,   8.159,
  11.759,  13.557,  15.509,  17.075,  17.775,  17.659,  16.057,  14.875,  16.592,  16.253,
  13.397,  11.571,   9.936,   8.339,   7.087,   5.988,   4.839,   3.761,   2.813,   1.384,
  -0.340,  -0.942,  -1.849,  -3.477,  -4.837,  -8.360, -11.292, -12.301, -13.780, -15.826,
 -18.229, -20.339, -22.436, -24.373, -26.227, -27.857, -29.239, -30.099, -30.959, -31.667,
 -32.281, -33.051, -33.982, -35.235, -36.937, -38.416, -39.423, -40.796, -41.730, -42.800,
 -43.942, -45.064, -46.482, -47.628, -48.272, -48.825, -49.048, -49.025, -48.183, -47.059,
 -46.288, -45.840, -45.667, -45.258, -44.299, -43.133, -41.695, -40.230, -38.486, -37.456,
 -37.052, -35.179, -32.870, -30.756, -28.520, -26.520, -24.772, -23.865, -23.018, -21.837,
 -21.136, -20.236, -17.851, -13.422,  -7.775,  -2.945,  -0.118,   2.905,   6.180,   7.709,
   9.235,  10.948,  13.325,  15.825,  18.568,  20.742,  23.070,  25.318,  27.169,  29.294,
  30.952,  33.424,  35.925,  37.863,  42.487,  46.079,  45.544,  48.108,  48.978,  48.657,
  49.709,  50.758,  52.192,  53.335,  54.796,  56.319,  58.308,  59.914,  61.512,  62.916,
  63.958,  64.668,  64.842,  64.648,  64.173,  63.895,  63.291,  62.714,  62.002,  61.386,
  60.804,  60.486,  60.074,  59.725,  59.590,  59.457,  59.579,  59.157,  58.759,  58.202,
  58.142,  56.954,  56.582,  55.350,  53.215,  51.715,  50.656,  51.020,  50.555,  49.454},
{
  50.172,  48.633,  47.789,  47.165,  45.910,  45.463,  45.242,  45.572,  46.153,  44.902,
  41.717,  38.809,  37.018,  35.295,  33.962,  32.361,  30.000,  26.646,  23.422,  22.235,
  21.007,  19.376,  18.814,  18.589,  18.311,  18.450,  18.547,  18.502,  18.459,  18.097,
  17.598,  17.312,  16.828,  15.970,  14.830,  13.858,  13.963,  13.703,  13.187,  12.332,
  11.463,  11.614,  11.129,  10.939,  10.435,   9.264,   8.839,   8.392,   7.757,   6.612,
   5.519,   3.714,   1.886,   0.766,  -0.365,  -0.786,  -1.360,  -2.284,  -3.499,  -2.669,
  -3.062,  -6.149,  -8.677, -10.157, -10.621, -11.395, -13.040, -15.481, -16.786, -17.749,
 -19.060, -19.989, -21.209, -22.263, -22.341, -22.791, -23.820, -24.810, -25.154, -25.472,
 -26.460, -26.996, -27.362, -27.648, -27.473, -27.193, -27.108, -26.824, -27.062, -27.302,
 -27.699, -28.572, -28.934, -29.115, -29.247, -29.256, -29.653, -30.492, -31.112, -31.170,
 -31.189, -31.171, -30.621, -30.238, -29.799, -29.484, -29.213, -29.087, -28.990, -27.869,
 -26.561, -25.328, -24.612, -24.417, -24.151, -23.446, -21.780, -19.634, -19.420, -21.728,
 -21.518, -17.633, -14.181, -12.461, -11.581, -10.835, -10.100,  -9.007,  -7.586,  -6.124,
  -4.620,  -2.727,  -1.126,  -0.111,   0.589,   1.190,   2.216,   4.471,   7.059,   8.351,
   9.443,  10.625,  11.531,  12.119,  12.246,  12.285,  12.497,  12.585,  12.938,  13.570,
  14.059,  14.215,  14.389,  14.033,  13.956,  14.672,  14.479,  13.624,  12.953,  12.618,
  12.705,  13.234,  12.968,  11.723,  11.001,  11.038,  10.664,  10.132,  10.334,  10.908,
  11.187,  11.027,   9.539,   7.722,   6.123,   3.903,   2.486,   3.025,   5.297,   5.337,   4.519,
   4.562,   4.289,   4.718,   5.023,   5.595,   6.047,   6.161,   6.257,   6.412,   6.530,
   6.910,   7.745,   8.485,   9.000,   9.666,  10.394,  10.852,  11.345,  11.740,  11.882,
  12.085,  12.202,  11.863,  12.308,  13.820,  15.555,  14.936,  11.894,   9.594,  11.089,
  15.668,  18.170,  18.245,  16.728,  16.607,  18.466,  18.224,  16.794,  14.831,  12.676,
  11.698,  10.715,   9.463,   8.072,   7.012,   6.019,   4.869,   3.596,   2.150,   1.615,
   0.893,   0.089,  -0.838,  -2.404,  -4.186,  -7.232, -10.679, -13.053, -13.991, -14.970,
 -17.646, -20.085, -21.966, -23.907, -25.922, -27.691, -28.653, -29.475, -30.378, -31.316,
 -32.401, -33.514, -34.307, -35.338, -36.738, -38.447, -40.248, -41.514, -42.415, -43.070,
 -43.972, -45.213, -46.299, -47.166, -47.833, -47.891, -47.934, -48.042, -47.693, -46.916,
 -45.726, -44.863, -44.080, -43.525, -42.748, -41.520, -40.264, -39.083, -37.536, -35.750,
 -34.275, -32.503, -30.521, -28.570, -27.072, -25.359, -24.250, -24.414, -23.163, -21.056,
 -19.181, -16.441, -13.022,  -9.334,  -5.407,  -0.931,   2.316,   6.036,   9.512,   9.260,
  10.175,  12.388,  14.705,  17.506,  19.687,  21.892,  23.352,  25.262,  28.940,  31.305,
  33.432,  37.363,  41.538,  45.059,  47.700,  49.067,  46.690,  46.762,  49.703,  49.850,
  50.119,  50.909,  51.711,  52.781,  54.032,  55.348,  56.727,  58.614,  60.290,  61.684,
  62.918,  64.002,  64.829,  65.107,  65.045,  64.795,  64.377,  64.034,  63.594,  63.098,
  62.499,  61.703,  61.009,  60.534,  60.126,  60.326,  60.434,  60.176,  59.913,  59.395,
  59.318,  58.283,  57.260,  56.109,  54.483,  52.420,  50.047,  49.287,  49.938,  50.172},
{
  49.071,  48.060,  47.210,  46.513,  45.481,  45.186,  45.196,  44.547,  43.589,  42.426,
  41.486,  39.656,  37.262,  34.957,  33.061,  31.863,  30.666,  28.517,  25.233,  22.149,
  19.592,  18.308,  17.958,  17.872,  18.137,  18.282,  18.202,  18.171,  18.188,  17.648,
  17.140,  16.460,  15.455,  15.009,  14.503,  13.708,  13.918,  13.837,  12.845,  12.532,
  12.092,  11.629,  11.223,  11.149,  10.682,   9.205,   8.968,   9.374,   9.298,   8.245,
   6.126,   4.107,   2.411,   1.304,   1.186,   0.755,  -0.318,  -1.662,  -2.347,  -1.889,
  -3.265,  -5.758,  -8.116,  -9.843, -10.584, -11.199, -11.661, -13.026, -14.412, -15.634,
 -16.973, -18.181, -19.272, -20.015, -20.741, -21.271, -21.919, -22.578, -22.649, -23.072,
 -23.575, -23.725, -24.354, -24.501, -24.133, -24.964, -25.856, -25.567, -26.107, -26.615,
 -26.437, -26.711, -26.708, -26.873, -27.720, -28.663, -29.144, -29.264, -29.376, -29.917,
 -30.141, -29.516, -29.387, -29.846, -29.637, -29.113, -28.993, -28.584, -27.786, -26.492,
 -25.129, -24.313, -23.674, -22.946, -22.260, -22.406, -22.797, -21.904, -19.685, -18.300,
 -18.298, -17.356, -15.497, -14.288, -13.594, -12.971, -12.563, -12.051, -11.288, -10.505,
  -9.764,  -8.459,  -6.843,  -5.893,  -5.060,  -4.091,  -2.428,   1.053,   4.638,   6.973,
   8.262,   8.483,   8.641,   8.851,   9.730,  10.999,  10.926,  10.866,  11.658,  11.960,
  11.874,  11.917,  11.642,  11.388,  12.087,  12.720,  12.798,  13.186,  13.373,  13.050,
  12.651,  12.295,  11.797,  11.470,  11.512,  11.509,  10.998,  10.102,   9.753,   9.733,
   9.455,   9.098,   8.573,   8.095,   7.473,   6.902,   6.254,   5.247,   5.137,   4.989,   4.329,
   4.534,   4.586,   4.166,   4.717,   5.136,   5.087,   5.404,   5.695,   6.084,   6.318,
   6.529,   6.664,   6.815,   7.167,   7.774,   8.633,   9.546,  10.506,  11.028,  10.899,
  10.576,  10.959,  11.575,  11.582,  11.550,  12.245,  13.687,  15.889,  16.664,  15.457,
  15.483,  16.314,  16.709,  17.119,  17.233,  16.335,  14.517,  13.144,  12.449,  11.608,
  10.744,   9.690,   8.467,   7.398,   6.294,   5.211,   4.344,   3.308,   2.252,   1.796,
   1.144,   0.349,  -0.897,  -3.023,  -5.275,  -7.912, -10.569, -12.405, -13.612, -15.791,
 -18.289, -19.716, -21.500, -23.452, -25.296, -26.527, -27.362, -28.429, -29.080, -29.993,
 -30.993, -32.429, -33.828, -34.678, -36.202, -37.955, -39.343, -40.867, -42.331, -43.434,
 -44.341, -44.887, -45.287, -45.396, -45.358, -45.238, -45.141, -44.871, -44.411, -44.116,
 -43.497, -42.770, -42.038, -41.417, -40.737, -40.030, -39.063, -38.102, -37.323, -36.062,
 -34.543, -33.046, -32.246, -31.190, -29.377, -27.720, -25.980, -23.308, -19.568, -15.822,
 -12.803, -10.835,  -8.640,  -5.760,  -3.618,  -1.488,   2.019,   6.855,  10.087,  10.859,
  12.072,  14.050,  16.150,  18.157,  20.247,  22.070,  24.755,  28.261,  30.034,  32.328,
  36.870,  41.199,  44.442,  47.180,  49.366,  50.233,  47.998,  45.931,  47.648,  49.628,
  50.213,  50.849,  51.522,  52.240,  53.456,  54.622,  55.710,  57.190,  58.946,  60.554,
  61.867,  63.030,  63.803,  64.711,  65.390,  65.492,  65.152,  64.895,  64.563,  63.360,
  62.976,  63.022,  62.232,  61.279,  60.621,  60.914,  61.582,  61.560,  61.037,  60.167,
  59.534,  58.715,  57.102,  55.515,  53.572,  52.165,  51.446,  50.315,  49.639,  49.071},
{
  49.781,  48.638,  46.716,  44.688,  42.712,  42.236,  43.352,  43.190,  42.118,  41.314,
  40.655,  38.901,  36.107,  33.958,  32.366,  31.097,  30.251,  29.642,  28.403,  26.188,
  23.769,  21.194,  19.017,  17.835,  17.417,  17.783,  17.929,  17.715,  17.720,  17.735,
  17.631,  16.743,  16.032,  16.029,  15.710,  15.205,  14.955,  14.804,  14.364,  13.946,
  13.486,  12.712,  11.873,  11.040,   9.930,   8.955,   8.897,   8.911,   8.036,   6.773,
   5.733,   4.751,   3.643,   2.293,   0.881,  -0.018,  -0.531,  -1.967,  -2.919,  -1.580,
  -1.684,  -4.098,  -6.056,  -7.841,  -9.891, -11.341, -11.872, -12.322, -13.143, -14.125,
 -14.992, -15.897, -16.896, -17.919, -19.192, -20.223, -20.757, -20.799, -20.729, -21.002,
 -21.127, -21.568, -22.256, -22.198, -22.296, -23.282, -24.251, -24.776, -25.047, -24.889,
 -24.570, -24.256, -24.372, -25.379, -26.684, -27.504, -27.532, -27.563, -27.748, -28.082,
 -28.782, -29.027, -28.645, -27.999, -27.156, -26.474, -26.248, -25.985, -25.231, -23.962,
 -22.617, -21.517, -20.655, -19.971, -19.649, -20.202, -20.460, -19.804, -18.951, -18.219,
 -17.948, -17.576, -16.966, -16.323, -15.668, -15.213, -14.259, -12.509, -10.556,  -9.451,
  -9.524,  -8.661,  -5.920,  -2.888,  -0.371,   1.385,   2.690,   3.945,   4.982,   5.963,
   6.389,   6.209,   6.516,   7.767,   8.719,   8.699,   9.416,  10.537,  10.571,  10.385,
  10.468,  10.113,   9.552,   9.600,  10.032,  10.565,  11.023,  11.486,  11.907,  11.940,
  12.208,  12.205,  11.853,  11.674,  11.224,  11.129,  10.961,  10.484,   9.954,   9.136,
   8.703,   8.213,   7.851,   7.501,   6.811,   6.142,   4.721,   4.071,   4.181,   3.684,   3.863,
   4.197,   4.351,   4.644,   4.598,   4.779,   4.886,   4.602,   4.994,   5.477,   5.306,
   5.281,   5.250,   5.358,   5.890,   6.395,   6.845,   7.453,   8.305,   9.233,   9.817,
   9.681,   9.695,  10.209,  10.430,  10.838,  11.682,  11.957,  11.952,  12.723,  13.592,
  14.322,  15.168,  14.701,  13.548,  13.111,  12.765,  12.648,  12.355,  11.574,  10.950,
  10.301,   9.695,   8.981,   8.054,   7.170,   5.922,   5.087,   4.844,   4.189,   3.261,
   1.620,  -0.344,  -2.581,  -5.699,  -8.097,  -9.404, -10.811, -12.589, -14.384, -16.173,
 -17.658, -19.022, -20.754, -22.188, -23.361, -24.183, -25.288, -27.000, -28.177, -29.015,
 -30.171, -32.087, -33.627, -34.121, -35.067, -37.035, -38.903, -40.242, -41.730, -42.936,
 -43.711, -43.823, -43.206, -42.814, -42.183, -41.318, -40.800, -40.262, -40.000, -39.909,
 -39.699, -39.476, -38.955, -38.259, -37.459, -36.709, -36.116, -35.295, -34.359, -33.837,
 -33.020, -31.718, -30.695, -29.379, -27.798, -25.809, -23.077, -19.839, -16.167, -13.286,
 -11.370,  -8.344,  -4.428,  -2.366,  -1.804,  -1.412,   0.900,   6.745,  11.569,  13.594,
  14.919,  16.309,  18.747,  20.795,  22.495,  25.512,  27.946,  28.382,  29.388,  32.519,
  36.246,  39.911,  43.236,  46.319,  48.852,  50.119,  50.219,  48.001,  45.436,  46.650,
  49.921,  52.085,  53.348,  53.724,  53.505,  54.034,  55.143,  56.232,  57.613,  59.219,
  60.636,  61.847,  63.320,  64.505,  64.819,  65.224,  65.897,  66.313,  66.295,  66.727,
  67.146,  66.196,  65.209,  64.681,  64.009,  63.172,  62.669,  62.383,  61.443,  59.940,
  58.487,  57.285,  55.464,  53.497,  52.335,  51.892,  51.761,  51.386,  50.646,  49.781},
{
  50.804,  49.669,  48.096,  46.149,  44.601,  43.936,  42.783,  41.290,  40.714,  40.404,
  40.107,  39.329,  37.513,  35.803,  34.155,  32.129,  30.617,  29.746,  28.766,  27.286,
  25.206,  22.680,  20.332,  18.747,  17.957,  17.704,  17.735,  17.541,  17.417,  18.036,
  17.970,  16.842,  16.576,  17.028,  17.237,  17.406,  17.469,  16.686,  15.003,  13.497,
  12.542,  12.019,  11.342,  10.008,   8.882,   8.185,   7.565,   7.061,   6.593,   6.452,
   6.356,   5.378,   3.543,   1.839,   0.680,  -0.103,  -0.445,  -1.598,  -2.903,  -2.044,
  -1.065,  -1.867,  -3.810,  -5.945,  -7.771,  -9.810, -10.880, -11.423, -12.695, -13.421,
 -14.057, -15.061, -16.102, -17.129, -17.698, -18.438, -19.152, -19.320, -19.095, -18.871,
 -19.713, -20.565, -20.318, -20.095, -20.243, -20.859, -21.998, -22.899, -23.219, -23.117,
 -22.756, -22.492, -23.214, -24.639, -25.130, -24.801, -24.911, -25.127, -24.601, -23.850,
 -23.926, -24.867, -25.148, -23.958, -22.647, -22.040, -21.897, -21.697, -20.881, -19.927,
 -19.136, -18.281, -17.586, -17.084, -17.235, -18.006, -18.224, -17.951, -17.522, -17.091,
 -17.007, -16.981, -16.995, -17.012, -17.111, -16.393, -13.592, -10.005,  -6.848,  -4.765,
  -4.145,  -3.655,  -2.813,  -1.704,  -0.341,   0.480,   1.184,   2.274,   2.998,   3.325,
   4.132,   5.402,   6.112,   6.376,   7.335,   8.831,   9.148,   8.850,   9.672,   9.259,
   7.218,   6.852,   7.702,   8.092,   8.176,   8.414,   8.861,   9.082,   9.289,   9.572,
   9.546,   9.655,  10.450,  11.004,  10.875,  10.821,  10.646,  10.163,   9.517,   9.107,
   8.587,   7.476,   7.113,   6.927,   6.365,   6.031,   5.147,   4.661,   4.859,   4.845,   4.744,
   4.414,   4.325,   4.597,   4.797,   4.712,   4.594,   4.446,   3.963,   4.115,   4.407,
   4.308,   4.882,   5.672,   6.309,   6.706,   6.806,   7.198,   7.640,   8.024,   8.565,
   8.828,   8.595,   8.599,   9.101,   9.545,   9.873,   9.983,  10.257,  10.748,  10.782,
  10.945,  11.331,  11.598,  11.991,  12.111,  11.911,  11.502,  10.850,  10.708,  11.002,
  10.679,   9.920,   9.297,   8.587,   7.694,   7.032,   6.699,   5.413,   3.362,   1.904,
  -0.088,  -3.180,  -6.719,  -9.565, -10.599, -11.337, -12.860, -14.575, -15.633, -16.214,
 -17.786, -19.380, -19.812, -20.466, -21.530, -22.335, -23.587, -25.140, -26.480, -27.900,
 -29.337, -31.311, -33.250, -33.864, -34.444, -35.819, -37.102, -38.283, -39.447, -40.302,
 -40.841, -41.090, -41.155, -40.852, -40.193, -39.419, -38.484, -37.880, -37.650, -37.315,
 -37.028, -36.871, -36.589, -35.866, -34.623, -33.332, -33.129, -33.421, -32.546, -31.898,
 -31.504, -29.657, -27.773, -26.423, -24.864, -23.133, -20.943, -18.630, -16.301, -13.472,
 -10.569,  -7.986,  -6.275,  -5.160,  -2.351,   1.061,   2.885,   6.037,  11.805,  16.237,
  17.583,  18.725,  20.498,  22.436,  25.775,  27.877,  27.195,  27.312,  29.372,  32.779,
  36.792,  40.070,  42.997,  45.749,  47.873,  49.201,  49.146,  47.725,  45.461,  44.363,
  45.926,  48.357,  51.239,  54.430,  55.972,  56.018,  55.682,  56.101,  57.526,  58.595,
  59.422,  60.867,  62.647,  63.786,  64.343,  65.146,  65.723,  66.007,  66.509,  66.912,
  67.182,  67.258,  66.703,  66.280,  65.880,  64.338,  63.069,  62.496,  61.378,  59.593,
  57.650,  56.716,  55.868,  54.303,  53.278,  52.695,  52.046,  51.542,  51.296,  50.804},
{
  50.320,  49.180,  47.357,  45.965,  44.929,  43.966,  43.393,  42.911,  41.948,  40.709,
  39.595,  38.713,  37.149,  34.847,  33.515,  32.982,  32.282,  31.455,  30.359,  29.188,
  27.593,  25.398,  23.352,  21.557,  20.311,  19.611,  18.865,  18.272,  18.128,  18.038,
  17.548,  17.264,  17.422,  17.535,  17.690,  17.259,  16.189,  15.355,  14.820,  14.646,
  14.135,  12.902,  12.034,  11.859,  11.771,  10.472,   7.839,   6.174,   6.313,   6.558,
   5.744,   4.229,   2.659,   1.503,   1.004,   0.495,  -0.568,  -1.759,  -2.640,  -3.545,
  -4.951,  -6.142,  -6.034,  -4.781,  -4.423,  -6.297,  -8.476,  -9.751, -11.018, -12.368,
 -13.434, -14.389, -15.571, -16.582, -17.064, -17.195, -17.184, -17.229, -17.008, -17.202,
 -18.158, -18.652, -18.806, -18.600, -18.393, -19.244, -19.993, -20.612, -21.671, -22.042,
 -22.112, -22.444, -22.667, -22.729, -22.335, -21.957, -22.295, -22.698, -22.534, -22.143,
 -22.068, -22.245, -21.991, -21.239, -20.701, -20.446, -20.052, -19.256, -18.121, -17.400,
 -17.047, -16.263, -15.560, -15.357, -15.222, -15.363, -15.696, -15.497, -14.939, -14.678,
 -14.534, -14.356, -14.403, -14.589, -14.867, -14.209, -11.527,  -8.162,  -5.585,  -4.273,
  -4.341,  -4.668,  -4.363,  -3.435,  -2.088,  -0.946,  -0.228,   0.602,   1.864,   3.224,
   4.383,   5.025,   4.787,   4.609,   5.482,   6.992,   8.181,   8.183,   7.048,   5.336,
   4.385,   5.388,   6.608,   6.938,   7.260,   7.138,   6.936,   7.419,   7.793,   8.089,
   8.730,   9.330,   9.518,   9.522,   9.785,   9.818,   9.597,   9.664,   9.486,   8.864,
   8.377,   8.219,   7.547,   6.324,   6.085,   6.180,   5.640,   5.319,   5.186,   5.191,   4.960,
   4.354,   4.248,   4.123,   3.782,   3.610,   3.360,   3.446,   3.877,   3.868,   3.528,
   3.665,   4.071,   4.429,   4.894,   5.090,   5.257,   5.577,   5.933,   6.423,   6.702,
   7.110,   7.475,   7.417,   7.668,   8.143,   8.561,   9.029,   9.513,   9.698,   9.346,
   9.250,   9.528,   9.516,   9.235,   8.982,   8.896,   8.756,   8.676,   8.761,   8.532,
   7.798,   6.855,   6.255,   5.715,   4.299,   2.057,  -0.125,  -1.825,  -3.528,  -5.103,
  -6.017,  -6.732,  -7.441,  -8.504, -10.366, -11.967, -13.278, -15.060, -16.489, -17.296,
 -18.020, -18.670, -18.828, -18.727, -19.660, -21.264, -22.558, -23.878, -25.209, -26.769,
 -28.654, -30.340, -31.737, -32.899, -34.183, -35.511, -36.375, -36.989, -37.612, -38.258,
 -38.844, -39.193, -38.983, -38.174, -37.292, -36.598, -36.365, -36.348, -35.993, -35.612,
 -34.990, -34.212, -34.016, -33.967, -33.493, -32.725, -32.068, -31.242, -30.228, -29.717,
 -29.046, -28.222, -27.517, -25.997, -24.243, -22.590, -20.522, -18.303, -16.029, -13.493,
 -10.655,  -8.059,  -4.957,  -0.638,   3.416,   6.444,   7.608,   8.528,  12.763,  17.391,
  19.477,  20.608,  21.681,  24.306,  27.264,  28.289,  29.648,  31.441,  32.417,  33.877,
  36.234,  39.109,  41.712,  43.608,  45.463,  47.251,  48.111,  47.701,  47.545,  48.339,
  48.629,  49.125,  51.005,  53.028,  54.952,  56.866,  57.679,  58.018,  58.964,  59.796,
  60.199,  60.641,  61.270,  62.245,  63.207,  63.817,  64.557,  65.173,  65.289,  65.564,
  65.816,  65.467,  65.239,  65.252,  64.406,  63.308,  62.915,  61.853,  59.850,  58.435,
  57.548,  56.051,  54.348,  53.632,  52.998,  52.000,  51.661,  51.267,  50.668,  50.320},
{
  48.953,  48.457,  48.183,  46.815,  44.814,  43.324,  42.573,  42.322,  42.618,  42.485,
  41.118,  39.795,  38.287,  35.543,  32.860,  31.089,  30.483,  30.975,  31.118,  30.557,
  29.653,  28.263,  26.427,  24.381,  22.889,  21.791,  20.431,  19.308,  18.370,  17.625,
  17.566,  17.381,  16.817,  16.827,  17.295,  17.671,  17.718,  16.833,  15.517,  14.880,
  14.141,  12.850,  12.111,  11.854,  11.037,   9.382,   7.806,   7.332,   7.508,   7.222,
   5.885,   3.987,   2.651,   1.932,   1.271,   0.475,  -0.434,  -1.241,  -2.023,  -2.800,
  -3.634,  -5.067,  -6.610,  -6.994,  -5.956,  -4.716,  -5.011,  -6.683,  -8.499, -10.276,
 -11.727, -12.898, -14.186, -14.973, -15.127, -15.191, -15.320, -15.350, -15.372, -15.908,
 -16.363, -16.460, -16.735, -16.644, -16.923, -17.693, -17.473, -17.539, -18.339, -18.805,
 -19.549, -19.825, -19.411, -19.500, -19.521, -19.538, -19.850, -19.590, -19.122, -19.302,
 -19.930, -19.959, -18.991, -18.312, -18.449, -18.576, -18.500, -18.334, -17.904, -17.384,
 -17.036, -16.455, -15.422, -14.514, -13.924, -13.584, -13.326, -12.716, -12.185, -12.066,
 -11.932, -11.993, -12.476, -13.340, -14.176, -13.626, -11.231,  -7.986,  -5.498,  -4.901,
  -5.111,  -5.029,  -4.852,  -4.239,  -3.273,  -2.358,  -1.761,  -1.454,   0.157,   2.525,
   3.045,   2.526,   2.977,   4.586,   5.797,   4.531,   2.419,   2.301,   3.501,   4.613,
   5.277,   5.549,   5.742,   5.805,   5.794,   5.877,   6.019,   6.375,   6.875,   7.083,
   7.138,   7.429,   7.553,   7.534,   7.870,   8.155,   8.378,   8.670,   8.573,   8.321,
   8.001,   7.527,   7.113,   6.602,   6.298,   6.334,   6.376,   6.493,   6.276,   5.571,   4.990,
   4.613,   4.205,   3.627,   2.901,   2.318,   2.044,   2.103,   2.375,   2.438,   2.173,
   1.995,   2.074,   2.175,   2.235,   2.521,   3.079,   3.599,   4.142,   4.754,   5.193,
   5.484,   5.782,   6.370,   7.154,   7.801,   8.369,   8.572,   8.547,   8.876,   9.220,
   9.229,   8.787,   7.922,   7.294,   7.069,   6.900,   6.986,   7.261,   7.258,   6.985,
   5.975,   4.545,   4.374,   4.097,   1.866,  -0.162,  -1.058,  -2.034,  -2.957,  -3.825,
  -4.723,  -5.466,  -6.371,  -7.527,  -8.985, -10.438, -11.536, -12.946, -14.709, -15.903,
 -16.494, -17.255, -18.256, -19.004, -19.988, -21.720, -23.414, -24.683, -26.011, -27.586,
 -29.505, -31.333, -32.212, -32.754, -33.766, -34.699, -35.542, -36.342, -36.560, -36.629,
 -36.819, -36.767, -36.638, -36.242, -35.204, -34.185, -33.581, -32.787, -31.899, -31.426,
 -31.390, -31.841, -32.073, -31.487, -30.651, -29.632, -28.589, -28.139, -27.893, -27.239,
 -26.567, -26.358, -25.586, -23.690, -22.136, -21.016, -18.985, -16.323, -13.651, -10.362,
  -6.799,  -3.459,   1.297,   6.181,   7.642,   7.310,   8.245,  11.162,  15.159,  17.733,
  19.391,  21.406,  22.930,  24.644,  25.947,  26.441,  28.063,  29.960,  31.408,  33.422,
  35.818,  38.617,  41.185,  42.885,  44.498,  46.048,  47.075,  47.573,  47.887,  48.558,
  50.070,  52.646,  55.134,  56.307,  56.595,  56.764,  57.100,  57.396,  57.683,  58.397,
  59.220,  59.864,  60.220,  60.108,  60.535,  61.916,  62.938,  63.265,  63.553,  63.774,
  63.806,  63.680,  63.422,  63.186,  62.836,  61.989,  60.339,  58.504,  57.679,  57.274,
  56.376,  55.497,  54.278,  52.719,  52.028,  51.623,  50.914,  50.641,  50.105,  48.953},
{
  48.675,  48.343,  48.079,  47.523,  46.503,  45.195,  43.844,  42.287,  41.043,  40.898,
  41.162,  41.126,  40.858,  39.433,  35.936,  31.938,  30.005,  30.383,  31.451,  31.729,
  30.598,  29.132,  28.319,  27.109,  25.180,  23.652,  22.584,  21.524,  20.550,  19.620,
  18.711,  18.130,  18.074,  18.217,  18.053,  17.548,  16.862,  15.930,  14.902,  13.927,
  12.661,  11.325,  10.702,  10.566,  10.274,   9.889,   9.418,   8.659,   7.539,   6.025,
   4.626,   3.837,   3.008,   1.727,   0.666,  -0.011,  -0.635,  -1.283,  -2.005,  -2.544,
  -2.992,  -3.904,  -4.993,  -5.900,  -6.602,  -6.502,  -5.810,  -6.006,  -7.484,  -9.148,
 -10.591, -12.200, -13.357, -13.598, -13.661, -13.799, -13.944, -14.119, -14.135, -14.382,
 -14.875, -15.165, -15.414, -15.538, -15.529, -15.541, -15.403, -15.709, -16.600, -17.244,
 -17.678, -17.891, -17.453, -16.751, -16.306, -16.238, -16.473, -16.635, -16.445, -16.268,
 -16.496, -16.665, -16.376, -16.028, -16.088, -16.493, -16.601, -16.405, -16.551, -16.489,
 -15.764, -15.205, -14.963, -14.619, -14.107, -13.326, -12.407, -11.613, -11.103, -10.941,
 -10.928, -11.111, -11.911, -13.506, -15.190, -15.526, -13.753, -10.280,  -6.653,  -4.993,
  -5.184,  -5.191,  -4.645,  -4.158,  -3.565,  -3.020,  -2.445,  -0.938,   0.831,   1.329,
   1.325,   1.543,   0.949,   0.232,   0.770,   1.810,   2.602,   3.204,   3.700,   4.289,
   4.652,   4.667,   4.745,   4.815,   4.856,   5.002,   5.102,   5.232,   5.463,   5.681,
   5.855,   5.947,   6.093,   6.518,   7.067,   7.427,   7.656,   7.968,   8.083,   7.517,
   6.534,   5.824,   5.656,   6.006,   6.452,   6.408,   6.212,   6.293,   6.185,   5.665,   4.952,
   4.211,   3.594,   2.983,   2.322,   1.757,   1.305,   1.081,   0.955,   0.690,   0.576,
   0.599,   0.577,   0.816,   1.145,   1.372,   1.986,   2.890,   3.599,   4.065,   4.153,
   3.940,   4.072,   5.005,   6.470,   7.600,   7.719,   7.272,   7.336,   8.217,   9.153,
   9.481,   9.312,   8.877,   8.363,   8.089,   7.828,   7.267,   6.558,   5.621,   4.635,
   3.853,   2.882,   1.683,   0.111,  -1.696,  -2.305,  -1.997,  -2.462,  -3.484,  -4.530,
  -5.528,  -6.073,  -6.645,  -7.660,  -8.482,  -8.945,  -9.557, -10.451, -11.310, -12.456,
 -14.513, -16.567, -17.551, -18.519, -20.734, -23.396, -25.166, -26.392, -27.907, -29.326,
 -30.356, -31.172, -31.693, -32.304, -33.052, -33.462, -34.025, -34.869, -35.316, -35.446,
 -35.343, -34.940, -34.598, -34.265, -33.767, -32.961, -31.549, -30.197, -29.732, -29.761,
 -30.077, -30.803, -30.921, -29.513, -27.500, -26.357, -26.263, -26.437, -26.272, -25.470,
 -24.340, -23.448, -22.349, -20.599, -18.618, -16.532, -14.321, -11.824,  -8.677,  -5.340,
  -2.283,   0.335,   1.995,   3.427,   5.990,   9.118,  11.916,  14.446,  16.569,  19.044,
  22.299,  24.482,  25.184,  25.488,  25.211,  25.298,  26.408,  27.378,  29.002,  31.841,
  34.233,  36.012,  37.930,  39.923,  41.852,  43.444,  44.766,  46.036,  47.008,  47.891,
  49.302,  51.305,  53.646,  55.771,  56.435,  55.750,  55.820,  56.915,  57.766,  58.560,
  59.039,  58.873,  59.273,  59.755,  59.200,  58.915,  59.650,  60.598,  61.375,  61.869,
  62.236,  62.347,  61.881,  61.392,  60.854,  59.959,  59.364,  58.763,  57.650,  56.638,
  55.690,  54.711,  53.771,  52.324,  51.032,  50.830,  50.823,  50.215,  49.348,  48.675},
{
  48.539,  47.985,  47.224,  46.682,  46.011,  45.051,  44.224,  43.497,  42.604,  41.457,
  40.202,  39.381,  39.063,  38.724,  38.192,  37.221,  35.460,  33.375,  31.776,  31.040,
  30.833,  30.210,  28.939,  27.757,  26.797,  25.774,  24.876,  23.858,  22.339,  21.021,
  20.313,  19.436,  18.287,  17.488,  16.774,  15.743,  14.696,  13.761,  13.001,  12.549,
  11.933,  11.024,  10.395,   9.818,   8.894,   8.150,   7.649,   6.777,   5.557,   4.489,
   3.798,   3.135,   2.111,   1.109,   0.454,  -0.275,  -1.208,  -1.980,  -2.517,  -3.129,
  -4.036,  -4.794,  -4.995,  -5.227,  -5.890,  -6.551,  -7.092,  -7.696,  -8.387,  -9.235,
 -10.126, -10.768, -11.174, -11.525, -11.787, -11.920, -12.127, -12.448, -12.763, -13.125,
 -13.457, -13.744, -14.136, -14.442, -14.741, -15.199, -15.121, -14.474, -14.472, -15.374,
 -16.097, -15.785, -14.925, -14.673, -14.704, -14.242, -14.309, -15.145, -15.404, -15.697,
 -16.892, -17.349, -16.463, -15.690, -15.428, -15.221, -14.970, -14.578, -14.051, -13.575,
 -13.300, -12.968, -12.317, -12.024, -12.349, -12.318, -11.848, -11.571, -11.175, -10.455,
 -10.066, -10.484, -11.522, -12.715, -13.783, -14.299, -13.408, -10.790,  -7.492,  -5.426,
  -5.331,  -5.617,  -5.020,  -4.312,  -3.823,  -3.053,  -2.307,  -1.616,  -0.631,   0.078,
   0.181,   0.174,   0.296,   0.575,   0.913,   1.168,   1.590,   2.229,   2.780,   3.231,
   3.570,   3.567,   3.319,   3.250,   3.481,   3.706,   3.804,   3.914,   4.093,   4.327,
   4.505,   4.749,   5.309,   5.846,   6.110,   6.364,   6.438,   6.218,   5.990,   5.612,
   5.117,   4.905,   4.878,   4.941,   5.133,   5.259,   5.333,   5.217,   4.673,   4.086,   3.662,
   3.147,   2.601,   2.076,   1.507,   1.057,   0.801,   0.681,   0.684,   0.849,   1.079,
   1.046,   0.867,   0.914,   0.899,   0.831,   1.372,   2.435,   3.438,   4.099,   4.186,
   3.801,   3.315,   2.864,   2.605,   2.666,   2.863,   2.936,   2.869,   2.975,   3.334,
   3.816,   4.762,   6.070,   6.870,   7.104,   7.104,   6.527,   5.381,   4.213,   2.968,
   0.977,  -1.670,  -3.747,  -4.829,  -5.555,  -5.808,  -5.549,  -5.538,  -5.915,  -6.581,
  -7.513,  -8.233,  -8.705,  -8.988,  -8.640,  -8.163,  -8.491,  -9.410, -10.522, -11.914,
 -13.570, -15.385, -17.083, -18.363, -19.416, -20.349, -20.928, -21.645, -23.400, -26.025,
 -28.424, -29.823, -30.393, -30.797, -31.272, -31.525, -31.355, -31.052, -31.161, -31.611,
 -31.801, -31.670, -31.522, -31.381, -31.092, -30.403, -29.300, -28.309, -27.907, -27.929,
 -27.857, -27.745, -27.845, -27.209, -25.251, -23.597, -23.392, -23.549, -23.086, -22.237,
 -21.422, -20.519, -19.044, -16.884, -14.227, -11.411,  -8.851,  -6.354,  -3.902,  -1.888,
  -0.149,   0.661,   0.754,   3.618,   9.371,  13.256,  14.700,  16.286,  18.127,  20.710,
  23.868,  25.395,  26.111,  27.180,  26.358,  23.893,  22.776,  23.593,  25.739,  28.667,
  31.371,  33.721,  35.956,  37.855,  39.427,  40.699,  41.927,  43.598,  45.263,  46.528,
  48.016,  49.707,  51.354,  53.358,  54.637,  54.443,  55.191,  57.468,  58.676,  59.130,
  60.417,  61.196,  60.677,  59.801,  58.973,  58.607,  58.779,  59.004,  59.378,  59.761,
  60.002,  60.517,  60.835,  60.478,  60.054,  59.606,  58.906,  58.320,  57.719,  56.787,
  55.596,  54.056,  52.483,  51.546,  51.089,  50.572,  49.899,  49.210,  48.775,  48.539},
{
  48.836,  48.113,  47.473,  46.880,  46.213,  45.466,  44.808,  44.151,  43.381,  42.597,
  41.764,  40.837,  40.008,  39.415,  38.900,  38.095,  36.785,  35.149,  33.355,  31.377,
  29.535,  28.462,  27.925,  26.790,  25.140,  24.482,  24.746,  24.197,  22.661,  21.217,
  19.931,  18.624,  17.561,  16.603,  15.675,  14.992,  14.334,  13.591,  12.898,  11.909,
  10.668,   9.925,   9.573,   8.889,   8.001,   7.254,   6.466,   5.591,   4.745,   3.860,
   3.015,   2.412,   1.894,   1.067,  -0.084,  -1.118,  -1.694,  -1.776,  -1.638,  -1.957,
  -2.966,  -4.077,  -5.141,  -6.299,  -6.987,  -7.150,  -7.573,  -8.217,  -8.623,  -8.946,
  -9.252,  -9.502,  -9.822, -10.018, -10.056, -10.306, -10.707, -11.029, -11.458, -12.046,
 -12.638, -13.227, -13.732, -14.014, -14.149, -14.307, -14.506, -14.452, -13.940, -13.437,
 -13.267, -12.903, -12.301, -12.297, -13.004, -13.799, -14.364, -14.577, -14.592, -14.797,
 -15.010, -14.966, -14.906, -14.633, -13.722, -12.615, -11.895, -11.311, -10.346,  -9.317,
  -9.214, -10.057, -10.791, -11.124, -11.402, -11.437, -11.145, -10.776, -10.263,  -9.515,
  -8.772,  -8.418,  -8.628,  -9.063,  -9.408,  -9.715,  -9.831,  -9.507,  -8.653,  -7.267,
  -6.067,  -5.829,  -5.810,  -5.047,  -4.001,  -3.196,  -2.466,  -1.874,  -1.458,  -0.989,
  -0.683,  -0.800,  -0.953,  -0.819,  -0.553,  -0.135,   0.607,   1.419,   1.893,   2.066,
   2.234,   2.382,   2.318,   2.184,   2.219,   2.331,   2.421,   2.616,   2.928,   3.153,
   3.246,   3.453,   3.848,   4.146,   4.261,   4.262,   4.018,   3.790,   4.045,   4.376,
   4.223,   3.909,   3.782,   3.685,   3.612,   3.618,   3.478,   3.132,   2.873,   2.766,   2.600,
   2.377,   2.090,   1.525,   0.838,   0.483,   0.563,   0.879,   1.138,   1.055,   0.787,
   0.756,   0.936,   0.934,   0.628,   0.296,   0.098,  -0.156,  -0.522,  -0.742,  -0.723,
  -0.666,  -0.680,  -0.666,  -0.541,  -0.329,  -0.127,  -0.135,  -0.489,  -0.985,  -1.373,
  -1.643,  -1.717,  -1.556,  -1.498,  -1.711,  -1.849,  -1.818,  -2.020,  -2.501,  -2.948,
  -3.450,  -4.042,  -4.247,  -4.104,  -4.321,  -5.122,  -6.100,  -6.815,  -7.273,  -7.954,
  -8.850,  -9.484, -10.034, -10.769, -11.468, -12.205, -12.967, -13.447, -14.126, -15.490,
 -16.892, -17.729, -18.059, -18.034, -18.038, -18.443, -19.184, -20.370, -22.347, -24.836,
 -26.921, -27.923, -28.025, -27.993, -28.302, -28.727, -28.738, -28.210, -27.658, -27.524,
 -27.621, -27.592, -27.306, -26.710, -25.898, -25.242, -25.046, -25.234, -25.515, -25.671,
 -25.648, -25.430, -24.803, -23.593, -22.267, -21.466, -21.056, -20.484, -19.567, -18.443,
 -17.118, -15.364, -13.192, -10.755,  -7.833,  -4.749,  -2.459,  -0.692,   1.281,   2.847,
   3.569,   3.997,   4.750,   6.817,  10.390,  13.794,  16.262,  18.736,  21.210,  22.996,
  24.360,  25.777,  27.078,  27.821,  27.861,  27.319,  26.390,  25.997,  27.302,  29.860,
  32.384,  34.567,  36.487,  38.034,  39.311,  40.460,  41.720,  43.177,  44.430,  45.572,
  47.015,  48.372,  49.691,  51.632,  53.506,  54.539,  55.437,  56.384,  56.922,  57.182,
  56.887,  56.232,  56.652,  57.775,  57.764,  56.806,  56.253,  56.840,  58.478,  59.554,
  59.039,  58.503,  58.974,  59.433,  59.323,  58.949,  58.517,  57.925,  56.832,  55.741,
  55.297,  54.392,  52.474,  51.265,  51.329,  51.300,  50.816,  50.298,  49.634,  48.836},
{
  50.053,  49.355,  48.408,  47.220,  46.320,  45.571,  44.702,  44.039,  43.593,  42.993,
  42.337,  41.728,  40.944,  40.161,  39.591,  38.734,  37.200,  35.266,  33.175,  31.067,
  29.403,  28.437,  27.731,  26.739,  25.454,  24.184,  23.079,  22.156,  21.328,  20.225,
  18.703,  17.337,  16.605,  16.215,  15.790,  15.254,  14.508,  13.446,  12.109,  10.737,
   9.716,   9.131,   8.710,   8.347,   8.013,   7.508,   6.782,   5.828,   4.602,   3.481,
   2.791,   2.177,   1.499,   1.105,   0.795,   0.164,  -0.520,  -1.109,  -1.973,  -3.023,
  -3.868,  -4.724,  -5.818,  -6.664,  -6.924,  -7.006,  -7.262,  -7.611,  -7.986,  -8.366,
  -8.741,  -9.232,  -9.726,  -9.888,  -9.776,  -9.609,  -9.396,  -9.408,  -9.968, -10.734,
 -11.288, -11.800, -12.400, -12.878, -13.235, -13.651, -14.031, -14.197, -14.207, -14.112,
 -13.929, -13.960, -14.377, -14.740, -14.736, -14.575, -14.276, -13.798, -13.426, -13.029,
 -12.318, -11.930, -12.350, -12.793, -12.558, -11.841, -10.908,  -9.947,  -9.239,  -8.902,
  -9.026,  -9.596, -10.173, -10.366, -10.225,  -9.968,  -9.779,  -9.629,  -9.275,  -8.694,
  -8.081,  -7.508,  -7.050,  -6.802,  -6.763,  -6.975,  -7.401,  -7.759,  -7.838,  -7.618,
  -7.105,  -6.426,  -5.757,  -5.119,  -4.540,  -4.083,  -3.691,  -3.299,  -2.896,  -2.427,
  -1.937,  -1.592,  -1.403,  -1.267,  -1.103,  -0.725,  -0.074,   0.518,   0.861,   1.198,
   1.498,   1.407,   1.119,   1.140,   1.381,   1.519,   1.647,   1.817,   1.890,   1.983,
   2.242,   2.584,   3.048,   3.534,   3.553,   3.118,   2.940,   3.174,   3.313,   3.312,
   3.405,   3.334,   2.904,   2.503,   2.426,   2.541,   2.648,   2.592,   2.391,   2.352,   2.522,
   2.472,   2.001,   1.419,   1.018,   0.847,   0.854,   0.960,   1.153,   1.417,   1.529,
   1.306,   0.907,   0.528,   0.165,  -0.104,  -0.131,  -0.080,  -0.313,  -0.846,  -1.246,
  -1.196,  -0.984,  -1.060,  -1.386,  -1.726,  -2.091,  -2.458,  -2.736,  -3.072,  -3.575,
  -4.043,  -4.222,  -3.876,  -2.971,  -2.227,  -2.574,  -3.887,  -5.207,  -5.998,  -6.382,
  -6.476,  -6.144,  -5.404,  -4.607,  -3.947,  -3.364,  -3.136,  -3.801,  -5.373,  -7.267,
  -8.884, -10.091, -11.124, -11.956, -12.315, -12.521, -13.165, -14.024, -14.539, -14.878,
 -15.423, -16.010, -16.429, -16.819, -17.170, -17.315, -17.635, -18.877, -21.034, -23.116,
 -24.263, -24.578, -24.637, -24.720, -24.948, -25.443, -25.878, -25.716, -25.017, -24.172,
 -23.354, -22.754, -22.232, -21.124, -19.674, -19.291, -20.263, -21.205, -21.251, -20.799,
 -20.245, -19.578, -18.838, -18.148, -17.510, -16.747, -15.728, -14.581, -13.427, -12.124,
 -10.484,  -8.458,  -6.214,  -4.083,  -2.148,  -0.239,   1.414,   2.300,   2.579,   3.268,
   5.104,   7.560,   9.700,  11.589,  13.675,  15.661,  17.389,  19.477,  22.026,  24.228,
  25.599,  26.442,  27.288,  28.285,  28.837,  28.336,  27.509,  27.890,  29.794,  32.108,
  33.988,  35.510,  36.862,  37.981,  38.881,  39.743,  40.710,  41.863,  43.282,  44.846,
  46.268,  47.498,  48.625,  49.681,  50.912,  52.165,  52.676,  52.908,  53.985,  54.734,
  53.925,  53.539,  54.403,  54.037,  52.664,  53.191,  54.973,  55.864,  56.486,  57.054,
  56.532,  55.992,  56.659,  57.349,  57.319,  57.373,  57.582,  57.299,  56.598,  56.082,
  55.880,  55.402,  54.457,  53.747,  53.474,  53.068,  52.457,  51.798,  50.929,  50.053},
{
  50.553,  49.970,  49.272,  48.557,  47.778,  46.719,  45.589,  44.665,  43.736,  42.837,
  42.334,  41.908,  41.245,  40.962,  41.002,  40.058,  37.936,  35.894,  34.316,  32.646,
  30.863,  29.084,  27.179,  25.502,  24.385,  23.380,  22.188,  21.102,  20.158,  19.223,
  18.483,  17.852,  16.974,  16.058,  15.456,  14.892,  14.083,  13.152,  12.082,  10.847,
   9.785,   9.056,   8.446,   7.877,   7.372,   6.884,   6.431,   5.862,   4.869,   3.696,
   2.931,   2.576,   2.338,   2.201,   1.873,   0.781,  -0.846,  -2.057,  -2.487,  -2.683,
  -3.199,  -4.003,  -4.744,  -5.248,  -5.580,  -5.843,  -6.165,  -6.609,  -6.958,  -7.054,
  -7.215,  -7.682,  -8.104,  -8.173,  -8.061,  -7.965,  -8.018,  -8.488,  -9.335, -10.094,
 -10.519, -10.731, -10.879, -11.174, -11.717, -12.203, -12.408, -12.562, -12.746, -12.772,
 -12.769, -12.988, -13.323, -13.708, -14.132, -14.131, -13.475, -12.869, -12.722, -12.534,
 -12.421, -13.140, -14.193, -14.274, -13.369, -12.377, -11.551, -10.758, -10.183, -10.024,
 -10.181, -10.306, -10.094,  -9.742,  -9.637,  -9.592,  -9.258,  -8.913,  -8.924,  -9.043,
  -8.870,  -8.352,  -7.609,  -6.871,  -6.517,  -6.686,  -7.027,  -7.123,  -6.916,  -6.562,
  -6.162,  -5.732,  -5.240,  -4.676,  -4.136,  -3.695,  -3.318,  -3.017,  -2.820,  -2.576,
  -2.198,  -1.888,  -1.708,  -1.396,  -0.901,  -0.504,  -0.259,   0.048,   0.387,   0.496,
   0.369,   0.368,   0.739,   1.212,   1.358,   1.249,   1.259,   1.446,   1.701,   2.012,
   2.148,   1.939,   1.893,   2.396,   2.826,   2.646,   2.348,   2.381,   2.485,   2.498,
   2.451,   2.170,   1.801,   1.824,   2.126,   2.291,   2.466,   2.772,   2.786,   2.437,   2.154,
   1.929,   1.482,   0.943,   0.514,   0.186,   0.072,   0.194,   0.257,   0.217,   0.319,
   0.424,   0.247,  -0.010,  -0.074,  -0.038,   0.080,   0.371,   0.578,   0.442,   0.161,
  -0.107,  -0.586,  -1.342,  -2.061,  -2.494,  -2.721,  -3.188,  -4.410,  -6.176,  -7.528,
  -8.005,  -8.083,  -8.190,  -8.290,  -8.363,  -8.382,  -8.208,  -7.948,  -7.852,  -7.834,
  -7.696,  -7.500,  -7.373,  -7.301,  -7.214,  -6.996,  -6.419,  -5.357,  -4.422,  -4.962,
  -7.393, -10.015, -10.948, -10.474,  -9.943,  -9.683,  -9.480,  -9.643, -10.408, -11.449,
 -12.420, -13.256, -14.078, -15.098, -16.243, -17.221, -18.149, -19.282, -20.276, -20.819,
 -21.294, -21.824, -21.916, -21.769, -22.231, -23.070, -23.178, -22.388, -21.327, -20.141,
 -18.930, -18.270, -18.119, -17.656, -16.816, -16.436, -16.773, -17.207, -17.110, -16.244,
 -14.867, -13.685, -13.146, -12.913, -12.333, -11.181,  -9.887,  -9.026,  -8.485,  -7.539,
  -5.807,  -3.509,  -1.126,   0.615,   1.351,   2.042,   3.946,   6.708,   8.986,  10.356,
  11.222,  11.890,  12.729,  14.067,  15.596,  16.976,  18.572,  20.672,  22.842,  24.553,
  25.519,  25.639,  25.416,  25.728,  26.779,  28.198,  29.934,  31.962,  33.744,  34.932,
  35.832,  36.671,  37.299,  37.814,  38.556,  39.417,  40.066,  40.741,  41.832,  43.093,
  44.266,  45.642,  47.098,  48.021,  48.576,  49.259,  49.823,  50.248,  50.900,  51.123,
  50.587,  50.955,  52.854,  54.057,  53.458,  52.770,  52.981,  53.440,  54.112,  55.069,
  55.313,  54.463,  53.664,  53.767,  54.251,  54.290,  53.843,  53.460,  53.406,  53.410,
  53.222,  52.937,  52.727,  52.593,  52.424,  52.194,  51.944,  51.599,  51.099,  50.553},
{
  48.925,  48.839,  48.533,  48.093,  47.537,  46.813,  46.096,  45.508,  44.866,  44.091,
  43.339,  42.652,  42.070,  41.673,  41.056,  39.600,  37.440,  35.212,  33.169,  31.432,
  30.238,  29.237,  27.709,  25.747,  24.068,  22.890,  21.948,  21.056,  20.065,  18.931,
  17.921,  17.225,  16.696,  16.207,  15.730,  15.114,  14.322,  13.530,  12.691,  11.612,
  10.462,   9.558,   8.827,   8.039,   7.231,   6.545,   5.960,   5.308,   4.430,   3.396,
   2.555,   2.128,   1.958,   1.883,   1.899,   1.742,   0.936,  -0.431,  -1.663,  -2.384,
  -2.859,  -3.289,  -3.567,  -3.772,  -4.152,  -4.618,  -4.914,  -5.152,  -5.601,  -6.139,
  -6.468,  -6.646,  -6.947,  -7.309,  -7.456,  -7.468,  -7.701,  -8.173,  -8.596,  -8.912,
  -9.200,  -9.324,  -9.324,  -9.619, -10.279, -10.843, -11.172, -11.508, -11.712, -11.520,
 -11.233, -11.166, -11.083, -10.924, -11.146, -11.848, -12.515, -12.768, -12.602, -12.152,
 -11.763, -11.848, -12.381, -12.942, -13.151, -12.810, -12.005, -11.185, -10.723, -10.533,
 -10.394, -10.247, -10.015,  -9.612,  -9.187,  -8.978,  -8.953,  -8.856,  -8.539,  -8.112,
  -7.726,  -7.362,  -6.987,  -6.747,  -6.716,  -6.665,  -6.424,  -6.153,  -5.979,  -5.753,
  -5.316,  -4.664,  -3.919,  -3.348,  -3.125,  -3.033,  -2.792,  -2.501,  -2.267,  -1.870,
  -1.300,  -0.927,  -0.795,  -0.567,  -0.285,  -0.268,  -0.370,  -0.198,   0.180,   0.458,
   0.655,   0.934,   1.206,   1.332,   1.475,   1.873,   2.296,   2.187,   1.559,   1.166,
   1.388,   1.665,   1.521,   1.229,   1.103,   1.089,   1.217,   1.530,   1.706,   1.535,
   1.354,   1.414,   1.464,   1.351,   1.289,   1.345,   1.352,   1.319,   1.380,   1.478,   1.519,
   1.566,   1.564,   1.243,   0.591,   0.055,  -0.083,  -0.096,  -0.254,  -0.371,  -0.285,
  -0.200,  -0.249,  -0.274,  -0.206,  -0.156,  -0.180,  -0.340,  -0.638,  -0.792,  -0.694,
  -0.935,  -2.002,  -3.376,  -4.363,  -5.144,  -6.127,  -7.061,  -7.610,  -7.956,  -8.292,
  -8.425,  -8.206,  -7.801,  -7.447,  -7.282,  -7.364,  -7.622,  -7.889,  -8.071,  -8.192,
  -8.293,  -8.434,  -8.708,  -9.056,  -9.207,  -9.014,  -8.616,  -8.078,  -7.361,  -6.853,
  -7.203,  -8.378,  -9.591, -10.218, -10.195,  -9.703,  -9.067,  -8.735,  -8.987,  -9.807,
 -10.983, -12.155, -13.078, -13.918, -14.947, -16.133, -17.311, -18.391, -19.249, -19.800,
 -20.089, -20.092, -19.892, -19.966, -20.472, -20.643, -19.883, -18.717, -17.715, -16.653,
 -15.538, -14.947, -14.731, -14.005, -12.748, -11.863, -11.712, -11.874, -11.988, -11.846,
 -11.318, -10.816, -10.928, -11.211, -10.413,  -8.295,  -6.209,  -5.274,  -4.801,  -3.310,
  -0.686,   1.653,   2.854,   3.904,   6.148,   9.241,  11.652,  12.766,  13.221,  13.506,
  13.664,  14.060,  14.973,  15.989,  17.129,  19.336,  22.307,  23.697,  22.319,  20.224,
  19.847,  21.004,  22.193,  23.068,  24.314,  26.265,  28.550,  30.633,  32.387,  34.082,
  35.695,  36.683,  36.850,  36.818,  37.228,  38.081,  39.103,  40.118,  40.972,  41.642,
  42.400,  43.491,  44.851,  46.257,  47.432,  48.082,  48.329,  48.683,  49.073,  48.809,
  48.115,  48.266,  49.461,  50.252,  49.865,  49.360,  49.779,  50.686,  51.183,  51.295,
  51.704,  52.524,  52.958,  52.380,  51.273,  50.566,  50.459,  50.533,  50.555,  50.579,
  50.635,  50.680,  50.632,  50.398,  50.058,  49.783,  49.542,  49.227,  48.981,  48.925},
{
  46.724,  46.055,  45.403,  44.877,  44.576,  44.714,  45.145,  45.203,  44.488,  43.422,
  42.586,  41.955,  41.202,  40.302,  39.422,  38.571,  37.622,  36.423,  34.863,  33.025,
  31.164,  29.419,  27.760,  26.211,  24.864,  23.754,  22.882,  22.161,  21.326,  20.259,
  19.281,  18.675,  18.151,  17.301,  16.271,  15.423,  14.709,  13.889,  13.008,  12.180,
  11.285,  10.249,   9.279,   8.501,   7.768,   7.057,   6.545,   6.129,   5.419,   4.432,
   3.646,   3.180,   2.635,   1.938,   1.572,   1.694,   1.734,   1.168,   0.121,  -0.988,
  -1.915,  -2.586,  -3.024,  -3.340,  -3.609,  -3.797,  -3.925,  -4.125,  -4.434,  -4.736,
  -4.953,  -5.153,  -5.446,  -5.874,  -6.381,  -6.848,  -7.210,  -7.528,  -7.854,  -8.118,
  -8.273,  -8.431,  -8.699,  -8.998,  -9.215,  -9.405,  -9.644,  -9.850,  -9.946, -10.027,
 -10.167, -10.261, -10.214, -10.057,  -9.827,  -9.614,  -9.667, -10.114, -10.693, -11.098,
 -11.309, -11.322, -11.006, -10.527, -10.363, -10.604, -10.780, -10.559, -10.110,  -9.694,
  -9.348,  -9.017,  -8.653,  -8.216,  -7.750,  -7.361,  -7.058,  -6.777,  -6.516,  -6.276,
  -5.969,  -5.587,  -5.260,  -5.021,  -4.750,  -4.474,  -4.339,  -4.260,  -3.989,  -3.581,
  -3.314,  -3.189,  -2.966,  -2.611,  -2.269,  -1.972,  -1.724,  -1.604,  -1.524,  -1.273,
  -0.945,  -0.831,  -0.801,  -0.435,   0.192,   0.514,   0.345,   0.161,   0.358,   0.729,
   0.904,   0.967,   1.298,   1.958,   2.585,   2.883,   2.829,   2.422,   1.696,   0.996,
   0.735,   0.846,   0.941,   0.919,   0.981,   1.186,   1.398,   1.520,   1.524,   1.432,
   1.364,   1.387,   1.366,   1.217,   1.126,   1.240,   1.382,   1.346,   1.214,   1.173,   1.252,
   1.386,   1.507,   1.537,   1.466,   1.411,   1.437,   1.469,   1.542,   1.857,   2.392,
   2.704,   2.412,   1.664,   0.953,   0.575,   0.413,   0.150,  -0.436,  -1.390,  -2.636,
  -3.945,  -4.924,  -5.340,  -5.461,  -5.728,  -6.133,  -6.307,  -6.186,  -6.097,  -6.173,
  -6.114,  -5.629,  -4.856,  -4.306,  -4.431,  -5.187,  -6.080,  -6.745,  -7.310,  -7.965,
  -8.559,  -8.948,  -9.317,  -9.756,  -9.977,  -9.813,  -9.441,  -8.767,  -7.382,  -5.596,
  -4.684,  -5.436,  -7.138,  -8.526,  -9.136,  -9.277,  -9.349,  -9.668, -10.394, -11.245,
 -11.759, -12.067, -12.790, -13.982, -14.932, -15.275, -15.479, -15.941, -16.404, -16.627,
 -16.799, -17.020, -17.107, -17.042, -16.942, -16.694, -16.185, -15.571, -14.870, -13.895,
 -12.891, -12.386, -12.213, -11.700, -10.850, -10.272, -10.063,  -9.870,  -9.731,  -9.787,
  -9.662,  -9.059,  -8.340,  -7.763,  -6.984,  -5.894,  -4.968,  -4.277,  -3.175,  -1.352,
   0.779,   2.990,   5.545,   8.389,  10.873,  12.550,  13.631,  14.513,  15.267,  15.740,
  15.944,  16.211,  16.943,  18.230,  19.680,  20.570,  20.306,  19.068,  17.971,  18.133,
  19.522,  21.155,  22.420,  23.616,  25.168,  26.961,  28.658,  30.181,  31.585,  32.760,
  33.562,  34.117,  34.705,  35.394,  36.046,  36.613,  37.143,  37.567,  37.820,  38.090,
  38.684,  39.708,  41.043,  42.414,  43.481,  44.179,  44.896,  45.908,  46.851,  47.217,
  47.093,  46.909,  46.877,  47.066,  47.533,  48.041,  48.211,  48.117,  48.186,  48.541,
  49.043,  49.708,  50.441,  50.719,  50.195,  49.281,  48.634,  48.413,  48.366,  48.271,
  48.051,  47.785,  47.673,  47.744,  47.734,  47.489,  47.246,  47.207,  47.136,  46.724},
{
  43.723,  43.870,  44.082,  43.904,  43.583,  43.615,  43.790,  43.430,  42.496,  41.650,
  41.173,  40.662,  39.991,  39.725,  40.130,  40.408,  39.508,  37.348,  34.797,  32.658,
  31.083,  29.850,  28.802,  27.858,  26.858,  25.647,  24.296,  23.060,  22.093,  21.322,
  20.608,  19.889,  19.144,  18.348,  17.501,  16.604,  15.621,  14.560,  13.525,  12.584,
  11.666,  10.708,   9.819,   9.132,   8.587,   8.009,   7.337,   6.639,   5.939,   5.202,
   4.464,   3.832,   3.284,   2.651,   1.934,   1.527,   1.815,   2.528,   2.805,   2.082,
   0.688,  -0.638,  -1.555,  -2.202,  -2.761,  -3.273,  -3.738,  -4.100,  -4.270,  -4.328,
  -4.521,  -4.915,  -5.296,  -5.537,  -5.766,  -6.055,  -6.269,  -6.367,  -6.522,  -6.800,
  -7.042,  -7.152,  -7.232,  -7.329,  -7.329,  -7.194,  -7.061,  -7.028,  -7.065,  -7.152,
  -7.311,  -7.507,  -7.675,  -7.864,  -8.178,  -8.580,  -8.906,  -9.088,  -9.219,  -9.333,
  -9.290,  -9.029,  -8.803,  -8.908,  -9.190,  -9.138,  -8.554,  -7.793,  -7.210,  -6.738,
  -6.234,  -5.845,  -5.719,  -5.668,  -5.452,  -5.168,  -5.032,  -4.953,  -4.641,  -4.064,
  -3.511,  -3.231,  -3.168,  -3.069,  -2.784,  -2.413,  -2.167,  -2.118,  -2.122,  -2.001,
  -1.753,  -1.529,  -1.456,  -1.498,  -1.491,  -1.301,  -0.952,  -0.587,  -0.304,  -0.088,
   0.107,   0.278,   0.392,   0.458,   0.545,   0.729,   1.008,   1.247,   1.271,   1.069,
   0.864,   0.913,   1.207,   1.454,   1.425,   1.216,   1.092,   1.164,   1.338,   1.493,
   1.549,   1.482,   1.410,   1.536,   1.841,   2.032,   1.980,   1.966,   2.237,   2.566,
   2.571,   2.277,   2.023,   1.966,   1.961,   1.869,   1.739,   1.675,   1.707,   1.789,   1.845,
   1.840,   1.829,   1.894,   2.053,   2.266,   2.494,   2.700,   2.830,   2.834,   2.674,
   2.304,   1.742,   1.136,   0.614,   0.113,  -0.484,  -1.065,  -1.400,  -1.544,  -1.824,
  -2.402,  -3.129,  -3.840,  -4.442,  -4.714,  -4.454,  -3.923,  -3.681,  -3.875,  -4.100,
  -4.070,  -3.934,  -3.864,  -3.837,  -3.976,  -4.561,  -5.528,  -6.410,  -6.949,  -7.358,
  -7.868,  -8.409,  -8.834,  -9.095,  -9.129,  -8.850,  -8.239,  -7.227,  -5.724,  -4.111,
  -3.375,  -4.212,  -6.173,  -8.160,  -9.569, -10.496, -11.098, -11.378, -11.496, -11.694,
 -11.918, -11.898, -11.633, -11.418, -11.427, -11.582, -11.798, -12.109, -12.557, -13.096,
 -13.603, -13.940, -14.051, -13.995, -13.827, -13.517, -13.056, -12.482, -11.739, -10.768,
  -9.762,  -9.037,  -8.569,  -8.050,  -7.441,  -7.006,  -6.718,  -6.152,  -5.127,  -3.988,
  -3.021,  -2.024,  -0.726,   0.753,   2.204,   3.763,   5.547,   6.980,   7.147,   6.127,
   5.445,   6.580,   9.291,  12.036,  13.871,  15.117,  16.256,  17.160,  17.691,  18.173,
  18.804,  19.240,  19.164,  18.752,  18.246,  17.600,  16.928,  16.791,  17.579,  18.941,
  20.206,  21.156,  22.084,  23.293,  24.827,  26.565,  28.303,  29.798,  30.900,  31.651,
  32.213,  32.760,  33.417,  34.200,  34.965,  35.530,  35.908,  36.314,  36.883,  37.476,
  37.907,  38.291,  38.966,  40.012,  41.079,  41.816,  42.300,  42.878,  43.743,  44.749,
  45.510,  45.668,  45.318,  45.163,  45.834,  46.974,  47.452,  46.664,  45.233,  44.217,
  43.999,  44.258,  44.692,  45.317,  46.125,  46.823,  47.057,  46.784,  46.289,  45.894,
  45.720,  45.699,  45.686,  45.581,  45.418,  45.286,  45.108,  44.677,  44.066,  43.723},
{
  43.370,  43.379,  42.920,  42.567,  42.642,  42.818,  42.637,  42.155,  41.729,  41.334,
  40.599,  39.610,  39.113,  39.522,  40.056,  39.403,  37.215,  34.492,  32.418,  31.231,
  30.399,  29.511,  28.607,  27.800,  26.994,  26.068,  25.076,  24.124,  23.166,  22.082,
  20.910,  19.845,  19.013,  18.338,  17.656,  16.887,  16.049,  15.168,  14.236,  13.243,
  12.222,  11.274,  10.518,   9.975,   9.503,   8.934,   8.269,   7.652,   7.129,   6.563,
   5.867,   5.199,   4.776,   4.559,   4.279,   3.769,   3.122,   2.520,   2.081,   1.877,
   1.921,   2.026,   1.875,   1.333,   0.595,  -0.093,  -0.728,  -1.443,  -2.226,  -2.912,
  -3.400,  -3.756,  -4.108,  -4.523,  -4.995,  -5.483,  -5.953,  -6.381,  -6.698,  -6.779,
  -6.579,  -6.286,  -6.206,  -6.467,  -6.911,  -7.298,  -7.509,  -7.534,  -7.416,  -7.270,
  -7.223,  -7.243,  -7.162,  -6.955,  -6.879,  -7.164,  -7.662,  -7.998,  -8.019,  -7.918,
  -7.880,  -7.815,  -7.546,  -7.096,  -6.644,  -6.264,  -5.849,  -5.288,  -4.637,  -4.075,
  -3.732,  -3.546,  -3.330,  -3.013,  -2.773,  -2.821,  -3.067,  -3.144,  -2.847,  -2.388,
  -2.104,  -2.014,  -1.841,  -1.438,  -1.009,  -0.818,  -0.830,  -0.777,  -0.529,  -0.249,
  -0.146,  -0.198,  -0.252,  -0.291,  -0.425,  -0.585,  -0.466,   0.071,   0.693,   0.922,
   0.715,   0.498,   0.589,   0.846,   0.983,   1.017,   1.176,   1.505,   1.829,   2.050,
   2.243,   2.420,   2.447,   2.279,   2.101,   2.076,   2.129,   2.135,   2.191,   2.476,
   2.886,   3.084,   2.932,   2.666,   2.560,   2.608,   2.648,   2.636,   2.632,   2.665,
   2.733,   2.865,   3.021,   3.025,   2.748,   2.296,   1.879,   1.567,   1.343,   1.305,   1.546,
   1.811,   1.587,   0.723,  -0.241,  -0.612,  -0.312,   0.135,   0.341,   0.449,   0.729,
   1.116,   1.371,   1.460,   1.449,   1.182,   0.444,  -0.559,  -1.244,  -1.285,  -0.954,
  -0.740,  -0.877,  -1.344,  -2.031,  -2.718,  -3.097,  -3.076,  -2.931,  -2.985,  -3.232,
  -3.459,  -3.604,  -3.812,  -4.178,  -4.650,  -5.148,  -5.640,  -6.096,  -6.499,  -6.878,
  -7.270,  -7.653,  -7.961,  -8.123,  -8.029,  -7.560,  -6.761,  -5.909,  -5.296,  -4.983,
  -4.866,  -4.965,  -5.454,  -6.338,  -7.240,  -7.700,  -7.665,  -7.540,  -7.702,  -8.097,
  -8.399,  -8.437,  -8.300,  -8.112,  -7.954,  -7.995,  -8.438,  -9.205,  -9.859, -10.041,
  -9.865,  -9.732,  -9.817,  -9.908,  -9.739,  -9.303,  -8.774,  -8.261,  -7.734,  -7.165,
  -6.625,  -6.211,  -5.923,  -5.644,  -5.223,  -4.568,  -3.697,  -2.766,  -2.019,  -1.623,
  -1.521,  -1.467,  -1.233,  -0.739,  -0.012,   0.886,   1.840,   2.731,   3.595,   4.686,
   6.209,   8.024,   9.756,  11.192,  12.398,  13.422,  14.128,  14.470,  14.726,  15.212,
  15.881,  16.433,  16.757,  17.045,  17.480,  18.046,  18.661,  19.307,  19.965,  20.621,
  21.378,  22.427,  23.818,  25.360,  26.805,  28.030,  29.025,  29.830,  30.526,  31.171,
  31.708,  32.038,  32.231,  32.550,  33.166,  33.985,  34.824,  35.664,  36.587,  37.543,
  38.378,  39.062,  39.748,  40.532,  41.334,  42.060,  42.804,  43.732,  44.857,  45.970,
  46.743,  46.818,  45.919,  44.090,  41.835,  39.898,  38.762,  38.350,  38.217,  37.992,
  37.676,  37.615,  38.165,  39.329,  40.665,  41.645,  42.162,  42.595,  43.304,  44.143,
  44.614,  44.469,  43.961,  43.461,  43.036,  42.578,  42.212,  42.277,  42.825,  43.370},
{
  41.560,  41.425,  41.345,  41.466,  41.609,  41.516,  41.178,  40.799,  40.488,  40.126,
  39.589,  38.958,  38.369,  37.711,  36.661,  35.074,  33.223,  31.588,  30.438,  29.706,
  29.188,  28.740,  28.277,  27.722,  27.029,  26.207,  25.285,  24.263,  23.159,  22.047,
  21.007,  20.038,  19.095,  18.197,  17.413,  16.748,  16.098,  15.381,  14.641,  13.961,
  13.316,  12.609,  11.833,  11.113,  10.530,   9.998,   9.374,   8.659,   7.989,   7.452,
   6.981,   6.469,   5.914,   5.406,   4.993,   4.618,   4.187,   3.673,   3.133,   2.635,
   2.198,   1.800,   1.454,   1.240,   1.220,   1.316,   1.293,   0.914,   0.129,  -0.877,
  -1.811,  -2.451,  -2.754,  -2.844,  -2.906,  -3.056,  -3.277,  -3.466,  -3.552,  -3.581,
  -3.689,  -3.980,  -4.432,  -4.910,  -5.271,  -5.464,  -5.525,  -5.528,  -5.533,  -5.577,
  -5.675,  -5.813,  -5.973,  -6.153,  -6.348,  -6.527,  -6.661,  -6.785,  -6.964,  -7.148,
  -7.115,  -6.678,  -5.939,  -5.239,  -4.806,  -4.538,  -4.225,  -3.883,  -3.703,  -3.688,
  -3.528,  -2.975,  -2.254,  -1.874,  -2.028,  -2.324,  -2.220,  -1.609,  -0.860,  -0.329,
  -0.048,   0.102,   0.107,  -0.110,  -0.421,  -0.520,  -0.266,   0.136,   0.416,   0.584,
   0.866,   1.338,   1.753,   1.795,   1.405,   0.817,   0.332,   0.146,   0.317,   0.756,
   1.216,   1.420,   1.307,   1.125,   1.191,   1.553,   1.947,   2.101,   2.007,   1.866,
   1.849,   1.972,   2.172,   2.366,   2.463,   2.424,   2.365,   2.493,   2.846,   3.189,
   3.279,   3.189,   3.213,   3.428,   3.572,   3.446,   3.260,   3.379,   3.781,   4.034,
   3.839,   3.390,   3.061,   2.920,   2.739,   2.429,   2.181,   2.163,   2.270,   2.294,   2.211,
   2.129,   2.058,   1.870,   1.508,   1.081,   0.733,   0.554,   0.619,   0.985,   1.556,
   2.065,   2.317,   2.394,   2.476,   2.519,   2.295,   1.739,   1.095,   0.605,   0.225,
  -0.207,  -0.646,  -0.902,  -0.999,  -1.248,  -1.822,  -2.447,  -2.686,  -2.510,  -2.362,
  -2.636,  -3.231,  -3.754,  -4.048,  -4.329,  -4.799,  -5.331,  -5.665,  -5.796,  -5.991,
  -6.411,  -6.893,  -7.166,  -7.157,  -6.988,  -6.756,  -6.463,  -6.119,  -5.738,  -5.223,
  -4.429,  -3.449,  -2.723,  -2.687,  -3.321,  -4.168,  -4.796,  -5.125,  -5.296,  -5.397,
  -5.445,  -5.534,  -5.819,  -6.327,  -6.891,  -7.327,  -7.597,  -7.779,  -7.923,  -8.003,
  -7.944,  -7.678,  -7.213,  -6.719,  -6.464,  -6.565,  -6.823,  -6.907,  -6.699,  -6.348,
  -5.945,  -5.329,  -4.348,  -3.239,  -2.519,  -2.428,  -2.638,  -2.603,  -2.095,  -1.251,
  -0.175,   1.255,   3.063,   4.846,   6.021,   6.436,   6.583,   7.076,   8.010,   8.922,
   9.362,   9.361,   9.300,   9.441,   9.741,  10.082,  10.542,  11.340,  12.576,  14.105,
  15.705,  17.255,  18.736,  20.112,  21.296,  22.229,  22.917,  23.394,  23.702,  23.946,
  24.295,  24.872,  25.651,  26.509,  27.356,  28.185,  28.993,  29.718,  30.260,  30.564,
  30.678,  30.772,  31.071,  31.700,  32.558,  33.395,  34.053,  34.597,  35.198,  35.921,
  36.709,  37.528,  38.406,  39.338,  40.226,  40.989,  41.659,  42.271,  42.734,  42.917,
  42.857,  42.698,  42.341,  41.278,  39.028,  35.842,  32.837,  31.251,  31.438,  32.594,
  33.470,  33.415,  32.824,  32.679,  33.675,  35.670,  37.859,  39.434,  40.170,  40.466,
  40.877,  41.566,  42.189,  42.295,  41.825,  41.210,  40.953,  41.153,  41.472,  41.560},
{
  38.495,  38.629,  38.921,  39.437,  39.981,  40.197,  39.905,  39.322,  38.887,  38.836,
  38.901,  38.492,  37.226,  35.339,  33.539,  32.406,  31.917,  31.558,  30.915,  30.074,
  29.428,  29.137,  28.892,  28.234,  27.053,  25.669,  24.466,  23.520,  22.636,  21.664,
  20.679,  19.852,  19.220,  18.663,  18.068,  17.447,  16.874,  16.341,  15.745,  15.017,
  14.226,  13.523,  12.990,  12.556,  12.083,  11.507,  10.881,  10.294,   9.759,   9.216,
   8.625,   8.024,   7.485,   7.036,   6.645,   6.269,   5.876,   5.438,   4.925,   4.358,
   3.830,   3.444,   3.195,   2.947,   2.559,   2.020,   1.435,   0.905,   0.432,  -0.034,
  -0.493,  -0.893,  -1.201,  -1.436,  -1.616,  -1.721,  -1.738,  -1.753,  -1.904,  -2.244,
  -2.661,  -2.990,  -3.193,  -3.383,  -3.663,  -3.979,  -4.183,  -4.209,  -4.148,  -4.124,
  -4.157,  -4.180,  -4.167,  -4.189,  -4.323,  -4.558,  -4.821,  -5.063,  -5.252,  -5.290,
  -5.014,  -4.355,  -3.474,  -2.676,  -2.158,  -1.866,  -1.630,  -1.387,  -1.239,  -1.296,
  -1.518,  -1.740,  -1.817,  -1.726,  -1.566,  -1.506,  -1.699,  -2.166,  -2.715,  -3.039,
  -2.960,  -2.590,  -2.203,  -1.977,  -1.884,  -1.848,  -1.873,  -1.947,  -1.874,  -1.361,
  -0.353,   0.799,   1.612,   1.928,   2.019,   2.241,   2.635,   2.935,   2.897,   2.534,
   2.065,   1.717,   1.612,   1.746,   1.989,   2.135,   2.073,   1.943,   2.041,   2.487,
   3.023,   3.226,   2.952,   2.508,   2.342,   2.599,   3.036,   3.325,   3.380,   3.348,
   3.367,   3.404,   3.355,   3.223,   3.157,   3.291,   3.568,   3.762,   3.684,   3.377,
   3.065,   2.902,   2.801,   2.556,   2.141,   1.829,   1.942,   2.499,   3.145,   3.448,   3.267,
   2.814,   2.422,   2.272,   2.316,   2.379,   2.333,   2.198,   2.131,   2.297,   2.721,
   3.255,   3.686,   3.887,   3.852,   3.627,   3.252,   2.772,   2.239,   1.678,   1.078,
   0.480,   0.028,  -0.151,  -0.156,  -0.328,  -0.944,  -1.889,  -2.701,  -2.994,  -2.828,
  -2.616,  -2.720,  -3.154,  -3.692,  -4.143,  -4.490,  -4.790,  -5.048,  -5.225,  -5.327,
  -5.428,  -5.592,  -5.802,  -5.976,  -6.034,  -5.963,  -5.812,  -5.632,  -5.421,  -5.114,
  -4.666,  -4.119,  -3.583,  -3.149,  -2.833,  -2.634,  -2.605,  -2.805,  -3.157,  -3.423,
  -3.383,  -3.059,  -2.712,  -2.586,  -2.695,  -2.865,  -2.982,  -3.131,  -3.464,  -3.961,
  -4.393,  -4.527,  -4.349,  -4.062,  -3.868,  -3.808,  -3.808,  -3.832,  -3.886,  -3.885,
  -3.598,  -2.845,  -1.777,  -0.859,  -0.503,  -0.706,  -1.096,  -1.302,  -1.231,  -0.955,
  -0.432,   0.521,   1.969,   3.664,   5.219,   6.455,   7.512,   8.581,   9.604,  10.279,
  10.363,   9.910,   9.246,   8.750,   8.705,   9.264,  10.444,  12.102,  13.949,  15.684,
  17.146,  18.355,  19.433,  20.492,  21.582,  22.695,  23.777,  24.757,  25.597,  26.330,
  27.027,  27.718,  28.363,  28.904,  29.337,  29.701,  30.020,  30.279,  30.470,  30.640,
  30.872,  31.209,  31.627,  32.072,  32.517,  32.957,  33.376,  33.746,  34.074,  34.423,
  34.868,  35.422,  36.017,  36.587,  37.148,  37.761,  38.408,  38.917,  39.058,  38.732,
  38.043,  37.174,  36.197,  35.040,  33.630,  32.065,  30.627,  29.653,  29.351,  29.676,
  30.301,  30.765,  30.784,  30.528,  30.568,  31.400,  32.949,  34.568,  35.616,  36.056,
  36.433,  37.259,  38.452,  39.420,  39.673,  39.279,  38.726,  38.423,  38.405,  38.495},
{
  36.215,  36.716,  36.961,  37.085,  37.323,  37.757,  38.250,  38.584,  38.637,  38.412,
  37.953,  37.293,  36.511,  35.782,  35.295,  35.063,  34.829,  34.236,  33.125,  31.689,
  30.305,  29.215,  28.367,  27.549,  26.650,  25.749,  24.979,  24.345,  23.723,  23.020,
  22.283,  21.624,  21.063,  20.490,  19.790,  18.993,  18.255,  17.689,  17.239,  16.737,
  16.084,  15.364,  14.761,  14.384,  14.160,  13.899,  13.454,  12.817,  12.089,  11.375,
  10.714,  10.092,   9.504,   8.993,   8.612,   8.354,   8.126,   7.812,   7.368,   6.871,
   6.441,   6.107,   5.758,   5.247,   4.554,   3.832,   3.283,   2.968,   2.765,   2.496,
   2.111,   1.710,   1.422,   1.273,   1.181,   1.069,   0.935,   0.817,   0.711,   0.560,
   0.341,   0.119,   0.004,   0.029,   0.086,   0.012,  -0.270,  -0.680,  -1.049,  -1.256,
  -1.307,  -1.296,  -1.304,  -1.341,  -1.380,  -1.413,  -1.468,  -1.554,  -1.612,  -1.533,
  -1.243,  -0.780,  -0.270,   0.163,   0.479,   0.710,   0.889,   0.984,   0.898,   0.530,
  -0.158,  -1.088,  -2.053,  -2.785,  -3.087,  -2.966,  -2.634,  -2.366,  -2.293,  -2.338,
  -2.346,  -2.266,  -2.207,  -2.296,  -2.506,  -2.653,  -2.569,  -2.276,  -1.968,  -1.818,
  -1.811,  -1.769,  -1.556,  -1.241,  -1.050,  -1.125,  -1.319,  -1.247,  -0.592,   0.600,
   1.900,   2.806,   3.114,   3.017,   2.885,   2.927,   3.067,   3.096,   2.921,   2.644,
   2.452,   2.438,   2.565,   2.745,   2.943,   3.174,   3.433,   3.632,   3.631,   3.340,
   2.810,   2.227,   1.796,   1.595,   1.513,   1.339,   0.949,   0.441,   0.073,   0.045,
   0.332,   0.705,   0.952,   1.070,   1.234,   1.599,   2.117,   2.586,   2.855,   2.968,   3.097,
   3.335,   3.586,   3.669,   3.512,   3.223,   2.976,   2.836,   2.731,   2.577,   2.408,
   2.352,   2.487,   2.733,   2.907,   2.890,   2.723,   2.530,   2.350,   2.069,   1.527,
   0.715,  -0.168,  -0.862,  -1.261,  -1.475,  -1.684,  -1.932,  -2.100,  -2.070,  -1.896,
  -1.788,  -1.923,  -2.276,  -2.662,  -2.920,  -3.046,  -3.154,  -3.328,  -3.541,  -3.712,
  -3.808,  -3.870,  -3.943,  -4.022,  -4.066,  -4.070,  -4.078,  -4.117,  -4.147,  -4.096,
  -3.956,  -3.827,  -3.827,  -3.951,  -4.031,  -3.867,  -3.391,  -2.720,  -2.035,  -1.430,
  -0.876,  -0.322,   0.186,   0.503,   0.504,   0.194,  -0.283,  -0.742,  -1.070,  -1.258,
  -1.351,  -1.362,  -1.227,  -0.857,  -0.250,   0.425,   0.896,   0.974,   0.735,   0.492,
   0.548,   0.946,   1.454,   1.841,   2.160,   2.720,   3.753,   5.082,   6.161,   6.491,
   6.028,   5.222,   4.660,   4.685,   5.291,   6.303,   7.566,   8.945,  10.211,  11.028,
  11.134,  10.537,   9.527,   8.475,   7.626,   7.061,   6.829,   7.041,   7.816,   9.134,
  10.777,  12.454,  13.990,  15.408,  16.817,  18.250,  19.619,  20.810,  21.821,  22.763,
  23.739,  24.737,  25.642,  26.360,  26.897,  27.333,  27.742,  28.148,  28.566,  29.038,
  29.608,  30.266,  30.935,  31.538,  32.070,  32.596,  33.162,  33.737,  34.233,  34.597,
  34.869,  35.139,  35.469,  35.841,  36.195,  36.475,  36.650,  36.684,  36.512,  36.057,
  35.289,  34.265,  33.114,  31.981,  30.955,  30.052,  29.264,  28.635,  28.283,  28.326,
  28.752,  29.373,  29.917,  30.227,  30.356,  30.499,  30.811,  31.315,  31.977,  32.826,
  33.936,  35.256,  36.485,  37.183,  37.096,  36.408,  35.655,  35.351,  35.629,  36.215},
{
  33.895,  34.188,  34.468,  34.733,  34.987,  35.205,  35.352,  35.413,  35.418,  35.411,
  35.405,  35.353,  35.172,  34.795,  34.207,  33.444,  32.579,  31.701,  30.905,  30.265,
  29.804,  29.462,  29.120,  28.662,  28.038,  27.288,  26.508,  25.790,  25.182,  24.685,
  24.270,  23.889,  23.487,  23.014,  22.451,  21.816,  21.146,  20.464,  19.762,  19.029,
  18.290,  17.619,  17.085,  16.698,  16.381,  16.019,  15.546,  14.986,  14.422,  13.911,
  13.438,  12.939,  12.384,  11.826,  11.373,  11.096,  10.954,  10.809,  10.517,  10.033,
   9.427,   8.821,   8.295,   7.842,   7.401,   6.917,   6.383,   5.827,   5.276,   4.746,
   4.259,   3.864,   3.624,   3.571,   3.661,   3.780,   3.793,   3.623,   3.280,   2.846,
   2.407,   2.005,   1.630,   1.267,   0.949,   0.751,   0.739,   0.907,   1.160,   1.378,
   1.501,   1.563,   1.648,   1.800,   1.985,   2.124,   2.180,   2.188,   2.226,   2.329,
   2.463,   2.552,   2.551,   2.466,   2.311,   2.062,   1.668,   1.119,   0.494,  -0.063,
  -0.449,  -0.669,  -0.810,  -0.938,  -1.036,  -1.028,  -0.885,  -0.680,  -0.535,  -0.509,
  -0.539,  -0.497,  -0.317,  -0.064,   0.111,   0.095,  -0.106,  -0.388,  -0.643,  -0.822,
  -0.931,  -0.981,  -0.964,  -0.877,  -0.757,  -0.672,  -0.659,  -0.667,  -0.572,  -0.261,
   0.277,   0.931,   1.529,   1.941,   2.142,   2.215,   2.286,   2.449,   2.717,   3.012,
   3.203,   3.179,   2.912,   2.483,   2.033,   1.680,   1.450,   1.282,   1.094,   0.862,
   0.634,   0.483,   0.438,   0.459,   0.496,   0.547,   0.677,   0.954,   1.355,   1.747,
   1.962,   1.930,   1.749,   1.635,   1.779,   2.219,   2.834,   3.440,   3.911,   4.222,   4.410,
   4.493,   4.451,   4.258,   3.933,   3.546,   3.182,   2.899,   2.719,   2.649,   2.677,
   2.756,   2.791,   2.672,   2.346,   1.873,   1.398,   1.051,   0.866,   0.769,   0.658,
   0.492,   0.310,   0.173,   0.092,   0.026,  -0.069,  -0.175,  -0.231,  -0.191,  -0.090,
  -0.025,  -0.078,  -0.250,  -0.469,  -0.658,  -0.802,  -0.945,  -1.133,  -1.355,  -1.548,
  -1.647,  -1.649,  -1.616,  -1.642,  -1.791,  -2.069,  -2.417,  -2.744,  -2.961,  -3.014,
  -2.913,  -2.728,  -2.559,  -2.477,  -2.476,  -2.463,  -2.314,  -1.944,  -1.361,  -0.649,
   0.092,   0.817,   1.552,   2.344,   3.179,   3.949,   4.489,   4.679,   4.522,   4.132,
   3.657,   3.184,   2.718,   2.241,   1.796,   1.508,   1.516,   1.878,   2.512,   3.250,
   3.947,   4.587,   5.296,   6.247,   7.516,   8.982,  10.335,  11.211,  11.379,  10.855,
   9.892,   8.851,   8.052,   7.703,   7.905,   8.685,   9.984,  11.619,  13.270,  14.574,
  15.267,  15.291,  14.774,  13.902,  12.802,  11.528,  10.145,   8.811,   7.768,   7.236,
   7.300,   7.888,   8.835,   9.987,  11.250,  12.592,  14.001,  15.466,  16.966,  18.462,
  19.893,  21.184,  22.289,  23.221,  24.063,  24.904,  25.774,  26.623,  27.367,  27.971,
  28.485,  29.014,  29.639,  30.365,  31.126,  31.843,  32.469,  33.002,  33.461,  33.872,
  34.268,  34.700,  35.219,  35.841,  36.501,  37.061,  37.369,  37.338,  36.987,  36.412,
  35.719,  34.970,  34.184,  33.393,  32.679,  32.153,  31.888,  31.851,  31.910,  31.914,
  31.791,  31.581,  31.395,  31.317,  31.348,  31.426,  31.501,  31.607,  31.853,  32.334,
  33.031,  33.777,  34.336,  34.533,  34.365,  34.000,  33.672,  33.545,  33.647,  33.895},
{
  32.344,  32.741,  32.717,  32.390,  31.998,  31.739,  31.672,  31.721,  31.760,  31.706,
  31.561,  31.390,  31.258,  31.184,  31.135,  31.056,  30.905,  30.669,  30.357,  29.980,
  29.555,  29.105,  28.674,  28.307,  28.028,  27.814,  27.589,  27.267,  26.800,  26.210,
  25.578,  24.998,  24.520,  24.121,  23.732,  23.282,  22.738,  22.115,  21.456,  20.803,
  20.184,  19.621,  19.136,  18.754,  18.480,  18.280,  18.083,  17.815,  17.435,  16.961,
  16.458,  15.993,  15.604,  15.274,  14.962,  14.637,  14.304,  13.996,  13.737,  13.505,
  13.232,  12.836,  12.274,  11.578,  10.840,  10.160,   9.593,   9.124,   8.696,   8.264,
   7.837,   7.469,   7.220,   7.100,   7.061,   7.025,   6.936,   6.790,   6.620,   6.449,
   6.254,   5.984,   5.602,   5.132,   4.663,   4.292,   4.068,   3.965,   3.917,   3.884,
   3.895,   4.025,   4.318,   4.732,   5.142,   5.410,   5.475,   5.386,   5.257,   5.181,
   5.160,   5.103,   4.889,   4.446,   3.790,   3.018,   2.251,   1.594,   1.103,   0.788,
   0.629,   0.593,   0.654,   0.796,   1.002,   1.243,   1.464,   1.598,   1.603,   1.498,
   1.371,   1.340,   1.477,   1.759,   2.068,   2.258,   2.237,   2.013,   1.678,   1.342,
   1.072,   0.875,   0.729,   0.632,   0.621,   0.755,   1.070,   1.543,   2.093,   2.610,
   2.994,   3.185,   3.168,   2.971,   2.658,   2.314,   2.029,   1.867,   1.837,   1.891,
   1.950,   1.952,   1.887,   1.798,   1.740,   1.736,   1.756,   1.741,   1.653,   1.506,
   1.369,   1.313,   1.374,   1.529,   1.723,   1.907,   2.062,   2.200,   2.334,   2.458,
   2.555,   2.617,   2.662,   2.733,   2.872,   3.089,   3.357,   3.624,   3.841,   3.986,   4.058,
   4.072,   4.035,   3.945,   3.800,   3.611,   3.408,   3.232,   3.116,   3.064,   3.052,
   3.042,   3.005,   2.937,   2.853,   2.765,   2.667,   2.535,   2.346,   2.111,   1.883,
   1.725,   1.673,   1.699,   1.730,   1.685,   1.538,   1.326,   1.127,   1.007,   0.980,
   1.010,   1.042,   1.039,   0.998,   0.940,   0.883,   0.841,   0.824,   0.842,   0.907,
   1.017,   1.146,   1.256,   1.316,   1.316,   1.268,   1.186,   1.073,   0.923,   0.737,
   0.545,   0.390,   0.307,   0.299,   0.332,   0.366,   0.385,   0.412,   0.489,   0.639,
   0.854,   1.103,   1.368,   1.665,   2.033,   2.504,   3.067,   3.674,   4.266,   4.800,
   5.261,   5.642,   5.922,   6.066,   6.058,   5.930,   5.773,   5.698,   5.788,   6.065,
   6.495,   7.030,   7.644,   8.333,   9.081,   9.824,  10.455,  10.874,  11.051,  11.074,
  11.121,  11.388,  11.996,  12.934,  14.058,  15.153,  16.008,  16.492,  16.601,  16.463,
  16.292,  16.311,  16.667,  17.351,  18.192,  18.910,  19.218,  18.929,  18.024,  16.643,
  15.040,  13.500,  12.272,  11.525,  11.330,  11.670,  12.457,  13.563,  14.853,  16.202,
  17.515,  18.724,  19.791,  20.705,  21.479,  22.148,  22.761,  23.367,  24.002,  24.682,
  25.403,  26.150,  26.916,  27.705,  28.539,  29.441,  30.416,  31.436,  32.430,  33.305,
  33.973,  34.385,  34.548,  34.530,  34.436,  34.379,  34.437,  34.620,  34.856,  35.010,
  34.936,  34.552,  33.889,  33.092,  32.352,  31.818,  31.525,  31.394,  31.301,  31.170,
  31.012,  30.904,  30.907,  31.006,  31.124,  31.193,  31.229,  31.341,  31.647,  32.155,
  32.700,  33.010,  32.862,  32.247,  31.408,  30.733,  30.541,  30.908,  31.629,  32.344},
{
  30.978,  31.166,  31.112,  30.947,  30.825,  30.846,  31.018,  31.268,  31.488,  31.596,
  31.559,  31.388,  31.105,  30.720,  30.225,  29.615,  28.914,  28.187,  27.525,  27.005,
  26.652,  26.424,  26.224,  25.942,  25.508,  24.916,  24.233,  23.562,  22.999,  22.593,
  22.328,  22.138,  21.942,  21.677,  21.329,  20.924,  20.521,  20.179,  19.944,  19.837,
  19.860,  19.992,  20.185,  20.366,  20.440,  20.321,  19.964,  19.400,  18.736,  18.121,
  17.683,  17.470,  17.432,  17.451,  17.408,  17.242,  16.978,  16.692,  16.448,  16.245,
  16.004,  15.611,  14.989,  14.149,  13.202,  12.306,  11.601,  11.148,  10.917,  10.823,
  10.773,  10.704,  10.595,  10.444,  10.246,   9.989,   9.662,   9.280,   8.891,   8.561,
   8.341,   8.236,   8.203,   8.165,   8.060,   7.871,   7.646,   7.470,   7.422,   7.525,
   7.724,   7.898,   7.903,   7.637,   7.089,   6.350,   5.587,   4.974,   4.626,   4.556,
   4.670,   4.816,   4.850,   4.699,   4.383,   4.002,   3.682,   3.520,   3.541,   3.698,
   3.901,   4.065,   4.146,   4.157,   4.149,   4.168,   4.227,   4.293,   4.311,   4.236,
   4.067,   3.849,   3.648,   3.512,   3.445,   3.406,   3.342,   3.226,   3.077,   2.949,
   2.894,   2.920,   2.988,   3.034,   3.017,   2.959,   2.948,   3.094,   3.463,   4.022,
   4.635,   5.113,   5.289,   5.103,   4.623,   4.017,   3.476,   3.130,   3.005,   3.025,
   3.066,   3.025,   2.863,   2.616,   2.364,   2.187,   2.127,   2.178,   2.304,   2.460,
   2.615,   2.758,   2.890,   3.013,   3.123,   3.210,   3.264,   3.285,   3.282,   3.277,
   3.291,   3.343,   3.439,   3.570,   3.718,   3.856,   3.965,   4.039,   4.088,   4.131,   4.186,
   4.256,   4.328,   4.374,   4.370,   4.306,   4.194,   4.064,   3.950,   3.876,   3.841,
   3.830,   3.816,   3.782,   3.725,   3.661,   3.617,   3.614,   3.661,   3.744,   3.837,
   3.906,   3.929,   3.905,   3.848,   3.784,   3.734,   3.700,   3.665,   3.600,   3.484,
   3.315,   3.115,   2.919,   2.759,   2.650,   2.586,   2.552,   2.536,   2.539,   2.569,
   2.635,   2.731,   2.839,   2.937,   3.013,   3.071,   3.132,   3.217,   3.333,   3.469,
   3.597,   3.689,   3.732,   3.735,   3.724,   3.731,   3.773,   3.843,   3.915,   3.952,
   3.933,   3.864,   3.782,   3.741,   3.785,   3.931,   4.157,   4.419,   4.672,   4.894,
   5.097,   5.313,   5.587,   5.950,   6.422,   7.001,   7.672,   8.399,   9.129,   9.791,
  10.319,  10.679,  10.892,  11.033,  11.202,  11.476,  11.870,  12.326,  12.755,  13.091,
  13.338,  13.581,  13.936,  14.484,  15.214,  16.023,  16.771,  17.361,  17.799,  18.191,
  18.678,  19.339,  20.127,  20.875,  21.375,  21.481,  21.193,  20.655,  20.082,  19.653,
  19.426,  19.318,  19.166,  18.825,  18.244,  17.494,  16.732,  16.126,  15.789,  15.746,
  15.946,  16.301,  16.725,  17.158,  17.570,  17.951,  18.301,  18.629,  18.959,  19.333,
  19.809,  20.446,  21.285,  22.335,  23.559,  24.887,  26.224,  27.475,  28.568,  29.464,
  30.162,  30.688,  31.083,  31.383,  31.610,  31.766,  31.842,  31.828,  31.717,  31.511,
  31.220,  30.850,  30.405,  29.881,  29.275,  28.601,  27.901,  27.253,  26.764,  26.544,
  26.667,  27.148,  27.922,  28.864,  29.823,  30.661,  31.283,  31.638,  31.707,  31.496,
  31.036,  30.396,  29.699,  29.111,  28.793,  28.845,  29.252,  29.877,  30.516,  30.978},
{
  28.477,  28.361,  28.276,  28.200,  28.104,  27.965,  27.779,  27.559,  27.340,  27.162,
  27.055,  27.031,  27.072,  27.135,  27.167,  27.120,  26.970,  26.717,  26.390,  26.028,
  25.672,  25.349,  25.069,  24.823,  24.594,  24.357,  24.092,  23.787,  23.439,  23.058,
  22.666,  22.289,  21.950,  21.659,  21.404,  21.159,  20.889,  20.569,  20.193,  19.785,
  19.385,  19.031,  18.738,  18.488,  18.236,  17.928,  17.534,  17.066,  16.576,  16.140,
  15.825,  15.657,  15.616,  15.643,  15.669,  15.647,  15.567,  15.450,  15.327,  15.217,
  15.105,  14.950,  14.705,  14.342,  13.870,  13.334,  12.798,  12.321,  11.934,  11.635,
  11.398,  11.186,  10.974,  10.750,  10.519,  10.284,  10.046,   9.792,   9.508,   9.178,
   8.806,   8.406,   8.011,   7.652,   7.356,   7.135,   6.985,   6.891,   6.832,   6.791,
   6.758,   6.731,   6.714,   6.711,   6.724,   6.747,   6.766,   6.768,   6.740,   6.679,
   6.600,   6.530,   6.502,   6.543,   6.658,   6.828,   7.003,   7.126,   7.150,   7.063,
   6.895,   6.712,   6.597,   6.610,   6.768,   7.029,   7.306,   7.497,   7.523,   7.357,
   7.029,   6.614,   6.199,   5.853,   5.607,   5.454,   5.362,   5.299,   5.242,   5.183,
   5.117,   5.033,   4.915,   4.749,   4.547,   4.357,   4.264,   4.362,   4.716,   5.325,
   6.102,   6.888,   7.496,   7.773,   7.647,   7.153,   6.420,   5.626,   4.942,   4.477,
   4.260,   4.242,   4.331,   4.430,   4.473,   4.437,   4.341,   4.222,   4.114,   4.035,
   3.979,   3.931,   3.875,   3.810,   3.751,   3.719,   3.734,   3.798,   3.897,   4.003,
   4.087,   4.132,   4.142,   4.138,   4.153,   4.214,   4.332,   4.502,   4.698,   4.888,   5.041,
   5.137,   5.168,   5.142,   5.075,   4.988,   4.903,   4.840,   4.813,   4.828,   4.883,
   4.967,   5.064,   5.159,   5.242,   5.310,   5.365,   5.410,   5.445,   5.467,   5.470,
   5.452,   5.418,   5.381,   5.361,   5.376,   5.432,   5.522,   5.625,   5.714,   5.768,
   5.779,   5.755,   5.711,   5.665,   5.624,   5.581,   5.524,   5.436,   5.313,   5.165,
   5.016,   4.899,   4.842,   4.857,   4.942,   5.074,   5.223,   5.360,   5.467,   5.540,
   5.595,   5.654,   5.740,   5.866,   6.032,   6.219,   6.404,   6.562,   6.676,   6.749,
   6.793,   6.831,   6.885,   6.969,   7.086,   7.234,   7.406,   7.597,   7.802,   8.015,
   8.226,   8.421,   8.589,   8.728,   8.853,   8.998,   9.200,   9.485,   9.858,  10.292,
  10.743,  11.167,  11.545,  11.889,  12.240,  12.644,  13.125,  13.671,  14.232,  14.742,
  15.153,  15.456,  15.687,  15.911,  16.190,  16.550,  16.968,  17.384,  17.728,  17.953,
  18.060,  18.097,  18.139,  18.253,  18.475,  18.792,  19.151,  19.487,  19.744,  19.898,
  19.961,  19.967,  19.959,  19.966,  19.997,  20.042,  20.081,  20.104,  20.111,  20.117,
  20.146,  20.214,  20.326,  20.472,  20.634,  20.797,  20.959,  21.135,  21.349,  21.627,
  21.980,  22.401,  22.867,  23.347,  23.820,  24.279,  24.737,  25.216,  25.738,  26.308,
  26.910,  27.512,  28.075,  28.566,  28.970,  29.291,  29.545,  29.752,  29.920,  30.042,
  30.098,  30.063,  29.919,  29.669,  29.337,  28.971,  28.626,  28.353,  28.175,  28.088,
  28.057,  28.034,  27.970,  27.839,  27.652,  27.448,  27.292,  27.243,  27.341,  27.587,
  27.941,  28.337,  28.698,  28.961,  29.090,  29.087,  28.978,  28.811,  28.632,  28.477},
{
  27.419,  27.379,  27.259,  27.066,  26.822,  26.564,  26.327,  26.134,  25.997,  25.906,
  25.843,  25.781,  25.702,  25.595,  25.460,  25.308,  25.151,  25.000,  24.860,  24.728,
  24.597,  24.459,  24.310,  24.147,  23.975,  23.797,  23.615,  23.429,  23.237,  23.034,
  22.817,  22.586,  22.345,  22.099,  21.850,  21.602,  21.352,  21.097,  20.835,  20.563,
  20.285,  20.005,  19.728,  19.458,  19.196,  18.938,  18.678,  18.412,  18.136,  17.851,
  17.561,  17.273,  16.992,  16.723,  16.467,  16.220,  15.977,  15.734,  15.486,  15.231,
  14.970,  14.710,  14.455,  14.212,  13.986,  13.778,  13.583,  13.394,  13.203,  13.003,
  12.790,  12.566,  12.336,  12.109,  11.890,  11.683,  11.486,  11.294,  11.099,  10.898,
  10.692,  10.486,  10.289,  10.112,   9.961,   9.839,   9.745,   9.674,   9.623,   9.588,
   9.569,   9.566,   9.573,   9.585,   9.588,   9.572,   9.531,   9.465,   9.387,   9.320,
   9.285,   9.299,   9.364,   9.461,   9.560,   9.623,   9.620,   9.541,   9.403,   9.250,
   9.135,   9.108,   9.197,   9.391,   9.644,   9.880,  10.018,   9.989,   9.763,   9.354,
   8.818,   8.239,   7.706,   7.292,   7.034,   6.929,   6.940,   7.004,   7.054,   7.036,
   6.916,   6.695,   6.403,   6.092,   5.828,   5.676,   5.688,   5.886,   6.261,   6.767,
   7.329,   7.859,   8.275,   8.516,   8.560,   8.425,   8.161,   7.838,   7.524,   7.267,
   7.082,   6.956,   6.853,   6.730,   6.556,   6.318,   6.029,   5.718,   5.422,   5.176,
   4.997,   4.891,   4.846,   4.846,   4.873,   4.916,   4.971,   5.037,   5.113,   5.198,
   5.284,   5.364,   5.431,   5.484,   5.528,   5.575,   5.637,   5.725,   5.845,   5.992,   6.156,
   6.326,   6.487,   6.632,   6.758,   6.868,   6.967,   7.057,   7.142,   7.219,   7.287,
   7.346,   7.400,   7.458,   7.531,   7.629,   7.757,   7.908,   8.069,   8.219,   8.334,
   8.395,   8.394,   8.335,   8.232,   8.110,   7.994,   7.908,   7.867,   7.878,   7.938,
   8.038,   8.165,   8.305,   8.441,   8.558,   8.643,   8.685,   8.679,   8.626,   8.538,
   8.427,   8.314,   8.217,   8.153,   8.130,   8.151,   8.212,   8.304,   8.419,   8.544,
   8.671,   8.792,   8.900,   8.993,   9.072,   9.140,   9.207,   9.286,   9.389,   9.525,
   9.695,   9.894,  10.109,  10.327,  10.535,  10.729,  10.912,  11.095,  11.289,  11.503,
  11.737,  11.981,  12.219,  12.434,  12.614,  12.757,  12.872,  12.975,  13.084,  13.212,
  13.362,  13.531,  13.707,  13.881,  14.047,  14.203,  14.356,  14.511,  14.673,  14.841,
  15.012,  15.177,  15.334,  15.484,  15.635,  15.803,  16.005,  16.257,  16.566,  16.931,
  17.337,  17.762,  18.181,  18.568,  18.902,  19.177,  19.395,  19.574,  19.736,  19.905,
  20.101,  20.329,  20.583,  20.847,  21.100,  21.326,  21.518,  21.687,  21.852,  22.038,
  22.268,  22.547,  22.869,  23.211,  23.544,  23.842,  24.094,  24.300,  24.475,  24.639,
  24.807,  24.979,  25.144,  25.278,  25.361,  25.379,  25.337,  25.261,  25.188,  25.160,
  25.207,  25.341,  25.553,  25.813,  26.081,  26.320,  26.503,  26.621,  26.679,  26.696,
  26.691,  26.682,  26.679,  26.680,  26.679,  26.667,  26.634,  26.579,  26.504,  26.421,
  26.344,  26.289,  26.271,  26.299,  26.378,  26.502,  26.656,  26.819,  26.967,  27.081,
  27.150,  27.174,  27.168,  27.153,  27.152,  27.181,  27.242,  27.321,  27.390,  27.419},
{
  25.076,  25.135,  25.151,  25.126,  25.069,  24.992,  24.912,  24.841,  24.790,  24.764,
  24.760,  24.769,  24.780,  24.780,  24.759,  24.710,  24.633,  24.533,  24.421,  24.308,
  24.207,  24.124,  24.061,  24.014,  23.969,  23.911,  23.822,  23.687,  23.496,  23.247,
  22.948,  22.614,  22.264,  21.918,  21.597,  21.310,  21.064,  20.852,  20.667,  20.494,
  20.321,  20.139,  19.943,  19.733,  19.515,  19.294,  19.079,  18.871,  18.673,  18.481,
  18.291,  18.098,  17.898,  17.690,  17.474,  17.254,  17.033,  16.813,  16.596,  16.380,
  16.164,  15.945,  15.724,  15.502,  15.284,  15.076,  14.886,  14.719,  14.577,  14.456,
  14.351,  14.253,  14.150,  14.035,  13.905,  13.763,  13.616,  13.476,  13.355,  13.261,
  13.196,  13.157,  13.131,  13.102,  13.051,  12.965,  12.838,  12.672,  12.480,  12.283,
  12.105,  11.967,  11.885,  11.864,  11.897,  11.967,  12.050,  12.121,  12.155,  12.139,
  12.068,  11.947,  11.789,  11.614,  11.441,  11.286,  11.159,  11.062,  10.989,  10.928,
  10.861,  10.774,  10.652,  10.486,  10.277,  10.031,   9.763,   9.493,   9.243,   9.033,
   8.879,   8.786,   8.749,   8.751,   8.767,   8.767,   8.722,   8.609,   8.421,   8.164,
   7.864,   7.557,   7.292,   7.115,   7.065,   7.162,   7.407,   7.774,   8.218,   8.679,
   9.091,   9.395,   9.547,   9.527,   9.341,   9.017,   8.603,   8.156,   7.729,   7.366,
   7.093,   6.914,   6.815,   6.769,   6.741,   6.700,   6.621,   6.496,   6.330,   6.141,
   5.956,   5.803,   5.709,   5.691,   5.753,   5.889,   6.081,   6.305,   6.536,   6.748,
   6.926,   7.060,   7.152,   7.213,   7.256,   7.299,   7.356,   7.439,   7.550,   7.687,   7.842,
   8.004,   8.163,   8.309,   8.439,   8.553,   8.653,   8.747,   8.841,   8.939,   9.043,
   9.151,   9.260,   9.362,   9.453,   9.527,   9.582,   9.621,   9.645,   9.662,   9.678,
   9.699,   9.730,   9.771,   9.822,   9.877,   9.930,   9.973,  10.001,  10.011,  10.001,
   9.979,   9.951,   9.930,   9.929,   9.960,  10.029,  10.140,  10.287,  10.463,  10.652,
  10.838,  11.006,  11.145,  11.248,  11.316,  11.355,  11.375,  11.388,  11.407,  11.441,
  11.494,  11.568,  11.661,  11.771,  11.892,  12.023,  12.162,  12.308,  12.463,  12.625,
  12.790,  12.954,  13.111,  13.253,  13.376,  13.479,  13.564,  13.637,  13.706,  13.781,
  13.870,  13.977,  14.103,  14.245,  14.398,  14.555,  14.712,  14.866,  15.016,  15.163,
  15.309,  15.454,  15.597,  15.734,  15.860,  15.971,  16.061,  16.131,  16.183,  16.224,
  16.265,  16.317,  16.390,  16.492,  16.626,  16.791,  16.982,  17.190,  17.408,  17.628,
  17.845,  18.059,  18.273,  18.490,  18.715,  18.950,  19.197,  19.452,  19.707,  19.956,
  20.189,  20.402,  20.594,  20.766,  20.926,  21.081,  21.237,  21.397,  21.556,  21.704,
  21.826,  21.903,  21.920,  21.868,  21.747,  21.570,  21.358,  21.139,  20.943,  20.796,
  20.713,  20.701,  20.757,  20.867,  21.016,  21.191,  21.382,  21.586,  21.806,  22.048,
  22.317,  22.611,  22.925,  23.243,  23.547,  23.817,  24.041,  24.211,  24.332,  24.420,
  24.495,  24.582,  24.703,  24.868,  25.079,  25.327,  25.590,  25.843,  26.058,  26.214,
  26.294,  26.293,  26.214,  26.068,  25.873,  25.646,  25.407,  25.172,  24.955,  24.770,
  24.624,  24.525,  24.476,  24.480,  24.531,  24.621,  24.737,  24.863,  24.982,  25.076},
{
  22.659,  22.684,  22.681,  22.654,  22.607,  22.551,  22.497,  22.454,  22.432,  22.435,
  22.466,  22.520,  22.591,  22.671,  22.748,  22.812,  22.857,  22.876,  22.868,  22.834,
  22.776,  22.700,  22.610,  22.512,  22.407,  22.298,  22.183,  22.063,  21.935,  21.798,
  21.652,  21.497,  21.338,  21.176,  21.016,  20.863,  20.720,  20.588,  20.467,  20.356,
  20.251,  20.147,  20.040,  19.925,  19.798,  19.656,  19.499,  19.329,  19.147,  18.958,
  18.767,  18.579,  18.397,  18.226,  18.069,  17.926,  17.798,  17.685,  17.585,  17.495,
  17.414,  17.338,  17.266,  17.195,  17.122,  17.044,  16.958,  16.861,  16.751,  16.625,
  16.483,  16.326,  16.155,  15.975,  15.790,  15.607,  15.431,  15.268,  15.121,  14.992,
  14.879,  14.779,  14.686,  14.594,  14.498,  14.392,  14.274,  14.146,  14.009,  13.868,
  13.731,  13.602,  13.487,  13.388,  13.305,  13.234,  13.169,  13.102,  13.024,  12.927,
  12.807,  12.660,  12.486,  12.289,  12.074,  11.850,  11.624,  11.406,  11.204,  11.024,
  10.870,  10.746,  10.651,  10.586,  10.547,  10.528,  10.522,  10.520,  10.511,  10.483,
  10.426,  10.329,  10.186,   9.997,   9.765,   9.503,   9.230,   8.968,   8.746,   8.592,
   8.529,   8.575,   8.738,   9.012,   9.381,   9.815,  10.273,  10.713,  11.088,  11.358,
  11.492,  11.472,  11.295,  10.976,  10.542,  10.030,   9.484,   8.949,   8.461,   8.052,
   7.737,   7.520,   7.394,   7.339,   7.333,   7.349,   7.365,   7.363,   7.336,   7.284,
   7.214,   7.142,   7.085,   7.059,   7.078,   7.150,   7.274,   7.446,   7.653,   7.879,
   8.106,   8.318,   8.502,   8.650,   8.760,   8.835,   8.883,   8.913,   8.938,   8.969,   9.013,
   9.076,   9.159,   9.260,   9.375,   9.496,   9.617,   9.728,   9.826,   9.904,   9.962,
   9.998,  10.014,  10.013,   9.998,   9.972,   9.940,   9.904,   9.869,   9.836,   9.809,
   9.791,   9.784,   9.792,   9.816,   9.861,   9.928,  10.019,  10.134,  10.273,  10.433,
  10.612,  10.806,  11.008,  11.213,  11.417,  11.614,  11.800,  11.973,  12.133,  12.280,
  12.417,  12.547,  12.672,  12.796,  12.920,  13.046,  13.172,  13.297,  13.416,  13.528,
  13.627,  13.711,  13.780,  13.834,  13.875,  13.908,  13.938,  13.970,  14.008,  14.057,
  14.118,  14.191,  14.274,  14.364,  14.457,  14.548,  14.633,  14.710,  14.779,  14.839,
  14.893,  14.944,  14.995,  15.050,  15.112,  15.182,  15.261,  15.348,  15.443,  15.541,
  15.641,  15.739,  15.831,  15.916,  15.992,  16.056,  16.109,  16.152,  16.185,  16.211,
  16.235,  16.261,  16.293,  16.336,  16.396,  16.476,  16.579,  16.704,  16.850,  17.015,
  17.191,  17.374,  17.556,  17.733,  17.900,  18.059,  18.212,  18.366,  18.532,  18.721,
  18.944,  19.213,  19.533,  19.904,  20.318,  20.763,  21.216,  21.653,  22.045,  22.365,
  22.591,  22.707,  22.706,  22.593,  22.382,  22.095,  21.763,  21.417,  21.087,  20.801,
  20.577,  20.428,  20.356,  20.355,  20.412,  20.513,  20.641,  20.780,  20.918,  21.045,
  21.159,  21.257,  21.342,  21.416,  21.483,  21.545,  21.603,  21.656,  21.705,  21.747,
  21.784,  21.817,  21.848,  21.882,  21.923,  21.976,  22.044,  22.128,  22.226,  22.333,
  22.441,  22.542,  22.626,  22.685,  22.713,  22.708,  22.672,  22.612,  22.535,  22.455,
  22.383,  22.330,  22.303,  22.307,  22.340,  22.397,  22.468,  22.543,  22.610,  22.659},
{
  20.147,  20.156,  20.170,  20.185,  20.201,  20.214,  20.221,  20.222,  20.213,  20.195,
  20.166,  20.128,  20.083,  20.031,  19.975,  19.917,  19.861,  19.808,  19.761,  19.720,
  19.687,  19.661,  19.644,  19.633,  19.628,  19.628,  19.629,  19.632,  19.634,  19.634,
  19.631,  19.623,  19.610,  19.590,  19.564,  19.529,  19.485,  19.430,  19.363,  19.284,
  19.190,  19.082,  18.960,  18.824,  18.676,  18.519,  18.356,  18.191,  18.028,  17.872,
  17.729,  17.600,  17.491,  17.402,  17.335,  17.288,  17.259,  17.243,  17.237,  17.233,
  17.226,  17.209,  17.177,  17.125,  17.049,  16.947,  16.819,  16.666,  16.491,  16.299,
  16.094,  15.882,  15.668,  15.458,  15.258,  15.070,  14.897,  14.742,  14.603,  14.482,
  14.375,  14.281,  14.197,  14.120,  14.047,  13.977,  13.906,  13.833,  13.756,  13.676,
  13.590,  13.499,  13.403,  13.300,  13.191,  13.075,  12.954,  12.827,  12.697,  12.565,
  12.432,  12.304,  12.181,  12.069,  11.970,  11.888,  11.824,  11.779,  11.751,  11.739,
  11.738,  11.741,  11.742,  11.731,  11.702,  11.647,  11.559,  11.437,  11.279,  11.088,
  10.872,  10.640,  10.406,  10.186,   9.995,   9.851,   9.769,   9.760,   9.835,   9.994,
  10.237,  10.555,  10.934,  11.355,  11.795,  12.228,  12.629,  12.973,  13.237,  13.404,
  13.461,  13.403,  13.231,  12.954,  12.584,  12.142,  11.649,  11.131,  10.611,  10.114,
   9.660,   9.266,   8.944,   8.700,   8.537,   8.451,   8.435,   8.479,   8.571,   8.698,
   8.846,   9.004,   9.161,   9.309,   9.442,   9.557,   9.652,   9.728,   9.788,   9.836,
   9.874,   9.908,   9.941,   9.976,  10.015,  10.059,  10.108,  10.163,  10.220,  10.278,  10.335,
  10.388,  10.435,  10.475,  10.507,  10.531,  10.547,  10.556,  10.561,  10.563,  10.565,
  10.569,  10.577,  10.590,  10.610,  10.637,  10.671,  10.710,  10.755,  10.804,  10.855,
  10.907,  10.960,  11.013,  11.066,  11.121,  11.179,  11.240,  11.306,  11.380,  11.461,
  11.551,  11.649,  11.755,  11.868,  11.984,  12.102,  12.219,  12.333,  12.440,  12.539,
  12.629,  12.710,  12.781,  12.846,  12.905,  12.963,  13.023,  13.087,  13.160,  13.244,
  13.341,  13.452,  13.577,  13.714,  13.862,  14.017,  14.175,  14.332,  14.484,  14.624,
  14.751,  14.861,  14.951,  15.020,  15.070,  15.100,  15.114,  15.113,  15.103,  15.087,
  15.069,  15.053,  15.041,  15.036,  15.040,  15.053,  15.076,  15.107,  15.144,  15.188,
  15.234,  15.284,  15.334,  15.384,  15.435,  15.487,  15.540,  15.596,  15.657,  15.725,
  15.799,  15.882,  15.974,  16.075,  16.183,  16.298,  16.418,  16.540,  16.663,  16.785,
  16.906,  17.025,  17.143,  17.262,  17.383,  17.511,  17.648,  17.798,  17.963,  18.146,
  18.347,  18.566,  18.799,  19.042,  19.291,  19.537,  19.774,  19.993,  20.186,  20.348,
  20.473,  20.558,  20.602,  20.607,  20.576,  20.515,  20.431,  20.331,  20.224,  20.118,
  20.020,  19.935,  19.868,  19.821,  19.795,  19.790,  19.802,  19.828,  19.865,  19.908,
  19.952,  19.994,  20.031,  20.060,  20.079,  20.088,  20.088,  20.079,  20.063,  20.043,
  20.020,  19.999,  19.980,  19.967,  19.962,  19.965,  19.978,  20.001,  20.032,  20.071,
  20.115,  20.162,  20.209,  20.252,  20.291,  20.321,  20.342,  20.353,  20.353,  20.342,
  20.324,  20.298,  20.269,  20.238,  20.208,  20.183,  20.163,  20.150,  20.145,  20.147},
{
  17.632,  17.609,  17.589,  17.570,  17.555,  17.541,  17.529,  17.519,  17.509,  17.500,
  17.492,  17.484,  17.475,  17.465,  17.454,  17.443,  17.430,  17.415,  17.399,  17.382,
  17.363,  17.343,  17.322,  17.299,  17.275,  17.250,  17.224,  17.196,  17.168,  17.137,
  17.105,  17.071,  17.035,  16.997,  16.956,  16.913,  16.866,  16.816,  16.763,  16.707,
  16.647,  16.585,  16.519,  16.452,  16.382,  16.311,  16.239,  16.166,  16.093,  16.021,
  15.950,  15.880,  15.812,  15.746,  15.682,  15.620,  15.561,  15.503,  15.447,  15.393,
  15.339,  15.286,  15.233,  15.180,  15.126,  15.071,  15.014,  14.956,  14.895,  14.833,
  14.768,  14.702,  14.633,  14.563,  14.491,  14.418,  14.344,  14.270,  14.196,  14.122,
  14.049,  13.978,  13.908,  13.841,  13.777,  13.715,  13.657,  13.603,  13.552,  13.504,
  13.461,  13.420,  13.383,  13.347,  13.314,  13.281,  13.247,  13.213,  13.175,  13.134,
  13.087,  13.033,  12.971,  12.901,  12.820,  12.729,  12.627,  12.515,  12.392,  12.261,
  12.122,  11.978,  11.833,  11.687,  11.547,  11.414,  11.294,  11.191,  11.109,  11.051,
  11.022,  11.024,  11.060,  11.132,  11.240,  11.383,  11.561,  11.770,  12.006,  12.266,
  12.542,  12.830,  13.121,  13.409,  13.685,  13.942,  14.174,  14.374,  14.535,  14.655,
  14.729,  14.755,  14.732,  14.661,  14.545,  14.385,  14.188,  13.957,  13.699,  13.421,
  13.129,  12.831,  12.533,  12.242,  11.963,  11.702,  11.464,  11.250,  11.065,  10.910,
  10.784,  10.688,  10.621,  10.580,  10.564,  10.568,  10.591,  10.629,  10.678,  10.735,
  10.796,  10.859,  10.921,  10.980,  11.034,  11.081,  11.120,  11.152,  11.175,  11.190,  11.198,
  11.198,  11.194,  11.184,  11.172,  11.159,  11.145,  11.133,  11.123,  11.118,  11.118,
  11.124,  11.137,  11.157,  11.184,  11.219,  11.262,  11.310,  11.365,  11.426,  11.490,
  11.557,  11.626,  11.696,  11.765,  11.833,  11.898,  11.959,  12.016,  12.069,  12.117,
  12.159,  12.197,  12.231,  12.261,  12.288,  12.312,  12.335,  12.359,  12.383,  12.409,
  12.438,  12.470,  12.507,  12.549,  12.596,  12.648,  12.705,  12.766,  12.831,  12.898,
  12.968,  13.038,  13.108,  13.176,  13.241,  13.303,  13.360,  13.412,  13.458,  13.498,
  13.533,  13.563,  13.589,  13.612,  13.633,  13.655,  13.677,  13.704,  13.735,  13.773,
  13.819,  13.875,  13.940,  14.018,  14.106,  14.206,  14.316,  14.437,  14.567,  14.704,
  14.847,  14.994,  15.143,  15.292,  15.439,  15.582,  15.721,  15.852,  15.977,  16.093,
  16.200,  16.299,  16.391,  16.475,  16.553,  16.627,  16.697,  16.767,  16.837,  16.909,
  16.985,  17.066,  17.153,  17.247,  17.349,  17.457,  17.573,  17.694,  17.819,  17.947,
  18.076,  18.202,  18.324,  18.439,  18.545,  18.638,  18.716,  18.778,  18.822,  18.847,
  18.851,  18.836,  18.801,  18.748,  18.677,  18.592,  18.494,  18.386,  18.272,  18.153,
  18.034,  17.917,  17.805,  17.700,  17.606,  17.524,  17.454,  17.400,  17.359,  17.334,
  17.323,  17.327,  17.342,  17.370,  17.407,  17.452,  17.504,  17.561,  17.621,  17.681,
  17.742,  17.801,  17.857,  17.909,  17.956,  17.997,  18.032,  18.062,  18.084,  18.100,
  18.109,  18.112,  18.109,  18.100,  18.086,  18.068,  18.045,  18.018,  17.989,  17.957,
  17.923,  17.888,  17.852,  17.817,  17.782,  17.748,  17.716,  17.686,  17.658,  17.632},
{
  15.452,  15.446,  15.439,  15.431,  15.423,  15.414,  15.404,  15.393,  15.382,  15.371,
  15.359,  15.346,  15.333,  15.319,  15.305,  15.290,  15.275,  15.260,  15.244,  15.227,
  15.210,  15.193,  15.176,  15.158,  15.140,  15.121,  15.103,  15.084,  15.065,  15.045,
  15.026,  15.006,  14.987,  14.967,  14.947,  14.927,  14.907,  14.887,  14.867,  14.848,
  14.828,  14.808,  14.788,  14.768,  14.748,  14.728,  14.708,  14.687,  14.667,  14.647,
  14.626,  14.605,  14.583,  14.561,  14.539,  14.517,  14.494,  14.470,  14.446,  14.422,
  14.396,  14.371,  14.344,  14.317,  14.289,  14.261,  14.232,  14.203,  14.172,  14.142,
  14.110,  14.079,  14.046,  14.014,  13.980,  13.947,  13.913,  13.879,  13.844,  13.809,
  13.774,  13.739,  13.703,  13.668,  13.632,  13.596,  13.560,  13.524,  13.489,  13.453,
  13.417,  13.381,  13.345,  13.310,  13.275,  13.240,  13.206,  13.172,  13.139,  13.107,
  13.075,  13.045,  13.015,  12.987,  12.960,  12.935,  12.912,  12.890,  12.871,  12.853,
  12.838,  12.826,  12.815,  12.808,  12.803,  12.801,  12.802,  12.805,  12.812,  12.821,
  12.832,  12.846,  12.862,  12.881,  12.901,  12.923,  12.947,  12.971,  12.997,  13.023,
  13.049,  13.076,  13.102,  13.127,  13.152,  13.175,  13.196,  13.216,  13.234,  13.250,
  13.263,  13.274,  13.282,  13.288,  13.291,  13.291,  13.289,  13.284,  13.277,  13.268,
  13.256,  13.243,  13.228,  13.211,  13.194,  13.175,  13.156,  13.136,  13.115,  13.095,
  13.075,  13.055,  13.036,  13.018,  13.000,  12.983,  12.968,  12.953,  12.940,  12.928,
  12.917,  12.907,  12.899,  12.892,  12.885,  12.880,  12.876,  12.873,  12.870,  12.868,  12.867,
  12.867,  12.866,  12.867,  12.867,  12.868,  12.869,  12.871,  12.873,  12.875,  12.877,
  12.879,  12.882,  12.885,  12.889,  12.892,  12.897,  12.902,  12.907,  12.913,  12.920,
  12.928,  12.937,  12.946,  12.957,  12.969,  12.981,  12.995,  13.010,  13.026,  13.044,
  13.062,  13.082,  13.103,  13.125,  13.148,  13.173,  13.198,  13.224,  13.251,  13.279,
  13.308,  13.337,  13.367,  13.398,  13.429,  13.461,  13.493,  13.526,  13.559,  13.592,
  13.626,  13.659,  13.693,  13.728,  13.762,  13.797,  13.832,  13.868,  13.903,  13.939,
  13.976,  14.013,  14.050,  14.088,  14.126,  14.165,  14.205,  14.245,  14.286,  14.327,
  14.369,  14.412,  14.455,  14.499,  14.544,  14.589,  14.634,  14.680,  14.726,  14.773,
  14.819,  14.866,  14.913,  14.959,  15.006,  15.052,  15.098,  15.143,  15.188,  15.231,
  15.274,  15.317,  15.358,  15.397,  15.436,  15.474,  15.510,  15.544,  15.577,  15.608,
  15.638,  15.665,  15.691,  15.715,  15.737,  15.757,  15.774,  15.790,  15.803,  15.814,
  15.823,  15.829,  15.834,  15.836,  15.835,  15.833,  15.829,  15.822,  15.814,  15.803,
  15.791,  15.777,  15.762,  15.745,  15.727,  15.708,  15.688,  15.667,  15.646,  15.624,
  15.602,  15.581,  15.560,  15.539,  15.518,  15.499,  15.480,  15.463,  15.446,  15.431,
  15.418,  15.406,  15.395,  15.386,  15.378,  15.373,  15.368,  15.365,  15.364,  15.364,
  15.365,  15.367,  15.370,  15.375,  15.380,  15.386,  15.392,  15.399,  15.406,  15.413,
  15.420,  15.427,  15.434,  15.441,  15.447,  15.453,  15.458,  15.462,  15.466,  15.469,
  15.471,  15.472,  15.473,  15.472,  15.471,  15.469,  15.466,  15.462,  15.457,  15.452},
{
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,
  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606,  13.606}};

/** modip2001_wrapped.asc preloaded as constant. */
const double_t g_corrected_modip_degree
    [NEQUICK_G_JRC_MODIP_GRID_LAT_POINTS_COUNT]
    [NEQUICK_G_JRC_MODIP_GRID_LONG_POINTS_COUNT] = {
  { -76.37, -76.34, -76.30, -76.25, -76.55, -76.47, -76.75, -76.65, -76.90, -76.79, -77.04, -77.27, -77.19, -77.44, -77.37, -77.33, -77.30, -77.30, -77.31, -77.34, -77.40, -77.13, -77.21, -76.96, -77.06, -76.81, -76.91, -76.66, -76.76, -76.48, -76.56, -76.25, -76.31, -76.34, -76.37, -76.37, -76.37, -76.34, -76.30 },
  { -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00, -90.00 },
  { -77.31, -77.34, -77.40, -77.13, -77.21, -76.96, -77.06, -76.81, -76.91, -76.66, -76.76, -76.48, -76.56, -76.25, -76.31, -76.34, -76.37, -76.37, -76.37, -76.34, -76.30, -76.25, -76.55, -76.47, -76.75, -76.65, -76.90, -76.79, -77.04, -77.27, -77.19, -77.44, -77.37, -77.33, -77.30, -77.30, -77.31, -77.34, -77.40 },
  { -72.98, -73.10, -72.84, -72.61, -72.40, -72.20, -72.00, -71.80, -71.59, -71.36, -71.11, -70.82, -70.49, -70.64, -70.23, -70.31, -70.35, -70.36, -70.34, -70.28, -70.18, -70.59, -70.42, -70.74, -71.02, -71.26, -71.47, -71.66, -72.30, -72.47, -72.65, -72.86, -73.10, -73.38, -73.32, -73.32, -72.98, -73.10, -72.84 },
  { -70.32, -70.13, -69.97, -69.37, -69.25, -68.64, -67.99, -67.89, -67.19, -67.04, -66.22, -65.96, -65.61, -65.84, -65.33, -65.41, -65.43, -65.42, -65.35, -65.26, -65.78, -65.58, -65.97, -66.28, -66.51, -67.28, -67.95, -68.01, -68.58, -69.12, -69.65, -70.20, -70.34, -70.58, -70.51, -70.58, -70.32, -70.13, -69.97 },
  { -67.88, -67.33, -66.76, -66.71, -66.11, -65.50, -64.87, -64.23, -63.55, -62.82, -62.72, -61.76, -61.41, -60.88, -61.01, -61.01, -60.95, -60.84, -61.51, -61.36, -61.95, -61.71, -62.13, -62.42, -63.30, -64.03, -64.63, -65.15, -66.19, -66.60, -67.52, -67.93, -68.39, -68.55, -68.58, -68.42, -67.88, -67.33, -66.76 },
  { -65.43, -64.90, -64.33, -63.71, -63.06, -62.40, -61.75, -61.11, -60.47, -59.79, -58.18, -58.04, -57.67, -57.02, -57.01, -57.72, -57.46, -58.06, -57.84, -58.51, -58.32, -58.92, -59.38, -59.66, -60.54, -61.22, -61.75, -62.83, -63.80, -64.68, -65.51, -66.31, -66.61, -67.17, -66.81, -65.90, -65.43, -64.90, -64.33 },
  { -63.10, -62.52, -61.89, -61.21, -60.49, -59.77, -59.08, -58.42, -57.00, -56.36, -54.71, -53.70, -53.29, -53.45, -53.20, -53.67, -55.00, -55.41, -55.94, -55.71, -55.57, -56.29, -55.98, -57.17, -58.06, -58.70, -59.90, -60.92, -61.83, -62.70, -63.55, -64.45, -64.56, -64.70, -64.14, -63.63, -63.10, -62.52, -61.89 },
  { -60.55, -59.89, -59.18, -58.41, -57.60, -56.79, -56.00, -55.29, -53.77, -53.17, -51.50, -50.51, -50.08, -50.08, -50.56, -51.68, -52.65, -52.78, -54.07, -53.77, -53.68, -54.54, -54.38, -54.82, -55.78, -56.42, -57.62, -58.65, -60.26, -61.19, -61.61, -62.29, -62.19, -61.90, -61.80, -61.18, -60.55, -59.89, -59.18 },
  { -57.52, -56.76, -56.72, -55.88, -54.99, -54.10, -53.23, -52.44, -50.80, -49.16, -48.52, -46.43, -45.95, -45.73, -47.00, -48.82, -50.45, -51.19, -52.26, -52.75, -52.68, -52.80, -52.84, -53.46, -53.66, -54.34, -55.58, -57.41, -58.42, -58.87, -59.57, -59.97, -60.07, -59.79, -58.99, -58.25, -57.52, -56.76, -56.72 },
  { -55.35, -54.54, -53.68, -52.76, -51.79, -50.80, -49.83, -48.94, -48.17, -46.46, -44.70, -42.50, -41.96, -42.75, -44.84, -46.20, -48.49, -49.87, -51.57, -51.85, -51.71, -51.95, -51.34, -51.31, -52.58, -53.34, -54.63, -55.78, -56.24, -57.62, -57.94, -57.92, -57.47, -57.17, -56.26, -56.16, -55.35, -54.54, -53.68 },
  { -52.61, -51.73, -50.81, -49.82, -48.79, -47.75, -46.73, -45.73, -44.88, -43.03, -41.20, -38.94, -38.31, -38.86, -41.88, -43.95, -46.94, -48.96, -50.32, -51.24, -51.85, -51.24, -50.87, -50.24, -50.85, -51.80, -52.39, -53.74, -54.41, -55.38, -55.25, -55.40, -54.97, -54.65, -53.63, -53.49, -52.61, -51.73, -50.81 },
  { -50.08, -48.20, -47.20, -47.17, -46.12, -45.07, -42.87, -41.78, -40.82, -38.82, -36.89, -34.51, -33.78, -35.48, -38.12, -41.01, -44.83, -47.57, -49.58, -50.20, -50.62, -50.87, -49.77, -49.45, -49.44, -49.72, -51.40, -52.11, -53.01, -52.69, -52.77, -53.07, -52.64, -51.44, -51.17, -50.13, -50.08, -48.20, -47.20 },
  { -46.90, -44.89, -43.84, -42.71, -41.57, -40.48, -39.39, -36.91, -35.88, -33.72, -31.73, -29.23, -28.33, -29.76, -33.48, -38.67, -42.12, -45.72, -47.53, -48.89, -50.02, -50.20, -49.21, -48.22, -47.56, -48.11, -49.16, -50.11, -50.41, -50.29, -50.57, -50.12, -49.68, -49.27, -48.03, -47.90, -46.90, -44.89, -43.84 },
  { -42.94, -40.78, -39.69, -38.48, -37.26, -36.16, -33.63, -32.45, -29.87, -27.64, -23.91, -21.22, -21.91, -22.92, -29.50, -34.27, -40.00, -43.37, -46.09, -47.36, -48.44, -48.57, -47.60, -45.76, -45.35, -46.19, -46.51, -46.76, -47.29, -47.35, -46.83, -46.42, -45.93, -45.45, -45.13, -43.97, -42.94, -40.78, -39.69 },
  { -38.00, -37.05, -34.60, -33.29, -32.00, -29.31, -28.16, -25.31, -22.60, -20.39, -16.58, -13.78, -12.32, -16.69, -22.92, -30.45, -35.94, -40.47, -43.26, -45.67, -45.82, -45.96, -44.97, -43.15, -42.92, -42.92, -43.46, -43.88, -44.62, -43.76, -43.30, -42.89, -42.33, -41.78, -40.28, -39.04, -38.00, -37.05, -34.60 },
  { -33.30, -30.99, -28.38, -26.97, -25.53, -22.62, -19.66, -18.45, -15.74, -11.72, -7.87, -4.92, -3.12, -7.12, -14.93, -24.08, -32.47, -36.99, -40.00, -41.64, -43.06, -42.18, -41.13, -40.42, -39.12, -38.01, -38.74, -39.34, -40.26, -39.38, -38.94, -38.49, -37.82, -37.16, -35.51, -34.24, -33.30, -30.99, -28.38 },
  { -25.81, -23.45, -20.73, -19.23, -15.79, -14.45, -11.27, -8.07, -5.36, -1.38, 1.54, 4.46, 4.34, -0.15, -7.30, -18.17, -26.70, -32.91, -36.30, -38.36, -38.91, -38.05, -36.96, -34.95, -33.56, -32.35, -33.22, -33.93, -35.02, -34.07, -33.58, -33.04, -32.18, -31.36, -29.52, -28.25, -25.81, -23.45, -20.73 },
  { -18.21, -14.07, -11.29, -9.72, -7.95, -4.47, -1.12, -0.13, 3.20, 7.24, 11.16, 13.63, 13.20, 8.68, -0.02, -8.99, -19.80, -26.62, -30.66, -31.94, -32.88, -32.04, -29.36, -27.06, -25.49, -25.75, -26.72, -27.51, -27.17, -27.63, -25.41, -24.66, -25.14, -24.09, -22.00, -20.69, -18.21, -14.07, -11.29 },
  { -8.65, -4.46, -1.73, 0.13, 2.22, 4.54, 7.05, 9.67, 12.82, 16.46, 19.79, 21.67, 21.01, 16.82, 8.72, -0.07, -11.56, -19.23, -22.42, -24.32, -24.10, -23.26, -20.19, -17.56, -15.77, -16.03, -17.09, -17.97, -17.55, -17.99, -17.17, -16.10, -16.39, -14.98, -14.45, -11.16, -8.65, -4.46, -1.73 },
  { 0.78, 4.62, 7.90, 10.21, 12.09, 14.13, 16.34, 18.68, 21.42, 24.45, 27.06, 28.37, 27.58, 23.89, 16.91, 7.39, -1.72, -8.45, -12.47, -13.20, -13.31, -12.51, -9.01, -7.96, -5.94, -4.22, -5.36, -6.30, -7.78, -6.17, -5.05, -5.58, -5.56, -5.74, -4.93, -1.35, 0.78, 4.62, 7.90 },
  { 10.53, 14.10, 17.12, 19.21, 20.83, 22.54, 24.39, 26.39, 28.66, 31.06, 33.02, 33.86, 32.99, 29.88, 24.18, 16.35, 8.11, 1.73, -0.45, -1.65, -0.05, -0.74, 1.49, 4.17, 5.96, 6.21, 5.41, 4.44, 3.94, 4.19, 4.79, 4.99, 4.61, 4.42, 5.26, 7.37, 10.53, 14.10, 17.12 },
  { 19.28, 22.29, 24.88, 26.70, 28.07, 29.49, 31.06, 32.75, 34.60, 36.47, 37.90, 38.39, 37.50, 34.90, 30.42, 24.42, 18.03, 12.98, 10.38, 9.92, 10.50, 11.69, 13.59, 15.77, 17.22, 17.43, 16.81, 16.02, 15.59, 15.74, 16.07, 15.96, 15.30, 14.76, 15.12, 16.69, 19.28, 22.29, 24.88 },
  { 26.70, 29.05, 31.16, 32.70, 33.90, 35.12, 36.47, 37.92, 39.45, 40.91, 41.96, 42.19, 41.32, 39.16, 35.70, 31.30, 26.70, 23.08, 21.25, 21.01, 21.61, 22.63, 24.09, 25.69, 26.74, 26.93, 26.51, 25.95, 25.66, 25.72, 25.83, 25.49, 24.69, 23.90, 23.83, 24.79, 26.70, 29.05, 31.16 },
  { 32.85, 34.58, 36.23, 37.53, 38.62, 39.71, 40.91, 42.19, 43.48, 44.66, 45.43, 45.49, 44.65, 42.86, 40.21, 37.04, 33.89, 31.44, 30.23, 30.14, 30.66, 31.45, 32.51, 33.62, 34.37, 34.53, 34.28, 33.93, 33.76, 33.79, 33.75, 33.31, 32.45, 31.54, 31.14, 31.57, 32.85, 34.58, 36.23 },
  { 37.97, 39.18, 40.43, 41.54, 42.53, 43.56, 44.66, 45.80, 46.94, 47.90, 48.48, 48.44, 47.65, 46.16, 44.13, 41.87, 39.72, 38.12, 37.34, 37.33, 37.74, 38.35, 39.10, 39.85, 40.37, 40.53, 40.41, 40.25, 40.16, 40.17, 40.07, 39.59, 38.73, 37.78, 37.19, 37.23, 37.97, 39.18, 40.43 },
  { 42.38, 43.15, 44.10, 45.03, 45.95, 46.92, 47.95, 49.00, 50.00, 50.82, 51.27, 51.16, 50.44, 49.18, 47.61, 45.97, 44.51, 43.46, 42.97, 42.99, 43.30, 43.76, 44.30, 44.81, 45.18, 45.33, 45.33, 45.29, 45.30, 45.31, 45.15, 44.67, 43.86, 42.93, 42.23, 42.02, 42.38, 43.15, 44.10 },
  { 46.29, 46.76, 47.46, 48.25, 49.09, 50.00, 50.97, 51.93, 52.82, 53.52, 53.87, 53.73, 53.07, 52.02, 50.78, 49.57, 48.55, 47.84, 47.52, 47.55, 47.79, 48.13, 48.51, 48.87, 49.16, 49.32, 49.40, 49.46, 49.53, 49.55, 49.38, 48.91, 48.15, 47.29, 46.56, 46.20, 46.29, 46.76, 47.46 },
  { 49.92, 50.19, 50.69, 51.34, 52.10, 52.95, 53.84, 54.71, 55.50, 56.10, 56.38, 56.21, 55.61, 54.73, 53.75, 52.83, 52.07, 51.57, 51.35, 51.37, 51.54, 51.80, 52.07, 52.34, 52.58, 52.76, 52.91, 53.05, 53.17, 53.19, 53.02, 52.58, 51.89, 51.10, 50.41, 49.98, 49.92, 50.19, 50.69 },
  { 53.37, 53.50, 53.87, 54.42, 55.09, 55.84, 56.64, 57.42, 58.11, 58.62, 58.82, 58.64, 58.11, 57.37, 56.57, 55.85, 55.28, 54.90, 54.72, 54.72, 54.84, 55.02, 55.22, 55.44, 55.65, 55.86, 56.06, 56.26, 56.41, 56.45, 56.28, 55.87, 55.26, 54.56, 53.93, 53.52, 53.37, 53.50, 53.87 },
  { 56.74, 56.80, 57.06, 57.50, 58.07, 58.72, 59.42, 60.10, 60.69, 61.11, 61.25, 61.05, 60.58, 59.96, 59.31, 58.73, 58.28, 57.98, 57.82, 57.79, 57.86, 57.98, 58.14, 58.33, 58.53, 58.75, 59.00, 59.23, 59.41, 59.46, 59.31, 58.95, 58.41, 57.82, 57.28, 56.90, 56.74, 56.80, 57.06 },
  { 60.07, 60.09, 60.28, 60.62, 61.09, 61.63, 62.21, 62.78, 63.28, 63.61, 63.69, 63.48, 63.06, 62.54, 62.02, 61.56, 61.19, 60.93, 60.79, 60.74, 60.77, 60.85, 60.98, 61.14, 61.34, 61.58, 61.83, 62.09, 62.28, 62.34, 62.21, 61.91, 61.47, 60.98, 60.54, 60.22, 60.07, 60.09, 60.28 },
  { 63.41, 63.42, 63.55, 63.81, 64.16, 64.59, 65.05, 65.50, 65.91, 66.17, 66.19, 65.97, 65.60, 65.17, 64.76, 64.39, 64.10, 63.88, 63.75, 63.69, 63.70, 63.75, 63.85, 64.00, 64.19, 64.42, 64.67, 64.92, 65.11, 65.17, 65.08, 64.84, 64.50, 64.13, 63.79, 63.54, 63.41, 63.42, 63.55 },
  { 66.83, 66.83, 66.92, 67.10, 67.35, 67.65, 68.00, 68.34, 68.64, 68.83, 68.80, 68.57, 68.26, 67.93, 67.61, 67.34, 67.11, 66.94, 66.82, 66.77, 66.76, 66.80, 66.88, 67.01, 67.18, 67.38, 67.60, 67.82, 67.99, 68.07, 68.02, 67.85, 67.60, 67.34, 67.10, 66.93, 66.83, 66.83, 66.92 },
  { 70.42, 70.42, 70.48, 70.58, 70.74, 70.94, 71.16, 71.40, 71.61, 71.72, 71.63, 71.42, 71.18, 70.94, 70.72, 70.53, 70.37, 70.25, 70.16, 70.12, 70.11, 70.14, 70.20, 70.30, 70.43, 70.58, 70.76, 70.93, 71.07, 71.15, 71.14, 71.04, 70.90, 70.74, 70.59, 70.48, 70.42, 70.42, 70.48 },
  { 74.37, 74.37, 74.40, 74.46, 74.54, 74.65, 74.77, 74.90, 75.03, 75.03, 74.90, 74.75, 74.60, 74.46, 74.33, 74.21, 74.12, 74.05, 73.99, 73.97, 73.96, 73.98, 74.02, 74.08, 74.16, 74.26, 74.36, 74.47, 74.57, 74.64, 74.66, 74.64, 74.58, 74.52, 74.45, 74.40, 74.37, 74.37, 74.40 },
  { 79.15, 79.15, 79.17, 79.19, 79.22, 79.25, 79.29, 79.31, 79.28, 79.23, 79.18, 79.12, 79.06, 79.01, 78.96, 78.92, 78.88, 78.85, 78.83, 78.82, 78.82, 78.83, 78.84, 78.87, 78.90, 78.93, 78.97, 79.02, 79.06, 79.10, 79.13, 79.15, 79.16, 79.16, 79.15, 79.15, 79.15, 79.15, 79.17 },
  { 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00, 90.00 },
  { 78.83, 78.82, 78.82, 78.83, 78.84, 78.87, 78.90, 78.93, 78.97, 79.02, 79.06, 79.10, 79.13, 79.15, 79.16, 79.16, 79.15, 79.15, 79.15, 79.15, 79.17, 79.19, 79.22, 79.25, 79.29, 79.31, 79.28, 79.23, 79.18, 79.12, 79.06, 79.01, 78.96, 78.92, 78.88, 78.85, 78.83, 78.82, 78.82 },
};

const F2_coefficient_array_t g_NeQuickG_ccir11_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.52396593E+01, -0.56523629E-01, -0.18704617E-01, 0.12128915E-01, 0.79412190E-02, -0.10031431E-01, 0.21567261E-01, -0.68602660E-02, 0.37022342E-02, 0.78359339E-02, 0.63161603E-02, -0.10695397E-01, 0.29390149E-01, 
    },
    {
      0.93325400E+00, -0.28997503E-01, 0.10946778E+00, -0.30769527E+00, -0.37993383E+00, -0.23273268E+00, 0.89480691E-01, 0.33896305E-01, 0.32839003E+00, -0.81993349E-01, -0.14348941E+00, -0.27823284E-01, 0.11266428E-01, 
    },
    {
      0.80531130E+01, 0.13981724E+01, 0.47361958E+00, -0.11388183E+00, 0.77816737E+00, -0.17388150E+00, 0.29099104E+00, 0.29059123E-01, -0.37210885E+00, -0.11191850E+00, -0.43733008E-01, 0.12193082E+00, -0.32639468E+00, 
    },
    {
      -0.13390853E+02, 0.62356526E+00, 0.24597554E+01, 0.44970918E+01, 0.85659552E+01, 0.40355296E+01, -0.79231381E+00, -0.67203265E+00, -0.47442737E+01, 0.21074810E+01, 0.33913586E+01, -0.10161761E+00, -0.69574153E+00, 
    },
    {
      -0.32273560E+02, -0.56248417E+01, -0.58702965E+01, -0.43174982E+01, -0.32914643E+01, 0.14731911E+01, -0.84157687E+00, -0.18629679E+00, 0.36630037E+01, -0.49589828E+00, 0.42669845E+00, -0.39217409E+00, 0.15235479E+01, 
    },
    {
      -0.10364820E+00, -0.17449905E+02, -0.18375044E+02, -0.16429411E+02, -0.42843178E+02, -0.18512333E+02, 0.29954524E+01, 0.44427662E+01, 0.22331482E+02, -0.11312363E+02, -0.17385393E+02, 0.13743354E+01, 0.49287653E+01, 
    },
    {
      0.35317551E+02, 0.78040113E+01, 0.22017223E+02, 0.19007372E+02, 0.51486945E+01, -0.32355106E+01, -0.70289177E+00, -0.18949838E-01, -0.10833933E+02, 0.30533838E+01, -0.15614654E+01, 0.34174812E+00, -0.33425496E+01, 
    },
    {
      0.10955015E+03, 0.64495300E+02, 0.39307384E+02, 0.21447481E+02, 0.88321732E+02, 0.36022030E+02, -0.49279366E+01, -0.11266347E+02, -0.45660416E+02, 0.24368761E+02, 0.36402855E+02, -0.37914002E+01, -0.12354781E+02, 
    },
    {
      -0.15069658E+02, -0.54928751E+01, -0.30891773E+02, -0.26297506E+02, -0.28576374E+01, 0.23907247E+01, 0.27161193E+01, 0.10669554E+01, 0.12239296E+02, -0.43662004E+01, 0.20293267E+01, 0.21851628E+00, 0.34469433E+01, 
    },
    {
      -0.18371085E+03, -0.78854736E+02, -0.32724571E+02, -0.86944246E+01, -0.80539780E+02, -0.31638607E+02, 0.32864110E+01, 0.12187758E+02, 0.42042595E+02, -0.23398046E+02, -0.33907742E+02, 0.42148323E+01, 0.12834593E+02, 
    },
    {
      0.22537875E+01, 0.19722799E+01, 0.14214872E+02, 0.11752549E+02, 0.97573824E-01, -0.41853258E+00, -0.14487737E+01, -0.91950291E+00, -0.47050514E+01, 0.19138944E+01, -0.85748482E+00, -0.30080923E+00, -0.13293968E+01, 
    },
    {
      0.86102806E+02, 0.31215933E+02, 0.92085867E+01, -0.52750909E+00, 0.26795364E+02, 0.10308076E+02, -0.60259044E+00, -0.47401175E+01, -0.14323977E+02, 0.83509979E+01, 0.11643826E+02, -0.16808408E+01, -0.47212696E+01, 
    },
    {
      0.53214836E-02, 0.17079000E+01, 0.16235816E+01, 0.33274181E-01, -0.12954264E-01, -0.36316011E-02, 0.45543164E-01, -0.43191575E-02, -0.25532667E-01, 0.89464840E-02, 0.23123844E-01, 0.16921325E-01, 0.33485752E-01, 
    },
    {
      -0.22577250E+00, -0.16961778E+01, 0.17585011E+01, -0.36464985E-01, -0.46494432E-01, -0.19080851E-02, -0.31143922E-01, 0.10176010E-01, -0.69593927E-02, -0.66427351E-02, 0.23588389E-02, -0.55617536E-02, -0.19860143E-01, 
    },
    {
      -0.17363087E+01, 0.18277783E+01, 0.16650461E+01, -0.48928317E-01, -0.43136498E+00, -0.47681876E-01, 0.25482827E+00, -0.18712924E+00, 0.47013280E+00, -0.17417172E+00, 0.78237861E-01, 0.10381135E+00, 0.17819388E-01, 
    },
    {
      -0.14371649E+01, -0.80177718E+00, 0.61360812E+00, -0.59573984E+00, 0.52904451E+00, -0.10058827E+00, 0.21083134E+00, 0.78204811E-01, 0.61967522E-02, 0.11077180E+00, -0.40769756E-01, 0.21807816E-01, 0.20648648E-02, 
    },
    {
      0.46911736E+01, 0.78173614E+01, 0.12199426E+02, -0.16422656E+01, 0.37755153E+01, -0.40263680E+00, -0.52920246E+00, -0.21391793E+00, 0.57929236E+00, 0.80246784E-01, -0.18910319E+00, -0.50804520E+00, -0.88265932E+00, 
    },
    {
      0.69275224E+00, -0.12002475E+02, 0.57374411E+01, 0.13528378E+01, 0.44981871E+01, -0.47366244E+00, 0.36056972E+00, 0.79039741E+00, -0.22344507E-01, -0.64753883E-01, 0.46970192E+00, -0.40660375E+00, 0.44191428E-01, 
    },
    {
      0.39014656E+02, -0.28619871E+02, -0.15200562E+01, -0.19682378E+01, 0.13416008E+02, 0.27089481E+01, -0.59305787E+01, 0.31192074E+01, -0.91367588E+01, 0.26589971E+01, -0.11606407E+00, -0.35579140E+01, -0.89003706E+00, 
    },
    {
      0.34678040E+02, -0.52889085E+00, -0.57936335E+01, 0.87519836E+01, -0.10267422E+02, 0.12299103E+01, -0.30112040E+01, 0.11044223E+01, -0.21466107E+01, 0.45218792E-01, 0.30407925E+01, -0.71307260E+00, -0.15156338E+01, 
    },
    {
      -0.52359825E+02, -0.65902946E+02, -0.30542086E+02, 0.45587659E+01, -0.33256973E+02, 0.32021556E+01, 0.60264249E+01, 0.39931140E+01, -0.21957397E+01, -0.14460717E+01, -0.62892991E+00, 0.19364247E+01, 0.71591692E+01, 
    },
    {
      0.99375763E+01, 0.36078590E+02, -0.38465443E+02, -0.34042759E+01, -0.46145630E+02, 0.20836172E+01, -0.37764747E+01, -0.62155070E+01, -0.36285675E+00, 0.83748114E+00, -0.11298743E+01, 0.38863559E+01, 0.14642847E+01, 
    },
    {
      -0.22182343E+03, 0.11115142E+03, -0.16951048E+02, 0.17158335E+02, -0.61797066E+02, -0.21725040E+02, 0.33751064E+02, -0.20662024E+02, 0.52077259E+02, -0.10576082E+02, 0.19523579E+00, 0.24283333E+02, 0.39174995E+01, 
    },
    {
      -0.22131866E+03, 0.17304182E+02, -0.28861012E+02, -0.44697052E+02, 0.53147095E+02, -0.61317348E+01, 0.16753126E+02, -0.89905128E+01, 0.20872980E+02, -0.63277907E+01, -0.26276550E+02, 0.64804440E+01, 0.11094950E+02, 
    },
    {
      0.17550497E+03, 0.16372993E+03, -0.21463629E+02, 0.27199810E+01, 0.99251183E+02, -0.79424796E+01, -0.21751915E+02, -0.18336351E+02, 0.69923621E+00, 0.50953045E+01, 0.72605686E+01, -0.15412294E+01, -0.22803534E+02, 
    },
    {
      -0.56229095E+02, 0.14818197E+01, 0.69807755E+02, 0.33564434E+01, 0.14655196E+03, -0.27966247E+01, 0.12579897E+02, 0.17929468E+02, 0.43725858E+01, -0.22176220E+01, -0.37626834E+01, -0.10943099E+02, -0.52088380E+01, 
    },
    {
      0.53665625E+03, -0.20060770E+03, 0.17310247E+02, -0.54237488E+02, 0.99595345E+02, 0.66468887E+02, -0.80023064E+02, 0.58134552E+02, -0.12709147E+03, 0.15493525E+02, -0.31282263E+01, -0.65706467E+02, -0.49680519E+01, 
    },
    {
      0.55392444E+03, -0.46329002E+02, 0.14449484E+03, 0.10233988E+03, -0.11556944E+03, 0.13084748E+02, -0.44251144E+02, 0.22732943E+02, -0.62439163E+02, 0.21050632E+02, 0.77459229E+02, -0.19942715E+02, -0.28916513E+02, 
    },
    {
      -0.23239845E+03, -0.18225694E+03, 0.10545853E+03, -0.16537470E+02, -0.12225861E+03, 0.72478418E+01, 0.29107874E+02, 0.29596212E+02, 0.43757663E+01, -0.56305561E+01, -0.14893911E+02, -0.16775293E+01, 0.30470707E+02, 
    },
    {
      0.95573692E+02, -0.82114143E+02, -0.54173027E+02, -0.28702703E+01, -0.18174019E+03, -0.33480763E+00, -0.15407213E+02, -0.20705093E+02, -0.86905804E+01, 0.13393050E+01, 0.11856750E+02, 0.11862597E+02, 0.56412406E+01, 
    },
    {
      -0.58746796E+03, 0.17872774E+03, 0.44170815E+02, 0.71773781E+02, -0.55947525E+02, -0.85275742E+02, 0.84508293E+02, -0.71050819E+02, 0.13868575E+03, -0.73421726E+01, 0.69255672E+01, 0.76826965E+02, 0.51061606E+00, 
    },
    {
      -0.58909216E+03, 0.41609570E+02, -0.19081412E+03, -0.10350248E+03, 0.11202155E+03, -0.11497140E+02, 0.53446163E+02, -0.24178514E+02, 0.73052414E+02, -0.24337910E+02, -0.93695763E+02, 0.24911678E+02, 0.31728121E+02, 
    },
    {
      0.10805177E+03, 0.77080933E+02, -0.67569344E+02, 0.11093394E+02, 0.53251934E+02, -0.18096049E+01, -0.12874349E+02, -0.15589551E+02, -0.36182413E+01, 0.18153871E+01, 0.87414274E+01, 0.19512452E+01, -0.14214747E+02, 
    },
    {
      -0.53066933E+02, 0.58636524E+02, 0.16770851E+02, 0.15051444E+01, 0.77045029E+02, 0.18778749E+01, 0.61496325E+01, 0.81532841E+01, 0.46847639E+01, 0.32412454E+00, -0.78524041E+01, -0.42911925E+01, -0.19247583E+01, 
    },
    {
      0.23769147E+03, -0.61533985E+02, -0.47820061E+02, -0.34214542E+02, 0.40193143E+01, 0.38487625E+02, -0.32399452E+02, 0.30905638E+02, -0.55419487E+02, -0.24066789E+00, -0.41797304E+01, -0.32275726E+02, 0.15222611E+01, 
    },
    {
      0.22423343E+03, -0.12581701E+02, 0.80397057E+02, 0.37280334E+02, -0.41082157E+02, 0.31536264E+01, -0.23678221E+02, 0.94953613E+01, -0.29353230E+02, 0.94907122E+01, 0.40155750E+02, -0.10989323E+02, -0.12388154E+02, 
    },
    {
      0.20844085E-01, -0.80697089E-01, 0.30533599E-01, -0.64598656E+00, -0.16657443E+00, -0.21123195E-01, 0.92248200E-02, -0.25574602E-02, -0.73007592E-02, 0.12166507E-01, -0.12164736E-01, 0.68045142E-02, -0.77586402E-02, 
    },
    {
      -0.69821566E-01, 0.10285664E-01, 0.30625435E-01, 0.12556811E+00, -0.77144945E+00, 0.14403922E-01, 0.32232669E-02, -0.18554740E-01, 0.41907090E-02, -0.10089985E-02, 0.43220972E-02, 0.18566685E-02, -0.25719402E-01, 
    },
    {
      -0.36431205E+00, -0.18156898E+00, -0.11474409E+00, 0.64775425E+00, 0.20361809E+00, 0.58565461E-02, -0.85165583E-01, -0.11972090E+00, -0.25655270E+00, -0.12931146E+00, -0.37106670E-01, -0.15892604E+00, -0.45120206E-01, 
    },
    {
      0.10515987E+01, 0.61001372E+00, 0.23907986E+00, -0.96539223E+00, 0.35442737E+00, -0.22381260E+00, 0.19057482E+00, -0.75943330E-02, 0.47220100E-01, -0.86490205E-02, 0.44078577E-01, -0.37896223E-01, -0.78807129E-02, 
    },
    {
      0.27450144E+00, -0.69624442E+00, -0.14316448E+01, 0.18575617E+01, 0.37408116E+01, -0.57409221E+00, -0.48675659E+00, 0.22489709E+00, -0.55995405E-01, 0.24701399E+00, 0.39949957E-01, 0.95097050E-02, 0.15562671E+00, 
    },
    {
      0.29598875E+01, -0.22428038E+01, 0.16291965E+01, -0.38134632E+01, 0.55432544E+01, -0.76043957E+00, -0.81076163E+00, 0.97178990E+00, 0.18208525E+00, -0.16950764E-01, 0.58102113E+00, -0.36129999E+00, 0.14220279E+00, 
    },
    {
      0.34898536E+01, 0.21215761E+01, 0.22773728E+01, -0.46729574E+01, 0.32191403E+01, 0.10207694E+01, 0.40245906E+00, 0.71286887E+00, 0.25590413E+01, 0.10251747E+01, -0.47245717E+00, 0.82578117E+00, -0.63747823E-01, 
    },
    {
      -0.69263606E+01, -0.71795802E+01, 0.26847789E+01, 0.42841086E+01, -0.33002298E+01, 0.25364313E+01, -0.16150088E+01, 0.56352544E+00, 0.55770433E+00, 0.44689545E+00, -0.62278980E+00, -0.38403094E+00, -0.24888094E+00, 
    },
    {
      -0.28185658E+01, 0.10438052E+02, 0.53413100E+01, 0.12666808E+02, 0.10997744E+02, 0.27631536E+01, 0.28622413E+01, -0.12079479E+01, 0.16985104E+01, -0.21927736E+01, 0.75003189E+00, 0.98793454E-01, -0.10628952E+01, 
    },
    {
      -0.15150089E+02, 0.23410828E+02, -0.12058938E+02, -0.97661982E+01, -0.59994440E+01, 0.12919568E+01, 0.10240656E+01, -0.46654758E+01, -0.17236120E+01, 0.58620757E+00, -0.33529608E+01, 0.26049089E+01, -0.11627923E+00, 
    },
    {
      -0.76791863E+01, -0.53989253E+01, -0.93963947E+01, 0.89348888E+01, -0.13714699E+02, -0.40793180E+01, -0.13703756E+01, -0.11329166E+01, -0.66214280E+01, -0.19741457E+01, 0.19676812E+01, -0.15802444E+01, 0.74486417E+00, 
    },
    {
      0.10005783E+02, 0.22066399E+02, -0.93126631E+01, -0.34220302E+01, 0.73809199E+01, -0.56643796E+01, 0.28039861E+01, -0.20841560E+01, -0.25536554E+01, -0.20768545E+01, 0.14609894E+01, 0.22773330E+01, 0.10445656E+01, 
    },
    {
      0.88902111E+01, -0.21831713E+02, -0.75237002E+01, -0.45947899E+02, -0.52254841E+02, -0.43777122E+01, -0.65832415E+01, 0.27934115E+01, -0.51957211E+01, 0.45358448E+01, -0.29035003E+01, -0.80293524E+00, 0.31588891E+01, 
    },
    {
      0.30450523E+02, -0.57946457E+02, 0.24090494E+02, 0.50759991E+02, -0.16110195E+02, 0.35295739E+01, 0.28397505E+01, 0.93683434E+01, 0.46508346E+01, -0.98472559E+00, 0.71794248E+01, -0.56259923E+01, 0.42344132E+00, 
    },
    {
      0.37496376E+01, 0.41398544E+01, 0.89929867E+01, -0.21592083E+01, 0.15710482E+02, 0.37366185E+01, 0.12688895E+01, 0.56009912E+00, 0.45878363E+01, 0.10043666E+01, -0.15857160E+01, 0.11541928E+01, -0.63341975E+00, 
    },
    {
      -0.16841691E+01, -0.19766069E+02, 0.61021037E+01, -0.47736259E+01, -0.30741682E+01, 0.37985227E+01, -0.14054892E+01, 0.19982285E+01, 0.24144678E+01, 0.21827767E+01, -0.78220928E+00, -0.24638696E+01, -0.81352621E+00, 
    },
    {
      -0.86120291E+01, 0.12768568E+02, 0.40449328E+01, 0.35735741E+02, 0.45281525E+02, 0.14536483E+01, 0.48580317E+01, -0.21178229E+01, 0.40526733E+01, -0.26759613E+01, 0.23512807E+01, 0.86008638E+00, -0.27273736E+01, 
    },
    {
      -0.23491453E+02, 0.42392094E+02, -0.13589549E+02, -0.45290329E+02, 0.22402483E+02, -0.52239609E+01, -0.39634597E+01, -0.66355700E+01, -0.30386250E+01, 0.24522121E+00, -0.52094841E+01, 0.38612597E+01, -0.76240212E+00, 
    },
    {
      -0.14890511E+00, -0.24325432E-01, -0.38893390E-01, 0.66723570E-01, 0.59333660E-01, -0.36929664E+00, -0.60848296E+00, 0.94836196E-02, 0.54205763E-02, -0.15333358E-01, -0.11106117E-02, 0.22575622E-02, 0.23593083E-02, 
    },
    {
      -0.22660287E-01, 0.45373976E-01, 0.50671652E-01, -0.36677148E-01, 0.63135192E-01, 0.61916250E+00, -0.39381266E+00, -0.22737507E-01, 0.37895903E-01, -0.11128758E-01, -0.91112573E-02, -0.10985604E-01, 0.15489354E-01, 
    },
    {
      -0.22802021E+00, 0.26512158E+00, -0.22216140E+00, 0.10585243E+00, -0.16293724E+00, -0.69953322E+00, -0.20355964E+00, 0.23573788E-01, 0.29221939E-01, -0.58141425E-01, -0.10479219E+00, 0.38031027E-01, 0.34238581E-01, 
    },
    {
      -0.32278037E+00, 0.10433105E+00, -0.70636101E-01, -0.45763806E-01, 0.49903285E-01, 0.49318537E+00, -0.80187595E+00, -0.83227634E-01, 0.23710161E-01, -0.30420637E-01, -0.10782068E-01, 0.52331045E-01, 0.40990692E-01, 
    },
    {
      0.75721157E+00, -0.10054682E+00, -0.43790352E+00, 0.35622928E-01, -0.68008196E+00, 0.13344156E+01, 0.10325201E+01, 0.21211481E+00, -0.52811496E-01, 0.58822650E-01, -0.29041562E-01, 0.13278130E+00, 0.60832474E-01, 
    },
    {
      -0.51921064E+00, 0.73404215E-01, -0.14045806E+00, 0.56511676E+00, -0.10780083E+01, -0.90128416E+00, 0.18347827E+01, 0.23589993E-01, 0.19664790E+00, 0.10371149E+00, 0.16768134E+00, 0.25422749E+00, -0.27516479E-01, 
    },
    {
      0.47642750E+00, -0.12712888E+01, 0.50919574E+00, -0.37558299E-01, 0.41861004E+00, 0.72344553E+00, 0.58304822E+00, -0.37630741E-01, -0.21993767E-01, 0.15301681E+00, 0.14597137E+00, -0.88676058E-01, -0.34135389E-02, 
    },
    {
      0.10227042E+01, -0.10414064E+00, 0.21676075E+00, -0.18772216E+00, 0.27981552E+00, -0.11495651E+01, 0.94243652E+00, 0.13660996E+00, -0.18090218E+00, 0.32121949E-01, 0.75706489E-01, -0.24605010E+00, -0.13862634E+00, 
    },
    {
      -0.76268691E+00, 0.10962813E+01, 0.10399640E+01, 0.96028954E-01, 0.77267742E+00, -0.18069416E+01, -0.60100126E+00, -0.48312178E+00, -0.72206438E-01, 0.13322057E-02, -0.17916759E-01, -0.28112444E+00, -0.19206632E-02, 
    },
    {
      0.68990123E+00, -0.17288908E+00, 0.42733526E+00, -0.42356339E+00, 0.20443857E+01, 0.25324249E+00, -0.22968805E+01, 0.21025427E+00, -0.64606416E+00, -0.20289107E+00, -0.37085348E+00, -0.52354544E+00, -0.47585368E-01, 
    },
    {
      0.59121337E-01, 0.75014145E-02, 0.32973826E-01, 0.16474618E-01, 0.32099034E-02, 0.20523887E-01, 0.14850210E-01, 0.15989561E+00, -0.22012538E+00, 0.11682349E-01, 0.37532360E-02, 0.28839896E-02, 0.90501476E-02, 
    },
    {
      0.25123442E-01, -0.54974113E-01, -0.40254004E-01, -0.64294338E-01, 0.18476812E-01, 0.54143360E-02, 0.26773563E-01, 0.22185430E+00, 0.18453135E+00, 0.40824950E-03, -0.20066464E-01, 0.10584317E-01, -0.42071915E-02, 
    },
    {
      -0.11119860E-01, -0.84521756E-01, -0.72065502E-01, -0.10357253E+00, 0.25481060E-01, -0.31697039E-01, -0.34495186E-01, -0.12500069E+00, -0.29316071E+00, -0.50189231E-01, 0.20690957E-01, 0.20984372E-01, 0.46142563E-01, 
    },
    {
      0.31204326E-02, -0.15957630E+00, -0.45523688E-01, 0.11955297E-02, 0.21015655E-01, 0.35082828E-01, -0.32445975E-01, 0.22741143E+00, -0.11440502E+00, 0.16439863E-01, -0.42733331E-02, 0.36579225E-01, -0.35476536E-01, 
    },
    {
      0.20448800E-01, 0.88741951E-01, -0.45464959E-01, -0.15399729E-01, 0.49765386E-01, -0.19545292E-01, 0.20340797E-01, -0.11448290E-01, 0.21605456E-01, 0.11625878E+00, 0.15443808E+00, -0.68518179E-02, 0.68731960E-02, 
    },
    {
      0.44177953E-01, 0.17190436E-01, -0.34843344E-01, -0.55796955E-01, -0.18618047E-01, -0.13446327E-02, -0.13807907E-01, 0.16442720E-01, -0.63370585E-01, -0.13994679E+00, 0.10775572E+00, 0.18059421E-01, -0.11631262E-02, 
    },
    {
      0.56581877E-01, -0.43296669E-01, 0.87594809E-02, -0.19183895E-01, 0.22148756E-01, 0.16138326E-01, 0.35961125E-01, 0.62827743E-02, -0.29091935E-02, 0.17914118E-01, 0.23472050E-01, -0.40293451E-01, 0.52220732E-01, 
    },
    {
      -0.28325748E-01, 0.18121796E-01, -0.55669233E-01, -0.43550450E-01, -0.41511986E-01, 0.28219909E-03, -0.45169748E-01, 0.30541521E-01, -0.75302455E-02, 0.11561214E-01, -0.71855863E-02, -0.42805452E-01, -0.74895360E-01, 
    },
    {
      -0.40362079E-01, 0.39089788E-02, -0.43586571E-01, -0.13430195E-01, 0.19074345E-01, -0.14927471E-02, -0.83385846E-02, 0.12412880E-01, -0.39789891E-02, -0.93350494E-02, 0.31809614E-03, 0.69165085E-02, 0.16948413E-01, 
    },
    {
      0.40239409E-01, 0.21818378E-02, -0.23886045E-01, -0.19706257E-02, 0.17834106E-01, 0.75256284E-02, -0.88512655E-02, 0.25555318E-01, 0.24699425E-01, -0.68906411E-02, -0.15585792E-02, 0.12531394E-01, -0.15920257E-01, 
    },
    {
      0.71578033E-01, 0.36469545E-01, -0.18202318E-01, 0.23860386E-01, -0.15428271E-02, -0.31324875E-01, 0.12997811E-01, 0.49806456E-02, -0.13169400E-01, 0.15163812E-01, 0.45060008E-02, 0.96705575E-02, 0.32339197E-01, 
    },
    {
      -0.71236528E-01, 0.24793777E-03, 0.38303085E-01, 0.23001375E-01, -0.34666590E-02, 0.19711124E-01, 0.14156577E-01, -0.15163544E-02, 0.14914329E-01, -0.16906057E-01, 0.68336162E-02, 0.72287614E-02, -0.10352528E-01, 
    },
  },
  {
    {
      0.84432182E+01, -0.17606992E-01, -0.12995030E-01, 0.30051462E-01, 0.45477942E-01, 0.12136347E-01, -0.20886991E-01, -0.25218520E-01, 0.38802756E-02, 0.50634779E-02, 0.17426111E-01, -0.19343946E-01, 0.21120729E-01, 
    },
    {
      0.13067631E+01, -0.19031613E+00, -0.47110280E+00, 0.15309569E+00, -0.25000983E+00, -0.21639533E+00, 0.13572131E+00, 0.16442597E-01, 0.29920632E+00, -0.35695668E-01, 0.72213798E-02, 0.36948250E-03, -0.19350252E+00, 
    },
    {
      0.20192820E+02, 0.10371314E+00, -0.12936909E+01, -0.47384906E+00, 0.43617851E+00, -0.88357627E+00, 0.84384918E+00, 0.40130630E+00, -0.24254835E+00, 0.19504283E+00, -0.32403544E+00, 0.68768036E+00, -0.70447671E+00, 
    },
    {
      0.10385976E+02, 0.26844606E+01, 0.91865940E+01, -0.28847570E+01, 0.69040632E+01, 0.30880930E+01, -0.16791331E+01, -0.77926904E+00, -0.40792027E+01, 0.16587726E+01, 0.11048369E+01, 0.56877963E-01, 0.26553495E+01, 
    },
    {
      -0.94894234E+02, -0.74020910E+00, 0.97572689E+01, -0.13988616E+01, -0.48308020E+01, 0.61252832E+01, -0.29142206E+01, -0.31117852E+01, 0.10199926E+01, -0.25860064E+01, 0.15488667E+01, -0.52199402E+01, 0.62431121E+01, 
    },
    {
      -0.15947099E+03, -0.33153683E+02, -0.38277382E+02, 0.16296080E+02, -0.37176064E+02, -0.16504719E+02, 0.97844954E+01, 0.60243282E+01, 0.18602617E+02, -0.88784866E+01, -0.70023069E+01, -0.85667688E+00, -0.10270858E+02, 
    },
    {
      0.14434691E+03, 0.40043654E+01, -0.22543146E+02, 0.10819620E+02, 0.13452056E+02, -0.14324649E+02, 0.14989966E+01, 0.83210297E+01, -0.11521393E+01, 0.86834040E+01, -0.32280769E+01, 0.13568054E+02, -0.17570894E+02, 
    },
    {
      0.46371674E+03, 0.11013431E+03, 0.54929173E+02, -0.38762981E+02, 0.82882599E+02, 0.38311665E+02, -0.25024160E+02, -0.15210931E+02, -0.36635254E+02, 0.18874268E+02, 0.14921486E+02, 0.31062043E+01, 0.17111540E+02, 
    },
    {
      -0.10361338E+03, -0.10559738E+02, 0.20353815E+02, -0.17083015E+02, -0.12879775E+02, 0.13625260E+02, 0.28179276E+01, -0.85639162E+01, -0.39836466E+00, -0.10996532E+02, 0.32205973E+01, -0.14092367E+02, 0.19663298E+02, 
    },
    {
      -0.51334308E+03, -0.13410316E+03, -0.27003073E+02, 0.40749172E+02, -0.81037025E+02, -0.38719093E+02, 0.27226254E+02, 0.15593900E+02, 0.32171307E+02, -0.18132610E+02, -0.12962888E+02, -0.39625747E+01, -0.13030040E+02, 
    },
    {
      0.30432690E+02, 0.74267569E+01, -0.65251384E+01, 0.81919823E+01, 0.35932457E+01, -0.45463281E+01, -0.21767290E+01, 0.29457211E+01, 0.77555543E+00, 0.47232051E+01, -0.12480311E+01, 0.50437918E+01, -0.76494961E+01, 
    },
    {
      0.19720856E+03, 0.54842201E+02, 0.14811361E+01, -0.15508772E+02, 0.28584126E+02, 0.13968620E+02, -0.10371297E+02, -0.56411085E+01, -0.10385420E+02, 0.65630751E+01, 0.39054112E+01, 0.16419491E+01, 0.37288151E+01, 
    },
    {
      0.73617727E-01, 0.15868278E+01, 0.15948020E+01, -0.51909208E-01, 0.10965753E+00, 0.42061005E-01, 0.86369058E-02, -0.35956562E-01, -0.24439633E-01, 0.13813865E-01, 0.33473477E-01, -0.12541716E-01, 0.13831583E-01, 
    },
    {
      -0.22122271E+00, -0.16716869E+01, 0.17568765E+01, -0.10806262E+00, -0.60551725E-01, 0.32627724E-01, -0.40097758E-01, 0.32800335E-01, -0.22351671E-01, -0.30436333E-01, 0.85034929E-02, 0.44959588E-02, -0.57923377E-01, 
    },
    {
      -0.13319713E+01, 0.20320992E+00, 0.28417442E+01, -0.26053941E+00, 0.41736737E-01, 0.48073849E+00, 0.30752096E+00, 0.27773425E+00, 0.58028370E+00, -0.27988768E+00, 0.20987786E+00, 0.21454689E+00, -0.21824501E+00, 
    },
    {
      -0.32366142E+00, -0.18102030E+01, 0.73649567E+00, -0.12784365E+01, 0.13987561E+00, 0.29401943E+00, -0.11734486E+00, 0.30285731E+00, 0.47187995E-01, -0.40725458E+00, 0.69726735E-01, -0.14392924E-01, -0.16440730E+00, 
    },
    {
      0.43222728E+01, 0.25277067E+02, -0.94375896E+01, 0.12424446E+01, 0.19729581E-01, -0.89677191E+00, -0.36436412E+00, 0.90389627E+00, -0.13076818E+00, 0.15358768E+00, -0.36496457E+00, 0.83983088E+00, -0.97137824E-01, 
    },
    {
      -0.39563198E+01, 0.94551926E+01, 0.17336029E+02, 0.19375322E+01, 0.29811153E+01, 0.36434084E+00, 0.96826065E+00, 0.76319680E-01, 0.63908470E+00, 0.66578782E+00, -0.31862479E+00, -0.63156575E+00, 0.88741764E-01, 
    },
    {
      0.31899084E+02, 0.26338453E+02, -0.26924992E+02, -0.28954151E+01, 0.11245874E+01, -0.62383070E+01, -0.71671276E+01, -0.44009361E+01, -0.11893276E+02, 0.59641714E+01, -0.12007704E+01, -0.48183508E+01, 0.41470280E+01, 
    },
    {
      0.29249310E+01, 0.10189669E+02, 0.64222164E+01, 0.16846102E+02, -0.56195569E+01, -0.38048911E+01, 0.29025004E+01, -0.38915317E+01, -0.16548055E+01, 0.71972709E+01, 0.18794578E+01, 0.20752627E+00, 0.15136871E+01, 
    },
    {
      -0.47960709E+02, -0.14527844E+03, 0.13558260E+03, -0.15012978E+02, -0.89980879E+01, 0.33358767E+01, 0.65263834E+01, -0.57562585E+01, 0.23737571E+01, -0.32089403E+01, 0.66337514E+00, -0.98810377E+01, -0.12185401E+01, 
    },
    {
      0.54375412E+02, -0.13382069E+03, -0.69432632E+02, -0.15594945E+02, -0.28076584E+02, -0.55643253E+01, -0.81322355E+01, -0.26028937E+00, -0.52387090E+01, -0.39641640E+01, 0.58208838E+01, 0.58059587E+01, 0.43214092E+01, 
    },
    {
      -0.18676460E+03, -0.16449841E+03, 0.18679526E+03, 0.44028065E+02, 0.67239275E+01, 0.20764881E+02, 0.43572437E+02, 0.15105406E+02, 0.72933273E+02, -0.35417358E+02, -0.33102422E+01, 0.33491127E+02, -0.28131252E+02, 
    },
    {
      0.13132272E+02, -0.71697983E+02, -0.14838391E+02, -0.88297775E+02, 0.35739063E+02, 0.24429436E+02, -0.19637230E+02, 0.15767020E+02, 0.12106133E+02, -0.38460899E+02, -0.17444717E+02, -0.57627004E+00, -0.71105013E+01, 
    },
    {
      0.16358971E+03, 0.30532544E+03, -0.45578516E+03, 0.51725403E+02, 0.42052006E+02, -0.12835170E+01, -0.28778358E+02, 0.11572843E+02, -0.11792854E+02, 0.97184315E+01, 0.28023026E+01, 0.34010338E+02, 0.76306810E+01, 
    },
    {
      -0.19060997E+03, 0.45207547E+03, 0.60605202E+02, 0.49964787E+02, 0.83181343E+02, 0.17889486E+02, 0.23796255E+02, -0.10524486E+01, 0.17759972E+02, 0.98242826E+01, -0.24883154E+02, -0.17036594E+02, -0.18363985E+02, 
    },
    {
      0.47393274E+03, 0.36210413E+03, -0.54417682E+03, -0.15719064E+03, -0.48869652E+02, -0.19227535E+02, -0.11137081E+03, -0.14016641E+02, -0.18484474E+03, 0.86145981E+02, 0.26243605E+02, -0.93299721E+02, 0.76153831E+02, 
    },
    {
      -0.10607853E+03, 0.18373109E+03, -0.71254395E+02, 0.21552979E+03, -0.88725677E+02, -0.66744774E+02, 0.48730526E+02, -0.28787111E+02, -0.34096436E+02, 0.84370399E+02, 0.52670143E+02, -0.46487325E+00, 0.18449718E+02, 
    },
    {
      -0.22080930E+03, -0.29196872E+03, 0.61042780E+03, -0.71494499E+02, -0.68313850E+02, -0.85744381E+01, 0.44273830E+02, -0.82676754E+01, 0.20448513E+02, -0.80523005E+01, -0.84449396E+01, -0.45622372E+02, -0.12797120E+02, 
    },
    {
      0.25156476E+03, -0.59986975E+03, 0.34923374E+02, -0.63288609E+02, -0.98415474E+02, -0.20724583E+02, -0.28114664E+02, 0.44651847E+01, -0.23789305E+02, -0.11882715E+02, 0.37959538E+02, 0.19314947E+02, 0.25713858E+02, 
    },
    {
      -0.54166937E+03, -0.33657809E+03, 0.70614557E+03, 0.21372903E+03, 0.80236282E+02, -0.66625953E+01, 0.12710117E+03, -0.65174923E+01, 0.20572482E+03, -0.91904800E+02, -0.43451694E+02, 0.11079334E+03, -0.87262184E+02, 
    },
    {
      0.19350169E+03, -0.21782204E+03, 0.19717972E+03, -0.23897552E+03, 0.95249496E+02, 0.76332985E+02, -0.49345135E+02, 0.26176369E+02, 0.39268757E+02, -0.81754929E+02, -0.66510406E+02, 0.24783335E+01, -0.23144762E+02, 
    },
    {
      0.10398347E+03, 0.10713031E+03, -0.28724725E+03, 0.34307171E+02, 0.36974197E+02, 0.81658230E+01, -0.22095419E+02, 0.13036780E+01, -0.11200826E+02, 0.97571468E+00, 0.54882064E+01, 0.21151997E+02, 0.67120991E+01, 
    },
    {
      -0.11383261E+03, 0.27709833E+03, -0.45287556E+02, 0.26823547E+02, 0.40176460E+02, 0.79526610E+01, 0.11477461E+02, -0.35184112E+01, 0.10715305E+02, 0.56945720E+01, -0.19063519E+02, -0.72694077E+01, -0.11852451E+02, 
    },
    {
      0.22522438E+03, 0.11369879E+03, -0.33281451E+03, -0.10020547E+03, -0.39448936E+02, 0.11468914E+02, -0.53004147E+02, 0.10001610E+02, -0.83203133E+02, 0.35805698E+02, 0.21695692E+02, -0.46803165E+02, 0.35515347E+02, 
    },
    {
      -0.10584347E+03, 0.98553574E+02, -0.12213636E+03, 0.96482056E+02, -0.38166542E+02, -0.30556103E+02, 0.17174086E+02, -0.97377214E+01, -0.15434036E+02, 0.29186680E+02, 0.30085251E+02, -0.17299670E+01, 0.10782919E+02, 
    },
    {
      -0.27865281E-01, -0.14160953E+00, -0.46167344E-01, -0.83359647E+00, 0.54028988E+00, 0.17667174E-01, -0.23999680E-02, -0.50582860E-01, -0.34205623E-01, 0.24929952E-01, -0.20637264E-01, -0.54828986E-02, -0.16638445E-02, 
    },
    {
      -0.99061303E-01, -0.61199144E-02, 0.16169669E-01, -0.50423771E+00, -0.87746030E+00, -0.14626432E-02, 0.91258921E-02, 0.41810900E-01, 0.27339887E-02, -0.15622666E-02, 0.32905261E-02, 0.18039158E-01, -0.38937837E-01, 
    },
    {
      -0.11977397E+00, -0.37973851E-01, 0.51543677E+00, -0.18759683E+01, 0.63566899E+00, 0.28123093E+00, -0.29566622E+00, -0.18840617E-01, -0.36266062E-01, -0.26874974E+00, -0.17266946E+00, 0.10599524E-01, -0.83736926E-01, 
    },
    {
      0.49956435E+00, 0.68339938E+00, 0.18975423E+00, -0.11983259E+01, -0.20601561E+01, 0.58113441E-01, 0.15165167E+00, 0.12512311E+00, 0.19267312E+00, -0.55146988E-01, 0.15980954E+00, 0.82305297E-01, -0.42632781E-01, 
    },
    {
      0.23057373E+01, 0.76537257E+00, 0.43733449E+01, -0.50523310E+01, 0.28448110E+01, 0.51284897E+00, -0.16459599E+01, 0.74293798E+00, 0.66224128E-01, -0.41510153E+00, 0.26375905E+00, 0.37355468E+00, -0.17590469E+00, 
    },
    {
      0.11372099E+01, -0.12647940E+00, 0.25441618E+01, -0.39932528E+01, -0.41477337E+01, 0.13616855E+01, -0.13583241E+00, -0.12699013E+01, -0.66975029E-02, -0.43774159E-02, 0.72404754E+00, -0.79435611E+00, 0.15491714E+00, 
    },
    {
      0.19683877E+01, 0.29083934E+01, -0.72367439E+01, 0.59465566E+01, -0.19175700E+01, -0.14585569E+01, 0.16693540E+01, -0.56365860E+00, 0.17722517E+01, 0.32326121E+01, 0.53729033E+00, -0.88220710E+00, -0.56293678E+00, 
    },
    {
      -0.41296825E+01, -0.31065300E+01, 0.19166176E+01, 0.91364346E+01, 0.66980739E+01, 0.77398407E+00, -0.30196124E+00, -0.13194689E+01, -0.16660863E+01, 0.88819361E+00, -0.11270262E+01, -0.17106761E+01, 0.95284730E-01, 
    },
    {
      -0.11591570E+02, 0.10206953E+02, -0.44386238E+02, 0.47346340E+02, -0.79038370E+00, -0.90883474E+01, 0.11115670E+02, -0.19898731E+01, 0.37733054E+00, 0.13326268E+01, -0.29739472E+00, -0.21107407E+01, 0.19746448E+01, 
    },
    {
      -0.77565346E+01, 0.19142870E+02, -0.19318724E+02, 0.79751496E+01, 0.45732677E+02, -0.11700178E+02, -0.21409187E+01, 0.95194101E+01, -0.12891829E+01, -0.70309061E+00, -0.49068875E+01, 0.54949450E+01, 0.72287112E+00, 
    },
    {
      -0.51350603E+01, -0.88256512E+01, 0.16213120E+02, -0.95362787E+01, -0.49770875E+01, 0.10020863E+01, -0.38989060E+01, 0.27658885E+01, -0.65731020E+01, -0.83604755E+01, -0.15559013E+00, 0.25949471E+01, 0.29289787E+01, 
    },
    {
      0.96731539E+01, 0.52264996E+01, -0.65303130E+01, -0.14551664E+02, -0.86602306E+01, -0.16371939E+01, -0.16511111E+01, 0.34544625E+01, 0.35504460E+01, -0.31492507E+01, 0.25533750E+01, 0.52310605E+01, -0.48787493E-01, 
    },
    {
      0.20374853E+02, -0.30132755E+02, 0.10810119E+03, -0.10528988E+03, -0.18006445E+02, 0.24039795E+02, -0.24310305E+02, 0.27824404E+01, -0.91876078E+00, -0.13190975E+01, -0.37454176E+00, 0.39671352E+01, -0.47484903E+01, 
    },
    {
      0.22527531E+02, -0.60154240E+02, 0.38476719E+02, 0.45293006E+00, -0.10749376E+03, 0.29633495E+02, 0.56231313E+01, -0.19200817E+02, 0.49150085E+01, 0.26779370E+01, 0.99957170E+01, -0.12485248E+02, -0.21405015E+01, 
    },
    {
      0.15377275E+01, 0.74568505E+01, -0.10354218E+02, 0.10922906E+02, 0.13785338E+02, 0.83337110E+00, 0.27636354E+01, -0.27910285E+01, 0.50508113E+01, 0.60138631E+01, -0.15283442E+00, -0.17226232E+01, -0.27341883E+01, 
    },
    {
      -0.54970169E+01, -0.42347341E+01, 0.50668473E+01, 0.30705661E+00, 0.80789337E+01, 0.52706480E+00, 0.30988753E+01, -0.23904357E+01, -0.23165927E+01, 0.27464480E+01, -0.15124305E+01, -0.44464526E+01, 0.11784693E+00, 
    },
    {
      -0.12952716E+02, 0.21251139E+02, -0.73948364E+02, 0.71904915E+02, 0.22982941E+02, -0.17152084E+02, 0.16589590E+02, -0.19711390E+01, 0.14029673E+00, 0.53307498E+00, 0.45210415E+00, -0.22865257E+01, 0.31447535E+01, 
    },
    {
      -0.21541494E+02, 0.48531136E+02, -0.20743677E+02, -0.97520990E+01, 0.76132256E+02, -0.22496307E+02, -0.30357060E+01, 0.11248891E+02, -0.39207585E+01, -0.22813702E+01, -0.63165536E+01, 0.86511211E+01, 0.12434313E+01, 
    },
    {
      -0.15622102E+00, -0.59632380E-01, -0.90208650E-01, 0.19414164E-01, 0.73269002E-01, -0.56129473E+00, -0.63541585E+00, -0.16588993E-01, -0.34606468E-01, -0.21938419E-02, -0.21088144E-01, -0.10219838E-01, 0.62936130E-02, 
    },
    {
      0.10229758E+00, 0.31424812E-02, 0.12619580E+00, -0.10478984E-01, 0.10643311E+00, 0.63880980E+00, -0.58100104E+00, -0.80984796E-03, -0.32982638E-02, -0.10336742E-01, -0.12900264E-01, -0.13739295E-01, 0.22639144E-01, 
    },
    {
      0.11596611E+00, 0.24706095E+00, -0.15466779E+00, -0.18783578E+00, -0.65919489E-01, -0.62681031E+00, -0.66545194E+00, 0.32444045E-01, 0.13160099E+00, 0.13321731E-01, -0.10927579E+00, 0.71161732E-01, 0.11962270E-01, 
    },
    {
      -0.13758811E+00, -0.70412166E-01, 0.23057573E+00, -0.63733466E-01, -0.23197371E-02, 0.77353519E+00, -0.66607153E+00, -0.11856486E+00, 0.19452100E+00, 0.44829205E-01, 0.53664832E-02, 0.81992827E-01, -0.22440353E-01, 
    },
    {
      0.17074040E+01, 0.56844890E+00, -0.19279140E+00, -0.48811182E+00, -0.82032895E+00, 0.17906553E+01, 0.46447378E+00, 0.62934273E+00, 0.11496823E+00, 0.72347268E-01, 0.16785286E+00, 0.25202709E+00, 0.64511701E-01, 
    },
    {
      -0.21264596E+01, 0.78551662E+00, -0.27990738E+00, 0.20393796E+00, -0.17369808E+01, -0.49138665E+00, 0.25081100E+01, -0.13376953E+00, 0.84815377E+00, 0.67785099E-01, -0.21081494E-01, 0.27407336E+00, -0.19572496E+00, 
    },
    {
      0.22916576E+00, -0.89228165E+00, 0.14403734E+00, 0.62578565E+00, 0.42053275E-01, -0.14299740E+00, 0.10011269E+01, -0.10695215E+00, -0.26902208E+00, 0.31509314E-01, 0.18015152E+00, -0.22677259E+00, 0.45944661E-01, 
    },
    {
      0.78097403E+00, 0.17563751E+00, -0.79983306E+00, -0.23202594E-01, 0.51307887E+00, -0.12497675E+01, 0.63095026E-01, 0.27101776E+00, -0.64966005E+00, -0.12470666E+00, 0.57328966E-01, -0.25540158E+00, 0.30125923E-01, 
    },
    {
      -0.36262932E+01, 0.51153991E-04, 0.65369076E+00, 0.12630241E+01, 0.59161860E+00, -0.33515153E+01, 0.12459384E+00, -0.92280024E+00, 0.16300046E+00, -0.98683946E-02, -0.39698192E+00, -0.33663470E+00, -0.30247193E-01, 
    },
    {
      0.35409751E+01, -0.94228369E+00, 0.59330601E+00, -0.47572304E-01, 0.26288931E+01, -0.11341671E+00, -0.40985465E+01, 0.55726737E+00, -0.15782918E+01, -0.92821464E-01, -0.86877599E-01, -0.41220182E+00, 0.23814109E+00, 
    },
    {
      0.40369980E-01, 0.69174074E-01, 0.90490915E-02, 0.33020664E-01, 0.11527482E-01, 0.12345885E-01, 0.26705259E-01, 0.19554184E+00, -0.36173043E+00, 0.22097254E-01, -0.12866687E-01, 0.62658121E-02, 0.24714425E-01, 
    },
    {
      0.16548369E-01, -0.36080949E-01, -0.25610084E-01, -0.21156788E-01, 0.43353230E-01, -0.31337328E-01, 0.13042549E-01, 0.34459081E+00, 0.21506374E+00, 0.16185859E-01, -0.13301040E-01, 0.14696219E-01, 0.17140709E-01, 
    },
    {
      0.11640607E+00, -0.35488162E-01, -0.61050300E-01, -0.10260254E+00, 0.45410369E-01, -0.78796092E-02, -0.90719406E-02, -0.14850970E+00, -0.58185756E+00, -0.61733756E-01, -0.32026162E-02, 0.21957425E-01, 0.52263454E-01, 
    },
    {
      0.55307180E-01, -0.17401587E-01, -0.41585356E-01, 0.76646626E-01, -0.57702390E-02, -0.12208957E+00, 0.50012428E-01, 0.47269493E+00, 0.20059482E-02, 0.50178025E-01, -0.18123774E-02, 0.58514785E-01, -0.23311080E-01, 
    },
    {
      0.39524496E-01, 0.80552071E-01, -0.42998645E-01, 0.16749263E-01, 0.31579159E-01, -0.44717076E-02, -0.84887655E-03, -0.49433589E-01, 0.10765205E-01, 0.21353492E+00, 0.13728055E+00, -0.10284929E-01, 0.20548493E-01, 
    },
    {
      0.99338487E-01, 0.17302408E-02, 0.41451098E-04, -0.84679388E-01, -0.11557483E-01, -0.22574198E-01, -0.23921095E-01, -0.82226349E-02, -0.69644853E-01, -0.13849451E+00, 0.20419706E+00, 0.16953168E-01, 0.82117142E-02, 
    },
    {
      0.72829023E-01, -0.41823316E-01, 0.10375716E-01, -0.56614611E-01, 0.41825403E-01, 0.20270178E-01, 0.12032609E-01, -0.24085421E-01, -0.12075845E-01, 0.61627743E-02, 0.59546165E-01, -0.69686007E-02, 0.10623483E+00, 
    },
    {
      0.54462738E-02, -0.64947754E-02, -0.79987003E-02, -0.12688862E-01, -0.63994959E-01, -0.21709125E-01, -0.43794498E-01, 0.31835452E-01, -0.27470827E-01, -0.70910002E-02, 0.91979457E-02, -0.11307489E+00, -0.22885030E-01, 
    },
    {
      -0.13536142E-01, 0.18244112E-01, -0.90564989E-01, 0.45806570E-02, 0.17037233E-01, 0.58723753E-02, -0.15382156E-01, -0.81376935E-03, -0.10111791E-01, -0.13766921E-01, 0.88332891E-02, -0.10245698E-01, 0.28617415E-02, 
    },
    {
      0.21085773E-01, 0.25080677E-01, -0.33543896E-01, 0.24984222E-01, -0.45952465E-01, -0.19729603E-01, 0.25285017E-01, 0.50980866E-01, -0.13921154E-01, -0.91980677E-02, 0.12159725E-01, -0.17162612E-03, -0.10890553E-01, 
    },
    {
      0.40087324E-01, 0.82861669E-01, -0.52096929E-01, -0.26000416E-01, -0.59353504E-01, -0.32921359E-01, 0.14816161E-01, -0.14302545E-01, -0.12839498E-01, 0.24855364E-01, -0.25451362E-01, 0.88601554E-03, 0.27688909E-01, 
    },
    {
      -0.54229580E-01, 0.40404364E-01, 0.11072406E+00, 0.53175781E-01, -0.27590901E-01, 0.21070372E-01, 0.30208087E-01, 0.12800893E-01, 0.44277799E-02, -0.28233783E-01, 0.11693147E-02, 0.20908635E-01, -0.11176485E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir11_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30831590E+01, 0.36423469E-02, 0.95374268E-02, -0.43797647E-02, -0.18526374E-02, 0.12529454E-02, -0.21983865E-02, 0.29568156E-03, 0.14598125E-02, 
    },
    {
      0.46192214E+00, 0.57096808E-03, 0.68593971E-01, 0.65610953E-01, -0.10962675E-01, -0.29499030E-01, -0.40161178E-01, 0.90372534E-02, 0.10792332E-01, 
    },
    {
      0.59456217E+00, -0.58928527E-01, -0.10835457E+00, 0.15614104E-01, 0.38582690E-01, 0.23725567E-01, 0.37873339E-01, -0.45640215E-01, -0.25683858E-01, 
    },
    {
      -0.58345467E+00, -0.27269027E+00, -0.13519897E+00, -0.14928319E+00, 0.25888463E-01, 0.10878225E+00, 0.17128439E+00, -0.17555457E-01, -0.41636076E-01, 
    },
    {
      -0.47605807E+00, 0.18886384E+00, 0.18850458E+00, 0.15938819E-02, -0.82812428E-01, -0.16693853E-01, -0.74811786E-01, 0.13277733E+00, 0.68222940E-01, 
    },
    {
      0.97611375E-01, 0.27176991E+00, 0.73444307E-01, 0.79687804E-01, -0.56013255E-02, -0.85544743E-01, -0.14200655E+00, 0.61242115E-02, 0.28192669E-01, 
    },
    {
      -0.89809746E-01, -0.11081778E+00, -0.60269386E-01, -0.10322292E-01, 0.58627542E-01, -0.12565220E-01, 0.40889379E-01, -0.88073589E-01, -0.47065701E-01, 
    },
    {
      0.50374349E-02, -0.12933061E+00, -0.36687320E+00, 0.13463667E-02, -0.39664879E-02, 0.91926754E-02, -0.25788704E-02, 0.42144703E-02, 0.51479922E-02, 
    },
    {
      0.33288829E-01, 0.38233611E+00, -0.14407398E+00, 0.14271198E-01, 0.11049573E-01, -0.93471622E-02, -0.82618268E-02, -0.22260593E-02, -0.10530115E-02, 
    },
    {
      -0.19392334E-02, 0.48438212E+00, 0.80777749E-01, 0.84854387E-01, 0.67027077E-01, 0.96795214E-02, -0.77769174E-02, -0.18757645E-01, -0.34126833E-02, 
    },
    {
      0.10197649E-01, -0.10353541E-01, 0.50584143E+00, -0.11593369E+00, -0.44522554E-01, 0.55989962E-01, -0.29334029E-01, 0.31971790E-01, -0.19206982E-01, 
    },
    {
      -0.19130749E+00, 0.67898899E+00, 0.13500415E+01, -0.41417088E-01, -0.41389436E-01, -0.11354158E+00, 0.12479115E+00, -0.15671545E+00, -0.16941839E+00, 
    },
    {
      -0.69109064E+00, -0.15827085E+01, 0.77499503E+00, -0.54360222E-01, -0.76154426E-01, 0.21804528E-01, 0.12542558E+00, 0.97710431E-01, 0.15336756E-01, 
    },
    {
      0.46886271E+00, -0.25143082E+01, 0.51227117E+00, 0.20739187E+00, -0.24356622E+00, -0.41955873E+00, -0.88255458E-01, -0.11599702E+00, -0.16374972E+00, 
    },
    {
      0.60307807E+00, -0.83648968E+00, -0.31556292E+01, 0.47265437E+00, 0.62394643E+00, -0.54705179E+00, 0.20412104E+00, -0.19938691E+00, 0.61526425E-01, 
    },
    {
      0.66733706E+00, -0.84022284E+00, -0.36363804E+00, -0.47948975E-01, 0.91792926E-01, 0.36664313E+00, -0.50559223E+00, 0.43860993E+00, 0.55686808E+00, 
    },
    {
      0.20355382E+01, 0.71406555E+00, -0.12435471E+01, 0.21483092E+00, -0.53310115E-01, -0.59347369E-01, -0.33315611E+00, -0.25482255E+00, -0.24464151E-02, 
    },
    {
      -0.78910613E+00, 0.37705996E+01, -0.86769462E+00, -0.12256935E+01, 0.33561519E+00, 0.13111099E+01, 0.17731486E+00, 0.52530521E+00, 0.42007926E+00, 
    },
    {
      -0.23053181E+01, 0.16463206E+01, 0.60083599E+01, -0.67770940E+00, -0.16302882E+01, 0.13140205E+01, -0.39495325E+00, 0.35773265E+00, -0.20983079E-01, 
    },
    {
      -0.48089162E+00, 0.39525765E+00, -0.48140696E+00, 0.16104813E+00, -0.11935355E-01, -0.33358160E+00, 0.43848920E+00, -0.33588856E+00, -0.45440468E+00, 
    },
    {
      -0.16640769E+01, 0.70364147E+00, 0.72135723E+00, -0.26408896E+00, 0.13320316E+00, 0.60090333E-01, 0.20128188E+00, 0.18900408E+00, -0.34622148E-01, 
    },
    {
      0.22574455E+00, -0.15068560E+01, 0.90492386E+00, 0.10402929E+01, -0.16921978E+00, -0.10404446E+01, -0.66429354E-01, -0.44397834E+00, -0.24560237E+00, 
    },
    {
      0.19569390E+01, -0.13762980E+01, -0.33979950E+01, 0.37711099E+00, 0.11398213E+01, -0.88892829E+00, 0.24327640E+00, -0.19420274E+00, -0.70902784E-02, 
    },
    {
      -0.20724727E-01, -0.35504639E-01, -0.82354777E-03, 0.36352221E-01, -0.11773663E+00, -0.11033534E-01, 0.95688127E-03, -0.97791515E-02, -0.58923671E-02, 
    },
    {
      0.11510791E-01, -0.53077191E-02, -0.21679617E-01, 0.12743253E+00, 0.43944508E-01, -0.99055991E-02, -0.21604956E-02, -0.19020389E-02, 0.15028869E-02, 
    },
    {
      0.34215365E-01, 0.19715223E+00, 0.12483910E+00, -0.11022672E+00, -0.31943321E-01, 0.11501916E-02, -0.29224209E-01, -0.21727003E-01, -0.48656184E-01, 
    },
    {
      -0.12077173E+00, 0.44407237E-01, -0.50276604E-01, 0.74237496E-01, -0.13876393E+00, 0.50061855E-01, -0.89441317E-04, -0.19618968E-01, 0.77808690E-02, 
    },
    {
      0.11498174E+00, 0.30138892E+00, 0.84145188E-01, 0.52387480E-01, -0.16962323E+00, 0.13028961E-01, 0.83306350E-01, 0.12248267E+00, -0.22928387E-01, 
    },
    {
      -0.18141045E+00, -0.73630512E-01, 0.12350223E-01, 0.22824010E+00, 0.88408612E-01, -0.56969866E-01, 0.44197991E-01, 0.42088978E-01, -0.97412907E-01, 
    },
    {
      -0.15992832E+00, -0.10725375E+01, -0.37769392E+00, 0.20365877E+00, -0.10712838E+00, -0.52147180E-01, 0.11348375E+00, 0.12272292E+00, 0.21829335E+00, 
    },
    {
      0.90741438E+00, 0.27918214E+00, 0.21284491E+00, -0.16413610E+00, 0.20872620E+00, -0.26607731E-01, -0.13638993E+00, 0.10097218E+00, -0.90743758E-01, 
    },
    {
      -0.16678724E+00, -0.49267471E+00, -0.25228688E+00, -0.20995039E+00, 0.35344949E+00, -0.42392898E-01, -0.14291544E+00, -0.18131788E+00, 0.37399542E-01, 
    },
    {
      0.22921523E+00, -0.11525336E-01, 0.10362252E+00, -0.50349134E+00, -0.24221870E+00, 0.16809084E+00, -0.94889142E-01, -0.70770048E-01, 0.13145503E+00, 
    },
    {
      0.21850903E+00, 0.14192160E+01, 0.37012824E+00, -0.17118199E-01, 0.49541610E+00, 0.17686056E+00, -0.86773776E-01, -0.13545221E+00, -0.20977929E+00, 
    },
    {
      -0.10618936E+01, -0.52702361E+00, -0.14100182E+00, -0.16452017E+00, 0.87792031E-01, -0.11034585E+00, 0.22220004E+00, -0.97200051E-01, 0.10797976E+00, 
    },
    {
      -0.20363385E-01, -0.53697578E-02, 0.68575093E-02, 0.68112370E-02, 0.10468053E-01, 0.19557819E-01, -0.29966215E-01, 0.26615837E-02, -0.39665838E-03, 
    },
    {
      -0.27748954E-02, -0.12629431E-01, -0.34522379E-02, -0.19603441E-01, -0.74741081E-02, 0.36278360E-01, 0.38917966E-01, 0.62133339E-02, 0.52181524E-02, 
    },
    {
      0.47905695E-01, 0.42130247E-01, 0.10669809E-01, -0.42659171E-01, -0.48365989E-02, -0.12414151E+00, 0.83759082E-02, 0.88774720E-02, -0.24281682E-02, 
    },
    {
      0.65088286E-02, -0.18100817E-01, -0.99380631E-02, -0.49212319E-02, 0.54735452E-03, -0.14428278E-01, -0.12441483E+00, -0.92067998E-02, -0.31998189E-03, 
    },
    {
      0.24574004E-01, -0.10163294E+00, 0.31056121E-01, 0.11012295E-01, -0.45570403E-01, -0.14161536E+00, -0.41898366E-01, 0.22085367E-02, -0.13771434E-01, 
    },
    {
      -0.37635207E-01, -0.79316162E-02, -0.12239707E-01, 0.54416325E-01, -0.23432048E-01, 0.64405173E-01, -0.15231898E+00, -0.39993403E-02, 0.10142059E-01, 
    },
    {
      -0.18101418E-01, -0.84799835E-02, 0.34984993E-03, 0.89346385E-02, 0.47468557E-03, 0.28572041E-02, -0.35473125E-02, 0.14134602E-01, -0.84402598E-02, 
    },
    {
      -0.24490006E-01, 0.59575271E-02, -0.83975289E-02, -0.10783280E-01, 0.10519189E-01, 0.37295434E-02, 0.78448253E-02, 0.11155130E-01, 0.20005925E-01, 
    },
    {
      0.62797785E-01, 0.11900892E-01, -0.27039561E-01, -0.19209160E-01, 0.50548892E-02, -0.49631188E-04, -0.94857474E-03, -0.61566597E-02, -0.72145365E-01, 
    },
    {
      0.32358173E-01, -0.47734085E-01, 0.26209857E-01, 0.28988877E-01, 0.74443221E-02, -0.39548054E-02, 0.37050038E-02, 0.76541923E-01, -0.17525801E-02, 
    },
    {
      0.52079731E-02, -0.92345122E-02, 0.14466225E-02, -0.19333605E-02, -0.76571708E-02, 0.36316388E-02, -0.43371287E-02, -0.36561657E-02, 0.67056152E-02, 
    },
    {
      0.16525778E-02, 0.18420495E-01, -0.17318184E-01, -0.72389068E-02, -0.36425991E-02, 0.67126690E-02, 0.62284856E-02, -0.25143016E-02, -0.54201265E-02, 
    },
    {
      -0.49585272E-02, 0.72677066E-02, 0.95763793E-02, 0.33443144E-02, 0.68702078E-02, 0.38347980E-02, -0.47801477E-02, -0.56901621E-02, 0.15504239E-02, 
    },
    {
      -0.65200361E-02, -0.13484796E-01, -0.16534761E-01, -0.93058543E-02, -0.70031788E-02, 0.61732728E-03, 0.68263346E-02, 0.85278694E-02, 0.13020536E-03, 
    },
  },
  {
    {
      0.26661365E+01, -0.11117877E-02, 0.34222321E-02, -0.20794710E-02, 0.24010155E-02, -0.40831408E-02, -0.47719344E-02, -0.29675064E-02, -0.20296483E-02, 
    },
    {
      0.41117510E+00, 0.10291318E-01, 0.47319636E-01, 0.31215122E-01, 0.78935334E-02, -0.25743436E-01, -0.38405076E-01, 0.98825051E-02, 0.67149191E-02, 
    },
    {
      0.74506956E+00, 0.48187740E-01, -0.71195945E-01, 0.25079396E-01, -0.21259580E-01, 0.67766912E-01, 0.48439432E-01, -0.66096769E-02, 0.20697504E-01, 
    },
    {
      -0.39045444E+00, -0.26972118E+00, -0.70841588E-01, -0.73414013E-01, -0.12673208E-01, 0.81827506E-01, 0.14244869E+00, -0.26457297E-01, -0.33021890E-01, 
    },
    {
      -0.66143775E+00, -0.15479179E+00, 0.11289432E+00, -0.54646458E-01, 0.28970562E-01, -0.14369094E+00, -0.10134431E+00, 0.25818255E-01, -0.48019908E-01, 
    },
    {
      0.53726800E-01, 0.25038397E+00, 0.37336465E-01, 0.40437397E-01, 0.14270316E-01, -0.59865799E-01, -0.11061210E+00, 0.18749857E-01, 0.24053220E-01, 
    },
    {
      -0.40976707E-01, 0.14224784E+00, -0.30323656E-01, 0.30423727E-01, -0.44295667E-02, 0.87912515E-01, 0.58721401E-01, -0.13916852E-01, 0.28919090E-01, 
    },
    {
      0.74064271E-02, -0.27877745E+00, -0.28779912E+00, 0.39281864E-02, -0.14419556E-02, -0.11547093E-02, 0.31535963E-02, -0.72778366E-03, 0.34911863E-02, 
    },
    {
      0.19075764E-01, 0.28820461E+00, -0.28949183E+00, 0.22228628E-01, 0.63713072E-02, -0.21684170E-02, -0.55248677E-02, -0.13032343E-02, 0.91081448E-02, 
    },
    {
      -0.81327841E-01, 0.27268976E+00, 0.13500665E+00, 0.54546636E-01, 0.37922066E-01, -0.32600060E-01, -0.10949604E-01, 0.19963372E-01, 0.26835289E-01, 
    },
    {
      0.87298211E-02, -0.22350943E-01, 0.29820484E+00, 0.20726329E-01, -0.52162141E-01, 0.49364645E-01, -0.39931941E-02, 0.40589698E-01, -0.22952456E-02, 
    },
    {
      -0.10790928E+00, 0.83676583E+00, 0.63060778E+00, -0.68363361E-01, -0.99837855E-02, -0.66430098E-02, -0.12207449E-01, -0.76347262E-01, -0.15965886E+00, 
    },
    {
      -0.35241818E+00, -0.65478712E+00, 0.11052341E+01, -0.12858538E+00, 0.48155166E-01, 0.30722397E-02, 0.79846002E-01, 0.70853211E-01, -0.98708250E-01, 
    },
    {
      0.10073023E+01, -0.17845337E+01, -0.80092168E+00, -0.46188459E-01, -0.14296983E+00, 0.64327158E-01, -0.35780571E-01, -0.21253628E+00, -0.38209790E+00, 
    },
    {
      0.28986445E+00, 0.20012251E+00, -0.21629679E+01, -0.28332931E+00, 0.53812021E+00, -0.36693877E+00, 0.39688271E-01, -0.34019363E+00, -0.44371117E-01, 
    },
    {
      0.21445227E+00, -0.62239212E+00, 0.50726289E+00, 0.12845744E+00, -0.50081648E-01, 0.10264898E-01, -0.10842126E-01, 0.25324762E+00, 0.47339380E+00, 
    },
    {
      0.98013788E+00, -0.77179390E+00, -0.15949878E+01, 0.26540795E+00, -0.23207283E+00, -0.71822710E-01, -0.27238446E+00, -0.25130799E+00, 0.26421374E+00, 
    },
    {
      -0.21762612E+01, 0.31028821E+01, 0.20128744E+01, -0.15856740E+00, 0.15671612E+00, -0.83239703E-02, 0.12464356E+00, 0.52468282E+00, 0.88304269E+00, 
    },
    {
      -0.13928696E+01, -0.54484439E+00, 0.43236532E+01, 0.72512794E+00, -0.13013941E+01, 0.80244911E+00, -0.15787378E+00, 0.71791559E+00, 0.87640740E-01, 
    },
    {
      -0.90506822E-01, 0.29343450E+00, -0.64716011E+00, -0.85126469E-02, 0.86700074E-01, -0.19545069E-01, 0.22539426E-01, -0.21825901E+00, -0.35119224E+00, 
    },
    {
      -0.77049851E+00, 0.13025295E+01, 0.10195137E+01, -0.17070553E+00, 0.19100779E+00, 0.10148965E+00, 0.19473387E+00, 0.22605897E+00, -0.21595261E+00, 
    },
    {
      0.12541903E+01, -0.14151653E+01, -0.76735270E+00, 0.20479736E+00, -0.57303514E-01, -0.61315097E-01, -0.67472473E-01, -0.35682014E+00, -0.55211598E+00, 
    },
    {
      0.12503939E+01, -0.19497246E+00, -0.25232329E+01, -0.48834419E+00, 0.95700103E+00, -0.50639701E+00, 0.15538096E+00, -0.43736929E+00, -0.24403557E-02, 
    },
    {
      0.58859168E-02, -0.32317564E-01, 0.18789638E-02, 0.81562921E-01, -0.10188310E+00, -0.16970166E-02, 0.10929941E-01, -0.80213174E-02, -0.64339088E-02, 
    },
    {
      0.12977694E-01, -0.20233879E-01, -0.87966016E-02, 0.10940550E+00, 0.84022991E-01, -0.43360288E-02, 0.49938275E-02, -0.24055217E-02, 0.34415221E-02, 
    },
    {
      0.13273510E-01, 0.11690062E+00, 0.11301346E+00, -0.53855572E-01, 0.71545341E-02, 0.26712853E-02, -0.65927980E-02, -0.43166103E-02, -0.16314678E-01, 
    },
    {
      -0.76610744E-01, 0.31374011E-01, -0.51425263E-01, 0.29726738E-01, -0.10468841E+00, 0.35839197E-02, 0.31228568E-01, -0.15449940E-02, 0.10545071E-01, 
    },
    {
      0.31877279E-01, 0.21496819E+00, 0.69084771E-01, -0.39078769E+00, 0.10721206E+00, -0.50794119E-02, -0.28492521E-01, 0.88592894E-01, 0.17780269E-01, 
    },
    {
      -0.14039850E+00, -0.13015021E-01, -0.97526431E-01, -0.13987683E+00, -0.35487783E+00, -0.44129573E-01, 0.44898996E-02, 0.47013644E-01, -0.73569492E-01, 
    },
    {
      -0.59302974E-01, -0.65177011E+00, -0.36309096E+00, -0.23424511E+00, -0.13430981E+00, -0.32746039E-01, 0.60699228E-01, 0.33282310E-01, 0.85893124E-01, 
    },
    {
      0.49057353E+00, 0.21044838E+00, 0.23838104E+00, 0.10888345E-01, -0.39147530E-01, 0.10908667E+00, -0.18488690E+00, 0.53481314E-01, -0.13602662E+00, 
    },
    {
      -0.10266943E+00, -0.34125081E+00, -0.22209170E+00, 0.21323264E+00, -0.73764898E-01, -0.42999983E-01, 0.28910641E-01, -0.12368568E+00, -0.24522578E-01, 
    },
    {
      0.17222378E+00, -0.45565147E-01, 0.26528457E+00, 0.10084933E+00, 0.18232718E+00, 0.10636824E+00, -0.40960647E-01, -0.71175881E-01, 0.91745026E-01, 
    },
    {
      0.48435759E-01, 0.88375968E+00, 0.36494932E+00, 0.37522808E+00, 0.38870537E+00, 0.11076623E+00, -0.87760679E-01, -0.44668164E-01, -0.12374668E+00, 
    },
    {
      -0.56823593E+00, -0.31550634E+00, -0.23784252E+00, -0.28017625E+00, 0.20619076E+00, -0.18672663E+00, 0.23123896E+00, -0.66964269E-01, 0.16415413E+00, 
    },
    {
      -0.73595000E-02, -0.61682197E-02, 0.89407256E-02, 0.50872318E-02, 0.11877637E-01, 0.35297502E-01, -0.39307773E-01, 0.14423406E-02, 0.44433944E-03, 
    },
    {
      -0.74655698E-04, -0.16662820E-02, -0.57203509E-02, -0.14137458E-01, -0.46855919E-02, 0.45804936E-01, 0.35633929E-01, 0.30760856E-02, -0.23319600E-02, 
    },
    {
      0.12941354E-01, 0.44963390E-01, 0.54211006E-02, -0.12041884E-01, -0.95001562E-02, -0.11241351E+00, -0.19248098E-01, 0.27269039E-02, -0.30035619E-02, 
    },
    {
      0.59964410E-02, -0.13195523E-02, -0.46101669E-02, 0.39474596E-02, -0.64694295E-02, 0.15680529E-01, -0.11271558E+00, -0.32720950E-02, -0.51417295E-02, 
    },
    {
      0.48678972E-01, -0.73544301E-01, 0.30441817E-01, 0.96263876E-02, -0.39520711E-01, -0.17735021E+00, -0.45016337E-01, 0.51114690E-02, -0.11839547E-01, 
    },
    {
      -0.81728892E-02, -0.23926420E-01, 0.60986411E-02, 0.40335935E-01, -0.35920464E-02, 0.41258320E-01, -0.14260805E+00, -0.34774819E-02, 0.18476740E-01, 
    },
    {
      -0.10441234E-01, -0.13517550E-01, 0.34676618E-02, -0.14249106E-02, 0.14872112E-02, -0.32633506E-02, -0.21993518E-02, 0.17342664E-01, -0.17490787E-02, 
    },
    {
      -0.15205597E-01, 0.49438239E-02, -0.87333824E-02, -0.87372363E-02, 0.14296352E-02, 0.37051606E-03, 0.55464786E-02, 0.15304491E-03, 0.21174494E-01, 
    },
    {
      0.39107330E-01, 0.80265282E-02, -0.22919752E-01, -0.76330574E-02, -0.77880705E-02, -0.70988761E-04, 0.10868077E-01, 0.58998396E-02, -0.64436145E-01, 
    },
    {
      0.41344304E-01, -0.33050463E-01, 0.35609432E-01, 0.15621077E-01, 0.52525150E-02, -0.76148482E-02, 0.10637491E-01, 0.63588083E-01, 0.85605606E-02, 
    },
    {
      0.37736224E-02, -0.12569182E-02, -0.73067157E-03, -0.55869166E-02, -0.57651363E-02, -0.35552061E-03, -0.52995873E-02, -0.63682459E-02, 0.34755855E-02, 
    },
    {
      0.33890572E-02, 0.20352295E-01, -0.10258360E-01, -0.16343703E-02, -0.76323333E-02, 0.61682411E-02, 0.74977572E-02, -0.10669790E-02, -0.31994097E-02, 
    },
    {
      -0.56459666E-02, 0.10195384E-01, 0.95578721E-02, 0.39921715E-02, 0.18846111E-02, -0.28111471E-03, -0.14161777E-03, -0.63958345E-02, -0.16010746E-03, 
    },
    {
      0.83076116E-02, -0.32676375E-02, -0.81046103E-02, -0.20230603E-02, -0.86054020E-02, 0.55735982E-02, 0.49315910E-02, 0.46230308E-02, 0.80582756E-03, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir12_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.58852773E+01, -0.96291333E-01, -0.24346260E-02, 0.59140641E-01, -0.63641076E-02, -0.26449412E-02, -0.31942986E-01, -0.10096008E-02, 0.14919642E-01, -0.31421110E-02, -0.46499600E-02, -0.67822053E-02, 0.15032781E-01, 
    },
    {
      0.16458440E+01, -0.18067592E+00, 0.30474961E+00, -0.21605857E+00, -0.49374822E+00, -0.62353790E-01, -0.34158051E-01, -0.12912387E+00, 0.10613752E+00, 0.29181531E-01, -0.23050727E-01, 0.40541522E-01, -0.11224715E+00, 
    },
    {
      0.10501716E+02, 0.15741066E+01, -0.14308903E+01, -0.48761311E+00, 0.94652450E+00, -0.30434468E+00, 0.10839933E+01, -0.10534965E+00, -0.51447499E+00, -0.91966093E-01, 0.29239580E+00, 0.20149432E-01, -0.39589244E+00, 
    },
    {
      -0.23581297E+02, 0.36677198E+01, -0.80632642E-01, 0.43846774E+01, 0.81138687E+01, 0.20209019E+01, 0.32430317E+01, 0.27812088E+01, -0.52667046E+00, 0.26251921E+00, 0.75711900E+00, -0.62242740E+00, 0.95958692E+00, 
    },
    {
      -0.55303654E+02, -0.63127465E+01, 0.83682976E+01, -0.49186325E+01, -0.69345760E+01, 0.62089677E+01, -0.60723081E+01, 0.13838018E+01, 0.38213999E+01, -0.14319529E+00, -0.18019359E+01, 0.50669366E+00, 0.23809679E+01, 
    },
    {
      0.59686283E+02, -0.30755751E+02, -0.12747594E+02, -0.16911802E+02, -0.32807728E+02, -0.12712322E+02, -0.21192408E+02, -0.12855251E+02, -0.71364388E-01, -0.14239579E+01, -0.39535253E+01, 0.33811421E+01, -0.32126002E+01, 
    },
    {
      0.87687027E+02, 0.13251464E+02, -0.15531278E+02, 0.25012060E+02, 0.18020908E+02, -0.21276220E+02, 0.13908660E+02, -0.47290397E+01, -0.11022418E+02, 0.16285949E+01, 0.42395010E+01, -0.24118855E+01, -0.54046264E+01, 
    },
    {
      -0.23993408E+02, 0.89517502E+02, 0.42279510E+02, 0.22479494E+02, 0.54392288E+02, 0.30301125E+02, 0.50255486E+02, 0.23656767E+02, 0.32564282E+01, 0.28625791E+01, 0.82307177E+01, -0.78404694E+01, 0.53785987E+01, 
    },
    {
      -0.66045982E+02, -0.16009594E+02, 0.10073885E+02, -0.35849022E+02, -0.18812138E+02, 0.25561331E+02, -0.14358003E+02, 0.62677326E+01, 0.13022478E+02, -0.24407010E+01, -0.45216794E+01, 0.35019362E+01, 0.52684317E+01, 
    },
    {
      -0.57186363E+02, -0.10306866E+03, -0.47426033E+02, -0.90657454E+01, -0.39237549E+02, -0.30380714E+02, -0.50804012E+02, -0.19044092E+02, -0.49842196E+01, -0.29108648E+01, -0.74697671E+01, 0.79306436E+01, -0.42882524E+01, 
    },
    {
      0.20778374E+02, 0.76332736E+01, -0.15185139E+01, 0.16232574E+02, 0.66888428E+01, -0.10187649E+02, 0.54846730E+01, -0.28345735E+01, -0.53408065E+01, 0.10432419E+01, 0.18286176E+01, -0.16316253E+01, -0.18638300E+01, 
    },
    {
      0.42999359E+02, 0.40907448E+02, 0.17649250E+02, -0.67038941E+00, 0.99691086E+01, 0.10768867E+02, 0.18582966E+02, 0.55952592E+01, 0.21914599E+01, 0.12064843E+01, 0.24707305E+01, -0.28870997E+01, 0.12680758E+01, 
    },
    {
      0.10030658E+00, 0.17061027E+01, 0.14686522E+01, 0.67665540E-01, 0.21641403E-01, 0.18892784E-01, 0.10023636E-01, -0.12924850E-01, -0.28599406E-01, -0.64662793E-02, 0.11706679E-01, -0.93984343E-02, 0.17487410E-01, 
    },
    {
      -0.23059970E+00, -0.14013827E+01, 0.18092790E+01, -0.91769502E-01, 0.27110616E-01, 0.17508439E-02, -0.26233229E-02, 0.22305546E-01, 0.89679928E-02, 0.32947335E-01, 0.14318242E-01, 0.24811620E-01, -0.15941351E-02, 
    },
    {
      -0.90277523E+00, 0.23125539E+01, 0.16092964E+01, -0.14453238E+00, -0.20665842E+00, 0.51915783E+00, 0.49450490E+00, -0.19916792E+00, 0.18426664E+00, 0.25234304E-01, 0.38722817E-01, -0.42649661E-02, -0.26182031E-01, 
    },
    {
      -0.56744313E+00, 0.44611600E+00, 0.11730177E+01, -0.47925121E+00, 0.65121371E+00, -0.45957804E-01, 0.45501205E+00, 0.25943026E+00, 0.15283296E+00, 0.99556014E-01, -0.21310527E-01, 0.37541655E+00, -0.42286009E-01, 
    },
    {
      0.20739954E+01, 0.21575235E+02, 0.34665103E+01, -0.20396402E+01, 0.26834292E+01, 0.28675493E-01, 0.42604324E+00, 0.28829846E+00, 0.78976136E+00, 0.17124640E-01, -0.31698342E-01, 0.52003258E+00, -0.44703317E+00, 
    },
    {
      -0.10670326E+00, -0.74782805E+01, 0.15177947E+02, 0.26179512E+01, 0.20770681E+01, -0.10762349E+01, -0.44632065E+00, 0.60240370E+00, -0.58553094E+00, -0.10602016E+01, -0.88225491E-01, -0.11664101E+01, -0.34999716E+00, 
    },
    {
      0.25958813E+02, -0.27454485E+02, -0.54103351E+01, 0.15304747E+01, 0.93126850E+01, -0.87061596E+01, -0.62785430E+01, 0.45617943E+01, -0.37337399E+01, -0.49913475E+00, 0.41627291E+00, -0.20554547E+00, 0.59320398E-01, 
    },
    {
      0.14931066E+02, -0.24322105E+02, -0.18978451E+02, 0.66896195E+01, -0.11191239E+02, 0.25101948E+01, -0.91349773E+01, -0.48819108E+01, -0.36994448E+01, 0.46167007E+00, 0.15925080E+01, -0.77618051E+01, -0.19510709E+01, 
    },
    {
      -0.18649813E+02, -0.15214691E+03, 0.53326965E+02, 0.57327814E+01, -0.26576279E+02, -0.19274548E+00, 0.15754475E+00, 0.28831229E+01, -0.84430466E+01, -0.36592329E+00, 0.27528572E+01, -0.42047176E+01, 0.30515225E+01, 
    },
    {
      0.14867091E+02, -0.28944750E+02, -0.87118622E+02, -0.89018812E+01, -0.27807983E+02, 0.10147781E+02, 0.44683981E+01, -0.29643438E+01, 0.63139424E+01, 0.88454924E+01, -0.19191800E+01, 0.97981710E+01, 0.39162433E+01, 
    },
    {
      -0.15608711E+03, 0.10050726E+03, 0.31733864E+02, 0.67168102E+01, -0.41432091E+02, 0.47361160E+02, 0.28008722E+02, -0.28612593E+02, 0.24360193E+02, 0.73379183E+01, -0.59432845E+01, 0.31006491E+01, 0.61421555E+00, 
    },
    {
      -0.10763945E+03, 0.12835529E+03, 0.69382744E+02, -0.43361359E+02, 0.56389217E+02, -0.13272401E+02, 0.51118816E+02, 0.27908688E+02, 0.29593452E+02, -0.90813322E+01, -0.13741107E+02, 0.45609165E+02, 0.17076014E+02, 
    },
    {
      0.49405956E+02, 0.36311176E+03, -0.26420923E+03, 0.41176744E+01, 0.81804565E+02, -0.21720428E+01, -0.76053777E+01, -0.19046164E+02, 0.28341925E+02, 0.31033659E+01, -0.15050348E+02, 0.12041607E+02, -0.86199579E+01, 
    },
    {
      -0.58249134E+02, 0.21876076E+03, 0.15515247E+03, 0.21008591E+02, 0.99964432E+02, -0.32595772E+02, -0.15503604E+02, 0.48927865E+01, -0.20159599E+02, -0.26689791E+02, 0.95943127E+01, -0.29238544E+02, -0.13632691E+02, 
    },
    {
      0.39192972E+03, -0.18186565E+03, -0.12129234E+03, -0.57435074E+02, 0.56430367E+02, -0.11213171E+03, -0.54315083E+02, 0.70347794E+02, -0.67493736E+02, -0.30095579E+02, 0.18334648E+02, -0.94074984E+01, -0.31521268E+01, 
    },
    {
      0.27239188E+03, -0.28597064E+03, -0.12101062E+03, 0.12336749E+03, -0.12470569E+03, 0.26024103E+02, -0.12104427E+03, -0.59888294E+02, -0.85995743E+02, 0.29719172E+02, 0.37946323E+02, -0.11199417E+03, -0.46891853E+02, 
    },
    {
      -0.50657452E+02, -0.37947943E+03, 0.38515732E+03, -0.23199799E+02, -0.10095910E+03, 0.50711770E+01, 0.14311047E+02, 0.33573044E+02, -0.37079788E+02, -0.60635815E+01, 0.25830387E+02, -0.14170277E+02, 0.10315926E+02, 
    },
    {
      0.79237267E+02, -0.35955161E+03, -0.10834153E+03, -0.33607841E+02, -0.13744238E+03, 0.42688866E+02, 0.22823519E+02, -0.24746652E+01, 0.25041647E+02, 0.33539783E+02, -0.14638429E+02, 0.35408268E+02, 0.18950537E+02, 
    },
    {
      -0.44368756E+03, 0.16978250E+03, 0.19025841E+03, 0.10587995E+03, -0.16499306E+02, 0.12066158E+03, 0.45466454E+02, -0.74402390E+02, 0.81906021E+02, 0.44979858E+02, -0.21132385E+02, 0.10045572E+02, 0.49591422E+01, 
    },
    {
      -0.27212534E+03, 0.29296640E+03, 0.10778039E+03, -0.14709779E+03, 0.12872285E+03, -0.21798769E+02, 0.12947227E+03, 0.51874657E+02, 0.10150548E+03, -0.36791374E+02, -0.41881748E+02, 0.12185918E+03, 0.52135880E+02, 
    },
    {
      0.17340685E+02, 0.14838107E+03, -0.17882977E+03, 0.15429540E+02, 0.43413673E+02, -0.26469874E+01, -0.74090319E+01, -0.18376600E+02, 0.16659889E+02, 0.33483205E+01, -0.13998299E+02, 0.58607697E+01, -0.43770027E+01, 
    },
    {
      -0.36554958E+02, 0.17909183E+03, 0.23982536E+02, 0.19982359E+02, 0.63960342E+02, -0.19503317E+02, -0.11998042E+02, -0.21119624E+00, -0.10841630E+02, -0.14879585E+02, 0.71827703E+01, -0.14873817E+02, -0.91093102E+01, 
    },
    {
      0.18463353E+03, -0.63689983E+02, -0.10000366E+03, -0.59487541E+02, -0.83337955E+01, -0.48281857E+02, -0.12824591E+02, 0.28396828E+02, -0.35691708E+02, -0.22169371E+02, 0.81618071E+01, -0.34571950E+01, -0.25228517E+01, 
    },
    {
      0.91842926E+02, -0.11355025E+03, -0.39262459E+02, 0.60758072E+02, -0.51565228E+02, 0.64481277E+01, -0.51903679E+02, -0.14617392E+02, -0.41567421E+02, 0.15849867E+02, 0.16073263E+02, -0.48541443E+02, -0.20367502E+02, 
    },
    {
      0.33230793E-01, -0.53760577E-02, -0.43271579E-01, -0.82253754E+00, 0.10927148E+00, -0.62830341E-02, 0.32827888E-01, 0.20675356E-02, 0.11074953E-01, 0.14030251E-01, -0.93817338E-02, 0.58153598E-02, -0.90415105E-02, 
    },
    {
      -0.72160540E-02, 0.44827171E-01, 0.94338991E-01, -0.15674749E+00, -0.88599294E+00, 0.21040715E-01, 0.11656060E-01, -0.43813023E-02, 0.28248457E-02, 0.18651808E-01, 0.15717305E-01, 0.10290854E-01, -0.74585471E-02, 
    },
    {
      0.32696807E+00, 0.55384904E+00, -0.20601006E+00, 0.22753012E+00, -0.38133556E+00, 0.28293085E+00, -0.22165217E+00, -0.19600345E+00, -0.30326167E+00, -0.53036101E-01, -0.64323887E-01, -0.25258502E-01, 0.27758265E-01, 
    },
    {
      0.84868371E+00, 0.71870285E+00, 0.10409747E+00, -0.34669647E+00, 0.68305522E-01, 0.17950530E+00, 0.17103158E+00, -0.16034061E+00, 0.11909673E+00, -0.90939038E-01, 0.77516790E-02, 0.50704526E-02, -0.44594165E-01, 
    },
    {
      0.19773264E+01, -0.63938504E+00, 0.54355240E+00, 0.34600697E+01, 0.14479409E+01, -0.65015870E+00, -0.12161551E+01, 0.79876208E+00, -0.60558927E+00, 0.30918708E+00, 0.64186800E+00, 0.23445684E+00, 0.25567999E+00, 
    },
    {
      -0.23858164E+00, -0.21537189E+01, -0.21538348E+00, 0.19827148E+00, 0.62378974E+01, -0.14610817E+01, -0.79967242E+00, 0.97321606E+00, 0.28241432E+00, -0.71557999E+00, -0.40822551E-01, -0.35631898E+00, -0.30096257E+00, 
    },
    {
      0.27642247E+00, -0.35841665E+01, 0.20953624E+01, 0.21262348E+01, 0.82936811E+01, -0.25825257E+01, 0.12673424E+01, 0.22965710E+01, 0.32046340E+01, 0.59451181E+00, 0.29327148E+00, 0.60926776E-01, -0.18911134E+00, 
    },
    {
      -0.77316437E+01, -0.86913109E+01, 0.70329523E+00, 0.29154107E+00, 0.37658503E+01, -0.15267447E+01, -0.19037495E+01, 0.18705777E+01, -0.38128939E+00, 0.16374694E+01, 0.94264515E-01, -0.12111475E+01, 0.54479063E+00, 
    },
    {
      -0.11841908E+02, 0.12982108E+02, -0.46878176E+01, 0.11409554E+02, 0.11456189E+02, 0.28962121E+01, 0.70374961E+01, -0.53894963E+01, 0.32251689E+01, -0.26807201E+01, -0.33563082E+01, -0.14099928E+01, -0.16444161E+01, 
    },
    {
      0.45781121E+01, 0.22667099E+02, 0.25675006E+01, -0.19983486E+02, -0.88170242E+00, 0.60731277E+01, 0.46235948E+01, -0.60956397E+01, -0.36376829E+01, 0.50197077E+01, 0.35221526E+00, 0.22886684E+01, 0.17762291E+01, 
    },
    {
      -0.40170121E+01, 0.62177038E+01, -0.81950235E+01, -0.85565033E+01, -0.26114573E+02, 0.52662058E+01, -0.33671968E+01, -0.57750220E+01, -0.89394646E+01, -0.96039099E+00, -0.44176674E+00, -0.38639593E+00, 0.38240880E+00, 
    },
    {
      0.16040434E+02, 0.22648796E+02, -0.38160076E+01, 0.32987599E+01, -0.11050160E+02, 0.41704860E+01, 0.42172189E+01, -0.41252751E+01, 0.30131927E+00, -0.52030268E+01, -0.83111125E+00, 0.49567075E+01, -0.11959333E+01, 
    },
    {
      0.23355219E+02, -0.34282352E+02, 0.11538242E+02, -0.49848263E+02, -0.40598743E+02, -0.39050004E+01, -0.14526564E+02, 0.10634094E+02, -0.49407711E+01, 0.58253703E+01, 0.57160439E+01, 0.24022803E+01, 0.39250429E+01, 
    },
    {
      -0.89171867E+01, -0.57658352E+02, -0.79275866E+01, 0.56511944E+02, -0.33630772E+02, -0.73403568E+01, -0.97860651E+01, 0.12670938E+02, 0.95948782E+01, -0.10938619E+02, -0.13207883E+01, -0.39759080E+01, -0.31204889E+01, 
    },
    {
      0.31317453E+01, -0.26248660E+01, 0.75624285E+01, 0.10054381E+02, 0.23560154E+02, -0.29456441E+01, 0.29910414E+01, 0.40049529E+01, 0.69776058E+01, 0.23051453E+00, 0.35383257E+00, 0.45363480E+00, -0.17931676E+00, 
    },
    {
      -0.85573292E+01, -0.16594076E+02, 0.34918075E+01, -0.66410427E+01, 0.96621037E+01, -0.31863921E+01, -0.24729841E+01, 0.26573007E+01, 0.11516500E+00, 0.40185380E+01, 0.10184885E+01, -0.46337085E+01, 0.70501119E+00, 
    },
    {
      -0.15520618E+02, 0.24251535E+02, -0.78417487E+01, 0.39622379E+02, 0.31915640E+02, 0.84966403E+00, 0.10047779E+02, -0.63451653E+01, 0.19262159E+01, -0.35821593E+01, -0.31948893E+01, -0.13733149E+01, -0.27612700E+01, 
    },
    {
      0.26133041E+01, 0.42539505E+02, 0.70027027E+01, -0.41647896E+02, 0.34347599E+02, 0.24487431E+01, 0.63871117E+01, -0.86457167E+01, -0.66765871E+01, 0.73050599E+01, 0.12362036E+01, 0.21852674E+01, 0.16885310E+01, 
    },
    {
      -0.16071810E+00, 0.14123527E-01, -0.13230798E+00, 0.34138896E-01, 0.59729513E-01, -0.50168329E+00, -0.57280529E+00, 0.43802693E-01, -0.13694671E-01, 0.17453475E-01, 0.33055807E-02, 0.23122183E-02, 0.62880949E-02, 
    },
    {
      0.53357907E-01, -0.19458123E-01, 0.69014654E-01, 0.63686967E-02, 0.54668400E-01, 0.58147484E+00, -0.50669259E+00, -0.12543346E-01, 0.45777824E-01, -0.29289942E-01, -0.17757392E-01, -0.45786155E-03, -0.10528302E-01, 
    },
    {
      -0.54030935E-02, 0.18703187E+00, -0.11604700E+00, -0.90121031E-01, -0.24077909E+00, -0.58423525E+00, -0.78927018E-01, -0.25518278E-01, 0.17570584E+00, 0.64967521E-01, -0.28771313E-01, 0.21496497E-01, 0.58259081E-01, 
    },
    {
      -0.21440378E+00, 0.46478152E-01, 0.59997041E-01, -0.10545159E+00, 0.87879352E-01, 0.31408742E+00, -0.76883548E+00, -0.22707506E+00, 0.97065166E-01, -0.84792078E-01, -0.42242599E-02, -0.36972161E-01, 0.13577213E-01, 
    },
    {
      0.13984452E+01, 0.21562593E+00, 0.81785202E+00, 0.42129427E-01, -0.45740336E+00, 0.19257108E+01, 0.10962818E+01, -0.13253964E+00, 0.64441629E-01, -0.76614603E-01, 0.80381967E-01, 0.60075335E-01, -0.42204820E-02, 
    },
    {
      -0.88858914E+00, 0.68705350E+00, -0.11165400E+00, -0.39081043E-02, -0.11788681E+01, -0.96058542E+00, 0.22603960E+01, 0.79304010E-01, -0.20290291E+00, 0.10524015E+00, 0.22646762E+00, 0.12000297E-01, 0.20639112E-02, 
    },
    {
      -0.38604550E-01, -0.64558333E+00, 0.15047131E+00, 0.23604788E+00, 0.51776189E+00, 0.65631807E+00, -0.64743772E-01, 0.23293781E+00, -0.31635416E+00, -0.24677651E+00, 0.73522218E-01, -0.12383265E+00, -0.23125414E-01, 
    },
    {
      0.63801259E+00, -0.17993659E+00, 0.41443504E-01, -0.30626392E+00, 0.13222785E+00, -0.38505167E+00, 0.11026433E+01, 0.38552052E+00, -0.14338492E+00, 0.16477841E+00, 0.44638623E-01, -0.27766574E-01, -0.14409494E-01, 
    },
    {
      -0.16779745E+01, 0.35500047E+00, -0.75000972E+00, 0.44501577E-01, 0.16619511E+00, -0.29565687E+01, -0.11524156E+01, 0.10906345E+00, 0.63353598E-01, 0.74749373E-01, -0.98273449E-01, -0.23684590E+00, 0.99928290E-01, 
    },
    {
      0.12862395E+01, -0.10268888E+01, -0.34922286E-03, 0.42514104E+00, 0.18869687E+01, 0.92553735E+00, -0.33864348E+01, -0.96434183E-01, 0.88395439E-01, -0.27669014E-01, -0.47842038E+00, -0.87917507E-01, -0.50919611E-01, 
    },
    {
      0.67794308E-01, 0.43764468E-01, -0.85234791E-02, 0.55448759E-01, -0.75007221E-02, 0.26949584E-02, 0.40005311E-01, 0.17787793E+00, -0.21966216E+00, 0.13119747E-01, 0.17033216E-01, 0.48258617E-02, 0.93731331E-02, 
    },
    {
      0.41226272E-01, -0.45745876E-01, -0.44874894E-02, -0.56989055E-01, 0.66878618E-03, 0.49436111E-02, -0.93553104E-02, 0.20852254E+00, 0.17780304E+00, -0.10008897E-01, 0.40826789E-03, -0.44579688E-02, -0.96555213E-02, 
    },
    {
      0.32994185E-01, -0.50840318E-01, -0.82209967E-01, -0.16789146E+00, -0.35853390E-01, -0.74349888E-01, 0.36870703E-01, -0.67959286E-01, -0.23842365E+00, -0.75435199E-01, 0.58410540E-01, 0.19143853E-01, 0.72812706E-01, 
    },
    {
      -0.35581663E-01, -0.75313292E-01, -0.37497939E-02, -0.53745732E-02, 0.86594537E-01, -0.85272796E-01, -0.75303704E-01, 0.15578288E+00, -0.57227295E-01, -0.19145612E-01, 0.38526919E-01, -0.33226525E-02, -0.25027826E-01, 
    },
    {
      -0.24258215E-01, 0.56879047E-01, -0.87987788E-01, -0.18437518E-01, -0.26629930E-02, 0.94258245E-02, 0.11299050E-01, -0.10933794E-01, 0.11623712E-02, 0.17373119E+00, 0.16964652E+00, -0.10166550E-01, -0.62725325E-02, 
    },
    {
      0.30380953E-01, -0.23155006E-01, -0.33405375E-01, -0.44372205E-01, 0.21909758E-01, -0.50504357E-02, -0.34750249E-01, 0.32919927E-02, -0.52613996E-01, -0.17605481E+00, 0.17251751E+00, 0.14515134E-01, 0.11617797E-01, 
    },
    {
      0.42745028E-01, -0.67311842E-02, -0.12199854E-01, -0.23588119E-01, 0.42619001E-01, 0.23207851E-01, 0.16206969E-01, -0.71131140E-02, -0.15745321E-01, 0.93277032E-02, 0.54416276E-03, -0.92576854E-01, 0.75096644E-01, 
    },
    {
      -0.57170991E-01, -0.11325012E-01, -0.45626007E-01, -0.51696181E-01, -0.54596667E-02, 0.18047408E-02, -0.57304051E-01, 0.28688202E-01, -0.29154113E-02, -0.48920629E-02, -0.11745230E-02, -0.72187759E-01, -0.10533948E+00, 
    },
    {
      -0.46112318E-01, -0.93352431E-02, -0.77548169E-01, 0.10806181E-01, 0.16295983E-01, -0.91340821E-02, -0.37854966E-01, 0.59323609E-02, -0.70786551E-02, 0.43499945E-02, -0.12444421E-01, -0.23848300E-02, 0.11797853E-01, 
    },
    {
      -0.69076233E-02, -0.14922633E-01, -0.16607357E-01, -0.28959988E-01, 0.37629224E-01, 0.27335903E-02, -0.21362375E-01, 0.14246407E-01, 0.54619005E-02, 0.60001425E-02, 0.29482830E-01, -0.15042797E-01, -0.14008236E-01, 
    },
    {
      -0.39528869E-02, 0.18516526E-02, -0.39649025E-01, 0.28208220E-01, -0.11501377E-01, -0.40830545E-01, 0.14463947E-01, 0.18475296E-02, -0.75901337E-02, 0.18725945E-01, 0.10570663E-01, -0.12890499E-04, 0.23008062E-01, 
    },
    {
      -0.51339738E-01, 0.28110512E-01, -0.61789113E-02, 0.81891287E-02, -0.29036777E-01, 0.28971057E-01, 0.64823423E-02, 0.76583945E-02, 0.27483683E-02, -0.17355928E-01, 0.15277426E-01, 0.51793233E-02, -0.15141426E-01, 
    },
  },
  {
    {
      0.88938065E+01, -0.15078264E-01, -0.25140066E-01, 0.22020804E-01, 0.25030833E-02, -0.20642305E-01, -0.30851010E-01, 0.35875994E-02, -0.87216607E-03, 0.24480885E-01, 0.15102470E-01, -0.19631794E-01, 0.20391542E-01, 
    },
    {
      0.61357880E+00, 0.13776004E+00, -0.57032752E+00, -0.28906608E-01, -0.35331145E+00, -0.21978439E+00, -0.14012583E+00, 0.54715615E-01, 0.14554127E+00, -0.24320902E-02, -0.52757952E-01, -0.18868657E-01, -0.78913517E-01, 
    },
    {
      0.26584980E+02, -0.91476983E+00, -0.10767018E+01, 0.83116639E+00, 0.18041835E+01, -0.63989371E+00, 0.51932621E+00, 0.50343698E+00, -0.63642442E+00, -0.97818619E+00, -0.29772168E+00, 0.35066122E+00, -0.68079138E+00, 
    },
    {
      0.31529946E+01, -0.22581005E+01, 0.95738211E+01, 0.81222594E+00, 0.97320185E+01, 0.23350554E+01, 0.16431698E+01, 0.61718774E+00, -0.20582492E+01, -0.12018207E-01, 0.19452527E+01, -0.24020819E+00, 0.93675071E+00, 
    },
    {
      -0.12845282E+03, 0.76379828E+01, 0.81046038E+01, -0.12054643E+02, -0.15387237E+02, 0.68140392E+01, -0.15058470E+01, -0.48598957E+01, 0.49816790E+01, 0.52011762E+01, 0.22433043E+01, -0.60794902E+00, 0.40673318E+01, 
    },
    {
      -0.78451271E+02, -0.14396160E+02, -0.34657215E+02, -0.25212047E+01, -0.49333851E+02, -0.11871571E+02, -0.53334084E+01, -0.59590955E+01, 0.88634005E+01, 0.10821542E+01, -0.11399439E+02, 0.47652416E+01, -0.32855816E+01, 
    },
    {
      0.21243150E+03, -0.14004525E+02, -0.17843674E+02, 0.43910168E+02, 0.43445202E+02, -0.19488865E+02, 0.13466698E+01, 0.12091710E+02, -0.13215184E+02, -0.10340988E+02, -0.57268853E+01, -0.15506169E+01, -0.88107119E+01, 
    },
    {
      0.24853136E+03, 0.78768333E+02, 0.34186371E+02, 0.19544324E+01, 0.10425354E+03, 0.27955261E+02, 0.85331831E+01, 0.15262192E+02, -0.16240059E+02, -0.35532913E+01, 0.25211710E+02, -0.16891586E+02, 0.51664090E+01, 
    },
    {
      -0.16578394E+03, 0.30796990E+01, 0.14645020E+02, -0.58538593E+02, -0.48412685E+02, 0.21371475E+02, -0.61169732E+00, -0.10880286E+02, 0.14118615E+02, 0.89713211E+01, 0.56740236E+01, 0.40566039E+01, 0.80941858E+01, 
    },
    {
      -0.28710046E+03, -0.11082728E+03, 0.47830529E+01, 0.12881908E+01, -0.99203979E+02, -0.28727617E+02, -0.78436465E+01, -0.15181051E+02, 0.13331136E+02, 0.39052055E+01, -0.23992605E+02, 0.21218477E+02, -0.36269221E+01, 
    },
    {
      0.51397850E+02, 0.44691315E+01, -0.40350652E+01, 0.25961407E+02, 0.18408005E+02, -0.80434484E+01, 0.32910940E+00, 0.31243167E+01, -0.52550664E+01, -0.28733051E+01, -0.18864105E+01, -0.22640367E+01, -0.26912494E+01, 
    },
    {
      0.11311129E+03, 0.48871571E+02, -0.13456262E+02, -0.14472980E+01, 0.34871849E+02, 0.10456634E+02, 0.32485316E+01, 0.52217917E+01, -0.40892677E+01, -0.13860580E+01, 0.83094473E+01, -0.88487806E+01, 0.87602508E+00, 
    },
    {
      0.69597103E-01, 0.16769730E+01, 0.15545819E+01, -0.55449349E-02, 0.69705747E-01, 0.23653561E-01, 0.12845129E-01, 0.86718919E-02, -0.45932859E-01, 0.31570107E-01, 0.45554500E-01, -0.21642283E-01, 0.12889816E-01, 
    },
    {
      -0.21921287E+00, -0.16087893E+01, 0.17624192E+01, -0.13460632E+00, -0.59290775E-02, -0.83502904E-02, -0.34621302E-01, 0.11610980E-01, -0.14968859E-01, -0.25111865E-01, 0.17095666E-01, -0.55068266E-02, -0.23103973E-01, 
    },
    {
      -0.48869601E+00, 0.17512943E+01, 0.14671189E+01, -0.29812467E+00, 0.43485570E+00, 0.74070942E+00, 0.47886825E+00, -0.15530139E-01, 0.28587043E+00, -0.90405986E-01, 0.12089233E+00, -0.10466526E+00, -0.22864467E+00, 
    },
    {
      0.26475123E+00, -0.10255442E+01, 0.12023131E+01, -0.79612106E+00, -0.14503480E+00, 0.48347488E-01, 0.24352114E+00, 0.33930928E+00, 0.17734116E+00, -0.29564553E+00, -0.23605779E-01, 0.26259270E+00, -0.14518072E+00, 
    },
    {
      0.32442267E+01, 0.19304132E+02, -0.17855993E+02, 0.15392791E+01, 0.12891808E+01, -0.13034651E+01, 0.78133240E-01, -0.28128734E+00, 0.13791049E+01, -0.29651865E+00, -0.13774163E+01, 0.90447134E+00, -0.48520055E+00, 
    },
    {
      -0.38545475E+01, 0.17391617E+02, 0.15856245E+02, 0.32506826E+01, 0.17471519E+01, 0.54408711E+00, -0.87338269E-01, 0.11546892E+01, -0.39241776E+00, 0.82400136E-01, -0.41304421E+00, -0.31734011E+00, -0.47280574E+00, 
    },
    {
      0.15891956E+02, -0.71328111E+01, -0.82937202E+01, -0.10363106E+01, -0.37894361E+01, -0.13196981E+02, -0.97192078E+01, 0.57880574E+00, -0.50687685E+01, 0.15093237E+00, -0.15454117E+01, 0.22428346E+01, 0.37367301E+01, 
    },
    {
      -0.37251842E+01, 0.30780756E+00, -0.38300519E+01, 0.13541659E+02, -0.41292334E+01, 0.24990664E+01, -0.36101906E+01, -0.68872380E+01, -0.33844929E+01, 0.48587584E+01, 0.16653656E+01, -0.54408031E+01, -0.14610332E+01, 
    },
    {
      -0.35525444E+02, -0.55185818E+02, 0.18234506E+03, -0.17191862E+02, -0.18231380E+02, 0.10715060E+02, 0.41820860E+00, 0.49597664E+01, -0.10105695E+02, 0.44890794E+00, 0.11448022E+02, -0.59232845E+01, 0.21196377E+01, 
    },
    {
      0.48971519E+02, -0.17984668E+03, -0.10086093E+02, -0.28173100E+02, -0.18571411E+02, -0.34781046E+01, 0.46446908E+00, -0.91365309E+01, 0.32334213E+01, 0.29283717E+00, 0.53791575E+01, 0.33645318E+01, 0.52979574E+01, 
    },
    {
      -0.98831841E+02, 0.35010017E+02, 0.81809647E+02, 0.42378185E+02, 0.32127880E+02, 0.71276741E+02, 0.47292530E+02, -0.79754658E+01, 0.26064089E+02, 0.57404060E+01, 0.62453079E+01, -0.13935455E+02, -0.23157148E+02, 
    },
    {
      0.37681507E+02, -0.22015532E+02, 0.41568684E+02, -0.89377480E+02, 0.34201309E+02, -0.14441486E+02, 0.12126778E+02, 0.37069691E+02, 0.21612261E+02, -0.29737093E+02, -0.11194304E+02, 0.31738426E+02, 0.18420141E+02, 
    },
    {
      0.11772128E+03, -0.16025244E+02, -0.54052008E+03, 0.52717884E+02, 0.66749588E+02, -0.31852327E+02, -0.27905669E+01, -0.18864178E+02, 0.26960806E+02, 0.23407796E+00, -0.36213844E+02, 0.12973186E+02, -0.41116295E+01, 
    },
    {
      -0.15707156E+03, 0.53367804E+03, -0.18824304E+03, 0.96771606E+02, 0.59688560E+02, 0.96216154E+01, 0.33605173E+01, 0.27191839E+02, -0.63597393E+01, -0.57947147E+00, -0.20041174E+02, -0.10695272E+02, -0.15405689E+02, 
    },
    {
      0.25402480E+03, -0.12169586E+03, -0.29696808E+03, -0.17765755E+03, -0.10246051E+03, -0.16805841E+03, -0.91728348E+02, 0.29707182E+02, -0.58331951E+02, -0.25739067E+02, -0.14582021E+02, 0.34748569E+02, 0.59372269E+02, 
    },
    {
      -0.14180334E+03, 0.90961342E+02, -0.17497099E+03, 0.24299307E+03, -0.93683586E+02, 0.25937101E+02, -0.10940236E+02, -0.83610329E+02, -0.59765331E+02, 0.78051521E+02, 0.30359779E+02, -0.77565620E+02, -0.57330196E+02, 
    },
    {
      -0.15175063E+03, 0.14136122E+03, 0.67243622E+03, -0.63942410E+02, -0.95142883E+02, 0.37566513E+02, 0.25746937E+01, 0.25226887E+02, -0.30201876E+02, 0.11459297E+01, 0.47971264E+02, -0.10312352E+02, 0.43723435E+01, 
    },
    {
      0.18913745E+03, -0.64964203E+03, 0.40084787E+03, -0.13516272E+03, -0.75677673E+02, -0.12140692E+02, -0.98627472E+01, -0.34648926E+02, 0.30085969E+01, -0.86283243E+00, 0.28143951E+02, 0.13283780E+02, 0.17091984E+02, 
    },
    {
      -0.29154059E+03, 0.17865944E+03, 0.44016376E+03, 0.25832797E+03, 0.12581499E+03, 0.18186659E+03, 0.75668427E+02, -0.42528137E+02, 0.58712421E+02, 0.38270672E+02, 0.19351198E+02, -0.37923859E+02, -0.65868088E+02, 
    },
    {
      0.21195496E+03, -0.14264771E+03, 0.25605206E+03, -0.27863412E+03, 0.10939057E+03, -0.14928993E+02, -0.43717604E+01, 0.85104988E+02, 0.72570885E+02, -0.90291161E+02, -0.36729885E+02, 0.84085503E+02, 0.70076569E+02, 
    },
    {
      0.67721512E+02, -0.90769173E+02, -0.30035886E+03, 0.26721811E+02, 0.47102436E+02, -0.15028220E+02, -0.20477100E+00, -0.11208223E+02, 0.12107578E+02, -0.20266440E+01, -0.22364059E+02, 0.21444893E+01, -0.19921613E+01, 
    },
    {
      -0.77146408E+02, 0.27997412E+03, -0.22317946E+03, 0.64471100E+02, 0.32618526E+02, 0.58598595E+01, 0.59874973E+01, 0.15801620E+02, 0.81914449E+00, 0.12750365E+01, -0.13325417E+02, -0.56135464E+01, -0.63814468E+01, 
    },
    {
      0.12182439E+03, -0.87194756E+02, -0.22319070E+03, -0.12528762E+03, -0.51943058E+02, -0.73651070E+02, -0.21513721E+02, 0.20706060E+02, -0.21602533E+02, -0.18770321E+02, -0.10114022E+02, 0.15116210E+02, 0.26225403E+02, 
    },
    {
      -0.10590909E+03, 0.75341286E+02, -0.12101205E+03, 0.11209223E+03, -0.47404507E+02, 0.42671257E+00, 0.63914065E+01, -0.32075302E+02, -0.31445272E+02, 0.37990154E+02, 0.16272755E+02, -0.33418064E+02, -0.29758162E+02, 
    },
    {
      0.24743646E-01, -0.24780034E+00, 0.32477383E-01, -0.70758212E+00, 0.77269542E+00, 0.21704473E-01, 0.73262020E-02, -0.54433793E-02, -0.36484734E-02, 0.15552454E-01, 0.11511402E-01, 0.95175020E-02, -0.77193896E-02, 
    },
    {
      -0.48369352E-01, -0.53220801E-01, 0.27144585E-01, -0.81413311E+00, -0.71973467E+00, -0.70226772E-04, -0.60768458E-02, 0.16783381E-01, -0.50986288E-02, -0.15607438E-01, 0.57154228E-02, 0.16545281E-01, -0.27624261E-01, 
    },
    {
      0.42223513E+00, -0.98611221E-01, 0.50277317E+00, -0.16155951E+01, 0.44995168E+00, 0.60947478E+00, -0.26007742E+00, -0.34951422E+00, -0.61507735E-01, -0.29398164E+00, -0.57447203E-01, -0.17995073E-01, -0.72372705E-01, 
    },
    {
      0.12762964E+01, 0.26164582E+00, 0.47269124E+00, -0.10257167E+01, -0.19976238E+01, 0.18526942E+00, 0.29332149E+00, -0.20163956E+00, 0.15762205E+00, -0.21016265E+00, -0.20759076E-01, 0.75109035E-01, -0.30883980E+00, 
    },
    {
      0.23981123E+01, 0.29899118E+01, 0.38449600E+01, -0.60386233E+01, 0.53788686E+01, -0.93953538E+00, -0.15277848E+01, 0.59264237E+00, 0.45906556E+00, 0.24649295E+00, -0.10305211E+00, 0.32214963E+00, 0.46169207E+00, 
    },
    {
      0.16696631E+01, 0.25706765E+00, 0.28278365E+01, -0.46109791E+01, -0.72205920E+01, 0.34772471E+00, -0.62179738E+00, -0.14041898E-01, 0.11932110E+01, 0.11926308E+00, 0.61168361E-01, -0.12217858E+00, 0.74488103E-01, 
    },
    {
      -0.42425346E+01, 0.35861960E+00, -0.69054904E+01, 0.77307906E+01, 0.14447305E+01, -0.51289444E+01, 0.42940632E+00, 0.37216494E+01, 0.11370006E+01, 0.24206262E+01, -0.55311292E-01, 0.61113574E-01, 0.43589985E+00, 
    },
    {
      -0.11186955E+02, 0.14551842E+01, -0.46096087E+01, 0.70924859E+01, 0.82799072E+01, 0.54562348E+00, -0.19091228E+01, 0.63213462E+00, -0.97554809E+00, 0.27594109E+01, 0.64862490E-01, -0.13592871E+01, 0.23405161E+01, 
    },
    {
      -0.17754059E+02, -0.32034547E+01, -0.41930241E+02, 0.41079243E+02, -0.18051516E+02, 0.25384116E+01, 0.90485382E+01, -0.25573390E+01, -0.52479219E+01, -0.22733593E+01, 0.16716585E+01, -0.19437449E+01, -0.34734039E+01, 
    },
    {
      -0.11303488E+02, 0.17556179E+02, -0.22370165E+02, 0.15778043E+02, 0.52471008E+02, -0.31812954E+01, 0.27775536E+01, -0.23356988E+00, -0.10081247E+02, 0.12472678E+01, -0.56565022E+00, 0.53981122E-01, 0.12582926E+00, 
    },
    {
      0.12362515E+02, -0.29632866E+01, 0.15124092E+02, -0.18801899E+02, -0.13831272E+02, 0.12536344E+02, 0.10004860E+01, -0.93945045E+01, -0.36175582E+01, -0.51194916E+01, 0.14203264E+01, -0.24957810E-01, -0.76139688E+00, 
    },
    {
      0.25606970E+02, -0.82127504E+01, 0.11993701E+02, -0.10493736E+02, -0.15479230E+02, -0.24436879E+01, 0.37523029E+01, 0.70967054E+00, 0.28187840E+01, -0.75910454E+01, 0.22493189E-01, 0.46240821E+01, -0.53346510E+01, 
    },
    {
      0.38124302E+02, -0.81857109E+01, 0.10164820E+03, -0.78550056E+02, 0.12163009E+02, 0.46451297E-01, -0.17560549E+02, 0.37782583E+01, 0.13861211E+02, 0.52900162E+01, -0.42033973E+01, 0.37402480E+01, 0.76729035E+01, 
    },
    {
      0.27311661E+02, -0.54518929E+02, 0.45628468E+02, -0.14268176E+02, -0.10502533E+03, 0.78711667E+01, -0.59193988E+01, 0.21227322E+01, 0.23136654E+02, -0.59244928E+01, 0.94676512E+00, 0.14095348E+01, 0.77875800E-01, 
    },
    {
      -0.12606331E+02, 0.52987642E+01, -0.10864910E+02, 0.19476118E+02, 0.18500385E+02, -0.88927727E+01, -0.17200902E+01, 0.67320638E+01, 0.26062746E+01, 0.29826615E+01, -0.15832294E+01, -0.24634519E+00, 0.30302539E+00, 
    },
    {
      -0.16824041E+02, 0.83268948E+01, -0.88709040E+01, 0.17254645E+00, 0.14521973E+02, 0.19288079E+01, -0.16344380E+01, -0.16025696E+01, -0.28889387E+01, 0.53860931E+01, 0.74160628E-01, -0.41329098E+01, 0.36643453E+01, 
    },
    {
      -0.25644232E+02, 0.11694123E+02, -0.68230270E+02, 0.47731461E+02, 0.49299583E+01, -0.33478410E+01, 0.10701688E+02, -0.15946394E+01, -0.10438581E+02, -0.35907745E+01, 0.26091225E+01, -0.24007337E+01, -0.50249095E+01, 
    },
    {
      -0.22422714E+02, 0.41399738E+02, -0.25003633E+02, 0.22523800E+00, 0.66093285E+02, -0.54813409E+01, 0.45410838E+01, -0.25613079E+01, -0.15686406E+02, 0.53152466E+01, -0.38827559E+00, -0.17855603E+01, -0.56493247E+00, 
    },
    {
      -0.25739521E+00, -0.51444806E-01, -0.39007924E-01, 0.43471336E-01, 0.74577101E-01, -0.69018120E+00, -0.77091724E+00, 0.18789891E-01, -0.30679217E-01, 0.34278046E-01, 0.70777461E-02, 0.17007646E-02, 0.12116459E-01, 
    },
    {
      0.30931152E-01, 0.82304358E-01, 0.23780592E-01, -0.61587142E-02, 0.14854147E+00, 0.73189169E+00, -0.68215245E+00, 0.67674126E-02, 0.98737925E-02, -0.13763633E-01, -0.28919926E-01, -0.11603950E-02, 0.28128898E-01, 
    },
    {
      0.15583675E+00, -0.97631486E-02, -0.42719718E-01, 0.55609424E-01, -0.76401152E-01, -0.72838420E+00, -0.35610983E+00, 0.81875414E-01, 0.17080715E+00, -0.30333633E-01, -0.31908482E-01, 0.57391785E-01, 0.27769849E-01, 
    },
    {
      -0.52173845E-01, 0.24948262E+00, 0.12350982E+00, -0.15202740E+00, 0.21870527E+00, 0.53911513E+00, -0.62828916E+00, -0.21726982E+00, -0.81490837E-02, -0.29228795E-01, -0.10110049E-01, -0.15235767E-01, 0.31288322E-01, 
    },
    {
      0.17942052E+01, -0.47697356E+00, -0.24683045E-01, -0.73759890E+00, -0.31411227E-01, 0.23973184E+01, 0.15254406E+01, 0.46166003E+00, -0.19089140E+00, -0.40453413E+00, 0.24325491E+00, 0.47338758E-01, -0.12414197E+00, 
    },
    {
      -0.10261521E+01, 0.34687656E+00, -0.13719416E+00, -0.56109297E+00, -0.19106919E+01, -0.12889602E+01, 0.28899510E+01, 0.31973645E+00, 0.39896643E+00, 0.91862530E-01, 0.74149244E-01, 0.17311741E+00, -0.22030464E+00, 
    },
    {
      0.19139686E+00, -0.29637602E+00, -0.89744121E-01, 0.34440780E-03, 0.12633979E+00, 0.17837116E+00, 0.34260508E+00, -0.52784592E-01, -0.35431859E+00, 0.11950550E+00, 0.18184212E+00, -0.15287527E+00, -0.12142629E-02, 
    },
    {
      0.38059655E+00, -0.70801187E+00, -0.47004503E+00, 0.36706582E-01, -0.10262799E+00, -0.90509468E+00, -0.71891233E-01, 0.41418588E+00, 0.51909392E-02, 0.73546261E-01, 0.61649170E-01, -0.10198504E+00, -0.11583487E+00, 
    },
    {
      -0.30275664E+01, 0.20220103E+01, 0.34253958E+00, 0.11947123E+01, -0.63510936E+00, -0.44640880E+01, -0.18234404E+01, -0.84736711E+00, 0.57081997E+00, 0.75149286E+00, -0.48013139E+00, -0.24022443E-01, 0.20686179E+00, 
    },
    {
      0.15877696E+01, -0.71108854E+00, 0.99898326E+00, 0.13696940E+01, 0.26519346E+01, 0.18384987E+01, -0.53081517E+01, -0.50006735E+00, -0.82502490E+00, -0.16206498E+00, -0.16357306E+00, -0.40980041E+00, 0.16522449E+00, 
    },
    {
      -0.80586374E-02, 0.50992239E-01, -0.46632238E-01, 0.57858139E-01, 0.29589590E-01, -0.29866926E-01, 0.44593327E-01, 0.20724247E+00, -0.33956417E+00, 0.48757136E-01, 0.48252712E-02, 0.75157746E-02, 0.82745338E-02, 
    },
    {
      0.56242988E-01, -0.98413797E-02, -0.85247047E-02, -0.81652971E-02, 0.10179361E-01, -0.32878406E-01, 0.63941120E-02, 0.30926964E+00, 0.19301084E+00, 0.51544094E-02, -0.95572891E-02, 0.14922715E-01, -0.64172191E-02, 
    },
    {
      0.19160556E-02, -0.24454545E-02, -0.16382514E+00, -0.21728696E-01, 0.84944502E-01, -0.70919216E-01, 0.61331507E-01, -0.12808955E+00, -0.38097063E+00, -0.58370437E-01, 0.67949370E-01, 0.16139146E-01, 0.56471489E-01, 
    },
    {
      -0.82536153E-02, -0.59724025E-01, 0.79807997E-01, 0.33527713E-01, 0.72642304E-01, -0.10801914E+00, -0.34897956E-02, 0.36352578E+00, -0.57958063E-01, -0.35505172E-01, 0.12366363E-02, 0.40800598E-01, -0.18490655E-01, 
    },
    {
      -0.46349317E-02, 0.14545528E-01, -0.12120489E+00, 0.26173390E-01, 0.17550857E-02, -0.28536871E-01, 0.22547282E-01, -0.36351152E-01, -0.60174363E-02, 0.28420082E+00, 0.16398448E+00, -0.12288245E-01, 0.14442492E-01, 
    },
    {
      0.60371339E-01, 0.38244709E-01, 0.32651130E-01, -0.77243634E-01, -0.35003498E-01, 0.14984454E-01, 0.15117691E-01, -0.19099513E-01, -0.49697436E-01, -0.15599817E+00, 0.25729859E+00, 0.33024959E-02, -0.99182769E-03, 
    },
    {
      0.71319506E-01, -0.75824670E-01, 0.36760561E-01, -0.49347986E-01, 0.15868425E-01, 0.38285628E-01, 0.23324162E-01, -0.39113779E-01, -0.30983375E-01, 0.13260117E-01, 0.37964746E-01, -0.74268147E-01, 0.11632289E+00, 
    },
    {
      -0.17346490E-01, 0.15348023E-01, -0.14875906E-01, -0.52558880E-01, -0.71550235E-01, -0.61885449E-02, -0.24182137E-01, 0.37889171E-02, -0.25238233E-01, -0.21548860E-02, 0.12035027E-01, -0.14105137E+00, -0.92943028E-01, 
    },
    {
      -0.53878661E-01, -0.32849424E-01, -0.84194556E-01, 0.31174347E-01, -0.46395538E-02, 0.12672110E-01, -0.28512165E-01, -0.13265373E-01, -0.14615447E-01, 0.36233847E-02, -0.12471826E-01, 0.34783301E-02, 0.54068253E-02, 
    },
    {
      -0.35035559E-02, 0.55042438E-01, 0.23584509E-01, -0.25574235E-01, 0.69150315E-02, -0.15520843E-01, 0.11701881E-01, 0.12512361E-01, -0.23976658E-01, 0.11502410E-02, 0.91396850E-02, 0.85549625E-02, -0.30781742E-01, 
    },
    {
      0.11815184E-01, 0.77101663E-01, 0.19891476E-01, -0.14418868E-01, -0.48151329E-01, -0.46777282E-01, 0.32677677E-01, 0.10109333E-01, 0.22690778E-03, 0.29887985E-01, -0.23127506E-01, 0.21396790E-01, 0.23422489E-01, 
    },
    {
      -0.10521832E+00, 0.46248104E-01, -0.85954962E-04, 0.24128256E-01, 0.10098126E-01, 0.15944390E-01, 0.24146425E-01, 0.21453386E-01, 0.39750654E-02, -0.21658875E-01, 0.59070806E-02, 0.15631018E-01, -0.13315894E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir12_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30724745E+01, 0.76932684E-02, 0.10582104E-01, -0.58678435E-02, 0.86863544E-02, 0.23237041E-02, -0.16314250E-02, 0.22063474E-02, 0.30517471E-02, 
    },
    {
      0.31540647E+00, -0.10881314E-01, 0.51532704E-01, 0.60170814E-01, 0.84916512E-02, -0.18651348E-01, -0.24970861E-01, 0.24080003E-01, -0.41201661E-03, 
    },
    {
      0.70605505E+00, -0.14244080E+00, -0.29646529E-01, -0.12706631E-01, -0.95824571E-02, -0.33969921E-02, 0.17386436E-01, -0.78530975E-01, -0.11982952E-01, 
    },
    {
      -0.42120489E+00, -0.20498924E+00, -0.70140377E-01, -0.13760924E+00, -0.42224452E-01, 0.85511714E-01, 0.73424190E-01, -0.67239940E-01, 0.15587630E-01, 
    },
    {
      -0.89627123E+00, 0.41788438E+00, -0.50209500E-01, 0.68723954E-01, 0.28690726E-01, 0.34792382E-01, -0.43728635E-01, 0.22312871E+00, 0.29185694E-01, 
    },
    {
      0.65906473E-01, 0.19188379E+00, 0.25963048E-01, 0.69747426E-01, 0.35327666E-01, -0.73112071E-01, -0.53085875E-01, 0.43332018E-01, -0.21014430E-01, 
    },
    {
      0.26427588E+00, -0.27628177E+00, 0.10059510E+00, -0.50584864E-01, -0.29864950E-01, -0.44461355E-01, 0.33647098E-01, -0.14697210E+00, -0.26645195E-01, 
    },
    {
      0.16860802E-01, -0.18605264E+00, -0.38714597E+00, -0.19328046E-03, 0.40253001E-03, 0.75271179E-03, 0.54476932E-02, 0.79560261E-02, 0.55059040E-03, 
    },
    {
      0.20514991E-01, 0.39720568E+00, -0.19083656E+00, -0.12558009E-01, 0.56398655E-02, -0.12297944E-01, -0.14556028E-01, -0.61569130E-02, 0.10029857E-01, 
    },
    {
      0.10009073E-01, 0.30681634E+00, 0.34232047E+00, 0.98892421E-01, 0.69215246E-01, 0.82950536E-02, -0.18969623E-01, 0.55852693E-01, -0.31100798E-01, 
    },
    {
      0.77327024E-02, -0.19496195E+00, 0.34152469E+00, -0.10556676E+00, -0.18567836E-01, 0.55548723E-03, -0.31484790E-01, 0.53078610E-01, 0.60641043E-01, 
    },
    {
      -0.30647790E+00, 0.81640363E+00, 0.18649312E+01, -0.98093107E-01, 0.15951095E-01, 0.17376818E-01, -0.39560460E-01, -0.20320363E+00, -0.74637689E-01, 
    },
    {
      -0.73014015E+00, -0.19144686E+01, 0.81253171E+00, 0.19620964E+00, -0.52492674E-01, 0.88517673E-01, 0.17049298E+00, 0.12281351E+00, -0.20673332E+00, 
    },
    {
      0.48533568E+00, -0.15024738E+01, -0.17705139E+01, -0.81606116E-03, -0.53666764E+00, -0.37227347E+00, 0.75344280E-02, -0.40956607E+00, 0.40251580E+00, 
    },
    {
      0.89858896E+00, 0.10686440E+01, -0.16356423E+01, 0.29005781E+00, 0.37916356E+00, -0.13871599E-01, 0.33012232E+00, -0.34580472E+00, -0.33433393E+00, 
    },
    {
      0.10481230E+01, -0.77934951E+00, -0.19115534E+01, 0.12393999E+00, 0.55828355E-01, -0.70680223E-01, 0.11313992E+00, 0.58049357E+00, 0.31924832E+00, 
    },
    {
      0.22701538E+01, 0.16197052E+01, -0.96915621E+00, -0.33665845E+00, -0.14552668E+00, -0.21495460E+00, -0.52067465E+00, -0.32353064E+00, 0.57284433E+00, 
    },
    {
      -0.12134343E+01, 0.19404646E+01, 0.37635965E+01, -0.62870938E+00, 0.10511738E+01, 0.11378073E+01, -0.29707570E-01, 0.86152846E+00, -0.11126765E+01, 
    },
    {
      -0.29988582E+01, -0.20589967E+01, 0.24830685E+01, -0.22141758E-01, -0.89734024E+00, 0.17300197E+00, -0.71507871E+00, 0.67783284E+00, 0.56816947E+00, 
    },
    {
      -0.79848272E+00, 0.23965521E+00, 0.80937093E+00, 0.18386710E-01, -0.11181234E+00, 0.65899365E-01, -0.10855011E+00, -0.42357123E+00, -0.30929345E+00, 
    },
    {
      -0.19041708E+01, -0.49399413E-01, 0.43354791E+00, 0.93478829E-01, 0.24635497E+00, 0.13785353E+00, 0.41312331E+00, 0.21510728E+00, -0.42922243E+00, 
    },
    {
      0.70117193E+00, -0.47726721E+00, -0.19355594E+01, 0.54693979E+00, -0.56250012E+00, -0.84684205E+00, 0.39356496E-01, -0.54770625E+00, 0.81452745E+00, 
    },
    {
      0.24183495E+01, 0.72689992E+00, -0.10589318E+01, -0.23238608E+00, 0.54398221E+00, -0.20291325E+00, 0.43034214E+00, -0.41715705E+00, -0.29115310E+00, 
    },
    {
      -0.90023009E-02, -0.35391398E-01, -0.47154608E-02, 0.19855529E-01, -0.11723321E+00, -0.10261102E-01, 0.60877344E-02, -0.11590884E-01, -0.67245788E-02, 
    },
    {
      0.96493689E-02, -0.22004317E-01, -0.56211897E-02, 0.11911504E+00, 0.25769388E-01, -0.13206676E-01, -0.57847810E-03, -0.63171289E-02, -0.14825935E-02, 
    },
    {
      0.97160034E-01, 0.14064242E+00, 0.12654921E+00, 0.90184147E-02, 0.10824892E+00, 0.98466850E-03, -0.39058451E-01, 0.21360939E-01, -0.68189781E-02, 
    },
    {
      -0.69581389E-01, 0.23204980E-01, -0.33772092E-01, -0.23201568E-01, -0.16431106E-01, 0.35661325E-01, -0.24787607E-01, -0.15851652E-01, 0.41659139E-01, 
    },
    {
      -0.95431058E-03, 0.32939354E+00, 0.11949618E+00, -0.34582306E-01, -0.11245041E+00, 0.58247324E-01, 0.44278201E-01, 0.32697480E-01, 0.79446398E-02, 
    },
    {
      -0.27802193E+00, 0.99982269E-01, -0.18884204E+00, 0.14989616E+00, -0.30045878E-01, -0.12822229E-01, 0.70558451E-01, 0.67012236E-02, -0.54927107E-01, 
    },
    {
      -0.52163047E+00, -0.76617777E+00, -0.38275307E+00, 0.93988568E-01, -0.80029559E+00, -0.46852093E-01, 0.19828326E+00, -0.19820786E+00, -0.20384854E-01, 
    },
    {
      0.73622519E+00, 0.23355688E+00, 0.24254851E+00, 0.40519950E+00, 0.11749786E-01, -0.76143019E-01, 0.65943278E-01, 0.78227133E-01, -0.24041747E+00, 
    },
    {
      -0.48859108E-01, -0.53506434E+00, -0.33032450E+00, -0.10122342E+00, 0.79143211E-01, -0.14064236E+00, -0.84364615E-01, 0.29860295E-02, -0.22253854E-01, 
    },
    {
      0.39246804E+00, -0.24812710E+00, 0.40944198E+00, -0.17673942E+00, -0.78235663E-01, 0.72178833E-01, -0.16227914E+00, 0.10302575E-01, 0.85755751E-01, 
    },
    {
      0.64045298E+00, 0.11233286E+01, 0.36713567E+00, -0.16331942E+00, 0.12432953E+01, 0.20128106E+00, -0.23927681E+00, 0.27385575E+00, 0.72312951E-01, 
    },
    {
      -0.94199336E+00, -0.33690757E+00, -0.17330156E+00, -0.81343490E+00, 0.69725722E-01, 0.41456182E-01, -0.43618813E-03, -0.84605128E-01, 0.24982797E+00, 
    },
    {
      -0.22805035E-01, -0.59726271E-02, 0.38705333E-02, 0.30002026E-02, 0.19159815E-02, 0.18420041E-01, -0.37724916E-01, 0.35234531E-02, -0.43825819E-02, 
    },
    {
      -0.67331134E-02, -0.10876593E-01, -0.79003610E-02, -0.13564546E-01, -0.40359148E-02, 0.42511482E-01, 0.42690061E-01, 0.78176362E-02, -0.44241268E-02, 
    },
    {
      0.61757568E-01, 0.38225956E-01, 0.16911132E-01, -0.45007326E-01, 0.18286875E-01, -0.83427288E-01, 0.43182421E-01, 0.80693150E-02, -0.15373636E-02, 
    },
    {
      0.29610468E-01, -0.18918579E-01, -0.18943191E-01, -0.15242503E-01, -0.12218372E-01, -0.21460386E-01, -0.98122187E-01, -0.44038394E-03, 0.17151907E-01, 
    },
    {
      0.91914488E-02, -0.88108525E-01, 0.29710799E-01, 0.37761092E-01, -0.42912245E-01, -0.14925539E+00, -0.86556077E-01, 0.32728366E-02, -0.97887143E-02, 
    },
    {
      -0.35554547E-01, -0.30815808E-01, -0.29502632E-01, 0.11506282E-01, -0.96649574E-02, 0.11226377E+00, -0.22778022E+00, 0.95754518E-03, 0.19735217E-01, 
    },
    {
      -0.19949598E-01, -0.14575391E-01, 0.80723669E-02, 0.10257792E-01, 0.98831486E-03, 0.21253310E-02, 0.44467166E-03, 0.30331459E-01, -0.68215737E-02, 
    },
    {
      -0.16475892E-01, 0.66600149E-02, -0.96909413E-02, -0.23368848E-02, 0.10803202E-01, 0.18352592E-03, 0.94600813E-02, 0.34825657E-02, 0.29143477E-01, 
    },
    {
      0.68724401E-01, 0.15268687E-01, -0.34583665E-02, -0.26248895E-01, -0.41180071E-02, -0.69923680E-02, 0.78354776E-02, -0.15067245E-01, -0.50502356E-01, 
    },
    {
      0.26632205E-01, -0.38117900E-01, 0.12489504E-01, 0.38269710E-01, -0.23778917E-01, -0.45978669E-02, -0.10741331E-01, 0.55188518E-01, -0.19089222E-01, 
    },
    {
      0.50233081E-02, -0.68038455E-02, 0.13171528E-01, -0.79407496E-02, -0.15828494E-02, 0.69405884E-02, -0.26731647E-02, -0.18833794E-02, 0.10679700E-01, 
    },
    {
      0.77929986E-02, 0.25170868E-01, -0.12785459E-01, -0.85407840E-02, -0.88489456E-02, 0.11496038E-01, -0.64758729E-03, -0.79743098E-02, -0.92306349E-03, 
    },
    {
      -0.26192807E-03, 0.58631459E-02, 0.11213572E-01, 0.88310177E-02, 0.19518093E-02, 0.98038837E-03, -0.22055174E-02, 0.15052836E-02, 0.49419207E-02, 
    },
    {
      -0.49593612E-02, -0.28198517E-02, -0.11953374E-01, -0.12248964E-01, -0.72131357E-02, 0.10401811E-01, -0.17990725E-02, -0.28054372E-02, 0.18662043E-02, 
    },
  },
  {
    {
      0.27312081E+01, -0.36635617E-03, 0.16809828E-02, 0.35598597E-02, 0.75279046E-02, 0.53007563E-03, 0.95740217E-03, -0.22847112E-02, -0.28182086E-03, 
    },
    {
      0.21439911E+00, 0.15668221E-01, 0.39947953E-01, 0.18864455E-01, 0.18349148E-01, -0.73838946E-02, -0.29964812E-01, 0.16750785E-01, 0.58596502E-02, 
    },
    {
      0.98324704E+00, -0.38888179E-01, -0.89483615E-02, -0.30634863E-01, -0.57423547E-01, 0.13029807E-01, 0.31853255E-01, -0.39872225E-01, 0.66098981E-02, 
    },
    {
      -0.71482211E-01, -0.26525632E+00, -0.19658718E-01, -0.33415090E-01, -0.51787488E-01, 0.35274658E-01, 0.85669100E-01, -0.48909988E-01, -0.21501379E-01, 
    },
    {
      -0.13290060E+01, 0.17191923E+00, -0.47023818E-01, 0.52566610E-01, 0.13115127E+00, -0.12865191E-01, -0.10077955E+00, 0.11846802E+00, 0.88568972E-02, 
    },
    {
      -0.13872875E+00, 0.21173365E+00, -0.49534766E-02, 0.75430055E-02, 0.31798299E-01, -0.33703383E-01, -0.59091616E-01, 0.35570819E-01, 0.11184796E-01, 
    },
    {
      0.40687090E+00, -0.11250211E+00, 0.64376906E-01, -0.24939286E-01, -0.85930578E-01, -0.26775303E-02, 0.72211958E-01, -0.74090555E-01, -0.24171596E-01, 
    },
    {
      0.18235693E-01, -0.32615599E+00, -0.29267040E+00, 0.70642778E-02, 0.11300360E-02, -0.19541867E-02, 0.65887659E-02, 0.25641054E-03, 0.29901171E-03, 
    },
    {
      0.10992618E-01, 0.29850316E+00, -0.32224572E+00, -0.13626590E-01, -0.15000732E-02, -0.56463708E-02, -0.13415891E-01, 0.31011614E-02, 0.13834988E-01, 
    },
    {
      -0.92273327E-02, 0.24425700E+00, 0.56881353E-01, 0.36258149E-03, 0.63643664E-01, -0.37808134E-03, -0.32947324E-01, 0.38247403E-01, 0.19941090E-01, 
    },
    {
      -0.57335582E-01, 0.87612174E-01, 0.25150666E+00, -0.19707844E-01, -0.44863459E-01, 0.16167358E-01, -0.10398516E-01, 0.50344802E-01, 0.68640865E-01, 
    },
    {
      -0.23492485E+00, 0.11852225E+01, 0.87067664E+00, -0.62521182E-01, 0.42712395E-02, 0.31809572E-01, 0.21999139E-01, -0.92716359E-01, -0.10461421E+00, 
    },
    {
      -0.57572049E+00, -0.91379106E+00, 0.12106148E+01, 0.14713533E+00, 0.49721982E-01, 0.31173525E-01, 0.14988387E+00, 0.74236974E-01, -0.24920762E+00, 
    },
    {
      0.37146536E+00, -0.15066245E+01, -0.33582768E+00, 0.13240382E+00, -0.49595794E+00, -0.17519762E+00, 0.14915365E-01, -0.31701419E+00, -0.11277883E+00, 
    },
    {
      0.11722958E+01, -0.33704272E+00, -0.16070234E+01, -0.73385984E-01, 0.44775617E+00, -0.89496598E-01, 0.11555021E+00, -0.42885178E+00, -0.43839565E+00, 
    },
    {
      0.61520100E+00, -0.11530799E+01, 0.13902967E+00, 0.11537025E+00, -0.13625573E-01, -0.11211418E+00, -0.90618208E-01, 0.34417906E+00, 0.43181029E+00, 
    },
    {
      0.19090471E+01, -0.30285779E+00, -0.14655380E+01, -0.29066685E+00, -0.22451898E+00, -0.12008682E+00, -0.50865197E+00, -0.31707212E+00, 0.65367371E+00, 
    },
    {
      -0.86626554E+00, 0.25727086E+01, 0.11919091E+01, -0.25942805E+00, 0.98266923E+00, 0.50214672E+00, 0.11235259E+00, 0.75492084E+00, 0.13290592E+00, 
    },
    {
      -0.35606391E+01, 0.50929976E+00, 0.30332544E+01, 0.46690202E+00, -0.93843836E+00, 0.33491963E+00, -0.23293011E+00, 0.92977613E+00, 0.81740057E+00, 
    },
    {
      -0.38443810E+00, 0.52484918E+00, -0.42313319E+00, -0.31547219E-01, -0.23082055E-01, 0.98553799E-01, 0.48045248E-01, -0.29371861E+00, -0.38383645E+00, 
    },
    {
      -0.16131206E+01, 0.93166280E+00, 0.74328411E+00, 0.15688561E+00, 0.21910542E+00, 0.11658286E+00, 0.44385305E+00, 0.28195140E+00, -0.46593127E+00, 
    },
    {
      0.43430272E+00, -0.11887656E+01, -0.43771577E+00, 0.14430624E+00, -0.56061119E+00, -0.36049953E+00, -0.12475537E+00, -0.52396154E+00, -0.43358773E-01, 
    },
    {
      0.28472085E+01, -0.78438592E+00, -0.16446294E+01, -0.40643516E+00, 0.59259111E+00, -0.32095200E+00, 0.95902123E-01, -0.59256905E+00, -0.44462144E+00, 
    },
    {
      0.10362735E-02, -0.26521128E-01, 0.73475315E-03, 0.78305684E-01, -0.87907061E-01, -0.87850457E-02, 0.68914322E-02, -0.58439686E-02, -0.10781294E-01, 
    },
    {
      0.83343238E-02, -0.22290355E-01, 0.50302162E-02, 0.79541668E-01, 0.76331988E-01, -0.69553126E-02, -0.10722160E-02, -0.23844005E-02, -0.24587914E-03, 
    },
    {
      0.73374845E-01, 0.11703365E+00, 0.75943522E-01, -0.60151584E-01, 0.15490431E+00, 0.13585863E-01, -0.18192250E-01, 0.11195567E-01, 0.70080832E-02, 
    },
    {
      -0.29040046E-01, 0.32617461E-01, -0.60124036E-01, -0.11059102E+00, -0.43481726E-01, 0.27728144E-01, 0.21420596E-01, -0.26929742E-01, 0.35609290E-01, 
    },
    {
      -0.37986711E-01, 0.21261556E+00, 0.65299794E-01, -0.47157770E+00, 0.18923242E+00, 0.89936638E-02, 0.30793028E-02, 0.21662014E-01, 0.45056507E-01, 
    },
    {
      -0.18105258E+00, 0.52625749E-01, -0.21388224E+00, -0.12751260E+00, -0.44269633E+00, -0.68748944E-01, 0.48228782E-01, 0.46695493E-01, -0.61157331E-01, 
    },
    {
      -0.31314614E+00, -0.65667820E+00, -0.16174650E+00, 0.18607917E+00, -0.82700682E+00, -0.49523894E-01, 0.11539982E+00, -0.11283565E+00, -0.42533660E-02, 
    },
    {
      0.31864041E+00, 0.13107453E+00, 0.33589387E+00, 0.72368956E+00, 0.48834573E-01, -0.47290842E-02, -0.90480998E-01, 0.11117186E+00, -0.19858018E+00, 
    },
    {
      0.16404297E-01, -0.35947222E+00, -0.23393697E+00, 0.24125105E+00, -0.40167990E+00, -0.59507195E-01, -0.12976588E-01, 0.16296888E-02, -0.64027645E-01, 
    },
    {
      0.21480629E+00, -0.16956283E+00, 0.42124182E+00, 0.31029803E+00, 0.24321973E+00, 0.14924949E+00, -0.11708120E+00, -0.62503330E-01, 0.91057926E-01, 
    },
    {
      0.29138744E+00, 0.99653769E+00, 0.11597233E+00, -0.23700126E+00, 0.11012855E+01, 0.13280974E+00, -0.17497236E+00, 0.16628051E+00, -0.11821241E-01, 
    },
    {
      -0.40243325E+00, -0.14012007E+00, -0.34927446E+00, -0.10288903E+01, -0.43111134E-01, -0.44211321E-01, 0.12559946E+00, -0.10422827E+00, 0.19184583E+00, 
    },
    {
      -0.12884370E-01, -0.28929866E-02, 0.62589170E-02, -0.43417537E-02, 0.16956747E-02, 0.30290781E-01, -0.54206152E-01, 0.47972486E-02, -0.54383771E-02, 
    },
    {
      -0.18483959E-02, -0.49267081E-02, -0.35054756E-02, -0.10494787E-02, -0.31706650E-03, 0.53946707E-01, 0.46015918E-01, 0.77006100E-02, -0.74489904E-02, 
    },
    {
      0.28184274E-01, 0.48737317E-01, 0.56083137E-02, -0.11658832E-01, 0.10535412E-01, -0.77447735E-01, 0.14809875E-01, -0.54364088E-02, 0.26957258E-02, 
    },
    {
      0.12003129E-01, -0.48028270E-03, -0.13672524E-01, -0.10574406E-01, -0.94266171E-02, -0.67775045E-02, -0.85156538E-01, 0.66752793E-04, 0.83940942E-02, 
    },
    {
      0.25840148E-01, -0.82805589E-01, 0.31506196E-01, 0.41513219E-01, -0.28803995E-01, -0.18279599E+00, -0.70030414E-01, 0.43120938E-02, -0.44541989E-03, 
    },
    {
      -0.10199835E-01, -0.34188926E-01, -0.19336939E-01, 0.11237562E-01, -0.19958831E-01, 0.87408088E-01, -0.20651507E+00, -0.62467060E-02, 0.28983794E-01, 
    },
    {
      -0.82333675E-02, -0.15517812E-01, 0.84783202E-02, 0.52443543E-03, 0.13247513E-02, 0.27244100E-02, 0.20155314E-03, 0.27045680E-01, 0.32722121E-02, 
    },
    {
      -0.98290918E-02, 0.24848564E-02, -0.12258626E-01, 0.43838951E-02, 0.21730985E-02, 0.14691027E-03, 0.10895627E-01, -0.70574167E-02, 0.27331531E-01, 
    },
    {
      0.34686182E-01, 0.11546563E-01, -0.10812423E-01, -0.11996835E-01, -0.61492608E-02, -0.16286846E-01, 0.90343263E-02, -0.16508182E-02, -0.44392467E-01, 
    },
    {
      0.24883408E-01, -0.41023172E-01, 0.30125706E-01, 0.19361736E-01, -0.84361620E-02, -0.86162388E-02, -0.12216836E-01, 0.51103357E-01, -0.75027507E-04, 
    },
    {
      0.11774480E-01, -0.37965027E-02, 0.73318430E-02, -0.37572165E-02, -0.29526060E-02, 0.51077693E-02, -0.20714712E-02, -0.67285765E-02, 0.69300095E-02, 
    },
    {
      0.76303934E-02, 0.18072840E-01, -0.92502227E-02, -0.23002878E-02, -0.76491982E-02, 0.66094738E-02, 0.57324427E-02, -0.36398084E-02, 0.18937787E-02, 
    },
    {
      0.12494961E-03, 0.95274886E-02, 0.79445858E-02, 0.87751141E-02, -0.22462981E-02, 0.10836435E-02, 0.13684512E-02, -0.88894583E-03, 0.21372179E-02, 
    },
    {
      -0.11187471E-02, -0.17962216E-02, -0.84776217E-02, -0.32081937E-02, -0.57096705E-02, 0.51954137E-02, 0.20163737E-02, -0.19013528E-02, 0.13429531E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir13_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.65998969E+01, -0.96180730E-01, -0.18148595E-01, 0.64034291E-01, -0.14069435E-01, -0.11540678E-01, 0.66070836E-02, 0.26406836E-01, -0.39686030E-03, 0.23210552E-01, -0.22450823E-02, -0.10104239E-01, -0.12631579E-01, 
    },
    {
      -0.42413343E-01, 0.80642931E-01, 0.27393836E+00, -0.16345827E+00, -0.16899700E+00, -0.32240987E+00, 0.70448697E-01, -0.77763997E-01, -0.13244358E-01, 0.30214982E-01, -0.13424671E+00, -0.17871724E-01, 0.15353709E-01, 
    },
    {
      0.17452030E+02, 0.14823139E+01, -0.10489835E+01, -0.31613788E+00, 0.14407470E+01, -0.17907827E+00, 0.29510629E+00, -0.22444358E-01, 0.21991573E+00, -0.10238969E+01, 0.21552382E+00, 0.43258831E+00, 0.30413870E-01, 
    },
    {
      -0.42824140E+01, -0.43370719E+01, 0.27843063E+01, 0.32597795E+01, 0.52011843E+01, 0.38541880E+01, -0.19715509E+01, 0.21992905E+01, 0.23029692E+01, -0.60228556E+00, 0.22066264E+01, 0.23972336E-01, -0.80908841E+00, 
    },
    {
      -0.11069540E+03, -0.47234511E+01, 0.81807337E+01, -0.57918835E+01, -0.15985530E+02, 0.41825180E+01, 0.31950969E+00, -0.15979006E+01, -0.29805335E+00, 0.65984240E+01, -0.15701975E+01, -0.28786166E+01, -0.55098748E+00, 
    },
    {
      0.53160348E+01, 0.72372947E+01, -0.31718040E+02, -0.84363041E+01, -0.20782209E+02, -0.15876794E+02, 0.10275279E+02, -0.12953021E+02, -0.18293257E+02, 0.30118990E+01, -0.10225961E+02, 0.10002422E+01, 0.57159657E+01, 
    },
    {
      0.22316840E+03, 0.88046808E+01, -0.19268261E+02, 0.27535299E+02, 0.49745659E+02, -0.14210365E+02, -0.63509369E+01, 0.59100389E+01, -0.75693429E+00, -0.15410734E+02, 0.43323641E+01, 0.69434185E+01, 0.26285546E+01, 
    },
    {
      0.29579544E+02, 0.18237551E+02, 0.82377556E+02, 0.84427804E+00, 0.30343164E+02, 0.29563707E+02, -0.19898548E+02, 0.30717596E+02, 0.48212425E+02, -0.59438219E+01, 0.19914419E+02, -0.42029824E+01, -0.14977191E+02, 
    },
    {
      -0.20669727E+03, -0.11006776E+02, 0.17986933E+02, -0.39454205E+02, -0.59185928E+02, 0.16635517E+02, 0.11510051E+02, -0.69961228E+01, 0.14310032E+01, 0.15335248E+02, -0.50885744E+01, -0.70506644E+01, -0.39190328E+01, 
    },
    {
      -0.65092957E+02, -0.45056271E+02, -0.80958855E+02, 0.13474419E+02, -0.17513617E+02, -0.25489252E+02, 0.16221642E+02, -0.31921728E+02, -0.51784859E+02, 0.50485601E+01, -0.17261196E+02, 0.55830841E+01, 0.16545364E+02, 
    },
    {
      0.74110527E+02, 0.55856385E+01, -0.58897896E+01, 0.18034472E+02, 0.23959206E+02, -0.63647890E+01, -0.57852674E+01, 0.26369393E+01, -0.59193814E+00, -0.55258875E+01, 0.21282353E+01, 0.25619016E+01, 0.18311784E+01, 
    },
    {
      0.34186604E+02, 0.23974506E+02, 0.27236195E+02, -0.89847775E+01, 0.29195013E+01, 0.82554817E+01, -0.46716986E+01, 0.12058103E+02, 0.19580334E+02, -0.15368044E+01, 0.54791775E+01, -0.23879936E+01, -0.65005140E+01, 
    },
    {
      -0.24007400E-01, 0.21675975E+01, 0.11725757E+01, 0.50758090E-01, 0.13093476E-01, -0.23201449E-01, 0.38342886E-01, 0.48389513E-01, 0.63410657E-02, 0.44442564E-01, 0.19522274E-01, -0.27899284E-01, -0.14491891E-02, 
    },
    {
      -0.34519996E-01, -0.11496038E+01, 0.21471732E+01, -0.14192484E+00, 0.68746604E-01, -0.33114925E-01, -0.61117645E-01, 0.42638432E-01, 0.90309959E-02, -0.28206121E-01, -0.93916431E-02, -0.10890185E-01, 0.86654341E-02, 
    },
    {
      -0.62337194E-01, 0.79887128E+00, 0.91313583E+00, -0.18493378E+00, 0.26627880E+00, 0.28355405E+00, 0.51949298E+00, 0.11120747E+00, 0.27354652E+00, -0.12354252E+00, -0.27485764E-01, -0.33132059E-02, 0.10892819E+00, 
    },
    {
      -0.26281741E+00, 0.44478518E+00, 0.60648137E+00, -0.41248816E+00, 0.25587833E+00, -0.95881902E-01, -0.33539433E-01, 0.57145733E+00, 0.13682042E+00, -0.25388572E+00, -0.10539193E+00, 0.25895587E+00, -0.85921496E-01, 
    },
    {
      0.27832451E+01, 0.16864960E+02, 0.40619507E+01, -0.20387976E+01, 0.93537444E+00, 0.16030459E+01, 0.35865560E+00, -0.70939946E+00, 0.63179672E+00, -0.16262046E+01, 0.22060844E+00, 0.10916179E+01, -0.26890588E+00, 
    },
    {
      -0.38157611E+01, -0.31960356E+01, 0.17024626E+02, 0.35198452E+01, 0.14320093E+01, -0.42652923E+00, 0.44432792E+00, -0.12067032E+00, -0.51051909E+00, 0.10688665E+01, 0.34004807E+00, 0.25413656E+00, -0.20112856E+00, 
    },
    {
      0.69541912E+01, -0.18067736E+02, 0.91149321E+01, 0.38701576E+00, 0.36137350E+01, -0.53189068E+01, -0.11495462E+02, -0.10826284E+01, -0.27792187E+01, 0.27802820E+01, 0.12037225E+01, -0.14036686E+01, -0.12014115E+01, 
    },
    {
      0.12424923E+02, -0.22949505E+02, -0.12072407E+02, 0.37552617E+01, -0.35413680E+01, 0.57876983E+01, 0.79935116E+00, -0.93507385E+01, -0.10703611E+01, 0.37773585E+01, 0.11066062E+01, -0.40062089E+01, -0.51845562E+00, 
    },
    {
      -0.17640097E+02, -0.10663563E+03, 0.76676788E+02, 0.15154503E+02, -0.11779245E+02, -0.14249015E+02, -0.23536947E+01, 0.50884314E+01, -0.77559652E+01, 0.10476252E+02, -0.28326011E+01, -0.80197392E+01, 0.17742889E+01, 
    },
    {
      0.41513287E+02, -0.97553841E+02, -0.95567078E+02, -0.19530249E+02, -0.16382975E+02, 0.50004249E+01, 0.46859521E+00, 0.59307051E+00, 0.33864222E+01, -0.85503206E+01, -0.49351854E+01, -0.36469856E+00, 0.17972651E+01, 
    },
    {
      -0.50485680E+02, 0.11387515E+03, -0.35196548E+02, 0.13647220E+02, -0.18866581E+02, 0.35348217E+02, 0.61966629E+02, 0.28466079E+01, 0.12943146E+02, -0.17173031E+02, -0.69480538E+01, 0.11160711E+02, 0.54356551E+01, 
    },
    {
      -0.10127907E+03, 0.11782328E+03, 0.68567123E+02, -0.27491049E+02, 0.10630718E+02, -0.37230160E+02, -0.21067333E+01, 0.51232288E+02, 0.83989620E+01, -0.19422342E+02, -0.21974294E+01, 0.19062983E+02, 0.78485546E+01, 
    },
    {
      0.37536377E+02, 0.21429555E+03, -0.35160019E+03, -0.42891769E+02, 0.40961739E+02, 0.44047215E+02, 0.81962329E+00, -0.13827250E+02, 0.28208015E+02, -0.26970650E+02, 0.91622019E+01, 0.21888941E+02, -0.39068761E+01, 
    },
    {
      -0.13467381E+03, 0.43239706E+03, 0.16293761E+03, 0.55668144E+02, 0.53183456E+02, -0.12969423E+02, -0.54627337E+01, -0.22841866E+01, -0.78471823E+01, 0.25674978E+02, 0.17792437E+02, -0.14446113E+01, -0.57079659E+01, 
    },
    {
      0.15792365E+03, -0.30741367E+03, 0.12066244E+02, -0.73300522E+02, 0.21158731E+02, -0.10123546E+03, -0.13733231E+03, 0.33357602E-01, -0.29374790E+02, 0.42309250E+02, 0.14585931E+02, -0.30827669E+02, -0.13427900E+02, 
    },
    {
      0.27333496E+03, -0.24579082E+03, -0.17028867E+03, 0.87392265E+02, -0.19682516E+02, 0.87100410E+02, 0.28480635E+01, -0.12049124E+03, -0.31395855E+02, 0.46579369E+02, -0.16486955E+01, -0.39088516E+02, -0.23570745E+02, 
    },
    {
      -0.31926086E+02, -0.18912199E+03, 0.49377734E+03, 0.51254738E+02, -0.55447197E+02, -0.56824291E+02, 0.64984670E+01, 0.14788384E+02, -0.39874298E+02, 0.31675289E+02, -0.11539883E+02, -0.25417494E+02, 0.36541965E+01, 
    },
    {
      0.17872554E+03, -0.60567108E+03, -0.10871112E+03, -0.77059822E+02, -0.71071373E+02, 0.11870968E+02, 0.10181161E+02, 0.38203001E+01, 0.61157165E+01, -0.31524561E+02, -0.23383059E+02, 0.37743144E+01, 0.61833124E+01, 
    },
    {
      -0.22133595E+03, 0.37228479E+03, 0.54766731E+02, 0.12162727E+03, 0.21483109E+01, 0.12726328E+03, 0.13672382E+03, -0.70452433E+01, 0.30163021E+02, -0.44665485E+02, -0.12745265E+02, 0.35419662E+02, 0.16225985E+02, 
    },
    {
      -0.29494229E+03, 0.23506215E+03, 0.19200447E+03, -0.10832544E+03, 0.29829510E+02, -0.84662743E+02, -0.59570804E+01, 0.12735707E+03, 0.48301052E+02, -0.52132950E+02, 0.77952442E+01, 0.36343361E+02, 0.26745314E+02, 
    },
    {
      0.91377707E+01, 0.66141739E+02, -0.22206757E+03, -0.21895306E+02, 0.25961342E+02, 0.26107534E+02, -0.57225599E+01, -0.50952439E+01, 0.19250315E+02, -0.14170377E+02, 0.50944929E+01, 0.10638086E+02, -0.13205591E+01, 
    },
    {
      -0.84391235E+02, 0.27314240E+03, 0.24791868E+02, 0.38730942E+02, 0.33692513E+02, -0.33924561E+01, -0.61104679E+01, -0.23151672E+01, -0.92081583E+00, 0.13415736E+02, 0.10256060E+02, -0.23337302E+01, -0.19138920E+01, 
    },
    {
      0.11088846E+03, -0.16574039E+03, -0.44082218E+02, -0.64954468E+02, -0.86842918E+01, -0.57481571E+02, -0.51243740E+02, 0.55801930E+01, -0.11060497E+02, 0.16775259E+02, 0.39023740E+01, -0.14520377E+02, -0.73669624E+01, 
    },
    {
      0.11177200E+03, -0.84463303E+02, -0.81101242E+02, 0.44699486E+02, -0.19768431E+02, 0.28911020E+02, 0.49595222E+01, -0.49829720E+02, -0.25354982E+02, 0.21757786E+02, -0.52241926E+01, -0.12592972E+02, -0.10411530E+02, 
    },
    {
      0.30574529E-01, -0.31148717E-01, -0.33901516E-01, -0.87049460E+00, 0.20368204E+00, -0.32607842E-01, -0.16634515E-01, 0.27669171E-01, 0.36947880E-01, 0.47631604E-02, 0.34091142E-02, -0.23116291E-01, 0.21287642E-01, 
    },
    {
      0.58895037E-01, 0.15586518E-01, 0.69088846E-01, -0.21127127E+00, -0.80712038E+00, -0.14075822E-01, -0.17555034E-01, -0.16835840E-01, 0.10679643E-01, -0.12162154E-01, -0.73469672E-02, -0.37808849E-02, 0.10991140E-02, 
    },
    {
      0.62006795E+00, 0.13560387E+00, 0.16468419E+00, 0.99273920E-01, -0.26552621E+00, 0.37287197E+00, -0.35812023E+00, -0.19624999E-01, 0.44019237E-01, -0.16461620E+00, 0.20510979E-01, -0.11776922E+00, 0.38004402E-01, 
    },
    {
      0.10694618E+01, 0.31023821E+00, 0.14203931E+00, -0.43358117E-01, 0.10166671E+00, 0.41650143E+00, 0.48760930E+00, -0.12005379E+00, 0.97967356E-01, -0.15563516E+00, -0.22586857E+00, 0.49013641E-01, -0.20813970E-01, 
    },
    {
      0.18206578E+01, -0.41084620E+00, 0.27921236E+01, 0.24209125E+01, 0.16685303E+01, -0.28808552E+00, -0.90014291E+00, -0.50827670E+00, -0.45300135E+00, -0.35001542E-01, 0.36338080E-01, 0.24220033E+00, -0.31733364E+00, 
    },
    {
      0.43253836E+00, -0.63174385E+00, 0.50231868E+00, -0.17871255E+01, 0.42565956E+01, -0.41577759E+00, -0.12443371E-01, 0.90635931E+00, -0.81871720E-02, 0.12240939E+00, -0.58664852E+00, 0.16192600E-01, -0.34534615E-01, 
    },
    {
      -0.39647148E+01, 0.33445108E+00, -0.17875113E+01, 0.35496018E+01, 0.51491537E+01, -0.33950496E+01, 0.13043875E+01, 0.25871399E+00, 0.68523347E-01, 0.13834658E+01, 0.79948354E+00, 0.91803938E+00, -0.45803875E-01, 
    },
    {
      -0.64467540E+01, -0.20690763E+01, 0.49838591E+00, -0.82725680E+00, 0.50504017E+01, -0.24308875E+01, -0.33799708E+01, 0.58349323E+00, -0.12550083E+01, 0.10462335E+01, 0.23345358E+01, -0.70762014E+00, 0.51471233E+00, 
    },
    {
      -0.15008719E+02, 0.14641236E+02, -0.20451660E+02, 0.19731981E+02, 0.55634995E+01, 0.25407569E+01, 0.62443280E+01, 0.28349652E+01, 0.24219623E+01, -0.96738261E+00, -0.41830257E+00, -0.83132789E-01, 0.83228946E+00, 
    },
    {
      -0.45921922E+01, 0.10040653E+02, -0.35055692E+01, -0.40767598E+01, 0.60929508E+01, 0.75083405E+00, 0.31848945E-01, -0.46174974E+01, -0.34868580E+00, 0.49051598E+00, 0.39066932E+01, 0.55248684E+00, -0.44293070E+00, 
    },
    {
      0.63590775E+01, -0.54019079E+01, 0.35216527E+01, -0.11525576E+02, -0.16511587E+02, 0.75384369E+01, -0.19127115E+01, -0.46861181E+00, -0.16359047E+01, -0.28776963E+01, -0.27371666E+01, -0.23561432E+01, -0.15143438E+00, 
    },
    {
      0.10229012E+02, 0.11192629E+01, -0.24896045E+01, 0.21574695E+01, -0.16082794E+02, 0.50335708E+01, 0.60995483E+01, 0.46528451E-01, 0.44228539E+01, -0.24760296E+01, -0.58701057E+01, 0.17978436E+01, -0.14681013E+01, 
    },
    {
      0.34721951E+02, -0.42912174E+02, 0.43504799E+02, -0.63420731E+02, -0.24599474E+02, -0.63787656E+01, -0.11826031E+02, -0.60871482E+01, -0.45813766E+01, 0.32331924E+01, 0.11356505E+01, -0.13086146E+01, -0.20494525E+00, 
    },
    {
      0.13705240E+02, -0.26223623E+02, 0.63174248E+01, 0.21032364E+02, -0.33421906E+02, 0.69371116E+00, 0.33373719E+00, 0.82973967E+01, 0.57793902E-02, -0.35463052E+01, -0.80484915E+01, -0.15746634E+01, 0.11740108E+01, 
    },
    {
      -0.34624617E+01, 0.80138206E+01, -0.30727046E+01, 0.99697714E+01, 0.14296803E+02, -0.47378635E+01, 0.91392756E+00, 0.25849405E+00, 0.23256736E+01, 0.17453274E+01, 0.22083228E+01, 0.17524531E+01, 0.13290420E+00, 
    },
    {
      -0.44886589E+01, 0.27600312E+01, 0.26766870E+01, -0.16193961E+01, 0.12873305E+02, -0.35492828E+01, -0.32382324E+01, -0.82338858E+00, -0.40521255E+01, 0.16536398E+01, 0.42015963E+01, -0.11780934E+01, 0.11460971E+01, 
    },
    {
      -0.24425175E+02, 0.32594193E+02, -0.26974190E+02, 0.44860214E+02, 0.18794016E+02, 0.41103082E+01, 0.71137915E+01, 0.43121295E+01, 0.26971650E+01, -0.24290116E+01, -0.88320661E+00, 0.13616098E+01, -0.47190940E+00, 
    },
    {
      -0.12868726E+02, 0.17487787E+02, -0.35085900E+01, -0.16604246E+02, 0.24305555E+02, -0.93955803E+00, -0.42676285E+00, -0.49368086E+01, 0.68438685E+00, 0.36197155E+01, 0.52519341E+01, 0.11007043E+01, -0.74931699E+00, 
    },
    {
      -0.15135266E+00, -0.28601021E-02, -0.89310557E-01, -0.33103667E-01, 0.79293363E-01, -0.50901473E+00, -0.82532144E+00, 0.27812915E-01, 0.20369733E-01, 0.24551898E-01, 0.18150743E-01, -0.15981859E-01, 0.17871849E-01, 
    },
    {
      0.47308975E-02, -0.10820430E+00, 0.30828558E-01, 0.20033805E-01, 0.29000906E-01, 0.75493073E+00, -0.47701424E+00, 0.32587885E-02, 0.46372235E-01, -0.10304347E-01, -0.20734092E-01, 0.25965128E-01, -0.58355234E-02, 
    },
    {
      -0.30175501E-01, 0.29149768E+00, -0.15922238E+00, -0.23880073E+00, -0.31934550E+00, -0.21410786E+00, -0.14080049E+00, -0.13714211E+00, 0.11907785E+00, 0.39285958E-01, 0.14231239E+00, 0.47506593E-01, -0.21241790E-01, 
    },
    {
      -0.11470716E+00, 0.12420133E+00, -0.40786356E-01, -0.17982394E+00, 0.13026784E+00, 0.35654244E+00, -0.31161553E+00, -0.22107504E+00, 0.22054937E-01, -0.22563413E-01, -0.27322270E-01, 0.78594657E-02, 0.29850721E-01, 
    },
    {
      0.91586936E+00, 0.45026922E+00, 0.85786003E+00, 0.57185078E+00, -0.34676486E+00, 0.13745323E+01, 0.23583150E+01, -0.16617003E+00, -0.55329833E-01, -0.98012596E-01, -0.80390774E-01, 0.74126594E-01, -0.24377277E-01, 
    },
    {
      -0.41545358E+00, 0.12106431E+01, -0.93434161E+00, 0.17056143E+00, -0.64522797E+00, -0.23836486E+01, 0.16357204E+01, -0.10281745E+00, -0.50418901E+00, 0.29982610E-01, -0.91422498E-01, -0.29078582E+00, 0.26920128E-02, 
    },
    {
      -0.32012030E+00, -0.33131790E+00, 0.44400707E+00, 0.35786450E+00, 0.46332926E+00, 0.39637893E+00, 0.22233669E+00, 0.42943186E+00, -0.11153209E+00, -0.72281212E-01, -0.15658084E+00, -0.14401633E+00, 0.95212579E-01, 
    },
    {
      0.34326747E+00, -0.40543489E-01, 0.20303135E+00, -0.67969397E-01, -0.69856882E-01, -0.80425566E+00, 0.52323300E+00, 0.22219919E+00, 0.59583042E-01, 0.29945867E-01, 0.43922316E-01, -0.15566495E-01, -0.91002490E-02, 
    },
    {
      -0.44131538E+00, -0.30327237E+00, -0.92330801E+00, -0.77402663E+00, 0.13792758E+00, -0.20475721E+01, -0.27376485E+01, 0.10802826E+00, -0.65257788E-01, 0.78867815E-01, 0.10399983E+00, 0.14911513E-01, -0.53068724E-01, 
    },
    {
      0.66848314E+00, -0.16395499E+01, 0.17826374E+01, -0.61707690E-01, 0.10942951E+01, 0.34093125E+01, -0.27390239E+01, 0.11946615E+00, 0.77287668E+00, 0.62772550E-01, 0.15083092E+00, 0.45553604E+00, -0.37019581E-01, 
    },
    {
      0.21583674E-01, 0.64543181E-03, -0.51576743E-03, 0.58179975E-01, 0.27645398E-01, 0.17042680E-01, 0.23695769E-01, 0.30250359E+00, -0.15979145E+00, 0.11132521E-01, 0.16834151E-01, -0.14365312E-01, -0.20031256E-02, 
    },
    {
      0.37123345E-01, 0.31039350E-01, -0.32179166E-01, 0.90723867E-02, -0.57725228E-01, 0.15170093E-01, 0.13725638E-01, 0.11982990E+00, 0.30412433E+00, 0.16133197E-01, 0.21785465E-02, 0.67538638E-02, -0.11140665E-01, 
    },
    {
      -0.14433794E-02, 0.11381988E+00, -0.10765873E+00, -0.10239576E+00, -0.81107587E-01, -0.61356392E-01, 0.38089089E-01, 0.56286909E-01, 0.27529791E-01, -0.90354122E-01, 0.81266269E-01, 0.29821135E-03, 0.13847735E-01, 
    },
    {
      0.18724556E+00, 0.95570683E-02, -0.21760223E-01, 0.13237505E+00, 0.12431903E+00, -0.66988245E-01, -0.10429621E+00, -0.10871855E-01, -0.82563132E-03, -0.46499770E-01, -0.35336360E-01, 0.74498006E-02, -0.50428421E-02, 
    },
    {
      -0.87156013E-01, 0.10371949E-01, -0.10520714E+00, 0.13535339E-01, -0.12469280E-01, 0.89040743E-02, 0.31624246E-01, 0.37560670E-02, -0.11854895E-01, 0.15747434E+00, 0.22601800E+00, -0.16246822E-01, -0.11745606E-01, 
    },
    {
      0.61789237E-01, 0.55235770E-01, -0.75306934E-02, -0.16015360E-01, -0.22144977E-01, 0.42483758E-01, -0.94584087E-02, -0.29955334E-04, -0.20362385E-01, -0.22954023E+00, 0.14123079E+00, 0.11846971E-01, -0.48866938E-02, 
    },
    {
      0.41830935E-01, -0.25836842E-01, 0.33995997E-01, -0.30328436E-02, -0.43513537E-02, 0.32394242E-01, 0.24555236E-01, 0.15828688E-01, -0.14602708E-01, 0.79171928E-02, -0.24579484E-02, -0.12720145E+00, 0.61430719E-01, 
    },
    {
      0.11831303E-01, 0.48629563E-01, -0.36074687E-01, 0.10915471E-01, 0.65989299E-02, 0.81955465E-02, -0.26336169E-01, -0.81899911E-02, -0.11951329E-01, -0.18219600E-01, -0.14673570E-01, -0.55252794E-01, -0.12856482E+00, 
    },
    {
      -0.18628046E-01, -0.58445823E-02, -0.82438067E-01, 0.24269113E-01, 0.24777478E-01, -0.10445634E-01, -0.56395708E-02, 0.88527258E-02, -0.72764154E-02, 0.19107535E-03, -0.24804156E-01, 0.88188238E-02, 0.33777929E-02, 
    },
    {
      0.22041786E-01, 0.75239100E-01, 0.32701142E-01, -0.36235764E-02, 0.28052095E-01, 0.17236874E-01, 0.44026822E-03, -0.37383433E-01, 0.11369155E-02, 0.21209640E-02, 0.54265251E-02, 0.13008385E-02, 0.85698189E-02, 
    },
    {
      0.10458428E-01, -0.10118044E-02, -0.77243894E-01, 0.30200144E-04, 0.16539751E-01, -0.37585138E-02, 0.45206282E-01, -0.19975781E-01, 0.64657149E-02, 0.45526298E-02, -0.13365079E-01, 0.17035095E-01, 0.60681752E-02, 
    },
    {
      -0.31197891E-01, 0.27110072E-01, -0.81724217E-02, 0.53124398E-01, -0.37347758E-03, -0.60616857E-02, 0.15380551E-01, 0.24982348E-01, 0.43096016E-02, -0.35201445E-01, -0.23587819E-01, -0.47319382E-02, 0.29201475E-02, 
    },
  },
  {
    {
      0.93594637E+01, 0.11758081E-01, 0.84277675E-01, 0.73569529E-01, 0.11903162E-01, -0.44155386E-02, -0.15138906E-01, 0.74017155E-02, -0.15588634E-01, 0.36506962E-01, -0.14251899E-02, -0.65558045E-02, -0.18775851E-02, 
    },
    {
      0.88674414E+00, 0.51994503E-01, 0.62100869E-02, -0.30662936E+00, -0.13489079E+00, -0.16742069E+00, -0.27921739E+00, -0.73416762E-01, -0.51260434E-01, 0.34743927E-01, -0.64074233E-01, -0.14990614E-01, -0.33514559E-01, 
    },
    {
      0.29466911E+02, -0.10430388E+01, -0.17538368E+01, 0.64629501E+00, 0.15527004E+01, -0.81919432E+00, 0.24933697E+00, 0.21972425E+00, 0.14215918E+00, -0.14595740E+01, 0.24961294E+00, 0.30417073E+00, -0.55456981E-01, 
    },
    {
      -0.67446027E+01, 0.40613857E+00, 0.46058164E+01, 0.29142275E+01, 0.43435960E+01, 0.19936315E+01, 0.27405980E+01, 0.19893142E+01, 0.26587334E+01, -0.36289024E+00, 0.87958097E+00, 0.43432501E+00, 0.45760855E+00, 
    },
    {
      -0.14369508E+03, 0.82375011E+01, 0.86440678E+01, -0.11147148E+02, -0.15416179E+02, 0.78732028E+01, 0.98172742E+00, -0.10455229E+01, 0.79332030E+00, 0.88577414E+01, -0.20928278E+01, -0.17485448E+01, -0.37013805E+00, 
    },
    {
      -0.38804274E+01, -0.24045439E+02, -0.22236847E+02, -0.39119647E+01, -0.14422157E+02, -0.11314327E+02, -0.11605090E+02, -0.11023640E+02, -0.19990587E+02, 0.18404472E+01, -0.36433988E+01, -0.19584160E+01, -0.16701306E+01, 
    },
    {
      0.24726463E+03, -0.13912273E+02, -0.16638878E+02, 0.39193115E+02, 0.45437634E+02, -0.20975456E+02, -0.71870809E+01, 0.10358952E+01, -0.41288266E+01, -0.20344755E+02, 0.61743193E+01, 0.37249963E+01, 0.23654854E+01, 
    },
    {
      0.73438400E+02, 0.85878952E+02, 0.28821287E+02, -0.12087472E+02, 0.15061758E+02, 0.27009087E+02, 0.25284414E+02, 0.25017273E+02, 0.52237434E+02, -0.39412918E+01, 0.60525670E+01, 0.32629383E+01, 0.20258255E+01, 
    },
    {
      -0.20673763E+03, 0.37566130E+01, 0.14237963E+02, -0.50470264E+02, -0.52279778E+02, 0.21282154E+02, 0.11163842E+02, 0.52002543E+00, 0.55029087E+01, 0.20240622E+02, -0.74278593E+01, -0.33429351E+01, -0.35941296E+01, 
    },
    {
      -0.11962262E+03, -0.10679832E+03, -0.54719863E+01, 0.29441711E+02, -0.20143695E+01, -0.27501472E+02, -0.26564098E+02, -0.25437061E+02, -0.56287434E+02, 0.34835873E+01, -0.41210556E+01, -0.22583303E+01, -0.41311565E+00, 
    },
    {
      0.69616798E+02, 0.29591665E+01, -0.46051245E+01, 0.21710459E+02, 0.20603981E+02, -0.72812696E+01, -0.51570253E+01, -0.75040758E+00, -0.23034899E+01, -0.73399339E+01, 0.31145618E+01, 0.10590287E+01, 0.16675494E+01, 
    },
    {
      0.55920959E+02, 0.44685562E+02, -0.57169757E+01, -0.16177469E+02, -0.28279176E+01, 0.99622288E+01, 0.10504270E+02, 0.95728416E+01, 0.21457167E+02, -0.10309284E+01, 0.88172013E+00, 0.53643727E+00, -0.38652459E+00, 
    },
    {
      -0.91051519E-01, 0.16604546E+01, 0.15699625E+01, 0.76715745E-01, 0.43537375E-01, -0.13598571E-01, 0.49196340E-01, 0.54983906E-02, -0.43234859E-01, 0.35610504E-01, 0.33476464E-01, -0.23877524E-01, -0.35067947E-03, 
    },
    {
      0.12912868E+00, -0.15785012E+01, 0.15362040E+01, -0.14700162E+00, 0.59193391E-01, -0.35832047E-01, -0.21853298E-01, 0.66302014E-02, 0.16508466E-01, 0.10124203E-01, 0.11756688E-01, -0.16161023E-01, -0.59979502E-02, 
    },
    {
      -0.84079015E+00, 0.12545848E+01, -0.53319615E+00, -0.38466421E+00, 0.71295905E+00, 0.44927454E+00, 0.46575686E+00, 0.96385442E-01, 0.24827370E+00, -0.49310956E-01, -0.60446393E-01, -0.82794242E-01, 0.53030346E-01, 
    },
    {
      0.93256634E+00, -0.20340858E+00, 0.24310616E+00, -0.10172791E+00, 0.75127995E+00, 0.97861588E-01, 0.30854341E+00, 0.39208809E+00, 0.25966071E-01, -0.73293775E-01, -0.10588093E+00, 0.91211334E-01, -0.16803257E+00, 
    },
    {
      0.31397121E+01, 0.17180098E+02, -0.13800441E+02, 0.70740975E-01, 0.46341306E+00, 0.14315274E+00, -0.44296429E-01, 0.59252042E+00, 0.15348610E+01, -0.11630143E+01, -0.22830707E+00, 0.94050133E+00, -0.27207699E+00, 
    },
    {
      -0.83226357E+01, 0.16027706E+02, 0.18227182E+02, 0.24127638E+01, 0.16124903E+01, 0.64143932E+00, -0.48962837E+00, 0.14949485E+01, -0.68322599E+00, -0.83393383E+00, -0.21246544E+00, 0.36369675E+00, 0.15471225E-01, 
    },
    {
      0.25553751E+02, -0.20131190E+01, 0.26365784E+02, -0.37747600E+01, -0.64699745E+01, -0.78558393E+01, -0.10889877E+02, -0.23848505E+00, -0.30194924E+01, 0.15077212E+01, 0.17867323E+01, 0.61363947E+00, -0.83897740E+00, 
    },
    {
      -0.97228460E+01, -0.83611345E+01, 0.63290434E+01, 0.16351886E+01, -0.20032827E+02, 0.44843369E+01, -0.30204687E+01, -0.67688684E+01, 0.79144502E+00, 0.25467318E+00, 0.99470073E+00, -0.17565970E+01, 0.14626296E+01, 
    },
    {
      -0.19355894E+02, -0.47454216E+02, 0.12098093E+03, -0.89022665E+01, -0.42745748E+01, -0.45515871E+01, 0.30617993E-01, -0.31281152E+01, -0.12787308E+02, 0.67629743E+01, 0.81136113E+00, -0.72239995E+01, 0.21611481E+01, 
    },
    {
      0.77719887E+02, -0.15284087E+03, -0.28094072E+02, -0.12935537E+02, -0.18593321E+02, -0.32531598E+01, 0.55940328E+01, -0.11318358E+02, 0.32409825E+01, 0.75526614E+01, 0.48537755E+00, -0.15223314E+01, 0.88688987E+00, 
    },
    {
      -0.16509648E+03, -0.13600343E+02, -0.14347475E+03, 0.67285439E+02, 0.29329416E+02, 0.45942951E+02, 0.55822559E+02, -0.64459381E+01, 0.13185073E+02, -0.11480048E+02, -0.14453205E+02, -0.30418870E+00, 0.47256918E+01, 
    },
    {
      0.55850414E+02, 0.47293289E+02, -0.38787140E+02, -0.14584154E+02, 0.12430083E+03, -0.31622999E+02, 0.13420869E+02, 0.35762138E+02, -0.36125880E+00, -0.15479681E+01, -0.12744308E+01, 0.86030378E+01, -0.44953423E+01, 
    },
    {
      0.38700024E+02, -0.28222193E+02, -0.30419247E+03, 0.39386944E+02, 0.15715430E+02, 0.20906570E+02, -0.27541671E+01, 0.65622759E+01, 0.40180477E+02, -0.16013824E+02, -0.15492724E+01, 0.20418951E+02, -0.60262113E+01, 
    },
    {
      -0.24083032E+03, 0.41435196E+03, -0.13463223E+03, 0.38008816E+02, 0.62823856E+02, 0.78923759E+01, -0.17364424E+02, 0.29752794E+02, -0.36583562E+01, -0.22033194E+02, -0.86091971E+00, 0.23167665E+01, -0.40712471E+01, 
    },
    {
      0.43789868E+03, 0.37758240E+02, 0.30222437E+03, -0.25211705E+03, -0.68283501E+02, -0.12247941E+03, -0.11544855E+03, 0.29691120E+02, -0.26862541E+02, 0.32532574E+02, 0.40526028E+02, -0.55368447E+01, -0.12097772E+02, 
    },
    {
      -0.16987953E+03, -0.98175888E+02, 0.93504860E+02, 0.44424995E+02, -0.31534625E+03, 0.75678284E+02, -0.27929533E+02, -0.75439156E+02, -0.14984094E+02, 0.86548948E+01, -0.38562276E+01, -0.19092201E+02, 0.82323360E+01, 
    },
    {
      -0.28264292E+02, 0.16086139E+03, 0.33062650E+03, -0.60156818E+02, -0.25561420E+02, -0.33909615E+02, 0.69856234E+01, -0.67728324E+01, -0.52560555E+02, 0.17889961E+02, 0.16138614E+01, -0.24223919E+02, 0.70134716E+01, 
    },
    {
      0.30368719E+03, -0.46959631E+03, 0.33666513E+03, -0.55255219E+02, -0.86959816E+02, -0.90647640E+01, 0.22722628E+02, -0.31628498E+02, -0.20313892E+01, 0.26141891E+02, 0.15387295E+01, -0.10764118E+01, 0.55649915E+01, 
    },
    {
      -0.51269446E+03, -0.24774500E+02, -0.26924051E+03, 0.35456693E+03, 0.72959488E+02, 0.14845412E+03, 0.10644101E+03, -0.44661758E+02, 0.25646076E+02, -0.38323627E+02, -0.45829033E+02, 0.11802900E+02, 0.13949016E+02, 
    },
    {
      0.23760629E+03, 0.85215302E+02, -0.10850819E+03, -0.47027061E+02, 0.35732697E+03, -0.74357651E+02, 0.24879040E+02, 0.67221268E+02, 0.34626274E+02, -0.15760853E+02, 0.95299788E+01, 0.20570358E+02, -0.99656239E+01, 
    },
    {
      0.57932053E+01, -0.10340132E+03, -0.13253651E+03, 0.30239813E+02, 0.14212955E+02, 0.18223082E+02, -0.44180136E+01, 0.29476216E+01, 0.24200890E+02, -0.79207344E+01, -0.70972711E+00, 0.10236231E+02, -0.29160755E+01, 
    },
    {
      -0.13413786E+03, 0.19037309E+03, -0.19506718E+03, 0.28727461E+02, 0.42836147E+02, 0.36783912E+01, -0.11200485E+02, 0.11366664E+02, 0.33454440E+01, -0.10977008E+02, -0.10146281E+01, -0.15341072E+00, -0.23435018E+01, 
    },
    {
      0.21834229E+03, -0.54074174E+00, 0.81919060E+02, -0.17072711E+03, -0.27865332E+02, -0.65729042E+02, -0.36894901E+02, 0.22338470E+02, -0.91411514E+01, 0.15863236E+02, 0.18038870E+02, -0.67405629E+01, -0.59032989E+01, 
    },
    {
      -0.11627517E+03, -0.24264631E+02, 0.48176723E+02, 0.13980582E+02, -0.15089851E+03, 0.25507414E+02, -0.73551226E+01, -0.20702442E+02, -0.20904865E+02, 0.87487831E+01, -0.55933886E+01, -0.86189528E+01, 0.53419061E+01, 
    },
    {
      0.78886345E-01, -0.29637566E+00, -0.60464900E-01, -0.71206862E+00, 0.59372735E+00, -0.20558244E-01, 0.48958015E-01, 0.17571801E-01, 0.42676944E-02, -0.61578038E-02, 0.21608399E-01, 0.51746960E-03, -0.57962183E-02, 
    },
    {
      -0.11099774E-01, -0.77226818E-01, -0.35840608E-01, -0.68324232E+00, -0.71531457E+00, -0.41512378E-01, -0.81707286E-02, -0.76341666E-02, 0.20536058E-01, 0.75761302E-04, 0.10275762E-01, -0.93738467E-03, 0.10511260E-02, 
    },
    {
      0.13842398E+01, -0.49234234E-01, 0.12677437E-01, -0.10571032E+01, 0.10398120E+00, 0.59301776E+00, -0.14417675E+00, -0.12642118E+00, -0.50918173E-01, -0.13518077E+00, 0.30521570E-01, -0.12413521E+00, -0.55086080E-01, 
    },
    {
      0.13468280E+01, 0.45321479E+00, 0.33050817E-01, -0.19486007E+00, -0.84212863E+00, 0.33315787E+00, 0.56156081E+00, -0.21251965E+00, 0.18369167E+00, -0.45716692E-01, -0.15525161E+00, -0.21887375E-02, -0.13814416E-01, 
    },
    {
      0.13547544E+01, 0.35052302E+01, 0.39791517E+01, -0.39452157E+01, 0.91488066E+01, -0.59917301E+00, -0.23404021E+01, 0.19902436E+00, 0.28914595E+00, 0.32449687E+00, -0.30973130E+00, -0.13093148E+00, 0.24543998E+00, 
    },
    {
      0.93270177E+00, 0.10100644E+01, 0.21535051E+01, -0.79689913E+01, -0.27236195E+01, 0.71202469E+00, -0.46545181E+00, 0.11434031E+01, -0.18245190E+00, -0.40822005E+00, -0.56782430E+00, -0.14069527E-01, -0.91296121E-01, 
    },
    {
      -0.10542483E+02, 0.11633001E+00, -0.27338638E+01, 0.60308566E+01, 0.38688369E+01, -0.47813067E+01, -0.80322623E-01, 0.11000689E+01, 0.13760308E+01, 0.10471877E+01, -0.10195704E+00, 0.84878850E+00, 0.67848611E+00, 
    },
    {
      -0.11052456E+02, -0.35547700E+01, -0.17437516E+01, -0.14164073E+01, 0.29000556E+01, -0.17477272E+00, -0.31872108E+01, 0.15479761E+01, -0.11199074E+01, 0.35770193E-01, 0.15186522E+01, -0.51870143E+00, 0.42309412E+00, 
    },
    {
      -0.12578057E+02, -0.68359613E+01, -0.30894140E+02, 0.29231524E+02, -0.37497784E+02, 0.42103472E+01, 0.13634277E+02, -0.23105447E+01, -0.25411386E+01, -0.20200241E+01, 0.17991817E+01, 0.12031139E+01, -0.22861779E+01, 
    },
    {
      -0.45937929E+01, 0.20954571E+01, -0.12322830E+02, 0.30053967E+02, 0.19175726E+02, -0.63447361E+01, 0.20542054E+01, -0.71390734E+01, -0.82132196E+00, 0.34279492E+01, 0.34167354E+01, 0.95250791E+00, -0.24310125E+00, 
    },
    {
      0.24285011E+02, -0.33831913E+01, 0.87184391E+01, -0.14099850E+02, -0.16883636E+02, 0.11646293E+02, 0.59411901E+00, -0.24929807E+01, -0.56281371E+01, -0.20639105E+01, 0.36040890E+00, -0.18116419E+01, -0.17643514E+01, 
    },
    {
      0.24702211E+02, 0.38919735E+01, 0.61029816E+01, 0.65146122E+01, -0.22756846E+01, -0.12218447E+01, 0.49939814E+01, -0.24445646E+01, 0.28680899E+01, -0.10368446E+00, -0.39225392E+01, 0.16193440E+01, -0.12233027E+01, 
    },
    {
      0.28191441E+02, 0.78694773E+00, 0.64988586E+02, -0.59622383E+02, 0.50546680E+02, -0.93377047E+01, -0.25691008E+02, 0.49692044E+01, 0.66733646E+01, 0.38342435E+01, -0.33193257E+01, -0.28077452E+01, 0.59086132E+01, 
    },
    {
      0.68308530E+01, -0.11145398E+02, 0.20142715E+02, -0.35770203E+02, -0.34637836E+02, 0.14650714E+02, -0.28380070E+01, 0.14030188E+02, 0.30224795E+01, -0.93840456E+01, -0.67304063E+01, -0.30187075E+01, 0.13366692E+01, 
    },
    {
      -0.19719049E+02, 0.68867254E+01, -0.97942524E+01, 0.12571602E+02, 0.16866741E+02, -0.80991716E+01, -0.60612911E+00, 0.14781255E+01, 0.55502706E+01, 0.12907434E+01, -0.25038749E+00, 0.10915368E+01, 0.11269934E+01, 
    },
    {
      -0.16954172E+02, 0.31211653E+01, -0.45404301E+01, -0.60521469E+01, 0.81650001E+00, 0.11353194E+01, -0.19928211E+01, 0.10430180E+01, -0.25809250E+01, 0.46962984E-01, 0.30409884E+01, -0.11915781E+01, 0.96175605E+00, 
    },
    {
      -0.18254309E+02, 0.40652580E+01, -0.38976635E+02, 0.37998058E+02, -0.22253872E+02, 0.54544687E+01, 0.15347375E+02, -0.28659797E+01, -0.52428341E+01, -0.22536197E+01, 0.18733271E+01, 0.18808317E+01, -0.43471789E+01, 
    },
    {
      -0.42598801E+01, 0.68740029E+01, -0.86680889E+01, 0.13091699E+02, 0.19776112E+02, -0.94230890E+01, 0.93041652E+00, -0.87512522E+01, -0.19371971E+01, 0.73051834E+01, 0.41424298E+01, 0.23743393E+01, -0.12347583E+01, 
    },
    {
      -0.10308004E+00, -0.11345814E+00, -0.42388607E-01, -0.54478396E-01, 0.37894849E-01, -0.71249014E+00, -0.99690706E+00, 0.42135198E-01, -0.17244507E-01, 0.18175337E-01, 0.22615835E-01, 0.40037422E-02, -0.35507255E-02, 
    },
    {
      -0.27820380E-01, -0.77944390E-01, 0.14688867E+00, 0.16636178E-01, 0.53211000E-01, 0.95628875E+00, -0.70051461E+00, 0.17104762E-01, 0.31993400E-01, -0.93262866E-02, -0.18731829E-01, 0.20413624E-01, -0.50872733E-03, 
    },
    {
      0.29756284E+00, 0.12934245E+00, -0.18983012E+00, 0.71824878E-01, -0.41808215E+00, -0.11720640E+00, -0.41143602E+00, -0.28631784E-01, 0.21437567E+00, 0.25426921E-01, 0.33951670E-01, 0.75097620E-01, -0.22934020E-01, 
    },
    {
      0.60831457E-01, 0.26703396E+00, 0.11492737E+00, -0.14872836E+00, 0.15820681E+00, 0.73452604E+00, -0.15326773E+00, -0.20186380E+00, 0.79028957E-01, -0.39597573E-02, -0.69052177E-02, -0.35170142E-02, 0.39248485E-01, 
    },
    {
      0.77993286E+00, 0.11083744E+01, 0.43744072E+00, 0.59175074E+00, 0.27468514E+00, 0.27732401E+01, 0.21687400E+01, -0.25144777E+00, 0.20495271E-01, -0.25761271E-01, 0.13951994E-02, -0.70093565E-01, 0.10886874E+00, 
    },
    {
      -0.10175124E+01, 0.14403582E+01, -0.19194679E+01, 0.23482410E-01, -0.27122623E+00, -0.25169380E+01, 0.30849991E+01, -0.93018368E-01, -0.32557723E+00, -0.32845687E-01, 0.71069151E-01, -0.18071032E+00, -0.11195786E+00, 
    },
    {
      -0.68460888E+00, 0.29701370E+00, 0.60023886E+00, 0.74086197E-01, 0.98181754E+00, 0.17232137E-01, 0.71498251E+00, 0.15950713E+00, -0.32126063E+00, -0.25459521E-01, 0.10034926E+00, -0.21438281E+00, 0.80607370E-01, 
    },
    {
      0.29350964E-01, -0.52844387E+00, -0.36183974E+00, -0.28149527E+00, 0.92464551E-01, -0.15573705E+01, -0.51209643E-01, 0.32866797E+00, -0.38855325E-01, 0.12696608E-01, 0.45670737E-01, -0.69922097E-02, -0.22250839E-01, 
    },
    {
      -0.49448678E+00, -0.11309274E+01, -0.43056571E+00, -0.97913963E+00, -0.75081772E+00, -0.50914221E+01, -0.24884198E+01, 0.29986811E+00, 0.74715838E-02, -0.13069284E-01, -0.12120257E+00, 0.20640135E+00, -0.19321387E+00, 
    },
    {
      0.16908376E+01, -0.17448000E+01, 0.30932491E+01, -0.34096658E+00, 0.17136328E-01, 0.35928824E+01, -0.54552412E+01, 0.32137137E-01, 0.51082975E+00, 0.19546562E+00, -0.10672491E+00, 0.28139511E+00, 0.41073278E-01, 
    },
    {
      -0.29096520E-01, 0.21864381E-01, 0.12243150E-01, 0.32921236E-01, 0.26921557E-01, -0.18893564E-01, 0.16802772E-02, 0.28694057E+00, -0.29211941E+00, 0.16393879E-01, 0.94393184E-02, -0.60703461E-02, -0.10121063E-02, 
    },
    {
      0.22870019E-01, 0.40416889E-01, 0.20207793E-02, 0.28009053E-01, -0.47511235E-01, 0.16151302E-01, 0.41487203E-02, 0.24980807E+00, 0.29251510E+00, -0.41429587E-02, 0.31435888E-02, 0.86209821E-02, -0.10191371E-01, 
    },
    {
      -0.31774394E-01, 0.34211457E-01, -0.82030557E-01, 0.80120198E-01, 0.52378416E-01, -0.12363158E+00, 0.10513846E+00, 0.79557329E-01, -0.65318130E-01, -0.10646767E+00, 0.76470129E-01, 0.25804231E-02, 0.20964760E-01, 
    },
    {
      0.12791419E+00, 0.45520272E-02, 0.28757790E-01, 0.62985376E-01, 0.84955081E-01, -0.12073985E+00, -0.98188937E-01, 0.52158490E-01, 0.69911316E-01, -0.42228647E-01, -0.41942727E-01, -0.53663673E-04, 0.20629911E-01, 
    },
    {
      -0.84972277E-01, 0.74791508E-02, -0.72876997E-01, 0.54825362E-01, 0.78998553E-02, 0.21651424E-02, 0.11295615E-01, -0.22708066E-01, -0.24967110E-01, 0.23709922E+00, 0.20721011E+00, -0.75820675E-02, -0.66711716E-02, 
    },
    {
      0.91338418E-01, 0.75325817E-01, -0.11707541E-01, -0.18225781E-02, -0.68179360E-02, 0.26803995E-01, 0.29882748E-01, -0.50773583E-02, -0.38290662E-02, -0.20912644E+00, 0.21416292E+00, -0.53014228E-03, 0.60032657E-02, 
    },
    {
      0.42857230E-01, -0.81398398E-01, -0.70250421E-02, 0.28814852E-01, 0.10183385E-01, 0.45627076E-01, 0.34210838E-01, -0.13469204E-01, -0.16962208E-01, 0.95584551E-02, 0.76674414E-02, -0.11811175E+00, 0.10944934E+00, 
    },
    {
      0.37916854E-01, 0.34320731E-01, -0.38995236E-01, -0.11582067E-01, 0.56862938E-02, 0.19178797E-01, -0.19646440E-01, -0.19834880E-02, -0.70472504E-02, -0.87856855E-02, -0.38532328E-02, -0.12558445E+00, -0.99743478E-01, 
    },
    {
      -0.44582158E-01, -0.84324472E-01, -0.50034769E-01, 0.46942960E-01, 0.46193022E-02, 0.71702013E-02, -0.12611548E-01, 0.95665868E-03, -0.12036865E-01, -0.67090063E-04, -0.20105047E-01, 0.25720501E-02, 0.19281610E-02, 
    },
    {
      -0.26651995E-01, 0.87436706E-01, 0.20321304E-01, -0.69048993E-01, 0.10091052E-02, 0.10883666E-01, -0.17108636E-01, -0.87443851E-02, 0.57800952E-02, 0.46012336E-02, 0.20599568E-01, -0.53997170E-02, 0.58125644E-02, 
    },
    {
      -0.57532750E-02, 0.34642294E-01, 0.31471007E-01, -0.11714598E-01, -0.16319931E-01, -0.50641306E-01, 0.25461499E-01, 0.27247524E-01, 0.11968782E-01, 0.18001130E-01, -0.22403058E-01, 0.21385524E-01, 0.13279283E-01, 
    },
    {
      -0.25084522E-01, 0.21860242E-01, 0.28222946E-02, 0.35862558E-01, 0.60812447E-01, 0.43446873E-02, 0.19660845E-01, 0.30596415E-01, -0.88513829E-02, -0.24848016E-01, -0.96459324E-02, -0.90757394E-02, -0.20335526E-02, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir13_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.29977396E+01, -0.81961853E-02, 0.18906312E-01, 0.31932625E-02, -0.40838039E-02, -0.23357221E-03, 0.76571512E-02, -0.79752160E-02, 0.34400288E-03, 
    },
    {
      0.20014945E+00, -0.29716199E-01, 0.52193105E-01, 0.37738375E-01, -0.50039222E-02, -0.29981710E-01, -0.13986629E-01, 0.22926540E-02, -0.21308563E-02, 
    },
    {
      0.13428990E+01, 0.44498194E-01, -0.18523586E+00, -0.17925259E+00, 0.45948785E-01, 0.47860365E-01, -0.33176940E-01, 0.42967819E-01, 0.43051202E-01, 
    },
    {
      -0.47216845E+00, -0.16368178E+00, -0.10566967E+00, -0.59037436E-01, -0.33112377E-01, 0.11583864E+00, 0.14226819E-01, 0.78643629E-04, 0.69223372E-02, 
    },
    {
      -0.23492818E+01, 0.67165658E-01, 0.47879127E+00, 0.52358502E+00, -0.15943331E+00, -0.14764257E+00, 0.48829701E-01, -0.51714528E-01, -0.10612427E+00, 
    },
    {
      0.24342524E+00, 0.16540094E+00, 0.64394549E-01, 0.85904840E-02, 0.40568676E-01, -0.88596366E-01, -0.20141602E-02, -0.10498879E-02, -0.26079479E-02, 
    },
    {
      0.11590286E+01, -0.87880895E-01, -0.30248937E+00, -0.36426729E+00, 0.13489676E+00, 0.10174665E+00, -0.20617876E-01, 0.17740101E-01, 0.64807892E-01, 
    },
    {
      0.14391306E-01, -0.21309042E+00, -0.33884123E+00, 0.10892765E-01, -0.42811530E-02, -0.25248926E-02, 0.12565411E-01, -0.44771405E-02, 0.69199535E-02, 
    },
    {
      0.26753715E-02, 0.35680091E+00, -0.20014954E+00, -0.98086214E-02, 0.87794177E-02, -0.34690208E-02, -0.91818944E-02, 0.56939302E-02, 0.10811757E-01, 
    },
    {
      -0.10302394E+00, 0.19964647E+00, 0.12284161E+00, 0.62227905E-01, 0.39830696E-01, -0.31292230E-01, -0.18277895E-01, 0.34071945E-01, -0.24559934E-01, 
    },
    {
      0.73183656E-01, -0.31019021E-01, 0.22225291E+00, -0.79470731E-01, -0.39866902E-01, 0.20936769E-01, -0.13295945E-01, 0.57061128E-01, 0.55259675E-01, 
    },
    {
      -0.22905074E+00, 0.90782636E+00, 0.16330606E+01, 0.19163447E-02, 0.35776563E-01, 0.49570456E-01, -0.11902388E+00, 0.45402091E-01, -0.11314365E+00, 
    },
    {
      -0.39456546E+00, -0.17365061E+01, 0.86957514E+00, 0.16775037E+00, -0.85004143E-01, 0.42578947E-01, 0.72186656E-01, -0.78504875E-01, -0.12058398E+00, 
    },
    {
      0.13912698E+01, -0.96152496E+00, -0.60672927E+00, -0.13106681E-01, -0.41902941E+00, 0.11963305E+00, 0.62974300E-02, -0.18483321E+00, 0.25544432E+00, 
    },
    {
      0.54170632E+00, 0.34240028E+00, -0.99347818E+00, 0.24775450E+00, 0.47574925E+00, -0.18045484E+00, 0.13595828E+00, -0.61139083E+00, -0.26010543E+00, 
    },
    {
      0.74438137E+00, -0.80979353E+00, -0.17163999E+01, -0.35758594E+00, -0.14151997E-01, -0.13134675E+00, 0.27793384E+00, -0.11998469E+00, 0.27071473E+00, 
    },
    {
      0.13506950E+01, 0.16267354E+01, -0.12598553E+01, -0.31339702E+00, 0.68517044E-01, -0.18701576E+00, -0.86923413E-01, 0.14469229E+00, 0.22702868E+00, 
    },
    {
      -0.34414785E+01, 0.10698559E+01, 0.14466070E+01, -0.78836642E-01, 0.10688086E+01, -0.78987226E-01, 0.20188455E+00, 0.36264941E+00, -0.72390538E+00, 
    },
    {
      -0.21533570E+01, -0.86708015E+00, 0.16319656E+01, -0.28797507E+00, -0.10868007E+01, 0.50646031E+00, -0.39900690E+00, 0.15120949E+01, 0.29860815E+00, 
    },
    {
      -0.48577079E+00, 0.13015094E+00, 0.95722884E+00, 0.44391346E+00, -0.32366667E-01, 0.12413590E+00, -0.17610566E+00, 0.96929386E-01, -0.16703936E+00, 
    },
    {
      -0.11894507E+01, -0.38839000E+00, 0.71710134E+00, 0.12543808E+00, 0.23892194E-01, 0.17996967E+00, -0.33655792E-01, -0.64367644E-01, -0.11553254E+00, 
    },
    {
      0.22641020E+01, -0.19882113E-01, -0.99476331E+00, -0.88516802E-01, -0.73695379E+00, 0.47051879E-02, -0.22439910E+00, -0.23656236E+00, 0.55085856E+00, 
    },
    {
      0.17834074E+01, 0.62063682E+00, -0.85358530E+00, 0.12144197E+00, 0.67288089E+00, -0.39924723E+00, 0.35274625E+00, -0.10284065E+01, -0.55081531E-01, 
    },
    {
      -0.42461534E-02, -0.19090576E-01, -0.21750687E-02, 0.77972598E-01, -0.10230899E+00, -0.61414940E-02, -0.60802395E-02, -0.60765678E-02, -0.31842440E-02, 
    },
    {
      0.36602344E-02, -0.20476623E-01, 0.19559988E-02, 0.12126583E+00, 0.80235869E-01, -0.91823516E-02, -0.76314104E-02, 0.35730617E-02, -0.18788265E-02, 
    },
    {
      0.73799282E-01, 0.18658288E+00, 0.25844129E-01, 0.38106717E-01, 0.13187513E-01, 0.33492066E-01, -0.36260985E-01, -0.13196414E-01, -0.99324873E-02, 
    },
    {
      0.14192873E-01, -0.25389520E-01, -0.18034412E-01, 0.66986732E-01, 0.37657667E-01, 0.83836555E-01, -0.19943716E-01, -0.24610309E-01, 0.22301378E-01, 
    },
    {
      -0.17031059E+00, 0.24293567E+00, -0.47562119E-01, -0.18710293E+00, -0.21173187E+00, 0.45424961E-01, 0.10562468E+00, 0.11877582E-01, 0.27717501E-01, 
    },
    {
      -0.23334758E+00, 0.18515380E-01, -0.18031678E+00, 0.16339774E+00, -0.16562817E+00, 0.22401256E-01, 0.12215625E+00, -0.20599376E-01, -0.22505935E-01, 
    },
    {
      -0.23864314E+00, -0.10410759E+01, 0.20875806E+00, 0.41495808E-02, -0.14400753E+00, -0.13934614E-01, 0.98016091E-01, 0.64151272E-01, -0.90540049E-03, 
    },
    {
      0.43185464E+00, 0.47100690E+00, 0.23188816E+00, -0.12820059E+00, 0.30625871E+00, -0.37186635E+00, 0.27746350E-01, 0.18685450E+00, -0.14788827E+00, 
    },
    {
      0.31099787E+00, -0.51439357E+00, 0.33028484E-02, -0.17644372E-01, 0.89867815E-01, -0.63249975E-01, -0.14556959E+00, -0.11969471E-01, -0.39798412E-01, 
    },
    {
      0.37246844E+00, -0.99704146E-01, 0.28348404E+00, -0.27310552E-01, -0.17952356E-02, 0.30666633E-01, -0.20847893E+00, -0.28230352E-02, 0.19477863E-01, 
    },
    {
      0.20202787E+00, 0.14707644E+01, -0.31790829E+00, -0.87093808E-01, 0.18147080E+00, -0.21470476E-01, -0.88236138E-01, -0.84560640E-01, 0.35590544E-01, 
    },
    {
      -0.68945140E+00, -0.61575735E+00, -0.83507515E-01, 0.61113432E-01, -0.57205093E+00, 0.39900196E+00, 0.70120350E-01, -0.22128372E+00, 0.17749649E+00, 
    },
    {
      -0.23316426E-01, 0.12048603E-01, 0.23783310E-02, 0.92227728E-03, 0.47883224E-02, 0.22830907E-01, -0.44083364E-01, -0.45512458E-02, 0.17387259E-02, 
    },
    {
      -0.70657874E-02, -0.15885700E-01, -0.40105851E-02, -0.69217128E-02, -0.58886250E-02, 0.35231218E-01, 0.42952802E-01, 0.33667521E-02, -0.20392444E-02, 
    },
    {
      0.59208658E-01, 0.63783467E-01, 0.12437732E-01, -0.64228050E-01, 0.13783633E-01, -0.51416833E-01, 0.18383345E-01, 0.69502792E-02, 0.81083290E-02, 
    },
    {
      0.40749948E-01, -0.75332168E-02, -0.27416045E-01, -0.33444658E-01, -0.19954119E-01, 0.60441419E-02, -0.56821901E-01, 0.11292337E-02, 0.17509496E-01, 
    },
    {
      -0.67723908E-01, -0.14539140E+00, -0.30557400E-01, 0.45087757E-02, -0.46182219E-01, -0.15280160E+00, -0.10240623E+00, 0.26795991E-01, -0.21112165E-02, 
    },
    {
      -0.47966205E-01, 0.10730190E-01, -0.31358361E-01, 0.15586140E-01, -0.61677694E-02, 0.13321459E+00, -0.20194066E+00, 0.10347781E-01, -0.12390453E-01, 
    },
    {
      -0.18478956E-01, -0.15848843E-01, 0.85952645E-02, 0.16202679E-01, 0.12485572E-02, -0.19915239E-02, 0.31446822E-02, 0.44241477E-01, -0.30390822E-03, 
    },
    {
      -0.94869835E-02, 0.91496035E-02, -0.60107494E-02, -0.70505976E-02, -0.37957488E-02, 0.63174270E-03, 0.55268044E-02, -0.33853750E-03, 0.48207454E-01, 
    },
    {
      0.65101489E-01, 0.29953860E-01, -0.47032386E-02, -0.26011158E-01, -0.31840345E-02, 0.30755354E-02, 0.13277406E-02, 0.22132441E-01, 0.82672704E-02, 
    },
    {
      -0.78570237E-02, -0.45788117E-01, 0.51256320E-02, 0.33361401E-01, -0.11823519E-01, -0.16863131E-02, -0.15676618E-01, 0.17375020E-01, 0.15721165E-01, 
    },
    {
      0.64684800E-02, -0.96570859E-02, 0.89065656E-02, -0.50537391E-02, -0.49813055E-02, 0.20753632E-02, 0.76234308E-02, -0.90143485E-02, 0.38806573E-02, 
    },
    {
      0.70466385E-02, 0.23270652E-01, -0.98308958E-02, -0.23814661E-02, -0.12868521E-02, 0.14961665E-01, 0.29478413E-02, -0.25250819E-02, 0.35902048E-02, 
    },
    {
      0.68825735E-02, 0.69126990E-02, 0.12519725E-01, 0.12947415E-01, -0.24092742E-02, 0.11634666E-02, -0.38903584E-02, -0.92853449E-03, 0.12939579E-02, 
    },
    {
      -0.27606012E-02, -0.23899183E-02, -0.13654518E-01, -0.14725209E-02, 0.18952837E-02, 0.95515465E-02, 0.27220524E-02, -0.41225250E-02, 0.64941798E-02, 
    },
  },
  {
    {
      0.26891699E+01, -0.77640088E-02, 0.11127225E-01, 0.56581018E-02, 0.55255590E-03, 0.19188958E-02, 0.38992222E-02, -0.63712653E-02, -0.10736038E-02, 
    },
    {
      0.89743264E-01, 0.20814057E-01, 0.49664252E-01, 0.12002765E-01, 0.16872061E-01, -0.14828340E-01, -0.20098081E-01, -0.71201292E-02, -0.10063166E-01, 
    },
    {
      0.14451987E+01, 0.18207384E-01, -0.12469007E+00, -0.77557176E-01, 0.96017895E-02, 0.50489246E-02, 0.77703637E-02, 0.43605078E-01, 0.20670578E-01, 
    },
    {
      -0.10420049E+00, -0.30086073E+00, -0.98388202E-01, -0.17734094E-01, -0.10223059E+00, 0.44458374E-01, 0.44485763E-01, 0.28927306E-01, 0.21576228E-01, 
    },
    {
      -0.26058183E+01, 0.11795595E+00, 0.28667596E+00, 0.16859519E+00, -0.69919094E-01, -0.51369867E-02, -0.43885093E-01, -0.75206392E-01, -0.15287774E-01, 
    },
    {
      0.39084069E-02, 0.23944090E+00, 0.64139619E-01, 0.46739038E-02, 0.91723315E-01, -0.28238665E-01, -0.27856352E-01, -0.21242142E-01, -0.13344122E-01, 
    },
    {
      0.13095468E+01, -0.10214921E+00, -0.16770977E+00, -0.98850533E-01, 0.73869824E-01, -0.47578220E-02, 0.37430078E-01, 0.40278375E-01, -0.92505598E-02, 
    },
    {
      0.23524527E-01, -0.34971893E+00, -0.28617576E+00, 0.24448626E-02, 0.10531447E-02, -0.24429685E-02, 0.79786703E-02, -0.32044223E-02, 0.65662437E-02, 
    },
    {
      -0.49985661E-02, 0.30120927E+00, -0.33876929E+00, -0.98431129E-02, -0.94157760E-03, -0.78012803E-04, -0.10437771E-01, 0.98567735E-02, 0.11818309E-01, 
    },
    {
      -0.57822518E-01, 0.21672082E+00, 0.88802382E-01, -0.38946371E-02, 0.58250647E-01, -0.24885843E-01, -0.17417107E-01, 0.15775116E-01, -0.64615905E-02, 
    },
    {
      0.32139458E-01, 0.32517146E-01, 0.21158411E+00, -0.15073596E-01, -0.36884345E-01, 0.46076268E-01, 0.17812065E-02, 0.35551559E-01, 0.46800584E-01, 
    },
    {
      -0.24651740E+00, 0.11098006E+01, 0.93898606E+00, 0.13167660E+00, 0.78329649E-02, 0.21393843E-01, 0.26152806E-01, 0.99524148E-02, -0.16420493E+00, 
    },
    {
      -0.35180086E+00, -0.94679403E+00, 0.12385637E+01, 0.81863344E-01, 0.52147958E-01, -0.94125001E-02, 0.86879626E-01, -0.13224545E+00, -0.17574912E+00, 
    },
    {
      0.77992594E+00, -0.13512278E+01, -0.66809833E+00, 0.17822575E-01, -0.50182998E+00, 0.64230546E-01, -0.65235958E-01, -0.17756595E+00, 0.55901699E-01, 
    },
    {
      0.60895658E+00, -0.39232608E-01, -0.12162886E+01, -0.19391639E-01, 0.24424776E+00, -0.32840198E+00, 0.16300706E-01, -0.27074844E+00, -0.22464836E+00, 
    },
    {
      0.55456519E+00, -0.69234842E+00, -0.40266684E+00, -0.63226157E+00, 0.10810311E-01, -0.15206470E-01, -0.15813012E+00, 0.76063015E-01, 0.54846871E+00, 
    },
    {
      0.12965285E+01, -0.14871565E+00, -0.16574056E+01, -0.77499501E-01, -0.24706507E+00, -0.49161386E-01, -0.18116425E+00, 0.29508817E+00, 0.42510369E+00, 
    },
    {
      -0.19151396E+01, 0.23017201E+01, 0.19343640E+01, 0.18132132E+00, 0.97290933E+00, -0.32005332E-01, 0.30257878E+00, 0.47717166E+00, -0.22451325E+00, 
    },
    {
      -0.23164320E+01, -0.23464796E-02, 0.23696494E+01, 0.11396227E+00, -0.30829358E+00, 0.81678474E+00, -0.14368920E+00, 0.57714278E+00, 0.30708337E+00, 
    },
    {
      -0.28491089E+00, 0.36383953E-01, 0.31162423E+00, 0.57425320E+00, -0.42527661E-01, 0.73678866E-02, 0.16507234E+00, -0.12021207E+00, -0.42351913E+00, 
    },
    {
      -0.11174608E+01, 0.66397589E+00, 0.92702854E+00, -0.17574133E-01, 0.21021613E+00, 0.75819165E-01, 0.74323684E-01, -0.16377261E+00, -0.30484450E+00, 
    },
    {
      0.11824274E+01, -0.11248287E+01, -0.14524229E+01, -0.21500514E+00, -0.54644227E+00, -0.13700272E-01, -0.25452733E+00, -0.35112619E+00, 0.19990946E+00, 
    },
    {
      0.20050185E+01, 0.13328118E-01, -0.14572992E+01, -0.51557049E-01, 0.12367892E+00, -0.57814145E+00, 0.17719761E+00, -0.37120983E+00, -0.99483021E-01, 
    },
    {
      -0.84307930E-03, -0.17258719E-01, -0.53407173E-02, 0.81290640E-01, -0.76485693E-01, -0.47752857E-02, 0.17251115E-02, -0.37566551E-02, -0.43733157E-02, 
    },
    {
      0.26733198E-02, -0.18344466E-01, 0.50127055E-02, 0.71141437E-01, 0.87314069E-01, -0.61744666E-02, -0.70299716E-02, 0.26243473E-02, -0.23496000E-02, 
    },
    {
      0.28444359E-01, 0.11505325E+00, -0.33933113E-02, 0.28973157E-02, 0.92887551E-01, 0.31396382E-01, -0.23507848E-01, -0.87130796E-02, -0.41673426E-02, 
    },
    {
      0.46873268E-01, 0.13442473E-01, -0.25793346E-01, -0.76118015E-01, 0.18148199E-01, 0.74972913E-01, 0.14298553E-01, -0.31826302E-01, 0.11931223E-01, 
    },
    {
      -0.10709218E+00, 0.17803708E+00, 0.89736991E-02, -0.42456055E+00, 0.21876013E+00, -0.10819510E-01, 0.28348537E-01, -0.17255733E-01, 0.18319869E-01, 
    },
    {
      -0.19474298E+00, 0.41860364E-01, -0.13339257E+00, -0.18719760E+00, -0.42420575E+00, -0.77197952E-02, 0.99332146E-01, -0.31550217E-01, -0.35813104E-01, 
    },
    {
      0.40004592E-01, -0.68339717E+00, 0.27681801E+00, -0.17132305E-01, -0.32585517E+00, 0.31075010E-02, 0.79180390E-01, 0.84248841E-01, -0.97170295E-02, 
    },
    {
      0.93430899E-01, 0.14520720E+00, 0.17805339E+00, 0.36324471E+00, 0.13691723E+00, -0.32484740E+00, -0.81565991E-01, 0.24830067E+00, -0.83268017E-01, 
    },
    {
      0.17182921E+00, -0.39017051E+00, -0.75088419E-01, 0.10350134E+00, -0.59314120E+00, 0.26216136E-01, -0.36645230E-01, 0.51128760E-01, -0.23116680E-01, 
    },
    {
      0.29115209E+00, -0.16628721E+00, 0.17287152E+00, 0.57455140E+00, 0.10316521E+00, 0.24566704E-01, -0.18611246E+00, 0.36977723E-01, 0.51306799E-01, 
    },
    {
      -0.20561792E+00, 0.10611655E+01, -0.37965357E+00, 0.59568016E-02, 0.27356151E+00, -0.27140494E-01, -0.93819186E-01, -0.12059696E+00, 0.53143971E-01, 
    },
    {
      -0.24612656E+00, -0.89114860E-01, -0.68045080E-01, -0.36101612E+00, -0.25655359E+00, 0.38696358E+00, 0.15423262E+00, -0.32276914E+00, 0.91858663E-01, 
    },
    {
      -0.19476928E-01, 0.73561035E-02, 0.17873573E-02, -0.35678830E-02, -0.33237985E-02, 0.33766452E-01, -0.63296571E-01, -0.19452671E-02, -0.13239883E-02, 
    },
    {
      -0.42364779E-02, -0.10886639E-01, 0.13287691E-02, 0.24953796E-03, -0.21379865E-02, 0.55756997E-01, 0.50980330E-01, 0.43777414E-02, -0.78862291E-02, 
    },
    {
      0.35293180E-01, 0.57927500E-01, 0.91379685E-02, -0.24990179E-01, 0.15874520E-01, -0.32756235E-01, -0.91940789E-02, 0.63448343E-02, 0.59108045E-02, 
    },
    {
      0.20577457E-01, 0.78460500E-02, -0.80252262E-02, -0.28437277E-01, -0.12757894E-01, 0.18312816E-01, -0.39697696E-01, -0.13474633E-02, 0.13109337E-01, 
    },
    {
      -0.83176987E-02, -0.11117948E+00, -0.57078397E-03, 0.10578160E-01, -0.20841746E-01, -0.17377385E+00, -0.81262551E-01, 0.16026536E-01, 0.58632647E-02, 
    },
    {
      -0.14602411E-01, -0.52710978E-03, -0.37079420E-01, 0.15373140E-01, 0.46844836E-03, 0.11484052E+00, -0.20478855E+00, -0.60416027E-02, 0.15950160E-01, 
    },
    {
      -0.92366859E-02, -0.13327929E-01, 0.76846951E-02, 0.44647725E-02, 0.99554425E-03, -0.25372533E-02, 0.11280873E-02, 0.34911953E-01, 0.61604744E-02, 
    },
    {
      -0.65696407E-02, 0.27554489E-02, -0.89025469E-02, -0.58459694E-03, -0.36023045E-02, 0.17695033E-02, 0.71463874E-02, -0.67619323E-02, 0.35612658E-01, 
    },
    {
      0.37786972E-01, 0.27549380E-01, -0.10036921E-01, -0.17617412E-01, -0.51135998E-02, -0.85136052E-02, 0.82170218E-02, 0.34347039E-01, 0.35131623E-02, 
    },
    {
      0.34137515E-02, -0.39485246E-01, 0.19134531E-01, 0.13404503E-01, -0.16808830E-01, -0.10705445E-01, -0.19780356E-01, 0.15510093E-01, 0.21839621E-01, 
    },
    {
      0.12911428E-01, -0.73543200E-02, 0.50285622E-02, 0.59204112E-03, -0.34981014E-02, 0.37478306E-02, 0.20511295E-02, -0.60793972E-02, 0.34043847E-02, 
    },
    {
      0.10783734E-01, 0.17769093E-01, -0.77495449E-02, 0.85491629E-03, -0.31315386E-02, 0.84124962E-02, 0.57312618E-02, 0.12188161E-02, 0.74360068E-02, 
    },
    {
      0.63715721E-02, 0.54682349E-02, 0.96778516E-02, 0.52643046E-02, -0.51632943E-02, 0.32400473E-02, -0.58748294E-03, -0.52953459E-03, 0.25913932E-02, 
    },
    {
      0.41622827E-02, -0.12601028E-02, -0.82769860E-02, 0.12908098E-02, -0.11074101E-02, 0.28636472E-02, 0.26873229E-02, -0.14917087E-02, 0.50641391E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir14_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.64934897E+01, -0.60000785E-01, 0.15284570E+00, 0.42398740E-01, 0.80449143E-02, -0.12943848E-02, -0.18471053E-01, 0.14364937E-01, -0.27232725E-01, 0.34205071E-02, 0.44790650E-03, 0.49032934E-03, -0.11499410E-01, 
    },
    {
      -0.88309318E+00, -0.13880616E+00, 0.72883040E+00, -0.23786409E+00, 0.56492403E-01, -0.44968966E+00, 0.27552197E-01, -0.10558825E+00, -0.26403388E+00, -0.71968939E-02, -0.76252706E-01, -0.75245716E-01, 0.36521424E-01, 
    },
    {
      0.10725157E+02, 0.11520151E+01, -0.29717491E+01, 0.77741110E+00, 0.10456994E+01, -0.56724101E+00, 0.10612051E+01, -0.19255696E-01, 0.20865384E+00, 0.83688861E-02, 0.17841756E+00, 0.38568929E-01, -0.13791542E+00, 
    },
    {
      0.65042410E+01, -0.35215244E+01, -0.41706271E+01, 0.52065024E+01, 0.25852561E+00, 0.73438549E+01, -0.40593761E+00, 0.99215275E+00, 0.34198158E+01, 0.59768510E+00, 0.13837192E+01, 0.21450295E+01, -0.30246952E+00, 
    },
    {
      -0.73952766E+02, -0.10371047E+02, 0.20006662E+02, -0.58430343E+01, -0.17244541E+02, 0.25785179E+01, -0.66231856E+01, -0.10432464E+01, 0.73055929E+00, 0.10648139E+01, -0.16124163E+01, 0.77939086E-01, 0.78449029E+00, 
    },
    {
      0.16648845E+00, 0.12172329E+02, -0.27850802E+01, -0.25546795E+02, 0.91742992E+01, -0.32795715E+02, 0.28933215E+00, -0.56761956E+01, -0.18448353E+02, -0.33526461E+01, -0.67829127E+01, -0.12907632E+02, -0.90398967E+00, 
    },
    {
      0.15081433E+03, 0.33412430E+02, -0.50513580E+02, 0.15113485E+02, 0.58774567E+02, -0.36906857E+01, 0.14214766E+02, 0.30996857E+01, -0.45489955E+01, -0.52512527E+01, 0.42320633E+01, -0.65626699E+00, -0.12561433E+01, 
    },
    {
      -0.37230701E+02, -0.76325502E+01, 0.37164608E+02, 0.49207062E+02, -0.43934433E+02, 0.60087318E+02, 0.41347008E+01, 0.15909638E+02, 0.43547508E+02, 0.70065861E+01, 0.12691342E+02, 0.29472052E+02, 0.48083038E+01, 
    },
    {
      -0.14633238E+03, -0.42288403E+02, 0.54256516E+02, -0.17434313E+02, -0.73183670E+02, 0.16437659E+01, -0.12896775E+02, -0.25577700E+01, 0.64732404E+01, 0.78860893E+01, -0.44402709E+01, 0.10190881E+01, 0.83452731E+00, 
    },
    {
      0.52198563E+02, -0.78648896E+01, -0.54521660E+02, -0.41549355E+02, 0.62281200E+02, -0.48637409E+02, -0.83191090E+01, -0.19453861E+02, -0.45155918E+02, -0.66038699E+01, -0.10019678E+02, -0.28811026E+02, -0.57783585E+01, 
    },
    {
      0.55789318E+02, 0.18161959E+02, -0.20868525E+02, 0.74471416E+01, 0.30560638E+02, 0.63638061E-01, 0.42821302E+01, 0.46981829E+00, -0.28367572E+01, -0.37410161E+01, 0.16539934E+01, -0.48298737E+00, -0.21111564E+00, 
    },
    {
      -0.20312288E+02, 0.70529175E+01, 0.23697180E+02, 0.12980741E+02, -0.27796104E+02, 0.14427422E+02, 0.43019180E+01, 0.83882170E+01, 0.16908133E+02, 0.23868093E+01, 0.27971561E+01, 0.10196642E+02, 0.21167967E+01, 
    },
    {
      -0.23634213E+00, 0.20315306E+01, 0.14751929E+01, 0.51506594E-01, -0.41991584E-01, -0.66530528E-02, 0.60239173E-01, 0.42047124E-01, 0.14694092E-01, 0.14501678E-01, 0.22909958E-01, 0.50705005E-02, -0.13813758E-01, 
    },
    {
      0.24778880E-01, -0.13336085E+01, 0.20401320E+01, -0.29250428E-01, 0.11255962E+00, -0.27715957E-01, -0.61806727E-01, 0.33551306E-02, 0.49672600E-01, -0.88648414E-02, 0.15672091E-01, -0.17772952E-01, 0.25840361E-01, 
    },
    {
      -0.94002926E+00, 0.96187347E+00, -0.48695213E+00, -0.14855824E-01, 0.85204057E-01, 0.68687528E-01, 0.59923762E+00, -0.17397812E+00, 0.52780789E+00, -0.97602285E-01, -0.50955068E-01, 0.97110130E-01, 0.78837104E-01, 
    },
    {
      -0.15056743E+00, 0.25372839E+00, 0.77975589E+00, 0.44844508E-01, 0.66648591E+00, 0.29808590E-01, -0.28804496E-01, 0.27313724E+00, 0.57353723E+00, -0.14783476E+00, -0.54335695E-01, 0.18213356E+00, -0.17622781E+00, 
    },
    {
      0.44821749E+01, 0.12306819E+02, 0.71697435E+01, -0.92180455E+00, 0.17149884E+01, 0.12936370E+01, -0.14673869E+00, -0.16734644E+01, -0.90978986E+00, -0.64830935E+00, -0.10648321E+01, 0.25684693E+00, -0.10137751E+00, 
    },
    {
      0.85512914E-01, -0.10835258E+02, 0.85481062E+01, 0.11747242E+01, 0.18459188E+01, 0.27393374E+00, 0.21028144E+01, -0.48073640E+00, -0.68387592E+00, 0.73008960E+00, -0.79940535E-01, 0.11825157E+01, -0.94839269E+00, 
    },
    {
      0.13821822E+02, -0.18135992E+02, 0.25114826E+02, 0.74971118E+01, 0.18632157E+01, 0.13507416E+01, -0.10556159E+02, 0.19317433E+01, -0.81330175E+01, 0.36078362E+01, 0.21540523E+00, -0.27909291E+01, -0.10886441E+01, 
    },
    {
      0.62252507E+01, -0.13018351E+02, -0.10101440E+02, -0.10007334E+01, -0.14628852E+02, 0.11357673E+01, 0.36046584E+01, -0.42910991E+01, -0.89776678E+01, -0.89598792E-02, -0.42933542E+00, -0.15849814E+01, 0.34161615E+01, 
    },
    {
      -0.17371538E+02, -0.66748375E+02, 0.62607899E+02, 0.20953897E+02, -0.16884054E+02, -0.15206277E+02, -0.11621350E+01, 0.13002543E+02, 0.13039217E+02, 0.52742462E+01, 0.93021603E+01, -0.34476848E+01, 0.56655067E+00, 
    },
    {
      -0.56244588E+00, -0.33539303E+02, -0.27122942E+02, -0.17939253E+02, -0.18481981E+02, -0.38500461E+00, -0.14842595E+02, -0.31616006E+01, 0.47638435E+01, -0.45677834E+01, -0.55476618E+00, -0.94443140E+01, 0.57917881E+01, 
    },
    {
      -0.74924995E+02, 0.14071364E+03, -0.14038925E+03, -0.58120224E+02, 0.71764197E+01, -0.49531412E+01, 0.62206001E+02, -0.40747013E+01, 0.32049103E+02, -0.27928366E+02, 0.17490292E+01, 0.17929644E+02, 0.32690539E+01, 
    },
    {
      -0.65806305E+02, 0.54215317E+02, 0.98710693E+02, -0.88018885E+01, 0.75170929E+02, -0.11637873E+02, -0.30119808E+02, 0.31769405E+02, 0.47102097E+02, 0.81819410E+01, 0.76862321E+01, 0.71875310E+01, -0.20791466E+02, 
    },
    {
      0.89678268E+01, 0.10135834E+03, -0.35098581E+03, -0.86679405E+02, 0.51673248E+02, 0.53351051E+02, 0.76851803E+00, -0.36545685E+02, -0.47825413E+02, -0.15288162E+02, -0.24478294E+02, 0.11542589E+02, 0.20585668E+01, 
    },
    {
      0.16705963E+02, 0.25283752E+03, -0.40345543E+02, 0.73295845E+02, 0.65839951E+02, 0.55188236E+01, 0.37025734E+02, 0.24288336E+02, -0.10951288E+02, 0.91241064E+01, 0.23609199E+00, 0.27590469E+02, -0.15265544E+02, 
    },
    {
      0.19687543E+03, -0.38724133E+03, 0.30177777E+03, 0.15378911E+03, -0.65229530E+02, -0.34766839E+01, -0.15835892E+03, 0.39712989E+01, -0.47045303E+02, 0.82022125E+02, -0.10522800E+02, -0.45835922E+02, -0.15153832E+01, 
    },
    {
      0.19607632E+03, -0.66794991E+02, -0.28728564E+03, 0.45004837E+02, -0.16881575E+03, 0.29734268E+02, 0.87140694E+02, -0.93818230E+02, -0.11390152E+03, -0.33501694E+02, -0.25811827E+02, -0.17500467E+02, 0.51393471E+02, 
    },
    {
      0.36721821E+02, -0.58378582E+02, 0.52954004E+03, 0.12650403E+03, -0.61309460E+02, -0.72215317E+02, 0.55067840E+01, 0.43052776E+02, 0.65146362E+02, 0.18467991E+02, 0.23828384E+02, -0.14804021E+02, -0.66496506E+01, 
    },
    {
      -0.41053894E+02, -0.39513184E+03, 0.14958513E+03, -0.11126383E+03, -0.10337821E+03, -0.17559303E+02, -0.34940395E+02, -0.43032936E+02, 0.85262613E+01, -0.69466352E+01, 0.36357474E+01, -0.34809723E+02, 0.18168346E+02, 
    },
    {
      -0.24683015E+03, 0.44674811E+03, -0.29277634E+03, -0.16985976E+03, 0.11071173E+03, 0.20931087E+02, 0.18121008E+03, -0.38683255E+01, 0.23711794E+02, -0.10216876E+03, 0.16843342E+02, 0.51287098E+02, -0.42596765E+01, 
    },
    {
      -0.21369885E+03, 0.15854466E+02, 0.32373230E+03, -0.56573444E+02, 0.18293462E+03, -0.24894985E+02, -0.10696294E+03, 0.11558176E+03, 0.12890714E+03, 0.48120480E+02, 0.32083454E+02, 0.20970270E+02, -0.55741127E+02, 
    },
    {
      -0.36825500E+02, 0.12671242E+02, -0.24902853E+03, -0.60534359E+02, 0.24415829E+02, 0.33357067E+02, -0.56761322E+01, -0.18138479E+02, -0.29999857E+02, -0.79702997E+01, -0.73311758E+01, 0.64502964E+01, 0.43373952E+01, 
    },
    {
      0.26744591E+02, 0.18625591E+03, -0.93930122E+02, 0.56447613E+02, 0.58127296E+02, 0.13249874E+02, 0.98932362E+01, 0.23001041E+02, -0.14396309E+01, 0.16029669E+01, -0.36048818E+01, 0.15927850E+02, -0.79116158E+01, 
    },
    {
      0.11669273E+03, -0.18738850E+03, 0.10586761E+03, 0.65915176E+02, -0.55447582E+02, -0.14503896E+02, -0.76371826E+02, 0.23939581E+01, -0.49679360E+00, 0.45396572E+02, -0.83436613E+01, -0.20962414E+02, 0.35527298E+01, 
    },
    {
      0.76042992E+02, 0.11706135E+02, -0.12662367E+03, 0.20020100E+02, -0.78973969E+02, 0.45572462E+01, 0.47576389E+02, -0.49966724E+02, -0.54524891E+02, -0.23131273E+02, -0.13527062E+02, -0.95180111E+01, 0.22111988E+02, 
    },
    {
      0.19034376E-01, 0.98584309E-01, 0.10540415E-01, -0.10316912E+01, -0.50497755E-01, -0.28466880E-02, 0.14425301E-01, 0.89508243E-03, 0.40291805E-01, -0.58668968E-02, -0.37617797E-02, 0.81910536E-03, 0.14230346E-01, 
    },
    {
      0.55068702E-01, -0.61659035E-02, 0.25711088E-01, 0.34259062E-01, -0.75383478E+00, 0.40433593E-02, -0.13901708E-01, -0.40712301E-01, -0.29276965E-01, -0.10735627E-01, -0.72969968E-03, -0.98825945E-02, -0.17654425E-02, 
    },
    {
      0.12820829E+00, 0.71660620E+00, 0.71653306E+00, 0.37673324E+00, -0.52130508E+00, 0.26109612E+00, -0.33626252E+00, 0.12658983E+00, 0.23496895E+00, 0.45112655E-01, 0.10876163E-01, -0.79171509E-01, -0.49059778E-01, 
    },
    {
      0.84820724E+00, 0.11983800E+00, 0.93958420E+00, 0.63321990E+00, 0.95848083E+00, 0.16888373E+00, 0.38933194E+00, -0.13175000E+00, -0.13978599E+00, -0.28500167E+00, -0.16895266E+00, 0.28273294E-01, -0.13272938E+00, 
    },
    {
      0.10719464E+01, -0.10473355E+01, 0.16151112E+01, 0.88533993E+01, 0.29226627E+01, -0.44834813E+00, -0.18926809E+01, -0.10330509E+01, 0.70661664E-01, 0.34604049E+00, 0.24440205E+00, -0.55603737E+00, -0.15387839E+00, 
    },
    {
      0.15618073E+01, 0.14701888E+00, -0.74968077E-01, -0.39555376E+01, 0.69049392E+01, -0.19420260E+00, 0.57892245E+00, -0.79809892E+00, 0.60062045E+00, 0.22479720E-01, -0.32704258E+00, 0.23038152E+00, 0.36810508E+00, 
    },
    {
      -0.29060516E+01, -0.35181143E+01, -0.57364321E+01, 0.55235324E+01, 0.45822948E+00, -0.15032139E+01, 0.21021144E+01, -0.25328857E+00, -0.18473371E+01, -0.84338069E-01, 0.39706534E+00, 0.18311095E+00, 0.12498163E+01, 
    },
    {
      -0.99295187E+01, -0.10716367E+01, -0.49983807E+01, -0.20832248E+01, 0.18830384E+01, 0.86717820E+00, -0.15240573E+01, 0.21804466E+01, 0.58699912E+00, 0.14913265E+01, 0.91169506E+00, 0.18899290E+00, 0.18160110E+01, 
    },
    {
      -0.94457245E+01, 0.17658319E+02, -0.11660739E+02, -0.68379431E+01, -0.55005379E+01, 0.43817210E+01, 0.11818675E+02, 0.47689600E+01, 0.28519669E+00, -0.19000353E+01, -0.25939715E+01, 0.37662926E+01, 0.75847077E+00, 
    },
    {
      -0.56515350E+01, 0.91205730E+01, -0.19686940E+01, 0.13128128E+02, -0.68311524E+00, 0.99185145E+00, -0.16249559E+01, 0.51543722E+01, -0.11013597E+01, -0.84779017E-01, 0.18501135E+01, -0.38142797E+00, -0.30122881E+01, 
    },
    {
      0.75019464E+01, 0.35646596E+01, 0.11501075E+02, -0.18637688E+02, -0.69856608E+00, 0.35613034E+01, -0.47264366E+01, -0.46824551E+00, 0.42374134E+01, -0.10472504E+00, -0.11728551E+01, -0.23341645E+00, -0.39178739E+01, 
    },
    {
      0.26873817E+02, -0.89032382E+00, 0.74988832E+01, 0.33407247E+01, -0.96325502E+01, -0.34091041E+01, 0.21045973E+01, -0.56645589E+01, 0.19026136E+00, -0.22414279E+01, -0.13311152E+01, -0.59583026E+00, -0.46499009E+01, 
    },
    {
      0.21163473E+02, -0.48009987E+02, 0.25364653E+02, -0.28718987E+02, 0.28193629E+00, -0.14275020E+02, -0.20585054E+02, -0.64663568E+01, -0.17794694E+01, 0.29850135E+01, 0.60325065E+01, -0.71860552E+01, -0.16418031E+01, 
    },
    {
      0.39766910E+01, -0.27315857E+02, 0.82747126E+01, -0.15480450E+02, -0.37333546E+02, -0.14049913E+01, -0.78134075E-01, -0.88510962E+01, -0.22243371E+01, 0.97602409E+00, -0.25714512E+01, -0.92084908E+00, 0.73255792E+01, 
    },
    {
      -0.59779377E+01, 0.12392722E+01, -0.78840604E+01, 0.12213486E+02, -0.45400086E+00, -0.31542869E+01, 0.34382269E+01, 0.87975073E+00, -0.27262006E+01, 0.43419167E+00, 0.68464506E+00, 0.22095339E+00, 0.30523138E+01, 
    },
    {
      -0.20467285E+02, 0.39570870E+01, -0.23940721E+01, 0.22183473E+00, 0.45692406E+01, 0.23968148E+01, -0.16324166E+01, 0.37603843E+01, -0.13567244E+01, 0.12711176E+01, 0.64020485E+00, 0.41453242E+00, 0.32386906E+01, 
    },
    {
      -0.14195784E+02, 0.34288692E+02, -0.15830330E+02, 0.32217907E+02, 0.34044676E+01, 0.12452764E+02, 0.10785024E+02, 0.26360891E+01, 0.13603797E+01, -0.15146811E+01, -0.39355423E+01, 0.41184163E+01, 0.11697747E+01, 
    },
    {
      0.83907932E+00, 0.17864233E+02, -0.90199413E+01, 0.53105712E+01, 0.36594128E+02, 0.70343792E+00, 0.17940173E+01, 0.47861300E+01, 0.37236953E+01, -0.12472179E+01, 0.73165965E+00, 0.12939287E+01, -0.52513423E+01, 
    },
    {
      -0.64029008E-01, 0.89368589E-01, -0.10483639E+00, -0.41617226E-01, 0.42851824E-01, -0.37522101E+00, -0.74830681E+00, -0.35888155E-02, 0.50473481E-01, 0.39554734E-01, 0.17244495E-01, -0.62780860E-02, 0.40925271E-02, 
    },
    {
      0.47223818E-01, -0.15007500E+00, -0.77078842E-01, 0.23586081E-01, -0.40981129E-01, 0.73113239E+00, -0.37994549E+00, -0.29909762E-01, 0.14559225E-01, 0.11881361E-01, -0.21748168E-01, 0.17178882E-01, 0.25832307E-01, 
    },
    {
      -0.12422315E+00, 0.75573784E+00, 0.11975990E+00, -0.26279071E+00, -0.63855439E+00, -0.10346416E+00, -0.24757487E+00, 0.83832882E-01, -0.37987664E-01, 0.47420461E-01, 0.51827621E-01, -0.82543463E-01, -0.48918199E-01, 
    },
    {
      -0.25143120E-01, -0.32514688E+00, -0.23532422E+00, -0.13309413E-01, -0.85015304E-01, 0.38035685E+00, 0.60264971E-01, -0.13409159E-01, -0.31570796E-01, -0.10386517E+00, -0.10889217E-01, 0.21442087E-01, -0.58925855E-02, 
    },
    {
      0.51993853E+00, -0.57693455E-01, 0.28187773E+00, -0.73443003E-01, -0.19520199E+00, -0.11351212E-01, 0.58080941E+00, -0.54905391E+00, 0.22085562E+00, -0.66554368E-01, -0.17385939E+00, 0.18870950E-03, -0.11992918E+00, 
    },
    {
      -0.21958467E+00, 0.21047795E+01, -0.20394135E+00, -0.57054359E+00, 0.54538924E-01, -0.12259457E+01, 0.53822305E-01, -0.47958690E+00, -0.65027052E+00, -0.24385016E-01, 0.14002994E+00, -0.22192319E+00, -0.13525110E+00, 
    },
    {
      -0.54393631E+00, -0.13514569E+01, 0.26100159E+00, 0.43233761E+00, 0.97404122E+00, 0.40088597E+00, 0.80763835E+00, -0.98028913E-01, 0.11027520E+00, -0.10748786E+00, -0.36647193E-01, 0.19190548E+00, 0.88344067E-01, 
    },
    {
      -0.25658110E+00, 0.72416735E+00, 0.57697695E+00, -0.34691748E+00, 0.32392946E+00, -0.13627889E+01, 0.11442543E+00, -0.10774225E+00, 0.10521625E+00, 0.98740987E-01, 0.29976496E-01, 0.26290998E-01, -0.56078196E-01, 
    },
    {
      0.27019140E+00, 0.45024633E+00, -0.17958543E+00, 0.12233192E+00, 0.51881069E+00, 0.42524201E+00, 0.30084953E+00, 0.94178420E+00, -0.65111023E+00, -0.27122760E-01, 0.76508380E-01, -0.85360545E-04, 0.12641452E+00, 
    },
    {
      0.79744786E-01, -0.28821568E+01, 0.73260629E+00, 0.53177392E+00, 0.33241987E+00, 0.14034445E+01, 0.37998933E+00, 0.94033825E+00, 0.12633942E+01, -0.60186643E-01, -0.11780193E+00, 0.26741916E+00, 0.93822643E-01, 
    },
    {
      -0.40619418E-01, 0.80723641E-02, -0.33849481E-01, 0.31438507E-01, 0.26720112E-01, 0.34434196E-01, 0.15387811E-01, 0.30153814E+00, -0.68199985E-01, 0.15819095E-01, 0.84541179E-02, -0.12130376E-01, -0.27178498E-02, 
    },
    {
      0.10475045E+00, 0.49542687E-02, 0.35022680E-01, 0.87533630E-01, -0.61479896E-01, -0.14135311E-01, -0.12770660E-01, 0.36308039E-01, 0.26710710E+00, 0.10513934E-01, -0.55483915E-02, 0.76266043E-02, -0.12753191E-01, 
    },
    {
      -0.21390147E-02, 0.58438715E-01, -0.11032433E+00, 0.12841828E+00, -0.22302932E+00, -0.96188486E-01, 0.17974999E-01, 0.29908022E+00, 0.44135422E+00, -0.46303540E-01, 0.41154433E-01, -0.14057259E-01, 0.87432843E-02, 
    },
    {
      0.30396742E+00, 0.75595453E-01, -0.15522315E+00, -0.17690331E-02, 0.20859620E+00, -0.87131839E-02, -0.10404948E+00, -0.36777064E+00, 0.11611282E+00, -0.29208448E-02, -0.80868147E-01, 0.16911151E-01, 0.13700172E-01, 
    },
    {
      -0.13175465E+00, -0.61796591E-01, -0.48755106E-01, -0.12370612E-01, -0.54931860E-01, -0.18966548E-01, 0.25214590E-01, 0.38970198E-01, -0.42744722E-01, 0.83304532E-01, 0.22285062E+00, -0.54338336E-03, 0.11550711E-02, 
    },
    {
      0.12497649E-01, 0.33643261E-01, -0.49855210E-01, 0.77534869E-01, -0.20823942E-01, 0.56232158E-02, -0.27867474E-01, 0.31055154E-01, 0.45609649E-01, -0.22880867E+00, 0.88289447E-01, 0.14045586E-01, -0.77454462E-02, 
    },
    {
      -0.15834920E-01, -0.47239352E-01, 0.55206638E-01, -0.21747222E-01, 0.19386897E-01, 0.35340972E-01, 0.19234702E-01, 0.55831406E-01, 0.12493882E-01, -0.18281758E-01, -0.14992007E-01, -0.12327413E+00, 0.10836018E-01, 
    },
    {
      0.29655593E-01, 0.94482973E-02, -0.85593509E-02, 0.62130149E-02, 0.62354233E-01, 0.73285494E-02, -0.53194746E-01, -0.14881391E-01, 0.54990896E-02, -0.13634078E-01, -0.12771355E-01, 0.14701579E-02, -0.12577879E+00, 
    },
    {
      0.20759542E-01, 0.37858854E-02, -0.38624063E-01, 0.25079571E-01, 0.28830525E-01, 0.14798064E-01, 0.11077589E-01, 0.17425489E-01, 0.35621811E-01, -0.25656393E-01, -0.24061346E-01, 0.10520253E-01, 0.45533017E-02, 
    },
    {
      0.66021487E-01, -0.21064222E-01, -0.72724107E-02, 0.14031813E-01, 0.59114888E-01, 0.15700987E-01, 0.68930597E-02, -0.33762053E-01, -0.16519623E-01, -0.90381289E-02, -0.59326985E-02, 0.11942729E-01, 0.13869079E-01, 
    },
    {
      0.63189775E-01, 0.68125673E-01, -0.60883962E-01, -0.21250362E-01, -0.66251666E-02, -0.73823389E-02, 0.39793655E-01, -0.44493157E-01, 0.10462924E-01, -0.12947634E-01, -0.25500301E-02, -0.15238015E-01, 0.15929751E-01, 
    },
    {
      -0.17856486E-01, -0.52744098E-01, -0.53626947E-01, 0.39534133E-01, 0.30350417E-01, 0.46890290E-02, -0.20292008E-02, 0.65144026E-02, 0.41108787E-01, -0.98453555E-02, -0.10881119E-01, 0.51510613E-03, 0.14489324E-01, 
    },
  },
  {
    {
      0.97736454E+01, 0.14032146E+00, 0.20471638E+00, 0.45970183E-01, 0.19347202E-01, 0.34263041E-02, -0.56153454E-01, 0.16434982E-01, -0.13611992E-01, 0.26763946E-01, 0.57403501E-02, 0.63922638E-02, -0.11557203E-01, 
    },
    {
      0.21873847E+00, -0.40463670E-02, 0.25696555E+00, -0.28938836E+00, 0.35062745E+00, 0.15157066E+00, -0.29073560E+00, 0.10218962E+00, -0.17300349E+00, -0.40608421E-01, -0.18231006E-01, -0.10038075E-01, 0.41188154E-01, 
    },
    {
      0.30748093E+02, -0.22748404E+01, -0.31234460E+01, 0.27615957E+01, 0.15846493E+01, -0.34613076E+00, 0.56108177E+00, 0.11689792E+00, 0.14760590E+00, -0.10534291E+01, -0.31982246E+00, 0.22981022E+00, -0.44722270E-01, 
    },
    {
      -0.14903673E+02, 0.10495951E+00, 0.34792299E+01, 0.47500277E+01, -0.45592580E+01, -0.62566572E+00, 0.40229535E+01, -0.11005296E+01, 0.26008980E+01, -0.20069683E+00, 0.20519117E+00, 0.83855027E+00, -0.12517659E+00, 
    },
    {
      -0.15924529E+03, 0.13519261E+02, 0.10932475E+02, -0.29060259E+02, -0.17306383E+02, 0.23026779E+01, -0.11505018E+01, 0.89456850E+00, 0.19126704E+00, 0.56772127E+01, 0.24429238E+01, -0.23409443E+01, 0.10148532E+01, 
    },
    {
      0.11584834E+03, -0.17900440E+02, -0.13041323E+02, -0.16672869E+02, 0.35105484E+02, -0.30701976E+01, -0.21194830E+02, 0.64325070E+01, -0.15774626E+02, 0.32018068E+01, 0.94925404E+00, -0.46957250E+01, -0.13125659E+01, 
    },
    {
      0.29437540E+03, -0.25372292E+02, -0.15352101E+02, 0.91963020E+02, 0.53816139E+02, -0.39363284E+01, -0.11707987E+01, -0.62234478E+01, -0.21492360E+01, -0.11163957E+02, -0.70858817E+01, 0.70973535E+01, -0.29095371E+01, 
    },
    {
      -0.29224625E+03, 0.61603146E+02, -0.42039781E+01, 0.19493481E+02, -0.98674675E+02, 0.13333192E+02, 0.47858620E+02, -0.16327671E+02, 0.39196560E+02, -0.93861637E+01, -0.60532708E+01, 0.96561794E+01, 0.45459685E+01, 
    },
    {
      -0.26182547E+03, 0.17990252E+02, 0.11552253E+02, -0.11288570E+03, -0.65355690E+02, 0.19973891E+01, 0.40330243E+01, 0.10058838E+02, 0.31824422E+01, 0.93387594E+01, 0.84235001E+01, -0.84775333E+01, 0.29146054E+01, 
    },
    {
      0.29994781E+03, -0.74359123E+02, 0.36368931E+02, -0.57280774E+01, 0.11249139E+03, -0.15975358E+02, -0.48075100E+02, 0.17203218E+02, -0.41371578E+02, 0.10335050E+02, 0.90007267E+01, -0.86931496E+01, -0.47164741E+01, 
    },
    {
      0.91385628E+02, -0.41139150E+01, -0.41653366E+01, 0.47360004E+02, 0.27213379E+02, -0.10529798E-02, -0.22038872E+01, -0.48922997E+01, -0.13381232E+01, -0.28368783E+01, -0.34559801E+01, 0.34892297E+01, -0.93806028E+00, 
    },
    {
      -0.10869084E+03, 0.30786814E+02, -0.22838774E+02, -0.15038877E+01, -0.44714828E+02, 0.61636329E+01, 0.17758163E+02, -0.62771416E+01, 0.15508571E+02, -0.38858967E+01, -0.40972447E+01, 0.29156959E+01, 0.15558701E+01, 
    },
    {
      -0.13369410E+00, 0.17268354E+01, 0.14351931E+01, 0.76106519E-01, -0.28437419E-01, -0.11006788E-01, 0.26756642E-01, 0.27960541E-01, 0.89221104E-03, 0.14117885E-01, 0.26185049E-01, -0.14624753E-02, -0.14191492E-01, 
    },
    {
      0.24574746E+00, -0.14273102E+01, 0.15711827E+01, -0.83912954E-01, 0.46123277E-01, -0.25364889E-01, -0.29690681E-01, -0.29524004E-01, 0.19983063E-01, -0.17720954E-01, 0.16691638E-01, -0.70638601E-02, -0.72944113E-02, 
    },
    {
      0.39862120E+00, -0.13170791E+00, -0.64909148E+00, -0.32600096E+00, 0.21109910E+00, -0.12451132E+00, 0.43011883E+00, -0.18416311E+00, 0.24612106E+00, -0.10645313E+00, 0.16143291E+00, -0.21634487E-01, 0.63935556E-01, 
    },
    {
      0.92620957E+00, -0.49406093E+00, 0.64722395E+00, -0.72104082E-01, 0.28068078E+00, 0.60120350E+00, 0.27893120E+00, 0.30756903E+00, 0.17852144E-01, -0.17372723E+00, -0.12843073E-01, 0.60005605E-01, -0.44218376E+00, 
    },
    {
      0.10201948E+01, 0.13495907E+02, -0.10630937E+02, -0.10683585E+01, 0.58663207E+00, -0.10076985E+01, -0.52236602E-01, -0.47220096E+00, 0.46965632E+00, -0.10257162E+01, -0.77649891E+00, 0.21886133E+00, -0.13192793E-01, 
    },
    {
      -0.63123832E+01, 0.98021412E+01, 0.17635956E+02, 0.41804461E+01, 0.17669466E+01, 0.96404308E+00, -0.45744419E-01, 0.97980118E+00, -0.73717684E+00, -0.27884778E+00, -0.14686050E+00, 0.75753582E+00, 0.43112877E+00, 
    },
    {
      -0.29365113E+01, -0.14322780E+01, 0.14004352E+02, 0.79823327E+00, -0.58754277E+01, 0.25628228E+01, -0.86770887E+01, 0.29869020E+01, -0.79320574E+00, 0.71983641E+00, -0.20778570E+01, 0.96054983E+00, -0.10125142E+01, 
    },
    {
      -0.14787666E+02, 0.89715195E+01, -0.16094648E+02, 0.62427344E+01, -0.78737440E+01, -0.20880353E+01, -0.53182220E+00, -0.72572389E+01, 0.20186639E+01, -0.28493693E+00, -0.16519644E+01, -0.56926078E+00, 0.75286026E+01, 
    },
    {
      0.38437169E+01, 0.23742902E+01, 0.10843651E+03, 0.32352233E+01, 0.29753842E+01, 0.82216110E+01, -0.24196384E+01, 0.54554963E+01, -0.16184169E+00, 0.67759509E+01, 0.53559599E+01, -0.44717285E+00, -0.89659041E+00, 
    },
    {
      0.51552006E+02, -0.11602977E+03, -0.19592438E+02, -0.36331181E+02, -0.12795036E+02, -0.63453674E+01, 0.43711376E+01, -0.10106875E+02, 0.51247444E+01, 0.35840931E+01, 0.18366957E+01, -0.68241401E+01, -0.46224647E+01, 
    },
    {
      -0.85184917E+01, 0.16360107E+02, -0.82426064E+02, 0.12709790E+02, 0.42946472E+02, -0.15440465E+02, 0.50130878E+02, -0.12952686E+02, -0.69280114E+01, -0.39795120E+01, 0.73542218E+01, -0.55133834E+01, 0.81558447E+01, 
    },
    {
      0.96721992E+02, -0.39015026E+02, 0.11213722E+03, -0.45417725E+02, 0.37435158E+02, -0.13082533E+01, -0.10213682E+02, 0.49293102E+02, -0.15153333E+02, 0.73512311E+01, 0.14430303E+02, 0.19037811E+01, -0.38897942E+02, 
    },
    {
      -0.43353905E+02, -0.20246556E+03, -0.28853781E+03, 0.21355968E+01, -0.19108803E+02, -0.21631128E+02, 0.12639037E+02, -0.18742435E+02, -0.96571703E+01, -0.16555702E+02, -0.14417870E+02, 0.12324684E+00, 0.51283345E+01, 
    },
    {
      -0.14933170E+03, 0.32959485E+03, -0.18182323E+03, 0.11664076E+03, 0.33371815E+02, 0.17004457E+02, -0.19989794E+02, 0.32714710E+02, -0.11863819E+02, -0.12437937E+02, -0.82264671E+01, 0.19701031E+02, 0.14783766E+02, 
    },
    {
      0.80268501E+02, -0.52966335E+02, 0.19053438E+03, -0.56769230E+02, -0.11784628E+03, 0.38850357E+02, -0.12247976E+03, 0.22450397E+02, 0.35795444E+02, 0.12220045E+02, -0.89107647E+01, 0.82319517E+01, -0.27355555E+02, 
    },
    {
      -0.27648264E+03, 0.89760971E+02, -0.27464084E+03, 0.11196627E+03, -0.69628716E+02, 0.97918758E+01, 0.44225208E+02, -0.12664152E+03, 0.35483727E+02, -0.18650520E+02, -0.41325413E+02, -0.23082571E+01, 0.85997475E+02, 
    },
    {
      0.88594856E+02, 0.39292142E+03, 0.33457990E+03, -0.12988249E+02, 0.27334524E+02, 0.24206512E+02, -0.20228155E+02, 0.24981485E+02, 0.22630320E+02, 0.17233301E+02, 0.16416178E+02, -0.32048091E+00, -0.83845139E+01, 
    },
    {
      0.18617633E+03, -0.39346274E+03, 0.42378693E+03, -0.15928719E+03, -0.40951443E+02, -0.20998573E+02, 0.32289707E+02, -0.42694340E+02, 0.11828459E+02, 0.17220362E+02, 0.13173159E+02, -0.22972109E+02, -0.18645224E+02, 
    },
    {
      -0.14528056E+03, 0.72072128E+02, -0.21711536E+03, 0.77514603E+02, 0.13667310E+03, -0.45385811E+02, 0.13474231E+03, -0.16270960E+02, -0.53262211E+02, -0.17100195E+02, 0.23942339E+01, -0.10488991E+01, 0.37738712E+02, 
    },
    {
      0.35145352E+03, -0.85622932E+02, 0.26572525E+03, -0.10509781E+03, 0.58970482E+02, -0.46502337E+01, -0.66478355E+02, 0.13732690E+03, -0.33438587E+02, 0.15821194E+02, 0.48546875E+02, 0.14652005E+01, -0.85408585E+02, 
    },
    {
      -0.52937946E+02, -0.20882060E+03, -0.14351868E+03, 0.98669796E+01, -0.11305204E+02, -0.10061109E+02, 0.10184311E+02, -0.11492324E+02, -0.13785835E+02, -0.64393106E+01, -0.66763282E+01, 0.56769723E+00, 0.43437328E+01, 
    },
    {
      -0.82705460E+02, 0.16970370E+03, -0.24499571E+03, 0.77111313E+02, 0.21147839E+02, 0.97495165E+01, -0.17787193E+02, 0.19586727E+02, -0.44074783E+01, -0.82658730E+01, -0.68889670E+01, 0.94197063E+01, 0.82319374E+01, 
    },
    {
      0.79956360E+02, -0.38184700E+02, 0.95916229E+02, -0.35911259E+02, -0.56887783E+02, 0.20183622E+02, -0.54853699E+02, 0.38453090E+01, 0.25204611E+02, 0.86927290E+01, 0.11558259E+01, -0.29337678E+01, -0.18029606E+02, 
    },
    {
      -0.15973770E+03, 0.28304308E+02, -0.87137711E+02, 0.30577978E+02, -0.21899446E+02, -0.35690374E+01, 0.34283478E+02, -0.53517769E+02, 0.10929289E+02, -0.38510764E+01, -0.20168242E+02, -0.64146167E+00, 0.31352381E+02, 
    },
    {
      0.29265393E-01, -0.24591343E+00, 0.14553222E+00, -0.84675032E+00, 0.61966658E+00, -0.31919140E-01, 0.56226421E-01, 0.23503466E-01, 0.53592073E-02, -0.20415217E-01, 0.15285027E-01, -0.63779615E-02, 0.55775273E-03, 
    },
    {
      0.90108313E-01, -0.20454387E+00, -0.73520780E-01, -0.64477378E+00, -0.82056785E+00, -0.32559566E-01, -0.23924757E-01, -0.15155605E-01, -0.23472775E-01, -0.29544417E-01, 0.71342289E-02, -0.47043166E-02, -0.47027934E-02, 
    },
    {
      0.52882087E+00, 0.75878024E+00, 0.37392050E+00, 0.52214783E+00, -0.73013431E+00, 0.44710663E+00, 0.54939032E-01, 0.22959816E+00, 0.20311107E+00, -0.81182420E-01, -0.88519581E-01, -0.12205946E+00, -0.90271063E-01, 
    },
    {
      0.76053393E+00, 0.47488001E+00, 0.52843851E+00, 0.69257706E+00, 0.27872598E+00, 0.68038344E-01, 0.28850755E+00, -0.78320973E-01, 0.16159263E+00, -0.23097005E+00, -0.32177996E-01, 0.43377087E-01, -0.46754386E-01, 
    },
    {
      0.45139912E+00, 0.33630769E+01, -0.12169751E+01, -0.37143755E+01, 0.72003312E+01, -0.30781597E+00, -0.21212363E+01, -0.96283758E+00, 0.84034169E+00, 0.23972341E+00, -0.28099918E+00, 0.56343917E-01, -0.53734177E+00, 
    },
    {
      -0.14882063E+01, 0.13920959E+01, 0.12130704E+01, -0.79513636E+01, -0.25826595E+01, -0.42678672E+00, 0.51285851E+00, 0.19756992E+00, 0.12423240E+01, 0.54061139E+00, -0.37421107E-01, 0.29954505E+00, 0.11390962E+00, 
    },
    {
      -0.56358533E+01, -0.57385144E+01, -0.52712636E+01, -0.10486714E+01, 0.53347268E+01, -0.35143905E+01, -0.12000906E+01, -0.28657036E+01, -0.12631081E+01, 0.94290608E+00, 0.92692262E+00, 0.13676528E+01, 0.12747043E+01, 
    },
    {
      -0.10084604E+02, -0.83411729E+00, -0.68800192E+01, -0.56028543E+01, -0.31238866E+00, 0.13777543E+01, -0.14698855E+01, 0.52827829E+00, -0.40625021E+00, 0.18055102E+01, -0.78214049E+00, -0.69355123E-01, 0.65176880E+00, 
    },
    {
      0.11241800E+01, -0.65113230E+01, -0.21158099E+01, 0.29776365E+02, -0.24963110E+02, 0.54256268E+01, 0.11766754E+02, 0.49691448E+01, -0.47331429E+01, 0.25715870E+00, 0.14093016E+01, -0.11038080E+01, 0.35657845E+01, 
    },
    {
      0.14730919E+02, -0.30367286E+01, -0.11354827E+02, 0.30109867E+02, 0.22487963E+02, 0.39788237E+01, -0.24708533E+01, -0.19585212E+01, -0.73399401E+01, -0.29647706E+01, 0.20387471E+00, -0.30212843E+01, -0.14654021E+01, 
    },
    {
      0.14754129E+02, 0.11049473E+02, 0.13400968E+02, 0.58017559E+01, -0.10485391E+02, 0.91833000E+01, 0.26157830E+01, 0.80078135E+01, 0.25257504E+01, -0.23380883E+01, -0.18219419E+01, -0.34472325E+01, -0.37856765E+01, 
    },
    {
      0.27359425E+02, -0.55443020E+01, 0.20240791E+02, 0.79822865E+01, 0.63297043E+01, -0.35024116E+01, 0.38299043E+01, 0.60368818E+00, 0.41653635E-02, -0.47033868E+01, 0.24803977E+01, -0.54888380E+00, -0.18400835E+01, 
    },
    {
      -0.71220417E+01, -0.11084239E+01, 0.93044796E+01, -0.59619061E+02, 0.21860548E+02, -0.16687052E+02, -0.19368769E+02, -0.89723320E+01, 0.81729593E+01, -0.26830549E+01, -0.22497013E+01, 0.32522850E+01, -0.77899055E+01, 
    },
    {
      -0.42522598E+02, 0.84326611E+01, 0.25657207E+02, -0.35328533E+02, -0.45498726E+02, -0.12340075E+02, 0.45494366E+01, 0.38336654E+01, 0.14014010E+02, 0.50293469E+01, -0.80758733E+00, 0.73919759E+01, 0.33322282E+01, 
    },
    {
      -0.14168424E+02, -0.38411441E+01, -0.11940656E+02, -0.86237640E+01, 0.33731959E+01, -0.65763731E+01, -0.14734651E+01, -0.60773010E+01, -0.16627012E+01, 0.18981390E+01, 0.91158038E+00, 0.23963671E+01, 0.29741066E+01, 
    },
    {
      -0.22353624E+02, 0.11713838E+02, -0.14435802E+02, 0.11054903E+01, -0.10635337E+02, 0.18735406E+01, -0.32202849E+01, -0.17535979E+01, 0.22505578E+00, 0.38518550E+01, -0.17070565E+01, 0.92721128E+00, 0.15198689E+01, 
    },
    {
      0.53186111E+01, 0.61862016E+01, -0.31186655E+01, 0.37600212E+02, -0.35645136E+00, 0.12633383E+02, 0.96007862E+01, 0.52255836E+01, -0.41970530E+01, 0.22828405E+01, 0.11919301E+01, -0.24510043E+01, 0.52333202E+01, 
    },
    {
      0.34840076E+02, -0.12539686E+02, -0.16330975E+02, 0.10885726E+02, 0.29065947E+02, 0.10912358E+02, -0.25189373E+01, -0.15795352E+01, -0.82213926E+01, -0.28210602E+01, 0.71790761E+00, -0.55111480E+01, -0.20255141E+01, 
    },
    {
      -0.19228468E-01, -0.13114972E+00, 0.15714884E-01, -0.59424613E-01, 0.61318103E-01, -0.58697629E+00, -0.10644103E+01, 0.20145899E-01, -0.13622616E-01, -0.18642439E-01, 0.22151224E-01, 0.47150520E-02, 0.11869197E-02, 
    },
    {
      0.58520351E-01, -0.16852538E+00, -0.50097305E-01, 0.11990506E-01, -0.65762773E-02, 0.10272388E+01, -0.58508122E+00, 0.10597577E-01, 0.35774235E-01, -0.15361957E-02, -0.45238879E-01, -0.18923517E-01, -0.59281811E-02, 
    },
    {
      0.25976369E+00, 0.61606842E+00, 0.13208783E-01, 0.22894323E+00, -0.39829186E+00, 0.10561230E+00, -0.10849887E+00, 0.61109383E-01, 0.13916729E+00, 0.60511831E-01, 0.12184703E-01, 0.34330254E-02, 0.49652576E-01, 
    },
    {
      0.33294815E+00, 0.93975160E-02, -0.15443997E-01, 0.12950849E-01, -0.53923000E-01, 0.16462359E+00, 0.16396619E+00, -0.73470592E-01, 0.80365367E-01, -0.35155836E-01, -0.88913888E-01, -0.47349852E-01, 0.39145373E-01, 
    },
    {
      0.11885962E+01, 0.70961654E+00, 0.16401079E+00, 0.73770237E+00, -0.50291091E+00, 0.21559596E+01, 0.16197531E+01, -0.89583582E+00, 0.20669129E+00, 0.29641485E+00, 0.17593935E+00, 0.21165982E-01, -0.99231154E-02, 
    },
    {
      -0.12865351E+01, 0.23329020E+01, -0.46353462E+00, -0.75524963E-01, 0.23183054E+00, -0.19695925E+01, 0.19060562E+01, -0.13436994E+00, -0.10136840E+01, 0.37376542E-01, 0.39470100E+00, 0.21998765E+00, -0.11666382E-01, 
    },
    {
      -0.71975428E+00, -0.56198847E+00, 0.62459812E-01, -0.95555380E-01, 0.97378629E+00, 0.58153582E+00, 0.55926204E+00, -0.48008882E-01, -0.39891291E+00, -0.46037629E-01, 0.33642620E-01, 0.59914336E-01, -0.43716524E-01, 
    },
    {
      -0.10643944E+01, -0.18275885E+00, -0.33081704E+00, -0.55505341E+00, 0.37782079E+00, -0.80053759E+00, 0.45424259E+00, -0.85584493E-02, -0.85956886E-01, 0.12678237E+00, 0.19393581E+00, 0.18442687E+00, -0.24153914E-01, 
    },
    {
      -0.15753589E+01, -0.46785453E+00, -0.12589587E+01, -0.16451536E+01, 0.12443190E+01, -0.42283177E+01, -0.14672083E+01, 0.14503810E+01, -0.24458455E+00, -0.60518199E+00, -0.42549494E+00, -0.76019049E-01, -0.11163369E+00, 
    },
    {
      0.21944077E+01, -0.25643041E+01, 0.99713415E+00, -0.11333123E+00, -0.71992117E+00, 0.24609313E+01, -0.36890426E+01, 0.18535003E+00, 0.17068619E+01, -0.43102510E-01, -0.49058500E+00, -0.35204008E+00, -0.33120699E-01, 
    },
    {
      -0.58546931E-01, -0.61199530E-02, 0.13920870E-01, -0.22593593E-01, 0.60600404E-01, -0.26639467E-01, -0.16022459E-01, 0.32998070E+00, -0.22570471E+00, 0.51776585E-02, 0.17488200E-01, -0.34345095E-02, -0.48399698E-02, 
    },
    {
      0.23667336E-01, 0.10178508E+00, -0.29486222E-01, 0.57037488E-01, -0.34963217E-01, 0.11159874E-01, -0.30660912E-01, 0.19712472E+00, 0.30310646E+00, -0.16301766E-01, 0.55935443E-03, -0.90922392E-03, 0.42374507E-02, 
    },
    {
      -0.16880974E+00, 0.95758550E-01, -0.14629462E+00, 0.12069291E+00, 0.56210205E-01, -0.12945679E+00, 0.17435406E+00, 0.96609093E-01, 0.45201021E+00, -0.36387533E-01, -0.25372004E-01, 0.18637785E-02, -0.35717620E-02, 
    },
    {
      0.18735176E+00, -0.15473643E-01, 0.31935621E-01, 0.47426168E-02, 0.11311485E+00, -0.13744952E+00, -0.12981303E+00, -0.40770856E+00, 0.16419126E-01, 0.27462881E-03, -0.82576886E-01, -0.18355066E-01, -0.51394594E-02, 
    },
    {
      -0.22159746E-01, -0.74835721E-03, -0.36677279E-01, 0.25976226E-01, 0.33930016E-02, 0.16898612E-01, 0.23251474E-02, 0.38304754E-01, -0.42265587E-01, 0.20900027E+00, 0.22713178E+00, -0.21249074E-02, -0.10710889E-01, 
    },
    {
      0.22197489E-01, 0.83272249E-01, -0.76565236E-01, 0.21317841E-02, -0.13129898E-01, 0.16477922E-01, 0.48060704E-01, 0.35126519E-02, 0.52965067E-01, -0.20686191E+00, 0.17197247E+00, -0.42362534E-02, -0.16804585E-01, 
    },
    {
      0.21828212E-01, -0.31586766E-01, 0.81997104E-01, 0.30568482E-01, 0.17527269E-01, 0.51112954E-01, -0.15168884E-02, 0.15925393E-02, -0.10897530E-01, -0.70789969E-02, 0.73832730E-02, -0.12407729E+00, 0.85962117E-01, 
    },
    {
      0.66300765E-01, 0.51962450E-01, -0.13351660E-01, -0.23629975E-01, 0.20142781E-01, 0.38463738E-01, -0.13412732E-01, -0.93066981E-02, -0.41893595E-02, 0.41461214E-02, -0.17357355E-01, -0.11255018E+00, -0.12668885E+00, 
    },
    {
      -0.59766680E-01, -0.32285359E-01, -0.46988070E-01, 0.39100412E-01, 0.98173115E-02, 0.30527687E-01, -0.36675990E-01, -0.82789976E-02, 0.30238917E-01, -0.19059830E-02, -0.11280150E-01, -0.14733355E-01, -0.83649457E-02, 
    },
    {
      0.45781240E-01, 0.46038061E-01, 0.41310016E-01, -0.11225826E+00, -0.28089361E-01, 0.13309368E-01, -0.41037355E-01, -0.30710904E-01, -0.95726689E-02, -0.23963351E-01, 0.16758924E-01, 0.11331435E-01, 0.23001438E-01, 
    },
    {
      0.68212301E-02, 0.87390676E-01, 0.28464913E-01, -0.32604266E-01, -0.19913841E-01, -0.69480464E-01, -0.14800048E-01, 0.22902381E-01, 0.37979659E-01, 0.20491343E-01, -0.21194531E-01, -0.35256979E-02, -0.19661651E-02, 
    },
    {
      -0.39640926E-01, 0.19599617E-01, -0.60924832E-01, -0.32056216E-01, 0.47802251E-01, 0.17946452E-01, -0.80087418E-02, 0.32511208E-01, -0.46744170E-02, 0.21583992E-02, -0.19618925E-01, -0.14446047E-01, -0.65722289E-02, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir14_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30586901E+01, -0.16637227E-01, 0.11471306E-01, 0.27490901E-02, 0.25678012E-02, 0.37677912E-02, 0.98502170E-02, -0.31296548E-02, -0.10140381E-02, 
    },
    {
      -0.13978875E+00, 0.40312845E-01, 0.71543925E-01, 0.19734301E-01, -0.78674164E-02, -0.28102145E-01, -0.14239720E-01, -0.12600211E-01, 0.16756015E-01, 
    },
    {
      0.90984887E+00, 0.30035329E+00, -0.44702735E-01, -0.11969292E+00, -0.10557233E+00, -0.41341111E-01, -0.80306910E-01, 0.44076685E-01, 0.34757949E-01, 
    },
    {
      0.86042881E-01, -0.43596470E+00, -0.28588781E+00, -0.25438232E-01, 0.25069173E-01, 0.82588397E-01, 0.67566939E-01, 0.33206843E-01, -0.41824643E-01, 
    },
    {
      -0.12932888E+01, -0.72340810E+00, 0.77358051E-02, 0.28751051E+00, 0.31436363E+00, 0.10259442E+00, 0.24149394E+00, -0.12717050E+00, -0.11702042E+00, 
    },
    {
      -0.22331176E-01, 0.38658845E+00, 0.28259498E+00, 0.19711521E-02, -0.24714001E-01, -0.56775019E-01, -0.61311480E-01, -0.21844221E-01, 0.23831621E-01, 
    },
    {
      0.39463902E+00, 0.47750130E+00, 0.76429829E-01, -0.16718166E+00, -0.21960321E+00, -0.68737417E-01, -0.17916369E+00, 0.90928286E-01, 0.88356271E-01, 
    },
    {
      0.15348783E-01, -0.19566035E+00, -0.33543330E+00, 0.15999714E-01, -0.25417623E-02, 0.83187250E-02, 0.10059298E-01, -0.18240046E-02, -0.25734284E-02, 
    },
    {
      0.16147807E-01, 0.36533716E+00, -0.16898610E+00, -0.20461163E-01, 0.10633610E-01, -0.42438932E-03, -0.20938257E-01, -0.21641953E-02, 0.55525000E-02, 
    },
    {
      -0.45814320E-01, 0.26786938E-01, 0.10879262E+00, 0.84931575E-01, 0.15261321E-01, -0.21805953E-01, -0.14455589E-01, 0.77350605E-02, 0.63701227E-01, 
    },
    {
      0.53684928E-01, 0.14995110E-01, -0.27922710E-01, -0.33047084E-01, 0.29755291E-01, -0.18964428E-01, -0.24144307E-01, 0.18880604E-01, 0.58857322E-01, 
    },
    {
      -0.22821718E+00, 0.75741506E+00, 0.14800967E+01, 0.46163443E-01, -0.14732170E+00, -0.15568149E+00, -0.56558043E-01, 0.70298553E-01, 0.11026065E+00, 
    },
    {
      -0.79187071E+00, -0.15295109E+01, 0.65210593E+00, 0.17441598E+00, 0.41551366E-01, -0.66639587E-01, 0.12556066E+00, -0.61385304E-01, 0.27926465E-01, 
    },
    {
      0.87673557E+00, -0.39955699E+00, -0.11206909E+01, -0.28584492E+00, -0.59748320E-02, 0.23729971E+00, -0.19509710E+00, 0.81449501E-01, -0.73962659E+00, 
    },
    {
      0.28284615E+00, 0.89057451E+00, 0.18807463E+00, -0.24254571E+00, -0.17398730E-01, 0.79301950E-02, 0.20122550E+00, 0.40711228E-01, -0.34922609E+00, 
    },
    {
      0.67434180E+00, -0.67298329E+00, -0.74801010E+00, -0.38288945E+00, 0.36076793E+00, 0.47676265E+00, 0.13727280E+00, -0.16624431E+00, -0.43988281E+00, 
    },
    {
      0.26705718E+01, 0.34237185E+00, -0.97559363E+00, -0.27165601E+00, -0.16340867E+00, 0.21915103E+00, -0.12436141E+00, 0.95994301E-01, -0.12902556E+00, 
    },
    {
      -0.24572020E+01, 0.30924758E+00, 0.28079758E+01, 0.50721401E+00, -0.17149813E+00, -0.52098680E+00, 0.85138261E+00, -0.28973010E+00, 0.17749029E+01, 
    },
    {
      -0.11304549E+01, -0.23089054E+01, -0.42302907E+00, 0.84942168E+00, 0.32151359E+00, 0.73436782E-01, -0.39552429E+00, -0.25851825E+00, 0.62756401E+00, 
    },
    {
      -0.39601427E+00, 0.25266504E+00, -0.31376165E+00, 0.29245359E+00, -0.21280552E+00, -0.37739208E+00, -0.73750824E-01, 0.96640058E-01, 0.40330350E+00, 
    },
    {
      -0.21711268E+01, 0.10492510E+01, 0.67013448E+00, 0.88845745E-01, 0.71819268E-01, -0.15133914E+00, -0.11708713E-01, -0.57785516E-02, 0.95103070E-01, 
    },
    {
      0.18162587E+01, 0.38207385E+00, -0.23117478E+01, -0.38997951E+00, 0.19229445E+00, 0.28452805E+00, -0.74004084E+00, 0.22057386E+00, -0.11552705E+01, 
    },
    {
      0.10284771E+01, 0.18343008E+01, 0.33841428E+00, -0.62726456E+00, -0.48801002E+00, -0.87205552E-01, 0.21157789E+00, 0.21693024E+00, -0.32993799E+00, 
    },
    {
      -0.18698256E-01, -0.71124723E-02, -0.16465351E-02, 0.46282008E-01, -0.15294395E+00, -0.13111427E-03, -0.99337585E-02, -0.43608667E-02, -0.90814838E-02, 
    },
    {
      0.45116893E-02, -0.97028855E-02, 0.12505715E-01, 0.16136165E+00, 0.81213906E-01, -0.58312347E-03, -0.18872097E-01, -0.72849258E-02, 0.14784365E-02, 
    },
    {
      0.11196061E+00, 0.20533665E+00, 0.13476495E+00, 0.33945632E+00, -0.11701981E+00, 0.27578942E-01, -0.47568362E-01, 0.25997533E-01, 0.16754409E-01, 
    },
    {
      -0.93068995E-01, 0.18283080E-01, 0.31589814E-01, 0.22644915E+00, 0.42842603E+00, 0.72851159E-01, 0.11030859E-01, 0.23866531E-02, 0.14397724E-01, 
    },
    {
      0.61859138E-01, 0.55517521E-01, -0.38489584E-01, -0.42788506E-01, -0.22592755E+00, -0.99609442E-01, 0.99531204E-01, 0.29087320E-01, 0.16361941E+00, 
    },
    {
      -0.23381135E+00, 0.14897207E-01, -0.29369393E+00, 0.21109748E+00, -0.18685365E+00, -0.38392194E-01, 0.10634159E+00, -0.28300032E-02, 0.66435995E-03, 
    },
    {
      -0.47440812E+00, -0.11250492E+01, -0.27856839E+00, -0.13569555E+01, 0.21064770E+00, -0.44342969E-01, 0.76229990E-01, -0.24290409E-01, -0.11261434E+00, 
    },
    {
      0.10018148E+01, 0.22767542E+00, -0.75611047E-01, -0.59486103E+00, -0.14215121E+01, -0.24381903E+00, -0.25383967E+00, 0.49857941E-01, -0.13947746E+00, 
    },
    {
      -0.18549798E+00, -0.21002911E+00, -0.15347001E-01, -0.16834821E+00, 0.21803890E+00, 0.20929639E+00, -0.10061080E+00, -0.31904507E-01, -0.28207496E+00, 
    },
    {
      0.25608513E+00, -0.18180054E+00, 0.42851421E+00, -0.17312264E+00, -0.17903680E-01, 0.86439624E-01, -0.80150083E-01, 0.17098183E-01, -0.30419683E-01, 
    },
    {
      0.51279235E+00, 0.15700254E+01, 0.14896685E+00, 0.12490706E+01, -0.17830937E+00, 0.75656466E-01, -0.44224098E-01, -0.61300103E-01, 0.12280394E+00, 
    },
    {
      -0.14596163E+01, -0.35159817E+00, 0.27934030E+00, 0.59198815E+00, 0.11541491E+01, 0.17868274E+00, 0.44150430E+00, -0.87764300E-01, 0.17973503E+00, 
    },
    {
      -0.24559209E-01, 0.12929835E-01, 0.68207067E-02, -0.12768010E-01, -0.38065603E-02, 0.26851548E-01, -0.10830905E-01, 0.77535969E-03, -0.17365512E-02, 
    },
    {
      -0.15774000E-01, -0.33989731E-01, -0.11249687E-01, 0.58662505E-02, -0.10439691E-01, 0.19332505E-02, 0.34207720E-01, -0.10329138E-01, -0.22380955E-02, 
    },
    {
      0.55513047E-01, 0.53344555E-01, 0.27011273E-01, -0.84630392E-01, 0.11442303E-01, 0.11869611E-01, 0.18336920E-01, 0.21410547E-02, 0.52863415E-02, 
    },
    {
      0.21707227E-01, -0.38661321E-02, -0.25270691E-01, -0.34020606E-01, -0.14334359E-02, -0.38540121E-02, 0.28737679E-01, 0.37115016E-02, -0.18286736E-02, 
    },
    {
      -0.17681256E-02, -0.14221156E+00, -0.59075002E-01, 0.19836439E-01, -0.76047629E-02, -0.12995811E+00, -0.11560405E+00, -0.16809883E-01, 0.35188191E-01, 
    },
    {
      0.23344200E-01, 0.10418752E+00, -0.43988552E-01, -0.21166109E-01, -0.33218442E-02, 0.14282972E+00, -0.15798105E+00, 0.98103182E-02, 0.10149175E-01, 
    },
    {
      -0.20782379E-01, -0.12256938E-01, 0.90859635E-02, 0.14533856E-01, -0.24522969E-02, 0.11254709E-01, 0.13583394E-02, 0.43356489E-01, 0.26643327E-02, 
    },
    {
      -0.14128673E-01, 0.47900616E-02, -0.10882850E-01, -0.47548087E-02, -0.53821979E-02, 0.31365620E-02, 0.12613773E-01, -0.60953344E-02, 0.45044251E-01, 
    },
    {
      0.63900851E-01, -0.59823156E-02, 0.71507203E-02, -0.16160239E-01, -0.33681545E-01, -0.40392941E-02, 0.60944878E-02, 0.60271841E-01, 0.34214541E-01, 
    },
    {
      0.23164382E-01, -0.68351336E-01, 0.12046548E-01, 0.31958293E-01, 0.36123714E-02, -0.11482013E-01, -0.12627610E-01, -0.46289042E-01, 0.64354017E-01, 
    },
    {
      0.10392379E-02, -0.16863159E-02, 0.73957373E-02, -0.10698938E-02, -0.64807963E-02, 0.99908616E-02, 0.22006985E-02, -0.50359694E-02, -0.14203113E-02, 
    },
    {
      0.12199912E-01, 0.20179294E-01, -0.62138624E-02, 0.78815650E-02, -0.74640587E-02, 0.73885913E-02, 0.22004370E-02, 0.85453289E-02, 0.76826192E-02, 
    },
    {
      0.31520717E-02, -0.12767011E-02, 0.30879308E-02, 0.64275470E-02, -0.43601147E-02, 0.98584695E-02, 0.17043862E-02, 0.84848665E-02, 0.36111218E-02, 
    },
    {
      -0.43704277E-02, -0.78293349E-03, -0.17360970E-01, -0.11529316E-02, 0.11252892E-01, -0.21595629E-02, -0.14455176E-02, 0.98986849E-04, 0.87585188E-02, 
    },
  },
  {
    {
      0.26772809E+01, -0.86198850E-02, 0.94552236E-02, 0.55961576E-02, 0.43859920E-02, 0.41296384E-02, 0.46669282E-02, -0.20577968E-02, -0.24387249E-03, 
    },
    {
      -0.65158509E-01, 0.74515760E-01, 0.72828352E-01, 0.12489909E-01, 0.58226571E-02, -0.11427455E-01, -0.15949465E-01, -0.65154112E-02, 0.17882379E-02, 
    },
    {
      0.12957149E+01, 0.13960780E+00, -0.44050530E-01, -0.51063329E-01, -0.63741840E-01, -0.59596483E-01, -0.53228300E-01, 0.46188686E-01, 0.11657926E-01, 
    },
    {
      -0.15501007E+00, -0.51885819E+00, -0.24664447E+00, -0.34158930E-01, -0.38116850E-01, 0.12921122E-01, 0.49244039E-01, 0.17124636E-01, -0.53521986E-02, 
    },
    {
      -0.20537572E+01, -0.26769561E+00, 0.49208454E-02, 0.11124684E+00, 0.16032118E+00, 0.16892996E+00, 0.16521209E+00, -0.12701942E+00, -0.60623229E-01, 
    },
    {
      0.17818429E+00, 0.42037153E+00, 0.23076217E+00, 0.26306644E-01, 0.27854430E-01, -0.14802126E-02, -0.40314917E-01, -0.73871380E-02, 0.24690884E-02, 
    },
    {
      0.82360959E+00, 0.18576214E+00, 0.74848384E-01, -0.66371389E-01, -0.10588888E+00, -0.11814051E+00, -0.11794991E+00, 0.87646529E-01, 0.56451142E-01, 
    },
    {
      0.19192193E-01, -0.34562644E+00, -0.26493758E+00, 0.78390101E-02, -0.45823748E-02, 0.45295575E-03, 0.74164444E-02, -0.11921156E-02, 0.17643599E-02, 
    },
    {
      0.59155165E-04, 0.28991437E+00, -0.32774398E+00, -0.21361131E-01, 0.55374289E-02, 0.53592436E-02, -0.18224832E-01, -0.22406343E-02, 0.48251497E-02, 
    },
    {
      0.43235659E-02, 0.13838615E+00, 0.46312861E-01, 0.21590933E-01, 0.16815718E-01, -0.19970588E-01, 0.64824068E-04, -0.19745342E-03, 0.47364235E-01, 
    },
    {
      0.35681468E-01, 0.49083747E-01, 0.10480716E+00, 0.58407634E-02, 0.63248561E-02, 0.33869767E-02, -0.22746570E-01, 0.86412365E-02, 0.40659290E-01, 
    },
    {
      -0.36743322E+00, 0.11957974E+01, 0.91784447E+00, 0.13865001E+00, -0.27913701E-01, -0.31462729E-01, -0.16293770E-01, 0.48463598E-01, -0.73108380E-03, 
    },
    {
      -0.61130720E+00, -0.93153477E+00, 0.12323002E+01, 0.17962562E+00, 0.92122052E-02, -0.88324815E-01, 0.12294372E+00, -0.67933217E-01, -0.18161764E-01, 
    },
    {
      0.24850307E+00, -0.11966048E+01, -0.56054062E+00, -0.38083974E-01, -0.12794180E+00, 0.15602164E+00, -0.32316092E+00, 0.64251274E-01, -0.52134240E+00, 
    },
    {
      0.38644767E+00, 0.52516818E+00, -0.71278828E+00, -0.22031011E+00, 0.48846070E-01, -0.63622043E-01, 0.20911933E+00, 0.87847181E-01, -0.20944715E+00, 
    },
    {
      0.10246786E+01, -0.12937899E+01, -0.40522784E+00, -0.58681512E+00, 0.13400623E+00, 0.13890813E+00, 0.50930053E-01, -0.12004627E-01, -0.18637219E-01, 
    },
    {
      0.22995977E+01, -0.29405490E+00, -0.19752228E+01, -0.36606318E+00, -0.36992472E-01, 0.18337315E+00, -0.16064256E+00, 0.12423857E+00, 0.50655352E-02, 
    },
    {
      -0.79235864E+00, 0.22702663E+01, 0.19082406E+01, 0.38968597E-01, 0.33631064E-01, -0.34791511E+00, 0.11422637E+01, -0.18621103E+00, 0.13048165E+01, 
    },
    {
      -0.16866664E+01, -0.17086961E+01, 0.15004869E+01, 0.64295954E+00, 0.89812361E-01, 0.19795394E+00, -0.47849467E+00, -0.32040602E+00, 0.35970518E+00, 
    },
    {
      -0.68489271E+00, 0.76091516E+00, 0.12355461E+00, 0.44687083E+00, -0.14212961E+00, -0.13827266E+00, -0.91967387E-02, -0.65421730E-01, 0.32495581E-01, 
    },
    {
      -0.19574792E+01, 0.92517078E+00, 0.13603250E+01, 0.23316291E+00, -0.27497089E-01, -0.11032374E+00, 0.18738519E-01, -0.33233728E-01, -0.10697253E-01, 
    },
    {
      0.59273267E+00, -0.12416369E+01, -0.19759161E+01, 0.18526608E-01, 0.10760016E+00, 0.19402064E+00, -0.95974118E+00, 0.11718221E+00, -0.88802713E+00, 
    },
    {
      0.15903203E+01, 0.16918579E+01, -0.10219685E+01, -0.45557439E+00, -0.17555867E+00, -0.13787279E+00, 0.30494553E+00, 0.23491570E+00, -0.17701098E+00, 
    },
    {
      -0.17107019E-01, -0.11062065E-01, 0.93461676E-02, 0.72881304E-01, -0.11495988E+00, -0.24580231E-02, -0.34615425E-02, -0.27623498E-02, -0.68666781E-02, 
    },
    {
      -0.95973676E-02, -0.17075194E-01, 0.91627231E-02, 0.10242627E+00, 0.92916168E-01, 0.24848136E-02, -0.11218095E-01, -0.55938442E-02, 0.10607603E-03, 
    },
    {
      0.10907742E+00, 0.13264808E+00, 0.73107131E-01, 0.13814703E+00, 0.24720404E-01, 0.25445605E-01, -0.19353466E-01, -0.68268122E-03, 0.79990206E-02, 
    },
    {
      0.44637113E-02, 0.49218766E-01, -0.18194420E-01, 0.33405833E-01, 0.20319274E+00, 0.36236916E-01, 0.17274117E-01, -0.63722855E-02, 0.33721287E-01, 
    },
    {
      0.25666658E-01, 0.92364043E-01, -0.91084383E-01, -0.36218074E+00, 0.19792330E+00, -0.52057378E-01, 0.40467951E-01, -0.31956337E-01, 0.12260315E+00, 
    },
    {
      -0.15529650E+00, 0.11941776E+00, -0.13577776E+00, -0.17587660E+00, -0.45683962E+00, -0.65770708E-01, 0.69410942E-01, -0.18389063E-01, 0.85385451E-02, 
    },
    {
      -0.37895921E+00, -0.73540902E+00, -0.56422465E-01, -0.47111732E+00, -0.20011030E+00, -0.17941417E-01, 0.27807336E-01, 0.76329634E-01, -0.42905848E-01, 
    },
    {
      0.33812630E+00, -0.11257858E+00, 0.52691784E-01, 0.99168252E-02, -0.50797993E+00, -0.69310442E-01, -0.12250749E+00, 0.10490370E+00, -0.24429388E+00, 
    },
    {
      -0.96905559E-01, -0.29989800E+00, 0.95130719E-01, 0.83204031E-01, -0.46254903E+00, 0.10771059E+00, -0.16102755E-01, 0.71079575E-01, -0.19646353E+00, 
    },
    {
      0.21520536E+00, -0.41078597E+00, 0.15975517E+00, 0.48532072E+00, 0.16240959E+00, 0.74154310E-01, -0.57246845E-01, 0.45167182E-01, -0.30693915E-01, 
    },
    {
      0.27020487E+00, 0.10784189E+01, -0.53013716E-01, 0.46552581E+00, 0.92620172E-01, 0.51729448E-01, -0.36019329E-01, -0.14987502E+00, 0.62644877E-01, 
    },
    {
      -0.55153680E+00, 0.28454646E+00, 0.94641276E-01, 0.93170024E-01, 0.38448083E+00, 0.47209587E-01, 0.20970257E+00, -0.13942525E+00, 0.29898921E+00, 
    },
    {
      -0.25534306E-01, 0.54413616E-02, 0.14065277E-01, -0.76298043E-02, -0.10420343E-01, 0.29778404E-01, -0.41098725E-01, -0.40198048E-02, -0.51609301E-02, 
    },
    {
      -0.10902795E-01, -0.23414372E-01, -0.62101181E-02, 0.97503457E-02, -0.91798007E-02, 0.33085518E-01, 0.36749907E-01, -0.47427085E-02, -0.77581666E-02, 
    },
    {
      0.33572678E-01, 0.60052566E-01, 0.14006686E-01, -0.36195546E-01, 0.13402758E-01, 0.33575904E-01, 0.35400663E-02, -0.38376546E-02, 0.42043719E-02, 
    },
    {
      0.25882047E-01, 0.10295448E-01, -0.20480629E-02, -0.22618541E-01, -0.37360494E-02, 0.12422180E-02, 0.29957978E-01, 0.19331203E-02, 0.71227252E-02, 
    },
    {
      0.44606984E-01, -0.10304903E+00, -0.50200410E-01, -0.11642319E-02, 0.27468766E-02, -0.13490731E+00, -0.85447587E-01, -0.19251921E-02, 0.38483702E-01, 
    },
    {
      0.11273085E-01, 0.55817258E-01, -0.40558748E-01, -0.12057238E-02, 0.39492734E-02, 0.10076770E+00, -0.14639939E+00, -0.27176535E-02, 0.11188758E-01, 
    },
    {
      -0.19127030E-01, -0.16049713E-01, 0.92721246E-02, 0.23639505E-02, -0.56997489E-03, 0.52273925E-02, 0.54570334E-03, 0.32571722E-01, 0.92297168E-02, 
    },
    {
      -0.76535931E-02, 0.41365675E-02, -0.10613388E-01, 0.34963775E-02, -0.13046856E-02, 0.11616315E-02, 0.74619632E-02, -0.10147394E-01, 0.32333374E-01, 
    },
    {
      0.38845401E-01, 0.55554281E-02, -0.41929283E-02, -0.15542838E-01, -0.18608436E-01, -0.68834010E-02, 0.12991249E-01, 0.47244698E-01, 0.51083591E-01, 
    },
    {
      0.33533398E-01, -0.61769698E-01, 0.22310423E-01, 0.10246638E-01, -0.62199375E-02, -0.11598877E-01, -0.16985392E-01, -0.50428756E-01, 0.41603886E-01, 
    },
    {
      0.17688370E-02, -0.36227442E-02, 0.84820492E-02, 0.28933492E-02, -0.58887932E-02, 0.68608033E-02, 0.30614494E-03, -0.15316576E-02, -0.24033547E-02, 
    },
    {
      0.92781493E-02, 0.14051887E-01, -0.10239172E-01, 0.53211111E-02, -0.21397618E-02, 0.44980948E-02, 0.54727206E-02, 0.70043919E-02, 0.11711870E-01, 
    },
    {
      0.68015773E-02, 0.21883957E-02, 0.35151727E-02, 0.67476844E-02, -0.67525292E-02, 0.54354891E-02, 0.19866449E-02, 0.76344414E-02, 0.59919464E-02, 
    },
    {
      -0.26257581E-02, -0.48037629E-04, -0.10798140E-01, 0.56925311E-03, 0.80262767E-02, -0.51455793E-03, -0.93577243E-03, -0.71163801E-03, 0.78903772E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir15_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.56421361E+01, -0.81572428E-01, 0.20138618E+00, 0.42779669E-01, 0.44377025E-01, 0.38025160E-02, -0.87150786E-03, 0.21708569E-01, -0.23025220E-01, -0.33594195E-02, 0.21446483E-03, -0.18263943E-02, -0.12104558E-01, 
    },
    {
      -0.10026588E+01, -0.77602589E+00, 0.80115181E+00, -0.27182043E+00, 0.54614949E+00, -0.18997940E+00, 0.82413152E-01, -0.38998052E-01, -0.97885408E-01, 0.13837354E+00, 0.10652158E+00, -0.15692274E+00, 0.99289812E-01, 
    },
    {
      0.80236692E+01, 0.19602764E+01, -0.35586731E+01, 0.28472129E+00, -0.56597894E+00, -0.44183967E+00, 0.13710301E+00, 0.35544112E-01, 0.41852084E+00, 0.63865769E+00, -0.19955039E+00, 0.11715313E+00, 0.86535461E-01, 
    },
    {
      0.14375471E+02, 0.64524765E+01, -0.62800088E+01, 0.55624423E+01, -0.81383467E+01, 0.28389122E+01, -0.18190529E+01, 0.17703345E+00, 0.13019991E+01, -0.19010282E+01, -0.25744607E+01, 0.22809548E+01, -0.81978828E+00, 
    },
    {
      -0.53995434E+02, -0.20322048E+02, 0.18375986E+02, -0.74895936E+00, -0.51071935E+01, 0.81174737E+00, -0.39087531E+00, -0.14415859E+01, -0.10194235E+01, -0.38520768E+01, 0.11969794E+01, -0.14105285E+01, -0.62901032E+00, 
    },
    {
      -0.18940151E+02, -0.23667875E+02, 0.88806610E+01, -0.27459188E+02, 0.41921669E+02, -0.13132250E+02, 0.69411273E+01, -0.15238301E+01, -0.77653646E+01, 0.84578733E+01, 0.15979069E+02, -0.12099964E+02, 0.12503768E+01, 
    },
    {
      0.11738895E+03, 0.68091232E+02, -0.39820190E+02, -0.16654693E+01, 0.25928150E+02, 0.13787634E+01, -0.43914926E+00, 0.40016642E+01, 0.47450104E+00, 0.83207235E+01, -0.34342322E+01, 0.43038507E+01, 0.21257403E+01, 
    },
    {
      -0.37460907E+02, 0.41022415E+02, 0.10812599E+02, 0.51976315E+02, -0.94334923E+02, 0.25150223E+02, -0.89467859E+01, 0.58774261E+01, 0.20332344E+02, -0.17085917E+02, -0.38310886E+02, 0.27109829E+02, 0.87873173E+00, 
    },
    {
      -0.12372593E+03, -0.87597099E+02, 0.38698399E+02, 0.42182999E+01, -0.37059380E+02, -0.34062238E+01, 0.15188286E+01, -0.37199290E+01, 0.36815086E+00, -0.74235358E+01, 0.42371931E+01, -0.49250031E+01, -0.28798954E+01, 
    },
    {
      0.87714088E+02, -0.30394653E+02, -0.28824894E+02, -0.42016949E+02, 0.95767807E+02, -0.21253895E+02, 0.34847987E+01, -0.83288565E+01, -0.22973545E+02, 0.15845953E+02, 0.39268745E+02, -0.26755207E+02, -0.26882515E+01, 
    },
    {
      0.49962570E+02, 0.37968533E+02, -0.13776732E+02, -0.21174350E+01, 0.16793774E+02, 0.16195391E+01, -0.80588609E+00, 0.10947104E+01, -0.22045150E+00, 0.22938368E+01, -0.17940499E+01, 0.19093058E+01, 0.13130230E+01, 
    },
    {
      -0.43963936E+02, 0.73540716E+01, 0.14685716E+02, 0.12209099E+02, -0.35778717E+02, 0.65739970E+01, 0.25819159E+00, 0.38785126E+01, 0.92260513E+01, -0.54415417E+01, -0.14486173E+02, 0.96487923E+01, 0.12716563E+01, 
    },
    {
      -0.21124800E+00, 0.17043127E+01, 0.20464387E+01, 0.56599015E-02, -0.10268541E+00, 0.23923576E-01, 0.46511523E-01, 0.47292676E-01, -0.43041897E-02, -0.13252716E-01, 0.11830020E-02, 0.10728045E-01, -0.11813324E-01, 
    },
    {
      0.11713128E+00, -0.18258561E+01, 0.17008826E+01, -0.15215686E-01, 0.11784354E+00, -0.43866588E-02, -0.22057734E-01, -0.63331686E-02, 0.49470387E-01, -0.17963713E-01, -0.59963893E-02, -0.16910573E-01, 0.36876634E-01, 
    },
    {
      -0.67201293E+00, 0.13338754E+00, -0.21074166E+01, 0.51338410E+00, -0.38658473E+00, 0.81445105E-01, -0.95247170E-02, -0.41881129E+00, 0.26037970E+00, 0.10410904E+00, 0.31723064E+00, 0.19097517E+00, 0.45013998E-01, 
    },
    {
      -0.20128696E+00, 0.15844090E+01, -0.83317590E+00, 0.54410454E-01, 0.24460571E+00, -0.36854930E-01, -0.48869914E+00, 0.82281917E-01, 0.58834499E+00, 0.92693098E-01, 0.62349554E-01, 0.23411550E+00, 0.29134646E+00, 
    },
    {
      0.55470839E+01, 0.96163158E+01, 0.50997686E+01, 0.68665922E+00, 0.65423720E-01, -0.24311846E+00, -0.12406635E+01, -0.11371589E+01, 0.11067604E+01, -0.51531062E-01, -0.74737412E+00, 0.54248173E-01, 0.63131139E-01, 
    },
    {
      -0.68635064E+00, -0.12532818E+02, 0.31381702E+01, 0.58695197E+00, -0.89948498E-01, -0.52485317E+00, 0.11644162E+01, -0.24193650E+00, 0.99850368E+00, 0.73659611E+00, 0.30205196E+00, 0.79808694E+00, -0.56268388E+00, 
    },
    {
      0.84152431E+01, -0.18351845E+01, 0.22851471E+02, -0.62919526E+01, 0.38386998E+00, -0.11887341E+01, -0.32209761E+01, 0.62786980E+01, -0.32539845E+01, 0.33874832E-01, -0.68638234E+01, -0.39917052E+01, 0.50753254E+00, 
    },
    {
      0.54675989E+01, -0.14342614E+02, 0.20977676E+02, -0.40119495E-01, -0.43394451E+01, 0.13764828E+01, 0.10244990E+02, 0.31333181E+00, -0.92915649E+01, -0.30012243E+01, -0.67161626E+00, -0.32468977E+01, -0.40313520E+01, 
    },
    {
      -0.35370174E+02, -0.69950500E+02, 0.26502401E+02, 0.31762896E+01, -0.11033854E+01, -0.33225198E+01, 0.66381388E+01, 0.81779928E+01, -0.61022053E+01, -0.22130617E+00, 0.75198293E+01, 0.25410479E+00, 0.76021153E+00, 
    },
    {
      -0.13441768E+01, 0.27343557E+02, 0.47622552E+01, -0.13991564E+02, 0.15639170E+01, 0.82513838E+01, -0.82063255E+01, -0.34090364E+01, -0.11663890E+02, -0.46683278E+01, -0.29037738E+01, -0.57367353E+01, 0.13532746E+01, 
    },
    {
      -0.48274441E+02, 0.83131844E+02, -0.96900970E+02, 0.27448704E+02, 0.21074257E+02, 0.43562355E+01, 0.23491089E+02, -0.30603300E+02, 0.98027277E+01, -0.17388166E+01, 0.41459393E+02, 0.21956772E+02, -0.74577127E+01, 
    },
    {
      -0.57677406E+02, 0.61015911E+02, -0.67599236E+02, -0.76322885E+01, 0.15360656E+02, -0.13954189E+02, -0.63971962E+02, -0.57594633E+00, 0.46648548E+02, 0.18534496E+02, -0.11813200E+01, 0.14227717E+02, 0.16818104E+02, 
    },
    {
      0.80797073E+02, 0.17325121E+03, -0.21808232E+03, -0.25066273E+02, 0.77090278E+01, 0.19229862E+02, -0.16147860E+02, -0.21793207E+02, 0.11739250E+02, 0.27427473E+01, -0.23950100E+02, -0.14955986E+01, -0.43279033E+01, 
    },
    {
      0.23412638E+02, 0.63171879E+02, -0.99681198E+02, 0.62766190E+02, -0.79878149E+01, -0.30435446E+02, 0.21497587E+02, 0.20054741E+02, 0.38359371E+02, 0.92688770E+01, 0.10108423E+02, 0.14417062E+02, -0.20493574E+01, 
    },
    {
      0.14545422E+03, -0.31993140E+03, 0.20626822E+03, -0.59260872E+02, -0.98615173E+02, -0.91029177E+01, -0.61238270E+02, 0.69641464E+02, -0.53348732E+01, 0.55038424E+01, -0.10444603E+03, -0.51997677E+02, 0.24604845E+02, 
    },
    {
      0.17732971E+03, -0.14906339E+03, 0.97182152E+02, 0.28640442E+02, -0.17308891E+02, 0.45831375E+02, 0.16195528E+03, -0.62566624E+01, -0.10428041E+03, -0.45196327E+02, 0.11912194E+02, -0.28687082E+02, -0.28267488E+02, 
    },
    {
      -0.71918938E+02, -0.20264278E+03, 0.37785242E+03, 0.43567059E+02, -0.15192982E+02, -0.31795984E+02, 0.19159107E+02, 0.24578568E+02, -0.10395302E+02, -0.54584193E+01, 0.29443462E+02, 0.18643064E+01, 0.73603716E+01, 
    },
    {
      -0.46939167E+02, -0.19401114E+03, 0.19549149E+03, -0.98742599E+02, 0.65470691E+01, 0.41411484E+02, -0.23003340E+02, -0.32262764E+02, -0.48849609E+02, -0.63948112E+01, -0.14122622E+02, -0.15711078E+02, 0.33360174E+01, 
    },
    {
      -0.20310941E+03, 0.43382684E+03, -0.22873076E+03, 0.61049896E+02, 0.15113927E+03, 0.12178139E+02, 0.69404495E+02, -0.76307945E+02, -0.99914246E+01, -0.87610970E+01, 0.11714094E+03, 0.56842529E+02, -0.31153524E+02, 
    },
    {
      -0.20313609E+03, 0.19473550E+03, -0.81566399E+02, -0.31190516E+02, 0.46536431E+01, -0.60174667E+02, -0.17905360E+03, 0.15700250E+02, 0.10812027E+03, 0.48076645E+02, -0.19411180E+02, 0.28377907E+02, 0.18879490E+02, 
    },
    {
      0.18400129E+02, 0.92124519E+02, -0.19595229E+03, -0.22818579E+02, 0.91224928E+01, 0.16459476E+02, -0.86445189E+01, -0.10113220E+02, 0.36470227E+01, 0.31096530E+01, -0.12310074E+02, -0.65811318E+00, -0.39445660E+01, 
    },
    {
      0.27785633E+02, 0.12070577E+03, -0.10845207E+03, 0.50968784E+02, 0.20172384E+01, -0.18991886E+02, 0.84511404E+01, 0.16190018E+02, 0.21269348E+02, 0.98832011E+00, 0.67662487E+01, 0.63501000E+01, -0.22222323E+01, 
    },
    {
      0.10348282E+03, -0.20003775E+03, 0.10056244E+03, -0.24191299E+02, -0.75551025E+02, -0.66165800E+01, -0.28941280E+02, 0.32210011E+02, 0.89458332E+01, 0.52425108E+01, -0.48151104E+02, -0.23473763E+02, 0.13704450E+02, 
    },
    {
      0.76933624E+02, -0.97199905E+02, 0.34218758E+02, 0.94315853E+01, -0.39759991E+00, 0.27437757E+02, 0.72066147E+02, -0.95865564E+01, -0.42264874E+02, -0.18651270E+02, 0.93191538E+01, -0.11183516E+02, -0.35364709E+01, 
    },
    {
      0.97871020E-01, 0.20281778E+00, 0.90489715E-01, -0.95290929E+00, -0.27933320E+00, 0.52461393E-01, 0.72199428E-02, 0.15705125E-01, 0.41039664E-01, -0.36827065E-01, -0.15596761E-01, -0.21006119E-01, -0.74302865E-03, 
    },
    {
      0.18413863E+00, -0.22145297E-01, -0.21845268E-01, 0.28004029E+00, -0.71528941E+00, -0.12112075E-01, -0.43858774E-02, -0.32223654E-02, -0.29513298E-02, -0.20680975E-01, -0.50281603E-02, -0.97669261E-02, 0.16754756E-01, 
    },
    {
      -0.50228141E-01, 0.86401850E+00, 0.37329832E+00, -0.43455347E-01, -0.37337765E+00, 0.31667180E-01, -0.33294806E+00, 0.19361353E+00, -0.14481428E+00, 0.17382690E+00, -0.27535016E-01, -0.52372944E-01, 0.51665086E-01, 
    },
    {
      0.77041340E+00, -0.10802412E+00, 0.78209734E+00, 0.82496119E+00, -0.11585006E+00, -0.95668048E-01, 0.54860033E-01, 0.16625006E+00, -0.91622919E-01, -0.65615319E-01, 0.60415332E-03, 0.15374319E+00, 0.46706021E-01, 
    },
    {
      0.54401726E+00, -0.93265307E+00, -0.56378675E+00, 0.10645507E+02, 0.57229429E+00, -0.12630768E+01, -0.64789832E+00, -0.60353708E+00, -0.18424869E+00, 0.13993631E+00, 0.48576195E-01, 0.22784832E+00, -0.42130151E+00, 
    },
    {
      0.27307931E+00, 0.10153091E+01, 0.11288965E+01, -0.25885277E+01, 0.76768188E+01, 0.73958021E+00, 0.39947239E+00, -0.11303554E+01, 0.74285984E+00, 0.42960650E+00, -0.46359366E-02, 0.57497424E+00, -0.24941255E+00, 
    },
    {
      -0.27576047E+00, -0.49189267E+01, -0.30248418E+01, 0.82247725E+01, -0.40766196E+01, 0.37377694E+00, 0.15920104E+01, -0.10867124E+01, 0.17805321E+01, -0.15497746E+01, 0.73062021E+00, 0.45829788E-01, 0.16331393E+00, 
    },
    {
      -0.90494785E+01, -0.15745066E+00, -0.52452645E+01, -0.91057497E+00, 0.86765213E+01, 0.23414004E+01, -0.92353123E+00, -0.28763181E+00, 0.22783259E-01, -0.82801855E+00, -0.28931600E+00, -0.10174736E+01, 0.16624613E+00, 
    },
    {
      -0.10378919E+02, 0.58001833E+01, -0.22736102E+00, -0.23410555E+02, 0.69524856E+01, 0.71683693E+01, 0.60270047E+01, 0.22774773E+01, 0.14830394E+01, 0.16389118E+01, -0.30775643E-02, -0.78709882E+00, 0.30579998E+01, 
    },
    {
      -0.15465743E-02, -0.94203407E+00, -0.70205545E+01, 0.62845588E+01, -0.72446790E+01, -0.44833145E+01, -0.13172768E+01, 0.54477401E+01, -0.26917975E+01, -0.22063324E+01, 0.88429081E+00, -0.25067618E+01, 0.70779270E+00, 
    },
    {
      -0.41653025E+00, 0.77432489E+01, 0.44009709E+01, -0.26288893E+02, 0.11670718E+02, -0.90083349E+00, -0.26713226E+01, 0.27358816E+01, -0.44873104E+01, 0.38639853E+01, -0.20221062E+01, 0.50795317E+00, -0.10303019E+01, 
    },
    {
      0.25454908E+02, -0.35198110E+00, 0.91766758E+01, 0.36100191E+00, -0.26752708E+02, -0.79657893E+01, 0.40400887E+01, -0.13829775E+00, 0.47660896E+00, 0.37345989E+01, 0.49525353E+00, 0.16495343E+01, -0.32004577E+00, 
    },
    {
      0.26923609E+02, -0.12705561E+02, 0.77154827E+01, 0.47243571E+01, -0.21181990E+02, -0.15098260E+02, -0.13009450E+02, -0.24404039E+01, -0.28010442E+01, -0.60208449E+01, -0.58497292E+00, 0.13716583E+01, -0.60833530E+01, 
    },
    {
      -0.75704546E+01, -0.52916546E+01, 0.16575653E+02, -0.68116546E+01, -0.28722118E+02, 0.83101597E+01, 0.55782187E+00, -0.92281694E+01, 0.28159809E+01, 0.46327081E+01, -0.30307541E+01, 0.37459197E+01, -0.88170326E+00, 
    },
    {
      0.22301376E+00, -0.31962299E+01, -0.10003719E+01, 0.17513151E+02, -0.10250562E+02, 0.30598539E+00, 0.16885284E+01, -0.19954596E+01, 0.31069324E+01, -0.28071928E+01, 0.13887666E+01, -0.67095476E+00, 0.10415512E+01, 
    },
    {
      -0.20710796E+02, 0.89437598E+00, -0.42360606E+01, 0.20117793E+01, 0.17224691E+02, 0.66144185E+01, -0.45690055E+01, -0.18276887E+00, -0.40452820E+00, -0.32121379E+01, -0.92587329E-01, -0.48688596E+00, -0.14324804E+00, 
    },
    {
      -0.18712111E+02, 0.78072848E+01, -0.87452765E+01, 0.12498962E+02, 0.17300501E+02, 0.10375720E+02, 0.82283955E+01, 0.58782387E+00, 0.11830972E+01, 0.48081841E+01, 0.77937973E+00, -0.73815542E+00, 0.35488698E+01, 
    },
    {
      0.98545418E+01, 0.57349968E+01, -0.14113446E+02, 0.14990284E+01, 0.34328514E+02, -0.51231594E+01, 0.11489596E+01, 0.57513089E+01, -0.11141500E+01, -0.30930831E+01, 0.25668890E+01, -0.20892141E+01, 0.63647670E+00, 
    },
    {
      -0.29918633E-01, 0.19186273E+00, -0.70395455E-01, -0.77079989E-01, 0.31567018E-01, -0.38534543E+00, -0.50261647E+00, -0.21860762E-01, 0.62904462E-01, -0.54080831E-02, -0.92705982E-02, -0.40175516E-01, -0.11157575E-01, 
    },
    {
      0.72898664E-01, -0.50168093E-01, -0.10570346E+00, -0.67150467E-02, -0.14133672E+00, 0.50753140E+00, -0.42581895E+00, -0.46848822E-01, 0.22408159E-02, 0.83338562E-02, 0.41946103E-02, -0.58955308E-02, -0.26281827E-03, 
    },
    {
      -0.26040110E+00, 0.76159227E+00, 0.23537507E-01, -0.80746651E-01, -0.17011431E+00, 0.11547460E+00, -0.15361232E+00, 0.17724212E+00, 0.76225460E-01, -0.13381564E-01, -0.60893386E-02, -0.11518117E+00, -0.31129291E-01, 
    },
    {
      -0.51667415E-01, -0.24441694E+00, -0.31773397E+00, 0.81835248E-01, -0.46549350E+00, 0.17142838E+00, 0.30523732E+00, -0.57707924E-01, -0.20797607E-01, -0.15126070E-01, -0.33176012E-01, -0.42346753E-01, -0.11050799E-01, 
    },
    {
      0.59186459E-01, -0.11252819E+01, 0.44319230E+00, -0.11809949E+00, -0.14963280E+00, 0.52666683E-01, -0.52684325E+00, -0.40326202E+00, 0.12269677E+00, 0.28846711E+00, 0.12978290E+00, 0.32258502E+00, -0.14862968E+00, 
    },
    {
      -0.80570722E+00, 0.17221979E+01, 0.46162105E+00, -0.34327608E+00, 0.13253006E+01, 0.44412488E+00, 0.28846559E+00, -0.34307161E+00, -0.36704317E+00, 0.56617562E-01, 0.18177399E+00, 0.25099775E+00, -0.99185131E-01, 
    },
    {
      -0.19901596E+00, -0.17126943E+01, 0.37629980E+00, 0.45409298E+00, 0.27510104E+00, 0.15234837E-01, 0.65682179E+00, -0.45590812E+00, -0.21512812E+00, 0.90877265E-01, -0.33986196E-01, 0.28594887E+00, 0.92194878E-01, 
    },
    {
      -0.24209818E+00, 0.65725392E+00, 0.75583166E+00, -0.39450571E+00, 0.12069463E+01, -0.75257272E+00, -0.52532136E+00, 0.99424221E-01, 0.93963951E-01, 0.45424551E-02, 0.14965154E+00, 0.19529074E+00, 0.89346744E-01, 
    },
    {
      0.76194292E+00, 0.21998913E+01, -0.61996305E+00, 0.38401404E+00, 0.65990424E+00, 0.10685669E+01, 0.16874920E+01, 0.10111738E+01, -0.47734886E+00, -0.55955160E+00, -0.36574565E-01, -0.54719472E+00, 0.21022716E+00, 
    },
    {
      0.96725523E+00, -0.21709857E+01, -0.58654177E+00, 0.52371088E-01, -0.19939440E+01, -0.13083379E+01, 0.97265768E+00, 0.10472490E+01, 0.79788148E+00, -0.23003300E+00, -0.31732571E+00, -0.59758753E+00, 0.17752629E-01, 
    },
    {
      -0.48324671E-01, -0.39871577E-01, 0.35558338E-02, -0.26979163E-01, 0.43359697E-01, 0.12917671E-01, 0.17010419E-01, 0.27531299E+00, -0.38564362E-01, -0.34878931E-02, -0.61485823E-02, -0.97274072E-02, -0.18774368E-01, 
    },
    {
      0.92524938E-01, -0.57997428E-01, 0.36104083E-01, 0.66196665E-01, -0.40818259E-01, 0.68578757E-02, 0.56706844E-02, 0.14657995E-01, 0.25169230E+00, 0.17018361E-01, -0.13085437E-02, 0.40550678E-03, -0.75565781E-02, 
    },
    {
      -0.27891763E-01, -0.61812479E-01, -0.66995524E-01, 0.60951114E-01, -0.16235442E+00, -0.18117538E+00, -0.63721985E-01, 0.19130240E+00, 0.34930453E+00, 0.10013775E-03, -0.18040525E-01, 0.93262065E-02, 0.29071178E-01, 
    },
    {
      0.37352061E+00, 0.23084080E+00, -0.11973947E+00, 0.39940480E-01, 0.16606483E+00, -0.25743233E-01, -0.50882917E-01, -0.38867268E+00, 0.90072393E-01, -0.18342920E-01, -0.16830478E-01, 0.65812245E-02, -0.11528600E-01, 
    },
    {
      -0.17556234E+00, -0.39191529E-01, 0.14456748E-01, -0.31045396E-01, -0.29462801E-01, 0.17225457E-02, 0.81582461E-02, 0.32372467E-01, -0.37683394E-01, 0.90681992E-01, 0.16984081E+00, -0.56379475E-02, -0.16191302E-01, 
    },
    {
      0.32786489E-02, 0.13823365E-01, -0.58228850E-01, 0.46045277E-01, -0.32285951E-01, -0.29173903E-01, -0.13254940E-02, 0.33482354E-01, 0.14193523E-01, -0.18655913E+00, 0.75768754E-01, 0.28235951E-01, -0.12667628E-01, 
    },
    {
      -0.63085563E-01, -0.16578859E-01, -0.65637636E-02, 0.90851039E-02, -0.29999416E-01, 0.40609598E-01, -0.12365533E-01, 0.32348711E-01, 0.48897574E-02, -0.33822318E-02, -0.95927659E-02, -0.10472462E+00, -0.49497215E-02, 
    },
    {
      0.13566672E-01, 0.51126843E-02, 0.69563393E-02, 0.21742018E-01, 0.64151838E-01, -0.15110991E-01, -0.11155133E-01, -0.23500491E-02, -0.49487525E-02, -0.34136429E-01, -0.13634215E-01, -0.51828269E-02, -0.11964961E+00, 
    },
    {
      0.16280914E-01, -0.15323651E-01, 0.11045886E-01, 0.15388856E-01, 0.71778158E-02, 0.13341840E-01, 0.86160125E-02, 0.35415865E-01, 0.20430239E-01, -0.12541118E-01, 0.64188964E-02, 0.69954526E-02, 0.14519762E-01, 
    },
    {
      0.26640084E-01, -0.81357203E-01, -0.82582489E-01, 0.33618689E-02, -0.22538289E-01, -0.19978173E-01, 0.16238056E-01, -0.12985623E-01, 0.16977180E-01, -0.14756378E-01, -0.48074991E-03, 0.21765309E-02, 0.14442624E-01, 
    },
    {
      0.29858338E-01, 0.81244618E-01, -0.43053649E-01, 0.27717860E-01, 0.12420647E-01, -0.21559792E-01, 0.41899722E-01, -0.27755309E-01, 0.32173231E-01, -0.28853366E-01, 0.96155703E-02, 0.35324396E-03, -0.10962152E-01, 
    },
    {
      -0.35660431E-01, -0.20827377E-01, -0.27171995E-01, -0.39691202E-01, 0.78898314E-02, -0.13305612E-01, -0.33514686E-02, -0.12471182E-01, 0.22468641E-01, 0.17353624E-01, -0.10118406E-01, -0.17869116E-02, 0.15280945E-01, 
    },
  },
  {
    {
      0.90913858E+01, 0.84599406E-01, 0.16230763E+00, 0.69487393E-01, -0.29570591E-02, -0.92661269E-02, -0.40300194E-01, 0.30152716E-01, -0.22117848E-01, 0.44199214E-02, -0.37470509E-02, 0.57405089E-02, -0.12053542E-01, 
    },
    {
      -0.76227540E+00, -0.81212771E+00, 0.16045764E-01, -0.18294992E-01, 0.17572674E+00, 0.21314566E+00, 0.27046995E-01, 0.82527503E-01, -0.28845841E+00, 0.24496240E-02, 0.97445555E-01, -0.39677333E-01, 0.67152999E-01, 
    },
    {
      0.24650427E+02, -0.44397229E+00, -0.17870303E+01, 0.55547500E+00, 0.10984249E+01, 0.28475463E+00, 0.69275779E+00, -0.18453959E+00, 0.40410057E+00, 0.59705932E-01, -0.23535267E+00, 0.43878641E-01, 0.11331538E+00, 
    },
    {
      -0.10896392E+02, 0.80607300E+01, 0.99370747E+01, 0.28089604E+00, -0.36042504E+01, -0.39281123E+01, -0.43405023E+00, -0.12527297E+00, 0.35088420E+01, -0.13881062E+01, -0.10145779E+01, 0.22173715E+00, -0.23965409E+00, 
    },
    {
      -0.13814932E+03, -0.35444760E+01, 0.49864116E+01, -0.73278651E+01, -0.15269502E+02, -0.34637457E+00, -0.28207886E+01, 0.86998552E+00, -0.19370019E+01, 0.66754687E+00, 0.17929994E+01, -0.26302782E+00, -0.35202691E+00, 
    },
    {
      0.14604482E+03, -0.40838085E+02, -0.57767654E+02, 0.57909651E+01, 0.29187016E+02, 0.18583534E+02, 0.96246183E+00, -0.31996384E+01, -0.18213390E+02, 0.86978769E+01, 0.45727201E+01, 0.29856852E+00, -0.23486688E+01, 
    },
    {
      0.25891187E+03, 0.25923477E+02, -0.39683554E+01, 0.22933840E+02, 0.49968544E+02, -0.23614740E+01, 0.36843927E+01, -0.18787698E+01, 0.36446035E+01, -0.35390708E+01, -0.51340876E+01, 0.95419985E+00, 0.99823630E+00, 
    },
    {
      -0.41596628E+03, 0.86813042E+02, 0.11431066E+03, -0.30645103E+02, -0.82427086E+02, -0.37951321E+02, 0.10900590E+01, 0.11784755E+02, 0.41145611E+02, -0.19994535E+02, -0.94752426E+01, -0.28417885E+01, 0.96225624E+01, 
    },
    {
      -0.22837682E+03, -0.43467033E+02, 0.16895311E+01, -0.27585829E+02, -0.61419991E+02, 0.55255609E+01, -0.12157919E+01, 0.21039855E+01, -0.30910072E+01, 0.53918390E+01, 0.60075970E+01, -0.13708730E+01, -0.16586262E+01, 
    },
    {
      0.45341833E+03, -0.79230301E+02, -0.95844460E+02, 0.45088043E+02, 0.94350952E+02, 0.35473499E+02, -0.38808095E+01, -0.14583707E+02, -0.40570156E+02, 0.19583448E+02, 0.89746466E+01, 0.42096753E+01, -0.11760089E+02, 
    },
    {
      0.78894501E+02, 0.21365410E+02, -0.97209376E+00, 0.11413410E+02, 0.25590933E+02, -0.31731806E+01, -0.29654628E+00, -0.94074392E+00, 0.10037756E+01, -0.26313784E+01, -0.24205432E+01, 0.61006391E+00, 0.94114125E+00, 
    },
    {
      -0.17161839E+03, 0.26147404E+02, 0.29388216E+02, -0.20523869E+02, -0.37670624E+02, -0.12390260E+02, 0.22395270E+01, 0.60695238E+01, 0.14398425E+02, -0.68785167E+01, -0.31523423E+01, -0.18456868E+01, 0.46553698E+01, 
    },
    {
      -0.17224808E+00, 0.15295334E+01, 0.16319659E+01, 0.16117379E-01, -0.11191823E+00, -0.55230491E-01, -0.11635203E-01, 0.23330303E-01, 0.14298861E-01, 0.73818795E-04, -0.13618111E-01, 0.40551308E-02, -0.13926427E-01, 
    },
    {
      0.25908184E+00, -0.15589666E+01, 0.15277148E+01, -0.52449096E-01, 0.15074499E+00, 0.48045823E-02, 0.26224246E-01, -0.31880584E-01, 0.30712897E-01, -0.15925413E-01, 0.30503899E-01, -0.15572473E-01, 0.10377554E-01, 
    },
    {
      -0.44175947E+00, -0.14467432E+00, -0.23857009E+01, -0.93523920E-01, 0.39645356E+00, 0.34704551E+00, -0.23056675E+00, -0.74715160E-01, -0.10230460E-01, 0.67601532E-01, 0.24271284E+00, -0.83723627E-02, 0.27769657E-01, 
    },
    {
      0.43120182E+00, 0.15971814E+01, -0.11812301E+01, -0.19957671E-01, 0.79163861E+00, 0.32819316E+00, 0.47173291E-01, 0.10105979E+00, 0.29491344E+00, 0.12614381E+00, -0.77709101E-01, 0.28120133E+00, -0.83021149E-01, 
    },
    {
      0.28239892E+01, 0.20100613E+02, -0.53325434E+01, 0.45510817E+00, 0.10075530E+01, 0.19475687E+01, 0.87764084E+00, -0.28519857E+00, -0.21241654E+00, -0.31646341E+00, -0.12565173E+00, -0.32961044E-01, 0.23946792E+00, 
    },
    {
      -0.30458715E+01, 0.34423187E+01, 0.15719871E+02, 0.27821069E+01, -0.24117477E+01, -0.53116071E+00, -0.93346071E+00, 0.38857710E+00, -0.41621810E+00, 0.30858749E+00, -0.85805213E+00, 0.56975406E+00, -0.45797646E+00, 
    },
    {
      0.64062252E+01, -0.17833441E+02, 0.38745689E+02, 0.15681111E+01, -0.13188357E+02, -0.35303750E+01, 0.42505393E+01, 0.23458998E+01, 0.36971977E+00, -0.15250417E+01, -0.25250385E+01, 0.11713949E+01, -0.25048587E+00, 
    },
    {
      -0.47601018E+01, -0.14131312E+02, 0.16657829E+02, 0.26456211E+01, -0.20797195E+02, -0.33104246E+01, 0.13996363E+01, -0.86525661E+00, -0.32111042E+01, -0.41198349E+01, -0.53699982E+00, -0.44991617E+01, 0.22725008E+01, 
    },
    {
      -0.15645193E+02, -0.83720718E+02, 0.81215302E+02, -0.40556893E+01, 0.73948927E+01, -0.12809274E+02, -0.93950243E+01, 0.31676583E+01, 0.58742332E+01, 0.21229396E+01, 0.13009325E+01, 0.20239291E+01, -0.53447652E+00, 
    },
    {
      0.26276909E+02, -0.75438370E+02, -0.32025635E+02, -0.30627806E+02, 0.26027321E+02, 0.33012908E+01, 0.51888561E+01, -0.51801095E+01, 0.33771536E+01, -0.15328786E+01, 0.53555956E+01, -0.47251158E+01, 0.42387228E+01, 
    },
    {
      -0.45288258E+02, 0.15141853E+03, -0.23644781E+03, -0.37938159E+01, 0.85512016E+02, 0.71108932E+01, -0.15790138E+02, -0.14872607E+02, -0.79568539E+01, 0.77140079E+01, 0.65682583E+01, -0.10868766E+02, -0.11020887E+01, 
    },
    {
      0.15142425E+02, 0.92271660E+02, -0.60111969E+02, -0.21405069E+02, 0.12360047E+03, 0.14346115E+02, -0.23227024E+02, 0.11077701E+02, 0.13304357E+02, 0.28648264E+02, 0.70955434E+01, 0.23707230E+02, -0.13942078E+02, 
    },
    {
      0.22773668E+02, 0.94970367E+02, -0.27014450E+03, 0.13208767E+02, -0.51395786E+02, 0.31416155E+02, 0.32453068E+02, -0.10277439E+02, -0.23690439E+02, -0.66481109E+01, -0.47724204E+01, -0.81575747E+01, -0.47356147E-01, 
    },
    {
      -0.76574455E+02, 0.26398227E+03, -0.96341965E+02, 0.11079182E+03, -0.88914604E+02, -0.30642164E+01, -0.13225583E+02, 0.17785391E+02, -0.11198152E+02, 0.20659568E+01, -0.13020063E+02, 0.14299355E+02, -0.16164017E+02, 
    },
    {
      0.15472884E+03, -0.41988278E+03, 0.56520874E+03, 0.14640667E+01, -0.21045242E+03, 0.49664984E+01, 0.16372763E+02, 0.38117653E+02, 0.34166718E+02, -0.15126781E+02, -0.92465889E+00, 0.30510057E+02, 0.76909704E+01, 
    },
    {
      -0.28304167E+02, -0.24170132E+03, 0.12176568E+03, 0.48061871E+02, -0.29643774E+03, -0.33637665E+02, 0.83702583E+02, -0.42672482E+02, -0.25475426E+02, -0.74082893E+02, -0.21089073E+02, -0.52561008E+02, 0.36209114E+02, 
    },
    {
      0.58211665E+01, -0.10134482E+02, 0.36446863E+03, -0.14913530E+02, 0.88607018E+02, -0.31856615E+02, -0.45381077E+02, 0.13251899E+02, 0.32890629E+02, 0.99184685E+01, 0.62132945E+01, 0.10968967E+02, 0.11417336E+01, 
    },
    {
      0.98582275E+02, -0.37036868E+03, 0.28047604E+03, -0.16251318E+03, 0.11509062E+03, -0.60012803E+01, 0.18846554E+02, -0.23316351E+02, 0.17238712E+02, -0.60839558E+00, 0.14205902E+02, -0.18185123E+02, 0.25047588E+02, 
    },
    {
      -0.21569096E+03, 0.48272430E+03, -0.61942755E+03, -0.80701262E-01, 0.22225163E+03, -0.23381893E+02, 0.46022134E+01, -0.43023670E+02, -0.49467533E+02, 0.12000541E+02, -0.11647184E+02, -0.33798725E+02, -0.12033131E+02, 
    },
    {
      0.35190132E+02, 0.30408795E+03, -0.15077676E+03, -0.33036247E+02, 0.31228766E+03, 0.41757622E+02, -0.11409177E+03, 0.61402256E+02, 0.21682520E+02, 0.81727142E+02, 0.24596991E+02, 0.52422810E+02, -0.42727203E+02, 
    },
    {
      -0.19147324E+02, -0.20564623E+02, -0.17347337E+03, 0.61360617E+01, -0.46458740E+02, 0.11141960E+02, 0.22063715E+02, -0.60768661E+01, -0.15222783E+02, -0.52150383E+01, -0.25772462E+01, -0.48167362E+01, -0.76443511E+00, 
    },
    {
      -0.44532223E+02, 0.18376117E+03, -0.17429190E+03, 0.82279617E+02, -0.48177139E+02, 0.71741357E+01, -0.10823287E+02, 0.10417562E+02, -0.95174761E+01, -0.11475333E+00, -0.57856946E+01, 0.82347708E+01, -0.13050280E+02, 
    },
    {
      0.10409344E+03, -0.20178984E+03, 0.25817352E+03, -0.14944650E+00, -0.85363350E+02, 0.14998612E+02, -0.99766254E+01, 0.17693336E+02, 0.23184240E+02, -0.29896560E+01, 0.85577002E+01, 0.12933767E+02, 0.56063805E+01, 
    },
    {
      -0.18578611E+02, -0.14634050E+03, 0.76995132E+02, 0.21686547E+01, -0.12222349E+03, -0.20490656E+02, 0.53517189E+02, -0.29491062E+02, -0.64766569E+01, -0.32628654E+02, -0.10126882E+02, -0.19581978E+02, 0.18641266E+02, 
    },
    {
      0.73967703E-01, -0.14112779E+00, 0.17739686E+00, -0.10380149E+01, 0.46758762E+00, -0.52029893E-01, 0.38553629E-01, -0.19935209E-01, 0.24950037E-01, -0.20095922E-01, -0.28559000E-02, -0.27403604E-01, -0.97574195E-03, 
    },
    {
      0.11055296E+00, -0.16600928E+00, -0.32541618E-01, -0.43835488E+00, -0.84406424E+00, -0.22534750E-01, 0.29131178E-01, -0.11667867E-01, -0.34039460E-01, -0.32340210E-01, 0.11184864E-01, -0.23590121E-01, 0.87832250E-02, 
    },
    {
      0.91149993E-01, 0.97471708E+00, 0.51438361E+00, 0.13833370E+01, -0.50698072E+00, 0.27014571E+00, 0.11282755E+00, 0.54294676E-01, -0.75866021E-02, 0.12068818E+00, 0.16455030E-01, 0.90567581E-01, 0.75802878E-01, 
    },
    {
      0.34389898E+00, 0.55841506E+00, 0.90160257E+00, 0.12392099E+01, 0.17093930E+01, 0.13808005E+00, 0.10166630E+00, 0.20844191E+00, -0.15551122E+00, 0.10595880E-01, -0.34317158E-01, 0.91806263E-01, 0.14963972E-01, 
    },
    {
      0.16836854E+01, 0.24811437E+01, -0.16475384E+01, 0.17249215E+01, 0.30530896E+01, 0.75067306E+00, -0.15300773E+01, -0.39954576E+00, 0.12609008E+00, -0.23109366E+00, -0.14283510E+00, 0.11207330E+00, -0.23422834E+00, 
    },
    {
      0.22920127E+01, -0.15677989E+00, 0.10574989E+01, -0.47031479E+01, 0.83145690E+00, 0.60532081E+00, -0.75335938E+00, 0.72133221E-01, 0.13619457E+01, 0.90937072E+00, -0.11166501E+00, 0.18261883E+00, -0.90696365E-01, 
    },
    {
      -0.31183465E+01, -0.60279012E+01, -0.40491314E+01, 0.21722064E+01, -0.14826589E+01, -0.20925791E+01, -0.12620353E+01, -0.50687373E+00, -0.72735786E+00, -0.18294009E+01, 0.10776019E+01, -0.65747297E+00, -0.52606915E-02, 
    },
    {
      -0.83851500E+01, -0.95547777E+00, -0.60184135E+01, -0.85619497E+01, 0.61534148E+00, 0.10074309E+01, -0.85972464E+00, -0.75723243E+00, 0.14236768E+01, -0.71778768E+00, -0.21167088E-01, -0.60213548E+00, 0.36463749E+00, 
    },
    {
      -0.12601962E+02, -0.60230761E+01, -0.29406281E+01, 0.11637723E+02, -0.90889711E+01, -0.18631748E+01, 0.10832180E+02, 0.28277893E+01, 0.15355181E+01, 0.26887271E+01, -0.92505731E-01, 0.11178880E+00, 0.12021805E+01, 
    },
    {
      -0.11314941E+02, 0.27979467E+01, -0.14773956E+02, 0.20741819E+02, 0.14875138E+02, -0.32533488E+01, 0.46425400E+01, -0.30427511E+01, -0.68344517E+01, -0.52124023E+01, -0.58438945E+00, -0.31132221E+00, -0.11308281E+01, 
    },
    {
      0.78239317E+01, 0.12827962E+02, 0.60509434E+01, -0.11737222E+02, 0.10860994E+02, 0.64112983E+01, 0.23935566E+01, 0.18167053E+01, 0.33579202E+01, 0.52217593E+01, -0.41448812E+01, 0.17888294E+01, -0.79165143E+00, 
    },
    {
      0.24079092E+02, -0.46660581E+01, 0.13022430E+02, 0.13900978E+02, -0.91128855E+01, -0.40291896E+01, 0.34799480E+01, 0.18368626E+01, -0.34240825E+01, 0.28355694E+01, 0.16054691E+00, 0.70973271E+00, -0.11734304E+01, 
    },
    {
      0.24227192E+02, 0.57047434E+01, 0.18991457E+02, -0.39872898E+02, 0.15416595E+01, -0.74747860E+00, -0.19494642E+02, -0.59575405E+01, -0.70911956E+01, -0.68836184E+01, 0.16940249E+01, -0.13717374E+01, -0.21887500E+01, 
    },
    {
      0.13924574E+02, -0.16587017E+01, 0.38369900E+02, -0.27500231E+02, -0.43234756E+02, 0.34732475E+01, -0.10127350E+02, 0.84686279E+01, 0.12072734E+02, 0.89818316E+01, 0.23602653E+01, -0.55452925E+00, 0.37565517E+01, 
    },
    {
      -0.70313997E+01, -0.75539041E+01, -0.29214725E+01, 0.47263317E+01, -0.15518660E+02, -0.53729682E+01, -0.81937367E+00, -0.13753450E+01, -0.31669335E+01, -0.39806004E+01, 0.32713692E+01, -0.14239472E+01, 0.86429405E+00, 
    },
    {
      -0.19807907E+02, 0.82111835E+01, -0.75345130E+01, -0.12458882E+01, 0.27180896E+01, 0.28920496E+01, -0.37135229E+01, -0.16157656E+01, 0.26609850E+01, -0.22828846E+01, -0.28764549E+00, 0.47844201E-01, 0.77955002E+00, 
    },
    {
      -0.14075713E+02, -0.23986549E+01, -0.14268916E+02, 0.32392254E+02, 0.10021720E+02, 0.22754657E+01, 0.98033123E+01, 0.34065833E+01, 0.62487297E+01, 0.50818644E+01, -0.16391834E+01, 0.16020304E+01, 0.12680324E+01, 
    },
    {
      -0.18211569E+01, -0.56535420E+01, -0.28800457E+02, 0.73658066E+01, 0.32735832E+02, -0.24511378E+00, 0.70626469E+01, -0.58881278E+01, -0.75028081E+01, -0.46169567E+01, -0.17588331E+01, 0.63301569E+00, -0.28427830E+01, 
    },
    {
      0.76349199E-01, -0.75198315E-01, -0.50871223E-02, -0.72312318E-01, 0.48052013E-01, -0.58207494E+00, -0.75327557E+00, -0.42022016E-01, 0.18120332E-01, -0.16880617E-01, 0.73444042E-02, -0.28194983E-01, -0.15638337E-01, 
    },
    {
      0.39461084E-01, -0.13488993E+00, -0.10055972E+00, 0.46493847E-01, -0.93732879E-01, 0.74668914E+00, -0.58071917E+00, -0.26067192E-01, 0.20636749E-01, 0.86305896E-02, 0.54170616E-03, -0.23148715E-01, -0.52713491E-02, 
    },
    {
      -0.97694404E-01, 0.93344510E+00, 0.29393701E-01, 0.73514744E-02, -0.34737319E+00, 0.26163518E+00, 0.36970001E+00, 0.14785510E+00, -0.55417154E-01, 0.31292816E-02, 0.78705736E-01, -0.81819005E-01, 0.33747096E-01, 
    },
    {
      -0.56032654E-01, -0.46219949E-01, -0.64890347E-01, 0.13689564E+00, -0.56143868E+00, -0.29345471E+00, 0.32432055E+00, 0.15689895E-01, 0.65602973E-01, 0.60567804E-01, -0.10864880E-01, -0.10147065E+00, -0.35914693E-01, 
    },
    {
      0.67012018E+00, 0.42257118E+00, 0.19467359E-01, 0.89575797E-02, -0.10516388E+01, 0.78924227E+00, 0.37479475E+00, -0.40912420E+00, 0.42230487E+00, 0.17606600E+00, 0.89392833E-01, 0.30513483E+00, -0.95638916E-01, 
    },
    {
      -0.71605802E+00, 0.26266012E+01, 0.52779651E+00, -0.18734942E+00, 0.10323124E+01, -0.91403717E+00, 0.53446960E+00, -0.13627081E+00, -0.66635752E+00, -0.12717301E+00, 0.21606894E+00, 0.22295696E+00, 0.47271583E-01, 
    },
    {
      -0.82142699E+00, -0.15194725E+01, 0.36633497E+00, 0.13400576E+00, 0.29195154E+00, 0.53095353E+00, 0.40886421E-01, -0.17716572E+00, 0.17803183E+00, 0.71642324E-01, -0.24857517E+00, 0.35629210E+00, -0.15162340E-01, 
    },
    {
      -0.43191445E+00, 0.30332151E+00, 0.67259073E-01, -0.42061138E+00, 0.12614172E+01, -0.33652726E+00, 0.57137632E+00, -0.20697652E+00, -0.75853616E-01, -0.11044272E+00, 0.11320280E+00, 0.41475806E+00, 0.16160649E+00, 
    },
    {
      -0.10161493E+00, 0.58983707E+00, -0.84979522E+00, 0.30489936E+00, 0.25749736E+01, -0.91752881E+00, -0.53623151E-02, 0.80623126E+00, -0.10490124E+01, -0.38548842E+00, 0.44795712E-02, -0.58598870E+00, 0.16041683E+00, 
    },
    {
      0.97775286E+00, -0.35922375E+01, -0.79198337E+00, -0.14946298E+00, -0.18443063E+01, 0.13141613E+01, -0.25741628E+00, 0.35331422E+00, 0.10484736E+01, 0.61409295E-01, -0.38274255E+00, -0.52102482E+00, -0.16893646E+00, 
    },
    {
      0.26682647E-01, -0.44999905E-01, -0.30429330E-01, -0.24290942E-01, 0.55203110E-01, -0.96770590E-02, -0.37339274E-01, 0.31498066E+00, -0.17359960E+00, -0.28275497E-01, 0.76047601E-02, -0.17485568E-01, -0.21701658E-01, 
    },
    {
      0.19067446E-01, 0.82480900E-01, -0.24417976E-01, 0.60341705E-01, -0.76737002E-01, 0.30098597E-01, -0.14800011E-01, 0.16694994E+00, 0.29754844E+00, 0.60499981E-02, -0.65341271E-02, -0.12154019E-02, -0.32120442E-02, 
    },
    {
      -0.31834999E+00, -0.10176746E+00, -0.12919512E+00, 0.57349347E-01, -0.38777705E-01, -0.16754001E+00, 0.71787603E-01, 0.55727322E-01, 0.48854449E+00, 0.43341171E-01, -0.26241627E-01, 0.33198107E-01, 0.25002802E-01, 
    },
    {
      0.26534075E+00, 0.13981649E+00, 0.81587816E-03, -0.29231967E-01, 0.14419927E+00, -0.16403934E+00, -0.73877536E-01, -0.48972750E+00, -0.57703011E-01, -0.12619291E-01, -0.23607044E-01, -0.51552523E-01, -0.31952363E-01, 
    },
    {
      -0.93206882E-01, -0.45213788E-02, 0.22315690E-02, 0.23447252E-02, -0.55114511E-01, 0.20760903E-01, 0.28086805E-02, 0.39981101E-01, -0.25381871E-01, 0.16350323E+00, 0.20666330E+00, 0.96418187E-02, -0.18114807E-01, 
    },
    {
      0.34151793E-01, 0.49445402E-01, -0.71330421E-01, -0.45091002E-02, -0.17531471E-01, -0.32351278E-01, 0.23442954E-02, 0.62810970E-02, 0.84683262E-01, -0.18843912E+00, 0.15088977E+00, 0.14799476E-01, -0.90895388E-02, 
    },
    {
      0.39233349E-01, 0.67935758E-02, 0.17726658E-01, 0.13225300E-01, 0.48658162E-01, 0.31417664E-01, -0.39813090E-01, 0.35820976E-01, 0.27609885E-01, 0.68485285E-02, -0.42618301E-01, -0.11170206E+00, 0.59159048E-01, 
    },
    {
      0.40888876E-01, 0.11376599E+00, 0.16162302E-01, -0.29816031E-01, 0.62418427E-01, -0.13326484E-01, -0.32052249E-01, -0.99713812E-02, 0.28484600E-01, -0.10098713E-02, -0.12852327E-01, -0.85706830E-01, -0.11604415E+00, 
    },
    {
      -0.68437657E-02, 0.27410646E-02, -0.60129546E-01, 0.18074039E-01, 0.18273348E-01, 0.19582406E-01, -0.92187580E-02, 0.13423235E-01, 0.26095778E-01, 0.18187763E-01, -0.19140057E-01, -0.84481649E-02, 0.25512718E-02, 
    },
    {
      0.86681031E-01, -0.42338647E-01, -0.20124212E-01, -0.68186219E-02, -0.59652332E-01, 0.32581393E-01, -0.32793738E-01, -0.16700966E-01, 0.26844621E-01, -0.54568769E-02, 0.35884832E-02, -0.91454759E-02, 0.69946642E-02, 
    },
    {
      0.57878252E-01, 0.23915995E-01, 0.12776043E-02, 0.33935215E-01, -0.24363872E-01, -0.43215800E-01, -0.14799070E-01, -0.12832981E-01, 0.26710723E-01, 0.12159403E-02, -0.57877437E-02, -0.20406270E-01, -0.79121506E-02, 
    },
    {
      -0.46333019E-01, 0.68355016E-01, -0.24444332E-01, -0.56031704E-01, 0.86708978E-01, 0.14629347E-01, 0.63353748E-03, 0.32699492E-03, 0.16688421E-01, 0.15034921E-01, -0.95454603E-03, -0.14304600E-01, 0.29492131E-02, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir15_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30829437E+01, -0.95623417E-03, 0.14071414E-01, 0.67336978E-02, 0.16362238E-02, 0.31438053E-02, 0.11660823E-01, -0.21267915E-02, -0.41597779E-02, 
    },
    {
      -0.42748913E+00, 0.24467647E-01, 0.60202345E-01, 0.19219477E-01, -0.36522668E-01, -0.30637039E-02, -0.21544660E-01, -0.25534466E-01, 0.19556645E-01, 
    },
    {
      0.83016407E+00, 0.10246895E+00, -0.80737822E-01, -0.11050681E+00, -0.88695347E-01, -0.43044806E-01, -0.54705851E-01, 0.12666488E+00, 0.29847229E-01, 
    },
    {
      0.79621750E+00, -0.35017520E+00, -0.20669055E+00, -0.10942011E+00, 0.13936011E+00, 0.12803758E-02, 0.82603864E-01, 0.60857706E-01, -0.45860834E-01, 
    },
    {
      -0.10769858E+01, -0.21080144E+00, 0.11818612E+00, 0.17931351E+00, 0.22275142E+00, 0.15942883E+00, 0.81478491E-01, -0.35534877E+00, -0.45454629E-01, 
    },
    {
      -0.48894572E+00, 0.31695545E+00, 0.18177135E+00, 0.96843764E-01, -0.10721572E+00, 0.11368042E-02, -0.67227729E-01, -0.28057726E-01, 0.24773201E-01, 
    },
    {
      0.21707885E+00, 0.13230781E+00, 0.58709583E-02, -0.66172384E-01, -0.14059685E+00, -0.13175756E+00, -0.37203427E-01, 0.24195625E+00, 0.17272590E-01, 
    },
    {
      0.22961402E-01, -0.11863714E+00, -0.35007322E+00, 0.12821091E-01, -0.12189340E-01, -0.68265543E-03, 0.62941569E-02, -0.44815377E-02, -0.26578419E-02, 
    },
    {
      0.14728447E-01, 0.37473145E+00, -0.96780010E-01, -0.27384372E-01, 0.10581867E-01, 0.22666370E-02, -0.17687570E-01, 0.68567963E-02, 0.72228303E-02, 
    },
    {
      -0.12491125E+00, -0.71489297E-01, 0.97123340E-01, 0.34888726E-01, 0.35563158E-02, -0.51503751E-01, -0.51774520E-01, -0.26012940E-01, 0.61413504E-01, 
    },
    {
      0.35958029E-01, 0.11307500E-01, -0.75540908E-01, 0.68435147E-02, 0.20799374E-01, -0.41120008E-01, 0.33827305E-01, 0.36062390E-01, 0.14471351E-01, 
    },
    {
      -0.33956641E+00, 0.54326749E+00, 0.19545006E+01, 0.74205472E-03, -0.80778316E-01, -0.77715360E-01, -0.50010346E-01, 0.22858028E+00, 0.87244682E-01, 
    },
    {
      -0.63539225E+00, -0.19151980E+01, 0.43575630E+00, 0.33090886E+00, -0.24209768E-02, -0.12964347E+00, 0.15143993E+00, -0.32082044E-01, 0.21434763E-01, 
    },
    {
      0.14364071E+01, 0.52485019E-01, -0.16226660E+01, -0.25793177E+00, -0.36766967E+00, 0.67354387E+00, 0.12053780E+00, 0.38931045E+00, -0.57873863E+00, 
    },
    {
      0.38684511E+00, 0.14660304E+01, 0.38292372E+00, 0.73202103E-02, 0.29484910E+00, 0.21471724E+00, -0.17507131E+00, -0.18583956E+00, -0.20088874E+00, 
    },
    {
      0.86518139E+00, -0.70394367E+00, -0.22149346E+01, -0.19477078E+00, 0.22440962E+00, 0.24766719E+00, 0.11958317E+00, -0.63776320E+00, -0.41532379E+00, 
    },
    {
      0.21289845E+01, 0.17928966E+01, -0.76989651E+00, -0.93120408E+00, -0.15367731E+00, 0.42717677E+00, -0.41677803E+00, 0.60080402E-01, -0.10584719E+00, 
    },
    {
      -0.35660176E+01, -0.61952978E+00, 0.36163361E+01, 0.49679780E+00, 0.11595428E+01, -0.17288561E+01, 0.13930221E+00, -0.92723536E+00, 0.12680798E+01, 
    },
    {
      -0.96336013E+00, -0.35305631E+01, -0.10435582E+01, -0.91957055E-01, -0.86425048E+00, -0.35838720E+00, 0.15090416E+00, 0.41967389E+00, 0.41672745E+00, 
    },
    {
      -0.42783877E+00, 0.46139610E+00, 0.61414963E+00, 0.16084965E+00, -0.10869228E+00, -0.21873859E+00, -0.10980464E+00, 0.42953312E+00, 0.41603583E+00, 
    },
    {
      -0.16682128E+01, -0.58972288E-01, 0.53959042E+00, 0.71032184E+00, 0.12974617E+00, -0.31926504E+00, 0.34129691E+00, -0.30647589E-01, 0.58991492E-01, 
    },
    {
      0.24329290E+01, 0.10756953E+01, -0.27146754E+01, -0.27721620E+00, -0.90134066E+00, 0.11445333E+01, -0.27922100E+00, 0.58598697E+00, -0.78309441E+00, 
    },
    {
      0.52151656E+00, 0.27146018E+01, 0.98834449E+00, 0.13908830E+00, 0.59049463E+00, 0.22968982E+00, -0.30222910E-01, -0.30343634E+00, -0.20505387E+00, 
    },
    {
      -0.91355713E-02, 0.29706708E-02, -0.31272422E-02, 0.22377572E-02, -0.16692504E+00, -0.71510244E-02, -0.62363604E-02, 0.17446436E-02, -0.74244705E-02, 
    },
    {
      -0.95079606E-03, -0.17749453E-01, 0.95796930E-02, 0.16611043E+00, 0.34745239E-01, -0.31304036E-02, -0.20624200E-01, -0.34306971E-02, 0.38484319E-02, 
    },
    {
      0.50973926E-01, 0.18939672E+00, 0.45086402E-01, 0.27533084E+00, -0.25414023E+00, 0.19070908E-01, -0.36821201E-01, 0.30477703E-01, 0.12377640E-01, 
    },
    {
      -0.62895715E-01, 0.65114923E-01, 0.16306715E-01, 0.36566287E+00, 0.41203955E+00, 0.24824467E-01, -0.14574272E-01, -0.27033810E-02, -0.31155946E-01, 
    },
    {
      0.35454988E-01, 0.60413000E-02, -0.18199088E-01, 0.76078407E-01, -0.16603535E+00, 0.15787179E-01, 0.39357701E-02, -0.27306402E-01, 0.80539152E-01, 
    },
    {
      -0.28398255E+00, 0.76125801E-01, -0.20132874E+00, 0.19440579E+00, -0.70742458E-01, 0.12295902E-01, 0.14769582E+00, 0.96357893E-02, 0.31613465E-01, 
    },
    {
      -0.23468396E+00, -0.99008822E+00, 0.71097992E-01, -0.12017087E+01, 0.50148809E+00, -0.10378635E+00, 0.77810965E-01, -0.91363624E-01, -0.11122116E+00, 
    },
    {
      0.67196661E+00, -0.91745377E-01, 0.13793392E+00, -0.85245258E+00, -0.14946311E+01, -0.28373862E-01, -0.16134294E-01, 0.76116845E-01, 0.61346978E-01, 
    },
    {
      -0.15103582E+00, -0.18279678E+00, -0.22154266E-01, -0.34824383E+00, 0.38926736E+00, 0.52207313E-01, 0.68610243E-01, 0.53184792E-01, -0.12627009E+00, 
    },
    {
      0.39844444E+00, -0.27108863E+00, 0.32306048E+00, -0.42485851E+00, -0.19756165E+00, -0.24037840E-01, -0.15082456E+00, 0.23419553E-03, -0.81081510E-01, 
    },
    {
      0.24415769E+00, 0.13665473E+01, -0.11141641E+00, 0.11919484E+01, -0.32598576E+00, 0.14885177E+00, -0.22232583E-01, 0.77452868E-01, 0.11268586E+00, 
    },
    {
      -0.99769139E+00, 0.24313284E-01, -0.25469135E-01, 0.58631575E+00, 0.13545561E+01, 0.77170582E-03, 0.12235342E+00, -0.77348314E-01, 0.12083619E-01, 
    },
    {
      -0.14412698E-01, 0.11934513E-01, 0.10500170E-01, -0.19674214E-01, -0.95874164E-02, 0.18518902E-01, 0.13004042E-01, 0.67524160E-02, 0.29727495E-02, 
    },
    {
      -0.11646161E-01, -0.37059538E-01, -0.13889452E-01, 0.17757954E-01, -0.19775612E-01, -0.20395935E-01, 0.16103785E-01, -0.75430768E-02, 0.19750383E-02, 
    },
    {
      0.60488075E-01, 0.41547071E-01, 0.15083943E-01, -0.27579306E-01, 0.15362628E-01, 0.45602046E-01, 0.11825938E-01, 0.46184431E-02, 0.33622717E-02, 
    },
    {
      0.20156203E-01, 0.11467540E-01, -0.14934420E-01, -0.22406610E-01, 0.51867850E-02, -0.21814423E-01, 0.54861706E-01, -0.98177865E-02, 0.35655340E-02, 
    },
    {
      -0.18438589E-01, -0.11210715E+00, -0.77830158E-01, 0.20839000E-01, 0.28409932E-02, -0.75307317E-01, -0.12357456E+00, -0.36384016E-01, 0.30860552E-02, 
    },
    {
      0.94958721E-02, 0.14035341E+00, -0.26280927E-01, -0.39201953E-01, 0.32175556E-01, 0.13319464E+00, -0.63215457E-01, 0.32919317E-02, 0.36654861E-02, 
    },
    {
      -0.24621338E-01, -0.67079356E-02, 0.15665390E-01, -0.16562494E-02, -0.48217624E-02, 0.42397496E-02, -0.47429493E-02, 0.29692261E-01, 0.93461300E-03, 
    },
    {
      -0.17562089E-01, 0.82685566E-02, -0.17116127E-02, 0.37179261E-02, -0.40929308E-02, 0.27845460E-02, 0.11334968E-01, -0.89295959E-03, 0.25569776E-01, 
    },
    {
      0.54277506E-01, -0.23874678E-01, 0.16083041E-01, -0.17546413E-01, -0.15783805E-01, 0.44590095E-02, 0.11278612E-01, 0.31083820E-01, 0.20818269E-01, 
    },
    {
      0.31938594E-01, -0.67811191E-01, 0.65194801E-02, 0.25008731E-01, 0.63123470E-02, 0.44389213E-02, -0.66350435E-03, -0.64168908E-01, 0.25633674E-01, 
    },
    {
      0.53379317E-02, 0.13877489E-03, 0.42313864E-02, -0.18789083E-02, -0.55933762E-02, 0.72038206E-02, -0.30773245E-02, -0.21693164E-02, -0.23185892E-02, 
    },
    {
      0.16698711E-01, 0.15778277E-01, -0.79104668E-02, 0.66635874E-02, -0.56811650E-02, 0.11245554E-02, 0.19182810E-02, 0.65909019E-02, -0.23176591E-02, 
    },
    {
      0.10475382E-01, 0.17943383E-02, 0.12603857E-01, 0.36077972E-02, -0.10011117E-01, 0.22700520E-02, 0.29119304E-02, 0.28559086E-02, -0.42581139E-02, 
    },
    {
      0.80381511E-02, 0.52013630E-02, -0.15140937E-01, 0.62007951E-02, 0.13680374E-01, -0.36166117E-02, -0.32075443E-02, -0.44594007E-02, 0.68366108E-03, 
    },
  },
  {
    {
      0.27214456E+01, -0.59877997E-02, 0.48487321E-02, 0.11528316E-01, 0.83281239E-03, 0.11694297E-02, 0.40655504E-02, -0.12881265E-02, -0.53722416E-02, 
    },
    {
      -0.32532126E+00, 0.63253954E-01, 0.72628997E-01, 0.17123209E-01, -0.41282631E-01, -0.95136929E-03, -0.11847535E-01, -0.15334175E-01, 0.25392929E-02, 
    },
    {
      0.11310599E+01, 0.56544151E-01, -0.42942472E-01, -0.46731703E-01, -0.11823614E-01, -0.18031187E-01, -0.14277424E-01, 0.86189993E-01, 0.44520613E-01, 
    },
    {
      0.39499840E+00, -0.47727019E+00, -0.22805473E+00, -0.68672247E-01, 0.14230086E+00, -0.33062331E-01, 0.33439875E-01, 0.30319517E-01, -0.60412716E-02, 
    },
    {
      -0.15841742E+01, 0.48886333E-01, 0.48712164E-01, 0.24080880E-01, -0.28131995E-01, 0.11339512E+00, 0.49910150E-01, -0.21996433E+00, -0.95440678E-01, 
    },
    {
      -0.17072472E+00, 0.39422408E+00, 0.19510727E+00, 0.54062922E-01, -0.10195932E+00, 0.36064230E-01, -0.24015130E-01, -0.89633605E-02, 0.65473379E-02, 
    },
    {
      0.46993488E+00, -0.63174419E-01, 0.38383942E-01, 0.14852938E-01, 0.32509878E-01, -0.10623866E+00, -0.43911122E-01, 0.13939916E+00, 0.55493016E-01, 
    },
    {
      0.23913419E-01, -0.28132305E+00, -0.27718166E+00, 0.13927994E-01, -0.82893772E-02, -0.20239395E-02, 0.55825445E-02, -0.44232481E-02, -0.53990549E-04, 
    },
    {
      -0.56995903E-02, 0.29952064E+00, -0.25091380E+00, -0.14151807E-01, 0.12410180E-01, 0.29630894E-02, -0.19999536E-01, 0.42727250E-02, 0.10325929E-01, 
    },
    {
      0.28787147E-01, 0.21169320E-01, -0.18652616E-01, -0.22639667E-02, 0.15749900E-01, -0.35053190E-01, -0.11501440E-01, -0.28702248E-01, 0.50729234E-01, 
    },
    {
      0.13469434E-01, 0.11982938E+00, -0.71723476E-01, 0.38758967E-01, 0.65331496E-02, -0.35491824E-01, 0.78479424E-02, 0.58513922E-02, 0.13235763E-01, 
    },
    {
      -0.45237130E+00, 0.98882270E+00, 0.10768780E+01, -0.19403292E-01, 0.32672111E-01, 0.33263270E-01, -0.44367231E-01, 0.15512651E+00, -0.33647958E-02, 
    },
    {
      -0.50860500E+00, -0.11076277E+01, 0.85086280E+00, 0.18959832E+00, -0.88933870E-01, -0.87038308E-01, 0.14959402E+00, -0.85117795E-01, -0.10071342E+00, 
    },
    {
      0.16840751E+00, -0.55341917E+00, -0.18681356E+00, 0.46795160E-01, -0.36465406E+00, 0.34773061E+00, -0.56421123E-01, 0.41546983E+00, -0.37943959E+00, 
    },
    {
      0.48950046E+00, 0.17589639E+00, 0.51375562E+00, -0.19277123E+00, 0.27112055E+00, 0.27518129E+00, 0.19615915E-03, 0.11075651E+00, -0.41206829E-01, 
    },
    {
      0.12937419E+01, -0.11020116E+01, -0.64798677E+00, -0.75120479E-01, -0.96557952E-01, -0.93400031E-01, 0.93651280E-01, -0.38147634E+00, 0.10117626E-01, 
    },
    {
      0.18461685E+01, 0.29588601E+00, -0.11936399E+01, -0.57821631E+00, 0.12842540E+00, 0.21288218E+00, -0.31878108E+00, 0.19306193E+00, 0.21284461E+00, 
    },
    {
      -0.75016105E+00, 0.11026198E+01, 0.58122629E+00, -0.15455300E+00, 0.91149658E+00, -0.80885851E+00, 0.42632446E+00, -0.10352906E+01, 0.75210643E+00, 
    },
    {
      -0.14370759E+01, -0.69056630E+00, -0.11955296E+01, 0.39424512E+00, -0.76573443E+00, -0.49984264E+00, -0.18155196E+00, -0.34026536E+00, -0.81930935E-01, 
    },
    {
      -0.85781968E+00, 0.72457278E+00, -0.22283772E-02, 0.71129121E-01, 0.41598022E-01, 0.46153598E-01, -0.39733808E-01, 0.24633887E+00, -0.84825670E-02, 
    },
    {
      -0.14810368E+01, 0.54091024E+00, 0.79484820E+00, 0.49584129E+00, -0.65858573E-01, -0.14391363E+00, 0.22043347E+00, -0.11268485E+00, -0.12970915E+00, 
    },
    {
      0.58107245E+00, -0.53847790E+00, -0.96545213E+00, 0.15799177E+00, -0.56166148E+00, 0.47301409E+00, -0.46004382E+00, 0.66261995E+00, -0.41957995E+00, 
    },
    {
      0.10437909E+01, 0.10547332E+01, 0.84057873E+00, -0.25376251E+00, 0.53126270E+00, 0.30007434E+00, 0.16683571E+00, 0.24391644E+00, 0.14497797E+00, 
    },
    {
      -0.10355590E-01, -0.59964433E-02, 0.24918828E-02, 0.40248983E-01, -0.11974522E+00, -0.68878066E-02, 0.52973459E-03, 0.16610487E-02, -0.47631026E-02, 
    },
    {
      -0.80973590E-02, -0.23301819E-01, 0.16890654E-01, 0.10705962E+00, 0.74931182E-01, 0.11066468E-02, -0.15185948E-01, -0.26813566E-02, 0.41845441E-02, 
    },
    {
      0.87334573E-01, 0.17434020E+00, 0.29009607E-01, 0.16647445E+00, -0.85534535E-01, -0.42410148E-02, -0.62163905E-02, 0.25638817E-01, -0.72888704E-02, 
    },
    {
      0.21463536E-01, 0.82178161E-01, -0.59707514E-02, 0.14109133E+00, 0.29117456E+00, -0.18742292E-02, -0.28855884E-04, 0.78286685E-03, -0.20326763E-01, 
    },
    {
      -0.23505557E-01, 0.50627660E-01, -0.42973336E-01, -0.24614643E+00, 0.68570442E-01, 0.36111508E-01, -0.65725469E-02, -0.73447876E-01, 0.73012590E-01, 
    },
    {
      -0.22136711E+00, 0.12092493E+00, -0.19790530E+00, -0.13838495E-01, -0.37348422E+00, -0.22489665E-01, 0.12434304E+00, -0.84679516E-04, 0.23986567E-02, 
    },
    {
      -0.23487779E+00, -0.88596624E+00, 0.60993783E-01, -0.40152115E+00, 0.38432073E-01, 0.25088888E-01, -0.47009036E-01, -0.10053905E+00, 0.56201056E-01, 
    },
    {
      0.26335394E+00, -0.29947615E+00, 0.83331093E-01, -0.21753137E+00, -0.68338555E+00, 0.12236043E+00, -0.42072836E-01, 0.58362342E-01, 0.13270973E-01, 
    },
    {
      -0.91692032E-02, -0.24499676E+00, 0.31347323E-01, 0.55245612E-01, -0.95426172E-01, -0.17419253E-01, 0.71632415E-01, 0.13672970E+00, -0.10337002E+00, 
    },
    {
      0.32925227E+00, -0.42062321E+00, 0.24600771E+00, 0.48706904E-01, 0.17111088E+00, -0.23537112E-01, -0.12686042E+00, 0.21002434E-01, -0.16690811E-01, 
    },
    {
      0.69676243E-01, 0.11886312E+01, -0.81857055E-01, 0.28926817E+00, -0.38532227E-01, 0.34839232E-01, 0.56403339E-01, 0.85259937E-01, -0.75022250E-01, 
    },
    {
      -0.51328564E+00, 0.48400918E+00, 0.22977814E-01, 0.13689673E+00, 0.42785177E+00, -0.13421567E+00, 0.10484700E+00, -0.73551662E-01, 0.30889113E-01, 
    },
    {
      -0.18197976E-01, 0.72729443E-02, 0.12053806E-01, -0.14377148E-01, -0.17075215E-01, 0.18109780E-01, -0.17175833E-01, -0.27712116E-02, -0.34635665E-03, 
    },
    {
      -0.13854426E-01, -0.35453096E-01, -0.12559115E-01, 0.11651063E-01, -0.17490545E-01, 0.12271131E-01, 0.17724304E-01, -0.37607695E-02, -0.41570398E-02, 
    },
    {
      0.43509074E-01, 0.63937917E-01, 0.10682798E-01, -0.17563932E-01, 0.50647352E-02, 0.68679936E-01, 0.10368119E-02, 0.18375491E-03, 0.41448488E-02, 
    },
    {
      0.25249584E-01, 0.15560197E-01, 0.24144184E-02, -0.15688732E-01, 0.11401277E-02, -0.10834954E-01, 0.62144563E-01, -0.63146567E-02, 0.39895601E-02, 
    },
    {
      0.55234623E-02, -0.11332307E+00, -0.56373134E-01, 0.11459470E-01, 0.22336069E-01, -0.10773219E+00, -0.61456602E-01, -0.12642963E-01, 0.19825991E-01, 
    },
    {
      0.55266679E-02, 0.10650305E+00, -0.20077813E-01, -0.12482653E-01, 0.13799644E-01, 0.59804551E-01, -0.83235875E-01, -0.92901830E-02, 0.25422184E-03, 
    },
    {
      -0.16490787E-01, -0.12288227E-01, 0.15624606E-01, -0.15590733E-02, -0.44173814E-03, 0.32792636E-02, -0.54032323E-02, 0.23649098E-01, 0.65248716E-02, 
    },
    {
      -0.10436538E-01, 0.51402622E-02, -0.10056754E-01, 0.46160161E-02, -0.17697068E-02, 0.29858900E-02, 0.63185222E-02, -0.59197075E-02, 0.20331379E-01, 
    },
    {
      0.39676785E-01, -0.77944589E-02, 0.29491605E-02, -0.97978339E-02, 0.13357931E-02, 0.31436787E-02, 0.15301985E-01, 0.28650684E-01, 0.43868300E-01, 
    },
    {
      0.33646043E-01, -0.66603132E-01, 0.23036173E-01, 0.62092445E-02, -0.26638312E-02, -0.65160957E-02, -0.45676758E-02, -0.67288466E-01, 0.19755878E-01, 
    },
    {
      0.65249880E-02, -0.74825557E-02, 0.43182308E-02, 0.69805081E-02, -0.67913304E-02, 0.64756777E-02, -0.16087731E-02, 0.93622401E-03, -0.20369282E-02, 
    },
    {
      0.11612056E-01, 0.10057593E-01, -0.81270486E-02, 0.65685674E-02, 0.31718882E-02, 0.40957397E-02, 0.22977332E-02, 0.49984353E-02, 0.35048136E-02, 
    },
    {
      0.72671412E-02, 0.54456848E-02, 0.84348572E-02, 0.48800143E-02, -0.11012992E-01, -0.88449667E-03, 0.15507995E-02, 0.17194895E-02, -0.80987962E-03, 
    },
    {
      -0.22724713E-04, 0.86002506E-03, -0.96425554E-02, 0.45933654E-02, 0.12882300E-01, 0.65020460E-03, -0.29192658E-02, -0.44633718E-02, 0.21139076E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir16_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.53135514E+01, -0.61482415E-01, 0.75109832E-01, 0.34710813E-01, 0.22684038E-01, -0.57048872E-02, -0.84173474E-02, 0.44471744E-01, -0.16074698E-01, 0.85667998E-03, 0.22791712E-03, 0.92006745E-02, -0.97777117E-02, 
    },
    {
      -0.19316612E+01, -0.42007375E+00, 0.51300526E+00, -0.14258102E+00, 0.31701854E+00, -0.88729514E-02, 0.24743211E+00, -0.80617800E-01, -0.12699057E+00, -0.17045531E-01, 0.84008286E-02, -0.75328685E-01, 0.18828200E-01, 
    },
    {
      0.46763873E+01, 0.11320518E+01, -0.18580437E+01, -0.27329683E+00, -0.90213805E+00, -0.61084539E+00, 0.23694906E+00, -0.46116051E+00, 0.34168661E+00, 0.33637661E+00, -0.19169091E+00, -0.11774270E+00, 0.24918503E+00, 
    },
    {
      0.28097292E+02, 0.26754222E+01, -0.14834358E+01, 0.31701112E+01, -0.53412337E+01, 0.76283532E+00, -0.51419859E+01, 0.73323458E+00, 0.33136436E+00, 0.64914894E+00, 0.56351509E-01, 0.17945926E+01, 0.51091266E+00, 
    },
    {
      -0.29699144E+02, -0.10468776E+02, 0.11762651E+02, 0.23983593E+01, -0.62706685E+00, 0.43435154E+01, -0.74329567E+00, 0.25396516E+01, -0.11957142E+01, -0.22866728E+01, 0.13279625E+01, 0.59059352E+00, -0.12482961E+01, 
    },
    {
      -0.74450401E+02, -0.65370393E+01, -0.13106404E+02, -0.16474710E+02, 0.30279955E+02, -0.35481794E+01, 0.26479380E+02, -0.35002706E+01, 0.19646826E+01, -0.25204167E+01, -0.10204962E+01, -0.11526494E+02, -0.46005092E+01, 
    },
    {
      0.57885025E+02, 0.38123917E+02, -0.29894230E+02, -0.82557545E+01, 0.11295097E+02, -0.11428127E+02, -0.59597737E+00, -0.69582043E+01, 0.21485796E+01, 0.51051588E+01, -0.42251997E+01, -0.11586436E+01, 0.25822914E+01, 
    },
    {
      0.64745682E+02, 0.56215553E+01, 0.52016819E+02, 0.30449734E+02, -0.71321335E+02, 0.47431431E+01, -0.54951233E+02, 0.77081962E+01, -0.85594635E+01, 0.24655552E+01, 0.34039283E+01, 0.28274258E+02, 0.11945585E+02, 
    },
    {
      -0.61265751E+02, -0.52918648E+02, 0.32995888E+02, 0.10468823E+02, -0.18554152E+02, 0.12917597E+02, 0.30349872E+01, 0.83194733E+01, -0.23343971E+01, -0.44490862E+01, 0.54735060E+01, 0.93166763E+00, -0.25764549E+01, 
    },
    {
      0.18634470E+01, 0.28486409E+01, -0.62127373E+02, -0.22524527E+02, 0.73664352E+02, -0.11603888E+01, 0.50629311E+02, -0.74867887E+01, 0.10972483E+02, 0.44247410E+00, -0.41348748E+01, -0.29279104E+02, -0.12263268E+02, 
    },
    {
      0.26340647E+02, 0.24130861E+02, -0.13063382E+02, -0.43806276E+01, 0.87424278E+01, -0.52607641E+01, -0.19757253E+01, -0.35192804E+01, 0.10667630E+01, 0.12665039E+01, -0.23884971E+01, -0.25039810E+00, 0.10172306E+01, 
    },
    {
      -0.17573338E+02, -0.41650991E+01, 0.24150318E+02, 0.54928117E+01, -0.27598324E+02, -0.83179241E+00, -0.17326010E+02, 0.26191671E+01, -0.46037669E+01, -0.10153755E+01, 0.16939994E+01, 0.10823385E+02, 0.43800306E+01, 
    },
    {
      -0.18334110E+00, 0.16058656E+01, 0.18451380E+01, 0.23984671E-01, -0.77645235E-01, -0.71523450E-02, -0.39501633E-02, 0.48090607E-01, 0.10382491E-02, -0.10124075E-02, -0.98467534E-02, 0.16033929E-01, -0.86858859E-02, 
    },
    {
      0.11624959E+00, -0.18097712E+01, 0.16314628E+01, -0.21510409E-01, 0.10980082E+00, 0.23303006E-01, -0.40793312E-02, -0.20446820E-01, 0.77379011E-01, -0.18653555E-01, 0.93311332E-02, -0.12022865E-02, 0.20225255E-01, 
    },
    {
      -0.59684879E+00, -0.33499169E+00, -0.13415632E+01, 0.50301838E+00, -0.44692445E+00, 0.90694763E-01, -0.62957235E-01, -0.15851028E+00, 0.21268469E+00, 0.11720025E+00, 0.27619761E+00, 0.42360511E-01, 0.29771827E-01, 
    },
    {
      -0.34701595E+00, 0.15442055E+01, -0.20425926E+00, 0.16489644E+00, 0.11285269E+00, 0.20746952E+00, -0.82578465E-01, 0.18210283E+00, 0.43843952E+00, 0.80315150E-01, 0.14707121E+00, 0.18282767E+00, 0.11922825E+00, 
    },
    {
      0.64125867E+01, 0.73759170E+01, 0.92153292E+01, -0.22129582E+00, -0.95621055E+00, 0.99270296E+00, -0.32091096E+00, -0.50337327E+00, 0.87898248E+00, -0.46062678E+00, -0.19489811E+00, -0.40433988E+00, 0.12082740E+00, 
    },
    {
      -0.33119065E+00, -0.12071861E+02, 0.27483230E+01, -0.50224137E+00, -0.92419714E+00, -0.15161744E+01, 0.67563355E+00, 0.55171251E+00, -0.55627638E+00, 0.47423026E+00, 0.14856794E+00, 0.11773849E+00, -0.19019896E+00, 
    },
    {
      0.57699900E+01, 0.14562860E+01, 0.80994415E+01, -0.70281959E+01, 0.21438148E+01, 0.44327801E+00, -0.11645912E+01, 0.24570403E+01, -0.35969975E+01, -0.75412291E+00, -0.54496884E+01, -0.50559856E-01, 0.71804142E+00, 
    },
    {
      0.59823146E+01, -0.11598491E+02, 0.66550407E+01, 0.14041714E+00, -0.62558069E+01, -0.17235842E+01, 0.17961844E+01, -0.20923460E+01, -0.71783361E+01, -0.25725663E+01, -0.15988770E+01, -0.18210272E+01, -0.11821673E+01, 
    },
    {
      -0.42838211E+02, -0.50422752E+02, -0.27591564E+02, 0.72362642E+01, 0.67548871E+01, -0.13753628E+02, 0.42630401E+01, 0.29996941E+01, -0.36755970E+01, 0.23023808E+01, 0.17583880E+01, 0.28728533E+01, 0.73049802E+00, 
    },
    {
      -0.27694557E+01, 0.49150024E+02, -0.67315660E+01, 0.31819255E+01, 0.76478081E+01, 0.15908955E+02, -0.70412917E+01, -0.81684484E+01, 0.24692976E+01, -0.28794234E+01, -0.24862816E+01, -0.21675971E+00, -0.54332763E+00, 
    },
    {
      -0.26377701E+02, 0.61225288E+02, -0.36126202E+02, 0.34660233E+02, 0.78493519E+01, -0.54490952E+01, 0.13300788E+02, -0.13281824E+02, 0.18924454E+02, 0.30061636E+01, 0.32234249E+02, -0.37474296E+01, -0.85291586E+01, 
    },
    {
      -0.58997833E+02, 0.78057648E+02, 0.12184751E+02, -0.11472317E+02, 0.42532295E+02, 0.64326682E+01, -0.15016547E+02, 0.18534817E+02, 0.30961195E+02, 0.14846240E+02, 0.53940797E+01, 0.65931735E+01, 0.31726680E+01, 
    },
    {
      0.10302815E+03, 0.10765425E+03, -0.18549156E+02, -0.29647400E+02, -0.18905441E+02, 0.51032181E+02, -0.18045750E+02, -0.76261740E+01, 0.33367565E+01, -0.33117504E+01, -0.46762824E+01, -0.70676723E+01, -0.52129989E+01, 
    },
    {
      0.19405296E+02, -0.37196632E+02, -0.30845879E+02, -0.12045132E+01, -0.24903931E+02, -0.52441917E+02, 0.21243822E+02, 0.32386875E+02, -0.61232271E+01, 0.69945936E+01, 0.10995952E+02, -0.16689281E+01, 0.32438049E+01, 
    },
    {
      0.83281013E+02, -0.24887656E+03, 0.94408211E+02, -0.81634117E+02, -0.56857452E+02, 0.11548672E+02, -0.39307831E+02, 0.35773415E+02, -0.44039265E+02, -0.80293608E+01, -0.81032394E+02, 0.15755465E+02, 0.29912725E+02, 
    },
    {
      0.17956213E+03, -0.25433278E+03, -0.10402868E+03, 0.38665817E+02, -0.10846452E+03, -0.16294807E+02, 0.43562103E+02, -0.65742599E+02, -0.52942467E+02, -0.32878132E+02, -0.90460014E+01, -0.10402137E+02, -0.99966192E+00, 
    },
    {
      -0.99776047E+02, -0.10776513E+03, 0.93763397E+02, 0.41308418E+02, 0.23109884E+02, -0.71873962E+02, 0.29615770E+02, 0.85741634E+01, 0.15823607E+01, 0.56734997E+00, 0.39712129E+01, 0.65275764E+01, 0.91736221E+01, 
    },
    {
      -0.30059841E+02, -0.42904842E+02, 0.75276306E+02, -0.11297899E+02, 0.29414515E+02, 0.68314957E+02, -0.24166710E+02, -0.47956135E+02, 0.79749370E+01, -0.81939878E+01, -0.17279491E+02, 0.42007875E+01, -0.36308939E+01, 
    },
    {
      -0.12822026E+03, 0.33393097E+03, -0.12717429E+03, 0.89460068E+02, 0.95602333E+02, -0.56442857E+01, 0.46425873E+02, -0.45104004E+02, 0.47921806E+02, 0.10032402E+02, 0.92091537E+02, -0.21654915E+02, -0.41252964E+02, 
    },
    {
      -0.20665901E+03, 0.35750424E+03, 0.15833980E+03, -0.42081184E+02, 0.11878425E+03, 0.21206402E+02, -0.51555927E+02, 0.93430550E+02, 0.37203754E+02, 0.32163887E+02, 0.86474857E+01, 0.69033074E+01, -0.54782982E+01, 
    },
    {
      0.31105921E+02, 0.42761517E+02, -0.58713291E+02, -0.18812452E+02, -0.99255018E+01, 0.34368755E+02, -0.16233343E+02, -0.36081984E+01, -0.22873654E+01, 0.10863211E+01, -0.70161188E+00, -0.17797345E+01, -0.49097610E+01, 
    },
    {
      0.14821376E+02, 0.46732437E+02, -0.43093964E+02, 0.10727091E+02, -0.10283649E+02, -0.30857159E+02, 0.94555550E+01, 0.23968437E+02, -0.39821031E+01, 0.38799286E+01, 0.87662678E+01, -0.24545429E+01, 0.99674916E+00, 
    },
    {
      0.70579102E+02, -0.14992651E+03, 0.64224586E+02, -0.36984921E+02, -0.49639442E+02, -0.14587626E+01, -0.19475060E+02, 0.20724131E+02, -0.19756748E+02, -0.44517136E+01, -0.38782745E+02, 0.96913366E+01, 0.19572958E+02, 
    },
    {
      0.79299797E+02, -0.17635577E+03, -0.73442924E+02, 0.14171571E+02, -0.48244438E+02, -0.10028917E+02, 0.21243670E+02, -0.45515968E+02, -0.82364569E+01, -0.11869068E+02, -0.36658957E+01, -0.14194759E+01, 0.46354752E+01, 
    },
    {
      0.10682828E+00, 0.15335342E+00, 0.24009713E-02, -0.94681269E+00, -0.20006712E+00, 0.35301302E-01, -0.26023917E-01, -0.17329574E-01, 0.51625643E-01, -0.21257993E-01, -0.38532102E-02, -0.74463747E-02, 0.90640932E-02, 
    },
    {
      0.19327568E+00, -0.18736569E-01, 0.22726156E-01, 0.25009137E+00, -0.72113854E+00, 0.99601615E-02, -0.55031674E-02, -0.26947472E-01, 0.28790315E-02, -0.16322194E-01, -0.36808793E-03, -0.56895283E-02, 0.75899302E-02, 
    },
    {
      -0.55829501E+00, 0.74548769E+00, 0.33699301E+00, -0.38933989E+00, -0.11029156E+01, 0.14290258E+00, -0.13941409E-01, 0.13658305E+00, -0.56958534E-02, 0.14612201E+00, -0.36390368E-01, 0.34695819E-04, 0.22696797E-01, 
    },
    {
      -0.19936664E-01, 0.11800451E+00, 0.74418652E+00, 0.16391089E+01, -0.22145706E+00, 0.27896428E-01, 0.86451977E-01, 0.23700632E+00, -0.18414943E-01, 0.22026940E-03, -0.35097837E-01, 0.70859253E-01, 0.16451158E-01, 
    },
    {
      0.78674906E+00, 0.40604296E+00, 0.13178561E+01, 0.94937210E+01, 0.35943595E+00, -0.45602840E+00, 0.23751763E+00, 0.16641727E+00, -0.49453366E+00, -0.13433233E+00, -0.40668243E+00, -0.21240635E-01, -0.36346906E+00, 
    },
    {
      0.72496071E-01, 0.11956791E+01, 0.17677049E+00, -0.21435833E+01, 0.64546747E+01, 0.49960443E+00, 0.50610906E+00, -0.10921649E+00, 0.65923679E+00, 0.13551632E+00, -0.21280682E+00, 0.16277494E+00, 0.11476960E+00, 
    },
    {
      0.45960088E+01, -0.63736382E+01, -0.17190388E+01, 0.82465286E+01, 0.17915984E+01, -0.91923392E+00, -0.91493428E+00, -0.46042120E+00, -0.24589279E+00, -0.14275963E+01, 0.90770066E+00, 0.12878948E+00, 0.27167577E+00, 
    },
    {
      -0.16134089E+01, -0.31690547E+01, -0.59042554E+01, -0.66804452E+01, 0.78570251E+01, 0.24046221E+01, -0.80951697E+00, -0.12263418E+01, -0.75266880E+00, -0.16484741E+01, 0.21213526E+00, -0.13156597E+00, 0.31026939E+00, 
    },
    {
      -0.10191396E+02, -0.35821579E+01, -0.67020483E+01, -0.21100809E+02, 0.50712733E+01, 0.15026681E+01, 0.15030552E+01, -0.14051753E+01, 0.43638811E+01, 0.14188287E+01, 0.20353065E+01, 0.83522245E-01, 0.18233762E+01, 
    },
    {
      -0.14807859E+00, -0.29033175E+01, 0.49544346E+00, 0.72752118E+01, -0.59718432E+01, -0.30980995E+01, -0.21497185E+01, -0.39590228E+00, -0.21902370E+01, -0.58068059E-01, 0.13292863E+01, -0.71125287E+00, -0.20930350E+01, 
    },
    {
      -0.11946964E+02, 0.14842890E+02, -0.15085144E+00, -0.26015095E+02, -0.11381587E+00, 0.23868535E+01, 0.37237940E+01, 0.10595403E+01, 0.15183357E+01, 0.36690791E+01, -0.32947772E+01, -0.13435316E+00, -0.86445820E+00, 
    },
    {
      0.70006118E+01, 0.88694229E+01, 0.11823129E+02, 0.99858646E+01, -0.25837082E+02, -0.93706493E+01, 0.37652469E+01, 0.22050340E+01, 0.24930382E+01, 0.60392127E+01, -0.10250652E+01, -0.25152349E+00, -0.87281799E+00, 
    },
    {
      0.25511642E+02, 0.61066298E+01, 0.16452080E+02, 0.43983407E+01, -0.15019115E+02, -0.25166245E+01, -0.56915488E+01, 0.42986555E+01, -0.11266659E+02, -0.37101965E+01, -0.29099541E+01, 0.31099221E+00, -0.24174685E+01, 
    },
    {
      -0.37529130E+01, -0.29049420E+01, -0.49277145E+00, -0.12283454E+02, -0.22927643E+02, 0.58630066E+01, 0.24241989E+01, 0.20458815E+01, 0.14258064E+01, -0.27732790E+00, -0.25777242E+01, 0.11977600E+01, 0.57707081E+01, 
    },
    {
      0.85478735E+01, -0.10505727E+02, 0.35088875E+01, 0.18057755E+02, -0.31088126E+01, -0.20258300E+01, -0.33768830E+01, -0.75583994E+00, -0.16262639E+01, -0.27519293E+01, 0.28147068E+01, -0.59420045E-01, 0.66975570E+00, 
    },
    {
      -0.69251733E+01, -0.75538425E+01, -0.68501964E+01, -0.25737543E+01, 0.18046787E+02, 0.80261040E+01, -0.44093361E+01, -0.15761824E+01, -0.17725418E+01, -0.48975239E+01, 0.11016632E+01, 0.48760176E+00, 0.51885700E+00, 
    },
    {
      -0.17836386E+02, -0.32042096E+01, -0.13130729E+02, 0.11606612E+02, 0.12947696E+02, 0.19340641E+01, 0.47053881E+01, -0.35811558E+01, 0.81174049E+01, 0.28834205E+01, 0.10804650E+01, -0.32320696E+00, 0.74705285E+00, 
    },
    {
      0.50985780E+01, 0.60607548E+01, -0.27176561E+01, 0.50116320E+01, 0.27655184E+02, -0.36654398E+01, 0.24630740E+00, -0.14518211E+01, 0.14266887E+00, 0.26619384E+00, 0.16181467E+01, -0.81003147E+00, -0.41553168E+01, 
    },
    {
      0.44094801E-01, 0.12518990E+00, -0.13474181E+00, -0.11440214E+00, 0.75728670E-01, -0.35541511E+00, -0.50099200E+00, -0.55849489E-01, 0.62280174E-01, -0.11978948E-01, -0.18889341E-03, -0.37581697E-01, -0.79547213E-02, 
    },
    {
      0.80496319E-01, -0.75641625E-01, -0.15093005E+00, -0.28906606E-01, -0.13633297E+00, 0.48335886E+00, -0.39124358E+00, -0.51529985E-01, -0.16674163E-01, 0.13538418E-01, 0.16163324E-02, -0.82436623E-02, -0.84085455E-04, 
    },
    {
      -0.27754477E+00, 0.71311891E+00, -0.89700386E-01, -0.21103679E+00, -0.17687225E+00, 0.15123169E+00, 0.22682615E-01, 0.21121208E+00, -0.16462816E-01, -0.60570084E-01, 0.39293535E-01, -0.10021122E+00, 0.95546981E-02, 
    },
    {
      -0.29436976E+00, -0.28597829E+00, -0.23569670E+00, -0.61088581E-01, -0.30247924E+00, 0.42297073E-01, 0.29653084E+00, -0.27596239E-01, -0.62871657E-01, -0.64206328E-02, -0.10770185E+00, -0.35808660E-01, -0.16216652E-01, 
    },
    {
      -0.14272901E+00, -0.95003837E+00, 0.14099309E+01, 0.34587759E+00, -0.58685106E+00, -0.53814942E+00, -0.16351078E+00, 0.94250903E-01, 0.10865152E+00, 0.25800949E+00, 0.30995958E-01, 0.31330404E+00, -0.10657041E+00, 
    },
    {
      -0.97229975E+00, 0.19243021E+01, 0.96367663E+00, -0.21867950E+00, 0.12699347E+01, 0.10597437E+00, -0.20831758E+00, -0.16814458E+00, 0.64819470E-01, -0.26897462E-01, 0.20455129E+00, 0.10330691E+00, -0.44147037E-01, 
    },
    {
      -0.15029612E+00, -0.14936835E+01, 0.13109908E+00, 0.61500740E+00, 0.24869451E+00, -0.14464337E+00, 0.21722548E+00, -0.44552392E+00, 0.24300979E-02, 0.12858246E+00, -0.11394108E+00, 0.26291013E+00, -0.14204117E-01, 
    },
    {
      0.23137939E+00, 0.89680660E+00, 0.70877725E+00, 0.72342165E-01, 0.90643167E+00, -0.44160914E+00, -0.38550001E+00, 0.47713518E-01, 0.66411771E-01, -0.97876545E-02, 0.23822626E+00, 0.20062061E+00, 0.58113962E-01, 
    },
    {
      0.85560966E+00, 0.20959485E+01, -0.21306541E+01, -0.25514224E+00, 0.12787257E+01, 0.20966003E+01, 0.11189801E+01, 0.11804737E+00, -0.47789565E+00, -0.44577163E+00, 0.65643959E-01, -0.56391513E+00, 0.21226558E+00, 
    },
    {
      0.11625001E+01, -0.24074843E+01, -0.14329216E+01, -0.33291936E-01, -0.19074674E+01, -0.82054245E+00, 0.16421086E+01, 0.59356934E+00, -0.19039523E-01, -0.14523591E+00, -0.37697402E+00, -0.36098781E+00, 0.90530626E-01, 
    },
    {
      0.41392002E-01, -0.41367438E-01, -0.56517832E-01, -0.29723020E-01, 0.45230564E-01, 0.29130369E-01, -0.52787438E-02, 0.22493723E+00, -0.37028406E-01, -0.76311780E-02, 0.66862693E-02, -0.14504310E-01, -0.11098138E-01, 
    },
    {
      0.60032666E-01, -0.52753162E-01, 0.69005713E-02, 0.17640466E-01, -0.42814646E-01, 0.10649805E-02, -0.16114800E-02, 0.41139279E-01, 0.21040384E+00, 0.91859568E-02, -0.34672103E-02, 0.49885898E-02, -0.36266081E-02, 
    },
    {
      -0.22909485E-01, -0.57594564E-01, -0.49758352E-01, 0.40956624E-01, -0.18536091E+00, -0.12244110E+00, 0.80677725E-01, 0.27833149E+00, 0.26716733E+00, -0.18367633E-01, -0.48983745E-01, -0.45404307E-03, 0.61136829E-02, 
    },
    {
      0.23329218E+00, 0.21748865E+00, -0.60024496E-01, -0.60173009E-01, 0.15808193E+00, 0.48826919E-02, -0.13135211E-01, -0.34629282E+00, 0.13930210E+00, -0.11054378E-01, -0.39946426E-01, -0.16428996E-01, -0.25691027E-01, 
    },
    {
      -0.12797327E+00, -0.13176106E-01, 0.16579621E-01, -0.26598467E-01, -0.44626016E-01, 0.11000606E-01, -0.68844962E-02, 0.30088242E-01, -0.21648988E-01, 0.85243776E-01, 0.13738230E+00, 0.71503147E-02, -0.82375538E-02, 
    },
    {
      0.41204984E-02, -0.28827868E-02, -0.10035298E-01, 0.48501052E-01, -0.72628535E-01, -0.31241590E-01, -0.27600962E-02, 0.50012227E-01, 0.15929714E-01, -0.15858254E+00, 0.73988296E-01, 0.60939570E-02, -0.19759694E-01, 
    },
    {
      -0.29281452E-01, -0.43308523E-01, -0.50946777E-02, 0.35335834E-03, -0.43678591E-02, 0.27276650E-02, -0.24702514E-01, 0.39685573E-01, 0.26893996E-01, 0.11443176E-02, 0.35606234E-03, -0.97904116E-01, 0.17392356E-01, 
    },
    {
      0.29230772E-01, 0.30779630E-01, 0.66181175E-01, 0.33190589E-01, 0.54724183E-01, -0.13985308E-01, -0.75963368E-02, -0.36636444E-02, -0.11163543E-01, -0.28938515E-01, -0.19950645E-01, -0.18290415E-01, -0.11451925E+00, 
    },
    {
      0.25555792E-02, 0.33711719E-02, -0.17088523E-01, 0.12914094E-01, 0.22545582E-01, -0.11638348E-01, 0.84269810E-02, 0.31605013E-01, 0.43525486E-02, -0.14523819E-01, -0.11092685E-01, 0.72272657E-03, 0.65107662E-02, 
    },
    {
      0.73336594E-01, -0.65333776E-01, -0.23503136E-01, 0.45837099E-02, 0.15234247E-01, -0.57932669E-02, 0.15654413E-02, -0.23898358E-01, 0.22590397E-01, -0.16863436E-01, -0.25549070E-02, -0.54761791E-02, 0.18356221E-01, 
    },
    {
      0.17226540E-01, 0.12007785E+00, -0.12394503E+00, 0.27426206E-01, 0.75115939E-02, -0.23556633E-01, 0.14968693E-01, -0.76391217E-02, 0.15911356E-01, -0.28241746E-01, -0.31705524E-02, -0.28219888E-01, -0.10984738E-01, 
    },
    {
      -0.51771577E-01, -0.57098405E-02, 0.31669147E-01, -0.37942089E-01, 0.34906853E-01, 0.10111721E-01, -0.26458624E-03, -0.22664923E-01, 0.10375086E-01, 0.15457364E-01, -0.95300321E-02, 0.46850368E-02, 0.10677025E-01, 
    },
  },
  {
    {
      0.80510998E+01, -0.25949422E-01, 0.21837340E+00, 0.76524161E-01, -0.26052505E-01, -0.19464865E-01, -0.11594040E-01, 0.40525004E-01, -0.83783437E-02, 0.77255415E-02, -0.43546511E-02, 0.13457161E-01, -0.19088319E-01, 
    },
    {
      -0.62016821E+00, -0.67613447E+00, 0.24612109E+00, -0.29176807E+00, 0.24085705E+00, -0.57791073E-01, 0.25819287E+00, 0.50516576E-01, -0.14634442E+00, 0.24345282E-02, 0.78144493E-02, -0.72793506E-01, 0.15928514E-01, 
    },
    {
      0.17819746E+02, 0.14743440E+01, -0.27259543E+01, -0.13425981E+00, 0.37347874E+00, 0.50698123E-02, 0.27457744E+00, -0.32091236E+00, 0.19228476E+00, 0.13757847E+00, -0.13359857E+00, -0.24793717E+00, 0.29714280E+00, 
    },
    {
      -0.11858767E+02, 0.69419298E+01, 0.46298509E+01, 0.42721686E+01, -0.35434074E+01, 0.75084180E+00, -0.42302465E+01, -0.70975661E+00, 0.99484891E+00, -0.40930372E+00, -0.47602188E-01, 0.14345567E+01, 0.58592618E+00, 
    },
    {
      -0.97920853E+02, -0.13692502E+02, 0.91235781E+01, -0.14460938E+01, -0.85480909E+01, 0.18136225E+01, -0.40499854E+00, 0.23285608E+01, 0.21228252E+00, -0.17068177E+01, 0.10739388E+01, 0.17055205E+01, -0.15911946E+01, 
    },
    {
      0.13621458E+03, -0.31045456E+02, -0.30203131E+02, -0.16606495E+02, 0.24689091E+02, -0.46569004E+01, 0.19129166E+02, 0.27777817E+01, -0.42672701E+01, 0.31055405E+01, 0.20506692E+00, -0.88754292E+01, -0.49245520E+01, 
    },
    {
      0.18260400E+03, 0.48245602E+02, -0.12680835E+02, 0.67058239E+01, 0.30385052E+02, -0.77466555E+01, -0.24252276E+01, -0.74331207E+01, -0.30216341E+01, 0.50820293E+01, -0.34738324E+01, -0.39474041E+01, 0.36654823E+01, 
    },
    {
      -0.37284259E+03, 0.57534088E+02, 0.59615005E+02, 0.21586212E+02, -0.65765862E+02, 0.11232314E+02, -0.35051090E+02, -0.48223314E+01, 0.10247663E+02, -0.89555550E+01, 0.19287635E-01, 0.21166868E+02, 0.12303819E+02, 
    },
    {
      -0.16647343E+03, -0.65894676E+02, 0.92080431E+01, -0.95255365E+01, -0.38242653E+02, 0.10814571E+02, 0.54566917E+01, 0.95939255E+01, 0.48733854E+01, -0.54909315E+01, 0.43026366E+01, 0.36398020E+01, -0.39649410E+01, 
    },
    {
      0.40097147E+03, -0.42800575E+02, -0.49323971E+02, -0.69891477E+01, 0.72595741E+02, -0.11016485E+02, 0.28427053E+02, 0.38504643E+01, -0.11370510E+02, 0.10496172E+02, -0.38129082E+00, -0.21485941E+02, -0.12095145E+02, 
    },
    {
      0.60411118E+02, 0.29807955E+02, -0.30553248E+01, 0.43338375E+01, 0.16042967E+02, -0.49058828E+01, -0.29349284E+01, -0.42607260E+01, -0.22481444E+01, 0.19431134E+01, -0.17624894E+01, -0.11618649E+01, 0.16431694E+01, 
    },
    {
      -0.15156676E+03, 0.10127841E+02, 0.15031334E+02, -0.20478544E+01, -0.28253370E+02, 0.37018678E+01, -0.85423803E+01, -0.11203144E+01, 0.45561490E+01, -0.42304101E+01, 0.19091371E+00, 0.78487253E+01, 0.40897169E+01, 
    },
    {
      -0.36684188E+00, 0.12926129E+01, 0.19531890E+01, 0.42412702E-01, -0.14427768E+00, -0.23068365E-01, 0.24992963E-02, 0.48195336E-01, -0.24228871E-02, 0.57797058E-03, -0.24764627E-01, 0.62236316E-02, -0.12336963E-01, 
    },
    {
      0.35460928E+00, -0.17541237E+01, 0.13221329E+01, -0.24655612E-04, 0.17179137E+00, 0.25261961E-01, 0.78048166E-02, -0.28375529E-01, 0.45270234E-01, -0.58390521E-02, 0.73639741E-02, -0.46915747E-02, 0.16659260E-01, 
    },
    {
      -0.69456911E+00, 0.28679106E+00, -0.34354627E+01, 0.52327073E+00, 0.16810082E+00, -0.47624176E-02, -0.91403551E-01, -0.16691753E+00, 0.12325200E-01, 0.15972370E+00, 0.14241000E+00, 0.19505944E-01, 0.13720220E+00, 
    },
    {
      -0.31720053E-02, 0.21107395E+01, -0.78734863E+00, -0.80523528E-02, 0.38587731E+00, 0.15747765E+00, 0.84543660E-01, 0.20151427E+00, 0.25630075E+00, 0.16800867E+00, 0.43668777E-01, 0.15243682E+00, 0.24071131E-01, 
    },
    {
      0.73923025E+01, 0.16647623E+02, -0.37959800E+01, 0.78674835E+00, 0.54041618E+00, 0.10678436E+01, -0.19122691E+00, -0.94048089E+00, 0.97269678E+00, -0.43552950E+00, 0.79522692E-01, -0.37138738E-01, -0.29479025E-01, 
    },
    {
      -0.34290457E+01, -0.85238671E+00, 0.14262810E+02, 0.15189956E+01, -0.36586006E+01, -0.14248925E+01, -0.20277178E+00, 0.64990950E+00, -0.36768889E+00, 0.84792279E-01, -0.21742781E+00, -0.16557669E-01, 0.12499936E+00, 
    },
    {
      0.73744383E+01, -0.12027929E+02, 0.55477055E+02, -0.12374017E+02, -0.89327345E+01, 0.21117227E+01, -0.63658524E+00, 0.36965864E+01, -0.23416463E+00, -0.27068796E+01, -0.28190041E+01, 0.65999812E+00, -0.12989928E+01, 
    },
    {
      0.79730363E+01, -0.21980247E+02, 0.14381426E+02, 0.35661685E+01, -0.12320098E+02, -0.91450326E-01, 0.10625300E+01, -0.12245990E+01, -0.24598341E+01, -0.45958142E+01, -0.12662302E+01, -0.16750321E+01, 0.63961756E+00, 
    },
    {
      -0.44315506E+02, -0.68935883E+02, 0.51209675E+02, -0.96216927E+01, 0.28568304E+01, -0.88125429E+01, 0.41833620E+01, 0.88361034E+01, -0.57595100E+01, 0.20263133E+01, 0.30380830E+00, 0.13347017E+01, 0.20728569E+01, 
    },
    {
      0.27656660E+02, -0.24090044E+02, -0.49807522E+02, -0.16521885E+02, 0.37152962E+02, 0.12428595E+02, 0.23935599E+01, -0.96629171E+01, 0.31866329E+01, -0.14910942E+00, -0.12049040E+01, 0.13381022E+01, -0.26929400E+01, 
    },
    {
      -0.27525063E+02, 0.94771980E+02, -0.32496066E+03, 0.86580139E+02, 0.72477280E+02, -0.22168686E+02, 0.94462023E+01, -0.26052528E+02, 0.84192997E+00, 0.15367923E+02, 0.14202004E+02, -0.84568520E+01, 0.30221646E+01, 
    },
    {
      -0.95830826E+02, 0.13345288E+03, -0.56603416E+02, -0.34000439E+02, 0.71806602E+02, -0.46853423E+01, -0.18473898E+02, 0.10775526E+02, 0.38866134E+01, 0.27761209E+02, 0.10534242E+02, 0.68109131E+01, -0.58825417E+01, 
    },
    {
      0.10576842E+03, 0.75757675E+02, -0.18107726E+03, 0.35365593E+02, -0.19049076E+02, 0.26869581E+02, -0.20695049E+02, -0.29340977E+02, 0.12610971E+02, -0.15228617E+01, -0.24438107E+01, -0.53549356E+01, -0.92385302E+01, 
    },
    {
      -0.89127052E+02, 0.11649306E+03, 0.99241447E+01, 0.59149529E+02, -0.12863148E+03, -0.36652939E+02, -0.66132178E+01, 0.37069744E+02, -0.13132428E+02, -0.28511498E+01, 0.12527537E+02, -0.64466619E+01, 0.80187778E+01, 
    },
    {
      0.66444099E+02, -0.26490027E+03, 0.78100250E+03, -0.24294917E+03, -0.21248206E+03, 0.73090462E+02, -0.29586170E+02, 0.75238579E+02, 0.11143709E+01, -0.39969158E+02, -0.29317850E+02, 0.27842539E+02, 0.15493094E+01, 
    },
    {
      0.31403052E+03, -0.33532126E+03, 0.10940713E+03, 0.94474915E+02, -0.15778313E+03, 0.12858322E+02, 0.64601768E+02, -0.41679054E+02, 0.12144041E+02, -0.63168148E+02, -0.34294418E+02, -0.11578428E+02, 0.18609144E+02, 
    },
    {
      -0.10748645E+03, -0.73261108E+01, 0.25742560E+03, -0.49046207E+02, 0.29797882E+02, -0.34313000E+02, 0.34743034E+02, 0.39293594E+02, -0.13067580E+02, -0.31035950E+01, 0.34878824E+01, 0.69859462E+01, 0.13996960E+02, 
    },
    {
      0.12487012E+03, -0.19734895E+03, 0.10182087E+03, -0.88575996E+02, 0.17575227E+03, 0.44216854E+02, 0.75572181E+01, -0.53957497E+02, 0.22110167E+02, 0.75596704E+01, -0.24996717E+02, 0.10142770E+02, -0.73031473E+01, 
    },
    {
      -0.81616684E+02, 0.30472897E+03, -0.86462183E+03, 0.28986890E+03, 0.26276410E+03, -0.95904716E+02, 0.36672890E+02, -0.93806793E+02, -0.43140082E+01, 0.47584782E+02, 0.27736128E+02, -0.35119770E+02, -0.99922857E+01, 
    },
    {
      -0.39365311E+03, 0.40164426E+03, -0.12205155E+03, -0.98447792E+02, 0.14291895E+03, -0.10083131E+02, -0.84789528E+02, 0.63032337E+02, -0.34860493E+02, 0.60776482E+02, 0.46811893E+02, 0.81084528E+01, -0.25287830E+02, 
    },
    {
      0.36671532E+02, -0.15971799E+02, -0.12797734E+03, 0.23251619E+02, -0.13679494E+02, 0.15429979E+02, -0.18713863E+02, -0.18289385E+02, 0.52698426E+01, 0.33367665E+01, -0.13960910E+01, -0.28204956E+01, -0.69832816E+01, 
    },
    {
      -0.59758083E+02, 0.11212505E+03, -0.80015900E+02, 0.45999847E+02, -0.81153366E+02, -0.18856775E+02, -0.28614230E+01, 0.26849733E+02, -0.12275286E+02, -0.47784538E+01, 0.14424383E+02, -0.51218362E+01, 0.15706040E+01, 
    },
    {
      0.38644615E+02, -0.12562950E+03, 0.36313803E+03, -0.12401902E+03, -0.11666864E+03, 0.43718800E+02, -0.16167812E+02, 0.41872963E+02, 0.27231398E+01, -0.20851528E+02, -0.99890919E+01, 0.15157324E+02, 0.68729215E+01, 
    },
    {
      0.16797034E+03, -0.18619617E+03, 0.57807312E+02, 0.33903023E+02, -0.45472706E+02, 0.14138937E+01, 0.37569668E+02, -0.32199303E+02, 0.21854696E+02, -0.20931599E+02, -0.22431284E+02, -0.17632341E+01, 0.12280620E+02, 
    },
    {
      0.67577185E-02, -0.38690273E-01, 0.12910472E+00, -0.96712691E+00, 0.28352141E+00, -0.23064802E-02, -0.67802047E-03, -0.94842110E-02, 0.16349863E-01, -0.21919638E-01, 0.69284305E-03, -0.18507710E-01, 0.65062190E-02, 
    },
    {
      0.98213874E-01, -0.86133815E-01, -0.43852173E-01, -0.25530991E+00, -0.75969112E+00, 0.14430066E-02, 0.11108719E-01, -0.26231494E-01, -0.16866442E-01, -0.13983527E-01, 0.39212308E-02, -0.13768057E-01, 0.86285966E-02, 
    },
    {
      0.52556640E+00, 0.14199088E+01, 0.23482423E+00, 0.12398671E+01, -0.11701978E+01, 0.22943036E+00, 0.19906037E-01, 0.14976382E+00, -0.83794696E-02, 0.12005270E+00, -0.11553056E+00, 0.17222632E-01, 0.61729096E-01, 
    },
    {
      0.25419283E+00, 0.50038338E+00, 0.77746254E+00, 0.17081389E+01, 0.14862339E+01, -0.36264654E-01, 0.18140624E+00, 0.25878024E+00, 0.94825074E-01, 0.19079769E-01, -0.11765101E-01, 0.11579243E+00, 0.53563844E-01, 
    },
    {
      0.29051917E+01, 0.17165275E+01, -0.23254345E+00, 0.30379848E+01, 0.28690646E+01, 0.66818811E-01, -0.14130628E+00, -0.79569364E+00, 0.35532689E+00, -0.21288022E+00, -0.17567311E+00, 0.18693624E-01, -0.56478500E+00, 
    },
    {
      0.17097101E+01, -0.15680478E+01, 0.19375648E+01, -0.37112069E+01, 0.15600331E+01, 0.15556124E+00, -0.19353969E+00, -0.24991187E+00, 0.85250163E+00, 0.35902387E+00, -0.43576801E+00, 0.26326782E+00, 0.53666756E-01, 
    },
    {
      -0.71911683E+01, -0.10550578E+02, 0.53046633E-01, 0.34881504E+01, 0.18754406E+01, -0.14867971E+01, -0.13179111E+01, -0.11313314E+01, -0.38169405E+00, -0.14188350E+01, 0.18088046E+01, 0.11685710E+00, 0.16416541E+00, 
    },
    {
      -0.77558784E+01, -0.51033101E+01, -0.53945317E+01, -0.79216189E+01, 0.20668960E+01, 0.22548456E+01, -0.17693250E+01, -0.83705181E+00, -0.15927854E+01, -0.11773171E+01, -0.69688298E-01, -0.17553112E+00, 0.24276218E+00, 
    },
    {
      -0.18529564E+02, -0.49476762E+01, -0.80366945E+01, -0.59576052E+00, -0.12046759E+02, 0.84566593E+00, 0.37427566E+01, 0.49305139E+01, -0.11143556E+01, 0.25452406E+01, -0.23028915E+00, 0.14220794E+00, 0.31511154E+01, 
    },
    {
      -0.19755463E+00, 0.14758796E+02, -0.15246205E+02, 0.17099882E+02, 0.68206816E+01, -0.16745818E+01, 0.30730574E+01, 0.30909929E+00, -0.25502460E+01, -0.20344477E+01, 0.21614161E+01, -0.16361508E+01, -0.19083089E+01, 
    },
    {
      0.17850922E+02, 0.23550045E+02, -0.43193951E+01, -0.17244564E+02, 0.45090499E+01, 0.36393521E+01, 0.38667064E+01, 0.35419881E+01, 0.14072553E+01, 0.40633569E+01, -0.58945403E+01, -0.35010526E+00, -0.12329621E+01, 
    },
    {
      0.25905699E+02, 0.12087793E+02, 0.13750561E+02, 0.92878151E+01, -0.13878139E+02, -0.71122155E+01, 0.60348301E+01, 0.10128769E+01, 0.49910145E+01, 0.45441375E+01, -0.48133820E+00, -0.69957542E+00, -0.12999972E+01, 
    },
    {
      0.35084892E+02, 0.58390584E+01, 0.27396364E+02, -0.14684110E+02, 0.16126560E+02, -0.30175450E+01, -0.90232534E+01, -0.10018678E+02, 0.10488937E+01, -0.65118580E+01, 0.26540687E+01, -0.18480808E+00, -0.52721381E+01, 
    },
    {
      -0.17716702E+02, -0.33102341E+02, 0.34046539E+02, -0.25810240E+02, -0.28638424E+02, 0.30514753E+01, -0.93732405E+01, 0.12338361E+00, 0.99789155E+00, 0.36990767E+01, -0.32632139E+01, 0.32731326E+01, 0.55128732E+01, 
    },
    {
      -0.13897304E+02, -0.15651490E+02, 0.54399266E+01, 0.10493177E+02, -0.10791659E+02, -0.24484687E+01, -0.28806729E+01, -0.28858256E+01, -0.10917011E+01, -0.32969794E+01, 0.47821879E+01, 0.21285339E+00, 0.13126817E+01, 
    },
    {
      -0.23277504E+02, -0.86345367E+01, -0.10629125E+02, 0.21820576E+01, 0.74627986E+01, 0.50995250E+01, -0.59439478E+01, -0.66435319E+00, -0.40929780E+01, -0.37793703E+01, 0.67755610E+00, 0.11436670E+01, 0.11544902E+01, 
    },
    {
      -0.19324162E+02, -0.31898675E+01, -0.20808168E+02, 0.17447025E+02, -0.33315101E+01, 0.21073177E+01, 0.58234086E+01, 0.62877498E+01, -0.35331634E+00, 0.48601122E+01, -0.26413305E+01, 0.93884856E-01, 0.25990388E+01, 
    },
    {
      0.21725630E+02, 0.20295200E+02, -0.24284405E+02, 0.83008165E+01, 0.26109770E+02, -0.13780729E+01, 0.82305450E+01, 0.23272064E+00, 0.11319129E+01, -0.20607281E+01, 0.17115957E+01, -0.22145329E+01, -0.40296369E+01, 
    },
    {
      0.11448875E+00, 0.53372294E-01, -0.10701045E-01, -0.74131079E-01, 0.42126238E-01, -0.47183973E+00, -0.62683088E+00, -0.49725015E-01, 0.23906445E-01, -0.18615820E-01, 0.69484608E-02, -0.43719456E-01, -0.12143527E-01, 
    },
    {
      0.18777304E-03, -0.18922636E+00, -0.75974211E-01, -0.15254694E-01, -0.11909236E+00, 0.59676683E+00, -0.50491786E+00, -0.35921060E-01, 0.58115725E-02, 0.59416122E-02, 0.95220320E-02, -0.17321225E-01, 0.72834064E-03, 
    },
    {
      -0.31445304E+00, 0.10098319E+01, -0.87438963E-01, -0.22055499E+00, -0.36901802E+00, 0.44668826E+00, 0.42858791E+00, 0.21535960E+00, -0.43418493E-01, -0.49117277E-02, 0.10932283E+00, -0.98627552E-01, 0.37286639E-01, 
    },
    {
      -0.16686460E+00, -0.16635004E+00, -0.32959878E-01, -0.37202913E-01, -0.50687611E+00, -0.32216674E+00, 0.34823060E+00, 0.52472215E-01, 0.84805384E-01, 0.22642313E-01, -0.84570244E-01, -0.48971225E-01, 0.45566838E-01, 
    },
    {
      -0.63517235E-01, -0.32054949E+00, 0.87246031E-01, 0.14539485E+00, -0.43560812E+00, 0.26938108E+00, 0.18754629E-01, -0.34433404E+00, 0.29875723E+00, 0.17472742E+00, 0.54090742E-01, 0.38365376E+00, -0.13843028E+00, 
    },
    {
      -0.48946375E+00, 0.28971550E+01, -0.20743754E-01, -0.83255611E-01, 0.11938328E+01, 0.55411711E-01, 0.60760695E+00, -0.29698908E+00, -0.55650949E+00, 0.20695293E-01, 0.36676876E-01, 0.14274941E+00, -0.98814443E-01, 
    },
    {
      -0.46403563E+00, -0.19422119E+01, 0.36596009E+00, 0.68322176E+00, 0.45094913E+00, 0.16012102E+00, -0.40091038E+00, -0.46582344E+00, 0.89832827E-01, 0.94689272E-01, -0.34453103E+00, 0.32739177E+00, -0.56781594E-01, 
    },
    {
      -0.38054120E-01, 0.52844387E+00, 0.19522023E+00, 0.33836249E-01, 0.12656829E+01, -0.14137593E-01, 0.38812596E+00, -0.22139823E+00, -0.17572556E+00, -0.40329311E-01, 0.20087887E+00, 0.30446243E+00, -0.45953188E-01, 
    },
    {
      0.10138659E+01, 0.14339803E+01, -0.45818272E+00, -0.24776068E+00, 0.95378476E+00, 0.11445934E+00, 0.66230971E+00, 0.79100257E+00, -0.76573831E+00, -0.34316742E+00, 0.80294624E-01, -0.70623606E+00, 0.29327330E+00, 
    },
    {
      0.63751262E+00, -0.37546885E+01, -0.70940018E-01, -0.13639070E+00, -0.21821494E+01, -0.36584163E+00, -0.49187639E+00, 0.82650185E+00, 0.85966331E+00, -0.19263405E+00, -0.79641826E-01, -0.47789237E+00, 0.15595616E+00, 
    },
    {
      0.34587827E-01, -0.33244256E-01, -0.54627232E-01, -0.24314072E-01, 0.39300185E-01, 0.12838095E-01, -0.27576180E-01, 0.22983019E+00, -0.15771292E+00, -0.23307070E-01, 0.17552825E-01, -0.18896224E-01, -0.17950874E-01, 
    },
    {
      0.24897996E-01, -0.81121093E-02, -0.34756999E-01, 0.31306081E-01, -0.67682348E-01, 0.25938660E-01, -0.15375219E-01, 0.15197173E+00, 0.24201533E+00, -0.26705752E-02, 0.45459899E-02, 0.56280023E-02, -0.77642538E-02, 
    },
    {
      -0.17782798E+00, -0.95260762E-01, 0.24179123E-01, 0.14758726E-01, -0.92235729E-01, -0.15296949E+00, 0.94236672E-01, 0.82640037E-01, 0.50521767E+00, 0.33969387E-01, -0.62877722E-01, 0.20034479E-01, 0.95262146E-02, 
    },
    {
      0.32055417E+00, 0.20525834E+00, -0.76902891E-02, -0.64364143E-01, 0.19882102E+00, -0.81640065E-01, -0.71830384E-01, -0.49157616E+00, -0.57283647E-01, 0.12709680E-01, -0.30256083E-01, -0.39767653E-01, -0.55746748E-02, 
    },
    {
      -0.16962911E+00, -0.24519250E-01, 0.53539060E-01, -0.22426639E-01, -0.52650891E-01, 0.57139243E-02, -0.12633940E-01, 0.43505147E-01, -0.18997150E-01, 0.14223887E+00, 0.16691910E+00, 0.28885426E-02, -0.21186968E-01, 
    },
    {
      0.75518645E-01, 0.52836422E-01, -0.81654236E-01, 0.27446862E-01, -0.50890606E-01, -0.17863950E-01, -0.73899115E-02, 0.33153400E-01, 0.59054721E-01, -0.16145012E+00, 0.14286152E+00, 0.14969439E-01, -0.21417528E-01, 
    },
    {
      -0.22653613E-01, -0.57655737E-01, -0.41951939E-01, 0.21979176E-01, 0.39826494E-01, 0.30946506E-02, -0.23654750E-01, 0.27493823E-01, 0.20307899E-01, 0.68600680E-03, -0.30355731E-01, -0.89675814E-01, 0.58423135E-01, 
    },
    {
      0.93922287E-01, 0.68419144E-01, 0.33733767E-01, -0.72773839E-02, 0.48324339E-01, -0.24635356E-02, -0.89860288E-02, -0.79068150E-02, 0.14755895E-01, -0.16021999E-01, -0.71704332E-02, -0.60977753E-01, -0.10361492E+00, 
    },
    {
      0.12631768E-01, -0.49197290E-01, 0.24660140E-01, 0.43001071E-01, 0.69695385E-02, -0.31781915E-03, 0.53595994E-02, 0.34497995E-01, 0.24565097E-01, -0.50819240E-03, -0.10135818E-01, -0.16772541E-02, 0.32331720E-02, 
    },
    {
      0.38050801E-01, -0.56574538E-01, -0.31323060E-01, -0.14262537E-01, -0.38575940E-01, 0.97641703E-02, -0.77285576E-02, -0.27659681E-01, 0.20871909E-01, -0.15837995E-01, 0.18971974E-01, -0.38362518E-02, 0.23249378E-01, 
    },
    {
      0.32181796E-01, 0.80461048E-01, -0.23041502E-01, 0.28507439E-01, -0.57646193E-01, -0.25178518E-01, 0.10214640E-01, -0.14632375E-02, 0.22060374E-01, -0.57708239E-02, 0.16520452E-01, -0.30608475E-01, -0.77883643E-02, 
    },
    {
      -0.38133394E-01, 0.41286569E-01, 0.27755771E-01, -0.51213089E-01, 0.75025320E-01, 0.91396663E-02, -0.18474469E-02, -0.56077321E-02, 0.64123259E-02, 0.22182658E-01, -0.65464103E-02, -0.81443517E-02, 0.73059131E-02, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir16_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30632045E+01, -0.57797851E-02, 0.91647813E-02, 0.41370955E-02, -0.19437239E-02, 0.24532698E-03, 0.60106777E-02, -0.21756680E-02, -0.31438239E-02, 
    },
    {
      -0.41220626E+00, 0.29579294E-02, 0.18636411E-01, 0.37512437E-01, -0.35412803E-01, 0.96220765E-02, -0.19383488E-01, -0.20679396E-02, 0.28241726E-02, 
    },
    {
      0.10044123E+01, 0.77929936E-01, -0.67139186E-01, -0.42513080E-01, -0.39895583E-01, -0.43510310E-01, 0.13425289E-01, 0.35680491E-01, 0.17047495E-01, 
    },
    {
      0.43130204E+00, -0.24383721E+00, -0.48306435E-01, -0.16724011E+00, 0.17723440E+00, -0.52420892E-01, 0.50915804E-01, -0.49679428E-02, 0.87580746E-02, 
    },
    {
      -0.12934322E+01, 0.17085487E-01, 0.12944695E+00, -0.12789106E-01, 0.11405879E+00, 0.17716755E+00, -0.93909204E-01, -0.11515923E+00, -0.44955883E-01, 
    },
    {
      -0.30758275E-01, 0.21726264E+00, 0.53814311E-01, 0.14063519E+00, -0.15107477E+00, 0.41386593E-01, -0.35365798E-01, 0.10928157E-01, -0.13283265E-01, 
    },
    {
      0.26758054E+00, -0.95864914E-01, -0.27972437E-01, 0.68929851E-01, -0.87709561E-01, -0.14526750E+00, 0.82934819E-01, 0.83224915E-01, 0.33448048E-01, 
    },
    {
      0.21264717E-01, -0.10639626E+00, -0.32184544E+00, 0.10739344E-01, -0.15501696E-01, 0.23128907E-02, -0.49819713E-02, 0.26865129E-02, 0.11654526E-02, 
    },
    {
      0.12853952E-02, 0.34511042E+00, -0.64418115E-01, -0.11180845E-01, 0.79391981E-02, 0.22869739E-02, -0.77759842E-02, 0.74287609E-03, 0.91153430E-02, 
    },
    {
      -0.58623929E-01, -0.22482648E+00, -0.12649196E+00, 0.18788284E-01, -0.15773132E-01, -0.67276560E-01, -0.85366756E-01, 0.54488275E-01, 0.43759987E-01, 
    },
    {
      0.15944013E-01, 0.91821142E-01, -0.17644204E+00, 0.27534859E-01, 0.95351227E-02, -0.37926875E-01, -0.12159072E-02, 0.34995351E-01, 0.62579088E-01, 
    },
    {
      -0.14691392E+00, 0.48885843E+00, 0.11924573E+01, -0.97168460E-01, -0.39609086E-01, -0.14198712E+00, 0.43608997E-01, 0.92370085E-01, -0.18335907E-01, 
    },
    {
      -0.61437911E+00, -0.14346495E+01, -0.14602734E-01, 0.17535535E+00, -0.40383213E-02, -0.13587080E+00, -0.25987262E-01, -0.29818421E-01, -0.44571441E-01, 
    },
    {
      0.85389566E+00, 0.11196742E+01, -0.79580927E+00, 0.13261000E-01, -0.15927894E+00, 0.72401887E+00, 0.55747724E+00, -0.33719021E+00, -0.25143474E+00, 
    },
    {
      0.10345604E+01, 0.16206180E+01, 0.10684304E+01, -0.67587040E-01, 0.19474277E+00, 0.26491645E+00, 0.16167685E-01, -0.36782002E+00, -0.53570050E+00, 
    },
    {
      0.32261285E+00, -0.64737231E+00, -0.90475298E-01, 0.17187940E+00, 0.13722499E+00, 0.50946969E+00, -0.93138762E-01, -0.32658824E+00, -0.66461273E-01, 
    },
    {
      0.22725639E+01, 0.69083470E+00, 0.48083431E+00, -0.49758098E+00, -0.84927388E-01, 0.52185225E+00, 0.19340727E+00, 0.32786693E-01, 0.78221820E-01, 
    },
    {
      -0.20661349E+01, -0.26972232E+01, 0.25274463E+01, -0.41764119E+00, 0.75514770E+00, -0.17189139E+01, -0.10526706E+01, 0.63819683E+00, 0.37426770E+00, 
    },
    {
      -0.32621279E+01, -0.44142714E+01, -0.25552752E+01, -0.18484338E+00, -0.58688587E+00, -0.73055243E+00, -0.17521368E+00, 0.10259843E+01, 0.11375989E+01, 
    },
    {
      -0.73171481E-01, 0.42583603E+00, -0.10637463E+01, -0.14348125E+00, -0.74651487E-01, -0.45749301E+00, 0.43486070E-01, 0.28321269E+00, 0.10276017E+00, 
    },
    {
      -0.18371630E+01, 0.76073569E+00, -0.44730103E+00, 0.35803327E+00, 0.80352314E-01, -0.40243587E+00, -0.17068449E+00, 0.65292539E-02, -0.63528381E-01, 
    },
    {
      0.13972975E+01, 0.22084525E+01, -0.22679477E+01, 0.48937282E+00, -0.68422282E+00, 0.11014562E+01, 0.64520347E+00, -0.39747542E+00, -0.17840618E+00, 
    },
    {
      0.23154075E+01, 0.34323814E+01, 0.20233507E+01, 0.38661447E+00, 0.40558794E+00, 0.58963746E+00, 0.15238740E+00, -0.77325076E+00, -0.66182941E+00, 
    },
    {
      0.40216600E-02, 0.81703132E-02, -0.19011486E-01, -0.47476284E-01, -0.14503382E+00, -0.59457687E-02, -0.60895034E-02, 0.35825928E-02, -0.50007328E-02, 
    },
    {
      -0.10892768E-01, -0.62809291E-03, -0.17020314E-03, 0.16835703E+00, -0.23484120E-01, 0.19963286E-02, -0.22885388E-01, 0.47483742E-02, 0.63064597E-02, 
    },
    {
      0.95548488E-01, 0.15581782E+00, 0.46200905E-01, 0.19114015E+00, -0.25267342E+00, 0.37239853E-02, 0.66554118E-02, 0.20674267E-02, -0.10625846E-01, 
    },
    {
      0.15886465E-01, -0.15695851E-01, -0.18671377E-01, 0.40752470E+00, 0.28736994E+00, 0.15669392E-01, -0.44707544E-01, 0.49465522E-02, 0.24645288E-01, 
    },
    {
      -0.16001102E+00, 0.39749850E-01, 0.11758360E+00, 0.14307669E+00, -0.21471813E+00, 0.79647778E-02, 0.41321028E-01, -0.32722354E-02, 0.57371218E-01, 
    },
    {
      -0.38416374E+00, -0.51390864E-01, -0.17938653E+00, 0.15849835E+00, 0.80724046E-01, -0.29265678E-02, 0.96612900E-01, -0.67667104E-01, 0.22370834E-01, 
    },
    {
      -0.27011070E+00, -0.86970943E+00, -0.26902573E-01, -0.77681959E+00, 0.69076121E+00, 0.41663114E-01, -0.10570973E+00, 0.84764361E-01, 0.69376291E-02, 
    },
    {
      0.42020214E+00, 0.23183244E+00, 0.38358140E+00, -0.11147493E+01, -0.10411042E+01, 0.53920228E-01, 0.16102199E+00, -0.14279890E-01, -0.27331126E+00, 
    },
    {
      0.14651980E+00, -0.24204558E+00, -0.19398539E+00, -0.30382502E+00, 0.62044007E+00, 0.47480121E-01, -0.15849266E-01, -0.10659328E-01, -0.80701113E-01, 
    },
    {
      0.72861421E+00, -0.92455231E-01, 0.29580644E+00, -0.54386604E+00, -0.24819843E+00, 0.39313916E-01, -0.49355179E-01, 0.44585172E-01, -0.85791960E-01, 
    },
    {
      0.11903773E+00, 0.12535146E+01, 0.82356326E-01, 0.67133921E+00, -0.63358366E+00, -0.49258057E-01, 0.10747476E+00, -0.10950665E+00, 0.64388467E-02, 
    },
    {
      -0.83330280E+00, -0.29580790E+00, -0.38530514E+00, 0.86506933E+00, 0.80623895E+00, -0.13946185E+00, -0.10462884E+00, 0.11260598E-01, 0.38791931E+00, 
    },
    {
      -0.10402253E-01, 0.13346081E-01, 0.63475878E-02, -0.27291005E-01, -0.13256154E-02, 0.17733403E-01, 0.11105449E-01, -0.21581961E-03, -0.40799612E-03, 
    },
    {
      -0.13009773E-01, -0.28726665E-01, -0.28361091E-01, 0.41243983E-02, -0.13775193E-01, -0.92879981E-02, 0.12643198E-01, -0.23260165E-02, 0.20666828E-02, 
    },
    {
      0.73535770E-01, 0.55688366E-01, 0.99641196E-02, -0.35065774E-01, 0.18284770E-01, 0.44654209E-01, 0.16372960E-01, -0.12499312E-02, 0.70338654E-04, 
    },
    {
      0.43763254E-01, -0.15740022E-01, -0.13505678E-01, -0.87119667E-02, 0.10261146E-01, -0.21680510E-01, 0.51671389E-01, -0.18428918E-01, 0.62357122E-02, 
    },
    {
      -0.52719351E-01, -0.13156223E+00, -0.79338402E-01, 0.49488008E-01, 0.24957510E-01, -0.92251182E-01, -0.94008923E-01, 0.44068401E-02, 0.33692043E-01, 
    },
    {
      -0.60246535E-01, 0.86852990E-01, 0.20347420E-01, -0.17174080E-02, 0.17042400E-01, 0.74888632E-01, -0.72979771E-01, 0.60993590E-03, -0.11566933E-02, 
    },
    {
      -0.14518532E-01, -0.12442473E-01, 0.85507222E-02, 0.22372473E-02, -0.15677895E-02, 0.13057685E-01, -0.67228526E-02, 0.22530334E-01, 0.29998149E-02, 
    },
    {
      -0.19739617E-01, 0.10189696E-01, -0.83538266E-02, 0.88368857E-03, -0.46331584E-02, 0.45361207E-03, 0.11164929E-01, -0.60726954E-02, 0.22200333E-01, 
    },
    {
      0.76710410E-01, 0.17538734E-01, 0.20874044E-01, -0.30737719E-01, -0.15248102E-01, 0.13212540E-01, 0.14518229E-01, 0.43812137E-01, 0.32291189E-01, 
    },
    {
      -0.13844927E-01, -0.34613054E-01, 0.59420760E-02, 0.18336574E-01, 0.12362840E-01, 0.78165001E-03, 0.94650965E-02, -0.54954968E-01, 0.48140671E-01, 
    },
    {
      0.21998519E-02, -0.30175578E-02, -0.21670593E-02, 0.24198068E-02, -0.30340797E-02, 0.98152589E-02, -0.65381783E-02, 0.18372700E-02, -0.32104759E-02, 
    },
    {
      0.15909625E-01, 0.17480101E-01, 0.73182187E-03, 0.53907954E-02, -0.10893663E-01, -0.14815384E-02, 0.32584106E-02, 0.13903450E-02, -0.29880791E-02, 
    },
    {
      0.18037504E-01, -0.34718488E-02, 0.86077778E-02, 0.51989104E-02, -0.11354344E-01, 0.33079886E-02, -0.26848447E-02, 0.73557561E-02, 0.26780984E-03, 
    },
    {
      0.12174977E-02, 0.38145680E-02, -0.32383250E-02, 0.11604379E-01, 0.10995209E-01, -0.69180368E-02, 0.18747009E-02, -0.57764836E-02, -0.31420870E-02, 
    },
  },
  {
    {
      0.27280636E+01, -0.75169103E-02, 0.63383449E-02, 0.95558316E-02, -0.28542201E-02, -0.33672303E-02, 0.22478853E-02, 0.12390611E-02, -0.25694482E-02, 
    },
    {
      -0.34184477E+00, 0.26675481E-01, 0.36192123E-01, 0.17344089E-01, -0.27918063E-01, 0.46071778E-02, -0.19322153E-01, 0.14321898E-02, -0.45877192E-02, 
    },
    {
      0.11131347E+01, 0.26039990E-01, -0.86778462E-01, -0.36485463E-01, 0.39370175E-01, 0.96002901E-02, 0.23496447E-01, 0.10478078E-01, 0.15071011E-01, 
    },
    {
      0.26609913E+00, -0.37570027E+00, -0.76007955E-01, -0.84936224E-01, 0.13073616E+00, -0.46946578E-01, 0.45213178E-01, -0.10088617E-01, 0.15226900E-01, 
    },
    {
      -0.13583761E+01, 0.24481584E+00, 0.19391088E+00, -0.25062395E-01, -0.17285113E+00, 0.35910089E-01, -0.83230503E-01, -0.49273174E-01, -0.39427180E-01, 
    },
    {
      -0.86215436E-02, 0.33480579E+00, 0.59670329E-01, 0.72379611E-01, -0.10527881E+00, 0.40745031E-01, -0.27480911E-01, 0.10371030E-01, -0.11401369E-01, 
    },
    {
      0.24424902E+00, -0.25925013E+00, -0.73127270E-01, 0.64353183E-01, 0.13510229E+00, -0.47481738E-01, 0.65662742E-01, 0.38441211E-01, 0.27695235E-01, 
    },
    {
      0.26992045E-01, -0.27138302E+00, -0.25009316E+00, 0.13358043E-01, -0.99321315E-02, -0.14139374E-02, -0.65408170E-03, 0.11991658E-02, 0.26755203E-02, 
    },
    {
      -0.18171994E-01, 0.28280777E+00, -0.23699513E+00, -0.49411627E-02, 0.77506038E-02, 0.32914598E-02, -0.10480203E-01, 0.53377454E-02, 0.11933497E-01, 
    },
    {
      0.55335838E-01, -0.93998998E-01, -0.40524747E-01, -0.86126318E-02, 0.32151006E-01, -0.25418943E-01, -0.56042250E-01, 0.22723371E-01, 0.48075102E-01, 
    },
    {
      0.21400840E-02, 0.54945960E-02, -0.11576591E+00, 0.33427548E-01, 0.10076365E-01, -0.27823556E-01, 0.87937471E-02, 0.16767824E-01, 0.51116429E-01, 
    },
    {
      -0.27947503E+00, 0.93270457E+00, 0.78121406E+00, -0.63873820E-01, 0.13776053E-01, -0.41885123E-01, -0.14760400E-01, 0.49246795E-01, -0.38276151E-01, 
    },
    {
      -0.38543758E+00, -0.12443132E+01, 0.63409525E+00, 0.10198897E+00, -0.18275555E-01, -0.13154987E+00, 0.44033036E-01, -0.12012166E+00, -0.11587386E+00, 
    },
    {
      -0.24930301E+00, 0.41436785E+00, -0.42938063E+00, 0.14058124E+00, -0.51638871E+00, 0.26499906E+00, 0.30564678E+00, -0.35653755E-01, -0.30761656E+00, 
    },
    {
      0.96339178E+00, 0.17102709E+01, 0.64794040E+00, -0.14055821E+00, 0.10397163E+00, 0.27128929E+00, -0.49434416E-01, -0.15722241E+00, -0.39522058E+00, 
    },
    {
      0.71117145E+00, -0.79698235E+00, 0.35370767E+00, 0.88413291E-01, 0.65246043E-02, 0.25285155E+00, 0.90608612E-01, -0.12623823E+00, 0.72437324E-01, 
    },
    {
      0.16189957E+01, 0.92443705E+00, -0.42564002E+00, -0.32307896E+00, -0.47557414E-01, 0.47201774E+00, -0.36377423E-01, 0.27063537E+00, 0.27386218E+00, 
    },
    {
      0.49492338E+00, -0.11964587E+01, 0.11970491E+01, -0.51224530E+00, 0.14279464E+01, -0.65490454E+00, -0.49377796E+00, -0.79987206E-01, 0.57029313E+00, 
    },
    {
      -0.29584346E+01, -0.44350519E+01, -0.13688515E+01, 0.90590253E-01, -0.28306559E+00, -0.70212859E+00, -0.48325151E-01, 0.42926690E+00, 0.80071324E+00, 
    },
    {
      -0.36841384E+00, 0.36017466E+00, -0.10382642E+01, -0.75529769E-01, -0.65040000E-01, -0.29723659E+00, -0.82611166E-01, 0.96393012E-01, -0.45457043E-01, 
    },
    {
      -0.13651522E+01, 0.12593678E+00, 0.79192951E-01, 0.25124004E+00, 0.63250221E-01, -0.38624293E+00, 0.94562238E-02, -0.15385661E+00, -0.18089309E+00, 
    },
    {
      -0.32429627E+00, 0.10307864E+01, -0.13849024E+01, 0.49472693E+00, -0.10159550E+01, 0.43086207E+00, 0.25678623E+00, 0.93890756E-01, -0.33852810E+00, 
    },
    {
      0.21680369E+01, 0.36000590E+01, 0.99755543E+00, 0.11148506E+00, 0.17308816E+00, 0.52738160E+00, 0.95052890E-01, -0.32424158E+00, -0.48464921E+00, 
    },
    {
      -0.40881429E-02, 0.31389890E-02, 0.18125862E-02, 0.19379327E-01, -0.11340187E+00, -0.64218352E-02, -0.21440715E-02, 0.86666498E-03, -0.48338207E-02, 
    },
    {
      -0.12114726E-01, -0.90989377E-02, 0.15213010E-01, 0.11901259E+00, 0.44593573E-01, 0.35579519E-02, -0.16553327E-01, 0.31294818E-02, 0.35295503E-02, 
    },
    {
      0.96299417E-01, 0.16172042E+00, 0.36571477E-01, 0.13853014E+00, -0.11524634E+00, 0.13647977E-01, 0.85650487E-02, 0.21059543E-02, -0.88550262E-02, 
    },
    {
      0.78764156E-01, 0.23560308E-01, -0.20360565E-01, 0.20476563E+00, 0.23432060E+00, 0.29904551E-02, -0.21854280E-01, 0.12916533E-02, 0.17185966E-01, 
    },
    {
      -0.13958269E+00, 0.61252635E-01, -0.22732537E-01, -0.22977982E+00, -0.42527966E-01, 0.48952647E-01, 0.13864973E-01, -0.26283702E-01, 0.68163075E-01, 
    },
    {
      -0.25024179E+00, -0.60532674E-01, -0.19260155E+00, 0.39119691E-01, -0.26626396E+00, -0.44200186E-01, 0.90043500E-01, -0.57880651E-01, 0.20709954E-01, 
    },
    {
      -0.24203920E+00, -0.77736235E+00, -0.10913799E-01, -0.15193807E+00, 0.20603824E+00, -0.31180762E-01, -0.11680545E+00, 0.67744017E-01, 0.44562139E-01, 
    },
    {
      0.29965173E-01, 0.71288981E-01, 0.24611737E+00, -0.42808515E+00, -0.44561577E+00, 0.12265429E+00, 0.95303804E-01, 0.38496863E-01, -0.18627241E+00, 
    },
    {
      0.16500773E+00, -0.28346860E+00, 0.19127062E-01, 0.85767560E-01, 0.18231672E+00, -0.66293105E-01, 0.29998733E-01, 0.59395704E-01, -0.82480334E-01, 
    },
    {
      0.46248379E+00, -0.12290555E+00, 0.23955330E+00, -0.17114018E+00, 0.85420050E-01, 0.51798742E-01, -0.63089624E-01, 0.69193467E-01, -0.49095791E-01, 
    },
    {
      0.49044739E-01, 0.10698808E+01, 0.55030648E-01, -0.20411465E-01, -0.20164534E+00, 0.81138603E-01, 0.11107465E+00, -0.91995150E-01, -0.63594230E-01, 
    },
    {
      -0.31178156E+00, 0.25681297E-02, -0.21641961E+00, 0.26155630E+00, 0.19081634E+00, -0.18793817E+00, -0.63471727E-01, -0.60597725E-01, 0.24745049E+00, 
    },
    {
      -0.20798016E-01, 0.15350298E-01, 0.10828707E-01, -0.18462822E-01, -0.11293224E-01, 0.18085541E-01, -0.18076560E-01, -0.26808286E-02, -0.14639653E-02, 
    },
    {
      -0.94376355E-02, -0.31904358E-01, -0.19534955E-01, 0.60758656E-02, -0.13963651E-01, 0.17930444E-01, 0.15960965E-01, -0.58656308E-03, -0.17829806E-02, 
    },
    {
      0.50320923E-01, 0.85855640E-01, -0.37249629E-03, -0.27981598E-01, 0.27195103E-02, 0.81393354E-01, 0.11739279E-01, -0.58815950E-02, -0.22160930E-02, 
    },
    {
      0.30873053E-01, -0.98756980E-02, 0.89320733E-03, -0.10296550E-01, -0.80322934E-03, -0.16821079E-01, 0.74325636E-01, -0.80107599E-02, 0.32815721E-02, 
    },
    {
      -0.20960687E-01, -0.15283193E+00, -0.57994306E-01, 0.26183322E-01, 0.26019171E-01, -0.12097655E+00, -0.35915900E-01, -0.33830630E-02, 0.31820308E-01, 
    },
    {
      -0.43779407E-01, 0.92425197E-01, -0.36082782E-02, 0.10040569E-01, 0.69529493E-02, 0.14791758E-01, -0.99151865E-01, -0.59039001E-02, -0.60759163E-02, 
    },
    {
      -0.11967236E-01, -0.15639639E-01, 0.11332542E-01, 0.19568345E-03, -0.19746975E-02, 0.52477429E-02, -0.68195597E-02, 0.18582528E-01, 0.10505051E-01, 
    },
    {
      -0.10374394E-01, 0.77973618E-02, -0.14768258E-01, 0.33124851E-02, -0.50527342E-02, 0.13333512E-02, 0.80117164E-02, -0.10721104E-01, 0.16201984E-01, 
    },
    {
      0.51714681E-01, 0.20969117E-01, 0.20908019E-01, -0.21666186E-01, -0.65546874E-02, 0.10858584E-01, 0.19092079E-01, 0.34162614E-01, 0.63728623E-01, 
    },
    {
      -0.17780149E-01, -0.39414037E-01, 0.19805619E-01, 0.27769320E-02, 0.32051459E-02, -0.13837150E-02, 0.38510279E-02, -0.78942724E-01, 0.31475380E-01, 
    },
    {
      0.72736945E-02, -0.99098906E-02, 0.94700296E-03, 0.49195858E-02, -0.48104613E-02, 0.46512815E-02, -0.50494503E-02, 0.46232669E-02, -0.99206646E-03, 
    },
    {
      0.11354908E-01, 0.14260411E-01, -0.41098539E-02, 0.63321404E-02, -0.46809567E-02, 0.28002951E-02, 0.52125915E-02, 0.95991232E-03, 0.42521297E-02, 
    },
    {
      0.14690466E-01, 0.30270962E-02, 0.53093689E-02, 0.74441070E-02, -0.92850029E-02, 0.12141704E-02, -0.84077264E-03, 0.41119582E-02, 0.30829709E-02, 
    },
    {
      -0.49859304E-02, -0.33042641E-03, -0.43395390E-02, 0.45590200E-02, 0.91798063E-02, -0.12648865E-02, 0.12946193E-02, -0.44219219E-02, 0.19899376E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir17_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.47669554E+01, -0.72017536E-01, 0.13942049E+00, 0.19542389E-01, -0.16752634E-01, -0.16510993E-01, -0.43158536E-03, 0.28941497E-01, -0.32931969E-01, 0.10719537E-02, 0.56658965E-02, 0.12240363E-01, -0.93685742E-02, 
    },
    {
      -0.99394917E+00, -0.35927990E+00, 0.43325570E+00, -0.22593535E+00, 0.49836481E+00, -0.91735348E-01, 0.10013926E+00, -0.16090275E+00, -0.16553175E+00, -0.32432500E-01, -0.38297892E-01, -0.82239382E-01, -0.42720828E-02, 
    },
    {
      0.12144961E+02, 0.79801214E+00, -0.25601573E+01, -0.68445534E-01, 0.18775876E+00, 0.16825244E+00, 0.18251759E+00, -0.42203125E+00, 0.31813495E-01, 0.75639360E-01, -0.42337403E+00, -0.16204225E+00, 0.93350336E-01, 
    },
    {
      0.12359418E+02, 0.26116188E+01, -0.71795487E+00, 0.47592411E+01, -0.74272795E+01, 0.17137471E+01, -0.29525971E+01, 0.15546530E+01, 0.12446737E+01, 0.55519021E+00, -0.42093495E+00, 0.19277966E+01, 0.44809118E+00, 
    },
    {
      -0.81837517E+02, -0.65713782E+01, 0.11112719E+02, 0.46813384E+00, -0.61830668E+01, -0.12024459E+01, -0.17724164E+01, 0.16484222E+01, 0.20026155E+01, -0.97772136E-01, 0.29807785E+01, 0.10582246E+01, -0.70731813E+00, 
    },
    {
      -0.72209411E+01, -0.77260118E+01, -0.15136260E+02, -0.22104202E+02, 0.39128777E+02, -0.83586140E+01, 0.14526752E+02, -0.68457294E+01, -0.29666450E+01, -0.23397076E+01, 0.42374215E+01, -0.12021458E+02, -0.40962782E+01, 
    },
    {
      0.19595665E+03, 0.24501345E+02, -0.20282061E+02, -0.20203459E+01, 0.23160069E+02, 0.31945641E+01, 0.48780818E+01, -0.33812382E+01, -0.74402051E+01, -0.85799468E+00, -0.81885357E+01, -0.29702859E+01, 0.22015822E+01, 
    },
    {
      -0.64242828E+02, 0.68489776E+01, 0.53867455E+02, 0.37935581E+02, -0.88741364E+02, 0.16450567E+02, -0.26521566E+02, 0.13563215E+02, 0.29131477E+01, 0.35748591E+01, -0.11355737E+02, 0.28946178E+02, 0.11574268E+02, 
    },
    {
      -0.21447914E+03, -0.34901211E+02, 0.17094313E+02, 0.27981231E+01, -0.29804564E+02, -0.32461698E+01, -0.52834835E+01, 0.37156496E+01, 0.90832834E+01, 0.20808840E+01, 0.93321514E+01, 0.36256206E+01, -0.27270098E+01, 
    },
    {
      0.11643464E+03, 0.41456490E+01, -0.62626999E+02, -0.26506950E+02, 0.89930573E+02, -0.14143736E+02, 0.20422012E+02, -0.11808813E+02, -0.10933266E+01, -0.21905138E+01, 0.12235158E+02, -0.29777472E+02, -0.13072017E+02, 
    },
    {
      0.86740616E+02, 0.16090628E+02, -0.54182315E+01, -0.11867726E+01, 0.12633715E+02, 0.11003165E+01, 0.20218921E+01, -0.16267680E+01, -0.36661091E+01, -0.12221850E+01, -0.37074435E+01, -0.15805085E+01, 0.11606567E+01, 
    },
    {
      -0.55534855E+02, -0.55097966E+01, 0.24141411E+02, 0.61227908E+01, -0.33446136E+02, 0.44106121E+01, -0.55466518E+01, 0.36874857E+01, 0.85909329E-01, 0.44831970E+00, -0.46812630E+01, 0.11031492E+02, 0.51603775E+01, 
    },
    {
      -0.17869551E+00, 0.13910967E+01, 0.19192034E+01, 0.11573006E-01, -0.15976201E+00, -0.94560003E-02, 0.15693951E-01, 0.52722126E-01, -0.85014366E-02, -0.97233402E-02, 0.14873112E-02, -0.89176523E-03, 0.52959193E-02, 
    },
    {
      0.14833479E+00, -0.18277854E+01, 0.14350961E+01, 0.30985156E-01, 0.15200675E+00, 0.21492884E-01, -0.96385013E-02, -0.21557527E-01, 0.68189330E-01, 0.68523240E-03, -0.16917478E-02, -0.20286474E-01, 0.11845784E-01, 
    },
    {
      -0.40418360E+00, 0.81630504E+00, -0.15781780E+01, 0.64889038E+00, -0.37128392E+00, 0.77008136E-01, 0.69604263E-01, -0.34736013E+00, 0.32418558E+00, 0.33138849E-01, 0.30220786E+00, -0.44437163E-01, 0.15520661E+00, 
    },
    {
      -0.41341978E+00, 0.15498554E+01, -0.13387534E+00, 0.92664844E+00, -0.41743796E-01, 0.40571007E+00, -0.87894253E-01, -0.86675435E-01, 0.41113243E+00, 0.72546639E-01, 0.17126939E+00, 0.89121640E-01, 0.15699174E+00, 
    },
    {
      0.41381030E+01, 0.69626184E+01, 0.53660226E+01, 0.43502375E+00, 0.14352136E+01, 0.10049860E+01, -0.11870514E+01, -0.10593939E+01, 0.46142328E+00, 0.64426951E-01, 0.14919201E+00, 0.15421334E+00, -0.24009842E+00, 
    },
    {
      -0.21884575E+01, -0.73620200E+01, 0.25206361E+01, 0.66348322E-01, -0.16518012E+01, -0.94297117E+00, 0.68582207E+00, 0.67841366E-01, -0.10439430E+01, 0.77655651E-01, 0.36230335E+00, 0.63397574E+00, -0.54638183E+00, 
    },
    {
      0.53683534E+01, -0.93815575E+01, 0.17646755E+02, -0.46907468E+01, 0.32573049E+01, 0.13132706E+00, -0.36337693E+01, 0.69651113E+01, -0.32140777E+01, 0.10166423E+01, -0.68326788E+01, -0.89097261E+00, -0.18256758E+01, 
    },
    {
      0.64158630E+01, -0.17649351E+02, 0.56288085E+01, -0.14704763E+02, -0.18327432E+01, -0.44618912E+01, 0.16981554E+01, 0.28276656E+01, -0.73030243E+01, -0.36753397E+01, -0.28609183E+01, -0.84641004E+00, -0.27507968E+01, 
    },
    {
      -0.22282154E+02, -0.46678989E+02, -0.15916713E+02, 0.96986282E+00, -0.12215256E+02, -0.11022174E+02, 0.93760414E+01, 0.61606741E+01, -0.46595785E+00, -0.15873922E+01, -0.12295980E+01, -0.86355746E+00, 0.31697853E+01, 
    },
    {
      0.11682695E+02, 0.20335405E+02, -0.30051084E+01, -0.24479063E+01, 0.95660801E+01, 0.90430307E+01, -0.53547616E+01, -0.27718673E+01, 0.75180225E+01, 0.14638613E+00, -0.37047281E+01, -0.48737941E+01, 0.41346002E+01, 
    },
    {
      -0.28579910E+02, 0.87348892E+02, -0.86842323E+02, 0.14033395E+02, -0.40439680E-01, 0.79252595E+00, 0.26493780E+02, -0.41706631E+02, 0.64251800E+01, -0.86240749E+01, 0.41827084E+02, 0.86550608E+01, 0.81083364E+01, 
    },
    {
      -0.62369034E+02, 0.10966636E+03, 0.57230368E+01, 0.72633636E+02, 0.11584758E+02, 0.18804466E+02, -0.13280591E+02, -0.17550068E+02, 0.34799728E+02, 0.25178934E+02, 0.15181874E+02, 0.13891364E+01, 0.13887790E+02, 
    },
    {
      0.38604603E+02, 0.10207992E+03, -0.29230875E+02, -0.10485903E+02, 0.40149216E+02, 0.39026348E+02, -0.27193357E+02, -0.14548298E+02, -0.53476868E+01, 0.69869070E+01, 0.42715588E+01, 0.28882394E+01, -0.10630176E+02, 
    },
    {
      -0.17743690E+02, 0.38365269E+02, -0.44383980E+02, 0.14417901E+02, -0.26099285E+02, -0.29533575E+02, 0.11660138E+02, 0.13445920E+02, -0.22865898E+02, -0.37443502E+01, 0.11432396E+02, 0.13121328E+02, -0.13328956E+02, 
    },
    {
      0.83295593E+02, -0.27551205E+03, 0.20297762E+03, -0.27005564E+02, -0.36729797E+02, -0.95780029E+01, -0.72624352E+02, 0.10681412E+03, 0.42937274E+01, 0.23383394E+02, -0.10540522E+03, -0.25344908E+02, -0.16387964E+02, 
    },
    {
      0.19820599E+03, -0.31570514E+03, -0.72060852E+02, -0.15624391E+03, -0.26451447E+02, -0.37654526E+02, 0.41089275E+02, 0.39953007E+02, -0.67739883E+02, -0.63928482E+02, -0.35235271E+02, 0.25527909E+01, -0.27388004E+02, 
    },
    {
      -0.16694485E+02, -0.10372476E+03, 0.99343521E+02, 0.17906614E+02, -0.54899918E+02, -0.54147346E+02, 0.33478874E+02, 0.15286818E+02, 0.11120189E+02, -0.10826112E+02, -0.68150578E+01, -0.49306626E+01, 0.13716017E+02, 
    },
    {
      0.77929759E+01, -0.13942567E+03, 0.99441032E+02, -0.28234491E+02, 0.29203568E+02, 0.39281445E+02, -0.71675448E+01, -0.22050737E+02, 0.30663433E+02, 0.86184502E+01, -0.13269634E+02, -0.14924605E+02, 0.19169945E+02, 
    },
    {
      -0.11517070E+03, 0.34904071E+03, -0.23037421E+03, 0.29416445E+02, 0.71594467E+02, 0.19957907E+02, 0.86077232E+02, -0.12215864E+03, -0.19588717E+02, -0.26783237E+02, 0.11752483E+03, 0.30323790E+02, 0.14876017E+02, 
    },
    {
      -0.23966219E+03, 0.41143326E+03, 0.11223404E+03, 0.15691782E+03, 0.27349728E+02, 0.35871052E+02, -0.54274956E+02, -0.36422298E+02, 0.56854435E+02, 0.68983749E+02, 0.36825100E+02, -0.73194642E+01, 0.21632607E+02, 
    },
    {
      -0.64012694E+01, 0.41037807E+02, -0.62049179E+02, -0.87074289E+01, 0.26557791E+02, 0.25656239E+02, -0.14669993E+02, -0.59758453E+01, -0.57773538E+01, 0.55395350E+01, 0.37282646E+01, 0.29195380E+01, -0.61014247E+01, 
    },
    {
      0.11938375E+01, 0.93779282E+02, -0.58239254E+02, 0.16913059E+02, -0.99648829E+01, -0.18439867E+02, 0.14224555E-02, 0.11951701E+02, -0.14707023E+02, -0.52538481E+01, 0.50785527E+01, 0.62320628E+01, -0.98415794E+01, 
    },
    {
      0.59114883E+02, -0.15509395E+03, 0.99913818E+02, -0.13110655E+02, -0.38926403E+02, -0.11956065E+02, -0.37092178E+02, 0.51071323E+02, 0.12083856E+02, 0.11144695E+02, -0.47916153E+02, -0.12860417E+02, -0.49043174E+01, 
    },
    {
      0.97261864E+02, -0.19500647E+03, -0.51127705E+02, -0.60541309E+02, -0.11995492E+02, -0.12932009E+02, 0.25311089E+02, 0.10825487E+02, -0.16840401E+02, -0.26814421E+02, -0.14202532E+02, 0.40943551E+01, -0.52332826E+01, 
    },
    {
      0.59779093E-01, 0.60269125E-01, 0.43412182E-02, -0.86104780E+00, -0.28442442E+00, 0.38802601E-01, -0.24069672E-01, 0.61457353E-02, 0.26447732E-01, -0.10484729E-01, -0.16964691E-01, -0.17687945E-01, 0.11072901E-01, 
    },
    {
      0.88018358E-01, -0.24598600E-01, -0.90877013E-03, 0.27645341E+00, -0.60979271E+00, -0.11868040E-01, -0.52261841E-02, -0.38976759E-01, 0.42610688E-03, 0.50539705E-02, -0.27756458E-02, -0.37858905E-02, 0.15917744E-01, 
    },
    {
      0.24479559E-01, 0.64580876E+00, 0.88117726E-01, -0.19167215E+00, -0.63693744E+00, 0.20008095E+00, 0.13274206E+00, 0.14893281E+00, 0.11748412E+00, 0.23986910E+00, -0.11419377E+00, 0.55741396E-01, 0.13194546E+00, 
    },
    {
      0.31363299E+00, 0.28544968E+00, 0.47836196E+00, 0.13769221E+01, -0.44855464E+00, 0.14112352E+00, 0.25663987E+00, 0.10438135E+00, -0.10474421E+00, 0.16550185E-01, -0.23519803E-01, 0.12023649E+00, 0.87688930E-01, 
    },
    {
      0.15380675E+01, 0.14324548E+01, 0.13127990E+01, 0.88563509E+01, 0.12689075E+01, -0.85907197E+00, -0.37547376E-01, -0.15601891E+00, 0.33903081E-01, 0.14706784E+00, 0.37223580E+00, 0.14558417E+00, -0.20162193E+00, 
    },
    {
      0.20672157E+01, 0.85868543E+00, 0.20100529E+01, -0.15593593E+01, 0.54521675E+01, 0.74721974E+00, 0.24748008E+00, 0.22941186E+00, 0.41224465E+00, 0.10947402E+00, 0.19377372E+00, 0.53652622E-01, -0.26892835E+00, 
    },
    {
      -0.85527790E+00, -0.59873829E+01, 0.82442212E+00, 0.72160811E+01, -0.55486888E+00, -0.11447293E+01, -0.21035647E+01, -0.41455364E+00, -0.13146853E+01, -0.19809873E+01, 0.17242732E+01, -0.57977825E+00, -0.89161646E+00, 
    },
    {
      -0.37543094E+01, -0.52248607E+01, -0.48182120E+01, -0.46627312E+01, 0.96200695E+01, 0.14079694E+01, -0.24980745E+01, -0.34461054E+00, 0.31529835E+00, -0.14311078E+01, 0.84543675E+00, -0.79500079E+00, 0.87758198E-01, 
    },
    {
      -0.14999284E+02, -0.91342783E+01, -0.57922735E+01, -0.19058977E+02, 0.25543198E+00, 0.45159869E+01, 0.29069436E+01, 0.59061384E+00, -0.35207078E+00, -0.85799408E+00, -0.22677574E+01, -0.35971817E+00, 0.99695718E+00, 
    },
    {
      -0.13004522E+02, -0.14058595E+01, -0.92240829E+01, 0.49196196E+01, -0.47094283E+01, -0.47235041E+01, -0.11311457E+01, -0.21206169E+01, -0.10868499E+01, -0.12165101E+01, -0.93641424E+00, -0.54299787E-01, 0.10610462E+01, 
    },
    {
      0.39302894E+00, 0.14020730E+02, -0.63046670E+01, -0.23180601E+02, 0.46417074E+01, 0.19860531E+01, 0.61314015E+01, 0.10478191E+01, 0.39981802E+01, 0.44316478E+01, -0.51703086E+01, 0.15147123E+01, 0.18553241E+01, 
    },
    {
      0.10761917E+02, 0.15200618E+02, 0.10380513E+02, 0.55815134E+01, -0.28787245E+02, -0.61082196E+01, 0.76987019E+01, 0.63724285E+00, -0.10391539E+00, 0.52535143E+01, -0.30186067E+01, 0.21368234E+01, -0.69694448E+00, 
    },
    {
      0.33612606E+02, 0.17597853E+02, 0.12092536E+02, 0.56274945E+00, -0.44205179E+01, -0.89462976E+01, -0.83171101E+01, -0.39124992E+00, 0.78949368E+00, 0.13598933E+01, 0.44765353E+01, 0.34951830E+00, -0.16460216E+01, 
    },
    {
      0.24521830E+02, -0.49467854E+01, 0.14729335E+02, -0.94777203E+01, -0.24623814E+02, 0.99676733E+01, 0.11067480E+01, 0.53398237E+01, 0.20645846E+00, 0.34774735E+01, 0.12603258E+01, -0.11288598E+00, -0.20366173E+01, 
    },
    {
      0.48746821E+00, -0.93833942E+01, 0.79512820E+01, 0.16153667E+02, -0.65026054E+01, -0.10142500E+01, -0.47579007E+01, -0.93598449E+00, -0.33259747E+01, -0.30019803E+01, 0.39455864E+01, -0.10857137E+01, -0.10412169E+01, 
    },
    {
      -0.85157776E+01, -0.12449425E+02, -0.60487118E+01, 0.28868362E+00, 0.19360544E+02, 0.52751713E+01, -0.72028527E+01, -0.67789567E+00, -0.26337564E+00, -0.44146733E+01, 0.24460630E+01, -0.18024994E+01, 0.45497325E+00, 
    },
    {
      -0.22045063E+02, -0.11040361E+02, -0.89635525E+01, 0.14214537E+02, 0.54145708E+01, 0.58426008E+01, 0.63665800E+01, -0.24814361E+00, -0.69569623E+00, -0.63997215E+00, -0.28150806E+01, 0.38981918E-01, 0.74708533E+00, 
    },
    {
      -0.14405212E+02, 0.67465463E+01, -0.91972771E+01, 0.44302187E+01, 0.29674311E+02, -0.69829130E+01, 0.78397954E+00, -0.36686268E+01, 0.63624251E+00, -0.25999174E+01, -0.36813948E+00, 0.81634127E-01, 0.15097922E+01, 
    },
    {
      0.19783931E-01, 0.11059830E+00, -0.11023654E+00, -0.75270355E-01, 0.54907449E-01, -0.38886008E+00, -0.41173506E+00, -0.22116266E-01, 0.64385377E-01, 0.88824835E-02, -0.96755140E-02, -0.36646105E-01, -0.59598195E-02, 
    },
    {
      0.64539403E-01, -0.11785930E+00, -0.12389943E+00, -0.48941523E-01, -0.10718920E+00, 0.40435502E+00, -0.40478399E+00, -0.61611693E-01, 0.38958149E-03, 0.99619920E-03, -0.42985836E-02, 0.18320024E-02, 0.19367568E-01, 
    },
    {
      -0.32544461E+00, 0.51384830E+00, -0.25873068E+00, -0.29849005E+00, -0.10000073E+00, -0.15430472E-01, 0.58037211E-03, 0.10718274E+00, -0.16415427E-01, -0.15686333E-01, 0.29702716E-01, -0.29885124E-01, -0.15864320E-01, 
    },
    {
      -0.14868063E+00, -0.11868401E+00, -0.14044303E+00, -0.17538846E+00, -0.13476025E+00, 0.57173114E-01, 0.18623371E+00, -0.32387108E-01, -0.34499161E-01, -0.32388661E-01, -0.80245815E-01, 0.56267913E-01, 0.15138485E-01, 
    },
    {
      -0.47861069E-01, -0.77162355E+00, 0.16360250E+01, 0.50554228E+00, -0.63924365E-01, -0.10863320E+00, -0.60092205E+00, 0.49819637E-01, -0.82271285E-02, 0.98940387E-01, -0.35185188E-01, 0.12983482E+00, -0.19542508E+00, 
    },
    {
      -0.76338345E+00, 0.19630162E+01, -0.10691432E+00, -0.60503505E-01, 0.13289427E+01, 0.43475190E+00, 0.11329929E+00, -0.10821795E-01, -0.62644839E-01, 0.12987889E+00, 0.13167375E+00, 0.77864654E-01, -0.24797525E+00, 
    },
    {
      0.24173480E+00, -0.10378166E+01, 0.63027930E+00, 0.74880987E+00, 0.33605361E+00, 0.35392138E+00, 0.33974937E+00, -0.25777468E+00, 0.61715249E-03, 0.44684775E-01, -0.11695752E+00, 0.11965482E+00, 0.14105978E-01, 
    },
    {
      -0.24457288E+00, 0.41991296E+00, 0.31134152E+00, 0.22164924E+00, 0.72564882E+00, -0.43179068E+00, -0.61885383E-01, 0.49443066E-01, 0.54404356E-01, 0.31738911E-01, 0.19574244E+00, 0.27847867E-02, -0.16355213E-01, 
    },
    {
      0.63239706E+00, 0.16469691E+01, -0.27302732E+01, -0.85581708E+00, 0.37954584E+00, 0.11399052E+01, 0.17348310E+01, 0.94211541E-01, -0.21935873E+00, -0.23572168E+00, 0.16577183E+00, -0.19861729E+00, 0.35016689E+00, 
    },
    {
      0.68477434E+00, -0.23878093E+01, 0.44941813E+00, -0.84276356E-01, -0.20325942E+01, -0.13022912E+01, 0.10930405E+01, 0.38093248E+00, 0.19024202E+00, -0.33430806E+00, -0.23511814E+00, -0.28587615E+00, 0.32850736E+00, 
    },
    {
      0.67689978E-01, -0.20780521E-01, -0.29032283E-01, -0.13323180E-01, 0.28254269E-01, 0.84414100E-02, -0.15326820E-01, 0.25174028E+00, -0.40093143E-01, 0.27145154E-02, -0.66505717E-02, -0.13159065E-01, -0.18460391E-01, 
    },
    {
      0.69634892E-01, -0.45518778E-01, 0.13114978E-03, -0.10398069E-01, -0.32790072E-01, -0.10675941E-01, -0.56565069E-02, 0.37150957E-01, 0.23021942E+00, 0.76370039E-02, -0.43002558E-02, 0.90373280E-02, -0.44722832E-02, 
    },
    {
      -0.81032980E-02, -0.43692399E-01, -0.10039860E-01, 0.17762598E-01, -0.13719068E+00, -0.41710932E-01, 0.31929690E-01, 0.24983566E+00, 0.21474642E+00, -0.19966746E-02, 0.15482435E-02, -0.80155046E-03, 0.30076524E-01, 
    },
    {
      0.23039341E+00, 0.19240947E+00, -0.78822196E-01, -0.48037753E-01, 0.23808508E+00, -0.85829999E-02, -0.15761973E-01, -0.27143875E+00, 0.14898765E+00, 0.10034035E-01, -0.17262492E-01, 0.28459651E-02, -0.29851828E-01, 
    },
    {
      -0.92098333E-01, -0.30134434E-01, -0.82727504E-03, -0.17205035E-01, -0.25473502E-01, 0.82982341E-02, -0.11164435E-01, 0.28170865E-01, -0.21050289E-01, 0.71516864E-01, 0.11752138E+00, -0.58911447E-02, -0.18760277E-01, 
    },
    {
      0.83058653E-02, 0.17495196E-01, -0.28589493E-01, 0.34168422E-01, -0.72476611E-01, -0.29382722E-02, 0.76811700E-02, 0.45096446E-01, 0.30133612E-02, -0.14231350E+00, 0.89629523E-01, 0.86848531E-02, -0.99676484E-02, 
    },
    {
      -0.47373831E-01, -0.30634219E-01, 0.16800816E-02, 0.18063534E-01, -0.41148610E-01, 0.12769086E-01, -0.67251893E-02, 0.29845899E-01, 0.57053417E-02, -0.11818508E-02, -0.55645364E-02, -0.10559253E+00, 0.37416615E-01, 
    },
    {
      0.55516444E-01, 0.15834859E-01, 0.20975683E-01, 0.30448465E-01, 0.37793189E-01, 0.77342098E-02, -0.49891463E-02, -0.35565118E-02, -0.14904852E-01, -0.27537171E-01, -0.54248427E-02, -0.32515921E-01, -0.11524391E+00, 
    },
    {
      -0.27006501E-01, -0.25322475E-01, 0.14191590E-01, 0.12788468E-01, 0.57791113E-02, -0.54317171E-03, -0.15069270E-02, 0.22591196E-01, -0.38332655E-02, -0.25445297E-01, 0.13055151E-01, 0.71395235E-02, 0.97428244E-02, 
    },
    {
      0.17405979E-01, -0.11919928E-01, -0.45696460E-02, -0.54349615E-02, -0.26084450E-02, -0.11787944E-01, -0.67170402E-02, -0.12048977E-01, 0.12912758E-01, -0.51113814E-02, 0.12004339E-02, 0.12741836E-01, 0.13009419E-01, 
    },
    {
      0.38855115E-02, 0.78748085E-01, -0.10889452E+00, -0.12783662E-01, -0.66305078E-02, -0.20848872E-01, 0.30470356E-01, -0.30254409E-01, -0.69153430E-02, -0.26659120E-01, 0.13413356E-01, 0.29060601E-02, -0.56839683E-02, 
    },
    {
      -0.37495904E-01, 0.66720471E-02, -0.68216706E-02, 0.11667379E-01, 0.37586469E-01, 0.30343615E-01, -0.39814594E-02, -0.21826498E-01, 0.24977464E-01, 0.14075888E-01, -0.11293185E-01, -0.50641843E-02, 0.14215916E-01, 
    },
  },
  {
    {
      0.80071468E+01, 0.23540020E-01, 0.14707807E+00, 0.61012927E-01, -0.19648800E-01, -0.79102488E-02, -0.12570567E-01, 0.60084999E-01, -0.31929113E-01, 0.15448347E-01, -0.50893174E-02, 0.16869301E-01, -0.57898671E-02, 
    },
    {
      -0.21232369E+01, -0.46706444E+00, -0.10000438E+00, -0.10564299E+00, 0.28760383E+00, 0.85470997E-01, 0.23103155E+00, 0.48029888E-02, -0.14235696E+00, 0.14473069E+00, 0.36030032E-01, -0.35992347E-01, -0.17253455E-01, 
    },
    {
      0.16923178E+02, 0.12374104E+01, -0.25139482E+01, 0.29843819E+00, 0.55985516E+00, 0.69062692E+00, 0.48571926E+00, -0.87643987E+00, 0.75207931E+00, 0.19640107E-01, -0.28679374E+00, -0.58915937E+00, -0.28327841E+00, 
    },
    {
      0.14983639E+02, 0.24837055E+01, 0.10404690E+02, 0.21364896E+01, -0.29136975E+01, -0.39514872E+00, -0.35909803E+01, 0.20056395E+00, 0.15654199E+01, -0.34472339E+01, -0.59020323E+00, 0.87846285E+00, 0.14721909E+01, 
    },
    {
      -0.94911194E+02, -0.13008690E+02, 0.14315582E+02, -0.42997451E+01, -0.10056245E+02, -0.40604191E+01, -0.27439721E+01, 0.63601122E+01, -0.44736714E+01, -0.78184110E+00, 0.28047204E+01, 0.52159157E+01, 0.21123674E+01, 
    },
    {
      0.30943274E+01, -0.11885342E+02, -0.64855026E+02, -0.10654855E+02, 0.19201283E+02, -0.13944058E+01, 0.16869150E+02, -0.27809067E+01, -0.75264554E+01, 0.20107191E+02, 0.25993383E+01, -0.68293734E+01, -0.10687795E+02, 
    },
    {
      0.16992487E+03, 0.46145271E+02, -0.36491428E+02, 0.13116936E+02, 0.34286945E+02, 0.78730626E+01, 0.54953218E+01, -0.18412989E+02, 0.11174599E+02, 0.26052146E+01, -0.87555580E+01, -0.14695855E+02, -0.49863572E+01, 
    },
    {
      -0.86712395E+02, 0.28112967E+02, 0.14459599E+03, 0.19851633E+02, -0.51848770E+02, 0.83207064E+01, -0.31751226E+02, 0.81608305E+01, 0.16190430E+02, -0.46680210E+02, -0.44772248E+01, 0.18257042E+02, 0.26758467E+02, 
    },
    {
      -0.14461774E+03, -0.62235676E+02, 0.42808392E+02, -0.15741500E+02, -0.42179947E+02, -0.63173599E+01, -0.47827044E+01, 0.21872601E+02, -0.12338368E+02, -0.28421400E+01, 0.10545931E+02, 0.16512472E+02, 0.45623174E+01, 
    },
    {
      0.12349963E+03, -0.25484196E+02, -0.13910371E+03, -0.15541058E+02, 0.57951984E+02, -0.11481497E+02, 0.25678280E+02, -0.90002031E+01, -0.15541973E+02, 0.46808899E+02, 0.34660118E+01, -0.19979811E+02, -0.27687761E+02, 
    },
    {
      0.49242916E+02, 0.27691017E+02, -0.18267742E+02, 0.66256733E+01, 0.17394239E+02, 0.18268524E+01, 0.15869142E+01, -0.90256739E+01, 0.49349022E+01, 0.95967901E+00, -0.43152213E+01, -0.64874339E+01, -0.13755007E+01, 
    },
    {
      -0.52184998E+02, 0.73005271E+01, 0.49102627E+02, 0.43267603E+01, -0.22683687E+02, 0.48243718E+01, -0.74036064E+01, 0.34105940E+01, 0.54719052E+01, -0.16946058E+02, -0.10452632E+01, 0.77386184E+01, 0.10162491E+02, 
    },
    {
      -0.28234264E+00, 0.15047417E+01, 0.16470916E+01, -0.48499373E-02, -0.99328689E-01, -0.19733939E-01, -0.57898001E-02, 0.54610927E-01, 0.14524219E-02, 0.27774854E-02, 0.19980098E-03, 0.76767760E-02, -0.83661648E-02, 
    },
    {
      0.36049801E+00, -0.15487713E+01, 0.14352394E+01, -0.28597059E-01, 0.10506235E+00, -0.50372839E-01, -0.33528510E-01, -0.47886573E-01, 0.57634916E-01, -0.36588252E-01, 0.12803383E-01, -0.29328017E-01, 0.12255316E-01, 
    },
    {
      -0.45544675E+00, 0.12475321E+01, -0.27271132E+01, 0.53092772E+00, -0.37300754E-01, -0.15156125E+00, -0.12215277E+00, -0.25236827E+00, 0.13521501E+00, 0.36329892E+00, 0.22915353E+00, -0.17122564E+00, 0.20348294E+00, 
    },
    {
      0.99758154E+00, 0.18983068E+01, -0.12141097E+00, 0.98354695E-02, 0.15336272E+00, 0.17870606E+00, -0.72830565E-01, 0.17440289E+00, 0.29191229E+00, -0.11306345E-01, 0.49621547E-02, -0.26877809E-01, -0.59098396E-02, 
    },
    {
      0.66613059E+01, 0.10923632E+02, -0.65970397E+00, 0.15522877E+01, -0.40954612E-01, 0.10254952E+01, -0.66243565E+00, -0.82247669E+00, 0.49483865E+00, -0.44243667E+00, -0.27341887E+00, -0.15122773E+00, -0.44921702E-02, 
    },
    {
      -0.38886521E+01, 0.54575652E+00, 0.10046308E+02, 0.23151209E+01, -0.10080287E+01, 0.23143321E+00, 0.23112388E+01, 0.82673109E+00, -0.13583063E+01, 0.51237226E+00, -0.59203404E+00, 0.10641146E+01, -0.16020350E-01, 
    },
    {
      0.38286211E+01, -0.29793097E+02, 0.47293575E+02, -0.99862328E+01, -0.12771469E+01, 0.57090254E+01, 0.90686393E+00, 0.46967688E+01, -0.48249456E+00, -0.78903108E+01, -0.37669451E+01, 0.33611767E+01, -0.28549347E+01, 
    },
    {
      -0.42156825E+01, -0.18497114E+02, 0.43299971E+01, 0.12616365E+01, -0.74922557E+01, 0.18494391E+01, 0.33060460E+01, -0.33028397E+00, -0.44510746E+01, -0.27160680E+00, 0.20754679E+00, 0.16191085E+00, 0.37127814E+00, 
    },
    {
      -0.39047291E+02, -0.20917377E+02, 0.41411633E+02, -0.88546057E+01, 0.36347630E+01, -0.46501169E+01, 0.43165107E+01, 0.63377647E+01, -0.79014248E+00, 0.20996790E+01, 0.26977179E+01, 0.49547371E-01, -0.80331719E+00, 
    },
    {
      0.24417505E+02, -0.50597050E+02, -0.12887446E+02, -0.25855635E+02, 0.15060377E+02, 0.16648369E+01, -0.19734503E+02, -0.85645714E+01, 0.11742890E+02, 0.24708913E+00, 0.30122676E+01, -0.88011379E+01, -0.45888126E-01, 
    },
    {
      -0.12037977E+02, 0.18195720E+03, -0.30125452E+03, 0.56718910E+02, 0.25235325E+02, -0.37847782E+02, -0.31361133E+00, -0.27802044E+02, -0.43944149E+01, 0.44352974E+02, 0.19799467E+02, -0.20778854E+02, 0.12326245E+02, 
    },
    {
      -0.41416504E+02, 0.12508257E+03, -0.62992988E+01, -0.16054337E+02, 0.42519859E+02, -0.22424593E+02, -0.25135471E+02, 0.28161285E+01, 0.19682430E+02, 0.15035913E+01, -0.35815947E+01, 0.37708406E+01, -0.11817546E+01, 
    },
    {
      0.89285217E+02, -0.69640999E+02, -0.17817468E+03, 0.21492142E+02, -0.17860403E+02, 0.67075515E+01, -0.15617906E+02, -0.20148720E+02, -0.40488234E+01, -0.17107679E+01, -0.83144178E+01, 0.30443573E+01, 0.37432699E+01, 
    },
    {
      -0.67462074E+02, 0.21829535E+03, -0.10266725E+03, 0.90444672E+02, -0.59647964E+02, -0.10293538E+02, 0.61297581E+02, 0.29657206E+02, -0.38800804E+02, -0.93168879E+01, -0.34729412E+01, 0.26925245E+02, -0.22017689E+01, 
    },
    {
      0.41186665E+02, -0.43663583E+03, 0.76321771E+03, -0.13260320E+03, -0.96994392E+02, 0.98809723E+02, -0.51419187E+01, 0.72419868E+02, 0.21797657E+02, -0.10391363E+03, -0.44690708E+02, 0.50902679E+02, -0.22424061E+02, 
    },
    {
      0.19840215E+03, -0.35004047E+03, -0.27337153E+01, 0.44635445E+02, -0.85559502E+02, 0.67447411E+02, 0.67286049E+02, -0.17907366E+02, -0.35397381E+02, 0.19058682E+01, 0.13237237E+02, -0.18305573E+02, 0.25211602E+00, 
    },
    {
      -0.86747185E+02, 0.17347545E+03, 0.27344232E+03, -0.22800623E+02, 0.27914158E+02, -0.33933644E+01, 0.26151587E+02, 0.27388149E+02, 0.97586889E+01, -0.34674821E+01, 0.91410255E+01, -0.73688040E+01, -0.56035948E+01, 
    },
    {
      0.89096458E+02, -0.34042059E+03, 0.24030518E+03, -0.12773231E+03, 0.86151863E+02, 0.17995501E+02, -0.78346107E+02, -0.41468628E+02, 0.53289478E+02, 0.18957485E+02, -0.17327203E+01, -0.35620258E+02, 0.71820140E+01, 
    },
    {
      -0.68477409E+02, 0.45522491E+03, -0.86998053E+03, 0.13320493E+03, 0.13828078E+03, -0.11296691E+03, 0.83289499E+01, -0.84006424E+02, -0.30078983E+02, 0.10894444E+03, 0.46073669E+02, -0.53107388E+02, 0.18520054E+02, 
    },
    {
      -0.27914142E+03, 0.44985599E+03, -0.11727840E+01, -0.40542358E+02, 0.65773628E+02, -0.78929565E+02, -0.74833633E+02, 0.33005203E+02, 0.26883795E+02, -0.96850271E+01, -0.17542385E+02, 0.28257006E+02, 0.11094317E+01, 
    },
    {
      0.28169247E+02, -0.94577866E+02, -0.13985548E+03, 0.92719965E+01, -0.13244056E+02, 0.46153149E+00, -0.14713109E+02, -0.13108560E+02, -0.55679369E+01, 0.39470751E+01, -0.31173911E+01, 0.46963744E+01, 0.28085086E+01, 
    },
    {
      -0.41424831E+02, 0.17922871E+03, -0.13824788E+03, 0.62719311E+02, -0.40361134E+02, -0.99571943E+01, 0.35196617E+02, 0.20173746E+02, -0.25507154E+02, -0.10779361E+02, 0.30331249E+01, 0.17037985E+02, -0.52984161E+01, 
    },
    {
      0.38881653E+02, -0.17535535E+03, 0.36977036E+03, -0.48800972E+02, -0.67534874E+02, 0.46994465E+02, -0.36084449E+01, 0.35268539E+02, 0.12913198E+02, -0.42146160E+02, -0.17815418E+02, 0.19775288E+02, -0.58432074E+01, 
    },
    {
      0.12580310E+03, -0.21518388E+03, 0.61443510E+01, 0.98084249E+01, -0.15733668E+02, 0.32032814E+02, 0.29372606E+02, -0.18224518E+02, -0.68510780E+01, 0.68954396E+01, 0.76545310E+01, -0.14303968E+02, -0.45466852E+00, 
    },
    {
      0.14912950E+00, -0.37197623E-01, 0.86089209E-01, -0.99689066E+00, 0.34358004E+00, 0.41757068E-02, 0.11042265E-01, -0.31665795E-01, 0.28434340E-01, -0.40736802E-01, 0.25358051E-01, -0.47866986E-02, -0.79530235E-02, 
    },
    {
      0.48351955E-01, -0.13903272E+00, 0.35033248E-01, -0.24141113E+00, -0.73474455E+00, -0.21165458E-02, 0.14233136E-01, -0.48750002E-01, -0.17404562E-01, -0.13931587E-01, 0.45621013E-02, -0.16081575E-01, 0.18449978E-02, 
    },
    {
      0.15439525E+00, 0.10494125E+01, 0.53804463E+00, 0.16673172E+01, -0.14299715E+01, 0.68726435E-01, -0.16137257E+00, 0.13222758E+00, 0.19399203E-01, -0.11860881E-01, -0.47297385E-01, 0.52892003E-01, 0.16677070E-01, 
    },
    {
      0.36067051E+00, 0.11057934E+01, 0.61169720E+00, 0.19141263E+01, 0.20964134E+01, 0.23646539E+00, 0.27404869E+00, 0.19712859E+00, -0.11474232E+00, 0.12215359E+00, -0.76190814E-01, 0.15382162E+00, 0.43123838E-01, 
    },
    {
      0.10588363E+01, 0.11870986E+01, 0.88391542E+00, 0.23195529E+01, 0.29367788E+01, 0.42034310E+00, -0.13270929E+01, -0.13814926E+00, 0.26789281E+00, -0.76729983E-01, -0.62709659E+00, -0.13922800E+00, -0.22054443E+00, 
    },
    {
      0.26687746E+01, 0.14925655E+01, -0.10293322E+01, -0.60973663E+01, 0.13214234E-01, -0.80502622E-01, -0.20025112E+00, 0.51012868E+00, 0.36622775E+00, 0.50678074E+00, -0.61252302E+00, 0.36497283E+00, 0.66087967E+00, 
    },
    {
      -0.18513010E+01, -0.84644527E+01, -0.22534673E+01, 0.16729214E+01, 0.59963422E+01, 0.83362734E+00, -0.32418764E+00, -0.12634010E+01, -0.21761373E+00, -0.32083201E+00, 0.10582196E+01, -0.40086418E-01, 0.34409949E+00, 
    },
    {
      -0.56983318E+01, -0.85301552E+01, -0.62716293E+01, -0.10336222E+02, -0.12337216E+01, 0.14525561E+01, -0.27670050E+01, -0.61272615E+00, 0.89260936E-02, -0.17194246E+01, 0.48456728E+00, -0.75325119E+00, -0.12380633E+00, 
    },
    {
      -0.10879601E+02, 0.18890285E+00, -0.12624624E+02, 0.85096445E+01, -0.14647897E+02, -0.11804613E+01, 0.11204756E+02, 0.11301163E+01, -0.65179712E+00, 0.20850821E+01, 0.33798697E+01, 0.68264633E+00, 0.13606979E+01, 
    },
    {
      -0.41109333E+01, -0.76429524E+01, 0.31357608E+01, 0.33289852E+02, 0.18736217E+02, 0.52604097E+00, 0.32423337E+01, -0.39312522E+01, -0.42085305E+00, -0.20370185E+01, 0.46622753E+01, -0.20514877E+01, -0.67121148E+01, 
    },
    {
      0.33330195E+01, 0.19760462E+02, 0.13707290E+01, -0.15091409E+02, -0.44948006E+01, -0.37715893E+01, 0.19507859E+01, 0.41595783E+01, 0.84614331E+00, 0.17263756E+01, -0.36580293E+01, -0.27227202E+00, -0.15076882E+01, 
    },
    {
      0.17205370E+02, 0.15522379E+02, 0.18444885E+02, 0.14610297E+02, -0.95373640E+01, -0.77903051E+01, 0.75731368E+01, 0.10421705E+01, 0.14117401E+01, 0.51428719E+01, -0.15598035E+01, 0.11522195E+01, 0.33083048E+00, 
    },
    {
      0.23841166E+02, -0.88205357E+01, 0.33722401E+02, -0.35630077E+02, 0.22245089E+02, 0.98581381E-01, -0.23630219E+02, -0.24590368E+01, -0.60612363E+00, -0.56737971E+01, -0.57873211E+01, -0.94312793E+00, -0.20955913E+01, 
    },
    {
      -0.14294488E+02, 0.16591463E+02, -0.36211152E+01, -0.58140308E+02, -0.52787479E+02, -0.87697142E+00, -0.86255169E+01, 0.79262271E+01, -0.16103133E+01, 0.19649650E+01, -0.10198181E+02, 0.28549724E+01, 0.16818821E+02, 
    },
    {
      -0.28521106E+01, -0.13874240E+02, 0.60465282E+00, 0.88589296E+01, -0.56679926E+01, 0.32743344E+01, -0.16497432E+01, -0.34437387E+01, -0.67558551E+00, -0.17088107E+01, 0.29519806E+01, 0.42225948E+00, 0.13573281E+01, 
    },
    {
      -0.15705806E+02, -0.77159276E+01, -0.15102554E+02, -0.16817579E+01, 0.56376777E+01, 0.67993746E+01, -0.60265536E+01, -0.10358450E+01, -0.14570483E+01, -0.36746924E+01, 0.12936862E+01, -0.37998509E+00, -0.32503825E+00, 
    },
    {
      -0.14731427E+02, 0.91323881E+01, -0.23522758E+02, 0.31281187E+02, -0.66921196E+01, 0.88023257E+00, 0.14684581E+02, 0.14542351E+01, 0.11787513E+01, 0.40096455E+01, 0.31103556E+01, 0.42140290E+00, 0.75577301E+00, 
    },
    {
      0.22154350E+02, -0.13379253E+02, 0.43898410E+00, 0.28501005E+02, 0.40360855E+02, 0.30411851E+00, 0.64322257E+01, -0.44322767E+01, 0.17975950E+01, -0.39312956E+00, 0.68860784E+01, -0.11438121E+01, -0.11947880E+02, 
    },
    {
      0.27685111E-01, -0.40053260E-01, -0.17839098E-01, -0.57812698E-01, 0.70409186E-01, -0.50432491E+00, -0.61618298E+00, -0.51988091E-01, 0.36189247E-01, -0.57952993E-01, 0.35770409E-01, -0.30052310E-01, -0.15522750E-01, 
    },
    {
      -0.99234711E-02, -0.25411305E+00, -0.98816864E-01, 0.27671959E-01, -0.90068385E-01, 0.64124870E+00, -0.49534273E+00, -0.23844037E-01, 0.90174731E-02, 0.16274264E-01, 0.13823560E-02, -0.16471019E-01, 0.77255028E-02, 
    },
    {
      -0.29735389E+00, 0.89244604E+00, 0.47042780E-01, -0.19908190E+00, -0.20957834E+00, 0.48498312E+00, 0.34726390E+00, 0.14254756E+00, -0.51067263E-01, -0.74147590E-01, 0.43129999E-01, -0.94624739E-02, 0.69399804E-01, 
    },
    {
      -0.27613023E+00, 0.39687451E-01, -0.22952503E+00, -0.17188592E+00, -0.37711376E+00, -0.37344018E+00, 0.44496518E+00, 0.49825110E-01, 0.50268643E-01, 0.10268872E-01, -0.15096737E+00, -0.18276261E-02, 0.57382856E-01, 
    },
    {
      0.55527079E+00, 0.49626112E+00, 0.41093677E+00, -0.36513752E+00, -0.92810732E+00, 0.45406213E+00, 0.39471567E+00, -0.30232587E+00, 0.54300815E+00, 0.59649146E+00, -0.42978968E-01, 0.16226175E+00, -0.37429817E-01, 
    },
    {
      -0.20489936E+00, 0.37278998E+01, -0.70701146E+00, -0.19198212E+00, 0.71857435E+00, -0.99595582E+00, 0.54645717E+00, -0.43600601E+00, -0.46274224E+00, 0.24707445E-02, 0.20912142E+00, 0.10534024E+00, -0.15043882E-01, 
    },
    {
      -0.54460263E+00, -0.15500151E+01, 0.16229698E-01, 0.42883778E+00, 0.12063434E+00, -0.49502440E-01, -0.30448535E+00, -0.20664476E+00, 0.19829325E+00, 0.26275936E+00, -0.19895925E+00, 0.21524690E+00, -0.15297687E+00, 
    },
    {
      0.29179329E+00, 0.17002033E+00, 0.72335619E+00, 0.31211472E+00, 0.96879709E+00, 0.17340130E+00, 0.76188266E-01, -0.18247452E+00, -0.35377108E-01, -0.23166701E-01, 0.31477252E+00, 0.19326968E+00, -0.67122035E-01, 
    },
    {
      0.96205816E-01, 0.37230721E+00, -0.99004132E+00, 0.10045286E+01, 0.21211245E+01, -0.90331696E-01, 0.57627037E-01, 0.65079659E+00, -0.12693567E+01, -0.96175170E+00, 0.58023080E-01, -0.35038722E+00, 0.14534500E+00, 
    },
    {
      -0.79609752E-01, -0.49538903E+01, 0.14437783E+01, 0.20078922E-01, -0.11813406E+01, 0.14132198E+01, -0.34010509E+00, 0.10204923E+01, 0.75110745E+00, -0.16915599E+00, -0.33175179E+00, -0.33036423E+00, -0.20429282E-01, 
    },
    {
      -0.10897010E-01, -0.49416184E-01, -0.44178925E-01, -0.84936572E-02, 0.51277783E-01, 0.19495703E-01, -0.32920949E-01, 0.21518527E+00, -0.17939478E+00, -0.32887351E-01, 0.33539165E-01, -0.19388679E-01, -0.20020414E-01, 
    },
    {
      0.20877963E-01, 0.27526090E-01, -0.37257954E-01, 0.32353066E-01, -0.61392702E-01, 0.98865917E-02, -0.21773189E-01, 0.17957857E+00, 0.23242453E+00, 0.58700140E-02, 0.39826785E-02, 0.25550261E-01, 0.60049252E-03, 
    },
    {
      -0.15558755E+00, -0.54346319E-01, -0.56051023E-01, 0.19358875E-01, -0.15182693E+00, -0.10598335E+00, 0.37775725E-01, 0.16895783E+00, 0.51459491E+00, -0.84568355E-02, -0.67650639E-01, 0.27996898E-02, -0.11083936E-01, 
    },
    {
      0.30534738E+00, 0.23073228E+00, -0.16615911E+00, -0.73981099E-01, 0.17580722E+00, -0.86328842E-01, -0.72317007E-02, -0.50074303E+00, -0.23022939E-02, 0.22110448E-02, -0.61923102E-01, -0.39555937E-01, 0.32912306E-01, 
    },
    {
      -0.12910460E+00, -0.22992328E-01, -0.24645030E-01, -0.16976858E-01, -0.48633263E-01, 0.24791237E-01, -0.44906000E-02, 0.44549942E-01, -0.13423373E-01, 0.14574012E+00, 0.18071000E+00, 0.22294498E-02, -0.21888547E-01, 
    },
    {
      0.82140617E-01, 0.60347050E-01, -0.69422051E-01, 0.34339996E-02, -0.79234779E-01, -0.41237101E-01, -0.28514843E-01, 0.24108853E-01, 0.40748831E-01, -0.15513334E+00, 0.15307407E+00, 0.99850390E-02, -0.27682120E-01, 
    },
    {
      0.54472767E-01, -0.43857154E-01, -0.53891961E-01, -0.13284815E-01, 0.35775442E-01, 0.40358426E-02, -0.27592603E-01, 0.31445228E-01, 0.15491653E-01, 0.28190129E-02, -0.24603689E-01, -0.92626899E-01, 0.67674026E-01, 
    },
    {
      0.71794264E-01, 0.95135733E-01, 0.50922348E-02, -0.13623560E-01, 0.23757331E-01, -0.47744387E-02, 0.46965736E-02, -0.67392141E-02, -0.48691975E-02, -0.11122855E-02, -0.13454689E-01, -0.77131763E-01, -0.98624110E-01, 
    },
    {
      0.48160724E-01, 0.16543763E-01, -0.49138770E-01, 0.17461605E-01, 0.26750349E-01, 0.39088377E-02, 0.55514900E-02, 0.43032456E-01, 0.60738609E-02, -0.25063993E-02, -0.15855387E-01, -0.94612045E-02, -0.41879309E-03, 
    },
    {
      0.32233037E-01, -0.19281844E-01, 0.20792220E-01, 0.94251204E-02, -0.47330942E-01, 0.23148457E-01, 0.49970322E-02, -0.24953449E-01, 0.21466170E-01, -0.37387648E-03, 0.19118547E-01, 0.32155174E-02, 0.16403269E-01, 
    },
    {
      0.35219163E-01, 0.13113920E+00, -0.96734829E-01, 0.13746214E-01, -0.28147859E-01, 0.16202064E-01, 0.51043965E-01, -0.35983652E-02, -0.52156085E-02, -0.96815266E-02, -0.55470653E-02, -0.20149490E-01, -0.10250242E-01, 
    },
    {
      -0.27565159E-01, 0.56680135E-01, -0.28701194E-01, -0.58173505E-02, 0.27876016E-01, 0.27989307E-01, -0.13116185E-01, -0.12744517E-01, 0.25175970E-01, 0.43719960E-02, 0.16126093E-02, -0.25717752E-01, 0.67674026E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir17_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30802848E+01, -0.46980400E-02, 0.10328545E-01, -0.13646303E-02, -0.13921593E-02, 0.46656518E-02, 0.75738039E-02, -0.13919175E-02, 0.18014414E-02, 
    },
    {
      -0.73226583E+00, 0.12279886E-01, 0.34578342E-01, 0.33941451E-01, 0.73212688E-03, -0.21163456E-01, 0.32289596E-02, -0.18333461E-01, 0.91965441E-02, 
    },
    {
      0.79729569E+00, 0.13542826E+00, -0.11436112E+00, -0.70079923E-01, 0.79188850E-02, -0.77631488E-01, -0.55221070E-01, 0.45033410E-01, -0.30216960E-01, 
    },
    {
      0.18268728E+01, -0.31209853E+00, -0.70338599E-01, -0.13058031E+00, 0.52127116E-02, 0.59936713E-01, -0.18583011E-01, 0.56642335E-01, 0.17623593E-02, 
    },
    {
      -0.78476065E+00, -0.25712866E+00, 0.25612435E+00, 0.15128808E+00, -0.38681533E-01, 0.24744213E+00, 0.14993742E+00, -0.11375405E+00, 0.66822320E-01, 
    },
    {
      -0.13493091E+01, 0.30580220E+00, 0.73491491E-01, 0.89370720E-01, 0.54565580E-02, -0.46355233E-01, 0.13624964E-01, -0.36061008E-01, -0.18066941E-01, 
    },
    {
      -0.12063475E+00, 0.14794455E+00, -0.10244069E+00, -0.80330491E-01, 0.43026242E-01, -0.18751287E+00, -0.10860286E+00, 0.70955373E-01, -0.43552943E-01, 
    },
    {
      0.86388132E-02, -0.10469254E+00, -0.35336661E+00, 0.68503232E-02, -0.22832712E-01, 0.74300254E-02, -0.25587229E-02, 0.51693595E-02, -0.12979574E-03, 
    },
    {
      0.82351686E-02, 0.37010682E+00, -0.82846217E-01, -0.23642721E-01, 0.21374708E-01, 0.24881929E-02, -0.14439701E-01, 0.34175650E-02, 0.15466106E-01, 
    },
    {
      -0.13728087E+00, -0.90724178E-01, -0.25918829E+00, 0.55246770E-01, 0.28049927E-01, -0.48293497E-01, -0.40341385E-01, 0.25727993E-01, 0.12118476E-01, 
    },
    {
      0.87273709E-01, 0.31198278E+00, -0.16230865E+00, 0.16893700E-01, 0.17322816E-02, -0.83513986E-02, 0.22217995E-01, 0.15569704E-01, 0.63724130E-01, 
    },
    {
      0.55732697E-01, 0.28494719E+00, 0.15282595E+01, -0.20563173E+00, 0.30643355E-01, -0.10549944E+00, 0.10248962E+00, 0.31379037E-01, -0.74367374E-02, 
    },
    {
      -0.61427230E+00, -0.15684907E+01, 0.10502628E+00, 0.25299487E+00, -0.15594783E+00, -0.50039280E-01, 0.76203002E-02, -0.49784355E-01, -0.12944193E+00, 
    },
    {
      0.15699542E+01, 0.26262048E+00, 0.62663382E+00, 0.22692803E-01, -0.51594186E+00, 0.51364642E+00, 0.29827582E-01, -0.80818407E-01, -0.30650470E+00, 
    },
    {
      0.24862222E+00, -0.59349144E+00, 0.11204494E+01, -0.97968966E-01, 0.33110031E+00, 0.38167056E-01, -0.16308677E+00, 0.54526995E-02, -0.50980163E+00, 
    },
    {
      -0.47659847E+00, -0.55125346E-02, -0.95217299E+00, 0.47925776E+00, -0.42554024E-02, 0.30441788E+00, -0.30924687E+00, -0.12380421E+00, -0.70827238E-01, 
    },
    {
      0.20973926E+01, 0.99478257E+00, 0.27451342E+00, -0.55617458E+00, 0.29574648E+00, 0.15973371E+00, 0.80506027E-01, 0.57603039E-01, 0.32256550E+00, 
    },
    {
      -0.38655005E+01, -0.90660799E+00, -0.59841734E+00, -0.50917405E+00, 0.14092973E+01, -0.11617564E+01, 0.28998095E+00, 0.36635254E-01, 0.80191475E+00, 
    },
    {
      -0.10721380E+01, 0.55519497E+00, -0.26079962E+01, 0.24563982E+00, -0.10116768E+01, -0.10343829E+00, 0.17780776E+00, -0.81209339E-01, 0.10334632E+01, 
    },
    {
      0.69367838E+00, -0.10222288E+00, -0.55200565E+00, -0.34082192E+00, 0.15705701E-01, -0.24950403E+00, 0.24499443E+00, 0.78426957E-01, 0.83322436E-01, 
    },
    {
      -0.16341933E+01, 0.60357797E+00, -0.35521424E+00, 0.34881729E+00, -0.18382023E+00, -0.85821614E-01, -0.79250164E-01, 0.24138987E-02, -0.23403056E+00, 
    },
    {
      0.27014959E+01, 0.11459872E+01, -0.33060220E+00, 0.46477214E+00, -0.97078705E+00, 0.67035884E+00, -0.35578537E+00, 0.13218268E-01, -0.56398720E+00, 
    },
    {
      0.62231594E+00, 0.28681967E+00, 0.19950296E+01, -0.16715036E+00, 0.70967084E+00, 0.80056027E-01, -0.58295801E-01, 0.67210004E-01, -0.60891330E+00, 
    },
    {
      -0.86633563E-02, -0.13671836E-02, -0.19294512E-01, -0.29737541E-01, -0.15660122E+00, -0.86753992E-02, 0.21384902E-03, 0.20422880E-02, -0.96033914E-02, 
    },
    {
      0.10736866E-02, 0.13963470E-02, 0.80128183E-03, 0.16088326E+00, -0.86053144E-02, 0.60172728E-03, -0.22198915E-01, -0.37762453E-03, 0.10805884E-02, 
    },
    {
      0.39740406E-01, 0.17161980E+00, 0.12527364E+00, 0.13852577E+00, -0.44440904E+00, -0.17778879E-01, -0.20324275E-01, 0.42344291E-01, -0.20028139E-01, 
    },
    {
      -0.42737368E-01, 0.44157285E-01, 0.24216682E-01, 0.59995002E+00, 0.19375703E+00, 0.41468341E-01, -0.40872436E-01, -0.31848896E-01, -0.14190458E-01, 
    },
    {
      -0.27813461E-01, 0.34570392E-01, 0.91654710E-01, 0.33465780E-01, -0.93962334E-01, 0.60697462E-01, -0.45351591E-01, -0.58084037E-02, 0.61074000E-01, 
    },
    {
      -0.37716964E+00, -0.15043245E+00, -0.14817756E+00, 0.47543745E-01, -0.86728200E-01, 0.59861321E-01, 0.14234369E+00, 0.12710453E-01, 0.17827239E-01, 
    },
    {
      -0.24680117E+00, -0.86096442E+00, -0.37145874E+00, -0.49700645E+00, 0.13920882E+01, 0.93504988E-01, 0.44289514E-01, -0.18447290E+00, 0.17054563E-01, 
    },
    {
      0.70487481E+00, -0.10964639E-01, 0.19685283E+00, -0.18399014E+01, -0.80342036E+00, -0.18387800E+00, 0.39359804E-01, 0.17004316E+00, -0.85180625E-01, 
    },
    {
      -0.19522337E-01, -0.21377654E+00, -0.14989559E+00, -0.97864687E-01, 0.37941441E+00, -0.63822173E-01, 0.14001627E+00, 0.39011661E-02, -0.85912690E-01, 
    },
    {
      0.60058129E+00, 0.10312814E+00, 0.23901564E+00, -0.29843274E+00, 0.10723197E-01, -0.10886107E+00, -0.15636367E+00, -0.39202116E-01, -0.35129011E-01, 
    },
    {
      0.25636184E+00, 0.11735374E+01, 0.45907420E+00, 0.33341327E+00, -0.12187766E+01, -0.59096042E-01, -0.59066504E-01, 0.20547853E+00, 0.14204452E-01, 
    },
    {
      -0.11265887E+01, -0.48129063E-01, -0.23462927E+00, 0.14852360E+01, 0.70783460E+00, 0.19520159E+00, 0.13287680E+00, -0.18437938E+00, 0.19562475E+00, 
    },
    {
      -0.32194812E-01, 0.65475008E-02, 0.37131528E-02, -0.14261339E-01, -0.12782280E-02, 0.23382235E-01, 0.15923720E-01, 0.40608902E-04, 0.66276081E-03, 
    },
    {
      0.31999865E-03, -0.42041160E-01, -0.23349874E-01, 0.11027430E-01, -0.22927441E-01, -0.20414442E-01, 0.19795489E-01, -0.69397581E-02, 0.38036958E-02, 
    },
    {
      0.65187119E-01, 0.50789274E-01, 0.12329996E-01, -0.14014617E-01, 0.19207697E-01, 0.54233067E-01, 0.19744506E-01, 0.45834291E-02, 0.72671259E-02, 
    },
    {
      0.29568112E-01, -0.95958523E-02, -0.18973609E-01, -0.41401144E-02, 0.17457606E-01, -0.15092639E-01, 0.75680852E-01, -0.13022618E-01, 0.67012999E-02, 
    },
    {
      0.77274968E-02, -0.82368053E-01, -0.58961112E-01, 0.20770151E-01, -0.13150994E-02, -0.13049465E+00, -0.10597041E+00, 0.16197849E-01, 0.15678909E-01, 
    },
    {
      -0.72845094E-01, 0.12490302E+00, -0.20829698E-02, -0.25820648E-01, 0.39102312E-01, 0.10707958E+00, -0.13309467E+00, 0.15458102E-01, 0.37044154E-02, 
    },
    {
      -0.31673741E-01, -0.62940731E-02, 0.84093548E-02, -0.41974136E-02, -0.33746944E-02, 0.80712140E-02, -0.77740918E-02, 0.31657413E-01, 0.36994985E-02, 
    },
    {
      -0.28113980E-01, 0.10798969E-01, -0.56396737E-02, -0.16900625E-02, -0.34986283E-02, 0.40706303E-02, 0.11442027E-01, -0.97587286E-02, 0.27690606E-01, 
    },
    {
      0.79382271E-01, -0.30479811E-01, 0.32784846E-01, -0.26296770E-01, -0.10788017E-01, 0.10917043E-01, 0.14642074E-01, 0.22470199E-01, 0.14176787E-01, 
    },
    {
      0.31412635E-01, -0.57085998E-01, 0.26893871E-01, 0.37701238E-01, -0.21372926E-02, -0.14738973E-01, 0.80593152E-03, -0.38642716E-01, 0.31024845E-01, 
    },
    {
      0.50734030E-03, -0.64163841E-02, -0.12042396E-01, 0.22190325E-02, -0.28449393E-03, 0.62294272E-02, -0.91660712E-02, 0.21545021E-02, -0.10553715E-02, 
    },
    {
      0.14355371E-01, 0.12396547E-01, -0.10789397E-01, -0.44431705E-02, -0.94738062E-02, 0.47621662E-02, 0.40972382E-02, 0.32042819E-02, -0.99020661E-03, 
    },
    {
      0.49181809E-02, 0.23562347E-02, 0.40638875E-02, 0.27082455E-02, -0.65866150E-02, 0.62013259E-02, 0.16731292E-02, 0.18732373E-02, -0.48593818E-02, 
    },
    {
      0.13873120E-01, -0.13107671E-02, -0.19071303E-01, 0.38780418E-03, 0.61205011E-02, -0.50388812E-02, 0.68964978E-03, -0.54935296E-02, 0.18295188E-02, 
    },
  },
  {
    {
      0.27496719E+01, -0.11850115E-01, 0.73418082E-02, 0.64126854E-02, -0.29800744E-02, 0.10921643E-02, 0.41902238E-02, 0.14759843E-02, -0.22041446E-02, 
    },
    {
      -0.51172680E+00, 0.35100680E-01, 0.40524267E-01, 0.30474843E-01, -0.65487823E-02, -0.66860467E-02, -0.82540857E-02, -0.16374506E-01, 0.66717844E-02, 
    },
    {
      0.10756888E+01, 0.95197663E-01, -0.99020548E-01, -0.55270936E-01, 0.56841303E-01, -0.30476358E-01, -0.44919342E-01, 0.41535314E-01, 0.14882590E-01, 
    },
    {
      0.96904379E+00, -0.45030373E+00, -0.81589915E-01, -0.10445018E+00, 0.13437195E-01, -0.15732037E-01, 0.23781391E-01, 0.42177223E-01, -0.97438293E-02, 
    },
    {
      -0.12668188E+01, -0.14799927E-01, 0.23314634E+00, 0.87992266E-01, -0.20218949E+00, 0.13027802E+00, 0.13615835E+00, -0.11343505E+00, -0.31153904E-01, 
    },
    {
      -0.63311547E+00, 0.41507858E+00, 0.64436302E-01, 0.67502894E-01, 0.85417582E-02, 0.24351122E-01, -0.25230847E-01, -0.23825040E-01, 0.86125854E-03, 
    },
    {
      0.16201258E+00, -0.35472833E-01, -0.86220764E-01, -0.37761636E-01, 0.15207741E+00, -0.11216554E+00, -0.97735316E-01, 0.73955752E-01, 0.15572720E-01, 
    },
    {
      0.19004470E-01, -0.27081847E+00, -0.25790614E+00, 0.95325941E-02, -0.15033994E-01, 0.27574443E-02, -0.25284474E-02, 0.38412474E-02, 0.23758332E-02, 
    },
    {
      -0.15574141E-01, 0.28764451E+00, -0.24723634E+00, -0.21190660E-01, 0.19187475E-01, -0.14577296E-02, -0.13602387E-01, 0.47102780E-02, 0.15887067E-01, 
    },
    {
      0.15916567E-01, -0.13094133E-01, -0.18341503E+00, -0.20860847E-01, 0.39675791E-01, -0.68267519E-02, -0.66249087E-01, -0.62741726E-02, 0.41506059E-01, 
    },
    {
      0.33226557E-01, 0.21242505E+00, -0.11735039E+00, 0.20106252E-01, 0.14337774E-01, 0.52691698E-02, 0.18621637E-01, -0.20861311E-02, 0.50332192E-01, 
    },
    {
      -0.32793215E+00, 0.81487375E+00, 0.84182167E+00, -0.12698558E+00, 0.33453379E-01, -0.25510749E-01, 0.89533143E-01, 0.24488725E-01, -0.31261619E-01, 
    },
    {
      -0.37190884E+00, -0.10941412E+01, 0.68853664E+00, 0.24537718E+00, -0.12240302E+00, -0.18654227E-01, 0.40675946E-01, -0.11932018E+00, -0.14532743E+00, 
    },
    {
      0.39973900E-01, -0.20130175E+00, 0.98582131E+00, 0.46249530E+00, -0.45445198E+00, 0.95881127E-01, 0.30371648E+00, 0.18240003E+00, -0.36862892E+00, 
    },
    {
      0.49964002E+00, -0.55297488E+00, 0.81649625E+00, -0.15544052E+00, 0.17243408E+00, -0.80211408E-01, -0.12309635E+00, 0.14403760E+00, -0.36307234E+00, 
    },
    {
      0.92932731E+00, -0.43920514E+00, 0.27505100E+00, 0.24573041E+00, 0.31845255E-02, 0.79882346E-01, -0.29909259E+00, -0.48250318E-01, 0.84246330E-01, 
    },
    {
      0.15329826E+01, 0.19085544E+00, -0.40697181E+00, -0.55546308E+00, 0.18732370E+00, 0.17146086E-01, 0.14358910E-01, 0.27570710E+00, 0.33728591E+00, 
    },
    {
      -0.16367953E+00, 0.20679381E+00, -0.20812731E+01, -0.13375096E+01, 0.10512470E+01, -0.18085140E+00, -0.35574475E+00, -0.50738007E+00, 0.78306991E+00, 
    },
    {
      -0.15351163E+01, 0.11400683E+01, -0.19037578E+01, 0.40057018E+00, -0.52411419E+00, 0.25429851E+00, 0.26375519E-01, -0.37648410E+00, 0.65987456E+00, 
    },
    {
      -0.54566395E+00, 0.74088342E-01, -0.10294523E+01, -0.16286100E+00, -0.72648101E-01, -0.86193547E-01, 0.27236110E+00, 0.20283155E-01, -0.70224538E-01, 
    },
    {
      -0.12850333E+01, 0.83220738E+00, -0.85517123E-01, 0.36889663E+00, -0.78059815E-01, 0.37251141E-01, -0.59753969E-01, -0.16131189E+00, -0.22452673E+00, 
    },
    {
      0.14172664E+00, 0.15790805E+00, 0.79502696E+00, 0.10112376E+01, -0.61172265E+00, 0.34335397E-01, 0.59816524E-01, 0.33198726E+00, -0.48746240E+00, 
    },
    {
      0.10491930E+01, -0.24701853E+00, 0.15002282E+01, -0.29250708E+00, 0.31989691E+00, -0.20179859E+00, 0.11427309E+00, 0.24893668E+00, -0.36306697E+00, 
    },
    {
      -0.13550294E-01, -0.45300750E-02, 0.10528672E-02, 0.16152516E-01, -0.12370551E+00, -0.59309341E-02, 0.40165852E-02, 0.45750680E-03, -0.59628934E-02, 
    },
    {
      -0.12362761E-01, -0.79272036E-02, 0.50656572E-02, 0.11978005E+00, 0.42122323E-01, 0.62554074E-03, -0.15799779E-01, -0.46376414E-02, 0.30326883E-02, 
    },
    {
      0.40974230E-01, 0.18386219E+00, 0.66352732E-01, 0.13339929E+00, -0.20539436E+00, 0.68436977E-02, 0.56280051E-02, 0.26018705E-01, -0.77441498E-02, 
    },
    {
      0.30746153E-01, 0.64624652E-01, 0.41689617E-02, 0.32169458E+00, 0.19580451E+00, 0.16591450E-01, -0.15418100E-01, -0.22828069E-01, 0.79204692E-02, 
    },
    {
      -0.79213560E-01, 0.82305610E-01, -0.45376107E-01, -0.24927042E+00, 0.59723146E-01, 0.52313942E-01, -0.31539883E-01, -0.39117865E-01, 0.51201828E-01, 
    },
    {
      -0.16604353E+00, -0.34118190E-01, -0.95448852E-01, -0.24441021E-01, -0.38288221E+00, 0.89257881E-02, 0.12591472E+00, 0.38770281E-01, 0.11474446E-01, 
    },
    {
      -0.95860370E-01, -0.84688097E+00, -0.15334813E+00, -0.17581967E+00, 0.49173304E+00, -0.18058453E-01, -0.11134413E+00, -0.12453195E+00, 0.18030409E-01, 
    },
    {
      0.21763852E+00, -0.97603641E-01, 0.88954560E-01, -0.84687620E+00, -0.32350314E+00, 0.37710898E-01, 0.73295427E-02, 0.13443394E+00, -0.16885924E+00, 
    },
    {
      0.10223650E+00, -0.29878554E+00, 0.48473828E-01, 0.14142914E+00, -0.96604209E-02, -0.75192235E-01, 0.95178723E-01, 0.81541315E-01, -0.61776754E-01, 
    },
    {
      0.29203841E+00, -0.17904034E+00, 0.66509947E-01, -0.33817422E-01, 0.29854304E+00, -0.61735559E-01, -0.15191223E+00, -0.53076677E-01, -0.26007971E-01, 
    },
    {
      -0.22676747E-01, 0.11153493E+01, 0.22664391E+00, -0.13381992E-01, -0.44242853E+00, 0.69775105E-01, 0.11397256E+00, 0.13336641E+00, -0.25085449E-01, 
    },
    {
      -0.44642001E+00, 0.18625054E+00, -0.79894532E-02, 0.63482380E+00, 0.70025556E-01, -0.66104084E-01, 0.85890360E-01, -0.14158376E+00, 0.24665558E+00, 
    },
    {
      -0.29041942E-01, 0.10848622E-01, 0.89042038E-02, -0.12965817E-01, -0.79534082E-02, 0.20859802E-01, -0.14016238E-01, -0.29190006E-02, -0.91342576E-03, 
    },
    {
      -0.55187498E-02, -0.40372211E-01, -0.21171888E-01, 0.10458578E-01, -0.23211660E-01, 0.95567387E-02, 0.19028161E-01, -0.29276188E-02, -0.41621849E-02, 
    },
    {
      0.39665487E-01, 0.72864085E-01, 0.10825509E-01, -0.18992426E-01, 0.10280470E-01, 0.83170965E-01, 0.42992984E-02, 0.93182252E-03, 0.22329555E-02, 
    },
    {
      0.22737475E-01, -0.66708000E-02, -0.27572964E-02, -0.16470218E-01, 0.29193019E-02, -0.60508568E-02, 0.88432148E-01, -0.63352720E-02, 0.46102949E-02, 
    },
    {
      0.47638151E-02, -0.11769389E+00, -0.53772289E-01, 0.17596751E-01, 0.23333792E-03, -0.14802746E+00, -0.58708407E-01, 0.86851157E-02, 0.21828359E-01, 
    },
    {
      -0.43172285E-01, 0.11592652E+00, -0.16070418E-02, -0.11035030E-01, 0.17440675E-01, 0.57708938E-01, -0.13491406E+00, -0.18783395E-03, 0.19392852E-01, 
    },
    {
      -0.20160457E-01, -0.94742822E-02, 0.71759149E-02, 0.22431763E-03, -0.19019587E-02, 0.46471963E-02, -0.72945268E-02, 0.25696186E-01, 0.10441859E-01, 
    },
    {
      -0.15380017E-01, 0.67414884E-02, -0.12804105E-01, 0.24585058E-02, -0.51557599E-02, 0.37254030E-02, 0.76002465E-02, -0.13046168E-01, 0.20730097E-01, 
    },
    {
      0.57154182E-01, -0.87949578E-02, 0.30844212E-01, -0.17324708E-01, -0.67216763E-03, 0.66427323E-02, 0.19945210E-01, 0.24971852E-01, 0.30873109E-01, 
    },
    {
      0.13570253E-01, -0.68256535E-01, 0.30832380E-01, 0.22241361E-01, -0.16376127E-02, -0.12042126E-01, -0.16067507E-02, -0.53055108E-01, 0.24462879E-01, 
    },
    {
      0.78554545E-02, -0.98696323E-02, -0.61809998E-02, 0.83083156E-02, -0.34189986E-02, 0.37059986E-02, -0.70023979E-02, 0.45059728E-02, -0.20708663E-02, 
    },
    {
      0.94396900E-02, 0.11253028E-01, -0.92065604E-02, 0.81361667E-03, -0.75608422E-03, 0.33125300E-02, 0.41626222E-02, 0.37212600E-02, 0.41885455E-02, 
    },
    {
      0.60033705E-02, 0.81377747E-02, 0.13521597E-02, 0.25659315E-02, -0.80228671E-02, 0.31734100E-02, 0.90754894E-03, 0.38459085E-03, -0.12874377E-02, 
    },
    {
      -0.29407945E-02, -0.58028940E-02, -0.11697264E-01, 0.18782510E-02, 0.10815329E-01, -0.28420230E-02, 0.76707790E-03, -0.40857978E-02, 0.25540697E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir18_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.51813760E+01, -0.10739883E+00, 0.18743367E+00, 0.31032026E-01, -0.31483453E-01, 0.46369401E-02, -0.17831052E-01, 0.25944985E-01, -0.39442893E-01, 0.11673883E-01, 0.13706181E-01, 0.52697590E-03, -0.25141253E-02, 
    },
    {
      -0.16610043E+01, -0.63904011E+00, 0.52825075E+00, -0.35923615E+00, 0.32526955E+00, 0.38260385E-01, 0.74715331E-01, -0.12071173E+00, -0.12027262E+00, -0.22109205E-01, -0.18548496E-02, 0.25158874E-02, 0.43518778E-01, 
    },
    {
      0.11134710E+02, 0.21947019E+01, -0.38336279E+01, 0.41743028E+00, 0.46625581E+00, -0.36733770E+00, 0.96669585E+00, -0.55345803E+00, 0.46823847E+00, -0.48935539E+00, -0.64198560E+00, -0.11552801E+00, 0.34517925E-01, 
    },
    {
      0.20840195E+02, 0.53288760E+01, -0.30419121E+01, 0.70807428E+01, -0.30199430E+01, 0.25945339E-01, -0.33274748E+01, 0.13975244E+01, 0.53362620E+00, 0.27826801E+00, 0.33125132E-01, 0.62786597E+00, -0.52923745E+00, 
    },
    {
      -0.79337822E+02, -0.14883757E+02, 0.22834850E+02, -0.51182194E+01, -0.11177418E+02, 0.16302793E+01, -0.79739599E+01, 0.32296658E+01, -0.12089566E+01, 0.36283174E+01, 0.44965773E+01, 0.86827540E+00, -0.31795710E+00, 
    },
    {
      -0.60320335E+02, -0.19182150E+02, -0.33956966E+01, -0.32669117E+02, 0.17268829E+02, -0.60134012E+00, 0.18945456E+02, -0.86268539E+01, -0.11386640E+01, -0.19328128E+01, 0.17548665E+01, -0.65107803E+01, 0.11776744E+01, 
    },
    {
      0.18447604E+03, 0.42060055E+02, -0.56183029E+02, 0.14291060E+02, 0.41250343E+02, -0.34360697E+01, 0.21798124E+02, -0.76566525E+01, 0.13683516E+01, -0.94252605E+01, -0.11468615E+02, -0.22709808E+01, 0.12029746E+01, 
    },
    {
      0.74724304E+02, 0.34598545E+02, 0.32904587E+02, 0.59504917E+02, -0.45993324E+02, -0.96503580E+00, -0.41161522E+02, 0.22533396E+02, 0.26252196E+01, 0.52893238E+01, -0.82092028E+01, 0.18131802E+02, -0.36847976E+00, 
    },
    {
      -0.19413287E+03, -0.49194946E+02, 0.61412071E+02, -0.15853179E+02, -0.53150242E+02, 0.39649019E+01, -0.23896528E+02, 0.79412022E+01, -0.10084863E+01, 0.10269410E+02, 0.12102466E+02, 0.25096750E+01, -0.16858004E+01, 
    },
    {
      -0.42284904E+02, -0.29988331E+02, -0.47852016E+02, -0.47470257E+02, 0.53297440E+02, 0.41365795E+01, 0.38657391E+02, -0.24937689E+02, -0.36407509E+01, -0.61866503E+01, 0.11591171E+02, -0.19596478E+02, -0.89709473E+00, 
    },
    {
      0.75968063E+02, 0.19798077E+02, -0.24477795E+02, 0.62807326E+01, 0.22585154E+02, -0.18408319E+01, 0.91339474E+01, -0.30013289E+01, 0.41616115E+00, -0.40143032E+01, -0.44904690E+01, -0.10010843E+01, 0.77694696E+00, 
    },
    {
      0.95251894E+01, 0.10003698E+02, 0.20788961E+02, 0.13931005E+02, -0.21926090E+02, -0.26575904E+01, -0.13174495E+02, 0.97861652E+01, 0.17670221E+01, 0.25980010E+01, -0.51927876E+01, 0.73529782E+01, 0.56848937E+00, 
    },
    {
      -0.20119770E+00, 0.15843167E+01, 0.18475239E+01, -0.47732759E-03, -0.13149090E+00, -0.13344645E-01, 0.26832491E-01, 0.30770525E-01, -0.20653838E-01, 0.10311009E-02, 0.70003867E-02, -0.62839808E-02, -0.99203400E-02, 
    },
    {
      0.70720062E-01, -0.16220536E+01, 0.15745577E+01, 0.24463184E-01, 0.20518380E+00, 0.19001585E-01, 0.56619728E-02, -0.39436924E-02, 0.51251244E-01, -0.10332140E-01, 0.11076461E-01, -0.14877492E-01, 0.19403499E-01, 
    },
    {
      -0.90413511E+00, 0.15951061E+01, -0.32156929E+00, 0.67641193E+00, -0.58352768E+00, 0.14265889E+00, 0.43839268E-01, -0.37467858E+00, 0.20368481E+00, 0.66211246E-01, 0.16507994E+00, 0.45068026E-01, 0.21851064E+00, 
    },
    {
      -0.25522104E+00, 0.10744387E+01, 0.77017993E+00, 0.52008659E+00, 0.31209913E+00, 0.15328670E+00, -0.16450721E+00, 0.30181509E-01, 0.98174199E-01, -0.20944310E-01, -0.68538651E-01, 0.44294223E-01, 0.10606261E+00, 
    },
    {
      0.43100529E+01, 0.10209758E+02, 0.10042840E+02, 0.17070240E+01, 0.11407772E+00, 0.15409266E+01, -0.69876671E+00, -0.72542679E+00, 0.74991882E+00, -0.28893912E+00, -0.34130789E-01, 0.30699372E+00, 0.49604353E+00, 
    },
    {
      -0.37056020E+00, -0.13278563E+02, 0.55308051E+01, 0.54367453E+00, -0.10339448E+01, -0.52278483E+00, 0.11216925E+01, -0.80142033E+00, -0.27339029E+00, 0.27401743E-01, 0.37330829E-02, 0.88551909E+00, -0.19130912E+00, 
    },
    {
      0.15317790E+02, -0.13764858E+02, 0.45487285E+01, -0.40844550E+01, 0.10961870E+02, -0.27308998E+01, -0.39546218E+01, 0.59212508E+01, -0.20589113E+01, -0.41854739E+00, -0.23430898E+01, -0.11522582E+01, -0.30893290E+01, 
    },
    {
      0.34036417E+01, -0.14668043E+02, -0.73346114E+01, -0.88265228E+01, -0.11719311E+02, -0.30766165E+01, 0.42914267E+01, -0.19813076E+01, -0.91186047E+00, -0.14789377E+01, 0.34925082E+00, 0.82151622E+00, -0.14842312E+01, 
    },
    {
      -0.21855589E+02, -0.65027206E+02, -0.30625284E+02, -0.61926742E+01, -0.35101626E+01, -0.17476919E+02, 0.71970272E+00, 0.40459504E+01, -0.35212376E+01, 0.19424437E+01, -0.18655061E+01, -0.29581990E+01, -0.35298369E+01, 
    },
    {
      0.45670981E+01, 0.37643353E+02, -0.25258648E+02, -0.83348694E+01, 0.25785327E+01, 0.53328462E+01, -0.97375669E+01, 0.63686438E+01, 0.81883103E+00, -0.19984325E+01, -0.17359768E+01, -0.85691080E+01, 0.44630432E+00, 
    },
    {
      -0.83068283E+02, 0.81330490E+02, -0.13745291E+02, -0.39484618E+01, -0.48257332E+02, 0.27937885E+02, 0.28801771E+02, -0.31976980E+02, 0.32318890E+01, 0.10558324E+01, 0.14404617E+02, 0.91635990E+01, 0.14174400E+02, 
    },
    {
      -0.37831264E+02, 0.88615479E+02, 0.85113358E+02, 0.45270039E+02, 0.76213264E+02, 0.17117207E+02, -0.29024532E+02, 0.17087751E+02, 0.68407333E+00, 0.12642257E+02, 0.57222117E-01, -0.97055807E+01, 0.69868464E+01, 
    },
    {
      0.33955742E+02, 0.13318353E+03, -0.21473759E+02, 0.25355750E+00, 0.13797762E+02, 0.59645039E+02, 0.50354247E+01, -0.86467590E+01, 0.35548909E+01, -0.50609388E+01, 0.10475528E+02, 0.11849051E+02, 0.96025867E+01, 
    },
    {
      -0.88309050E+01, 0.25575996E+02, 0.12401312E+02, 0.32052364E+02, -0.87352157E+00, -0.18661947E+02, 0.23053522E+02, -0.18508951E+02, -0.16507331E+01, 0.99664297E+01, 0.73132496E+01, 0.25814552E+02, -0.15154300E+01, 
    },
    {
      0.21601781E+03, -0.21782782E+03, 0.13970534E+02, 0.51641003E+02, 0.82136398E+02, -0.96018127E+02, -0.80168480E+02, 0.78118790E+02, 0.89466515E+01, 0.96090563E-01, -0.40393982E+02, -0.29548920E+02, -0.28057667E+02, 
    },
    {
      0.11636549E+03, -0.24491010E+03, -0.28724524E+03, -0.10040260E+03, -0.19873129E+03, -0.43567253E+02, 0.75220818E+02, -0.51479694E+02, 0.52787185E+01, -0.37214550E+02, -0.12265165E+01, 0.29498684E+02, -0.13845214E+02, 
    },
    {
      -0.66059847E+01, -0.12299647E+03, 0.11331442E+03, 0.16227594E+02, -0.18005981E+02, -0.79721268E+02, -0.89493876E+01, 0.85854511E+01, 0.19708027E+01, 0.53805790E+01, -0.17981739E+02, -0.18486019E+02, -0.10671741E+02, 
    },
    {
      0.65107279E+01, -0.14133873E+03, 0.35762993E+02, -0.47331753E+02, -0.99259748E+01, 0.24949978E+02, -0.18619167E+02, 0.21813089E+02, 0.29833362E+01, -0.15825300E+02, -0.94992371E+01, -0.31681845E+02, 0.30539851E+01, 
    },
    {
      -0.26602240E+03, 0.25943344E+03, -0.85672159E+01, -0.89114342E+02, -0.59151909E+02, 0.12887794E+03, 0.96713982E+02, -0.87241043E+02, -0.24014769E+02, -0.39270320E+01, 0.50494331E+02, 0.40384113E+02, 0.24323547E+02, 
    },
    {
      -0.12623465E+03, 0.30866837E+03, 0.37077533E+03, 0.10442806E+03, 0.23109668E+03, 0.52738197E+02, -0.84341431E+02, 0.64608040E+02, -0.10220551E+02, 0.46504440E+02, 0.22121127E+00, -0.34443577E+02, 0.11775063E+02, 
    },
    {
      -0.12208918E+02, 0.45470352E+02, -0.71684982E+02, -0.12210121E+02, 0.75946212E+01, 0.36560390E+02, 0.36604092E+01, -0.33766203E+01, -0.28442261E+01, -0.19091837E+01, 0.96889944E+01, 0.94606447E+01, 0.40775514E+01, 
    },
    {
      -0.14950647E+01, 0.93858498E+02, -0.31066874E+02, 0.23789320E+02, 0.11685203E+02, -0.11222384E+02, 0.36598971E+01, -0.87570076E+01, -0.21024263E+01, 0.80605850E+01, 0.36843083E+01, 0.13835386E+02, -0.17752295E+01, 
    },
    {
      0.12323454E+03, -0.11393223E+03, 0.28638356E+01, 0.45115490E+02, 0.14615677E+02, -0.59010235E+02, -0.42132607E+02, 0.35980793E+02, 0.14087955E+02, 0.33986540E+01, -0.22851339E+02, -0.19381165E+02, -0.74500413E+01, 
    },
    {
      0.42877964E+02, -0.14127107E+03, -0.16386626E+03, -0.42104141E+02, -0.10067329E+03, -0.24085188E+02, 0.34423313E+02, -0.28762308E+02, 0.50822926E+01, -0.20915609E+02, 0.88597655E+00, 0.13771997E+02, -0.35257232E+01, 
    },
    {
      0.42188969E-01, 0.94234280E-01, -0.10187069E-02, -0.95930177E+00, -0.19991234E+00, 0.58659386E-01, 0.31168277E-02, -0.84970305E-02, 0.53929232E-01, -0.65914979E-02, 0.57587521E-02, -0.14161857E-01, 0.11686263E-03, 
    },
    {
      0.43799311E-01, 0.11469412E-01, 0.45734111E-01, 0.27424732E+00, -0.63086385E+00, 0.69261733E-02, 0.10186278E-02, -0.43821827E-01, 0.95467009E-02, -0.14789114E-01, 0.76593342E-02, -0.46657752E-02, 0.73617133E-02, 
    },
    {
      0.49559942E+00, 0.82897675E+00, 0.47035179E+00, 0.17087352E+00, -0.12786779E+01, 0.15497789E-01, -0.29272199E+00, 0.50515074E-01, 0.95190831E-01, 0.12434431E+00, 0.34540424E-02, -0.43415297E-01, 0.11710780E+00, 
    },
    {
      0.91401197E-01, 0.34288970E+00, 0.77033257E+00, 0.19686368E+01, 0.12333189E-02, 0.10140464E+00, 0.14774442E+00, -0.73128659E-02, -0.67382760E-01, -0.95939159E-01, 0.81297606E-01, 0.10078107E+00, 0.75467929E-01, 
    },
    {
      0.15210741E+01, 0.97265679E+00, -0.11593559E+00, 0.81999540E+01, 0.12909287E+00, -0.11842973E+01, -0.96201396E+00, 0.31217158E+00, -0.70322382E+00, -0.43442760E-01, -0.39182207E+00, -0.10794251E+00, -0.17477946E-01, 
    },
    {
      0.28209305E+01, 0.63783765E+00, 0.71414834E+00, -0.10531855E+01, 0.46761045E+01, 0.80202073E+00, 0.12529795E-01, 0.56188387E+00, -0.26510638E+00, 0.43705702E+00, 0.11708755E+00, -0.95809540E-02, 0.27964649E+00, 
    },
    {
      -0.60612836E+01, -0.50309405E+01, -0.43330379E+01, 0.93652105E+01, 0.22353725E+01, 0.19453353E+00, 0.13129736E+01, -0.80170625E+00, -0.83791220E+00, -0.26327208E+00, 0.91290653E+00, 0.65617907E+00, -0.94384933E+00, 
    },
    {
      -0.78995579E+00, -0.36866460E+01, -0.43037763E+01, -0.60018635E+01, 0.95902605E+01, 0.52022409E+00, -0.14935763E+01, -0.27530113E+00, 0.18274926E+00, 0.21199745E+00, -0.55878675E+00, -0.25891751E+00, 0.83397602E-03, 
    },
    {
      -0.13319073E+02, -0.63696756E+01, 0.26987333E+01, -0.92368822E+01, 0.32915807E+01, 0.60375071E+01, 0.76720295E+01, -0.18980306E+01, 0.46611791E+01, 0.43055254E+00, 0.20732810E+01, 0.14235936E+01, 0.31788766E+00, 
    },
    {
      -0.14097284E+02, -0.46897783E+01, -0.82707202E+00, -0.41121367E+00, 0.23824086E+01, -0.26390638E+01, 0.35508257E+00, -0.23547945E+01, 0.10697405E+01, -0.21197121E+01, -0.27898791E+01, 0.12480121E+01, -0.18391019E+01, 
    },
    {
      0.15204087E+02, 0.78556147E+01, 0.88966312E+01, -0.30797503E+02, 0.23227117E+01, -0.58457482E+00, -0.71814245E+00, 0.30694103E+01, 0.22050869E+01, -0.34793630E+00, -0.29825599E+01, -0.19286658E+01, 0.21556101E+01, 
    },
    {
      0.39697590E+01, 0.58501611E+01, 0.62696495E+01, 0.49510407E+01, -0.31089190E+02, -0.24617631E+01, 0.39447894E+01, 0.16149390E+01, -0.56727923E-01, 0.75868088E+00, 0.79545927E+00, 0.32469195E+00, -0.90026625E-01, 
    },
    {
      0.27027122E+02, 0.11467045E+02, -0.84330359E+01, -0.24010477E+02, -0.80087805E+01, -0.12543470E+02, -0.15589190E+02, 0.42210331E+01, -0.10726813E+02, -0.14858339E+01, -0.43507633E+01, -0.34340184E+01, -0.92487770E+00, 
    },
    {
      0.23143436E+02, 0.78808756E+01, -0.27895360E+01, 0.60927572E+01, -0.40888386E+02, 0.94554806E+00, -0.24960201E+01, 0.34536648E+01, -0.18722475E+01, 0.37619510E+01, 0.80734539E+01, -0.44939146E+01, 0.28656080E+01, 
    },
    {
      -0.11953168E+02, -0.27238479E+01, -0.52568307E+01, 0.21050129E+02, -0.68679848E+01, 0.32113910E+00, -0.58306420E+00, -0.27074990E+01, -0.17911482E+01, 0.77064550E+00, 0.22215376E+01, 0.15297557E+01, -0.14433203E+01, 
    },
    {
      -0.47204428E+01, -0.10654535E+01, -0.18083954E+01, 0.15844553E+01, 0.20997583E+02, 0.19634615E+01, -0.33727379E+01, -0.18170273E+01, -0.17987987E+00, -0.10957071E+01, -0.20409736E+00, -0.20305949E+00, -0.28999722E+00, 
    },
    {
      -0.15884968E+02, -0.64883618E+01, 0.73605752E+01, 0.31496780E+02, 0.69600091E+01, 0.86443911E+01, 0.96710367E+01, -0.29579740E+01, 0.75072083E+01, 0.12868271E+01, 0.29590070E+01, 0.23414221E+01, 0.84001136E+00, 
    },
    {
      -0.11925214E+02, -0.50585051E+01, 0.18607297E+01, -0.73451662E+01, 0.40590572E+02, 0.12435257E+01, 0.30310061E+01, -0.15552270E+01, 0.12240763E+01, -0.21083274E+01, -0.60944834E+01, 0.37612362E+01, -0.10206302E+01, 
    },
    {
      0.64615183E-01, 0.78795552E-01, -0.15892947E+00, -0.10464617E+00, 0.10353974E+00, -0.41167954E+00, -0.43749309E+00, 0.20638651E-02, 0.74383467E-01, -0.43644945E-03, -0.90350350E-03, -0.19853504E-01, -0.59061525E-02, 
    },
    {
      0.81961580E-01, -0.21425916E+00, -0.38655456E-01, -0.88595822E-02, -0.14763263E+00, 0.46912336E+00, -0.45955449E+00, -0.42392034E-01, -0.80482624E-02, 0.18471889E-01, -0.14141251E-01, 0.55350456E-02, 0.16753854E-01, 
    },
    {
      -0.24682516E+00, 0.65146202E+00, -0.45703122E+00, -0.23654184E+00, -0.30039293E+00, -0.82151897E-01, -0.20593224E+00, -0.13945434E-01, 0.16855291E+00, 0.32676257E-01, 0.80637790E-01, -0.79479448E-01, -0.61515283E-01, 
    },
    {
      -0.29577321E+00, -0.25834915E+00, 0.53513624E-01, 0.29484292E-01, -0.17856269E+00, 0.19976071E+00, 0.16096137E+00, -0.13107130E+00, -0.10473933E+00, -0.19589696E-01, 0.30329490E-01, 0.64674303E-01, -0.38300402E-01, 
    },
    {
      -0.60273278E+00, -0.20669493E+00, 0.82746637E+00, 0.46272588E+00, -0.10339417E+00, 0.15778165E+00, -0.30000055E+00, -0.28547913E+00, 0.21799399E-01, -0.91843419E-01, -0.55076419E-02, 0.13837788E-01, -0.14949028E+00, 
    },
    {
      -0.13389989E+01, 0.26402633E+01, -0.13628585E+01, -0.36753692E-01, 0.15853026E+01, 0.13760415E+00, 0.79008830E+00, -0.23089467E+00, 0.14146335E+00, -0.11153022E+00, 0.23030405E+00, -0.12568083E+00, -0.15787458E+00, 
    },
    {
      -0.16602457E+00, -0.13449451E+01, 0.96737868E+00, 0.40215719E+00, 0.72849149E+00, 0.41462141E+00, 0.71418375E+00, 0.10896401E+00, -0.35342368E+00, -0.11979083E+00, -0.10750519E+00, 0.23451699E+00, 0.13568750E+00, 
    },
    {
      0.14382593E-01, 0.82095027E+00, -0.72794154E-01, -0.23630470E+00, 0.75462168E+00, -0.65143400E+00, 0.49917012E-01, 0.24157877E+00, 0.31928110E+00, -0.23850443E-01, -0.63626952E-01, -0.70730507E-01, 0.35635930E-01, 
    },
    {
      0.16258487E+01, 0.63477248E+00, -0.11753286E+01, -0.52329248E+00, 0.51585090E+00, 0.42656636E+00, 0.12154765E+01, 0.54212022E+00, -0.33579046E+00, 0.13298018E+00, -0.58626540E-01, -0.93792476E-01, 0.21695583E+00, 
    },
    {
      0.19641677E+01, -0.34773593E+01, 0.24826541E+01, -0.47208840E+00, -0.22139666E+01, -0.79506546E+00, -0.24327517E+00, 0.67086709E+00, -0.30561268E-01, 0.17824247E+00, -0.29080093E+00, 0.13225445E+00, 0.22372004E+00, 
    },
    {
      0.78067645E-01, -0.53872090E-01, -0.60876328E-01, -0.20090356E-01, 0.19590052E-01, 0.37616719E-01, -0.21996699E-01, 0.30491668E+00, -0.21350857E-01, 0.73222145E-02, 0.10664701E-01, -0.17186703E-01, -0.79840180E-02, 
    },
    {
      0.54252598E-01, -0.42442694E-01, 0.41468631E-01, 0.11270581E-01, -0.79623416E-01, -0.27057895E-01, 0.20517232E-01, 0.19652473E-01, 0.27900285E+00, 0.76455669E-02, -0.26138762E-02, 0.26082578E-02, -0.48174490E-02, 
    },
    {
      0.37738148E-01, 0.49505927E-01, 0.45610685E-01, -0.40265694E-01, -0.23026168E+00, -0.11767189E+00, -0.47708195E-01, 0.25821838E+00, 0.16750067E+00, -0.34544688E-01, -0.48518635E-01, 0.20903081E-01, 0.94813956E-02, 
    },
    {
      0.18504436E+00, 0.13707677E+00, -0.76823071E-01, 0.73215045E-01, 0.25763610E+00, -0.57229973E-01, -0.43995194E-01, -0.27386624E+00, 0.97568236E-01, 0.11287920E-01, -0.45958824E-01, -0.17281851E-01, -0.18543916E-01, 
    },
    {
      -0.10852473E+00, -0.31554475E-01, -0.29181827E-01, -0.35038237E-01, -0.30168429E-01, 0.12917520E-01, 0.94219139E-02, 0.38543744E-02, -0.17429788E-01, 0.97568251E-01, 0.17038140E+00, -0.22542167E-02, -0.68581281E-02, 
    },
    {
      -0.16812027E-01, 0.67123897E-01, -0.14856836E-01, 0.52298285E-01, -0.10310156E+00, -0.40235251E-01, -0.89705503E-02, 0.38926877E-01, -0.10838482E-02, -0.16985069E+00, 0.76311469E-01, 0.69150580E-02, -0.63182632E-02, 
    },
    {
      -0.48873704E-01, -0.22341227E-01, 0.31494450E-01, -0.74725705E-02, -0.32325108E-01, 0.47166054E-02, 0.15750030E-01, 0.19931326E-01, 0.61885430E-02, -0.89711556E-02, -0.22252218E-02, -0.13308422E+00, 0.20251697E-01, 
    },
    {
      -0.45145783E-02, 0.35775520E-01, -0.10440513E-01, 0.63793033E-01, 0.43009203E-01, -0.34634177E-01, -0.14203522E-01, 0.64703203E-02, -0.13139685E-01, -0.32075319E-01, -0.19045237E-01, -0.12680903E-01, -0.13717172E+00, 
    },
    {
      0.29166550E-02, 0.11035411E-01, 0.24293691E-01, 0.25893958E-01, -0.13120287E-02, -0.17062493E-01, -0.69651459E-02, 0.15115200E-01, -0.12452925E-01, -0.28048914E-01, -0.81404569E-02, 0.14331407E-01, 0.12644852E-01, 
    },
    {
      0.24846908E-01, -0.53721450E-01, 0.16348697E-01, 0.18280599E-01, -0.24520375E-01, -0.36987716E-02, 0.18260539E-01, -0.72437376E-02, 0.14151407E-01, -0.22939686E-01, 0.24036895E-02, 0.15635617E-01, 0.12054204E-01, 
    },
    {
      0.11425200E-01, 0.64702697E-01, -0.10536327E+00, 0.15012830E-01, -0.28842267E-01, 0.11042488E-01, 0.13311166E-01, -0.21481495E-02, 0.73922775E-02, -0.16001927E-01, 0.10468615E-01, -0.66283136E-02, -0.15343135E-02, 
    },
    {
      -0.30380554E-01, -0.51950170E-02, -0.93821893E-02, 0.29672759E-01, 0.30377174E-01, 0.16042337E-01, -0.11458547E-01, -0.95287329E-02, 0.13136099E-01, 0.89334687E-02, -0.17647015E-01, -0.20479793E-02, 0.17140299E-01, 
    },
  },
  {
    {
      0.84415445E+01, 0.26355572E-01, 0.16610390E+00, 0.69635049E-01, 0.77834222E-02, -0.25760174E-01, -0.21441121E-01, 0.21861907E-01, -0.43986645E-01, 0.11674735E-01, 0.56952261E-02, 0.28399102E-01, -0.13653146E-01, 
    },
    {
      -0.68217140E+00, -0.62160009E+00, 0.28212479E+00, -0.24901924E+00, 0.21900199E+00, 0.24887722E-01, 0.72718799E-01, -0.13336083E+00, -0.78881793E-01, 0.14562245E+00, 0.74521393E-01, -0.27514610E-01, 0.10257617E+00, 
    },
    {
      0.17437782E+02, 0.79321623E+00, -0.35100787E+01, 0.84811628E-01, 0.70552386E-01, 0.47351521E+00, 0.83660269E+00, -0.47608498E+00, 0.52839464E+00, -0.69492771E-02, -0.20629799E+00, -0.32985020E+00, 0.31779465E+00, 
    },
    {
      -0.81596603E+01, 0.80805292E+01, 0.59097290E+01, 0.42979136E+01, -0.42025328E+01, -0.14056302E+01, -0.12244359E+01, 0.21699390E+01, 0.52206904E+00, -0.24186170E+01, -0.21609974E+01, 0.93807650E+00, -0.84871131E+00, 
    },
    {
      -0.89978989E+02, -0.97939730E+01, 0.20942669E+02, -0.55313463E+01, -0.76858745E+01, -0.21651825E+00, -0.45114717E+01, 0.38345082E+01, -0.16243830E+01, -0.89623278E+00, 0.11692472E+01, 0.93201983E+00, -0.17749968E+01, 
    },
    {
      0.10546338E+03, -0.44166641E+02, -0.43877007E+02, -0.16669228E+02, 0.34683037E+02, 0.49835873E+01, 0.66487484E+01, -0.12427212E+02, -0.19051713E+01, 0.11908665E+02, 0.12564396E+02, -0.36486394E+01, 0.21449244E+01, 
    },
    {
      0.13710806E+03, 0.36479836E+02, -0.52778606E+02, 0.21492033E+02, 0.30154282E+02, -0.53919759E+01, 0.90821161E+01, -0.11428968E+02, 0.14775454E+01, 0.36835520E+01, -0.33135943E+01, -0.38186175E+00, 0.41576314E+01, 
    },
    {
      -0.28806189E+03, 0.10074989E+03, 0.97997749E+02, 0.22048018E+02, -0.10003945E+03, -0.56709027E+01, -0.15209171E+02, 0.29163034E+02, 0.34512579E+01, -0.24924685E+02, -0.28658514E+02, 0.41394582E+01, -0.29731119E+01, 
    },
    {
      -0.92897758E+02, -0.49600414E+02, 0.61100552E+02, -0.29214361E+02, -0.39369938E+02, 0.10823676E+02, -0.76445956E+01, 0.13982789E+02, 0.32520753E+00, -0.46346078E+01, 0.43214622E+01, -0.86068255E+00, -0.42901020E+01, 
    },
    {
      0.30256940E+03, -0.99049271E+02, -0.91175804E+02, -0.90334110E+01, 0.11542465E+03, 0.17341838E+01, 0.15239948E+02, -0.29818983E+02, -0.28255911E+01, 0.23323519E+02, 0.28412251E+02, -0.63854128E+00, 0.26830745E+01, 
    },
    {
      0.24605694E+02, 0.22045202E+02, -0.26002682E+02, 0.13213135E+02, 0.16756659E+02, -0.57091608E+01, 0.22745020E+01, -0.59658070E+01, -0.66697073E+00, 0.18124434E+01, -0.19950210E+01, 0.59119934E+00, 0.16144733E+01, 
    },
    {
      -0.11073072E+03, 0.35112747E+02, 0.30812071E+02, -0.38981718E+00, -0.46086620E+02, 0.31463280E+00, -0.55018616E+01, 0.11076973E+02, 0.82669556E+00, -0.80218763E+01, -0.10252825E+02, -0.76681507E+00, -0.11369579E+01, 
    },
    {
      -0.24553841E+00, 0.17193551E+01, 0.16381207E+01, 0.33417907E-01, -0.87884702E-01, -0.44835400E-01, 0.15925823E-02, 0.49698550E-01, -0.38588453E-01, -0.72526657E-02, 0.15999507E-01, 0.17281709E-01, -0.23321385E-01, 
    },
    {
      0.24167123E+00, -0.16023287E+01, 0.16606361E+01, -0.45682013E-01, 0.11704248E+00, -0.18465983E-01, 0.11106687E-01, -0.60868151E-02, 0.61217647E-01, -0.20567151E-01, 0.24912491E-01, -0.28536351E-01, 0.29306814E-01, 
    },
    {
      -0.10224926E+01, 0.14925382E+01, -0.19260072E+01, 0.61678416E+00, 0.13985469E-01, 0.10479474E+00, -0.49587530E-02, -0.18433394E+00, 0.11050394E+00, 0.21527071E+00, 0.22213718E+00, 0.34993712E-01, 0.57391070E-01, 
    },
    {
      0.10419737E+01, 0.66445446E+00, 0.14948963E+01, 0.38029797E-01, 0.27230132E+00, 0.42719600E+00, -0.11771204E+00, 0.30687597E+00, 0.19784185E+00, 0.22538267E+00, 0.75966686E-01, 0.23985893E-01, 0.45384716E-01, 
    },
    {
      0.53208251E+01, 0.75190620E+01, -0.22172494E+01, -0.12778753E+00, 0.56682128E-01, 0.18827899E+01, 0.72104907E+00, -0.80092132E+00, 0.77964395E+00, -0.24134891E+00, -0.46352610E+00, 0.28311205E+00, 0.10742375E+01, 
    },
    {
      -0.41045341E+01, 0.57498431E+01, 0.83945713E+01, 0.23423393E+01, -0.14150887E+01, -0.54918933E+00, -0.10221728E+01, 0.26822236E+00, -0.12929893E+01, 0.97145742E+00, -0.42177832E+00, 0.68537563E+00, -0.58753085E+00, 
    },
    {
      0.14998870E+02, -0.16657772E+02, 0.37549644E+02, -0.95424957E+01, -0.54443731E+01, -0.14911759E+01, 0.17992316E+01, 0.51960616E+01, 0.45644295E+00, -0.39563315E+01, -0.49043798E+01, 0.12105607E+00, 0.70748419E+00, 
    },
    {
      -0.60123081E+01, 0.27362645E+01, -0.13724414E+02, 0.12894962E+01, -0.14900351E+02, -0.52518587E+01, 0.34944665E+01, -0.55088906E+01, -0.13576542E+01, -0.56548223E+01, -0.14211199E+01, -0.34091601E+00, 0.21933839E+00, 
    },
    {
      -0.29305424E+02, 0.13374522E+02, 0.44242222E+02, 0.50953383E+01, 0.65631719E+01, -0.14111629E+02, -0.65622892E+01, 0.39515307E+01, -0.34545369E+01, -0.12505515E+00, 0.29348438E+01, -0.32276697E+01, -0.90113583E+01, 
    },
    {
      0.43037151E+02, -0.89085785E+02, 0.11644012E+02, -0.24075853E+02, 0.19164951E+02, 0.77643185E+01, 0.86020861E+01, -0.46943688E+01, 0.93667440E+01, -0.74655390E+01, 0.14679497E+01, -0.47342935E+01, 0.18102796E+01, 
    },
    {
      -0.85623055E+02, 0.92817024E+02, -0.23375591E+03, 0.44671661E+02, 0.45343346E+02, 0.83043785E+01, -0.13587065E+02, -0.29296474E+02, -0.10942899E+02, 0.20286388E+02, 0.29787083E+02, -0.44528899E+00, -0.79419918E+01, 
    },
    {
      -0.14370024E+02, -0.23429798E+02, 0.72968773E+02, -0.15687455E+02, 0.10109998E+03, 0.26097256E+02, -0.26932791E+02, 0.35635754E+02, -0.34211284E+00, 0.35934021E+02, 0.61775665E+01, 0.22777042E+01, -0.23924460E+01, 
    },
    {
      0.57428696E+02, -0.17855730E+03, -0.15708559E+03, -0.22408588E+02, -0.40592297E+02, 0.40333138E+02, 0.14209915E+02, -0.89817600E+01, 0.57742748E+01, 0.57205853E+01, -0.63581610E+01, 0.10133678E+02, 0.26636444E+02, 
    },
    {
      -0.14794983E+03, 0.30004440E+03, -0.19929813E+03, 0.84621696E+02, -0.69814247E+02, -0.29103554E+02, -0.25958654E+02, 0.20315407E+02, -0.28830307E+02, 0.23460487E+02, -0.31325624E+01, 0.13810139E+02, -0.21702385E+01, 
    },
    {
      0.23797343E+03, -0.23116394E+03, 0.57393121E+03, -0.83598228E+02, -0.12368264E+03, -0.19707600E+02, 0.34384033E+02, 0.70111160E+02, 0.39034550E+02, -0.41825325E+02, -0.75330421E+02, -0.22089810E+01, 0.25101606E+02, 
    },
    {
      0.10284250E+03, 0.69117165E+02, -0.15603969E+03, 0.39777988E+02, -0.25606186E+03, -0.60250473E+02, 0.77766205E+02, -0.99222710E+02, 0.16326077E+02, -0.91902000E+02, -0.83012609E+01, -0.85460234E+01, 0.51328115E+01, 
    },
    {
      -0.39939327E+02, 0.31202109E+03, 0.21929730E+03, 0.33807743E+02, 0.73635201E+02, -0.49816437E+02, -0.86225977E+01, 0.10290629E+02, -0.33790231E+01, -0.12311239E+02, 0.48300257E+01, -0.12229668E+02, -0.32109089E+02, 
    },
    {
      0.20875215E+03, -0.39679617E+03, 0.38521735E+03, -0.12168412E+03, 0.91634483E+02, 0.41888870E+02, 0.34228813E+02, -0.33702576E+02, 0.39293419E+02, -0.32169025E+02, 0.41854286E+01, -0.18185402E+02, 0.11737410E+01, 
    },
    {
      -0.29656155E+03, 0.26091183E+03, -0.63637982E+03, 0.62630611E+02, 0.13731348E+03, 0.20191547E+02, -0.35748486E+02, -0.76228554E+02, -0.50450886E+02, 0.36267136E+02, 0.83756058E+02, 0.57476335E+01, -0.32509529E+02, 
    },
    {
      -0.15768517E+03, -0.66360435E+02, 0.12792127E+03, -0.30300962E+02, 0.27846167E+03, 0.64009613E+02, -0.94658752E+02, 0.12312606E+03, -0.32563572E+02, 0.10454442E+03, 0.17428432E+01, 0.15118477E+02, -0.33578801E+01, 
    },
    {
      0.47491126E+01, -0.15489160E+03, -0.10394126E+03, -0.16034281E+02, -0.40621128E+02, 0.22269804E+02, -0.31044576E+00, -0.47275276E+01, 0.55015851E-01, 0.71959572E+01, -0.86488312E+00, 0.50432911E+01, 0.13676483E+02, 
    },
    {
      -0.10171490E+03, 0.18161353E+03, -0.21165048E+03, 0.60878799E+02, -0.37977203E+02, -0.20529871E+02, -0.16520407E+02, 0.18732552E+02, -0.19229500E+02, 0.15826124E+02, -0.22000675E+01, 0.87317457E+01, -0.31610513E+00, 
    },
    {
      0.13370317E+03, -0.11240935E+03, 0.26245120E+03, -0.15337643E+02, -0.54398438E+02, -0.74560561E+01, 0.13048158E+02, 0.30778091E+02, 0.21956047E+02, -0.10538049E+02, -0.33810032E+02, -0.32820210E+01, 0.14922835E+02, 
    },
    {
      0.75884071E+02, 0.17452677E+02, -0.31479095E+02, 0.36800358E+01, -0.11180386E+03, -0.25325590E+02, 0.41215767E+02, -0.55677891E+02, 0.18475767E+02, -0.44003674E+02, 0.18647236E+01, -0.90560875E+01, 0.22126122E+00, 
    },
    {
      0.59121020E-01, -0.38882446E-01, 0.10154726E+00, -0.99066663E+00, 0.50862730E+00, 0.37865343E-02, -0.13369663E-01, -0.14525892E-01, 0.73863156E-02, -0.30057946E-01, 0.92732534E-02, -0.24269311E-01, 0.67565674E-02, 
    },
    {
      -0.24163937E-01, -0.20808218E+00, 0.18645240E-02, -0.51261657E+00, -0.78584683E+00, -0.29707689E-01, -0.29643602E-02, -0.43101724E-01, 0.92033194E-02, -0.20032544E-01, 0.12454584E-01, -0.10859430E-01, 0.14212405E-01, 
    },
    {
      -0.20849434E+00, 0.50730848E+00, 0.54870909E+00, 0.17681440E+01, -0.10749120E+01, 0.42078006E+00, -0.26223657E+00, 0.33963424E+00, 0.82626760E-01, 0.92897750E-01, -0.60189381E-01, -0.76034263E-01, 0.18571973E-01, 
    },
    {
      0.42013156E+00, 0.22726022E+00, 0.90818733E+00, 0.16409394E+01, 0.20787294E+01, 0.16166556E+00, 0.25318131E+00, 0.91371499E-03, -0.12026191E+00, 0.11957009E-01, -0.17968928E-01, 0.20824531E+00, 0.10581681E+00, 
    },
    {
      0.18979921E+01, 0.71559265E-01, 0.89219481E+00, -0.64803618E+00, 0.19108589E+01, 0.61305553E-01, -0.76099968E+00, 0.66523030E-02, 0.38209528E+00, -0.37404042E+00, -0.24693862E-01, -0.15020821E-01, -0.21455880E+00, 
    },
    {
      0.42901478E+01, 0.27137227E+01, 0.18731354E+01, -0.25761666E+01, -0.37876692E-01, 0.42818061E+00, -0.39081356E+00, 0.64281893E+00, 0.43348350E-01, 0.28045836E+00, -0.51489584E-01, 0.33649018E+00, 0.13021860E+00, 
    },
    {
      0.61204880E+00, -0.18564703E+01, -0.39929540E+01, 0.96122243E-01, 0.16075824E+01, -0.39626312E+01, 0.11777747E+01, -0.33050480E+01, -0.11639585E+01, -0.89641178E+00, 0.77148652E+00, 0.10184772E+01, 0.13489571E+00, 
    },
    {
      -0.77427740E+01, -0.11954879E+01, -0.10486695E+02, -0.58320417E+01, -0.18617257E+01, 0.74399304E+00, -0.28453026E+01, -0.58667099E+00, 0.11166003E+01, -0.77045232E+00, 0.64419031E+00, -0.96420527E+00, -0.97287238E+00, 
    },
    {
      -0.14225009E+02, 0.82643967E+01, -0.12470506E+02, 0.23383005E+02, -0.65144601E+01, 0.16238860E+01, 0.87209654E+01, 0.49000898E+00, -0.16810759E+01, 0.40386558E+01, -0.54124254E+00, 0.10109010E+00, 0.19018241E+01, 
    },
    {
      -0.21727472E+02, -0.96902895E+01, -0.16478348E+02, 0.13098322E+02, 0.13914478E+02, -0.20104990E+01, 0.37096817E+01, -0.27436793E+01, -0.43885440E-01, -0.13556918E+01, 0.84448802E+00, -0.14343554E+01, -0.20107965E+01, 
    },
    {
      -0.89594114E+00, 0.10083544E+01, 0.83134689E+01, -0.92639027E+01, 0.61116376E+01, 0.10380750E+02, -0.26651862E+01, 0.87228279E+01, 0.38335390E+01, 0.22555156E+01, -0.13236306E+01, -0.31068852E+01, -0.87246418E+00, 
    },
    {
      0.23570587E+02, -0.42556520E+01, 0.28354004E+02, 0.33738234E+01, -0.40777335E+01, -0.38387177E+01, 0.82230272E+01, 0.30383515E+01, -0.22789021E+01, 0.26750424E+01, -0.16830761E+01, 0.11110405E+01, 0.25620415E+01, 
    },
    {
      0.29727594E+02, -0.24605175E+02, 0.28340710E+02, -0.59987556E+02, -0.22684792E+00, -0.73963881E+01, -0.18819962E+02, -0.99585795E+00, 0.24640315E+01, -0.93489466E+01, 0.17905924E+01, 0.29973716E+00, -0.48281670E+01, 
    },
    {
      0.32501541E+02, 0.18787926E+02, 0.36723728E+02, -0.15535311E+02, -0.37230057E+02, 0.20521283E+01, -0.89159498E+01, 0.26074269E+01, -0.13309879E+01, 0.20207708E+01, -0.26565912E+01, 0.20663900E+01, 0.51164727E+01, 
    },
    {
      -0.40171739E+00, 0.11958817E+01, -0.70398526E+01, 0.46672320E+01, -0.12406041E+02, -0.77181621E+01, 0.25220163E+01, -0.60711088E+01, -0.32043297E+01, -0.13135966E+01, 0.27531669E+00, 0.25722663E+01, 0.94680876E+00, 
    },
    {
      -0.20347054E+02, 0.94924231E+01, -0.21017023E+02, 0.51578898E+01, 0.91074489E-01, 0.31714413E+01, -0.65355906E+01, -0.32284212E+01, 0.13073928E+01, -0.20606008E+01, 0.10267882E+01, -0.13762021E+00, -0.18398629E+01, 
    },
    {
      -0.18292107E+02, 0.18030376E+02, -0.15053249E+02, 0.42048473E+02, 0.93764210E+01, 0.66455450E+01, 0.11065763E+02, 0.21237047E+00, -0.12657822E+01, 0.58830400E+01, -0.11597257E+01, -0.38042429E+00, 0.34563775E+01, 
    },
    {
      -0.12867409E+02, -0.16812653E+02, -0.25532946E+02, 0.17166659E+01, 0.26918324E+02, -0.10111470E+00, 0.59210205E+01, 0.11573966E+00, 0.15903757E+01, -0.98362416E+00, 0.20214050E+01, -0.12598089E+01, -0.36495934E+01, 
    },
    {
      0.28310280E-01, 0.48973437E-01, -0.48619840E-01, -0.72511427E-01, 0.58058472E-02, -0.57069111E+00, -0.71659225E+00, -0.18692106E-02, 0.42772554E-01, -0.31100757E-01, 0.13589856E-01, -0.41641410E-01, -0.74718855E-02, 
    },
    {
      -0.28761316E-01, -0.29289359E+00, -0.91049284E-01, 0.28099608E-01, -0.29132949E-01, 0.69334489E+00, -0.55960542E+00, -0.21222344E-01, 0.24635848E-01, 0.21211600E-01, -0.15531605E-02, 0.26932620E-02, 0.16090758E-01, 
    },
    {
      0.20954400E+00, 0.73779237E+00, -0.42621502E+00, -0.23507123E+00, -0.37294054E+00, 0.17519316E+00, 0.57431810E-01, -0.47046978E-01, 0.15076196E-01, -0.54287319E-02, 0.58389850E-01, -0.86825825E-01, -0.17713912E-01, 
    },
    {
      -0.23256958E+00, -0.20628646E+00, 0.74116617E-01, 0.43207306E-01, -0.11664169E+00, -0.47020305E-01, 0.75928085E-01, -0.98211251E-01, 0.93043060E-03, 0.55244029E-01, -0.46837240E-01, -0.95012728E-02, -0.13064935E-01, 
    },
    {
      0.87182951E+00, 0.98459274E-02, 0.13477666E+01, 0.84805533E-01, -0.90552852E-01, 0.11711707E+01, 0.11748557E+01, -0.12035688E+00, 0.12323441E+00, 0.49199241E+00, -0.20814496E+00, 0.27044243E+00, -0.61024815E-01, 
    },
    {
      -0.74558862E-01, 0.37783556E+01, -0.26144087E+00, 0.38623143E-01, 0.53248072E+00, -0.10437831E+01, 0.90481389E+00, -0.44594201E+00, -0.40208280E+00, -0.60124949E-01, -0.48902173E-01, -0.11441462E+00, -0.19880749E+00, 
    },
    {
      -0.13543634E+01, -0.10415508E+01, 0.12016990E+01, 0.31708425E+00, 0.61210239E+00, 0.58425140E+00, 0.41064870E+00, 0.33918127E+00, 0.11345342E-01, 0.93263805E-01, -0.17820968E+00, 0.18072975E+00, 0.96731916E-01, 
    },
    {
      0.32778770E+00, 0.61795223E+00, 0.38936533E-01, -0.50377148E+00, 0.31532276E+00, -0.38423333E+00, 0.74498868E+00, 0.17342281E+00, 0.20069090E+00, -0.13456018E+00, 0.16104504E+00, 0.72342478E-01, 0.79420209E-01, 
    },
    {
      -0.31249234E+00, 0.92828923E+00, -0.26748803E+01, 0.70246682E-01, 0.58934122E+00, -0.16105400E+01, -0.14977306E+01, -0.18931707E-01, -0.31899455E+00, -0.82864511E+00, 0.36326492E+00, -0.42674014E+00, -0.46573725E-01, 
    },
    {
      0.81409059E-01, -0.52734656E+01, 0.76651889E+00, -0.32308713E+00, -0.10347863E+01, 0.13800038E+01, -0.11256927E+01, 0.89915168E+00, 0.71173912E+00, 0.48457179E-01, 0.25199077E+00, 0.14797410E+00, 0.20402835E+00, 
    },
    {
      0.24246454E-01, 0.33298835E-01, -0.51717691E-01, -0.14058649E-01, 0.17725911E-01, -0.72037068E-03, -0.37894547E-02, 0.30802149E+00, -0.16611566E+00, -0.17199373E-01, 0.13114575E-01, -0.76011149E-02, -0.20986781E-01, 
    },
    {
      0.43441545E-01, 0.33683844E-01, -0.34691116E-02, 0.59443168E-01, -0.51950090E-01, 0.14247886E-01, -0.17108589E-01, 0.14792427E+00, 0.28507856E+00, 0.35790657E-02, 0.32111532E-02, 0.13911074E-02, -0.73967529E-02, 
    },
    {
      -0.19056965E+00, -0.73565212E-02, -0.24821372E+00, 0.15964326E-01, -0.97779892E-01, -0.89113951E-01, 0.58998290E-01, 0.14977667E+00, 0.37016949E+00, -0.46572641E-01, -0.31782843E-01, -0.32574493E-02, 0.12899466E-01, 
    },
    {
      0.37356037E+00, 0.15084852E+00, -0.12477553E+00, 0.29166037E-01, 0.21163251E+00, -0.11922070E+00, -0.11764953E+00, -0.38922527E+00, 0.26210438E-01, 0.71545686E-02, -0.61420333E-01, -0.20083405E-01, -0.11609336E-01, 
    },
    {
      -0.15726325E+00, 0.20085622E-01, -0.33992976E-01, -0.16686320E-01, -0.15578196E-01, -0.89996764E-02, 0.52281570E-01, 0.24140203E-01, -0.13938872E-01, 0.15512536E+00, 0.19358249E+00, 0.49229502E-02, -0.39371615E-02, 
    },
    {
      0.14834774E+00, 0.25739465E-01, -0.83402745E-01, 0.18065708E-01, -0.55616297E-01, -0.52165301E-02, -0.17087549E-01, 0.21788107E-01, 0.38194988E-01, -0.19343592E+00, 0.16540502E+00, -0.46783774E-02, -0.27857188E-01, 
    },
    {
      0.57891991E-01, -0.67955382E-01, 0.42646252E-01, 0.22683810E-01, 0.79109436E-02, 0.28160231E-01, -0.27141111E-01, 0.41366834E-01, -0.74150283E-02, 0.70931404E-02, -0.23410007E-01, -0.12303041E+00, 0.72064400E-01, 
    },
    {
      0.10394365E+00, 0.57514329E-01, -0.21865284E-01, 0.21552900E-02, 0.56937981E-01, -0.20307319E-02, 0.91909728E-03, -0.14761715E-01, 0.11957444E-01, -0.14504174E-01, -0.67575327E-02, -0.97156227E-01, -0.12803575E+00, 
    },
    {
      -0.39362207E-01, -0.32046013E-01, -0.55135102E-02, 0.48773892E-01, 0.35422139E-01, -0.15434346E-01, -0.21533076E-01, 0.26582329E-01, -0.12274754E-01, -0.12937960E-01, -0.11175702E-01, 0.53416751E-02, 0.67404481E-02, 
    },
    {
      0.73514357E-01, -0.38842116E-01, 0.64909115E-01, 0.56321872E-02, -0.52885335E-01, 0.41649371E-01, -0.11770996E-02, -0.26402390E-01, 0.25709886E-01, -0.13926198E-02, 0.27511716E-01, 0.84407153E-02, 0.18437810E-01, 
    },
    {
      0.43839946E-01, 0.51004354E-01, -0.11362996E+00, -0.25646824E-01, -0.38826078E-01, -0.55158515E-02, 0.62045278E-02, -0.30429460E-01, 0.18208286E-01, -0.11582433E-01, 0.17949345E-01, -0.15534251E-01, 0.51459920E-03, 
    },
    {
      -0.43065980E-01, 0.55039483E-02, 0.27189227E-02, 0.13957473E-01, 0.87091923E-01, 0.10603105E-01, 0.77987841E-03, -0.24959501E-02, 0.29643327E-01, 0.11370151E-01, -0.18617267E-01, -0.15555475E-01, 0.72064400E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir18_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30766914E+01, -0.92550134E-02, 0.13718505E-01, 0.25895583E-02, 0.33574869E-03, 0.78745289E-02, 0.86438525E-02, -0.60896222E-02, -0.74246735E-03, 
    },
    {
      -0.50321376E+00, 0.22363435E-01, 0.60551178E-01, 0.54579623E-01, -0.33105440E-01, -0.27306173E-01, -0.16869433E-01, -0.11755095E-01, 0.34014808E-02, 
    },
    {
      0.10928845E+01, 0.24034415E+00, -0.11382686E+00, -0.20328368E+00, -0.24367245E-01, -0.92177808E-01, -0.59212130E-01, 0.92386231E-01, 0.51138084E-02, 
    },
    {
      0.11338211E+01, -0.36324242E+00, -0.20184082E+00, -0.19428015E+00, 0.12432440E+00, 0.92264690E-01, 0.44125546E-01, 0.36674652E-01, 0.35087094E-02, 
    },
    {
      -0.19719753E+01, -0.64643306E+00, 0.26046559E+00, 0.49098560E+00, 0.92803150E-01, 0.28286198E+00, 0.15693869E+00, -0.19568925E+00, -0.20824329E-02, 
    },
    {
      -0.74346578E+00, 0.34690487E+00, 0.19643439E+00, 0.14222169E+00, -0.96713386E-01, -0.67438833E-01, -0.23977172E-01, -0.20441962E-01, -0.10335058E-01, 
    },
    {
      0.92457843E+00, 0.46374896E+00, -0.11210353E+00, -0.29201853E+00, -0.79533793E-01, -0.21276265E+00, -0.10755368E+00, 0.10385718E+00, -0.11526103E-02, 
    },
    {
      0.57956658E-03, -0.12857877E+00, -0.37720919E+00, 0.71668392E-02, -0.15985873E-01, 0.35551095E-02, -0.34057416E-03, 0.14429940E-04, 0.77074156E-02, 
    },
    {
      0.16609097E-01, 0.39636013E+00, -0.11226575E+00, -0.17938472E-01, 0.14924302E-01, -0.20589654E-02, -0.79512065E-02, 0.89895235E-04, 0.10084794E-01, 
    },
    {
      -0.84668107E-01, 0.76170884E-01, 0.11493012E+00, 0.58139663E-01, -0.90520024E-01, -0.49981788E-01, -0.52669294E-01, 0.67823958E-02, 0.27890796E-01, 
    },
    {
      0.11318666E+00, -0.73709123E-01, 0.20983368E-01, -0.12963885E-01, 0.24992188E-01, -0.32588847E-02, 0.19097012E-02, 0.37765738E-01, 0.79129241E-01, 
    },
    {
      0.79023778E-01, 0.42699623E+00, 0.21301763E+01, -0.24369007E+00, -0.14716794E+00, -0.96097685E-01, -0.25467388E-01, 0.65074660E-01, -0.70087478E-01, 
    },
    {
      -0.67493325E+00, -0.20482519E+01, 0.44933867E+00, 0.34022969E+00, -0.98558955E-01, 0.39792750E-01, 0.53783178E-01, -0.29130100E-03, -0.84323861E-01, 
    },
    {
      0.10565306E+01, -0.90792572E+00, -0.15945410E+01, 0.52398443E-01, 0.60421151E+00, 0.41048077E+00, -0.23732608E-01, 0.20138571E+00, -0.27786821E+00, 
    },
    {
      -0.57795197E-01, 0.20736563E+01, -0.39351213E+00, 0.12668329E+00, 0.84951319E-01, 0.28178258E-01, -0.23711748E-01, -0.16345520E+00, -0.57944131E+00, 
    },
    {
      -0.20411162E+00, -0.13376068E+00, -0.29449091E+01, 0.41258937E+00, 0.56290030E+00, 0.31165561E+00, 0.20544067E+00, -0.21796127E+00, 0.47752853E-01, 
    },
    {
      0.20179317E+01, 0.24129705E+01, -0.70617986E+00, -0.83160406E+00, 0.22868297E+00, -0.11227272E+00, -0.42454861E-01, -0.91994748E-01, 0.24011090E+00, 
    },
    {
      -0.28024802E+01, 0.13774416E+01, 0.36972830E+01, -0.59003949E+00, -0.13606600E+01, -0.84793615E+00, 0.57913697E+00, -0.74747223E+00, 0.55327827E+00, 
    },
    {
      -0.22295168E+00, -0.52054830E+01, 0.60161120E+00, -0.30097663E+00, -0.41190431E+00, -0.10745584E+00, -0.66798493E-01, 0.33825681E+00, 0.10680364E+01, 
    },
    {
      0.27225360E+00, -0.13481601E+00, 0.12781663E+01, -0.17782266E+00, -0.43654701E+00, -0.28031033E+00, -0.21654996E+00, 0.18418586E+00, 0.50952464E-01, 
    },
    {
      -0.13763983E+01, -0.62647796E+00, 0.42017722E+00, 0.53561598E+00, -0.22507687E+00, 0.83096340E-01, -0.32817584E-01, 0.12693739E+00, -0.21098842E+00, 
    },
    {
      0.21004002E+01, -0.19308670E+00, -0.28420036E+01, 0.45582688E+00, 0.97629905E+00, 0.47537673E+00, -0.57095116E+00, 0.58855778E+00, -0.29641420E+00, 
    },
    {
      0.67884624E-01, 0.40048332E+01, -0.36370404E-01, 0.15666211E+00, 0.35521010E+00, 0.11299833E+00, 0.86527780E-01, -0.22095665E+00, -0.55126411E+00, 
    },
    {
      -0.88387989E-02, -0.10312352E-01, -0.61113387E-02, -0.20343257E-01, -0.15003505E+00, -0.10647348E-01, -0.78879838E-03, 0.66963251E-03, -0.76038600E-02, 
    },
    {
      0.41395663E-02, -0.34112083E-02, 0.10996495E-01, 0.17316753E+00, -0.10291023E-01, -0.73754769E-02, -0.14292508E-01, -0.21934167E-02, 0.18509438E-02, 
    },
    {
      0.63759461E-01, 0.23891982E+00, 0.63553944E-01, 0.18635046E+00, -0.57529569E+00, 0.23344524E-01, -0.40877607E-01, 0.17844956E-01, -0.39308965E-01, 
    },
    {
      -0.51171139E-01, 0.10197276E+00, 0.71252682E-02, 0.72727478E+00, 0.15855432E+00, 0.49558837E-01, -0.45434017E-01, -0.18780010E-01, 0.50435401E-02, 
    },
    {
      0.15530051E-01, 0.14170410E+00, -0.61096840E-01, -0.37090287E-01, -0.10134680E+00, 0.43753270E-01, 0.16147131E-01, 0.11646984E-01, 0.15542092E-01, 
    },
    {
      -0.36590850E+00, -0.32496876E-02, -0.23358011E+00, -0.45408837E-01, -0.12131418E+00, 0.92704944E-01, 0.10798852E+00, -0.10122197E-01, 0.33092615E-03, 
    },
    {
      -0.40528265E+00, -0.12369304E+01, -0.24165794E-01, -0.74351096E+00, 0.17940987E+01, -0.18066379E+00, 0.10244836E+00, -0.25582047E-01, 0.19211514E+00, 
    },
    {
      0.76301038E+00, -0.20792608E+00, 0.27201289E+00, -0.23394880E+01, -0.62358546E+00, -0.24235266E+00, 0.63761316E-01, 0.15015985E+00, -0.11706612E+00, 
    },
    {
      -0.10061766E+00, -0.37011167E+00, 0.14104469E+00, -0.26392074E-01, 0.25978979E+00, 0.53946935E-02, 0.15027906E-01, -0.45017198E-01, 0.68766944E-03, 
    },
    {
      0.53705126E+00, -0.14921010E+00, 0.35961419E+00, -0.32242209E-01, 0.42051267E-01, -0.12954605E+00, -0.89045294E-01, 0.19702064E-04, -0.23211284E-01, 
    },
    {
      0.49777216E+00, 0.15834332E+01, 0.39794613E-01, 0.63238031E+00, -0.15466079E+01, 0.28480631E+00, -0.73767789E-01, 0.22335300E-01, -0.21944922E+00, 
    },
    {
      -0.12544603E+01, 0.71595207E-01, -0.24489100E+00, 0.20173306E+01, 0.51294672E+00, 0.28983241E+00, 0.99735051E-01, -0.18368557E+00, 0.18341650E+00, 
    },
    {
      -0.42410571E-01, -0.52171294E-02, 0.19085895E-02, -0.95514078E-02, 0.75655179E-02, 0.16171318E-01, 0.16939323E-01, 0.22008182E-02, -0.19602047E-02, 
    },
    {
      -0.60261209E-02, -0.52024610E-01, -0.89434842E-02, 0.10327846E-01, -0.25033876E-01, -0.14511695E-01, 0.23334123E-01, -0.61932360E-02, 0.28024365E-02, 
    },
    {
      0.82942434E-01, 0.45942049E-01, 0.93686730E-02, -0.41635599E-01, 0.30467382E-01, 0.41355919E-01, 0.26757890E-01, 0.10544643E-01, 0.57960446E-02, 
    },
    {
      0.58556028E-01, -0.45138933E-02, -0.20666830E-01, -0.21553095E-01, 0.11791925E-01, -0.26258841E-01, 0.53928331E-01, -0.11998913E-01, 0.21382065E-01, 
    },
    {
      0.77182382E-01, -0.48827209E-01, -0.91188706E-01, 0.14237206E-01, -0.10213549E-01, -0.94213583E-01, -0.15220276E+00, -0.12513259E-02, 0.21464357E-01, 
    },
    {
      -0.15465503E-01, 0.13954592E+00, -0.39753947E-01, -0.45017153E-01, 0.57116088E-01, 0.15541755E+00, -0.12666646E+00, 0.11608529E-01, 0.80760103E-02, 
    },
    {
      -0.41384086E-01, -0.10227784E-01, 0.80588767E-02, -0.46442337E-02, -0.21833975E-02, 0.53496994E-02, -0.68548233E-02, 0.41308586E-01, 0.59219892E-02, 
    },
    {
      -0.27170403E-01, 0.62266630E-02, -0.46112244E-02, -0.12002573E-01, 0.16872867E-02, 0.85898452E-02, 0.16054403E-01, -0.50134775E-02, 0.41393556E-01, 
    },
    {
      0.69590531E-01, -0.37438667E-02, 0.20416295E-01, -0.48384883E-01, -0.18482743E-01, -0.26152568E-03, 0.12390666E-01, 0.16554868E-01, 0.15767444E-01, 
    },
    {
      0.27102111E-01, -0.50822388E-01, 0.19041890E-01, 0.45237873E-01, -0.61440300E-02, -0.86206961E-02, -0.41923290E-02, -0.37406992E-01, 0.99841179E-02, 
    },
    {
      -0.49510780E-02, -0.68882159E-02, -0.39861626E-02, -0.19369667E-02, -0.11413929E-02, 0.49859667E-02, -0.92119956E-02, 0.47657973E-03, 0.34508729E-03, 
    },
    {
      0.18523857E-01, 0.17887400E-01, -0.11042763E-01, -0.10345280E-01, -0.11657286E-01, 0.51686917E-02, 0.27109310E-02, 0.35306651E-02, 0.11049252E-02, 
    },
    {
      0.36995022E-02, 0.33790923E-02, 0.36772604E-02, -0.24134079E-02, -0.38045826E-02, 0.39723795E-02, 0.10341560E-02, 0.54815668E-02, -0.16316446E-02, 
    },
    {
      0.22392916E-01, 0.57393694E-02, -0.18493544E-01, 0.23872721E-03, 0.48304014E-02, -0.32244358E-03, 0.17225273E-02, -0.32498674E-02, -0.64918457E-03, 
    },
  },
  {
    {
      0.27327437E+01, -0.15785564E-01, 0.97018043E-02, 0.48172278E-02, -0.28546432E-02, 0.45274911E-02, 0.34621279E-02, -0.15102919E-02, -0.27329123E-02, 
    },
    {
      -0.25903657E+00, 0.55780217E-01, 0.65728329E-01, 0.32537010E-01, -0.25246153E-01, -0.58772466E-02, -0.21703182E-01, -0.43969019E-02, -0.69272384E-03, 
    },
    {
      0.11714929E+01, 0.16338734E+00, -0.89373522E-01, -0.10331998E+00, 0.13328508E-01, -0.56765247E-01, -0.25298420E-01, 0.44980552E-01, 0.17588584E-01, 
    },
    {
      0.33108100E+00, -0.49494028E+00, -0.18589833E+00, -0.10610254E+00, 0.78505076E-01, -0.88712927E-02, 0.33990052E-01, 0.10129429E-01, 0.43824976E-02, 
    },
    {
      -0.17302923E+01, -0.27473363E+00, 0.17417511E+00, 0.22498931E+00, -0.11839691E-01, 0.17832267E+00, 0.89640006E-01, -0.92783161E-01, -0.12903627E-01, 
    },
    {
      -0.15330756E+00, 0.43055135E+00, 0.16011506E+00, 0.73783807E-01, -0.51124744E-01, 0.16581139E-01, -0.11291335E-01, -0.22684636E-02, -0.43424224E-02, 
    },
    {
      0.63714796E+00, 0.18350784E+00, -0.42438373E-01, -0.12726226E+00, -0.10468658E-01, -0.13165224E+00, -0.66993155E-01, 0.44999924E-01, -0.64978427E-02, 
    },
    {
      0.79358937E-02, -0.29707021E+00, -0.28992331E+00, 0.80166208E-02, -0.15802497E-01, -0.19063946E-02, 0.49845497E-02, 0.27219285E-03, 0.67873574E-02, 
    },
    {
      0.62516500E-03, 0.32199055E+00, -0.27686164E+00, -0.15336918E-01, 0.17286168E-01, 0.12508899E-02, -0.14910120E-01, 0.12231509E-02, 0.14304347E-01, 
    },
    {
      0.16834082E-01, 0.10278944E+00, -0.63203834E-01, 0.10091765E-01, -0.13688602E-01, -0.95551871E-02, -0.36618944E-01, -0.14316996E-01, 0.35773881E-01, 
    },
    {
      0.76245844E-01, 0.77808850E-01, -0.12024795E-01, 0.12149659E-01, 0.17534763E-01, 0.12295829E-02, -0.60971705E-02, 0.14225282E-01, 0.82393371E-01, 
    },
    {
      -0.15534867E+00, 0.80440164E+00, 0.12503365E+01, -0.12217534E+00, -0.85220747E-01, -0.29276054E-01, -0.48868369E-01, 0.28717835E-01, -0.10331116E+00, 
    },
    {
      -0.52347887E+00, -0.13314192E+01, 0.89782566E+00, 0.28488895E+00, -0.12012097E+00, -0.21137984E-01, 0.12435994E+00, -0.14031794E-01, -0.16417257E+00, 
    },
    {
      0.76694377E-01, -0.93182540E+00, 0.25457191E+00, 0.19702044E+00, 0.48383981E-01, 0.41341368E-01, -0.14367604E-01, 0.36359438E+00, -0.25973442E+00, 
    },
    {
      0.16930665E+00, 0.43561876E+00, 0.67502417E-01, -0.75919375E-01, -0.20566436E-01, 0.45070466E-01, 0.74356943E-01, -0.47366299E-01, -0.56628042E+00, 
    },
    {
      0.58145106E+00, -0.34549749E+00, -0.13091027E+01, 0.17559958E+00, 0.43319133E+00, 0.16244146E+00, 0.20602815E+00, -0.41776557E-01, 0.30710989E+00, 
    },
    {
      0.17932121E+01, 0.97565883E+00, -0.11525728E+01, -0.74104398E+00, 0.24997143E+00, 0.52191265E-01, -0.24265365E+00, -0.61648171E-01, 0.39528716E+00, 
    },
    {
      -0.39080113E+00, 0.15962688E+01, -0.17618662E+00, -0.74545747E+00, -0.20102003E+00, -0.15913907E-01, 0.36910924E+00, -0.10008817E+01, 0.41477209E+00, 
    },
    {
      -0.10000013E+01, -0.15626558E+01, -0.33813745E+00, 0.20472082E+00, 0.47138073E-01, -0.14620386E+00, -0.33509147E+00, 0.74930072E-01, 0.10780282E+01, 
    },
    {
      -0.40786570E+00, 0.21695097E-02, 0.61571985E+00, -0.87222412E-01, -0.40252182E+00, -0.18569061E+00, -0.16891396E+00, 0.19968783E-01, -0.22394384E+00, 
    },
    {
      -0.13754953E+01, -0.63793459E-02, 0.60612941E+00, 0.53394854E+00, -0.20959119E+00, -0.59920955E-01, 0.10459418E+00, 0.97062908E-01, -0.27976444E+00, 
    },
    {
      0.39142886E+00, -0.65925485E+00, -0.55752856E+00, 0.61146140E+00, 0.21277002E+00, -0.66732608E-01, -0.38353363E+00, 0.67082864E+00, -0.16077600E+00, 
    },
    {
      0.87837434E+00, 0.18233119E+01, 0.43625033E+00, -0.18887997E+00, -0.19754047E-01, 0.14874180E+00, 0.31835011E+00, -0.25374444E-01, -0.59807515E+00, 
    },
    {
      -0.35858257E-02, -0.18426422E-01, 0.11940397E-01, 0.35249278E-01, -0.11406703E+00, -0.12094727E-01, 0.12369556E-02, 0.28486021E-02, -0.41241115E-02, 
    },
    {
      -0.44571543E-02, -0.12385954E-01, 0.10621492E-01, 0.11439821E+00, 0.52063718E-01, -0.31545039E-02, -0.91192275E-02, -0.23166044E-02, 0.18062951E-02, 
    },
    {
      0.77467859E-01, 0.19298270E+00, 0.57610362E-02, 0.17876288E+00, -0.28498143E+00, 0.33583455E-01, -0.21250669E-02, 0.40939944E-02, -0.17955484E-01, 
    },
    {
      0.18605869E-01, 0.88920861E-01, -0.16398907E-01, 0.37881327E+00, 0.19283801E+00, 0.35964742E-01, -0.57503371E-03, -0.16725548E-01, 0.14335946E-01, 
    },
    {
      -0.39326932E-01, 0.15444148E+00, -0.13056353E+00, -0.27956468E+00, 0.53259443E-01, 0.70716918E-01, 0.21401586E-01, -0.33642214E-01, 0.21223655E-01, 
    },
    {
      -0.20482560E+00, 0.36653712E-01, -0.11544807E+00, -0.96398272E-01, -0.40236861E+00, 0.14114498E-01, 0.74856617E-01, 0.13251967E-01, 0.58944183E-02, 
    },
    {
      -0.22919865E+00, -0.93992049E+00, 0.22374582E+00, -0.53515422E+00, 0.77677137E+00, -0.16614506E+00, -0.86691096E-01, 0.29328503E-01, 0.12252356E+00, 
    },
    {
      0.30133858E+00, -0.19458076E+00, 0.14997202E+00, -0.10709431E+01, -0.48711553E+00, -0.11093743E+00, -0.55874560E-01, 0.11211873E+00, -0.14200687E+00, 
    },
    {
      0.17139059E-01, -0.40209791E+00, 0.19722582E+00, 0.75891189E-01, -0.96643463E-01, -0.82315318E-01, -0.36470420E-02, 0.47883753E-01, -0.17165303E-01, 
    },
    {
      0.32197592E+00, -0.29618844E+00, 0.99401399E-01, 0.21706817E+00, 0.22001296E+00, -0.53230442E-01, -0.68153217E-01, -0.20923691E-01, -0.37170000E-01, 
    },
    {
      0.10419927E+00, 0.12077615E+01, -0.25412044E+00, 0.44525284E+00, -0.71311611E+00, 0.24982084E+00, 0.11983401E+00, -0.38160987E-01, -0.16348417E+00, 
    },
    {
      -0.60997337E+00, 0.27142987E+00, -0.35729647E-01, 0.90510237E+00, 0.34794793E+00, 0.13886355E+00, 0.16090356E+00, -0.10832496E+00, 0.19660074E+00, 
    },
    {
      -0.28633850E-01, -0.51487325E-03, 0.12486504E-01, -0.59260265E-02, -0.30928615E-02, 0.16670747E-01, -0.18133426E-01, 0.13137640E-02, -0.29191643E-02, 
    },
    {
      -0.73401621E-02, -0.43932945E-01, -0.60264287E-02, 0.11825816E-01, -0.20122048E-01, 0.13404754E-01, 0.25719734E-01, -0.27513448E-02, -0.39640418E-02, 
    },
    {
      0.54108258E-01, 0.62069230E-01, 0.32028572E-02, -0.30705713E-01, 0.21409197E-01, 0.68470776E-01, 0.18883282E-01, 0.12196345E-02, -0.59491460E-03, 
    },
    {
      0.44463780E-01, 0.15322219E-02, -0.11973175E-01, -0.23743760E-01, 0.23724630E-02, -0.18886378E-01, 0.65881431E-01, -0.61001866E-02, 0.12625720E-01, 
    },
    {
      0.42619128E-01, -0.87218329E-01, -0.71985729E-01, 0.61123907E-02, -0.80230413E-02, -0.94753467E-01, -0.10718403E+00, 0.44616999E-03, 0.18477114E-01, 
    },
    {
      -0.49196170E-02, 0.10727092E+00, -0.36542352E-01, -0.19530224E-01, 0.29819097E-01, 0.11121344E+00, -0.11566354E+00, -0.12570542E-02, 0.71676592E-02, 
    },
    {
      -0.28708642E-01, -0.12917655E-01, 0.10160701E-01, -0.15056655E-02, -0.51682289E-02, 0.13105695E-02, -0.69624763E-02, 0.31344198E-01, 0.13837171E-01, 
    },
    {
      -0.15254789E-01, 0.17759033E-02, -0.10812398E-01, -0.43772645E-02, -0.18710297E-02, 0.43232990E-02, 0.13123615E-01, -0.15094435E-01, 0.28170321E-01, 
    },
    {
      0.47917835E-01, 0.23561502E-02, 0.78526502E-02, -0.31751979E-01, -0.63430383E-02, -0.38912459E-02, 0.14013788E-01, 0.21223193E-01, 0.36825228E-01, 
    },
    {
      0.24586657E-01, -0.54429136E-01, 0.31562790E-01, 0.22680206E-01, 0.70566137E-03, -0.58405814E-02, -0.98891770E-02, -0.44559006E-01, 0.10059295E-01, 
    },
    {
      0.96068269E-03, -0.11706586E-01, 0.22053521E-02, 0.32256660E-02, -0.55763084E-02, 0.31189667E-02, -0.83871549E-02, 0.37916502E-03, 0.65405335E-03, 
    },
    {
      0.91339229E-02, 0.15419393E-01, -0.98529020E-02, -0.21402466E-02, -0.34411531E-02, 0.44119060E-02, 0.64586932E-02, 0.16170664E-02, 0.65145995E-02, 
    },
    {
      0.20680716E-02, 0.92757810E-02, 0.41878442E-02, 0.99740468E-03, -0.10964575E-01, 0.11091142E-03, 0.14305855E-02, 0.14857621E-02, 0.76322776E-03, 
    },
    {
      0.71551851E-02, -0.16748263E-02, -0.11113628E-01, 0.20171816E-02, 0.70238686E-02, 0.14452589E-02, 0.47303335E-03, -0.38445606E-02, 0.29371867E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir19_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.60445695E+01, -0.13494897E+00, 0.96523471E-01, 0.30800914E-01, -0.25691688E-02, -0.24381364E-01, -0.20869752E-01, 0.36958508E-01, -0.24559852E-01, 0.99607911E-02, 0.18155642E-02, 0.16870741E-04, -0.14172103E-01, 
    },
    {
      -0.72683138E+00, -0.18670639E+00, 0.51046336E+00, -0.36924446E+00, -0.19951835E+00, -0.26578641E+00, 0.36712724E+00, -0.13204169E+00, -0.27474661E-02, 0.22398602E-01, -0.15810436E+00, 0.48819307E-01, 0.84509522E-01, 
    },
    {
      0.13992862E+02, 0.23779745E+01, -0.22226188E+01, 0.20291131E+01, 0.90447301E+00, 0.73813629E+00, 0.10136298E+01, -0.49498451E+00, 0.72660756E+00, -0.77001888E+00, -0.32229844E+00, 0.19383496E+00, 0.34973943E+00, 
    },
    {
      0.33502454E+00, -0.41584191E+01, -0.30603714E+01, 0.73569479E+01, 0.64584503E+01, 0.61389766E+01, -0.67864485E+01, 0.15375274E+01, 0.11104467E+01, 0.45180345E+00, 0.31834459E+01, -0.39800099E+00, -0.16252377E+01, 
    },
    {
      -0.98461967E+02, -0.11804735E+02, 0.13932345E+02, -0.21968882E+02, -0.18346836E+02, -0.64648104E+01, -0.44017797E+01, 0.21018789E+01, -0.50670667E+01, 0.55518346E+01, 0.14706390E+01, -0.12416881E+01, -0.32495730E+01, 
    },
    {
      0.24509441E+02, 0.18994970E+02, -0.48482704E+01, -0.31010715E+02, -0.24011599E+02, -0.33442169E+02, 0.34668297E+02, -0.85968723E+01, -0.92274466E+01, -0.22783225E+01, -0.16358894E+02, 0.99501222E+00, 0.78107009E+01, 
    },
    {
      0.21546281E+03, 0.30437609E+02, -0.30536016E+02, 0.66628876E+02, 0.65871796E+02, 0.18404390E+02, 0.62645345E+01, -0.42978358E+01, 0.12689314E+02, -0.13794364E+02, -0.23129451E+01, 0.28023379E+01, 0.10037551E+02, 
    },
    {
      -0.63695538E+02, -0.22283279E+02, 0.36996841E+02, 0.50817825E+02, 0.28390305E+02, 0.72428009E+02, -0.71910110E+02, 0.21617928E+02, 0.24796774E+02, 0.37806880E+01, 0.33501648E+02, -0.13203067E+01, -0.16414623E+02, 
    },
    {
      -0.21614815E+03, -0.35832664E+02, 0.27470688E+02, -0.79447624E+02, -0.83696152E+02, -0.21164886E+02, -0.25409181E+01, 0.43153014E+01, -0.12989493E+02, 0.14197570E+02, 0.14302781E+01, -0.27501328E+01, -0.12051160E+02, 
    },
    {
      0.50732521E+02, 0.19075854E+01, -0.50831139E+02, -0.35323662E+02, -0.81260357E+01, -0.68726357E+02, 0.65579124E+02, -0.23836714E+02, -0.26761435E+02, -0.26312616E+01, -0.30050358E+02, 0.11140531E+01, 0.15877548E+02, 
    },
    {
      0.83067856E+02, 0.14992576E+02, -0.86386423E+01, 0.32848270E+02, 0.35260712E+02, 0.85520496E+01, -0.31619912E+00, -0.16816278E+01, 0.46327853E+01, -0.52116356E+01, -0.27778086E+00, 0.10079776E+01, 0.49246526E+01, 
    },
    {
      -0.10440203E+02, 0.58246608E+01, 0.21354177E+02, 0.85120945E+01, -0.24493570E+01, 0.23852114E+02, -0.21972420E+02, 0.94112186E+01, 0.10090175E+02, 0.65944129E+00, 0.98899641E+01, -0.44724047E+00, -0.57489376E+01, 
    },
    {
      -0.63381091E-01, 0.20475085E+01, 0.15625848E+01, -0.13729405E-01, -0.65938756E-01, -0.35532121E-01, -0.29464888E-02, 0.79189196E-01, 0.18604737E-02, 0.96583366E-02, 0.30324928E-01, 0.14240874E-02, -0.19405013E-01, 
    },
    {
      -0.23046065E-01, -0.14310176E+01, 0.19285048E+01, -0.16601916E-01, 0.18041897E+00, -0.12310948E-01, -0.66375285E-01, 0.31838231E-01, -0.65732598E-02, -0.13759874E-01, 0.18407125E-01, -0.24185097E-01, 0.10961994E-01, 
    },
    {
      -0.74787430E-01, 0.16353496E+01, -0.22924951E+00, 0.47504836E+00, -0.96047330E+00, -0.33705956E+00, 0.11720680E+00, -0.35202015E-01, 0.46376839E+00, -0.10265652E-01, -0.14193834E+00, 0.55663763E-02, 0.19628496E+00, 
    },
    {
      -0.12596104E+00, 0.14582266E+01, -0.98657086E-01, 0.94749078E-01, 0.78419155E+00, -0.29752413E-01, -0.19540650E+00, 0.36470249E+00, 0.27288207E+00, 0.13509390E+00, -0.56428380E-01, -0.11677991E+00, -0.53491343E-01, 
    },
    {
      0.19545885E+01, 0.11435436E+02, 0.84868221E+01, 0.32103235E+00, 0.12959203E+01, 0.88761157E+00, 0.49820134E+00, -0.11861235E+01, -0.71958087E-01, -0.25178078E+00, -0.12668631E+01, -0.35598647E-01, 0.32306072E+00, 
    },
    {
      -0.32294741E+01, -0.95716171E+01, 0.12515850E+02, 0.23365383E+01, 0.91575392E-01, -0.88255316E+00, 0.17265298E+01, -0.47084525E+00, 0.12570558E+01, 0.15405376E+00, 0.85922658E+00, 0.83317697E+00, -0.39139026E+00, 
    },
    {
      0.11067981E+01, -0.20162714E+02, 0.14454757E+02, -0.86500759E+01, 0.16342585E+02, 0.64312339E+01, -0.42658453E+01, -0.13463537E+00, -0.77128196E+01, 0.24014208E+01, 0.39491782E+01, 0.46762183E+00, -0.34915750E+01, 
    },
    {
      0.46240864E+01, -0.27222712E+02, 0.80155439E+01, -0.32276535E+01, -0.17326429E+02, 0.45502667E+01, 0.41100821E+01, -0.76522098E+01, -0.50376501E+01, -0.25007749E+01, 0.33137116E+00, 0.32376506E+01, 0.32104585E+00, 
    },
    {
      0.24788020E+01, -0.65718430E+02, 0.30561613E+02, 0.72700047E+01, -0.19604420E+02, -0.72012863E+01, -0.44401312E+01, 0.93486271E+01, 0.33760922E+01, 0.81467517E-01, 0.93344440E+01, 0.74155289E+00, -0.13300089E+01, 
    },
    {
      0.34563679E+02, -0.28905020E+02, -0.76245705E+02, -0.21742231E+02, -0.73547716E+01, 0.10644203E+02, -0.10774131E+02, 0.45778413E+01, -0.11885620E+02, -0.60007966E+00, -0.11258645E+02, -0.67847219E+01, 0.38254924E+01, 
    },
    {
      -0.19042896E+02, 0.11207795E+03, -0.28545294E+02, 0.46980576E+02, -0.61684097E+02, -0.24524796E+02, 0.27351971E+02, 0.67426047E+01, 0.44498322E+02, -0.20013243E+02, -0.25409010E+02, -0.27430830E+01, 0.19376114E+02, 
    },
    {
      -0.52723091E+02, 0.99575500E+02, -0.13353540E+02, 0.95136633E+01, 0.99035027E+02, -0.30965115E+02, -0.18851999E+02, 0.40043255E+02, 0.25284658E+02, 0.14075030E+02, 0.11309584E+01, -0.20960064E+02, -0.15822057E+01, 
    },
    {
      -0.48169975E+02, 0.94947174E+02, -0.23541681E+03, -0.43397125E+02, 0.64458717E+02, 0.22686687E+02, 0.10289151E+02, -0.30886675E+02, -0.14552165E+02, 0.30687971E+01, -0.24410389E+02, -0.24599547E+01, 0.13256648E+01, 
    },
    {
      -0.10717301E+03, 0.22662590E+03, 0.13886285E+03, 0.76127411E+02, 0.35750031E+02, -0.36212978E+02, 0.21812931E+02, -0.18747749E+02, 0.38481537E+02, 0.29832237E+01, 0.37119461E+02, 0.18805193E+02, -0.13755708E+02, 
    },
    {
      0.97456581E+02, -0.27367038E+03, -0.62144730E+02, -0.11783699E+03, 0.86211739E+02, 0.28845253E+02, -0.65351234E+02, -0.25876451E+02, -0.10882457E+03, 0.56715382E+02, 0.64410553E+02, 0.43075614E+01, -0.45408981E+02, 
    },
    {
      0.15016524E+03, -0.12772850E+03, -0.25741722E+02, -0.16592799E+01, -0.24547108E+03, 0.70262955E+02, 0.35663345E+02, -0.83696991E+02, -0.53332424E+02, -0.35324783E+02, -0.94713192E+01, 0.54111397E+02, 0.49722629E+01, 
    },
    {
      0.10506150E+03, -0.35246891E+02, 0.37882608E+03, 0.76265465E+02, -0.77831039E+02, -0.30161860E+02, -0.64913325E+01, 0.43002579E+02, 0.21760742E+02, -0.60918565E+01, 0.27057970E+02, 0.25830400E+01, 0.15233262E+01, 
    },
    {
      0.13615527E+03, -0.35940277E+03, -0.10190280E+03, -0.10915962E+03, -0.63814045E+02, 0.46641560E+02, -0.12899118E+02, 0.29911598E+02, -0.51161182E+02, -0.58345952E+01, -0.46785748E+02, -0.20363272E+02, 0.19093109E+02, 
    },
    {
      -0.17207220E+03, 0.30198877E+03, 0.17715979E+03, 0.13847585E+03, -0.45248798E+02, -0.24912093E+01, 0.66965683E+02, 0.35798885E+02, 0.11744838E+03, -0.66378052E+02, -0.71170624E+02, -0.11229161E+01, 0.46958759E+02, 
    },
    {
      -0.15008180E+03, 0.48165199E+02, 0.66421715E+02, -0.11201522E+02, 0.28107013E+03, -0.61901672E+02, -0.33674416E+02, 0.75491982E+02, 0.52029320E+02, 0.41378738E+02, 0.16554775E+02, -0.61346951E+02, -0.63890009E+01, 
    },
    {
      -0.64716988E+02, -0.60834203E+01, -0.18212560E+03, -0.42207047E+02, 0.31869562E+02, 0.14135931E+02, -0.47822112E+00, -0.21003960E+02, -0.10799576E+02, 0.31178300E+01, -0.10885387E+02, -0.69965732E+00, -0.20979679E+01, 
    },
    {
      -0.62125351E+02, 0.17012572E+03, 0.26723660E+02, 0.53963669E+02, 0.37830799E+02, -0.20305946E+02, -0.10621082E+01, -0.15930128E+02, 0.23917713E+02, 0.34242952E+01, 0.20282827E+02, 0.73727045E+01, -0.88406191E+01, 
    },
    {
      0.97107460E+02, -0.12479255E+03, -0.10509427E+03, -0.61543755E+02, 0.48847432E+01, -0.85133734E+01, -0.24883598E+02, -0.16887283E+02, -0.46089825E+02, 0.27499294E+02, 0.28551277E+02, -0.10849844E+01, -0.17723860E+02, 
    },
    {
      0.46780483E+02, 0.82450275E+01, -0.36148045E+02, 0.53822312E+01, -0.12206293E+03, 0.17168350E+02, 0.13732118E+02, -0.24196617E+02, -0.19565023E+02, -0.18214315E+02, -0.86782436E+01, 0.25485151E+02, 0.25449452E+01, 
    },
    {
      0.95903993E-01, 0.12766166E+00, -0.18699701E+00, -0.10404546E+01, -0.33024654E-01, 0.64295433E-01, -0.23013584E-01, -0.10285793E-01, 0.37326079E-01, 0.31895945E-02, 0.30733405E-01, -0.10633793E-01, 0.97853728E-02, 
    },
    {
      0.52426510E-01, -0.93061021E-02, -0.71888246E-01, 0.19813684E+00, -0.72505492E+00, -0.19345094E-02, -0.19068209E-01, -0.36769819E-01, 0.53982814E-02, -0.14521895E-01, 0.23320390E-01, -0.12640823E-02, 0.60315416E-02, 
    },
    {
      0.46948263E+00, 0.12817408E+01, 0.52438563E+00, 0.58714247E+00, -0.23273043E+00, 0.46726218E+00, -0.28142142E+00, 0.50233081E-01, -0.11847415E+00, -0.78053989E-01, -0.36326803E-01, -0.14682151E+00, 0.96851006E-01, 
    },
    {
      0.68315560E+00, 0.54635304E+00, 0.54994017E+00, 0.64071083E+00, 0.38787290E+00, 0.12569439E+00, 0.33838964E+00, -0.25519842E+00, -0.18448704E+00, -0.18360077E+00, -0.28392771E+00, 0.12267494E+00, 0.58491524E-01, 
    },
    {
      0.66295362E+00, -0.11465982E+01, 0.44839840E+01, 0.63690023E+01, -0.21015067E+01, -0.15754128E+01, 0.21561065E+00, 0.82991850E+00, -0.24949779E+00, 0.12380320E+00, -0.32161883E+00, -0.23786856E+00, -0.31155005E+00, 
    },
    {
      0.14305441E+01, 0.57962000E+00, 0.32569954E+01, -0.28792398E+01, 0.34138973E+01, -0.71248835E+00, -0.62034968E-01, 0.14145756E+01, 0.23729173E-01, -0.19871333E+00, -0.66435295E+00, 0.17309137E-01, 0.12985046E-02, 
    },
    {
      -0.47963367E+01, -0.89218206E+01, -0.36218772E+01, 0.90268250E+01, -0.87401462E+00, -0.29467752E+01, 0.10264589E+01, -0.11269215E+01, 0.13553836E+01, 0.10891480E+01, 0.17733830E+01, 0.15696553E+01, -0.47471687E+00, 
    },
    {
      -0.50660028E+01, -0.33092887E+01, -0.32716267E+01, -0.14229895E+01, 0.11620011E+02, 0.96331590E+00, -0.26582198E+01, 0.10442986E+01, 0.50223678E+00, 0.14703425E+01, 0.40092273E+01, -0.53163874E+00, -0.32783002E+00, 
    },
    {
      -0.91051664E+01, 0.17604088E+02, -0.22866295E+02, -0.28144238E+01, 0.20614243E+02, 0.94886522E+01, -0.77871466E+00, -0.35348945E+01, 0.17768615E+01, -0.16262704E+01, -0.12231375E+00, 0.21009269E+01, 0.29577000E+01, 
    },
    {
      -0.74458656E+01, 0.18080215E+01, -0.13265327E+02, 0.81773024E+01, 0.70696664E+01, 0.32722304E+01, 0.41111064E+01, -0.59407196E+01, -0.56068152E+00, 0.22052250E+01, 0.37547357E+01, -0.12558265E+01, 0.11403291E+01, 
    },
    {
      0.10656018E+02, 0.17880117E+02, 0.63293304E+01, -0.30708920E+02, 0.19644450E+01, 0.56080647E+01, -0.20938640E+01, 0.40487800E+01, -0.39456055E+01, -0.30156834E+01, -0.52191610E+01, -0.42830544E+01, 0.71889979E+00, 
    },
    {
      0.10659558E+02, 0.14922249E+01, 0.54132767E+01, 0.28934119E+01, -0.37208477E+02, -0.38359447E+01, 0.57908921E+01, -0.21831961E+00, 0.91132575E+00, -0.33404295E+01, -0.10999828E+02, 0.95123515E-01, 0.62245464E+00, 
    },
    {
      0.21184179E+02, -0.49009296E+02, 0.41135643E+02, -0.23144958E+02, -0.44450764E+02, -0.21969942E+02, 0.31099708E+01, 0.57433114E+01, -0.42183104E+01, 0.41270490E+01, 0.23304346E+01, -0.47546310E+01, -0.74363651E+01, 
    },
    {
      0.14705822E+02, -0.11686625E+02, 0.15904526E+02, -0.11272635E+02, -0.37194714E+02, -0.40984364E+01, -0.12301613E+02, 0.77951894E+01, 0.56480533E+00, -0.55646262E+01, -0.60771265E+01, 0.43992543E+01, -0.39650044E+01, 
    },
    {
      -0.80576096E+01, -0.10058566E+02, -0.42752757E+01, 0.22247026E+02, -0.14402002E+01, -0.33199692E+01, 0.17421428E+01, -0.34508293E+01, 0.32978044E+01, 0.24247870E+01, 0.38595335E+01, 0.32630551E+01, -0.44972682E+00, 
    },
    {
      -0.69929409E+01, 0.46824846E+01, -0.20134742E+01, -0.25459471E+01, 0.26119644E+02, 0.29910052E+01, -0.36391468E+01, -0.93128592E+00, -0.17195288E+01, 0.21541712E+01, 0.83281469E+01, 0.72017050E+00, -0.27138808E+00, 
    },
    {
      -0.15138537E+02, 0.36185287E+02, -0.22484388E+02, 0.22932758E+02, 0.28119076E+02, 0.15488616E+02, -0.30330801E+01, -0.34491429E+01, 0.28721263E+01, -0.30168860E+01, -0.23003569E+01, 0.31897192E+01, 0.54570012E+01, 
    },
    {
      -0.10177624E+02, 0.91815348E+01, -0.54682875E+01, 0.65655260E+01, 0.30150761E+02, 0.12102582E+01, 0.89690657E+01, -0.31353066E+01, 0.28226671E+00, 0.41323018E+01, 0.27218044E+01, -0.37113197E+01, 0.31892061E+01, 
    },
    {
      -0.56241829E-01, 0.73317058E-01, -0.23292311E+00, -0.61917238E-01, 0.11740326E+00, -0.28218538E+00, -0.69195360E+00, 0.36205132E-01, 0.38509484E-01, 0.25892833E-01, 0.39562065E-01, -0.88990778E-02, -0.71549118E-02, 
    },
    {
      0.10465758E+00, -0.25458544E+00, -0.65254413E-01, 0.92262980E-02, -0.97085372E-01, 0.68069780E+00, -0.31799850E+00, -0.16797258E-01, 0.35726070E-01, 0.86577646E-02, -0.35667919E-01, 0.55345451E-02, -0.37628175E-02, 
    },
    {
      -0.15063886E+00, 0.63173950E+00, 0.58640532E-01, -0.11310439E+00, -0.56205934E+00, -0.12007036E+00, -0.60520858E+00, -0.13778128E+00, 0.53372163E-01, 0.55302665E-01, 0.43478370E-01, 0.42523559E-01, -0.29654521E-01, 
    },
    {
      -0.46249099E-01, -0.18232642E+00, 0.18445110E+00, 0.11651336E+00, -0.13888021E+00, 0.77584624E+00, -0.20802420E+00, -0.27784950E+00, -0.12532072E+00, -0.95613144E-01, -0.53957444E-01, 0.21425946E-01, -0.73493482E-02, 
    },
    {
      0.32895875E+00, 0.57959902E+00, 0.17125003E+01, 0.64916812E-01, -0.66671216E+00, 0.42497212E+00, 0.10747848E+01, -0.29168382E+00, -0.33504639E-02, -0.39606754E-01, -0.48538277E+00, -0.46966869E-01, 0.79471610E-01, 
    },
    {
      -0.63520753E+00, 0.35119760E+01, -0.12184794E+01, 0.14708529E+00, 0.32728174E+00, -0.13452771E+01, 0.91351074E+00, -0.10723860E+00, -0.46364650E+00, -0.17759089E+00, 0.33546487E-03, -0.68300068E-01, 0.33596814E-01, 
    },
    {
      -0.32629791E+00, -0.10383365E+01, 0.18383627E+00, -0.12193326E+00, 0.95989013E+00, 0.17567472E+00, 0.10174921E+01, 0.54608107E+00, -0.69135964E-01, -0.16146509E+00, -0.69958568E-01, -0.12977321E+00, 0.74624196E-01, 
    },
    {
      -0.34265670E+00, 0.70992500E+00, -0.20130862E+00, -0.87979764E+00, 0.27189708E+00, -0.16422273E+01, 0.42830017E+00, 0.45904860E+00, 0.39268953E+00, 0.97692028E-01, 0.53528871E-01, -0.25272183E-02, -0.82533369E-02, 
    },
    {
      0.30163246E+00, -0.75499952E+00, -0.23318834E+01, 0.28877845E+00, 0.12862626E+01, -0.79786736E+00, -0.10814675E+01, 0.10795548E+00, -0.14806841E+00, 0.30484816E-01, 0.60364807E+00, 0.24701132E+00, -0.16554859E+00, 
    },
    {
      0.75564712E+00, -0.49911842E+01, 0.25221035E+01, -0.50133252E+00, -0.11462419E+00, 0.19772404E+01, -0.17406584E+01, 0.17858028E+00, 0.75754690E+00, 0.32222256E+00, 0.12655260E+00, 0.14917502E+00, -0.11977191E+00, 
    },
    {
      0.76102793E-01, -0.43147720E-01, -0.81818521E-01, 0.19644612E-01, 0.33825956E-01, 0.25999684E-01, 0.32359265E-01, 0.29988968E+00, -0.63901916E-01, 0.23167776E-02, 0.22926815E-01, -0.77330978E-02, 0.10746138E-02, 
    },
    {
      0.12264740E+00, 0.20447599E-01, 0.59562452E-01, 0.19856852E-01, -0.11300006E+00, -0.29754141E-01, -0.44352575E-02, 0.56839380E-01, 0.25084686E+00, -0.56135736E-03, -0.54703793E-02, 0.48170928E-02, -0.36291028E-02, 
    },
    {
      0.24246810E-01, 0.65477557E-01, 0.12209361E+00, -0.99905565E-01, -0.29400787E+00, -0.25406672E-01, 0.31711336E-01, 0.17040537E+00, 0.24152596E+00, -0.77344932E-01, 0.33627842E-01, -0.97614238E-02, 0.20799657E-01, 
    },
    {
      0.12910360E+00, 0.27060756E-01, -0.15518239E+00, 0.15885320E+00, 0.30369675E+00, -0.49701214E-01, -0.10029802E+00, -0.28022015E+00, 0.36929357E-02, 0.24210740E-01, -0.16487345E-01, 0.94972895E-02, -0.56310687E-02, 
    },
    {
      -0.11048770E+00, -0.25132010E-01, -0.74848481E-01, -0.36955964E-01, -0.84306849E-02, 0.59861795E-03, 0.43695938E-01, 0.46816003E-02, -0.12226735E-01, 0.69531336E-01, 0.23915982E+00, -0.52309283E-02, -0.14996283E-01, 
    },
    {
      0.29184736E-01, 0.55157773E-01, 0.20157386E-01, 0.30791920E-01, -0.58250558E-01, -0.56456961E-02, -0.21537563E-01, 0.28616695E-01, -0.26698606E-02, -0.25978240E+00, 0.62796988E-01, 0.22548081E-02, -0.23580464E-02, 
    },
    {
      0.47859617E-01, -0.17690947E-01, -0.53106416E-02, -0.21001626E-01, -0.39911261E-02, 0.21716015E-01, -0.17059458E-01, 0.28004622E-01, 0.10446360E-01, -0.44645024E-02, -0.42491471E-02, -0.13451762E+00, -0.59785281E-03, 
    },
    {
      -0.11083220E-01, 0.54354630E-01, -0.40297960E-02, 0.53323619E-01, 0.48851684E-01, 0.93341740E-02, -0.30359656E-05, -0.13635743E-01, -0.13263462E-01, -0.91869375E-02, -0.23183506E-01, 0.10026116E-02, -0.14598797E+00, 
    },
    {
      -0.39637778E-01, 0.34675904E-01, -0.13779508E-01, 0.31649983E-02, 0.24430851E-01, -0.10841477E-01, -0.87321568E-02, 0.24531895E-01, -0.39002334E-02, -0.10708721E-01, -0.32987449E-01, 0.29985204E-02, 0.48053171E-02, 
    },
    {
      0.86180255E-01, 0.39627044E-02, 0.22435104E-01, 0.21348750E-01, -0.34228362E-01, -0.53658383E-03, 0.33991758E-01, -0.37611157E-01, 0.23497487E-02, -0.50931578E-02, 0.65411784E-03, 0.11545370E-01, -0.11957943E-02, 
    },
    {
      -0.27704688E-01, 0.11326194E-01, -0.12809221E+00, 0.25609735E-01, -0.43673404E-01, -0.19919191E-01, 0.57441801E-01, 0.13651283E-02, 0.99236434E-02, 0.11110033E-02, -0.88214837E-02, 0.10099307E-01, 0.23763576E-03, 
    },
    {
      -0.32913532E-01, 0.35655253E-01, -0.93667448E-01, 0.22362642E-01, 0.35769939E-01, 0.23009181E-01, 0.10751249E-01, 0.21540695E-02, -0.16343425E-02, -0.27889894E-01, -0.19826155E-01, -0.16842147E-02, 0.94351284E-02, 
    },
  },
  {
    {
      0.89623709E+01, 0.19730670E-01, 0.21596248E+00, 0.50009090E-01, -0.21762836E-02, -0.36576837E-01, -0.37319519E-01, 0.31727146E-01, -0.24356844E-01, 0.29576533E-01, 0.85489517E-02, -0.13777472E-01, -0.97738234E-02, 
    },
    {
      0.33469999E+00, -0.24697515E+00, 0.48163033E+00, -0.35931331E+00, -0.15751718E-01, -0.13071296E+00, 0.29412560E-01, -0.13171205E+00, -0.74888915E-01, -0.31300247E-01, -0.69453977E-01, 0.31701785E-01, 0.48052177E-01, 
    },
    {
      0.24180008E+02, -0.59165543E+00, -0.37481437E+01, 0.27094533E+01, 0.17237262E+01, 0.33680838E+00, 0.12361391E+01, -0.40080991E+00, 0.24777052E+00, -0.11386623E+01, -0.26397088E+00, 0.51696777E+00, 0.26516542E+00, 
    },
    {
      -0.45573082E+01, 0.22108464E+01, -0.33646340E+01, 0.54219594E+01, 0.28218365E+01, 0.23740580E+01, -0.16911587E+01, 0.25113804E+01, 0.21021688E+01, 0.10237397E+00, 0.14552329E+01, -0.97442865E-01, -0.11147535E+01, 
    },
    {
      -0.11681603E+03, 0.18470577E+01, 0.18251852E+02, -0.25811161E+02, -0.20601797E+02, -0.18826652E+01, -0.56787043E+01, 0.29563882E+01, -0.94193526E-01, 0.73134317E+01, 0.93215400E+00, -0.28211844E+01, -0.22883477E+01, 
    },
    {
      0.19154861E+02, -0.18602942E+02, 0.20144350E+02, -0.16414492E+02, -0.60010877E+01, -0.12694320E+02, 0.97623625E+01, -0.14265865E+02, -0.14238819E+02, 0.14062835E+00, -0.69437256E+01, -0.15784086E+00, 0.63612428E+01, 
    },
    {
      0.18642741E+03, 0.41710634E+01, -0.36739273E+02, 0.75611092E+02, 0.65239151E+02, 0.62713485E+01, 0.90178194E+01, -0.86719627E+01, -0.24287522E+01, -0.17567657E+02, -0.38892239E+00, 0.58335896E+01, 0.67666426E+01, 
    },
    {
      -0.25843210E+02, 0.54234547E+02, -0.64408768E+02, 0.10873847E+02, -0.72628431E+01, 0.26647579E+02, -0.19763281E+02, 0.33570370E+02, 0.35707272E+02, -0.12532645E-01, 0.12849774E+02, 0.64608556E+00, -0.15312591E+02, 
    },
    {
      -0.14083356E+03, -0.12821358E+02, 0.34492329E+02, -0.88502136E+02, -0.78154541E+02, -0.92759399E+01, -0.48082018E+01, 0.10566218E+02, 0.48087645E+01, 0.18066017E+02, -0.13238797E+01, -0.51708665E+01, -0.80225296E+01, 
    },
    {
      0.59901586E+01, -0.64209244E+02, 0.83126961E+02, 0.10725402E+02, 0.23379948E+02, -0.23843412E+02, 0.16165714E+02, -0.34889957E+02, -0.37377113E+02, -0.10515053E+01, -0.10207627E+02, -0.58196390E+00, 0.16365570E+02, 
    },
    {
      0.43592426E+02, 0.73433652E+01, -0.12409451E+02, 0.35991268E+02, 0.31727673E+02, 0.46962137E+01, 0.23304205E+00, -0.45286808E+01, -0.25499396E+01, -0.67354417E+01, 0.10447700E+01, 0.16423652E+01, 0.32990837E+01, 
    },
    {
      0.50915036E+01, 0.26708822E+02, -0.35995022E+02, -0.10378606E+02, -0.12888997E+02, 0.75558434E+01, -0.44697514E+01, 0.13253450E+02, 0.13915670E+02, 0.88236904E+00, 0.29039440E+01, 0.16834038E+00, -0.63849540E+01, 
    },
    {
      -0.26160195E+00, 0.16475309E+01, 0.18607447E+01, 0.27880883E-01, -0.80495954E-01, -0.39654482E-01, 0.19887313E-01, 0.60452126E-01, -0.19153394E-01, 0.13035006E-01, 0.31249687E-01, -0.75434586E-02, -0.15125268E-01, 
    },
    {
      0.27377906E+00, -0.17763206E+01, 0.15071578E+01, -0.19758046E-01, 0.10182646E+00, -0.26982676E-01, -0.23550984E-01, 0.36379886E-02, -0.36529673E-02, 0.11719952E-01, 0.14171905E-01, -0.14677171E-01, -0.14988882E-02, 
    },
    {
      -0.11673796E+01, 0.21813788E+01, -0.90088534E+00, 0.22404641E+00, -0.33333287E+00, -0.52520949E-01, 0.29254219E+00, 0.10636131E+00, 0.68046279E-01, 0.12188297E-02, -0.82268119E-01, -0.86118124E-01, 0.65291584E-01, 
    },
    {
      0.17419060E+01, 0.69445306E+00, 0.54693860E+00, -0.17051892E-01, 0.58553481E+00, -0.22795539E-01, 0.11075487E+00, 0.25393158E+00, 0.15103640E+00, 0.17391805E+00, -0.11656546E+00, -0.77986225E-01, -0.58590692E-01, 
    },
    {
      0.73945637E+01, 0.12272065E+02, -0.10930670E+02, 0.13587241E+01, 0.17232165E+01, 0.35481679E+00, 0.61459994E+00, -0.78423440E+00, 0.48749742E+00, -0.26622641E+00, -0.94569325E+00, 0.14302284E+00, 0.17925684E+00, 
    },
    {
      -0.94064894E+01, 0.11073936E+02, 0.13409169E+02, 0.13028538E+01, 0.65245140E+00, 0.19972022E+00, 0.22320145E+00, 0.11397737E+01, 0.71284533E+00, -0.11684898E+01, 0.29646003E+00, 0.81502265E+00, -0.25439862E-01, 
    },
    {
      0.23263647E+02, -0.16287771E+02, 0.19715300E+02, -0.70920930E+01, 0.32130179E+01, 0.31002333E+01, -0.58355947E+01, -0.65897214E+00, 0.11365482E+00, -0.26643327E+00, 0.25300441E+01, 0.20917065E+01, -0.21054513E+01, 
    },
    {
      -0.25114740E+02, -0.66058850E+01, 0.86541386E+01, -0.41537158E-01, -0.17567032E+02, 0.47047324E+01, -0.51503760E+00, -0.63852286E+01, -0.29741337E+01, -0.44764342E+01, 0.62373561E+00, 0.24865687E+01, 0.47261235E+00, 
    },
    {
      -0.46814045E+02, -0.12041885E+02, 0.99752289E+02, -0.64495907E+01, -0.92173414E+01, -0.46798882E+01, -0.54954991E+01, 0.61131954E+01, -0.16522959E+01, -0.54570884E+00, 0.57664261E+01, -0.13666019E+01, -0.71777904E+00, 
    },
    {
      0.84022469E+02, -0.11194759E+03, -0.39507432E+01, -0.16203300E+02, -0.66421180E+01, -0.14821443E+00, -0.78339767E+00, -0.83775568E+01, -0.87370195E+01, 0.92152634E+01, -0.33313084E+01, -0.74441700E+01, 0.18862633E+00, 
    },
    {
      -0.14273080E+03, 0.62041721E+02, -0.79482536E+02, 0.51631123E+02, -0.57821947E+00, -0.19178213E+02, 0.29178068E+02, 0.33035190E+01, -0.28263030E+01, 0.34217317E-01, -0.18571112E+02, -0.10850002E+02, 0.14315627E+02, 
    },
    {
      0.12893007E+03, -0.33273048E+01, -0.50037964E+02, -0.18768166E+01, 0.10671618E+03, -0.27278667E+02, 0.50431371E+01, 0.36296223E+02, 0.19505959E+02, 0.25242184E+02, 0.24296870E+01, -0.16299253E+02, -0.30654261E+01, 
    },
    {
      0.11099849E+03, -0.12857597E+03, -0.27917538E+03, 0.11415404E+02, 0.13686012E+02, 0.19129147E+02, 0.14277333E+02, -0.19621246E+02, 0.15207518E+01, 0.56465182E+01, -0.13264232E+02, 0.55628519E+01, 0.15896335E+01, 
    },
    {
      -0.26209372E+03, 0.31271347E+03, -0.18712296E+03, 0.61504059E+02, 0.26230099E+02, -0.22624297E+01, 0.91929603E+00, 0.19573435E+02, 0.32240540E+02, -0.23829140E+02, 0.96026239E+01, 0.23164410E+02, -0.11098813E+01, 
    },
    {
      0.37926721E+03, -0.13912029E+03, 0.11385747E+03, -0.15078870E+03, -0.28678265E+02, 0.41981728E+02, -0.61708336E+02, -0.99502068E+01, 0.10072963E+02, 0.43251381E+01, 0.50344307E+02, 0.20761368E+02, -0.38384724E+02, 
    },
    {
      -0.31183640E+03, 0.80717461E+02, 0.11177303E+03, 0.91798124E+01, -0.26519113E+03, 0.57131924E+02, -0.17250916E+02, -0.80624062E+02, -0.53986412E+02, -0.57842266E+02, -0.15444742E+02, 0.40284004E+02, 0.98377438E+01, 
    },
    {
      -0.10636666E+03, 0.27624472E+03, 0.33518640E+03, -0.90313139E+01, -0.29404740E+01, -0.29340532E+02, -0.13130365E+02, 0.27019056E+02, 0.83808237E+00, -0.92540722E+01, 0.13138417E+02, -0.91429548E+01, -0.16192185E+01, 
    },
    {
      0.33913440E+03, -0.35920078E+03, 0.38479999E+03, -0.92527618E+02, -0.46624046E+02, 0.41366453E+01, 0.18275695E+01, -0.17915739E+02, -0.46472927E+02, 0.25177954E+02, -0.10816146E+02, -0.29169777E+02, 0.18046658E+01, 
    },
    {
      -0.44847644E+03, 0.16557732E+03, -0.59701851E+02, 0.19198531E+03, 0.51862465E+02, -0.38352905E+02, 0.59586216E+02, 0.13888144E+02, -0.13456491E+02, -0.92949009E+01, -0.56630112E+02, -0.15786907E+02, 0.44390171E+02, 
    },
    {
      0.35611423E+03, -0.13994145E+03, -0.12520427E+03, -0.78124218E+01, 0.29974640E+03, -0.48495983E+02, 0.19427450E+02, 0.77247574E+02, 0.66220634E+02, 0.59678249E+02, 0.23805574E+02, -0.41955189E+02, -0.13117631E+02, 
    },
    {
      0.34567719E+02, -0.14869554E+03, -0.14342838E+03, 0.30032880E+01, -0.36467550E+01, 0.15035783E+02, 0.32215376E+01, -0.13186405E+02, -0.12452888E+01, 0.43014369E+01, -0.46899452E+01, 0.50963411E+01, 0.54561722E+00, 
    },
    {
      -0.15541122E+03, 0.14388922E+03, -0.21015965E+03, 0.47509956E+02, 0.29349554E+02, -0.19194942E+01, -0.30574071E+01, 0.53069601E+01, 0.22960079E+02, -0.95035095E+01, 0.42254791E+01, 0.12840061E+02, -0.78551632E+00, 
    },
    {
      0.19336099E+03, -0.79162109E+02, 0.28272572E+01, -0.89598595E+02, -0.26287643E+02, 0.12450287E+02, -0.21709846E+02, -0.69439921E+01, 0.60890908E+01, 0.54649796E+01, 0.22320847E+02, 0.36850126E+01, -0.18504169E+02, 
    },
    {
      -0.15068378E+03, 0.72892563E+02, 0.55283623E+02, -0.76779521E+00, -0.12916338E+03, 0.13477338E+02, -0.61723790E+01, -0.26636496E+02, -0.29567837E+02, -0.23007311E+02, -0.11395707E+02, 0.15494074E+02, 0.59043760E+01, 
    },
    {
      0.97768597E-01, -0.21706764E+00, -0.14598057E+00, -0.98067135E+00, 0.56929809E+00, 0.43482479E-01, 0.11748247E-01, -0.63081672E-02, 0.20523420E-01, -0.19572712E-01, 0.17729735E-03, 0.11943498E-02, 0.66722082E-02, 
    },
    {
      -0.22512754E-01, -0.17544612E+00, -0.19724552E+00, -0.56630099E+00, -0.77338606E+00, -0.35187386E-01, 0.10012046E-01, -0.27610054E-01, 0.89784892E-03, -0.95248339E-03, 0.29487306E-01, 0.15489024E-01, 0.32603152E-02, 
    },
    {
      0.90035915E+00, 0.50192165E+00, 0.39231557E+00, 0.72755909E+00, 0.34172869E+00, 0.64472872E+00, -0.24112161E-01, -0.37124623E-01, -0.18246342E+00, -0.95317356E-01, -0.20045063E-01, -0.81074119E-01, 0.11073198E-01, 
    },
    {
      0.55172008E+00, 0.20146661E+00, 0.44691011E+00, 0.65697357E-01, 0.77415019E+00, 0.60446817E-01, 0.51282352E+00, -0.15366629E+00, -0.22197446E+00, -0.59145335E-01, -0.21178664E+00, 0.11207618E+00, -0.13108963E-01, 
    },
    {
      0.10676382E+01, 0.29645822E+01, 0.37952478E+01, -0.78678852E+00, 0.32439947E+01, -0.16431471E+01, -0.11421096E+01, 0.17130587E+00, 0.37664506E+00, 0.55448914E+00, 0.10976688E+00, -0.48771808E+00, -0.32246506E+00, 
    },
    {
      0.27298226E+01, 0.24008121E+01, 0.60759602E+01, -0.47121968E+01, -0.16266237E+01, -0.13724278E+00, -0.14916685E+01, 0.16014088E+01, 0.21585748E+00, -0.35771871E+00, -0.89440823E+00, -0.45400912E+00, -0.29001749E+00, 
    },
    {
      -0.74832025E+01, -0.29968793E+01, -0.42970152E+01, 0.20118179E+01, -0.21555855E+01, -0.44935822E+01, -0.38260967E+00, -0.88122147E+00, 0.21959436E+01, 0.85298419E+00, 0.65500510E+00, 0.64462608E+00, 0.34080345E-01, 
    },
    {
      -0.58604355E+01, -0.17552921E+01, -0.46983767E+01, -0.19348935E+01, 0.24940045E+01, 0.21718979E+01, -0.40221734E+01, 0.73567528E+00, 0.19161574E+01, -0.13397838E+00, 0.25529764E+01, -0.50712591E+00, 0.23999058E+00, 
    },
    {
      -0.10549742E+02, -0.37017422E+01, -0.24552065E+02, 0.14676774E+02, -0.11914489E+02, 0.10997478E+02, 0.74938984E+01, -0.25663313E+00, -0.35813992E+01, -0.28400996E+01, -0.93953723E+00, 0.33387854E+01, 0.20984364E+01, 
    },
    {
      -0.14897774E+02, -0.62908635E+01, -0.32800335E+02, 0.18528563E+02, 0.12417139E+02, -0.69420964E+00, 0.11475649E+02, -0.82309322E+01, -0.28264165E+01, 0.27254322E+01, 0.45255299E+01, 0.22271841E+01, 0.21908436E+01, 
    },
    {
      0.17852583E+02, 0.30649593E+01, 0.11477947E+02, -0.67365303E+01, 0.28456860E+01, 0.10238112E+02, 0.33093038E+00, 0.40141964E+01, -0.65592384E+01, -0.24417315E+01, -0.14772207E+01, -0.15737491E+01, -0.74067026E-01, 
    },
    {
      0.15307250E+02, -0.26185241E+01, 0.12405679E+02, 0.62960510E+01, -0.67733917E+01, -0.70792394E+01, 0.91405792E+01, 0.88415743E-03, -0.37179437E+01, 0.90935570E+00, -0.69149547E+01, 0.18560132E+00, -0.67503846E+00, 
    },
    {
      0.21599215E+02, -0.46994729E+01, 0.46027321E+02, -0.36294868E+02, 0.96739960E+01, -0.24945257E+02, -0.13759644E+02, -0.24616983E-01, 0.89603739E+01, 0.54308796E+01, 0.24014454E+01, -0.70196056E+01, -0.44327598E+01, 
    },
    {
      0.24843397E+02, 0.92978916E+01, 0.56579742E+02, -0.19334846E+02, -0.24256990E+02, 0.30407271E+01, -0.25068012E+02, 0.13572435E+02, 0.68599362E+01, -0.67986002E+01, -0.75372796E+01, -0.38506787E+01, -0.46823874E+01, 
    },
    {
      -0.14843636E+02, 0.19837296E+01, -0.11109138E+02, 0.31957550E+01, -0.17539186E+01, -0.67461467E+01, 0.35601419E+00, -0.37653399E+01, 0.55061636E+01, 0.21600037E+01, 0.83115494E+00, 0.11146947E+01, -0.17095937E+00, 
    },
    {
      -0.12286270E+02, 0.98170996E+01, -0.81901531E+01, -0.38168242E+01, 0.12576302E+01, 0.55010004E+01, -0.60531683E+01, -0.10407491E+01, 0.18965904E+01, -0.94105482E+00, 0.53762965E+01, 0.51917458E+00, 0.65497190E+00, 
    },
    {
      -0.13637304E+02, 0.65925388E+01, -0.23569542E+02, 0.25342230E+02, 0.71028113E+00, 0.16614403E+02, 0.75715604E+01, 0.19589573E-01, -0.64481812E+01, -0.36762245E+01, -0.19363295E+01, 0.46145444E+01, 0.30328979E+01, 
    },
    {
      -0.13089158E+02, -0.10095703E+02, -0.30684874E+02, 0.39198732E+01, 0.14732669E+02, -0.24675801E+01, 0.16160910E+02, -0.72000704E+01, -0.45575032E+01, 0.50762730E+01, 0.37089520E+01, 0.21473510E+01, 0.29847896E+01, 
    },
    {
      -0.14737675E-01, -0.25679311E-01, -0.16136080E+00, -0.33218786E-01, 0.27244406E-01, -0.47106108E+00, -0.91915029E+00, 0.38202498E-01, 0.29088611E-01, 0.31560920E-02, 0.18914854E-01, 0.52554188E-02, -0.11544408E-02, 
    },
    {
      0.72901808E-01, -0.22339118E+00, 0.53658471E-01, 0.85221520E-02, -0.27125636E-01, 0.88938648E+00, -0.52879435E+00, 0.87448508E-02, 0.30740051E-01, 0.10852590E-01, -0.27724359E-01, 0.31586990E-02, -0.96590295E-02, 
    },
    {
      0.20115213E+00, 0.69436580E+00, -0.17900576E+00, 0.66310994E-01, -0.75164104E+00, 0.13775501E+00, -0.21529463E+00, -0.14966124E+00, 0.16769144E+00, 0.77871263E-01, 0.51278979E-01, 0.15873125E-01, 0.21192819E-01, 
    },
    {
      0.77322774E-01, -0.42061176E-01, 0.45077723E+00, 0.66685408E-01, -0.15208274E+00, 0.44830185E+00, -0.12592126E+00, -0.13366275E+00, -0.90220653E-01, -0.61910573E-01, -0.25508961E-01, 0.12957442E-01, -0.12587028E-01, 
    },
    {
      0.74074906E+00, 0.65904367E+00, 0.15433567E+01, -0.63742287E-02, -0.98790757E-01, 0.18118404E+01, 0.15583321E+01, -0.43290365E+00, -0.20866205E+00, 0.95957085E-01, -0.65422811E-01, -0.14005326E+00, -0.20921381E-01, 
    },
    {
      -0.11043867E+01, 0.32856829E+01, -0.19886183E+01, 0.40116802E+00, 0.26653954E+00, -0.17676287E+01, 0.24231007E+01, -0.21656232E+00, -0.57338196E+00, -0.14849383E+00, 0.14925669E+00, -0.52936438E-01, 0.73655564E-02, 
    },
    {
      -0.71928334E+00, -0.94114923E+00, 0.78959525E+00, -0.13111615E+00, 0.16068588E+01, -0.33002041E-01, 0.42055130E+00, 0.56223714E+00, -0.28989446E+00, -0.17253461E+00, 0.78965602E-02, -0.94987266E-01, -0.70541412E-01, 
    },
    {
      -0.24910378E+00, 0.25402689E+00, -0.95344973E+00, -0.80586588E+00, 0.43270764E+00, -0.10406414E+01, 0.31384522E+00, 0.15633456E+00, 0.34068555E+00, 0.12060667E+00, 0.44971276E-01, 0.29588670E-01, 0.46364356E-01, 
    },
    {
      -0.51294982E+00, -0.29964095E+00, -0.28561344E+01, -0.11806239E+00, 0.18849920E+00, -0.34402866E+01, -0.21050479E+01, 0.34388900E+00, 0.30473164E+00, -0.14386865E+00, -0.43631759E-01, 0.33641663E+00, -0.22661610E-01, 
    },
    {
      0.15949913E+01, -0.41355290E+01, 0.32266247E+01, -0.85630828E+00, -0.50795478E+00, 0.28716083E+01, -0.42127266E+01, 0.28837740E+00, 0.93107861E+00, 0.27240244E+00, -0.13775127E+00, 0.13223094E+00, -0.13103773E+00, 
    },
    {
      -0.49044993E-01, -0.24295989E-02, -0.71430229E-01, 0.35413116E-01, 0.47498848E-01, -0.13454735E-01, 0.98276092E-02, 0.32753569E+00, -0.18088369E+00, 0.19360161E-02, 0.13374137E-01, -0.94158677E-02, -0.30937691E-02, 
    },
    {
      0.55371165E-01, 0.24911972E-01, 0.41393071E-01, 0.39408021E-01, -0.64998358E-01, -0.18124787E-01, -0.27285462E-01, 0.16500179E+00, 0.29878315E+00, 0.13220957E-02, -0.14755523E-02, 0.31277381E-02, 0.15555121E-03, 
    },
    {
      -0.56562969E-02, 0.10716920E-01, 0.58698531E-01, -0.22612128E-02, -0.11211977E+00, -0.11326517E+00, 0.88223264E-01, 0.13487896E+00, 0.22745983E+00, -0.64147316E-01, 0.13011377E-01, -0.15176603E-01, 0.37889306E-01, 
    },
    {
      0.18017766E+00, 0.76493047E-01, -0.57414532E-01, -0.15384177E-01, 0.19273417E+00, -0.67927875E-01, -0.18422922E+00, -0.22047012E+00, 0.14405789E-01, 0.12134630E-01, -0.35937384E-01, -0.15340685E-02, 0.19010877E-01, 
    },
    {
      -0.14448559E+00, -0.33150800E-02, -0.10132004E-01, 0.44721063E-01, -0.38510407E-02, 0.65288940E-02, 0.30064268E-01, -0.13620453E-01, -0.25559587E-01, 0.13974449E+00, 0.27066708E+00, -0.86301874E-03, -0.15927410E-01, 
    },
    {
      0.85362434E-01, 0.48635934E-01, -0.30831050E-01, 0.35901424E-02, -0.16469320E-01, -0.93949521E-02, 0.62366962E-02, 0.42513665E-02, 0.16225785E-01, -0.26053265E+00, 0.12521908E+00, -0.24030069E-02, -0.12820498E-02, 
    },
    {
      0.42795010E-01, -0.88504680E-01, -0.15416031E-01, 0.35100665E-01, 0.10118750E-01, 0.25384672E-01, -0.40251631E-02, 0.18797457E-02, -0.41919933E-02, -0.62470161E-02, -0.65072584E-02, -0.15718541E+00, 0.42815138E-01, 
    },
    {
      0.51770523E-01, 0.23641726E-01, -0.22212747E-01, 0.52224710E-02, 0.48885681E-01, 0.18276574E-01, -0.21028422E-01, -0.14288486E-01, -0.28683112E-02, 0.25293042E-03, -0.83165923E-02, -0.44567961E-01, -0.16046225E+00, 
    },
    {
      -0.36254805E-01, -0.75680152E-01, 0.11980511E-01, 0.15882116E-03, 0.22734884E-01, -0.10775125E-01, -0.22750691E-01, 0.10671292E-01, -0.29391658E-02, -0.48969565E-02, -0.22469327E-01, 0.37602372E-02, 0.80211431E-03, 
    },
    {
      0.21320259E-01, -0.17746110E-01, 0.20988828E-02, -0.17451579E-01, -0.29435143E-01, 0.10905027E-01, -0.15734235E-01, -0.13093308E-01, 0.74879108E-02, -0.76296972E-02, 0.12714958E-01, 0.12751132E-01, 0.24891517E-02, 
    },
    {
      -0.13963948E-01, 0.43464817E-01, -0.57666734E-01, 0.98980134E-02, -0.56504559E-01, -0.33702858E-01, 0.28680092E-01, 0.24618777E-01, 0.21952307E-01, 0.17210137E-01, -0.88103274E-02, 0.80375150E-02, 0.10196300E-01, 
    },
    {
      -0.33362068E-01, -0.18013632E-01, -0.50624602E-01, 0.88627376E-02, 0.10242135E+00, 0.11714541E-01, 0.93254829E-02, 0.93967225E-02, -0.15307956E-01, -0.24555840E-01, -0.17302781E-01, 0.19131057E-02, 0.42815138E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir19_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30401049E+01, -0.11382078E-01, 0.13699237E-01, 0.13619756E-01, -0.15344954E-03, -0.39020018E-03, 0.33493678E-02, -0.11851718E-02, 0.34948639E-03, 
    },
    {
      -0.82672872E-01, -0.32174747E-01, 0.38680803E-01, 0.55962011E-01, 0.32362810E-03, -0.41486818E-01, -0.11401795E-01, -0.68960893E-02, 0.10805070E-01, 
    },
    {
      0.14065900E+01, 0.15397176E+00, -0.18210937E+00, -0.31921729E+00, -0.54489747E-02, 0.57063509E-01, 0.12826703E-01, -0.53385347E-02, 0.62804609E-02, 
    },
    {
      0.22763485E+00, -0.14647670E+00, -0.14446510E+00, -0.12818100E+00, 0.18414596E-01, 0.14890826E+00, 0.45125451E-01, 0.18452298E-01, -0.38831923E-01, 
    },
    {
      -0.28037069E+01, -0.28717342E+00, 0.53387308E+00, 0.77303892E+00, 0.49879473E-01, -0.18796812E+00, -0.46676476E-01, -0.11675812E-02, -0.18786209E-01, 
    },
    {
      -0.20530869E+00, 0.14871514E+00, 0.13572893E+00, 0.65998554E-01, -0.31058760E-01, -0.12001267E+00, -0.36234986E-01, -0.10483883E-01, 0.33103835E-01, 
    },
    {
      0.15812294E+01, 0.16315398E+00, -0.34262037E+00, -0.47133777E+00, -0.53432129E-01, 0.13809934E+00, 0.39644614E-01, 0.98345075E-02, 0.13404462E-01, 
    },
    {
      0.11054382E-01, -0.15015076E+00, -0.34573376E+00, 0.15864993E-01, 0.12899942E-02, 0.38709242E-02, -0.75407006E-04, 0.21587815E-02, 0.61787381E-02, 
    },
    {
      0.11099253E-01, 0.36393246E+00, -0.14289594E+00, -0.51004002E-02, 0.82212426E-02, 0.36659569E-02, -0.50246581E-02, 0.64045354E-03, 0.93177753E-02, 
    },
    {
      -0.18676747E+00, 0.13912369E+00, 0.92345759E-01, 0.45837265E-01, -0.24934011E-01, -0.26602110E-01, -0.30030532E-01, 0.49917046E-02, 0.12582732E-01, 
    },
    {
      0.13173369E+00, -0.10819068E+00, 0.15030783E+00, -0.23040416E-01, 0.13999088E-01, -0.53261607E-02, -0.43439619E-01, -0.90245353E-02, 0.10945592E+00, 
    },
    {
      0.33914685E-01, 0.47822487E+00, 0.16972907E+01, -0.97674638E-01, -0.25269112E+00, -0.44677760E-01, 0.67510251E-02, 0.25692284E-01, -0.82349569E-01, 
    },
    {
      -0.35780886E+00, -0.18699245E+01, 0.49952665E+00, 0.71142484E-02, 0.22229070E-01, -0.12980822E+00, 0.70356480E-02, -0.14538261E-01, -0.14582597E+00, 
    },
    {
      0.21972475E+01, -0.59621954E+00, -0.96930653E+00, -0.94870478E-01, 0.64373899E+00, 0.14172481E+00, -0.12882969E+00, 0.14451261E+00, -0.11963338E+00, 
    },
    {
      -0.23363806E+00, 0.16209319E+01, -0.70229024E+00, 0.12993802E+00, -0.48182800E-01, 0.18711828E+00, 0.38275993E+00, 0.53427994E-01, -0.70009434E+00, 
    },
    {
      -0.24579667E+00, 0.24078412E-01, -0.16323544E+01, -0.19505771E+00, 0.68362409E+00, 0.15810391E+00, 0.11939487E+00, -0.77476732E-01, 0.17349911E+00, 
    },
    {
      0.10717825E+01, 0.19524435E+01, -0.70215613E+00, 0.40747777E+00, -0.14923325E+00, 0.51048869E+00, 0.74851766E-01, -0.32693017E-01, 0.38650307E+00, 
    },
    {
      -0.58413815E+01, 0.16597109E+00, 0.27413237E+01, 0.22096719E+00, -0.18455598E+01, -0.18398350E+00, 0.69714391E+00, -0.38460970E+00, 0.25037786E+00, 
    },
    {
      0.16911915E+00, -0.42069697E+01, 0.91982204E+00, -0.53290057E+00, 0.25880048E+00, -0.68226004E+00, -0.97130626E+00, 0.19385042E-01, 0.14533879E+01, 
    },
    {
      0.40299878E+00, -0.48112997E+00, 0.52594155E+00, 0.29526755E+00, -0.43825641E+00, -0.12281952E+00, -0.15333907E+00, 0.56682952E-01, -0.86398900E-01, 
    },
    {
      -0.75084627E+00, -0.45546556E+00, 0.44570705E+00, -0.53101832E+00, 0.12392121E-01, -0.47902694E+00, -0.12242783E+00, 0.79856642E-01, -0.28643292E+00, 
    },
    {
      0.42676086E+01, 0.75077915E+00, -0.22261817E+01, -0.27955213E+00, 0.13397520E+01, 0.14111669E+00, -0.55246305E+00, 0.23521648E+00, -0.18291542E+00, 
    },
    {
      -0.12076446E+00, 0.32277856E+01, -0.13160841E+00, 0.44515896E+00, -0.28191578E+00, 0.62046933E+00, 0.68323207E+00, -0.10502839E+00, -0.91661125E+00, 
    },
    {
      0.27267998E-02, -0.37316999E-02, 0.35979025E-03, 0.47388568E-03, -0.13095340E+00, 0.21076865E-03, 0.52354215E-02, 0.26401207E-02, -0.60893549E-02, 
    },
    {
      -0.61952616E-02, 0.12154616E-02, 0.40964223E-02, 0.15712863E+00, 0.24770111E-01, 0.38494696E-02, -0.92166457E-02, 0.62270334E-03, -0.54044821E-02, 
    },
    {
      0.42403363E-01, 0.16611996E+00, 0.80097161E-01, 0.56755996E+00, -0.33011404E+00, 0.29526917E-01, -0.65970644E-01, -0.28990030E-01, 0.46952786E-02, 
    },
    {
      -0.80264293E-01, 0.20948894E-01, -0.29410383E-01, 0.50466293E+00, 0.66722238E+00, 0.76955371E-01, -0.33780285E-02, -0.26567908E-01, 0.40409271E-01, 
    },
    {
      -0.15232775E+00, 0.13101333E+00, -0.91356911E-01, 0.89222513E-01, -0.36949329E-02, -0.17322950E-01, 0.55812236E-01, -0.38744360E-01, 0.16779214E-01, 
    },
    {
      -0.24620678E+00, -0.79602763E-01, -0.20739435E+00, -0.15338077E+00, -0.47849964E-01, -0.14388233E-01, 0.94271958E-01, -0.29404018E-01, 0.18598402E-01, 
    },
    {
      -0.18558075E+00, -0.83184230E+00, 0.13150723E-01, -0.22259934E+01, 0.66283786E+00, -0.31915799E-01, 0.28466687E+00, 0.21174742E+00, -0.12655893E-01, 
    },
    {
      0.94725484E+00, 0.19020595E+00, 0.40080431E+00, -0.13141518E+01, -0.27080829E+01, -0.34854394E+00, -0.15705083E+00, 0.14084280E+00, -0.21892302E+00, 
    },
    {
      0.17485859E+00, -0.35491335E+00, 0.11176544E+00, -0.19686691E+00, -0.70896432E-01, 0.47556851E-01, -0.74512035E-01, 0.51269028E-01, -0.11006738E-01, 
    },
    {
      0.36457384E+00, -0.93383693E-04, 0.29480577E+00, 0.35257223E+00, -0.10449865E+00, 0.26363550E-01, -0.74307755E-01, 0.54344729E-01, -0.63038334E-01, 
    },
    {
      0.20447034E+00, 0.11452413E+01, -0.16736257E+00, 0.20206418E+01, -0.30654338E+00, 0.28216397E-01, -0.26526570E+00, -0.28426144E+00, 0.90763699E-02, 
    },
    {
      -0.14910082E+01, -0.37116313E+00, -0.24615522E+00, 0.88091087E+00, 0.25278471E+01, 0.36680901E+00, 0.36790523E+00, -0.11417954E+00, 0.26816338E+00, 
    },
    {
      -0.24069045E-01, -0.37140232E-02, 0.59421258E-02, -0.14975986E-01, 0.38167016E-03, 0.24848035E-01, -0.14859644E-01, 0.81578468E-03, -0.90723382E-02, 
    },
    {
      -0.75797653E-02, -0.47509167E-01, -0.53818156E-02, 0.15393181E-01, -0.22171626E-01, 0.20891447E-01, 0.37496552E-01, -0.90530521E-03, -0.27469555E-02, 
    },
    {
      0.87826133E-01, 0.61553407E-01, 0.20494960E-01, -0.68606257E-01, 0.31154953E-01, -0.22373168E-01, 0.19868614E-01, -0.36711791E-02, 0.80109155E-02, 
    },
    {
      0.58815431E-01, -0.10893079E-02, -0.23224903E-01, -0.16784027E-01, -0.12939301E-02, -0.17937096E-01, -0.30362142E-01, -0.11328247E-01, 0.18137729E-01, 
    },
    {
      -0.83675431E-02, -0.66944212E-01, -0.99108607E-01, 0.53046636E-01, 0.11491635E-02, -0.97788624E-01, -0.19152731E+00, -0.73964251E-02, 0.37721645E-01, 
    },
    {
      -0.16437825E-01, 0.10426622E+00, -0.13097627E-01, -0.31341720E-01, 0.33724148E-01, 0.20263115E+00, -0.17441693E+00, -0.13966219E-01, 0.81566256E-03, 
    },
    {
      -0.30277975E-01, -0.14341920E-01, 0.34723729E-02, 0.66115148E-02, -0.40517421E-02, 0.80762738E-02, 0.77816513E-02, 0.50833821E-01, 0.10282873E-01, 
    },
    {
      -0.26500834E-01, 0.64238147E-02, -0.28022844E-02, -0.14088667E-01, -0.54610550E-03, 0.28645764E-02, 0.12587353E-01, -0.20940429E-01, 0.47031794E-01, 
    },
    {
      0.57847179E-01, 0.32054439E-01, 0.12395522E-01, -0.61136406E-01, -0.11976909E-01, 0.27411242E-03, 0.54735653E-02, 0.39643612E-01, 0.14503116E-01, 
    },
    {
      -0.14846103E-01, -0.43443698E-01, -0.62784306E-02, 0.66622496E-01, 0.10906368E-01, -0.16276255E-01, -0.74495091E-02, -0.27881341E-01, 0.35522427E-01, 
    },
    {
      -0.26395649E-02, -0.95528336E-02, -0.20187974E-02, 0.48234980E-02, -0.36433239E-02, 0.73964102E-02, 0.43250853E-02, -0.13721441E-04, -0.82670292E-03, 
    },
    {
      0.13747018E-01, 0.26495285E-01, -0.12389235E-01, 0.13934656E-03, -0.12922279E-01, 0.92590265E-02, 0.30354068E-02, 0.11110849E-02, 0.40769838E-02, 
    },
    {
      0.33883248E-02, 0.76762335E-02, 0.87840520E-02, 0.22252684E-02, -0.17388515E-02, 0.65976731E-02, -0.13436445E-02, 0.53181122E-02, 0.17218638E-02, 
    },
    {
      0.12250467E-01, 0.88767400E-02, -0.20778641E-01, 0.91415541E-02, -0.31969624E-02, 0.82466059E-03, 0.72336243E-02, -0.23431671E-02, 0.52077025E-02, 
    },
  },
  {
    {
      0.27045255E+01, -0.13861433E-01, 0.10853585E-01, 0.13674901E-01, 0.45255958E-02, 0.96232325E-04, 0.28539707E-02, -0.50630595E-04, -0.10792224E-02, 
    },
    {
      0.33882726E-01, 0.16726006E-01, 0.41411079E-01, 0.21762999E-01, 0.13571894E-01, -0.18095523E-01, -0.13972076E-01, -0.26165762E-02, -0.44640922E-03, 
    },
    {
      0.13936986E+01, 0.10243688E+00, -0.12836154E+00, -0.18260194E+00, -0.46525557E-01, 0.44762935E-01, 0.57317317E-02, 0.11237833E-01, 0.40574851E-02, 
    },
    {
      -0.14981104E+00, -0.25437304E+00, -0.14702123E+00, -0.51582754E-01, -0.47143351E-01, 0.47940180E-01, 0.27028264E-01, 0.77501182E-02, -0.80215288E-02, 
    },
    {
      -0.24028523E+01, -0.13707371E+00, 0.35199860E+00, 0.39832273E+00, 0.85571006E-01, -0.12622128E+00, -0.67097712E-02, -0.48906956E-01, 0.55381274E-02, 
    },
    {
      0.77955551E-01, 0.19398201E+00, 0.13605739E+00, 0.29710919E-01, 0.28151933E-01, -0.32603387E-01, -0.11237349E-01, -0.34127152E-02, 0.13379143E-01, 
    },
    {
      0.11976627E+01, 0.86586386E-01, -0.22247823E+00, -0.23125169E+00, -0.42487212E-01, 0.83616689E-01, 0.73067646E-03, 0.40582355E-01, -0.96359570E-02, 
    },
    {
      0.17483035E-01, -0.30285546E+00, -0.29345882E+00, 0.99322964E-02, -0.17893454E-02, 0.10022582E-02, 0.11504443E-02, 0.16644030E-02, 0.72499416E-02, 
    },
    {
      0.13540727E-02, 0.32227382E+00, -0.29328412E+00, -0.80812844E-02, 0.39960979E-02, 0.65438594E-02, -0.11646176E-01, 0.39998051E-02, 0.15032205E-01, 
    },
    {
      -0.36074266E-01, 0.11459168E+00, 0.90131857E-01, 0.35216373E-02, -0.96506746E-02, -0.11520932E-01, -0.12881164E-01, 0.99653052E-03, 0.34462839E-01, 
    },
    {
      0.89647539E-01, 0.63354815E-02, 0.12553091E+00, 0.24806114E-01, 0.23215953E-01, 0.13266702E-01, 0.39352668E-02, 0.90735722E-02, 0.86630374E-01, 
    },
    {
      -0.22010055E+00, 0.76199836E+00, 0.10411052E+01, 0.79058155E-01, -0.99691860E-01, -0.41243132E-01, 0.57794455E-01, -0.16476742E-02, -0.94741918E-01, 
    },
    {
      -0.32839933E+00, -0.12193962E+01, 0.87339061E+00, -0.75224414E-02, 0.83490223E-01, -0.14356729E+00, 0.13259062E+00, -0.44737361E-01, -0.20198126E+00, 
    },
    {
      0.54963994E+00, -0.52403355E+00, -0.74558896E+00, -0.96110024E-01, 0.27973866E+00, 0.15445522E-01, -0.24832048E+00, 0.65243810E-01, -0.22887573E+00, 
    },
    {
      -0.19111110E-01, 0.47740996E+00, -0.63515687E+00, -0.18528543E+00, -0.21004455E+00, 0.61389271E-01, 0.34061927E-01, -0.37664600E-01, -0.56488287E+00, 
    },
    {
      0.61485970E+00, -0.31153601E+00, -0.85383409E+00, -0.55827850E+00, 0.24451426E+00, 0.20632704E+00, -0.12735528E+00, 0.49136013E-01, 0.29886141E+00, 
    },
    {
      0.12342763E+01, 0.80860949E+00, -0.12855700E+01, 0.31017226E+00, -0.24868594E+00, 0.44440651E+00, -0.30813733E+00, 0.23854533E-01, 0.45323089E+00, 
    },
    {
      -0.15821571E+01, 0.43670562E+00, 0.22294593E+01, 0.54407924E+00, -0.86818773E+00, -0.27222033E-02, 0.87385136E+00, -0.10560450E+00, 0.39706621E+00, 
    },
    {
      -0.65258825E+00, -0.15170186E+01, 0.12435969E+01, 0.19180834E+00, 0.70927751E+00, -0.27004617E+00, -0.20135774E+00, 0.98844670E-01, 0.11511528E+01, 
    },
    {
      -0.35777101E+00, -0.10984187E+00, 0.54976183E+00, 0.51074892E+00, -0.15392701E+00, -0.20111597E+00, 0.95634453E-01, -0.59104066E-01, -0.22803462E+00, 
    },
    {
      -0.10545499E+01, -0.27652279E-01, 0.84639686E+00, -0.34254837E+00, 0.83599433E-01, -0.39442429E+00, 0.17786869E+00, 0.49440950E-01, -0.29044276E+00, 
    },
    {
      0.11647863E+01, 0.19355892E+00, -0.19237444E+01, -0.51354492E+00, 0.63591230E+00, 0.36586221E-01, -0.67268568E+00, 0.16708907E-01, -0.21348131E+00, 
    },
    {
      0.78157800E+00, 0.13963382E+01, -0.70771319E+00, -0.20638825E-01, -0.56146246E+00, 0.25821999E+00, 0.18338437E+00, -0.10435095E+00, -0.70114332E+00, 
    },
    {
      0.33377009E-02, -0.85557243E-02, 0.53263279E-02, 0.48212931E-01, -0.96097551E-01, 0.16747259E-02, 0.45310468E-02, 0.21136762E-02, -0.20766114E-02, 
    },
    {
      -0.66840369E-02, -0.10856335E-01, -0.23787534E-02, 0.92386574E-01, 0.68768784E-01, 0.10402149E-02, -0.86118113E-02, 0.32925498E-03, 0.57025033E-03, 
    },
    {
      0.51790807E-01, 0.13119574E+00, 0.47563083E-01, 0.28507447E+00, -0.11513667E+00, 0.31583652E-01, -0.17520612E-01, -0.31817120E-01, 0.82542188E-02, 
    },
    {
      0.19457083E-01, 0.16882727E-01, -0.44377230E-01, 0.21419054E+00, 0.39532006E+00, 0.44337548E-01, 0.33838246E-01, -0.26187995E-01, 0.26135845E-01, 
    },
    {
      -0.13655464E+00, 0.12530467E+00, -0.43442167E-01, -0.23065427E+00, 0.20435360E+00, -0.63712358E-01, 0.23470277E-01, -0.50777260E-01, 0.77459794E-02, 
    },
    {
      -0.15135796E+00, 0.11079714E-01, -0.65361388E-01, -0.20480011E+00, -0.31951770E+00, -0.43420333E-01, 0.58943622E-01, -0.19627888E-01, -0.11768199E-01, 
    },
    {
      -0.15009120E+00, -0.68192893E+00, 0.62561356E-01, -0.10389147E+01, 0.10595896E+00, -0.14849147E-01, 0.53832039E-01, 0.23531134E+00, -0.12635956E-01, 
    },
    {
      0.28327590E+00, 0.38813151E-01, 0.31918576E+00, -0.46337992E+00, -0.14411012E+01, -0.15281472E+00, -0.21893854E+00, 0.14097042E+00, -0.14544623E+00, 
    },
    {
      0.15800914E+00, -0.34267643E+00, 0.18755988E-02, -0.56562793E-01, -0.48462987E+00, 0.95949911E-01, -0.30947629E-01, 0.87747708E-01, -0.19303991E-01, 
    },
    {
      0.24605586E+00, -0.13650826E+00, 0.21064965E-01, 0.58473659E+00, -0.75870594E-02, 0.39007965E-01, -0.65296404E-01, 0.41345578E-01, -0.19478334E-01, 
    },
    {
      0.78472018E-01, 0.91951132E+00, -0.16342767E+00, 0.91248029E+00, 0.87954272E-02, -0.59618726E-02, -0.40209666E-01, -0.31057978E+00, 0.18426050E-01, 
    },
    {
      -0.60459119E+00, -0.10414231E-01, -0.18814395E+00, 0.27712169E+00, 0.13081124E+01, 0.17085785E+00, 0.34071115E+00, -0.14231364E+00, 0.17387938E+00, 
    },
    {
      -0.20386552E-01, 0.21182620E-02, 0.10177098E-01, -0.11378626E-01, -0.98456312E-02, 0.34398962E-01, -0.42581428E-01, 0.34251876E-03, -0.68609137E-02, 
    },
    {
      -0.90750623E-02, -0.32029830E-01, -0.47561587E-02, 0.14946333E-01, -0.16567832E-01, 0.42375118E-01, 0.41583791E-01, 0.13377547E-03, -0.61774603E-02, 
    },
    {
      0.45967977E-01, 0.61088778E-01, 0.14942785E-01, -0.37811134E-01, 0.15868828E-01, 0.47008302E-02, 0.46885707E-02, 0.11044603E-02, 0.83790062E-03, 
    },
    {
      0.33965636E-01, 0.12285841E-01, -0.15363421E-01, -0.20721937E-01, -0.11257234E-01, -0.34942648E-02, -0.12950346E-01, -0.59640319E-02, 0.12831651E-01, 
    },
    {
      0.60491147E-03, -0.74729361E-01, -0.69819920E-01, 0.26307790E-01, 0.17833915E-01, -0.10292423E+00, -0.14374946E+00, -0.44204816E-02, 0.32788616E-01, 
    },
    {
      -0.15078791E-01, 0.62297020E-01, -0.27057173E-01, -0.22026932E-01, 0.26395356E-01, 0.16220757E+00, -0.12398390E+00, -0.14915660E-01, 0.10779776E-01, 
    },
    {
      -0.22808822E-01, -0.15335506E-01, 0.21026686E-02, 0.44900649E-02, -0.13017196E-02, 0.27200764E-02, 0.10563135E-02, 0.33134330E-01, 0.16995138E-01, 
    },
    {
      -0.15292642E-01, 0.24188119E-02, -0.86551392E-02, -0.19452726E-02, -0.34849253E-02, 0.35679105E-02, 0.10176241E-01, -0.22009404E-01, 0.29706346E-01, 
    },
    {
      0.43338250E-01, 0.19471711E-01, 0.19346902E-03, -0.33331003E-01, -0.12334583E-01, -0.12015219E-01, 0.14972220E-02, 0.39651532E-01, 0.19308904E-01, 
    },
    {
      0.14977311E-02, -0.34002740E-01, 0.13617956E-02, 0.26084950E-01, 0.37475578E-02, -0.10403191E-01, -0.17990541E-01, -0.22997433E-01, 0.27994959E-01, 
    },
    {
      0.22576404E-03, -0.16536338E-01, -0.75725868E-03, 0.69075981E-02, -0.14428189E-02, 0.61770566E-02, 0.60216250E-03, -0.43525919E-03, -0.13888464E-02, 
    },
    {
      0.90428498E-02, 0.19430207E-01, -0.10300137E-01, 0.35099406E-02, -0.42426982E-02, 0.60566198E-02, 0.52108942E-02, -0.11467599E-03, 0.65590143E-02, 
    },
    {
      0.49651740E-02, 0.11645567E-01, 0.83609400E-02, 0.52899164E-02, -0.47522220E-02, 0.55126837E-02, 0.35592569E-02, 0.34613642E-02, 0.36597401E-02, 
    },
    {
      0.45108609E-02, 0.48094862E-02, -0.14215787E-01, 0.53761145E-02, 0.45948825E-02, 0.83546591E-03, 0.53303521E-02, -0.26749300E-02, 0.29056252E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir20_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.67417641E+01, -0.11025394E+00, 0.42666834E-01, 0.69521941E-01, 0.38874124E-02, 0.29258806E-01, -0.31417802E-01, 0.35931028E-01, 0.11918333E-01, 0.29975636E-01, 0.11452777E-01, -0.29228499E-01, -0.76055541E-02, 
    },
    {
      0.17436563E+01, 0.40657678E+00, 0.59440559E+00, -0.81161804E-01, -0.12061444E+00, -0.42257652E-01, -0.16804051E-01, -0.13078281E+00, 0.99271558E-01, -0.51369164E-01, -0.85818343E-01, 0.20449582E-01, -0.34776747E-01, 
    },
    {
      0.15906835E+02, 0.21122408E+01, -0.30194516E+01, -0.18164214E+00, 0.21544621E+01, -0.77579105E+00, 0.10776812E+01, -0.65275151E+00, -0.18329655E+00, -0.10741085E+01, -0.21992724E+00, 0.54272276E+00, 0.26474300E+00, 
    },
    {
      -0.21567724E+02, -0.10071519E+02, -0.46050444E+01, 0.22639155E+01, 0.77192564E+01, 0.70966089E+00, -0.21318839E+01, 0.47265267E+00, -0.24995817E-01, 0.65379196E+00, 0.20589564E+01, -0.17014968E+00, 0.56020799E-02, 
    },
    {
      -0.10173342E+03, -0.68225641E+01, 0.18339714E+02, -0.96500959E+01, -0.22774458E+02, 0.70135794E+01, -0.57714853E+01, 0.20234113E+01, 0.98741543E+00, 0.70813155E+01, 0.75227982E+00, -0.31741402E+01, -0.28904328E+01, 
    },
    {
      0.65307861E+02, 0.35732182E+02, 0.99326601E+01, -0.70676713E+01, -0.35474052E+02, -0.48456573E+01, 0.10269098E+02, 0.14416782E+01, -0.37578681E+01, -0.17822330E+01, -0.10741420E+02, 0.18747391E+01, 0.18222950E+01, 
    },
    {
      0.20845084E+03, 0.83894491E+01, -0.37941162E+02, 0.42702366E+02, 0.68335434E+02, -0.22162090E+02, 0.11464567E+02, -0.11181591E+01, -0.30024517E+01, -0.17626598E+02, -0.13377714E+00, 0.74559193E+01, 0.94865475E+01, 
    },
    {
      -0.69698105E+02, -0.40740063E+02, -0.12088245E+02, 0.55066094E+01, 0.60161949E+02, 0.12805878E+02, -0.14906711E+02, -0.69451771E+01, 0.13351114E+02, 0.15886544E+01, 0.22924721E+02, -0.65486469E+01, -0.64081602E+01, 
    },
    {
      -0.19967574E+03, -0.47658625E+01, 0.32122219E+02, -0.59845806E+02, -0.79271217E+02, 0.26973694E+02, -0.94852943E+01, -0.13761377E+01, 0.42310858E+01, 0.18761662E+02, -0.14811943E+01, -0.75402808E+01, -0.11900676E+02, 
    },
    {
      0.14291371E+02, 0.10281889E+02, 0.12892766E+02, 0.20141015E+01, -0.43246784E+02, -0.13862740E+02, 0.66849313E+01, 0.83780088E+01, -0.16522486E+02, -0.29590270E+00, -0.21744144E+02, 0.82283382E+01, 0.76309428E+01, 
    },
    {
      0.74218613E+02, 0.12529610E+01, -0.98135452E+01, 0.27000690E+02, 0.31554647E+02, -0.11102576E+02, 0.27184381E+01, 0.10648201E+01, -0.20641975E+01, -0.71944165E+01, 0.10889814E+01, 0.27322817E+01, 0.50644026E+01, 
    },
    {
      0.97484446E+01, 0.43847666E+01, -0.68491726E+01, -0.25755978E+01, 0.10981911E+02, 0.51441727E+01, 0.12638378E+00, -0.31936567E+01, 0.68834958E+01, -0.72948217E-01, 0.75957398E+01, -0.34138052E+01, -0.30203903E+01, 
    },
    {
      -0.17938741E-01, 0.19350545E+01, 0.15287569E+01, 0.26359454E-01, -0.69553903E-02, 0.48199920E-02, 0.15078617E-01, 0.53562142E-01, 0.21571161E-01, 0.32001279E-01, 0.37023298E-01, -0.33349071E-01, 0.10461510E-01, 
    },
    {
      -0.71816571E-01, -0.14951198E+01, 0.19070232E+01, -0.90120353E-01, 0.83764896E-01, -0.22245713E-01, -0.61692059E-01, 0.29028347E-01, -0.15067049E-03, 0.12577567E-01, -0.22862754E-02, 0.29112538E-01, -0.23672460E-01, 
    },
    {
      -0.51051003E+00, 0.87592429E+00, 0.11411304E+01, 0.22836588E+00, 0.10007082E+00, -0.85580871E-02, 0.14482203E+00, 0.23493987E+00, 0.12527980E+00, -0.16819620E+00, -0.48568074E-01, 0.24008988E+00, 0.96231982E-01, 
    },
    {
      0.10973550E+00, 0.93418971E-01, -0.70641696E-01, -0.53768361E+00, 0.48843423E+00, 0.34973875E+00, -0.26007736E+00, 0.36855582E-01, 0.41532871E+00, 0.89809448E-01, -0.23388444E-01, 0.20598353E+00, 0.13492872E+00, 
    },
    {
      0.54216437E+01, 0.19527584E+02, 0.53556581E+01, -0.20208032E+01, 0.27149031E+01, 0.11244015E+01, -0.42923173E+00, -0.90001625E+00, -0.22988829E+00, -0.79169679E+00, 0.95905103E-01, 0.54770768E+00, 0.40796738E-01, 
    },
    {
      -0.49918337E+01, -0.68213148E+01, 0.20141775E+02, 0.33453064E+01, 0.20416534E+01, -0.18846661E+00, 0.78771967E+00, 0.12231276E+01, 0.63756788E+00, -0.58407247E+00, -0.35017237E+00, -0.12065351E+01, -0.24117194E+00, 
    },
    {
      0.15664848E+02, -0.58232422E+01, 0.16471388E+01, -0.85785646E+01, 0.65734754E+01, 0.10979421E+01, -0.61561337E+01, -0.41306434E+01, -0.25767187E-01, 0.22381325E+01, 0.16197391E+01, -0.47094798E+01, -0.19486105E+01, 
    },
    {
      0.54404092E+01, -0.23112692E+02, 0.87230244E+01, 0.41379757E+01, -0.82567225E+01, -0.59944463E+00, 0.49454799E+01, 0.98528072E-01, -0.69704366E+01, -0.26175516E+01, 0.21983194E+01, -0.85137612E+00, -0.48446989E+01, 
    },
    {
      -0.34786564E+02, -0.13385123E+03, 0.53099270E+02, 0.98495770E+01, -0.21858513E+02, -0.12796616E+02, 0.21972885E+01, 0.71941557E+01, 0.17378716E+01, 0.38053644E+01, -0.41692481E+01, -0.43841825E+01, -0.76313013E+00, 
    },
    {
      0.52958324E+02, -0.62617264E+02, -0.13794221E+03, -0.20468203E+02, -0.26598701E+02, 0.76006281E+00, -0.45324337E+00, -0.73798909E+01, -0.68499522E+01, 0.43220587E+01, 0.26352139E+01, 0.11029831E+02, 0.39386227E+01, 
    },
    {
      -0.10278553E+03, -0.11709525E+02, 0.43092846E+02, 0.72482162E+02, -0.24375433E+02, -0.46230431E+01, 0.37790916E+02, 0.18252401E+02, 0.32258278E+00, -0.98822670E+01, -0.65691562E+01, 0.26157690E+02, 0.14764631E+02, 
    },
    {
      -0.63290554E+02, 0.11773388E+03, -0.50047344E+02, -0.31375063E+02, 0.31207630E+02, -0.13806309E+02, -0.21457752E+02, -0.76790338E+01, 0.33817787E+02, 0.16895014E+02, -0.18257301E+02, -0.33989453E+01, 0.31438547E+02, 
    },
    {
      0.73195847E+02, 0.30527811E+03, -0.27625671E+03, -0.13173969E+02, 0.52327705E+02, 0.43725922E+02, -0.53056192E+01, -0.26080379E+02, -0.80343571E+01, -0.59193101E+01, 0.15645749E+02, 0.14612919E+02, 0.19771434E+01, 
    },
    {
      -0.16780865E+03, 0.33273395E+03, 0.31980917E+03, 0.61927944E+02, 0.10291243E+03, -0.41476986E+00, -0.10810330E+02, 0.16019861E+02, 0.22894283E+02, -0.96919661E+01, -0.76568937E+01, -0.35144093E+02, -0.13620372E+02, 
    },
    {
      0.27711322E+03, 0.76293732E+02, -0.24416869E+03, -0.23412332E+03, 0.59856181E+01, 0.27036533E+01, -0.85468987E+02, -0.32569973E+02, -0.93770189E+01, 0.22770227E+02, 0.79303355E+01, -0.62524040E+02, -0.43537697E+02, 
    },
    {
      0.18402339E+03, -0.25016977E+03, 0.72703697E+02, 0.10605528E+03, -0.51503254E+02, 0.54460190E+02, 0.38909683E+02, 0.36796562E+02, -0.68024529E+02, -0.41103397E+02, 0.53216114E+02, 0.20626389E+02, -0.76586349E+02, 
    },
    {
      -0.53355736E+02, -0.31135284E+03, 0.40809729E+03, 0.22601173E+01, -0.48498844E+02, -0.60832169E+02, 0.60337324E+01, 0.38279335E+02, 0.14871668E+02, 0.28498704E+01, -0.20230135E+02, -0.19879827E+02, -0.18497735E+01, 
    },
    {
      0.21465652E+03, -0.48994763E+03, -0.32978488E+03, -0.86516556E+02, -0.15399602E+03, -0.53305179E+00, 0.26561083E+02, -0.14605586E+02, -0.29598482E+02, 0.74684277E+01, 0.98743944E+01, 0.45331917E+02, 0.17090784E+02, 
    },
    {
      -0.33723950E+03, -0.97724586E+02, 0.38985550E+03, 0.31390698E+03, 0.46843403E+02, 0.78864837E+01, 0.82173851E+02, 0.24357899E+02, 0.20966228E+02, -0.26853407E+02, -0.14723213E+01, 0.67029839E+02, 0.53401054E+02, 
    },
    {
      -0.18740387E+03, 0.25239400E+03, -0.10743400E+02, -0.13630225E+03, 0.44591602E+02, -0.69500046E+02, -0.34749329E+02, -0.57859837E+02, 0.60593739E+02, 0.43181362E+02, -0.64212608E+02, -0.31543680E+02, 0.79737732E+02, 
    },
    {
      0.79219427E+01, 0.12230792E+03, -0.19162122E+03, 0.30479712E+01, 0.14701435E+02, 0.29905254E+02, -0.25530334E+01, -0.18944752E+02, -0.87824144E+01, -0.98065168E-01, 0.87094259E+01, 0.93237696E+01, 0.71487147E+00, 
    },
    {
      -0.98071724E+02, 0.22642126E+03, 0.13037717E+03, 0.42995682E+02, 0.77473969E+02, 0.44055769E+00, -0.17098782E+02, 0.44686174E+01, 0.12881230E+02, -0.12471923E+01, -0.46582799E+01, -0.20446924E+02, -0.70700450E+01, 
    },
    {
      0.15115994E+03, 0.37530956E+02, -0.19583463E+03, -0.14785715E+03, -0.36896564E+02, -0.74833941E+01, -0.28641958E+02, -0.58377399E+01, -0.12369586E+02, 0.12254857E+02, -0.14293088E+01, -0.26398525E+02, -0.23067722E+02, 
    },
    {
      0.59227791E+02, -0.99076767E+02, -0.24037251E+02, 0.57410080E+02, -0.17986750E+02, 0.29093678E+02, 0.13013299E+02, 0.29289719E+02, -0.19803263E+02, -0.16676153E+02, 0.27448059E+02, 0.15189789E+02, -0.29994450E+02, 
    },
    {
      0.11391176E+00, 0.11553436E+00, -0.13704097E+00, -0.10008467E+01, -0.24247713E-01, 0.66881222E-02, -0.69679026E-02, -0.18138342E-01, 0.27570529E-01, 0.35871942E-01, 0.73015732E-02, -0.69541764E-02, 0.26003541E-01, 
    },
    {
      -0.63337181E-02, 0.59983626E-01, -0.62968001E-01, -0.78753326E-02, -0.75065094E+00, -0.37398331E-01, -0.44260029E-01, -0.36373198E-01, 0.74006170E-02, -0.10022034E-02, 0.77789617E-02, 0.12514464E-01, -0.10064091E-01, 
    },
    {
      0.86539829E+00, 0.32539025E+00, -0.91669343E-01, -0.73207444E+00, 0.26933301E-01, 0.13879360E+00, -0.26311162E+00, -0.37850922E+00, -0.21304099E+00, -0.14258491E+00, -0.13731001E+00, -0.30468129E-01, 0.92138372E-01, 
    },
    {
      0.98262489E+00, -0.15070742E+00, 0.71532860E-01, -0.49188623E+00, -0.57117575E+00, 0.43504018E+00, 0.19652760E+00, -0.13652391E-02, -0.67386083E-01, -0.13397166E+00, -0.20012897E+00, 0.14705022E+00, 0.10473770E-01, 
    },
    {
      0.65290123E+00, -0.26946282E+01, 0.31813004E+01, 0.30024579E+01, -0.75441279E-01, -0.15592440E+01, -0.21163237E+00, -0.12697160E+00, -0.19921806E+00, -0.66176975E+00, 0.10346518E+00, -0.18251108E+00, 0.17797954E-01, 
    },
    {
      0.18768592E+00, -0.25953684E+01, 0.43567185E+01, -0.59679967E+00, 0.19172314E+01, -0.20841491E+00, 0.28366706E+00, 0.23806677E+01, -0.16260111E+00, 0.13395765E+00, -0.83577555E+00, -0.23089887E+00, -0.83205447E-03, 
    },
    {
      -0.53307676E+01, 0.61688054E+00, 0.14718913E+01, 0.10001513E+02, 0.29644825E+01, -0.11001223E+01, 0.14558415E+01, 0.26624279E+01, 0.21787982E+01, 0.13549827E+01, 0.16713557E+01, -0.15284111E+00, -0.52187556E+00, 
    },
    {
      -0.72269955E+01, 0.91789746E+00, 0.22233882E+01, -0.45572978E+00, 0.10729968E+02, -0.29951901E+01, -0.21556865E+00, 0.35207564E+00, 0.13136928E+01, 0.57696944E+00, 0.23771760E+01, -0.17125853E+01, -0.86974633E+00, 
    },
    {
      -0.36376338E+01, 0.24080009E+02, -0.22185778E+02, 0.18825312E+02, 0.14852665E+02, 0.13149285E+02, 0.78031552E+00, 0.31474850E+01, 0.90485042E+00, 0.16362022E+01, -0.48030415E+00, 0.18465122E+01, -0.83016783E+00, 
    },
    {
      -0.18757300E+01, 0.21337709E+02, -0.17421923E+02, -0.10823549E+02, 0.15082483E+02, -0.15371397E+01, -0.41927969E+00, -0.11322726E+02, 0.11449955E+01, -0.39108786E+00, 0.54798636E+01, 0.16054937E+01, 0.82940835E+00, 
    },
    {
      0.90842657E+01, -0.63919301E+01, -0.66001620E+01, -0.23100986E+02, -0.12172362E+02, 0.22952433E+01, -0.42773733E+01, -0.51716614E+01, -0.57398787E+01, -0.32952104E+01, -0.44157414E+01, 0.97596109E+00, 0.49134457E+00, 
    },
    {
      0.14462518E+02, -0.52479367E+01, -0.77367229E+01, 0.63962259E+01, -0.26448858E+02, 0.69845829E+01, -0.13445467E+01, -0.86556637E+00, -0.37851818E+01, -0.59372693E+00, -0.62924967E+01, 0.47564263E+01, 0.29345119E+01, 
    },
    {
      0.36482801E+01, -0.54122810E+02, 0.51179451E+02, -0.67288895E+02, -0.36518002E+02, -0.31562344E+02, -0.62648946E+00, -0.97915506E+01, -0.14420407E+01, -0.11084070E+01, 0.88024229E+00, -0.53258376E+01, 0.19962699E+01, 
    },
    {
      0.75823536E+01, -0.47701122E+02, 0.20454948E+02, 0.31747019E+02, -0.50241886E+02, 0.81888828E+01, 0.43072930E+00, 0.17209768E+02, -0.32296555E+01, -0.40862790E+00, -0.11610971E+02, -0.27689962E+01, -0.20991716E+01, 
    },
    {
      -0.50960884E+01, 0.83177357E+01, 0.55298500E+01, 0.15825745E+02, 0.13216482E+02, -0.14331723E+01, 0.36789041E+01, 0.29861968E+01, 0.41942501E+01, 0.21892414E+01, 0.33499987E+01, -0.10724663E+01, 0.32862183E-01, 
    },
    {
      -0.85228519E+01, 0.73134866E+01, 0.61695185E+01, -0.86385641E+01, 0.17545185E+02, -0.52412457E+01, 0.18066072E+01, 0.71043044E+00, 0.29276321E+01, -0.61344326E-01, 0.45447369E+01, -0.35246754E+01, -0.23307092E+01, 
    },
    {
      -0.14690654E+01, 0.35724678E+02, -0.35846252E+02, 0.50520844E+02, 0.23054344E+02, 0.21617199E+02, 0.42514130E+00, 0.79941406E+01, 0.58302629E+00, 0.34610984E+00, -0.60215044E+00, 0.39643900E+01, -0.13188493E+01, 
    },
    {
      -0.92348213E+01, 0.30514215E+02, -0.52573867E+01, -0.22224283E+02, 0.36993191E+02, -0.71565070E+01, -0.77821231E+00, -0.85328360E+01, 0.27982295E+01, 0.89062178E+00, 0.76935654E+01, 0.13601083E+01, 0.11444426E+01, 
    },
    {
      -0.74351348E-01, 0.11168007E+00, -0.12678429E+00, -0.30625785E-01, 0.81547037E-01, -0.26790616E+00, -0.77349359E+00, 0.18570652E-01, 0.38729448E-01, 0.77069057E-02, 0.12018093E-01, -0.19658063E-01, -0.86077964E-02, 
    },
    {
      0.12409092E+00, -0.10666509E+00, 0.45101717E-01, 0.15352032E-01, 0.32349147E-01, 0.78267974E+00, -0.29498765E+00, -0.52898012E-01, 0.61348882E-01, 0.11466916E-01, 0.13180029E-02, 0.22439856E-01, 0.54106647E-02, 
    },
    {
      0.23398851E+00, 0.64493269E+00, -0.25810570E+00, -0.12363273E+00, -0.66201246E+00, -0.27819201E+00, -0.30108163E+00, -0.24624993E+00, 0.83835699E-01, -0.74149808E-02, 0.13186117E+00, -0.16377572E-01, -0.61401571E-02, 
    },
    {
      -0.10936357E+00, -0.29475041E-01, 0.53700852E+00, 0.50398592E-01, -0.11941080E+00, 0.84602892E+00, -0.45599923E+00, -0.29877833E+00, -0.10263894E+00, -0.88044167E-01, -0.85591517E-01, 0.64158496E-02, 0.54729246E-02, 
    },
    {
      0.34892559E+00, 0.13616563E-01, 0.68630129E+00, 0.92597771E+00, -0.78667665E+00, 0.75528562E+00, 0.13148088E+01, 0.89991748E-01, -0.31526265E+00, 0.76205105E-01, -0.15212275E+00, 0.11211587E+00, 0.32542087E-01, 
    },
    {
      -0.11553695E+01, 0.24110079E+01, -0.11516421E+01, 0.59397310E+00, -0.87734061E+00, -0.15284475E+01, 0.15883173E+01, 0.54522204E+00, -0.40533349E+00, 0.17719109E-01, -0.13833664E+00, -0.10189562E+00, -0.79786859E-01, 
    },
    {
      -0.51425898E+00, -0.15127133E+01, 0.96630716E+00, 0.11464612E+00, 0.11833749E+01, 0.31068289E+00, -0.74194200E-01, 0.72967577E+00, -0.67832701E-01, -0.77846984E-03, -0.20988442E+00, -0.51187862E-01, 0.38249537E-01, 
    },
    {
      -0.78501157E-01, 0.32682669E+00, -0.10028919E+01, -0.68409115E+00, 0.19638067E+00, -0.12582335E+01, 0.71608400E+00, 0.40842831E+00, 0.26023892E+00, 0.20802367E+00, 0.18907779E+00, -0.95071457E-01, -0.14524133E-01, 
    },
    {
      0.70622310E-01, -0.23873582E-01, -0.46930566E+00, -0.11361483E+01, 0.12044519E+01, -0.13642393E+01, -0.16782405E+01, -0.27571243E+00, 0.62265939E+00, -0.27575913E+00, 0.26088655E+00, -0.17131169E+00, -0.96791862E-02, 
    },
    {
      0.16397073E+01, -0.35624073E+01, 0.21964977E+01, -0.69169527E+00, 0.17876031E+01, 0.23638010E+01, -0.29829438E+01, -0.65094346E+00, 0.50062829E+00, -0.78845382E-01, 0.20135802E+00, 0.35707735E-01, 0.43332722E-01, 
    },
    {
      -0.73506460E-02, -0.20373339E-01, -0.59456583E-02, 0.47092013E-01, 0.59794255E-01, -0.22777202E-01, 0.34936242E-01, 0.28207561E+00, -0.12549119E+00, 0.68888762E-02, 0.73659769E-02, -0.16957685E-01, -0.69591464E-02, 
    },
    {
      0.60888998E-01, 0.48170742E-01, 0.90240315E-02, -0.23745389E-02, -0.88016756E-01, 0.73502366E-02, -0.11265669E-02, 0.12217470E+00, 0.24448656E+00, -0.23252072E-02, -0.70067816E-02, -0.43141777E-02, -0.13638792E-01, 
    },
    {
      -0.21918522E-01, 0.39478119E-01, 0.21353072E-01, -0.17231616E+00, -0.24383087E+00, -0.18098043E+00, 0.78112222E-01, -0.96717022E-01, -0.21291029E+00, -0.78414723E-01, 0.27862417E-01, 0.96148206E-02, 0.23770265E-01, 
    },
    {
      -0.13406623E-01, -0.97504027E-01, -0.22518177E-01, 0.13743083E+00, 0.13689157E+00, -0.19897047E-01, -0.10379756E+00, 0.18668231E+00, -0.28371491E-01, -0.66896580E-01, -0.45131464E-01, 0.52789491E-01, -0.40041968E-01, 
    },
    {
      -0.66618256E-01, 0.42511921E-01, -0.38794965E-01, -0.10955274E-01, 0.30871356E-01, 0.52086008E-02, 0.24081491E-01, -0.67935861E-02, -0.13217943E-01, 0.82531139E-01, 0.28200027E+00, -0.76102950E-02, -0.90683848E-02, 
    },
    {
      0.23629055E-02, -0.10296953E-02, -0.14055242E-01, -0.31999350E-01, -0.57006430E-01, 0.18182561E-01, -0.55214246E-02, 0.49740784E-02, -0.54625701E-01, -0.27094826E+00, 0.51207963E-01, 0.45274422E-02, 0.16952408E-01, 
    },
    {
      0.41054886E-01, -0.54486718E-01, -0.10627154E-01, -0.28977985E-01, 0.76224119E-02, 0.58471583E-01, -0.90225777E-02, 0.22696028E-01, -0.16175291E-01, 0.21533300E-02, 0.36474864E-02, -0.14331755E+00, -0.11911416E-01, 
    },
    {
      -0.35727635E-01, -0.25858898E-01, -0.81544844E-02, 0.45217113E-02, 0.82371887E-02, 0.27359465E-01, -0.12122327E-01, -0.90430006E-02, -0.25242612E-01, -0.10398197E-01, -0.11534831E-01, 0.29744126E-01, -0.14588501E+00, 
    },
    {
      -0.27973065E-01, -0.17960723E-01, -0.72880231E-01, -0.13211422E-01, 0.74935732E-02, 0.40451571E-01, -0.13288229E-02, 0.23200864E-01, -0.21652490E-01, 0.10630924E-01, -0.19871013E-01, -0.89811655E-02, 0.44909436E-02, 
    },
    {
      0.15431918E-01, 0.17651659E-01, 0.37659165E-01, 0.40000812E-02, 0.67411782E-02, 0.39261781E-01, 0.64045368E-02, 0.15160269E-01, 0.13118289E-01, -0.23408243E-01, -0.30519590E-02, -0.95153069E-02, -0.35423130E-01, 
    },
    {
      0.18906865E-01, 0.64482591E-02, -0.96611172E-01, 0.15293910E-01, -0.54000065E-01, 0.30757926E-01, 0.59357233E-01, 0.36398362E-01, 0.25581049E-02, 0.11665761E-01, 0.56256289E-02, 0.57567651E-02, 0.31496417E-02, 
    },
    {
      -0.46144247E-01, 0.41687142E-01, -0.34603193E-01, 0.53481121E-01, 0.11668256E-01, -0.49698236E-02, 0.57190461E-02, 0.17202956E-02, -0.39180489E-02, -0.26724448E-01, -0.32743976E-01, 0.17875217E-01, -0.13148283E-01, 
    },
  },
  {
    {
      0.97347717E+01, 0.80199363E-02, 0.88190660E-01, 0.34591865E-01, -0.20683805E-01, 0.29348714E-01, -0.21960216E-01, -0.94146077E-02, 0.15242774E-02, 0.18784342E-01, -0.42741881E-02, -0.11803485E-01, 0.89114830E-02, 
    },
    {
      0.17238567E+01, -0.14606644E+00, -0.23781560E+00, -0.18891615E+00, -0.83614878E-01, 0.48989985E-01, -0.23082246E+00, -0.45915204E+00, 0.52839011E-01, 0.65929607E-01, -0.27021214E+00, 0.59767954E-01, 0.45384470E-01, 
    },
    {
      0.26139549E+02, -0.26177337E+01, -0.18818336E+01, 0.28918128E+01, 0.26626935E+01, -0.13387190E+01, 0.35804051E+00, 0.61114317E+00, -0.39724249E+00, -0.34554461E+00, 0.11107710E-01, 0.46736574E+00, -0.14963689E+00, 
    },
    {
      -0.14329063E+02, 0.17789959E+01, 0.11053965E+02, 0.40939870E+01, 0.74688954E+01, 0.81202614E+00, 0.18834094E+01, 0.78406534E+01, 0.12138880E+01, -0.18455257E+01, 0.44079385E+01, -0.87374485E+00, -0.19130919E+01, 
    },
    {
      -0.12161191E+03, 0.20769333E+02, 0.10549552E+02, -0.25936485E+02, -0.22852907E+02, 0.10790629E+02, 0.59407198E+00, -0.15447999E+01, 0.52851825E+01, 0.17490312E+00, 0.17242504E+01, -0.27120554E+01, 0.20165932E+00, 
    },
    {
      0.56903176E+01, -0.30547829E+02, -0.53477966E+02, -0.13067127E+02, -0.37363407E+02, -0.77788720E+01, -0.83473549E+01, -0.36846283E+02, -0.10496927E+02, 0.98110600E+01, -0.20411940E+02, 0.33325286E+01, 0.11927084E+02, 
    },
    {
      0.18244954E+03, -0.48165424E+02, -0.24666487E+02, 0.77004440E+02, 0.64335304E+02, -0.29243431E+02, -0.68443341E+01, -0.68185717E+00, -0.17612892E+02, 0.27443240E+01, -0.67463579E+01, 0.52837248E+01, 0.84665340E+00, 
    },
    {
      0.83355324E+02, 0.10590702E+03, 0.85216660E+02, 0.10837219E+02, 0.71135712E+02, 0.20155458E+02, 0.19259611E+02, 0.72910683E+02, 0.27517429E+02, -0.19637804E+02, 0.41009953E+02, -0.52474489E+01, -0.27441065E+02, 
    },
    {
      -0.12251460E+03, 0.41699898E+02, 0.27045620E+02, -0.92491447E+02, -0.72843239E+02, 0.32114170E+02, 0.11262262E+02, 0.40949612E+01, 0.22106949E+02, -0.48072939E+01, 0.86024685E+01, -0.41298609E+01, -0.18827738E+01, 
    },
    {
      -0.14486723E+03, -0.13257008E+03, -0.48221794E+02, 0.35999949E+01, -0.58804531E+02, -0.19716801E+02, -0.20199421E+02, -0.64824928E+02, -0.29064964E+02, 0.16903173E+02, -0.37667240E+02, 0.34804225E+01, 0.26898163E+02, 
    },
    {
      0.30944574E+02, -0.11548166E+02, -0.11504376E+02, 0.38749798E+02, 0.28696703E+02, -0.12365782E+02, -0.53707695E+01, -0.24697113E+01, -0.94244776E+01, 0.22112002E+01, -0.35927393E+01, 0.10858067E+01, 0.97322190E+00, 
    },
    {
      0.68757813E+02, 0.55788937E+02, 0.54013424E+01, -0.52038832E+01, 0.17593607E+02, 0.63505483E+01, 0.76194353E+01, 0.21402803E+02, 0.10745119E+02, -0.52722220E+01, 0.12947476E+02, -0.74531823E+00, -0.95251179E+01, 
    },
    {
      -0.11819898E+00, 0.16981829E+01, 0.17257081E+01, -0.46752773E-01, 0.31282842E-01, -0.48106876E-02, 0.42170644E-01, 0.12124905E-01, -0.19874832E-01, 0.12629480E-01, 0.39389998E-01, -0.12933319E-01, -0.14372598E-01, 
    },
    {
      0.10223607E+00, -0.16555769E+01, 0.15552585E+01, -0.12264324E+00, 0.65852158E-01, -0.47787797E-03, -0.17770162E-01, 0.24633724E-01, -0.14257873E-01, -0.19858574E-01, 0.25402587E-01, -0.80514261E-02, 0.68274671E-02, 
    },
    {
      -0.99085343E+00, 0.17199130E+01, -0.47460574E+00, -0.74100316E+00, 0.10574152E+01, 0.11296365E+00, 0.82022703E+00, -0.18898286E+00, 0.36352620E+00, 0.80850042E-01, -0.19919536E+00, -0.40538307E-01, 0.16510320E+00, 
    },
    {
      0.11990190E+01, 0.51340390E-01, 0.75055975E+00, -0.12824067E+01, 0.86559242E+00, -0.28388435E-02, 0.14062119E-01, 0.24629071E+00, 0.20781083E+00, -0.11486780E+00, 0.91260616E-02, 0.73089264E-01, -0.38038802E-02, 
    },
    {
      0.51901360E+01, 0.11364080E+02, -0.14781400E+02, 0.34115415E+01, 0.17582846E+01, 0.49295741E+00, 0.74259001E+00, 0.60993052E+00, 0.43994176E+00, 0.49398413E+00, -0.82156599E-01, 0.19255295E+00, 0.72789848E+00, 
    },
    {
      -0.10661777E+02, 0.13685346E+02, 0.17330500E+02, 0.17598805E+01, 0.15718225E+01, -0.35067117E+00, -0.24391625E+01, 0.87025601E+00, 0.10643625E+01, 0.49751109E+00, -0.65840626E+00, 0.33091986E+00, -0.13313615E+01, 
    },
    {
      0.21538338E+02, -0.17283089E+02, 0.23691774E+02, 0.43560233E+01, -0.13055443E+02, 0.54472351E+00, -0.12533862E+02, 0.69577451E+01, -0.59899955E+01, -0.24915967E+01, 0.48114142E+01, -0.15907498E+00, -0.37714081E+01, 
    },
    {
      -0.12313134E+02, -0.22617418E+02, 0.24490969E+01, 0.22901756E+02, -0.17509409E+02, 0.51498137E+01, 0.57923192E+00, -0.68807111E+01, -0.37194932E+01, 0.33164258E+01, -0.77732426E+00, -0.33252787E-01, -0.22301903E+01, 
    },
    {
      -0.31336721E+02, 0.28626883E+01, 0.13524348E+03, -0.29654032E+02, -0.11670376E+02, -0.11883139E+02, -0.67173586E+01, -0.26616092E+01, -0.46543760E+01, -0.62572255E+01, -0.36758001E+01, -0.12831590E+00, -0.83014069E+01, 
    },
    {
      0.10399525E+03, -0.14304045E+03, -0.14625183E+02, -0.12641206E+02, -0.18179705E+02, 0.63950963E+01, 0.24394505E+02, -0.26621683E+01, -0.96777048E+01, -0.36132805E+01, 0.60474663E+01, -0.52178507E+01, 0.14800138E+02, 
    },
    {
      -0.13617873E+03, 0.98202393E+02, -0.52166687E+02, 0.82724130E+00, 0.79597404E+02, -0.97800264E+01, 0.51829216E+02, -0.55843555E+02, 0.30942001E+02, 0.15340666E+02, -0.32310665E+02, 0.47805400E+01, 0.16678785E+02, 
    },
    {
      0.71018166E+02, 0.11442221E+03, 0.75697632E+01, -0.13351144E+03, 0.80042564E+02, -0.19554031E+02, -0.23364079E+01, 0.45533413E+02, 0.25092596E+02, -0.27485874E+02, 0.11551980E+02, -0.85284805E+01, 0.22155190E+02, 
    },
    {
      0.72216705E+02, -0.18108649E+03, -0.35639267E+03, 0.91767868E+02, 0.27649412E+02, 0.51445087E+02, 0.16393131E+02, 0.84795976E+00, 0.11156778E+02, 0.21300432E+02, 0.15415002E+02, -0.20493467E+01, 0.29226500E+02, 
    },
    {
      -0.33076926E+03, 0.39775549E+03, -0.19134296E+03, 0.43584145E+02, 0.63088699E+02, -0.25370869E+02, -0.73108444E+02, 0.75435942E+00, 0.30592009E+02, 0.11364077E+02, -0.18345686E+02, 0.19485804E+02, -0.48085079E+02, 
    },
    {
      0.35883054E+03, -0.27099554E+03, -0.71656799E+02, -0.42929935E+02, -0.21730473E+03, 0.31643620E+02, -0.90837830E+02, 0.16265657E+03, -0.70991798E+02, -0.36282787E+02, 0.85148849E+02, -0.18301117E+02, -0.27730669E+02, 
    },
    {
      -0.21271365E+03, -0.24034344E+03, -0.90665169E+02, 0.33108362E+03, -0.14937883E+03, 0.15747417E+02, 0.58647780E+01, -0.11471978E+03, -0.71360039E+02, 0.83697266E+02, -0.38742779E+02, 0.33933643E+02, -0.68442345E+02, 
    },
    {
      -0.65536263E+02, 0.34354947E+03, 0.40484613E+03, -0.11488379E+03, -0.30950413E+02, -0.81028809E+02, -0.16655785E+02, 0.55178194E+01, -0.82527332E+01, -0.27838104E+02, -0.20135136E+02, 0.50235019E+01, -0.40293743E+02, 
    },
    {
      0.42467825E+03, -0.45499557E+03, 0.41682208E+03, -0.65872879E+02, -0.86527466E+02, 0.35884357E+02, 0.88252113E+02, 0.32965231E+01, -0.39156567E+02, -0.14947558E+02, 0.21975187E+02, -0.26912621E+02, 0.60366070E+02, 
    },
    {
      -0.41629010E+03, 0.33380893E+03, 0.27333212E+03, 0.82168999E+02, 0.25991666E+03, -0.40061562E+02, 0.73255257E+02, -0.19695262E+03, 0.74450417E+02, 0.36899387E+02, -0.95648888E+02, 0.25082708E+02, 0.18237814E+02, 
    },
    {
      0.29320822E+03, 0.22077357E+03, 0.16044160E+03, -0.35729318E+03, 0.12526577E+03, 0.13969834E+02, -0.11527355E+02, 0.12376456E+03, 0.87678589E+02, -0.10517470E+03, 0.47443066E+02, -0.45187057E+02, 0.84595184E+02, 
    },
    {
      0.20050932E+02, -0.17802786E+03, -0.16851566E+03, 0.49642143E+02, 0.13681282E+02, 0.42664768E+02, 0.62386675E+01, -0.44329839E+01, 0.11935354E+01, 0.12374495E+02, 0.84155293E+01, -0.32429960E+01, 0.19101906E+02, 
    },
    {
      -0.19180679E+03, 0.18370607E+03, -0.23252710E+03, 0.34046066E+02, 0.40737797E+02, -0.16559662E+02, -0.37816505E+02, -0.23657224E+01, 0.17279062E+02, 0.67948833E+01, -0.89706383E+01, 0.12578118E+02, -0.25909058E+02, 
    },
    {
      0.17511346E+03, -0.14709790E+03, -0.17955647E+03, -0.46925884E+02, -0.11192869E+03, 0.18169199E+02, -0.22695395E+02, 0.84683502E+02, -0.29096594E+02, -0.13504688E+02, 0.38522060E+02, -0.11487776E+02, -0.34478912E+01, 
    },
    {
      -0.14182312E+03, -0.71934738E+02, -0.81841087E+02, 0.13787189E+03, -0.40718723E+02, -0.16116909E+02, 0.78342290E+01, -0.48472740E+02, -0.38284924E+02, 0.46472549E+02, -0.19452414E+02, 0.19880957E+02, -0.36534100E+02, 
    },
    {
      0.72543673E-01, -0.91503099E-01, -0.11593870E+00, -0.89834797E+00, 0.61833626E+00, 0.47313958E-01, 0.14711141E-02, 0.10285539E-01, -0.22091963E-02, -0.20506838E-01, 0.34120739E-01, -0.22531549E-01, -0.78072742E-03, 
    },
    {
      -0.96214540E-01, -0.54838866E-01, -0.14777187E+00, -0.64081067E+00, -0.69528621E+00, -0.30307973E-01, -0.33019181E-01, 0.77059716E-02, -0.97604236E-02, 0.16391926E-02, 0.17542893E-02, 0.45385696E-02, -0.20407379E-01, 
    },
    {
      0.10522875E+01, 0.18861838E+00, -0.25081074E-01, -0.14823078E+01, 0.23610927E+00, 0.65894437E+00, -0.24199405E+00, -0.87921269E-01, -0.54795289E+00, -0.14422113E+00, -0.10366818E+00, -0.18718950E+00, 0.21545326E-01, 
    },
    {
      0.42069238E+00, -0.22901510E+00, -0.91535009E-01, 0.25210852E-01, -0.13556504E+01, 0.37180072E+00, 0.70798683E+00, 0.22921542E-01, 0.14435732E+00, 0.44424638E-01, -0.13737643E+00, 0.19060686E-01, 0.10951717E+00, 
    },
    {
      0.14380064E+01, -0.18441191E+00, 0.46994653E+01, -0.21012435E+01, 0.58279996E+01, -0.13754472E+01, -0.14721313E+01, 0.12456217E+00, -0.15493175E+00, 0.83261234E+00, -0.55525309E+00, -0.44475816E-01, 0.58509374E+00, 
    },
    {
      0.19653126E+01, -0.75903958E+00, 0.46786284E+01, -0.82793903E+01, -0.22516363E+01, 0.15342182E+01, -0.47626281E+00, 0.35320488E+00, 0.90423584E+00, -0.91187485E-01, -0.61380106E+00, -0.55390352E+00, 0.45646802E+00, 
    },
    {
      -0.69860544E+01, -0.81822312E+00, -0.18467587E+01, 0.76645608E+01, 0.19130971E+01, -0.54477286E+01, 0.15927750E+01, 0.45357141E+00, 0.58195391E+01, 0.14146249E+01, 0.78945208E+00, 0.14831167E+01, -0.24750523E+00, 
    },
    {
      -0.21675032E+00, 0.16968607E+01, -0.38511195E+01, -0.57537832E+01, 0.97383308E+01, -0.13728514E+01, -0.47436857E+01, -0.27548201E+01, -0.15055325E+01, 0.76347262E-01, 0.40537009E+00, -0.25368646E+00, -0.12546912E+01, 
    },
    {
      -0.13612036E+02, 0.10055150E+02, -0.42518852E+02, 0.18920910E+02, -0.25449715E+02, 0.61947532E+01, 0.10574360E+02, 0.17470124E+01, -0.16524613E+00, -0.48321452E+01, 0.26428554E+01, 0.16906739E+01, -0.48421707E+01, 
    },
    {
      -0.15996868E+02, 0.18508348E+02, -0.23283045E+02, 0.35497883E+02, 0.12486041E+02, -0.13592185E+02, 0.23057771E+01, -0.33479509E+01, -0.91297178E+01, 0.20603864E+01, 0.30273182E+01, 0.47495813E+01, -0.24556153E+01, 
    },
    {
      0.14484992E+02, -0.12824783E+01, 0.20726955E+01, -0.13500336E+02, -0.13579147E+02, 0.12507603E+02, -0.44871597E+01, -0.13822854E+01, -0.16225626E+02, -0.32869406E+01, -0.13542657E+01, -0.36956699E+01, 0.49570027E+00, 
    },
    {
      -0.28008471E+01, -0.72233214E+01, 0.16578836E+02, 0.20905169E+02, -0.18679207E+02, 0.21797931E+01, 0.85508757E+01, 0.10466122E+02, 0.37626474E+01, -0.83420086E+00, -0.52542210E+00, 0.81247646E+00, 0.30173633E+01, 
    },
    {
      0.29996330E+02, -0.24423285E+02, 0.98381447E+02, -0.40458511E+02, 0.31081457E+02, -0.10232323E+02, -0.22379601E+02, -0.84146442E+01, 0.16865023E+01, 0.94843407E+01, -0.36858835E+01, -0.51620426E+01, 0.11019040E+02, 
    },
    {
      0.38974052E+02, -0.47553543E+02, 0.32823261E+02, -0.45407307E+02, -0.18267160E+02, 0.32208038E+02, -0.38544316E+01, 0.74274793E+01, 0.22328951E+02, -0.70309253E+01, -0.59846663E+01, -0.10427266E+02, 0.56140108E+01, 
    },
    {
      -0.11142906E+02, 0.50988245E+01, -0.87233227E+00, 0.10051833E+02, 0.17175941E+02, -0.81438684E+01, 0.35580683E+01, 0.14108999E+01, 0.12470690E+02, 0.20998702E+01, 0.89491493E+00, 0.26432147E+01, -0.34282866E+00, 
    },
    {
      0.32492125E+01, 0.95376787E+01, -0.15285566E+02, -0.19982004E+02, 0.12006571E+02, -0.14975014E+01, -0.40201006E+01, -0.91834297E+01, -0.27835941E+01, 0.51126963E+00, 0.23976563E+00, -0.81174558E+00, -0.18622959E+01, 
    },
    {
      -0.20657648E+02, 0.16039457E+02, -0.63869217E+02, 0.26869730E+02, -0.87734652E+01, 0.46892033E+01, 0.14171228E+02, 0.80630875E+01, -0.16122298E+01, -0.58862243E+01, 0.13981988E+01, 0.40675755E+01, -0.73429766E+01, 
    },
    {
      -0.29788891E+02, 0.30969652E+02, -0.10377958E+02, 0.15621193E+02, 0.83370752E+01, -0.22060728E+02, 0.21002872E+01, -0.47829390E+01, -0.15391683E+02, 0.58026686E+01, 0.41257248E+01, 0.70328178E+01, -0.43682208E+01, 
    },
    {
      -0.13153026E+00, 0.52396119E-01, -0.78708120E-01, -0.31984996E-01, 0.66879772E-01, -0.44668841E+00, -0.11997780E+01, 0.44919256E-01, -0.21612016E-01, -0.85290587E-02, 0.40041875E-01, -0.18274922E-01, -0.12108567E-01, 
    },
    {
      0.68851382E-01, -0.16586338E+00, 0.10567039E+00, -0.35489373E-01, 0.10825994E+00, 0.11110218E+01, -0.48024797E+00, -0.32602344E-01, 0.30053170E-01, 0.14435068E-01, -0.43142173E-01, 0.28743999E-01, -0.16354781E-03, 
    },
    {
      0.43814236E+00, 0.41187561E+00, -0.30113024E+00, 0.17926472E+00, -0.76448750E+00, -0.22111153E-01, -0.31990391E+00, -0.39657485E-01, 0.26836851E+00, 0.31483594E-01, 0.17101410E+00, -0.48243791E-01, 0.58657944E-01, 
    },
    {
      0.29777789E+00, 0.70694327E-01, 0.46847123E+00, 0.16531505E+00, -0.68414465E-01, 0.86738020E+00, -0.25612119E+00, -0.27293500E+00, 0.91270700E-01, 0.98998006E-02, -0.54274175E-01, 0.55843040E-01, 0.87239005E-01, 
    },
    {
      0.14875149E+01, 0.18058693E+00, 0.96503448E+00, 0.38504772E-01, -0.81080878E+00, 0.23628514E+01, 0.31373041E+01, 0.35168904E+00, 0.17957918E+00, 0.99717319E-01, 0.12352602E+00, -0.28581694E-01, 0.32931817E+00, 
    },
    {
      -0.18351794E+01, 0.33663924E+01, -0.17299489E+01, 0.30873060E+00, -0.10293868E+01, -0.27952218E+01, 0.29568193E+01, 0.23304486E+00, 0.48831758E+00, -0.16388200E-01, 0.26351678E+00, -0.23727889E+00, 0.21022537E+00, 
    },
    {
      -0.37114632E+00, -0.61679882E+00, 0.70735806E+00, -0.31027344E+00, 0.14951218E+01, -0.35590678E+00, -0.21566993E-02, 0.32540190E+00, -0.41896856E+00, -0.11011457E+00, -0.22550862E+00, -0.13376597E+00, -0.11999568E+00, 
    },
    {
      -0.62257117E+00, -0.84297568E-01, -0.13334845E+01, -0.10768685E+01, 0.16588950E+00, -0.11461290E+01, -0.81029370E-01, 0.43433130E+00, -0.16242850E+00, -0.12601896E-01, 0.15622550E+00, -0.24858958E+00, -0.19455425E+00, 
    },
    {
      -0.25591044E+01, 0.25164348E+00, -0.15365639E+01, -0.79542905E-01, 0.11995534E+01, -0.45441856E+01, -0.45316739E+01, -0.10119162E+01, -0.80300719E-01, -0.52209429E-01, -0.35192704E+00, 0.85065439E-01, -0.56523353E+00, 
    },
    {
      0.30118430E+01, -0.47145052E+01, 0.29010694E+01, -0.36481506E+00, 0.11654499E+01, 0.43329029E+01, -0.52999201E+01, -0.24453495E+00, -0.10941246E+01, 0.39213534E-01, -0.34023368E+00, 0.27661625E+00, -0.51622367E+00, 
    },
    {
      -0.63605845E-01, 0.50896596E-01, 0.11700895E-01, 0.43969844E-01, 0.13695911E+00, -0.28410681E-01, -0.93208402E-02, 0.34158775E+00, -0.23341532E+00, 0.18170308E-01, 0.19188978E-02, -0.67197052E-02, 0.79181036E-02, 
    },
    {
      0.70990942E-01, 0.41841712E-01, 0.20218294E-02, 0.81035674E-01, -0.52733000E-01, -0.91757327E-02, -0.15658233E-01, 0.20307919E+00, 0.37615097E+00, -0.61855074E-02, 0.74328687E-02, 0.91638947E-02, -0.92550581E-02, 
    },
    {
      0.22771759E+00, -0.20004077E-01, -0.19363023E+00, -0.27965687E-02, 0.73775589E-02, -0.16548780E+00, 0.24623014E-01, 0.59029967E-01, -0.32041299E+00, -0.62310666E-01, -0.22864670E-01, 0.68012923E-02, 0.41548245E-01, 
    },
    {
      0.16248761E+00, -0.53943157E-01, -0.10124869E+00, 0.88020787E-01, 0.84402587E-03, -0.98183513E-01, -0.16950405E+00, 0.16585672E+00, 0.22753301E+00, -0.47868274E-01, -0.62641405E-01, 0.39978638E-01, -0.17434882E-01, 
    },
    {
      -0.12619768E+00, 0.46246566E-01, -0.21488875E-01, 0.72626024E-01, 0.24535481E-01, 0.37770525E-01, 0.66443910E-02, -0.34785960E-01, 0.48899220E-03, 0.11222392E+00, 0.31389341E+00, -0.94639612E-02, 0.29937252E-02, 
    },
    {
      0.18342680E+00, -0.42065293E-01, 0.56304694E-02, -0.50527258E-02, -0.67958511E-01, -0.14655957E-01, 0.18668728E-01, 0.40048421E-02, -0.52483652E-01, -0.29388085E+00, 0.83209299E-01, 0.99197831E-02, 0.39488561E-02, 
    },
    {
      0.73701084E-01, -0.11414203E+00, 0.50474752E-01, 0.52794013E-02, 0.51152506E-02, 0.31500902E-01, 0.14944745E-01, 0.27513613E-02, -0.20441337E-01, -0.17187309E-01, 0.23239015E-01, -0.15333304E+00, 0.10820904E-01, 
    },
    {
      0.10246071E+00, -0.10813680E-01, -0.28045259E-01, 0.85988389E-02, -0.66896558E-01, 0.20317493E-01, -0.25311206E-01, 0.21137524E-01, -0.59515103E-02, -0.73466594E-02, -0.46075849E-03, -0.14882740E-01, -0.15001261E+00, 
    },
    {
      -0.62333126E-01, -0.11848945E+00, -0.74348420E-01, -0.13404631E-02, -0.22817828E-01, -0.10830126E-01, -0.48877955E-02, 0.12237122E-01, -0.30541401E-01, 0.57756705E-02, -0.10295756E-01, 0.12080544E-01, -0.14406471E-01, 
    },
    {
      0.41248910E-01, 0.44591889E-01, 0.36592875E-01, 0.55447145E-03, 0.19915281E-01, -0.63508269E-02, -0.47519132E-01, 0.28801892E-01, 0.77612209E-02, -0.59790271E-02, 0.73863971E-02, 0.16617060E-01, -0.22494165E-01, 
    },
    {
      -0.71962871E-01, -0.41219030E-01, -0.70982397E-01, 0.12210120E-01, -0.87522060E-01, -0.47411237E-01, 0.57425763E-01, 0.45690551E-01, 0.26248202E-01, 0.28013919E-01, -0.16828964E-02, 0.15796809E-01, 0.20284114E-01, 
    },
    {
      -0.47389165E-01, 0.37087288E-01, -0.58108825E-01, -0.54313848E-02, 0.10483098E+00, 0.13532057E-01, -0.20340554E-01, -0.50312970E-02, -0.42387201E-02, -0.89632040E-02, -0.32773022E-01, 0.21721894E-01, 0.10820904E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir20_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30495803E+01, 0.35076546E-02, 0.76724417E-02, 0.87806750E-02, 0.47458550E-02, 0.25473754E-02, 0.29852835E-02, -0.12277642E-03, -0.32025625E-03, 
    },
    {
      0.10599100E+00, -0.21536589E-01, 0.46008382E-01, 0.78201421E-01, 0.22680618E-01, -0.46905663E-01, -0.96271224E-02, 0.12562236E-01, 0.24986399E-01, 
    },
    {
      0.10458715E+01, -0.25089502E-01, -0.17962436E+00, -0.17404799E+00, -0.22410804E-01, 0.27813714E-01, -0.18103598E-01, -0.58276366E-01, 0.26434981E-02, 
    },
    {
      0.17178220E+00, -0.15449823E+00, -0.10333893E+00, -0.18914133E+00, -0.13402216E+00, 0.16679610E+00, 0.46815984E-01, -0.91324374E-02, -0.10088239E+00, 
    },
    {
      -0.18385901E+01, 0.77625453E-01, 0.55216676E+00, 0.35753915E+00, -0.87231752E-02, -0.51372502E-01, 0.34072302E-01, 0.17208675E+00, -0.13502965E-01, 
    },
    {
      -0.32891473E+00, 0.15413617E+00, 0.46012118E-01, 0.10862464E+00, 0.12537768E+00, -0.12952726E+00, -0.47462337E-01, -0.35870597E-02, 0.86061329E-01, 
    },
    {
      0.89103073E+00, -0.37754770E-01, -0.38592705E+00, -0.17952418E+00, 0.40533390E-01, 0.47559445E-02, -0.18648680E-01, -0.11294576E+00, 0.12865890E-01, 
    },
    {
      -0.69007248E-03, -0.18691148E+00, -0.33661303E+00, 0.48614144E-02, 0.11631002E-01, -0.95324073E-03, 0.85280649E-02, 0.36868376E-02, 0.60911463E-02, 
    },
    {
      0.32471757E-01, 0.35279295E+00, -0.18637425E+00, -0.75623016E-02, -0.97096683E-02, 0.31888080E-02, -0.51301736E-02, -0.62356037E-02, 0.11819322E-01, 
    },
    {
      -0.84096253E-01, 0.25142324E+00, -0.35519764E-01, 0.26370287E-01, 0.97039536E-01, -0.31170715E-01, -0.37270427E-01, 0.46906836E-01, 0.65398109E-02, 
    },
    {
      0.45785110E-01, 0.13975869E+00, 0.26513165E+00, -0.29776646E-01, -0.25009252E-01, 0.36971882E-01, -0.30679654E-01, 0.35375282E-01, 0.11662492E+00, 
    },
    {
      0.15254179E+00, 0.67530102E+00, 0.15002052E+01, 0.58707792E-01, -0.19350794E+00, 0.69749169E-01, -0.31615756E-01, -0.64090937E-01, -0.20062795E+00, 
    },
    {
      -0.83293790E+00, -0.15851485E+01, 0.79647267E+00, 0.14819029E+00, 0.13229819E+00, -0.22473099E-01, 0.55494748E-01, 0.13741881E+00, -0.30662367E+00, 
    },
    {
      0.15882361E+01, -0.10474651E+01, 0.56707895E+00, 0.21353865E+00, -0.70312959E+00, 0.67766070E-01, 0.13480249E+00, -0.26861522E+00, 0.10558596E+00, 
    },
    {
      0.42528209E+00, -0.81927925E+00, -0.13487948E+01, -0.28105810E+00, 0.56431311E+00, -0.22810178E+00, 0.15924788E+00, -0.40170720E+00, -0.70476109E+00, 
    },
    {
      -0.50016040E+00, -0.57117230E+00, -0.95467204E+00, -0.63309115E+00, 0.40046284E+00, -0.30428410E+00, 0.42816602E-01, 0.86010098E-01, 0.67586058E+00, 
    },
    {
      0.25389247E+01, 0.97007108E+00, -0.13116745E+01, -0.12303472E+00, -0.53631330E+00, 0.64055860E-01, -0.72015218E-01, -0.42832959E+00, 0.86087793E+00, 
    },
    {
      -0.44755645E+01, 0.10564411E+01, -0.66151208E+00, -0.45737475E+00, 0.15295455E+01, 0.55482633E-01, -0.10898267E+00, 0.44265333E+00, -0.39993185E+00, 
    },
    {
      -0.14845783E+01, 0.12734603E+01, 0.23328686E+01, 0.94133908E+00, -0.13304437E+01, 0.53523225E+00, -0.28171411E+00, 0.10899057E+01, 0.11757891E+01, 
    },
    {
      0.43657407E+00, 0.14657229E+00, 0.85669927E-01, 0.69275594E+00, -0.22207183E+00, 0.28957173E+00, 0.16122398E-02, 0.24003390E-01, -0.55424219E+00, 
    },
    {
      -0.20080192E+01, 0.30063149E+00, 0.88270497E+00, -0.69684193E-01, 0.51002681E+00, -0.63769467E-01, 0.87251076E-02, 0.35308439E+00, -0.61427212E+00, 
    },
    {
      0.32225552E+01, -0.25058147E-02, 0.50169885E+00, 0.13464716E+00, -0.10105824E+01, -0.71476690E-01, 0.32011800E-01, -0.21982677E+00, 0.30498490E+00, 
    },
    {
      0.11992681E+01, -0.96706653E+00, -0.11388683E+01, -0.73485720E+00, 0.80431402E+00, -0.39570120E+00, 0.16278829E+00, -0.83236420E+00, -0.58864450E+00, 
    },
    {
      -0.14675558E-01, -0.23623409E-01, -0.17950738E-01, 0.66034064E-01, -0.11922826E+00, -0.54179137E-02, 0.14375948E-01, -0.60841944E-02, -0.26708576E-02, 
    },
    {
      0.43763379E-02, -0.10887975E-01, -0.71096844E-02, 0.11585610E+00, 0.90701804E-01, 0.55126945E-03, 0.56899362E-02, 0.31451138E-02, -0.52986522E-02, 
    },
    {
      0.10125032E+00, 0.17959675E+00, 0.10625629E+00, -0.50184567E-01, 0.10304492E+00, 0.24058525E-01, -0.32390207E-01, 0.75553027E-02, -0.20764861E-01, 
    },
    {
      -0.11460812E+00, 0.79707839E-01, -0.29399197E-02, -0.67522340E-01, 0.13258697E+00, 0.49736477E-01, -0.78869835E-02, -0.39157858E-02, 0.24828205E-01, 
    },
    {
      0.10480853E+00, 0.19343019E+00, 0.17289962E+00, -0.31678677E+00, -0.17360756E+00, 0.37843499E-01, 0.62656075E-01, -0.18113919E-02, -0.78926899E-01, 
    },
    {
      -0.26774183E+00, -0.22751540E-01, -0.13300699E+00, 0.24421448E+00, -0.40151697E+00, -0.44680607E-01, 0.90509988E-01, -0.45797311E-01, -0.19253580E-01, 
    },
    {
      -0.53374827E+00, -0.90245152E+00, -0.14048404E+00, 0.31762448E+00, -0.75603151E+00, -0.17422198E+00, 0.16830654E+00, -0.67688487E-01, 0.83993100E-01, 
    },
    {
      0.92552233E+00, -0.17285006E-01, 0.26109925E+00, 0.74910748E+00, -0.50195232E-01, -0.18097821E+00, -0.59655149E-01, 0.36261849E-01, -0.16167022E+00, 
    },
    {
      -0.18626483E+00, -0.31044501E+00, -0.34045997E+00, 0.30551198E+00, 0.83555371E-01, -0.10926255E+00, -0.14400959E+00, 0.99276425E-02, 0.17333107E+00, 
    },
    {
      0.29973921E+00, -0.38495310E-01, 0.25320417E+00, -0.16226833E+00, 0.34776387E+00, 0.68651758E-01, -0.18148483E+00, 0.94451770E-01, 0.44157453E-01, 
    },
    {
      0.68308866E+00, 0.12142456E+01, 0.23941616E-01, -0.49178913E+00, 0.10317271E+01, 0.30297923E+00, -0.16258200E+00, 0.87277547E-01, -0.87765060E-01, 
    },
    {
      -0.12280098E+01, -0.17898367E+00, -0.24613486E+00, -0.11506540E+01, -0.37660384E+00, 0.14181741E+00, 0.15465291E+00, -0.45992363E-01, 0.17110234E+00, 
    },
    {
      -0.21097982E-01, 0.60008105E-03, -0.18510208E-02, -0.62552849E-02, -0.93422104E-02, 0.60288724E-01, -0.40334366E-01, -0.43196464E-02, -0.61084577E-02, 
    },
    {
      -0.83072307E-02, -0.34225453E-01, -0.11824392E-01, -0.13751618E-02, -0.17261762E-01, 0.43971878E-01, 0.76659046E-01, 0.33106832E-02, 0.45388611E-02, 
    },
    {
      0.75911164E-01, 0.46962660E-01, 0.21635788E-01, -0.77638470E-01, 0.22221388E-01, -0.79871550E-01, -0.12672771E-01, 0.25498134E-02, 0.12646123E-02, 
    },
    {
      0.51471148E-01, 0.33606584E-02, -0.13351038E-01, -0.24537463E-01, -0.35670411E-01, 0.14631917E-01, -0.90372808E-01, 0.99615008E-03, 0.22813002E-01, 
    },
    {
      0.21750353E-01, -0.83270617E-01, -0.18755151E-02, 0.53586468E-01, 0.93389477E-03, -0.19740278E+00, -0.11719642E+00, 0.33343490E-01, 0.22319425E-01, 
    },
    {
      -0.29066566E-02, 0.41417848E-01, -0.10801823E-01, 0.28476706E-01, 0.10270041E-01, 0.15855379E+00, -0.24726029E+00, 0.14735273E-01, -0.27862186E-01, 
    },
    {
      -0.23555486E-01, -0.40901415E-02, 0.60059442E-02, 0.97556552E-02, 0.91812024E-02, -0.14488046E-02, 0.16592178E-02, 0.36357131E-01, 0.13751269E-01, 
    },
    {
      -0.29764045E-01, 0.92727989E-02, -0.86509585E-02, -0.18609468E-01, -0.27944657E-03, 0.31447997E-02, 0.75850124E-02, -0.18039657E-01, 0.38740944E-01, 
    },
    {
      0.79681963E-01, 0.48902291E-02, 0.57677603E-02, -0.33417843E-01, -0.68668877E-02, 0.22215862E-02, -0.40455712E-02, 0.36775917E-02, -0.22278046E-01, 
    },
    {
      0.10155325E-01, -0.54381907E-01, 0.22469945E-01, 0.53363744E-01, -0.10555824E-01, -0.46219719E-02, -0.16068283E-02, 0.35136204E-01, 0.11071405E-01, 
    },
    {
      0.18289034E-02, -0.75299381E-02, -0.14482981E-03, 0.35247686E-02, 0.42061866E-02, 0.51727840E-02, 0.17811510E-02, -0.10641967E-01, 0.48779794E-02, 
    },
    {
      0.15797319E-01, 0.18416291E-01, 0.16457362E-05, -0.16946379E-02, -0.99487100E-02, 0.13055210E-01, -0.14801215E-02, -0.56088273E-02, -0.40782345E-02, 
    },
    {
      -0.35834808E-02, 0.71111410E-02, 0.12105495E-01, -0.45448786E-03, 0.62631634E-02, 0.36953585E-02, -0.28422426E-02, -0.19898398E-02, 0.48590722E-02, 
    },
    {
      0.11628112E-01, 0.28534012E-03, -0.12218224E-01, 0.24151376E-02, -0.16151644E-03, 0.75605242E-02, 0.45826556E-02, -0.46821288E-02, 0.12317485E-02, 
    },
  },
  {
    {
      0.26966269E+01, -0.14578769E-05, 0.50493279E-02, 0.65831305E-02, 0.43787737E-02, 0.39015978E-02, 0.89700258E-03, 0.85402245E-03, -0.68476156E-03, 
    },
    {
      0.11744526E+00, 0.12448850E-01, 0.51003762E-01, 0.40138353E-01, 0.26125014E-01, -0.16533557E-01, -0.29090958E-01, -0.17052057E-02, 0.80826618E-02, 
    },
    {
      0.11973118E+01, -0.65078013E-01, -0.99028312E-01, -0.81597298E-01, -0.32280181E-01, -0.22603732E-01, -0.10447549E-01, -0.41825704E-01, 0.23424169E-02, 
    },
    {
      0.84804960E-01, -0.19440410E+00, -0.12508395E+00, -0.10045742E+00, -0.13333432E+00, 0.45456056E-01, 0.10219767E+00, 0.23704873E-01, -0.32805737E-01, 
    },
    {
      -0.18993030E+01, 0.22647896E+00, 0.26471585E+00, 0.15323690E+00, 0.31060221E-01, 0.91907240E-01, 0.40878788E-01, 0.95126577E-01, 0.45649102E-02, 
    },
    {
      -0.19986725E+00, 0.13906373E+00, 0.85632414E-01, 0.60318705E-01, 0.11465530E+00, -0.33805624E-01, -0.83187349E-01, -0.24020078E-01, 0.27057443E-01, 
    },
    {
      0.82170755E+00, -0.12471489E+00, -0.17876776E+00, -0.71481489E-01, 0.27813890E-02, -0.87954432E-01, -0.34027167E-01, -0.52087925E-01, -0.90684919E-02, 
    },
    {
      0.43652612E-02, -0.33954412E+00, -0.28059766E+00, -0.48793945E-03, 0.57276776E-02, -0.32016952E-02, 0.50360449E-02, 0.29207736E-02, 0.43764645E-02, 
    },
    {
      0.85188961E-02, 0.30349514E+00, -0.33877447E+00, -0.28169504E-02, -0.11583043E-01, 0.14881253E-02, -0.76294229E-02, -0.18964122E-02, 0.11345063E-01, 
    },
    {
      -0.12247941E-01, 0.17714764E+00, -0.34096591E-01, -0.22695521E-01, 0.67024931E-01, -0.39165456E-01, -0.30592242E-01, 0.23541650E-01, 0.10291560E-01, 
    },
    {
      0.50195772E-01, 0.15112165E+00, 0.20722289E+00, 0.34769014E-01, -0.29959407E-01, 0.41639198E-01, -0.29532194E-01, 0.25578812E-01, 0.78463838E-01, 
    },
    {
      -0.60606830E-01, 0.10363140E+01, 0.86959022E+00, 0.15487760E+00, -0.99790499E-01, 0.41185062E-01, 0.15034962E-01, -0.47623415E-01, -0.14509769E+00, 
    },
    {
      -0.57434636E+00, -0.97933829E+00, 0.12581741E+01, 0.25207497E-01, 0.18004075E+00, -0.21284509E-01, 0.77311493E-01, 0.64989790E-01, -0.20855135E+00, 
    },
    {
      0.47468442E+00, -0.10885421E+01, 0.38821119E+00, 0.23283729E+00, -0.50951570E+00, 0.23506916E+00, 0.35243019E-01, -0.17403507E+00, -0.23718674E-01, 
    },
    {
      0.26270482E+00, -0.70497757E+00, -0.13304210E+01, -0.53473514E+00, 0.28150737E+00, -0.26304531E+00, 0.18870257E+00, -0.23507071E+00, -0.43555272E+00, 
    },
    {
      0.11272904E+00, -0.99514472E+00, -0.34449834E+00, -0.71797216E+00, 0.18339764E+00, -0.18274480E+00, -0.63537367E-01, 0.10146949E+00, 0.48926970E+00, 
    },
    {
      0.20371649E+01, 0.22109611E+00, -0.20750749E+01, 0.11111619E+00, -0.56142098E+00, 0.42209644E-01, -0.19956113E+00, -0.19299477E+00, 0.54810613E+00, 
    },
    {
      -0.14952495E+01, 0.17684418E+01, -0.49218816E+00, -0.28709847E+00, 0.11091636E+01, -0.42815658E+00, 0.12084533E+00, 0.37503195E+00, -0.81880488E-01, 
    },
    {
      -0.14637735E+01, 0.10478177E+01, 0.26719489E+01, 0.13296674E+01, -0.46114388E+00, 0.65784061E+00, -0.44182307E+00, 0.56460422E+00, 0.67381424E+00, 
    },
    {
      -0.28567558E-01, 0.46754566E+00, 0.17383933E+00, 0.67441458E+00, -0.10554819E+00, 0.19837098E+00, 0.47730759E-01, -0.50233286E-01, -0.40128100E+00, 
    },
    {
      -0.17196779E+01, 0.35429364E+00, 0.14362482E+01, -0.16386373E+00, 0.45979795E+00, -0.21841004E-01, 0.13400900E+00, 0.14516471E+00, -0.39125723E+00, 
    },
    {
      0.10732856E+01, -0.75785559E+00, 0.52866328E+00, 0.71639992E-01, -0.73546070E+00, 0.24522778E+00, -0.13535358E+00, -0.22979885E+00, 0.11654176E+00, 
    },
    {
      0.15114303E+01, -0.77876258E+00, -0.15880039E+01, -0.89949375E+00, 0.22964896E+00, -0.49463138E+00, 0.30698773E+00, -0.40895736E+00, -0.29977861E+00, 
    },
    {
      -0.79477411E-02, -0.18472305E-01, -0.12967197E-01, 0.94843686E-01, -0.88151120E-01, -0.51070806E-02, 0.13504693E-01, -0.41395505E-02, 0.24751091E-03, 
    },
    {
      -0.44946186E-02, -0.16740741E-01, -0.41124807E-02, 0.76202437E-01, 0.10344878E+00, -0.22537799E-02, 0.43183658E-02, 0.64705717E-04, -0.59640518E-03, 
    },
    {
      0.95618159E-01, 0.92954643E-01, 0.61767250E-01, -0.79024971E-01, 0.11346289E+00, 0.15616628E-01, -0.56508440E-02, 0.42031724E-02, 0.24400151E-02, 
    },
    {
      0.96911388E-02, 0.29432042E-01, -0.46471480E-01, -0.12849204E+00, 0.22337712E-01, 0.15172200E-01, 0.14756573E-01, -0.10275276E-01, 0.32632593E-01, 
    },
    {
      0.90240715E-02, 0.98447271E-01, 0.13910589E+00, -0.49684674E+00, 0.73422439E-01, 0.23828603E-02, 0.33573525E-02, -0.72285314E-02, -0.70686325E-01, 
    },
    {
      -0.16957556E+00, 0.50366666E-01, -0.11485599E+00, -0.34002479E-01, -0.50488788E+00, -0.42284265E-01, 0.40896658E-01, -0.22395281E-02, -0.41483179E-01, 
    },
    {
      -0.39100450E+00, -0.52364653E+00, 0.81161167E-02, 0.23826459E+00, -0.69935983E+00, -0.49828898E-01, 0.58817074E-01, -0.30699246E-01, 0.18457994E-02, 
    },
    {
      0.18260169E+00, 0.12305915E-01, 0.35234398E+00, 0.78882325E+00, 0.17531563E+00, 0.12401869E-01, -0.94772555E-01, 0.70517570E-01, -0.18616913E+00, 
    },
    {
      -0.20249626E-01, -0.18173020E+00, -0.28802195E+00, 0.27763471E+00, -0.33017445E+00, -0.40717527E-01, -0.36766961E-01, 0.36674999E-01, 0.13604657E+00, 
    },
    {
      0.24803011E+00, -0.15678068E+00, 0.20198597E+00, 0.30329382E+00, 0.25746182E+00, 0.61767373E-01, -0.93926691E-01, 0.18352006E-01, 0.62546603E-01, 
    },
    {
      0.35427305E+00, 0.75491345E+00, -0.11663892E+00, -0.31980485E+00, 0.90525955E+00, 0.89059651E-01, -0.78885183E-01, 0.31197926E-01, -0.51024295E-02, 
    },
    {
      -0.33548266E+00, 0.36326089E-03, -0.39684558E+00, -0.10250442E+01, -0.45261109E+00, -0.34287624E-01, 0.12944759E+00, -0.95721230E-01, 0.19820654E+00, 
    },
    {
      -0.17688449E-01, 0.15564424E-02, 0.39224289E-02, -0.38294392E-02, -0.75140465E-02, 0.59528705E-01, -0.48654277E-01, -0.22800365E-02, -0.31075822E-02, 
    },
    {
      -0.96650571E-02, -0.21461820E-01, 0.77744690E-03, 0.16463975E-02, -0.11097819E-01, 0.51459324E-01, 0.68256907E-01, 0.39644483E-02, -0.15337553E-02, 
    },
    {
      0.34351684E-01, 0.43128196E-01, 0.19956021E-01, -0.36301594E-01, 0.99364547E-02, -0.61444476E-01, -0.24400521E-01, 0.44092600E-03, 0.13028901E-02, 
    },
    {
      0.22622351E-01, 0.11014455E-01, -0.81500681E-02, -0.17104017E-01, -0.29482944E-01, 0.26210617E-01, -0.75117700E-01, -0.28545759E-03, 0.11840697E-01, 
    },
    {
      0.17646819E-01, -0.77693120E-01, 0.38775124E-02, 0.28927619E-01, 0.27532193E-02, -0.16366033E+00, -0.96197531E-01, 0.21273768E-01, 0.86615868E-02, 
    },
    {
      0.72159092E-02, 0.12788316E-01, -0.31918742E-01, 0.10852541E-01, 0.17516689E-01, 0.12468155E+00, -0.17470151E+00, -0.41061863E-02, -0.12752778E-02, 
    },
    {
      -0.18442726E-01, -0.91943759E-02, 0.67788684E-02, 0.14435091E-03, 0.78480067E-02, 0.98523422E-04, 0.22894247E-02, 0.25241941E-01, 0.15407712E-01, 
    },
    {
      -0.16817285E-01, 0.31028625E-02, -0.10441542E-01, -0.79444088E-02, -0.17821643E-02, 0.52367494E-03, 0.82076322E-02, -0.18905919E-01, 0.29213514E-01, 
    },
    {
      0.44744335E-01, -0.13247139E-02, -0.37604002E-02, -0.16222805E-01, -0.89065684E-02, -0.51037166E-02, -0.28314355E-02, 0.18366748E-01, -0.27277077E-01, 
    },
    {
      0.19684851E-01, -0.35009183E-01, 0.23218991E-01, 0.28978951E-01, -0.10054998E-01, -0.95502073E-02, -0.97779622E-02, 0.31590566E-01, 0.18811375E-01, 
    },
    {
      0.27152249E-02, -0.10953413E-01, 0.43081818E-02, 0.26744686E-02, 0.12518663E-02, 0.79071680E-02, -0.18357651E-02, -0.90767127E-02, 0.21241107E-02, 
    },
    {
      0.64085410E-02, 0.12395338E-01, -0.43195123E-02, 0.27304080E-02, -0.55480246E-02, 0.65472890E-02, 0.11660623E-02, -0.26218852E-02, -0.10962778E-02, 
    },
    {
      0.18541915E-03, 0.11536265E-01, 0.10362318E-01, 0.24194026E-02, -0.11380968E-02, 0.27860261E-02, 0.36632185E-03, -0.13429242E-02, 0.37920703E-02, 
    },
    {
      0.48007774E-02, 0.10027210E-02, -0.10543189E-01, 0.72369934E-02, -0.16516771E-02, 0.35164719E-02, 0.81353862E-03, -0.30095526E-02, 0.13707984E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir21_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.63694148E+01, -0.34792226E-01, -0.74101150E-01, 0.35132650E-01, 0.24849806E-01, 0.11001032E-01, -0.19688884E-01, 0.26043868E-01, 0.10386006E-01, -0.32306511E-02, 0.13962947E-01, -0.10103378E-01, 0.18723549E-01, 
    },
    {
      0.21955316E+01, -0.39681266E-02, 0.18197685E+00, 0.11352257E+00, -0.31861094E+00, -0.15868060E-02, 0.39689042E-01, 0.17209752E-01, 0.19868961E+00, -0.80792844E-01, -0.11767738E+00, 0.12445180E+00, -0.55129856E-01, 
    },
    {
      0.97781305E+01, 0.63362435E-01, -0.53614479E+00, -0.49567866E+00, 0.11741468E+01, -0.63698125E+00, 0.52839708E+00, -0.11437672E+00, -0.15261206E+00, -0.13389719E+00, -0.36991826E+00, 0.11427352E+00, -0.65866762E+00, 
    },
    {
      -0.32809158E+02, -0.52905566E+00, 0.56250453E+00, 0.25619733E+00, 0.76901011E+01, -0.76617765E+00, 0.13428245E+00, -0.12091900E+01, -0.15285416E+01, 0.67826205E+00, 0.23068154E+01, -0.12790626E+01, -0.21327354E+00, 
    },
    {
      -0.58018158E+02, 0.77569704E+01, 0.58701005E+01, -0.37597203E+01, -0.11486585E+02, 0.50069208E+01, -0.19475280E+01, -0.26410751E+01, 0.11136000E+01, 0.12512259E+00, 0.17224880E+01, -0.64933658E+00, 0.42313733E+01, 
    },
    {
      0.81844147E+02, -0.18177500E+02, -0.10697339E+02, 0.40396552E+01, -0.31275728E+02, 0.73205338E+01, -0.39177847E+01, 0.58170772E+01, 0.57865572E+01, -0.22364533E+01, -0.11624065E+02, 0.43845687E+01, 0.30469108E+01, 
    },
    {
      0.10854137E+03, -0.30877756E+02, -0.10837447E+02, 0.22226160E+02, 0.34705902E+02, -0.13085101E+02, 0.15620646E+01, 0.11406905E+02, -0.31094887E+01, 0.85883933E+00, -0.26042700E+01, 0.11922537E+01, -0.10150283E+02, 
    },
    {
      -0.44580307E+02, 0.79383812E+02, 0.24413515E+02, -0.21338619E+02, 0.50065228E+02, -0.21513105E+02, 0.15347665E+02, -0.10429506E+02, -0.11754140E+02, 0.34370358E+01, 0.24936632E+02, -0.73416333E+01, -0.82282839E+01, 
    },
    {
      -0.10028139E+03, 0.37588978E+02, 0.43946161E+01, -0.33876656E+02, -0.40892410E+02, 0.13890780E+02, 0.78339022E+00, -0.14895038E+02, 0.35314486E+01, -0.15978297E+01, 0.11808842E+01, -0.71185213E+00, 0.10538705E+02, 
    },
    {
      -0.50124062E+02, -0.10354129E+03, -0.18342447E+02, 0.30102610E+02, -0.34501987E+02, 0.25331213E+02, -0.20694256E+02, 0.83078852E+01, 0.11584894E+02, -0.26216393E+01, -0.24104069E+02, 0.61813402E+01, 0.85840073E+01, 
    },
    {
      0.37392719E+02, -0.14476900E+02, 0.99496180E+00, 0.15965063E+02, 0.16501513E+02, -0.52093725E+01, -0.93629724E+00, 0.62317777E+01, -0.14138385E+01, 0.73162246E+00, 0.71325324E-01, 0.48246250E-01, -0.39943352E+01, 
    },
    {
      0.43037422E+02, 0.42916580E+02, 0.38246312E+01, -0.13166988E+02, 0.83462524E+01, -0.10465692E+02, 0.90876627E+01, -0.24948664E+01, -0.43121109E+01, 0.83627832E+00, 0.86244040E+01, -0.20590615E+01, -0.31272595E+01, 
    },
    {
      -0.40967345E-01, 0.18442563E+01, 0.17633563E+01, -0.41515935E-01, 0.14598442E-01, 0.52640415E-02, 0.33678550E-01, 0.18317280E-01, -0.30224023E-01, -0.20597542E-01, 0.24734342E-01, -0.10589097E-01, 0.24328345E-01, 
    },
    {
      -0.22360490E+00, -0.19206479E+01, 0.18999895E+01, -0.99210821E-01, 0.13044933E-01, 0.23915380E-01, 0.10539087E-01, 0.34584550E-02, 0.12267416E-01, 0.57058195E-02, 0.42165192E-02, 0.10772912E-01, -0.26614875E-01, 
    },
    {
      -0.84906530E+00, 0.15865498E+01, 0.22927699E+01, -0.33673322E+00, 0.19166987E+00, 0.64818418E+00, 0.42091948E+00, -0.16618843E+00, 0.78511827E-01, -0.64768136E-01, -0.39221168E-01, -0.89224152E-01, 0.13744977E+00, 
    },
    {
      -0.14338664E+01, -0.13094720E+01, 0.17051713E+01, -0.10853502E+01, 0.19608435E+00, 0.16197301E+00, 0.21957265E+00, -0.31434265E+00, 0.37495288E+00, 0.34630176E-01, -0.90445802E-01, 0.12040331E+00, 0.17347548E-01, 
    },
    {
      0.52947850E+01, 0.10844605E+02, 0.37868409E+01, -0.10532480E+01, 0.36165705E+01, -0.88409197E+00, 0.49687558E+00, -0.61486661E+00, 0.60421681E+00, 0.21092279E+00, -0.32828119E+00, -0.47752506E+00, -0.94027740E+00, 
    },
    {
      -0.25222619E+01, -0.46317253E+01, 0.80145826E+01, 0.24095893E+01, 0.39263377E+01, 0.25864679E+00, -0.13400965E+01, 0.52035743E+00, 0.38246587E+00, -0.92407811E+00, -0.74844398E-01, -0.70236456E+00, 0.10894667E+00, 
    },
    {
      0.22070520E+02, -0.17723101E+02, -0.98362646E+01, 0.23332732E+01, 0.64828506E+01, -0.10151888E+02, -0.41051803E+01, 0.79824430E+00, -0.15293568E+01, -0.26827996E+01, 0.52764028E+00, 0.14580549E+01, -0.35977669E+01, 
    },
    {
      0.25647627E+02, -0.80371361E+01, -0.18844767E+02, 0.13793572E+02, 0.82348639E+00, 0.21338460E+01, -0.32136986E+01, 0.57288928E+01, -0.49189863E+01, 0.13956453E+01, 0.25248592E+01, -0.16098522E+00, -0.32238259E+01, 
    },
    {
      -0.37831741E+02, -0.82785957E+02, 0.40492912E+02, 0.49759912E+01, -0.32693497E+02, 0.38788466E+01, -0.55101314E+01, 0.52948971E+01, -0.33069551E+01, -0.27049429E+01, 0.15614638E+01, 0.44462080E+01, 0.58480530E+01, 
    },
    {
      0.43074551E+02, -0.35585445E+02, -0.63008469E+02, -0.10148061E+02, -0.38213272E+02, -0.12484533E+01, 0.11518609E+02, -0.16986067E+01, -0.44723053E+01, 0.66183820E+01, 0.27471104E+00, 0.51703401E+01, 0.16862423E+01, 
    },
    {
      -0.12163273E+03, 0.35768089E+02, 0.79624352E+02, -0.42255273E+00, -0.29189657E+02, 0.46464653E+02, 0.11797403E+02, -0.35026555E+01, 0.71749539E+01, 0.28277515E+02, 0.89573674E-01, -0.11701196E+02, 0.20683783E+02, 
    },
    {
      -0.14859586E+03, 0.73930824E+02, 0.46461655E+02, -0.73587746E+02, -0.38935528E+02, -0.21367317E+02, 0.22717581E+02, -0.38625607E+02, 0.22611212E+02, -0.13838289E+02, -0.19563032E+02, -0.54408684E+01, 0.22669931E+02, 
    },
    {
      0.96786781E+02, 0.19943217E+03, -0.22654292E+03, -0.25076122E+01, 0.99311028E+02, -0.32880912E+01, 0.18322615E+02, -0.17872757E+02, 0.32761443E+01, 0.10708526E+02, -0.29951110E+01, -0.12494042E+02, -0.13923942E+02, 
    },
    {
      -0.15908409E+03, 0.22274886E+03, 0.14706615E+03, 0.20086601E+02, 0.12031584E+03, -0.89654094E+00, -0.32396168E+02, -0.99818575E+00, 0.18598759E+02, -0.17126841E+02, 0.17582542E+00, -0.13738128E+02, -0.75380640E+01, 
    },
    {
      0.28013394E+03, -0.26947153E+02, -0.30688678E+03, -0.24610758E+02, 0.25527044E+02, -0.90596825E+02, -0.53650866E+01, 0.11049999E+02, -0.16220327E+02, -0.88406433E+02, -0.52042227E+01, 0.37062027E+02, -0.48077995E+02, 
    },
    {
      0.34042722E+03, -0.20572089E+03, -0.69570465E+02, 0.17793367E+03, 0.15674860E+03, 0.55941879E+02, -0.64308510E+02, 0.11241123E+03, -0.44273415E+02, 0.41367416E+02, 0.56130985E+02, 0.22120152E+02, -0.57441715E+02, 
    },
    {
      -0.10435168E+03, -0.22239352E+03, 0.35573297E+03, -0.10317237E+02, -0.12436950E+03, -0.36222422E+01, -0.26919477E+02, 0.25896906E+02, 0.43949404E+01, -0.15245827E+02, 0.24285519E+01, 0.14503664E+02, 0.14726946E+02, 
    },
    {
      0.22333339E+03, -0.35747256E+03, -0.15495203E+03, -0.19590113E+02, -0.15038443E+03, 0.54151440E+01, 0.37789562E+02, 0.71866360E+01, -0.28933886E+02, 0.18707529E+02, -0.73916203E+00, 0.14262703E+02, 0.99755716E+01, 
    },
    {
      -0.29553113E+03, 0.18611118E+02, 0.45680972E+03, 0.50157089E+02, 0.23859198E+02, 0.80800880E+02, -0.16151365E+02, -0.16563154E+02, 0.17871490E+02, 0.11023319E+03, 0.84442797E+01, -0.48368690E+02, 0.49598927E+02, 
    },
    {
      -0.32021780E+03, 0.24028613E+03, 0.77732681E+02, -0.19011111E+03, -0.21534552E+03, -0.59049309E+02, 0.76389923E+02, -0.14133389E+03, 0.37059875E+02, -0.50378239E+02, -0.66598373E+02, -0.30074066E+02, 0.62589401E+02, 
    },
    {
      0.41164993E+02, 0.96164986E+02, -0.17758118E+03, 0.96721039E+01, 0.54648762E+02, 0.41163754E+01, 0.14427828E+02, -0.13174432E+02, -0.55630002E+01, 0.72903724E+01, -0.63024837E+00, -0.61418190E+01, -0.56469097E+01, 
    },
    {
      -0.10944715E+03, 0.17815541E+03, 0.63606415E+02, 0.73619609E+01, 0.64106522E+02, -0.35921402E+01, -0.15911514E+02, -0.52352424E+01, 0.14696769E+02, -0.72762308E+01, 0.22202390E+00, -0.47575579E+01, -0.41712513E+01, 
    },
    {
      0.11644561E+03, -0.11094306E+02, -0.22673264E+03, -0.28864283E+02, -0.28778976E+02, -0.27439207E+02, 0.14110433E+02, 0.86871843E+01, -0.75935063E+01, -0.47815193E+02, -0.36480932E+01, 0.22022610E+02, -0.18813353E+02, 
    },
    {
      0.10188473E+03, -0.10228133E+03, -0.39122402E+02, 0.73224983E+02, 0.96420273E+02, 0.22319057E+02, -0.32373447E+02, 0.63122818E+02, -0.10419510E+02, 0.21842417E+02, 0.27861082E+02, 0.13659900E+02, -0.24893913E+02, 
    },
    {
      0.17436193E-01, -0.93482248E-01, -0.10793444E+00, -0.82785124E+00, -0.84534422E-01, -0.26766161E-03, 0.16654745E-01, -0.30994598E-01, -0.11604786E-01, 0.13113019E-01, -0.68038772E-03, 0.19080637E-02, 0.26341386E-01, 
    },
    {
      0.13251559E-01, -0.39075937E-01, 0.66035584E-01, 0.18244596E-01, -0.78173661E+00, -0.11768348E-01, -0.53608539E-02, -0.28954308E-01, 0.22645099E-01, 0.34665675E-02, 0.17233107E-01, 0.14634843E-01, -0.29612660E-01, 
    },
    {
      0.50487894E+00, 0.22164828E+00, -0.24378799E+00, -0.20211710E+00, 0.94408840E-01, 0.48972866E+00, -0.23864597E-01, -0.27299488E+00, -0.34467158E+00, -0.14131469E+00, -0.20775779E+00, -0.61760709E-01, -0.61581101E-01, 
    },
    {
      0.99109298E+00, 0.17821749E+00, 0.82714492E+00, -0.94977474E+00, -0.61216378E+00, 0.36868054E+00, 0.22957662E+00, -0.12434895E+00, 0.49470678E-01, -0.90065990E-02, -0.12500758E+00, 0.15639452E-01, -0.57548754E-01, 
    },
    {
      0.26369982E+01, 0.56819504E+00, 0.27612624E+01, 0.47631841E+01, 0.26233945E+01, -0.35643741E+00, -0.49008372E+00, 0.86467236E+00, 0.32177621E+00, -0.11061669E+00, 0.54952288E+00, 0.32997720E-01, -0.70062494E+00, 
    },
    {
      0.11445436E+01, -0.10629308E+01, 0.18839436E+01, -0.18157120E+01, 0.57845397E+01, -0.11576208E+00, 0.76202619E+00, 0.12073851E+01, 0.22669688E-01, -0.58742684E+00, -0.53316754E+00, -0.86090761E+00, 0.12829594E+00, 
    },
    {
      -0.12164210E+00, -0.11267498E+00, 0.39087934E+01, 0.14429278E+01, 0.49848485E+01, -0.38739052E+01, 0.32264167E+00, 0.28122408E+01, 0.33922617E+01, 0.10338417E+01, 0.14853586E+01, 0.56888652E+00, 0.14110199E+00, 
    },
    {
      -0.83919153E+01, -0.36753819E+01, -0.42955184E+01, 0.33921528E+01, 0.80299025E+01, -0.31461635E+01, -0.15146472E+01, 0.62426537E+00, 0.49127710E+00, 0.91559422E+00, 0.12294981E+01, -0.11871793E+01, -0.17221960E+00, 
    },
    {
      -0.10749142E+02, 0.90448103E+01, -0.22625639E+02, -0.52351284E+01, 0.13716090E+02, 0.13832469E+00, 0.71097839E+00, -0.13742863E+01, -0.29770541E+01, -0.49114960E+00, -0.18319699E+01, 0.83288515E+00, 0.42117639E+01, 
    },
    {
      -0.74518118E+01, 0.20720312E+02, -0.12630376E+02, -0.16478212E+02, -0.81918631E+01, -0.70487130E+00, -0.88454456E+01, -0.63079481E+01, -0.16885533E+01, 0.26288922E+01, 0.37352061E+01, 0.57972150E+01, 0.16058006E+01, 
    },
    {
      -0.51203427E+01, -0.25459335E+01, -0.16263302E+02, -0.22843137E+01, -0.19856340E+02, 0.82789288E+01, -0.49226767E+00, -0.64728532E+01, -0.86539001E+01, -0.22938893E+01, -0.35961869E+01, -0.14950491E+01, 0.24064369E+00, 
    },
    {
      0.16074701E+02, 0.12596722E+02, 0.77019687E+01, -0.36065149E+01, -0.19908415E+02, 0.71198821E+01, 0.27998273E+01, -0.71557932E-01, -0.21365404E+01, -0.29742064E+01, -0.35151474E+01, 0.40621052E+01, 0.22940133E+01, 
    },
    {
      0.13587055E+02, -0.25019388E+02, 0.54193542E+02, -0.97644215E+01, -0.52721302E+02, 0.34096539E+01, 0.70795673E+00, -0.32761893E+01, 0.75477500E+01, 0.20922942E+01, 0.12183950E+01, -0.31462874E+01, -0.88257294E+01, 
    },
    {
      0.19144426E+02, -0.55912872E+02, 0.20548437E+02, 0.58517231E+02, -0.10258888E+02, 0.43046956E+01, 0.22909641E+02, 0.11171076E+02, 0.56796517E+01, -0.34555612E+01, -0.77860579E+01, -0.11080912E+02, -0.49921532E+01, 
    },
    {
      0.50711031E+01, 0.36411526E+01, 0.15481095E+02, 0.27508340E+01, 0.21559464E+02, -0.52467356E+01, -0.18340144E+00, 0.42208877E+01, 0.59293571E+01, 0.12054518E+01, 0.28709695E+01, 0.10667130E+01, -0.38529634E+00, 
    },
    {
      -0.73677697E+01, -0.12092901E+02, -0.48593469E+01, -0.27885952E+01, 0.14997449E+02, -0.45441275E+01, -0.15541584E+01, -0.77496177E+00, 0.20404882E+01, 0.22645693E+01, 0.28346663E+01, -0.32707062E+01, -0.27267530E+01, 
    },
    {
      -0.64316149E+01, 0.16303364E+02, -0.38149586E+02, 0.12910197E+02, 0.42223900E+02, -0.46120901E+01, -0.76996171E+00, 0.53413100E+01, -0.56315417E+01, -0.12916847E+01, 0.31741199E+00, 0.23768795E+01, 0.58335042E+01, 
    },
    {
      -0.17753616E+02, 0.41826595E+02, -0.77663765E+01, -0.46996101E+02, 0.16903778E+02, -0.39931262E+01, -0.17180531E+02, -0.67168827E+01, -0.45003757E+01, 0.13160985E+01, 0.48622327E+01, 0.64542894E+01, 0.36038089E+01, 
    },
    {
      -0.22608569E+00, 0.22383168E-01, -0.21983469E+00, 0.38383208E-01, 0.10683269E+00, -0.27958038E+00, -0.72073120E+00, -0.13659807E-01, -0.94429441E-02, -0.40688934E-02, 0.14465627E-01, -0.12307435E-01, 0.84743900E-02, 
    },
    {
      0.77261649E-01, 0.52328430E-01, -0.33979189E-01, 0.24548575E-01, 0.15614870E+00, 0.68249077E+00, -0.32020846E+00, -0.23883097E-01, 0.48189063E-01, -0.71453056E-02, 0.65714070E-02, -0.42875452E-03, -0.49787788E-02, 
    },
    {
      0.16318841E+00, 0.40782076E+00, -0.13131221E+00, -0.78088120E-01, -0.43878078E+00, -0.45414293E+00, -0.18014473E+00, 0.51335122E-01, 0.92360303E-01, 0.70260219E-01, 0.40587511E-01, 0.57136446E-01, -0.38172573E-01, 
    },
    {
      -0.32295987E+00, -0.36206935E-01, 0.54615408E+00, 0.15255278E+00, 0.10668567E-02, 0.59954655E+00, -0.58115119E+00, -0.14064194E+00, -0.59324689E-02, 0.17862687E-02, -0.16501710E+00, -0.68641931E-01, 0.14574032E-01, 
    },
    {
      0.22033205E+01, -0.42505041E+00, 0.14705453E+01, 0.99113494E-01, -0.91881227E+00, 0.16044205E+01, 0.17543232E+01, 0.45916408E+00, 0.27655859E-02, -0.15211327E+00, 0.62531821E-01, 0.22174665E+00, -0.89840271E-01, 
    },
    {
      -0.98237640E+00, 0.61098063E+00, -0.93203463E-01, 0.24999487E+00, -0.16881289E+01, -0.12344168E+01, 0.17829297E+01, 0.27928630E+00, -0.16599249E-01, 0.71338467E-01, -0.15047702E+00, 0.21899357E+00, 0.51836595E-01, 
    },
    {
      0.34149848E-01, -0.11778055E+01, 0.41532448E+00, 0.15049255E+00, 0.86920917E+00, 0.51964009E+00, -0.20884304E+00, 0.89135788E-01, -0.16401221E+00, -0.20817281E+00, -0.11379710E+00, -0.21582046E+00, 0.54201890E-01, 
    },
    {
      0.65804678E+00, 0.50968662E-01, -0.83104873E+00, -0.68715566E+00, 0.75480483E-01, -0.84142131E+00, 0.86892831E+00, 0.25009820E+00, -0.44691864E-01, 0.41970626E-01, 0.42782474E+00, 0.50102556E-02, -0.18369292E+00, 
    },
    {
      -0.30072827E+01, 0.10538676E+01, -0.18198307E+01, 0.25968856E+00, 0.99997920E+00, -0.23609509E+01, -0.17743442E+01, -0.73138219E+00, 0.11850427E+00, 0.35192615E+00, -0.23847862E+00, -0.57276577E+00, 0.14691012E+00, 
    },
    {
      0.11595041E+01, -0.71592605E+00, 0.47009271E+00, 0.22766799E+00, 0.25950408E+01, 0.12390486E+01, -0.26281445E+01, -0.37698331E+00, -0.21818750E+00, -0.15531531E+00, 0.19898316E+00, -0.42630193E+00, -0.24251965E+00, 
    },
    {
      0.75564645E-01, -0.15819146E-02, -0.17070128E-01, 0.62731832E-01, 0.80122389E-01, -0.19053893E-01, -0.75502822E-03, 0.20281678E+00, -0.16062942E+00, -0.61040525E-02, 0.32858420E-01, -0.53751818E-02, 0.11075356E-01, 
    },
    {
      0.84059753E-01, -0.26932992E-02, 0.44004381E-01, -0.30800799E-01, 0.69791381E-02, 0.19021643E-01, -0.30156031E-01, 0.18149000E+00, 0.21475334E+00, -0.98471157E-02, -0.55861087E-02, 0.48251478E-02, -0.33545550E-02, 
    },
    {
      0.15172923E+00, -0.10189057E+00, 0.25645059E-01, -0.52431297E-01, -0.31783830E-01, -0.11050274E+00, 0.11658210E-01, -0.22356764E-02, -0.46832764E+00, -0.37870400E-01, -0.47675385E-02, -0.38657390E-01, 0.74187763E-01, 
    },
    {
      0.38575072E-01, -0.21887577E+00, 0.43412637E-01, 0.79081804E-01, 0.39201990E-01, 0.11154531E-01, -0.64383946E-01, 0.39611334E+00, -0.98804720E-02, 0.39850544E-01, -0.27133226E-01, 0.39764382E-01, -0.42043138E-01, 
    },
    {
      -0.27693447E-02, 0.97743109E-01, -0.95391117E-01, 0.11043071E-01, 0.31084713E-01, -0.40372331E-02, 0.13213970E-01, -0.24805332E-01, 0.15137902E-01, 0.56641247E-01, 0.22465751E+00, -0.23782956E-02, 0.96832955E-03, 
    },
    {
      0.56767878E-02, 0.27197037E-01, 0.41327935E-01, -0.92214338E-01, -0.47889195E-01, -0.41702380E-02, -0.81137735E-02, 0.41315701E-01, -0.75267494E-01, -0.20449100E+00, 0.59991557E-01, -0.68433122E-02, 0.14528462E-01, 
    },
    {
      0.14191374E-01, -0.28540870E-01, 0.25324695E-01, -0.79455242E-01, -0.23113195E-01, 0.44245183E-01, 0.34820773E-01, -0.90411389E-02, -0.25287878E-01, 0.31344425E-02, 0.10355319E-01, -0.10612031E+00, 0.21018282E-01, 
    },
    {
      -0.77975765E-02, -0.76840571E-02, 0.15142567E-01, -0.22948675E-01, -0.47588389E-01, 0.11881715E-01, -0.63865617E-01, 0.30425686E-01, -0.82677780E-02, 0.21936271E-01, 0.12532286E-01, -0.44801785E-02, -0.13164043E+00, 
    },
    {
      -0.46902079E-01, -0.13513894E-01, 0.16624622E-03, -0.26316926E-01, -0.12186967E-02, 0.28734948E-01, -0.22959083E-02, 0.10504897E-01, -0.78018638E-02, 0.84232548E-02, -0.90076961E-02, -0.66834749E-02, 0.13984973E-01, 
    },
    {
      -0.61138693E-01, 0.11805169E-01, 0.78700714E-01, 0.18015917E-01, 0.28624056E-01, -0.76267403E-02, -0.46754625E-01, -0.22486155E-02, 0.15880592E-01, 0.64636157E-02, 0.14192480E-01, -0.67947758E-03, -0.35506718E-01, 
    },
    {
      -0.25783082E-01, 0.39371252E-01, -0.79181977E-02, -0.47959839E-02, -0.34519922E-01, -0.25952054E-01, -0.92375092E-02, 0.23945650E-01, -0.36872244E-02, 0.38071219E-01, 0.14242783E-01, 0.10948616E-01, 0.21540886E-01, 
    },
    {
      -0.42456158E-01, 0.11031415E-01, 0.34051061E-01, 0.57469562E-01, 0.45886550E-01, 0.13773274E-01, -0.64351404E-03, -0.16607651E-01, 0.18327732E-01, -0.26219755E-02, -0.77091181E-02, 0.13056586E-01, -0.54514217E-02, 
    },
  },
  {
    {
      0.94509640E+01, 0.10107969E+00, 0.95546111E-01, 0.34959666E-01, -0.25399696E-01, 0.18447235E-01, -0.18753277E-01, -0.21336963E-01, 0.20273566E-01, -0.74352040E-02, 0.86263977E-02, -0.19587195E-01, -0.25863366E-02, 
    },
    {
      0.17576997E+01, -0.20958441E+00, 0.20107350E+00, 0.26883551E+00, -0.69596463E+00, 0.10864495E+00, -0.57094645E-01, -0.11880032E+00, 0.32511491E+00, -0.73277988E-01, -0.19179016E+00, 0.55194344E-01, -0.48167195E-01, 
    },
    {
      0.29966228E+02, -0.28514745E+01, -0.27479420E+01, 0.16417269E+00, 0.23762665E+01, -0.24790712E+00, 0.38372591E-01, 0.68017262E+00, -0.64208418E+00, 0.19077405E+00, -0.19376384E+00, 0.37586835E+00, 0.30274448E+00, 
    },
    {
      -0.79763260E+01, 0.40185118E+01, 0.27582934E+01, -0.44167314E+01, 0.16556219E+02, -0.10127513E+01, -0.14381729E-01, 0.10526127E+01, -0.36751299E+01, 0.16371242E+01, 0.41598487E+01, -0.55727178E+00, 0.25336441E+00, 
    },
    {
      -0.15115533E+03, 0.20107765E+02, 0.16070599E+02, -0.69206400E+01, -0.15509940E+02, -0.65745495E-01, 0.61048579E+00, -0.44110274E+01, 0.59234338E+01, -0.19652240E+01, 0.83969170E+00, -0.19654928E+01, -0.22424846E+01, 
    },
    {
      -0.51554680E+02, -0.49286888E+02, -0.15149562E+02, 0.23192854E+02, -0.80976891E+02, 0.17976050E+01, 0.31771953E+01, -0.22073927E+01, 0.17565590E+02, -0.79421062E+01, -0.22982227E+02, 0.20799625E+01, 0.11556675E+01, 
    },
    {
      0.26708896E+03, -0.49775471E+02, -0.31839897E+02, 0.28108734E+02, 0.37422558E+02, 0.36079819E+01, -0.32567403E+01, 0.10001911E+02, -0.17851255E+02, 0.60564766E+01, -0.76352984E+00, 0.37349446E+01, 0.60045552E+01, 
    },
    {
      0.22564854E+03, 0.16248386E+03, 0.12195727E+02, -0.51379269E+02, 0.16426907E+03, 0.24047847E+01, -0.12291703E+02, -0.43176427E+00, -0.39148830E+02, 0.16007685E+02, 0.51836742E+02, -0.38026702E+01, -0.54714437E+01, 
    },
    {
      -0.22505072E+03, 0.46291023E+02, 0.24638092E+02, -0.39034763E+02, -0.37432152E+02, -0.72609534E+01, 0.49040074E+01, -0.89484711E+01, 0.20959160E+02, -0.69502692E+01, -0.58399487E+00, -0.28382370E+01, -0.64945631E+01, 
    },
    {
      -0.28448126E+03, -0.19788374E+03, 0.13808954E+02, 0.50430801E+02, -0.14956322E+03, -0.69601116E+01, 0.16451906E+02, 0.45007586E+01, 0.39577290E+02, -0.14559422E+02, -0.51503658E+02, 0.33227439E+01, 0.67428279E+01, 
    },
    {
      0.75200890E+02, -0.13670172E+02, -0.64882317E+01, 0.17861347E+02, 0.13128088E+02, 0.39957590E+01, -0.23009706E+01, 0.26648011E+01, -0.84337873E+01, 0.26480327E+01, 0.69651747E+00, 0.70226914E+00, 0.24378769E+01, 
    },
    {
      0.11641260E+03, 0.81135353E+02, -0.14079868E+02, -0.17968302E+02, 0.50422318E+02, 0.35681407E+01, -0.73284211E+01, -0.27828982E+01, -0.14672050E+02, 0.49403863E+01, 0.18719210E+02, -0.10883954E+01, -0.26159148E+01, 
    },
    {
      -0.12469147E-01, 0.13936483E+01, 0.20418575E+01, -0.22939935E-01, 0.33805311E-01, 0.10366075E-01, 0.52215006E-01, -0.36718503E-01, -0.45215521E-01, 0.20861069E-02, 0.31817243E-01, -0.47385283E-02, -0.63512935E-02, 
    },
    {
      -0.79864658E-01, -0.21967120E+01, 0.13169786E+01, -0.49137101E-01, 0.79261839E-01, 0.56432478E-01, -0.11213575E-01, 0.39652806E-01, -0.29792653E-01, 0.64580194E-02, -0.23676369E-02, 0.21759911E-01, -0.17042236E-01, 
    },
    {
      -0.10706545E+01, 0.23217793E+01, 0.22575028E+01, -0.52305853E+00, 0.33758911E+00, 0.40237907E+00, 0.70796597E+00, -0.25653306E+00, 0.18688981E+00, -0.25361739E-01, 0.87322183E-01, 0.24367885E+00, 0.40593699E-01, 
    },
    {
      -0.15108147E+00, -0.16911287E+01, 0.46142992E+00, -0.10303921E+01, 0.93612778E+00, -0.41574039E-02, 0.11723827E+00, -0.15480372E+00, 0.25974426E-01, -0.10854245E+00, -0.14955416E+00, 0.20788647E+00, 0.12922181E+00, 
    },
    {
      0.45731859E+01, 0.24417906E+02, -0.21377110E+02, -0.15830410E+00, 0.28918290E+01, -0.90342098E+00, -0.10628958E+01, 0.70779812E+00, 0.28248230E+00, 0.51816136E+00, -0.36014694E+00, -0.21408524E+00, 0.44734198E+00, 
    },
    {
      -0.10066270E+02, 0.24201607E+02, 0.22228245E+02, 0.18787283E+01, 0.35598913E+00, -0.81920505E+00, -0.45490566E+00, 0.21012618E+00, 0.17021557E+01, -0.23625430E+00, -0.15785757E+00, -0.44948339E+00, -0.50422323E+00, 
    },
    {
      0.25363052E+02, -0.22388351E+02, -0.88549417E+00, -0.22121559E+00, 0.23813438E+01, -0.34516563E+01, -0.12210044E+02, 0.55636535E+01, -0.32453895E+01, 0.19233570E+00, -0.14384522E+01, -0.48079529E+01, -0.13344591E+01, 
    },
    {
      0.43639555E+01, -0.82982798E+01, 0.11524791E+02, 0.18317818E+02, -0.13572060E+02, 0.43871188E+01, -0.21051328E+01, 0.19665152E+01, -0.18783902E+01, 0.23498127E+01, 0.49273577E+01, -0.22936010E+01, -0.35927687E+01, 
    },
    {
      -0.33341885E+02, -0.11396944E+03, 0.19362669E+03, 0.21216767E+01, -0.21863722E+02, 0.87932034E+01, 0.66870275E+01, -0.20778823E+01, 0.39056785E+01, -0.55137281E+01, 0.73257333E+00, 0.10095549E+01, -0.58631840E+01, 
    },
    {
      0.10840975E+03, -0.21727190E+03, -0.68914238E+02, -0.21362246E+02, -0.86977644E+01, 0.82116756E+01, 0.49080110E+01, 0.17874563E+00, -0.12050731E+02, 0.51833916E+00, 0.31218398E+01, 0.36935799E+01, 0.80403614E+01, 
    },
    {
      -0.15367139E+03, 0.11878262E+03, 0.51804714E+02, 0.41489388E+02, -0.92071486E+01, 0.51096478E+01, 0.64470383E+02, -0.44764999E+02, 0.16702759E+02, 0.28305548E+00, 0.86546431E+01, 0.29645947E+02, 0.80854740E+01, 
    },
    {
      -0.86085358E+01, 0.41116173E+02, -0.47850399E+02, -0.11364082E+03, 0.49419270E+02, -0.27409754E+02, 0.12218876E+02, -0.15123549E+02, 0.17315872E+02, -0.15716614E+02, -0.34133495E+02, 0.81285315E+01, 0.23876568E+02, 
    },
    {
      0.82036057E+02, 0.18187785E+03, -0.55807098E+03, -0.10879265E+02, 0.61412319E+02, -0.27611143E+02, -0.19147324E+02, -0.71760750E+00, -0.27551128E+02, 0.17640097E+02, -0.56628734E-01, -0.14020643E+01, 0.21747429E+02, 
    },
    {
      -0.34376413E+03, 0.61911011E+03, 0.88020509E+00, 0.80663330E+02, 0.33479347E+02, -0.29412020E+02, -0.12201406E+02, -0.44523067E+01, 0.36756096E+02, 0.16952411E+01, -0.14172467E+02, -0.11445619E+02, -0.29532711E+02, 
    },
    {
      0.39695712E+03, -0.31497382E+03, -0.28771411E+03, -0.17258989E+03, -0.86242523E+01, 0.11642606E+02, -0.14586752E+03, 0.13591988E+03, -0.39550259E+02, -0.15524073E+01, -0.22016012E+02, -0.76522240E+02, -0.19635260E+02, 
    },
    {
      -0.40375984E+02, -0.67061432E+02, 0.28259424E+02, 0.29492319E+03, -0.67607613E+02, 0.62555237E+02, -0.28627241E+02, 0.50263222E+02, -0.52197632E+02, 0.39175499E+02, 0.90898705E+02, -0.12516282E+02, -0.61318790E+02, 
    },
    {
      -0.77070221E+02, -0.11974761E+03, 0.69099872E+03, 0.18479025E+02, -0.77280106E+02, 0.33186859E+02, 0.23168581E+02, 0.75805955E+01, 0.50765301E+02, -0.22174515E+02, -0.21018499E+00, 0.45265010E+00, -0.30652090E+02, 
    },
    {
      0.42491214E+03, -0.73491571E+03, 0.15072490E+03, -0.11781622E+03, -0.46392895E+02, 0.41384529E+02, 0.10866575E+02, 0.84837999E+01, -0.48217793E+02, -0.58226819E+01, 0.23120079E+02, 0.13756743E+02, 0.40079506E+02, 
    },
    {
      -0.45912537E+03, 0.38094446E+03, 0.50466553E+03, 0.24710139E+03, 0.43194183E+02, -0.31068718E+02, 0.14789594E+03, -0.17133972E+03, 0.43186558E+02, 0.93547171E+00, 0.24350426E+02, 0.87021637E+02, 0.20689638E+02, 
    },
    {
      0.12518957E+03, 0.21702751E+02, 0.59070374E+02, -0.33266949E+03, 0.31520334E+02, -0.63018490E+02, 0.29843792E+02, -0.66842224E+02, 0.62429546E+02, -0.40911572E+02, -0.10445174E+03, 0.77937250E+01, 0.66863708E+02, 
    },
    {
      0.23716257E+02, 0.26841919E+02, -0.31155273E+03, -0.99036121E+01, 0.36229214E+02, -0.13366701E+02, -0.94885092E+01, -0.60028815E+01, -0.28343122E+02, 0.97320232E+01, -0.21416096E+00, 0.32326192E+00, 0.14711193E+02, 
    },
    {
      -0.18264751E+03, 0.31139908E+03, -0.10778708E+03, 0.57750397E+02, 0.20604429E+02, -0.19741055E+02, -0.31399972E+01, -0.45069766E+01, 0.22200363E+02, 0.40623951E+01, -0.12418812E+02, -0.53976035E+01, -0.18255384E+02, 
    },
    {
      0.19313271E+03, -0.16603587E+03, -0.27830505E+03, -0.11844001E+03, -0.29174128E+02, 0.17652523E+02, -0.55044979E+02, 0.75791016E+02, -0.17480881E+02, 0.37095240E+00, -0.96177435E+01, -0.35863873E+02, -0.79011192E+01, 
    },
    {
      -0.83767029E+02, 0.14769975E+02, -0.53073029E+02, 0.13477794E+03, -0.10499868E+01, 0.24099756E+02, -0.11736704E+02, 0.30020025E+02, -0.25523891E+02, 0.15339497E+02, 0.43580658E+02, -0.11596680E+01, -0.26010191E+02, 
    },
    {
      -0.24157926E-01, -0.34682450E+00, -0.18939769E+00, -0.64226544E+00, 0.55502748E+00, -0.18019242E-02, 0.22090345E-02, -0.17449288E-02, -0.28906872E-01, 0.43518983E-01, -0.22777419E-02, 0.19499291E-01, 0.26698962E-01, 
    },
    {
      -0.66926338E-01, -0.13376318E+00, -0.66897385E-01, -0.60208154E+00, -0.70030004E+00, -0.11302370E-01, 0.22896841E-01, 0.18626200E-01, 0.16040893E-01, 0.18582191E-01, -0.50697462E-02, 0.14439798E-01, -0.25851393E-01, 
    },
    {
      0.14338864E+00, -0.20765391E+00, 0.20112073E-01, -0.19753712E+01, 0.36831617E+00, 0.33009917E+00, -0.17354338E+00, 0.28956002E-02, -0.41839543E+00, -0.34900645E-02, -0.25787419E+00, 0.39183512E-01, 0.11534757E+00, 
    },
    {
      0.57123256E+00, 0.13649093E+00, 0.26264960E-01, -0.47044334E+00, -0.22277884E+01, 0.34276772E+00, 0.26841480E+00, -0.31208158E+00, 0.21924429E+00, -0.73497251E-01, -0.54401502E-01, 0.14796603E-01, -0.10233092E-01, 
    },
    {
      0.26954265E+01, 0.43012781E+01, 0.75683284E+01, -0.86731186E+01, 0.55410342E+01, -0.34549937E+00, -0.13151923E+01, 0.79660070E+00, 0.91225111E+00, -0.41482779E+00, 0.31972158E+00, -0.65595812E+00, -0.98114915E-01, 
    },
    {
      0.17693305E+01, 0.10319415E+01, 0.57046089E+01, -0.44164987E+01, -0.76197066E+01, 0.69258058E+00, -0.58791065E+00, -0.44445005E+00, 0.10734792E+01, -0.10839006E+01, 0.18576832E+00, -0.56307358E+00, -0.81509627E-01, 
    },
    {
      0.19725923E+00, 0.28090067E+01, 0.29866964E-01, 0.65746832E+01, -0.11214925E+01, -0.20306814E+01, 0.75946403E+00, 0.37662435E+00, 0.41457515E+01, 0.45575309E+00, 0.18497028E+01, -0.14746182E+00, -0.16188046E+01, 
    },
    {
      -0.33737960E+01, -0.32643378E+00, 0.13863986E+01, 0.35408862E+01, 0.94330416E+01, -0.89115298E+00, -0.36629066E+00, 0.16838450E+01, -0.21706350E+01, 0.13435290E+01, 0.55284905E+00, -0.85979068E+00, -0.11613747E+01, 
    },
    {
      -0.12316348E+02, -0.62260914E+01, -0.65585999E+02, 0.51939068E+02, -0.18281101E+02, 0.18273875E+00, 0.63747220E+01, -0.31539526E+01, -0.72625136E+01, -0.33774602E+00, -0.64533216E+00, 0.45377502E+01, -0.94161624E+00, 
    },
    {
      -0.12553596E+02, 0.13752454E+02, -0.35581814E+02, 0.14409949E+02, 0.47376873E+02, -0.34375093E+01, 0.25728254E+01, 0.43016977E+01, -0.10728865E+02, 0.57186842E+01, -0.11502291E+01, 0.40079656E+01, 0.17250158E+01, 
    },
    {
      -0.35611405E+01, -0.97855129E+01, -0.59383321E+01, -0.11810493E+02, -0.76464243E+01, 0.37008057E+01, -0.16280932E+01, -0.65146512E+00, -0.10992208E+02, -0.11756465E+01, -0.40906620E+01, -0.35428894E+00, 0.46568985E+01, 
    },
    {
      0.40749516E+01, -0.54280490E+00, -0.29149284E+01, -0.16286656E+01, -0.16713770E+02, 0.47613055E+00, -0.20696537E+01, -0.19630735E+01, 0.53624787E+01, -0.44261832E+01, -0.13592387E+01, 0.33461597E+01, 0.42439528E+01, 
    },
    {
      0.19575254E+02, -0.26728251E+01, 0.15306610E+03, -0.97623291E+02, 0.11181700E+02, 0.33377397E+01, -0.11486933E+02, 0.41499352E+01, 0.16440626E+02, 0.37850554E+01, -0.18400379E+00, -0.10662279E+02, 0.31876299E+01, 
    },
    {
      0.31439281E+02, -0.47513309E+02, 0.64092300E+02, -0.10152738E+02, -0.89982887E+02, 0.79072332E+01, -0.60005236E+01, -0.98099136E+01, 0.26642679E+02, -0.94802303E+01, 0.32628336E+01, -0.80624046E+01, -0.41372290E+01, 
    },
    {
      0.24437640E+01, 0.10388881E+02, 0.68885064E+01, 0.12287474E+02, 0.17508669E+02, -0.15349623E+01, 0.13399680E+01, 0.89213789E-01, 0.77907591E+01, 0.79130781E+00, 0.28016553E+01, 0.53495371E+00, -0.36855667E+01, 
    },
    {
      0.68383670E+00, 0.14101259E+00, 0.12219146E+01, -0.87792912E+01, 0.14048577E+02, -0.40416885E-01, 0.34328499E+01, 0.43522057E+00, -0.37442031E+01, 0.36730547E+01, 0.10144058E+01, -0.31850066E+01, -0.35276756E+01, 
    },
    {
      -0.11864492E+02, 0.65092659E+01, -0.10362085E+03, 0.60457390E+02, 0.88302031E+01, -0.48034692E+01, 0.68933558E+01, -0.17215779E+01, -0.11243499E+02, -0.31618342E+01, 0.66602868E+00, 0.76481528E+01, -0.22676835E+01, 
    },
    {
      -0.26041988E+02, 0.39193073E+02, -0.33144814E+02, -0.53187180E+01, 0.56564884E+02, -0.62643838E+01, 0.45099568E+01, 0.61439466E+01, -0.18840761E+02, 0.46164675E+01, -0.28137391E+01, 0.49374666E+01, 0.28067627E+01, 
    },
    {
      -0.25802639E+00, -0.10362916E+00, -0.14398941E+00, 0.58753002E-01, 0.12338717E+00, -0.46615648E+00, -0.90974379E+00, 0.48583642E-01, -0.38379822E-01, 0.27855985E-01, 0.10477463E-02, -0.31467213E-02, 0.46931063E-02, 
    },
    {
      0.13942069E+00, 0.12052406E-01, 0.13556781E+00, -0.54092322E-01, 0.15451410E+00, 0.88882434E+00, -0.48357457E+00, -0.32771014E-01, 0.11079021E-01, -0.23551969E-01, -0.56003863E-02, -0.15953735E-01, 0.25094349E-02, 
    },
    {
      0.20348473E+00, 0.49829680E+00, -0.55318739E-01, -0.92830181E-01, -0.18149967E+00, -0.54599792E+00, -0.81165332E+00, 0.41849393E-01, 0.22468764E+00, 0.12768607E-01, -0.45675814E-01, 0.34527790E-01, 0.90832688E-01, 
    },
    {
      0.19808038E+00, 0.24423756E-01, 0.55058855E+00, 0.96684530E-01, 0.14878929E+00, 0.10547436E+01, -0.62313193E+00, -0.34745380E+00, 0.50657775E-01, -0.10458727E+00, -0.26829256E-01, -0.22105072E-01, 0.96443847E-01, 
    },
    {
      0.31103616E+01, 0.71752518E+00, 0.43198234E+00, -0.83313775E+00, -0.53591412E+00, 0.21094110E+01, 0.14961641E+01, 0.52341461E+00, 0.56767017E+00, -0.88358931E-01, -0.25685966E-01, -0.62317005E-02, 0.82329907E-01, 
    },
    {
      -0.23380406E+01, 0.12645091E+01, -0.13486705E+01, 0.41103089E+00, -0.19716953E+01, -0.13170983E+01, 0.23266773E+01, 0.89183159E-01, 0.99011421E+00, 0.14776289E+00, -0.96008055E-01, 0.22073540E+00, 0.14917378E-01, 
    },
    {
      0.83357565E-01, -0.13930918E+01, 0.16632780E+00, 0.31476656E+00, 0.29862192E+00, -0.63334607E-01, 0.78159118E+00, -0.13982970E-01, -0.37854099E+00, -0.40592089E-01, 0.14077954E+00, -0.17048199E+00, -0.20473443E+00, 
    },
    {
      -0.39951530E+00, 0.38091745E-01, -0.17263715E+01, -0.53047764E+00, -0.42756036E-01, -0.14003088E+01, 0.27357009E+00, 0.70787126E+00, -0.17673521E+00, 0.21879543E+00, 0.80344655E-01, -0.11453548E+00, -0.28751436E+00, 
    },
    {
      -0.55554628E+01, 0.63528180E-01, 0.22511274E+00, 0.16389220E+01, 0.68950057E-01, -0.38856034E+01, -0.18219573E+01, -0.12218167E+01, -0.72119141E+00, 0.12372471E-01, 0.60922831E-01, -0.13905819E-01, -0.64103089E-01, 
    },
    {
      0.34650526E+01, -0.16690048E+01, 0.22416811E+01, -0.33316332E+00, 0.29141891E+01, 0.15945005E+01, -0.39390171E+01, 0.14417328E+00, -0.18604945E+01, -0.11784468E+00, -0.16993752E-01, -0.41759604E+00, -0.18586282E+00, 
    },
    {
      0.81547573E-02, 0.49226344E-01, 0.81611373E-01, 0.27063418E-01, 0.56695234E-01, 0.29357164E-02, 0.14539908E-01, 0.30828521E+00, -0.25350037E+00, 0.32837324E-01, 0.15210060E-01, -0.29200683E-02, 0.44837701E-02, 
    },
    {
      0.63956775E-01, 0.18061155E-01, 0.10958803E-01, -0.12470478E-01, 0.35997184E-02, -0.25639519E-01, -0.23812983E-01, 0.26933238E+00, 0.31840917E+00, -0.11766966E-02, -0.27984953E-02, -0.12501192E-02, -0.80634849E-02, 
    },
    {
      0.14839776E+00, -0.10045361E+00, 0.11778341E-01, 0.67010596E-01, 0.10346636E+00, -0.71791507E-01, 0.17526034E-01, 0.13578543E+00, -0.68693244E+00, -0.82893714E-01, 0.53929277E-02, 0.23169667E-02, 0.82021438E-01, 
    },
    {
      0.72574534E-01, -0.60684308E-01, -0.86023420E-01, 0.90452895E-01, 0.34990810E-01, -0.15001757E+00, -0.22994665E-01, 0.55149466E+00, 0.16523851E+00, -0.21080254E-01, -0.10785421E-01, 0.24345174E-01, -0.54076161E-01, 
    },
    {
      -0.14327258E-01, 0.63852847E-01, -0.25183151E-01, 0.29911581E-01, 0.37023772E-01, 0.14488783E-01, -0.15269484E-01, -0.60717769E-01, 0.13028531E-01, 0.14004098E+00, 0.26168507E+00, -0.16771069E-01, 0.91431160E-02, 
    },
    {
      0.12888256E+00, 0.18391632E-02, -0.29771950E-01, -0.61108533E-01, 0.42996099E-02, -0.14745311E-01, -0.18162554E-01, 0.26890341E-01, -0.88421255E-01, -0.26770899E+00, 0.11803225E+00, -0.93428353E-02, 0.73491596E-02, 
    },
    {
      0.42300068E-01, -0.45630768E-01, 0.12768690E-01, -0.31926125E-01, 0.13576030E-01, 0.21456007E-01, 0.21313623E-01, -0.25462057E-01, -0.38112130E-01, -0.38594943E-02, 0.38873557E-01, -0.10141932E+00, 0.36759220E-01, 
    },
    {
      0.96034102E-01, 0.24284817E-01, -0.10095998E+00, -0.96809138E-02, -0.16556034E-01, 0.69916956E-02, -0.73311739E-01, 0.47385052E-01, -0.22124328E-01, -0.81969174E-02, 0.11834414E-01, -0.44802796E-01, -0.12607549E+00, 
    },
    {
      -0.29452965E-01, -0.18742386E-01, -0.76822206E-01, 0.27353512E-01, 0.59363917E-02, 0.14614475E-01, -0.26352966E-01, -0.18402827E-02, -0.14121162E-01, -0.11845143E-01, -0.38414609E-02, -0.61304052E-02, -0.21523267E-02, 
    },
    {
      0.61670173E-01, 0.75453207E-01, 0.65484559E-02, -0.33840731E-01, 0.21261007E-01, -0.11356534E-01, -0.17962535E-02, 0.23999993E-01, 0.22581514E-01, -0.91374582E-02, -0.62641094E-03, -0.11860891E-01, -0.49161039E-01, 
    },
    {
      -0.25092568E-01, 0.65070137E-01, -0.17101344E-01, -0.30995658E-01, -0.48828226E-01, -0.25985174E-01, 0.12641994E-01, 0.94131026E-02, 0.50771046E-01, 0.10564291E-01, -0.72512310E-02, -0.25239275E-02, 0.16289363E-01, 
    },
    {
      -0.34294993E-01, -0.21341737E-01, -0.14802243E-01, 0.52205998E-01, 0.39956398E-01, 0.15908470E-01, 0.47674975E-02, 0.53186305E-02, 0.79527143E-02, -0.18043905E-01, 0.74231257E-02, 0.87468559E-02, 0.36759220E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir21_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30515985E+01, 0.18279366E-01, 0.17321778E-01, 0.13490525E-02, 0.54005985E-02, -0.12811268E-02, -0.19620711E-02, 0.16095781E-02, 0.14036379E-02, 
    },
    {
      0.36024821E+00, -0.26675237E-01, 0.60383171E-01, 0.85761592E-01, 0.32015789E-01, -0.64746398E-02, -0.23822611E-01, 0.12654256E-02, 0.19042149E-01, 
    },
    {
      0.84638625E+00, -0.16989470E+00, -0.25357148E+00, -0.44812657E-01, -0.90480018E-02, 0.13199283E-01, -0.49734190E-02, -0.90144098E-01, -0.46962071E-01, 
    },
    {
      -0.31312117E+00, -0.13555598E+00, -0.11967962E+00, -0.20256715E+00, -0.13136697E+00, 0.48089355E-01, 0.87422676E-01, 0.93227997E-02, -0.46409301E-01, 
    },
    {
      -0.11959354E+01, 0.43471059E+00, 0.63000107E+00, 0.10890940E+00, -0.57202525E-01, -0.42029072E-01, 0.43181811E-01, 0.22912052E+00, 0.13201462E+00, 
    },
    {
      -0.73961414E-01, 0.12270621E+00, 0.51998906E-01, 0.10816619E+00, 0.11332868E+00, -0.50417311E-01, -0.68393320E-01, -0.12005514E-01, 0.23018699E-01, 
    },
    {
      0.42550960E+00, -0.27673936E+00, -0.39599004E+00, -0.66320188E-01, 0.74589588E-01, 0.29583050E-01, -0.38249172E-01, -0.14549865E+00, -0.96091665E-01, 
    },
    {
      -0.81040589E-02, -0.14526182E+00, -0.32619151E+00, 0.83900504E-02, 0.61066765E-02, -0.79448353E-02, 0.88535789E-02, 0.61606448E-02, 0.44978694E-02, 
    },
    {
      0.34204829E-01, 0.34175068E+00, -0.15657048E+00, -0.35750011E-02, 0.16515717E-02, 0.34812370E-02, -0.10691710E-01, -0.10663674E-01, 0.23136511E-02, 
    },
    {
      -0.77340782E-01, 0.29641938E+00, 0.46389941E-02, 0.60751557E-01, 0.13668638E+00, 0.73995506E-02, 0.36284521E-01, 0.43617327E-01, -0.78822457E-03, 
    },
    {
      -0.23608712E-01, 0.67810323E-02, 0.38915452E+00, -0.11137573E+00, -0.67211337E-01, 0.10261582E+00, -0.20004436E-01, -0.96985581E-03, 0.74450895E-01, 
    },
    {
      0.24482937E+00, 0.68339103E+00, 0.12487189E+01, -0.17763805E+00, -0.11472326E+00, 0.13915654E+00, -0.96532442E-01, -0.11791184E+00, -0.16605750E+00, 
    },
    {
      -0.76662284E+00, -0.14496020E+01, 0.84100413E+00, 0.16240078E+00, 0.76123066E-01, -0.33510583E-04, 0.83549894E-01, 0.23060144E+00, -0.16083211E+00, 
    },
    {
      0.14090075E+01, -0.13704493E+01, 0.82855475E+00, 0.23537932E+00, -0.70777285E+00, -0.28288743E+00, -0.19857667E+00, -0.45078415E+00, -0.24828870E-01, 
    },
    {
      0.88282007E+00, -0.41681415E+00, -0.20059464E+01, 0.15218285E+00, 0.75277686E+00, -0.67411071E+00, 0.11159578E+00, -0.46602711E-01, -0.56245166E+00, 
    },
    {
      -0.76950717E+00, -0.10647525E+01, -0.34677619E+00, 0.63541669E+00, 0.15894818E+00, -0.45485401E+00, 0.20325267E+00, 0.27888203E+00, 0.53756195E+00, 
    },
    {
      0.23074839E+01, 0.57445264E+00, -0.17256494E+01, -0.20322417E+00, -0.25591567E+00, -0.21160703E-01, -0.19882099E+00, -0.64506978E+00, 0.51714456E+00, 
    },
    {
      -0.38324997E+01, 0.20495992E+01, -0.12175716E+01, -0.10217855E+01, 0.12773292E+01, 0.99618143E+00, 0.27631152E+00, 0.99627531E+00, 0.13893232E-01, 
    },
    {
      -0.26958106E+01, 0.34171972E+00, 0.34515679E+01, 0.34862801E+00, -0.16408192E+01, 0.13690850E+01, -0.63893899E-01, 0.18333472E+00, 0.11947597E+01, 
    },
    {
      0.52510667E+00, 0.79661417E+00, -0.27414501E+00, -0.59034508E+00, -0.29286742E-01, 0.33197421E+00, -0.11383864E+00, -0.16956963E+00, -0.43929490E+00, 
    },
    {
      -0.18387926E+01, 0.63808429E+00, 0.13265423E+01, -0.18234240E-01, 0.18122619E+00, 0.92859911E-02, 0.90937927E-01, 0.46281740E+00, -0.37837014E+00, 
    },
    {
      0.27536213E+01, -0.96891516E+00, 0.93679023E+00, 0.81044078E+00, -0.73057920E+00, -0.86545914E+00, -0.13849419E+00, -0.62656611E+00, 0.88995732E-02, 
    },
    {
      0.20475008E+01, -0.35984865E+00, -0.18440652E+01, -0.44248223E+00, 0.98484981E+00, -0.84341449E+00, -0.35579182E-01, -0.14597042E+00, -0.76683801E+00, 
    },
    {
      -0.10489090E-01, -0.39877072E-01, -0.26295239E-01, 0.73930889E-01, -0.10012757E+00, -0.76692938E-02, 0.43066703E-02, -0.10926161E-01, -0.48446790E-02, 
    },
    {
      0.18568033E-01, -0.17637935E-01, -0.15594788E-01, 0.10350185E+00, 0.84533416E-01, -0.65629520E-02, -0.97999501E-03, 0.40845182E-02, -0.18347696E-02, 
    },
    {
      0.30939199E-01, 0.14365974E+00, 0.14420240E+00, -0.77935576E-01, 0.16382048E+00, -0.65682996E-02, -0.39891735E-01, 0.59632142E-02, -0.52234367E-01, 
    },
    {
      -0.16920580E+00, 0.51358785E-01, 0.32620181E-02, -0.74220739E-01, 0.15272126E-01, 0.58219019E-01, 0.18921616E-01, 0.17532919E-01, 0.32072235E-01, 
    },
    {
      0.11713439E-01, 0.24454077E+00, 0.22597702E+00, -0.16024730E+00, -0.13617696E+00, 0.23809751E-01, 0.65214179E-01, 0.66272601E-01, -0.77504419E-01, 
    },
    {
      -0.36200440E+00, 0.25575509E-01, -0.46344455E-01, 0.24129643E+00, -0.14844865E+00, -0.63828021E-01, 0.11078907E+00, -0.16530167E-01, -0.12369106E-01, 
    },
    {
      -0.18463311E+00, -0.74005991E+00, -0.46722332E+00, 0.24123137E+00, -0.84606230E+00, 0.38615707E-01, 0.21995303E+00, -0.47678851E-01, 0.23793846E+00, 
    },
    {
      0.10495080E+01, 0.73820189E-01, 0.17725090E-01, 0.56766880E+00, 0.60886084E-02, -0.12156435E+00, -0.15492921E+00, -0.12488942E+00, -0.24417043E+00, 
    },
    {
      0.22861373E-01, -0.35652456E+00, -0.40377608E+00, 0.73468529E-01, 0.20436759E+00, -0.63278258E-01, -0.14106607E+00, -0.82917646E-01, 0.15102735E+00, 
    },
    {
      0.48683444E+00, -0.13043688E+00, 0.14397170E+00, -0.34078661E+00, 0.98404596E-02, 0.12281654E+00, -0.21639533E+00, 0.21982001E-01, 0.36892843E-01, 
    },
    {
      0.22279333E+00, 0.97342873E+00, 0.53715861E+00, -0.29357713E+00, 0.11556523E+01, 0.34201711E-01, -0.19304839E+00, 0.81136659E-01, -0.24921562E+00, 
    },
    {
      -0.13376596E+01, -0.14052005E+00, 0.35047200E-01, -0.92397130E+00, -0.17662610E+00, -0.17981412E-01, 0.24125864E+00, 0.16264918E+00, 0.29551035E+00, 
    },
    {
      -0.21483745E-01, -0.36245286E-02, -0.66329092E-02, 0.25388638E-02, 0.10015321E-01, 0.40678948E-01, -0.17979136E-01, -0.35519518E-02, -0.52801091E-02, 
    },
    {
      0.23592529E-02, -0.10888859E-01, -0.72884997E-02, -0.15401221E-01, -0.82034264E-02, 0.22992514E-01, 0.55773426E-01, 0.39712242E-02, 0.38829453E-02, 
    },
    {
      0.69172211E-01, 0.48699997E-01, 0.27636139E-01, -0.61428368E-01, -0.52963924E-02, -0.12855162E+00, -0.45400564E-01, 0.16215649E-02, -0.12359414E-01, 
    },
    {
      0.48297949E-01, -0.49748858E-02, -0.98275091E-03, -0.19215355E-01, -0.17720040E-01, 0.45080960E-01, -0.12306380E+00, 0.50776438E-02, 0.61110142E-02, 
    },
    {
      0.41643456E-01, -0.11696455E+00, 0.58375955E-01, 0.44971615E-01, -0.28971106E-01, -0.11677378E+00, -0.12909587E+00, 0.26876843E-01, -0.87210005E-02, 
    },
    {
      -0.27737666E-01, -0.31307798E-01, -0.67466116E-02, 0.54379940E-01, -0.14625589E-02, 0.13169987E+00, -0.15068077E+00, 0.17126866E-01, -0.85981060E-02, 
    },
    {
      -0.20431230E-01, -0.40790266E-02, 0.63434639E-02, 0.17711798E-01, 0.10949640E-01, -0.42260378E-02, 0.21843901E-02, 0.74000163E-02, 0.10929640E-01, 
    },
    {
      -0.27122835E-01, 0.13527165E-01, -0.41228514E-02, -0.12625435E-01, -0.37137333E-02, -0.90636835E-02, 0.47529996E-02, -0.73364242E-02, 0.13638129E-01, 
    },
    {
      0.87972768E-01, 0.10461537E-01, -0.20852067E-01, -0.11302185E-01, -0.18262224E-01, -0.84132282E-02, -0.51975157E-02, 0.13170406E-02, -0.52717347E-01, 
    },
    {
      0.91586038E-02, -0.67112334E-01, 0.20436235E-01, 0.37548594E-01, -0.96599301E-02, 0.67724632E-02, -0.25683215E-02, 0.61321169E-01, 0.49731391E-02, 
    },
    {
      0.30659314E-03, -0.44321292E-02, 0.46804580E-02, -0.64085319E-03, 0.67579322E-02, 0.63220300E-02, -0.32938600E-02, -0.14760738E-01, 0.72151292E-02, 
    },
    {
      0.10289571E-01, 0.13664791E-01, -0.10606593E-01, -0.77491556E-02, -0.10887925E-01, 0.14865423E-01, 0.28957613E-02, -0.37721093E-02, -0.77854278E-02, 
    },
    {
      -0.13292247E-01, 0.18373661E-02, 0.13984727E-01, 0.48237410E-02, -0.23889195E-03, 0.75594457E-05, -0.27488936E-02, -0.73596179E-02, 0.46203346E-02, 
    },
    {
      0.75970124E-02, -0.96835792E-02, -0.18875452E-01, -0.29838530E-02, -0.54170112E-02, 0.12925678E-01, 0.10607135E-02, -0.30239653E-02, -0.59676557E-02, 
    },
  },
  {
    {
      0.26920488E+01, 0.11389978E-01, 0.98317824E-02, 0.31620429E-02, 0.71236417E-02, 0.11406111E-02, -0.56722080E-02, -0.48866012E-04, 0.11990918E-03, 
    },
    {
      0.32997811E+00, 0.15148186E-01, 0.51400617E-01, 0.34719396E-01, 0.39471131E-01, 0.79672048E-02, -0.34432359E-01, -0.70640777E-03, 0.12702998E-01, 
    },
    {
      0.10057878E+01, -0.11234321E+00, -0.14979783E+00, -0.79824487E-02, -0.41622274E-01, 0.64391503E-02, 0.55315364E-01, -0.57274159E-01, -0.10790603E-01, 
    },
    {
      -0.18790220E+00, -0.25727227E+00, -0.12212502E+00, -0.53840321E-01, -0.13131250E+00, -0.32435574E-01, 0.12784509E+00, 0.15400990E-01, -0.22077980E-01, 
    },
    {
      -0.11739050E+01, 0.24142879E+00, 0.30385274E+00, 0.25107432E-02, 0.12746825E-01, -0.26694337E-01, -0.99654526E-01, 0.15097810E+00, 0.38390014E-01, 
    },
    {
      -0.10511706E+00, 0.20301551E+00, 0.83070330E-01, 0.77093919E-02, 0.10205883E+00, 0.23708317E-01, -0.99959821E-01, -0.15592851E-01, 0.25066223E-02, 
    },
    {
      0.24683730E+00, -0.11758252E+00, -0.16027445E+00, 0.14066710E-02, 0.35752703E-01, 0.23185330E-01, 0.48782002E-01, -0.95009811E-01, -0.37437104E-01, 
    },
    {
      0.88183937E-03, -0.30879867E+00, -0.27182159E+00, 0.20393897E-02, 0.69514830E-02, -0.93122534E-02, 0.51981998E-02, 0.39402079E-02, 0.44796327E-02, 
    },
    {
      0.87271491E-02, 0.28625375E+00, -0.31935695E+00, 0.52287388E-02, -0.67754048E-02, 0.86630229E-02, -0.77996161E-02, -0.60802847E-02, 0.78002373E-02, 
    },
    {
      -0.49862538E-01, 0.26634568E+00, 0.43451972E-01, -0.30126229E-01, 0.80953494E-01, -0.31576782E-01, 0.30366642E-01, 0.42546973E-01, 0.15590327E-01, 
    },
    {
      -0.92389202E-02, 0.49252599E-01, 0.31745583E+00, -0.11614788E-01, -0.22583406E-01, 0.92642047E-01, -0.35509120E-02, 0.13614683E-01, 0.56305062E-01, 
    },
    {
      0.99343164E-02, 0.11374707E+01, 0.70455754E+00, 0.17588671E-01, -0.75575434E-01, 0.13827311E+00, -0.49321633E-01, -0.75004660E-01, -0.13349947E+00, 
    },
    {
      -0.42829427E+00, -0.84019780E+00, 0.13067011E+01, 0.34150391E-03, 0.15719564E+00, -0.40955950E-01, 0.75470291E-01, 0.13132016E+00, -0.17925552E+00, 
    },
    {
      0.84138435E+00, -0.15912498E+01, 0.10511795E+00, 0.54206944E+00, -0.38688609E+00, 0.17358598E+00, -0.36708114E+00, -0.45672667E+00, -0.17131349E+00, 
    },
    {
      0.51538002E+00, -0.22939309E+00, -0.19646254E+01, -0.31394237E+00, 0.19157244E+00, -0.59996855E+00, -0.45292720E-01, -0.13538666E+00, -0.45756468E+00, 
    },
    {
      -0.15232117E+00, -0.15745953E+01, 0.19509153E+00, 0.83998501E-01, 0.52878518E-01, -0.39462078E+00, 0.11502753E+00, 0.21379741E+00, 0.39088184E+00, 
    },
    {
      0.13416333E+01, -0.28084823E+00, -0.21697867E+01, 0.53006239E-01, -0.35013482E+00, 0.56974604E-02, -0.25478348E+00, -0.39096361E+00, 0.53477621E+00, 
    },
    {
      -0.23757496E+01, 0.26613855E+01, 0.50327115E-01, -0.13042608E+01, 0.67580837E+00, -0.25880238E+00, 0.89902413E+00, 0.10861776E+01, 0.29366979E+00, 
    },
    {
      -0.18501281E+01, 0.81386849E-01, 0.37214670E+01, 0.11096135E+01, -0.27120560E+00, 0.12532763E+01, 0.18778534E+00, 0.33471352E+00, 0.96769452E+00, 
    },
    {
      0.14908600E+00, 0.10968636E+01, -0.32124093E+00, -0.12884712E+00, 0.20785201E-01, 0.26191479E+00, -0.83107598E-01, -0.16051687E+00, -0.30176720E+00, 
    },
    {
      -0.10500849E+01, 0.92894047E+00, 0.14747008E+01, -0.54342680E-01, 0.21156456E+00, 0.44004083E-01, 0.19760658E+00, 0.30182552E+00, -0.39921561E+00, 
    },
    {
      0.17400059E+01, -0.12493542E+01, 0.42433235E+00, 0.91900015E+00, -0.38044804E+00, 0.54337408E-01, -0.64055729E+00, -0.72182900E+00, -0.13823685E+00, 
    },
    {
      0.16223973E+01, -0.39147210E+00, -0.21358540E+01, -0.86738002E+00, 0.11883757E+00, -0.81355983E+00, -0.14370240E+00, -0.23106740E+00, -0.60141885E+00, 
    },
    {
      -0.41002692E-02, -0.36998648E-01, -0.17324733E-01, 0.11522978E+00, -0.77249430E-01, -0.40560430E-02, 0.70232498E-02, -0.66030943E-02, -0.17369712E-02, 
    },
    {
      0.50536189E-02, -0.27059956E-01, -0.90134060E-02, 0.70231602E-01, 0.11070350E+00, -0.50629838E-02, 0.32779330E-02, -0.49949071E-03, 0.14835848E-02, 
    },
    {
      0.78637168E-01, 0.81079379E-01, 0.11170681E+00, -0.13736880E+00, 0.11822002E+00, -0.20365615E-02, 0.56556158E-03, 0.83278008E-02, -0.18613596E-01, 
    },
    {
      -0.28584436E-01, -0.58529172E-02, -0.46521258E-01, -0.11188855E+00, -0.86472198E-01, 0.11980531E-01, 0.47342207E-01, 0.28845016E-02, 0.28383438E-01, 
    },
    {
      0.11538816E-01, 0.23314317E+00, 0.19854650E+00, -0.54327124E+00, 0.38741905E-01, 0.24814070E-02, -0.65545565E-02, 0.54088697E-01, -0.54012053E-01, 
    },
    {
      -0.19324923E+00, 0.67288898E-01, -0.59418734E-01, -0.33771223E-02, -0.44705462E+00, -0.62183533E-01, 0.37801322E-01, 0.20709286E-01, -0.37422135E-01, 
    },
    {
      -0.37016487E+00, -0.48482874E+00, -0.33297902E+00, 0.26727724E+00, -0.59645987E+00, 0.24300680E-01, -0.60210197E-03, -0.69528729E-01, 0.92760593E-01, 
    },
    {
      0.19207671E+00, 0.27005580E+00, 0.30531994E+00, 0.62059212E+00, 0.20099737E+00, 0.89472935E-01, -0.26362386E+00, -0.48994768E-01, -0.22086474E+00, 
    },
    {
      0.96757943E-02, -0.35891131E+00, -0.41006809E+00, 0.39764428E+00, -0.12385339E+00, -0.45176540E-01, -0.23520408E-01, -0.71112022E-01, 0.90213269E-01, 
    },
    {
      0.28137875E+00, -0.18450230E+00, 0.15791111E+00, 0.12017469E+00, 0.28672829E+00, 0.10803044E+00, -0.11098100E+00, -0.24466727E-01, 0.60332600E-01, 
    },
    {
      0.31297663E+00, 0.70864713E+00, 0.34029236E+00, -0.29005626E+00, 0.79561400E+00, 0.39008945E-01, 0.32641575E-01, 0.97112492E-01, -0.11558538E+00, 
    },
    {
      -0.27881745E+00, -0.29964480E+00, -0.33964044E+00, -0.84683907E+00, -0.30862889E+00, -0.19399051E+00, 0.30272609E+00, 0.92557348E-01, 0.24453788E+00, 
    },
    {
      -0.15125610E-01, -0.41656359E-02, -0.18647785E-03, 0.33162646E-02, 0.50839242E-02, 0.39834607E-01, -0.29238209E-01, -0.27017531E-03, -0.12010454E-02, 
    },
    {
      -0.41408944E-02, -0.13943053E-02, 0.43188082E-02, -0.89726988E-02, -0.38272680E-02, 0.31500991E-01, 0.41226041E-01, 0.43756245E-02, -0.18315912E-03, 
    },
    {
      0.28158762E-01, 0.54722499E-01, 0.17690169E-01, -0.28583935E-01, -0.11942321E-01, -0.10031212E+00, -0.44474870E-01, 0.96164370E-03, -0.99050319E-02, 
    },
    {
      0.20313956E-01, 0.83643049E-02, 0.24804203E-02, -0.18918399E-02, -0.15124847E-01, 0.43693494E-01, -0.10090125E+00, 0.51212185E-02, 0.10049676E-02, 
    },
    {
      0.44255398E-01, -0.71331427E-01, 0.36873396E-01, 0.20218041E-01, -0.17782997E-01, -0.11550901E+00, -0.89868911E-01, 0.13469202E-01, -0.11585846E-01, 
    },
    {
      -0.35603975E-02, -0.43277528E-01, -0.30855784E-01, 0.36533542E-01, 0.49907360E-02, 0.10209156E+00, -0.10187620E+00, 0.46443972E-02, 0.65317191E-02, 
    },
    {
      -0.13696849E-01, -0.11126058E-01, 0.41133645E-02, 0.43337820E-02, 0.92311008E-02, -0.39588860E-02, 0.16066917E-02, 0.65573328E-02, 0.13953523E-01, 
    },
    {
      -0.12618544E-01, 0.56935060E-02, -0.90064071E-02, -0.48857494E-02, -0.64519118E-02, -0.68315426E-02, 0.63782968E-02, -0.12648813E-01, 0.11987502E-01, 
    },
    {
      0.40514901E-01, 0.76482454E-02, -0.16679959E-01, -0.22868705E-02, -0.15372105E-01, -0.84379911E-02, 0.35069739E-02, 0.13873799E-01, -0.56473475E-01, 
    },
    {
      0.20609386E-01, -0.42281467E-01, 0.22169946E-01, 0.14798193E-01, -0.78803375E-02, -0.38305379E-02, 0.14281864E-02, 0.57916559E-01, 0.14591149E-01, 
    },
    {
      0.25254390E-02, -0.38154577E-02, 0.52015893E-02, -0.14573345E-02, 0.38699114E-02, 0.55529494E-02, -0.52860053E-02, -0.12854257E-01, 0.48506586E-02, 
    },
    {
      0.82307914E-02, 0.14202736E-01, -0.94194580E-02, 0.12523378E-02, -0.89414315E-02, 0.87305941E-02, 0.66426750E-02, -0.21206427E-02, -0.49932897E-02, 
    },
    {
      -0.50855689E-02, 0.87064495E-02, 0.80340123E-02, 0.87542385E-02, -0.17597808E-02, 0.23883770E-03, 0.21871652E-03, -0.43078689E-02, 0.17831495E-02, 
    },
    {
      0.10356435E-01, -0.30754986E-02, -0.14237600E-01, 0.38996674E-02, -0.63463412E-02, 0.10078620E-01, 0.10248976E-02, 0.13287502E-04, -0.20263768E-02, 
    },
  },
};

const F2_coefficient_array_t g_NeQuickG_ccir22_F2[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.58184547E+01, 0.57026275E-01, -0.92279583E-01, 0.19769210E-02, -0.63549969E-02, 0.41342468E-03, -0.16677534E-01, 0.29840269E-02, 0.19395292E-01, -0.31526904E-02, 0.83086165E-02, -0.21525295E-01, -0.12466762E-02, 
    },
    {
      0.10098135E+01, -0.13401215E+00, -0.33394384E+00, 0.53832471E+00, -0.33256567E+00, -0.14199045E+00, 0.17917547E+00, 0.53553913E-01, 0.36476406E+00, -0.17950524E+00, 0.83029037E-02, 0.52691013E-01, -0.10528484E+00, 
    },
    {
      0.10512217E+02, -0.11169873E+01, 0.69445834E-01, -0.11768398E+01, 0.15896691E+01, -0.46408096E+00, 0.20050450E+00, -0.11498708E+00, -0.57279074E+00, -0.55499069E-01, -0.22052056E+00, 0.46984813E+00, -0.40550134E-02, 
    },
    {
      -0.88619490E+01, 0.20526533E+01, 0.73109612E+01, -0.79561610E+01, 0.10884787E+02, 0.16878909E+01, -0.36661744E+01, -0.14041631E+01, -0.40023403E+01, 0.28041875E+01, 0.11619287E+01, -0.47040588E+00, 0.75334054E+00, 
    },
    {
      -0.62287876E+02, 0.10522943E+02, 0.70986325E+00, 0.65067163E+01, -0.11479047E+02, 0.16810986E+01, 0.92432690E+00, -0.25512171E+00, 0.36646883E+01, -0.23676340E+00, 0.22571785E-01, -0.15189878E+01, 0.11747854E+01, 
    },
    {
      -0.51741142E+02, -0.26478231E+02, -0.41005913E+02, 0.41161312E+02, -0.59304089E+02, -0.69479570E+01, 0.18469704E+02, 0.71893792E+01, 0.17059683E+02, -0.13025754E+02, -0.89780149E+01, 0.14681500E+01, 0.57374662E+00, 
    },
    {
      0.13726009E+03, -0.30932428E+02, 0.32160757E+01, -0.12439873E+02, 0.30461220E+02, -0.47828910E+00, -0.68936372E+01, 0.25059500E+01, -0.91692266E+01, 0.18327304E+01, 0.26999156E+01, 0.10919596E+01, -0.47839489E+01, 
    },
    {
      0.24011169E+03, 0.88412567E+02, 0.86547073E+02, -0.93710808E+02, 0.12796277E+03, 0.12891210E+02, -0.38673927E+02, -0.13718105E+02, -0.34361622E+02, 0.25967365E+02, 0.23587971E+02, -0.25956814E+01, -0.78749714E+01, 
    },
    {
      -0.15224947E+03, 0.34043217E+02, -0.10899738E+02, 0.94556570E+01, -0.33112156E+02, -0.28386922E+01, 0.11093059E+02, -0.39166231E+01, 0.96845121E+01, -0.29332969E+01, -0.52627816E+01, 0.10346909E+01, 0.64608569E+01, 
    },
    {
      -0.31515054E+03, -0.10496167E+03, -0.78569489E+02, 0.95572739E+02, -0.12133438E+03, -0.10754658E+02, 0.36431686E+02, 0.10801248E+02, 0.32190319E+02, -0.23463760E+02, -0.25738005E+02, 0.27011609E+01, 0.12155417E+02, 
    },
    {
      0.65124084E+02, -0.12487495E+02, 0.69259715E+01, -0.23780632E+01, 0.12499528E+02, 0.21069338E+01, -0.53715777E+01, 0.17919514E+01, -0.36225002E+01, 0.14143248E+01, 0.27828605E+01, -0.10895599E+01, -0.28614464E+01, 
    },
    {
      0.13440363E+03, 0.41123314E+02, 0.26135244E+02, -0.35684410E+02, 0.42126480E+02, 0.32188382E+01, -0.12763022E+02, -0.28629279E+01, -0.11248032E+02, 0.79024072E+01, 0.99840336E+01, -0.11786457E+01, -0.55230813E+01, 
    },
    {
      0.46197325E-01, 0.18396910E+01, 0.18488530E+01, -0.72137311E-01, -0.22965059E-01, 0.49167216E-01, -0.88025229E-02, -0.14172700E-02, 0.23798379E-02, 0.10697534E-01, -0.86828216E-03, -0.11820203E-01, -0.45507951E-02, 
    },
    {
      -0.34407914E+00, -0.19840239E+01, 0.18451532E+01, -0.71499117E-01, -0.29341163E-01, -0.38065258E-02, 0.12800088E-02, 0.35247207E-01, -0.29336326E-01, -0.28935510E-02, 0.34494635E-01, 0.21810072E-01, -0.35707127E-01, 
    },
    {
      -0.14954211E+01, 0.16685070E+01, 0.26021440E+01, -0.25061376E-01, -0.11775342E+00, 0.18140525E+00, 0.39146900E+00, -0.52104045E-01, 0.17152113E+00, -0.24431722E+00, 0.83082154E-01, 0.10321481E+00, -0.50523303E-01, 
    },
    {
      -0.29371035E+01, -0.11318464E+01, 0.17996328E+01, -0.12266654E+01, 0.44546545E+00, -0.17469050E-01, 0.10839576E+00, 0.13065223E+00, -0.14671960E+00, -0.17891817E-01, 0.82936779E-01, -0.53874258E-01, 0.32858126E-01, 
    },
    {
      0.35966065E+01, 0.80477695E+01, 0.35369372E+01, -0.24907688E-01, 0.49452472E+01, -0.24985828E+01, 0.10750397E+01, 0.63011825E-01, -0.19682705E+00, 0.48551191E-01, -0.24114858E+00, 0.44627540E-01, 0.18766768E+00, 
    },
    {
      -0.88128072E+00, -0.42972250E+01, 0.32859540E+01, 0.79657030E+00, 0.30905221E+01, 0.28533414E-01, -0.50964904E+00, -0.72284150E+00, 0.72344583E+00, -0.41555977E+00, -0.27640811E+00, -0.11061147E+01, -0.24223343E+00, 
    },
    {
      0.26371164E+02, -0.15519335E+02, -0.16091352E+02, -0.61467066E+01, 0.11827186E+02, -0.18683293E+01, -0.10109256E+02, -0.41066766E+00, -0.26999035E+01, 0.11567605E+01, -0.66600835E+00, -0.19529310E+01, 0.17002133E+00, 
    },
    {
      0.52747944E+02, -0.59905572E+01, -0.19065039E+02, 0.19325472E+02, -0.37808802E+01, 0.25225880E+01, -0.50770384E+00, -0.24967108E+01, 0.18975933E+01, 0.15414751E+01, 0.46599278E+00, 0.26762016E+01, -0.23129802E+01, 
    },
    {
      -0.29581627E+02, -0.67340530E+02, 0.28578785E+02, -0.18400332E+01, -0.45590832E+02, 0.19642672E+02, -0.11388482E+02, -0.14384048E+01, 0.23746016E+01, -0.25869808E+01, 0.12232533E+01, -0.22738206E+00, -0.13948896E+01, 
    },
    {
      0.25422518E+02, -0.24761139E+02, -0.18647608E+02, -0.35908458E+01, -0.27576836E+02, 0.63180453E+00, 0.12264181E+01, 0.73593297E+01, -0.56782942E+01, 0.42023597E+01, 0.94908339E+00, 0.99610100E+01, 0.74102955E+01, 
    },
    {
      -0.13301096E+03, 0.29004534E+02, 0.96729050E+02, 0.52969921E+02, -0.57015472E+02, 0.43563184E+00, 0.61022591E+02, -0.60143486E-01, 0.14871103E+02, 0.36482079E+01, 0.28649278E+01, 0.86638842E+01, -0.38194213E+01, 
    },
    {
      -0.28812790E+03, 0.46469360E+02, 0.46227028E+02, -0.10875101E+03, -0.70202547E+00, -0.15389408E+02, 0.14878047E+01, 0.11239145E+02, -0.60284514E+01, -0.76406460E+01, -0.82917595E+01, -0.18456875E+02, 0.17502781E+02, 
    },
    {
      0.90780685E+02, 0.17434750E+03, -0.17242982E+03, 0.74431643E+01, 0.14255406E+03, -0.58141239E+02, 0.39145481E+02, 0.34620700E+01, -0.12430976E+02, 0.10840864E+02, -0.43649893E-01, 0.59502614E+00, 0.22078130E+01, 
    },
    {
      -0.10493087E+03, 0.18064041E+03, 0.12570059E+02, 0.93478651E+01, 0.80625313E+02, -0.71573329E+00, 0.36804316E+01, -0.22530621E+02, 0.18506762E+02, -0.12973787E+02, -0.15200816E+01, -0.31336437E+02, -0.28821777E+02, 
    },
    {
      0.29102225E+03, -0.17519779E+02, -0.29855054E+03, -0.15717352E+03, 0.88891266E+02, 0.25252760E+02, -0.15087782E+03, 0.76023388E+01, -0.37483158E+02, -0.23462990E+02, -0.69535470E+01, -0.14723260E+02, 0.18481215E+02, 
    },
    {
      0.65221857E+03, -0.11774995E+03, -0.61445251E+02, 0.26809067E+03, 0.39994270E+02, 0.31199490E+02, -0.64933023E+01, -0.18023024E+02, 0.64583664E+01, 0.12623926E+02, 0.27021164E+02, 0.45272427E+02, -0.46571922E+02, 
    },
    {
      -0.11661733E+03, -0.21271249E+03, 0.27106604E+03, -0.73082743E+01, -0.18257878E+03, 0.73000687E+02, -0.54757530E+02, -0.16007293E+01, 0.22605629E+02, -0.14365945E+02, -0.50904408E+01, -0.98563850E+00, 0.61151135E+00, 
    },
    {
      0.15628786E+03, -0.29974243E+03, 0.10640742E+02, -0.10775609E+02, -0.95066612E+02, -0.36515400E+01, -0.11473593E+02, 0.27438049E+02, -0.24867657E+02, 0.15603787E+02, 0.12937165E+01, 0.39795895E+02, 0.39510033E+02, 
    },
    {
      -0.29016129E+03, 0.28823552E+01, 0.39856253E+03, 0.19402602E+03, -0.45091358E+02, -0.53616592E+02, 0.16660764E+03, -0.15165434E+02, 0.42398968E+02, 0.36244057E+02, 0.75232472E+01, 0.95507259E+01, -0.28676897E+02, 
    },
    {
      -0.64159631E+03, 0.12781446E+03, 0.57046253E+02, -0.29240640E+03, -0.71489220E+02, -0.24908289E+02, 0.13778471E+02, 0.10944452E+02, -0.66494685E+00, -0.74688768E+01, -0.32826694E+02, -0.46453716E+02, 0.50829025E+02, 
    },
    {
      0.52968678E+02, 0.99466309E+02, -0.13427264E+03, 0.10737036E+01, 0.81903267E+02, -0.32730858E+02, 0.26825422E+02, -0.69327599E+00, -0.12845870E+02, 0.59912968E+01, 0.45008116E+01, 0.76359075E+00, -0.18522649E+01, 
    },
    {
      -0.80060051E+02, 0.15137988E+03, -0.81274099E+01, 0.41646919E+01, 0.38812267E+02, 0.42529483E+01, 0.73921485E+01, -0.11849993E+02, 0.11516630E+02, -0.64914279E+01, -0.51537508E+00, -0.17529139E+02, -0.18062862E+02, 
    },
    {
      0.10677107E+03, 0.52536047E+00, -0.18758057E+03, -0.85391655E+02, 0.35786375E+00, 0.30250124E+02, -0.67851524E+02, 0.84298763E+01, -0.17312017E+02, -0.17777077E+02, -0.27602580E+01, -0.14503183E+01, 0.14062240E+02, 
    },
    {
      0.22694662E+03, -0.50785782E+02, -0.25865496E+02, 0.11572225E+03, 0.35236393E+02, 0.62496061E+01, -0.90702019E+01, -0.16877422E+01, -0.16750351E+01, 0.10357891E+01, 0.13722246E+02, 0.17021441E+02, -0.19445265E+02, 
    },
    {
      -0.39433226E-01, -0.21142434E+00, -0.90724351E-02, -0.83350801E+00, -0.50699193E-01, 0.27291071E-01, -0.22223895E-01, -0.42128101E-01, 0.17714033E-01, 0.50699499E-01, -0.18712536E-01, -0.25406671E-02, 0.20049289E-01, 
    },
    {
      -0.49931601E-01, -0.20490995E-01, 0.89790568E-01, 0.51811721E-01, -0.87662041E+00, -0.31257950E-01, 0.40552104E-02, -0.14721376E-01, -0.69894153E-02, -0.72437213E-02, 0.23188520E-01, 0.26874647E-01, -0.39877024E-01, 
    },
    {
      -0.24794403E+00, -0.22231767E+00, 0.30068848E+00, 0.26426491E+00, 0.85828662E+00, 0.70318066E-01, -0.19365221E+00, -0.67670614E-01, -0.44199413E+00, -0.11491100E+00, -0.18138404E+00, 0.90832096E-02, 0.69318712E-02, 
    },
    {
      0.56248826E+00, 0.47823495E+00, 0.41573855E+00, -0.15119504E+01, 0.67963108E-01, -0.13004091E+00, 0.64840458E-01, -0.91815062E-01, -0.15601902E+00, -0.78684457E-01, -0.90477422E-01, 0.76826736E-01, -0.65387972E-01, 
    },
    {
      0.36128531E+01, 0.25138476E+01, 0.35854378E+00, 0.26517680E+01, 0.23037329E+01, -0.11147922E+01, -0.63770419E+00, 0.86852974E+00, -0.41232470E+00, -0.61307818E+00, 0.48212731E+00, -0.85765906E-01, -0.57749575E+00, 
    },
    {
      0.26269414E+01, -0.84287798E+00, 0.11433274E+01, -0.36928189E+01, 0.54051328E+01, 0.74675167E+00, -0.73716527E+00, -0.66338859E-01, 0.11560968E+01, -0.29391339E+00, 0.12733983E+00, -0.10972692E+01, 0.59222376E+00, 
    },
    {
      0.34999065E+01, 0.44444332E+01, -0.94481927E+00, -0.35326798E+01, -0.27436817E+01, 0.56982708E+00, 0.76956064E-01, 0.31902781E+00, 0.52106791E+01, 0.92866051E+00, 0.87636429E+00, -0.33926573E+00, -0.11445957E+01, 
    },
    {
      -0.32386546E+01, -0.44973111E+01, -0.15676624E+01, 0.10158686E+02, 0.16184865E+00, 0.16182853E+01, -0.21049386E+00, -0.28675815E+00, 0.18338075E+01, 0.79065192E+00, 0.12128632E+01, -0.13835089E+01, -0.13644750E-01, 
    },
    {
      -0.20781828E+02, 0.17002146E+01, -0.11406076E+02, 0.90912104E+01, 0.13688050E+02, 0.40151911E+01, 0.44199939E+01, -0.25502865E+01, 0.19699675E+01, 0.12704998E+01, -0.20803688E+01, 0.20040412E+01, 0.38943079E+01, 
    },
    {
      -0.19840216E+02, 0.19784559E+02, -0.86485300E+01, -0.49333038E+01, -0.39710073E+01, -0.40674095E+01, 0.28137231E+01, 0.38876519E+01, -0.90402184E+01, 0.19475243E+01, -0.10716798E+01, 0.62051377E+01, -0.18414352E+01, 
    },
    {
      -0.83312197E+01, -0.13289581E+02, -0.21736002E+01, 0.84501009E+01, 0.49111623E+00, -0.37395921E+01, 0.19115800E+01, -0.31825442E-01, -0.13818624E+02, -0.18343800E+01, -0.11181920E+01, 0.80240744E+00, 0.43642979E+01, 
    },
    {
      0.33570621E+01, 0.12437586E+02, 0.45811825E+01, -0.16603788E+02, 0.22749799E+00, -0.40944538E+01, -0.53237623E+00, 0.23649323E+01, -0.41487660E+01, -0.17448486E+01, -0.36165757E+01, 0.43271375E+01, 0.10559268E+01, 
    },
    {
      0.39056721E+02, -0.16090921E+02, 0.33891582E+02, -0.35210052E+02, -0.50438084E+02, -0.49294519E+01, -0.90522985E+01, 0.19701451E+01, -0.41518707E+01, -0.19429389E+00, 0.29413283E+01, -0.64302478E+01, -0.75884647E+01, 
    },
    {
      0.49585617E+02, -0.54486118E+02, 0.14613009E+02, 0.31883018E+02, -0.19175350E+02, 0.78750348E+01, -0.41738634E+01, -0.11004766E+02, 0.19877577E+02, -0.26470737E+01, 0.21634009E+01, -0.10771997E+02, 0.26160250E+01, 
    },
    {
      0.42416420E+01, 0.10615758E+02, 0.38284779E+01, -0.26582875E+01, 0.61848521E+01, 0.39846966E+01, -0.23324945E+01, -0.27493328E+00, 0.98930330E+01, 0.91430855E+00, 0.42881787E+00, -0.45336166E+00, -0.38754070E+01, 
    },
    {
      0.24860692E+01, -0.11103153E+02, -0.55986381E+01, 0.30626101E+01, 0.12912285E+01, 0.31475737E+01, 0.94185495E+00, -0.24471545E+01, 0.26365435E+01, 0.11146898E+01, 0.30409727E+01, -0.34792864E+01, -0.11173525E+01, 
    },
    {
      -0.24812998E+02, 0.12293331E+02, -0.24448639E+02, 0.25972631E+02, 0.41211086E+02, 0.13715389E+01, 0.59663386E+01, -0.71985565E-01, 0.29825644E+01, -0.41931191E+00, -0.13509992E+01, 0.51593585E+01, 0.43710918E+01, 
    },
    {
      -0.40055676E+02, 0.40998569E+02, -0.62064986E+01, -0.28469561E+02, 0.22959063E+02, -0.49932957E+01, 0.19243966E+01, 0.76674228E+01, -0.12809521E+02, 0.61497426E+00, -0.14364125E+01, 0.57304015E+01, -0.13400353E+01, 
    },
    {
      -0.21295029E+00, -0.11400111E+00, -0.14825793E+00, 0.73622137E-01, 0.64809769E-01, -0.29826665E+00, -0.60644519E+00, -0.13423300E-01, 0.10754849E-01, 0.98052928E-02, -0.27152034E-02, -0.18311009E-01, 0.64233541E-02, 
    },
    {
      -0.17358117E-01, 0.90460517E-01, -0.55983938E-01, -0.37680583E-02, 0.10464079E+00, 0.60645211E+00, -0.32411674E+00, -0.19527392E-01, 0.57433564E-01, -0.14313875E-01, 0.14264599E-01, -0.12788296E-01, 0.12462893E-01, 
    },
    {
      0.72629370E-01, 0.26425889E+00, -0.72310239E-01, 0.14027588E+00, -0.22800760E+00, -0.82698578E+00, -0.59944493E+00, 0.41647565E-01, 0.10132378E+00, 0.10920783E+00, -0.92539199E-01, 0.54342553E-01, 0.23273796E-01, 
    },
    {
      -0.79490080E-01, -0.41849073E-01, 0.11703410E+00, 0.22500437E+00, 0.79866014E-01, 0.79547960E+00, -0.68940175E+00, -0.91365695E-01, 0.26474414E-01, 0.18633839E-01, -0.69830693E-01, 0.42015348E-01, 0.70129656E-02, 
    },
    {
      0.20651813E+01, 0.10757015E+01, 0.79591483E+00, -0.10709347E+01, -0.39959517E+00, 0.12449454E+01, 0.65018201E+00, 0.57454801E+00, -0.10787249E+00, -0.11827233E+00, 0.89393541E-01, 0.38926962E+00, 0.20278770E-01, 
    },
    {
      -0.84885871E+00, 0.50201140E-01, 0.41724554E+00, 0.10930064E+00, -0.19860851E+01, -0.40959147E+00, 0.14579872E+01, 0.10777475E+00, 0.18702915E+00, 0.58261946E-01, -0.15377997E+00, 0.27000561E+00, -0.32803483E-01, 
    },
    {
      0.41750300E+00, -0.89985055E+00, 0.11675123E+00, 0.33188909E-01, 0.56400406E+00, 0.12467124E+01, 0.11199267E+01, -0.13293042E-01, -0.56087691E-01, -0.28695655E+00, 0.14087248E+00, -0.16224995E+00, -0.25811439E-01, 
    },
    {
      0.51898807E+00, 0.17738800E+00, -0.20662893E+00, -0.74402356E+00, -0.51481444E-02, -0.16159163E+01, 0.10972348E+01, 0.44918306E-01, -0.20699680E+00, -0.11631241E-01, 0.20945826E+00, -0.16005199E+00, -0.11305244E+00, 
    },
    {
      -0.41412516E+01, -0.14561827E+01, -0.11091290E+01, 0.17685617E+01, 0.66921726E-01, -0.16768463E+01, 0.23649099E+00, -0.10593188E+01, 0.12860221E+00, 0.10053814E+00, -0.29987147E+00, -0.75887227E+00, 0.55327021E-01, 
    },
    {
      0.18062499E+01, 0.15113790E+00, -0.36903124E-01, 0.28656241E+00, 0.33028457E+01, -0.43000481E+00, -0.17351261E+01, -0.96321538E-01, -0.62929451E+00, -0.54888729E-01, 0.19858769E+00, -0.39517498E+00, -0.23517720E-01, 
    },
    {
      0.74075326E-01, 0.55286419E-01, 0.37950128E-01, 0.51930584E-01, 0.40860217E-01, -0.80779605E-02, -0.85725700E-02, 0.18911599E+00, -0.20170498E+00, 0.79165380E-02, 0.14078500E-01, -0.48967579E-03, 0.23352634E-01, 
    },
    {
      0.89454815E-01, 0.45219720E-02, 0.19925103E-01, -0.36571357E-01, 0.47071739E-02, -0.22226086E-01, 0.12251128E-01, 0.23481008E+00, 0.19869673E+00, -0.52124937E-02, -0.27196895E-01, 0.11811548E-01, 0.11335118E-01, 
    },
    {
      0.20631571E+00, -0.40896397E-01, -0.66134050E-01, -0.51769245E-01, 0.11315651E+00, -0.29828817E-01, -0.70822775E-01, -0.71775690E-01, -0.42620280E+00, -0.58472637E-01, -0.11232210E-01, -0.11249758E-01, 0.20316778E-01, 
    },
    {
      0.17420180E-01, -0.14574118E+00, -0.10202885E+00, 0.72233818E-01, 0.11322709E-01, -0.20416332E-01, -0.24119224E-02, 0.33355662E+00, -0.65747298E-01, -0.34695771E-02, -0.40769741E-01, 0.66911586E-01, -0.37780598E-01, 
    },
    {
      0.43868124E-01, 0.14381094E+00, -0.61921760E-01, -0.12998655E-01, 0.31560197E-01, 0.31114614E-02, 0.38460427E-03, -0.28693313E-01, 0.34101967E-01, 0.90243518E-01, 0.18531388E+00, -0.13973088E-01, 0.10807764E-01, 
    },
    {
      0.87856077E-01, 0.35964809E-01, 0.23326820E-01, -0.84922142E-01, -0.19689238E-01, 0.94046257E-02, 0.14910235E-02, 0.26861498E-01, -0.73656917E-01, -0.17092833E+00, 0.78958347E-01, 0.87024560E-02, 0.12014927E-01, 
    },
    {
      0.82137808E-02, -0.21466317E-01, 0.53214639E-01, -0.74789733E-01, 0.19152014E-01, 0.21119369E-01, 0.25541728E-01, 0.88134193E-03, -0.14540546E-01, 0.78659160E-02, 0.15383034E-01, -0.64573243E-01, 0.21328622E-01, 
    },
    {
      0.28671172E-01, 0.20090424E-01, -0.94851188E-01, -0.24305644E-01, -0.72831862E-01, -0.18191807E-01, -0.27933948E-01, 0.21245504E-01, -0.38015399E-01, 0.13354953E-01, 0.61733276E-02, 0.53486875E-02, -0.67182191E-01, 
    },
    {
      0.49329074E-02, 0.14485019E-01, -0.57947930E-01, -0.16713040E-01, 0.52156751E-02, 0.26821880E-01, -0.57450021E-02, -0.63059921E-02, -0.14616561E-01, -0.94541983E-03, -0.15653398E-01, -0.63756146E-02, 0.19463095E-02, 
    },
    {
      0.91823854E-01, 0.34876380E-01, -0.81078401E-02, 0.40082075E-01, 0.11814687E-01, -0.21329770E-01, 0.22458946E-01, 0.65466752E-02, 0.32364109E-02, -0.13744369E-01, 0.66302335E-02, 0.16155699E-01, -0.14738605E-01, 
    },
    {
      0.45109160E-01, 0.52861501E-01, -0.88001676E-01, 0.24178792E-01, -0.26224427E-01, -0.27569726E-01, 0.28009566E-02, -0.12213375E-01, -0.14918844E-02, 0.38230043E-01, -0.10143310E-01, 0.97960318E-02, 0.72819605E-02, 
    },
    {
      -0.12117780E+00, 0.23717241E-01, -0.16449519E-01, 0.21570399E-01, -0.11265770E-02, 0.23510948E-01, 0.17620750E-01, -0.14561802E-01, 0.44145947E-02, -0.19181985E-01, 0.15150579E-01, 0.23060735E-01, -0.34930815E-02, 
    },
  },
  {
    {
      0.85668230E+01, 0.58971159E-01, -0.37359368E-01, 0.98172938E-02, 0.60634501E-02, -0.19980328E-01, -0.32437317E-01, 0.73017133E-02, 0.21622829E-01, -0.31204394E-02, 0.86839320E-02, -0.15189710E-01, 0.15739934E-02, 
    },
    {
      0.15422637E+01, -0.45887977E+00, -0.63496351E+00, 0.17954226E+00, -0.36354408E+00, -0.81500508E-01, 0.53280359E-02, 0.13243176E+00, 0.38387778E+00, -0.11102082E+00, -0.78610063E-01, 0.94472550E-01, -0.58259502E-01, 
    },
    {
      0.24310478E+02, -0.11059129E+01, -0.73916423E+00, -0.92551410E+00, 0.17832867E+01, 0.35105866E+00, 0.37704727E+00, -0.31123796E+00, -0.53970373E+00, 0.86808112E-02, -0.72665714E-01, 0.46467739E+00, -0.98924175E-01, 
    },
    {
      0.20755560E+02, 0.67352271E+01, 0.14065329E+02, -0.25990703E+01, 0.10409377E+02, 0.16016326E+01, -0.47564018E+00, -0.34755881E+01, -0.49684534E+01, 0.15748028E+01, 0.26321311E+01, -0.12655041E+01, 0.15727635E+00, 
    },
    {
      -0.12931139E+03, 0.12252737E+02, 0.60598216E+01, 0.50476084E+01, -0.11773267E+02, -0.44770432E+01, 0.58385706E+00, 0.23724539E+01, 0.37232220E+01, -0.10170099E+01, -0.81260860E-01, -0.29869394E+01, 0.22409637E+01, 
    },
    {
      -0.23669112E+03, -0.51905815E+02, -0.72058403E+02, 0.12511720E+02, -0.56770985E+02, -0.10931894E+02, 0.33293872E+01, 0.20718521E+02, 0.24318001E+02, -0.57881908E+01, -0.15459870E+02, 0.46247511E+01, 0.34875414E+01, 
    },
    {
      0.24901990E+03, -0.39882587E+02, -0.83957424E+01, -0.10404231E+02, 0.27809801E+02, 0.16726303E+02, -0.68507929E+01, -0.69119296E+01, -0.97388639E+01, 0.47765851E+01, 0.10716473E+01, 0.72030778E+01, -0.82405844E+01, 
    },
    {
      0.65161255E+03, 0.14923309E+03, 0.14101414E+03, -0.29154617E+02, 0.12519636E+03, 0.29351564E+02, -0.84194279E+01, -0.48260624E+02, -0.53140247E+02, 0.90673313E+01, 0.35299232E+02, -0.72856526E+01, -0.14410473E+02, 
    },
    {
      -0.23240266E+03, 0.45167839E+02, -0.96814597E+00, 0.99426565E+01, -0.26336687E+02, -0.23381857E+02, 0.11163598E+02, 0.86249828E+01, 0.10363331E+02, -0.69799447E+01, -0.19014372E+01, -0.71312876E+01, 0.10706991E+02, 
    },
    {
      -0.70311969E+03, -0.17120168E+03, -0.11967136E+03, 0.32664330E+02, -0.12050853E+03, -0.33125164E+02, 0.86441698E+01, 0.48691650E+02, 0.51921913E+02, -0.65182605E+01, -0.34950245E+02, 0.55114870E+01, 0.18803883E+02, 
    },
    {
      0.85038689E+02, -0.16290762E+02, 0.40023508E+01, -0.37510366E+01, 0.83846464E+01, 0.10919796E+02, -0.52057114E+01, -0.38280964E+01, -0.38365119E+01, 0.32329717E+01, 0.99229783E+00, 0.24394760E+01, -0.46375208E+01, 
    },
    {
      0.26612119E+03, 0.67721451E+02, 0.37346931E+02, -0.13712147E+02, 0.41968533E+02, 0.13200320E+02, -0.30124371E+01, -0.17847174E+02, -0.18521812E+02, 0.17858783E+01, 0.12569484E+02, -0.17031465E+01, -0.80002728E+01, 
    },
    {
      -0.13380323E+00, 0.13366905E+01, 0.19891349E+01, -0.10753992E-01, 0.60303468E-01, 0.12918174E-01, 0.21262087E-01, -0.86736670E-02, -0.13964107E-01, 0.11481032E-01, 0.10142981E-01, 0.72076609E-02, -0.85020000E-02, 
    },
    {
      -0.14243269E-01, -0.21352170E+01, 0.13255059E+01, -0.50178140E-01, -0.54384332E-01, 0.46576783E-01, 0.34636784E-01, 0.48959581E-02, -0.16821539E-01, 0.46431450E-02, 0.12051069E-01, 0.19954719E-01, -0.32330859E-01, 
    },
    {
      -0.12649831E+01, 0.14815663E+01, 0.32681105E+01, -0.28774256E+00, 0.42832842E+00, 0.53366369E+00, 0.69917524E+00, -0.12640499E+00, 0.27914149E+00, -0.86231783E-01, 0.41828956E-01, 0.22610457E+00, -0.11857753E+00, 
    },
    {
      -0.59289962E+00, -0.26590743E+01, 0.71710646E+00, -0.80467057E+00, 0.91069084E+00, 0.17542484E+00, 0.10409195E+00, -0.57189096E-01, 0.43423317E-01, -0.84960863E-01, -0.95633030E-01, 0.33903435E-01, -0.36507253E-01, 
    },
    {
      0.83079939E+01, 0.22010695E+02, -0.11524427E+02, -0.11183747E+00, 0.25016727E+01, -0.13266135E+01, 0.22817478E+00, 0.46875328E+00, 0.34386232E+00, 0.40068373E+00, -0.31280607E+00, -0.36988357E+00, 0.16799913E+00, 
    },
    {
      -0.83424835E+01, 0.11889293E+02, 0.16917753E+02, 0.39813349E+00, 0.21293635E+01, -0.16580629E+00, -0.91171318E+00, -0.19491872E-01, 0.62579691E+00, -0.31269869E+00, -0.14522402E+00, -0.69244891E+00, -0.35517254E+00, 
    },
    {
      0.23320499E+02, 0.19454935E+01, -0.32308598E+02, -0.22719015E+00, -0.26350050E+01, -0.75710707E+01, -0.13849431E+02, 0.13357112E+01, -0.52987828E+01, -0.12208441E+01, 0.56600362E+00, -0.38908503E+01, 0.15157390E+01, 
    },
    {
      0.16239506E+02, 0.26882658E+02, 0.11697211E+02, 0.13142451E+02, -0.13128712E+02, 0.12815480E+01, -0.26485260E+01, 0.10252703E+01, -0.71769851E+00, 0.26684380E+01, 0.33325875E+01, 0.19368449E+00, -0.84864247E+00, 
    },
    {
      -0.65441612E+02, -0.12118099E+03, 0.10928064E+03, -0.38240664E+01, -0.23437611E+02, 0.10801532E+02, -0.57041898E+01, -0.27291293E+01, -0.15678360E+01, -0.73939724E+01, 0.37768626E+00, 0.33962661E+00, -0.93000340E+00, 
    },
    {
      0.93348366E+02, -0.10296486E+03, -0.70953468E+02, -0.35590768E+00, -0.17237814E+02, 0.94308928E-01, 0.38772573E+01, 0.26037455E+01, -0.46013012E+01, 0.14253412E+01, 0.20683482E+01, 0.66785064E+01, 0.57986798E+01, 
    },
    {
      -0.11509787E+03, -0.40946972E+02, 0.20788942E+03, 0.28071583E+02, 0.24814009E+02, 0.32777500E+02, 0.80104485E+02, -0.11489333E+02, 0.29405239E+02, 0.16883720E+02, -0.67208619E+01, 0.20167179E+02, -0.86552134E+01, 
    },
    {
      -0.99561371E+02, -0.15522029E+03, -0.83673103E+02, -0.80494255E+02, 0.58310280E+02, -0.12448786E+02, 0.17183468E+02, -0.89103413E+01, 0.53705993E+01, -0.13733030E+02, -0.22880342E+02, -0.23245771E+01, 0.86334391E+01, 
    },
    {
      0.17955624E+03, 0.24499542E+03, -0.33905188E+03, 0.16499292E+02, 0.74469521E+02, -0.33439671E+02, 0.23443146E+02, 0.36955311E+01, -0.19126467E+00, 0.27561241E+02, 0.56984258E+01, 0.44640994E+01, 0.42115948E+00, 
    },
    {
      -0.30066251E+03, 0.29662677E+03, 0.81036797E+02, -0.27805751E-01, 0.43939407E+02, 0.27983515E+01, -0.60298972E+01, -0.10772779E+02, 0.15233603E+02, -0.58941245E+00, -0.87549553E+01, -0.21512075E+02, -0.19019295E+02, 
    },
    {
      0.25668509E+03, 0.95829285E+02, -0.60892999E+03, -0.11896392E+03, -0.95018196E+02, -0.59214664E+02, -0.19752927E+03, 0.38122326E+02, -0.69384659E+02, -0.56728783E+02, 0.22776094E+02, -0.43739380E+02, 0.22105671E+02, 
    },
    {
      0.22645404E+03, 0.37045316E+03, 0.17037305E+03, 0.21269356E+03, -0.11409075E+03, 0.32559406E+02, -0.47628788E+02, 0.30940620E+02, -0.16595905E+02, 0.25164560E+02, 0.59392143E+02, 0.48828721E+01, -0.25786690E+02, 
    },
    {
      -0.19930544E+03, -0.22929100E+03, 0.46098517E+03, -0.23051146E+02, -0.99358955E+02, 0.43438587E+02, -0.36220837E+02, 0.10952334E+01, 0.52028384E+01, -0.36286861E+02, -0.15389333E+02, -0.10731896E+02, 0.32523794E+01, 
    },
    {
      0.36962787E+03, -0.36779376E+03, -0.39542661E+01, -0.51138026E+00, -0.43776100E+02, -0.79711547E+01, 0.35742667E+01, 0.16140493E+02, -0.21209721E+02, -0.38114667E+01, 0.13219823E+02, 0.27240120E+02, 0.23259232E+02, 
    },
    {
      -0.26610745E+03, -0.69780006E+02, 0.81476440E+03, 0.17288275E+03, 0.13693073E+03, 0.46457741E+02, 0.21848393E+03, -0.50798943E+02, 0.72315247E+02, 0.74823509E+02, -0.31317862E+02, 0.41232723E+02, -0.24342672E+02, 
    },
    {
      -0.20460796E+03, -0.41652863E+03, -0.12226570E+03, -0.24583488E+03, 0.10287257E+03, -0.36260223E+02, 0.59697350E+02, -0.42112240E+02, 0.20332176E+02, -0.18127939E+02, -0.65813683E+02, -0.29394605E+01, 0.30914236E+02, 
    },
    {
      0.77954071E+02, 0.84637459E+02, -0.22712354E+03, 0.10199258E+02, 0.47287090E+02, -0.19806791E+02, 0.19231525E+02, -0.30108094E+01, -0.38124545E+01, 0.15798557E+02, 0.10139494E+02, 0.67266974E+01, -0.32354667E+01, 
    },
    {
      -0.15642979E+03, 0.16548746E+03, -0.24017464E+02, 0.17524619E+00, 0.14203983E+02, 0.56363974E+01, -0.46004733E+00, -0.83661394E+01, 0.10115620E+02, 0.35576582E+01, -0.64790502E+01, -0.11826179E+02, -0.96805029E+01, 
    },
    {
      0.10178647E+03, 0.11225664E+02, -0.39543845E+03, -0.84101143E+02, -0.65480949E+02, -0.12942043E+02, -0.88712433E+02, 0.23320219E+02, -0.27256409E+02, -0.34397594E+02, 0.15014293E+02, -0.13734142E+02, 0.94665298E+01, 
    },
    {
      0.60337921E+02, 0.17995731E+03, 0.21066154E+02, 0.10197303E+03, -0.35869057E+02, 0.15114552E+02, -0.27445200E+02, 0.19399147E+02, -0.83749247E+01, 0.40309701E+01, 0.26498037E+02, -0.13782138E-01, -0.13060991E+02, 
    },
    {
      -0.10923450E-01, -0.34306788E+00, -0.50401580E-01, -0.53487885E+00, 0.29230735E+00, 0.21476941E-01, 0.15144962E-01, -0.43948866E-01, -0.17394183E-01, 0.35130534E-01, -0.16292399E-01, 0.19373255E-01, -0.59662841E-03, 
    },
    {
      -0.71579292E-01, -0.52842975E-01, -0.63118017E-02, -0.39636719E+00, -0.73986328E+00, 0.37425833E-02, 0.59765212E-01, -0.49656001E-02, -0.27551942E-02, -0.15596611E-01, 0.61703688E-02, 0.14522822E-01, -0.32979961E-01, 
    },
    {
      -0.84795365E-02, 0.16810313E-02, -0.11047077E+00, -0.18073320E+01, 0.75538784E+00, 0.17500804E+00, -0.17733023E+00, -0.16572930E+00, -0.33800665E+00, -0.15694118E+00, -0.16335554E+00, 0.71416780E-01, -0.57195537E-01, 
    },
    {
      0.38296199E+00, 0.77749258E+00, -0.32756321E-01, -0.16789720E+01, -0.18577414E+01, 0.17541359E+00, -0.14601539E-01, 0.33322370E-02, 0.29737126E-01, -0.87182283E-01, -0.31497914E-01, 0.12684085E-01, -0.11295188E-01, 
    },
    {
      0.23203259E+01, 0.53199558E+01, 0.16164668E+01, -0.67294307E+01, 0.70609212E+01, -0.22695801E+00, -0.15492891E+01, 0.11764526E+01, 0.21189202E+00, -0.77632385E+00, 0.94330899E-01, -0.52504981E+00, 0.80921426E-01, 
    },
    {
      0.32638378E+01, -0.18616734E+00, 0.25722713E+01, -0.63221340E+01, -0.21414373E+01, 0.87206262E+00, -0.12179788E+01, 0.13763864E+00, 0.67562330E+00, -0.19984256E+00, 0.40309739E+00, -0.78906339E+00, 0.50545210E+00, 
    },
    {
      0.94253534E+00, 0.12683750E+01, 0.62755954E+00, 0.30156040E+01, 0.10006471E+01, 0.20007719E+00, 0.20150834E+00, 0.14695404E+01, 0.41846933E+01, 0.12572116E+01, 0.65958399E+00, -0.71903515E+00, -0.69292140E+00, 
    },
    {
      -0.96949649E+00, -0.45164185E+01, 0.29109006E+01, 0.85274162E+01, 0.41857915E+01, -0.27851799E+00, 0.91058570E+00, -0.63607663E+00, -0.55645533E-01, 0.10949482E+01, 0.38305825E+00, -0.72221702E+00, -0.53886354E+00, 
    },
    {
      -0.11483661E+02, -0.18448803E+02, -0.25420229E+02, 0.41966877E+02, -0.18985954E+02, -0.14435635E+01, 0.80026455E+01, -0.42318783E+01, -0.10838451E+01, 0.28619409E+01, 0.61609662E+00, 0.37960072E+01, -0.52905315E+00, 
    },
    {
      -0.17035336E+02, 0.18621317E+02, -0.21966009E+02, 0.18735846E+02, 0.16747011E+02, -0.54307652E+01, 0.46619482E+01, 0.13719676E+01, -0.60610504E+01, 0.19186600E+01, -0.33431869E+01, 0.48622260E+01, -0.19352443E+01, 
    },
    {
      -0.21383023E+01, -0.43643160E+01, -0.42401314E+01, -0.24073770E+01, -0.16128111E+02, -0.27644095E+01, 0.22688612E+00, -0.28903916E+01, -0.11218066E+02, -0.23697209E+01, -0.47793251E+00, 0.14431590E+01, 0.33283553E+01, 
    },
    {
      -0.29100385E+01, 0.97868328E+01, -0.62205763E+01, -0.72675920E+01, -0.31473615E+01, 0.19133319E+00, -0.38154783E+01, 0.26445303E+01, -0.82178062E+00, -0.30948792E+01, -0.84835994E+00, 0.25783031E+01, 0.23090553E+01, 
    },
    {
      0.19434116E+02, 0.30117950E+02, 0.67169952E+02, -0.83444626E+02, 0.97705097E+01, 0.59132814E+01, -0.15886705E+02, 0.57373362E+01, 0.21024210E+01, -0.35821984E+01, -0.25029049E+01, -0.88817778E+01, 0.14393607E+01, 
    },
    {
      0.31481159E+02, -0.55233753E+02, 0.44615513E+02, -0.14347218E+02, -0.38026085E+02, 0.13057509E+02, -0.83902435E+01, -0.42682991E+01, 0.14140669E+02, -0.35403111E+01, 0.79202223E+01, -0.92711287E+01, 0.31921399E+01, 
    },
    {
      -0.40178025E+00, 0.42423429E+01, 0.44495778E+01, 0.65041461E+01, 0.24042871E+02, 0.30938346E+01, -0.44691342E+00, 0.15059996E+01, 0.77578301E+01, 0.13139611E+01, -0.14930587E+00, -0.67745137E+00, -0.30641334E+01, 
    },
    {
      0.71513062E+01, -0.80155268E+01, 0.27227871E+01, -0.77887855E+01, 0.47823057E+01, -0.47966767E-01, 0.38374758E+01, -0.23107648E+01, 0.14483806E+01, 0.24450035E+01, 0.82637250E+00, -0.23080869E+01, -0.19962353E+01, 
    },
    {
      -0.12261738E+02, -0.18203463E+02, -0.47682377E+02, 0.55126305E+02, 0.94886637E+01, -0.50771351E+01, 0.10601325E+02, -0.30876780E+01, -0.16409191E+01, 0.16010164E+01, 0.18693397E+01, 0.63582640E+01, -0.12191840E+01, 
    },
    {
      -0.22224463E+02, 0.43476688E+02, -0.25098253E+02, -0.41798453E+01, 0.30020361E+02, -0.10327440E+02, 0.55453873E+01, 0.28609169E+01, -0.91859627E+01, 0.17712349E+01, -0.55576038E+01, 0.54289842E+01, -0.16860573E+01, 
    },
    {
      -0.18644652E+00, -0.13026768E+00, -0.51389337E-01, 0.13972405E+00, 0.26259443E-01, -0.48280844E+00, -0.59422386E+00, -0.21665383E-01, -0.14493400E-01, 0.15758840E-01, -0.17352704E-02, -0.41003264E-02, 0.18915856E-02, 
    },
    {
      0.57341747E-01, 0.42945143E-01, 0.96227340E-01, -0.34298085E-01, 0.13546824E+00, 0.63187897E+00, -0.47779465E+00, 0.20380679E-02, 0.22282990E-01, -0.25250018E-01, 0.17265640E-01, -0.25108820E-01, 0.94693340E-02, 
    },
    {
      -0.25878796E+00, 0.28488120E+00, -0.16580707E+00, 0.12818602E+00, 0.35997484E-01, -0.64699262E+00, -0.10717692E+01, 0.10131285E+00, 0.16753370E+00, 0.32581434E-01, -0.11407858E+00, 0.59960652E-01, -0.77041015E-02, 
    },
    {
      -0.41792631E-01, 0.13099888E+00, 0.21938293E+00, -0.11784480E+00, 0.17003253E+00, 0.11506057E+01, -0.71615899E+00, -0.11056098E+00, 0.15374933E+00, 0.68769425E-01, 0.45401078E-01, 0.56390647E-01, 0.55900361E-01, 
    },
    {
      0.28251824E+01, 0.44229245E+00, -0.81067193E+00, -0.12398359E+01, -0.28553767E-01, 0.22159548E+01, -0.54317749E+00, 0.75851673E+00, 0.27122670E+00, -0.17409934E+00, 0.50550703E-01, 0.21512508E+00, 0.78091890E-01, 
    },
    {
      -0.14035779E+01, 0.84380841E+00, -0.64104086E+00, 0.46829499E-01, -0.14327688E+01, 0.25394753E+00, 0.19480610E+01, -0.91148943E-01, 0.78488475E+00, 0.22460079E+00, -0.24968466E+00, 0.46386969E+00, -0.64009469E-03, 
    },
    {
      0.98801070E+00, -0.11307433E+01, 0.52810138E+00, 0.87130740E-01, 0.50161190E-01, -0.13517044E+00, 0.17106220E+01, -0.20004056E+00, -0.23602618E+00, -0.27457481E-01, 0.24470349E+00, -0.14111978E+00, 0.38745444E-01, 
    },
    {
      0.47567675E+00, -0.22344865E+00, -0.75762641E+00, -0.57572670E-01, 0.84235653E-01, -0.19831989E+01, 0.25760263E+00, 0.19424839E+00, -0.48381361E+00, -0.13142878E+00, -0.88607289E-01, -0.23240808E+00, -0.17395243E+00, 
    },
    {
      -0.52569547E+01, 0.56680167E+00, 0.16630803E+01, 0.17917036E+01, -0.68722159E+00, -0.39917276E+01, 0.18925166E+01, -0.13140194E+01, -0.42374143E+00, 0.26679030E+00, -0.19116408E+00, -0.40941295E+00, -0.66891648E-01, 
    },
    {
      0.19922812E+01, -0.13392367E+01, 0.13449210E+01, 0.34894937E+00, 0.20699844E+01, -0.15274907E+01, -0.31885908E+01, 0.30454439E+00, -0.15603627E+01, -0.34676239E+00, 0.25211805E+00, -0.71817315E+00, -0.18247532E-01, 
    },
    {
      0.26140720E-01, 0.61536264E-01, 0.58766458E-01, 0.60080968E-01, 0.25750723E-01, -0.20359196E-01, 0.31155759E-01, 0.21450184E+00, -0.28345245E+00, 0.30602589E-01, 0.73734480E-02, -0.73152762E-04, 0.24243210E-01, 
    },
    {
      0.19760441E-01, 0.36050626E-02, 0.31682935E-01, -0.42440861E-01, 0.21866074E-01, -0.36474712E-01, -0.28035508E-02, 0.29943243E+00, 0.22910598E+00, 0.11597273E-01, -0.11098400E-01, 0.64757066E-02, 0.79353563E-02, 
    },
    {
      0.14553243E+00, -0.74297085E-01, -0.25932562E-01, -0.46754810E-02, 0.16648775E+00, -0.20828387E-01, -0.15960755E-02, -0.84260628E-02, -0.61570007E+00, -0.76635070E-01, 0.38172156E-01, -0.21436466E-02, 0.33362638E-01, 
    },
    {
      0.16782244E+00, -0.23373500E+00, -0.76538667E-01, -0.46066750E-01, 0.44487558E-01, -0.98707631E-01, -0.62737954E-02, 0.47552589E+00, 0.33932384E-01, -0.10383098E-01, 0.54402561E-02, 0.42593040E-01, -0.40717162E-01, 
    },
    {
      -0.34468144E-01, 0.99032059E-01, -0.15680047E-01, 0.16594915E-01, 0.60266782E-01, -0.96315397E-02, -0.12239870E-01, -0.44720203E-01, 0.24779214E-01, 0.17988123E+00, 0.16323383E+00, -0.13639979E-01, 0.24354503E-01, 
    },
    {
      0.13295417E+00, 0.36607604E-01, 0.11295780E-01, -0.85317560E-01, 0.23435000E-01, -0.11025268E-01, -0.20789802E-01, 0.22085607E-01, -0.84365681E-01, -0.16373871E+00, 0.14517102E+00, -0.26270184E-02, 0.10327833E-01, 
    },
    {
      0.39756168E-01, -0.21894038E-01, 0.32299142E-01, -0.25709143E-01, 0.46975818E-01, 0.91380738E-02, 0.45137787E-02, -0.18453874E-01, -0.16742310E-01, 0.62881177E-02, 0.31858400E-01, -0.52852482E-01, 0.64126670E-01, 
    },
    {
      0.70708215E-01, 0.26414242E-01, -0.95734000E-01, -0.51715512E-01, -0.61222430E-01, 0.32585312E-02, -0.72614186E-01, 0.29171487E-01, -0.30535296E-01, -0.62957483E-05, 0.48962277E-02, -0.69674492E-01, -0.63961051E-01, 
    },
    {
      0.24603065E-02, 0.32394640E-01, -0.59477240E-01, -0.60001020E-02, -0.37934547E-03, 0.28015230E-01, -0.36198471E-01, 0.15542215E-01, -0.82169725E-02, -0.18009149E-01, -0.15005712E-01, -0.68974853E-02, -0.44691041E-02, 
    },
    {
      0.16535794E-01, 0.56647848E-01, -0.21545598E-01, -0.41591849E-01, -0.52183713E-02, -0.25275357E-01, 0.10705411E-01, 0.35601035E-01, 0.87656369E-02, -0.97603891E-02, 0.26778039E-02, 0.52772003E-03, -0.49273390E-02, 
    },
    {
      -0.10546016E-02, 0.60169496E-01, -0.38740799E-01, -0.21469520E-01, 0.29543184E-02, -0.40258944E-01, -0.13036972E-01, 0.18623425E-01, 0.90010501E-02, 0.14114746E-01, -0.19131180E-01, -0.65653445E-03, 0.10507869E-01, 
    },
    {
      0.22839565E-01, -0.56603742E-02, 0.63457802E-01, 0.13852383E-01, 0.19432070E-01, 0.25225332E-01, 0.21424852E-01, -0.25062202E-02, 0.82456255E-02, -0.16133005E-01, 0.12762873E-01, 0.22965312E-01, 0.64126670E-01, 
    },
  },
};

const Fm3_coefficient_array_t g_NeQuickG_ccir22_Fm3[ITU_F2_LAYER_SOLAR_ACTIVITY_CONDITIONS_COUNT] = {
  {
    {
      0.30660009E+01, 0.84245279E-02, 0.10416747E-01, -0.79152789E-02, 0.97902566E-02, 0.27440805E-02, -0.18010708E-02, -0.44805654E-02, 0.44099679E-02, 
    },
    {
      0.47937575E+00, -0.73863333E-02, 0.62492259E-01, 0.71070671E-01, 0.24661029E-01, -0.76728272E-02, -0.34576587E-01, -0.55801393E-02, 0.24911568E-01, 
    },
    {
      0.72640646E+00, -0.13610825E+00, -0.92151836E-01, 0.57366926E-01, -0.79799727E-01, 0.15536819E-01, 0.29552156E-01, -0.23152215E-01, -0.71591372E-02, 
    },
    {
      -0.52047455E+00, -0.18002425E+00, -0.14835778E+00, -0.15261011E+00, -0.84134936E-01, 0.74266516E-01, 0.12540767E+00, 0.69134543E-02, -0.10695435E+00, 
    },
    {
      -0.78587282E+00, 0.42692187E+00, 0.14050616E+00, -0.10526546E+00, 0.16430844E+00, -0.38810812E-01, -0.95510431E-01, 0.89601807E-01, 0.10011222E-01, 
    },
    {
      -0.18745262E-01, 0.15723652E+00, 0.10078381E+00, 0.68045869E-01, 0.75673684E-01, -0.72365344E-01, -0.91716684E-01, -0.11222346E-02, 0.89619219E-01, 
    },
    {
      0.83576031E-01, -0.30042040E+00, -0.12336441E-01, 0.59537537E-01, -0.84578089E-01, 0.17366186E-01, 0.79638004E-01, -0.66486605E-01, -0.10213455E-01, 
    },
    {
      -0.24552050E-02, -0.10713033E+00, -0.34063554E+00, 0.15499518E-03, 0.16007205E-01, -0.43798913E-02, 0.56936443E-02, 0.65028886E-02, 0.10104384E-01, 
    },
    {
      0.11042719E-01, 0.37260795E+00, -0.10771591E+00, 0.56015593E-02, -0.42281896E-02, -0.25236963E-02, -0.11927215E-02, -0.10831781E-01, 0.66218302E-02, 
    },
    {
      -0.12299374E+00, 0.37923911E+00, 0.25219777E+00, 0.50730992E-01, 0.70875049E-01, -0.12099678E-01, 0.36568884E-01, 0.57058934E-01, -0.18947596E-01, 
    },
    {
      -0.35674047E-01, -0.31694239E+00, 0.45843518E+00, -0.12794103E+00, -0.69748282E-01, 0.42822126E-01, 0.30879781E-01, -0.51876172E-01, 0.43667890E-01, 
    },
    {
      0.11940227E+00, 0.61209160E+00, 0.10119880E+01, -0.12578768E+00, -0.23280199E+00, 0.84285967E-01, -0.31325568E-01, -0.14911862E+00, -0.10552527E+00, 
    },
    {
      -0.42771018E+00, -0.15437865E+01, 0.72490811E+00, -0.41361243E-01, 0.11341509E+00, 0.58602658E-02, -0.10867080E-01, 0.23384985E+00, -0.18132639E+00, 
    },
    {
      0.14408733E+01, -0.15181103E+01, -0.70733821E+00, 0.22425801E+00, -0.20626904E+00, -0.23165996E+00, -0.24311604E+00, -0.75157189E+00, 0.32163310E+00, 
    },
    {
      0.11034918E+01, 0.19018683E+01, -0.25283425E+01, 0.54272127E+00, 0.81747144E+00, -0.33536407E+00, -0.28131485E+00, 0.44513670E+00, -0.38100833E+00, 
    },
    {
      -0.49559730E+00, -0.76724958E+00, 0.64328122E+00, 0.37971288E+00, 0.38863850E+00, -0.20415752E+00, -0.27450986E-01, 0.35379422E+00, 0.24708341E+00, 
    },
    {
      0.15503995E+01, 0.70807004E+00, -0.15437074E+01, 0.22590764E+00, -0.22412081E+00, 0.70388354E-01, 0.28750276E-01, -0.59116602E+00, 0.65819973E+00, 
    },
    {
      -0.31314054E+01, 0.15656358E+01, 0.18508843E+01, -0.10490452E+01, 0.13597289E+00, 0.96911991E+00, 0.34504092E+00, 0.18334630E+01, -0.94316882E+00, 
    },
    {
      -0.32625735E+01, -0.48891945E+01, 0.43436418E+01, -0.89692688E+00, -0.22180891E+01, 0.81275958E+00, 0.57701594E+00, -0.92923886E+00, 0.10023088E+01, 
    },
    {
      0.46660012E+00, 0.33790904E+00, -0.12450566E+01, -0.27982497E+00, -0.11594565E+00, 0.78048654E-01, 0.53353399E-01, -0.21656552E+00, -0.16291541E+00, 
    },
    {
      -0.14640079E+01, 0.67885804E+00, 0.12005756E+01, -0.24355866E+00, 0.63787431E-01, -0.11045193E+00, -0.53045399E-01, 0.37856793E+00, -0.54621232E+00, 
    },
    {
      0.18558358E+01, -0.15444054E+00, -0.95597863E+00, 0.83078742E+00, 0.86885981E-01, -0.86581147E+00, -0.98496750E-01, -0.12377061E+01, 0.71731114E+00, 
    },
    {
      0.23264565E+01, 0.32293904E+01, -0.22380531E+01, 0.50838184E+00, 0.16552072E+01, -0.55498052E+00, -0.29978287E+00, 0.58354074E+00, -0.75011027E+00, 
    },
    {
      -0.18672661E-02, -0.25456157E-01, 0.60223136E-02, 0.49720477E-01, -0.96478693E-01, -0.38978413E-02, -0.22143553E-03, -0.35477506E-02, -0.55410183E-03, 
    },
    {
      -0.34474754E-02, -0.11412907E-01, -0.14703928E-01, 0.10629820E+00, 0.38479723E-01, -0.11775076E-01, 0.24791460E-02, -0.13214642E-04, 0.54156147E-02, 
    },
    {
      0.17747490E-01, 0.85842289E-01, 0.12783460E-01, -0.80395052E-02, 0.82914688E-01, 0.52115154E-01, -0.59585259E-02, -0.34316640E-01, -0.31203479E-01, 
    },
    {
      -0.15008713E+00, 0.13022913E-01, -0.13114069E+00, -0.98083436E-01, 0.43264881E-01, 0.23754107E-01, 0.40249381E-01, -0.25870558E-01, 0.10604724E-01, 
    },
    {
      -0.11669767E+00, 0.18879850E+00, 0.85860342E-01, -0.80694489E-01, -0.14608112E+00, -0.38452525E-01, 0.11943401E+00, 0.36737777E-01, -0.83100669E-01, 
    },
    {
      -0.98170832E-01, -0.25824536E-01, -0.13470170E-01, 0.16278113E+00, 0.14445356E+00, -0.37913606E-01, 0.26959060E-01, 0.54383244E-01, -0.33156373E-01, 
    },
    {
      -0.69396205E-01, -0.29353243E+00, 0.12254417E+00, -0.11734211E-01, -0.65991062E+00, -0.22865216E+00, -0.37296597E-01, 0.19098966E+00, 0.12578192E+00, 
    },
    {
      0.10914133E+01, 0.47314721E+00, 0.71602601E+00, 0.67153996E+00, -0.39010674E+00, 0.86237691E-01, -0.38608781E+00, 0.10881895E+00, -0.13271818E-01, 
    },
    {
      0.24741711E+00, -0.29912946E+00, -0.27786815E+00, 0.58144867E-01, 0.35581678E+00, 0.53345066E-01, -0.22732742E+00, -0.30918676E-01, 0.11908955E+00, 
    },
    {
      0.87417528E-01, -0.19209103E-02, 0.12066233E+00, -0.37097234E+00, -0.29199532E+00, 0.99688828E-01, -0.63901879E-01, -0.95861524E-01, 0.24224207E-01, 
    },
    {
      -0.14814958E-01, 0.35126498E+00, -0.10779124E+00, -0.37479263E-01, 0.10758766E+01, 0.33757868E+00, 0.12894204E+00, -0.21282887E+00, -0.11465584E+00, 
    },
    {
      -0.14006575E+01, -0.77323854E+00, -0.76969266E+00, -0.10458283E+01, 0.45393547E+00, -0.20995061E+00, 0.51989406E+00, -0.76540701E-01, -0.34076910E-01, 
    },
    {
      -0.14458183E-01, 0.55377367E-02, 0.10118624E-01, 0.43726047E-02, 0.30203222E-02, 0.26917916E-01, -0.27824063E-01, 0.38613998E-02, -0.23288135E-02, 
    },
    {
      0.97673871E-02, -0.12221950E-01, -0.20525286E-02, -0.16824726E-01, 0.66913562E-04, 0.29891016E-01, 0.39633669E-01, 0.86441580E-02, 0.57085617E-02, 
    },
    {
      0.60076401E-01, 0.44391785E-01, 0.75440123E-02, -0.47574192E-01, -0.25747169E-01, -0.13596806E+00, -0.25348425E-01, -0.39103729E-03, -0.79124048E-02, 
    },
    {
      0.47816981E-01, -0.12213377E-01, -0.60907356E-02, -0.16249452E-01, -0.81891660E-02, 0.20783218E-01, -0.12515883E+00, -0.63580880E-02, -0.93365340E-02, 
    },
    {
      0.35757348E-01, -0.11628059E+00, 0.23941850E-01, 0.39817769E-01, 0.22874350E-01, -0.14632468E+00, -0.48552625E-01, 0.21140071E-01, -0.29413914E-01, 
    },
    {
      -0.10383629E+00, -0.25948141E-01, -0.37469424E-01, 0.13439357E-01, -0.65478263E-02, 0.93140125E-01, -0.17407711E+00, -0.22653142E-01, 0.20707196E-01, 
    },
    {
      -0.12050121E-01, -0.76118793E-03, 0.14033116E-02, 0.10667692E-01, 0.31927912E-02, -0.39311568E-02, 0.34947079E-02, 0.25789489E-02, 0.31379892E-02, 
    },
    {
      -0.14101000E-01, 0.78286082E-02, -0.55754250E-02, -0.79506766E-02, 0.20916839E-02, 0.22414438E-02, 0.73174662E-02, -0.58244873E-03, 0.37509066E-02, 
    },
    {
      0.89371942E-01, 0.39174829E-01, -0.15044765E-01, -0.11416791E-02, -0.28407650E-01, -0.12282236E-01, 0.21022290E-02, 0.14036150E-01, -0.10413602E+00, 
    },
    {
      -0.34115105E-02, -0.33254713E-01, 0.22143330E-01, 0.35554647E-01, -0.46583945E-02, 0.82491385E-02, 0.43415073E-02, 0.11600922E+00, 0.77101225E-02, 
    },
    {
      -0.15329777E-02, -0.16291803E-02, -0.26436499E-02, -0.49016397E-02, 0.12201109E-02, 0.16907734E-02, 0.32629367E-03, -0.15686175E-01, 0.86215120E-02, 
    },
    {
      0.87878658E-02, 0.13618468E-01, -0.13227126E-01, -0.82996050E-02, -0.15946573E-01, 0.71573956E-02, 0.56405202E-02, -0.54664928E-02, -0.11713528E-01, 
    },
    {
      -0.12917928E-02, 0.49148039E-02, 0.13711536E-01, 0.31793679E-02, 0.66184965E-02, 0.20440640E-02, -0.92955929E-03, -0.69237831E-02, -0.55623794E-03, 
    },
    {
      0.28278935E-03, -0.72326208E-02, -0.10317585E-01, 0.10494902E-03, -0.11032540E-01, 0.24561610E-02, 0.84329043E-02, -0.13802494E-02, -0.10716310E-01, 
    },
  },
  {
    {
      0.26969762E+01, 0.32681650E-02, 0.43882988E-02, -0.27695221E-02, 0.10836555E-01, -0.36222459E-03, -0.40186797E-02, -0.25219587E-02, -0.58713165E-04, 
    },
    {
      0.43824396E+00, 0.19661849E-02, 0.37536167E-01, 0.40424865E-01, 0.34500234E-01, -0.12376064E-01, -0.37711073E-01, -0.25790085E-02, 0.21678755E-01, 
    },
    {
      0.90846723E+00, -0.32327816E-01, -0.90745442E-01, 0.30759597E-01, -0.99251971E-01, 0.25366189E-01, 0.51679712E-01, -0.10518884E-01, 0.47520380E-01, 
    },
    {
      -0.36741725E+00, -0.22690707E+00, -0.80392964E-01, -0.10627037E+00, -0.96276939E-01, 0.56504987E-01, 0.13875304E+00, -0.96518273E-03, -0.10057620E+00, 
    },
    {
      -0.98529118E+00, 0.91059320E-01, 0.14998367E+00, -0.79339772E-01, 0.17853460E+00, -0.37562929E-01, -0.12783723E+00, 0.34081671E-01, -0.13274518E+00, 
    },
    {
      -0.14129069E-01, 0.19939514E+00, 0.66745520E-01, 0.63963860E-01, 0.75332023E-01, -0.48900589E-01, -0.10383722E+00, 0.38989398E-02, 0.86292423E-01, 
    },
    {
      0.10910988E+00, -0.51262561E-01, -0.39895140E-01, 0.59805356E-01, -0.85471421E-01, 0.12946113E-01, 0.86364143E-01, -0.23241360E-01, 0.85489251E-01, 
    },
    {
      0.68126281E-03, -0.26906112E+00, -0.25801969E+00, -0.36406654E-03, 0.15821831E-01, -0.58499649E-02, 0.49455757E-02, 0.19876708E-02, 0.70102816E-02, 
    },
    {
      0.10385459E-01, 0.28189835E+00, -0.27379474E+00, 0.99617587E-02, -0.94898148E-02, 0.23341458E-02, -0.17798019E-02, -0.35258995E-02, 0.95521472E-02, 
    },
    {
      -0.59638537E-01, 0.22104724E+00, 0.11115756E+00, 0.36462764E-02, 0.68865590E-01, -0.22449013E-01, 0.57117860E-02, 0.45325380E-01, 0.41912273E-02, 
    },
    {
      -0.27711060E-01, -0.10434671E+00, 0.26874310E+00, -0.27662922E-01, -0.48578799E-01, 0.34111135E-01, 0.36623143E-01, -0.33080466E-02, 0.34407139E-01, 
    },
    {
      -0.86527253E-02, 0.10166690E+01, 0.33211759E+00, -0.69901399E-01, -0.18758866E+00, 0.96067429E-01, -0.20856796E-01, -0.89351051E-01, -0.11115274E+00, 
    },
    {
      -0.27616104E+00, -0.77658612E+00, 0.11817397E+01, -0.40753499E-01, 0.15753447E+00, -0.38351923E-01, 0.15354547E-01, 0.11337685E+00, -0.17565700E+00, 
    },
    {
      0.70202410E+00, -0.12175164E+01, -0.54803377E+00, 0.31004715E+00, -0.24121673E+00, 0.54169376E-02, -0.16837779E+00, -0.57409120E+00, -0.40699039E-01, 
    },
    {
      0.73859042E+00, 0.11418145E+01, -0.18006949E+01, -0.92229664E-01, 0.44709337E+00, -0.23921192E+00, -0.38646665E+00, 0.41334989E-03, -0.26683557E+00, 
    },
    {
      -0.14896899E+00, -0.11365073E+01, 0.14847133E+01, 0.33789396E+00, 0.27518693E+00, -0.24134621E+00, -0.20836785E-01, 0.24304211E+00, 0.32803184E+00, 
    },
    {
      0.98338109E+00, -0.47739923E+00, -0.19460961E+01, 0.17087240E-01, -0.25491282E+00, 0.74909970E-01, -0.80111466E-01, -0.28729579E+00, 0.55402631E+00, 
    },
    {
      -0.14751050E+01, 0.17943174E+01, 0.17780763E+01, -0.91258276E+00, 0.23318025E+00, 0.18268088E+00, 0.40549916E+00, 0.14381841E+01, 0.42043656E-01, 
    },
    {
      -0.24171760E+01, -0.34031210E+01, 0.33957868E+01, 0.46847907E+00, -0.11220565E+01, 0.62169796E+00, 0.88092864E+00, 0.88447511E-01, 0.57009745E+00, 
    },
    {
      0.19369581E+00, 0.60008752E+00, -0.14411573E+01, -0.28396797E+00, -0.81249483E-01, 0.11603205E+00, 0.37423864E-01, -0.16771762E+00, -0.25206971E+00, 
    },
    {
      -0.87629116E+00, 0.11575415E+01, 0.13158292E+01, 0.61591063E-01, 0.67997426E-01, -0.37556443E-01, 0.51347751E-01, 0.18010110E+00, -0.45261535E+00, 
    },
    {
      0.80545521E+00, -0.58170259E+00, -0.83205831E+00, 0.67988551E+00, -0.48437756E-02, -0.24384318E+00, -0.24711609E+00, -0.99533564E+00, -0.74051530E-03, 
    },
    {
      0.19566715E+01, 0.21233406E+01, -0.18595179E+01, -0.41028827E+00, 0.87477672E+00, -0.47322938E+00, -0.53535795E+00, -0.10144442E+00, -0.34023517E+00, 
    },
    {
      0.19387048E-02, -0.24992781E-01, 0.41351970E-02, 0.82355030E-01, -0.10060746E+00, 0.10565569E-02, 0.61386190E-02, -0.56847963E-02, 0.28516669E-03, 
    },
    {
      0.18765301E-02, -0.18433956E-01, -0.57170093E-02, 0.96418172E-01, 0.74403822E-01, -0.49711904E-02, 0.73961457E-02, -0.26749184E-02, 0.35287787E-02, 
    },
    {
      0.44248577E-01, 0.44711590E-01, 0.35021894E-01, -0.10435295E+00, 0.12232722E+00, 0.40250495E-01, 0.19099291E-01, 0.10738099E-02, -0.69058025E-02, 
    },
    {
      -0.51917080E-01, -0.27200999E-01, -0.10501927E+00, -0.16478807E+00, -0.83345883E-01, -0.19203559E-02, 0.45351192E-01, -0.21076348E-01, 0.86140940E-02, 
    },
    {
      -0.10319305E+00, 0.13298593E+00, 0.73425964E-01, -0.46246278E+00, 0.13967551E+00, -0.45703530E-01, 0.76875174E-02, 0.67371532E-01, -0.46121821E-01, 
    },
    {
      -0.71389146E-01, -0.35766073E-01, -0.66930741E-01, -0.13037997E+00, -0.29186937E+00, -0.43604024E-01, -0.38124863E-01, 0.46981771E-01, -0.33583380E-01, 
    },
    {
      -0.22479945E+00, -0.17325100E+00, -0.42389341E-01, 0.63497208E-01, -0.65707499E+00, -0.18000965E+00, -0.13928051E+00, 0.92657882E-03, 0.32833587E-01, 
    },
    {
      0.35454518E+00, 0.56296510E+00, 0.54718381E+00, 0.87280089E+00, -0.95135480E-01, 0.18839495E+00, -0.28471857E+00, 0.99687912E-01, -0.74194811E-01, 
    },
    {
      0.22021563E+00, -0.19814751E+00, -0.26210853E+00, 0.38916111E+00, -0.11701600E+00, 0.43736532E-01, -0.51089481E-01, -0.89396305E-01, 0.56751575E-01, 
    },
    {
      0.84288068E-01, 0.46980049E-01, 0.16849743E+00, 0.14809281E+00, 0.15139215E+00, 0.80832951E-01, 0.28207000E-01, -0.66889457E-01, 0.30166352E-01, 
    },
    {
      0.12673169E+00, 0.26919132E+00, 0.78022078E-01, 0.44204059E-03, 0.88714296E+00, 0.26615989E+00, 0.19613440E+00, -0.19267652E-01, -0.56138389E-01, 
    },
    {
      -0.44904345E+00, -0.76422292E+00, -0.58892214E+00, -0.11297759E+01, 0.21760568E+00, -0.28634107E+00, 0.33985862E+00, -0.90833008E-01, 0.72062925E-01, 
    },
    {
      -0.51934798E-02, 0.74422540E-03, 0.79966486E-02, 0.72337938E-02, -0.33158786E-03, 0.32955918E-01, -0.36173228E-01, 0.37724695E-02, -0.41374913E-03, 
    },
    {
      -0.21712191E-02, -0.42578488E-03, 0.46815402E-02, -0.11746973E-01, 0.20427755E-02, 0.37616257E-01, 0.33100210E-01, 0.46383860E-02, 0.91368466E-03, 
    },
    {
      0.12895065E-01, 0.50809748E-01, 0.29385984E-02, -0.18984521E-01, -0.27869942E-01, -0.12483256E+00, -0.34310818E-01, -0.27956211E-02, -0.34609209E-04, 
    },
    {
      0.11988606E-01, -0.21017001E-02, 0.25302023E-02, -0.96779130E-02, -0.14333853E-01, 0.32475434E-01, -0.11795449E+00, -0.32607217E-02, -0.88285496E-02, 
    },
    {
      0.10866594E-01, -0.90187959E-01, 0.33737600E-01, 0.10237814E-01, 0.22867156E-01, -0.17877857E+00, -0.42381451E-01, 0.12767127E-01, -0.93460009E-02, 
    },
    {
      -0.30316098E-01, -0.55458799E-01, -0.43160249E-01, 0.22116292E-01, -0.88928174E-02, 0.67580663E-01, -0.13444774E+00, -0.15160068E-01, 0.16901061E-01, 
    },
    {
      -0.74068061E-02, -0.93656145E-02, 0.16188228E-02, 0.28720102E-02, 0.59500979E-02, -0.38546906E-02, 0.22760713E-02, 0.99050291E-02, 0.50792447E-02, 
    },
    {
      -0.11978722E-01, 0.60989028E-02, -0.10759148E-01, -0.51578460E-02, -0.61472147E-02, -0.13215247E-02, 0.77774036E-02, -0.41620326E-02, 0.12450544E-01, 
    },
    {
      0.33062402E-01, 0.28125867E-01, -0.74927374E-02, 0.32324712E-02, -0.23429561E-01, -0.13820422E-02, 0.71330271E-02, 0.28685573E-01, -0.87745167E-01, 
    },
    {
      0.55558784E-02, -0.37983757E-01, 0.23901636E-01, 0.20743832E-01, -0.81198774E-02, -0.32056179E-02, 0.81567001E-02, 0.96058153E-01, 0.33794027E-01, 
    },
    {
      0.55779296E-03, -0.44976049E-02, -0.40536787E-03, -0.57768053E-02, 0.36962274E-02, 0.12729283E-02, -0.32060328E-02, -0.11466083E-01, 0.39026923E-02, 
    },
    {
      0.30240275E-02, 0.14327827E-01, -0.91552837E-02, -0.15580239E-02, -0.14593671E-01, 0.58387034E-02, 0.71950480E-02, -0.30479324E-02, -0.67118248E-02, 
    },
    {
      -0.39168904E-02, 0.10714657E-01, 0.97695896E-02, 0.44566900E-02, 0.11315702E-02, -0.24044179E-03, -0.57686231E-03, -0.59982366E-02, -0.22883762E-02, 
    },
    {
      0.37045740E-02, -0.60077994E-02, -0.99031944E-02, 0.84312021E-03, -0.11128205E-01, 0.24483956E-02, 0.49151871E-02, 0.20255843E-03, -0.45773280E-02, 
    },
  },
};
